from .osifinance import Osifinance 
from pandas import DataFrame


class StateTaxes(Osifinance):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)


    def __fetch__(self, data, path):
        Osifinance.__fetch__(self, data, path)

    def alabama(self, sources=False, filing_status='single', county_residence='Other', capital_gains_short=0, capital_gains_long=0, credits_state=0, state_agi=75000, income=100000, dependents=0):
        """Taxes for Alabama
        
        Parameters:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2

        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        Sources:
        https://www.revenue.alabama.gov/ultraviewer/viewer/basic_viewer/index.html?form=2023/02/22f40.pdf
        https://www.revenue.alabama.gov/faqs/what-is-alabamas-individual-income-tax-rate/
        https://www.revenue.alabama.gov/ultraviewer/viewer/basic_viewer/index.html?form=2023/02/22f40.pdf
        https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-24.htm?taxmap=true
        https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-24.htm?taxmap=true
        https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-24.htm?taxmap=true
        https://www.revenue.alabama.gov/individual-corporate/income-to-be-reported-on-the-alabama-income-tax-return/
        https://www.almonline.org/TaxRates.aspx#OccupationalTax
        """
        
        # Check if the parameters are set upstream in the instance
        sources = self.sources if hasattr(self, 'sources') else sources
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        
        # State-specific parameters
        data = {
            'state_residence': 'Alabama', "sources": sources, "filing_status": filing_status, "county_residence": county_residence, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "credits_state": credits_state, "state_agi": state_agi, "income": income, "dependents": dependents
        }
        
        return self.__fetch__(data, 'state-taxes')


    def alaska(self, sources=False):
        """Taxes for Alaska
        
        Parameters:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.

        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        Sources:
        https://www.commerce.alaska.gov/web/dcra/officeofthestateassessor/alaskataxfacts.aspx
        """
        
        # Check if the parameters are set upstream in the instance
        sources = self.sources if hasattr(self, 'sources') else sources
        
        # State-specific parameters
        data = {
            'state_residence': 'Alaska', "sources": sources
        }
        
        return self.__fetch__(data, 'state-taxes')


    def arizona(self, sources=False, filing_status='single', capital_gains_short=0, capital_gains_long=0, state_agi=75000, income=100000, dependents=0, credits_state=0):
        """Taxes for Arizona
        
        Parameters:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        credits_state : int, optional
            Credits for state taxes. Default is 0.

        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        Sources:
        https://azdor.gov/forms/individual/form-140-x-y-tables
        https://azdor.gov/forms/individual/form-140-x-y-tables
        https://azdor.gov/forms/individual/form-140-x-y-tables
        https://azdor.gov/businesses-arizona/withholding-tax/withholding-exceptions
        "https://azdor.gov/individuals/income-tax-filing-assistance/identifying-other-taxable-income#:~:text=In%20regards%20to%20capital%20gains
       the%20individuals%20regular%20tax%20rate.
        """
        
        # Check if the parameters are set upstream in the instance
        sources = self.sources if hasattr(self, 'sources') else sources
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Arizona', "sources": sources, "filing_status": filing_status, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "state_agi": state_agi, "income": income, "dependents": dependents, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')


    def arkansas(self, sources=False, filing_status='single', capital_gains_short=0, capital_gains_long=0, credits_state=0, state_agi=75000, income=100000, dependents=0, filers_over_65=0):
        """Taxes for Arkansas
        
        Parameters:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.

        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        Sources:
        https://www.arkansasedc.com/why-arkansas/business-climate/tax-structure/personal-income-tax
        https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_and_AR1000NR_Instructions.pdf
        https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_FullYearResidentIndividualIncomeTaxReturn.pdf
        https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_and_AR1000NR_Instructions.pdf
        https://www.arkansasedc.com/why-arkansas/business-climate/tax-structure/capital-gains-tax-reduction
        https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_and_AR1000NR_Instructions.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        sources = self.sources if hasattr(self, 'sources') else sources
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        
        # State-specific parameters
        data = {
            'state_residence': 'Arkansas', "sources": sources, "filing_status": filing_status, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "credits_state": credits_state, "state_agi": state_agi, "income": income, "dependents": dependents, "filers_over_65": filers_over_65
        }
        
        return self.__fetch__(data, 'state-taxes')


    def california(self, sources=False, filing_status='single', capital_gains_short=0, capital_gains_long=0, credits_state=0, state_agi=75000, income=100000):
        """Taxes for California
        
        Parameters:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.

        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        Sources:
        https://www.ftb.ca.gov/forms/2022/2022-540-tax-rate-schedules.pdf
        https://www.ftb.ca.gov/file/personal/deductions/index.html
        https://www.ftb.ca.gov/about-ftb/newsroom/caleitc/eligibility-and-credit-information.html
        https://www.ftb.ca.gov/file/personal/income-types/capital-gains-and-losses.html#:~:text=California%20does%20not%20have%20a
       are%20taxed%20as%20ordinary%20income
        https://www.ftb.ca.gov/file/personal/
        """
        
        # Check if the parameters are set upstream in the instance
        sources = self.sources if hasattr(self, 'sources') else sources
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        income = self.income if hasattr(self, 'income') else income
        
        # State-specific parameters
        data = {
            'state_residence': 'California', "sources": sources, "filing_status": filing_status, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "credits_state": credits_state, "state_agi": state_agi, "income": income
        }
        
        return self.__fetch__(data, 'state-taxes')


    def colorado(self, sources=False, credits_state=0, county_residence='Other', capital_gains_short=0, capital_gains_long=0, state_agi=75000, income=100000):
        """Taxes for Colorado
        
        Parameters:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.

        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        Sources:
        https://leg.colorado.gov/agencies/legislative-council-staff/individual-income-tax%C2%A0
        https://leg.colorado.gov/agencies/legislative-council-staff/individual-income-tax%C2%A0
        https://tax.colorado.gov/sites/tax/files/documents/ITT_Capital_Gain_Subtraction_Feb_2022.pdf
        https://www.auroragov.org/business_services/taxes/occupational_privilege_tax
        https://www.denvergov.org/Government/Agencies-Departments-Offices/Agencies-Departments-Offices-Directory/Department-of-Finance/Our-Divisions/Treasury/Business-Tax-Information#:~:text=Occupational%20privilege%20tax
       -%28show%20below%29&text=Each%20taxable%20employee%20is%20liable
       month%20for%20each%20taxable%20employee
        https://www.glendale.co.us/355/Occupational-Privilege-Tax
        https://www.greenwoodvillage.com/1220/Occupational-Privilege-Tax-OPT
        https://ci.sheridan.co.us/288/Occupational-Privilege-Tax
        """
        
        # Check if the parameters are set upstream in the instance
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        income = self.income if hasattr(self, 'income') else income
        
        # State-specific parameters
        data = {
            'state_residence': 'Colorado', "sources": sources, "credits_state": credits_state, "county_residence": county_residence, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "state_agi": state_agi, "income": income
        }
        
        return self.__fetch__(data, 'state-taxes')


    def connecticut(self, sources=False, filing_status='single', capital_gains_short=0, capital_gains_long=0, credits_state=0, state_agi=75000, income=100000):
        """Taxes for Connecticut
        
        Parameters:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.

        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        Sources:
        https://portal.ct.gov/-/media/DRS/Forms/2022/Income/2022-CT-1040-Instructions_1222.pdf
        https://www.cga.ct.gov/2022/rpt/pdf/2022-R-0108.pdf
        https://www.cga.ct.gov/current/pub/chap_229.htm#sec_12-703
        https://portal.ct.gov/-/media/DRS/Forms/2022/Income/2022-CT-1040-Instructions_1222.pdf
        "https://portal.ct.gov/DRS/Publications/TSSNs/TSSN-29#:~:text=An%20individuals%20net%20capital%20gains
       portion%20of%20Social%20Security%20benefits"
        "https://portal.ct.gov/DRS/Publications/TSSNs/TSSN-29#:~:text=An%20individuals%20net%20capital%20gains
       portion%20of%20Social%20Security%20benefits"
        https://portal.ct.gov/DRS/Individuals/Resident-Income-Tax/Tax-Information
        """
        
        # Check if the parameters are set upstream in the instance
        sources = self.sources if hasattr(self, 'sources') else sources
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        income = self.income if hasattr(self, 'income') else income
        
        # State-specific parameters
        data = {
            'state_residence': 'Connecticut', "sources": sources, "filing_status": filing_status, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "credits_state": credits_state, "state_agi": state_agi, "income": income
        }
        
        return self.__fetch__(data, 'state-taxes')


    def delaware(self, sources=False, capital_gains_short=0, capital_gains_long=0, credits_state=0, federal_agi=75000, income=100000):
        """Taxes for Delaware
        
        Parameters:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        federal_agi : int, optional
            Adjusted Gross Income (AGI) for federal taxes. Default is 0. Sources - https://www.irs.gov/e-file-providers/definition-of-adjusted-gross-income, https://www.irs.gov/newsroom/taxpayers-should-know-the-difference-between-standard-and-itemized-deductions
        income : int, optional
            Gross income. Default is 0.

        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        Sources:
        https://revenue.delaware.gov/software-developer/tax-rate-changes/
        https://revenuefiles.delaware.gov/2022/PIT-RES_TY22_2022-01_PaperInteractive.pdf
        https://revenuefiles.delaware.gov/2022/PIT-RES_TY22_2022-02_Instructions.pdf
        https://revenue.delaware.gov/frequently-asked-questions/personal-income-tax-faqs/
        https://www.newcastlede.gov/DocumentCenter/View/20147/City-of-Wilmington-Wage-Tax-Information-PDF
        """
        
        # Check if the parameters are set upstream in the instance
        sources = self.sources if hasattr(self, 'sources') else sources
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        federal_agi = self.federal_agi if hasattr(self, 'federal_agi') else federal_agi
        income = self.income if hasattr(self, 'income') else income
        
        # State-specific parameters
        data = {
            'state_residence': 'Delaware', "sources": sources, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "credits_state": credits_state, "federal_agi": federal_agi, "income": income
        }
        
        return self.__fetch__(data, 'state-taxes')


    def florida(self, sources=False):
        """Taxes for Florida
        
        Parameters:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.

        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        Sources:
        https://floridarevenue.com/taxes/taxesfees/Pages/tax_interest_rates.aspx
        """
        
        # Check if the parameters are set upstream in the instance
        sources = self.sources if hasattr(self, 'sources') else sources
        
        # State-specific parameters
        data = {
            'state_residence': 'Florida', "sources": sources
        }
        
        return self.__fetch__(data, 'state-taxes')


    def georgia(self, sources=False, filing_status='single', capital_gains_short=0, capital_gains_long=0, credits_state=0, state_agi=75000, income=100000, dependents=0, filers_over_65=0):
        """Taxes for Georgia
        
        Parameters:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.

        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        Sources:
        https://dor.georgia.gov/tax-tables-georgia-tax-rate-schedule
        https://apps.dor.ga.gov/FillableForms/PDFViewer/Index?form=2022GA500
        https://apps.dor.ga.gov/FillableForms/PDFViewer/Index?form=2022GA500
        https://apps.dor.ga.gov/FillableForms/PDFViewer/Index?form=2022GA500
        https://dor.georgia.gov/years-individual-income-tax-forms
        https://dor.georgia.gov/years-individual-income-tax-forms
        """
        
        # Check if the parameters are set upstream in the instance
        sources = self.sources if hasattr(self, 'sources') else sources
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        
        # State-specific parameters
        data = {
            'state_residence': 'Georgia', "sources": sources, "filing_status": filing_status, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "credits_state": credits_state, "state_agi": state_agi, "income": income, "dependents": dependents, "filers_over_65": filers_over_65
        }
        
        return self.__fetch__(data, 'state-taxes')


    def hawaii(self, sources=False, filing_status='single', capital_gains_short=0, capital_gains_long=0, credits_state=0, state_agi=75000, income=100000, dependents=0, filers_over_65=0):
        """Taxes for Hawaii
        
        Parameters:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.

        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        Sources:
        https://tax.hawaii.gov/forms/d_18table-on/d_18table-on_p13/
        https://files.hawaii.gov/tax/forms/2018/18table-on.pdf
        https://files.hawaii.gov/tax/news/pubs/22outline.pdf
        https://files.hawaii.gov/tax/news/pubs/FastTaxRefGuide_2017.pdf
        https://www.tfhawaii.org/wordpress/state-tax-resources/mini-tax-guides/individual-income-tax-chapter-235/#:~:text=Long%20term%20capital%20gains%20are%20taxed%20at%20a%20maximum%20of%207.25%25
        https://www.grassrootinstitute.org/2023/02/hb337-increasing-tax-on-capital-gains-a-bad-risk/
        https://files.hawaii.gov/tax/news/pubs/22outline.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        sources = self.sources if hasattr(self, 'sources') else sources
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        
        # State-specific parameters
        data = {
            'state_residence': 'Hawaii', "sources": sources, "filing_status": filing_status, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "credits_state": credits_state, "state_agi": state_agi, "income": income, "dependents": dependents, "filers_over_65": filers_over_65
        }
        
        return self.__fetch__(data, 'state-taxes')


    def idaho(self, sources=False, filing_status='single', capital_gains_short=0, capital_gains_long=0, credits_state=0, state_agi=75000, income=100000, dependents=0, filers_over_65=0):
        """Taxes for Idaho
        
        Parameters:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.

        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        Sources:
        https://tax.idaho.gov/taxes/income-tax/individual-income/individual-income-tax-rate-schedule/
        https://tax.idaho.gov/wp-content/uploads/forms/EIN00046/EIN00046_03-01-2023.pdf
        https://tax.idaho.gov/taxes/income-tax/individual-income/idaho-source-income/
        https://tax.idaho.gov/taxes/income-tax/individual-income/idaho-source-income/
        """
        
        # Check if the parameters are set upstream in the instance
        sources = self.sources if hasattr(self, 'sources') else sources
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        
        # State-specific parameters
        data = {
            'state_residence': 'Idaho', "sources": sources, "filing_status": filing_status, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "credits_state": credits_state, "state_agi": state_agi, "income": income, "dependents": dependents, "filers_over_65": filers_over_65
        }
        
        return self.__fetch__(data, 'state-taxes')


    def illinois(self, sources=False, filing_status='single', capital_gains_short=0, capital_gains_long=0, credits_state=0, state_agi=75000, income=100000, dependents=0):
        """Taxes for Illinois
        
        Parameters:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2

        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        Sources:
        https://tax.illinois.gov/research/taxrates/income.html#IndividualIncome
        https://tax.illinois.gov/content/dam/soi/en/web/tax/forms/incometax/documents/currentyear/individual/il-1040-instr.pdf
        https://tax.illinois.gov/questionsandanswers/12.html
        https://tax.illinois.gov/content/dam/soi/en/web/tax/forms/incometax/documents/currentyear/individual/il-1040-instr.pdf
        https://tax.illinois.gov/research/taxrates/personalproperty.html
        https://tax.illinois.gov/localgovernments/income.html
        """
        
        # Check if the parameters are set upstream in the instance
        sources = self.sources if hasattr(self, 'sources') else sources
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        
        # State-specific parameters
        data = {
            'state_residence': 'Illinois', "sources": sources, "filing_status": filing_status, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "credits_state": credits_state, "state_agi": state_agi, "income": income, "dependents": dependents
        }
        
        return self.__fetch__(data, 'state-taxes')


    def indiana(self, sources=False, filing_status='single', county_residence='Other', capital_gains_short=0, capital_gains_long=0, credits_state=0, state_agi=75000, income=100000, dependents=0, filers_over_65=0):
        """Taxes for Indiana
        
        Parameters:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.

        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        Sources:
        https://www.in.gov/dor/tax-forms/2022-individual-income-tax-forms/
        https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-32.htm
        https://www.in.gov/dor/files/ib28.pdf
        https://www.in.gov/dor/tax-forms/2022-individual-income-tax-forms/
        https://www.in.gov/dor/files/2022-county-tax-rates-and-codes.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        sources = self.sources if hasattr(self, 'sources') else sources
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        
        # State-specific parameters
        data = {
            'state_residence': 'Indiana', "sources": sources, "filing_status": filing_status, "county_residence": county_residence, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "credits_state": credits_state, "state_agi": state_agi, "income": income, "dependents": dependents, "filers_over_65": filers_over_65
        }
        
        return self.__fetch__(data, 'state-taxes')


    def iowa(self, sources=False, filing_status='single', county_residence='Other', capital_gains_short=0, capital_gains_long=0, credits_state=0, state_agi=75000, income=100000, dependents=0):
        """Taxes for Iowa
        
        Parameters:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2

        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        Sources:
        https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
       who%20file%20a%20joint%20return
        https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
       who%20file%20a%20joint%20return
        https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
       who%20file%20a%20joint%20return
        https://tax.iowa.gov/iowa-illinois-reciprocal-agreement
        https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
       who%20file%20a%20joint%20return
        """
        
        # Check if the parameters are set upstream in the instance
        sources = self.sources if hasattr(self, 'sources') else sources
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        
        # State-specific parameters
        data = {
            'state_residence': 'Iowa', "sources": sources, "filing_status": filing_status, "county_residence": county_residence, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "credits_state": credits_state, "state_agi": state_agi, "income": income, "dependents": dependents
        }
        
        return self.__fetch__(data, 'state-taxes')


    def kansas(self, sources=False, filing_status='single', capital_gains_short=0, capital_gains_long=0, credits_state=0, state_agi=75000, income=100000, dependents=0, filers_over_65=0):
        """Taxes for Kansas
        
        Parameters:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.

        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        Sources:
        https://www.ksrevenue.gov/taxrates.html
        https://www.ksrevenue.gov/incomebook22.html#0
        https://www.ksrevenue.gov/incomebook22.html#0
        https://www.ksrevenue.gov/incomebook22.html#0
        https://www.ksrevenue.gov/incomebook22.html
        https://www.ksrevenue.gov/incomebook22.html
        """
        
        # Check if the parameters are set upstream in the instance
        sources = self.sources if hasattr(self, 'sources') else sources
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        
        # State-specific parameters
        data = {
            'state_residence': 'Kansas', "sources": sources, "filing_status": filing_status, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "credits_state": credits_state, "state_agi": state_agi, "income": income, "dependents": dependents, "filers_over_65": filers_over_65
        }
        
        return self.__fetch__(data, 'state-taxes')


    def kentucky(self, sources=False, filing_status='single', county_residence='Other', capital_gains_short=0, capital_gains_long=0, credits_state=0, state_agi=75000, income=100000):
        """Taxes for Kentucky
        
        Parameters:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.

        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        Sources:
        https://revenue.ky.gov/News/Pages/Dor-Announces-Updates-to-Individual-Income-Tax-for-2023-Tax-Year.aspx
        https://revenue.ky.gov/Individual/Individual-Income-Tax/Pages/default.aspx
        https://revenue.ky.gov/News/Pages/Dor-Announces-Updates-to-Individual-Income-Tax-for-2023-Tax-Year.aspx
        https://revenue.ky.gov/Individual/Individual-Income-Tax/Pages/default.aspx
        https://revenue.ky.gov/Dor%20Training%20Materials/103%20KAR%2017.140.%20Individual%20income%20tax%20-%20reciprocity%20-%20nonresidents.pdf
        https://revenue.ky.gov/Individual/Individual-Income-Tax/Pages/default.aspx
        https://nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2021/Taxes-21-32.htm?taxmap=true
        """
        
        # Check if the parameters are set upstream in the instance
        sources = self.sources if hasattr(self, 'sources') else sources
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        income = self.income if hasattr(self, 'income') else income
        
        # State-specific parameters
        data = {
            'state_residence': 'Kentucky', "sources": sources, "filing_status": filing_status, "county_residence": county_residence, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "credits_state": credits_state, "state_agi": state_agi, "income": income
        }
        
        return self.__fetch__(data, 'state-taxes')


    def louisiana(self, sources=False, filing_status='single', capital_gains_short=0, capital_gains_long=0, credits_state=0, state_agi=75000, income=100000, dependents=0, filers_over_65=0):
        """Taxes for Louisiana
        
        Parameters:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.

        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        Sources:
        https://revenue.louisiana.gov/individualincometax
        https://revenue.louisiana.gov/individualincometax
        https://www.revenue.louisiana.gov/taxforms/6935(11_02)F.pdf
        https://revenue.louisiana.gov/individualincometax
        https://revenue.louisiana.gov/individualincometax
        """
        
        # Check if the parameters are set upstream in the instance
        sources = self.sources if hasattr(self, 'sources') else sources
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        
        # State-specific parameters
        data = {
            'state_residence': 'Louisiana', "sources": sources, "filing_status": filing_status, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "credits_state": credits_state, "state_agi": state_agi, "income": income, "dependents": dependents, "filers_over_65": filers_over_65
        }
        
        return self.__fetch__(data, 'state-taxes')


    def maine(self, sources=False, filing_status='single', capital_gains_short=0, capital_gains_long=0, credits_state=0, state_agi=75000, income=100000, dependents=0, filers_over_65=0):
        """Taxes for Maine
        
        Parameters:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.

        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        Sources:
        https://www.maine.gov/revenue/tax-return-forms/individual-income-tax-2022
        https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/ind_tax_rate_sched_2022.pdf
        https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/22_1040me_book_gen_instruc_revisedFeb23.pdf
        https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/22_1040me_book_gen_instruc_revisedFeb23.pdf
        https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/22_1040me_book_gen_instruc_revisedFeb23.pdf
        https://www.maine.gov/revenue/tax-return-forms/individual-income-tax-2022
        """
        
        # Check if the parameters are set upstream in the instance
        sources = self.sources if hasattr(self, 'sources') else sources
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        
        # State-specific parameters
        data = {
            'state_residence': 'Maine', "sources": sources, "filing_status": filing_status, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "credits_state": credits_state, "state_agi": state_agi, "income": income, "dependents": dependents, "filers_over_65": filers_over_65
        }
        
        return self.__fetch__(data, 'state-taxes')


    def maryland(self, sources=False, filing_status='single', county_residence='Other', capital_gains_short=0, capital_gains_long=0, credits_state=0, state_agi=75000, income=100000, dependents=0):
        """Taxes for Maryland
        
        Parameters:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2

        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        Sources:
        https://www.marylandtaxes.gov/individual/income/tax-info/tax-rates.php
        https://www.marylandtaxes.gov/forms/22_forms/Resident_Booklet.pdf
        https://www.marylandtaxes.gov/forms/Tax_Publications/Administrative_Releases/Income_and_Estate_Tax_Releases/ar_it37.pdf
        https://www.marylandtaxes.gov/individual/index.php
        https://www.marylandtaxes.gov/individual/income/tax-info/tax-rates.php
        """
        
        # Check if the parameters are set upstream in the instance
        sources = self.sources if hasattr(self, 'sources') else sources
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        
        # State-specific parameters
        data = {
            'state_residence': 'Maryland', "sources": sources, "filing_status": filing_status, "county_residence": county_residence, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "credits_state": credits_state, "state_agi": state_agi, "income": income, "dependents": dependents
        }
        
        return self.__fetch__(data, 'state-taxes')


    def massachusetts(self, sources=False, filing_status='single', capital_gains_short=0, capital_gains_long=0, credits_state=0, state_agi=75000, income=100000, dependents=0):
        """Taxes for Massachusetts
        
        Parameters:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2

        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        Sources:
        https://www.mass.gov/service-details/massachusetts-tax-rates
        https://www.mass.gov/service-details/view-massachusetts-personal-income-tax-exemptions
        https://www.mass.gov/service-details/view-massachusetts-personal-income-tax-exemptions
        https://www.mass.gov/service-details/view-child-and-dependent-related-credits
        https://www.mass.gov/service-details/massachusetts-tax-rates
        https://www.mass.gov/personal-income-tax
        """
        
        # Check if the parameters are set upstream in the instance
        sources = self.sources if hasattr(self, 'sources') else sources
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        
        # State-specific parameters
        data = {
            'state_residence': 'Massachusetts', "sources": sources, "filing_status": filing_status, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "credits_state": credits_state, "state_agi": state_agi, "income": income, "dependents": dependents
        }
        
        return self.__fetch__(data, 'state-taxes')


    def michigan(self, sources=False, filing_status='single', county_residence='Other', county_occupation='Other', capital_gains_short=0, capital_gains_long=0, credits_state=0, state_agi=75000, income=100000, dependents=0):
        """Taxes for Michigan
        
        Parameters:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        county_occupation : str, optional
            County of occupation. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2

        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        Sources:
        https://www.michigan.gov/taxes/iit/new-developments-for-tax-year-2022
        "https://www.michigan.gov/taxes/-/media/Project/Websites/taxes/Forms/2023/2023-SUW/446_Withholding-Guide_2023.pdf?rev=177ca3d8cb034392bf84e317ed78c320&hash=19C5F42F43A89A91863D229349F48883#:~:text=The%20exemption%20amount%20is%20%245%2C400
       employees%20Michigan%20income%20tax%20return"
        "https://www.michigan.gov/taxes/-/media/Project/Websites/taxes/Forms/2023/2023-SUW/446_Withholding-Guide_2023.pdf?rev=177ca3d8cb034392bf84e317ed78c320&hash=19C5F42F43A89A91863D229349F48883#:~:text=The%20exemption%20amount%20is%20%245%2C400
       employees%20Michigan%20income%20tax%20return"
        https://www.michigan.gov/taxes/questions/iit/accordion/residency/are-my-wages-earned-in-another-state-taxable-in-michigan-if-i-am-a-michigan-resident-1
        https://www.michigan.gov/taxes/-/media/Project/Websites/taxes/Forms/2022/2022-IIT-Forms/BOOK_MI-1040.pdf?rev=4524f1cbbc0243b194ab05a6680be75b&hash=40DFE71E651F7C1D65C31F6CD03D944F
        https://www.michigan.gov/taxes/questions/iit/accordion/general/what-cities-impose-an-income-tax
        """
        
        # Check if the parameters are set upstream in the instance
        sources = self.sources if hasattr(self, 'sources') else sources
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        county_occupation = self.county_occupation if hasattr(self, 'county_occupation') else county_occupation
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        
        # State-specific parameters
        data = {
            'state_residence': 'Michigan', "sources": sources, "filing_status": filing_status, "county_residence": county_residence, "county_occupation": county_occupation, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "credits_state": credits_state, "state_agi": state_agi, "income": income, "dependents": dependents
        }
        
        return self.__fetch__(data, 'state-taxes')


    def minnesota(self, sources=False, filing_status='single', capital_gains_short=0, capital_gains_long=0, credits_state=0, state_agi=75000, income=100000, dependents=0):
        """Taxes for Minnesota
        
        Parameters:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2

        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        Sources:
        https://www.revenue.state.mn.us/minnesota-income-tax-rates-and-brackets
        https://www.revenue.state.mn.us/child-and-dependent-care-credit
        https://www.revenue.state.mn.us/dependent-exemptions
        https://www.revenue.state.mn.us/minnesota-standard-deduction
        https://mn.gov/mmb/assets/mwr_form_tcm1059-128581.pdf
        https://www.revenue.state.mn.us/individual-income-tax
        https://www.revenue.state.mn.us/individual-income-tax
        """
        
        # Check if the parameters are set upstream in the instance
        sources = self.sources if hasattr(self, 'sources') else sources
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        
        # State-specific parameters
        data = {
            'state_residence': 'Minnesota', "sources": sources, "filing_status": filing_status, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "credits_state": credits_state, "state_agi": state_agi, "income": income, "dependents": dependents
        }
        
        return self.__fetch__(data, 'state-taxes')


    def mississippi(self, sources=False, filing_status='single', capital_gains_short=0, capital_gains_long=0, credits_state=0, state_agi=75000, income=100000, dependents=0):
        """Taxes for Mississippi
        
        Parameters:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2

        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        Sources:
        https://www.dor.ms.gov/individual/tax-rates
        https://www.dor.ms.gov/individual/tax-rates
        https://www.dor.ms.gov/individual/tax-rates
        https://www.dor.ms.gov/individual
        https://www.dor.ms.gov/individual
        """
        
        # Check if the parameters are set upstream in the instance
        sources = self.sources if hasattr(self, 'sources') else sources
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        
        # State-specific parameters
        data = {
            'state_residence': 'Mississippi', "sources": sources, "filing_status": filing_status, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "credits_state": credits_state, "state_agi": state_agi, "income": income, "dependents": dependents
        }
        
        return self.__fetch__(data, 'state-taxes')


    def missouri(self, sources=False, filing_status='single', county_residence='Other', capital_gains_short=0, capital_gains_long=0, credits_state=0, state_agi=75000, income=100000, dependents=0, filers_over_65=0):
        """Taxes for Missouri
        
        Parameters:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.

        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        Sources:
        https://dor.mo.gov/taxation/individual/tax-types/income/year-changes/
        https://dor.mo.gov/taxation/individual/tax-types/income/year-changes/
        https://dor.mo.gov/taxation/individual/tax-types/income/year-changes/
        https://www.kcmo.gov/city-hall/departments/finance/earnings-tax
        https://www.stlouis-mo.gov/government/departments/collector/earnings-tax/individual-earnings-tax-info.cfm
        """
        
        # Check if the parameters are set upstream in the instance
        sources = self.sources if hasattr(self, 'sources') else sources
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        
        # State-specific parameters
        data = {
            'state_residence': 'Missouri', "sources": sources, "filing_status": filing_status, "county_residence": county_residence, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "credits_state": credits_state, "state_agi": state_agi, "income": income, "dependents": dependents, "filers_over_65": filers_over_65
        }
        
        return self.__fetch__(data, 'state-taxes')


    def montana(self, sources=False, filing_status='single', capital_gains_short=0, capital_gains_long=0, credits_state=0, state_agi=75000, income=100000, dependents=0, filers_over_65=0):
        """Taxes for Montana
        
        Parameters:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.

        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        Sources:
        https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
        https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
        https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
        https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
        https://mtrevenue.gov/wp-content/uploads/2022/12/montana-employees-withholding-allowance-and-exemption-certificate-form-mw-4.pdf
        https://mtrevenue.gov/2021/11/29/simplification-of-montana-income-taxation/?utm_source=rss&utm_medium=rss&utm_campaign=simplification-of-montana-income-taxation
        https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        sources = self.sources if hasattr(self, 'sources') else sources
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        
        # State-specific parameters
        data = {
            'state_residence': 'Montana', "sources": sources, "filing_status": filing_status, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "credits_state": credits_state, "state_agi": state_agi, "income": income, "dependents": dependents, "filers_over_65": filers_over_65
        }
        
        return self.__fetch__(data, 'state-taxes')


    def nebraska(self, sources=False, filing_status='single', capital_gains_short=0, capital_gains_long=0, credits_state=0, state_agi=75000, income=100000, dependents=0):
        """Taxes for Nebraska
        
        Parameters:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2

        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        Sources:
        https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2023_Circular_EN_8-429r2023_Whole_Book_11.pdf
        https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
        https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
        https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
        https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        sources = self.sources if hasattr(self, 'sources') else sources
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        
        # State-specific parameters
        data = {
            'state_residence': 'Nebraska', "sources": sources, "filing_status": filing_status, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "credits_state": credits_state, "state_agi": state_agi, "income": income, "dependents": dependents
        }
        
        return self.__fetch__(data, 'state-taxes')


    def nevada(self, sources=False):
        """Taxes for Nevada
        
        Parameters:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.

        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        Sources:
        https://www.nvsos.gov/sos/businesses/the-nevada-advantage
        """
        
        # Check if the parameters are set upstream in the instance
        sources = self.sources if hasattr(self, 'sources') else sources
        
        # State-specific parameters
        data = {
            'state_residence': 'Nevada', "sources": sources
        }
        
        return self.__fetch__(data, 'state-taxes')


    def new_hampshire(self, sources=False):
        """Taxes for New Hampshire
        
        Parameters:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.

        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        Sources:
        https://www.revenue.nh.gov/assistance/tax-overview.htm
        """
        
        # Check if the parameters are set upstream in the instance
        sources = self.sources if hasattr(self, 'sources') else sources
        
        # State-specific parameters
        data = {
            'state_residence': 'NewHampshire', "sources": sources
        }
        
        return self.__fetch__(data, 'state-taxes')


    def new_jersey(self, sources=False, filing_status='single', capital_gains_short=0, capital_gains_long=0, credits_state=0, state_agi=75000, income=100000, dependents=0, filers_over_65=0):
        """Taxes for New Jersey
        
        Parameters:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.

        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        Sources:
        https://www.state.nj.us/treasury/taxation/taxtables.shtml
        https://www.state.nj.us/treasury/taxation/njit13.shtml
        https://www.nj.gov/treasury/taxation/njit25.shtml
        https://www.state.nj.us/treasury/taxation/individuals/freqqiti.shtml
        https://www.state.nj.us/treasury/taxation/localtax.shtml
        """
        
        # Check if the parameters are set upstream in the instance
        sources = self.sources if hasattr(self, 'sources') else sources
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        
        # State-specific parameters
        data = {
            'state_residence': 'NewJersey', "sources": sources, "filing_status": filing_status, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "credits_state": credits_state, "state_agi": state_agi, "income": income, "dependents": dependents, "filers_over_65": filers_over_65
        }
        
        return self.__fetch__(data, 'state-taxes')


    def new_mexico(self, sources=False, filing_status='single', capital_gains_short=0, capital_gains_long=0, credits_state=0, state_agi=75000, income=100000, dependents=0):
        """Taxes for New Mexico
        
        Parameters:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2

        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        Sources:
        https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/fdf3c548-8aba-4b9c-9eb4-bb564c716015/FYI-104.pdf
        https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/0bf943fd-652e-4400-bb1b-cd8397eb5a95/2022pit-adj.pdf
        https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/2f1a6781-9534-4436-b427-1557f9592099/2022pit-adj-ins.pdf
        https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/0bf943fd-652e-4400-bb1b-cd8397eb5a95/2022pit-adj.pdf
        https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/1afc56af-ea90-4d48-82e5-1f9aeb43255a/PITbook2022.pdf
        https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/fdf3c548-8aba-4b9c-9eb4-bb564c716015/FYI-104.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        sources = self.sources if hasattr(self, 'sources') else sources
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        
        # State-specific parameters
        data = {
            'state_residence': 'NewMexico', "sources": sources, "filing_status": filing_status, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "credits_state": credits_state, "state_agi": state_agi, "income": income, "dependents": dependents
        }
        
        return self.__fetch__(data, 'state-taxes')


    def new_york(self, sources=False, filing_status='single', county_residence='Other', capital_gains_short=0, capital_gains_long=0, credits_state=0, state_agi=75000, income=100000, dependents=0):
        """Taxes for New York
        
        Parameters:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2

        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        Sources:
        https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nys-tax-rate-schedule
        https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nys-tax-rate-schedule
        https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nys-tax-rate-schedule
        https://www.tax.ny.gov/pit/
        https://www.tax.ny.gov/pit/file/nonresident-faqs.htm
        https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#yonkers-tax-rate-schedule
        https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nyc-tax-rate-schedule
        """
        
        # Check if the parameters are set upstream in the instance
        sources = self.sources if hasattr(self, 'sources') else sources
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        
        # State-specific parameters
        data = {
            'state_residence': 'NewYork', "sources": sources, "filing_status": filing_status, "county_residence": county_residence, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "credits_state": credits_state, "state_agi": state_agi, "income": income, "dependents": dependents
        }
        
        return self.__fetch__(data, 'state-taxes')


    def north_carolina(self, sources=False, filing_status='single', capital_gains_short=0, capital_gains_long=0, credits_state=0, state_agi=75000, income=100000):
        """Taxes for North Carolina
        
        Parameters:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.

        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        Sources:
        https://www.ncdor.gov/taxes-forms/tax-rate-schedules
        https://www.ncdor.gov/taxes-forms/individual-income-tax/north-carolina-standard-deduction-or-north-carolina-itemized-deductions
        https://www.ncdor.gov/taxes-forms/individual-income-tax
        https://www.ncdor.gov/taxes-forms/individual-income-tax
        """
        
        # Check if the parameters are set upstream in the instance
        sources = self.sources if hasattr(self, 'sources') else sources
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        income = self.income if hasattr(self, 'income') else income
        
        # State-specific parameters
        data = {
            'state_residence': 'NorthCarolina', "sources": sources, "filing_status": filing_status, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "credits_state": credits_state, "state_agi": state_agi, "income": income
        }
        
        return self.__fetch__(data, 'state-taxes')


    def north_dakota(self, sources=False, filing_status='single', capital_gains_short=0, capital_gains_long=0, credits_state=0, state_agi=75000, income=100000, dependents=0, filers_over_65=0):
        """Taxes for North Dakota
        
        Parameters:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.

        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        Sources:
        https://www.tax.nd.gov/forms
        https://www.tax.nd.gov/sites/www/files/documents/forms/individual/2022-iit/2022-individual-income-tax-booklet.pdf
        https://www.tax.nd.gov/sites/www/files/documents/forms/2022-individual-income-tax-booklet.pdf
        https://www.tax.nd.gov/tax-exemptions-credits/income-tax-exemptions-credits
        https://www.tax.nd.gov/sites/www/files/documents/forms/business/ndwrfillable.pdf
        https://www.tax.nd.gov/news/tax-legislative-changes/significant-changes-law/individual-income-tax-history
        https://www.tax.nd.gov/forms
        """
        
        # Check if the parameters are set upstream in the instance
        sources = self.sources if hasattr(self, 'sources') else sources
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        
        # State-specific parameters
        data = {
            'state_residence': 'NorthDakota', "sources": sources, "filing_status": filing_status, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "credits_state": credits_state, "state_agi": state_agi, "income": income, "dependents": dependents, "filers_over_65": filers_over_65
        }
        
        return self.__fetch__(data, 'state-taxes')


    def ohio(self, sources=False, filing_status='single', county_residence='Other', capital_gains_short=0, capital_gains_long=0, credits_state=0, state_agi=75000, income=100000, dependents=0):
        """Taxes for Ohio
        
        Parameters:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2

        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        Sources:
        https://tax.ohio.gov/individual/resources/annual-tax-rates
        https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
        https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
        https://tax.ohio.gov/static/forms/employer_withholding/generic/wth_it4nr.pdf
        https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
        https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
        https://www.ritaohio.com/TaxRatesTable
        """
        
        # Check if the parameters are set upstream in the instance
        sources = self.sources if hasattr(self, 'sources') else sources
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        
        # State-specific parameters
        data = {
            'state_residence': 'Ohio', "sources": sources, "filing_status": filing_status, "county_residence": county_residence, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "credits_state": credits_state, "state_agi": state_agi, "income": income, "dependents": dependents
        }
        
        return self.__fetch__(data, 'state-taxes')


    def oklahoma(self, sources=False, filing_status='single', capital_gains_short=0, capital_gains_long=0, credits_state=0, state_agi=75000, income=100000, dependents=0, filers_over_65=0):
        """Taxes for Oklahoma
        
        Parameters:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.

        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        Sources:
        https://oklahoma.gov/content/dam/ok/en/tax/documents/resources/publications/businesses/withholding-tables/WHTables-2022.pdf
        https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
        https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
        https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
        https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        sources = self.sources if hasattr(self, 'sources') else sources
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        
        # State-specific parameters
        data = {
            'state_residence': 'Oklahoma', "sources": sources, "filing_status": filing_status, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "credits_state": credits_state, "state_agi": state_agi, "income": income, "dependents": dependents, "filers_over_65": filers_over_65
        }
        
        return self.__fetch__(data, 'state-taxes')


    def oregon(self, sources=False, filing_status='single', capital_gains_short=0, capital_gains_long=0, credits_state=0, state_agi=75000, income=100000, dependents=0, filers_over_65=0):
        """Taxes for Oregon
        
        Parameters:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.

        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        Sources:
        https://www.oregon.gov/dor/programs/individuals/Pages/PIT.aspx
        https://www.oregon.gov/dor/programs/individuals/Documents/Full%20year%20resident
       %20Form%20or-40%20filers.pdf
        https://www.oregon.gov/dor/programs/individuals/Documents/Part-year%20and%20nonresident
       %20Form%20or-40-P%20and%20or-40-N%20filers.pdf.pdf
        https://www.oregon.gov/dor/forms/FormsPubs/form-or-40_101-040_2022.pdf
        https://www.oregon.gov/dor/forms/FormsPubs/form-or-40_101-040_2022.pdf
        https://www.oregon.gov/dor/forms/FormsPubs/form-or-40_101-040_2022.pdf
        https://www.portland.gov/revenue/personal-tax
        """
        
        # Check if the parameters are set upstream in the instance
        sources = self.sources if hasattr(self, 'sources') else sources
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        
        # State-specific parameters
        data = {
            'state_residence': 'Oregon', "sources": sources, "filing_status": filing_status, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "credits_state": credits_state, "state_agi": state_agi, "income": income, "dependents": dependents, "filers_over_65": filers_over_65
        }
        
        return self.__fetch__(data, 'state-taxes')


    def pennsylvania(self, sources=False, county_residence='Other', county_occupation='Other', capital_gains_short=0, capital_gains_long=0, credits_state=0, state_agi=75000, income=100000):
        """Taxes for Pennsylvania
        
        Parameters:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        county_residence : str, optional
            County of residence. Default is Other.
        county_occupation : str, optional
            County of occupation. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.

        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        Sources:
        https://www.revenue.pa.gov/Tax%20Rates/Pages/default.aspx
        https://www.revenue.pa.gov/FormsandPublications/FormsforIndividuals/PIT/Documents/2022/2022_pa-40in.pdf
        https://www.revenue.pa.gov/FormsandPublications/FormsforBusinesses/EmployerWithholding/Documents/rev-419.pdf
        https://www.revenue.pa.gov/FormsandPublications/FormsforIndividuals/PIT/Documents/2022/2022_pa-40in.pdf
        https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-25.htm
        """
        
        # Check if the parameters are set upstream in the instance
        sources = self.sources if hasattr(self, 'sources') else sources
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        county_occupation = self.county_occupation if hasattr(self, 'county_occupation') else county_occupation
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        income = self.income if hasattr(self, 'income') else income
        
        # State-specific parameters
        data = {
            'state_residence': 'Pennsylvania', "sources": sources, "county_residence": county_residence, "county_occupation": county_occupation, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "credits_state": credits_state, "state_agi": state_agi, "income": income
        }
        
        return self.__fetch__(data, 'state-taxes')


    def rhode_island(self, sources=False, filing_status='single', capital_gains_short=0, capital_gains_long=0, credits_state=0, state_agi=75000, income=100000, dependents=0):
        """Taxes for Rhode Island
        
        Parameters:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2

        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        Sources:
        https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
        https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
        https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2022_1040WE_w_0.pdf
        https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
        https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        sources = self.sources if hasattr(self, 'sources') else sources
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        
        # State-specific parameters
        data = {
            'state_residence': 'RhodeIsland', "sources": sources, "filing_status": filing_status, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "credits_state": credits_state, "state_agi": state_agi, "income": income, "dependents": dependents
        }
        
        return self.__fetch__(data, 'state-taxes')


    def south_carolina(self, sources=False, filing_status='single', capital_gains_short=0, capital_gains_long=0, credits_state=0, state_agi=75000, income=100000, dependents=0, filers_over_65=0):
        """Taxes for South Carolina
        
        Parameters:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.

        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        Sources:
        https://dor.sc.gov/resources-site/lawandpolicy/Advisory%20Opinions/IL22-15.pdf
        https://dor.sc.gov/forms-site/Forms/IITPacket_2022.pdf
        https://www.scstatehouse.gov/code/t12c006.php
        https://dor.sc.gov/resources-site/lawandpolicy/Advisory%20Opinions/IL22-15.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        sources = self.sources if hasattr(self, 'sources') else sources
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        
        # State-specific parameters
        data = {
            'state_residence': 'SouthCarolina', "sources": sources, "filing_status": filing_status, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "credits_state": credits_state, "state_agi": state_agi, "income": income, "dependents": dependents, "filers_over_65": filers_over_65
        }
        
        return self.__fetch__(data, 'state-taxes')


    def south_dakota(self, sources=False):
        """Taxes for South Dakota
        
        Parameters:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.

        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        Sources:
        https://dor.sd.gov/individuals/taxes/
        """
        
        # Check if the parameters are set upstream in the instance
        sources = self.sources if hasattr(self, 'sources') else sources
        
        # State-specific parameters
        data = {
            'state_residence': 'SouthDakota', "sources": sources
        }
        
        return self.__fetch__(data, 'state-taxes')


    def tennessee(self, sources=False):
        """Taxes for Tennessee
        
        Parameters:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.

        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        Sources:
        https://www.tn.gov/revenue/taxes.html
        """
        
        # Check if the parameters are set upstream in the instance
        sources = self.sources if hasattr(self, 'sources') else sources
        
        # State-specific parameters
        data = {
            'state_residence': 'Tennessee', "sources": sources
        }
        
        return self.__fetch__(data, 'state-taxes')


    def texas(self, sources=False):
        """Taxes for Texas
        
        Parameters:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.

        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        Sources:
        https://comptroller.texas.gov/taxes/
        """
        
        # Check if the parameters are set upstream in the instance
        sources = self.sources if hasattr(self, 'sources') else sources
        
        # State-specific parameters
        data = {
            'state_residence': 'Texas', "sources": sources
        }
        
        return self.__fetch__(data, 'state-taxes')


    def utah(self, sources=False, filing_status='single', capital_gains_short=0, capital_gains_long=0, credits_state=0, state_agi=75000, income=100000, dependents=0):
        """Taxes for Utah
        
        Parameters:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2

        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        Sources:
        https://incometax.utah.gov/paying/tax-rates
        https://tax.utah.gov/forms/current/tc-40.pdf
        https://tax.utah.gov/forms/current/tc-40.pdf
        https://tax.utah.gov/forms/current/tc-40.pdf
        https://tax.utah.gov/forms/current/tc-40.pdf
        https://tax.utah.gov/forms/current/tc-40.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        sources = self.sources if hasattr(self, 'sources') else sources
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        
        # State-specific parameters
        data = {
            'state_residence': 'Utah', "sources": sources, "filing_status": filing_status, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "credits_state": credits_state, "state_agi": state_agi, "income": income, "dependents": dependents
        }
        
        return self.__fetch__(data, 'state-taxes')


    def vermont(self, sources=False, filing_status='single', capital_gains_short=0, capital_gains_long=0, credits_state=0, state_agi=75000, income=100000, dependents=0, filers_over_65=0):
        """Taxes for Vermont
        
        Parameters:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.

        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        Sources:
        https://tax.vermont.gov/sites/tax/files/documents/GB-1210-2023.pdf
        https://tax.vermont.gov/sites/tax/files/documents/Income%20Booklet-2022.pdf
        https://tax.vermont.gov/sites/tax/files/documents/Income%20Booklet-2022.pdf
        https://tax.vermont.gov/sites/tax/files/documents/Income%20Booklet-2022.pdf
        https://tax.vermont.gov/individuals/personal-income-tax/tax-credits
        https://tax.vermont.gov/individuals/personal-income-tax/taxable-income
        https://tax.vermont.gov/individuals/personal-income-tax/local-option-tax
        """
        
        # Check if the parameters are set upstream in the instance
        sources = self.sources if hasattr(self, 'sources') else sources
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        
        # State-specific parameters
        data = {
            'state_residence': 'Vermont', "sources": sources, "filing_status": filing_status, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "credits_state": credits_state, "state_agi": state_agi, "income": income, "dependents": dependents, "filers_over_65": filers_over_65
        }
        
        return self.__fetch__(data, 'state-taxes')


    def virginia(self, sources=False, filing_status='single', capital_gains_short=0, capital_gains_long=0, credits_state=0, state_agi=75000, income=100000, dependents=0, filers_over_65=0):
        """Taxes for Virginia
        
        Parameters:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.

        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        Sources:
        https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
        https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
        https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
        https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
        https://www.tax.virginia.gov/reciprocity
        https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
        https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        sources = self.sources if hasattr(self, 'sources') else sources
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        
        # State-specific parameters
        data = {
            'state_residence': 'Virginia', "sources": sources, "filing_status": filing_status, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "credits_state": credits_state, "state_agi": state_agi, "income": income, "dependents": dependents, "filers_over_65": filers_over_65
        }
        
        return self.__fetch__(data, 'state-taxes')


    def washington(self, sources=False, capital_gains_short=0, capital_gains_long=0):
        """Taxes for Washington
        
        Parameters:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.

        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        Sources:
        https://dor.wa.gov/taxes-rates/income-tax
        https://dor.wa.gov/taxes-rates/other-taxes/capital-gains-tax
        https://dor.wa.gov/taxes-rates/income-tax
        """
        
        # Check if the parameters are set upstream in the instance
        sources = self.sources if hasattr(self, 'sources') else sources
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        
        # State-specific parameters
        data = {
            'state_residence': 'Washington', "sources": sources, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long
        }
        
        return self.__fetch__(data, 'state-taxes')


    def west_virginia(self, sources=False, filing_status='single', county_occupation='Other', pay_periods=24, capital_gains_short=0, capital_gains_long=0, credits_state=0, state_agi=75000, income=100000, dependents=0):
        """Taxes for West Virginia
        
        Parameters:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_occupation : str, optional
            County of occupation. Default is Other.
        pay_periods : int, optional
            Number of pay periods. Options are 13, 24, 26, 52. Default is 24.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2

        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        Sources:
        https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
        https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
        https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
        https://tax.wv.gov/Documents/TSD/tsd438.pdf
        https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
        https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2011/Taxes-11-35att.html
        """
        
        # Check if the parameters are set upstream in the instance
        sources = self.sources if hasattr(self, 'sources') else sources
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        county_occupation = self.county_occupation if hasattr(self, 'county_occupation') else county_occupation
        pay_periods = self.pay_periods if hasattr(self, 'pay_periods') else pay_periods
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        
        # State-specific parameters
        data = {
            'state_residence': 'WestVirginia', "sources": sources, "filing_status": filing_status, "county_occupation": county_occupation, "pay_periods": pay_periods, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "credits_state": credits_state, "state_agi": state_agi, "income": income, "dependents": dependents
        }
        
        return self.__fetch__(data, 'state-taxes')


    def wisconsin(self, sources=False, filing_status='single', capital_gains_short=0, capital_gains_long=0, credits_state=0, state_agi=75000, income=100000, dependents=0):
        """Taxes for Wisconsin
        
        Parameters:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2

        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        Sources:
        https://www.revenue.wi.gov/TaxForms2023/2023-Form1-ES-Inst.pdf
        https://www.revenue.wi.gov/TaxForms2023/2023-Form1-ES-Inst.pdf
        https://www.revenue.wi.gov/TaxForms2023/2023-Form1-ES-Inst.pdf
        https://www.revenue.wi.gov/Pages/FAQS/pcs-work.aspx
        https://www.revenue.wi.gov/Dor%20Publications/pb103.pdf
        https://www.revenue.wi.gov/Pages/FAQS/pcs-work.aspx
        """
        
        # Check if the parameters are set upstream in the instance
        sources = self.sources if hasattr(self, 'sources') else sources
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        
        # State-specific parameters
        data = {
            'state_residence': 'Wisconsin', "sources": sources, "filing_status": filing_status, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "credits_state": credits_state, "state_agi": state_agi, "income": income, "dependents": dependents
        }
        
        return self.__fetch__(data, 'state-taxes')


    def wyoming(self, sources=False):
        """Taxes for Wyoming
        
        Parameters:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.

        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        Sources:
        https://revenue.wyo.gov/
        """
        
        # Check if the parameters are set upstream in the instance
        sources = self.sources if hasattr(self, 'sources') else sources
        
        # State-specific parameters
        data = {
            'state_residence': 'Wyoming', "sources": sources
        }
        
        return self.__fetch__(data, 'state-taxes')


    def alabama_and_alaska(self, county_residence='Other', state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Alabama:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Alaska:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Alabama:
        - https://www.revenue.alabama.gov/ultraviewer/viewer/basic_viewer/index.html?form=2023/02/22f40.pdf
          https://www.revenue.alabama.gov/faqs/what-is-alabamas-individual-income-tax-rate/
          https://www.revenue.alabama.gov/ultraviewer/viewer/basic_viewer/index.html?form=2023/02/22f40.pdf
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-24.htm?taxmap=true
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-24.htm?taxmap=true
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-24.htm?taxmap=true
          https://www.revenue.alabama.gov/individual-corporate/income-to-be-reported-on-the-alabama-income-tax-return/
          https://www.almonline.org/TaxRates.aspx#OccupationalTax
        
        Alaska:
        -         
        - https://www.commerce.alaska.gov/web/dcra/officeofthestateassessor/alaskataxfacts.aspx
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Alabama', 'state_occupation': 'Alaska', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def alabama_and_arizona(self, county_residence='Other', state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Alabama:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Arizona:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        credits_state : int, optional
            Credits for state taxes. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Alabama:
        - https://www.revenue.alabama.gov/ultraviewer/viewer/basic_viewer/index.html?form=2023/02/22f40.pdf
          https://www.revenue.alabama.gov/faqs/what-is-alabamas-individual-income-tax-rate/
          https://www.revenue.alabama.gov/ultraviewer/viewer/basic_viewer/index.html?form=2023/02/22f40.pdf
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-24.htm?taxmap=true
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-24.htm?taxmap=true
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-24.htm?taxmap=true
          https://www.revenue.alabama.gov/individual-corporate/income-to-be-reported-on-the-alabama-income-tax-return/
          https://www.almonline.org/TaxRates.aspx#OccupationalTax
        
        Arizona:
        -         
        - https://azdor.gov/forms/individual/form-140-x-y-tables
          https://azdor.gov/forms/individual/form-140-x-y-tables
          https://azdor.gov/forms/individual/form-140-x-y-tables
          https://azdor.gov/businesses-arizona/withholding-tax/withholding-exceptions
          "https://azdor.gov/individuals/income-tax-filing-assistance/identifying-other-taxable-income#:~:text=In%20regards%20to%20capital%20gains
         the%20individuals%20regular%20tax%20rate.
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Alabama', 'state_occupation': 'Arizona', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def alabama_and_arkansas(self, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Alabama:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Arkansas:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Alabama:
        - https://www.revenue.alabama.gov/ultraviewer/viewer/basic_viewer/index.html?form=2023/02/22f40.pdf
          https://www.revenue.alabama.gov/faqs/what-is-alabamas-individual-income-tax-rate/
          https://www.revenue.alabama.gov/ultraviewer/viewer/basic_viewer/index.html?form=2023/02/22f40.pdf
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-24.htm?taxmap=true
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-24.htm?taxmap=true
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-24.htm?taxmap=true
          https://www.revenue.alabama.gov/individual-corporate/income-to-be-reported-on-the-alabama-income-tax-return/
          https://www.almonline.org/TaxRates.aspx#OccupationalTax
        
        Arkansas:
        -         
        - https://www.arkansasedc.com/why-arkansas/business-climate/tax-structure/personal-income-tax
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_and_AR1000NR_Instructions.pdf
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_FullYearResidentIndividualIncomeTaxReturn.pdf
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_and_AR1000NR_Instructions.pdf
          https://www.arkansasedc.com/why-arkansas/business-climate/tax-structure/capital-gains-tax-reduction
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_and_AR1000NR_Instructions.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Alabama', 'state_occupation': 'Arkansas', "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def alabama_and_california(self, county_residence='Other', state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Alabama:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        California:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Alabama:
        - https://www.revenue.alabama.gov/ultraviewer/viewer/basic_viewer/index.html?form=2023/02/22f40.pdf
          https://www.revenue.alabama.gov/faqs/what-is-alabamas-individual-income-tax-rate/
          https://www.revenue.alabama.gov/ultraviewer/viewer/basic_viewer/index.html?form=2023/02/22f40.pdf
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-24.htm?taxmap=true
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-24.htm?taxmap=true
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-24.htm?taxmap=true
          https://www.revenue.alabama.gov/individual-corporate/income-to-be-reported-on-the-alabama-income-tax-return/
          https://www.almonline.org/TaxRates.aspx#OccupationalTax
        
        California:
        -         
        - https://www.ftb.ca.gov/forms/2022/2022-540-tax-rate-schedules.pdf
          https://www.ftb.ca.gov/file/personal/deductions/index.html
          https://www.ftb.ca.gov/about-ftb/newsroom/caleitc/eligibility-and-credit-information.html
          https://www.ftb.ca.gov/file/personal/income-types/capital-gains-and-losses.html#:~:text=California%20does%20not%20have%20a
         are%20taxed%20as%20ordinary%20income
          https://www.ftb.ca.gov/file/personal/
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Alabama', 'state_occupation': 'California', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def alabama_and_colorado(self, county_residence='Other', state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Alabama:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Colorado:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Alabama:
        - https://www.revenue.alabama.gov/ultraviewer/viewer/basic_viewer/index.html?form=2023/02/22f40.pdf
          https://www.revenue.alabama.gov/faqs/what-is-alabamas-individual-income-tax-rate/
          https://www.revenue.alabama.gov/ultraviewer/viewer/basic_viewer/index.html?form=2023/02/22f40.pdf
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-24.htm?taxmap=true
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-24.htm?taxmap=true
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-24.htm?taxmap=true
          https://www.revenue.alabama.gov/individual-corporate/income-to-be-reported-on-the-alabama-income-tax-return/
          https://www.almonline.org/TaxRates.aspx#OccupationalTax
        
        Colorado:
        -         
        - https://leg.colorado.gov/agencies/legislative-council-staff/individual-income-tax%C2%A0
          https://leg.colorado.gov/agencies/legislative-council-staff/individual-income-tax%C2%A0
          https://tax.colorado.gov/sites/tax/files/documents/ITT_Capital_Gain_Subtraction_Feb_2022.pdf
          https://www.auroragov.org/business_services/taxes/occupational_privilege_tax
          https://www.denvergov.org/Government/Agencies-Departments-Offices/Agencies-Departments-Offices-Directory/Department-of-Finance/Our-Divisions/Treasury/Business-Tax-Information#:~:text=Occupational%20privilege%20tax
         -%28show%20below%29&text=Each%20taxable%20employee%20is%20liable
         month%20for%20each%20taxable%20employee
          https://www.glendale.co.us/355/Occupational-Privilege-Tax
          https://www.greenwoodvillage.com/1220/Occupational-Privilege-Tax-OPT
          https://ci.sheridan.co.us/288/Occupational-Privilege-Tax
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Alabama', 'state_occupation': 'Colorado', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def alabama_and_connecticut(self, county_residence='Other', state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Alabama:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Connecticut:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Alabama:
        - https://www.revenue.alabama.gov/ultraviewer/viewer/basic_viewer/index.html?form=2023/02/22f40.pdf
          https://www.revenue.alabama.gov/faqs/what-is-alabamas-individual-income-tax-rate/
          https://www.revenue.alabama.gov/ultraviewer/viewer/basic_viewer/index.html?form=2023/02/22f40.pdf
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-24.htm?taxmap=true
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-24.htm?taxmap=true
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-24.htm?taxmap=true
          https://www.revenue.alabama.gov/individual-corporate/income-to-be-reported-on-the-alabama-income-tax-return/
          https://www.almonline.org/TaxRates.aspx#OccupationalTax
        
        Connecticut:
        -         
        - https://portal.ct.gov/-/media/DRS/Forms/2022/Income/2022-CT-1040-Instructions_1222.pdf
          https://www.cga.ct.gov/2022/rpt/pdf/2022-R-0108.pdf
          https://www.cga.ct.gov/current/pub/chap_229.htm#sec_12-703
          https://portal.ct.gov/-/media/DRS/Forms/2022/Income/2022-CT-1040-Instructions_1222.pdf
          "https://portal.ct.gov/DRS/Publications/TSSNs/TSSN-29#:~:text=An%20individuals%20net%20capital%20gains
         portion%20of%20Social%20Security%20benefits"
          "https://portal.ct.gov/DRS/Publications/TSSNs/TSSN-29#:~:text=An%20individuals%20net%20capital%20gains
         portion%20of%20Social%20Security%20benefits"
          https://portal.ct.gov/DRS/Individuals/Resident-Income-Tax/Tax-Information
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Alabama', 'state_occupation': 'Connecticut', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def alabama_and_delaware(self, county_residence='Other', state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, federal_agi=75000, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Alabama:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Delaware:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        federal_agi : int, optional
            Adjusted Gross Income (AGI) for federal taxes. Default is 0. Sources - https://www.irs.gov/e-file-providers/definition-of-adjusted-gross-income, https://www.irs.gov/newsroom/taxpayers-should-know-the-difference-between-standard-and-itemized-deductions
        income : int, optional
            Gross income. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Alabama:
        - https://www.revenue.alabama.gov/ultraviewer/viewer/basic_viewer/index.html?form=2023/02/22f40.pdf
          https://www.revenue.alabama.gov/faqs/what-is-alabamas-individual-income-tax-rate/
          https://www.revenue.alabama.gov/ultraviewer/viewer/basic_viewer/index.html?form=2023/02/22f40.pdf
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-24.htm?taxmap=true
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-24.htm?taxmap=true
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-24.htm?taxmap=true
          https://www.revenue.alabama.gov/individual-corporate/income-to-be-reported-on-the-alabama-income-tax-return/
          https://www.almonline.org/TaxRates.aspx#OccupationalTax
        
        Delaware:
        -         
        - https://revenue.delaware.gov/software-developer/tax-rate-changes/
          https://revenuefiles.delaware.gov/2022/PIT-RES_TY22_2022-01_PaperInteractive.pdf
          https://revenuefiles.delaware.gov/2022/PIT-RES_TY22_2022-02_Instructions.pdf
          https://revenue.delaware.gov/frequently-asked-questions/personal-income-tax-faqs/
          https://www.newcastlede.gov/DocumentCenter/View/20147/City-of-Wilmington-Wage-Tax-Information-PDF
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        federal_agi = self.federal_agi if hasattr(self, 'federal_agi') else federal_agi
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Alabama', 'state_occupation': 'Delaware', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "federal_agi": federal_agi, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def alabama_and_florida(self, county_residence='Other', state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Alabama:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Florida:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Alabama:
        - https://www.revenue.alabama.gov/ultraviewer/viewer/basic_viewer/index.html?form=2023/02/22f40.pdf
          https://www.revenue.alabama.gov/faqs/what-is-alabamas-individual-income-tax-rate/
          https://www.revenue.alabama.gov/ultraviewer/viewer/basic_viewer/index.html?form=2023/02/22f40.pdf
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-24.htm?taxmap=true
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-24.htm?taxmap=true
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-24.htm?taxmap=true
          https://www.revenue.alabama.gov/individual-corporate/income-to-be-reported-on-the-alabama-income-tax-return/
          https://www.almonline.org/TaxRates.aspx#OccupationalTax
        
        Florida:
        -         
        - https://floridarevenue.com/taxes/taxesfees/Pages/tax_interest_rates.aspx
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Alabama', 'state_occupation': 'Florida', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def alabama_and_georgia(self, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Alabama:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Georgia:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Alabama:
        - https://www.revenue.alabama.gov/ultraviewer/viewer/basic_viewer/index.html?form=2023/02/22f40.pdf
          https://www.revenue.alabama.gov/faqs/what-is-alabamas-individual-income-tax-rate/
          https://www.revenue.alabama.gov/ultraviewer/viewer/basic_viewer/index.html?form=2023/02/22f40.pdf
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-24.htm?taxmap=true
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-24.htm?taxmap=true
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-24.htm?taxmap=true
          https://www.revenue.alabama.gov/individual-corporate/income-to-be-reported-on-the-alabama-income-tax-return/
          https://www.almonline.org/TaxRates.aspx#OccupationalTax
        
        Georgia:
        -         
        - https://dor.georgia.gov/tax-tables-georgia-tax-rate-schedule
          https://apps.dor.ga.gov/FillableForms/PDFViewer/Index?form=2022GA500
          https://apps.dor.ga.gov/FillableForms/PDFViewer/Index?form=2022GA500
          https://apps.dor.ga.gov/FillableForms/PDFViewer/Index?form=2022GA500
          https://dor.georgia.gov/years-individual-income-tax-forms
          https://dor.georgia.gov/years-individual-income-tax-forms
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Alabama', 'state_occupation': 'Georgia', "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def alabama_and_hawaii(self, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Alabama:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Hawaii:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Alabama:
        - https://www.revenue.alabama.gov/ultraviewer/viewer/basic_viewer/index.html?form=2023/02/22f40.pdf
          https://www.revenue.alabama.gov/faqs/what-is-alabamas-individual-income-tax-rate/
          https://www.revenue.alabama.gov/ultraviewer/viewer/basic_viewer/index.html?form=2023/02/22f40.pdf
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-24.htm?taxmap=true
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-24.htm?taxmap=true
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-24.htm?taxmap=true
          https://www.revenue.alabama.gov/individual-corporate/income-to-be-reported-on-the-alabama-income-tax-return/
          https://www.almonline.org/TaxRates.aspx#OccupationalTax
        
        Hawaii:
        -         
        - https://tax.hawaii.gov/forms/d_18table-on/d_18table-on_p13/
          https://files.hawaii.gov/tax/forms/2018/18table-on.pdf
          https://files.hawaii.gov/tax/news/pubs/22outline.pdf
          https://files.hawaii.gov/tax/news/pubs/FastTaxRefGuide_2017.pdf
          https://www.tfhawaii.org/wordpress/state-tax-resources/mini-tax-guides/individual-income-tax-chapter-235/#:~:text=Long%20term%20capital%20gains%20are%20taxed%20at%20a%20maximum%20of%207.25%25
          https://www.grassrootinstitute.org/2023/02/hb337-increasing-tax-on-capital-gains-a-bad-risk/
          https://files.hawaii.gov/tax/news/pubs/22outline.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Alabama', 'state_occupation': 'Hawaii', "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def alabama_and_idaho(self, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Alabama:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Idaho:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Alabama:
        - https://www.revenue.alabama.gov/ultraviewer/viewer/basic_viewer/index.html?form=2023/02/22f40.pdf
          https://www.revenue.alabama.gov/faqs/what-is-alabamas-individual-income-tax-rate/
          https://www.revenue.alabama.gov/ultraviewer/viewer/basic_viewer/index.html?form=2023/02/22f40.pdf
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-24.htm?taxmap=true
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-24.htm?taxmap=true
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-24.htm?taxmap=true
          https://www.revenue.alabama.gov/individual-corporate/income-to-be-reported-on-the-alabama-income-tax-return/
          https://www.almonline.org/TaxRates.aspx#OccupationalTax
        
        Idaho:
        -         
        - https://tax.idaho.gov/taxes/income-tax/individual-income/individual-income-tax-rate-schedule/
          https://tax.idaho.gov/wp-content/uploads/forms/EIN00046/EIN00046_03-01-2023.pdf
          https://tax.idaho.gov/taxes/income-tax/individual-income/idaho-source-income/
          https://tax.idaho.gov/taxes/income-tax/individual-income/idaho-source-income/
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Alabama', 'state_occupation': 'Idaho', "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def alabama_and_illinois(self, county_residence='Other', state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Alabama:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Illinois:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Alabama:
        - https://www.revenue.alabama.gov/ultraviewer/viewer/basic_viewer/index.html?form=2023/02/22f40.pdf
          https://www.revenue.alabama.gov/faqs/what-is-alabamas-individual-income-tax-rate/
          https://www.revenue.alabama.gov/ultraviewer/viewer/basic_viewer/index.html?form=2023/02/22f40.pdf
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-24.htm?taxmap=true
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-24.htm?taxmap=true
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-24.htm?taxmap=true
          https://www.revenue.alabama.gov/individual-corporate/income-to-be-reported-on-the-alabama-income-tax-return/
          https://www.almonline.org/TaxRates.aspx#OccupationalTax
        
        Illinois:
        -         
        - https://tax.illinois.gov/research/taxrates/income.html#IndividualIncome
          https://tax.illinois.gov/content/dam/soi/en/web/tax/forms/incometax/documents/currentyear/individual/il-1040-instr.pdf
          https://tax.illinois.gov/questionsandanswers/12.html
          https://tax.illinois.gov/content/dam/soi/en/web/tax/forms/incometax/documents/currentyear/individual/il-1040-instr.pdf
          https://tax.illinois.gov/research/taxrates/personalproperty.html
          https://tax.illinois.gov/localgovernments/income.html
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Alabama', 'state_occupation': 'Illinois', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def alabama_and_indiana(self, capital_gains_short=0, dependents=0, sources=False, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_long=0, filing_status='single', income=100000, credits_state=0):
        """
        --- Parameters ----
        
        Alabama:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Indiana:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Alabama:
        - https://www.revenue.alabama.gov/ultraviewer/viewer/basic_viewer/index.html?form=2023/02/22f40.pdf
          https://www.revenue.alabama.gov/faqs/what-is-alabamas-individual-income-tax-rate/
          https://www.revenue.alabama.gov/ultraviewer/viewer/basic_viewer/index.html?form=2023/02/22f40.pdf
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-24.htm?taxmap=true
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-24.htm?taxmap=true
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-24.htm?taxmap=true
          https://www.revenue.alabama.gov/individual-corporate/income-to-be-reported-on-the-alabama-income-tax-return/
          https://www.almonline.org/TaxRates.aspx#OccupationalTax
        
        Indiana:
        -         
        - https://www.in.gov/dor/tax-forms/2022-individual-income-tax-forms/
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-32.htm
          https://www.in.gov/dor/files/ib28.pdf
          https://www.in.gov/dor/tax-forms/2022-individual-income-tax-forms/
          https://www.in.gov/dor/files/2022-county-tax-rates-and-codes.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Alabama', 'state_occupation': 'Indiana', "capital_gains_short": capital_gains_short, "dependents": dependents, "sources": sources, "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def alabama_and_iowa(self, county_residence='Other', state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Alabama:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Iowa:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Alabama:
        - https://www.revenue.alabama.gov/ultraviewer/viewer/basic_viewer/index.html?form=2023/02/22f40.pdf
          https://www.revenue.alabama.gov/faqs/what-is-alabamas-individual-income-tax-rate/
          https://www.revenue.alabama.gov/ultraviewer/viewer/basic_viewer/index.html?form=2023/02/22f40.pdf
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-24.htm?taxmap=true
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-24.htm?taxmap=true
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-24.htm?taxmap=true
          https://www.revenue.alabama.gov/individual-corporate/income-to-be-reported-on-the-alabama-income-tax-return/
          https://www.almonline.org/TaxRates.aspx#OccupationalTax
        
        Iowa:
        -         
        - https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
          https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
          https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
          https://tax.iowa.gov/iowa-illinois-reciprocal-agreement
          https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Alabama', 'state_occupation': 'Iowa', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def alabama_and_kansas(self, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Alabama:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Kansas:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Alabama:
        - https://www.revenue.alabama.gov/ultraviewer/viewer/basic_viewer/index.html?form=2023/02/22f40.pdf
          https://www.revenue.alabama.gov/faqs/what-is-alabamas-individual-income-tax-rate/
          https://www.revenue.alabama.gov/ultraviewer/viewer/basic_viewer/index.html?form=2023/02/22f40.pdf
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-24.htm?taxmap=true
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-24.htm?taxmap=true
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-24.htm?taxmap=true
          https://www.revenue.alabama.gov/individual-corporate/income-to-be-reported-on-the-alabama-income-tax-return/
          https://www.almonline.org/TaxRates.aspx#OccupationalTax
        
        Kansas:
        -         
        - https://www.ksrevenue.gov/taxrates.html
          https://www.ksrevenue.gov/incomebook22.html#0
          https://www.ksrevenue.gov/incomebook22.html#0
          https://www.ksrevenue.gov/incomebook22.html#0
          https://www.ksrevenue.gov/incomebook22.html
          https://www.ksrevenue.gov/incomebook22.html
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Alabama', 'state_occupation': 'Kansas', "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def alabama_and_kentucky(self, county_residence='Other', state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Alabama:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Kentucky:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Alabama:
        - https://www.revenue.alabama.gov/ultraviewer/viewer/basic_viewer/index.html?form=2023/02/22f40.pdf
          https://www.revenue.alabama.gov/faqs/what-is-alabamas-individual-income-tax-rate/
          https://www.revenue.alabama.gov/ultraviewer/viewer/basic_viewer/index.html?form=2023/02/22f40.pdf
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-24.htm?taxmap=true
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-24.htm?taxmap=true
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-24.htm?taxmap=true
          https://www.revenue.alabama.gov/individual-corporate/income-to-be-reported-on-the-alabama-income-tax-return/
          https://www.almonline.org/TaxRates.aspx#OccupationalTax
        
        Kentucky:
        -         
        - https://revenue.ky.gov/News/Pages/Dor-Announces-Updates-to-Individual-Income-Tax-for-2023-Tax-Year.aspx
          https://revenue.ky.gov/Individual/Individual-Income-Tax/Pages/default.aspx
          https://revenue.ky.gov/News/Pages/Dor-Announces-Updates-to-Individual-Income-Tax-for-2023-Tax-Year.aspx
          https://revenue.ky.gov/Individual/Individual-Income-Tax/Pages/default.aspx
          https://revenue.ky.gov/Dor%20Training%20Materials/103%20KAR%2017.140.%20Individual%20income%20tax%20-%20reciprocity%20-%20nonresidents.pdf
          https://revenue.ky.gov/Individual/Individual-Income-Tax/Pages/default.aspx
          https://nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2021/Taxes-21-32.htm?taxmap=true
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Alabama', 'state_occupation': 'Kentucky', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def alabama_and_louisiana(self, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Alabama:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Louisiana:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Alabama:
        - https://www.revenue.alabama.gov/ultraviewer/viewer/basic_viewer/index.html?form=2023/02/22f40.pdf
          https://www.revenue.alabama.gov/faqs/what-is-alabamas-individual-income-tax-rate/
          https://www.revenue.alabama.gov/ultraviewer/viewer/basic_viewer/index.html?form=2023/02/22f40.pdf
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-24.htm?taxmap=true
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-24.htm?taxmap=true
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-24.htm?taxmap=true
          https://www.revenue.alabama.gov/individual-corporate/income-to-be-reported-on-the-alabama-income-tax-return/
          https://www.almonline.org/TaxRates.aspx#OccupationalTax
        
        Louisiana:
        -         
        - https://revenue.louisiana.gov/individualincometax
          https://revenue.louisiana.gov/individualincometax
          https://www.revenue.louisiana.gov/taxforms/6935(11_02)F.pdf
          https://revenue.louisiana.gov/individualincometax
          https://revenue.louisiana.gov/individualincometax
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Alabama', 'state_occupation': 'Louisiana', "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def alabama_and_maine(self, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Alabama:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Maine:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Alabama:
        - https://www.revenue.alabama.gov/ultraviewer/viewer/basic_viewer/index.html?form=2023/02/22f40.pdf
          https://www.revenue.alabama.gov/faqs/what-is-alabamas-individual-income-tax-rate/
          https://www.revenue.alabama.gov/ultraviewer/viewer/basic_viewer/index.html?form=2023/02/22f40.pdf
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-24.htm?taxmap=true
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-24.htm?taxmap=true
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-24.htm?taxmap=true
          https://www.revenue.alabama.gov/individual-corporate/income-to-be-reported-on-the-alabama-income-tax-return/
          https://www.almonline.org/TaxRates.aspx#OccupationalTax
        
        Maine:
        -         
        - https://www.maine.gov/revenue/tax-return-forms/individual-income-tax-2022
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/ind_tax_rate_sched_2022.pdf
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/22_1040me_book_gen_instruc_revisedFeb23.pdf
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/22_1040me_book_gen_instruc_revisedFeb23.pdf
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/22_1040me_book_gen_instruc_revisedFeb23.pdf
          https://www.maine.gov/revenue/tax-return-forms/individual-income-tax-2022
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Alabama', 'state_occupation': 'Maine', "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def alabama_and_maryland(self, county_residence='Other', state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Alabama:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Maryland:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Alabama:
        - https://www.revenue.alabama.gov/ultraviewer/viewer/basic_viewer/index.html?form=2023/02/22f40.pdf
          https://www.revenue.alabama.gov/faqs/what-is-alabamas-individual-income-tax-rate/
          https://www.revenue.alabama.gov/ultraviewer/viewer/basic_viewer/index.html?form=2023/02/22f40.pdf
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-24.htm?taxmap=true
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-24.htm?taxmap=true
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-24.htm?taxmap=true
          https://www.revenue.alabama.gov/individual-corporate/income-to-be-reported-on-the-alabama-income-tax-return/
          https://www.almonline.org/TaxRates.aspx#OccupationalTax
        
        Maryland:
        -         
        - https://www.marylandtaxes.gov/individual/income/tax-info/tax-rates.php
          https://www.marylandtaxes.gov/forms/22_forms/Resident_Booklet.pdf
          https://www.marylandtaxes.gov/forms/Tax_Publications/Administrative_Releases/Income_and_Estate_Tax_Releases/ar_it37.pdf
          https://www.marylandtaxes.gov/individual/index.php
          https://www.marylandtaxes.gov/individual/income/tax-info/tax-rates.php
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Alabama', 'state_occupation': 'Maryland', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def alabama_and_massachusetts(self, county_residence='Other', state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Alabama:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Massachusetts:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Alabama:
        - https://www.revenue.alabama.gov/ultraviewer/viewer/basic_viewer/index.html?form=2023/02/22f40.pdf
          https://www.revenue.alabama.gov/faqs/what-is-alabamas-individual-income-tax-rate/
          https://www.revenue.alabama.gov/ultraviewer/viewer/basic_viewer/index.html?form=2023/02/22f40.pdf
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-24.htm?taxmap=true
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-24.htm?taxmap=true
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-24.htm?taxmap=true
          https://www.revenue.alabama.gov/individual-corporate/income-to-be-reported-on-the-alabama-income-tax-return/
          https://www.almonline.org/TaxRates.aspx#OccupationalTax
        
        Massachusetts:
        -         
        - https://www.mass.gov/service-details/massachusetts-tax-rates
          https://www.mass.gov/service-details/view-massachusetts-personal-income-tax-exemptions
          https://www.mass.gov/service-details/view-massachusetts-personal-income-tax-exemptions
          https://www.mass.gov/service-details/view-child-and-dependent-related-credits
          https://www.mass.gov/service-details/massachusetts-tax-rates
          https://www.mass.gov/personal-income-tax
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Alabama', 'state_occupation': 'Massachusetts', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def alabama_and_michigan(self, capital_gains_short=0, county_occupation='Other', dependents=0, sources=False, county_residence='Other', state_agi=75000, capital_gains_long=0, filing_status='single', income=100000, credits_state=0):
        """
        --- Parameters ----
        
        Alabama:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Michigan:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        county_occupation : str, optional
            County of occupation. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Alabama:
        - https://www.revenue.alabama.gov/ultraviewer/viewer/basic_viewer/index.html?form=2023/02/22f40.pdf
          https://www.revenue.alabama.gov/faqs/what-is-alabamas-individual-income-tax-rate/
          https://www.revenue.alabama.gov/ultraviewer/viewer/basic_viewer/index.html?form=2023/02/22f40.pdf
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-24.htm?taxmap=true
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-24.htm?taxmap=true
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-24.htm?taxmap=true
          https://www.revenue.alabama.gov/individual-corporate/income-to-be-reported-on-the-alabama-income-tax-return/
          https://www.almonline.org/TaxRates.aspx#OccupationalTax
        
        Michigan:
        -         
        - https://www.michigan.gov/taxes/iit/new-developments-for-tax-year-2022
          "https://www.michigan.gov/taxes/-/media/Project/Websites/taxes/Forms/2023/2023-SUW/446_Withholding-Guide_2023.pdf?rev=177ca3d8cb034392bf84e317ed78c320&hash=19C5F42F43A89A91863D229349F48883#:~:text=The%20exemption%20amount%20is%20%245%2C400
         employees%20Michigan%20income%20tax%20return"
          "https://www.michigan.gov/taxes/-/media/Project/Websites/taxes/Forms/2023/2023-SUW/446_Withholding-Guide_2023.pdf?rev=177ca3d8cb034392bf84e317ed78c320&hash=19C5F42F43A89A91863D229349F48883#:~:text=The%20exemption%20amount%20is%20%245%2C400
         employees%20Michigan%20income%20tax%20return"
          https://www.michigan.gov/taxes/questions/iit/accordion/residency/are-my-wages-earned-in-another-state-taxable-in-michigan-if-i-am-a-michigan-resident-1
          https://www.michigan.gov/taxes/-/media/Project/Websites/taxes/Forms/2022/2022-IIT-Forms/BOOK_MI-1040.pdf?rev=4524f1cbbc0243b194ab05a6680be75b&hash=40DFE71E651F7C1D65C31F6CD03D944F
          https://www.michigan.gov/taxes/questions/iit/accordion/general/what-cities-impose-an-income-tax
        """
        
        # Check if the parameters are set upstream in the instance
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        county_occupation = self.county_occupation if hasattr(self, 'county_occupation') else county_occupation
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Alabama', 'state_occupation': 'Michigan', "capital_gains_short": capital_gains_short, "county_occupation": county_occupation, "dependents": dependents, "sources": sources, "county_residence": county_residence, "state_agi": state_agi, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def alabama_and_minnesota(self, county_residence='Other', state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Alabama:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Minnesota:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Alabama:
        - https://www.revenue.alabama.gov/ultraviewer/viewer/basic_viewer/index.html?form=2023/02/22f40.pdf
          https://www.revenue.alabama.gov/faqs/what-is-alabamas-individual-income-tax-rate/
          https://www.revenue.alabama.gov/ultraviewer/viewer/basic_viewer/index.html?form=2023/02/22f40.pdf
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-24.htm?taxmap=true
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-24.htm?taxmap=true
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-24.htm?taxmap=true
          https://www.revenue.alabama.gov/individual-corporate/income-to-be-reported-on-the-alabama-income-tax-return/
          https://www.almonline.org/TaxRates.aspx#OccupationalTax
        
        Minnesota:
        -         
        - https://www.revenue.state.mn.us/minnesota-income-tax-rates-and-brackets
          https://www.revenue.state.mn.us/child-and-dependent-care-credit
          https://www.revenue.state.mn.us/dependent-exemptions
          https://www.revenue.state.mn.us/minnesota-standard-deduction
          https://mn.gov/mmb/assets/mwr_form_tcm1059-128581.pdf
          https://www.revenue.state.mn.us/individual-income-tax
          https://www.revenue.state.mn.us/individual-income-tax
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Alabama', 'state_occupation': 'Minnesota', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def alabama_and_mississippi(self, county_residence='Other', state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Alabama:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Mississippi:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Alabama:
        - https://www.revenue.alabama.gov/ultraviewer/viewer/basic_viewer/index.html?form=2023/02/22f40.pdf
          https://www.revenue.alabama.gov/faqs/what-is-alabamas-individual-income-tax-rate/
          https://www.revenue.alabama.gov/ultraviewer/viewer/basic_viewer/index.html?form=2023/02/22f40.pdf
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-24.htm?taxmap=true
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-24.htm?taxmap=true
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-24.htm?taxmap=true
          https://www.revenue.alabama.gov/individual-corporate/income-to-be-reported-on-the-alabama-income-tax-return/
          https://www.almonline.org/TaxRates.aspx#OccupationalTax
        
        Mississippi:
        -         
        - https://www.dor.ms.gov/individual/tax-rates
          https://www.dor.ms.gov/individual/tax-rates
          https://www.dor.ms.gov/individual/tax-rates
          https://www.dor.ms.gov/individual
          https://www.dor.ms.gov/individual
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Alabama', 'state_occupation': 'Mississippi', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def alabama_and_missouri(self, capital_gains_short=0, dependents=0, sources=False, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_long=0, filing_status='single', income=100000, credits_state=0):
        """
        --- Parameters ----
        
        Alabama:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Missouri:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Alabama:
        - https://www.revenue.alabama.gov/ultraviewer/viewer/basic_viewer/index.html?form=2023/02/22f40.pdf
          https://www.revenue.alabama.gov/faqs/what-is-alabamas-individual-income-tax-rate/
          https://www.revenue.alabama.gov/ultraviewer/viewer/basic_viewer/index.html?form=2023/02/22f40.pdf
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-24.htm?taxmap=true
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-24.htm?taxmap=true
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-24.htm?taxmap=true
          https://www.revenue.alabama.gov/individual-corporate/income-to-be-reported-on-the-alabama-income-tax-return/
          https://www.almonline.org/TaxRates.aspx#OccupationalTax
        
        Missouri:
        -         
        - https://dor.mo.gov/taxation/individual/tax-types/income/year-changes/
          https://dor.mo.gov/taxation/individual/tax-types/income/year-changes/
          https://dor.mo.gov/taxation/individual/tax-types/income/year-changes/
          https://www.kcmo.gov/city-hall/departments/finance/earnings-tax
          https://www.stlouis-mo.gov/government/departments/collector/earnings-tax/individual-earnings-tax-info.cfm
        """
        
        # Check if the parameters are set upstream in the instance
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Alabama', 'state_occupation': 'Missouri', "capital_gains_short": capital_gains_short, "dependents": dependents, "sources": sources, "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def alabama_and_montana(self, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Alabama:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Montana:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Alabama:
        - https://www.revenue.alabama.gov/ultraviewer/viewer/basic_viewer/index.html?form=2023/02/22f40.pdf
          https://www.revenue.alabama.gov/faqs/what-is-alabamas-individual-income-tax-rate/
          https://www.revenue.alabama.gov/ultraviewer/viewer/basic_viewer/index.html?form=2023/02/22f40.pdf
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-24.htm?taxmap=true
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-24.htm?taxmap=true
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-24.htm?taxmap=true
          https://www.revenue.alabama.gov/individual-corporate/income-to-be-reported-on-the-alabama-income-tax-return/
          https://www.almonline.org/TaxRates.aspx#OccupationalTax
        
        Montana:
        -         
        - https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/2022/12/montana-employees-withholding-allowance-and-exemption-certificate-form-mw-4.pdf
          https://mtrevenue.gov/2021/11/29/simplification-of-montana-income-taxation/?utm_source=rss&utm_medium=rss&utm_campaign=simplification-of-montana-income-taxation
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Alabama', 'state_occupation': 'Montana', "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def alabama_and_nebraska(self, county_residence='Other', state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Alabama:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Nebraska:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Alabama:
        - https://www.revenue.alabama.gov/ultraviewer/viewer/basic_viewer/index.html?form=2023/02/22f40.pdf
          https://www.revenue.alabama.gov/faqs/what-is-alabamas-individual-income-tax-rate/
          https://www.revenue.alabama.gov/ultraviewer/viewer/basic_viewer/index.html?form=2023/02/22f40.pdf
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-24.htm?taxmap=true
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-24.htm?taxmap=true
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-24.htm?taxmap=true
          https://www.revenue.alabama.gov/individual-corporate/income-to-be-reported-on-the-alabama-income-tax-return/
          https://www.almonline.org/TaxRates.aspx#OccupationalTax
        
        Nebraska:
        -         
        - https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2023_Circular_EN_8-429r2023_Whole_Book_11.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Alabama', 'state_occupation': 'Nebraska', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def alabama_and_nevada(self, county_residence='Other', state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Alabama:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Nevada:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Alabama:
        - https://www.revenue.alabama.gov/ultraviewer/viewer/basic_viewer/index.html?form=2023/02/22f40.pdf
          https://www.revenue.alabama.gov/faqs/what-is-alabamas-individual-income-tax-rate/
          https://www.revenue.alabama.gov/ultraviewer/viewer/basic_viewer/index.html?form=2023/02/22f40.pdf
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-24.htm?taxmap=true
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-24.htm?taxmap=true
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-24.htm?taxmap=true
          https://www.revenue.alabama.gov/individual-corporate/income-to-be-reported-on-the-alabama-income-tax-return/
          https://www.almonline.org/TaxRates.aspx#OccupationalTax
        
        Nevada:
        -         
        - https://www.nvsos.gov/sos/businesses/the-nevada-advantage
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Alabama', 'state_occupation': 'Nevada', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def alabama_and_new_hampshire(self, county_residence='Other', state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Alabama:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        New Hampshire:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Alabama:
        - https://www.revenue.alabama.gov/ultraviewer/viewer/basic_viewer/index.html?form=2023/02/22f40.pdf
          https://www.revenue.alabama.gov/faqs/what-is-alabamas-individual-income-tax-rate/
          https://www.revenue.alabama.gov/ultraviewer/viewer/basic_viewer/index.html?form=2023/02/22f40.pdf
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-24.htm?taxmap=true
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-24.htm?taxmap=true
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-24.htm?taxmap=true
          https://www.revenue.alabama.gov/individual-corporate/income-to-be-reported-on-the-alabama-income-tax-return/
          https://www.almonline.org/TaxRates.aspx#OccupationalTax
        
        New Hampshire:
        -         
        - https://www.revenue.nh.gov/assistance/tax-overview.htm
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Alabama', 'state_occupation': 'NewHampshire', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def alabama_and_new_jersey(self, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Alabama:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        New Jersey:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Alabama:
        - https://www.revenue.alabama.gov/ultraviewer/viewer/basic_viewer/index.html?form=2023/02/22f40.pdf
          https://www.revenue.alabama.gov/faqs/what-is-alabamas-individual-income-tax-rate/
          https://www.revenue.alabama.gov/ultraviewer/viewer/basic_viewer/index.html?form=2023/02/22f40.pdf
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-24.htm?taxmap=true
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-24.htm?taxmap=true
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-24.htm?taxmap=true
          https://www.revenue.alabama.gov/individual-corporate/income-to-be-reported-on-the-alabama-income-tax-return/
          https://www.almonline.org/TaxRates.aspx#OccupationalTax
        
        New Jersey:
        -         
        - https://www.state.nj.us/treasury/taxation/taxtables.shtml
          https://www.state.nj.us/treasury/taxation/njit13.shtml
          https://www.nj.gov/treasury/taxation/njit25.shtml
          https://www.state.nj.us/treasury/taxation/individuals/freqqiti.shtml
          https://www.state.nj.us/treasury/taxation/localtax.shtml
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Alabama', 'state_occupation': 'NewJersey', "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def alabama_and_new_mexico(self, county_residence='Other', state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Alabama:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        New Mexico:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Alabama:
        - https://www.revenue.alabama.gov/ultraviewer/viewer/basic_viewer/index.html?form=2023/02/22f40.pdf
          https://www.revenue.alabama.gov/faqs/what-is-alabamas-individual-income-tax-rate/
          https://www.revenue.alabama.gov/ultraviewer/viewer/basic_viewer/index.html?form=2023/02/22f40.pdf
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-24.htm?taxmap=true
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-24.htm?taxmap=true
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-24.htm?taxmap=true
          https://www.revenue.alabama.gov/individual-corporate/income-to-be-reported-on-the-alabama-income-tax-return/
          https://www.almonline.org/TaxRates.aspx#OccupationalTax
        
        New Mexico:
        -         
        - https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/fdf3c548-8aba-4b9c-9eb4-bb564c716015/FYI-104.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/0bf943fd-652e-4400-bb1b-cd8397eb5a95/2022pit-adj.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/2f1a6781-9534-4436-b427-1557f9592099/2022pit-adj-ins.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/0bf943fd-652e-4400-bb1b-cd8397eb5a95/2022pit-adj.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/1afc56af-ea90-4d48-82e5-1f9aeb43255a/PITbook2022.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/fdf3c548-8aba-4b9c-9eb4-bb564c716015/FYI-104.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Alabama', 'state_occupation': 'NewMexico', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def alabama_and_new_york(self, county_residence='Other', state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Alabama:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        New York:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Alabama:
        - https://www.revenue.alabama.gov/ultraviewer/viewer/basic_viewer/index.html?form=2023/02/22f40.pdf
          https://www.revenue.alabama.gov/faqs/what-is-alabamas-individual-income-tax-rate/
          https://www.revenue.alabama.gov/ultraviewer/viewer/basic_viewer/index.html?form=2023/02/22f40.pdf
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-24.htm?taxmap=true
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-24.htm?taxmap=true
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-24.htm?taxmap=true
          https://www.revenue.alabama.gov/individual-corporate/income-to-be-reported-on-the-alabama-income-tax-return/
          https://www.almonline.org/TaxRates.aspx#OccupationalTax
        
        New York:
        -         
        - https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nys-tax-rate-schedule
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nys-tax-rate-schedule
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nys-tax-rate-schedule
          https://www.tax.ny.gov/pit/
          https://www.tax.ny.gov/pit/file/nonresident-faqs.htm
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#yonkers-tax-rate-schedule
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nyc-tax-rate-schedule
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Alabama', 'state_occupation': 'NewYork', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def alabama_and_north_carolina(self, county_residence='Other', state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Alabama:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        North Carolina:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Alabama:
        - https://www.revenue.alabama.gov/ultraviewer/viewer/basic_viewer/index.html?form=2023/02/22f40.pdf
          https://www.revenue.alabama.gov/faqs/what-is-alabamas-individual-income-tax-rate/
          https://www.revenue.alabama.gov/ultraviewer/viewer/basic_viewer/index.html?form=2023/02/22f40.pdf
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-24.htm?taxmap=true
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-24.htm?taxmap=true
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-24.htm?taxmap=true
          https://www.revenue.alabama.gov/individual-corporate/income-to-be-reported-on-the-alabama-income-tax-return/
          https://www.almonline.org/TaxRates.aspx#OccupationalTax
        
        North Carolina:
        -         
        - https://www.ncdor.gov/taxes-forms/tax-rate-schedules
          https://www.ncdor.gov/taxes-forms/individual-income-tax/north-carolina-standard-deduction-or-north-carolina-itemized-deductions
          https://www.ncdor.gov/taxes-forms/individual-income-tax
          https://www.ncdor.gov/taxes-forms/individual-income-tax
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Alabama', 'state_occupation': 'NorthCarolina', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def alabama_and_north_dakota(self, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Alabama:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        North Dakota:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Alabama:
        - https://www.revenue.alabama.gov/ultraviewer/viewer/basic_viewer/index.html?form=2023/02/22f40.pdf
          https://www.revenue.alabama.gov/faqs/what-is-alabamas-individual-income-tax-rate/
          https://www.revenue.alabama.gov/ultraviewer/viewer/basic_viewer/index.html?form=2023/02/22f40.pdf
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-24.htm?taxmap=true
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-24.htm?taxmap=true
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-24.htm?taxmap=true
          https://www.revenue.alabama.gov/individual-corporate/income-to-be-reported-on-the-alabama-income-tax-return/
          https://www.almonline.org/TaxRates.aspx#OccupationalTax
        
        North Dakota:
        -         
        - https://www.tax.nd.gov/forms
          https://www.tax.nd.gov/sites/www/files/documents/forms/individual/2022-iit/2022-individual-income-tax-booklet.pdf
          https://www.tax.nd.gov/sites/www/files/documents/forms/2022-individual-income-tax-booklet.pdf
          https://www.tax.nd.gov/tax-exemptions-credits/income-tax-exemptions-credits
          https://www.tax.nd.gov/sites/www/files/documents/forms/business/ndwrfillable.pdf
          https://www.tax.nd.gov/news/tax-legislative-changes/significant-changes-law/individual-income-tax-history
          https://www.tax.nd.gov/forms
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Alabama', 'state_occupation': 'NorthDakota', "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def alabama_and_ohio(self, county_residence='Other', state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Alabama:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Ohio:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Alabama:
        - https://www.revenue.alabama.gov/ultraviewer/viewer/basic_viewer/index.html?form=2023/02/22f40.pdf
          https://www.revenue.alabama.gov/faqs/what-is-alabamas-individual-income-tax-rate/
          https://www.revenue.alabama.gov/ultraviewer/viewer/basic_viewer/index.html?form=2023/02/22f40.pdf
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-24.htm?taxmap=true
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-24.htm?taxmap=true
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-24.htm?taxmap=true
          https://www.revenue.alabama.gov/individual-corporate/income-to-be-reported-on-the-alabama-income-tax-return/
          https://www.almonline.org/TaxRates.aspx#OccupationalTax
        
        Ohio:
        -         
        - https://tax.ohio.gov/individual/resources/annual-tax-rates
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://tax.ohio.gov/static/forms/employer_withholding/generic/wth_it4nr.pdf
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://www.ritaohio.com/TaxRatesTable
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Alabama', 'state_occupation': 'Ohio', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def alabama_and_oklahoma(self, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Alabama:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Oklahoma:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Alabama:
        - https://www.revenue.alabama.gov/ultraviewer/viewer/basic_viewer/index.html?form=2023/02/22f40.pdf
          https://www.revenue.alabama.gov/faqs/what-is-alabamas-individual-income-tax-rate/
          https://www.revenue.alabama.gov/ultraviewer/viewer/basic_viewer/index.html?form=2023/02/22f40.pdf
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-24.htm?taxmap=true
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-24.htm?taxmap=true
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-24.htm?taxmap=true
          https://www.revenue.alabama.gov/individual-corporate/income-to-be-reported-on-the-alabama-income-tax-return/
          https://www.almonline.org/TaxRates.aspx#OccupationalTax
        
        Oklahoma:
        -         
        - https://oklahoma.gov/content/dam/ok/en/tax/documents/resources/publications/businesses/withholding-tables/WHTables-2022.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Alabama', 'state_occupation': 'Oklahoma', "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def alabama_and_oregon(self, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Alabama:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Oregon:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Alabama:
        - https://www.revenue.alabama.gov/ultraviewer/viewer/basic_viewer/index.html?form=2023/02/22f40.pdf
          https://www.revenue.alabama.gov/faqs/what-is-alabamas-individual-income-tax-rate/
          https://www.revenue.alabama.gov/ultraviewer/viewer/basic_viewer/index.html?form=2023/02/22f40.pdf
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-24.htm?taxmap=true
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-24.htm?taxmap=true
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-24.htm?taxmap=true
          https://www.revenue.alabama.gov/individual-corporate/income-to-be-reported-on-the-alabama-income-tax-return/
          https://www.almonline.org/TaxRates.aspx#OccupationalTax
        
        Oregon:
        -         
        - https://www.oregon.gov/dor/programs/individuals/Pages/PIT.aspx
          https://www.oregon.gov/dor/programs/individuals/Documents/Full%20year%20resident
         %20Form%20or-40%20filers.pdf
          https://www.oregon.gov/dor/programs/individuals/Documents/Part-year%20and%20nonresident
         %20Form%20or-40-P%20and%20or-40-N%20filers.pdf.pdf
          https://www.oregon.gov/dor/forms/FormsPubs/form-or-40_101-040_2022.pdf
          https://www.oregon.gov/dor/forms/FormsPubs/form-or-40_101-040_2022.pdf
          https://www.oregon.gov/dor/forms/FormsPubs/form-or-40_101-040_2022.pdf
          https://www.portland.gov/revenue/personal-tax
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Alabama', 'state_occupation': 'Oregon', "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def alabama_and_pennsylvania(self, county_residence='Other', state_agi=75000, county_occupation='Other', capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Alabama:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Pennsylvania:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        county_residence : str, optional
            County of residence. Default is Other.
        county_occupation : str, optional
            County of occupation. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Alabama:
        - https://www.revenue.alabama.gov/ultraviewer/viewer/basic_viewer/index.html?form=2023/02/22f40.pdf
          https://www.revenue.alabama.gov/faqs/what-is-alabamas-individual-income-tax-rate/
          https://www.revenue.alabama.gov/ultraviewer/viewer/basic_viewer/index.html?form=2023/02/22f40.pdf
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-24.htm?taxmap=true
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-24.htm?taxmap=true
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-24.htm?taxmap=true
          https://www.revenue.alabama.gov/individual-corporate/income-to-be-reported-on-the-alabama-income-tax-return/
          https://www.almonline.org/TaxRates.aspx#OccupationalTax
        
        Pennsylvania:
        -         
        - https://www.revenue.pa.gov/Tax%20Rates/Pages/default.aspx
          https://www.revenue.pa.gov/FormsandPublications/FormsforIndividuals/PIT/Documents/2022/2022_pa-40in.pdf
          https://www.revenue.pa.gov/FormsandPublications/FormsforBusinesses/EmployerWithholding/Documents/rev-419.pdf
          https://www.revenue.pa.gov/FormsandPublications/FormsforIndividuals/PIT/Documents/2022/2022_pa-40in.pdf
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-25.htm
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        county_occupation = self.county_occupation if hasattr(self, 'county_occupation') else county_occupation
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Alabama', 'state_occupation': 'Pennsylvania', "county_residence": county_residence, "state_agi": state_agi, "county_occupation": county_occupation, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def alabama_and_rhode_island(self, county_residence='Other', state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Alabama:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Rhode Island:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Alabama:
        - https://www.revenue.alabama.gov/ultraviewer/viewer/basic_viewer/index.html?form=2023/02/22f40.pdf
          https://www.revenue.alabama.gov/faqs/what-is-alabamas-individual-income-tax-rate/
          https://www.revenue.alabama.gov/ultraviewer/viewer/basic_viewer/index.html?form=2023/02/22f40.pdf
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-24.htm?taxmap=true
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-24.htm?taxmap=true
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-24.htm?taxmap=true
          https://www.revenue.alabama.gov/individual-corporate/income-to-be-reported-on-the-alabama-income-tax-return/
          https://www.almonline.org/TaxRates.aspx#OccupationalTax
        
        Rhode Island:
        -         
        - https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2022_1040WE_w_0.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Alabama', 'state_occupation': 'RhodeIsland', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def alabama_and_south_carolina(self, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Alabama:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        South Carolina:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Alabama:
        - https://www.revenue.alabama.gov/ultraviewer/viewer/basic_viewer/index.html?form=2023/02/22f40.pdf
          https://www.revenue.alabama.gov/faqs/what-is-alabamas-individual-income-tax-rate/
          https://www.revenue.alabama.gov/ultraviewer/viewer/basic_viewer/index.html?form=2023/02/22f40.pdf
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-24.htm?taxmap=true
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-24.htm?taxmap=true
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-24.htm?taxmap=true
          https://www.revenue.alabama.gov/individual-corporate/income-to-be-reported-on-the-alabama-income-tax-return/
          https://www.almonline.org/TaxRates.aspx#OccupationalTax
        
        South Carolina:
        -         
        - https://dor.sc.gov/resources-site/lawandpolicy/Advisory%20Opinions/IL22-15.pdf
          https://dor.sc.gov/forms-site/Forms/IITPacket_2022.pdf
          https://www.scstatehouse.gov/code/t12c006.php
          https://dor.sc.gov/resources-site/lawandpolicy/Advisory%20Opinions/IL22-15.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Alabama', 'state_occupation': 'SouthCarolina', "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def alabama_and_south_dakota(self, county_residence='Other', state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Alabama:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        South Dakota:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Alabama:
        - https://www.revenue.alabama.gov/ultraviewer/viewer/basic_viewer/index.html?form=2023/02/22f40.pdf
          https://www.revenue.alabama.gov/faqs/what-is-alabamas-individual-income-tax-rate/
          https://www.revenue.alabama.gov/ultraviewer/viewer/basic_viewer/index.html?form=2023/02/22f40.pdf
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-24.htm?taxmap=true
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-24.htm?taxmap=true
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-24.htm?taxmap=true
          https://www.revenue.alabama.gov/individual-corporate/income-to-be-reported-on-the-alabama-income-tax-return/
          https://www.almonline.org/TaxRates.aspx#OccupationalTax
        
        South Dakota:
        -         
        - https://dor.sd.gov/individuals/taxes/
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Alabama', 'state_occupation': 'SouthDakota', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def alabama_and_tennessee(self, county_residence='Other', state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Alabama:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Tennessee:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Alabama:
        - https://www.revenue.alabama.gov/ultraviewer/viewer/basic_viewer/index.html?form=2023/02/22f40.pdf
          https://www.revenue.alabama.gov/faqs/what-is-alabamas-individual-income-tax-rate/
          https://www.revenue.alabama.gov/ultraviewer/viewer/basic_viewer/index.html?form=2023/02/22f40.pdf
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-24.htm?taxmap=true
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-24.htm?taxmap=true
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-24.htm?taxmap=true
          https://www.revenue.alabama.gov/individual-corporate/income-to-be-reported-on-the-alabama-income-tax-return/
          https://www.almonline.org/TaxRates.aspx#OccupationalTax
        
        Tennessee:
        -         
        - https://www.tn.gov/revenue/taxes.html
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Alabama', 'state_occupation': 'Tennessee', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def alabama_and_texas(self, county_residence='Other', state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Alabama:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Texas:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Alabama:
        - https://www.revenue.alabama.gov/ultraviewer/viewer/basic_viewer/index.html?form=2023/02/22f40.pdf
          https://www.revenue.alabama.gov/faqs/what-is-alabamas-individual-income-tax-rate/
          https://www.revenue.alabama.gov/ultraviewer/viewer/basic_viewer/index.html?form=2023/02/22f40.pdf
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-24.htm?taxmap=true
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-24.htm?taxmap=true
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-24.htm?taxmap=true
          https://www.revenue.alabama.gov/individual-corporate/income-to-be-reported-on-the-alabama-income-tax-return/
          https://www.almonline.org/TaxRates.aspx#OccupationalTax
        
        Texas:
        -         
        - https://comptroller.texas.gov/taxes/
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Alabama', 'state_occupation': 'Texas', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def alabama_and_utah(self, county_residence='Other', state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Alabama:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Utah:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Alabama:
        - https://www.revenue.alabama.gov/ultraviewer/viewer/basic_viewer/index.html?form=2023/02/22f40.pdf
          https://www.revenue.alabama.gov/faqs/what-is-alabamas-individual-income-tax-rate/
          https://www.revenue.alabama.gov/ultraviewer/viewer/basic_viewer/index.html?form=2023/02/22f40.pdf
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-24.htm?taxmap=true
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-24.htm?taxmap=true
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-24.htm?taxmap=true
          https://www.revenue.alabama.gov/individual-corporate/income-to-be-reported-on-the-alabama-income-tax-return/
          https://www.almonline.org/TaxRates.aspx#OccupationalTax
        
        Utah:
        -         
        - https://incometax.utah.gov/paying/tax-rates
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Alabama', 'state_occupation': 'Utah', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def alabama_and_vermont(self, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Alabama:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Vermont:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Alabama:
        - https://www.revenue.alabama.gov/ultraviewer/viewer/basic_viewer/index.html?form=2023/02/22f40.pdf
          https://www.revenue.alabama.gov/faqs/what-is-alabamas-individual-income-tax-rate/
          https://www.revenue.alabama.gov/ultraviewer/viewer/basic_viewer/index.html?form=2023/02/22f40.pdf
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-24.htm?taxmap=true
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-24.htm?taxmap=true
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-24.htm?taxmap=true
          https://www.revenue.alabama.gov/individual-corporate/income-to-be-reported-on-the-alabama-income-tax-return/
          https://www.almonline.org/TaxRates.aspx#OccupationalTax
        
        Vermont:
        -         
        - https://tax.vermont.gov/sites/tax/files/documents/GB-1210-2023.pdf
          https://tax.vermont.gov/sites/tax/files/documents/Income%20Booklet-2022.pdf
          https://tax.vermont.gov/sites/tax/files/documents/Income%20Booklet-2022.pdf
          https://tax.vermont.gov/sites/tax/files/documents/Income%20Booklet-2022.pdf
          https://tax.vermont.gov/individuals/personal-income-tax/tax-credits
          https://tax.vermont.gov/individuals/personal-income-tax/taxable-income
          https://tax.vermont.gov/individuals/personal-income-tax/local-option-tax
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Alabama', 'state_occupation': 'Vermont', "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def alabama_and_virginia(self, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Alabama:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Virginia:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Alabama:
        - https://www.revenue.alabama.gov/ultraviewer/viewer/basic_viewer/index.html?form=2023/02/22f40.pdf
          https://www.revenue.alabama.gov/faqs/what-is-alabamas-individual-income-tax-rate/
          https://www.revenue.alabama.gov/ultraviewer/viewer/basic_viewer/index.html?form=2023/02/22f40.pdf
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-24.htm?taxmap=true
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-24.htm?taxmap=true
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-24.htm?taxmap=true
          https://www.revenue.alabama.gov/individual-corporate/income-to-be-reported-on-the-alabama-income-tax-return/
          https://www.almonline.org/TaxRates.aspx#OccupationalTax
        
        Virginia:
        -         
        - https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/reciprocity
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Alabama', 'state_occupation': 'Virginia', "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def alabama_and_washington(self, county_residence='Other', state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Alabama:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Washington:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Alabama:
        - https://www.revenue.alabama.gov/ultraviewer/viewer/basic_viewer/index.html?form=2023/02/22f40.pdf
          https://www.revenue.alabama.gov/faqs/what-is-alabamas-individual-income-tax-rate/
          https://www.revenue.alabama.gov/ultraviewer/viewer/basic_viewer/index.html?form=2023/02/22f40.pdf
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-24.htm?taxmap=true
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-24.htm?taxmap=true
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-24.htm?taxmap=true
          https://www.revenue.alabama.gov/individual-corporate/income-to-be-reported-on-the-alabama-income-tax-return/
          https://www.almonline.org/TaxRates.aspx#OccupationalTax
        
        Washington:
        -         
        - https://dor.wa.gov/taxes-rates/income-tax
          https://dor.wa.gov/taxes-rates/other-taxes/capital-gains-tax
          https://dor.wa.gov/taxes-rates/income-tax
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Alabama', 'state_occupation': 'Washington', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def alabama_and_west_virginia(self, pay_periods=24, capital_gains_short=0, county_occupation='Other', dependents=0, sources=False, county_residence='Other', state_agi=75000, capital_gains_long=0, filing_status='single', income=100000, credits_state=0):
        """
        --- Parameters ----
        
        Alabama:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        West Virginia:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_occupation : str, optional
            County of occupation. Default is Other.
        pay_periods : int, optional
            Number of pay periods. Options are 13, 24, 26, 52. Default is 24.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Alabama:
        - https://www.revenue.alabama.gov/ultraviewer/viewer/basic_viewer/index.html?form=2023/02/22f40.pdf
          https://www.revenue.alabama.gov/faqs/what-is-alabamas-individual-income-tax-rate/
          https://www.revenue.alabama.gov/ultraviewer/viewer/basic_viewer/index.html?form=2023/02/22f40.pdf
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-24.htm?taxmap=true
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-24.htm?taxmap=true
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-24.htm?taxmap=true
          https://www.revenue.alabama.gov/individual-corporate/income-to-be-reported-on-the-alabama-income-tax-return/
          https://www.almonline.org/TaxRates.aspx#OccupationalTax
        
        West Virginia:
        -         
        - https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://tax.wv.gov/Documents/TSD/tsd438.pdf
          https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2011/Taxes-11-35att.html
        """
        
        # Check if the parameters are set upstream in the instance
        pay_periods = self.pay_periods if hasattr(self, 'pay_periods') else pay_periods
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        county_occupation = self.county_occupation if hasattr(self, 'county_occupation') else county_occupation
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Alabama', 'state_occupation': 'WestVirginia', "pay_periods": pay_periods, "capital_gains_short": capital_gains_short, "county_occupation": county_occupation, "dependents": dependents, "sources": sources, "county_residence": county_residence, "state_agi": state_agi, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def alabama_and_wisconsin(self, county_residence='Other', state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Alabama:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Wisconsin:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Alabama:
        - https://www.revenue.alabama.gov/ultraviewer/viewer/basic_viewer/index.html?form=2023/02/22f40.pdf
          https://www.revenue.alabama.gov/faqs/what-is-alabamas-individual-income-tax-rate/
          https://www.revenue.alabama.gov/ultraviewer/viewer/basic_viewer/index.html?form=2023/02/22f40.pdf
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-24.htm?taxmap=true
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-24.htm?taxmap=true
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-24.htm?taxmap=true
          https://www.revenue.alabama.gov/individual-corporate/income-to-be-reported-on-the-alabama-income-tax-return/
          https://www.almonline.org/TaxRates.aspx#OccupationalTax
        
        Wisconsin:
        -         
        - https://www.revenue.wi.gov/TaxForms2023/2023-Form1-ES-Inst.pdf
          https://www.revenue.wi.gov/TaxForms2023/2023-Form1-ES-Inst.pdf
          https://www.revenue.wi.gov/TaxForms2023/2023-Form1-ES-Inst.pdf
          https://www.revenue.wi.gov/Pages/FAQS/pcs-work.aspx
          https://www.revenue.wi.gov/Dor%20Publications/pb103.pdf
          https://www.revenue.wi.gov/Pages/FAQS/pcs-work.aspx
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Alabama', 'state_occupation': 'Wisconsin', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def alabama_and_wyoming(self, county_residence='Other', state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Alabama:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Wyoming:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Alabama:
        - https://www.revenue.alabama.gov/ultraviewer/viewer/basic_viewer/index.html?form=2023/02/22f40.pdf
          https://www.revenue.alabama.gov/faqs/what-is-alabamas-individual-income-tax-rate/
          https://www.revenue.alabama.gov/ultraviewer/viewer/basic_viewer/index.html?form=2023/02/22f40.pdf
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-24.htm?taxmap=true
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-24.htm?taxmap=true
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-24.htm?taxmap=true
          https://www.revenue.alabama.gov/individual-corporate/income-to-be-reported-on-the-alabama-income-tax-return/
          https://www.almonline.org/TaxRates.aspx#OccupationalTax
        
        Wyoming:
        -         
        - https://revenue.wyo.gov/
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Alabama', 'state_occupation': 'Wyoming', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def alaska_and_alabama(self, county_residence='Other', state_agi=75000, capital_gains_short=0, capital_gains_long=0, filing_status='single', income=100000, dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Alaska:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        Alabama:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Alaska:
        - https://www.commerce.alaska.gov/web/dcra/officeofthestateassessor/alaskataxfacts.aspx
        
        Alabama:
        -         
        - https://www.revenue.alabama.gov/ultraviewer/viewer/basic_viewer/index.html?form=2023/02/22f40.pdf
          https://www.revenue.alabama.gov/faqs/what-is-alabamas-individual-income-tax-rate/
          https://www.revenue.alabama.gov/ultraviewer/viewer/basic_viewer/index.html?form=2023/02/22f40.pdf
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-24.htm?taxmap=true
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-24.htm?taxmap=true
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-24.htm?taxmap=true
          https://www.revenue.alabama.gov/individual-corporate/income-to-be-reported-on-the-alabama-income-tax-return/
          https://www.almonline.org/TaxRates.aspx#OccupationalTax
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Alaska', 'state_occupation': 'Alabama', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def alaska_and_arizona(self, state_agi=75000, capital_gains_short=0, capital_gains_long=0, filing_status='single', income=100000, dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Alaska:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        Arizona:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        credits_state : int, optional
            Credits for state taxes. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Alaska:
        - https://www.commerce.alaska.gov/web/dcra/officeofthestateassessor/alaskataxfacts.aspx
        
        Arizona:
        -         
        - https://azdor.gov/forms/individual/form-140-x-y-tables
          https://azdor.gov/forms/individual/form-140-x-y-tables
          https://azdor.gov/forms/individual/form-140-x-y-tables
          https://azdor.gov/businesses-arizona/withholding-tax/withholding-exceptions
          "https://azdor.gov/individuals/income-tax-filing-assistance/identifying-other-taxable-income#:~:text=In%20regards%20to%20capital%20gains
         the%20individuals%20regular%20tax%20rate.
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Alaska', 'state_occupation': 'Arizona', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def alaska_and_arkansas(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, capital_gains_long=0, filing_status='single', income=100000, dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Alaska:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        Arkansas:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Alaska:
        - https://www.commerce.alaska.gov/web/dcra/officeofthestateassessor/alaskataxfacts.aspx
        
        Arkansas:
        -         
        - https://www.arkansasedc.com/why-arkansas/business-climate/tax-structure/personal-income-tax
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_and_AR1000NR_Instructions.pdf
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_FullYearResidentIndividualIncomeTaxReturn.pdf
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_and_AR1000NR_Instructions.pdf
          https://www.arkansasedc.com/why-arkansas/business-climate/tax-structure/capital-gains-tax-reduction
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_and_AR1000NR_Instructions.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Alaska', 'state_occupation': 'Arkansas', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def alaska_and_california(self, state_agi=75000, capital_gains_short=0, capital_gains_long=0, filing_status='single', income=100000, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Alaska:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        California:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Alaska:
        - https://www.commerce.alaska.gov/web/dcra/officeofthestateassessor/alaskataxfacts.aspx
        
        California:
        -         
        - https://www.ftb.ca.gov/forms/2022/2022-540-tax-rate-schedules.pdf
          https://www.ftb.ca.gov/file/personal/deductions/index.html
          https://www.ftb.ca.gov/about-ftb/newsroom/caleitc/eligibility-and-credit-information.html
          https://www.ftb.ca.gov/file/personal/income-types/capital-gains-and-losses.html#:~:text=California%20does%20not%20have%20a
         are%20taxed%20as%20ordinary%20income
          https://www.ftb.ca.gov/file/personal/
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Alaska', 'state_occupation': 'California', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def alaska_and_colorado(self, county_residence='Other', state_agi=75000, capital_gains_short=0, capital_gains_long=0, income=100000, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Alaska:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        Colorado:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Alaska:
        - https://www.commerce.alaska.gov/web/dcra/officeofthestateassessor/alaskataxfacts.aspx
        
        Colorado:
        -         
        - https://leg.colorado.gov/agencies/legislative-council-staff/individual-income-tax%C2%A0
          https://leg.colorado.gov/agencies/legislative-council-staff/individual-income-tax%C2%A0
          https://tax.colorado.gov/sites/tax/files/documents/ITT_Capital_Gain_Subtraction_Feb_2022.pdf
          https://www.auroragov.org/business_services/taxes/occupational_privilege_tax
          https://www.denvergov.org/Government/Agencies-Departments-Offices/Agencies-Departments-Offices-Directory/Department-of-Finance/Our-Divisions/Treasury/Business-Tax-Information#:~:text=Occupational%20privilege%20tax
         -%28show%20below%29&text=Each%20taxable%20employee%20is%20liable
         month%20for%20each%20taxable%20employee
          https://www.glendale.co.us/355/Occupational-Privilege-Tax
          https://www.greenwoodvillage.com/1220/Occupational-Privilege-Tax-OPT
          https://ci.sheridan.co.us/288/Occupational-Privilege-Tax
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        income = self.income if hasattr(self, 'income') else income
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Alaska', 'state_occupation': 'Colorado', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "income": income, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def alaska_and_connecticut(self, state_agi=75000, capital_gains_short=0, capital_gains_long=0, filing_status='single', income=100000, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Alaska:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        Connecticut:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Alaska:
        - https://www.commerce.alaska.gov/web/dcra/officeofthestateassessor/alaskataxfacts.aspx
        
        Connecticut:
        -         
        - https://portal.ct.gov/-/media/DRS/Forms/2022/Income/2022-CT-1040-Instructions_1222.pdf
          https://www.cga.ct.gov/2022/rpt/pdf/2022-R-0108.pdf
          https://www.cga.ct.gov/current/pub/chap_229.htm#sec_12-703
          https://portal.ct.gov/-/media/DRS/Forms/2022/Income/2022-CT-1040-Instructions_1222.pdf
          "https://portal.ct.gov/DRS/Publications/TSSNs/TSSN-29#:~:text=An%20individuals%20net%20capital%20gains
         portion%20of%20Social%20Security%20benefits"
          "https://portal.ct.gov/DRS/Publications/TSSNs/TSSN-29#:~:text=An%20individuals%20net%20capital%20gains
         portion%20of%20Social%20Security%20benefits"
          https://portal.ct.gov/DRS/Individuals/Resident-Income-Tax/Tax-Information
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Alaska', 'state_occupation': 'Connecticut', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def alaska_and_delaware(self, capital_gains_long=0, federal_agi=75000, income=100000, sources=False, capital_gains_short=0, credits_state=0):
        """
        --- Parameters ----
        
        Alaska:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        Delaware:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        federal_agi : int, optional
            Adjusted Gross Income (AGI) for federal taxes. Default is 0. Sources - https://www.irs.gov/e-file-providers/definition-of-adjusted-gross-income, https://www.irs.gov/newsroom/taxpayers-should-know-the-difference-between-standard-and-itemized-deductions
        income : int, optional
            Gross income. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Alaska:
        - https://www.commerce.alaska.gov/web/dcra/officeofthestateassessor/alaskataxfacts.aspx
        
        Delaware:
        -         
        - https://revenue.delaware.gov/software-developer/tax-rate-changes/
          https://revenuefiles.delaware.gov/2022/PIT-RES_TY22_2022-01_PaperInteractive.pdf
          https://revenuefiles.delaware.gov/2022/PIT-RES_TY22_2022-02_Instructions.pdf
          https://revenue.delaware.gov/frequently-asked-questions/personal-income-tax-faqs/
          https://www.newcastlede.gov/DocumentCenter/View/20147/City-of-Wilmington-Wage-Tax-Information-PDF
        """
        
        # Check if the parameters are set upstream in the instance
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        federal_agi = self.federal_agi if hasattr(self, 'federal_agi') else federal_agi
        income = self.income if hasattr(self, 'income') else income
        sources = self.sources if hasattr(self, 'sources') else sources
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Alaska', 'state_occupation': 'Delaware', "capital_gains_long": capital_gains_long, "federal_agi": federal_agi, "income": income, "sources": sources, "capital_gains_short": capital_gains_short, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def alaska_and_florida(self, sources=False):
        """
        --- Parameters ----
        
        Alaska:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        Florida:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Alaska:
        - https://www.commerce.alaska.gov/web/dcra/officeofthestateassessor/alaskataxfacts.aspx
        
        Florida:
        -         
        - https://floridarevenue.com/taxes/taxesfees/Pages/tax_interest_rates.aspx
        """
        
        # Check if the parameters are set upstream in the instance
        sources = self.sources if hasattr(self, 'sources') else sources
        
        # State-specific parameters
        data = {
            'state_residence': 'Alaska', 'state_occupation': 'Florida', "sources": sources
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def alaska_and_georgia(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, capital_gains_long=0, filing_status='single', income=100000, dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Alaska:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        Georgia:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Alaska:
        - https://www.commerce.alaska.gov/web/dcra/officeofthestateassessor/alaskataxfacts.aspx
        
        Georgia:
        -         
        - https://dor.georgia.gov/tax-tables-georgia-tax-rate-schedule
          https://apps.dor.ga.gov/FillableForms/PDFViewer/Index?form=2022GA500
          https://apps.dor.ga.gov/FillableForms/PDFViewer/Index?form=2022GA500
          https://apps.dor.ga.gov/FillableForms/PDFViewer/Index?form=2022GA500
          https://dor.georgia.gov/years-individual-income-tax-forms
          https://dor.georgia.gov/years-individual-income-tax-forms
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Alaska', 'state_occupation': 'Georgia', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def alaska_and_hawaii(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, capital_gains_long=0, filing_status='single', income=100000, dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Alaska:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        Hawaii:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Alaska:
        - https://www.commerce.alaska.gov/web/dcra/officeofthestateassessor/alaskataxfacts.aspx
        
        Hawaii:
        -         
        - https://tax.hawaii.gov/forms/d_18table-on/d_18table-on_p13/
          https://files.hawaii.gov/tax/forms/2018/18table-on.pdf
          https://files.hawaii.gov/tax/news/pubs/22outline.pdf
          https://files.hawaii.gov/tax/news/pubs/FastTaxRefGuide_2017.pdf
          https://www.tfhawaii.org/wordpress/state-tax-resources/mini-tax-guides/individual-income-tax-chapter-235/#:~:text=Long%20term%20capital%20gains%20are%20taxed%20at%20a%20maximum%20of%207.25%25
          https://www.grassrootinstitute.org/2023/02/hb337-increasing-tax-on-capital-gains-a-bad-risk/
          https://files.hawaii.gov/tax/news/pubs/22outline.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Alaska', 'state_occupation': 'Hawaii', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def alaska_and_idaho(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, capital_gains_long=0, filing_status='single', income=100000, dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Alaska:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        Idaho:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Alaska:
        - https://www.commerce.alaska.gov/web/dcra/officeofthestateassessor/alaskataxfacts.aspx
        
        Idaho:
        -         
        - https://tax.idaho.gov/taxes/income-tax/individual-income/individual-income-tax-rate-schedule/
          https://tax.idaho.gov/wp-content/uploads/forms/EIN00046/EIN00046_03-01-2023.pdf
          https://tax.idaho.gov/taxes/income-tax/individual-income/idaho-source-income/
          https://tax.idaho.gov/taxes/income-tax/individual-income/idaho-source-income/
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Alaska', 'state_occupation': 'Idaho', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def alaska_and_illinois(self, state_agi=75000, capital_gains_short=0, capital_gains_long=0, filing_status='single', income=100000, dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Alaska:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        Illinois:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Alaska:
        - https://www.commerce.alaska.gov/web/dcra/officeofthestateassessor/alaskataxfacts.aspx
        
        Illinois:
        -         
        - https://tax.illinois.gov/research/taxrates/income.html#IndividualIncome
          https://tax.illinois.gov/content/dam/soi/en/web/tax/forms/incometax/documents/currentyear/individual/il-1040-instr.pdf
          https://tax.illinois.gov/questionsandanswers/12.html
          https://tax.illinois.gov/content/dam/soi/en/web/tax/forms/incometax/documents/currentyear/individual/il-1040-instr.pdf
          https://tax.illinois.gov/research/taxrates/personalproperty.html
          https://tax.illinois.gov/localgovernments/income.html
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Alaska', 'state_occupation': 'Illinois', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def alaska_and_indiana(self, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_short=0, capital_gains_long=0, filing_status='single', income=100000, dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Alaska:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        Indiana:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Alaska:
        - https://www.commerce.alaska.gov/web/dcra/officeofthestateassessor/alaskataxfacts.aspx
        
        Indiana:
        -         
        - https://www.in.gov/dor/tax-forms/2022-individual-income-tax-forms/
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-32.htm
          https://www.in.gov/dor/files/ib28.pdf
          https://www.in.gov/dor/tax-forms/2022-individual-income-tax-forms/
          https://www.in.gov/dor/files/2022-county-tax-rates-and-codes.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Alaska', 'state_occupation': 'Indiana', "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def alaska_and_iowa(self, county_residence='Other', state_agi=75000, capital_gains_short=0, capital_gains_long=0, filing_status='single', income=100000, dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Alaska:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        Iowa:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Alaska:
        - https://www.commerce.alaska.gov/web/dcra/officeofthestateassessor/alaskataxfacts.aspx
        
        Iowa:
        -         
        - https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
          https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
          https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
          https://tax.iowa.gov/iowa-illinois-reciprocal-agreement
          https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Alaska', 'state_occupation': 'Iowa', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def alaska_and_kansas(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, capital_gains_long=0, filing_status='single', income=100000, dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Alaska:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        Kansas:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Alaska:
        - https://www.commerce.alaska.gov/web/dcra/officeofthestateassessor/alaskataxfacts.aspx
        
        Kansas:
        -         
        - https://www.ksrevenue.gov/taxrates.html
          https://www.ksrevenue.gov/incomebook22.html#0
          https://www.ksrevenue.gov/incomebook22.html#0
          https://www.ksrevenue.gov/incomebook22.html#0
          https://www.ksrevenue.gov/incomebook22.html
          https://www.ksrevenue.gov/incomebook22.html
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Alaska', 'state_occupation': 'Kansas', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def alaska_and_kentucky(self, county_residence='Other', state_agi=75000, capital_gains_short=0, capital_gains_long=0, filing_status='single', income=100000, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Alaska:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        Kentucky:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Alaska:
        - https://www.commerce.alaska.gov/web/dcra/officeofthestateassessor/alaskataxfacts.aspx
        
        Kentucky:
        -         
        - https://revenue.ky.gov/News/Pages/Dor-Announces-Updates-to-Individual-Income-Tax-for-2023-Tax-Year.aspx
          https://revenue.ky.gov/Individual/Individual-Income-Tax/Pages/default.aspx
          https://revenue.ky.gov/News/Pages/Dor-Announces-Updates-to-Individual-Income-Tax-for-2023-Tax-Year.aspx
          https://revenue.ky.gov/Individual/Individual-Income-Tax/Pages/default.aspx
          https://revenue.ky.gov/Dor%20Training%20Materials/103%20KAR%2017.140.%20Individual%20income%20tax%20-%20reciprocity%20-%20nonresidents.pdf
          https://revenue.ky.gov/Individual/Individual-Income-Tax/Pages/default.aspx
          https://nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2021/Taxes-21-32.htm?taxmap=true
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Alaska', 'state_occupation': 'Kentucky', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def alaska_and_louisiana(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, capital_gains_long=0, filing_status='single', income=100000, dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Alaska:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        Louisiana:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Alaska:
        - https://www.commerce.alaska.gov/web/dcra/officeofthestateassessor/alaskataxfacts.aspx
        
        Louisiana:
        -         
        - https://revenue.louisiana.gov/individualincometax
          https://revenue.louisiana.gov/individualincometax
          https://www.revenue.louisiana.gov/taxforms/6935(11_02)F.pdf
          https://revenue.louisiana.gov/individualincometax
          https://revenue.louisiana.gov/individualincometax
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Alaska', 'state_occupation': 'Louisiana', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def alaska_and_maine(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, capital_gains_long=0, filing_status='single', income=100000, dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Alaska:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        Maine:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Alaska:
        - https://www.commerce.alaska.gov/web/dcra/officeofthestateassessor/alaskataxfacts.aspx
        
        Maine:
        -         
        - https://www.maine.gov/revenue/tax-return-forms/individual-income-tax-2022
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/ind_tax_rate_sched_2022.pdf
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/22_1040me_book_gen_instruc_revisedFeb23.pdf
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/22_1040me_book_gen_instruc_revisedFeb23.pdf
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/22_1040me_book_gen_instruc_revisedFeb23.pdf
          https://www.maine.gov/revenue/tax-return-forms/individual-income-tax-2022
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Alaska', 'state_occupation': 'Maine', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def alaska_and_maryland(self, county_residence='Other', state_agi=75000, capital_gains_short=0, capital_gains_long=0, filing_status='single', income=100000, dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Alaska:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        Maryland:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Alaska:
        - https://www.commerce.alaska.gov/web/dcra/officeofthestateassessor/alaskataxfacts.aspx
        
        Maryland:
        -         
        - https://www.marylandtaxes.gov/individual/income/tax-info/tax-rates.php
          https://www.marylandtaxes.gov/forms/22_forms/Resident_Booklet.pdf
          https://www.marylandtaxes.gov/forms/Tax_Publications/Administrative_Releases/Income_and_Estate_Tax_Releases/ar_it37.pdf
          https://www.marylandtaxes.gov/individual/index.php
          https://www.marylandtaxes.gov/individual/income/tax-info/tax-rates.php
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Alaska', 'state_occupation': 'Maryland', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def alaska_and_massachusetts(self, state_agi=75000, capital_gains_short=0, capital_gains_long=0, filing_status='single', income=100000, dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Alaska:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        Massachusetts:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Alaska:
        - https://www.commerce.alaska.gov/web/dcra/officeofthestateassessor/alaskataxfacts.aspx
        
        Massachusetts:
        -         
        - https://www.mass.gov/service-details/massachusetts-tax-rates
          https://www.mass.gov/service-details/view-massachusetts-personal-income-tax-exemptions
          https://www.mass.gov/service-details/view-massachusetts-personal-income-tax-exemptions
          https://www.mass.gov/service-details/view-child-and-dependent-related-credits
          https://www.mass.gov/service-details/massachusetts-tax-rates
          https://www.mass.gov/personal-income-tax
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Alaska', 'state_occupation': 'Massachusetts', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def alaska_and_michigan(self, county_residence='Other', state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', county_occupation='Other', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Alaska:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        Michigan:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        county_occupation : str, optional
            County of occupation. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Alaska:
        - https://www.commerce.alaska.gov/web/dcra/officeofthestateassessor/alaskataxfacts.aspx
        
        Michigan:
        -         
        - https://www.michigan.gov/taxes/iit/new-developments-for-tax-year-2022
          "https://www.michigan.gov/taxes/-/media/Project/Websites/taxes/Forms/2023/2023-SUW/446_Withholding-Guide_2023.pdf?rev=177ca3d8cb034392bf84e317ed78c320&hash=19C5F42F43A89A91863D229349F48883#:~:text=The%20exemption%20amount%20is%20%245%2C400
         employees%20Michigan%20income%20tax%20return"
          "https://www.michigan.gov/taxes/-/media/Project/Websites/taxes/Forms/2023/2023-SUW/446_Withholding-Guide_2023.pdf?rev=177ca3d8cb034392bf84e317ed78c320&hash=19C5F42F43A89A91863D229349F48883#:~:text=The%20exemption%20amount%20is%20%245%2C400
         employees%20Michigan%20income%20tax%20return"
          https://www.michigan.gov/taxes/questions/iit/accordion/residency/are-my-wages-earned-in-another-state-taxable-in-michigan-if-i-am-a-michigan-resident-1
          https://www.michigan.gov/taxes/-/media/Project/Websites/taxes/Forms/2022/2022-IIT-Forms/BOOK_MI-1040.pdf?rev=4524f1cbbc0243b194ab05a6680be75b&hash=40DFE71E651F7C1D65C31F6CD03D944F
          https://www.michigan.gov/taxes/questions/iit/accordion/general/what-cities-impose-an-income-tax
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        county_occupation = self.county_occupation if hasattr(self, 'county_occupation') else county_occupation
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Alaska', 'state_occupation': 'Michigan', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "county_occupation": county_occupation, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def alaska_and_minnesota(self, state_agi=75000, capital_gains_short=0, capital_gains_long=0, filing_status='single', income=100000, dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Alaska:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        Minnesota:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Alaska:
        - https://www.commerce.alaska.gov/web/dcra/officeofthestateassessor/alaskataxfacts.aspx
        
        Minnesota:
        -         
        - https://www.revenue.state.mn.us/minnesota-income-tax-rates-and-brackets
          https://www.revenue.state.mn.us/child-and-dependent-care-credit
          https://www.revenue.state.mn.us/dependent-exemptions
          https://www.revenue.state.mn.us/minnesota-standard-deduction
          https://mn.gov/mmb/assets/mwr_form_tcm1059-128581.pdf
          https://www.revenue.state.mn.us/individual-income-tax
          https://www.revenue.state.mn.us/individual-income-tax
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Alaska', 'state_occupation': 'Minnesota', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def alaska_and_mississippi(self, state_agi=75000, capital_gains_short=0, capital_gains_long=0, filing_status='single', income=100000, dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Alaska:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        Mississippi:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Alaska:
        - https://www.commerce.alaska.gov/web/dcra/officeofthestateassessor/alaskataxfacts.aspx
        
        Mississippi:
        -         
        - https://www.dor.ms.gov/individual/tax-rates
          https://www.dor.ms.gov/individual/tax-rates
          https://www.dor.ms.gov/individual/tax-rates
          https://www.dor.ms.gov/individual
          https://www.dor.ms.gov/individual
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Alaska', 'state_occupation': 'Mississippi', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def alaska_and_missouri(self, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_short=0, capital_gains_long=0, filing_status='single', income=100000, dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Alaska:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        Missouri:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Alaska:
        - https://www.commerce.alaska.gov/web/dcra/officeofthestateassessor/alaskataxfacts.aspx
        
        Missouri:
        -         
        - https://dor.mo.gov/taxation/individual/tax-types/income/year-changes/
          https://dor.mo.gov/taxation/individual/tax-types/income/year-changes/
          https://dor.mo.gov/taxation/individual/tax-types/income/year-changes/
          https://www.kcmo.gov/city-hall/departments/finance/earnings-tax
          https://www.stlouis-mo.gov/government/departments/collector/earnings-tax/individual-earnings-tax-info.cfm
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Alaska', 'state_occupation': 'Missouri', "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def alaska_and_montana(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, capital_gains_long=0, filing_status='single', income=100000, dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Alaska:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        Montana:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Alaska:
        - https://www.commerce.alaska.gov/web/dcra/officeofthestateassessor/alaskataxfacts.aspx
        
        Montana:
        -         
        - https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/2022/12/montana-employees-withholding-allowance-and-exemption-certificate-form-mw-4.pdf
          https://mtrevenue.gov/2021/11/29/simplification-of-montana-income-taxation/?utm_source=rss&utm_medium=rss&utm_campaign=simplification-of-montana-income-taxation
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Alaska', 'state_occupation': 'Montana', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def alaska_and_nebraska(self, state_agi=75000, capital_gains_short=0, capital_gains_long=0, filing_status='single', income=100000, dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Alaska:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        Nebraska:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Alaska:
        - https://www.commerce.alaska.gov/web/dcra/officeofthestateassessor/alaskataxfacts.aspx
        
        Nebraska:
        -         
        - https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2023_Circular_EN_8-429r2023_Whole_Book_11.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Alaska', 'state_occupation': 'Nebraska', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def alaska_and_nevada(self, sources=False):
        """
        --- Parameters ----
        
        Alaska:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        Nevada:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Alaska:
        - https://www.commerce.alaska.gov/web/dcra/officeofthestateassessor/alaskataxfacts.aspx
        
        Nevada:
        -         
        - https://www.nvsos.gov/sos/businesses/the-nevada-advantage
        """
        
        # Check if the parameters are set upstream in the instance
        sources = self.sources if hasattr(self, 'sources') else sources
        
        # State-specific parameters
        data = {
            'state_residence': 'Alaska', 'state_occupation': 'Nevada', "sources": sources
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def alaska_and_new_hampshire(self, sources=False):
        """
        --- Parameters ----
        
        Alaska:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        New Hampshire:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Alaska:
        - https://www.commerce.alaska.gov/web/dcra/officeofthestateassessor/alaskataxfacts.aspx
        
        New Hampshire:
        -         
        - https://www.revenue.nh.gov/assistance/tax-overview.htm
        """
        
        # Check if the parameters are set upstream in the instance
        sources = self.sources if hasattr(self, 'sources') else sources
        
        # State-specific parameters
        data = {
            'state_residence': 'Alaska', 'state_occupation': 'NewHampshire', "sources": sources
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def alaska_and_new_jersey(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, capital_gains_long=0, filing_status='single', income=100000, dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Alaska:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        New Jersey:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Alaska:
        - https://www.commerce.alaska.gov/web/dcra/officeofthestateassessor/alaskataxfacts.aspx
        
        New Jersey:
        -         
        - https://www.state.nj.us/treasury/taxation/taxtables.shtml
          https://www.state.nj.us/treasury/taxation/njit13.shtml
          https://www.nj.gov/treasury/taxation/njit25.shtml
          https://www.state.nj.us/treasury/taxation/individuals/freqqiti.shtml
          https://www.state.nj.us/treasury/taxation/localtax.shtml
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Alaska', 'state_occupation': 'NewJersey', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def alaska_and_new_mexico(self, state_agi=75000, capital_gains_short=0, capital_gains_long=0, filing_status='single', income=100000, dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Alaska:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        New Mexico:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Alaska:
        - https://www.commerce.alaska.gov/web/dcra/officeofthestateassessor/alaskataxfacts.aspx
        
        New Mexico:
        -         
        - https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/fdf3c548-8aba-4b9c-9eb4-bb564c716015/FYI-104.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/0bf943fd-652e-4400-bb1b-cd8397eb5a95/2022pit-adj.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/2f1a6781-9534-4436-b427-1557f9592099/2022pit-adj-ins.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/0bf943fd-652e-4400-bb1b-cd8397eb5a95/2022pit-adj.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/1afc56af-ea90-4d48-82e5-1f9aeb43255a/PITbook2022.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/fdf3c548-8aba-4b9c-9eb4-bb564c716015/FYI-104.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Alaska', 'state_occupation': 'NewMexico', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def alaska_and_new_york(self, county_residence='Other', state_agi=75000, capital_gains_short=0, capital_gains_long=0, filing_status='single', income=100000, dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Alaska:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        New York:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Alaska:
        - https://www.commerce.alaska.gov/web/dcra/officeofthestateassessor/alaskataxfacts.aspx
        
        New York:
        -         
        - https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nys-tax-rate-schedule
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nys-tax-rate-schedule
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nys-tax-rate-schedule
          https://www.tax.ny.gov/pit/
          https://www.tax.ny.gov/pit/file/nonresident-faqs.htm
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#yonkers-tax-rate-schedule
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nyc-tax-rate-schedule
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Alaska', 'state_occupation': 'NewYork', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def alaska_and_north_carolina(self, state_agi=75000, capital_gains_short=0, capital_gains_long=0, filing_status='single', income=100000, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Alaska:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        North Carolina:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Alaska:
        - https://www.commerce.alaska.gov/web/dcra/officeofthestateassessor/alaskataxfacts.aspx
        
        North Carolina:
        -         
        - https://www.ncdor.gov/taxes-forms/tax-rate-schedules
          https://www.ncdor.gov/taxes-forms/individual-income-tax/north-carolina-standard-deduction-or-north-carolina-itemized-deductions
          https://www.ncdor.gov/taxes-forms/individual-income-tax
          https://www.ncdor.gov/taxes-forms/individual-income-tax
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Alaska', 'state_occupation': 'NorthCarolina', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def alaska_and_north_dakota(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, capital_gains_long=0, filing_status='single', income=100000, dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Alaska:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        North Dakota:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Alaska:
        - https://www.commerce.alaska.gov/web/dcra/officeofthestateassessor/alaskataxfacts.aspx
        
        North Dakota:
        -         
        - https://www.tax.nd.gov/forms
          https://www.tax.nd.gov/sites/www/files/documents/forms/individual/2022-iit/2022-individual-income-tax-booklet.pdf
          https://www.tax.nd.gov/sites/www/files/documents/forms/2022-individual-income-tax-booklet.pdf
          https://www.tax.nd.gov/tax-exemptions-credits/income-tax-exemptions-credits
          https://www.tax.nd.gov/sites/www/files/documents/forms/business/ndwrfillable.pdf
          https://www.tax.nd.gov/news/tax-legislative-changes/significant-changes-law/individual-income-tax-history
          https://www.tax.nd.gov/forms
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Alaska', 'state_occupation': 'NorthDakota', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def alaska_and_ohio(self, county_residence='Other', state_agi=75000, capital_gains_short=0, capital_gains_long=0, filing_status='single', income=100000, dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Alaska:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        Ohio:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Alaska:
        - https://www.commerce.alaska.gov/web/dcra/officeofthestateassessor/alaskataxfacts.aspx
        
        Ohio:
        -         
        - https://tax.ohio.gov/individual/resources/annual-tax-rates
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://tax.ohio.gov/static/forms/employer_withholding/generic/wth_it4nr.pdf
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://www.ritaohio.com/TaxRatesTable
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Alaska', 'state_occupation': 'Ohio', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def alaska_and_oklahoma(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, capital_gains_long=0, filing_status='single', income=100000, dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Alaska:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        Oklahoma:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Alaska:
        - https://www.commerce.alaska.gov/web/dcra/officeofthestateassessor/alaskataxfacts.aspx
        
        Oklahoma:
        -         
        - https://oklahoma.gov/content/dam/ok/en/tax/documents/resources/publications/businesses/withholding-tables/WHTables-2022.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Alaska', 'state_occupation': 'Oklahoma', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def alaska_and_oregon(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, capital_gains_long=0, filing_status='single', income=100000, dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Alaska:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        Oregon:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Alaska:
        - https://www.commerce.alaska.gov/web/dcra/officeofthestateassessor/alaskataxfacts.aspx
        
        Oregon:
        -         
        - https://www.oregon.gov/dor/programs/individuals/Pages/PIT.aspx
          https://www.oregon.gov/dor/programs/individuals/Documents/Full%20year%20resident
         %20Form%20or-40%20filers.pdf
          https://www.oregon.gov/dor/programs/individuals/Documents/Part-year%20and%20nonresident
         %20Form%20or-40-P%20and%20or-40-N%20filers.pdf.pdf
          https://www.oregon.gov/dor/forms/FormsPubs/form-or-40_101-040_2022.pdf
          https://www.oregon.gov/dor/forms/FormsPubs/form-or-40_101-040_2022.pdf
          https://www.oregon.gov/dor/forms/FormsPubs/form-or-40_101-040_2022.pdf
          https://www.portland.gov/revenue/personal-tax
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Alaska', 'state_occupation': 'Oregon', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def alaska_and_pennsylvania(self, county_residence='Other', state_agi=75000, county_occupation='Other', capital_gains_short=0, capital_gains_long=0, income=100000, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Alaska:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        Pennsylvania:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        county_residence : str, optional
            County of residence. Default is Other.
        county_occupation : str, optional
            County of occupation. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Alaska:
        - https://www.commerce.alaska.gov/web/dcra/officeofthestateassessor/alaskataxfacts.aspx
        
        Pennsylvania:
        -         
        - https://www.revenue.pa.gov/Tax%20Rates/Pages/default.aspx
          https://www.revenue.pa.gov/FormsandPublications/FormsforIndividuals/PIT/Documents/2022/2022_pa-40in.pdf
          https://www.revenue.pa.gov/FormsandPublications/FormsforBusinesses/EmployerWithholding/Documents/rev-419.pdf
          https://www.revenue.pa.gov/FormsandPublications/FormsforIndividuals/PIT/Documents/2022/2022_pa-40in.pdf
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-25.htm
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        county_occupation = self.county_occupation if hasattr(self, 'county_occupation') else county_occupation
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        income = self.income if hasattr(self, 'income') else income
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Alaska', 'state_occupation': 'Pennsylvania', "county_residence": county_residence, "state_agi": state_agi, "county_occupation": county_occupation, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "income": income, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def alaska_and_rhode_island(self, state_agi=75000, capital_gains_short=0, capital_gains_long=0, filing_status='single', income=100000, dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Alaska:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        Rhode Island:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Alaska:
        - https://www.commerce.alaska.gov/web/dcra/officeofthestateassessor/alaskataxfacts.aspx
        
        Rhode Island:
        -         
        - https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2022_1040WE_w_0.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Alaska', 'state_occupation': 'RhodeIsland', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def alaska_and_south_carolina(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, capital_gains_long=0, filing_status='single', income=100000, dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Alaska:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        South Carolina:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Alaska:
        - https://www.commerce.alaska.gov/web/dcra/officeofthestateassessor/alaskataxfacts.aspx
        
        South Carolina:
        -         
        - https://dor.sc.gov/resources-site/lawandpolicy/Advisory%20Opinions/IL22-15.pdf
          https://dor.sc.gov/forms-site/Forms/IITPacket_2022.pdf
          https://www.scstatehouse.gov/code/t12c006.php
          https://dor.sc.gov/resources-site/lawandpolicy/Advisory%20Opinions/IL22-15.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Alaska', 'state_occupation': 'SouthCarolina', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def alaska_and_south_dakota(self, sources=False):
        """
        --- Parameters ----
        
        Alaska:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        South Dakota:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Alaska:
        - https://www.commerce.alaska.gov/web/dcra/officeofthestateassessor/alaskataxfacts.aspx
        
        South Dakota:
        -         
        - https://dor.sd.gov/individuals/taxes/
        """
        
        # Check if the parameters are set upstream in the instance
        sources = self.sources if hasattr(self, 'sources') else sources
        
        # State-specific parameters
        data = {
            'state_residence': 'Alaska', 'state_occupation': 'SouthDakota', "sources": sources
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def alaska_and_tennessee(self, sources=False):
        """
        --- Parameters ----
        
        Alaska:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        Tennessee:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Alaska:
        - https://www.commerce.alaska.gov/web/dcra/officeofthestateassessor/alaskataxfacts.aspx
        
        Tennessee:
        -         
        - https://www.tn.gov/revenue/taxes.html
        """
        
        # Check if the parameters are set upstream in the instance
        sources = self.sources if hasattr(self, 'sources') else sources
        
        # State-specific parameters
        data = {
            'state_residence': 'Alaska', 'state_occupation': 'Tennessee', "sources": sources
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def alaska_and_texas(self, sources=False):
        """
        --- Parameters ----
        
        Alaska:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        Texas:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Alaska:
        - https://www.commerce.alaska.gov/web/dcra/officeofthestateassessor/alaskataxfacts.aspx
        
        Texas:
        -         
        - https://comptroller.texas.gov/taxes/
        """
        
        # Check if the parameters are set upstream in the instance
        sources = self.sources if hasattr(self, 'sources') else sources
        
        # State-specific parameters
        data = {
            'state_residence': 'Alaska', 'state_occupation': 'Texas', "sources": sources
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def alaska_and_utah(self, state_agi=75000, capital_gains_short=0, capital_gains_long=0, filing_status='single', income=100000, dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Alaska:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        Utah:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Alaska:
        - https://www.commerce.alaska.gov/web/dcra/officeofthestateassessor/alaskataxfacts.aspx
        
        Utah:
        -         
        - https://incometax.utah.gov/paying/tax-rates
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Alaska', 'state_occupation': 'Utah', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def alaska_and_vermont(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, capital_gains_long=0, filing_status='single', income=100000, dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Alaska:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        Vermont:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Alaska:
        - https://www.commerce.alaska.gov/web/dcra/officeofthestateassessor/alaskataxfacts.aspx
        
        Vermont:
        -         
        - https://tax.vermont.gov/sites/tax/files/documents/GB-1210-2023.pdf
          https://tax.vermont.gov/sites/tax/files/documents/Income%20Booklet-2022.pdf
          https://tax.vermont.gov/sites/tax/files/documents/Income%20Booklet-2022.pdf
          https://tax.vermont.gov/sites/tax/files/documents/Income%20Booklet-2022.pdf
          https://tax.vermont.gov/individuals/personal-income-tax/tax-credits
          https://tax.vermont.gov/individuals/personal-income-tax/taxable-income
          https://tax.vermont.gov/individuals/personal-income-tax/local-option-tax
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Alaska', 'state_occupation': 'Vermont', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def alaska_and_virginia(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, capital_gains_long=0, filing_status='single', income=100000, dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Alaska:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        Virginia:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Alaska:
        - https://www.commerce.alaska.gov/web/dcra/officeofthestateassessor/alaskataxfacts.aspx
        
        Virginia:
        -         
        - https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/reciprocity
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Alaska', 'state_occupation': 'Virginia', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def alaska_and_washington(self, capital_gains_long=0, sources=False, capital_gains_short=0):
        """
        --- Parameters ----
        
        Alaska:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        Washington:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Alaska:
        - https://www.commerce.alaska.gov/web/dcra/officeofthestateassessor/alaskataxfacts.aspx
        
        Washington:
        -         
        - https://dor.wa.gov/taxes-rates/income-tax
          https://dor.wa.gov/taxes-rates/other-taxes/capital-gains-tax
          https://dor.wa.gov/taxes-rates/income-tax
        """
        
        # Check if the parameters are set upstream in the instance
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        sources = self.sources if hasattr(self, 'sources') else sources
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        
        # State-specific parameters
        data = {
            'state_residence': 'Alaska', 'state_occupation': 'Washington', "capital_gains_long": capital_gains_long, "sources": sources, "capital_gains_short": capital_gains_short
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def alaska_and_west_virginia(self, pay_periods=24, state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', county_occupation='Other', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Alaska:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        West Virginia:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_occupation : str, optional
            County of occupation. Default is Other.
        pay_periods : int, optional
            Number of pay periods. Options are 13, 24, 26, 52. Default is 24.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Alaska:
        - https://www.commerce.alaska.gov/web/dcra/officeofthestateassessor/alaskataxfacts.aspx
        
        West Virginia:
        -         
        - https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://tax.wv.gov/Documents/TSD/tsd438.pdf
          https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2011/Taxes-11-35att.html
        """
        
        # Check if the parameters are set upstream in the instance
        pay_periods = self.pay_periods if hasattr(self, 'pay_periods') else pay_periods
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        county_occupation = self.county_occupation if hasattr(self, 'county_occupation') else county_occupation
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Alaska', 'state_occupation': 'WestVirginia', "pay_periods": pay_periods, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "county_occupation": county_occupation, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def alaska_and_wisconsin(self, state_agi=75000, capital_gains_short=0, capital_gains_long=0, filing_status='single', income=100000, dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Alaska:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        Wisconsin:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Alaska:
        - https://www.commerce.alaska.gov/web/dcra/officeofthestateassessor/alaskataxfacts.aspx
        
        Wisconsin:
        -         
        - https://www.revenue.wi.gov/TaxForms2023/2023-Form1-ES-Inst.pdf
          https://www.revenue.wi.gov/TaxForms2023/2023-Form1-ES-Inst.pdf
          https://www.revenue.wi.gov/TaxForms2023/2023-Form1-ES-Inst.pdf
          https://www.revenue.wi.gov/Pages/FAQS/pcs-work.aspx
          https://www.revenue.wi.gov/Dor%20Publications/pb103.pdf
          https://www.revenue.wi.gov/Pages/FAQS/pcs-work.aspx
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Alaska', 'state_occupation': 'Wisconsin', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def alaska_and_wyoming(self, sources=False):
        """
        --- Parameters ----
        
        Alaska:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        Wyoming:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Alaska:
        - https://www.commerce.alaska.gov/web/dcra/officeofthestateassessor/alaskataxfacts.aspx
        
        Wyoming:
        -         
        - https://revenue.wyo.gov/
        """
        
        # Check if the parameters are set upstream in the instance
        sources = self.sources if hasattr(self, 'sources') else sources
        
        # State-specific parameters
        data = {
            'state_residence': 'Alaska', 'state_occupation': 'Wyoming', "sources": sources
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def arizona_and_alabama(self, county_residence='Other', state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Arizona:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        
        Alabama:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Arizona:
        - https://azdor.gov/forms/individual/form-140-x-y-tables
          https://azdor.gov/forms/individual/form-140-x-y-tables
          https://azdor.gov/forms/individual/form-140-x-y-tables
          https://azdor.gov/businesses-arizona/withholding-tax/withholding-exceptions
          "https://azdor.gov/individuals/income-tax-filing-assistance/identifying-other-taxable-income#:~:text=In%20regards%20to%20capital%20gains
         the%20individuals%20regular%20tax%20rate.
        
        Alabama:
        -         
        - https://www.revenue.alabama.gov/ultraviewer/viewer/basic_viewer/index.html?form=2023/02/22f40.pdf
          https://www.revenue.alabama.gov/faqs/what-is-alabamas-individual-income-tax-rate/
          https://www.revenue.alabama.gov/ultraviewer/viewer/basic_viewer/index.html?form=2023/02/22f40.pdf
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-24.htm?taxmap=true
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-24.htm?taxmap=true
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-24.htm?taxmap=true
          https://www.revenue.alabama.gov/individual-corporate/income-to-be-reported-on-the-alabama-income-tax-return/
          https://www.almonline.org/TaxRates.aspx#OccupationalTax
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Arizona', 'state_occupation': 'Alabama', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def arizona_and_alaska(self, state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Arizona:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        
        Alaska:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Arizona:
        - https://azdor.gov/forms/individual/form-140-x-y-tables
          https://azdor.gov/forms/individual/form-140-x-y-tables
          https://azdor.gov/forms/individual/form-140-x-y-tables
          https://azdor.gov/businesses-arizona/withholding-tax/withholding-exceptions
          "https://azdor.gov/individuals/income-tax-filing-assistance/identifying-other-taxable-income#:~:text=In%20regards%20to%20capital%20gains
         the%20individuals%20regular%20tax%20rate.
        
        Alaska:
        -         
        - https://www.commerce.alaska.gov/web/dcra/officeofthestateassessor/alaskataxfacts.aspx
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Arizona', 'state_occupation': 'Alaska', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def arizona_and_arkansas(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Arizona:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        
        Arkansas:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Arizona:
        - https://azdor.gov/forms/individual/form-140-x-y-tables
          https://azdor.gov/forms/individual/form-140-x-y-tables
          https://azdor.gov/forms/individual/form-140-x-y-tables
          https://azdor.gov/businesses-arizona/withholding-tax/withholding-exceptions
          "https://azdor.gov/individuals/income-tax-filing-assistance/identifying-other-taxable-income#:~:text=In%20regards%20to%20capital%20gains
         the%20individuals%20regular%20tax%20rate.
        
        Arkansas:
        -         
        - https://www.arkansasedc.com/why-arkansas/business-climate/tax-structure/personal-income-tax
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_and_AR1000NR_Instructions.pdf
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_FullYearResidentIndividualIncomeTaxReturn.pdf
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_and_AR1000NR_Instructions.pdf
          https://www.arkansasedc.com/why-arkansas/business-climate/tax-structure/capital-gains-tax-reduction
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_and_AR1000NR_Instructions.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Arizona', 'state_occupation': 'Arkansas', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def arizona_and_california(self, state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Arizona:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        
        California:
        


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Arizona:
        - https://azdor.gov/forms/individual/form-140-x-y-tables
          https://azdor.gov/forms/individual/form-140-x-y-tables
          https://azdor.gov/forms/individual/form-140-x-y-tables
          https://azdor.gov/businesses-arizona/withholding-tax/withholding-exceptions
          "https://azdor.gov/individuals/income-tax-filing-assistance/identifying-other-taxable-income#:~:text=In%20regards%20to%20capital%20gains
         the%20individuals%20regular%20tax%20rate.
        
        California:
        -         
        - https://www.ftb.ca.gov/forms/2022/2022-540-tax-rate-schedules.pdf
          https://www.ftb.ca.gov/file/personal/deductions/index.html
          https://www.ftb.ca.gov/about-ftb/newsroom/caleitc/eligibility-and-credit-information.html
          https://www.ftb.ca.gov/file/personal/income-types/capital-gains-and-losses.html#:~:text=California%20does%20not%20have%20a
         are%20taxed%20as%20ordinary%20income
          https://www.ftb.ca.gov/file/personal/
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Arizona', 'state_occupation': 'California', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def arizona_and_colorado(self, county_residence='Other', state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Arizona:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        
        Colorado:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Arizona:
        - https://azdor.gov/forms/individual/form-140-x-y-tables
          https://azdor.gov/forms/individual/form-140-x-y-tables
          https://azdor.gov/forms/individual/form-140-x-y-tables
          https://azdor.gov/businesses-arizona/withholding-tax/withholding-exceptions
          "https://azdor.gov/individuals/income-tax-filing-assistance/identifying-other-taxable-income#:~:text=In%20regards%20to%20capital%20gains
         the%20individuals%20regular%20tax%20rate.
        
        Colorado:
        -         
        - https://leg.colorado.gov/agencies/legislative-council-staff/individual-income-tax%C2%A0
          https://leg.colorado.gov/agencies/legislative-council-staff/individual-income-tax%C2%A0
          https://tax.colorado.gov/sites/tax/files/documents/ITT_Capital_Gain_Subtraction_Feb_2022.pdf
          https://www.auroragov.org/business_services/taxes/occupational_privilege_tax
          https://www.denvergov.org/Government/Agencies-Departments-Offices/Agencies-Departments-Offices-Directory/Department-of-Finance/Our-Divisions/Treasury/Business-Tax-Information#:~:text=Occupational%20privilege%20tax
         -%28show%20below%29&text=Each%20taxable%20employee%20is%20liable
         month%20for%20each%20taxable%20employee
          https://www.glendale.co.us/355/Occupational-Privilege-Tax
          https://www.greenwoodvillage.com/1220/Occupational-Privilege-Tax-OPT
          https://ci.sheridan.co.us/288/Occupational-Privilege-Tax
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Arizona', 'state_occupation': 'Colorado', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def arizona_and_connecticut(self, state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Arizona:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        
        Connecticut:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Arizona:
        - https://azdor.gov/forms/individual/form-140-x-y-tables
          https://azdor.gov/forms/individual/form-140-x-y-tables
          https://azdor.gov/forms/individual/form-140-x-y-tables
          https://azdor.gov/businesses-arizona/withholding-tax/withholding-exceptions
          "https://azdor.gov/individuals/income-tax-filing-assistance/identifying-other-taxable-income#:~:text=In%20regards%20to%20capital%20gains
         the%20individuals%20regular%20tax%20rate.
        
        Connecticut:
        -         
        - https://portal.ct.gov/-/media/DRS/Forms/2022/Income/2022-CT-1040-Instructions_1222.pdf
          https://www.cga.ct.gov/2022/rpt/pdf/2022-R-0108.pdf
          https://www.cga.ct.gov/current/pub/chap_229.htm#sec_12-703
          https://portal.ct.gov/-/media/DRS/Forms/2022/Income/2022-CT-1040-Instructions_1222.pdf
          "https://portal.ct.gov/DRS/Publications/TSSNs/TSSN-29#:~:text=An%20individuals%20net%20capital%20gains
         portion%20of%20Social%20Security%20benefits"
          "https://portal.ct.gov/DRS/Publications/TSSNs/TSSN-29#:~:text=An%20individuals%20net%20capital%20gains
         portion%20of%20Social%20Security%20benefits"
          https://portal.ct.gov/DRS/Individuals/Resident-Income-Tax/Tax-Information
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Arizona', 'state_occupation': 'Connecticut', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def arizona_and_delaware(self, state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, federal_agi=75000, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Arizona:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        
        Delaware:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        federal_agi : int, optional
            Adjusted Gross Income (AGI) for federal taxes. Default is 0. Sources - https://www.irs.gov/e-file-providers/definition-of-adjusted-gross-income, https://www.irs.gov/newsroom/taxpayers-should-know-the-difference-between-standard-and-itemized-deductions
        income : int, optional
            Gross income. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Arizona:
        - https://azdor.gov/forms/individual/form-140-x-y-tables
          https://azdor.gov/forms/individual/form-140-x-y-tables
          https://azdor.gov/forms/individual/form-140-x-y-tables
          https://azdor.gov/businesses-arizona/withholding-tax/withholding-exceptions
          "https://azdor.gov/individuals/income-tax-filing-assistance/identifying-other-taxable-income#:~:text=In%20regards%20to%20capital%20gains
         the%20individuals%20regular%20tax%20rate.
        
        Delaware:
        -         
        - https://revenue.delaware.gov/software-developer/tax-rate-changes/
          https://revenuefiles.delaware.gov/2022/PIT-RES_TY22_2022-01_PaperInteractive.pdf
          https://revenuefiles.delaware.gov/2022/PIT-RES_TY22_2022-02_Instructions.pdf
          https://revenue.delaware.gov/frequently-asked-questions/personal-income-tax-faqs/
          https://www.newcastlede.gov/DocumentCenter/View/20147/City-of-Wilmington-Wage-Tax-Information-PDF
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        federal_agi = self.federal_agi if hasattr(self, 'federal_agi') else federal_agi
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Arizona', 'state_occupation': 'Delaware', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "federal_agi": federal_agi, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def arizona_and_florida(self, state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Arizona:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        
        Florida:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Arizona:
        - https://azdor.gov/forms/individual/form-140-x-y-tables
          https://azdor.gov/forms/individual/form-140-x-y-tables
          https://azdor.gov/forms/individual/form-140-x-y-tables
          https://azdor.gov/businesses-arizona/withholding-tax/withholding-exceptions
          "https://azdor.gov/individuals/income-tax-filing-assistance/identifying-other-taxable-income#:~:text=In%20regards%20to%20capital%20gains
         the%20individuals%20regular%20tax%20rate.
        
        Florida:
        -         
        - https://floridarevenue.com/taxes/taxesfees/Pages/tax_interest_rates.aspx
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Arizona', 'state_occupation': 'Florida', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def arizona_and_georgia(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Arizona:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        
        Georgia:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Arizona:
        - https://azdor.gov/forms/individual/form-140-x-y-tables
          https://azdor.gov/forms/individual/form-140-x-y-tables
          https://azdor.gov/forms/individual/form-140-x-y-tables
          https://azdor.gov/businesses-arizona/withholding-tax/withholding-exceptions
          "https://azdor.gov/individuals/income-tax-filing-assistance/identifying-other-taxable-income#:~:text=In%20regards%20to%20capital%20gains
         the%20individuals%20regular%20tax%20rate.
        
        Georgia:
        -         
        - https://dor.georgia.gov/tax-tables-georgia-tax-rate-schedule
          https://apps.dor.ga.gov/FillableForms/PDFViewer/Index?form=2022GA500
          https://apps.dor.ga.gov/FillableForms/PDFViewer/Index?form=2022GA500
          https://apps.dor.ga.gov/FillableForms/PDFViewer/Index?form=2022GA500
          https://dor.georgia.gov/years-individual-income-tax-forms
          https://dor.georgia.gov/years-individual-income-tax-forms
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Arizona', 'state_occupation': 'Georgia', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def arizona_and_hawaii(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Arizona:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        
        Hawaii:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Arizona:
        - https://azdor.gov/forms/individual/form-140-x-y-tables
          https://azdor.gov/forms/individual/form-140-x-y-tables
          https://azdor.gov/forms/individual/form-140-x-y-tables
          https://azdor.gov/businesses-arizona/withholding-tax/withholding-exceptions
          "https://azdor.gov/individuals/income-tax-filing-assistance/identifying-other-taxable-income#:~:text=In%20regards%20to%20capital%20gains
         the%20individuals%20regular%20tax%20rate.
        
        Hawaii:
        -         
        - https://tax.hawaii.gov/forms/d_18table-on/d_18table-on_p13/
          https://files.hawaii.gov/tax/forms/2018/18table-on.pdf
          https://files.hawaii.gov/tax/news/pubs/22outline.pdf
          https://files.hawaii.gov/tax/news/pubs/FastTaxRefGuide_2017.pdf
          https://www.tfhawaii.org/wordpress/state-tax-resources/mini-tax-guides/individual-income-tax-chapter-235/#:~:text=Long%20term%20capital%20gains%20are%20taxed%20at%20a%20maximum%20of%207.25%25
          https://www.grassrootinstitute.org/2023/02/hb337-increasing-tax-on-capital-gains-a-bad-risk/
          https://files.hawaii.gov/tax/news/pubs/22outline.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Arizona', 'state_occupation': 'Hawaii', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def arizona_and_idaho(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Arizona:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        
        Idaho:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Arizona:
        - https://azdor.gov/forms/individual/form-140-x-y-tables
          https://azdor.gov/forms/individual/form-140-x-y-tables
          https://azdor.gov/forms/individual/form-140-x-y-tables
          https://azdor.gov/businesses-arizona/withholding-tax/withholding-exceptions
          "https://azdor.gov/individuals/income-tax-filing-assistance/identifying-other-taxable-income#:~:text=In%20regards%20to%20capital%20gains
         the%20individuals%20regular%20tax%20rate.
        
        Idaho:
        -         
        - https://tax.idaho.gov/taxes/income-tax/individual-income/individual-income-tax-rate-schedule/
          https://tax.idaho.gov/wp-content/uploads/forms/EIN00046/EIN00046_03-01-2023.pdf
          https://tax.idaho.gov/taxes/income-tax/individual-income/idaho-source-income/
          https://tax.idaho.gov/taxes/income-tax/individual-income/idaho-source-income/
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Arizona', 'state_occupation': 'Idaho', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def arizona_and_illinois(self, state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Arizona:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        
        Illinois:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Arizona:
        - https://azdor.gov/forms/individual/form-140-x-y-tables
          https://azdor.gov/forms/individual/form-140-x-y-tables
          https://azdor.gov/forms/individual/form-140-x-y-tables
          https://azdor.gov/businesses-arizona/withholding-tax/withholding-exceptions
          "https://azdor.gov/individuals/income-tax-filing-assistance/identifying-other-taxable-income#:~:text=In%20regards%20to%20capital%20gains
         the%20individuals%20regular%20tax%20rate.
        
        Illinois:
        -         
        - https://tax.illinois.gov/research/taxrates/income.html#IndividualIncome
          https://tax.illinois.gov/content/dam/soi/en/web/tax/forms/incometax/documents/currentyear/individual/il-1040-instr.pdf
          https://tax.illinois.gov/questionsandanswers/12.html
          https://tax.illinois.gov/content/dam/soi/en/web/tax/forms/incometax/documents/currentyear/individual/il-1040-instr.pdf
          https://tax.illinois.gov/research/taxrates/personalproperty.html
          https://tax.illinois.gov/localgovernments/income.html
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Arizona', 'state_occupation': 'Illinois', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def arizona_and_indiana(self, state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Arizona:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        
        Indiana:
        


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Arizona:
        - https://azdor.gov/forms/individual/form-140-x-y-tables
          https://azdor.gov/forms/individual/form-140-x-y-tables
          https://azdor.gov/forms/individual/form-140-x-y-tables
          https://azdor.gov/businesses-arizona/withholding-tax/withholding-exceptions
          "https://azdor.gov/individuals/income-tax-filing-assistance/identifying-other-taxable-income#:~:text=In%20regards%20to%20capital%20gains
         the%20individuals%20regular%20tax%20rate.
        
        Indiana:
        -         
        - https://www.in.gov/dor/tax-forms/2022-individual-income-tax-forms/
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-32.htm
          https://www.in.gov/dor/files/ib28.pdf
          https://www.in.gov/dor/tax-forms/2022-individual-income-tax-forms/
          https://www.in.gov/dor/files/2022-county-tax-rates-and-codes.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Arizona', 'state_occupation': 'Indiana', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def arizona_and_iowa(self, county_residence='Other', state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Arizona:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        
        Iowa:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Arizona:
        - https://azdor.gov/forms/individual/form-140-x-y-tables
          https://azdor.gov/forms/individual/form-140-x-y-tables
          https://azdor.gov/forms/individual/form-140-x-y-tables
          https://azdor.gov/businesses-arizona/withholding-tax/withholding-exceptions
          "https://azdor.gov/individuals/income-tax-filing-assistance/identifying-other-taxable-income#:~:text=In%20regards%20to%20capital%20gains
         the%20individuals%20regular%20tax%20rate.
        
        Iowa:
        -         
        - https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
          https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
          https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
          https://tax.iowa.gov/iowa-illinois-reciprocal-agreement
          https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Arizona', 'state_occupation': 'Iowa', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def arizona_and_kansas(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Arizona:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        
        Kansas:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Arizona:
        - https://azdor.gov/forms/individual/form-140-x-y-tables
          https://azdor.gov/forms/individual/form-140-x-y-tables
          https://azdor.gov/forms/individual/form-140-x-y-tables
          https://azdor.gov/businesses-arizona/withholding-tax/withholding-exceptions
          "https://azdor.gov/individuals/income-tax-filing-assistance/identifying-other-taxable-income#:~:text=In%20regards%20to%20capital%20gains
         the%20individuals%20regular%20tax%20rate.
        
        Kansas:
        -         
        - https://www.ksrevenue.gov/taxrates.html
          https://www.ksrevenue.gov/incomebook22.html#0
          https://www.ksrevenue.gov/incomebook22.html#0
          https://www.ksrevenue.gov/incomebook22.html#0
          https://www.ksrevenue.gov/incomebook22.html
          https://www.ksrevenue.gov/incomebook22.html
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Arizona', 'state_occupation': 'Kansas', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def arizona_and_kentucky(self, county_residence='Other', state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Arizona:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        
        Kentucky:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Arizona:
        - https://azdor.gov/forms/individual/form-140-x-y-tables
          https://azdor.gov/forms/individual/form-140-x-y-tables
          https://azdor.gov/forms/individual/form-140-x-y-tables
          https://azdor.gov/businesses-arizona/withholding-tax/withholding-exceptions
          "https://azdor.gov/individuals/income-tax-filing-assistance/identifying-other-taxable-income#:~:text=In%20regards%20to%20capital%20gains
         the%20individuals%20regular%20tax%20rate.
        
        Kentucky:
        -         
        - https://revenue.ky.gov/News/Pages/Dor-Announces-Updates-to-Individual-Income-Tax-for-2023-Tax-Year.aspx
          https://revenue.ky.gov/Individual/Individual-Income-Tax/Pages/default.aspx
          https://revenue.ky.gov/News/Pages/Dor-Announces-Updates-to-Individual-Income-Tax-for-2023-Tax-Year.aspx
          https://revenue.ky.gov/Individual/Individual-Income-Tax/Pages/default.aspx
          https://revenue.ky.gov/Dor%20Training%20Materials/103%20KAR%2017.140.%20Individual%20income%20tax%20-%20reciprocity%20-%20nonresidents.pdf
          https://revenue.ky.gov/Individual/Individual-Income-Tax/Pages/default.aspx
          https://nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2021/Taxes-21-32.htm?taxmap=true
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Arizona', 'state_occupation': 'Kentucky', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def arizona_and_louisiana(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Arizona:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        
        Louisiana:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Arizona:
        - https://azdor.gov/forms/individual/form-140-x-y-tables
          https://azdor.gov/forms/individual/form-140-x-y-tables
          https://azdor.gov/forms/individual/form-140-x-y-tables
          https://azdor.gov/businesses-arizona/withholding-tax/withholding-exceptions
          "https://azdor.gov/individuals/income-tax-filing-assistance/identifying-other-taxable-income#:~:text=In%20regards%20to%20capital%20gains
         the%20individuals%20regular%20tax%20rate.
        
        Louisiana:
        -         
        - https://revenue.louisiana.gov/individualincometax
          https://revenue.louisiana.gov/individualincometax
          https://www.revenue.louisiana.gov/taxforms/6935(11_02)F.pdf
          https://revenue.louisiana.gov/individualincometax
          https://revenue.louisiana.gov/individualincometax
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Arizona', 'state_occupation': 'Louisiana', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def arizona_and_maine(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Arizona:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        
        Maine:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Arizona:
        - https://azdor.gov/forms/individual/form-140-x-y-tables
          https://azdor.gov/forms/individual/form-140-x-y-tables
          https://azdor.gov/forms/individual/form-140-x-y-tables
          https://azdor.gov/businesses-arizona/withholding-tax/withholding-exceptions
          "https://azdor.gov/individuals/income-tax-filing-assistance/identifying-other-taxable-income#:~:text=In%20regards%20to%20capital%20gains
         the%20individuals%20regular%20tax%20rate.
        
        Maine:
        -         
        - https://www.maine.gov/revenue/tax-return-forms/individual-income-tax-2022
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/ind_tax_rate_sched_2022.pdf
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/22_1040me_book_gen_instruc_revisedFeb23.pdf
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/22_1040me_book_gen_instruc_revisedFeb23.pdf
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/22_1040me_book_gen_instruc_revisedFeb23.pdf
          https://www.maine.gov/revenue/tax-return-forms/individual-income-tax-2022
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Arizona', 'state_occupation': 'Maine', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def arizona_and_maryland(self, county_residence='Other', state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Arizona:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        
        Maryland:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Arizona:
        - https://azdor.gov/forms/individual/form-140-x-y-tables
          https://azdor.gov/forms/individual/form-140-x-y-tables
          https://azdor.gov/forms/individual/form-140-x-y-tables
          https://azdor.gov/businesses-arizona/withholding-tax/withholding-exceptions
          "https://azdor.gov/individuals/income-tax-filing-assistance/identifying-other-taxable-income#:~:text=In%20regards%20to%20capital%20gains
         the%20individuals%20regular%20tax%20rate.
        
        Maryland:
        -         
        - https://www.marylandtaxes.gov/individual/income/tax-info/tax-rates.php
          https://www.marylandtaxes.gov/forms/22_forms/Resident_Booklet.pdf
          https://www.marylandtaxes.gov/forms/Tax_Publications/Administrative_Releases/Income_and_Estate_Tax_Releases/ar_it37.pdf
          https://www.marylandtaxes.gov/individual/index.php
          https://www.marylandtaxes.gov/individual/income/tax-info/tax-rates.php
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Arizona', 'state_occupation': 'Maryland', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def arizona_and_massachusetts(self, state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Arizona:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        
        Massachusetts:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Arizona:
        - https://azdor.gov/forms/individual/form-140-x-y-tables
          https://azdor.gov/forms/individual/form-140-x-y-tables
          https://azdor.gov/forms/individual/form-140-x-y-tables
          https://azdor.gov/businesses-arizona/withholding-tax/withholding-exceptions
          "https://azdor.gov/individuals/income-tax-filing-assistance/identifying-other-taxable-income#:~:text=In%20regards%20to%20capital%20gains
         the%20individuals%20regular%20tax%20rate.
        
        Massachusetts:
        -         
        - https://www.mass.gov/service-details/massachusetts-tax-rates
          https://www.mass.gov/service-details/view-massachusetts-personal-income-tax-exemptions
          https://www.mass.gov/service-details/view-massachusetts-personal-income-tax-exemptions
          https://www.mass.gov/service-details/view-child-and-dependent-related-credits
          https://www.mass.gov/service-details/massachusetts-tax-rates
          https://www.mass.gov/personal-income-tax
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Arizona', 'state_occupation': 'Massachusetts', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def arizona_and_michigan(self, county_occupation='Other', state_agi=75000, county_residence='Other', capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Arizona:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        
        Michigan:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        county_occupation : str, optional
            County of occupation. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Arizona:
        - https://azdor.gov/forms/individual/form-140-x-y-tables
          https://azdor.gov/forms/individual/form-140-x-y-tables
          https://azdor.gov/forms/individual/form-140-x-y-tables
          https://azdor.gov/businesses-arizona/withholding-tax/withholding-exceptions
          "https://azdor.gov/individuals/income-tax-filing-assistance/identifying-other-taxable-income#:~:text=In%20regards%20to%20capital%20gains
         the%20individuals%20regular%20tax%20rate.
        
        Michigan:
        -         
        - https://www.michigan.gov/taxes/iit/new-developments-for-tax-year-2022
          "https://www.michigan.gov/taxes/-/media/Project/Websites/taxes/Forms/2023/2023-SUW/446_Withholding-Guide_2023.pdf?rev=177ca3d8cb034392bf84e317ed78c320&hash=19C5F42F43A89A91863D229349F48883#:~:text=The%20exemption%20amount%20is%20%245%2C400
         employees%20Michigan%20income%20tax%20return"
          "https://www.michigan.gov/taxes/-/media/Project/Websites/taxes/Forms/2023/2023-SUW/446_Withholding-Guide_2023.pdf?rev=177ca3d8cb034392bf84e317ed78c320&hash=19C5F42F43A89A91863D229349F48883#:~:text=The%20exemption%20amount%20is%20%245%2C400
         employees%20Michigan%20income%20tax%20return"
          https://www.michigan.gov/taxes/questions/iit/accordion/residency/are-my-wages-earned-in-another-state-taxable-in-michigan-if-i-am-a-michigan-resident-1
          https://www.michigan.gov/taxes/-/media/Project/Websites/taxes/Forms/2022/2022-IIT-Forms/BOOK_MI-1040.pdf?rev=4524f1cbbc0243b194ab05a6680be75b&hash=40DFE71E651F7C1D65C31F6CD03D944F
          https://www.michigan.gov/taxes/questions/iit/accordion/general/what-cities-impose-an-income-tax
        """
        
        # Check if the parameters are set upstream in the instance
        county_occupation = self.county_occupation if hasattr(self, 'county_occupation') else county_occupation
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Arizona', 'state_occupation': 'Michigan', "county_occupation": county_occupation, "state_agi": state_agi, "county_residence": county_residence, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def arizona_and_minnesota(self, state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Arizona:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        
        Minnesota:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Arizona:
        - https://azdor.gov/forms/individual/form-140-x-y-tables
          https://azdor.gov/forms/individual/form-140-x-y-tables
          https://azdor.gov/forms/individual/form-140-x-y-tables
          https://azdor.gov/businesses-arizona/withholding-tax/withholding-exceptions
          "https://azdor.gov/individuals/income-tax-filing-assistance/identifying-other-taxable-income#:~:text=In%20regards%20to%20capital%20gains
         the%20individuals%20regular%20tax%20rate.
        
        Minnesota:
        -         
        - https://www.revenue.state.mn.us/minnesota-income-tax-rates-and-brackets
          https://www.revenue.state.mn.us/child-and-dependent-care-credit
          https://www.revenue.state.mn.us/dependent-exemptions
          https://www.revenue.state.mn.us/minnesota-standard-deduction
          https://mn.gov/mmb/assets/mwr_form_tcm1059-128581.pdf
          https://www.revenue.state.mn.us/individual-income-tax
          https://www.revenue.state.mn.us/individual-income-tax
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Arizona', 'state_occupation': 'Minnesota', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def arizona_and_mississippi(self, state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Arizona:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        
        Mississippi:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Arizona:
        - https://azdor.gov/forms/individual/form-140-x-y-tables
          https://azdor.gov/forms/individual/form-140-x-y-tables
          https://azdor.gov/forms/individual/form-140-x-y-tables
          https://azdor.gov/businesses-arizona/withholding-tax/withholding-exceptions
          "https://azdor.gov/individuals/income-tax-filing-assistance/identifying-other-taxable-income#:~:text=In%20regards%20to%20capital%20gains
         the%20individuals%20regular%20tax%20rate.
        
        Mississippi:
        -         
        - https://www.dor.ms.gov/individual/tax-rates
          https://www.dor.ms.gov/individual/tax-rates
          https://www.dor.ms.gov/individual/tax-rates
          https://www.dor.ms.gov/individual
          https://www.dor.ms.gov/individual
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Arizona', 'state_occupation': 'Mississippi', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def arizona_and_missouri(self, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Arizona:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        
        Missouri:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Arizona:
        - https://azdor.gov/forms/individual/form-140-x-y-tables
          https://azdor.gov/forms/individual/form-140-x-y-tables
          https://azdor.gov/forms/individual/form-140-x-y-tables
          https://azdor.gov/businesses-arizona/withholding-tax/withholding-exceptions
          "https://azdor.gov/individuals/income-tax-filing-assistance/identifying-other-taxable-income#:~:text=In%20regards%20to%20capital%20gains
         the%20individuals%20regular%20tax%20rate.
        
        Missouri:
        -         
        - https://dor.mo.gov/taxation/individual/tax-types/income/year-changes/
          https://dor.mo.gov/taxation/individual/tax-types/income/year-changes/
          https://dor.mo.gov/taxation/individual/tax-types/income/year-changes/
          https://www.kcmo.gov/city-hall/departments/finance/earnings-tax
          https://www.stlouis-mo.gov/government/departments/collector/earnings-tax/individual-earnings-tax-info.cfm
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Arizona', 'state_occupation': 'Missouri', "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def arizona_and_montana(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Arizona:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        
        Montana:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Arizona:
        - https://azdor.gov/forms/individual/form-140-x-y-tables
          https://azdor.gov/forms/individual/form-140-x-y-tables
          https://azdor.gov/forms/individual/form-140-x-y-tables
          https://azdor.gov/businesses-arizona/withholding-tax/withholding-exceptions
          "https://azdor.gov/individuals/income-tax-filing-assistance/identifying-other-taxable-income#:~:text=In%20regards%20to%20capital%20gains
         the%20individuals%20regular%20tax%20rate.
        
        Montana:
        -         
        - https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/2022/12/montana-employees-withholding-allowance-and-exemption-certificate-form-mw-4.pdf
          https://mtrevenue.gov/2021/11/29/simplification-of-montana-income-taxation/?utm_source=rss&utm_medium=rss&utm_campaign=simplification-of-montana-income-taxation
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Arizona', 'state_occupation': 'Montana', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def arizona_and_nebraska(self, state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Arizona:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        
        Nebraska:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Arizona:
        - https://azdor.gov/forms/individual/form-140-x-y-tables
          https://azdor.gov/forms/individual/form-140-x-y-tables
          https://azdor.gov/forms/individual/form-140-x-y-tables
          https://azdor.gov/businesses-arizona/withholding-tax/withholding-exceptions
          "https://azdor.gov/individuals/income-tax-filing-assistance/identifying-other-taxable-income#:~:text=In%20regards%20to%20capital%20gains
         the%20individuals%20regular%20tax%20rate.
        
        Nebraska:
        -         
        - https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2023_Circular_EN_8-429r2023_Whole_Book_11.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Arizona', 'state_occupation': 'Nebraska', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def arizona_and_nevada(self, state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Arizona:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        
        Nevada:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Arizona:
        - https://azdor.gov/forms/individual/form-140-x-y-tables
          https://azdor.gov/forms/individual/form-140-x-y-tables
          https://azdor.gov/forms/individual/form-140-x-y-tables
          https://azdor.gov/businesses-arizona/withholding-tax/withholding-exceptions
          "https://azdor.gov/individuals/income-tax-filing-assistance/identifying-other-taxable-income#:~:text=In%20regards%20to%20capital%20gains
         the%20individuals%20regular%20tax%20rate.
        
        Nevada:
        -         
        - https://www.nvsos.gov/sos/businesses/the-nevada-advantage
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Arizona', 'state_occupation': 'Nevada', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def arizona_and_new_hampshire(self, state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Arizona:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        
        New Hampshire:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Arizona:
        - https://azdor.gov/forms/individual/form-140-x-y-tables
          https://azdor.gov/forms/individual/form-140-x-y-tables
          https://azdor.gov/forms/individual/form-140-x-y-tables
          https://azdor.gov/businesses-arizona/withholding-tax/withholding-exceptions
          "https://azdor.gov/individuals/income-tax-filing-assistance/identifying-other-taxable-income#:~:text=In%20regards%20to%20capital%20gains
         the%20individuals%20regular%20tax%20rate.
        
        New Hampshire:
        -         
        - https://www.revenue.nh.gov/assistance/tax-overview.htm
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Arizona', 'state_occupation': 'NewHampshire', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def arizona_and_new_jersey(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Arizona:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        
        New Jersey:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Arizona:
        - https://azdor.gov/forms/individual/form-140-x-y-tables
          https://azdor.gov/forms/individual/form-140-x-y-tables
          https://azdor.gov/forms/individual/form-140-x-y-tables
          https://azdor.gov/businesses-arizona/withholding-tax/withholding-exceptions
          "https://azdor.gov/individuals/income-tax-filing-assistance/identifying-other-taxable-income#:~:text=In%20regards%20to%20capital%20gains
         the%20individuals%20regular%20tax%20rate.
        
        New Jersey:
        -         
        - https://www.state.nj.us/treasury/taxation/taxtables.shtml
          https://www.state.nj.us/treasury/taxation/njit13.shtml
          https://www.nj.gov/treasury/taxation/njit25.shtml
          https://www.state.nj.us/treasury/taxation/individuals/freqqiti.shtml
          https://www.state.nj.us/treasury/taxation/localtax.shtml
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Arizona', 'state_occupation': 'NewJersey', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def arizona_and_new_mexico(self, state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Arizona:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        
        New Mexico:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Arizona:
        - https://azdor.gov/forms/individual/form-140-x-y-tables
          https://azdor.gov/forms/individual/form-140-x-y-tables
          https://azdor.gov/forms/individual/form-140-x-y-tables
          https://azdor.gov/businesses-arizona/withholding-tax/withholding-exceptions
          "https://azdor.gov/individuals/income-tax-filing-assistance/identifying-other-taxable-income#:~:text=In%20regards%20to%20capital%20gains
         the%20individuals%20regular%20tax%20rate.
        
        New Mexico:
        -         
        - https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/fdf3c548-8aba-4b9c-9eb4-bb564c716015/FYI-104.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/0bf943fd-652e-4400-bb1b-cd8397eb5a95/2022pit-adj.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/2f1a6781-9534-4436-b427-1557f9592099/2022pit-adj-ins.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/0bf943fd-652e-4400-bb1b-cd8397eb5a95/2022pit-adj.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/1afc56af-ea90-4d48-82e5-1f9aeb43255a/PITbook2022.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/fdf3c548-8aba-4b9c-9eb4-bb564c716015/FYI-104.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Arizona', 'state_occupation': 'NewMexico', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def arizona_and_new_york(self, county_residence='Other', state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Arizona:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        
        New York:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Arizona:
        - https://azdor.gov/forms/individual/form-140-x-y-tables
          https://azdor.gov/forms/individual/form-140-x-y-tables
          https://azdor.gov/forms/individual/form-140-x-y-tables
          https://azdor.gov/businesses-arizona/withholding-tax/withholding-exceptions
          "https://azdor.gov/individuals/income-tax-filing-assistance/identifying-other-taxable-income#:~:text=In%20regards%20to%20capital%20gains
         the%20individuals%20regular%20tax%20rate.
        
        New York:
        -         
        - https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nys-tax-rate-schedule
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nys-tax-rate-schedule
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nys-tax-rate-schedule
          https://www.tax.ny.gov/pit/
          https://www.tax.ny.gov/pit/file/nonresident-faqs.htm
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#yonkers-tax-rate-schedule
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nyc-tax-rate-schedule
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Arizona', 'state_occupation': 'NewYork', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def arizona_and_north_carolina(self, state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Arizona:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        
        North Carolina:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Arizona:
        - https://azdor.gov/forms/individual/form-140-x-y-tables
          https://azdor.gov/forms/individual/form-140-x-y-tables
          https://azdor.gov/forms/individual/form-140-x-y-tables
          https://azdor.gov/businesses-arizona/withholding-tax/withholding-exceptions
          "https://azdor.gov/individuals/income-tax-filing-assistance/identifying-other-taxable-income#:~:text=In%20regards%20to%20capital%20gains
         the%20individuals%20regular%20tax%20rate.
        
        North Carolina:
        -         
        - https://www.ncdor.gov/taxes-forms/tax-rate-schedules
          https://www.ncdor.gov/taxes-forms/individual-income-tax/north-carolina-standard-deduction-or-north-carolina-itemized-deductions
          https://www.ncdor.gov/taxes-forms/individual-income-tax
          https://www.ncdor.gov/taxes-forms/individual-income-tax
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Arizona', 'state_occupation': 'NorthCarolina', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def arizona_and_north_dakota(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Arizona:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        
        North Dakota:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Arizona:
        - https://azdor.gov/forms/individual/form-140-x-y-tables
          https://azdor.gov/forms/individual/form-140-x-y-tables
          https://azdor.gov/forms/individual/form-140-x-y-tables
          https://azdor.gov/businesses-arizona/withholding-tax/withholding-exceptions
          "https://azdor.gov/individuals/income-tax-filing-assistance/identifying-other-taxable-income#:~:text=In%20regards%20to%20capital%20gains
         the%20individuals%20regular%20tax%20rate.
        
        North Dakota:
        -         
        - https://www.tax.nd.gov/forms
          https://www.tax.nd.gov/sites/www/files/documents/forms/individual/2022-iit/2022-individual-income-tax-booklet.pdf
          https://www.tax.nd.gov/sites/www/files/documents/forms/2022-individual-income-tax-booklet.pdf
          https://www.tax.nd.gov/tax-exemptions-credits/income-tax-exemptions-credits
          https://www.tax.nd.gov/sites/www/files/documents/forms/business/ndwrfillable.pdf
          https://www.tax.nd.gov/news/tax-legislative-changes/significant-changes-law/individual-income-tax-history
          https://www.tax.nd.gov/forms
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Arizona', 'state_occupation': 'NorthDakota', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def arizona_and_ohio(self, county_residence='Other', state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Arizona:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        
        Ohio:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Arizona:
        - https://azdor.gov/forms/individual/form-140-x-y-tables
          https://azdor.gov/forms/individual/form-140-x-y-tables
          https://azdor.gov/forms/individual/form-140-x-y-tables
          https://azdor.gov/businesses-arizona/withholding-tax/withholding-exceptions
          "https://azdor.gov/individuals/income-tax-filing-assistance/identifying-other-taxable-income#:~:text=In%20regards%20to%20capital%20gains
         the%20individuals%20regular%20tax%20rate.
        
        Ohio:
        -         
        - https://tax.ohio.gov/individual/resources/annual-tax-rates
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://tax.ohio.gov/static/forms/employer_withholding/generic/wth_it4nr.pdf
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://www.ritaohio.com/TaxRatesTable
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Arizona', 'state_occupation': 'Ohio', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def arizona_and_oklahoma(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Arizona:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        
        Oklahoma:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Arizona:
        - https://azdor.gov/forms/individual/form-140-x-y-tables
          https://azdor.gov/forms/individual/form-140-x-y-tables
          https://azdor.gov/forms/individual/form-140-x-y-tables
          https://azdor.gov/businesses-arizona/withholding-tax/withholding-exceptions
          "https://azdor.gov/individuals/income-tax-filing-assistance/identifying-other-taxable-income#:~:text=In%20regards%20to%20capital%20gains
         the%20individuals%20regular%20tax%20rate.
        
        Oklahoma:
        -         
        - https://oklahoma.gov/content/dam/ok/en/tax/documents/resources/publications/businesses/withholding-tables/WHTables-2022.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Arizona', 'state_occupation': 'Oklahoma', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def arizona_and_oregon(self, state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Arizona:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        
        Oregon:
        


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Arizona:
        - https://azdor.gov/forms/individual/form-140-x-y-tables
          https://azdor.gov/forms/individual/form-140-x-y-tables
          https://azdor.gov/forms/individual/form-140-x-y-tables
          https://azdor.gov/businesses-arizona/withholding-tax/withholding-exceptions
          "https://azdor.gov/individuals/income-tax-filing-assistance/identifying-other-taxable-income#:~:text=In%20regards%20to%20capital%20gains
         the%20individuals%20regular%20tax%20rate.
        
        Oregon:
        -         
        - https://www.oregon.gov/dor/programs/individuals/Pages/PIT.aspx
          https://www.oregon.gov/dor/programs/individuals/Documents/Full%20year%20resident
         %20Form%20or-40%20filers.pdf
          https://www.oregon.gov/dor/programs/individuals/Documents/Part-year%20and%20nonresident
         %20Form%20or-40-P%20and%20or-40-N%20filers.pdf.pdf
          https://www.oregon.gov/dor/forms/FormsPubs/form-or-40_101-040_2022.pdf
          https://www.oregon.gov/dor/forms/FormsPubs/form-or-40_101-040_2022.pdf
          https://www.oregon.gov/dor/forms/FormsPubs/form-or-40_101-040_2022.pdf
          https://www.portland.gov/revenue/personal-tax
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Arizona', 'state_occupation': 'Oregon', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def arizona_and_pennsylvania(self, county_residence='Other', state_agi=75000, county_occupation='Other', capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Arizona:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        
        Pennsylvania:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        county_residence : str, optional
            County of residence. Default is Other.
        county_occupation : str, optional
            County of occupation. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Arizona:
        - https://azdor.gov/forms/individual/form-140-x-y-tables
          https://azdor.gov/forms/individual/form-140-x-y-tables
          https://azdor.gov/forms/individual/form-140-x-y-tables
          https://azdor.gov/businesses-arizona/withholding-tax/withholding-exceptions
          "https://azdor.gov/individuals/income-tax-filing-assistance/identifying-other-taxable-income#:~:text=In%20regards%20to%20capital%20gains
         the%20individuals%20regular%20tax%20rate.
        
        Pennsylvania:
        -         
        - https://www.revenue.pa.gov/Tax%20Rates/Pages/default.aspx
          https://www.revenue.pa.gov/FormsandPublications/FormsforIndividuals/PIT/Documents/2022/2022_pa-40in.pdf
          https://www.revenue.pa.gov/FormsandPublications/FormsforBusinesses/EmployerWithholding/Documents/rev-419.pdf
          https://www.revenue.pa.gov/FormsandPublications/FormsforIndividuals/PIT/Documents/2022/2022_pa-40in.pdf
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-25.htm
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        county_occupation = self.county_occupation if hasattr(self, 'county_occupation') else county_occupation
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Arizona', 'state_occupation': 'Pennsylvania', "county_residence": county_residence, "state_agi": state_agi, "county_occupation": county_occupation, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def arizona_and_rhode_island(self, state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Arizona:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        
        Rhode Island:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Arizona:
        - https://azdor.gov/forms/individual/form-140-x-y-tables
          https://azdor.gov/forms/individual/form-140-x-y-tables
          https://azdor.gov/forms/individual/form-140-x-y-tables
          https://azdor.gov/businesses-arizona/withholding-tax/withholding-exceptions
          "https://azdor.gov/individuals/income-tax-filing-assistance/identifying-other-taxable-income#:~:text=In%20regards%20to%20capital%20gains
         the%20individuals%20regular%20tax%20rate.
        
        Rhode Island:
        -         
        - https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2022_1040WE_w_0.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Arizona', 'state_occupation': 'RhodeIsland', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def arizona_and_south_carolina(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Arizona:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        
        South Carolina:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Arizona:
        - https://azdor.gov/forms/individual/form-140-x-y-tables
          https://azdor.gov/forms/individual/form-140-x-y-tables
          https://azdor.gov/forms/individual/form-140-x-y-tables
          https://azdor.gov/businesses-arizona/withholding-tax/withholding-exceptions
          "https://azdor.gov/individuals/income-tax-filing-assistance/identifying-other-taxable-income#:~:text=In%20regards%20to%20capital%20gains
         the%20individuals%20regular%20tax%20rate.
        
        South Carolina:
        -         
        - https://dor.sc.gov/resources-site/lawandpolicy/Advisory%20Opinions/IL22-15.pdf
          https://dor.sc.gov/forms-site/Forms/IITPacket_2022.pdf
          https://www.scstatehouse.gov/code/t12c006.php
          https://dor.sc.gov/resources-site/lawandpolicy/Advisory%20Opinions/IL22-15.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Arizona', 'state_occupation': 'SouthCarolina', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def arizona_and_south_dakota(self, state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Arizona:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        
        South Dakota:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Arizona:
        - https://azdor.gov/forms/individual/form-140-x-y-tables
          https://azdor.gov/forms/individual/form-140-x-y-tables
          https://azdor.gov/forms/individual/form-140-x-y-tables
          https://azdor.gov/businesses-arizona/withholding-tax/withholding-exceptions
          "https://azdor.gov/individuals/income-tax-filing-assistance/identifying-other-taxable-income#:~:text=In%20regards%20to%20capital%20gains
         the%20individuals%20regular%20tax%20rate.
        
        South Dakota:
        -         
        - https://dor.sd.gov/individuals/taxes/
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Arizona', 'state_occupation': 'SouthDakota', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def arizona_and_tennessee(self, state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Arizona:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        
        Tennessee:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Arizona:
        - https://azdor.gov/forms/individual/form-140-x-y-tables
          https://azdor.gov/forms/individual/form-140-x-y-tables
          https://azdor.gov/forms/individual/form-140-x-y-tables
          https://azdor.gov/businesses-arizona/withholding-tax/withholding-exceptions
          "https://azdor.gov/individuals/income-tax-filing-assistance/identifying-other-taxable-income#:~:text=In%20regards%20to%20capital%20gains
         the%20individuals%20regular%20tax%20rate.
        
        Tennessee:
        -         
        - https://www.tn.gov/revenue/taxes.html
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Arizona', 'state_occupation': 'Tennessee', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def arizona_and_texas(self, state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Arizona:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        
        Texas:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Arizona:
        - https://azdor.gov/forms/individual/form-140-x-y-tables
          https://azdor.gov/forms/individual/form-140-x-y-tables
          https://azdor.gov/forms/individual/form-140-x-y-tables
          https://azdor.gov/businesses-arizona/withholding-tax/withholding-exceptions
          "https://azdor.gov/individuals/income-tax-filing-assistance/identifying-other-taxable-income#:~:text=In%20regards%20to%20capital%20gains
         the%20individuals%20regular%20tax%20rate.
        
        Texas:
        -         
        - https://comptroller.texas.gov/taxes/
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Arizona', 'state_occupation': 'Texas', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def arizona_and_utah(self, state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Arizona:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        
        Utah:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Arizona:
        - https://azdor.gov/forms/individual/form-140-x-y-tables
          https://azdor.gov/forms/individual/form-140-x-y-tables
          https://azdor.gov/forms/individual/form-140-x-y-tables
          https://azdor.gov/businesses-arizona/withholding-tax/withholding-exceptions
          "https://azdor.gov/individuals/income-tax-filing-assistance/identifying-other-taxable-income#:~:text=In%20regards%20to%20capital%20gains
         the%20individuals%20regular%20tax%20rate.
        
        Utah:
        -         
        - https://incometax.utah.gov/paying/tax-rates
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Arizona', 'state_occupation': 'Utah', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def arizona_and_vermont(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Arizona:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        
        Vermont:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Arizona:
        - https://azdor.gov/forms/individual/form-140-x-y-tables
          https://azdor.gov/forms/individual/form-140-x-y-tables
          https://azdor.gov/forms/individual/form-140-x-y-tables
          https://azdor.gov/businesses-arizona/withholding-tax/withholding-exceptions
          "https://azdor.gov/individuals/income-tax-filing-assistance/identifying-other-taxable-income#:~:text=In%20regards%20to%20capital%20gains
         the%20individuals%20regular%20tax%20rate.
        
        Vermont:
        -         
        - https://tax.vermont.gov/sites/tax/files/documents/GB-1210-2023.pdf
          https://tax.vermont.gov/sites/tax/files/documents/Income%20Booklet-2022.pdf
          https://tax.vermont.gov/sites/tax/files/documents/Income%20Booklet-2022.pdf
          https://tax.vermont.gov/sites/tax/files/documents/Income%20Booklet-2022.pdf
          https://tax.vermont.gov/individuals/personal-income-tax/tax-credits
          https://tax.vermont.gov/individuals/personal-income-tax/taxable-income
          https://tax.vermont.gov/individuals/personal-income-tax/local-option-tax
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Arizona', 'state_occupation': 'Vermont', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def arizona_and_virginia(self, state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Arizona:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        
        Virginia:
        


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Arizona:
        - https://azdor.gov/forms/individual/form-140-x-y-tables
          https://azdor.gov/forms/individual/form-140-x-y-tables
          https://azdor.gov/forms/individual/form-140-x-y-tables
          https://azdor.gov/businesses-arizona/withholding-tax/withholding-exceptions
          "https://azdor.gov/individuals/income-tax-filing-assistance/identifying-other-taxable-income#:~:text=In%20regards%20to%20capital%20gains
         the%20individuals%20regular%20tax%20rate.
        
        Virginia:
        -         
        - https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/reciprocity
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Arizona', 'state_occupation': 'Virginia', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def arizona_and_washington(self, state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Arizona:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        
        Washington:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Arizona:
        - https://azdor.gov/forms/individual/form-140-x-y-tables
          https://azdor.gov/forms/individual/form-140-x-y-tables
          https://azdor.gov/forms/individual/form-140-x-y-tables
          https://azdor.gov/businesses-arizona/withholding-tax/withholding-exceptions
          "https://azdor.gov/individuals/income-tax-filing-assistance/identifying-other-taxable-income#:~:text=In%20regards%20to%20capital%20gains
         the%20individuals%20regular%20tax%20rate.
        
        Washington:
        -         
        - https://dor.wa.gov/taxes-rates/income-tax
          https://dor.wa.gov/taxes-rates/other-taxes/capital-gains-tax
          https://dor.wa.gov/taxes-rates/income-tax
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Arizona', 'state_occupation': 'Washington', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def arizona_and_west_virginia(self, county_occupation='Other', state_agi=75000, pay_periods=24, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Arizona:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        
        West Virginia:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_occupation : str, optional
            County of occupation. Default is Other.
        pay_periods : int, optional
            Number of pay periods. Options are 13, 24, 26, 52. Default is 24.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Arizona:
        - https://azdor.gov/forms/individual/form-140-x-y-tables
          https://azdor.gov/forms/individual/form-140-x-y-tables
          https://azdor.gov/forms/individual/form-140-x-y-tables
          https://azdor.gov/businesses-arizona/withholding-tax/withholding-exceptions
          "https://azdor.gov/individuals/income-tax-filing-assistance/identifying-other-taxable-income#:~:text=In%20regards%20to%20capital%20gains
         the%20individuals%20regular%20tax%20rate.
        
        West Virginia:
        -         
        - https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://tax.wv.gov/Documents/TSD/tsd438.pdf
          https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2011/Taxes-11-35att.html
        """
        
        # Check if the parameters are set upstream in the instance
        county_occupation = self.county_occupation if hasattr(self, 'county_occupation') else county_occupation
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        pay_periods = self.pay_periods if hasattr(self, 'pay_periods') else pay_periods
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Arizona', 'state_occupation': 'WestVirginia', "county_occupation": county_occupation, "state_agi": state_agi, "pay_periods": pay_periods, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def arizona_and_wisconsin(self, state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Arizona:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        
        Wisconsin:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Arizona:
        - https://azdor.gov/forms/individual/form-140-x-y-tables
          https://azdor.gov/forms/individual/form-140-x-y-tables
          https://azdor.gov/forms/individual/form-140-x-y-tables
          https://azdor.gov/businesses-arizona/withholding-tax/withholding-exceptions
          "https://azdor.gov/individuals/income-tax-filing-assistance/identifying-other-taxable-income#:~:text=In%20regards%20to%20capital%20gains
         the%20individuals%20regular%20tax%20rate.
        
        Wisconsin:
        -         
        - https://www.revenue.wi.gov/TaxForms2023/2023-Form1-ES-Inst.pdf
          https://www.revenue.wi.gov/TaxForms2023/2023-Form1-ES-Inst.pdf
          https://www.revenue.wi.gov/TaxForms2023/2023-Form1-ES-Inst.pdf
          https://www.revenue.wi.gov/Pages/FAQS/pcs-work.aspx
          https://www.revenue.wi.gov/Dor%20Publications/pb103.pdf
          https://www.revenue.wi.gov/Pages/FAQS/pcs-work.aspx
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Arizona', 'state_occupation': 'Wisconsin', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def arizona_and_wyoming(self, state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Arizona:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        
        Wyoming:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Arizona:
        - https://azdor.gov/forms/individual/form-140-x-y-tables
          https://azdor.gov/forms/individual/form-140-x-y-tables
          https://azdor.gov/forms/individual/form-140-x-y-tables
          https://azdor.gov/businesses-arizona/withholding-tax/withholding-exceptions
          "https://azdor.gov/individuals/income-tax-filing-assistance/identifying-other-taxable-income#:~:text=In%20regards%20to%20capital%20gains
         the%20individuals%20regular%20tax%20rate.
        
        Wyoming:
        -         
        - https://revenue.wyo.gov/
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Arizona', 'state_occupation': 'Wyoming', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def arkansas_and_alabama(self, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Arkansas:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Alabama:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Arkansas:
        - https://www.arkansasedc.com/why-arkansas/business-climate/tax-structure/personal-income-tax
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_and_AR1000NR_Instructions.pdf
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_FullYearResidentIndividualIncomeTaxReturn.pdf
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_and_AR1000NR_Instructions.pdf
          https://www.arkansasedc.com/why-arkansas/business-climate/tax-structure/capital-gains-tax-reduction
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_and_AR1000NR_Instructions.pdf
        
        Alabama:
        -         
        - https://www.revenue.alabama.gov/ultraviewer/viewer/basic_viewer/index.html?form=2023/02/22f40.pdf
          https://www.revenue.alabama.gov/faqs/what-is-alabamas-individual-income-tax-rate/
          https://www.revenue.alabama.gov/ultraviewer/viewer/basic_viewer/index.html?form=2023/02/22f40.pdf
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-24.htm?taxmap=true
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-24.htm?taxmap=true
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-24.htm?taxmap=true
          https://www.revenue.alabama.gov/individual-corporate/income-to-be-reported-on-the-alabama-income-tax-return/
          https://www.almonline.org/TaxRates.aspx#OccupationalTax
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Arkansas', 'state_occupation': 'Alabama', "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def arkansas_and_alaska(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Arkansas:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Alaska:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Arkansas:
        - https://www.arkansasedc.com/why-arkansas/business-climate/tax-structure/personal-income-tax
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_and_AR1000NR_Instructions.pdf
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_FullYearResidentIndividualIncomeTaxReturn.pdf
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_and_AR1000NR_Instructions.pdf
          https://www.arkansasedc.com/why-arkansas/business-climate/tax-structure/capital-gains-tax-reduction
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_and_AR1000NR_Instructions.pdf
        
        Alaska:
        -         
        - https://www.commerce.alaska.gov/web/dcra/officeofthestateassessor/alaskataxfacts.aspx
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Arkansas', 'state_occupation': 'Alaska', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def arkansas_and_arizona(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Arkansas:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Arizona:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        credits_state : int, optional
            Credits for state taxes. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Arkansas:
        - https://www.arkansasedc.com/why-arkansas/business-climate/tax-structure/personal-income-tax
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_and_AR1000NR_Instructions.pdf
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_FullYearResidentIndividualIncomeTaxReturn.pdf
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_and_AR1000NR_Instructions.pdf
          https://www.arkansasedc.com/why-arkansas/business-climate/tax-structure/capital-gains-tax-reduction
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_and_AR1000NR_Instructions.pdf
        
        Arizona:
        -         
        - https://azdor.gov/forms/individual/form-140-x-y-tables
          https://azdor.gov/forms/individual/form-140-x-y-tables
          https://azdor.gov/forms/individual/form-140-x-y-tables
          https://azdor.gov/businesses-arizona/withholding-tax/withholding-exceptions
          "https://azdor.gov/individuals/income-tax-filing-assistance/identifying-other-taxable-income#:~:text=In%20regards%20to%20capital%20gains
         the%20individuals%20regular%20tax%20rate.
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Arkansas', 'state_occupation': 'Arizona', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def arkansas_and_california(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Arkansas:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        California:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Arkansas:
        - https://www.arkansasedc.com/why-arkansas/business-climate/tax-structure/personal-income-tax
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_and_AR1000NR_Instructions.pdf
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_FullYearResidentIndividualIncomeTaxReturn.pdf
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_and_AR1000NR_Instructions.pdf
          https://www.arkansasedc.com/why-arkansas/business-climate/tax-structure/capital-gains-tax-reduction
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_and_AR1000NR_Instructions.pdf
        
        California:
        -         
        - https://www.ftb.ca.gov/forms/2022/2022-540-tax-rate-schedules.pdf
          https://www.ftb.ca.gov/file/personal/deductions/index.html
          https://www.ftb.ca.gov/about-ftb/newsroom/caleitc/eligibility-and-credit-information.html
          https://www.ftb.ca.gov/file/personal/income-types/capital-gains-and-losses.html#:~:text=California%20does%20not%20have%20a
         are%20taxed%20as%20ordinary%20income
          https://www.ftb.ca.gov/file/personal/
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Arkansas', 'state_occupation': 'California', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def arkansas_and_colorado(self, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Arkansas:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Colorado:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Arkansas:
        - https://www.arkansasedc.com/why-arkansas/business-climate/tax-structure/personal-income-tax
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_and_AR1000NR_Instructions.pdf
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_FullYearResidentIndividualIncomeTaxReturn.pdf
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_and_AR1000NR_Instructions.pdf
          https://www.arkansasedc.com/why-arkansas/business-climate/tax-structure/capital-gains-tax-reduction
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_and_AR1000NR_Instructions.pdf
        
        Colorado:
        -         
        - https://leg.colorado.gov/agencies/legislative-council-staff/individual-income-tax%C2%A0
          https://leg.colorado.gov/agencies/legislative-council-staff/individual-income-tax%C2%A0
          https://tax.colorado.gov/sites/tax/files/documents/ITT_Capital_Gain_Subtraction_Feb_2022.pdf
          https://www.auroragov.org/business_services/taxes/occupational_privilege_tax
          https://www.denvergov.org/Government/Agencies-Departments-Offices/Agencies-Departments-Offices-Directory/Department-of-Finance/Our-Divisions/Treasury/Business-Tax-Information#:~:text=Occupational%20privilege%20tax
         -%28show%20below%29&text=Each%20taxable%20employee%20is%20liable
         month%20for%20each%20taxable%20employee
          https://www.glendale.co.us/355/Occupational-Privilege-Tax
          https://www.greenwoodvillage.com/1220/Occupational-Privilege-Tax-OPT
          https://ci.sheridan.co.us/288/Occupational-Privilege-Tax
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Arkansas', 'state_occupation': 'Colorado', "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def arkansas_and_connecticut(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Arkansas:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Connecticut:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Arkansas:
        - https://www.arkansasedc.com/why-arkansas/business-climate/tax-structure/personal-income-tax
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_and_AR1000NR_Instructions.pdf
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_FullYearResidentIndividualIncomeTaxReturn.pdf
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_and_AR1000NR_Instructions.pdf
          https://www.arkansasedc.com/why-arkansas/business-climate/tax-structure/capital-gains-tax-reduction
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_and_AR1000NR_Instructions.pdf
        
        Connecticut:
        -         
        - https://portal.ct.gov/-/media/DRS/Forms/2022/Income/2022-CT-1040-Instructions_1222.pdf
          https://www.cga.ct.gov/2022/rpt/pdf/2022-R-0108.pdf
          https://www.cga.ct.gov/current/pub/chap_229.htm#sec_12-703
          https://portal.ct.gov/-/media/DRS/Forms/2022/Income/2022-CT-1040-Instructions_1222.pdf
          "https://portal.ct.gov/DRS/Publications/TSSNs/TSSN-29#:~:text=An%20individuals%20net%20capital%20gains
         portion%20of%20Social%20Security%20benefits"
          "https://portal.ct.gov/DRS/Publications/TSSNs/TSSN-29#:~:text=An%20individuals%20net%20capital%20gains
         portion%20of%20Social%20Security%20benefits"
          https://portal.ct.gov/DRS/Individuals/Resident-Income-Tax/Tax-Information
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Arkansas', 'state_occupation': 'Connecticut', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def arkansas_and_delaware(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, federal_agi=75000, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Arkansas:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Delaware:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        federal_agi : int, optional
            Adjusted Gross Income (AGI) for federal taxes. Default is 0. Sources - https://www.irs.gov/e-file-providers/definition-of-adjusted-gross-income, https://www.irs.gov/newsroom/taxpayers-should-know-the-difference-between-standard-and-itemized-deductions
        income : int, optional
            Gross income. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Arkansas:
        - https://www.arkansasedc.com/why-arkansas/business-climate/tax-structure/personal-income-tax
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_and_AR1000NR_Instructions.pdf
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_FullYearResidentIndividualIncomeTaxReturn.pdf
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_and_AR1000NR_Instructions.pdf
          https://www.arkansasedc.com/why-arkansas/business-climate/tax-structure/capital-gains-tax-reduction
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_and_AR1000NR_Instructions.pdf
        
        Delaware:
        -         
        - https://revenue.delaware.gov/software-developer/tax-rate-changes/
          https://revenuefiles.delaware.gov/2022/PIT-RES_TY22_2022-01_PaperInteractive.pdf
          https://revenuefiles.delaware.gov/2022/PIT-RES_TY22_2022-02_Instructions.pdf
          https://revenue.delaware.gov/frequently-asked-questions/personal-income-tax-faqs/
          https://www.newcastlede.gov/DocumentCenter/View/20147/City-of-Wilmington-Wage-Tax-Information-PDF
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        federal_agi = self.federal_agi if hasattr(self, 'federal_agi') else federal_agi
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Arkansas', 'state_occupation': 'Delaware', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "federal_agi": federal_agi, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def arkansas_and_florida(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Arkansas:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Florida:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Arkansas:
        - https://www.arkansasedc.com/why-arkansas/business-climate/tax-structure/personal-income-tax
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_and_AR1000NR_Instructions.pdf
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_FullYearResidentIndividualIncomeTaxReturn.pdf
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_and_AR1000NR_Instructions.pdf
          https://www.arkansasedc.com/why-arkansas/business-climate/tax-structure/capital-gains-tax-reduction
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_and_AR1000NR_Instructions.pdf
        
        Florida:
        -         
        - https://floridarevenue.com/taxes/taxesfees/Pages/tax_interest_rates.aspx
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Arkansas', 'state_occupation': 'Florida', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def arkansas_and_georgia(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Arkansas:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Georgia:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Arkansas:
        - https://www.arkansasedc.com/why-arkansas/business-climate/tax-structure/personal-income-tax
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_and_AR1000NR_Instructions.pdf
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_FullYearResidentIndividualIncomeTaxReturn.pdf
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_and_AR1000NR_Instructions.pdf
          https://www.arkansasedc.com/why-arkansas/business-climate/tax-structure/capital-gains-tax-reduction
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_and_AR1000NR_Instructions.pdf
        
        Georgia:
        -         
        - https://dor.georgia.gov/tax-tables-georgia-tax-rate-schedule
          https://apps.dor.ga.gov/FillableForms/PDFViewer/Index?form=2022GA500
          https://apps.dor.ga.gov/FillableForms/PDFViewer/Index?form=2022GA500
          https://apps.dor.ga.gov/FillableForms/PDFViewer/Index?form=2022GA500
          https://dor.georgia.gov/years-individual-income-tax-forms
          https://dor.georgia.gov/years-individual-income-tax-forms
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Arkansas', 'state_occupation': 'Georgia', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def arkansas_and_hawaii(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Arkansas:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Hawaii:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Arkansas:
        - https://www.arkansasedc.com/why-arkansas/business-climate/tax-structure/personal-income-tax
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_and_AR1000NR_Instructions.pdf
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_FullYearResidentIndividualIncomeTaxReturn.pdf
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_and_AR1000NR_Instructions.pdf
          https://www.arkansasedc.com/why-arkansas/business-climate/tax-structure/capital-gains-tax-reduction
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_and_AR1000NR_Instructions.pdf
        
        Hawaii:
        -         
        - https://tax.hawaii.gov/forms/d_18table-on/d_18table-on_p13/
          https://files.hawaii.gov/tax/forms/2018/18table-on.pdf
          https://files.hawaii.gov/tax/news/pubs/22outline.pdf
          https://files.hawaii.gov/tax/news/pubs/FastTaxRefGuide_2017.pdf
          https://www.tfhawaii.org/wordpress/state-tax-resources/mini-tax-guides/individual-income-tax-chapter-235/#:~:text=Long%20term%20capital%20gains%20are%20taxed%20at%20a%20maximum%20of%207.25%25
          https://www.grassrootinstitute.org/2023/02/hb337-increasing-tax-on-capital-gains-a-bad-risk/
          https://files.hawaii.gov/tax/news/pubs/22outline.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Arkansas', 'state_occupation': 'Hawaii', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def arkansas_and_idaho(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Arkansas:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Idaho:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Arkansas:
        - https://www.arkansasedc.com/why-arkansas/business-climate/tax-structure/personal-income-tax
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_and_AR1000NR_Instructions.pdf
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_FullYearResidentIndividualIncomeTaxReturn.pdf
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_and_AR1000NR_Instructions.pdf
          https://www.arkansasedc.com/why-arkansas/business-climate/tax-structure/capital-gains-tax-reduction
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_and_AR1000NR_Instructions.pdf
        
        Idaho:
        -         
        - https://tax.idaho.gov/taxes/income-tax/individual-income/individual-income-tax-rate-schedule/
          https://tax.idaho.gov/wp-content/uploads/forms/EIN00046/EIN00046_03-01-2023.pdf
          https://tax.idaho.gov/taxes/income-tax/individual-income/idaho-source-income/
          https://tax.idaho.gov/taxes/income-tax/individual-income/idaho-source-income/
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Arkansas', 'state_occupation': 'Idaho', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def arkansas_and_illinois(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Arkansas:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Illinois:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Arkansas:
        - https://www.arkansasedc.com/why-arkansas/business-climate/tax-structure/personal-income-tax
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_and_AR1000NR_Instructions.pdf
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_FullYearResidentIndividualIncomeTaxReturn.pdf
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_and_AR1000NR_Instructions.pdf
          https://www.arkansasedc.com/why-arkansas/business-climate/tax-structure/capital-gains-tax-reduction
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_and_AR1000NR_Instructions.pdf
        
        Illinois:
        -         
        - https://tax.illinois.gov/research/taxrates/income.html#IndividualIncome
          https://tax.illinois.gov/content/dam/soi/en/web/tax/forms/incometax/documents/currentyear/individual/il-1040-instr.pdf
          https://tax.illinois.gov/questionsandanswers/12.html
          https://tax.illinois.gov/content/dam/soi/en/web/tax/forms/incometax/documents/currentyear/individual/il-1040-instr.pdf
          https://tax.illinois.gov/research/taxrates/personalproperty.html
          https://tax.illinois.gov/localgovernments/income.html
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Arkansas', 'state_occupation': 'Illinois', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def arkansas_and_indiana(self, capital_gains_short=0, dependents=0, sources=False, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_long=0, filing_status='single', income=100000, credits_state=0):
        """
        --- Parameters ----
        
        Arkansas:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Indiana:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Arkansas:
        - https://www.arkansasedc.com/why-arkansas/business-climate/tax-structure/personal-income-tax
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_and_AR1000NR_Instructions.pdf
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_FullYearResidentIndividualIncomeTaxReturn.pdf
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_and_AR1000NR_Instructions.pdf
          https://www.arkansasedc.com/why-arkansas/business-climate/tax-structure/capital-gains-tax-reduction
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_and_AR1000NR_Instructions.pdf
        
        Indiana:
        -         
        - https://www.in.gov/dor/tax-forms/2022-individual-income-tax-forms/
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-32.htm
          https://www.in.gov/dor/files/ib28.pdf
          https://www.in.gov/dor/tax-forms/2022-individual-income-tax-forms/
          https://www.in.gov/dor/files/2022-county-tax-rates-and-codes.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Arkansas', 'state_occupation': 'Indiana', "capital_gains_short": capital_gains_short, "dependents": dependents, "sources": sources, "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def arkansas_and_iowa(self, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Arkansas:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Iowa:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Arkansas:
        - https://www.arkansasedc.com/why-arkansas/business-climate/tax-structure/personal-income-tax
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_and_AR1000NR_Instructions.pdf
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_FullYearResidentIndividualIncomeTaxReturn.pdf
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_and_AR1000NR_Instructions.pdf
          https://www.arkansasedc.com/why-arkansas/business-climate/tax-structure/capital-gains-tax-reduction
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_and_AR1000NR_Instructions.pdf
        
        Iowa:
        -         
        - https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
          https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
          https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
          https://tax.iowa.gov/iowa-illinois-reciprocal-agreement
          https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Arkansas', 'state_occupation': 'Iowa', "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def arkansas_and_kansas(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Arkansas:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Kansas:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Arkansas:
        - https://www.arkansasedc.com/why-arkansas/business-climate/tax-structure/personal-income-tax
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_and_AR1000NR_Instructions.pdf
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_FullYearResidentIndividualIncomeTaxReturn.pdf
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_and_AR1000NR_Instructions.pdf
          https://www.arkansasedc.com/why-arkansas/business-climate/tax-structure/capital-gains-tax-reduction
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_and_AR1000NR_Instructions.pdf
        
        Kansas:
        -         
        - https://www.ksrevenue.gov/taxrates.html
          https://www.ksrevenue.gov/incomebook22.html#0
          https://www.ksrevenue.gov/incomebook22.html#0
          https://www.ksrevenue.gov/incomebook22.html#0
          https://www.ksrevenue.gov/incomebook22.html
          https://www.ksrevenue.gov/incomebook22.html
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Arkansas', 'state_occupation': 'Kansas', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def arkansas_and_kentucky(self, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Arkansas:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Kentucky:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Arkansas:
        - https://www.arkansasedc.com/why-arkansas/business-climate/tax-structure/personal-income-tax
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_and_AR1000NR_Instructions.pdf
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_FullYearResidentIndividualIncomeTaxReturn.pdf
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_and_AR1000NR_Instructions.pdf
          https://www.arkansasedc.com/why-arkansas/business-climate/tax-structure/capital-gains-tax-reduction
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_and_AR1000NR_Instructions.pdf
        
        Kentucky:
        -         
        - https://revenue.ky.gov/News/Pages/Dor-Announces-Updates-to-Individual-Income-Tax-for-2023-Tax-Year.aspx
          https://revenue.ky.gov/Individual/Individual-Income-Tax/Pages/default.aspx
          https://revenue.ky.gov/News/Pages/Dor-Announces-Updates-to-Individual-Income-Tax-for-2023-Tax-Year.aspx
          https://revenue.ky.gov/Individual/Individual-Income-Tax/Pages/default.aspx
          https://revenue.ky.gov/Dor%20Training%20Materials/103%20KAR%2017.140.%20Individual%20income%20tax%20-%20reciprocity%20-%20nonresidents.pdf
          https://revenue.ky.gov/Individual/Individual-Income-Tax/Pages/default.aspx
          https://nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2021/Taxes-21-32.htm?taxmap=true
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Arkansas', 'state_occupation': 'Kentucky', "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def arkansas_and_louisiana(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Arkansas:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Louisiana:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Arkansas:
        - https://www.arkansasedc.com/why-arkansas/business-climate/tax-structure/personal-income-tax
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_and_AR1000NR_Instructions.pdf
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_FullYearResidentIndividualIncomeTaxReturn.pdf
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_and_AR1000NR_Instructions.pdf
          https://www.arkansasedc.com/why-arkansas/business-climate/tax-structure/capital-gains-tax-reduction
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_and_AR1000NR_Instructions.pdf
        
        Louisiana:
        -         
        - https://revenue.louisiana.gov/individualincometax
          https://revenue.louisiana.gov/individualincometax
          https://www.revenue.louisiana.gov/taxforms/6935(11_02)F.pdf
          https://revenue.louisiana.gov/individualincometax
          https://revenue.louisiana.gov/individualincometax
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Arkansas', 'state_occupation': 'Louisiana', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def arkansas_and_maine(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Arkansas:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Maine:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Arkansas:
        - https://www.arkansasedc.com/why-arkansas/business-climate/tax-structure/personal-income-tax
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_and_AR1000NR_Instructions.pdf
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_FullYearResidentIndividualIncomeTaxReturn.pdf
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_and_AR1000NR_Instructions.pdf
          https://www.arkansasedc.com/why-arkansas/business-climate/tax-structure/capital-gains-tax-reduction
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_and_AR1000NR_Instructions.pdf
        
        Maine:
        -         
        - https://www.maine.gov/revenue/tax-return-forms/individual-income-tax-2022
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/ind_tax_rate_sched_2022.pdf
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/22_1040me_book_gen_instruc_revisedFeb23.pdf
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/22_1040me_book_gen_instruc_revisedFeb23.pdf
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/22_1040me_book_gen_instruc_revisedFeb23.pdf
          https://www.maine.gov/revenue/tax-return-forms/individual-income-tax-2022
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Arkansas', 'state_occupation': 'Maine', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def arkansas_and_maryland(self, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Arkansas:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Maryland:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Arkansas:
        - https://www.arkansasedc.com/why-arkansas/business-climate/tax-structure/personal-income-tax
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_and_AR1000NR_Instructions.pdf
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_FullYearResidentIndividualIncomeTaxReturn.pdf
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_and_AR1000NR_Instructions.pdf
          https://www.arkansasedc.com/why-arkansas/business-climate/tax-structure/capital-gains-tax-reduction
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_and_AR1000NR_Instructions.pdf
        
        Maryland:
        -         
        - https://www.marylandtaxes.gov/individual/income/tax-info/tax-rates.php
          https://www.marylandtaxes.gov/forms/22_forms/Resident_Booklet.pdf
          https://www.marylandtaxes.gov/forms/Tax_Publications/Administrative_Releases/Income_and_Estate_Tax_Releases/ar_it37.pdf
          https://www.marylandtaxes.gov/individual/index.php
          https://www.marylandtaxes.gov/individual/income/tax-info/tax-rates.php
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Arkansas', 'state_occupation': 'Maryland', "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def arkansas_and_massachusetts(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Arkansas:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Massachusetts:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Arkansas:
        - https://www.arkansasedc.com/why-arkansas/business-climate/tax-structure/personal-income-tax
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_and_AR1000NR_Instructions.pdf
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_FullYearResidentIndividualIncomeTaxReturn.pdf
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_and_AR1000NR_Instructions.pdf
          https://www.arkansasedc.com/why-arkansas/business-climate/tax-structure/capital-gains-tax-reduction
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_and_AR1000NR_Instructions.pdf
        
        Massachusetts:
        -         
        - https://www.mass.gov/service-details/massachusetts-tax-rates
          https://www.mass.gov/service-details/view-massachusetts-personal-income-tax-exemptions
          https://www.mass.gov/service-details/view-massachusetts-personal-income-tax-exemptions
          https://www.mass.gov/service-details/view-child-and-dependent-related-credits
          https://www.mass.gov/service-details/massachusetts-tax-rates
          https://www.mass.gov/personal-income-tax
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Arkansas', 'state_occupation': 'Massachusetts', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def arkansas_and_michigan(self, capital_gains_short=0, county_occupation='Other', dependents=0, sources=False, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_long=0, filing_status='single', income=100000, credits_state=0):
        """
        --- Parameters ----
        
        Arkansas:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Michigan:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        county_occupation : str, optional
            County of occupation. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Arkansas:
        - https://www.arkansasedc.com/why-arkansas/business-climate/tax-structure/personal-income-tax
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_and_AR1000NR_Instructions.pdf
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_FullYearResidentIndividualIncomeTaxReturn.pdf
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_and_AR1000NR_Instructions.pdf
          https://www.arkansasedc.com/why-arkansas/business-climate/tax-structure/capital-gains-tax-reduction
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_and_AR1000NR_Instructions.pdf
        
        Michigan:
        -         
        - https://www.michigan.gov/taxes/iit/new-developments-for-tax-year-2022
          "https://www.michigan.gov/taxes/-/media/Project/Websites/taxes/Forms/2023/2023-SUW/446_Withholding-Guide_2023.pdf?rev=177ca3d8cb034392bf84e317ed78c320&hash=19C5F42F43A89A91863D229349F48883#:~:text=The%20exemption%20amount%20is%20%245%2C400
         employees%20Michigan%20income%20tax%20return"
          "https://www.michigan.gov/taxes/-/media/Project/Websites/taxes/Forms/2023/2023-SUW/446_Withholding-Guide_2023.pdf?rev=177ca3d8cb034392bf84e317ed78c320&hash=19C5F42F43A89A91863D229349F48883#:~:text=The%20exemption%20amount%20is%20%245%2C400
         employees%20Michigan%20income%20tax%20return"
          https://www.michigan.gov/taxes/questions/iit/accordion/residency/are-my-wages-earned-in-another-state-taxable-in-michigan-if-i-am-a-michigan-resident-1
          https://www.michigan.gov/taxes/-/media/Project/Websites/taxes/Forms/2022/2022-IIT-Forms/BOOK_MI-1040.pdf?rev=4524f1cbbc0243b194ab05a6680be75b&hash=40DFE71E651F7C1D65C31F6CD03D944F
          https://www.michigan.gov/taxes/questions/iit/accordion/general/what-cities-impose-an-income-tax
        """
        
        # Check if the parameters are set upstream in the instance
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        county_occupation = self.county_occupation if hasattr(self, 'county_occupation') else county_occupation
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Arkansas', 'state_occupation': 'Michigan', "capital_gains_short": capital_gains_short, "county_occupation": county_occupation, "dependents": dependents, "sources": sources, "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def arkansas_and_minnesota(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Arkansas:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Minnesota:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Arkansas:
        - https://www.arkansasedc.com/why-arkansas/business-climate/tax-structure/personal-income-tax
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_and_AR1000NR_Instructions.pdf
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_FullYearResidentIndividualIncomeTaxReturn.pdf
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_and_AR1000NR_Instructions.pdf
          https://www.arkansasedc.com/why-arkansas/business-climate/tax-structure/capital-gains-tax-reduction
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_and_AR1000NR_Instructions.pdf
        
        Minnesota:
        -         
        - https://www.revenue.state.mn.us/minnesota-income-tax-rates-and-brackets
          https://www.revenue.state.mn.us/child-and-dependent-care-credit
          https://www.revenue.state.mn.us/dependent-exemptions
          https://www.revenue.state.mn.us/minnesota-standard-deduction
          https://mn.gov/mmb/assets/mwr_form_tcm1059-128581.pdf
          https://www.revenue.state.mn.us/individual-income-tax
          https://www.revenue.state.mn.us/individual-income-tax
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Arkansas', 'state_occupation': 'Minnesota', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def arkansas_and_mississippi(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Arkansas:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Mississippi:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Arkansas:
        - https://www.arkansasedc.com/why-arkansas/business-climate/tax-structure/personal-income-tax
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_and_AR1000NR_Instructions.pdf
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_FullYearResidentIndividualIncomeTaxReturn.pdf
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_and_AR1000NR_Instructions.pdf
          https://www.arkansasedc.com/why-arkansas/business-climate/tax-structure/capital-gains-tax-reduction
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_and_AR1000NR_Instructions.pdf
        
        Mississippi:
        -         
        - https://www.dor.ms.gov/individual/tax-rates
          https://www.dor.ms.gov/individual/tax-rates
          https://www.dor.ms.gov/individual/tax-rates
          https://www.dor.ms.gov/individual
          https://www.dor.ms.gov/individual
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Arkansas', 'state_occupation': 'Mississippi', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def arkansas_and_missouri(self, capital_gains_short=0, dependents=0, sources=False, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_long=0, filing_status='single', income=100000, credits_state=0):
        """
        --- Parameters ----
        
        Arkansas:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Missouri:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Arkansas:
        - https://www.arkansasedc.com/why-arkansas/business-climate/tax-structure/personal-income-tax
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_and_AR1000NR_Instructions.pdf
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_FullYearResidentIndividualIncomeTaxReturn.pdf
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_and_AR1000NR_Instructions.pdf
          https://www.arkansasedc.com/why-arkansas/business-climate/tax-structure/capital-gains-tax-reduction
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_and_AR1000NR_Instructions.pdf
        
        Missouri:
        -         
        - https://dor.mo.gov/taxation/individual/tax-types/income/year-changes/
          https://dor.mo.gov/taxation/individual/tax-types/income/year-changes/
          https://dor.mo.gov/taxation/individual/tax-types/income/year-changes/
          https://www.kcmo.gov/city-hall/departments/finance/earnings-tax
          https://www.stlouis-mo.gov/government/departments/collector/earnings-tax/individual-earnings-tax-info.cfm
        """
        
        # Check if the parameters are set upstream in the instance
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Arkansas', 'state_occupation': 'Missouri', "capital_gains_short": capital_gains_short, "dependents": dependents, "sources": sources, "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def arkansas_and_montana(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Arkansas:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Montana:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Arkansas:
        - https://www.arkansasedc.com/why-arkansas/business-climate/tax-structure/personal-income-tax
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_and_AR1000NR_Instructions.pdf
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_FullYearResidentIndividualIncomeTaxReturn.pdf
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_and_AR1000NR_Instructions.pdf
          https://www.arkansasedc.com/why-arkansas/business-climate/tax-structure/capital-gains-tax-reduction
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_and_AR1000NR_Instructions.pdf
        
        Montana:
        -         
        - https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/2022/12/montana-employees-withholding-allowance-and-exemption-certificate-form-mw-4.pdf
          https://mtrevenue.gov/2021/11/29/simplification-of-montana-income-taxation/?utm_source=rss&utm_medium=rss&utm_campaign=simplification-of-montana-income-taxation
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Arkansas', 'state_occupation': 'Montana', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def arkansas_and_nebraska(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Arkansas:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Nebraska:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Arkansas:
        - https://www.arkansasedc.com/why-arkansas/business-climate/tax-structure/personal-income-tax
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_and_AR1000NR_Instructions.pdf
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_FullYearResidentIndividualIncomeTaxReturn.pdf
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_and_AR1000NR_Instructions.pdf
          https://www.arkansasedc.com/why-arkansas/business-climate/tax-structure/capital-gains-tax-reduction
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_and_AR1000NR_Instructions.pdf
        
        Nebraska:
        -         
        - https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2023_Circular_EN_8-429r2023_Whole_Book_11.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Arkansas', 'state_occupation': 'Nebraska', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def arkansas_and_nevada(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Arkansas:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Nevada:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Arkansas:
        - https://www.arkansasedc.com/why-arkansas/business-climate/tax-structure/personal-income-tax
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_and_AR1000NR_Instructions.pdf
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_FullYearResidentIndividualIncomeTaxReturn.pdf
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_and_AR1000NR_Instructions.pdf
          https://www.arkansasedc.com/why-arkansas/business-climate/tax-structure/capital-gains-tax-reduction
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_and_AR1000NR_Instructions.pdf
        
        Nevada:
        -         
        - https://www.nvsos.gov/sos/businesses/the-nevada-advantage
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Arkansas', 'state_occupation': 'Nevada', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def arkansas_and_new_hampshire(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Arkansas:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        New Hampshire:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Arkansas:
        - https://www.arkansasedc.com/why-arkansas/business-climate/tax-structure/personal-income-tax
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_and_AR1000NR_Instructions.pdf
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_FullYearResidentIndividualIncomeTaxReturn.pdf
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_and_AR1000NR_Instructions.pdf
          https://www.arkansasedc.com/why-arkansas/business-climate/tax-structure/capital-gains-tax-reduction
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_and_AR1000NR_Instructions.pdf
        
        New Hampshire:
        -         
        - https://www.revenue.nh.gov/assistance/tax-overview.htm
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Arkansas', 'state_occupation': 'NewHampshire', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def arkansas_and_new_jersey(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Arkansas:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        New Jersey:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Arkansas:
        - https://www.arkansasedc.com/why-arkansas/business-climate/tax-structure/personal-income-tax
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_and_AR1000NR_Instructions.pdf
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_FullYearResidentIndividualIncomeTaxReturn.pdf
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_and_AR1000NR_Instructions.pdf
          https://www.arkansasedc.com/why-arkansas/business-climate/tax-structure/capital-gains-tax-reduction
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_and_AR1000NR_Instructions.pdf
        
        New Jersey:
        -         
        - https://www.state.nj.us/treasury/taxation/taxtables.shtml
          https://www.state.nj.us/treasury/taxation/njit13.shtml
          https://www.nj.gov/treasury/taxation/njit25.shtml
          https://www.state.nj.us/treasury/taxation/individuals/freqqiti.shtml
          https://www.state.nj.us/treasury/taxation/localtax.shtml
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Arkansas', 'state_occupation': 'NewJersey', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def arkansas_and_new_mexico(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Arkansas:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        New Mexico:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Arkansas:
        - https://www.arkansasedc.com/why-arkansas/business-climate/tax-structure/personal-income-tax
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_and_AR1000NR_Instructions.pdf
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_FullYearResidentIndividualIncomeTaxReturn.pdf
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_and_AR1000NR_Instructions.pdf
          https://www.arkansasedc.com/why-arkansas/business-climate/tax-structure/capital-gains-tax-reduction
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_and_AR1000NR_Instructions.pdf
        
        New Mexico:
        -         
        - https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/fdf3c548-8aba-4b9c-9eb4-bb564c716015/FYI-104.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/0bf943fd-652e-4400-bb1b-cd8397eb5a95/2022pit-adj.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/2f1a6781-9534-4436-b427-1557f9592099/2022pit-adj-ins.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/0bf943fd-652e-4400-bb1b-cd8397eb5a95/2022pit-adj.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/1afc56af-ea90-4d48-82e5-1f9aeb43255a/PITbook2022.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/fdf3c548-8aba-4b9c-9eb4-bb564c716015/FYI-104.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Arkansas', 'state_occupation': 'NewMexico', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def arkansas_and_new_york(self, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Arkansas:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        New York:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Arkansas:
        - https://www.arkansasedc.com/why-arkansas/business-climate/tax-structure/personal-income-tax
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_and_AR1000NR_Instructions.pdf
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_FullYearResidentIndividualIncomeTaxReturn.pdf
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_and_AR1000NR_Instructions.pdf
          https://www.arkansasedc.com/why-arkansas/business-climate/tax-structure/capital-gains-tax-reduction
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_and_AR1000NR_Instructions.pdf
        
        New York:
        -         
        - https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nys-tax-rate-schedule
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nys-tax-rate-schedule
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nys-tax-rate-schedule
          https://www.tax.ny.gov/pit/
          https://www.tax.ny.gov/pit/file/nonresident-faqs.htm
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#yonkers-tax-rate-schedule
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nyc-tax-rate-schedule
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Arkansas', 'state_occupation': 'NewYork', "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def arkansas_and_north_carolina(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Arkansas:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        North Carolina:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Arkansas:
        - https://www.arkansasedc.com/why-arkansas/business-climate/tax-structure/personal-income-tax
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_and_AR1000NR_Instructions.pdf
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_FullYearResidentIndividualIncomeTaxReturn.pdf
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_and_AR1000NR_Instructions.pdf
          https://www.arkansasedc.com/why-arkansas/business-climate/tax-structure/capital-gains-tax-reduction
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_and_AR1000NR_Instructions.pdf
        
        North Carolina:
        -         
        - https://www.ncdor.gov/taxes-forms/tax-rate-schedules
          https://www.ncdor.gov/taxes-forms/individual-income-tax/north-carolina-standard-deduction-or-north-carolina-itemized-deductions
          https://www.ncdor.gov/taxes-forms/individual-income-tax
          https://www.ncdor.gov/taxes-forms/individual-income-tax
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Arkansas', 'state_occupation': 'NorthCarolina', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def arkansas_and_north_dakota(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Arkansas:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        North Dakota:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Arkansas:
        - https://www.arkansasedc.com/why-arkansas/business-climate/tax-structure/personal-income-tax
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_and_AR1000NR_Instructions.pdf
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_FullYearResidentIndividualIncomeTaxReturn.pdf
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_and_AR1000NR_Instructions.pdf
          https://www.arkansasedc.com/why-arkansas/business-climate/tax-structure/capital-gains-tax-reduction
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_and_AR1000NR_Instructions.pdf
        
        North Dakota:
        -         
        - https://www.tax.nd.gov/forms
          https://www.tax.nd.gov/sites/www/files/documents/forms/individual/2022-iit/2022-individual-income-tax-booklet.pdf
          https://www.tax.nd.gov/sites/www/files/documents/forms/2022-individual-income-tax-booklet.pdf
          https://www.tax.nd.gov/tax-exemptions-credits/income-tax-exemptions-credits
          https://www.tax.nd.gov/sites/www/files/documents/forms/business/ndwrfillable.pdf
          https://www.tax.nd.gov/news/tax-legislative-changes/significant-changes-law/individual-income-tax-history
          https://www.tax.nd.gov/forms
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Arkansas', 'state_occupation': 'NorthDakota', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def arkansas_and_ohio(self, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Arkansas:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Ohio:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Arkansas:
        - https://www.arkansasedc.com/why-arkansas/business-climate/tax-structure/personal-income-tax
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_and_AR1000NR_Instructions.pdf
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_FullYearResidentIndividualIncomeTaxReturn.pdf
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_and_AR1000NR_Instructions.pdf
          https://www.arkansasedc.com/why-arkansas/business-climate/tax-structure/capital-gains-tax-reduction
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_and_AR1000NR_Instructions.pdf
        
        Ohio:
        -         
        - https://tax.ohio.gov/individual/resources/annual-tax-rates
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://tax.ohio.gov/static/forms/employer_withholding/generic/wth_it4nr.pdf
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://www.ritaohio.com/TaxRatesTable
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Arkansas', 'state_occupation': 'Ohio', "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def arkansas_and_oklahoma(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Arkansas:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Oklahoma:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Arkansas:
        - https://www.arkansasedc.com/why-arkansas/business-climate/tax-structure/personal-income-tax
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_and_AR1000NR_Instructions.pdf
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_FullYearResidentIndividualIncomeTaxReturn.pdf
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_and_AR1000NR_Instructions.pdf
          https://www.arkansasedc.com/why-arkansas/business-climate/tax-structure/capital-gains-tax-reduction
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_and_AR1000NR_Instructions.pdf
        
        Oklahoma:
        -         
        - https://oklahoma.gov/content/dam/ok/en/tax/documents/resources/publications/businesses/withholding-tables/WHTables-2022.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Arkansas', 'state_occupation': 'Oklahoma', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def arkansas_and_oregon(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Arkansas:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Oregon:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Arkansas:
        - https://www.arkansasedc.com/why-arkansas/business-climate/tax-structure/personal-income-tax
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_and_AR1000NR_Instructions.pdf
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_FullYearResidentIndividualIncomeTaxReturn.pdf
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_and_AR1000NR_Instructions.pdf
          https://www.arkansasedc.com/why-arkansas/business-climate/tax-structure/capital-gains-tax-reduction
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_and_AR1000NR_Instructions.pdf
        
        Oregon:
        -         
        - https://www.oregon.gov/dor/programs/individuals/Pages/PIT.aspx
          https://www.oregon.gov/dor/programs/individuals/Documents/Full%20year%20resident
         %20Form%20or-40%20filers.pdf
          https://www.oregon.gov/dor/programs/individuals/Documents/Part-year%20and%20nonresident
         %20Form%20or-40-P%20and%20or-40-N%20filers.pdf.pdf
          https://www.oregon.gov/dor/forms/FormsPubs/form-or-40_101-040_2022.pdf
          https://www.oregon.gov/dor/forms/FormsPubs/form-or-40_101-040_2022.pdf
          https://www.oregon.gov/dor/forms/FormsPubs/form-or-40_101-040_2022.pdf
          https://www.portland.gov/revenue/personal-tax
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Arkansas', 'state_occupation': 'Oregon', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def arkansas_and_pennsylvania(self, county_residence='Other', state_agi=75000, county_occupation='Other', filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Arkansas:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Pennsylvania:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        county_residence : str, optional
            County of residence. Default is Other.
        county_occupation : str, optional
            County of occupation. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Arkansas:
        - https://www.arkansasedc.com/why-arkansas/business-climate/tax-structure/personal-income-tax
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_and_AR1000NR_Instructions.pdf
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_FullYearResidentIndividualIncomeTaxReturn.pdf
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_and_AR1000NR_Instructions.pdf
          https://www.arkansasedc.com/why-arkansas/business-climate/tax-structure/capital-gains-tax-reduction
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_and_AR1000NR_Instructions.pdf
        
        Pennsylvania:
        -         
        - https://www.revenue.pa.gov/Tax%20Rates/Pages/default.aspx
          https://www.revenue.pa.gov/FormsandPublications/FormsforIndividuals/PIT/Documents/2022/2022_pa-40in.pdf
          https://www.revenue.pa.gov/FormsandPublications/FormsforBusinesses/EmployerWithholding/Documents/rev-419.pdf
          https://www.revenue.pa.gov/FormsandPublications/FormsforIndividuals/PIT/Documents/2022/2022_pa-40in.pdf
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-25.htm
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        county_occupation = self.county_occupation if hasattr(self, 'county_occupation') else county_occupation
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Arkansas', 'state_occupation': 'Pennsylvania', "county_residence": county_residence, "state_agi": state_agi, "county_occupation": county_occupation, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def arkansas_and_rhode_island(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Arkansas:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Rhode Island:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Arkansas:
        - https://www.arkansasedc.com/why-arkansas/business-climate/tax-structure/personal-income-tax
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_and_AR1000NR_Instructions.pdf
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_FullYearResidentIndividualIncomeTaxReturn.pdf
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_and_AR1000NR_Instructions.pdf
          https://www.arkansasedc.com/why-arkansas/business-climate/tax-structure/capital-gains-tax-reduction
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_and_AR1000NR_Instructions.pdf
        
        Rhode Island:
        -         
        - https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2022_1040WE_w_0.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Arkansas', 'state_occupation': 'RhodeIsland', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def arkansas_and_south_carolina(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Arkansas:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        South Carolina:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Arkansas:
        - https://www.arkansasedc.com/why-arkansas/business-climate/tax-structure/personal-income-tax
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_and_AR1000NR_Instructions.pdf
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_FullYearResidentIndividualIncomeTaxReturn.pdf
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_and_AR1000NR_Instructions.pdf
          https://www.arkansasedc.com/why-arkansas/business-climate/tax-structure/capital-gains-tax-reduction
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_and_AR1000NR_Instructions.pdf
        
        South Carolina:
        -         
        - https://dor.sc.gov/resources-site/lawandpolicy/Advisory%20Opinions/IL22-15.pdf
          https://dor.sc.gov/forms-site/Forms/IITPacket_2022.pdf
          https://www.scstatehouse.gov/code/t12c006.php
          https://dor.sc.gov/resources-site/lawandpolicy/Advisory%20Opinions/IL22-15.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Arkansas', 'state_occupation': 'SouthCarolina', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def arkansas_and_south_dakota(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Arkansas:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        South Dakota:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Arkansas:
        - https://www.arkansasedc.com/why-arkansas/business-climate/tax-structure/personal-income-tax
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_and_AR1000NR_Instructions.pdf
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_FullYearResidentIndividualIncomeTaxReturn.pdf
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_and_AR1000NR_Instructions.pdf
          https://www.arkansasedc.com/why-arkansas/business-climate/tax-structure/capital-gains-tax-reduction
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_and_AR1000NR_Instructions.pdf
        
        South Dakota:
        -         
        - https://dor.sd.gov/individuals/taxes/
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Arkansas', 'state_occupation': 'SouthDakota', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def arkansas_and_tennessee(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Arkansas:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Tennessee:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Arkansas:
        - https://www.arkansasedc.com/why-arkansas/business-climate/tax-structure/personal-income-tax
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_and_AR1000NR_Instructions.pdf
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_FullYearResidentIndividualIncomeTaxReturn.pdf
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_and_AR1000NR_Instructions.pdf
          https://www.arkansasedc.com/why-arkansas/business-climate/tax-structure/capital-gains-tax-reduction
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_and_AR1000NR_Instructions.pdf
        
        Tennessee:
        -         
        - https://www.tn.gov/revenue/taxes.html
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Arkansas', 'state_occupation': 'Tennessee', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def arkansas_and_texas(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Arkansas:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Texas:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Arkansas:
        - https://www.arkansasedc.com/why-arkansas/business-climate/tax-structure/personal-income-tax
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_and_AR1000NR_Instructions.pdf
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_FullYearResidentIndividualIncomeTaxReturn.pdf
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_and_AR1000NR_Instructions.pdf
          https://www.arkansasedc.com/why-arkansas/business-climate/tax-structure/capital-gains-tax-reduction
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_and_AR1000NR_Instructions.pdf
        
        Texas:
        -         
        - https://comptroller.texas.gov/taxes/
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Arkansas', 'state_occupation': 'Texas', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def arkansas_and_utah(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Arkansas:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Utah:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Arkansas:
        - https://www.arkansasedc.com/why-arkansas/business-climate/tax-structure/personal-income-tax
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_and_AR1000NR_Instructions.pdf
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_FullYearResidentIndividualIncomeTaxReturn.pdf
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_and_AR1000NR_Instructions.pdf
          https://www.arkansasedc.com/why-arkansas/business-climate/tax-structure/capital-gains-tax-reduction
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_and_AR1000NR_Instructions.pdf
        
        Utah:
        -         
        - https://incometax.utah.gov/paying/tax-rates
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Arkansas', 'state_occupation': 'Utah', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def arkansas_and_vermont(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Arkansas:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Vermont:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Arkansas:
        - https://www.arkansasedc.com/why-arkansas/business-climate/tax-structure/personal-income-tax
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_and_AR1000NR_Instructions.pdf
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_FullYearResidentIndividualIncomeTaxReturn.pdf
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_and_AR1000NR_Instructions.pdf
          https://www.arkansasedc.com/why-arkansas/business-climate/tax-structure/capital-gains-tax-reduction
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_and_AR1000NR_Instructions.pdf
        
        Vermont:
        -         
        - https://tax.vermont.gov/sites/tax/files/documents/GB-1210-2023.pdf
          https://tax.vermont.gov/sites/tax/files/documents/Income%20Booklet-2022.pdf
          https://tax.vermont.gov/sites/tax/files/documents/Income%20Booklet-2022.pdf
          https://tax.vermont.gov/sites/tax/files/documents/Income%20Booklet-2022.pdf
          https://tax.vermont.gov/individuals/personal-income-tax/tax-credits
          https://tax.vermont.gov/individuals/personal-income-tax/taxable-income
          https://tax.vermont.gov/individuals/personal-income-tax/local-option-tax
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Arkansas', 'state_occupation': 'Vermont', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def arkansas_and_virginia(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Arkansas:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Virginia:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Arkansas:
        - https://www.arkansasedc.com/why-arkansas/business-climate/tax-structure/personal-income-tax
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_and_AR1000NR_Instructions.pdf
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_FullYearResidentIndividualIncomeTaxReturn.pdf
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_and_AR1000NR_Instructions.pdf
          https://www.arkansasedc.com/why-arkansas/business-climate/tax-structure/capital-gains-tax-reduction
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_and_AR1000NR_Instructions.pdf
        
        Virginia:
        -         
        - https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/reciprocity
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Arkansas', 'state_occupation': 'Virginia', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def arkansas_and_washington(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Arkansas:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Washington:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Arkansas:
        - https://www.arkansasedc.com/why-arkansas/business-climate/tax-structure/personal-income-tax
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_and_AR1000NR_Instructions.pdf
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_FullYearResidentIndividualIncomeTaxReturn.pdf
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_and_AR1000NR_Instructions.pdf
          https://www.arkansasedc.com/why-arkansas/business-climate/tax-structure/capital-gains-tax-reduction
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_and_AR1000NR_Instructions.pdf
        
        Washington:
        -         
        - https://dor.wa.gov/taxes-rates/income-tax
          https://dor.wa.gov/taxes-rates/other-taxes/capital-gains-tax
          https://dor.wa.gov/taxes-rates/income-tax
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Arkansas', 'state_occupation': 'Washington', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def arkansas_and_west_virginia(self, pay_periods=24, capital_gains_short=0, county_occupation='Other', dependents=0, sources=False, state_agi=75000, filers_over_65=0, capital_gains_long=0, filing_status='single', income=100000, credits_state=0):
        """
        --- Parameters ----
        
        Arkansas:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        West Virginia:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_occupation : str, optional
            County of occupation. Default is Other.
        pay_periods : int, optional
            Number of pay periods. Options are 13, 24, 26, 52. Default is 24.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Arkansas:
        - https://www.arkansasedc.com/why-arkansas/business-climate/tax-structure/personal-income-tax
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_and_AR1000NR_Instructions.pdf
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_FullYearResidentIndividualIncomeTaxReturn.pdf
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_and_AR1000NR_Instructions.pdf
          https://www.arkansasedc.com/why-arkansas/business-climate/tax-structure/capital-gains-tax-reduction
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_and_AR1000NR_Instructions.pdf
        
        West Virginia:
        -         
        - https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://tax.wv.gov/Documents/TSD/tsd438.pdf
          https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2011/Taxes-11-35att.html
        """
        
        # Check if the parameters are set upstream in the instance
        pay_periods = self.pay_periods if hasattr(self, 'pay_periods') else pay_periods
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        county_occupation = self.county_occupation if hasattr(self, 'county_occupation') else county_occupation
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Arkansas', 'state_occupation': 'WestVirginia', "pay_periods": pay_periods, "capital_gains_short": capital_gains_short, "county_occupation": county_occupation, "dependents": dependents, "sources": sources, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def arkansas_and_wisconsin(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Arkansas:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Wisconsin:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Arkansas:
        - https://www.arkansasedc.com/why-arkansas/business-climate/tax-structure/personal-income-tax
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_and_AR1000NR_Instructions.pdf
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_FullYearResidentIndividualIncomeTaxReturn.pdf
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_and_AR1000NR_Instructions.pdf
          https://www.arkansasedc.com/why-arkansas/business-climate/tax-structure/capital-gains-tax-reduction
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_and_AR1000NR_Instructions.pdf
        
        Wisconsin:
        -         
        - https://www.revenue.wi.gov/TaxForms2023/2023-Form1-ES-Inst.pdf
          https://www.revenue.wi.gov/TaxForms2023/2023-Form1-ES-Inst.pdf
          https://www.revenue.wi.gov/TaxForms2023/2023-Form1-ES-Inst.pdf
          https://www.revenue.wi.gov/Pages/FAQS/pcs-work.aspx
          https://www.revenue.wi.gov/Dor%20Publications/pb103.pdf
          https://www.revenue.wi.gov/Pages/FAQS/pcs-work.aspx
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Arkansas', 'state_occupation': 'Wisconsin', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def arkansas_and_wyoming(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Arkansas:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Wyoming:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Arkansas:
        - https://www.arkansasedc.com/why-arkansas/business-climate/tax-structure/personal-income-tax
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_and_AR1000NR_Instructions.pdf
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_FullYearResidentIndividualIncomeTaxReturn.pdf
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_and_AR1000NR_Instructions.pdf
          https://www.arkansasedc.com/why-arkansas/business-climate/tax-structure/capital-gains-tax-reduction
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_and_AR1000NR_Instructions.pdf
        
        Wyoming:
        -         
        - https://revenue.wyo.gov/
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Arkansas', 'state_occupation': 'Wyoming', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def california_and_alabama(self, capital_gains_short=0, income=100000, dependents=0, sources=False, county_residence='Other', state_agi=75000, capital_gains_long=0, filing_status='single', credits_state=0):
        """
        --- Parameters ----
        
        California:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        
        Alabama:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        California:
        - https://www.ftb.ca.gov/forms/2022/2022-540-tax-rate-schedules.pdf
          https://www.ftb.ca.gov/file/personal/deductions/index.html
          https://www.ftb.ca.gov/about-ftb/newsroom/caleitc/eligibility-and-credit-information.html
          https://www.ftb.ca.gov/file/personal/income-types/capital-gains-and-losses.html#:~:text=California%20does%20not%20have%20a
         are%20taxed%20as%20ordinary%20income
          https://www.ftb.ca.gov/file/personal/
        
        Alabama:
        -         
        - https://www.revenue.alabama.gov/ultraviewer/viewer/basic_viewer/index.html?form=2023/02/22f40.pdf
          https://www.revenue.alabama.gov/faqs/what-is-alabamas-individual-income-tax-rate/
          https://www.revenue.alabama.gov/ultraviewer/viewer/basic_viewer/index.html?form=2023/02/22f40.pdf
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-24.htm?taxmap=true
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-24.htm?taxmap=true
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-24.htm?taxmap=true
          https://www.revenue.alabama.gov/individual-corporate/income-to-be-reported-on-the-alabama-income-tax-return/
          https://www.almonline.org/TaxRates.aspx#OccupationalTax
        """
        
        # Check if the parameters are set upstream in the instance
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'California', 'state_occupation': 'Alabama', "capital_gains_short": capital_gains_short, "income": income, "dependents": dependents, "sources": sources, "county_residence": county_residence, "state_agi": state_agi, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def california_and_alaska(self, capital_gains_long=0, state_agi=75000, filing_status='single', income=100000, sources=False, capital_gains_short=0, credits_state=0):
        """
        --- Parameters ----
        
        California:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        
        Alaska:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        California:
        - https://www.ftb.ca.gov/forms/2022/2022-540-tax-rate-schedules.pdf
          https://www.ftb.ca.gov/file/personal/deductions/index.html
          https://www.ftb.ca.gov/about-ftb/newsroom/caleitc/eligibility-and-credit-information.html
          https://www.ftb.ca.gov/file/personal/income-types/capital-gains-and-losses.html#:~:text=California%20does%20not%20have%20a
         are%20taxed%20as%20ordinary%20income
          https://www.ftb.ca.gov/file/personal/
        
        Alaska:
        -         
        - https://www.commerce.alaska.gov/web/dcra/officeofthestateassessor/alaskataxfacts.aspx
        """
        
        # Check if the parameters are set upstream in the instance
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        sources = self.sources if hasattr(self, 'sources') else sources
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'California', 'state_occupation': 'Alaska', "capital_gains_long": capital_gains_long, "state_agi": state_agi, "filing_status": filing_status, "income": income, "sources": sources, "capital_gains_short": capital_gains_short, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def california_and_arizona(self, state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        California:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        
        Arizona:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        credits_state : int, optional
            Credits for state taxes. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        California:
        - https://www.ftb.ca.gov/forms/2022/2022-540-tax-rate-schedules.pdf
          https://www.ftb.ca.gov/file/personal/deductions/index.html
          https://www.ftb.ca.gov/about-ftb/newsroom/caleitc/eligibility-and-credit-information.html
          https://www.ftb.ca.gov/file/personal/income-types/capital-gains-and-losses.html#:~:text=California%20does%20not%20have%20a
         are%20taxed%20as%20ordinary%20income
          https://www.ftb.ca.gov/file/personal/
        
        Arizona:
        -         
        - https://azdor.gov/forms/individual/form-140-x-y-tables
          https://azdor.gov/forms/individual/form-140-x-y-tables
          https://azdor.gov/forms/individual/form-140-x-y-tables
          https://azdor.gov/businesses-arizona/withholding-tax/withholding-exceptions
          "https://azdor.gov/individuals/income-tax-filing-assistance/identifying-other-taxable-income#:~:text=In%20regards%20to%20capital%20gains
         the%20individuals%20regular%20tax%20rate.
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'California', 'state_occupation': 'Arizona', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def california_and_arkansas(self, capital_gains_short=0, income=100000, dependents=0, sources=False, state_agi=75000, filers_over_65=0, capital_gains_long=0, filing_status='single', credits_state=0):
        """
        --- Parameters ----
        
        California:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        
        Arkansas:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        California:
        - https://www.ftb.ca.gov/forms/2022/2022-540-tax-rate-schedules.pdf
          https://www.ftb.ca.gov/file/personal/deductions/index.html
          https://www.ftb.ca.gov/about-ftb/newsroom/caleitc/eligibility-and-credit-information.html
          https://www.ftb.ca.gov/file/personal/income-types/capital-gains-and-losses.html#:~:text=California%20does%20not%20have%20a
         are%20taxed%20as%20ordinary%20income
          https://www.ftb.ca.gov/file/personal/
        
        Arkansas:
        -         
        - https://www.arkansasedc.com/why-arkansas/business-climate/tax-structure/personal-income-tax
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_and_AR1000NR_Instructions.pdf
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_FullYearResidentIndividualIncomeTaxReturn.pdf
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_and_AR1000NR_Instructions.pdf
          https://www.arkansasedc.com/why-arkansas/business-climate/tax-structure/capital-gains-tax-reduction
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_and_AR1000NR_Instructions.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'California', 'state_occupation': 'Arkansas', "capital_gains_short": capital_gains_short, "income": income, "dependents": dependents, "sources": sources, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def california_and_colorado(self, county_residence='Other', state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', sources=False, credits_state=0):
        """
        --- Parameters ----
        
        California:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        
        Colorado:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        California:
        - https://www.ftb.ca.gov/forms/2022/2022-540-tax-rate-schedules.pdf
          https://www.ftb.ca.gov/file/personal/deductions/index.html
          https://www.ftb.ca.gov/about-ftb/newsroom/caleitc/eligibility-and-credit-information.html
          https://www.ftb.ca.gov/file/personal/income-types/capital-gains-and-losses.html#:~:text=California%20does%20not%20have%20a
         are%20taxed%20as%20ordinary%20income
          https://www.ftb.ca.gov/file/personal/
        
        Colorado:
        -         
        - https://leg.colorado.gov/agencies/legislative-council-staff/individual-income-tax%C2%A0
          https://leg.colorado.gov/agencies/legislative-council-staff/individual-income-tax%C2%A0
          https://tax.colorado.gov/sites/tax/files/documents/ITT_Capital_Gain_Subtraction_Feb_2022.pdf
          https://www.auroragov.org/business_services/taxes/occupational_privilege_tax
          https://www.denvergov.org/Government/Agencies-Departments-Offices/Agencies-Departments-Offices-Directory/Department-of-Finance/Our-Divisions/Treasury/Business-Tax-Information#:~:text=Occupational%20privilege%20tax
         -%28show%20below%29&text=Each%20taxable%20employee%20is%20liable
         month%20for%20each%20taxable%20employee
          https://www.glendale.co.us/355/Occupational-Privilege-Tax
          https://www.greenwoodvillage.com/1220/Occupational-Privilege-Tax-OPT
          https://ci.sheridan.co.us/288/Occupational-Privilege-Tax
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'California', 'state_occupation': 'Colorado', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def california_and_connecticut(self, state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', sources=False, credits_state=0):
        """
        --- Parameters ----
        
        California:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        
        Connecticut:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        California:
        - https://www.ftb.ca.gov/forms/2022/2022-540-tax-rate-schedules.pdf
          https://www.ftb.ca.gov/file/personal/deductions/index.html
          https://www.ftb.ca.gov/about-ftb/newsroom/caleitc/eligibility-and-credit-information.html
          https://www.ftb.ca.gov/file/personal/income-types/capital-gains-and-losses.html#:~:text=California%20does%20not%20have%20a
         are%20taxed%20as%20ordinary%20income
          https://www.ftb.ca.gov/file/personal/
        
        Connecticut:
        -         
        - https://portal.ct.gov/-/media/DRS/Forms/2022/Income/2022-CT-1040-Instructions_1222.pdf
          https://www.cga.ct.gov/2022/rpt/pdf/2022-R-0108.pdf
          https://www.cga.ct.gov/current/pub/chap_229.htm#sec_12-703
          https://portal.ct.gov/-/media/DRS/Forms/2022/Income/2022-CT-1040-Instructions_1222.pdf
          "https://portal.ct.gov/DRS/Publications/TSSNs/TSSN-29#:~:text=An%20individuals%20net%20capital%20gains
         portion%20of%20Social%20Security%20benefits"
          "https://portal.ct.gov/DRS/Publications/TSSNs/TSSN-29#:~:text=An%20individuals%20net%20capital%20gains
         portion%20of%20Social%20Security%20benefits"
          https://portal.ct.gov/DRS/Individuals/Resident-Income-Tax/Tax-Information
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'California', 'state_occupation': 'Connecticut', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def california_and_delaware(self, state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, federal_agi=75000, filing_status='single', sources=False, credits_state=0):
        """
        --- Parameters ----
        
        California:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        
        Delaware:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        federal_agi : int, optional
            Adjusted Gross Income (AGI) for federal taxes. Default is 0. Sources - https://www.irs.gov/e-file-providers/definition-of-adjusted-gross-income, https://www.irs.gov/newsroom/taxpayers-should-know-the-difference-between-standard-and-itemized-deductions
        income : int, optional
            Gross income. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        California:
        - https://www.ftb.ca.gov/forms/2022/2022-540-tax-rate-schedules.pdf
          https://www.ftb.ca.gov/file/personal/deductions/index.html
          https://www.ftb.ca.gov/about-ftb/newsroom/caleitc/eligibility-and-credit-information.html
          https://www.ftb.ca.gov/file/personal/income-types/capital-gains-and-losses.html#:~:text=California%20does%20not%20have%20a
         are%20taxed%20as%20ordinary%20income
          https://www.ftb.ca.gov/file/personal/
        
        Delaware:
        -         
        - https://revenue.delaware.gov/software-developer/tax-rate-changes/
          https://revenuefiles.delaware.gov/2022/PIT-RES_TY22_2022-01_PaperInteractive.pdf
          https://revenuefiles.delaware.gov/2022/PIT-RES_TY22_2022-02_Instructions.pdf
          https://revenue.delaware.gov/frequently-asked-questions/personal-income-tax-faqs/
          https://www.newcastlede.gov/DocumentCenter/View/20147/City-of-Wilmington-Wage-Tax-Information-PDF
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        federal_agi = self.federal_agi if hasattr(self, 'federal_agi') else federal_agi
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'California', 'state_occupation': 'Delaware', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "federal_agi": federal_agi, "filing_status": filing_status, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def california_and_florida(self, capital_gains_long=0, state_agi=75000, filing_status='single', income=100000, sources=False, capital_gains_short=0, credits_state=0):
        """
        --- Parameters ----
        
        California:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        
        Florida:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        California:
        - https://www.ftb.ca.gov/forms/2022/2022-540-tax-rate-schedules.pdf
          https://www.ftb.ca.gov/file/personal/deductions/index.html
          https://www.ftb.ca.gov/about-ftb/newsroom/caleitc/eligibility-and-credit-information.html
          https://www.ftb.ca.gov/file/personal/income-types/capital-gains-and-losses.html#:~:text=California%20does%20not%20have%20a
         are%20taxed%20as%20ordinary%20income
          https://www.ftb.ca.gov/file/personal/
        
        Florida:
        -         
        - https://floridarevenue.com/taxes/taxesfees/Pages/tax_interest_rates.aspx
        """
        
        # Check if the parameters are set upstream in the instance
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        sources = self.sources if hasattr(self, 'sources') else sources
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'California', 'state_occupation': 'Florida', "capital_gains_long": capital_gains_long, "state_agi": state_agi, "filing_status": filing_status, "income": income, "sources": sources, "capital_gains_short": capital_gains_short, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def california_and_georgia(self, capital_gains_short=0, income=100000, dependents=0, sources=False, state_agi=75000, filers_over_65=0, capital_gains_long=0, filing_status='single', credits_state=0):
        """
        --- Parameters ----
        
        California:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        
        Georgia:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        California:
        - https://www.ftb.ca.gov/forms/2022/2022-540-tax-rate-schedules.pdf
          https://www.ftb.ca.gov/file/personal/deductions/index.html
          https://www.ftb.ca.gov/about-ftb/newsroom/caleitc/eligibility-and-credit-information.html
          https://www.ftb.ca.gov/file/personal/income-types/capital-gains-and-losses.html#:~:text=California%20does%20not%20have%20a
         are%20taxed%20as%20ordinary%20income
          https://www.ftb.ca.gov/file/personal/
        
        Georgia:
        -         
        - https://dor.georgia.gov/tax-tables-georgia-tax-rate-schedule
          https://apps.dor.ga.gov/FillableForms/PDFViewer/Index?form=2022GA500
          https://apps.dor.ga.gov/FillableForms/PDFViewer/Index?form=2022GA500
          https://apps.dor.ga.gov/FillableForms/PDFViewer/Index?form=2022GA500
          https://dor.georgia.gov/years-individual-income-tax-forms
          https://dor.georgia.gov/years-individual-income-tax-forms
        """
        
        # Check if the parameters are set upstream in the instance
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'California', 'state_occupation': 'Georgia', "capital_gains_short": capital_gains_short, "income": income, "dependents": dependents, "sources": sources, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def california_and_hawaii(self, capital_gains_short=0, income=100000, dependents=0, sources=False, state_agi=75000, filers_over_65=0, capital_gains_long=0, filing_status='single', credits_state=0):
        """
        --- Parameters ----
        
        California:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        
        Hawaii:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        California:
        - https://www.ftb.ca.gov/forms/2022/2022-540-tax-rate-schedules.pdf
          https://www.ftb.ca.gov/file/personal/deductions/index.html
          https://www.ftb.ca.gov/about-ftb/newsroom/caleitc/eligibility-and-credit-information.html
          https://www.ftb.ca.gov/file/personal/income-types/capital-gains-and-losses.html#:~:text=California%20does%20not%20have%20a
         are%20taxed%20as%20ordinary%20income
          https://www.ftb.ca.gov/file/personal/
        
        Hawaii:
        -         
        - https://tax.hawaii.gov/forms/d_18table-on/d_18table-on_p13/
          https://files.hawaii.gov/tax/forms/2018/18table-on.pdf
          https://files.hawaii.gov/tax/news/pubs/22outline.pdf
          https://files.hawaii.gov/tax/news/pubs/FastTaxRefGuide_2017.pdf
          https://www.tfhawaii.org/wordpress/state-tax-resources/mini-tax-guides/individual-income-tax-chapter-235/#:~:text=Long%20term%20capital%20gains%20are%20taxed%20at%20a%20maximum%20of%207.25%25
          https://www.grassrootinstitute.org/2023/02/hb337-increasing-tax-on-capital-gains-a-bad-risk/
          https://files.hawaii.gov/tax/news/pubs/22outline.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'California', 'state_occupation': 'Hawaii', "capital_gains_short": capital_gains_short, "income": income, "dependents": dependents, "sources": sources, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def california_and_idaho(self, capital_gains_short=0, income=100000, dependents=0, sources=False, state_agi=75000, filers_over_65=0, capital_gains_long=0, filing_status='single', credits_state=0):
        """
        --- Parameters ----
        
        California:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        
        Idaho:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        California:
        - https://www.ftb.ca.gov/forms/2022/2022-540-tax-rate-schedules.pdf
          https://www.ftb.ca.gov/file/personal/deductions/index.html
          https://www.ftb.ca.gov/about-ftb/newsroom/caleitc/eligibility-and-credit-information.html
          https://www.ftb.ca.gov/file/personal/income-types/capital-gains-and-losses.html#:~:text=California%20does%20not%20have%20a
         are%20taxed%20as%20ordinary%20income
          https://www.ftb.ca.gov/file/personal/
        
        Idaho:
        -         
        - https://tax.idaho.gov/taxes/income-tax/individual-income/individual-income-tax-rate-schedule/
          https://tax.idaho.gov/wp-content/uploads/forms/EIN00046/EIN00046_03-01-2023.pdf
          https://tax.idaho.gov/taxes/income-tax/individual-income/idaho-source-income/
          https://tax.idaho.gov/taxes/income-tax/individual-income/idaho-source-income/
        """
        
        # Check if the parameters are set upstream in the instance
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'California', 'state_occupation': 'Idaho', "capital_gains_short": capital_gains_short, "income": income, "dependents": dependents, "sources": sources, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def california_and_illinois(self, state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        California:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        
        Illinois:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        California:
        - https://www.ftb.ca.gov/forms/2022/2022-540-tax-rate-schedules.pdf
          https://www.ftb.ca.gov/file/personal/deductions/index.html
          https://www.ftb.ca.gov/about-ftb/newsroom/caleitc/eligibility-and-credit-information.html
          https://www.ftb.ca.gov/file/personal/income-types/capital-gains-and-losses.html#:~:text=California%20does%20not%20have%20a
         are%20taxed%20as%20ordinary%20income
          https://www.ftb.ca.gov/file/personal/
        
        Illinois:
        -         
        - https://tax.illinois.gov/research/taxrates/income.html#IndividualIncome
          https://tax.illinois.gov/content/dam/soi/en/web/tax/forms/incometax/documents/currentyear/individual/il-1040-instr.pdf
          https://tax.illinois.gov/questionsandanswers/12.html
          https://tax.illinois.gov/content/dam/soi/en/web/tax/forms/incometax/documents/currentyear/individual/il-1040-instr.pdf
          https://tax.illinois.gov/research/taxrates/personalproperty.html
          https://tax.illinois.gov/localgovernments/income.html
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'California', 'state_occupation': 'Illinois', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def california_and_indiana(self, capital_gains_short=0, income=100000, dependents=0, sources=False, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_long=0, filing_status='single', credits_state=0):
        """
        --- Parameters ----
        
        California:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        
        Indiana:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        California:
        - https://www.ftb.ca.gov/forms/2022/2022-540-tax-rate-schedules.pdf
          https://www.ftb.ca.gov/file/personal/deductions/index.html
          https://www.ftb.ca.gov/about-ftb/newsroom/caleitc/eligibility-and-credit-information.html
          https://www.ftb.ca.gov/file/personal/income-types/capital-gains-and-losses.html#:~:text=California%20does%20not%20have%20a
         are%20taxed%20as%20ordinary%20income
          https://www.ftb.ca.gov/file/personal/
        
        Indiana:
        -         
        - https://www.in.gov/dor/tax-forms/2022-individual-income-tax-forms/
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-32.htm
          https://www.in.gov/dor/files/ib28.pdf
          https://www.in.gov/dor/tax-forms/2022-individual-income-tax-forms/
          https://www.in.gov/dor/files/2022-county-tax-rates-and-codes.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'California', 'state_occupation': 'Indiana', "capital_gains_short": capital_gains_short, "income": income, "dependents": dependents, "sources": sources, "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def california_and_iowa(self, capital_gains_short=0, income=100000, dependents=0, sources=False, county_residence='Other', state_agi=75000, capital_gains_long=0, filing_status='single', credits_state=0):
        """
        --- Parameters ----
        
        California:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        
        Iowa:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        California:
        - https://www.ftb.ca.gov/forms/2022/2022-540-tax-rate-schedules.pdf
          https://www.ftb.ca.gov/file/personal/deductions/index.html
          https://www.ftb.ca.gov/about-ftb/newsroom/caleitc/eligibility-and-credit-information.html
          https://www.ftb.ca.gov/file/personal/income-types/capital-gains-and-losses.html#:~:text=California%20does%20not%20have%20a
         are%20taxed%20as%20ordinary%20income
          https://www.ftb.ca.gov/file/personal/
        
        Iowa:
        -         
        - https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
          https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
          https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
          https://tax.iowa.gov/iowa-illinois-reciprocal-agreement
          https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
        """
        
        # Check if the parameters are set upstream in the instance
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'California', 'state_occupation': 'Iowa', "capital_gains_short": capital_gains_short, "income": income, "dependents": dependents, "sources": sources, "county_residence": county_residence, "state_agi": state_agi, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def california_and_kansas(self, capital_gains_short=0, income=100000, dependents=0, sources=False, state_agi=75000, filers_over_65=0, capital_gains_long=0, filing_status='single', credits_state=0):
        """
        --- Parameters ----
        
        California:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        
        Kansas:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        California:
        - https://www.ftb.ca.gov/forms/2022/2022-540-tax-rate-schedules.pdf
          https://www.ftb.ca.gov/file/personal/deductions/index.html
          https://www.ftb.ca.gov/about-ftb/newsroom/caleitc/eligibility-and-credit-information.html
          https://www.ftb.ca.gov/file/personal/income-types/capital-gains-and-losses.html#:~:text=California%20does%20not%20have%20a
         are%20taxed%20as%20ordinary%20income
          https://www.ftb.ca.gov/file/personal/
        
        Kansas:
        -         
        - https://www.ksrevenue.gov/taxrates.html
          https://www.ksrevenue.gov/incomebook22.html#0
          https://www.ksrevenue.gov/incomebook22.html#0
          https://www.ksrevenue.gov/incomebook22.html#0
          https://www.ksrevenue.gov/incomebook22.html
          https://www.ksrevenue.gov/incomebook22.html
        """
        
        # Check if the parameters are set upstream in the instance
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'California', 'state_occupation': 'Kansas', "capital_gains_short": capital_gains_short, "income": income, "dependents": dependents, "sources": sources, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def california_and_kentucky(self, county_residence='Other', state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', sources=False, credits_state=0):
        """
        --- Parameters ----
        
        California:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        
        Kentucky:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        California:
        - https://www.ftb.ca.gov/forms/2022/2022-540-tax-rate-schedules.pdf
          https://www.ftb.ca.gov/file/personal/deductions/index.html
          https://www.ftb.ca.gov/about-ftb/newsroom/caleitc/eligibility-and-credit-information.html
          https://www.ftb.ca.gov/file/personal/income-types/capital-gains-and-losses.html#:~:text=California%20does%20not%20have%20a
         are%20taxed%20as%20ordinary%20income
          https://www.ftb.ca.gov/file/personal/
        
        Kentucky:
        -         
        - https://revenue.ky.gov/News/Pages/Dor-Announces-Updates-to-Individual-Income-Tax-for-2023-Tax-Year.aspx
          https://revenue.ky.gov/Individual/Individual-Income-Tax/Pages/default.aspx
          https://revenue.ky.gov/News/Pages/Dor-Announces-Updates-to-Individual-Income-Tax-for-2023-Tax-Year.aspx
          https://revenue.ky.gov/Individual/Individual-Income-Tax/Pages/default.aspx
          https://revenue.ky.gov/Dor%20Training%20Materials/103%20KAR%2017.140.%20Individual%20income%20tax%20-%20reciprocity%20-%20nonresidents.pdf
          https://revenue.ky.gov/Individual/Individual-Income-Tax/Pages/default.aspx
          https://nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2021/Taxes-21-32.htm?taxmap=true
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'California', 'state_occupation': 'Kentucky', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def california_and_louisiana(self, capital_gains_short=0, income=100000, dependents=0, sources=False, state_agi=75000, filers_over_65=0, capital_gains_long=0, filing_status='single', credits_state=0):
        """
        --- Parameters ----
        
        California:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        
        Louisiana:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        California:
        - https://www.ftb.ca.gov/forms/2022/2022-540-tax-rate-schedules.pdf
          https://www.ftb.ca.gov/file/personal/deductions/index.html
          https://www.ftb.ca.gov/about-ftb/newsroom/caleitc/eligibility-and-credit-information.html
          https://www.ftb.ca.gov/file/personal/income-types/capital-gains-and-losses.html#:~:text=California%20does%20not%20have%20a
         are%20taxed%20as%20ordinary%20income
          https://www.ftb.ca.gov/file/personal/
        
        Louisiana:
        -         
        - https://revenue.louisiana.gov/individualincometax
          https://revenue.louisiana.gov/individualincometax
          https://www.revenue.louisiana.gov/taxforms/6935(11_02)F.pdf
          https://revenue.louisiana.gov/individualincometax
          https://revenue.louisiana.gov/individualincometax
        """
        
        # Check if the parameters are set upstream in the instance
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'California', 'state_occupation': 'Louisiana', "capital_gains_short": capital_gains_short, "income": income, "dependents": dependents, "sources": sources, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def california_and_maine(self, capital_gains_short=0, income=100000, dependents=0, sources=False, state_agi=75000, filers_over_65=0, capital_gains_long=0, filing_status='single', credits_state=0):
        """
        --- Parameters ----
        
        California:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        
        Maine:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        California:
        - https://www.ftb.ca.gov/forms/2022/2022-540-tax-rate-schedules.pdf
          https://www.ftb.ca.gov/file/personal/deductions/index.html
          https://www.ftb.ca.gov/about-ftb/newsroom/caleitc/eligibility-and-credit-information.html
          https://www.ftb.ca.gov/file/personal/income-types/capital-gains-and-losses.html#:~:text=California%20does%20not%20have%20a
         are%20taxed%20as%20ordinary%20income
          https://www.ftb.ca.gov/file/personal/
        
        Maine:
        -         
        - https://www.maine.gov/revenue/tax-return-forms/individual-income-tax-2022
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/ind_tax_rate_sched_2022.pdf
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/22_1040me_book_gen_instruc_revisedFeb23.pdf
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/22_1040me_book_gen_instruc_revisedFeb23.pdf
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/22_1040me_book_gen_instruc_revisedFeb23.pdf
          https://www.maine.gov/revenue/tax-return-forms/individual-income-tax-2022
        """
        
        # Check if the parameters are set upstream in the instance
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'California', 'state_occupation': 'Maine', "capital_gains_short": capital_gains_short, "income": income, "dependents": dependents, "sources": sources, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def california_and_maryland(self, capital_gains_short=0, income=100000, dependents=0, sources=False, county_residence='Other', state_agi=75000, capital_gains_long=0, filing_status='single', credits_state=0):
        """
        --- Parameters ----
        
        California:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        
        Maryland:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        California:
        - https://www.ftb.ca.gov/forms/2022/2022-540-tax-rate-schedules.pdf
          https://www.ftb.ca.gov/file/personal/deductions/index.html
          https://www.ftb.ca.gov/about-ftb/newsroom/caleitc/eligibility-and-credit-information.html
          https://www.ftb.ca.gov/file/personal/income-types/capital-gains-and-losses.html#:~:text=California%20does%20not%20have%20a
         are%20taxed%20as%20ordinary%20income
          https://www.ftb.ca.gov/file/personal/
        
        Maryland:
        -         
        - https://www.marylandtaxes.gov/individual/income/tax-info/tax-rates.php
          https://www.marylandtaxes.gov/forms/22_forms/Resident_Booklet.pdf
          https://www.marylandtaxes.gov/forms/Tax_Publications/Administrative_Releases/Income_and_Estate_Tax_Releases/ar_it37.pdf
          https://www.marylandtaxes.gov/individual/index.php
          https://www.marylandtaxes.gov/individual/income/tax-info/tax-rates.php
        """
        
        # Check if the parameters are set upstream in the instance
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'California', 'state_occupation': 'Maryland', "capital_gains_short": capital_gains_short, "income": income, "dependents": dependents, "sources": sources, "county_residence": county_residence, "state_agi": state_agi, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def california_and_massachusetts(self, state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        California:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        
        Massachusetts:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        California:
        - https://www.ftb.ca.gov/forms/2022/2022-540-tax-rate-schedules.pdf
          https://www.ftb.ca.gov/file/personal/deductions/index.html
          https://www.ftb.ca.gov/about-ftb/newsroom/caleitc/eligibility-and-credit-information.html
          https://www.ftb.ca.gov/file/personal/income-types/capital-gains-and-losses.html#:~:text=California%20does%20not%20have%20a
         are%20taxed%20as%20ordinary%20income
          https://www.ftb.ca.gov/file/personal/
        
        Massachusetts:
        -         
        - https://www.mass.gov/service-details/massachusetts-tax-rates
          https://www.mass.gov/service-details/view-massachusetts-personal-income-tax-exemptions
          https://www.mass.gov/service-details/view-massachusetts-personal-income-tax-exemptions
          https://www.mass.gov/service-details/view-child-and-dependent-related-credits
          https://www.mass.gov/service-details/massachusetts-tax-rates
          https://www.mass.gov/personal-income-tax
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'California', 'state_occupation': 'Massachusetts', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def california_and_michigan(self, capital_gains_short=0, income=100000, county_occupation='Other', dependents=0, sources=False, county_residence='Other', state_agi=75000, capital_gains_long=0, filing_status='single', credits_state=0):
        """
        --- Parameters ----
        
        California:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        
        Michigan:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        county_occupation : str, optional
            County of occupation. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        California:
        - https://www.ftb.ca.gov/forms/2022/2022-540-tax-rate-schedules.pdf
          https://www.ftb.ca.gov/file/personal/deductions/index.html
          https://www.ftb.ca.gov/about-ftb/newsroom/caleitc/eligibility-and-credit-information.html
          https://www.ftb.ca.gov/file/personal/income-types/capital-gains-and-losses.html#:~:text=California%20does%20not%20have%20a
         are%20taxed%20as%20ordinary%20income
          https://www.ftb.ca.gov/file/personal/
        
        Michigan:
        -         
        - https://www.michigan.gov/taxes/iit/new-developments-for-tax-year-2022
          "https://www.michigan.gov/taxes/-/media/Project/Websites/taxes/Forms/2023/2023-SUW/446_Withholding-Guide_2023.pdf?rev=177ca3d8cb034392bf84e317ed78c320&hash=19C5F42F43A89A91863D229349F48883#:~:text=The%20exemption%20amount%20is%20%245%2C400
         employees%20Michigan%20income%20tax%20return"
          "https://www.michigan.gov/taxes/-/media/Project/Websites/taxes/Forms/2023/2023-SUW/446_Withholding-Guide_2023.pdf?rev=177ca3d8cb034392bf84e317ed78c320&hash=19C5F42F43A89A91863D229349F48883#:~:text=The%20exemption%20amount%20is%20%245%2C400
         employees%20Michigan%20income%20tax%20return"
          https://www.michigan.gov/taxes/questions/iit/accordion/residency/are-my-wages-earned-in-another-state-taxable-in-michigan-if-i-am-a-michigan-resident-1
          https://www.michigan.gov/taxes/-/media/Project/Websites/taxes/Forms/2022/2022-IIT-Forms/BOOK_MI-1040.pdf?rev=4524f1cbbc0243b194ab05a6680be75b&hash=40DFE71E651F7C1D65C31F6CD03D944F
          https://www.michigan.gov/taxes/questions/iit/accordion/general/what-cities-impose-an-income-tax
        """
        
        # Check if the parameters are set upstream in the instance
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        county_occupation = self.county_occupation if hasattr(self, 'county_occupation') else county_occupation
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'California', 'state_occupation': 'Michigan', "capital_gains_short": capital_gains_short, "income": income, "county_occupation": county_occupation, "dependents": dependents, "sources": sources, "county_residence": county_residence, "state_agi": state_agi, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def california_and_minnesota(self, state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        California:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        
        Minnesota:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        California:
        - https://www.ftb.ca.gov/forms/2022/2022-540-tax-rate-schedules.pdf
          https://www.ftb.ca.gov/file/personal/deductions/index.html
          https://www.ftb.ca.gov/about-ftb/newsroom/caleitc/eligibility-and-credit-information.html
          https://www.ftb.ca.gov/file/personal/income-types/capital-gains-and-losses.html#:~:text=California%20does%20not%20have%20a
         are%20taxed%20as%20ordinary%20income
          https://www.ftb.ca.gov/file/personal/
        
        Minnesota:
        -         
        - https://www.revenue.state.mn.us/minnesota-income-tax-rates-and-brackets
          https://www.revenue.state.mn.us/child-and-dependent-care-credit
          https://www.revenue.state.mn.us/dependent-exemptions
          https://www.revenue.state.mn.us/minnesota-standard-deduction
          https://mn.gov/mmb/assets/mwr_form_tcm1059-128581.pdf
          https://www.revenue.state.mn.us/individual-income-tax
          https://www.revenue.state.mn.us/individual-income-tax
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'California', 'state_occupation': 'Minnesota', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def california_and_mississippi(self, state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        California:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        
        Mississippi:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        California:
        - https://www.ftb.ca.gov/forms/2022/2022-540-tax-rate-schedules.pdf
          https://www.ftb.ca.gov/file/personal/deductions/index.html
          https://www.ftb.ca.gov/about-ftb/newsroom/caleitc/eligibility-and-credit-information.html
          https://www.ftb.ca.gov/file/personal/income-types/capital-gains-and-losses.html#:~:text=California%20does%20not%20have%20a
         are%20taxed%20as%20ordinary%20income
          https://www.ftb.ca.gov/file/personal/
        
        Mississippi:
        -         
        - https://www.dor.ms.gov/individual/tax-rates
          https://www.dor.ms.gov/individual/tax-rates
          https://www.dor.ms.gov/individual/tax-rates
          https://www.dor.ms.gov/individual
          https://www.dor.ms.gov/individual
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'California', 'state_occupation': 'Mississippi', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def california_and_missouri(self, capital_gains_short=0, income=100000, dependents=0, sources=False, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_long=0, filing_status='single', credits_state=0):
        """
        --- Parameters ----
        
        California:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        
        Missouri:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        California:
        - https://www.ftb.ca.gov/forms/2022/2022-540-tax-rate-schedules.pdf
          https://www.ftb.ca.gov/file/personal/deductions/index.html
          https://www.ftb.ca.gov/about-ftb/newsroom/caleitc/eligibility-and-credit-information.html
          https://www.ftb.ca.gov/file/personal/income-types/capital-gains-and-losses.html#:~:text=California%20does%20not%20have%20a
         are%20taxed%20as%20ordinary%20income
          https://www.ftb.ca.gov/file/personal/
        
        Missouri:
        -         
        - https://dor.mo.gov/taxation/individual/tax-types/income/year-changes/
          https://dor.mo.gov/taxation/individual/tax-types/income/year-changes/
          https://dor.mo.gov/taxation/individual/tax-types/income/year-changes/
          https://www.kcmo.gov/city-hall/departments/finance/earnings-tax
          https://www.stlouis-mo.gov/government/departments/collector/earnings-tax/individual-earnings-tax-info.cfm
        """
        
        # Check if the parameters are set upstream in the instance
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'California', 'state_occupation': 'Missouri', "capital_gains_short": capital_gains_short, "income": income, "dependents": dependents, "sources": sources, "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def california_and_montana(self, capital_gains_short=0, income=100000, dependents=0, sources=False, state_agi=75000, filers_over_65=0, capital_gains_long=0, filing_status='single', credits_state=0):
        """
        --- Parameters ----
        
        California:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        
        Montana:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        California:
        - https://www.ftb.ca.gov/forms/2022/2022-540-tax-rate-schedules.pdf
          https://www.ftb.ca.gov/file/personal/deductions/index.html
          https://www.ftb.ca.gov/about-ftb/newsroom/caleitc/eligibility-and-credit-information.html
          https://www.ftb.ca.gov/file/personal/income-types/capital-gains-and-losses.html#:~:text=California%20does%20not%20have%20a
         are%20taxed%20as%20ordinary%20income
          https://www.ftb.ca.gov/file/personal/
        
        Montana:
        -         
        - https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/2022/12/montana-employees-withholding-allowance-and-exemption-certificate-form-mw-4.pdf
          https://mtrevenue.gov/2021/11/29/simplification-of-montana-income-taxation/?utm_source=rss&utm_medium=rss&utm_campaign=simplification-of-montana-income-taxation
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'California', 'state_occupation': 'Montana', "capital_gains_short": capital_gains_short, "income": income, "dependents": dependents, "sources": sources, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def california_and_nebraska(self, state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        California:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        
        Nebraska:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        California:
        - https://www.ftb.ca.gov/forms/2022/2022-540-tax-rate-schedules.pdf
          https://www.ftb.ca.gov/file/personal/deductions/index.html
          https://www.ftb.ca.gov/about-ftb/newsroom/caleitc/eligibility-and-credit-information.html
          https://www.ftb.ca.gov/file/personal/income-types/capital-gains-and-losses.html#:~:text=California%20does%20not%20have%20a
         are%20taxed%20as%20ordinary%20income
          https://www.ftb.ca.gov/file/personal/
        
        Nebraska:
        -         
        - https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2023_Circular_EN_8-429r2023_Whole_Book_11.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'California', 'state_occupation': 'Nebraska', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def california_and_nevada(self, capital_gains_long=0, state_agi=75000, filing_status='single', income=100000, sources=False, capital_gains_short=0, credits_state=0):
        """
        --- Parameters ----
        
        California:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        
        Nevada:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        California:
        - https://www.ftb.ca.gov/forms/2022/2022-540-tax-rate-schedules.pdf
          https://www.ftb.ca.gov/file/personal/deductions/index.html
          https://www.ftb.ca.gov/about-ftb/newsroom/caleitc/eligibility-and-credit-information.html
          https://www.ftb.ca.gov/file/personal/income-types/capital-gains-and-losses.html#:~:text=California%20does%20not%20have%20a
         are%20taxed%20as%20ordinary%20income
          https://www.ftb.ca.gov/file/personal/
        
        Nevada:
        -         
        - https://www.nvsos.gov/sos/businesses/the-nevada-advantage
        """
        
        # Check if the parameters are set upstream in the instance
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        sources = self.sources if hasattr(self, 'sources') else sources
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'California', 'state_occupation': 'Nevada', "capital_gains_long": capital_gains_long, "state_agi": state_agi, "filing_status": filing_status, "income": income, "sources": sources, "capital_gains_short": capital_gains_short, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def california_and_new_hampshire(self, capital_gains_long=0, state_agi=75000, filing_status='single', income=100000, sources=False, capital_gains_short=0, credits_state=0):
        """
        --- Parameters ----
        
        California:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        
        New Hampshire:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        California:
        - https://www.ftb.ca.gov/forms/2022/2022-540-tax-rate-schedules.pdf
          https://www.ftb.ca.gov/file/personal/deductions/index.html
          https://www.ftb.ca.gov/about-ftb/newsroom/caleitc/eligibility-and-credit-information.html
          https://www.ftb.ca.gov/file/personal/income-types/capital-gains-and-losses.html#:~:text=California%20does%20not%20have%20a
         are%20taxed%20as%20ordinary%20income
          https://www.ftb.ca.gov/file/personal/
        
        New Hampshire:
        -         
        - https://www.revenue.nh.gov/assistance/tax-overview.htm
        """
        
        # Check if the parameters are set upstream in the instance
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        sources = self.sources if hasattr(self, 'sources') else sources
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'California', 'state_occupation': 'NewHampshire', "capital_gains_long": capital_gains_long, "state_agi": state_agi, "filing_status": filing_status, "income": income, "sources": sources, "capital_gains_short": capital_gains_short, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def california_and_new_jersey(self, capital_gains_short=0, income=100000, dependents=0, sources=False, state_agi=75000, filers_over_65=0, capital_gains_long=0, filing_status='single', credits_state=0):
        """
        --- Parameters ----
        
        California:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        
        New Jersey:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        California:
        - https://www.ftb.ca.gov/forms/2022/2022-540-tax-rate-schedules.pdf
          https://www.ftb.ca.gov/file/personal/deductions/index.html
          https://www.ftb.ca.gov/about-ftb/newsroom/caleitc/eligibility-and-credit-information.html
          https://www.ftb.ca.gov/file/personal/income-types/capital-gains-and-losses.html#:~:text=California%20does%20not%20have%20a
         are%20taxed%20as%20ordinary%20income
          https://www.ftb.ca.gov/file/personal/
        
        New Jersey:
        -         
        - https://www.state.nj.us/treasury/taxation/taxtables.shtml
          https://www.state.nj.us/treasury/taxation/njit13.shtml
          https://www.nj.gov/treasury/taxation/njit25.shtml
          https://www.state.nj.us/treasury/taxation/individuals/freqqiti.shtml
          https://www.state.nj.us/treasury/taxation/localtax.shtml
        """
        
        # Check if the parameters are set upstream in the instance
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'California', 'state_occupation': 'NewJersey', "capital_gains_short": capital_gains_short, "income": income, "dependents": dependents, "sources": sources, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def california_and_new_mexico(self, state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        California:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        
        New Mexico:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        California:
        - https://www.ftb.ca.gov/forms/2022/2022-540-tax-rate-schedules.pdf
          https://www.ftb.ca.gov/file/personal/deductions/index.html
          https://www.ftb.ca.gov/about-ftb/newsroom/caleitc/eligibility-and-credit-information.html
          https://www.ftb.ca.gov/file/personal/income-types/capital-gains-and-losses.html#:~:text=California%20does%20not%20have%20a
         are%20taxed%20as%20ordinary%20income
          https://www.ftb.ca.gov/file/personal/
        
        New Mexico:
        -         
        - https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/fdf3c548-8aba-4b9c-9eb4-bb564c716015/FYI-104.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/0bf943fd-652e-4400-bb1b-cd8397eb5a95/2022pit-adj.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/2f1a6781-9534-4436-b427-1557f9592099/2022pit-adj-ins.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/0bf943fd-652e-4400-bb1b-cd8397eb5a95/2022pit-adj.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/1afc56af-ea90-4d48-82e5-1f9aeb43255a/PITbook2022.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/fdf3c548-8aba-4b9c-9eb4-bb564c716015/FYI-104.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'California', 'state_occupation': 'NewMexico', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def california_and_new_york(self, capital_gains_short=0, income=100000, dependents=0, sources=False, county_residence='Other', state_agi=75000, capital_gains_long=0, filing_status='single', credits_state=0):
        """
        --- Parameters ----
        
        California:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        
        New York:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        California:
        - https://www.ftb.ca.gov/forms/2022/2022-540-tax-rate-schedules.pdf
          https://www.ftb.ca.gov/file/personal/deductions/index.html
          https://www.ftb.ca.gov/about-ftb/newsroom/caleitc/eligibility-and-credit-information.html
          https://www.ftb.ca.gov/file/personal/income-types/capital-gains-and-losses.html#:~:text=California%20does%20not%20have%20a
         are%20taxed%20as%20ordinary%20income
          https://www.ftb.ca.gov/file/personal/
        
        New York:
        -         
        - https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nys-tax-rate-schedule
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nys-tax-rate-schedule
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nys-tax-rate-schedule
          https://www.tax.ny.gov/pit/
          https://www.tax.ny.gov/pit/file/nonresident-faqs.htm
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#yonkers-tax-rate-schedule
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nyc-tax-rate-schedule
        """
        
        # Check if the parameters are set upstream in the instance
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'California', 'state_occupation': 'NewYork', "capital_gains_short": capital_gains_short, "income": income, "dependents": dependents, "sources": sources, "county_residence": county_residence, "state_agi": state_agi, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def california_and_north_carolina(self, state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', sources=False, credits_state=0):
        """
        --- Parameters ----
        
        California:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        
        North Carolina:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        California:
        - https://www.ftb.ca.gov/forms/2022/2022-540-tax-rate-schedules.pdf
          https://www.ftb.ca.gov/file/personal/deductions/index.html
          https://www.ftb.ca.gov/about-ftb/newsroom/caleitc/eligibility-and-credit-information.html
          https://www.ftb.ca.gov/file/personal/income-types/capital-gains-and-losses.html#:~:text=California%20does%20not%20have%20a
         are%20taxed%20as%20ordinary%20income
          https://www.ftb.ca.gov/file/personal/
        
        North Carolina:
        -         
        - https://www.ncdor.gov/taxes-forms/tax-rate-schedules
          https://www.ncdor.gov/taxes-forms/individual-income-tax/north-carolina-standard-deduction-or-north-carolina-itemized-deductions
          https://www.ncdor.gov/taxes-forms/individual-income-tax
          https://www.ncdor.gov/taxes-forms/individual-income-tax
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'California', 'state_occupation': 'NorthCarolina', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def california_and_north_dakota(self, capital_gains_short=0, income=100000, dependents=0, sources=False, state_agi=75000, filers_over_65=0, capital_gains_long=0, filing_status='single', credits_state=0):
        """
        --- Parameters ----
        
        California:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        
        North Dakota:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        California:
        - https://www.ftb.ca.gov/forms/2022/2022-540-tax-rate-schedules.pdf
          https://www.ftb.ca.gov/file/personal/deductions/index.html
          https://www.ftb.ca.gov/about-ftb/newsroom/caleitc/eligibility-and-credit-information.html
          https://www.ftb.ca.gov/file/personal/income-types/capital-gains-and-losses.html#:~:text=California%20does%20not%20have%20a
         are%20taxed%20as%20ordinary%20income
          https://www.ftb.ca.gov/file/personal/
        
        North Dakota:
        -         
        - https://www.tax.nd.gov/forms
          https://www.tax.nd.gov/sites/www/files/documents/forms/individual/2022-iit/2022-individual-income-tax-booklet.pdf
          https://www.tax.nd.gov/sites/www/files/documents/forms/2022-individual-income-tax-booklet.pdf
          https://www.tax.nd.gov/tax-exemptions-credits/income-tax-exemptions-credits
          https://www.tax.nd.gov/sites/www/files/documents/forms/business/ndwrfillable.pdf
          https://www.tax.nd.gov/news/tax-legislative-changes/significant-changes-law/individual-income-tax-history
          https://www.tax.nd.gov/forms
        """
        
        # Check if the parameters are set upstream in the instance
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'California', 'state_occupation': 'NorthDakota', "capital_gains_short": capital_gains_short, "income": income, "dependents": dependents, "sources": sources, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def california_and_ohio(self, capital_gains_short=0, income=100000, dependents=0, sources=False, county_residence='Other', state_agi=75000, capital_gains_long=0, filing_status='single', credits_state=0):
        """
        --- Parameters ----
        
        California:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        
        Ohio:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        California:
        - https://www.ftb.ca.gov/forms/2022/2022-540-tax-rate-schedules.pdf
          https://www.ftb.ca.gov/file/personal/deductions/index.html
          https://www.ftb.ca.gov/about-ftb/newsroom/caleitc/eligibility-and-credit-information.html
          https://www.ftb.ca.gov/file/personal/income-types/capital-gains-and-losses.html#:~:text=California%20does%20not%20have%20a
         are%20taxed%20as%20ordinary%20income
          https://www.ftb.ca.gov/file/personal/
        
        Ohio:
        -         
        - https://tax.ohio.gov/individual/resources/annual-tax-rates
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://tax.ohio.gov/static/forms/employer_withholding/generic/wth_it4nr.pdf
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://www.ritaohio.com/TaxRatesTable
        """
        
        # Check if the parameters are set upstream in the instance
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'California', 'state_occupation': 'Ohio', "capital_gains_short": capital_gains_short, "income": income, "dependents": dependents, "sources": sources, "county_residence": county_residence, "state_agi": state_agi, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def california_and_oklahoma(self, capital_gains_short=0, income=100000, dependents=0, sources=False, state_agi=75000, filers_over_65=0, capital_gains_long=0, filing_status='single', credits_state=0):
        """
        --- Parameters ----
        
        California:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        
        Oklahoma:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        California:
        - https://www.ftb.ca.gov/forms/2022/2022-540-tax-rate-schedules.pdf
          https://www.ftb.ca.gov/file/personal/deductions/index.html
          https://www.ftb.ca.gov/about-ftb/newsroom/caleitc/eligibility-and-credit-information.html
          https://www.ftb.ca.gov/file/personal/income-types/capital-gains-and-losses.html#:~:text=California%20does%20not%20have%20a
         are%20taxed%20as%20ordinary%20income
          https://www.ftb.ca.gov/file/personal/
        
        Oklahoma:
        -         
        - https://oklahoma.gov/content/dam/ok/en/tax/documents/resources/publications/businesses/withholding-tables/WHTables-2022.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'California', 'state_occupation': 'Oklahoma', "capital_gains_short": capital_gains_short, "income": income, "dependents": dependents, "sources": sources, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def california_and_oregon(self, capital_gains_short=0, income=100000, dependents=0, sources=False, state_agi=75000, filers_over_65=0, capital_gains_long=0, filing_status='single', credits_state=0):
        """
        --- Parameters ----
        
        California:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        
        Oregon:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        California:
        - https://www.ftb.ca.gov/forms/2022/2022-540-tax-rate-schedules.pdf
          https://www.ftb.ca.gov/file/personal/deductions/index.html
          https://www.ftb.ca.gov/about-ftb/newsroom/caleitc/eligibility-and-credit-information.html
          https://www.ftb.ca.gov/file/personal/income-types/capital-gains-and-losses.html#:~:text=California%20does%20not%20have%20a
         are%20taxed%20as%20ordinary%20income
          https://www.ftb.ca.gov/file/personal/
        
        Oregon:
        -         
        - https://www.oregon.gov/dor/programs/individuals/Pages/PIT.aspx
          https://www.oregon.gov/dor/programs/individuals/Documents/Full%20year%20resident
         %20Form%20or-40%20filers.pdf
          https://www.oregon.gov/dor/programs/individuals/Documents/Part-year%20and%20nonresident
         %20Form%20or-40-P%20and%20or-40-N%20filers.pdf.pdf
          https://www.oregon.gov/dor/forms/FormsPubs/form-or-40_101-040_2022.pdf
          https://www.oregon.gov/dor/forms/FormsPubs/form-or-40_101-040_2022.pdf
          https://www.oregon.gov/dor/forms/FormsPubs/form-or-40_101-040_2022.pdf
          https://www.portland.gov/revenue/personal-tax
        """
        
        # Check if the parameters are set upstream in the instance
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'California', 'state_occupation': 'Oregon', "capital_gains_short": capital_gains_short, "income": income, "dependents": dependents, "sources": sources, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def california_and_pennsylvania(self, county_residence='Other', state_agi=75000, county_occupation='Other', capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', sources=False, credits_state=0):
        """
        --- Parameters ----
        
        California:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        
        Pennsylvania:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        county_residence : str, optional
            County of residence. Default is Other.
        county_occupation : str, optional
            County of occupation. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        California:
        - https://www.ftb.ca.gov/forms/2022/2022-540-tax-rate-schedules.pdf
          https://www.ftb.ca.gov/file/personal/deductions/index.html
          https://www.ftb.ca.gov/about-ftb/newsroom/caleitc/eligibility-and-credit-information.html
          https://www.ftb.ca.gov/file/personal/income-types/capital-gains-and-losses.html#:~:text=California%20does%20not%20have%20a
         are%20taxed%20as%20ordinary%20income
          https://www.ftb.ca.gov/file/personal/
        
        Pennsylvania:
        -         
        - https://www.revenue.pa.gov/Tax%20Rates/Pages/default.aspx
          https://www.revenue.pa.gov/FormsandPublications/FormsforIndividuals/PIT/Documents/2022/2022_pa-40in.pdf
          https://www.revenue.pa.gov/FormsandPublications/FormsforBusinesses/EmployerWithholding/Documents/rev-419.pdf
          https://www.revenue.pa.gov/FormsandPublications/FormsforIndividuals/PIT/Documents/2022/2022_pa-40in.pdf
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-25.htm
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        county_occupation = self.county_occupation if hasattr(self, 'county_occupation') else county_occupation
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'California', 'state_occupation': 'Pennsylvania', "county_residence": county_residence, "state_agi": state_agi, "county_occupation": county_occupation, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def california_and_rhode_island(self, state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        California:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        
        Rhode Island:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        California:
        - https://www.ftb.ca.gov/forms/2022/2022-540-tax-rate-schedules.pdf
          https://www.ftb.ca.gov/file/personal/deductions/index.html
          https://www.ftb.ca.gov/about-ftb/newsroom/caleitc/eligibility-and-credit-information.html
          https://www.ftb.ca.gov/file/personal/income-types/capital-gains-and-losses.html#:~:text=California%20does%20not%20have%20a
         are%20taxed%20as%20ordinary%20income
          https://www.ftb.ca.gov/file/personal/
        
        Rhode Island:
        -         
        - https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2022_1040WE_w_0.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'California', 'state_occupation': 'RhodeIsland', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def california_and_south_carolina(self, capital_gains_short=0, income=100000, dependents=0, sources=False, state_agi=75000, filers_over_65=0, capital_gains_long=0, filing_status='single', credits_state=0):
        """
        --- Parameters ----
        
        California:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        
        South Carolina:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        California:
        - https://www.ftb.ca.gov/forms/2022/2022-540-tax-rate-schedules.pdf
          https://www.ftb.ca.gov/file/personal/deductions/index.html
          https://www.ftb.ca.gov/about-ftb/newsroom/caleitc/eligibility-and-credit-information.html
          https://www.ftb.ca.gov/file/personal/income-types/capital-gains-and-losses.html#:~:text=California%20does%20not%20have%20a
         are%20taxed%20as%20ordinary%20income
          https://www.ftb.ca.gov/file/personal/
        
        South Carolina:
        -         
        - https://dor.sc.gov/resources-site/lawandpolicy/Advisory%20Opinions/IL22-15.pdf
          https://dor.sc.gov/forms-site/Forms/IITPacket_2022.pdf
          https://www.scstatehouse.gov/code/t12c006.php
          https://dor.sc.gov/resources-site/lawandpolicy/Advisory%20Opinions/IL22-15.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'California', 'state_occupation': 'SouthCarolina', "capital_gains_short": capital_gains_short, "income": income, "dependents": dependents, "sources": sources, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def california_and_south_dakota(self, capital_gains_long=0, state_agi=75000, filing_status='single', income=100000, sources=False, capital_gains_short=0, credits_state=0):
        """
        --- Parameters ----
        
        California:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        
        South Dakota:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        California:
        - https://www.ftb.ca.gov/forms/2022/2022-540-tax-rate-schedules.pdf
          https://www.ftb.ca.gov/file/personal/deductions/index.html
          https://www.ftb.ca.gov/about-ftb/newsroom/caleitc/eligibility-and-credit-information.html
          https://www.ftb.ca.gov/file/personal/income-types/capital-gains-and-losses.html#:~:text=California%20does%20not%20have%20a
         are%20taxed%20as%20ordinary%20income
          https://www.ftb.ca.gov/file/personal/
        
        South Dakota:
        -         
        - https://dor.sd.gov/individuals/taxes/
        """
        
        # Check if the parameters are set upstream in the instance
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        sources = self.sources if hasattr(self, 'sources') else sources
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'California', 'state_occupation': 'SouthDakota', "capital_gains_long": capital_gains_long, "state_agi": state_agi, "filing_status": filing_status, "income": income, "sources": sources, "capital_gains_short": capital_gains_short, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def california_and_tennessee(self, capital_gains_long=0, state_agi=75000, filing_status='single', income=100000, sources=False, capital_gains_short=0, credits_state=0):
        """
        --- Parameters ----
        
        California:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        
        Tennessee:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        California:
        - https://www.ftb.ca.gov/forms/2022/2022-540-tax-rate-schedules.pdf
          https://www.ftb.ca.gov/file/personal/deductions/index.html
          https://www.ftb.ca.gov/about-ftb/newsroom/caleitc/eligibility-and-credit-information.html
          https://www.ftb.ca.gov/file/personal/income-types/capital-gains-and-losses.html#:~:text=California%20does%20not%20have%20a
         are%20taxed%20as%20ordinary%20income
          https://www.ftb.ca.gov/file/personal/
        
        Tennessee:
        -         
        - https://www.tn.gov/revenue/taxes.html
        """
        
        # Check if the parameters are set upstream in the instance
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        sources = self.sources if hasattr(self, 'sources') else sources
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'California', 'state_occupation': 'Tennessee', "capital_gains_long": capital_gains_long, "state_agi": state_agi, "filing_status": filing_status, "income": income, "sources": sources, "capital_gains_short": capital_gains_short, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def california_and_texas(self, capital_gains_long=0, state_agi=75000, filing_status='single', income=100000, sources=False, capital_gains_short=0, credits_state=0):
        """
        --- Parameters ----
        
        California:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        
        Texas:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        California:
        - https://www.ftb.ca.gov/forms/2022/2022-540-tax-rate-schedules.pdf
          https://www.ftb.ca.gov/file/personal/deductions/index.html
          https://www.ftb.ca.gov/about-ftb/newsroom/caleitc/eligibility-and-credit-information.html
          https://www.ftb.ca.gov/file/personal/income-types/capital-gains-and-losses.html#:~:text=California%20does%20not%20have%20a
         are%20taxed%20as%20ordinary%20income
          https://www.ftb.ca.gov/file/personal/
        
        Texas:
        -         
        - https://comptroller.texas.gov/taxes/
        """
        
        # Check if the parameters are set upstream in the instance
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        sources = self.sources if hasattr(self, 'sources') else sources
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'California', 'state_occupation': 'Texas', "capital_gains_long": capital_gains_long, "state_agi": state_agi, "filing_status": filing_status, "income": income, "sources": sources, "capital_gains_short": capital_gains_short, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def california_and_utah(self, state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        California:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        
        Utah:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        California:
        - https://www.ftb.ca.gov/forms/2022/2022-540-tax-rate-schedules.pdf
          https://www.ftb.ca.gov/file/personal/deductions/index.html
          https://www.ftb.ca.gov/about-ftb/newsroom/caleitc/eligibility-and-credit-information.html
          https://www.ftb.ca.gov/file/personal/income-types/capital-gains-and-losses.html#:~:text=California%20does%20not%20have%20a
         are%20taxed%20as%20ordinary%20income
          https://www.ftb.ca.gov/file/personal/
        
        Utah:
        -         
        - https://incometax.utah.gov/paying/tax-rates
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'California', 'state_occupation': 'Utah', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def california_and_vermont(self, capital_gains_short=0, income=100000, dependents=0, sources=False, state_agi=75000, filers_over_65=0, capital_gains_long=0, filing_status='single', credits_state=0):
        """
        --- Parameters ----
        
        California:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        
        Vermont:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        California:
        - https://www.ftb.ca.gov/forms/2022/2022-540-tax-rate-schedules.pdf
          https://www.ftb.ca.gov/file/personal/deductions/index.html
          https://www.ftb.ca.gov/about-ftb/newsroom/caleitc/eligibility-and-credit-information.html
          https://www.ftb.ca.gov/file/personal/income-types/capital-gains-and-losses.html#:~:text=California%20does%20not%20have%20a
         are%20taxed%20as%20ordinary%20income
          https://www.ftb.ca.gov/file/personal/
        
        Vermont:
        -         
        - https://tax.vermont.gov/sites/tax/files/documents/GB-1210-2023.pdf
          https://tax.vermont.gov/sites/tax/files/documents/Income%20Booklet-2022.pdf
          https://tax.vermont.gov/sites/tax/files/documents/Income%20Booklet-2022.pdf
          https://tax.vermont.gov/sites/tax/files/documents/Income%20Booklet-2022.pdf
          https://tax.vermont.gov/individuals/personal-income-tax/tax-credits
          https://tax.vermont.gov/individuals/personal-income-tax/taxable-income
          https://tax.vermont.gov/individuals/personal-income-tax/local-option-tax
        """
        
        # Check if the parameters are set upstream in the instance
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'California', 'state_occupation': 'Vermont', "capital_gains_short": capital_gains_short, "income": income, "dependents": dependents, "sources": sources, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def california_and_virginia(self, capital_gains_short=0, income=100000, dependents=0, sources=False, state_agi=75000, filers_over_65=0, capital_gains_long=0, filing_status='single', credits_state=0):
        """
        --- Parameters ----
        
        California:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        
        Virginia:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        California:
        - https://www.ftb.ca.gov/forms/2022/2022-540-tax-rate-schedules.pdf
          https://www.ftb.ca.gov/file/personal/deductions/index.html
          https://www.ftb.ca.gov/about-ftb/newsroom/caleitc/eligibility-and-credit-information.html
          https://www.ftb.ca.gov/file/personal/income-types/capital-gains-and-losses.html#:~:text=California%20does%20not%20have%20a
         are%20taxed%20as%20ordinary%20income
          https://www.ftb.ca.gov/file/personal/
        
        Virginia:
        -         
        - https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/reciprocity
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'California', 'state_occupation': 'Virginia', "capital_gains_short": capital_gains_short, "income": income, "dependents": dependents, "sources": sources, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def california_and_washington(self, state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', sources=False, credits_state=0):
        """
        --- Parameters ----
        
        California:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        
        Washington:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        California:
        - https://www.ftb.ca.gov/forms/2022/2022-540-tax-rate-schedules.pdf
          https://www.ftb.ca.gov/file/personal/deductions/index.html
          https://www.ftb.ca.gov/about-ftb/newsroom/caleitc/eligibility-and-credit-information.html
          https://www.ftb.ca.gov/file/personal/income-types/capital-gains-and-losses.html#:~:text=California%20does%20not%20have%20a
         are%20taxed%20as%20ordinary%20income
          https://www.ftb.ca.gov/file/personal/
        
        Washington:
        -         
        - https://dor.wa.gov/taxes-rates/income-tax
          https://dor.wa.gov/taxes-rates/other-taxes/capital-gains-tax
          https://dor.wa.gov/taxes-rates/income-tax
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'California', 'state_occupation': 'Washington', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def california_and_west_virginia(self, pay_periods=24, capital_gains_short=0, income=100000, county_occupation='Other', dependents=0, sources=False, state_agi=75000, capital_gains_long=0, filing_status='single', credits_state=0):
        """
        --- Parameters ----
        
        California:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        
        West Virginia:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_occupation : str, optional
            County of occupation. Default is Other.
        pay_periods : int, optional
            Number of pay periods. Options are 13, 24, 26, 52. Default is 24.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        California:
        - https://www.ftb.ca.gov/forms/2022/2022-540-tax-rate-schedules.pdf
          https://www.ftb.ca.gov/file/personal/deductions/index.html
          https://www.ftb.ca.gov/about-ftb/newsroom/caleitc/eligibility-and-credit-information.html
          https://www.ftb.ca.gov/file/personal/income-types/capital-gains-and-losses.html#:~:text=California%20does%20not%20have%20a
         are%20taxed%20as%20ordinary%20income
          https://www.ftb.ca.gov/file/personal/
        
        West Virginia:
        -         
        - https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://tax.wv.gov/Documents/TSD/tsd438.pdf
          https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2011/Taxes-11-35att.html
        """
        
        # Check if the parameters are set upstream in the instance
        pay_periods = self.pay_periods if hasattr(self, 'pay_periods') else pay_periods
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        county_occupation = self.county_occupation if hasattr(self, 'county_occupation') else county_occupation
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'California', 'state_occupation': 'WestVirginia', "pay_periods": pay_periods, "capital_gains_short": capital_gains_short, "income": income, "county_occupation": county_occupation, "dependents": dependents, "sources": sources, "state_agi": state_agi, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def california_and_wisconsin(self, state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        California:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        
        Wisconsin:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        California:
        - https://www.ftb.ca.gov/forms/2022/2022-540-tax-rate-schedules.pdf
          https://www.ftb.ca.gov/file/personal/deductions/index.html
          https://www.ftb.ca.gov/about-ftb/newsroom/caleitc/eligibility-and-credit-information.html
          https://www.ftb.ca.gov/file/personal/income-types/capital-gains-and-losses.html#:~:text=California%20does%20not%20have%20a
         are%20taxed%20as%20ordinary%20income
          https://www.ftb.ca.gov/file/personal/
        
        Wisconsin:
        -         
        - https://www.revenue.wi.gov/TaxForms2023/2023-Form1-ES-Inst.pdf
          https://www.revenue.wi.gov/TaxForms2023/2023-Form1-ES-Inst.pdf
          https://www.revenue.wi.gov/TaxForms2023/2023-Form1-ES-Inst.pdf
          https://www.revenue.wi.gov/Pages/FAQS/pcs-work.aspx
          https://www.revenue.wi.gov/Dor%20Publications/pb103.pdf
          https://www.revenue.wi.gov/Pages/FAQS/pcs-work.aspx
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'California', 'state_occupation': 'Wisconsin', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def california_and_wyoming(self, capital_gains_long=0, state_agi=75000, filing_status='single', income=100000, sources=False, capital_gains_short=0, credits_state=0):
        """
        --- Parameters ----
        
        California:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        
        Wyoming:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        California:
        - https://www.ftb.ca.gov/forms/2022/2022-540-tax-rate-schedules.pdf
          https://www.ftb.ca.gov/file/personal/deductions/index.html
          https://www.ftb.ca.gov/about-ftb/newsroom/caleitc/eligibility-and-credit-information.html
          https://www.ftb.ca.gov/file/personal/income-types/capital-gains-and-losses.html#:~:text=California%20does%20not%20have%20a
         are%20taxed%20as%20ordinary%20income
          https://www.ftb.ca.gov/file/personal/
        
        Wyoming:
        -         
        - https://revenue.wyo.gov/
        """
        
        # Check if the parameters are set upstream in the instance
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        sources = self.sources if hasattr(self, 'sources') else sources
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'California', 'state_occupation': 'Wyoming', "capital_gains_long": capital_gains_long, "state_agi": state_agi, "filing_status": filing_status, "income": income, "sources": sources, "capital_gains_short": capital_gains_short, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def colorado_and_alabama(self, capital_gains_short=0, dependents=0, sources=False, county_residence='Other', state_agi=75000, capital_gains_long=0, filing_status='single', income=100000, credits_state=0):
        """
        --- Parameters ----
        
        Colorado:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        
        Alabama:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Colorado:
        - https://leg.colorado.gov/agencies/legislative-council-staff/individual-income-tax%C2%A0
          https://leg.colorado.gov/agencies/legislative-council-staff/individual-income-tax%C2%A0
          https://tax.colorado.gov/sites/tax/files/documents/ITT_Capital_Gain_Subtraction_Feb_2022.pdf
          https://www.auroragov.org/business_services/taxes/occupational_privilege_tax
          https://www.denvergov.org/Government/Agencies-Departments-Offices/Agencies-Departments-Offices-Directory/Department-of-Finance/Our-Divisions/Treasury/Business-Tax-Information#:~:text=Occupational%20privilege%20tax
         -%28show%20below%29&text=Each%20taxable%20employee%20is%20liable
         month%20for%20each%20taxable%20employee
          https://www.glendale.co.us/355/Occupational-Privilege-Tax
          https://www.greenwoodvillage.com/1220/Occupational-Privilege-Tax-OPT
          https://ci.sheridan.co.us/288/Occupational-Privilege-Tax
        
        Alabama:
        -         
        - https://www.revenue.alabama.gov/ultraviewer/viewer/basic_viewer/index.html?form=2023/02/22f40.pdf
          https://www.revenue.alabama.gov/faqs/what-is-alabamas-individual-income-tax-rate/
          https://www.revenue.alabama.gov/ultraviewer/viewer/basic_viewer/index.html?form=2023/02/22f40.pdf
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-24.htm?taxmap=true
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-24.htm?taxmap=true
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-24.htm?taxmap=true
          https://www.revenue.alabama.gov/individual-corporate/income-to-be-reported-on-the-alabama-income-tax-return/
          https://www.almonline.org/TaxRates.aspx#OccupationalTax
        """
        
        # Check if the parameters are set upstream in the instance
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Colorado', 'state_occupation': 'Alabama', "capital_gains_short": capital_gains_short, "dependents": dependents, "sources": sources, "county_residence": county_residence, "state_agi": state_agi, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def colorado_and_alaska(self, county_residence='Other', state_agi=75000, capital_gains_long=0, income=100000, sources=False, capital_gains_short=0, credits_state=0):
        """
        --- Parameters ----
        
        Colorado:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        
        Alaska:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Colorado:
        - https://leg.colorado.gov/agencies/legislative-council-staff/individual-income-tax%C2%A0
          https://leg.colorado.gov/agencies/legislative-council-staff/individual-income-tax%C2%A0
          https://tax.colorado.gov/sites/tax/files/documents/ITT_Capital_Gain_Subtraction_Feb_2022.pdf
          https://www.auroragov.org/business_services/taxes/occupational_privilege_tax
          https://www.denvergov.org/Government/Agencies-Departments-Offices/Agencies-Departments-Offices-Directory/Department-of-Finance/Our-Divisions/Treasury/Business-Tax-Information#:~:text=Occupational%20privilege%20tax
         -%28show%20below%29&text=Each%20taxable%20employee%20is%20liable
         month%20for%20each%20taxable%20employee
          https://www.glendale.co.us/355/Occupational-Privilege-Tax
          https://www.greenwoodvillage.com/1220/Occupational-Privilege-Tax-OPT
          https://ci.sheridan.co.us/288/Occupational-Privilege-Tax
        
        Alaska:
        -         
        - https://www.commerce.alaska.gov/web/dcra/officeofthestateassessor/alaskataxfacts.aspx
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        income = self.income if hasattr(self, 'income') else income
        sources = self.sources if hasattr(self, 'sources') else sources
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Colorado', 'state_occupation': 'Alaska', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_long": capital_gains_long, "income": income, "sources": sources, "capital_gains_short": capital_gains_short, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def colorado_and_arizona(self, county_residence='Other', state_agi=75000, capital_gains_short=0, capital_gains_long=0, filing_status='single', income=100000, dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Colorado:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        
        Arizona:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        credits_state : int, optional
            Credits for state taxes. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Colorado:
        - https://leg.colorado.gov/agencies/legislative-council-staff/individual-income-tax%C2%A0
          https://leg.colorado.gov/agencies/legislative-council-staff/individual-income-tax%C2%A0
          https://tax.colorado.gov/sites/tax/files/documents/ITT_Capital_Gain_Subtraction_Feb_2022.pdf
          https://www.auroragov.org/business_services/taxes/occupational_privilege_tax
          https://www.denvergov.org/Government/Agencies-Departments-Offices/Agencies-Departments-Offices-Directory/Department-of-Finance/Our-Divisions/Treasury/Business-Tax-Information#:~:text=Occupational%20privilege%20tax
         -%28show%20below%29&text=Each%20taxable%20employee%20is%20liable
         month%20for%20each%20taxable%20employee
          https://www.glendale.co.us/355/Occupational-Privilege-Tax
          https://www.greenwoodvillage.com/1220/Occupational-Privilege-Tax-OPT
          https://ci.sheridan.co.us/288/Occupational-Privilege-Tax
        
        Arizona:
        -         
        - https://azdor.gov/forms/individual/form-140-x-y-tables
          https://azdor.gov/forms/individual/form-140-x-y-tables
          https://azdor.gov/forms/individual/form-140-x-y-tables
          https://azdor.gov/businesses-arizona/withholding-tax/withholding-exceptions
          "https://azdor.gov/individuals/income-tax-filing-assistance/identifying-other-taxable-income#:~:text=In%20regards%20to%20capital%20gains
         the%20individuals%20regular%20tax%20rate.
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Colorado', 'state_occupation': 'Arizona', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def colorado_and_arkansas(self, capital_gains_short=0, dependents=0, sources=False, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_long=0, filing_status='single', income=100000, credits_state=0):
        """
        --- Parameters ----
        
        Colorado:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        
        Arkansas:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Colorado:
        - https://leg.colorado.gov/agencies/legislative-council-staff/individual-income-tax%C2%A0
          https://leg.colorado.gov/agencies/legislative-council-staff/individual-income-tax%C2%A0
          https://tax.colorado.gov/sites/tax/files/documents/ITT_Capital_Gain_Subtraction_Feb_2022.pdf
          https://www.auroragov.org/business_services/taxes/occupational_privilege_tax
          https://www.denvergov.org/Government/Agencies-Departments-Offices/Agencies-Departments-Offices-Directory/Department-of-Finance/Our-Divisions/Treasury/Business-Tax-Information#:~:text=Occupational%20privilege%20tax
         -%28show%20below%29&text=Each%20taxable%20employee%20is%20liable
         month%20for%20each%20taxable%20employee
          https://www.glendale.co.us/355/Occupational-Privilege-Tax
          https://www.greenwoodvillage.com/1220/Occupational-Privilege-Tax-OPT
          https://ci.sheridan.co.us/288/Occupational-Privilege-Tax
        
        Arkansas:
        -         
        - https://www.arkansasedc.com/why-arkansas/business-climate/tax-structure/personal-income-tax
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_and_AR1000NR_Instructions.pdf
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_FullYearResidentIndividualIncomeTaxReturn.pdf
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_and_AR1000NR_Instructions.pdf
          https://www.arkansasedc.com/why-arkansas/business-climate/tax-structure/capital-gains-tax-reduction
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_and_AR1000NR_Instructions.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Colorado', 'state_occupation': 'Arkansas', "capital_gains_short": capital_gains_short, "dependents": dependents, "sources": sources, "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def colorado_and_california(self, county_residence='Other', state_agi=75000, capital_gains_short=0, capital_gains_long=0, filing_status='single', income=100000, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Colorado:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        
        California:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Colorado:
        - https://leg.colorado.gov/agencies/legislative-council-staff/individual-income-tax%C2%A0
          https://leg.colorado.gov/agencies/legislative-council-staff/individual-income-tax%C2%A0
          https://tax.colorado.gov/sites/tax/files/documents/ITT_Capital_Gain_Subtraction_Feb_2022.pdf
          https://www.auroragov.org/business_services/taxes/occupational_privilege_tax
          https://www.denvergov.org/Government/Agencies-Departments-Offices/Agencies-Departments-Offices-Directory/Department-of-Finance/Our-Divisions/Treasury/Business-Tax-Information#:~:text=Occupational%20privilege%20tax
         -%28show%20below%29&text=Each%20taxable%20employee%20is%20liable
         month%20for%20each%20taxable%20employee
          https://www.glendale.co.us/355/Occupational-Privilege-Tax
          https://www.greenwoodvillage.com/1220/Occupational-Privilege-Tax-OPT
          https://ci.sheridan.co.us/288/Occupational-Privilege-Tax
        
        California:
        -         
        - https://www.ftb.ca.gov/forms/2022/2022-540-tax-rate-schedules.pdf
          https://www.ftb.ca.gov/file/personal/deductions/index.html
          https://www.ftb.ca.gov/about-ftb/newsroom/caleitc/eligibility-and-credit-information.html
          https://www.ftb.ca.gov/file/personal/income-types/capital-gains-and-losses.html#:~:text=California%20does%20not%20have%20a
         are%20taxed%20as%20ordinary%20income
          https://www.ftb.ca.gov/file/personal/
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Colorado', 'state_occupation': 'California', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def colorado_and_connecticut(self, county_residence='Other', state_agi=75000, capital_gains_short=0, capital_gains_long=0, filing_status='single', income=100000, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Colorado:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        
        Connecticut:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Colorado:
        - https://leg.colorado.gov/agencies/legislative-council-staff/individual-income-tax%C2%A0
          https://leg.colorado.gov/agencies/legislative-council-staff/individual-income-tax%C2%A0
          https://tax.colorado.gov/sites/tax/files/documents/ITT_Capital_Gain_Subtraction_Feb_2022.pdf
          https://www.auroragov.org/business_services/taxes/occupational_privilege_tax
          https://www.denvergov.org/Government/Agencies-Departments-Offices/Agencies-Departments-Offices-Directory/Department-of-Finance/Our-Divisions/Treasury/Business-Tax-Information#:~:text=Occupational%20privilege%20tax
         -%28show%20below%29&text=Each%20taxable%20employee%20is%20liable
         month%20for%20each%20taxable%20employee
          https://www.glendale.co.us/355/Occupational-Privilege-Tax
          https://www.greenwoodvillage.com/1220/Occupational-Privilege-Tax-OPT
          https://ci.sheridan.co.us/288/Occupational-Privilege-Tax
        
        Connecticut:
        -         
        - https://portal.ct.gov/-/media/DRS/Forms/2022/Income/2022-CT-1040-Instructions_1222.pdf
          https://www.cga.ct.gov/2022/rpt/pdf/2022-R-0108.pdf
          https://www.cga.ct.gov/current/pub/chap_229.htm#sec_12-703
          https://portal.ct.gov/-/media/DRS/Forms/2022/Income/2022-CT-1040-Instructions_1222.pdf
          "https://portal.ct.gov/DRS/Publications/TSSNs/TSSN-29#:~:text=An%20individuals%20net%20capital%20gains
         portion%20of%20Social%20Security%20benefits"
          "https://portal.ct.gov/DRS/Publications/TSSNs/TSSN-29#:~:text=An%20individuals%20net%20capital%20gains
         portion%20of%20Social%20Security%20benefits"
          https://portal.ct.gov/DRS/Individuals/Resident-Income-Tax/Tax-Information
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Colorado', 'state_occupation': 'Connecticut', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def colorado_and_delaware(self, county_residence='Other', state_agi=75000, capital_gains_short=0, capital_gains_long=0, federal_agi=75000, income=100000, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Colorado:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        
        Delaware:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        federal_agi : int, optional
            Adjusted Gross Income (AGI) for federal taxes. Default is 0. Sources - https://www.irs.gov/e-file-providers/definition-of-adjusted-gross-income, https://www.irs.gov/newsroom/taxpayers-should-know-the-difference-between-standard-and-itemized-deductions
        income : int, optional
            Gross income. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Colorado:
        - https://leg.colorado.gov/agencies/legislative-council-staff/individual-income-tax%C2%A0
          https://leg.colorado.gov/agencies/legislative-council-staff/individual-income-tax%C2%A0
          https://tax.colorado.gov/sites/tax/files/documents/ITT_Capital_Gain_Subtraction_Feb_2022.pdf
          https://www.auroragov.org/business_services/taxes/occupational_privilege_tax
          https://www.denvergov.org/Government/Agencies-Departments-Offices/Agencies-Departments-Offices-Directory/Department-of-Finance/Our-Divisions/Treasury/Business-Tax-Information#:~:text=Occupational%20privilege%20tax
         -%28show%20below%29&text=Each%20taxable%20employee%20is%20liable
         month%20for%20each%20taxable%20employee
          https://www.glendale.co.us/355/Occupational-Privilege-Tax
          https://www.greenwoodvillage.com/1220/Occupational-Privilege-Tax-OPT
          https://ci.sheridan.co.us/288/Occupational-Privilege-Tax
        
        Delaware:
        -         
        - https://revenue.delaware.gov/software-developer/tax-rate-changes/
          https://revenuefiles.delaware.gov/2022/PIT-RES_TY22_2022-01_PaperInteractive.pdf
          https://revenuefiles.delaware.gov/2022/PIT-RES_TY22_2022-02_Instructions.pdf
          https://revenue.delaware.gov/frequently-asked-questions/personal-income-tax-faqs/
          https://www.newcastlede.gov/DocumentCenter/View/20147/City-of-Wilmington-Wage-Tax-Information-PDF
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        federal_agi = self.federal_agi if hasattr(self, 'federal_agi') else federal_agi
        income = self.income if hasattr(self, 'income') else income
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Colorado', 'state_occupation': 'Delaware', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "federal_agi": federal_agi, "income": income, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def colorado_and_florida(self, county_residence='Other', state_agi=75000, capital_gains_long=0, income=100000, sources=False, capital_gains_short=0, credits_state=0):
        """
        --- Parameters ----
        
        Colorado:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        
        Florida:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Colorado:
        - https://leg.colorado.gov/agencies/legislative-council-staff/individual-income-tax%C2%A0
          https://leg.colorado.gov/agencies/legislative-council-staff/individual-income-tax%C2%A0
          https://tax.colorado.gov/sites/tax/files/documents/ITT_Capital_Gain_Subtraction_Feb_2022.pdf
          https://www.auroragov.org/business_services/taxes/occupational_privilege_tax
          https://www.denvergov.org/Government/Agencies-Departments-Offices/Agencies-Departments-Offices-Directory/Department-of-Finance/Our-Divisions/Treasury/Business-Tax-Information#:~:text=Occupational%20privilege%20tax
         -%28show%20below%29&text=Each%20taxable%20employee%20is%20liable
         month%20for%20each%20taxable%20employee
          https://www.glendale.co.us/355/Occupational-Privilege-Tax
          https://www.greenwoodvillage.com/1220/Occupational-Privilege-Tax-OPT
          https://ci.sheridan.co.us/288/Occupational-Privilege-Tax
        
        Florida:
        -         
        - https://floridarevenue.com/taxes/taxesfees/Pages/tax_interest_rates.aspx
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        income = self.income if hasattr(self, 'income') else income
        sources = self.sources if hasattr(self, 'sources') else sources
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Colorado', 'state_occupation': 'Florida', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_long": capital_gains_long, "income": income, "sources": sources, "capital_gains_short": capital_gains_short, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def colorado_and_georgia(self, capital_gains_short=0, dependents=0, sources=False, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_long=0, filing_status='single', income=100000, credits_state=0):
        """
        --- Parameters ----
        
        Colorado:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        
        Georgia:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Colorado:
        - https://leg.colorado.gov/agencies/legislative-council-staff/individual-income-tax%C2%A0
          https://leg.colorado.gov/agencies/legislative-council-staff/individual-income-tax%C2%A0
          https://tax.colorado.gov/sites/tax/files/documents/ITT_Capital_Gain_Subtraction_Feb_2022.pdf
          https://www.auroragov.org/business_services/taxes/occupational_privilege_tax
          https://www.denvergov.org/Government/Agencies-Departments-Offices/Agencies-Departments-Offices-Directory/Department-of-Finance/Our-Divisions/Treasury/Business-Tax-Information#:~:text=Occupational%20privilege%20tax
         -%28show%20below%29&text=Each%20taxable%20employee%20is%20liable
         month%20for%20each%20taxable%20employee
          https://www.glendale.co.us/355/Occupational-Privilege-Tax
          https://www.greenwoodvillage.com/1220/Occupational-Privilege-Tax-OPT
          https://ci.sheridan.co.us/288/Occupational-Privilege-Tax
        
        Georgia:
        -         
        - https://dor.georgia.gov/tax-tables-georgia-tax-rate-schedule
          https://apps.dor.ga.gov/FillableForms/PDFViewer/Index?form=2022GA500
          https://apps.dor.ga.gov/FillableForms/PDFViewer/Index?form=2022GA500
          https://apps.dor.ga.gov/FillableForms/PDFViewer/Index?form=2022GA500
          https://dor.georgia.gov/years-individual-income-tax-forms
          https://dor.georgia.gov/years-individual-income-tax-forms
        """
        
        # Check if the parameters are set upstream in the instance
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Colorado', 'state_occupation': 'Georgia', "capital_gains_short": capital_gains_short, "dependents": dependents, "sources": sources, "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def colorado_and_hawaii(self, capital_gains_short=0, dependents=0, sources=False, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_long=0, filing_status='single', income=100000, credits_state=0):
        """
        --- Parameters ----
        
        Colorado:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        
        Hawaii:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Colorado:
        - https://leg.colorado.gov/agencies/legislative-council-staff/individual-income-tax%C2%A0
          https://leg.colorado.gov/agencies/legislative-council-staff/individual-income-tax%C2%A0
          https://tax.colorado.gov/sites/tax/files/documents/ITT_Capital_Gain_Subtraction_Feb_2022.pdf
          https://www.auroragov.org/business_services/taxes/occupational_privilege_tax
          https://www.denvergov.org/Government/Agencies-Departments-Offices/Agencies-Departments-Offices-Directory/Department-of-Finance/Our-Divisions/Treasury/Business-Tax-Information#:~:text=Occupational%20privilege%20tax
         -%28show%20below%29&text=Each%20taxable%20employee%20is%20liable
         month%20for%20each%20taxable%20employee
          https://www.glendale.co.us/355/Occupational-Privilege-Tax
          https://www.greenwoodvillage.com/1220/Occupational-Privilege-Tax-OPT
          https://ci.sheridan.co.us/288/Occupational-Privilege-Tax
        
        Hawaii:
        -         
        - https://tax.hawaii.gov/forms/d_18table-on/d_18table-on_p13/
          https://files.hawaii.gov/tax/forms/2018/18table-on.pdf
          https://files.hawaii.gov/tax/news/pubs/22outline.pdf
          https://files.hawaii.gov/tax/news/pubs/FastTaxRefGuide_2017.pdf
          https://www.tfhawaii.org/wordpress/state-tax-resources/mini-tax-guides/individual-income-tax-chapter-235/#:~:text=Long%20term%20capital%20gains%20are%20taxed%20at%20a%20maximum%20of%207.25%25
          https://www.grassrootinstitute.org/2023/02/hb337-increasing-tax-on-capital-gains-a-bad-risk/
          https://files.hawaii.gov/tax/news/pubs/22outline.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Colorado', 'state_occupation': 'Hawaii', "capital_gains_short": capital_gains_short, "dependents": dependents, "sources": sources, "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def colorado_and_idaho(self, capital_gains_short=0, dependents=0, sources=False, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_long=0, filing_status='single', income=100000, credits_state=0):
        """
        --- Parameters ----
        
        Colorado:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        
        Idaho:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Colorado:
        - https://leg.colorado.gov/agencies/legislative-council-staff/individual-income-tax%C2%A0
          https://leg.colorado.gov/agencies/legislative-council-staff/individual-income-tax%C2%A0
          https://tax.colorado.gov/sites/tax/files/documents/ITT_Capital_Gain_Subtraction_Feb_2022.pdf
          https://www.auroragov.org/business_services/taxes/occupational_privilege_tax
          https://www.denvergov.org/Government/Agencies-Departments-Offices/Agencies-Departments-Offices-Directory/Department-of-Finance/Our-Divisions/Treasury/Business-Tax-Information#:~:text=Occupational%20privilege%20tax
         -%28show%20below%29&text=Each%20taxable%20employee%20is%20liable
         month%20for%20each%20taxable%20employee
          https://www.glendale.co.us/355/Occupational-Privilege-Tax
          https://www.greenwoodvillage.com/1220/Occupational-Privilege-Tax-OPT
          https://ci.sheridan.co.us/288/Occupational-Privilege-Tax
        
        Idaho:
        -         
        - https://tax.idaho.gov/taxes/income-tax/individual-income/individual-income-tax-rate-schedule/
          https://tax.idaho.gov/wp-content/uploads/forms/EIN00046/EIN00046_03-01-2023.pdf
          https://tax.idaho.gov/taxes/income-tax/individual-income/idaho-source-income/
          https://tax.idaho.gov/taxes/income-tax/individual-income/idaho-source-income/
        """
        
        # Check if the parameters are set upstream in the instance
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Colorado', 'state_occupation': 'Idaho', "capital_gains_short": capital_gains_short, "dependents": dependents, "sources": sources, "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def colorado_and_illinois(self, county_residence='Other', state_agi=75000, capital_gains_short=0, capital_gains_long=0, filing_status='single', income=100000, dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Colorado:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        
        Illinois:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Colorado:
        - https://leg.colorado.gov/agencies/legislative-council-staff/individual-income-tax%C2%A0
          https://leg.colorado.gov/agencies/legislative-council-staff/individual-income-tax%C2%A0
          https://tax.colorado.gov/sites/tax/files/documents/ITT_Capital_Gain_Subtraction_Feb_2022.pdf
          https://www.auroragov.org/business_services/taxes/occupational_privilege_tax
          https://www.denvergov.org/Government/Agencies-Departments-Offices/Agencies-Departments-Offices-Directory/Department-of-Finance/Our-Divisions/Treasury/Business-Tax-Information#:~:text=Occupational%20privilege%20tax
         -%28show%20below%29&text=Each%20taxable%20employee%20is%20liable
         month%20for%20each%20taxable%20employee
          https://www.glendale.co.us/355/Occupational-Privilege-Tax
          https://www.greenwoodvillage.com/1220/Occupational-Privilege-Tax-OPT
          https://ci.sheridan.co.us/288/Occupational-Privilege-Tax
        
        Illinois:
        -         
        - https://tax.illinois.gov/research/taxrates/income.html#IndividualIncome
          https://tax.illinois.gov/content/dam/soi/en/web/tax/forms/incometax/documents/currentyear/individual/il-1040-instr.pdf
          https://tax.illinois.gov/questionsandanswers/12.html
          https://tax.illinois.gov/content/dam/soi/en/web/tax/forms/incometax/documents/currentyear/individual/il-1040-instr.pdf
          https://tax.illinois.gov/research/taxrates/personalproperty.html
          https://tax.illinois.gov/localgovernments/income.html
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Colorado', 'state_occupation': 'Illinois', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def colorado_and_indiana(self, capital_gains_short=0, dependents=0, sources=False, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_long=0, filing_status='single', income=100000, credits_state=0):
        """
        --- Parameters ----
        
        Colorado:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        
        Indiana:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Colorado:
        - https://leg.colorado.gov/agencies/legislative-council-staff/individual-income-tax%C2%A0
          https://leg.colorado.gov/agencies/legislative-council-staff/individual-income-tax%C2%A0
          https://tax.colorado.gov/sites/tax/files/documents/ITT_Capital_Gain_Subtraction_Feb_2022.pdf
          https://www.auroragov.org/business_services/taxes/occupational_privilege_tax
          https://www.denvergov.org/Government/Agencies-Departments-Offices/Agencies-Departments-Offices-Directory/Department-of-Finance/Our-Divisions/Treasury/Business-Tax-Information#:~:text=Occupational%20privilege%20tax
         -%28show%20below%29&text=Each%20taxable%20employee%20is%20liable
         month%20for%20each%20taxable%20employee
          https://www.glendale.co.us/355/Occupational-Privilege-Tax
          https://www.greenwoodvillage.com/1220/Occupational-Privilege-Tax-OPT
          https://ci.sheridan.co.us/288/Occupational-Privilege-Tax
        
        Indiana:
        -         
        - https://www.in.gov/dor/tax-forms/2022-individual-income-tax-forms/
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-32.htm
          https://www.in.gov/dor/files/ib28.pdf
          https://www.in.gov/dor/tax-forms/2022-individual-income-tax-forms/
          https://www.in.gov/dor/files/2022-county-tax-rates-and-codes.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Colorado', 'state_occupation': 'Indiana', "capital_gains_short": capital_gains_short, "dependents": dependents, "sources": sources, "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def colorado_and_iowa(self, capital_gains_short=0, dependents=0, sources=False, county_residence='Other', state_agi=75000, capital_gains_long=0, filing_status='single', income=100000, credits_state=0):
        """
        --- Parameters ----
        
        Colorado:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        
        Iowa:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Colorado:
        - https://leg.colorado.gov/agencies/legislative-council-staff/individual-income-tax%C2%A0
          https://leg.colorado.gov/agencies/legislative-council-staff/individual-income-tax%C2%A0
          https://tax.colorado.gov/sites/tax/files/documents/ITT_Capital_Gain_Subtraction_Feb_2022.pdf
          https://www.auroragov.org/business_services/taxes/occupational_privilege_tax
          https://www.denvergov.org/Government/Agencies-Departments-Offices/Agencies-Departments-Offices-Directory/Department-of-Finance/Our-Divisions/Treasury/Business-Tax-Information#:~:text=Occupational%20privilege%20tax
         -%28show%20below%29&text=Each%20taxable%20employee%20is%20liable
         month%20for%20each%20taxable%20employee
          https://www.glendale.co.us/355/Occupational-Privilege-Tax
          https://www.greenwoodvillage.com/1220/Occupational-Privilege-Tax-OPT
          https://ci.sheridan.co.us/288/Occupational-Privilege-Tax
        
        Iowa:
        -         
        - https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
          https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
          https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
          https://tax.iowa.gov/iowa-illinois-reciprocal-agreement
          https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
        """
        
        # Check if the parameters are set upstream in the instance
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Colorado', 'state_occupation': 'Iowa', "capital_gains_short": capital_gains_short, "dependents": dependents, "sources": sources, "county_residence": county_residence, "state_agi": state_agi, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def colorado_and_kansas(self, capital_gains_short=0, dependents=0, sources=False, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_long=0, filing_status='single', income=100000, credits_state=0):
        """
        --- Parameters ----
        
        Colorado:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        
        Kansas:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Colorado:
        - https://leg.colorado.gov/agencies/legislative-council-staff/individual-income-tax%C2%A0
          https://leg.colorado.gov/agencies/legislative-council-staff/individual-income-tax%C2%A0
          https://tax.colorado.gov/sites/tax/files/documents/ITT_Capital_Gain_Subtraction_Feb_2022.pdf
          https://www.auroragov.org/business_services/taxes/occupational_privilege_tax
          https://www.denvergov.org/Government/Agencies-Departments-Offices/Agencies-Departments-Offices-Directory/Department-of-Finance/Our-Divisions/Treasury/Business-Tax-Information#:~:text=Occupational%20privilege%20tax
         -%28show%20below%29&text=Each%20taxable%20employee%20is%20liable
         month%20for%20each%20taxable%20employee
          https://www.glendale.co.us/355/Occupational-Privilege-Tax
          https://www.greenwoodvillage.com/1220/Occupational-Privilege-Tax-OPT
          https://ci.sheridan.co.us/288/Occupational-Privilege-Tax
        
        Kansas:
        -         
        - https://www.ksrevenue.gov/taxrates.html
          https://www.ksrevenue.gov/incomebook22.html#0
          https://www.ksrevenue.gov/incomebook22.html#0
          https://www.ksrevenue.gov/incomebook22.html#0
          https://www.ksrevenue.gov/incomebook22.html
          https://www.ksrevenue.gov/incomebook22.html
        """
        
        # Check if the parameters are set upstream in the instance
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Colorado', 'state_occupation': 'Kansas', "capital_gains_short": capital_gains_short, "dependents": dependents, "sources": sources, "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def colorado_and_kentucky(self, county_residence='Other', state_agi=75000, capital_gains_short=0, capital_gains_long=0, filing_status='single', income=100000, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Colorado:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        
        Kentucky:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Colorado:
        - https://leg.colorado.gov/agencies/legislative-council-staff/individual-income-tax%C2%A0
          https://leg.colorado.gov/agencies/legislative-council-staff/individual-income-tax%C2%A0
          https://tax.colorado.gov/sites/tax/files/documents/ITT_Capital_Gain_Subtraction_Feb_2022.pdf
          https://www.auroragov.org/business_services/taxes/occupational_privilege_tax
          https://www.denvergov.org/Government/Agencies-Departments-Offices/Agencies-Departments-Offices-Directory/Department-of-Finance/Our-Divisions/Treasury/Business-Tax-Information#:~:text=Occupational%20privilege%20tax
         -%28show%20below%29&text=Each%20taxable%20employee%20is%20liable
         month%20for%20each%20taxable%20employee
          https://www.glendale.co.us/355/Occupational-Privilege-Tax
          https://www.greenwoodvillage.com/1220/Occupational-Privilege-Tax-OPT
          https://ci.sheridan.co.us/288/Occupational-Privilege-Tax
        
        Kentucky:
        -         
        - https://revenue.ky.gov/News/Pages/Dor-Announces-Updates-to-Individual-Income-Tax-for-2023-Tax-Year.aspx
          https://revenue.ky.gov/Individual/Individual-Income-Tax/Pages/default.aspx
          https://revenue.ky.gov/News/Pages/Dor-Announces-Updates-to-Individual-Income-Tax-for-2023-Tax-Year.aspx
          https://revenue.ky.gov/Individual/Individual-Income-Tax/Pages/default.aspx
          https://revenue.ky.gov/Dor%20Training%20Materials/103%20KAR%2017.140.%20Individual%20income%20tax%20-%20reciprocity%20-%20nonresidents.pdf
          https://revenue.ky.gov/Individual/Individual-Income-Tax/Pages/default.aspx
          https://nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2021/Taxes-21-32.htm?taxmap=true
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Colorado', 'state_occupation': 'Kentucky', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def colorado_and_louisiana(self, capital_gains_short=0, dependents=0, sources=False, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_long=0, filing_status='single', income=100000, credits_state=0):
        """
        --- Parameters ----
        
        Colorado:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        
        Louisiana:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Colorado:
        - https://leg.colorado.gov/agencies/legislative-council-staff/individual-income-tax%C2%A0
          https://leg.colorado.gov/agencies/legislative-council-staff/individual-income-tax%C2%A0
          https://tax.colorado.gov/sites/tax/files/documents/ITT_Capital_Gain_Subtraction_Feb_2022.pdf
          https://www.auroragov.org/business_services/taxes/occupational_privilege_tax
          https://www.denvergov.org/Government/Agencies-Departments-Offices/Agencies-Departments-Offices-Directory/Department-of-Finance/Our-Divisions/Treasury/Business-Tax-Information#:~:text=Occupational%20privilege%20tax
         -%28show%20below%29&text=Each%20taxable%20employee%20is%20liable
         month%20for%20each%20taxable%20employee
          https://www.glendale.co.us/355/Occupational-Privilege-Tax
          https://www.greenwoodvillage.com/1220/Occupational-Privilege-Tax-OPT
          https://ci.sheridan.co.us/288/Occupational-Privilege-Tax
        
        Louisiana:
        -         
        - https://revenue.louisiana.gov/individualincometax
          https://revenue.louisiana.gov/individualincometax
          https://www.revenue.louisiana.gov/taxforms/6935(11_02)F.pdf
          https://revenue.louisiana.gov/individualincometax
          https://revenue.louisiana.gov/individualincometax
        """
        
        # Check if the parameters are set upstream in the instance
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Colorado', 'state_occupation': 'Louisiana', "capital_gains_short": capital_gains_short, "dependents": dependents, "sources": sources, "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def colorado_and_maine(self, capital_gains_short=0, dependents=0, sources=False, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_long=0, filing_status='single', income=100000, credits_state=0):
        """
        --- Parameters ----
        
        Colorado:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        
        Maine:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Colorado:
        - https://leg.colorado.gov/agencies/legislative-council-staff/individual-income-tax%C2%A0
          https://leg.colorado.gov/agencies/legislative-council-staff/individual-income-tax%C2%A0
          https://tax.colorado.gov/sites/tax/files/documents/ITT_Capital_Gain_Subtraction_Feb_2022.pdf
          https://www.auroragov.org/business_services/taxes/occupational_privilege_tax
          https://www.denvergov.org/Government/Agencies-Departments-Offices/Agencies-Departments-Offices-Directory/Department-of-Finance/Our-Divisions/Treasury/Business-Tax-Information#:~:text=Occupational%20privilege%20tax
         -%28show%20below%29&text=Each%20taxable%20employee%20is%20liable
         month%20for%20each%20taxable%20employee
          https://www.glendale.co.us/355/Occupational-Privilege-Tax
          https://www.greenwoodvillage.com/1220/Occupational-Privilege-Tax-OPT
          https://ci.sheridan.co.us/288/Occupational-Privilege-Tax
        
        Maine:
        -         
        - https://www.maine.gov/revenue/tax-return-forms/individual-income-tax-2022
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/ind_tax_rate_sched_2022.pdf
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/22_1040me_book_gen_instruc_revisedFeb23.pdf
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/22_1040me_book_gen_instruc_revisedFeb23.pdf
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/22_1040me_book_gen_instruc_revisedFeb23.pdf
          https://www.maine.gov/revenue/tax-return-forms/individual-income-tax-2022
        """
        
        # Check if the parameters are set upstream in the instance
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Colorado', 'state_occupation': 'Maine', "capital_gains_short": capital_gains_short, "dependents": dependents, "sources": sources, "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def colorado_and_maryland(self, capital_gains_short=0, dependents=0, sources=False, county_residence='Other', state_agi=75000, capital_gains_long=0, filing_status='single', income=100000, credits_state=0):
        """
        --- Parameters ----
        
        Colorado:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        
        Maryland:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Colorado:
        - https://leg.colorado.gov/agencies/legislative-council-staff/individual-income-tax%C2%A0
          https://leg.colorado.gov/agencies/legislative-council-staff/individual-income-tax%C2%A0
          https://tax.colorado.gov/sites/tax/files/documents/ITT_Capital_Gain_Subtraction_Feb_2022.pdf
          https://www.auroragov.org/business_services/taxes/occupational_privilege_tax
          https://www.denvergov.org/Government/Agencies-Departments-Offices/Agencies-Departments-Offices-Directory/Department-of-Finance/Our-Divisions/Treasury/Business-Tax-Information#:~:text=Occupational%20privilege%20tax
         -%28show%20below%29&text=Each%20taxable%20employee%20is%20liable
         month%20for%20each%20taxable%20employee
          https://www.glendale.co.us/355/Occupational-Privilege-Tax
          https://www.greenwoodvillage.com/1220/Occupational-Privilege-Tax-OPT
          https://ci.sheridan.co.us/288/Occupational-Privilege-Tax
        
        Maryland:
        -         
        - https://www.marylandtaxes.gov/individual/income/tax-info/tax-rates.php
          https://www.marylandtaxes.gov/forms/22_forms/Resident_Booklet.pdf
          https://www.marylandtaxes.gov/forms/Tax_Publications/Administrative_Releases/Income_and_Estate_Tax_Releases/ar_it37.pdf
          https://www.marylandtaxes.gov/individual/index.php
          https://www.marylandtaxes.gov/individual/income/tax-info/tax-rates.php
        """
        
        # Check if the parameters are set upstream in the instance
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Colorado', 'state_occupation': 'Maryland', "capital_gains_short": capital_gains_short, "dependents": dependents, "sources": sources, "county_residence": county_residence, "state_agi": state_agi, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def colorado_and_massachusetts(self, county_residence='Other', state_agi=75000, capital_gains_short=0, capital_gains_long=0, filing_status='single', income=100000, dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Colorado:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        
        Massachusetts:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Colorado:
        - https://leg.colorado.gov/agencies/legislative-council-staff/individual-income-tax%C2%A0
          https://leg.colorado.gov/agencies/legislative-council-staff/individual-income-tax%C2%A0
          https://tax.colorado.gov/sites/tax/files/documents/ITT_Capital_Gain_Subtraction_Feb_2022.pdf
          https://www.auroragov.org/business_services/taxes/occupational_privilege_tax
          https://www.denvergov.org/Government/Agencies-Departments-Offices/Agencies-Departments-Offices-Directory/Department-of-Finance/Our-Divisions/Treasury/Business-Tax-Information#:~:text=Occupational%20privilege%20tax
         -%28show%20below%29&text=Each%20taxable%20employee%20is%20liable
         month%20for%20each%20taxable%20employee
          https://www.glendale.co.us/355/Occupational-Privilege-Tax
          https://www.greenwoodvillage.com/1220/Occupational-Privilege-Tax-OPT
          https://ci.sheridan.co.us/288/Occupational-Privilege-Tax
        
        Massachusetts:
        -         
        - https://www.mass.gov/service-details/massachusetts-tax-rates
          https://www.mass.gov/service-details/view-massachusetts-personal-income-tax-exemptions
          https://www.mass.gov/service-details/view-massachusetts-personal-income-tax-exemptions
          https://www.mass.gov/service-details/view-child-and-dependent-related-credits
          https://www.mass.gov/service-details/massachusetts-tax-rates
          https://www.mass.gov/personal-income-tax
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Colorado', 'state_occupation': 'Massachusetts', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def colorado_and_michigan(self, capital_gains_short=0, county_occupation='Other', dependents=0, sources=False, county_residence='Other', state_agi=75000, capital_gains_long=0, filing_status='single', income=100000, credits_state=0):
        """
        --- Parameters ----
        
        Colorado:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        
        Michigan:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        county_occupation : str, optional
            County of occupation. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Colorado:
        - https://leg.colorado.gov/agencies/legislative-council-staff/individual-income-tax%C2%A0
          https://leg.colorado.gov/agencies/legislative-council-staff/individual-income-tax%C2%A0
          https://tax.colorado.gov/sites/tax/files/documents/ITT_Capital_Gain_Subtraction_Feb_2022.pdf
          https://www.auroragov.org/business_services/taxes/occupational_privilege_tax
          https://www.denvergov.org/Government/Agencies-Departments-Offices/Agencies-Departments-Offices-Directory/Department-of-Finance/Our-Divisions/Treasury/Business-Tax-Information#:~:text=Occupational%20privilege%20tax
         -%28show%20below%29&text=Each%20taxable%20employee%20is%20liable
         month%20for%20each%20taxable%20employee
          https://www.glendale.co.us/355/Occupational-Privilege-Tax
          https://www.greenwoodvillage.com/1220/Occupational-Privilege-Tax-OPT
          https://ci.sheridan.co.us/288/Occupational-Privilege-Tax
        
        Michigan:
        -         
        - https://www.michigan.gov/taxes/iit/new-developments-for-tax-year-2022
          "https://www.michigan.gov/taxes/-/media/Project/Websites/taxes/Forms/2023/2023-SUW/446_Withholding-Guide_2023.pdf?rev=177ca3d8cb034392bf84e317ed78c320&hash=19C5F42F43A89A91863D229349F48883#:~:text=The%20exemption%20amount%20is%20%245%2C400
         employees%20Michigan%20income%20tax%20return"
          "https://www.michigan.gov/taxes/-/media/Project/Websites/taxes/Forms/2023/2023-SUW/446_Withholding-Guide_2023.pdf?rev=177ca3d8cb034392bf84e317ed78c320&hash=19C5F42F43A89A91863D229349F48883#:~:text=The%20exemption%20amount%20is%20%245%2C400
         employees%20Michigan%20income%20tax%20return"
          https://www.michigan.gov/taxes/questions/iit/accordion/residency/are-my-wages-earned-in-another-state-taxable-in-michigan-if-i-am-a-michigan-resident-1
          https://www.michigan.gov/taxes/-/media/Project/Websites/taxes/Forms/2022/2022-IIT-Forms/BOOK_MI-1040.pdf?rev=4524f1cbbc0243b194ab05a6680be75b&hash=40DFE71E651F7C1D65C31F6CD03D944F
          https://www.michigan.gov/taxes/questions/iit/accordion/general/what-cities-impose-an-income-tax
        """
        
        # Check if the parameters are set upstream in the instance
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        county_occupation = self.county_occupation if hasattr(self, 'county_occupation') else county_occupation
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Colorado', 'state_occupation': 'Michigan', "capital_gains_short": capital_gains_short, "county_occupation": county_occupation, "dependents": dependents, "sources": sources, "county_residence": county_residence, "state_agi": state_agi, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def colorado_and_minnesota(self, county_residence='Other', state_agi=75000, capital_gains_short=0, capital_gains_long=0, filing_status='single', income=100000, dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Colorado:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        
        Minnesota:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Colorado:
        - https://leg.colorado.gov/agencies/legislative-council-staff/individual-income-tax%C2%A0
          https://leg.colorado.gov/agencies/legislative-council-staff/individual-income-tax%C2%A0
          https://tax.colorado.gov/sites/tax/files/documents/ITT_Capital_Gain_Subtraction_Feb_2022.pdf
          https://www.auroragov.org/business_services/taxes/occupational_privilege_tax
          https://www.denvergov.org/Government/Agencies-Departments-Offices/Agencies-Departments-Offices-Directory/Department-of-Finance/Our-Divisions/Treasury/Business-Tax-Information#:~:text=Occupational%20privilege%20tax
         -%28show%20below%29&text=Each%20taxable%20employee%20is%20liable
         month%20for%20each%20taxable%20employee
          https://www.glendale.co.us/355/Occupational-Privilege-Tax
          https://www.greenwoodvillage.com/1220/Occupational-Privilege-Tax-OPT
          https://ci.sheridan.co.us/288/Occupational-Privilege-Tax
        
        Minnesota:
        -         
        - https://www.revenue.state.mn.us/minnesota-income-tax-rates-and-brackets
          https://www.revenue.state.mn.us/child-and-dependent-care-credit
          https://www.revenue.state.mn.us/dependent-exemptions
          https://www.revenue.state.mn.us/minnesota-standard-deduction
          https://mn.gov/mmb/assets/mwr_form_tcm1059-128581.pdf
          https://www.revenue.state.mn.us/individual-income-tax
          https://www.revenue.state.mn.us/individual-income-tax
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Colorado', 'state_occupation': 'Minnesota', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def colorado_and_mississippi(self, county_residence='Other', state_agi=75000, capital_gains_short=0, capital_gains_long=0, filing_status='single', income=100000, dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Colorado:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        
        Mississippi:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Colorado:
        - https://leg.colorado.gov/agencies/legislative-council-staff/individual-income-tax%C2%A0
          https://leg.colorado.gov/agencies/legislative-council-staff/individual-income-tax%C2%A0
          https://tax.colorado.gov/sites/tax/files/documents/ITT_Capital_Gain_Subtraction_Feb_2022.pdf
          https://www.auroragov.org/business_services/taxes/occupational_privilege_tax
          https://www.denvergov.org/Government/Agencies-Departments-Offices/Agencies-Departments-Offices-Directory/Department-of-Finance/Our-Divisions/Treasury/Business-Tax-Information#:~:text=Occupational%20privilege%20tax
         -%28show%20below%29&text=Each%20taxable%20employee%20is%20liable
         month%20for%20each%20taxable%20employee
          https://www.glendale.co.us/355/Occupational-Privilege-Tax
          https://www.greenwoodvillage.com/1220/Occupational-Privilege-Tax-OPT
          https://ci.sheridan.co.us/288/Occupational-Privilege-Tax
        
        Mississippi:
        -         
        - https://www.dor.ms.gov/individual/tax-rates
          https://www.dor.ms.gov/individual/tax-rates
          https://www.dor.ms.gov/individual/tax-rates
          https://www.dor.ms.gov/individual
          https://www.dor.ms.gov/individual
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Colorado', 'state_occupation': 'Mississippi', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def colorado_and_missouri(self, capital_gains_short=0, dependents=0, sources=False, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_long=0, filing_status='single', income=100000, credits_state=0):
        """
        --- Parameters ----
        
        Colorado:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        
        Missouri:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Colorado:
        - https://leg.colorado.gov/agencies/legislative-council-staff/individual-income-tax%C2%A0
          https://leg.colorado.gov/agencies/legislative-council-staff/individual-income-tax%C2%A0
          https://tax.colorado.gov/sites/tax/files/documents/ITT_Capital_Gain_Subtraction_Feb_2022.pdf
          https://www.auroragov.org/business_services/taxes/occupational_privilege_tax
          https://www.denvergov.org/Government/Agencies-Departments-Offices/Agencies-Departments-Offices-Directory/Department-of-Finance/Our-Divisions/Treasury/Business-Tax-Information#:~:text=Occupational%20privilege%20tax
         -%28show%20below%29&text=Each%20taxable%20employee%20is%20liable
         month%20for%20each%20taxable%20employee
          https://www.glendale.co.us/355/Occupational-Privilege-Tax
          https://www.greenwoodvillage.com/1220/Occupational-Privilege-Tax-OPT
          https://ci.sheridan.co.us/288/Occupational-Privilege-Tax
        
        Missouri:
        -         
        - https://dor.mo.gov/taxation/individual/tax-types/income/year-changes/
          https://dor.mo.gov/taxation/individual/tax-types/income/year-changes/
          https://dor.mo.gov/taxation/individual/tax-types/income/year-changes/
          https://www.kcmo.gov/city-hall/departments/finance/earnings-tax
          https://www.stlouis-mo.gov/government/departments/collector/earnings-tax/individual-earnings-tax-info.cfm
        """
        
        # Check if the parameters are set upstream in the instance
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Colorado', 'state_occupation': 'Missouri', "capital_gains_short": capital_gains_short, "dependents": dependents, "sources": sources, "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def colorado_and_montana(self, capital_gains_short=0, dependents=0, sources=False, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_long=0, filing_status='single', income=100000, credits_state=0):
        """
        --- Parameters ----
        
        Colorado:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        
        Montana:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Colorado:
        - https://leg.colorado.gov/agencies/legislative-council-staff/individual-income-tax%C2%A0
          https://leg.colorado.gov/agencies/legislative-council-staff/individual-income-tax%C2%A0
          https://tax.colorado.gov/sites/tax/files/documents/ITT_Capital_Gain_Subtraction_Feb_2022.pdf
          https://www.auroragov.org/business_services/taxes/occupational_privilege_tax
          https://www.denvergov.org/Government/Agencies-Departments-Offices/Agencies-Departments-Offices-Directory/Department-of-Finance/Our-Divisions/Treasury/Business-Tax-Information#:~:text=Occupational%20privilege%20tax
         -%28show%20below%29&text=Each%20taxable%20employee%20is%20liable
         month%20for%20each%20taxable%20employee
          https://www.glendale.co.us/355/Occupational-Privilege-Tax
          https://www.greenwoodvillage.com/1220/Occupational-Privilege-Tax-OPT
          https://ci.sheridan.co.us/288/Occupational-Privilege-Tax
        
        Montana:
        -         
        - https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/2022/12/montana-employees-withholding-allowance-and-exemption-certificate-form-mw-4.pdf
          https://mtrevenue.gov/2021/11/29/simplification-of-montana-income-taxation/?utm_source=rss&utm_medium=rss&utm_campaign=simplification-of-montana-income-taxation
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Colorado', 'state_occupation': 'Montana', "capital_gains_short": capital_gains_short, "dependents": dependents, "sources": sources, "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def colorado_and_nebraska(self, county_residence='Other', state_agi=75000, capital_gains_short=0, capital_gains_long=0, filing_status='single', income=100000, dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Colorado:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        
        Nebraska:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Colorado:
        - https://leg.colorado.gov/agencies/legislative-council-staff/individual-income-tax%C2%A0
          https://leg.colorado.gov/agencies/legislative-council-staff/individual-income-tax%C2%A0
          https://tax.colorado.gov/sites/tax/files/documents/ITT_Capital_Gain_Subtraction_Feb_2022.pdf
          https://www.auroragov.org/business_services/taxes/occupational_privilege_tax
          https://www.denvergov.org/Government/Agencies-Departments-Offices/Agencies-Departments-Offices-Directory/Department-of-Finance/Our-Divisions/Treasury/Business-Tax-Information#:~:text=Occupational%20privilege%20tax
         -%28show%20below%29&text=Each%20taxable%20employee%20is%20liable
         month%20for%20each%20taxable%20employee
          https://www.glendale.co.us/355/Occupational-Privilege-Tax
          https://www.greenwoodvillage.com/1220/Occupational-Privilege-Tax-OPT
          https://ci.sheridan.co.us/288/Occupational-Privilege-Tax
        
        Nebraska:
        -         
        - https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2023_Circular_EN_8-429r2023_Whole_Book_11.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Colorado', 'state_occupation': 'Nebraska', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def colorado_and_nevada(self, county_residence='Other', state_agi=75000, capital_gains_long=0, income=100000, sources=False, capital_gains_short=0, credits_state=0):
        """
        --- Parameters ----
        
        Colorado:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        
        Nevada:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Colorado:
        - https://leg.colorado.gov/agencies/legislative-council-staff/individual-income-tax%C2%A0
          https://leg.colorado.gov/agencies/legislative-council-staff/individual-income-tax%C2%A0
          https://tax.colorado.gov/sites/tax/files/documents/ITT_Capital_Gain_Subtraction_Feb_2022.pdf
          https://www.auroragov.org/business_services/taxes/occupational_privilege_tax
          https://www.denvergov.org/Government/Agencies-Departments-Offices/Agencies-Departments-Offices-Directory/Department-of-Finance/Our-Divisions/Treasury/Business-Tax-Information#:~:text=Occupational%20privilege%20tax
         -%28show%20below%29&text=Each%20taxable%20employee%20is%20liable
         month%20for%20each%20taxable%20employee
          https://www.glendale.co.us/355/Occupational-Privilege-Tax
          https://www.greenwoodvillage.com/1220/Occupational-Privilege-Tax-OPT
          https://ci.sheridan.co.us/288/Occupational-Privilege-Tax
        
        Nevada:
        -         
        - https://www.nvsos.gov/sos/businesses/the-nevada-advantage
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        income = self.income if hasattr(self, 'income') else income
        sources = self.sources if hasattr(self, 'sources') else sources
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Colorado', 'state_occupation': 'Nevada', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_long": capital_gains_long, "income": income, "sources": sources, "capital_gains_short": capital_gains_short, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def colorado_and_new_hampshire(self, county_residence='Other', state_agi=75000, capital_gains_long=0, income=100000, sources=False, capital_gains_short=0, credits_state=0):
        """
        --- Parameters ----
        
        Colorado:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        
        New Hampshire:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Colorado:
        - https://leg.colorado.gov/agencies/legislative-council-staff/individual-income-tax%C2%A0
          https://leg.colorado.gov/agencies/legislative-council-staff/individual-income-tax%C2%A0
          https://tax.colorado.gov/sites/tax/files/documents/ITT_Capital_Gain_Subtraction_Feb_2022.pdf
          https://www.auroragov.org/business_services/taxes/occupational_privilege_tax
          https://www.denvergov.org/Government/Agencies-Departments-Offices/Agencies-Departments-Offices-Directory/Department-of-Finance/Our-Divisions/Treasury/Business-Tax-Information#:~:text=Occupational%20privilege%20tax
         -%28show%20below%29&text=Each%20taxable%20employee%20is%20liable
         month%20for%20each%20taxable%20employee
          https://www.glendale.co.us/355/Occupational-Privilege-Tax
          https://www.greenwoodvillage.com/1220/Occupational-Privilege-Tax-OPT
          https://ci.sheridan.co.us/288/Occupational-Privilege-Tax
        
        New Hampshire:
        -         
        - https://www.revenue.nh.gov/assistance/tax-overview.htm
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        income = self.income if hasattr(self, 'income') else income
        sources = self.sources if hasattr(self, 'sources') else sources
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Colorado', 'state_occupation': 'NewHampshire', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_long": capital_gains_long, "income": income, "sources": sources, "capital_gains_short": capital_gains_short, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def colorado_and_new_jersey(self, capital_gains_short=0, dependents=0, sources=False, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_long=0, filing_status='single', income=100000, credits_state=0):
        """
        --- Parameters ----
        
        Colorado:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        
        New Jersey:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Colorado:
        - https://leg.colorado.gov/agencies/legislative-council-staff/individual-income-tax%C2%A0
          https://leg.colorado.gov/agencies/legislative-council-staff/individual-income-tax%C2%A0
          https://tax.colorado.gov/sites/tax/files/documents/ITT_Capital_Gain_Subtraction_Feb_2022.pdf
          https://www.auroragov.org/business_services/taxes/occupational_privilege_tax
          https://www.denvergov.org/Government/Agencies-Departments-Offices/Agencies-Departments-Offices-Directory/Department-of-Finance/Our-Divisions/Treasury/Business-Tax-Information#:~:text=Occupational%20privilege%20tax
         -%28show%20below%29&text=Each%20taxable%20employee%20is%20liable
         month%20for%20each%20taxable%20employee
          https://www.glendale.co.us/355/Occupational-Privilege-Tax
          https://www.greenwoodvillage.com/1220/Occupational-Privilege-Tax-OPT
          https://ci.sheridan.co.us/288/Occupational-Privilege-Tax
        
        New Jersey:
        -         
        - https://www.state.nj.us/treasury/taxation/taxtables.shtml
          https://www.state.nj.us/treasury/taxation/njit13.shtml
          https://www.nj.gov/treasury/taxation/njit25.shtml
          https://www.state.nj.us/treasury/taxation/individuals/freqqiti.shtml
          https://www.state.nj.us/treasury/taxation/localtax.shtml
        """
        
        # Check if the parameters are set upstream in the instance
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Colorado', 'state_occupation': 'NewJersey', "capital_gains_short": capital_gains_short, "dependents": dependents, "sources": sources, "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def colorado_and_new_mexico(self, county_residence='Other', state_agi=75000, capital_gains_short=0, capital_gains_long=0, filing_status='single', income=100000, dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Colorado:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        
        New Mexico:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Colorado:
        - https://leg.colorado.gov/agencies/legislative-council-staff/individual-income-tax%C2%A0
          https://leg.colorado.gov/agencies/legislative-council-staff/individual-income-tax%C2%A0
          https://tax.colorado.gov/sites/tax/files/documents/ITT_Capital_Gain_Subtraction_Feb_2022.pdf
          https://www.auroragov.org/business_services/taxes/occupational_privilege_tax
          https://www.denvergov.org/Government/Agencies-Departments-Offices/Agencies-Departments-Offices-Directory/Department-of-Finance/Our-Divisions/Treasury/Business-Tax-Information#:~:text=Occupational%20privilege%20tax
         -%28show%20below%29&text=Each%20taxable%20employee%20is%20liable
         month%20for%20each%20taxable%20employee
          https://www.glendale.co.us/355/Occupational-Privilege-Tax
          https://www.greenwoodvillage.com/1220/Occupational-Privilege-Tax-OPT
          https://ci.sheridan.co.us/288/Occupational-Privilege-Tax
        
        New Mexico:
        -         
        - https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/fdf3c548-8aba-4b9c-9eb4-bb564c716015/FYI-104.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/0bf943fd-652e-4400-bb1b-cd8397eb5a95/2022pit-adj.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/2f1a6781-9534-4436-b427-1557f9592099/2022pit-adj-ins.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/0bf943fd-652e-4400-bb1b-cd8397eb5a95/2022pit-adj.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/1afc56af-ea90-4d48-82e5-1f9aeb43255a/PITbook2022.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/fdf3c548-8aba-4b9c-9eb4-bb564c716015/FYI-104.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Colorado', 'state_occupation': 'NewMexico', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def colorado_and_new_york(self, capital_gains_short=0, dependents=0, sources=False, county_residence='Other', state_agi=75000, capital_gains_long=0, filing_status='single', income=100000, credits_state=0):
        """
        --- Parameters ----
        
        Colorado:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        
        New York:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Colorado:
        - https://leg.colorado.gov/agencies/legislative-council-staff/individual-income-tax%C2%A0
          https://leg.colorado.gov/agencies/legislative-council-staff/individual-income-tax%C2%A0
          https://tax.colorado.gov/sites/tax/files/documents/ITT_Capital_Gain_Subtraction_Feb_2022.pdf
          https://www.auroragov.org/business_services/taxes/occupational_privilege_tax
          https://www.denvergov.org/Government/Agencies-Departments-Offices/Agencies-Departments-Offices-Directory/Department-of-Finance/Our-Divisions/Treasury/Business-Tax-Information#:~:text=Occupational%20privilege%20tax
         -%28show%20below%29&text=Each%20taxable%20employee%20is%20liable
         month%20for%20each%20taxable%20employee
          https://www.glendale.co.us/355/Occupational-Privilege-Tax
          https://www.greenwoodvillage.com/1220/Occupational-Privilege-Tax-OPT
          https://ci.sheridan.co.us/288/Occupational-Privilege-Tax
        
        New York:
        -         
        - https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nys-tax-rate-schedule
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nys-tax-rate-schedule
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nys-tax-rate-schedule
          https://www.tax.ny.gov/pit/
          https://www.tax.ny.gov/pit/file/nonresident-faqs.htm
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#yonkers-tax-rate-schedule
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nyc-tax-rate-schedule
        """
        
        # Check if the parameters are set upstream in the instance
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Colorado', 'state_occupation': 'NewYork', "capital_gains_short": capital_gains_short, "dependents": dependents, "sources": sources, "county_residence": county_residence, "state_agi": state_agi, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def colorado_and_north_carolina(self, county_residence='Other', state_agi=75000, capital_gains_short=0, capital_gains_long=0, filing_status='single', income=100000, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Colorado:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        
        North Carolina:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Colorado:
        - https://leg.colorado.gov/agencies/legislative-council-staff/individual-income-tax%C2%A0
          https://leg.colorado.gov/agencies/legislative-council-staff/individual-income-tax%C2%A0
          https://tax.colorado.gov/sites/tax/files/documents/ITT_Capital_Gain_Subtraction_Feb_2022.pdf
          https://www.auroragov.org/business_services/taxes/occupational_privilege_tax
          https://www.denvergov.org/Government/Agencies-Departments-Offices/Agencies-Departments-Offices-Directory/Department-of-Finance/Our-Divisions/Treasury/Business-Tax-Information#:~:text=Occupational%20privilege%20tax
         -%28show%20below%29&text=Each%20taxable%20employee%20is%20liable
         month%20for%20each%20taxable%20employee
          https://www.glendale.co.us/355/Occupational-Privilege-Tax
          https://www.greenwoodvillage.com/1220/Occupational-Privilege-Tax-OPT
          https://ci.sheridan.co.us/288/Occupational-Privilege-Tax
        
        North Carolina:
        -         
        - https://www.ncdor.gov/taxes-forms/tax-rate-schedules
          https://www.ncdor.gov/taxes-forms/individual-income-tax/north-carolina-standard-deduction-or-north-carolina-itemized-deductions
          https://www.ncdor.gov/taxes-forms/individual-income-tax
          https://www.ncdor.gov/taxes-forms/individual-income-tax
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Colorado', 'state_occupation': 'NorthCarolina', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def colorado_and_north_dakota(self, capital_gains_short=0, dependents=0, sources=False, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_long=0, filing_status='single', income=100000, credits_state=0):
        """
        --- Parameters ----
        
        Colorado:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        
        North Dakota:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Colorado:
        - https://leg.colorado.gov/agencies/legislative-council-staff/individual-income-tax%C2%A0
          https://leg.colorado.gov/agencies/legislative-council-staff/individual-income-tax%C2%A0
          https://tax.colorado.gov/sites/tax/files/documents/ITT_Capital_Gain_Subtraction_Feb_2022.pdf
          https://www.auroragov.org/business_services/taxes/occupational_privilege_tax
          https://www.denvergov.org/Government/Agencies-Departments-Offices/Agencies-Departments-Offices-Directory/Department-of-Finance/Our-Divisions/Treasury/Business-Tax-Information#:~:text=Occupational%20privilege%20tax
         -%28show%20below%29&text=Each%20taxable%20employee%20is%20liable
         month%20for%20each%20taxable%20employee
          https://www.glendale.co.us/355/Occupational-Privilege-Tax
          https://www.greenwoodvillage.com/1220/Occupational-Privilege-Tax-OPT
          https://ci.sheridan.co.us/288/Occupational-Privilege-Tax
        
        North Dakota:
        -         
        - https://www.tax.nd.gov/forms
          https://www.tax.nd.gov/sites/www/files/documents/forms/individual/2022-iit/2022-individual-income-tax-booklet.pdf
          https://www.tax.nd.gov/sites/www/files/documents/forms/2022-individual-income-tax-booklet.pdf
          https://www.tax.nd.gov/tax-exemptions-credits/income-tax-exemptions-credits
          https://www.tax.nd.gov/sites/www/files/documents/forms/business/ndwrfillable.pdf
          https://www.tax.nd.gov/news/tax-legislative-changes/significant-changes-law/individual-income-tax-history
          https://www.tax.nd.gov/forms
        """
        
        # Check if the parameters are set upstream in the instance
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Colorado', 'state_occupation': 'NorthDakota', "capital_gains_short": capital_gains_short, "dependents": dependents, "sources": sources, "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def colorado_and_ohio(self, capital_gains_short=0, dependents=0, sources=False, county_residence='Other', state_agi=75000, capital_gains_long=0, filing_status='single', income=100000, credits_state=0):
        """
        --- Parameters ----
        
        Colorado:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        
        Ohio:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Colorado:
        - https://leg.colorado.gov/agencies/legislative-council-staff/individual-income-tax%C2%A0
          https://leg.colorado.gov/agencies/legislative-council-staff/individual-income-tax%C2%A0
          https://tax.colorado.gov/sites/tax/files/documents/ITT_Capital_Gain_Subtraction_Feb_2022.pdf
          https://www.auroragov.org/business_services/taxes/occupational_privilege_tax
          https://www.denvergov.org/Government/Agencies-Departments-Offices/Agencies-Departments-Offices-Directory/Department-of-Finance/Our-Divisions/Treasury/Business-Tax-Information#:~:text=Occupational%20privilege%20tax
         -%28show%20below%29&text=Each%20taxable%20employee%20is%20liable
         month%20for%20each%20taxable%20employee
          https://www.glendale.co.us/355/Occupational-Privilege-Tax
          https://www.greenwoodvillage.com/1220/Occupational-Privilege-Tax-OPT
          https://ci.sheridan.co.us/288/Occupational-Privilege-Tax
        
        Ohio:
        -         
        - https://tax.ohio.gov/individual/resources/annual-tax-rates
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://tax.ohio.gov/static/forms/employer_withholding/generic/wth_it4nr.pdf
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://www.ritaohio.com/TaxRatesTable
        """
        
        # Check if the parameters are set upstream in the instance
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Colorado', 'state_occupation': 'Ohio', "capital_gains_short": capital_gains_short, "dependents": dependents, "sources": sources, "county_residence": county_residence, "state_agi": state_agi, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def colorado_and_oklahoma(self, capital_gains_short=0, dependents=0, sources=False, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_long=0, filing_status='single', income=100000, credits_state=0):
        """
        --- Parameters ----
        
        Colorado:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        
        Oklahoma:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Colorado:
        - https://leg.colorado.gov/agencies/legislative-council-staff/individual-income-tax%C2%A0
          https://leg.colorado.gov/agencies/legislative-council-staff/individual-income-tax%C2%A0
          https://tax.colorado.gov/sites/tax/files/documents/ITT_Capital_Gain_Subtraction_Feb_2022.pdf
          https://www.auroragov.org/business_services/taxes/occupational_privilege_tax
          https://www.denvergov.org/Government/Agencies-Departments-Offices/Agencies-Departments-Offices-Directory/Department-of-Finance/Our-Divisions/Treasury/Business-Tax-Information#:~:text=Occupational%20privilege%20tax
         -%28show%20below%29&text=Each%20taxable%20employee%20is%20liable
         month%20for%20each%20taxable%20employee
          https://www.glendale.co.us/355/Occupational-Privilege-Tax
          https://www.greenwoodvillage.com/1220/Occupational-Privilege-Tax-OPT
          https://ci.sheridan.co.us/288/Occupational-Privilege-Tax
        
        Oklahoma:
        -         
        - https://oklahoma.gov/content/dam/ok/en/tax/documents/resources/publications/businesses/withholding-tables/WHTables-2022.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Colorado', 'state_occupation': 'Oklahoma', "capital_gains_short": capital_gains_short, "dependents": dependents, "sources": sources, "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def colorado_and_oregon(self, capital_gains_short=0, dependents=0, sources=False, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_long=0, filing_status='single', income=100000, credits_state=0):
        """
        --- Parameters ----
        
        Colorado:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        
        Oregon:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Colorado:
        - https://leg.colorado.gov/agencies/legislative-council-staff/individual-income-tax%C2%A0
          https://leg.colorado.gov/agencies/legislative-council-staff/individual-income-tax%C2%A0
          https://tax.colorado.gov/sites/tax/files/documents/ITT_Capital_Gain_Subtraction_Feb_2022.pdf
          https://www.auroragov.org/business_services/taxes/occupational_privilege_tax
          https://www.denvergov.org/Government/Agencies-Departments-Offices/Agencies-Departments-Offices-Directory/Department-of-Finance/Our-Divisions/Treasury/Business-Tax-Information#:~:text=Occupational%20privilege%20tax
         -%28show%20below%29&text=Each%20taxable%20employee%20is%20liable
         month%20for%20each%20taxable%20employee
          https://www.glendale.co.us/355/Occupational-Privilege-Tax
          https://www.greenwoodvillage.com/1220/Occupational-Privilege-Tax-OPT
          https://ci.sheridan.co.us/288/Occupational-Privilege-Tax
        
        Oregon:
        -         
        - https://www.oregon.gov/dor/programs/individuals/Pages/PIT.aspx
          https://www.oregon.gov/dor/programs/individuals/Documents/Full%20year%20resident
         %20Form%20or-40%20filers.pdf
          https://www.oregon.gov/dor/programs/individuals/Documents/Part-year%20and%20nonresident
         %20Form%20or-40-P%20and%20or-40-N%20filers.pdf.pdf
          https://www.oregon.gov/dor/forms/FormsPubs/form-or-40_101-040_2022.pdf
          https://www.oregon.gov/dor/forms/FormsPubs/form-or-40_101-040_2022.pdf
          https://www.oregon.gov/dor/forms/FormsPubs/form-or-40_101-040_2022.pdf
          https://www.portland.gov/revenue/personal-tax
        """
        
        # Check if the parameters are set upstream in the instance
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Colorado', 'state_occupation': 'Oregon', "capital_gains_short": capital_gains_short, "dependents": dependents, "sources": sources, "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def colorado_and_pennsylvania(self, county_residence='Other', state_agi=75000, county_occupation='Other', capital_gains_short=0, capital_gains_long=0, income=100000, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Colorado:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        
        Pennsylvania:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        county_residence : str, optional
            County of residence. Default is Other.
        county_occupation : str, optional
            County of occupation. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Colorado:
        - https://leg.colorado.gov/agencies/legislative-council-staff/individual-income-tax%C2%A0
          https://leg.colorado.gov/agencies/legislative-council-staff/individual-income-tax%C2%A0
          https://tax.colorado.gov/sites/tax/files/documents/ITT_Capital_Gain_Subtraction_Feb_2022.pdf
          https://www.auroragov.org/business_services/taxes/occupational_privilege_tax
          https://www.denvergov.org/Government/Agencies-Departments-Offices/Agencies-Departments-Offices-Directory/Department-of-Finance/Our-Divisions/Treasury/Business-Tax-Information#:~:text=Occupational%20privilege%20tax
         -%28show%20below%29&text=Each%20taxable%20employee%20is%20liable
         month%20for%20each%20taxable%20employee
          https://www.glendale.co.us/355/Occupational-Privilege-Tax
          https://www.greenwoodvillage.com/1220/Occupational-Privilege-Tax-OPT
          https://ci.sheridan.co.us/288/Occupational-Privilege-Tax
        
        Pennsylvania:
        -         
        - https://www.revenue.pa.gov/Tax%20Rates/Pages/default.aspx
          https://www.revenue.pa.gov/FormsandPublications/FormsforIndividuals/PIT/Documents/2022/2022_pa-40in.pdf
          https://www.revenue.pa.gov/FormsandPublications/FormsforBusinesses/EmployerWithholding/Documents/rev-419.pdf
          https://www.revenue.pa.gov/FormsandPublications/FormsforIndividuals/PIT/Documents/2022/2022_pa-40in.pdf
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-25.htm
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        county_occupation = self.county_occupation if hasattr(self, 'county_occupation') else county_occupation
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        income = self.income if hasattr(self, 'income') else income
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Colorado', 'state_occupation': 'Pennsylvania', "county_residence": county_residence, "state_agi": state_agi, "county_occupation": county_occupation, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "income": income, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def colorado_and_rhode_island(self, county_residence='Other', state_agi=75000, capital_gains_short=0, capital_gains_long=0, filing_status='single', income=100000, dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Colorado:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        
        Rhode Island:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Colorado:
        - https://leg.colorado.gov/agencies/legislative-council-staff/individual-income-tax%C2%A0
          https://leg.colorado.gov/agencies/legislative-council-staff/individual-income-tax%C2%A0
          https://tax.colorado.gov/sites/tax/files/documents/ITT_Capital_Gain_Subtraction_Feb_2022.pdf
          https://www.auroragov.org/business_services/taxes/occupational_privilege_tax
          https://www.denvergov.org/Government/Agencies-Departments-Offices/Agencies-Departments-Offices-Directory/Department-of-Finance/Our-Divisions/Treasury/Business-Tax-Information#:~:text=Occupational%20privilege%20tax
         -%28show%20below%29&text=Each%20taxable%20employee%20is%20liable
         month%20for%20each%20taxable%20employee
          https://www.glendale.co.us/355/Occupational-Privilege-Tax
          https://www.greenwoodvillage.com/1220/Occupational-Privilege-Tax-OPT
          https://ci.sheridan.co.us/288/Occupational-Privilege-Tax
        
        Rhode Island:
        -         
        - https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2022_1040WE_w_0.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Colorado', 'state_occupation': 'RhodeIsland', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def colorado_and_south_carolina(self, capital_gains_short=0, dependents=0, sources=False, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_long=0, filing_status='single', income=100000, credits_state=0):
        """
        --- Parameters ----
        
        Colorado:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        
        South Carolina:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Colorado:
        - https://leg.colorado.gov/agencies/legislative-council-staff/individual-income-tax%C2%A0
          https://leg.colorado.gov/agencies/legislative-council-staff/individual-income-tax%C2%A0
          https://tax.colorado.gov/sites/tax/files/documents/ITT_Capital_Gain_Subtraction_Feb_2022.pdf
          https://www.auroragov.org/business_services/taxes/occupational_privilege_tax
          https://www.denvergov.org/Government/Agencies-Departments-Offices/Agencies-Departments-Offices-Directory/Department-of-Finance/Our-Divisions/Treasury/Business-Tax-Information#:~:text=Occupational%20privilege%20tax
         -%28show%20below%29&text=Each%20taxable%20employee%20is%20liable
         month%20for%20each%20taxable%20employee
          https://www.glendale.co.us/355/Occupational-Privilege-Tax
          https://www.greenwoodvillage.com/1220/Occupational-Privilege-Tax-OPT
          https://ci.sheridan.co.us/288/Occupational-Privilege-Tax
        
        South Carolina:
        -         
        - https://dor.sc.gov/resources-site/lawandpolicy/Advisory%20Opinions/IL22-15.pdf
          https://dor.sc.gov/forms-site/Forms/IITPacket_2022.pdf
          https://www.scstatehouse.gov/code/t12c006.php
          https://dor.sc.gov/resources-site/lawandpolicy/Advisory%20Opinions/IL22-15.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Colorado', 'state_occupation': 'SouthCarolina', "capital_gains_short": capital_gains_short, "dependents": dependents, "sources": sources, "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def colorado_and_south_dakota(self, county_residence='Other', state_agi=75000, capital_gains_long=0, income=100000, sources=False, capital_gains_short=0, credits_state=0):
        """
        --- Parameters ----
        
        Colorado:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        
        South Dakota:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Colorado:
        - https://leg.colorado.gov/agencies/legislative-council-staff/individual-income-tax%C2%A0
          https://leg.colorado.gov/agencies/legislative-council-staff/individual-income-tax%C2%A0
          https://tax.colorado.gov/sites/tax/files/documents/ITT_Capital_Gain_Subtraction_Feb_2022.pdf
          https://www.auroragov.org/business_services/taxes/occupational_privilege_tax
          https://www.denvergov.org/Government/Agencies-Departments-Offices/Agencies-Departments-Offices-Directory/Department-of-Finance/Our-Divisions/Treasury/Business-Tax-Information#:~:text=Occupational%20privilege%20tax
         -%28show%20below%29&text=Each%20taxable%20employee%20is%20liable
         month%20for%20each%20taxable%20employee
          https://www.glendale.co.us/355/Occupational-Privilege-Tax
          https://www.greenwoodvillage.com/1220/Occupational-Privilege-Tax-OPT
          https://ci.sheridan.co.us/288/Occupational-Privilege-Tax
        
        South Dakota:
        -         
        - https://dor.sd.gov/individuals/taxes/
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        income = self.income if hasattr(self, 'income') else income
        sources = self.sources if hasattr(self, 'sources') else sources
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Colorado', 'state_occupation': 'SouthDakota', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_long": capital_gains_long, "income": income, "sources": sources, "capital_gains_short": capital_gains_short, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def colorado_and_tennessee(self, county_residence='Other', state_agi=75000, capital_gains_long=0, income=100000, sources=False, capital_gains_short=0, credits_state=0):
        """
        --- Parameters ----
        
        Colorado:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        
        Tennessee:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Colorado:
        - https://leg.colorado.gov/agencies/legislative-council-staff/individual-income-tax%C2%A0
          https://leg.colorado.gov/agencies/legislative-council-staff/individual-income-tax%C2%A0
          https://tax.colorado.gov/sites/tax/files/documents/ITT_Capital_Gain_Subtraction_Feb_2022.pdf
          https://www.auroragov.org/business_services/taxes/occupational_privilege_tax
          https://www.denvergov.org/Government/Agencies-Departments-Offices/Agencies-Departments-Offices-Directory/Department-of-Finance/Our-Divisions/Treasury/Business-Tax-Information#:~:text=Occupational%20privilege%20tax
         -%28show%20below%29&text=Each%20taxable%20employee%20is%20liable
         month%20for%20each%20taxable%20employee
          https://www.glendale.co.us/355/Occupational-Privilege-Tax
          https://www.greenwoodvillage.com/1220/Occupational-Privilege-Tax-OPT
          https://ci.sheridan.co.us/288/Occupational-Privilege-Tax
        
        Tennessee:
        -         
        - https://www.tn.gov/revenue/taxes.html
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        income = self.income if hasattr(self, 'income') else income
        sources = self.sources if hasattr(self, 'sources') else sources
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Colorado', 'state_occupation': 'Tennessee', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_long": capital_gains_long, "income": income, "sources": sources, "capital_gains_short": capital_gains_short, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def colorado_and_texas(self, county_residence='Other', state_agi=75000, capital_gains_long=0, income=100000, sources=False, capital_gains_short=0, credits_state=0):
        """
        --- Parameters ----
        
        Colorado:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        
        Texas:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Colorado:
        - https://leg.colorado.gov/agencies/legislative-council-staff/individual-income-tax%C2%A0
          https://leg.colorado.gov/agencies/legislative-council-staff/individual-income-tax%C2%A0
          https://tax.colorado.gov/sites/tax/files/documents/ITT_Capital_Gain_Subtraction_Feb_2022.pdf
          https://www.auroragov.org/business_services/taxes/occupational_privilege_tax
          https://www.denvergov.org/Government/Agencies-Departments-Offices/Agencies-Departments-Offices-Directory/Department-of-Finance/Our-Divisions/Treasury/Business-Tax-Information#:~:text=Occupational%20privilege%20tax
         -%28show%20below%29&text=Each%20taxable%20employee%20is%20liable
         month%20for%20each%20taxable%20employee
          https://www.glendale.co.us/355/Occupational-Privilege-Tax
          https://www.greenwoodvillage.com/1220/Occupational-Privilege-Tax-OPT
          https://ci.sheridan.co.us/288/Occupational-Privilege-Tax
        
        Texas:
        -         
        - https://comptroller.texas.gov/taxes/
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        income = self.income if hasattr(self, 'income') else income
        sources = self.sources if hasattr(self, 'sources') else sources
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Colorado', 'state_occupation': 'Texas', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_long": capital_gains_long, "income": income, "sources": sources, "capital_gains_short": capital_gains_short, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def colorado_and_utah(self, county_residence='Other', state_agi=75000, capital_gains_short=0, capital_gains_long=0, filing_status='single', income=100000, dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Colorado:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        
        Utah:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Colorado:
        - https://leg.colorado.gov/agencies/legislative-council-staff/individual-income-tax%C2%A0
          https://leg.colorado.gov/agencies/legislative-council-staff/individual-income-tax%C2%A0
          https://tax.colorado.gov/sites/tax/files/documents/ITT_Capital_Gain_Subtraction_Feb_2022.pdf
          https://www.auroragov.org/business_services/taxes/occupational_privilege_tax
          https://www.denvergov.org/Government/Agencies-Departments-Offices/Agencies-Departments-Offices-Directory/Department-of-Finance/Our-Divisions/Treasury/Business-Tax-Information#:~:text=Occupational%20privilege%20tax
         -%28show%20below%29&text=Each%20taxable%20employee%20is%20liable
         month%20for%20each%20taxable%20employee
          https://www.glendale.co.us/355/Occupational-Privilege-Tax
          https://www.greenwoodvillage.com/1220/Occupational-Privilege-Tax-OPT
          https://ci.sheridan.co.us/288/Occupational-Privilege-Tax
        
        Utah:
        -         
        - https://incometax.utah.gov/paying/tax-rates
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Colorado', 'state_occupation': 'Utah', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def colorado_and_vermont(self, capital_gains_short=0, dependents=0, sources=False, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_long=0, filing_status='single', income=100000, credits_state=0):
        """
        --- Parameters ----
        
        Colorado:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        
        Vermont:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Colorado:
        - https://leg.colorado.gov/agencies/legislative-council-staff/individual-income-tax%C2%A0
          https://leg.colorado.gov/agencies/legislative-council-staff/individual-income-tax%C2%A0
          https://tax.colorado.gov/sites/tax/files/documents/ITT_Capital_Gain_Subtraction_Feb_2022.pdf
          https://www.auroragov.org/business_services/taxes/occupational_privilege_tax
          https://www.denvergov.org/Government/Agencies-Departments-Offices/Agencies-Departments-Offices-Directory/Department-of-Finance/Our-Divisions/Treasury/Business-Tax-Information#:~:text=Occupational%20privilege%20tax
         -%28show%20below%29&text=Each%20taxable%20employee%20is%20liable
         month%20for%20each%20taxable%20employee
          https://www.glendale.co.us/355/Occupational-Privilege-Tax
          https://www.greenwoodvillage.com/1220/Occupational-Privilege-Tax-OPT
          https://ci.sheridan.co.us/288/Occupational-Privilege-Tax
        
        Vermont:
        -         
        - https://tax.vermont.gov/sites/tax/files/documents/GB-1210-2023.pdf
          https://tax.vermont.gov/sites/tax/files/documents/Income%20Booklet-2022.pdf
          https://tax.vermont.gov/sites/tax/files/documents/Income%20Booklet-2022.pdf
          https://tax.vermont.gov/sites/tax/files/documents/Income%20Booklet-2022.pdf
          https://tax.vermont.gov/individuals/personal-income-tax/tax-credits
          https://tax.vermont.gov/individuals/personal-income-tax/taxable-income
          https://tax.vermont.gov/individuals/personal-income-tax/local-option-tax
        """
        
        # Check if the parameters are set upstream in the instance
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Colorado', 'state_occupation': 'Vermont', "capital_gains_short": capital_gains_short, "dependents": dependents, "sources": sources, "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def colorado_and_virginia(self, capital_gains_short=0, dependents=0, sources=False, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_long=0, filing_status='single', income=100000, credits_state=0):
        """
        --- Parameters ----
        
        Colorado:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        
        Virginia:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Colorado:
        - https://leg.colorado.gov/agencies/legislative-council-staff/individual-income-tax%C2%A0
          https://leg.colorado.gov/agencies/legislative-council-staff/individual-income-tax%C2%A0
          https://tax.colorado.gov/sites/tax/files/documents/ITT_Capital_Gain_Subtraction_Feb_2022.pdf
          https://www.auroragov.org/business_services/taxes/occupational_privilege_tax
          https://www.denvergov.org/Government/Agencies-Departments-Offices/Agencies-Departments-Offices-Directory/Department-of-Finance/Our-Divisions/Treasury/Business-Tax-Information#:~:text=Occupational%20privilege%20tax
         -%28show%20below%29&text=Each%20taxable%20employee%20is%20liable
         month%20for%20each%20taxable%20employee
          https://www.glendale.co.us/355/Occupational-Privilege-Tax
          https://www.greenwoodvillage.com/1220/Occupational-Privilege-Tax-OPT
          https://ci.sheridan.co.us/288/Occupational-Privilege-Tax
        
        Virginia:
        -         
        - https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/reciprocity
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Colorado', 'state_occupation': 'Virginia', "capital_gains_short": capital_gains_short, "dependents": dependents, "sources": sources, "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def colorado_and_washington(self, county_residence='Other', state_agi=75000, capital_gains_short=0, capital_gains_long=0, income=100000, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Colorado:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        
        Washington:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Colorado:
        - https://leg.colorado.gov/agencies/legislative-council-staff/individual-income-tax%C2%A0
          https://leg.colorado.gov/agencies/legislative-council-staff/individual-income-tax%C2%A0
          https://tax.colorado.gov/sites/tax/files/documents/ITT_Capital_Gain_Subtraction_Feb_2022.pdf
          https://www.auroragov.org/business_services/taxes/occupational_privilege_tax
          https://www.denvergov.org/Government/Agencies-Departments-Offices/Agencies-Departments-Offices-Directory/Department-of-Finance/Our-Divisions/Treasury/Business-Tax-Information#:~:text=Occupational%20privilege%20tax
         -%28show%20below%29&text=Each%20taxable%20employee%20is%20liable
         month%20for%20each%20taxable%20employee
          https://www.glendale.co.us/355/Occupational-Privilege-Tax
          https://www.greenwoodvillage.com/1220/Occupational-Privilege-Tax-OPT
          https://ci.sheridan.co.us/288/Occupational-Privilege-Tax
        
        Washington:
        -         
        - https://dor.wa.gov/taxes-rates/income-tax
          https://dor.wa.gov/taxes-rates/other-taxes/capital-gains-tax
          https://dor.wa.gov/taxes-rates/income-tax
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        income = self.income if hasattr(self, 'income') else income
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Colorado', 'state_occupation': 'Washington', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "income": income, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def colorado_and_west_virginia(self, pay_periods=24, capital_gains_short=0, county_occupation='Other', dependents=0, sources=False, county_residence='Other', state_agi=75000, capital_gains_long=0, filing_status='single', income=100000, credits_state=0):
        """
        --- Parameters ----
        
        Colorado:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        
        West Virginia:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_occupation : str, optional
            County of occupation. Default is Other.
        pay_periods : int, optional
            Number of pay periods. Options are 13, 24, 26, 52. Default is 24.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Colorado:
        - https://leg.colorado.gov/agencies/legislative-council-staff/individual-income-tax%C2%A0
          https://leg.colorado.gov/agencies/legislative-council-staff/individual-income-tax%C2%A0
          https://tax.colorado.gov/sites/tax/files/documents/ITT_Capital_Gain_Subtraction_Feb_2022.pdf
          https://www.auroragov.org/business_services/taxes/occupational_privilege_tax
          https://www.denvergov.org/Government/Agencies-Departments-Offices/Agencies-Departments-Offices-Directory/Department-of-Finance/Our-Divisions/Treasury/Business-Tax-Information#:~:text=Occupational%20privilege%20tax
         -%28show%20below%29&text=Each%20taxable%20employee%20is%20liable
         month%20for%20each%20taxable%20employee
          https://www.glendale.co.us/355/Occupational-Privilege-Tax
          https://www.greenwoodvillage.com/1220/Occupational-Privilege-Tax-OPT
          https://ci.sheridan.co.us/288/Occupational-Privilege-Tax
        
        West Virginia:
        -         
        - https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://tax.wv.gov/Documents/TSD/tsd438.pdf
          https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2011/Taxes-11-35att.html
        """
        
        # Check if the parameters are set upstream in the instance
        pay_periods = self.pay_periods if hasattr(self, 'pay_periods') else pay_periods
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        county_occupation = self.county_occupation if hasattr(self, 'county_occupation') else county_occupation
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Colorado', 'state_occupation': 'WestVirginia', "pay_periods": pay_periods, "capital_gains_short": capital_gains_short, "county_occupation": county_occupation, "dependents": dependents, "sources": sources, "county_residence": county_residence, "state_agi": state_agi, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def colorado_and_wisconsin(self, county_residence='Other', state_agi=75000, capital_gains_short=0, capital_gains_long=0, filing_status='single', income=100000, dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Colorado:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        
        Wisconsin:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Colorado:
        - https://leg.colorado.gov/agencies/legislative-council-staff/individual-income-tax%C2%A0
          https://leg.colorado.gov/agencies/legislative-council-staff/individual-income-tax%C2%A0
          https://tax.colorado.gov/sites/tax/files/documents/ITT_Capital_Gain_Subtraction_Feb_2022.pdf
          https://www.auroragov.org/business_services/taxes/occupational_privilege_tax
          https://www.denvergov.org/Government/Agencies-Departments-Offices/Agencies-Departments-Offices-Directory/Department-of-Finance/Our-Divisions/Treasury/Business-Tax-Information#:~:text=Occupational%20privilege%20tax
         -%28show%20below%29&text=Each%20taxable%20employee%20is%20liable
         month%20for%20each%20taxable%20employee
          https://www.glendale.co.us/355/Occupational-Privilege-Tax
          https://www.greenwoodvillage.com/1220/Occupational-Privilege-Tax-OPT
          https://ci.sheridan.co.us/288/Occupational-Privilege-Tax
        
        Wisconsin:
        -         
        - https://www.revenue.wi.gov/TaxForms2023/2023-Form1-ES-Inst.pdf
          https://www.revenue.wi.gov/TaxForms2023/2023-Form1-ES-Inst.pdf
          https://www.revenue.wi.gov/TaxForms2023/2023-Form1-ES-Inst.pdf
          https://www.revenue.wi.gov/Pages/FAQS/pcs-work.aspx
          https://www.revenue.wi.gov/Dor%20Publications/pb103.pdf
          https://www.revenue.wi.gov/Pages/FAQS/pcs-work.aspx
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Colorado', 'state_occupation': 'Wisconsin', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def colorado_and_wyoming(self, county_residence='Other', state_agi=75000, capital_gains_long=0, income=100000, sources=False, capital_gains_short=0, credits_state=0):
        """
        --- Parameters ----
        
        Colorado:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        
        Wyoming:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Colorado:
        - https://leg.colorado.gov/agencies/legislative-council-staff/individual-income-tax%C2%A0
          https://leg.colorado.gov/agencies/legislative-council-staff/individual-income-tax%C2%A0
          https://tax.colorado.gov/sites/tax/files/documents/ITT_Capital_Gain_Subtraction_Feb_2022.pdf
          https://www.auroragov.org/business_services/taxes/occupational_privilege_tax
          https://www.denvergov.org/Government/Agencies-Departments-Offices/Agencies-Departments-Offices-Directory/Department-of-Finance/Our-Divisions/Treasury/Business-Tax-Information#:~:text=Occupational%20privilege%20tax
         -%28show%20below%29&text=Each%20taxable%20employee%20is%20liable
         month%20for%20each%20taxable%20employee
          https://www.glendale.co.us/355/Occupational-Privilege-Tax
          https://www.greenwoodvillage.com/1220/Occupational-Privilege-Tax-OPT
          https://ci.sheridan.co.us/288/Occupational-Privilege-Tax
        
        Wyoming:
        -         
        - https://revenue.wyo.gov/
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        income = self.income if hasattr(self, 'income') else income
        sources = self.sources if hasattr(self, 'sources') else sources
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Colorado', 'state_occupation': 'Wyoming', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_long": capital_gains_long, "income": income, "sources": sources, "capital_gains_short": capital_gains_short, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def connecticut_and_alabama(self, capital_gains_short=0, income=100000, dependents=0, sources=False, county_residence='Other', state_agi=75000, capital_gains_long=0, filing_status='single', credits_state=0):
        """
        --- Parameters ----
        
        Connecticut:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        
        Alabama:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Connecticut:
        - https://portal.ct.gov/-/media/DRS/Forms/2022/Income/2022-CT-1040-Instructions_1222.pdf
          https://www.cga.ct.gov/2022/rpt/pdf/2022-R-0108.pdf
          https://www.cga.ct.gov/current/pub/chap_229.htm#sec_12-703
          https://portal.ct.gov/-/media/DRS/Forms/2022/Income/2022-CT-1040-Instructions_1222.pdf
          "https://portal.ct.gov/DRS/Publications/TSSNs/TSSN-29#:~:text=An%20individuals%20net%20capital%20gains
         portion%20of%20Social%20Security%20benefits"
          "https://portal.ct.gov/DRS/Publications/TSSNs/TSSN-29#:~:text=An%20individuals%20net%20capital%20gains
         portion%20of%20Social%20Security%20benefits"
          https://portal.ct.gov/DRS/Individuals/Resident-Income-Tax/Tax-Information
        
        Alabama:
        -         
        - https://www.revenue.alabama.gov/ultraviewer/viewer/basic_viewer/index.html?form=2023/02/22f40.pdf
          https://www.revenue.alabama.gov/faqs/what-is-alabamas-individual-income-tax-rate/
          https://www.revenue.alabama.gov/ultraviewer/viewer/basic_viewer/index.html?form=2023/02/22f40.pdf
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-24.htm?taxmap=true
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-24.htm?taxmap=true
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-24.htm?taxmap=true
          https://www.revenue.alabama.gov/individual-corporate/income-to-be-reported-on-the-alabama-income-tax-return/
          https://www.almonline.org/TaxRates.aspx#OccupationalTax
        """
        
        # Check if the parameters are set upstream in the instance
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Connecticut', 'state_occupation': 'Alabama', "capital_gains_short": capital_gains_short, "income": income, "dependents": dependents, "sources": sources, "county_residence": county_residence, "state_agi": state_agi, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def connecticut_and_alaska(self, capital_gains_long=0, state_agi=75000, filing_status='single', income=100000, sources=False, capital_gains_short=0, credits_state=0):
        """
        --- Parameters ----
        
        Connecticut:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        
        Alaska:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Connecticut:
        - https://portal.ct.gov/-/media/DRS/Forms/2022/Income/2022-CT-1040-Instructions_1222.pdf
          https://www.cga.ct.gov/2022/rpt/pdf/2022-R-0108.pdf
          https://www.cga.ct.gov/current/pub/chap_229.htm#sec_12-703
          https://portal.ct.gov/-/media/DRS/Forms/2022/Income/2022-CT-1040-Instructions_1222.pdf
          "https://portal.ct.gov/DRS/Publications/TSSNs/TSSN-29#:~:text=An%20individuals%20net%20capital%20gains
         portion%20of%20Social%20Security%20benefits"
          "https://portal.ct.gov/DRS/Publications/TSSNs/TSSN-29#:~:text=An%20individuals%20net%20capital%20gains
         portion%20of%20Social%20Security%20benefits"
          https://portal.ct.gov/DRS/Individuals/Resident-Income-Tax/Tax-Information
        
        Alaska:
        -         
        - https://www.commerce.alaska.gov/web/dcra/officeofthestateassessor/alaskataxfacts.aspx
        """
        
        # Check if the parameters are set upstream in the instance
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        sources = self.sources if hasattr(self, 'sources') else sources
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Connecticut', 'state_occupation': 'Alaska', "capital_gains_long": capital_gains_long, "state_agi": state_agi, "filing_status": filing_status, "income": income, "sources": sources, "capital_gains_short": capital_gains_short, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def connecticut_and_arizona(self, state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Connecticut:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        
        Arizona:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        credits_state : int, optional
            Credits for state taxes. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Connecticut:
        - https://portal.ct.gov/-/media/DRS/Forms/2022/Income/2022-CT-1040-Instructions_1222.pdf
          https://www.cga.ct.gov/2022/rpt/pdf/2022-R-0108.pdf
          https://www.cga.ct.gov/current/pub/chap_229.htm#sec_12-703
          https://portal.ct.gov/-/media/DRS/Forms/2022/Income/2022-CT-1040-Instructions_1222.pdf
          "https://portal.ct.gov/DRS/Publications/TSSNs/TSSN-29#:~:text=An%20individuals%20net%20capital%20gains
         portion%20of%20Social%20Security%20benefits"
          "https://portal.ct.gov/DRS/Publications/TSSNs/TSSN-29#:~:text=An%20individuals%20net%20capital%20gains
         portion%20of%20Social%20Security%20benefits"
          https://portal.ct.gov/DRS/Individuals/Resident-Income-Tax/Tax-Information
        
        Arizona:
        -         
        - https://azdor.gov/forms/individual/form-140-x-y-tables
          https://azdor.gov/forms/individual/form-140-x-y-tables
          https://azdor.gov/forms/individual/form-140-x-y-tables
          https://azdor.gov/businesses-arizona/withholding-tax/withholding-exceptions
          "https://azdor.gov/individuals/income-tax-filing-assistance/identifying-other-taxable-income#:~:text=In%20regards%20to%20capital%20gains
         the%20individuals%20regular%20tax%20rate.
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Connecticut', 'state_occupation': 'Arizona', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def connecticut_and_arkansas(self, capital_gains_short=0, income=100000, dependents=0, sources=False, state_agi=75000, filers_over_65=0, capital_gains_long=0, filing_status='single', credits_state=0):
        """
        --- Parameters ----
        
        Connecticut:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        
        Arkansas:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Connecticut:
        - https://portal.ct.gov/-/media/DRS/Forms/2022/Income/2022-CT-1040-Instructions_1222.pdf
          https://www.cga.ct.gov/2022/rpt/pdf/2022-R-0108.pdf
          https://www.cga.ct.gov/current/pub/chap_229.htm#sec_12-703
          https://portal.ct.gov/-/media/DRS/Forms/2022/Income/2022-CT-1040-Instructions_1222.pdf
          "https://portal.ct.gov/DRS/Publications/TSSNs/TSSN-29#:~:text=An%20individuals%20net%20capital%20gains
         portion%20of%20Social%20Security%20benefits"
          "https://portal.ct.gov/DRS/Publications/TSSNs/TSSN-29#:~:text=An%20individuals%20net%20capital%20gains
         portion%20of%20Social%20Security%20benefits"
          https://portal.ct.gov/DRS/Individuals/Resident-Income-Tax/Tax-Information
        
        Arkansas:
        -         
        - https://www.arkansasedc.com/why-arkansas/business-climate/tax-structure/personal-income-tax
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_and_AR1000NR_Instructions.pdf
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_FullYearResidentIndividualIncomeTaxReturn.pdf
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_and_AR1000NR_Instructions.pdf
          https://www.arkansasedc.com/why-arkansas/business-climate/tax-structure/capital-gains-tax-reduction
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_and_AR1000NR_Instructions.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Connecticut', 'state_occupation': 'Arkansas', "capital_gains_short": capital_gains_short, "income": income, "dependents": dependents, "sources": sources, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def connecticut_and_california(self, state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Connecticut:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        
        California:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Connecticut:
        - https://portal.ct.gov/-/media/DRS/Forms/2022/Income/2022-CT-1040-Instructions_1222.pdf
          https://www.cga.ct.gov/2022/rpt/pdf/2022-R-0108.pdf
          https://www.cga.ct.gov/current/pub/chap_229.htm#sec_12-703
          https://portal.ct.gov/-/media/DRS/Forms/2022/Income/2022-CT-1040-Instructions_1222.pdf
          "https://portal.ct.gov/DRS/Publications/TSSNs/TSSN-29#:~:text=An%20individuals%20net%20capital%20gains
         portion%20of%20Social%20Security%20benefits"
          "https://portal.ct.gov/DRS/Publications/TSSNs/TSSN-29#:~:text=An%20individuals%20net%20capital%20gains
         portion%20of%20Social%20Security%20benefits"
          https://portal.ct.gov/DRS/Individuals/Resident-Income-Tax/Tax-Information
        
        California:
        -         
        - https://www.ftb.ca.gov/forms/2022/2022-540-tax-rate-schedules.pdf
          https://www.ftb.ca.gov/file/personal/deductions/index.html
          https://www.ftb.ca.gov/about-ftb/newsroom/caleitc/eligibility-and-credit-information.html
          https://www.ftb.ca.gov/file/personal/income-types/capital-gains-and-losses.html#:~:text=California%20does%20not%20have%20a
         are%20taxed%20as%20ordinary%20income
          https://www.ftb.ca.gov/file/personal/
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Connecticut', 'state_occupation': 'California', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def connecticut_and_colorado(self, county_residence='Other', state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Connecticut:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        
        Colorado:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Connecticut:
        - https://portal.ct.gov/-/media/DRS/Forms/2022/Income/2022-CT-1040-Instructions_1222.pdf
          https://www.cga.ct.gov/2022/rpt/pdf/2022-R-0108.pdf
          https://www.cga.ct.gov/current/pub/chap_229.htm#sec_12-703
          https://portal.ct.gov/-/media/DRS/Forms/2022/Income/2022-CT-1040-Instructions_1222.pdf
          "https://portal.ct.gov/DRS/Publications/TSSNs/TSSN-29#:~:text=An%20individuals%20net%20capital%20gains
         portion%20of%20Social%20Security%20benefits"
          "https://portal.ct.gov/DRS/Publications/TSSNs/TSSN-29#:~:text=An%20individuals%20net%20capital%20gains
         portion%20of%20Social%20Security%20benefits"
          https://portal.ct.gov/DRS/Individuals/Resident-Income-Tax/Tax-Information
        
        Colorado:
        -         
        - https://leg.colorado.gov/agencies/legislative-council-staff/individual-income-tax%C2%A0
          https://leg.colorado.gov/agencies/legislative-council-staff/individual-income-tax%C2%A0
          https://tax.colorado.gov/sites/tax/files/documents/ITT_Capital_Gain_Subtraction_Feb_2022.pdf
          https://www.auroragov.org/business_services/taxes/occupational_privilege_tax
          https://www.denvergov.org/Government/Agencies-Departments-Offices/Agencies-Departments-Offices-Directory/Department-of-Finance/Our-Divisions/Treasury/Business-Tax-Information#:~:text=Occupational%20privilege%20tax
         -%28show%20below%29&text=Each%20taxable%20employee%20is%20liable
         month%20for%20each%20taxable%20employee
          https://www.glendale.co.us/355/Occupational-Privilege-Tax
          https://www.greenwoodvillage.com/1220/Occupational-Privilege-Tax-OPT
          https://ci.sheridan.co.us/288/Occupational-Privilege-Tax
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Connecticut', 'state_occupation': 'Colorado', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def connecticut_and_delaware(self, state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, federal_agi=75000, filing_status='single', sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Connecticut:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        
        Delaware:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        federal_agi : int, optional
            Adjusted Gross Income (AGI) for federal taxes. Default is 0. Sources - https://www.irs.gov/e-file-providers/definition-of-adjusted-gross-income, https://www.irs.gov/newsroom/taxpayers-should-know-the-difference-between-standard-and-itemized-deductions
        income : int, optional
            Gross income. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Connecticut:
        - https://portal.ct.gov/-/media/DRS/Forms/2022/Income/2022-CT-1040-Instructions_1222.pdf
          https://www.cga.ct.gov/2022/rpt/pdf/2022-R-0108.pdf
          https://www.cga.ct.gov/current/pub/chap_229.htm#sec_12-703
          https://portal.ct.gov/-/media/DRS/Forms/2022/Income/2022-CT-1040-Instructions_1222.pdf
          "https://portal.ct.gov/DRS/Publications/TSSNs/TSSN-29#:~:text=An%20individuals%20net%20capital%20gains
         portion%20of%20Social%20Security%20benefits"
          "https://portal.ct.gov/DRS/Publications/TSSNs/TSSN-29#:~:text=An%20individuals%20net%20capital%20gains
         portion%20of%20Social%20Security%20benefits"
          https://portal.ct.gov/DRS/Individuals/Resident-Income-Tax/Tax-Information
        
        Delaware:
        -         
        - https://revenue.delaware.gov/software-developer/tax-rate-changes/
          https://revenuefiles.delaware.gov/2022/PIT-RES_TY22_2022-01_PaperInteractive.pdf
          https://revenuefiles.delaware.gov/2022/PIT-RES_TY22_2022-02_Instructions.pdf
          https://revenue.delaware.gov/frequently-asked-questions/personal-income-tax-faqs/
          https://www.newcastlede.gov/DocumentCenter/View/20147/City-of-Wilmington-Wage-Tax-Information-PDF
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        federal_agi = self.federal_agi if hasattr(self, 'federal_agi') else federal_agi
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Connecticut', 'state_occupation': 'Delaware', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "federal_agi": federal_agi, "filing_status": filing_status, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def connecticut_and_florida(self, capital_gains_long=0, state_agi=75000, filing_status='single', income=100000, sources=False, capital_gains_short=0, credits_state=0):
        """
        --- Parameters ----
        
        Connecticut:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        
        Florida:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Connecticut:
        - https://portal.ct.gov/-/media/DRS/Forms/2022/Income/2022-CT-1040-Instructions_1222.pdf
          https://www.cga.ct.gov/2022/rpt/pdf/2022-R-0108.pdf
          https://www.cga.ct.gov/current/pub/chap_229.htm#sec_12-703
          https://portal.ct.gov/-/media/DRS/Forms/2022/Income/2022-CT-1040-Instructions_1222.pdf
          "https://portal.ct.gov/DRS/Publications/TSSNs/TSSN-29#:~:text=An%20individuals%20net%20capital%20gains
         portion%20of%20Social%20Security%20benefits"
          "https://portal.ct.gov/DRS/Publications/TSSNs/TSSN-29#:~:text=An%20individuals%20net%20capital%20gains
         portion%20of%20Social%20Security%20benefits"
          https://portal.ct.gov/DRS/Individuals/Resident-Income-Tax/Tax-Information
        
        Florida:
        -         
        - https://floridarevenue.com/taxes/taxesfees/Pages/tax_interest_rates.aspx
        """
        
        # Check if the parameters are set upstream in the instance
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        sources = self.sources if hasattr(self, 'sources') else sources
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Connecticut', 'state_occupation': 'Florida', "capital_gains_long": capital_gains_long, "state_agi": state_agi, "filing_status": filing_status, "income": income, "sources": sources, "capital_gains_short": capital_gains_short, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def connecticut_and_georgia(self, capital_gains_short=0, income=100000, dependents=0, sources=False, state_agi=75000, filers_over_65=0, capital_gains_long=0, filing_status='single', credits_state=0):
        """
        --- Parameters ----
        
        Connecticut:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        
        Georgia:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Connecticut:
        - https://portal.ct.gov/-/media/DRS/Forms/2022/Income/2022-CT-1040-Instructions_1222.pdf
          https://www.cga.ct.gov/2022/rpt/pdf/2022-R-0108.pdf
          https://www.cga.ct.gov/current/pub/chap_229.htm#sec_12-703
          https://portal.ct.gov/-/media/DRS/Forms/2022/Income/2022-CT-1040-Instructions_1222.pdf
          "https://portal.ct.gov/DRS/Publications/TSSNs/TSSN-29#:~:text=An%20individuals%20net%20capital%20gains
         portion%20of%20Social%20Security%20benefits"
          "https://portal.ct.gov/DRS/Publications/TSSNs/TSSN-29#:~:text=An%20individuals%20net%20capital%20gains
         portion%20of%20Social%20Security%20benefits"
          https://portal.ct.gov/DRS/Individuals/Resident-Income-Tax/Tax-Information
        
        Georgia:
        -         
        - https://dor.georgia.gov/tax-tables-georgia-tax-rate-schedule
          https://apps.dor.ga.gov/FillableForms/PDFViewer/Index?form=2022GA500
          https://apps.dor.ga.gov/FillableForms/PDFViewer/Index?form=2022GA500
          https://apps.dor.ga.gov/FillableForms/PDFViewer/Index?form=2022GA500
          https://dor.georgia.gov/years-individual-income-tax-forms
          https://dor.georgia.gov/years-individual-income-tax-forms
        """
        
        # Check if the parameters are set upstream in the instance
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Connecticut', 'state_occupation': 'Georgia', "capital_gains_short": capital_gains_short, "income": income, "dependents": dependents, "sources": sources, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def connecticut_and_hawaii(self, capital_gains_short=0, income=100000, dependents=0, sources=False, state_agi=75000, filers_over_65=0, capital_gains_long=0, filing_status='single', credits_state=0):
        """
        --- Parameters ----
        
        Connecticut:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        
        Hawaii:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Connecticut:
        - https://portal.ct.gov/-/media/DRS/Forms/2022/Income/2022-CT-1040-Instructions_1222.pdf
          https://www.cga.ct.gov/2022/rpt/pdf/2022-R-0108.pdf
          https://www.cga.ct.gov/current/pub/chap_229.htm#sec_12-703
          https://portal.ct.gov/-/media/DRS/Forms/2022/Income/2022-CT-1040-Instructions_1222.pdf
          "https://portal.ct.gov/DRS/Publications/TSSNs/TSSN-29#:~:text=An%20individuals%20net%20capital%20gains
         portion%20of%20Social%20Security%20benefits"
          "https://portal.ct.gov/DRS/Publications/TSSNs/TSSN-29#:~:text=An%20individuals%20net%20capital%20gains
         portion%20of%20Social%20Security%20benefits"
          https://portal.ct.gov/DRS/Individuals/Resident-Income-Tax/Tax-Information
        
        Hawaii:
        -         
        - https://tax.hawaii.gov/forms/d_18table-on/d_18table-on_p13/
          https://files.hawaii.gov/tax/forms/2018/18table-on.pdf
          https://files.hawaii.gov/tax/news/pubs/22outline.pdf
          https://files.hawaii.gov/tax/news/pubs/FastTaxRefGuide_2017.pdf
          https://www.tfhawaii.org/wordpress/state-tax-resources/mini-tax-guides/individual-income-tax-chapter-235/#:~:text=Long%20term%20capital%20gains%20are%20taxed%20at%20a%20maximum%20of%207.25%25
          https://www.grassrootinstitute.org/2023/02/hb337-increasing-tax-on-capital-gains-a-bad-risk/
          https://files.hawaii.gov/tax/news/pubs/22outline.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Connecticut', 'state_occupation': 'Hawaii', "capital_gains_short": capital_gains_short, "income": income, "dependents": dependents, "sources": sources, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def connecticut_and_idaho(self, capital_gains_short=0, income=100000, dependents=0, sources=False, state_agi=75000, filers_over_65=0, capital_gains_long=0, filing_status='single', credits_state=0):
        """
        --- Parameters ----
        
        Connecticut:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        
        Idaho:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Connecticut:
        - https://portal.ct.gov/-/media/DRS/Forms/2022/Income/2022-CT-1040-Instructions_1222.pdf
          https://www.cga.ct.gov/2022/rpt/pdf/2022-R-0108.pdf
          https://www.cga.ct.gov/current/pub/chap_229.htm#sec_12-703
          https://portal.ct.gov/-/media/DRS/Forms/2022/Income/2022-CT-1040-Instructions_1222.pdf
          "https://portal.ct.gov/DRS/Publications/TSSNs/TSSN-29#:~:text=An%20individuals%20net%20capital%20gains
         portion%20of%20Social%20Security%20benefits"
          "https://portal.ct.gov/DRS/Publications/TSSNs/TSSN-29#:~:text=An%20individuals%20net%20capital%20gains
         portion%20of%20Social%20Security%20benefits"
          https://portal.ct.gov/DRS/Individuals/Resident-Income-Tax/Tax-Information
        
        Idaho:
        -         
        - https://tax.idaho.gov/taxes/income-tax/individual-income/individual-income-tax-rate-schedule/
          https://tax.idaho.gov/wp-content/uploads/forms/EIN00046/EIN00046_03-01-2023.pdf
          https://tax.idaho.gov/taxes/income-tax/individual-income/idaho-source-income/
          https://tax.idaho.gov/taxes/income-tax/individual-income/idaho-source-income/
        """
        
        # Check if the parameters are set upstream in the instance
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Connecticut', 'state_occupation': 'Idaho', "capital_gains_short": capital_gains_short, "income": income, "dependents": dependents, "sources": sources, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def connecticut_and_illinois(self, state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Connecticut:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        
        Illinois:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Connecticut:
        - https://portal.ct.gov/-/media/DRS/Forms/2022/Income/2022-CT-1040-Instructions_1222.pdf
          https://www.cga.ct.gov/2022/rpt/pdf/2022-R-0108.pdf
          https://www.cga.ct.gov/current/pub/chap_229.htm#sec_12-703
          https://portal.ct.gov/-/media/DRS/Forms/2022/Income/2022-CT-1040-Instructions_1222.pdf
          "https://portal.ct.gov/DRS/Publications/TSSNs/TSSN-29#:~:text=An%20individuals%20net%20capital%20gains
         portion%20of%20Social%20Security%20benefits"
          "https://portal.ct.gov/DRS/Publications/TSSNs/TSSN-29#:~:text=An%20individuals%20net%20capital%20gains
         portion%20of%20Social%20Security%20benefits"
          https://portal.ct.gov/DRS/Individuals/Resident-Income-Tax/Tax-Information
        
        Illinois:
        -         
        - https://tax.illinois.gov/research/taxrates/income.html#IndividualIncome
          https://tax.illinois.gov/content/dam/soi/en/web/tax/forms/incometax/documents/currentyear/individual/il-1040-instr.pdf
          https://tax.illinois.gov/questionsandanswers/12.html
          https://tax.illinois.gov/content/dam/soi/en/web/tax/forms/incometax/documents/currentyear/individual/il-1040-instr.pdf
          https://tax.illinois.gov/research/taxrates/personalproperty.html
          https://tax.illinois.gov/localgovernments/income.html
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Connecticut', 'state_occupation': 'Illinois', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def connecticut_and_indiana(self, capital_gains_short=0, income=100000, dependents=0, sources=False, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_long=0, filing_status='single', credits_state=0):
        """
        --- Parameters ----
        
        Connecticut:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        
        Indiana:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Connecticut:
        - https://portal.ct.gov/-/media/DRS/Forms/2022/Income/2022-CT-1040-Instructions_1222.pdf
          https://www.cga.ct.gov/2022/rpt/pdf/2022-R-0108.pdf
          https://www.cga.ct.gov/current/pub/chap_229.htm#sec_12-703
          https://portal.ct.gov/-/media/DRS/Forms/2022/Income/2022-CT-1040-Instructions_1222.pdf
          "https://portal.ct.gov/DRS/Publications/TSSNs/TSSN-29#:~:text=An%20individuals%20net%20capital%20gains
         portion%20of%20Social%20Security%20benefits"
          "https://portal.ct.gov/DRS/Publications/TSSNs/TSSN-29#:~:text=An%20individuals%20net%20capital%20gains
         portion%20of%20Social%20Security%20benefits"
          https://portal.ct.gov/DRS/Individuals/Resident-Income-Tax/Tax-Information
        
        Indiana:
        -         
        - https://www.in.gov/dor/tax-forms/2022-individual-income-tax-forms/
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-32.htm
          https://www.in.gov/dor/files/ib28.pdf
          https://www.in.gov/dor/tax-forms/2022-individual-income-tax-forms/
          https://www.in.gov/dor/files/2022-county-tax-rates-and-codes.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Connecticut', 'state_occupation': 'Indiana', "capital_gains_short": capital_gains_short, "income": income, "dependents": dependents, "sources": sources, "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def connecticut_and_iowa(self, capital_gains_short=0, income=100000, dependents=0, sources=False, county_residence='Other', state_agi=75000, capital_gains_long=0, filing_status='single', credits_state=0):
        """
        --- Parameters ----
        
        Connecticut:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        
        Iowa:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Connecticut:
        - https://portal.ct.gov/-/media/DRS/Forms/2022/Income/2022-CT-1040-Instructions_1222.pdf
          https://www.cga.ct.gov/2022/rpt/pdf/2022-R-0108.pdf
          https://www.cga.ct.gov/current/pub/chap_229.htm#sec_12-703
          https://portal.ct.gov/-/media/DRS/Forms/2022/Income/2022-CT-1040-Instructions_1222.pdf
          "https://portal.ct.gov/DRS/Publications/TSSNs/TSSN-29#:~:text=An%20individuals%20net%20capital%20gains
         portion%20of%20Social%20Security%20benefits"
          "https://portal.ct.gov/DRS/Publications/TSSNs/TSSN-29#:~:text=An%20individuals%20net%20capital%20gains
         portion%20of%20Social%20Security%20benefits"
          https://portal.ct.gov/DRS/Individuals/Resident-Income-Tax/Tax-Information
        
        Iowa:
        -         
        - https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
          https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
          https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
          https://tax.iowa.gov/iowa-illinois-reciprocal-agreement
          https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
        """
        
        # Check if the parameters are set upstream in the instance
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Connecticut', 'state_occupation': 'Iowa', "capital_gains_short": capital_gains_short, "income": income, "dependents": dependents, "sources": sources, "county_residence": county_residence, "state_agi": state_agi, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def connecticut_and_kansas(self, capital_gains_short=0, income=100000, dependents=0, sources=False, state_agi=75000, filers_over_65=0, capital_gains_long=0, filing_status='single', credits_state=0):
        """
        --- Parameters ----
        
        Connecticut:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        
        Kansas:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Connecticut:
        - https://portal.ct.gov/-/media/DRS/Forms/2022/Income/2022-CT-1040-Instructions_1222.pdf
          https://www.cga.ct.gov/2022/rpt/pdf/2022-R-0108.pdf
          https://www.cga.ct.gov/current/pub/chap_229.htm#sec_12-703
          https://portal.ct.gov/-/media/DRS/Forms/2022/Income/2022-CT-1040-Instructions_1222.pdf
          "https://portal.ct.gov/DRS/Publications/TSSNs/TSSN-29#:~:text=An%20individuals%20net%20capital%20gains
         portion%20of%20Social%20Security%20benefits"
          "https://portal.ct.gov/DRS/Publications/TSSNs/TSSN-29#:~:text=An%20individuals%20net%20capital%20gains
         portion%20of%20Social%20Security%20benefits"
          https://portal.ct.gov/DRS/Individuals/Resident-Income-Tax/Tax-Information
        
        Kansas:
        -         
        - https://www.ksrevenue.gov/taxrates.html
          https://www.ksrevenue.gov/incomebook22.html#0
          https://www.ksrevenue.gov/incomebook22.html#0
          https://www.ksrevenue.gov/incomebook22.html#0
          https://www.ksrevenue.gov/incomebook22.html
          https://www.ksrevenue.gov/incomebook22.html
        """
        
        # Check if the parameters are set upstream in the instance
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Connecticut', 'state_occupation': 'Kansas', "capital_gains_short": capital_gains_short, "income": income, "dependents": dependents, "sources": sources, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def connecticut_and_kentucky(self, county_residence='Other', state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Connecticut:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        
        Kentucky:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Connecticut:
        - https://portal.ct.gov/-/media/DRS/Forms/2022/Income/2022-CT-1040-Instructions_1222.pdf
          https://www.cga.ct.gov/2022/rpt/pdf/2022-R-0108.pdf
          https://www.cga.ct.gov/current/pub/chap_229.htm#sec_12-703
          https://portal.ct.gov/-/media/DRS/Forms/2022/Income/2022-CT-1040-Instructions_1222.pdf
          "https://portal.ct.gov/DRS/Publications/TSSNs/TSSN-29#:~:text=An%20individuals%20net%20capital%20gains
         portion%20of%20Social%20Security%20benefits"
          "https://portal.ct.gov/DRS/Publications/TSSNs/TSSN-29#:~:text=An%20individuals%20net%20capital%20gains
         portion%20of%20Social%20Security%20benefits"
          https://portal.ct.gov/DRS/Individuals/Resident-Income-Tax/Tax-Information
        
        Kentucky:
        -         
        - https://revenue.ky.gov/News/Pages/Dor-Announces-Updates-to-Individual-Income-Tax-for-2023-Tax-Year.aspx
          https://revenue.ky.gov/Individual/Individual-Income-Tax/Pages/default.aspx
          https://revenue.ky.gov/News/Pages/Dor-Announces-Updates-to-Individual-Income-Tax-for-2023-Tax-Year.aspx
          https://revenue.ky.gov/Individual/Individual-Income-Tax/Pages/default.aspx
          https://revenue.ky.gov/Dor%20Training%20Materials/103%20KAR%2017.140.%20Individual%20income%20tax%20-%20reciprocity%20-%20nonresidents.pdf
          https://revenue.ky.gov/Individual/Individual-Income-Tax/Pages/default.aspx
          https://nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2021/Taxes-21-32.htm?taxmap=true
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Connecticut', 'state_occupation': 'Kentucky', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def connecticut_and_louisiana(self, capital_gains_short=0, income=100000, dependents=0, sources=False, state_agi=75000, filers_over_65=0, capital_gains_long=0, filing_status='single', credits_state=0):
        """
        --- Parameters ----
        
        Connecticut:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        
        Louisiana:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Connecticut:
        - https://portal.ct.gov/-/media/DRS/Forms/2022/Income/2022-CT-1040-Instructions_1222.pdf
          https://www.cga.ct.gov/2022/rpt/pdf/2022-R-0108.pdf
          https://www.cga.ct.gov/current/pub/chap_229.htm#sec_12-703
          https://portal.ct.gov/-/media/DRS/Forms/2022/Income/2022-CT-1040-Instructions_1222.pdf
          "https://portal.ct.gov/DRS/Publications/TSSNs/TSSN-29#:~:text=An%20individuals%20net%20capital%20gains
         portion%20of%20Social%20Security%20benefits"
          "https://portal.ct.gov/DRS/Publications/TSSNs/TSSN-29#:~:text=An%20individuals%20net%20capital%20gains
         portion%20of%20Social%20Security%20benefits"
          https://portal.ct.gov/DRS/Individuals/Resident-Income-Tax/Tax-Information
        
        Louisiana:
        -         
        - https://revenue.louisiana.gov/individualincometax
          https://revenue.louisiana.gov/individualincometax
          https://www.revenue.louisiana.gov/taxforms/6935(11_02)F.pdf
          https://revenue.louisiana.gov/individualincometax
          https://revenue.louisiana.gov/individualincometax
        """
        
        # Check if the parameters are set upstream in the instance
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Connecticut', 'state_occupation': 'Louisiana', "capital_gains_short": capital_gains_short, "income": income, "dependents": dependents, "sources": sources, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def connecticut_and_maine(self, capital_gains_short=0, income=100000, dependents=0, sources=False, state_agi=75000, filers_over_65=0, capital_gains_long=0, filing_status='single', credits_state=0):
        """
        --- Parameters ----
        
        Connecticut:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        
        Maine:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Connecticut:
        - https://portal.ct.gov/-/media/DRS/Forms/2022/Income/2022-CT-1040-Instructions_1222.pdf
          https://www.cga.ct.gov/2022/rpt/pdf/2022-R-0108.pdf
          https://www.cga.ct.gov/current/pub/chap_229.htm#sec_12-703
          https://portal.ct.gov/-/media/DRS/Forms/2022/Income/2022-CT-1040-Instructions_1222.pdf
          "https://portal.ct.gov/DRS/Publications/TSSNs/TSSN-29#:~:text=An%20individuals%20net%20capital%20gains
         portion%20of%20Social%20Security%20benefits"
          "https://portal.ct.gov/DRS/Publications/TSSNs/TSSN-29#:~:text=An%20individuals%20net%20capital%20gains
         portion%20of%20Social%20Security%20benefits"
          https://portal.ct.gov/DRS/Individuals/Resident-Income-Tax/Tax-Information
        
        Maine:
        -         
        - https://www.maine.gov/revenue/tax-return-forms/individual-income-tax-2022
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/ind_tax_rate_sched_2022.pdf
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/22_1040me_book_gen_instruc_revisedFeb23.pdf
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/22_1040me_book_gen_instruc_revisedFeb23.pdf
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/22_1040me_book_gen_instruc_revisedFeb23.pdf
          https://www.maine.gov/revenue/tax-return-forms/individual-income-tax-2022
        """
        
        # Check if the parameters are set upstream in the instance
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Connecticut', 'state_occupation': 'Maine', "capital_gains_short": capital_gains_short, "income": income, "dependents": dependents, "sources": sources, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def connecticut_and_maryland(self, capital_gains_short=0, income=100000, dependents=0, sources=False, county_residence='Other', state_agi=75000, capital_gains_long=0, filing_status='single', credits_state=0):
        """
        --- Parameters ----
        
        Connecticut:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        
        Maryland:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Connecticut:
        - https://portal.ct.gov/-/media/DRS/Forms/2022/Income/2022-CT-1040-Instructions_1222.pdf
          https://www.cga.ct.gov/2022/rpt/pdf/2022-R-0108.pdf
          https://www.cga.ct.gov/current/pub/chap_229.htm#sec_12-703
          https://portal.ct.gov/-/media/DRS/Forms/2022/Income/2022-CT-1040-Instructions_1222.pdf
          "https://portal.ct.gov/DRS/Publications/TSSNs/TSSN-29#:~:text=An%20individuals%20net%20capital%20gains
         portion%20of%20Social%20Security%20benefits"
          "https://portal.ct.gov/DRS/Publications/TSSNs/TSSN-29#:~:text=An%20individuals%20net%20capital%20gains
         portion%20of%20Social%20Security%20benefits"
          https://portal.ct.gov/DRS/Individuals/Resident-Income-Tax/Tax-Information
        
        Maryland:
        -         
        - https://www.marylandtaxes.gov/individual/income/tax-info/tax-rates.php
          https://www.marylandtaxes.gov/forms/22_forms/Resident_Booklet.pdf
          https://www.marylandtaxes.gov/forms/Tax_Publications/Administrative_Releases/Income_and_Estate_Tax_Releases/ar_it37.pdf
          https://www.marylandtaxes.gov/individual/index.php
          https://www.marylandtaxes.gov/individual/income/tax-info/tax-rates.php
        """
        
        # Check if the parameters are set upstream in the instance
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Connecticut', 'state_occupation': 'Maryland', "capital_gains_short": capital_gains_short, "income": income, "dependents": dependents, "sources": sources, "county_residence": county_residence, "state_agi": state_agi, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def connecticut_and_massachusetts(self, state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Connecticut:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        
        Massachusetts:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Connecticut:
        - https://portal.ct.gov/-/media/DRS/Forms/2022/Income/2022-CT-1040-Instructions_1222.pdf
          https://www.cga.ct.gov/2022/rpt/pdf/2022-R-0108.pdf
          https://www.cga.ct.gov/current/pub/chap_229.htm#sec_12-703
          https://portal.ct.gov/-/media/DRS/Forms/2022/Income/2022-CT-1040-Instructions_1222.pdf
          "https://portal.ct.gov/DRS/Publications/TSSNs/TSSN-29#:~:text=An%20individuals%20net%20capital%20gains
         portion%20of%20Social%20Security%20benefits"
          "https://portal.ct.gov/DRS/Publications/TSSNs/TSSN-29#:~:text=An%20individuals%20net%20capital%20gains
         portion%20of%20Social%20Security%20benefits"
          https://portal.ct.gov/DRS/Individuals/Resident-Income-Tax/Tax-Information
        
        Massachusetts:
        -         
        - https://www.mass.gov/service-details/massachusetts-tax-rates
          https://www.mass.gov/service-details/view-massachusetts-personal-income-tax-exemptions
          https://www.mass.gov/service-details/view-massachusetts-personal-income-tax-exemptions
          https://www.mass.gov/service-details/view-child-and-dependent-related-credits
          https://www.mass.gov/service-details/massachusetts-tax-rates
          https://www.mass.gov/personal-income-tax
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Connecticut', 'state_occupation': 'Massachusetts', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def connecticut_and_michigan(self, capital_gains_short=0, income=100000, county_occupation='Other', dependents=0, sources=False, county_residence='Other', state_agi=75000, capital_gains_long=0, filing_status='single', credits_state=0):
        """
        --- Parameters ----
        
        Connecticut:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        
        Michigan:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        county_occupation : str, optional
            County of occupation. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Connecticut:
        - https://portal.ct.gov/-/media/DRS/Forms/2022/Income/2022-CT-1040-Instructions_1222.pdf
          https://www.cga.ct.gov/2022/rpt/pdf/2022-R-0108.pdf
          https://www.cga.ct.gov/current/pub/chap_229.htm#sec_12-703
          https://portal.ct.gov/-/media/DRS/Forms/2022/Income/2022-CT-1040-Instructions_1222.pdf
          "https://portal.ct.gov/DRS/Publications/TSSNs/TSSN-29#:~:text=An%20individuals%20net%20capital%20gains
         portion%20of%20Social%20Security%20benefits"
          "https://portal.ct.gov/DRS/Publications/TSSNs/TSSN-29#:~:text=An%20individuals%20net%20capital%20gains
         portion%20of%20Social%20Security%20benefits"
          https://portal.ct.gov/DRS/Individuals/Resident-Income-Tax/Tax-Information
        
        Michigan:
        -         
        - https://www.michigan.gov/taxes/iit/new-developments-for-tax-year-2022
          "https://www.michigan.gov/taxes/-/media/Project/Websites/taxes/Forms/2023/2023-SUW/446_Withholding-Guide_2023.pdf?rev=177ca3d8cb034392bf84e317ed78c320&hash=19C5F42F43A89A91863D229349F48883#:~:text=The%20exemption%20amount%20is%20%245%2C400
         employees%20Michigan%20income%20tax%20return"
          "https://www.michigan.gov/taxes/-/media/Project/Websites/taxes/Forms/2023/2023-SUW/446_Withholding-Guide_2023.pdf?rev=177ca3d8cb034392bf84e317ed78c320&hash=19C5F42F43A89A91863D229349F48883#:~:text=The%20exemption%20amount%20is%20%245%2C400
         employees%20Michigan%20income%20tax%20return"
          https://www.michigan.gov/taxes/questions/iit/accordion/residency/are-my-wages-earned-in-another-state-taxable-in-michigan-if-i-am-a-michigan-resident-1
          https://www.michigan.gov/taxes/-/media/Project/Websites/taxes/Forms/2022/2022-IIT-Forms/BOOK_MI-1040.pdf?rev=4524f1cbbc0243b194ab05a6680be75b&hash=40DFE71E651F7C1D65C31F6CD03D944F
          https://www.michigan.gov/taxes/questions/iit/accordion/general/what-cities-impose-an-income-tax
        """
        
        # Check if the parameters are set upstream in the instance
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        county_occupation = self.county_occupation if hasattr(self, 'county_occupation') else county_occupation
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Connecticut', 'state_occupation': 'Michigan', "capital_gains_short": capital_gains_short, "income": income, "county_occupation": county_occupation, "dependents": dependents, "sources": sources, "county_residence": county_residence, "state_agi": state_agi, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def connecticut_and_minnesota(self, state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Connecticut:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        
        Minnesota:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Connecticut:
        - https://portal.ct.gov/-/media/DRS/Forms/2022/Income/2022-CT-1040-Instructions_1222.pdf
          https://www.cga.ct.gov/2022/rpt/pdf/2022-R-0108.pdf
          https://www.cga.ct.gov/current/pub/chap_229.htm#sec_12-703
          https://portal.ct.gov/-/media/DRS/Forms/2022/Income/2022-CT-1040-Instructions_1222.pdf
          "https://portal.ct.gov/DRS/Publications/TSSNs/TSSN-29#:~:text=An%20individuals%20net%20capital%20gains
         portion%20of%20Social%20Security%20benefits"
          "https://portal.ct.gov/DRS/Publications/TSSNs/TSSN-29#:~:text=An%20individuals%20net%20capital%20gains
         portion%20of%20Social%20Security%20benefits"
          https://portal.ct.gov/DRS/Individuals/Resident-Income-Tax/Tax-Information
        
        Minnesota:
        -         
        - https://www.revenue.state.mn.us/minnesota-income-tax-rates-and-brackets
          https://www.revenue.state.mn.us/child-and-dependent-care-credit
          https://www.revenue.state.mn.us/dependent-exemptions
          https://www.revenue.state.mn.us/minnesota-standard-deduction
          https://mn.gov/mmb/assets/mwr_form_tcm1059-128581.pdf
          https://www.revenue.state.mn.us/individual-income-tax
          https://www.revenue.state.mn.us/individual-income-tax
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Connecticut', 'state_occupation': 'Minnesota', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def connecticut_and_mississippi(self, state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Connecticut:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        
        Mississippi:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Connecticut:
        - https://portal.ct.gov/-/media/DRS/Forms/2022/Income/2022-CT-1040-Instructions_1222.pdf
          https://www.cga.ct.gov/2022/rpt/pdf/2022-R-0108.pdf
          https://www.cga.ct.gov/current/pub/chap_229.htm#sec_12-703
          https://portal.ct.gov/-/media/DRS/Forms/2022/Income/2022-CT-1040-Instructions_1222.pdf
          "https://portal.ct.gov/DRS/Publications/TSSNs/TSSN-29#:~:text=An%20individuals%20net%20capital%20gains
         portion%20of%20Social%20Security%20benefits"
          "https://portal.ct.gov/DRS/Publications/TSSNs/TSSN-29#:~:text=An%20individuals%20net%20capital%20gains
         portion%20of%20Social%20Security%20benefits"
          https://portal.ct.gov/DRS/Individuals/Resident-Income-Tax/Tax-Information
        
        Mississippi:
        -         
        - https://www.dor.ms.gov/individual/tax-rates
          https://www.dor.ms.gov/individual/tax-rates
          https://www.dor.ms.gov/individual/tax-rates
          https://www.dor.ms.gov/individual
          https://www.dor.ms.gov/individual
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Connecticut', 'state_occupation': 'Mississippi', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def connecticut_and_missouri(self, capital_gains_short=0, income=100000, dependents=0, sources=False, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_long=0, filing_status='single', credits_state=0):
        """
        --- Parameters ----
        
        Connecticut:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        
        Missouri:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Connecticut:
        - https://portal.ct.gov/-/media/DRS/Forms/2022/Income/2022-CT-1040-Instructions_1222.pdf
          https://www.cga.ct.gov/2022/rpt/pdf/2022-R-0108.pdf
          https://www.cga.ct.gov/current/pub/chap_229.htm#sec_12-703
          https://portal.ct.gov/-/media/DRS/Forms/2022/Income/2022-CT-1040-Instructions_1222.pdf
          "https://portal.ct.gov/DRS/Publications/TSSNs/TSSN-29#:~:text=An%20individuals%20net%20capital%20gains
         portion%20of%20Social%20Security%20benefits"
          "https://portal.ct.gov/DRS/Publications/TSSNs/TSSN-29#:~:text=An%20individuals%20net%20capital%20gains
         portion%20of%20Social%20Security%20benefits"
          https://portal.ct.gov/DRS/Individuals/Resident-Income-Tax/Tax-Information
        
        Missouri:
        -         
        - https://dor.mo.gov/taxation/individual/tax-types/income/year-changes/
          https://dor.mo.gov/taxation/individual/tax-types/income/year-changes/
          https://dor.mo.gov/taxation/individual/tax-types/income/year-changes/
          https://www.kcmo.gov/city-hall/departments/finance/earnings-tax
          https://www.stlouis-mo.gov/government/departments/collector/earnings-tax/individual-earnings-tax-info.cfm
        """
        
        # Check if the parameters are set upstream in the instance
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Connecticut', 'state_occupation': 'Missouri', "capital_gains_short": capital_gains_short, "income": income, "dependents": dependents, "sources": sources, "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def connecticut_and_montana(self, capital_gains_short=0, income=100000, dependents=0, sources=False, state_agi=75000, filers_over_65=0, capital_gains_long=0, filing_status='single', credits_state=0):
        """
        --- Parameters ----
        
        Connecticut:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        
        Montana:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Connecticut:
        - https://portal.ct.gov/-/media/DRS/Forms/2022/Income/2022-CT-1040-Instructions_1222.pdf
          https://www.cga.ct.gov/2022/rpt/pdf/2022-R-0108.pdf
          https://www.cga.ct.gov/current/pub/chap_229.htm#sec_12-703
          https://portal.ct.gov/-/media/DRS/Forms/2022/Income/2022-CT-1040-Instructions_1222.pdf
          "https://portal.ct.gov/DRS/Publications/TSSNs/TSSN-29#:~:text=An%20individuals%20net%20capital%20gains
         portion%20of%20Social%20Security%20benefits"
          "https://portal.ct.gov/DRS/Publications/TSSNs/TSSN-29#:~:text=An%20individuals%20net%20capital%20gains
         portion%20of%20Social%20Security%20benefits"
          https://portal.ct.gov/DRS/Individuals/Resident-Income-Tax/Tax-Information
        
        Montana:
        -         
        - https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/2022/12/montana-employees-withholding-allowance-and-exemption-certificate-form-mw-4.pdf
          https://mtrevenue.gov/2021/11/29/simplification-of-montana-income-taxation/?utm_source=rss&utm_medium=rss&utm_campaign=simplification-of-montana-income-taxation
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Connecticut', 'state_occupation': 'Montana', "capital_gains_short": capital_gains_short, "income": income, "dependents": dependents, "sources": sources, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def connecticut_and_nebraska(self, state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Connecticut:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        
        Nebraska:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Connecticut:
        - https://portal.ct.gov/-/media/DRS/Forms/2022/Income/2022-CT-1040-Instructions_1222.pdf
          https://www.cga.ct.gov/2022/rpt/pdf/2022-R-0108.pdf
          https://www.cga.ct.gov/current/pub/chap_229.htm#sec_12-703
          https://portal.ct.gov/-/media/DRS/Forms/2022/Income/2022-CT-1040-Instructions_1222.pdf
          "https://portal.ct.gov/DRS/Publications/TSSNs/TSSN-29#:~:text=An%20individuals%20net%20capital%20gains
         portion%20of%20Social%20Security%20benefits"
          "https://portal.ct.gov/DRS/Publications/TSSNs/TSSN-29#:~:text=An%20individuals%20net%20capital%20gains
         portion%20of%20Social%20Security%20benefits"
          https://portal.ct.gov/DRS/Individuals/Resident-Income-Tax/Tax-Information
        
        Nebraska:
        -         
        - https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2023_Circular_EN_8-429r2023_Whole_Book_11.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Connecticut', 'state_occupation': 'Nebraska', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def connecticut_and_nevada(self, capital_gains_long=0, state_agi=75000, filing_status='single', income=100000, sources=False, capital_gains_short=0, credits_state=0):
        """
        --- Parameters ----
        
        Connecticut:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        
        Nevada:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Connecticut:
        - https://portal.ct.gov/-/media/DRS/Forms/2022/Income/2022-CT-1040-Instructions_1222.pdf
          https://www.cga.ct.gov/2022/rpt/pdf/2022-R-0108.pdf
          https://www.cga.ct.gov/current/pub/chap_229.htm#sec_12-703
          https://portal.ct.gov/-/media/DRS/Forms/2022/Income/2022-CT-1040-Instructions_1222.pdf
          "https://portal.ct.gov/DRS/Publications/TSSNs/TSSN-29#:~:text=An%20individuals%20net%20capital%20gains
         portion%20of%20Social%20Security%20benefits"
          "https://portal.ct.gov/DRS/Publications/TSSNs/TSSN-29#:~:text=An%20individuals%20net%20capital%20gains
         portion%20of%20Social%20Security%20benefits"
          https://portal.ct.gov/DRS/Individuals/Resident-Income-Tax/Tax-Information
        
        Nevada:
        -         
        - https://www.nvsos.gov/sos/businesses/the-nevada-advantage
        """
        
        # Check if the parameters are set upstream in the instance
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        sources = self.sources if hasattr(self, 'sources') else sources
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Connecticut', 'state_occupation': 'Nevada', "capital_gains_long": capital_gains_long, "state_agi": state_agi, "filing_status": filing_status, "income": income, "sources": sources, "capital_gains_short": capital_gains_short, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def connecticut_and_new_hampshire(self, capital_gains_long=0, state_agi=75000, filing_status='single', income=100000, sources=False, capital_gains_short=0, credits_state=0):
        """
        --- Parameters ----
        
        Connecticut:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        
        New Hampshire:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Connecticut:
        - https://portal.ct.gov/-/media/DRS/Forms/2022/Income/2022-CT-1040-Instructions_1222.pdf
          https://www.cga.ct.gov/2022/rpt/pdf/2022-R-0108.pdf
          https://www.cga.ct.gov/current/pub/chap_229.htm#sec_12-703
          https://portal.ct.gov/-/media/DRS/Forms/2022/Income/2022-CT-1040-Instructions_1222.pdf
          "https://portal.ct.gov/DRS/Publications/TSSNs/TSSN-29#:~:text=An%20individuals%20net%20capital%20gains
         portion%20of%20Social%20Security%20benefits"
          "https://portal.ct.gov/DRS/Publications/TSSNs/TSSN-29#:~:text=An%20individuals%20net%20capital%20gains
         portion%20of%20Social%20Security%20benefits"
          https://portal.ct.gov/DRS/Individuals/Resident-Income-Tax/Tax-Information
        
        New Hampshire:
        -         
        - https://www.revenue.nh.gov/assistance/tax-overview.htm
        """
        
        # Check if the parameters are set upstream in the instance
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        sources = self.sources if hasattr(self, 'sources') else sources
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Connecticut', 'state_occupation': 'NewHampshire', "capital_gains_long": capital_gains_long, "state_agi": state_agi, "filing_status": filing_status, "income": income, "sources": sources, "capital_gains_short": capital_gains_short, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def connecticut_and_new_jersey(self, capital_gains_short=0, income=100000, dependents=0, sources=False, state_agi=75000, filers_over_65=0, capital_gains_long=0, filing_status='single', credits_state=0):
        """
        --- Parameters ----
        
        Connecticut:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        
        New Jersey:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Connecticut:
        - https://portal.ct.gov/-/media/DRS/Forms/2022/Income/2022-CT-1040-Instructions_1222.pdf
          https://www.cga.ct.gov/2022/rpt/pdf/2022-R-0108.pdf
          https://www.cga.ct.gov/current/pub/chap_229.htm#sec_12-703
          https://portal.ct.gov/-/media/DRS/Forms/2022/Income/2022-CT-1040-Instructions_1222.pdf
          "https://portal.ct.gov/DRS/Publications/TSSNs/TSSN-29#:~:text=An%20individuals%20net%20capital%20gains
         portion%20of%20Social%20Security%20benefits"
          "https://portal.ct.gov/DRS/Publications/TSSNs/TSSN-29#:~:text=An%20individuals%20net%20capital%20gains
         portion%20of%20Social%20Security%20benefits"
          https://portal.ct.gov/DRS/Individuals/Resident-Income-Tax/Tax-Information
        
        New Jersey:
        -         
        - https://www.state.nj.us/treasury/taxation/taxtables.shtml
          https://www.state.nj.us/treasury/taxation/njit13.shtml
          https://www.nj.gov/treasury/taxation/njit25.shtml
          https://www.state.nj.us/treasury/taxation/individuals/freqqiti.shtml
          https://www.state.nj.us/treasury/taxation/localtax.shtml
        """
        
        # Check if the parameters are set upstream in the instance
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Connecticut', 'state_occupation': 'NewJersey', "capital_gains_short": capital_gains_short, "income": income, "dependents": dependents, "sources": sources, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def connecticut_and_new_mexico(self, state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Connecticut:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        
        New Mexico:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Connecticut:
        - https://portal.ct.gov/-/media/DRS/Forms/2022/Income/2022-CT-1040-Instructions_1222.pdf
          https://www.cga.ct.gov/2022/rpt/pdf/2022-R-0108.pdf
          https://www.cga.ct.gov/current/pub/chap_229.htm#sec_12-703
          https://portal.ct.gov/-/media/DRS/Forms/2022/Income/2022-CT-1040-Instructions_1222.pdf
          "https://portal.ct.gov/DRS/Publications/TSSNs/TSSN-29#:~:text=An%20individuals%20net%20capital%20gains
         portion%20of%20Social%20Security%20benefits"
          "https://portal.ct.gov/DRS/Publications/TSSNs/TSSN-29#:~:text=An%20individuals%20net%20capital%20gains
         portion%20of%20Social%20Security%20benefits"
          https://portal.ct.gov/DRS/Individuals/Resident-Income-Tax/Tax-Information
        
        New Mexico:
        -         
        - https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/fdf3c548-8aba-4b9c-9eb4-bb564c716015/FYI-104.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/0bf943fd-652e-4400-bb1b-cd8397eb5a95/2022pit-adj.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/2f1a6781-9534-4436-b427-1557f9592099/2022pit-adj-ins.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/0bf943fd-652e-4400-bb1b-cd8397eb5a95/2022pit-adj.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/1afc56af-ea90-4d48-82e5-1f9aeb43255a/PITbook2022.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/fdf3c548-8aba-4b9c-9eb4-bb564c716015/FYI-104.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Connecticut', 'state_occupation': 'NewMexico', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def connecticut_and_new_york(self, capital_gains_short=0, income=100000, dependents=0, sources=False, county_residence='Other', state_agi=75000, capital_gains_long=0, filing_status='single', credits_state=0):
        """
        --- Parameters ----
        
        Connecticut:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        
        New York:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Connecticut:
        - https://portal.ct.gov/-/media/DRS/Forms/2022/Income/2022-CT-1040-Instructions_1222.pdf
          https://www.cga.ct.gov/2022/rpt/pdf/2022-R-0108.pdf
          https://www.cga.ct.gov/current/pub/chap_229.htm#sec_12-703
          https://portal.ct.gov/-/media/DRS/Forms/2022/Income/2022-CT-1040-Instructions_1222.pdf
          "https://portal.ct.gov/DRS/Publications/TSSNs/TSSN-29#:~:text=An%20individuals%20net%20capital%20gains
         portion%20of%20Social%20Security%20benefits"
          "https://portal.ct.gov/DRS/Publications/TSSNs/TSSN-29#:~:text=An%20individuals%20net%20capital%20gains
         portion%20of%20Social%20Security%20benefits"
          https://portal.ct.gov/DRS/Individuals/Resident-Income-Tax/Tax-Information
        
        New York:
        -         
        - https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nys-tax-rate-schedule
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nys-tax-rate-schedule
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nys-tax-rate-schedule
          https://www.tax.ny.gov/pit/
          https://www.tax.ny.gov/pit/file/nonresident-faqs.htm
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#yonkers-tax-rate-schedule
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nyc-tax-rate-schedule
        """
        
        # Check if the parameters are set upstream in the instance
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Connecticut', 'state_occupation': 'NewYork', "capital_gains_short": capital_gains_short, "income": income, "dependents": dependents, "sources": sources, "county_residence": county_residence, "state_agi": state_agi, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def connecticut_and_north_carolina(self, state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Connecticut:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        
        North Carolina:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Connecticut:
        - https://portal.ct.gov/-/media/DRS/Forms/2022/Income/2022-CT-1040-Instructions_1222.pdf
          https://www.cga.ct.gov/2022/rpt/pdf/2022-R-0108.pdf
          https://www.cga.ct.gov/current/pub/chap_229.htm#sec_12-703
          https://portal.ct.gov/-/media/DRS/Forms/2022/Income/2022-CT-1040-Instructions_1222.pdf
          "https://portal.ct.gov/DRS/Publications/TSSNs/TSSN-29#:~:text=An%20individuals%20net%20capital%20gains
         portion%20of%20Social%20Security%20benefits"
          "https://portal.ct.gov/DRS/Publications/TSSNs/TSSN-29#:~:text=An%20individuals%20net%20capital%20gains
         portion%20of%20Social%20Security%20benefits"
          https://portal.ct.gov/DRS/Individuals/Resident-Income-Tax/Tax-Information
        
        North Carolina:
        -         
        - https://www.ncdor.gov/taxes-forms/tax-rate-schedules
          https://www.ncdor.gov/taxes-forms/individual-income-tax/north-carolina-standard-deduction-or-north-carolina-itemized-deductions
          https://www.ncdor.gov/taxes-forms/individual-income-tax
          https://www.ncdor.gov/taxes-forms/individual-income-tax
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Connecticut', 'state_occupation': 'NorthCarolina', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def connecticut_and_north_dakota(self, capital_gains_short=0, income=100000, dependents=0, sources=False, state_agi=75000, filers_over_65=0, capital_gains_long=0, filing_status='single', credits_state=0):
        """
        --- Parameters ----
        
        Connecticut:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        
        North Dakota:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Connecticut:
        - https://portal.ct.gov/-/media/DRS/Forms/2022/Income/2022-CT-1040-Instructions_1222.pdf
          https://www.cga.ct.gov/2022/rpt/pdf/2022-R-0108.pdf
          https://www.cga.ct.gov/current/pub/chap_229.htm#sec_12-703
          https://portal.ct.gov/-/media/DRS/Forms/2022/Income/2022-CT-1040-Instructions_1222.pdf
          "https://portal.ct.gov/DRS/Publications/TSSNs/TSSN-29#:~:text=An%20individuals%20net%20capital%20gains
         portion%20of%20Social%20Security%20benefits"
          "https://portal.ct.gov/DRS/Publications/TSSNs/TSSN-29#:~:text=An%20individuals%20net%20capital%20gains
         portion%20of%20Social%20Security%20benefits"
          https://portal.ct.gov/DRS/Individuals/Resident-Income-Tax/Tax-Information
        
        North Dakota:
        -         
        - https://www.tax.nd.gov/forms
          https://www.tax.nd.gov/sites/www/files/documents/forms/individual/2022-iit/2022-individual-income-tax-booklet.pdf
          https://www.tax.nd.gov/sites/www/files/documents/forms/2022-individual-income-tax-booklet.pdf
          https://www.tax.nd.gov/tax-exemptions-credits/income-tax-exemptions-credits
          https://www.tax.nd.gov/sites/www/files/documents/forms/business/ndwrfillable.pdf
          https://www.tax.nd.gov/news/tax-legislative-changes/significant-changes-law/individual-income-tax-history
          https://www.tax.nd.gov/forms
        """
        
        # Check if the parameters are set upstream in the instance
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Connecticut', 'state_occupation': 'NorthDakota', "capital_gains_short": capital_gains_short, "income": income, "dependents": dependents, "sources": sources, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def connecticut_and_ohio(self, capital_gains_short=0, income=100000, dependents=0, sources=False, county_residence='Other', state_agi=75000, capital_gains_long=0, filing_status='single', credits_state=0):
        """
        --- Parameters ----
        
        Connecticut:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        
        Ohio:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Connecticut:
        - https://portal.ct.gov/-/media/DRS/Forms/2022/Income/2022-CT-1040-Instructions_1222.pdf
          https://www.cga.ct.gov/2022/rpt/pdf/2022-R-0108.pdf
          https://www.cga.ct.gov/current/pub/chap_229.htm#sec_12-703
          https://portal.ct.gov/-/media/DRS/Forms/2022/Income/2022-CT-1040-Instructions_1222.pdf
          "https://portal.ct.gov/DRS/Publications/TSSNs/TSSN-29#:~:text=An%20individuals%20net%20capital%20gains
         portion%20of%20Social%20Security%20benefits"
          "https://portal.ct.gov/DRS/Publications/TSSNs/TSSN-29#:~:text=An%20individuals%20net%20capital%20gains
         portion%20of%20Social%20Security%20benefits"
          https://portal.ct.gov/DRS/Individuals/Resident-Income-Tax/Tax-Information
        
        Ohio:
        -         
        - https://tax.ohio.gov/individual/resources/annual-tax-rates
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://tax.ohio.gov/static/forms/employer_withholding/generic/wth_it4nr.pdf
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://www.ritaohio.com/TaxRatesTable
        """
        
        # Check if the parameters are set upstream in the instance
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Connecticut', 'state_occupation': 'Ohio', "capital_gains_short": capital_gains_short, "income": income, "dependents": dependents, "sources": sources, "county_residence": county_residence, "state_agi": state_agi, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def connecticut_and_oklahoma(self, capital_gains_short=0, income=100000, dependents=0, sources=False, state_agi=75000, filers_over_65=0, capital_gains_long=0, filing_status='single', credits_state=0):
        """
        --- Parameters ----
        
        Connecticut:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        
        Oklahoma:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Connecticut:
        - https://portal.ct.gov/-/media/DRS/Forms/2022/Income/2022-CT-1040-Instructions_1222.pdf
          https://www.cga.ct.gov/2022/rpt/pdf/2022-R-0108.pdf
          https://www.cga.ct.gov/current/pub/chap_229.htm#sec_12-703
          https://portal.ct.gov/-/media/DRS/Forms/2022/Income/2022-CT-1040-Instructions_1222.pdf
          "https://portal.ct.gov/DRS/Publications/TSSNs/TSSN-29#:~:text=An%20individuals%20net%20capital%20gains
         portion%20of%20Social%20Security%20benefits"
          "https://portal.ct.gov/DRS/Publications/TSSNs/TSSN-29#:~:text=An%20individuals%20net%20capital%20gains
         portion%20of%20Social%20Security%20benefits"
          https://portal.ct.gov/DRS/Individuals/Resident-Income-Tax/Tax-Information
        
        Oklahoma:
        -         
        - https://oklahoma.gov/content/dam/ok/en/tax/documents/resources/publications/businesses/withholding-tables/WHTables-2022.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Connecticut', 'state_occupation': 'Oklahoma', "capital_gains_short": capital_gains_short, "income": income, "dependents": dependents, "sources": sources, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def connecticut_and_oregon(self, capital_gains_short=0, income=100000, dependents=0, sources=False, state_agi=75000, filers_over_65=0, capital_gains_long=0, filing_status='single', credits_state=0):
        """
        --- Parameters ----
        
        Connecticut:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        
        Oregon:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Connecticut:
        - https://portal.ct.gov/-/media/DRS/Forms/2022/Income/2022-CT-1040-Instructions_1222.pdf
          https://www.cga.ct.gov/2022/rpt/pdf/2022-R-0108.pdf
          https://www.cga.ct.gov/current/pub/chap_229.htm#sec_12-703
          https://portal.ct.gov/-/media/DRS/Forms/2022/Income/2022-CT-1040-Instructions_1222.pdf
          "https://portal.ct.gov/DRS/Publications/TSSNs/TSSN-29#:~:text=An%20individuals%20net%20capital%20gains
         portion%20of%20Social%20Security%20benefits"
          "https://portal.ct.gov/DRS/Publications/TSSNs/TSSN-29#:~:text=An%20individuals%20net%20capital%20gains
         portion%20of%20Social%20Security%20benefits"
          https://portal.ct.gov/DRS/Individuals/Resident-Income-Tax/Tax-Information
        
        Oregon:
        -         
        - https://www.oregon.gov/dor/programs/individuals/Pages/PIT.aspx
          https://www.oregon.gov/dor/programs/individuals/Documents/Full%20year%20resident
         %20Form%20or-40%20filers.pdf
          https://www.oregon.gov/dor/programs/individuals/Documents/Part-year%20and%20nonresident
         %20Form%20or-40-P%20and%20or-40-N%20filers.pdf.pdf
          https://www.oregon.gov/dor/forms/FormsPubs/form-or-40_101-040_2022.pdf
          https://www.oregon.gov/dor/forms/FormsPubs/form-or-40_101-040_2022.pdf
          https://www.oregon.gov/dor/forms/FormsPubs/form-or-40_101-040_2022.pdf
          https://www.portland.gov/revenue/personal-tax
        """
        
        # Check if the parameters are set upstream in the instance
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Connecticut', 'state_occupation': 'Oregon', "capital_gains_short": capital_gains_short, "income": income, "dependents": dependents, "sources": sources, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def connecticut_and_pennsylvania(self, county_residence='Other', state_agi=75000, county_occupation='Other', capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Connecticut:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        
        Pennsylvania:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        county_residence : str, optional
            County of residence. Default is Other.
        county_occupation : str, optional
            County of occupation. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Connecticut:
        - https://portal.ct.gov/-/media/DRS/Forms/2022/Income/2022-CT-1040-Instructions_1222.pdf
          https://www.cga.ct.gov/2022/rpt/pdf/2022-R-0108.pdf
          https://www.cga.ct.gov/current/pub/chap_229.htm#sec_12-703
          https://portal.ct.gov/-/media/DRS/Forms/2022/Income/2022-CT-1040-Instructions_1222.pdf
          "https://portal.ct.gov/DRS/Publications/TSSNs/TSSN-29#:~:text=An%20individuals%20net%20capital%20gains
         portion%20of%20Social%20Security%20benefits"
          "https://portal.ct.gov/DRS/Publications/TSSNs/TSSN-29#:~:text=An%20individuals%20net%20capital%20gains
         portion%20of%20Social%20Security%20benefits"
          https://portal.ct.gov/DRS/Individuals/Resident-Income-Tax/Tax-Information
        
        Pennsylvania:
        -         
        - https://www.revenue.pa.gov/Tax%20Rates/Pages/default.aspx
          https://www.revenue.pa.gov/FormsandPublications/FormsforIndividuals/PIT/Documents/2022/2022_pa-40in.pdf
          https://www.revenue.pa.gov/FormsandPublications/FormsforBusinesses/EmployerWithholding/Documents/rev-419.pdf
          https://www.revenue.pa.gov/FormsandPublications/FormsforIndividuals/PIT/Documents/2022/2022_pa-40in.pdf
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-25.htm
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        county_occupation = self.county_occupation if hasattr(self, 'county_occupation') else county_occupation
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Connecticut', 'state_occupation': 'Pennsylvania', "county_residence": county_residence, "state_agi": state_agi, "county_occupation": county_occupation, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def connecticut_and_rhode_island(self, state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Connecticut:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        
        Rhode Island:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Connecticut:
        - https://portal.ct.gov/-/media/DRS/Forms/2022/Income/2022-CT-1040-Instructions_1222.pdf
          https://www.cga.ct.gov/2022/rpt/pdf/2022-R-0108.pdf
          https://www.cga.ct.gov/current/pub/chap_229.htm#sec_12-703
          https://portal.ct.gov/-/media/DRS/Forms/2022/Income/2022-CT-1040-Instructions_1222.pdf
          "https://portal.ct.gov/DRS/Publications/TSSNs/TSSN-29#:~:text=An%20individuals%20net%20capital%20gains
         portion%20of%20Social%20Security%20benefits"
          "https://portal.ct.gov/DRS/Publications/TSSNs/TSSN-29#:~:text=An%20individuals%20net%20capital%20gains
         portion%20of%20Social%20Security%20benefits"
          https://portal.ct.gov/DRS/Individuals/Resident-Income-Tax/Tax-Information
        
        Rhode Island:
        -         
        - https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2022_1040WE_w_0.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Connecticut', 'state_occupation': 'RhodeIsland', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def connecticut_and_south_carolina(self, capital_gains_short=0, income=100000, dependents=0, sources=False, state_agi=75000, filers_over_65=0, capital_gains_long=0, filing_status='single', credits_state=0):
        """
        --- Parameters ----
        
        Connecticut:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        
        South Carolina:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Connecticut:
        - https://portal.ct.gov/-/media/DRS/Forms/2022/Income/2022-CT-1040-Instructions_1222.pdf
          https://www.cga.ct.gov/2022/rpt/pdf/2022-R-0108.pdf
          https://www.cga.ct.gov/current/pub/chap_229.htm#sec_12-703
          https://portal.ct.gov/-/media/DRS/Forms/2022/Income/2022-CT-1040-Instructions_1222.pdf
          "https://portal.ct.gov/DRS/Publications/TSSNs/TSSN-29#:~:text=An%20individuals%20net%20capital%20gains
         portion%20of%20Social%20Security%20benefits"
          "https://portal.ct.gov/DRS/Publications/TSSNs/TSSN-29#:~:text=An%20individuals%20net%20capital%20gains
         portion%20of%20Social%20Security%20benefits"
          https://portal.ct.gov/DRS/Individuals/Resident-Income-Tax/Tax-Information
        
        South Carolina:
        -         
        - https://dor.sc.gov/resources-site/lawandpolicy/Advisory%20Opinions/IL22-15.pdf
          https://dor.sc.gov/forms-site/Forms/IITPacket_2022.pdf
          https://www.scstatehouse.gov/code/t12c006.php
          https://dor.sc.gov/resources-site/lawandpolicy/Advisory%20Opinions/IL22-15.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Connecticut', 'state_occupation': 'SouthCarolina', "capital_gains_short": capital_gains_short, "income": income, "dependents": dependents, "sources": sources, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def connecticut_and_south_dakota(self, capital_gains_long=0, state_agi=75000, filing_status='single', income=100000, sources=False, capital_gains_short=0, credits_state=0):
        """
        --- Parameters ----
        
        Connecticut:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        
        South Dakota:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Connecticut:
        - https://portal.ct.gov/-/media/DRS/Forms/2022/Income/2022-CT-1040-Instructions_1222.pdf
          https://www.cga.ct.gov/2022/rpt/pdf/2022-R-0108.pdf
          https://www.cga.ct.gov/current/pub/chap_229.htm#sec_12-703
          https://portal.ct.gov/-/media/DRS/Forms/2022/Income/2022-CT-1040-Instructions_1222.pdf
          "https://portal.ct.gov/DRS/Publications/TSSNs/TSSN-29#:~:text=An%20individuals%20net%20capital%20gains
         portion%20of%20Social%20Security%20benefits"
          "https://portal.ct.gov/DRS/Publications/TSSNs/TSSN-29#:~:text=An%20individuals%20net%20capital%20gains
         portion%20of%20Social%20Security%20benefits"
          https://portal.ct.gov/DRS/Individuals/Resident-Income-Tax/Tax-Information
        
        South Dakota:
        -         
        - https://dor.sd.gov/individuals/taxes/
        """
        
        # Check if the parameters are set upstream in the instance
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        sources = self.sources if hasattr(self, 'sources') else sources
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Connecticut', 'state_occupation': 'SouthDakota', "capital_gains_long": capital_gains_long, "state_agi": state_agi, "filing_status": filing_status, "income": income, "sources": sources, "capital_gains_short": capital_gains_short, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def connecticut_and_tennessee(self, capital_gains_long=0, state_agi=75000, filing_status='single', income=100000, sources=False, capital_gains_short=0, credits_state=0):
        """
        --- Parameters ----
        
        Connecticut:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        
        Tennessee:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Connecticut:
        - https://portal.ct.gov/-/media/DRS/Forms/2022/Income/2022-CT-1040-Instructions_1222.pdf
          https://www.cga.ct.gov/2022/rpt/pdf/2022-R-0108.pdf
          https://www.cga.ct.gov/current/pub/chap_229.htm#sec_12-703
          https://portal.ct.gov/-/media/DRS/Forms/2022/Income/2022-CT-1040-Instructions_1222.pdf
          "https://portal.ct.gov/DRS/Publications/TSSNs/TSSN-29#:~:text=An%20individuals%20net%20capital%20gains
         portion%20of%20Social%20Security%20benefits"
          "https://portal.ct.gov/DRS/Publications/TSSNs/TSSN-29#:~:text=An%20individuals%20net%20capital%20gains
         portion%20of%20Social%20Security%20benefits"
          https://portal.ct.gov/DRS/Individuals/Resident-Income-Tax/Tax-Information
        
        Tennessee:
        -         
        - https://www.tn.gov/revenue/taxes.html
        """
        
        # Check if the parameters are set upstream in the instance
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        sources = self.sources if hasattr(self, 'sources') else sources
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Connecticut', 'state_occupation': 'Tennessee', "capital_gains_long": capital_gains_long, "state_agi": state_agi, "filing_status": filing_status, "income": income, "sources": sources, "capital_gains_short": capital_gains_short, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def connecticut_and_texas(self, capital_gains_long=0, state_agi=75000, filing_status='single', income=100000, sources=False, capital_gains_short=0, credits_state=0):
        """
        --- Parameters ----
        
        Connecticut:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        
        Texas:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Connecticut:
        - https://portal.ct.gov/-/media/DRS/Forms/2022/Income/2022-CT-1040-Instructions_1222.pdf
          https://www.cga.ct.gov/2022/rpt/pdf/2022-R-0108.pdf
          https://www.cga.ct.gov/current/pub/chap_229.htm#sec_12-703
          https://portal.ct.gov/-/media/DRS/Forms/2022/Income/2022-CT-1040-Instructions_1222.pdf
          "https://portal.ct.gov/DRS/Publications/TSSNs/TSSN-29#:~:text=An%20individuals%20net%20capital%20gains
         portion%20of%20Social%20Security%20benefits"
          "https://portal.ct.gov/DRS/Publications/TSSNs/TSSN-29#:~:text=An%20individuals%20net%20capital%20gains
         portion%20of%20Social%20Security%20benefits"
          https://portal.ct.gov/DRS/Individuals/Resident-Income-Tax/Tax-Information
        
        Texas:
        -         
        - https://comptroller.texas.gov/taxes/
        """
        
        # Check if the parameters are set upstream in the instance
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        sources = self.sources if hasattr(self, 'sources') else sources
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Connecticut', 'state_occupation': 'Texas', "capital_gains_long": capital_gains_long, "state_agi": state_agi, "filing_status": filing_status, "income": income, "sources": sources, "capital_gains_short": capital_gains_short, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def connecticut_and_utah(self, state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Connecticut:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        
        Utah:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Connecticut:
        - https://portal.ct.gov/-/media/DRS/Forms/2022/Income/2022-CT-1040-Instructions_1222.pdf
          https://www.cga.ct.gov/2022/rpt/pdf/2022-R-0108.pdf
          https://www.cga.ct.gov/current/pub/chap_229.htm#sec_12-703
          https://portal.ct.gov/-/media/DRS/Forms/2022/Income/2022-CT-1040-Instructions_1222.pdf
          "https://portal.ct.gov/DRS/Publications/TSSNs/TSSN-29#:~:text=An%20individuals%20net%20capital%20gains
         portion%20of%20Social%20Security%20benefits"
          "https://portal.ct.gov/DRS/Publications/TSSNs/TSSN-29#:~:text=An%20individuals%20net%20capital%20gains
         portion%20of%20Social%20Security%20benefits"
          https://portal.ct.gov/DRS/Individuals/Resident-Income-Tax/Tax-Information
        
        Utah:
        -         
        - https://incometax.utah.gov/paying/tax-rates
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Connecticut', 'state_occupation': 'Utah', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def connecticut_and_vermont(self, capital_gains_short=0, income=100000, dependents=0, sources=False, state_agi=75000, filers_over_65=0, capital_gains_long=0, filing_status='single', credits_state=0):
        """
        --- Parameters ----
        
        Connecticut:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        
        Vermont:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Connecticut:
        - https://portal.ct.gov/-/media/DRS/Forms/2022/Income/2022-CT-1040-Instructions_1222.pdf
          https://www.cga.ct.gov/2022/rpt/pdf/2022-R-0108.pdf
          https://www.cga.ct.gov/current/pub/chap_229.htm#sec_12-703
          https://portal.ct.gov/-/media/DRS/Forms/2022/Income/2022-CT-1040-Instructions_1222.pdf
          "https://portal.ct.gov/DRS/Publications/TSSNs/TSSN-29#:~:text=An%20individuals%20net%20capital%20gains
         portion%20of%20Social%20Security%20benefits"
          "https://portal.ct.gov/DRS/Publications/TSSNs/TSSN-29#:~:text=An%20individuals%20net%20capital%20gains
         portion%20of%20Social%20Security%20benefits"
          https://portal.ct.gov/DRS/Individuals/Resident-Income-Tax/Tax-Information
        
        Vermont:
        -         
        - https://tax.vermont.gov/sites/tax/files/documents/GB-1210-2023.pdf
          https://tax.vermont.gov/sites/tax/files/documents/Income%20Booklet-2022.pdf
          https://tax.vermont.gov/sites/tax/files/documents/Income%20Booklet-2022.pdf
          https://tax.vermont.gov/sites/tax/files/documents/Income%20Booklet-2022.pdf
          https://tax.vermont.gov/individuals/personal-income-tax/tax-credits
          https://tax.vermont.gov/individuals/personal-income-tax/taxable-income
          https://tax.vermont.gov/individuals/personal-income-tax/local-option-tax
        """
        
        # Check if the parameters are set upstream in the instance
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Connecticut', 'state_occupation': 'Vermont', "capital_gains_short": capital_gains_short, "income": income, "dependents": dependents, "sources": sources, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def connecticut_and_virginia(self, capital_gains_short=0, income=100000, dependents=0, sources=False, state_agi=75000, filers_over_65=0, capital_gains_long=0, filing_status='single', credits_state=0):
        """
        --- Parameters ----
        
        Connecticut:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        
        Virginia:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Connecticut:
        - https://portal.ct.gov/-/media/DRS/Forms/2022/Income/2022-CT-1040-Instructions_1222.pdf
          https://www.cga.ct.gov/2022/rpt/pdf/2022-R-0108.pdf
          https://www.cga.ct.gov/current/pub/chap_229.htm#sec_12-703
          https://portal.ct.gov/-/media/DRS/Forms/2022/Income/2022-CT-1040-Instructions_1222.pdf
          "https://portal.ct.gov/DRS/Publications/TSSNs/TSSN-29#:~:text=An%20individuals%20net%20capital%20gains
         portion%20of%20Social%20Security%20benefits"
          "https://portal.ct.gov/DRS/Publications/TSSNs/TSSN-29#:~:text=An%20individuals%20net%20capital%20gains
         portion%20of%20Social%20Security%20benefits"
          https://portal.ct.gov/DRS/Individuals/Resident-Income-Tax/Tax-Information
        
        Virginia:
        -         
        - https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/reciprocity
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Connecticut', 'state_occupation': 'Virginia', "capital_gains_short": capital_gains_short, "income": income, "dependents": dependents, "sources": sources, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def connecticut_and_washington(self, state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Connecticut:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        
        Washington:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Connecticut:
        - https://portal.ct.gov/-/media/DRS/Forms/2022/Income/2022-CT-1040-Instructions_1222.pdf
          https://www.cga.ct.gov/2022/rpt/pdf/2022-R-0108.pdf
          https://www.cga.ct.gov/current/pub/chap_229.htm#sec_12-703
          https://portal.ct.gov/-/media/DRS/Forms/2022/Income/2022-CT-1040-Instructions_1222.pdf
          "https://portal.ct.gov/DRS/Publications/TSSNs/TSSN-29#:~:text=An%20individuals%20net%20capital%20gains
         portion%20of%20Social%20Security%20benefits"
          "https://portal.ct.gov/DRS/Publications/TSSNs/TSSN-29#:~:text=An%20individuals%20net%20capital%20gains
         portion%20of%20Social%20Security%20benefits"
          https://portal.ct.gov/DRS/Individuals/Resident-Income-Tax/Tax-Information
        
        Washington:
        -         
        - https://dor.wa.gov/taxes-rates/income-tax
          https://dor.wa.gov/taxes-rates/other-taxes/capital-gains-tax
          https://dor.wa.gov/taxes-rates/income-tax
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Connecticut', 'state_occupation': 'Washington', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def connecticut_and_west_virginia(self, pay_periods=24, capital_gains_short=0, income=100000, county_occupation='Other', dependents=0, sources=False, state_agi=75000, capital_gains_long=0, filing_status='single', credits_state=0):
        """
        --- Parameters ----
        
        Connecticut:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        
        West Virginia:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_occupation : str, optional
            County of occupation. Default is Other.
        pay_periods : int, optional
            Number of pay periods. Options are 13, 24, 26, 52. Default is 24.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Connecticut:
        - https://portal.ct.gov/-/media/DRS/Forms/2022/Income/2022-CT-1040-Instructions_1222.pdf
          https://www.cga.ct.gov/2022/rpt/pdf/2022-R-0108.pdf
          https://www.cga.ct.gov/current/pub/chap_229.htm#sec_12-703
          https://portal.ct.gov/-/media/DRS/Forms/2022/Income/2022-CT-1040-Instructions_1222.pdf
          "https://portal.ct.gov/DRS/Publications/TSSNs/TSSN-29#:~:text=An%20individuals%20net%20capital%20gains
         portion%20of%20Social%20Security%20benefits"
          "https://portal.ct.gov/DRS/Publications/TSSNs/TSSN-29#:~:text=An%20individuals%20net%20capital%20gains
         portion%20of%20Social%20Security%20benefits"
          https://portal.ct.gov/DRS/Individuals/Resident-Income-Tax/Tax-Information
        
        West Virginia:
        -         
        - https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://tax.wv.gov/Documents/TSD/tsd438.pdf
          https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2011/Taxes-11-35att.html
        """
        
        # Check if the parameters are set upstream in the instance
        pay_periods = self.pay_periods if hasattr(self, 'pay_periods') else pay_periods
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        county_occupation = self.county_occupation if hasattr(self, 'county_occupation') else county_occupation
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Connecticut', 'state_occupation': 'WestVirginia', "pay_periods": pay_periods, "capital_gains_short": capital_gains_short, "income": income, "county_occupation": county_occupation, "dependents": dependents, "sources": sources, "state_agi": state_agi, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def connecticut_and_wisconsin(self, state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Connecticut:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        
        Wisconsin:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Connecticut:
        - https://portal.ct.gov/-/media/DRS/Forms/2022/Income/2022-CT-1040-Instructions_1222.pdf
          https://www.cga.ct.gov/2022/rpt/pdf/2022-R-0108.pdf
          https://www.cga.ct.gov/current/pub/chap_229.htm#sec_12-703
          https://portal.ct.gov/-/media/DRS/Forms/2022/Income/2022-CT-1040-Instructions_1222.pdf
          "https://portal.ct.gov/DRS/Publications/TSSNs/TSSN-29#:~:text=An%20individuals%20net%20capital%20gains
         portion%20of%20Social%20Security%20benefits"
          "https://portal.ct.gov/DRS/Publications/TSSNs/TSSN-29#:~:text=An%20individuals%20net%20capital%20gains
         portion%20of%20Social%20Security%20benefits"
          https://portal.ct.gov/DRS/Individuals/Resident-Income-Tax/Tax-Information
        
        Wisconsin:
        -         
        - https://www.revenue.wi.gov/TaxForms2023/2023-Form1-ES-Inst.pdf
          https://www.revenue.wi.gov/TaxForms2023/2023-Form1-ES-Inst.pdf
          https://www.revenue.wi.gov/TaxForms2023/2023-Form1-ES-Inst.pdf
          https://www.revenue.wi.gov/Pages/FAQS/pcs-work.aspx
          https://www.revenue.wi.gov/Dor%20Publications/pb103.pdf
          https://www.revenue.wi.gov/Pages/FAQS/pcs-work.aspx
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Connecticut', 'state_occupation': 'Wisconsin', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def connecticut_and_wyoming(self, capital_gains_long=0, state_agi=75000, filing_status='single', income=100000, sources=False, capital_gains_short=0, credits_state=0):
        """
        --- Parameters ----
        
        Connecticut:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        
        Wyoming:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Connecticut:
        - https://portal.ct.gov/-/media/DRS/Forms/2022/Income/2022-CT-1040-Instructions_1222.pdf
          https://www.cga.ct.gov/2022/rpt/pdf/2022-R-0108.pdf
          https://www.cga.ct.gov/current/pub/chap_229.htm#sec_12-703
          https://portal.ct.gov/-/media/DRS/Forms/2022/Income/2022-CT-1040-Instructions_1222.pdf
          "https://portal.ct.gov/DRS/Publications/TSSNs/TSSN-29#:~:text=An%20individuals%20net%20capital%20gains
         portion%20of%20Social%20Security%20benefits"
          "https://portal.ct.gov/DRS/Publications/TSSNs/TSSN-29#:~:text=An%20individuals%20net%20capital%20gains
         portion%20of%20Social%20Security%20benefits"
          https://portal.ct.gov/DRS/Individuals/Resident-Income-Tax/Tax-Information
        
        Wyoming:
        -         
        - https://revenue.wyo.gov/
        """
        
        # Check if the parameters are set upstream in the instance
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        sources = self.sources if hasattr(self, 'sources') else sources
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Connecticut', 'state_occupation': 'Wyoming', "capital_gains_long": capital_gains_long, "state_agi": state_agi, "filing_status": filing_status, "income": income, "sources": sources, "capital_gains_short": capital_gains_short, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def delaware_and_alabama(self, county_residence='Other', state_agi=75000, capital_gains_short=0, capital_gains_long=0, filing_status='single', federal_agi=75000, income=100000, dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Delaware:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        federal_agi : int, optional
            Adjusted Gross Income (AGI) for federal taxes. Default is 0. Sources - https://www.irs.gov/e-file-providers/definition-of-adjusted-gross-income, https://www.irs.gov/newsroom/taxpayers-should-know-the-difference-between-standard-and-itemized-deductions
        income : int, optional
            Gross income. Default is 0.
        
        Alabama:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Delaware:
        - https://revenue.delaware.gov/software-developer/tax-rate-changes/
          https://revenuefiles.delaware.gov/2022/PIT-RES_TY22_2022-01_PaperInteractive.pdf
          https://revenuefiles.delaware.gov/2022/PIT-RES_TY22_2022-02_Instructions.pdf
          https://revenue.delaware.gov/frequently-asked-questions/personal-income-tax-faqs/
          https://www.newcastlede.gov/DocumentCenter/View/20147/City-of-Wilmington-Wage-Tax-Information-PDF
        
        Alabama:
        -         
        - https://www.revenue.alabama.gov/ultraviewer/viewer/basic_viewer/index.html?form=2023/02/22f40.pdf
          https://www.revenue.alabama.gov/faqs/what-is-alabamas-individual-income-tax-rate/
          https://www.revenue.alabama.gov/ultraviewer/viewer/basic_viewer/index.html?form=2023/02/22f40.pdf
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-24.htm?taxmap=true
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-24.htm?taxmap=true
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-24.htm?taxmap=true
          https://www.revenue.alabama.gov/individual-corporate/income-to-be-reported-on-the-alabama-income-tax-return/
          https://www.almonline.org/TaxRates.aspx#OccupationalTax
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        federal_agi = self.federal_agi if hasattr(self, 'federal_agi') else federal_agi
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Delaware', 'state_occupation': 'Alabama', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "federal_agi": federal_agi, "income": income, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def delaware_and_alaska(self, capital_gains_long=0, federal_agi=75000, income=100000, sources=False, capital_gains_short=0, credits_state=0):
        """
        --- Parameters ----
        
        Delaware:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        federal_agi : int, optional
            Adjusted Gross Income (AGI) for federal taxes. Default is 0. Sources - https://www.irs.gov/e-file-providers/definition-of-adjusted-gross-income, https://www.irs.gov/newsroom/taxpayers-should-know-the-difference-between-standard-and-itemized-deductions
        income : int, optional
            Gross income. Default is 0.
        
        Alaska:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Delaware:
        - https://revenue.delaware.gov/software-developer/tax-rate-changes/
          https://revenuefiles.delaware.gov/2022/PIT-RES_TY22_2022-01_PaperInteractive.pdf
          https://revenuefiles.delaware.gov/2022/PIT-RES_TY22_2022-02_Instructions.pdf
          https://revenue.delaware.gov/frequently-asked-questions/personal-income-tax-faqs/
          https://www.newcastlede.gov/DocumentCenter/View/20147/City-of-Wilmington-Wage-Tax-Information-PDF
        
        Alaska:
        -         
        - https://www.commerce.alaska.gov/web/dcra/officeofthestateassessor/alaskataxfacts.aspx
        """
        
        # Check if the parameters are set upstream in the instance
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        federal_agi = self.federal_agi if hasattr(self, 'federal_agi') else federal_agi
        income = self.income if hasattr(self, 'income') else income
        sources = self.sources if hasattr(self, 'sources') else sources
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Delaware', 'state_occupation': 'Alaska', "capital_gains_long": capital_gains_long, "federal_agi": federal_agi, "income": income, "sources": sources, "capital_gains_short": capital_gains_short, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def delaware_and_arizona(self, state_agi=75000, capital_gains_short=0, capital_gains_long=0, filing_status='single', federal_agi=75000, income=100000, dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Delaware:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        federal_agi : int, optional
            Adjusted Gross Income (AGI) for federal taxes. Default is 0. Sources - https://www.irs.gov/e-file-providers/definition-of-adjusted-gross-income, https://www.irs.gov/newsroom/taxpayers-should-know-the-difference-between-standard-and-itemized-deductions
        income : int, optional
            Gross income. Default is 0.
        
        Arizona:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        credits_state : int, optional
            Credits for state taxes. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Delaware:
        - https://revenue.delaware.gov/software-developer/tax-rate-changes/
          https://revenuefiles.delaware.gov/2022/PIT-RES_TY22_2022-01_PaperInteractive.pdf
          https://revenuefiles.delaware.gov/2022/PIT-RES_TY22_2022-02_Instructions.pdf
          https://revenue.delaware.gov/frequently-asked-questions/personal-income-tax-faqs/
          https://www.newcastlede.gov/DocumentCenter/View/20147/City-of-Wilmington-Wage-Tax-Information-PDF
        
        Arizona:
        -         
        - https://azdor.gov/forms/individual/form-140-x-y-tables
          https://azdor.gov/forms/individual/form-140-x-y-tables
          https://azdor.gov/forms/individual/form-140-x-y-tables
          https://azdor.gov/businesses-arizona/withholding-tax/withholding-exceptions
          "https://azdor.gov/individuals/income-tax-filing-assistance/identifying-other-taxable-income#:~:text=In%20regards%20to%20capital%20gains
         the%20individuals%20regular%20tax%20rate.
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        federal_agi = self.federal_agi if hasattr(self, 'federal_agi') else federal_agi
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Delaware', 'state_occupation': 'Arizona', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "federal_agi": federal_agi, "income": income, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def delaware_and_arkansas(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, capital_gains_long=0, filing_status='single', federal_agi=75000, income=100000, dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Delaware:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        federal_agi : int, optional
            Adjusted Gross Income (AGI) for federal taxes. Default is 0. Sources - https://www.irs.gov/e-file-providers/definition-of-adjusted-gross-income, https://www.irs.gov/newsroom/taxpayers-should-know-the-difference-between-standard-and-itemized-deductions
        income : int, optional
            Gross income. Default is 0.
        
        Arkansas:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Delaware:
        - https://revenue.delaware.gov/software-developer/tax-rate-changes/
          https://revenuefiles.delaware.gov/2022/PIT-RES_TY22_2022-01_PaperInteractive.pdf
          https://revenuefiles.delaware.gov/2022/PIT-RES_TY22_2022-02_Instructions.pdf
          https://revenue.delaware.gov/frequently-asked-questions/personal-income-tax-faqs/
          https://www.newcastlede.gov/DocumentCenter/View/20147/City-of-Wilmington-Wage-Tax-Information-PDF
        
        Arkansas:
        -         
        - https://www.arkansasedc.com/why-arkansas/business-climate/tax-structure/personal-income-tax
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_and_AR1000NR_Instructions.pdf
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_FullYearResidentIndividualIncomeTaxReturn.pdf
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_and_AR1000NR_Instructions.pdf
          https://www.arkansasedc.com/why-arkansas/business-climate/tax-structure/capital-gains-tax-reduction
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_and_AR1000NR_Instructions.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        federal_agi = self.federal_agi if hasattr(self, 'federal_agi') else federal_agi
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Delaware', 'state_occupation': 'Arkansas', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "federal_agi": federal_agi, "income": income, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def delaware_and_california(self, state_agi=75000, capital_gains_short=0, capital_gains_long=0, filing_status='single', federal_agi=75000, income=100000, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Delaware:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        federal_agi : int, optional
            Adjusted Gross Income (AGI) for federal taxes. Default is 0. Sources - https://www.irs.gov/e-file-providers/definition-of-adjusted-gross-income, https://www.irs.gov/newsroom/taxpayers-should-know-the-difference-between-standard-and-itemized-deductions
        income : int, optional
            Gross income. Default is 0.
        
        California:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Delaware:
        - https://revenue.delaware.gov/software-developer/tax-rate-changes/
          https://revenuefiles.delaware.gov/2022/PIT-RES_TY22_2022-01_PaperInteractive.pdf
          https://revenuefiles.delaware.gov/2022/PIT-RES_TY22_2022-02_Instructions.pdf
          https://revenue.delaware.gov/frequently-asked-questions/personal-income-tax-faqs/
          https://www.newcastlede.gov/DocumentCenter/View/20147/City-of-Wilmington-Wage-Tax-Information-PDF
        
        California:
        -         
        - https://www.ftb.ca.gov/forms/2022/2022-540-tax-rate-schedules.pdf
          https://www.ftb.ca.gov/file/personal/deductions/index.html
          https://www.ftb.ca.gov/about-ftb/newsroom/caleitc/eligibility-and-credit-information.html
          https://www.ftb.ca.gov/file/personal/income-types/capital-gains-and-losses.html#:~:text=California%20does%20not%20have%20a
         are%20taxed%20as%20ordinary%20income
          https://www.ftb.ca.gov/file/personal/
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        federal_agi = self.federal_agi if hasattr(self, 'federal_agi') else federal_agi
        income = self.income if hasattr(self, 'income') else income
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Delaware', 'state_occupation': 'California', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "federal_agi": federal_agi, "income": income, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def delaware_and_colorado(self, county_residence='Other', state_agi=75000, capital_gains_short=0, capital_gains_long=0, federal_agi=75000, income=100000, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Delaware:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        federal_agi : int, optional
            Adjusted Gross Income (AGI) for federal taxes. Default is 0. Sources - https://www.irs.gov/e-file-providers/definition-of-adjusted-gross-income, https://www.irs.gov/newsroom/taxpayers-should-know-the-difference-between-standard-and-itemized-deductions
        income : int, optional
            Gross income. Default is 0.
        
        Colorado:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Delaware:
        - https://revenue.delaware.gov/software-developer/tax-rate-changes/
          https://revenuefiles.delaware.gov/2022/PIT-RES_TY22_2022-01_PaperInteractive.pdf
          https://revenuefiles.delaware.gov/2022/PIT-RES_TY22_2022-02_Instructions.pdf
          https://revenue.delaware.gov/frequently-asked-questions/personal-income-tax-faqs/
          https://www.newcastlede.gov/DocumentCenter/View/20147/City-of-Wilmington-Wage-Tax-Information-PDF
        
        Colorado:
        -         
        - https://leg.colorado.gov/agencies/legislative-council-staff/individual-income-tax%C2%A0
          https://leg.colorado.gov/agencies/legislative-council-staff/individual-income-tax%C2%A0
          https://tax.colorado.gov/sites/tax/files/documents/ITT_Capital_Gain_Subtraction_Feb_2022.pdf
          https://www.auroragov.org/business_services/taxes/occupational_privilege_tax
          https://www.denvergov.org/Government/Agencies-Departments-Offices/Agencies-Departments-Offices-Directory/Department-of-Finance/Our-Divisions/Treasury/Business-Tax-Information#:~:text=Occupational%20privilege%20tax
         -%28show%20below%29&text=Each%20taxable%20employee%20is%20liable
         month%20for%20each%20taxable%20employee
          https://www.glendale.co.us/355/Occupational-Privilege-Tax
          https://www.greenwoodvillage.com/1220/Occupational-Privilege-Tax-OPT
          https://ci.sheridan.co.us/288/Occupational-Privilege-Tax
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        federal_agi = self.federal_agi if hasattr(self, 'federal_agi') else federal_agi
        income = self.income if hasattr(self, 'income') else income
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Delaware', 'state_occupation': 'Colorado', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "federal_agi": federal_agi, "income": income, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def delaware_and_connecticut(self, state_agi=75000, capital_gains_short=0, capital_gains_long=0, filing_status='single', federal_agi=75000, income=100000, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Delaware:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        federal_agi : int, optional
            Adjusted Gross Income (AGI) for federal taxes. Default is 0. Sources - https://www.irs.gov/e-file-providers/definition-of-adjusted-gross-income, https://www.irs.gov/newsroom/taxpayers-should-know-the-difference-between-standard-and-itemized-deductions
        income : int, optional
            Gross income. Default is 0.
        
        Connecticut:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Delaware:
        - https://revenue.delaware.gov/software-developer/tax-rate-changes/
          https://revenuefiles.delaware.gov/2022/PIT-RES_TY22_2022-01_PaperInteractive.pdf
          https://revenuefiles.delaware.gov/2022/PIT-RES_TY22_2022-02_Instructions.pdf
          https://revenue.delaware.gov/frequently-asked-questions/personal-income-tax-faqs/
          https://www.newcastlede.gov/DocumentCenter/View/20147/City-of-Wilmington-Wage-Tax-Information-PDF
        
        Connecticut:
        -         
        - https://portal.ct.gov/-/media/DRS/Forms/2022/Income/2022-CT-1040-Instructions_1222.pdf
          https://www.cga.ct.gov/2022/rpt/pdf/2022-R-0108.pdf
          https://www.cga.ct.gov/current/pub/chap_229.htm#sec_12-703
          https://portal.ct.gov/-/media/DRS/Forms/2022/Income/2022-CT-1040-Instructions_1222.pdf
          "https://portal.ct.gov/DRS/Publications/TSSNs/TSSN-29#:~:text=An%20individuals%20net%20capital%20gains
         portion%20of%20Social%20Security%20benefits"
          "https://portal.ct.gov/DRS/Publications/TSSNs/TSSN-29#:~:text=An%20individuals%20net%20capital%20gains
         portion%20of%20Social%20Security%20benefits"
          https://portal.ct.gov/DRS/Individuals/Resident-Income-Tax/Tax-Information
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        federal_agi = self.federal_agi if hasattr(self, 'federal_agi') else federal_agi
        income = self.income if hasattr(self, 'income') else income
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Delaware', 'state_occupation': 'Connecticut', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "federal_agi": federal_agi, "income": income, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def delaware_and_florida(self, capital_gains_long=0, federal_agi=75000, income=100000, sources=False, capital_gains_short=0, credits_state=0):
        """
        --- Parameters ----
        
        Delaware:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        federal_agi : int, optional
            Adjusted Gross Income (AGI) for federal taxes. Default is 0. Sources - https://www.irs.gov/e-file-providers/definition-of-adjusted-gross-income, https://www.irs.gov/newsroom/taxpayers-should-know-the-difference-between-standard-and-itemized-deductions
        income : int, optional
            Gross income. Default is 0.
        
        Florida:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Delaware:
        - https://revenue.delaware.gov/software-developer/tax-rate-changes/
          https://revenuefiles.delaware.gov/2022/PIT-RES_TY22_2022-01_PaperInteractive.pdf
          https://revenuefiles.delaware.gov/2022/PIT-RES_TY22_2022-02_Instructions.pdf
          https://revenue.delaware.gov/frequently-asked-questions/personal-income-tax-faqs/
          https://www.newcastlede.gov/DocumentCenter/View/20147/City-of-Wilmington-Wage-Tax-Information-PDF
        
        Florida:
        -         
        - https://floridarevenue.com/taxes/taxesfees/Pages/tax_interest_rates.aspx
        """
        
        # Check if the parameters are set upstream in the instance
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        federal_agi = self.federal_agi if hasattr(self, 'federal_agi') else federal_agi
        income = self.income if hasattr(self, 'income') else income
        sources = self.sources if hasattr(self, 'sources') else sources
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Delaware', 'state_occupation': 'Florida', "capital_gains_long": capital_gains_long, "federal_agi": federal_agi, "income": income, "sources": sources, "capital_gains_short": capital_gains_short, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def delaware_and_georgia(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, capital_gains_long=0, filing_status='single', federal_agi=75000, income=100000, dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Delaware:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        federal_agi : int, optional
            Adjusted Gross Income (AGI) for federal taxes. Default is 0. Sources - https://www.irs.gov/e-file-providers/definition-of-adjusted-gross-income, https://www.irs.gov/newsroom/taxpayers-should-know-the-difference-between-standard-and-itemized-deductions
        income : int, optional
            Gross income. Default is 0.
        
        Georgia:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Delaware:
        - https://revenue.delaware.gov/software-developer/tax-rate-changes/
          https://revenuefiles.delaware.gov/2022/PIT-RES_TY22_2022-01_PaperInteractive.pdf
          https://revenuefiles.delaware.gov/2022/PIT-RES_TY22_2022-02_Instructions.pdf
          https://revenue.delaware.gov/frequently-asked-questions/personal-income-tax-faqs/
          https://www.newcastlede.gov/DocumentCenter/View/20147/City-of-Wilmington-Wage-Tax-Information-PDF
        
        Georgia:
        -         
        - https://dor.georgia.gov/tax-tables-georgia-tax-rate-schedule
          https://apps.dor.ga.gov/FillableForms/PDFViewer/Index?form=2022GA500
          https://apps.dor.ga.gov/FillableForms/PDFViewer/Index?form=2022GA500
          https://apps.dor.ga.gov/FillableForms/PDFViewer/Index?form=2022GA500
          https://dor.georgia.gov/years-individual-income-tax-forms
          https://dor.georgia.gov/years-individual-income-tax-forms
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        federal_agi = self.federal_agi if hasattr(self, 'federal_agi') else federal_agi
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Delaware', 'state_occupation': 'Georgia', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "federal_agi": federal_agi, "income": income, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def delaware_and_hawaii(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, capital_gains_long=0, filing_status='single', federal_agi=75000, income=100000, dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Delaware:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        federal_agi : int, optional
            Adjusted Gross Income (AGI) for federal taxes. Default is 0. Sources - https://www.irs.gov/e-file-providers/definition-of-adjusted-gross-income, https://www.irs.gov/newsroom/taxpayers-should-know-the-difference-between-standard-and-itemized-deductions
        income : int, optional
            Gross income. Default is 0.
        
        Hawaii:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Delaware:
        - https://revenue.delaware.gov/software-developer/tax-rate-changes/
          https://revenuefiles.delaware.gov/2022/PIT-RES_TY22_2022-01_PaperInteractive.pdf
          https://revenuefiles.delaware.gov/2022/PIT-RES_TY22_2022-02_Instructions.pdf
          https://revenue.delaware.gov/frequently-asked-questions/personal-income-tax-faqs/
          https://www.newcastlede.gov/DocumentCenter/View/20147/City-of-Wilmington-Wage-Tax-Information-PDF
        
        Hawaii:
        -         
        - https://tax.hawaii.gov/forms/d_18table-on/d_18table-on_p13/
          https://files.hawaii.gov/tax/forms/2018/18table-on.pdf
          https://files.hawaii.gov/tax/news/pubs/22outline.pdf
          https://files.hawaii.gov/tax/news/pubs/FastTaxRefGuide_2017.pdf
          https://www.tfhawaii.org/wordpress/state-tax-resources/mini-tax-guides/individual-income-tax-chapter-235/#:~:text=Long%20term%20capital%20gains%20are%20taxed%20at%20a%20maximum%20of%207.25%25
          https://www.grassrootinstitute.org/2023/02/hb337-increasing-tax-on-capital-gains-a-bad-risk/
          https://files.hawaii.gov/tax/news/pubs/22outline.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        federal_agi = self.federal_agi if hasattr(self, 'federal_agi') else federal_agi
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Delaware', 'state_occupation': 'Hawaii', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "federal_agi": federal_agi, "income": income, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def delaware_and_idaho(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, capital_gains_long=0, filing_status='single', federal_agi=75000, income=100000, dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Delaware:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        federal_agi : int, optional
            Adjusted Gross Income (AGI) for federal taxes. Default is 0. Sources - https://www.irs.gov/e-file-providers/definition-of-adjusted-gross-income, https://www.irs.gov/newsroom/taxpayers-should-know-the-difference-between-standard-and-itemized-deductions
        income : int, optional
            Gross income. Default is 0.
        
        Idaho:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Delaware:
        - https://revenue.delaware.gov/software-developer/tax-rate-changes/
          https://revenuefiles.delaware.gov/2022/PIT-RES_TY22_2022-01_PaperInteractive.pdf
          https://revenuefiles.delaware.gov/2022/PIT-RES_TY22_2022-02_Instructions.pdf
          https://revenue.delaware.gov/frequently-asked-questions/personal-income-tax-faqs/
          https://www.newcastlede.gov/DocumentCenter/View/20147/City-of-Wilmington-Wage-Tax-Information-PDF
        
        Idaho:
        -         
        - https://tax.idaho.gov/taxes/income-tax/individual-income/individual-income-tax-rate-schedule/
          https://tax.idaho.gov/wp-content/uploads/forms/EIN00046/EIN00046_03-01-2023.pdf
          https://tax.idaho.gov/taxes/income-tax/individual-income/idaho-source-income/
          https://tax.idaho.gov/taxes/income-tax/individual-income/idaho-source-income/
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        federal_agi = self.federal_agi if hasattr(self, 'federal_agi') else federal_agi
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Delaware', 'state_occupation': 'Idaho', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "federal_agi": federal_agi, "income": income, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def delaware_and_illinois(self, state_agi=75000, capital_gains_short=0, capital_gains_long=0, filing_status='single', federal_agi=75000, income=100000, dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Delaware:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        federal_agi : int, optional
            Adjusted Gross Income (AGI) for federal taxes. Default is 0. Sources - https://www.irs.gov/e-file-providers/definition-of-adjusted-gross-income, https://www.irs.gov/newsroom/taxpayers-should-know-the-difference-between-standard-and-itemized-deductions
        income : int, optional
            Gross income. Default is 0.
        
        Illinois:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Delaware:
        - https://revenue.delaware.gov/software-developer/tax-rate-changes/
          https://revenuefiles.delaware.gov/2022/PIT-RES_TY22_2022-01_PaperInteractive.pdf
          https://revenuefiles.delaware.gov/2022/PIT-RES_TY22_2022-02_Instructions.pdf
          https://revenue.delaware.gov/frequently-asked-questions/personal-income-tax-faqs/
          https://www.newcastlede.gov/DocumentCenter/View/20147/City-of-Wilmington-Wage-Tax-Information-PDF
        
        Illinois:
        -         
        - https://tax.illinois.gov/research/taxrates/income.html#IndividualIncome
          https://tax.illinois.gov/content/dam/soi/en/web/tax/forms/incometax/documents/currentyear/individual/il-1040-instr.pdf
          https://tax.illinois.gov/questionsandanswers/12.html
          https://tax.illinois.gov/content/dam/soi/en/web/tax/forms/incometax/documents/currentyear/individual/il-1040-instr.pdf
          https://tax.illinois.gov/research/taxrates/personalproperty.html
          https://tax.illinois.gov/localgovernments/income.html
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        federal_agi = self.federal_agi if hasattr(self, 'federal_agi') else federal_agi
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Delaware', 'state_occupation': 'Illinois', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "federal_agi": federal_agi, "income": income, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def delaware_and_indiana(self, capital_gains_short=0, dependents=0, sources=False, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_long=0, filing_status='single', federal_agi=75000, income=100000, credits_state=0):
        """
        --- Parameters ----
        
        Delaware:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        federal_agi : int, optional
            Adjusted Gross Income (AGI) for federal taxes. Default is 0. Sources - https://www.irs.gov/e-file-providers/definition-of-adjusted-gross-income, https://www.irs.gov/newsroom/taxpayers-should-know-the-difference-between-standard-and-itemized-deductions
        income : int, optional
            Gross income. Default is 0.
        
        Indiana:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Delaware:
        - https://revenue.delaware.gov/software-developer/tax-rate-changes/
          https://revenuefiles.delaware.gov/2022/PIT-RES_TY22_2022-01_PaperInteractive.pdf
          https://revenuefiles.delaware.gov/2022/PIT-RES_TY22_2022-02_Instructions.pdf
          https://revenue.delaware.gov/frequently-asked-questions/personal-income-tax-faqs/
          https://www.newcastlede.gov/DocumentCenter/View/20147/City-of-Wilmington-Wage-Tax-Information-PDF
        
        Indiana:
        -         
        - https://www.in.gov/dor/tax-forms/2022-individual-income-tax-forms/
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-32.htm
          https://www.in.gov/dor/files/ib28.pdf
          https://www.in.gov/dor/tax-forms/2022-individual-income-tax-forms/
          https://www.in.gov/dor/files/2022-county-tax-rates-and-codes.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        federal_agi = self.federal_agi if hasattr(self, 'federal_agi') else federal_agi
        income = self.income if hasattr(self, 'income') else income
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Delaware', 'state_occupation': 'Indiana', "capital_gains_short": capital_gains_short, "dependents": dependents, "sources": sources, "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "federal_agi": federal_agi, "income": income, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def delaware_and_iowa(self, county_residence='Other', state_agi=75000, capital_gains_short=0, capital_gains_long=0, filing_status='single', federal_agi=75000, income=100000, dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Delaware:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        federal_agi : int, optional
            Adjusted Gross Income (AGI) for federal taxes. Default is 0. Sources - https://www.irs.gov/e-file-providers/definition-of-adjusted-gross-income, https://www.irs.gov/newsroom/taxpayers-should-know-the-difference-between-standard-and-itemized-deductions
        income : int, optional
            Gross income. Default is 0.
        
        Iowa:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Delaware:
        - https://revenue.delaware.gov/software-developer/tax-rate-changes/
          https://revenuefiles.delaware.gov/2022/PIT-RES_TY22_2022-01_PaperInteractive.pdf
          https://revenuefiles.delaware.gov/2022/PIT-RES_TY22_2022-02_Instructions.pdf
          https://revenue.delaware.gov/frequently-asked-questions/personal-income-tax-faqs/
          https://www.newcastlede.gov/DocumentCenter/View/20147/City-of-Wilmington-Wage-Tax-Information-PDF
        
        Iowa:
        -         
        - https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
          https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
          https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
          https://tax.iowa.gov/iowa-illinois-reciprocal-agreement
          https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        federal_agi = self.federal_agi if hasattr(self, 'federal_agi') else federal_agi
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Delaware', 'state_occupation': 'Iowa', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "federal_agi": federal_agi, "income": income, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def delaware_and_kansas(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, capital_gains_long=0, filing_status='single', federal_agi=75000, income=100000, dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Delaware:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        federal_agi : int, optional
            Adjusted Gross Income (AGI) for federal taxes. Default is 0. Sources - https://www.irs.gov/e-file-providers/definition-of-adjusted-gross-income, https://www.irs.gov/newsroom/taxpayers-should-know-the-difference-between-standard-and-itemized-deductions
        income : int, optional
            Gross income. Default is 0.
        
        Kansas:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Delaware:
        - https://revenue.delaware.gov/software-developer/tax-rate-changes/
          https://revenuefiles.delaware.gov/2022/PIT-RES_TY22_2022-01_PaperInteractive.pdf
          https://revenuefiles.delaware.gov/2022/PIT-RES_TY22_2022-02_Instructions.pdf
          https://revenue.delaware.gov/frequently-asked-questions/personal-income-tax-faqs/
          https://www.newcastlede.gov/DocumentCenter/View/20147/City-of-Wilmington-Wage-Tax-Information-PDF
        
        Kansas:
        -         
        - https://www.ksrevenue.gov/taxrates.html
          https://www.ksrevenue.gov/incomebook22.html#0
          https://www.ksrevenue.gov/incomebook22.html#0
          https://www.ksrevenue.gov/incomebook22.html#0
          https://www.ksrevenue.gov/incomebook22.html
          https://www.ksrevenue.gov/incomebook22.html
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        federal_agi = self.federal_agi if hasattr(self, 'federal_agi') else federal_agi
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Delaware', 'state_occupation': 'Kansas', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "federal_agi": federal_agi, "income": income, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def delaware_and_kentucky(self, county_residence='Other', state_agi=75000, capital_gains_short=0, capital_gains_long=0, filing_status='single', federal_agi=75000, income=100000, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Delaware:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        federal_agi : int, optional
            Adjusted Gross Income (AGI) for federal taxes. Default is 0. Sources - https://www.irs.gov/e-file-providers/definition-of-adjusted-gross-income, https://www.irs.gov/newsroom/taxpayers-should-know-the-difference-between-standard-and-itemized-deductions
        income : int, optional
            Gross income. Default is 0.
        
        Kentucky:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Delaware:
        - https://revenue.delaware.gov/software-developer/tax-rate-changes/
          https://revenuefiles.delaware.gov/2022/PIT-RES_TY22_2022-01_PaperInteractive.pdf
          https://revenuefiles.delaware.gov/2022/PIT-RES_TY22_2022-02_Instructions.pdf
          https://revenue.delaware.gov/frequently-asked-questions/personal-income-tax-faqs/
          https://www.newcastlede.gov/DocumentCenter/View/20147/City-of-Wilmington-Wage-Tax-Information-PDF
        
        Kentucky:
        -         
        - https://revenue.ky.gov/News/Pages/Dor-Announces-Updates-to-Individual-Income-Tax-for-2023-Tax-Year.aspx
          https://revenue.ky.gov/Individual/Individual-Income-Tax/Pages/default.aspx
          https://revenue.ky.gov/News/Pages/Dor-Announces-Updates-to-Individual-Income-Tax-for-2023-Tax-Year.aspx
          https://revenue.ky.gov/Individual/Individual-Income-Tax/Pages/default.aspx
          https://revenue.ky.gov/Dor%20Training%20Materials/103%20KAR%2017.140.%20Individual%20income%20tax%20-%20reciprocity%20-%20nonresidents.pdf
          https://revenue.ky.gov/Individual/Individual-Income-Tax/Pages/default.aspx
          https://nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2021/Taxes-21-32.htm?taxmap=true
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        federal_agi = self.federal_agi if hasattr(self, 'federal_agi') else federal_agi
        income = self.income if hasattr(self, 'income') else income
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Delaware', 'state_occupation': 'Kentucky', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "federal_agi": federal_agi, "income": income, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def delaware_and_louisiana(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, capital_gains_long=0, filing_status='single', federal_agi=75000, income=100000, dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Delaware:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        federal_agi : int, optional
            Adjusted Gross Income (AGI) for federal taxes. Default is 0. Sources - https://www.irs.gov/e-file-providers/definition-of-adjusted-gross-income, https://www.irs.gov/newsroom/taxpayers-should-know-the-difference-between-standard-and-itemized-deductions
        income : int, optional
            Gross income. Default is 0.
        
        Louisiana:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Delaware:
        - https://revenue.delaware.gov/software-developer/tax-rate-changes/
          https://revenuefiles.delaware.gov/2022/PIT-RES_TY22_2022-01_PaperInteractive.pdf
          https://revenuefiles.delaware.gov/2022/PIT-RES_TY22_2022-02_Instructions.pdf
          https://revenue.delaware.gov/frequently-asked-questions/personal-income-tax-faqs/
          https://www.newcastlede.gov/DocumentCenter/View/20147/City-of-Wilmington-Wage-Tax-Information-PDF
        
        Louisiana:
        -         
        - https://revenue.louisiana.gov/individualincometax
          https://revenue.louisiana.gov/individualincometax
          https://www.revenue.louisiana.gov/taxforms/6935(11_02)F.pdf
          https://revenue.louisiana.gov/individualincometax
          https://revenue.louisiana.gov/individualincometax
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        federal_agi = self.federal_agi if hasattr(self, 'federal_agi') else federal_agi
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Delaware', 'state_occupation': 'Louisiana', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "federal_agi": federal_agi, "income": income, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def delaware_and_maine(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, capital_gains_long=0, filing_status='single', federal_agi=75000, income=100000, dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Delaware:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        federal_agi : int, optional
            Adjusted Gross Income (AGI) for federal taxes. Default is 0. Sources - https://www.irs.gov/e-file-providers/definition-of-adjusted-gross-income, https://www.irs.gov/newsroom/taxpayers-should-know-the-difference-between-standard-and-itemized-deductions
        income : int, optional
            Gross income. Default is 0.
        
        Maine:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Delaware:
        - https://revenue.delaware.gov/software-developer/tax-rate-changes/
          https://revenuefiles.delaware.gov/2022/PIT-RES_TY22_2022-01_PaperInteractive.pdf
          https://revenuefiles.delaware.gov/2022/PIT-RES_TY22_2022-02_Instructions.pdf
          https://revenue.delaware.gov/frequently-asked-questions/personal-income-tax-faqs/
          https://www.newcastlede.gov/DocumentCenter/View/20147/City-of-Wilmington-Wage-Tax-Information-PDF
        
        Maine:
        -         
        - https://www.maine.gov/revenue/tax-return-forms/individual-income-tax-2022
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/ind_tax_rate_sched_2022.pdf
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/22_1040me_book_gen_instruc_revisedFeb23.pdf
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/22_1040me_book_gen_instruc_revisedFeb23.pdf
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/22_1040me_book_gen_instruc_revisedFeb23.pdf
          https://www.maine.gov/revenue/tax-return-forms/individual-income-tax-2022
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        federal_agi = self.federal_agi if hasattr(self, 'federal_agi') else federal_agi
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Delaware', 'state_occupation': 'Maine', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "federal_agi": federal_agi, "income": income, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def delaware_and_maryland(self, county_residence='Other', state_agi=75000, capital_gains_short=0, capital_gains_long=0, filing_status='single', federal_agi=75000, income=100000, dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Delaware:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        federal_agi : int, optional
            Adjusted Gross Income (AGI) for federal taxes. Default is 0. Sources - https://www.irs.gov/e-file-providers/definition-of-adjusted-gross-income, https://www.irs.gov/newsroom/taxpayers-should-know-the-difference-between-standard-and-itemized-deductions
        income : int, optional
            Gross income. Default is 0.
        
        Maryland:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Delaware:
        - https://revenue.delaware.gov/software-developer/tax-rate-changes/
          https://revenuefiles.delaware.gov/2022/PIT-RES_TY22_2022-01_PaperInteractive.pdf
          https://revenuefiles.delaware.gov/2022/PIT-RES_TY22_2022-02_Instructions.pdf
          https://revenue.delaware.gov/frequently-asked-questions/personal-income-tax-faqs/
          https://www.newcastlede.gov/DocumentCenter/View/20147/City-of-Wilmington-Wage-Tax-Information-PDF
        
        Maryland:
        -         
        - https://www.marylandtaxes.gov/individual/income/tax-info/tax-rates.php
          https://www.marylandtaxes.gov/forms/22_forms/Resident_Booklet.pdf
          https://www.marylandtaxes.gov/forms/Tax_Publications/Administrative_Releases/Income_and_Estate_Tax_Releases/ar_it37.pdf
          https://www.marylandtaxes.gov/individual/index.php
          https://www.marylandtaxes.gov/individual/income/tax-info/tax-rates.php
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        federal_agi = self.federal_agi if hasattr(self, 'federal_agi') else federal_agi
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Delaware', 'state_occupation': 'Maryland', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "federal_agi": federal_agi, "income": income, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def delaware_and_massachusetts(self, state_agi=75000, capital_gains_short=0, capital_gains_long=0, filing_status='single', federal_agi=75000, income=100000, dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Delaware:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        federal_agi : int, optional
            Adjusted Gross Income (AGI) for federal taxes. Default is 0. Sources - https://www.irs.gov/e-file-providers/definition-of-adjusted-gross-income, https://www.irs.gov/newsroom/taxpayers-should-know-the-difference-between-standard-and-itemized-deductions
        income : int, optional
            Gross income. Default is 0.
        
        Massachusetts:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Delaware:
        - https://revenue.delaware.gov/software-developer/tax-rate-changes/
          https://revenuefiles.delaware.gov/2022/PIT-RES_TY22_2022-01_PaperInteractive.pdf
          https://revenuefiles.delaware.gov/2022/PIT-RES_TY22_2022-02_Instructions.pdf
          https://revenue.delaware.gov/frequently-asked-questions/personal-income-tax-faqs/
          https://www.newcastlede.gov/DocumentCenter/View/20147/City-of-Wilmington-Wage-Tax-Information-PDF
        
        Massachusetts:
        -         
        - https://www.mass.gov/service-details/massachusetts-tax-rates
          https://www.mass.gov/service-details/view-massachusetts-personal-income-tax-exemptions
          https://www.mass.gov/service-details/view-massachusetts-personal-income-tax-exemptions
          https://www.mass.gov/service-details/view-child-and-dependent-related-credits
          https://www.mass.gov/service-details/massachusetts-tax-rates
          https://www.mass.gov/personal-income-tax
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        federal_agi = self.federal_agi if hasattr(self, 'federal_agi') else federal_agi
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Delaware', 'state_occupation': 'Massachusetts', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "federal_agi": federal_agi, "income": income, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def delaware_and_michigan(self, capital_gains_short=0, county_occupation='Other', dependents=0, sources=False, county_residence='Other', state_agi=75000, capital_gains_long=0, filing_status='single', federal_agi=75000, income=100000, credits_state=0):
        """
        --- Parameters ----
        
        Delaware:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        federal_agi : int, optional
            Adjusted Gross Income (AGI) for federal taxes. Default is 0. Sources - https://www.irs.gov/e-file-providers/definition-of-adjusted-gross-income, https://www.irs.gov/newsroom/taxpayers-should-know-the-difference-between-standard-and-itemized-deductions
        income : int, optional
            Gross income. Default is 0.
        
        Michigan:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        county_occupation : str, optional
            County of occupation. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Delaware:
        - https://revenue.delaware.gov/software-developer/tax-rate-changes/
          https://revenuefiles.delaware.gov/2022/PIT-RES_TY22_2022-01_PaperInteractive.pdf
          https://revenuefiles.delaware.gov/2022/PIT-RES_TY22_2022-02_Instructions.pdf
          https://revenue.delaware.gov/frequently-asked-questions/personal-income-tax-faqs/
          https://www.newcastlede.gov/DocumentCenter/View/20147/City-of-Wilmington-Wage-Tax-Information-PDF
        
        Michigan:
        -         
        - https://www.michigan.gov/taxes/iit/new-developments-for-tax-year-2022
          "https://www.michigan.gov/taxes/-/media/Project/Websites/taxes/Forms/2023/2023-SUW/446_Withholding-Guide_2023.pdf?rev=177ca3d8cb034392bf84e317ed78c320&hash=19C5F42F43A89A91863D229349F48883#:~:text=The%20exemption%20amount%20is%20%245%2C400
         employees%20Michigan%20income%20tax%20return"
          "https://www.michigan.gov/taxes/-/media/Project/Websites/taxes/Forms/2023/2023-SUW/446_Withholding-Guide_2023.pdf?rev=177ca3d8cb034392bf84e317ed78c320&hash=19C5F42F43A89A91863D229349F48883#:~:text=The%20exemption%20amount%20is%20%245%2C400
         employees%20Michigan%20income%20tax%20return"
          https://www.michigan.gov/taxes/questions/iit/accordion/residency/are-my-wages-earned-in-another-state-taxable-in-michigan-if-i-am-a-michigan-resident-1
          https://www.michigan.gov/taxes/-/media/Project/Websites/taxes/Forms/2022/2022-IIT-Forms/BOOK_MI-1040.pdf?rev=4524f1cbbc0243b194ab05a6680be75b&hash=40DFE71E651F7C1D65C31F6CD03D944F
          https://www.michigan.gov/taxes/questions/iit/accordion/general/what-cities-impose-an-income-tax
        """
        
        # Check if the parameters are set upstream in the instance
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        county_occupation = self.county_occupation if hasattr(self, 'county_occupation') else county_occupation
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        federal_agi = self.federal_agi if hasattr(self, 'federal_agi') else federal_agi
        income = self.income if hasattr(self, 'income') else income
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Delaware', 'state_occupation': 'Michigan', "capital_gains_short": capital_gains_short, "county_occupation": county_occupation, "dependents": dependents, "sources": sources, "county_residence": county_residence, "state_agi": state_agi, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "federal_agi": federal_agi, "income": income, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def delaware_and_minnesota(self, state_agi=75000, capital_gains_short=0, capital_gains_long=0, filing_status='single', federal_agi=75000, income=100000, dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Delaware:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        federal_agi : int, optional
            Adjusted Gross Income (AGI) for federal taxes. Default is 0. Sources - https://www.irs.gov/e-file-providers/definition-of-adjusted-gross-income, https://www.irs.gov/newsroom/taxpayers-should-know-the-difference-between-standard-and-itemized-deductions
        income : int, optional
            Gross income. Default is 0.
        
        Minnesota:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Delaware:
        - https://revenue.delaware.gov/software-developer/tax-rate-changes/
          https://revenuefiles.delaware.gov/2022/PIT-RES_TY22_2022-01_PaperInteractive.pdf
          https://revenuefiles.delaware.gov/2022/PIT-RES_TY22_2022-02_Instructions.pdf
          https://revenue.delaware.gov/frequently-asked-questions/personal-income-tax-faqs/
          https://www.newcastlede.gov/DocumentCenter/View/20147/City-of-Wilmington-Wage-Tax-Information-PDF
        
        Minnesota:
        -         
        - https://www.revenue.state.mn.us/minnesota-income-tax-rates-and-brackets
          https://www.revenue.state.mn.us/child-and-dependent-care-credit
          https://www.revenue.state.mn.us/dependent-exemptions
          https://www.revenue.state.mn.us/minnesota-standard-deduction
          https://mn.gov/mmb/assets/mwr_form_tcm1059-128581.pdf
          https://www.revenue.state.mn.us/individual-income-tax
          https://www.revenue.state.mn.us/individual-income-tax
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        federal_agi = self.federal_agi if hasattr(self, 'federal_agi') else federal_agi
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Delaware', 'state_occupation': 'Minnesota', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "federal_agi": federal_agi, "income": income, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def delaware_and_mississippi(self, state_agi=75000, capital_gains_short=0, capital_gains_long=0, filing_status='single', federal_agi=75000, income=100000, dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Delaware:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        federal_agi : int, optional
            Adjusted Gross Income (AGI) for federal taxes. Default is 0. Sources - https://www.irs.gov/e-file-providers/definition-of-adjusted-gross-income, https://www.irs.gov/newsroom/taxpayers-should-know-the-difference-between-standard-and-itemized-deductions
        income : int, optional
            Gross income. Default is 0.
        
        Mississippi:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Delaware:
        - https://revenue.delaware.gov/software-developer/tax-rate-changes/
          https://revenuefiles.delaware.gov/2022/PIT-RES_TY22_2022-01_PaperInteractive.pdf
          https://revenuefiles.delaware.gov/2022/PIT-RES_TY22_2022-02_Instructions.pdf
          https://revenue.delaware.gov/frequently-asked-questions/personal-income-tax-faqs/
          https://www.newcastlede.gov/DocumentCenter/View/20147/City-of-Wilmington-Wage-Tax-Information-PDF
        
        Mississippi:
        -         
        - https://www.dor.ms.gov/individual/tax-rates
          https://www.dor.ms.gov/individual/tax-rates
          https://www.dor.ms.gov/individual/tax-rates
          https://www.dor.ms.gov/individual
          https://www.dor.ms.gov/individual
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        federal_agi = self.federal_agi if hasattr(self, 'federal_agi') else federal_agi
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Delaware', 'state_occupation': 'Mississippi', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "federal_agi": federal_agi, "income": income, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def delaware_and_missouri(self, capital_gains_short=0, dependents=0, sources=False, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_long=0, filing_status='single', federal_agi=75000, income=100000, credits_state=0):
        """
        --- Parameters ----
        
        Delaware:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        federal_agi : int, optional
            Adjusted Gross Income (AGI) for federal taxes. Default is 0. Sources - https://www.irs.gov/e-file-providers/definition-of-adjusted-gross-income, https://www.irs.gov/newsroom/taxpayers-should-know-the-difference-between-standard-and-itemized-deductions
        income : int, optional
            Gross income. Default is 0.
        
        Missouri:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Delaware:
        - https://revenue.delaware.gov/software-developer/tax-rate-changes/
          https://revenuefiles.delaware.gov/2022/PIT-RES_TY22_2022-01_PaperInteractive.pdf
          https://revenuefiles.delaware.gov/2022/PIT-RES_TY22_2022-02_Instructions.pdf
          https://revenue.delaware.gov/frequently-asked-questions/personal-income-tax-faqs/
          https://www.newcastlede.gov/DocumentCenter/View/20147/City-of-Wilmington-Wage-Tax-Information-PDF
        
        Missouri:
        -         
        - https://dor.mo.gov/taxation/individual/tax-types/income/year-changes/
          https://dor.mo.gov/taxation/individual/tax-types/income/year-changes/
          https://dor.mo.gov/taxation/individual/tax-types/income/year-changes/
          https://www.kcmo.gov/city-hall/departments/finance/earnings-tax
          https://www.stlouis-mo.gov/government/departments/collector/earnings-tax/individual-earnings-tax-info.cfm
        """
        
        # Check if the parameters are set upstream in the instance
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        federal_agi = self.federal_agi if hasattr(self, 'federal_agi') else federal_agi
        income = self.income if hasattr(self, 'income') else income
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Delaware', 'state_occupation': 'Missouri', "capital_gains_short": capital_gains_short, "dependents": dependents, "sources": sources, "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "federal_agi": federal_agi, "income": income, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def delaware_and_montana(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, capital_gains_long=0, filing_status='single', federal_agi=75000, income=100000, dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Delaware:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        federal_agi : int, optional
            Adjusted Gross Income (AGI) for federal taxes. Default is 0. Sources - https://www.irs.gov/e-file-providers/definition-of-adjusted-gross-income, https://www.irs.gov/newsroom/taxpayers-should-know-the-difference-between-standard-and-itemized-deductions
        income : int, optional
            Gross income. Default is 0.
        
        Montana:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Delaware:
        - https://revenue.delaware.gov/software-developer/tax-rate-changes/
          https://revenuefiles.delaware.gov/2022/PIT-RES_TY22_2022-01_PaperInteractive.pdf
          https://revenuefiles.delaware.gov/2022/PIT-RES_TY22_2022-02_Instructions.pdf
          https://revenue.delaware.gov/frequently-asked-questions/personal-income-tax-faqs/
          https://www.newcastlede.gov/DocumentCenter/View/20147/City-of-Wilmington-Wage-Tax-Information-PDF
        
        Montana:
        -         
        - https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/2022/12/montana-employees-withholding-allowance-and-exemption-certificate-form-mw-4.pdf
          https://mtrevenue.gov/2021/11/29/simplification-of-montana-income-taxation/?utm_source=rss&utm_medium=rss&utm_campaign=simplification-of-montana-income-taxation
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        federal_agi = self.federal_agi if hasattr(self, 'federal_agi') else federal_agi
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Delaware', 'state_occupation': 'Montana', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "federal_agi": federal_agi, "income": income, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def delaware_and_nebraska(self, state_agi=75000, capital_gains_short=0, capital_gains_long=0, filing_status='single', federal_agi=75000, income=100000, dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Delaware:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        federal_agi : int, optional
            Adjusted Gross Income (AGI) for federal taxes. Default is 0. Sources - https://www.irs.gov/e-file-providers/definition-of-adjusted-gross-income, https://www.irs.gov/newsroom/taxpayers-should-know-the-difference-between-standard-and-itemized-deductions
        income : int, optional
            Gross income. Default is 0.
        
        Nebraska:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Delaware:
        - https://revenue.delaware.gov/software-developer/tax-rate-changes/
          https://revenuefiles.delaware.gov/2022/PIT-RES_TY22_2022-01_PaperInteractive.pdf
          https://revenuefiles.delaware.gov/2022/PIT-RES_TY22_2022-02_Instructions.pdf
          https://revenue.delaware.gov/frequently-asked-questions/personal-income-tax-faqs/
          https://www.newcastlede.gov/DocumentCenter/View/20147/City-of-Wilmington-Wage-Tax-Information-PDF
        
        Nebraska:
        -         
        - https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2023_Circular_EN_8-429r2023_Whole_Book_11.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        federal_agi = self.federal_agi if hasattr(self, 'federal_agi') else federal_agi
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Delaware', 'state_occupation': 'Nebraska', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "federal_agi": federal_agi, "income": income, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def delaware_and_nevada(self, capital_gains_long=0, federal_agi=75000, income=100000, sources=False, capital_gains_short=0, credits_state=0):
        """
        --- Parameters ----
        
        Delaware:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        federal_agi : int, optional
            Adjusted Gross Income (AGI) for federal taxes. Default is 0. Sources - https://www.irs.gov/e-file-providers/definition-of-adjusted-gross-income, https://www.irs.gov/newsroom/taxpayers-should-know-the-difference-between-standard-and-itemized-deductions
        income : int, optional
            Gross income. Default is 0.
        
        Nevada:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Delaware:
        - https://revenue.delaware.gov/software-developer/tax-rate-changes/
          https://revenuefiles.delaware.gov/2022/PIT-RES_TY22_2022-01_PaperInteractive.pdf
          https://revenuefiles.delaware.gov/2022/PIT-RES_TY22_2022-02_Instructions.pdf
          https://revenue.delaware.gov/frequently-asked-questions/personal-income-tax-faqs/
          https://www.newcastlede.gov/DocumentCenter/View/20147/City-of-Wilmington-Wage-Tax-Information-PDF
        
        Nevada:
        -         
        - https://www.nvsos.gov/sos/businesses/the-nevada-advantage
        """
        
        # Check if the parameters are set upstream in the instance
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        federal_agi = self.federal_agi if hasattr(self, 'federal_agi') else federal_agi
        income = self.income if hasattr(self, 'income') else income
        sources = self.sources if hasattr(self, 'sources') else sources
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Delaware', 'state_occupation': 'Nevada', "capital_gains_long": capital_gains_long, "federal_agi": federal_agi, "income": income, "sources": sources, "capital_gains_short": capital_gains_short, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def delaware_and_new_hampshire(self, capital_gains_long=0, federal_agi=75000, income=100000, sources=False, capital_gains_short=0, credits_state=0):
        """
        --- Parameters ----
        
        Delaware:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        federal_agi : int, optional
            Adjusted Gross Income (AGI) for federal taxes. Default is 0. Sources - https://www.irs.gov/e-file-providers/definition-of-adjusted-gross-income, https://www.irs.gov/newsroom/taxpayers-should-know-the-difference-between-standard-and-itemized-deductions
        income : int, optional
            Gross income. Default is 0.
        
        New Hampshire:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Delaware:
        - https://revenue.delaware.gov/software-developer/tax-rate-changes/
          https://revenuefiles.delaware.gov/2022/PIT-RES_TY22_2022-01_PaperInteractive.pdf
          https://revenuefiles.delaware.gov/2022/PIT-RES_TY22_2022-02_Instructions.pdf
          https://revenue.delaware.gov/frequently-asked-questions/personal-income-tax-faqs/
          https://www.newcastlede.gov/DocumentCenter/View/20147/City-of-Wilmington-Wage-Tax-Information-PDF
        
        New Hampshire:
        -         
        - https://www.revenue.nh.gov/assistance/tax-overview.htm
        """
        
        # Check if the parameters are set upstream in the instance
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        federal_agi = self.federal_agi if hasattr(self, 'federal_agi') else federal_agi
        income = self.income if hasattr(self, 'income') else income
        sources = self.sources if hasattr(self, 'sources') else sources
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Delaware', 'state_occupation': 'NewHampshire', "capital_gains_long": capital_gains_long, "federal_agi": federal_agi, "income": income, "sources": sources, "capital_gains_short": capital_gains_short, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def delaware_and_new_jersey(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, capital_gains_long=0, filing_status='single', federal_agi=75000, income=100000, dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Delaware:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        federal_agi : int, optional
            Adjusted Gross Income (AGI) for federal taxes. Default is 0. Sources - https://www.irs.gov/e-file-providers/definition-of-adjusted-gross-income, https://www.irs.gov/newsroom/taxpayers-should-know-the-difference-between-standard-and-itemized-deductions
        income : int, optional
            Gross income. Default is 0.
        
        New Jersey:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Delaware:
        - https://revenue.delaware.gov/software-developer/tax-rate-changes/
          https://revenuefiles.delaware.gov/2022/PIT-RES_TY22_2022-01_PaperInteractive.pdf
          https://revenuefiles.delaware.gov/2022/PIT-RES_TY22_2022-02_Instructions.pdf
          https://revenue.delaware.gov/frequently-asked-questions/personal-income-tax-faqs/
          https://www.newcastlede.gov/DocumentCenter/View/20147/City-of-Wilmington-Wage-Tax-Information-PDF
        
        New Jersey:
        -         
        - https://www.state.nj.us/treasury/taxation/taxtables.shtml
          https://www.state.nj.us/treasury/taxation/njit13.shtml
          https://www.nj.gov/treasury/taxation/njit25.shtml
          https://www.state.nj.us/treasury/taxation/individuals/freqqiti.shtml
          https://www.state.nj.us/treasury/taxation/localtax.shtml
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        federal_agi = self.federal_agi if hasattr(self, 'federal_agi') else federal_agi
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Delaware', 'state_occupation': 'NewJersey', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "federal_agi": federal_agi, "income": income, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def delaware_and_new_mexico(self, state_agi=75000, capital_gains_short=0, capital_gains_long=0, filing_status='single', federal_agi=75000, income=100000, dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Delaware:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        federal_agi : int, optional
            Adjusted Gross Income (AGI) for federal taxes. Default is 0. Sources - https://www.irs.gov/e-file-providers/definition-of-adjusted-gross-income, https://www.irs.gov/newsroom/taxpayers-should-know-the-difference-between-standard-and-itemized-deductions
        income : int, optional
            Gross income. Default is 0.
        
        New Mexico:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Delaware:
        - https://revenue.delaware.gov/software-developer/tax-rate-changes/
          https://revenuefiles.delaware.gov/2022/PIT-RES_TY22_2022-01_PaperInteractive.pdf
          https://revenuefiles.delaware.gov/2022/PIT-RES_TY22_2022-02_Instructions.pdf
          https://revenue.delaware.gov/frequently-asked-questions/personal-income-tax-faqs/
          https://www.newcastlede.gov/DocumentCenter/View/20147/City-of-Wilmington-Wage-Tax-Information-PDF
        
        New Mexico:
        -         
        - https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/fdf3c548-8aba-4b9c-9eb4-bb564c716015/FYI-104.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/0bf943fd-652e-4400-bb1b-cd8397eb5a95/2022pit-adj.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/2f1a6781-9534-4436-b427-1557f9592099/2022pit-adj-ins.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/0bf943fd-652e-4400-bb1b-cd8397eb5a95/2022pit-adj.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/1afc56af-ea90-4d48-82e5-1f9aeb43255a/PITbook2022.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/fdf3c548-8aba-4b9c-9eb4-bb564c716015/FYI-104.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        federal_agi = self.federal_agi if hasattr(self, 'federal_agi') else federal_agi
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Delaware', 'state_occupation': 'NewMexico', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "federal_agi": federal_agi, "income": income, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def delaware_and_new_york(self, county_residence='Other', state_agi=75000, capital_gains_short=0, capital_gains_long=0, filing_status='single', federal_agi=75000, income=100000, dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Delaware:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        federal_agi : int, optional
            Adjusted Gross Income (AGI) for federal taxes. Default is 0. Sources - https://www.irs.gov/e-file-providers/definition-of-adjusted-gross-income, https://www.irs.gov/newsroom/taxpayers-should-know-the-difference-between-standard-and-itemized-deductions
        income : int, optional
            Gross income. Default is 0.
        
        New York:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Delaware:
        - https://revenue.delaware.gov/software-developer/tax-rate-changes/
          https://revenuefiles.delaware.gov/2022/PIT-RES_TY22_2022-01_PaperInteractive.pdf
          https://revenuefiles.delaware.gov/2022/PIT-RES_TY22_2022-02_Instructions.pdf
          https://revenue.delaware.gov/frequently-asked-questions/personal-income-tax-faqs/
          https://www.newcastlede.gov/DocumentCenter/View/20147/City-of-Wilmington-Wage-Tax-Information-PDF
        
        New York:
        -         
        - https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nys-tax-rate-schedule
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nys-tax-rate-schedule
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nys-tax-rate-schedule
          https://www.tax.ny.gov/pit/
          https://www.tax.ny.gov/pit/file/nonresident-faqs.htm
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#yonkers-tax-rate-schedule
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nyc-tax-rate-schedule
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        federal_agi = self.federal_agi if hasattr(self, 'federal_agi') else federal_agi
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Delaware', 'state_occupation': 'NewYork', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "federal_agi": federal_agi, "income": income, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def delaware_and_north_carolina(self, state_agi=75000, capital_gains_short=0, capital_gains_long=0, filing_status='single', federal_agi=75000, income=100000, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Delaware:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        federal_agi : int, optional
            Adjusted Gross Income (AGI) for federal taxes. Default is 0. Sources - https://www.irs.gov/e-file-providers/definition-of-adjusted-gross-income, https://www.irs.gov/newsroom/taxpayers-should-know-the-difference-between-standard-and-itemized-deductions
        income : int, optional
            Gross income. Default is 0.
        
        North Carolina:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Delaware:
        - https://revenue.delaware.gov/software-developer/tax-rate-changes/
          https://revenuefiles.delaware.gov/2022/PIT-RES_TY22_2022-01_PaperInteractive.pdf
          https://revenuefiles.delaware.gov/2022/PIT-RES_TY22_2022-02_Instructions.pdf
          https://revenue.delaware.gov/frequently-asked-questions/personal-income-tax-faqs/
          https://www.newcastlede.gov/DocumentCenter/View/20147/City-of-Wilmington-Wage-Tax-Information-PDF
        
        North Carolina:
        -         
        - https://www.ncdor.gov/taxes-forms/tax-rate-schedules
          https://www.ncdor.gov/taxes-forms/individual-income-tax/north-carolina-standard-deduction-or-north-carolina-itemized-deductions
          https://www.ncdor.gov/taxes-forms/individual-income-tax
          https://www.ncdor.gov/taxes-forms/individual-income-tax
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        federal_agi = self.federal_agi if hasattr(self, 'federal_agi') else federal_agi
        income = self.income if hasattr(self, 'income') else income
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Delaware', 'state_occupation': 'NorthCarolina', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "federal_agi": federal_agi, "income": income, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def delaware_and_north_dakota(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, capital_gains_long=0, filing_status='single', federal_agi=75000, income=100000, dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Delaware:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        federal_agi : int, optional
            Adjusted Gross Income (AGI) for federal taxes. Default is 0. Sources - https://www.irs.gov/e-file-providers/definition-of-adjusted-gross-income, https://www.irs.gov/newsroom/taxpayers-should-know-the-difference-between-standard-and-itemized-deductions
        income : int, optional
            Gross income. Default is 0.
        
        North Dakota:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Delaware:
        - https://revenue.delaware.gov/software-developer/tax-rate-changes/
          https://revenuefiles.delaware.gov/2022/PIT-RES_TY22_2022-01_PaperInteractive.pdf
          https://revenuefiles.delaware.gov/2022/PIT-RES_TY22_2022-02_Instructions.pdf
          https://revenue.delaware.gov/frequently-asked-questions/personal-income-tax-faqs/
          https://www.newcastlede.gov/DocumentCenter/View/20147/City-of-Wilmington-Wage-Tax-Information-PDF
        
        North Dakota:
        -         
        - https://www.tax.nd.gov/forms
          https://www.tax.nd.gov/sites/www/files/documents/forms/individual/2022-iit/2022-individual-income-tax-booklet.pdf
          https://www.tax.nd.gov/sites/www/files/documents/forms/2022-individual-income-tax-booklet.pdf
          https://www.tax.nd.gov/tax-exemptions-credits/income-tax-exemptions-credits
          https://www.tax.nd.gov/sites/www/files/documents/forms/business/ndwrfillable.pdf
          https://www.tax.nd.gov/news/tax-legislative-changes/significant-changes-law/individual-income-tax-history
          https://www.tax.nd.gov/forms
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        federal_agi = self.federal_agi if hasattr(self, 'federal_agi') else federal_agi
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Delaware', 'state_occupation': 'NorthDakota', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "federal_agi": federal_agi, "income": income, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def delaware_and_ohio(self, county_residence='Other', state_agi=75000, capital_gains_short=0, capital_gains_long=0, filing_status='single', federal_agi=75000, income=100000, dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Delaware:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        federal_agi : int, optional
            Adjusted Gross Income (AGI) for federal taxes. Default is 0. Sources - https://www.irs.gov/e-file-providers/definition-of-adjusted-gross-income, https://www.irs.gov/newsroom/taxpayers-should-know-the-difference-between-standard-and-itemized-deductions
        income : int, optional
            Gross income. Default is 0.
        
        Ohio:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Delaware:
        - https://revenue.delaware.gov/software-developer/tax-rate-changes/
          https://revenuefiles.delaware.gov/2022/PIT-RES_TY22_2022-01_PaperInteractive.pdf
          https://revenuefiles.delaware.gov/2022/PIT-RES_TY22_2022-02_Instructions.pdf
          https://revenue.delaware.gov/frequently-asked-questions/personal-income-tax-faqs/
          https://www.newcastlede.gov/DocumentCenter/View/20147/City-of-Wilmington-Wage-Tax-Information-PDF
        
        Ohio:
        -         
        - https://tax.ohio.gov/individual/resources/annual-tax-rates
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://tax.ohio.gov/static/forms/employer_withholding/generic/wth_it4nr.pdf
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://www.ritaohio.com/TaxRatesTable
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        federal_agi = self.federal_agi if hasattr(self, 'federal_agi') else federal_agi
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Delaware', 'state_occupation': 'Ohio', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "federal_agi": federal_agi, "income": income, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def delaware_and_oklahoma(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, capital_gains_long=0, filing_status='single', federal_agi=75000, income=100000, dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Delaware:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        federal_agi : int, optional
            Adjusted Gross Income (AGI) for federal taxes. Default is 0. Sources - https://www.irs.gov/e-file-providers/definition-of-adjusted-gross-income, https://www.irs.gov/newsroom/taxpayers-should-know-the-difference-between-standard-and-itemized-deductions
        income : int, optional
            Gross income. Default is 0.
        
        Oklahoma:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Delaware:
        - https://revenue.delaware.gov/software-developer/tax-rate-changes/
          https://revenuefiles.delaware.gov/2022/PIT-RES_TY22_2022-01_PaperInteractive.pdf
          https://revenuefiles.delaware.gov/2022/PIT-RES_TY22_2022-02_Instructions.pdf
          https://revenue.delaware.gov/frequently-asked-questions/personal-income-tax-faqs/
          https://www.newcastlede.gov/DocumentCenter/View/20147/City-of-Wilmington-Wage-Tax-Information-PDF
        
        Oklahoma:
        -         
        - https://oklahoma.gov/content/dam/ok/en/tax/documents/resources/publications/businesses/withholding-tables/WHTables-2022.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        federal_agi = self.federal_agi if hasattr(self, 'federal_agi') else federal_agi
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Delaware', 'state_occupation': 'Oklahoma', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "federal_agi": federal_agi, "income": income, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def delaware_and_oregon(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, capital_gains_long=0, filing_status='single', federal_agi=75000, income=100000, dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Delaware:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        federal_agi : int, optional
            Adjusted Gross Income (AGI) for federal taxes. Default is 0. Sources - https://www.irs.gov/e-file-providers/definition-of-adjusted-gross-income, https://www.irs.gov/newsroom/taxpayers-should-know-the-difference-between-standard-and-itemized-deductions
        income : int, optional
            Gross income. Default is 0.
        
        Oregon:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Delaware:
        - https://revenue.delaware.gov/software-developer/tax-rate-changes/
          https://revenuefiles.delaware.gov/2022/PIT-RES_TY22_2022-01_PaperInteractive.pdf
          https://revenuefiles.delaware.gov/2022/PIT-RES_TY22_2022-02_Instructions.pdf
          https://revenue.delaware.gov/frequently-asked-questions/personal-income-tax-faqs/
          https://www.newcastlede.gov/DocumentCenter/View/20147/City-of-Wilmington-Wage-Tax-Information-PDF
        
        Oregon:
        -         
        - https://www.oregon.gov/dor/programs/individuals/Pages/PIT.aspx
          https://www.oregon.gov/dor/programs/individuals/Documents/Full%20year%20resident
         %20Form%20or-40%20filers.pdf
          https://www.oregon.gov/dor/programs/individuals/Documents/Part-year%20and%20nonresident
         %20Form%20or-40-P%20and%20or-40-N%20filers.pdf.pdf
          https://www.oregon.gov/dor/forms/FormsPubs/form-or-40_101-040_2022.pdf
          https://www.oregon.gov/dor/forms/FormsPubs/form-or-40_101-040_2022.pdf
          https://www.oregon.gov/dor/forms/FormsPubs/form-or-40_101-040_2022.pdf
          https://www.portland.gov/revenue/personal-tax
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        federal_agi = self.federal_agi if hasattr(self, 'federal_agi') else federal_agi
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Delaware', 'state_occupation': 'Oregon', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "federal_agi": federal_agi, "income": income, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def delaware_and_pennsylvania(self, county_residence='Other', state_agi=75000, county_occupation='Other', capital_gains_short=0, capital_gains_long=0, federal_agi=75000, income=100000, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Delaware:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        federal_agi : int, optional
            Adjusted Gross Income (AGI) for federal taxes. Default is 0. Sources - https://www.irs.gov/e-file-providers/definition-of-adjusted-gross-income, https://www.irs.gov/newsroom/taxpayers-should-know-the-difference-between-standard-and-itemized-deductions
        income : int, optional
            Gross income. Default is 0.
        
        Pennsylvania:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        county_residence : str, optional
            County of residence. Default is Other.
        county_occupation : str, optional
            County of occupation. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Delaware:
        - https://revenue.delaware.gov/software-developer/tax-rate-changes/
          https://revenuefiles.delaware.gov/2022/PIT-RES_TY22_2022-01_PaperInteractive.pdf
          https://revenuefiles.delaware.gov/2022/PIT-RES_TY22_2022-02_Instructions.pdf
          https://revenue.delaware.gov/frequently-asked-questions/personal-income-tax-faqs/
          https://www.newcastlede.gov/DocumentCenter/View/20147/City-of-Wilmington-Wage-Tax-Information-PDF
        
        Pennsylvania:
        -         
        - https://www.revenue.pa.gov/Tax%20Rates/Pages/default.aspx
          https://www.revenue.pa.gov/FormsandPublications/FormsforIndividuals/PIT/Documents/2022/2022_pa-40in.pdf
          https://www.revenue.pa.gov/FormsandPublications/FormsforBusinesses/EmployerWithholding/Documents/rev-419.pdf
          https://www.revenue.pa.gov/FormsandPublications/FormsforIndividuals/PIT/Documents/2022/2022_pa-40in.pdf
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-25.htm
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        county_occupation = self.county_occupation if hasattr(self, 'county_occupation') else county_occupation
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        federal_agi = self.federal_agi if hasattr(self, 'federal_agi') else federal_agi
        income = self.income if hasattr(self, 'income') else income
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Delaware', 'state_occupation': 'Pennsylvania', "county_residence": county_residence, "state_agi": state_agi, "county_occupation": county_occupation, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "federal_agi": federal_agi, "income": income, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def delaware_and_rhode_island(self, state_agi=75000, capital_gains_short=0, capital_gains_long=0, filing_status='single', federal_agi=75000, income=100000, dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Delaware:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        federal_agi : int, optional
            Adjusted Gross Income (AGI) for federal taxes. Default is 0. Sources - https://www.irs.gov/e-file-providers/definition-of-adjusted-gross-income, https://www.irs.gov/newsroom/taxpayers-should-know-the-difference-between-standard-and-itemized-deductions
        income : int, optional
            Gross income. Default is 0.
        
        Rhode Island:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Delaware:
        - https://revenue.delaware.gov/software-developer/tax-rate-changes/
          https://revenuefiles.delaware.gov/2022/PIT-RES_TY22_2022-01_PaperInteractive.pdf
          https://revenuefiles.delaware.gov/2022/PIT-RES_TY22_2022-02_Instructions.pdf
          https://revenue.delaware.gov/frequently-asked-questions/personal-income-tax-faqs/
          https://www.newcastlede.gov/DocumentCenter/View/20147/City-of-Wilmington-Wage-Tax-Information-PDF
        
        Rhode Island:
        -         
        - https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2022_1040WE_w_0.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        federal_agi = self.federal_agi if hasattr(self, 'federal_agi') else federal_agi
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Delaware', 'state_occupation': 'RhodeIsland', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "federal_agi": federal_agi, "income": income, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def delaware_and_south_carolina(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, capital_gains_long=0, filing_status='single', federal_agi=75000, income=100000, dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Delaware:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        federal_agi : int, optional
            Adjusted Gross Income (AGI) for federal taxes. Default is 0. Sources - https://www.irs.gov/e-file-providers/definition-of-adjusted-gross-income, https://www.irs.gov/newsroom/taxpayers-should-know-the-difference-between-standard-and-itemized-deductions
        income : int, optional
            Gross income. Default is 0.
        
        South Carolina:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Delaware:
        - https://revenue.delaware.gov/software-developer/tax-rate-changes/
          https://revenuefiles.delaware.gov/2022/PIT-RES_TY22_2022-01_PaperInteractive.pdf
          https://revenuefiles.delaware.gov/2022/PIT-RES_TY22_2022-02_Instructions.pdf
          https://revenue.delaware.gov/frequently-asked-questions/personal-income-tax-faqs/
          https://www.newcastlede.gov/DocumentCenter/View/20147/City-of-Wilmington-Wage-Tax-Information-PDF
        
        South Carolina:
        -         
        - https://dor.sc.gov/resources-site/lawandpolicy/Advisory%20Opinions/IL22-15.pdf
          https://dor.sc.gov/forms-site/Forms/IITPacket_2022.pdf
          https://www.scstatehouse.gov/code/t12c006.php
          https://dor.sc.gov/resources-site/lawandpolicy/Advisory%20Opinions/IL22-15.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        federal_agi = self.federal_agi if hasattr(self, 'federal_agi') else federal_agi
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Delaware', 'state_occupation': 'SouthCarolina', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "federal_agi": federal_agi, "income": income, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def delaware_and_south_dakota(self, capital_gains_long=0, federal_agi=75000, income=100000, sources=False, capital_gains_short=0, credits_state=0):
        """
        --- Parameters ----
        
        Delaware:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        federal_agi : int, optional
            Adjusted Gross Income (AGI) for federal taxes. Default is 0. Sources - https://www.irs.gov/e-file-providers/definition-of-adjusted-gross-income, https://www.irs.gov/newsroom/taxpayers-should-know-the-difference-between-standard-and-itemized-deductions
        income : int, optional
            Gross income. Default is 0.
        
        South Dakota:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Delaware:
        - https://revenue.delaware.gov/software-developer/tax-rate-changes/
          https://revenuefiles.delaware.gov/2022/PIT-RES_TY22_2022-01_PaperInteractive.pdf
          https://revenuefiles.delaware.gov/2022/PIT-RES_TY22_2022-02_Instructions.pdf
          https://revenue.delaware.gov/frequently-asked-questions/personal-income-tax-faqs/
          https://www.newcastlede.gov/DocumentCenter/View/20147/City-of-Wilmington-Wage-Tax-Information-PDF
        
        South Dakota:
        -         
        - https://dor.sd.gov/individuals/taxes/
        """
        
        # Check if the parameters are set upstream in the instance
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        federal_agi = self.federal_agi if hasattr(self, 'federal_agi') else federal_agi
        income = self.income if hasattr(self, 'income') else income
        sources = self.sources if hasattr(self, 'sources') else sources
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Delaware', 'state_occupation': 'SouthDakota', "capital_gains_long": capital_gains_long, "federal_agi": federal_agi, "income": income, "sources": sources, "capital_gains_short": capital_gains_short, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def delaware_and_tennessee(self, capital_gains_long=0, federal_agi=75000, income=100000, sources=False, capital_gains_short=0, credits_state=0):
        """
        --- Parameters ----
        
        Delaware:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        federal_agi : int, optional
            Adjusted Gross Income (AGI) for federal taxes. Default is 0. Sources - https://www.irs.gov/e-file-providers/definition-of-adjusted-gross-income, https://www.irs.gov/newsroom/taxpayers-should-know-the-difference-between-standard-and-itemized-deductions
        income : int, optional
            Gross income. Default is 0.
        
        Tennessee:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Delaware:
        - https://revenue.delaware.gov/software-developer/tax-rate-changes/
          https://revenuefiles.delaware.gov/2022/PIT-RES_TY22_2022-01_PaperInteractive.pdf
          https://revenuefiles.delaware.gov/2022/PIT-RES_TY22_2022-02_Instructions.pdf
          https://revenue.delaware.gov/frequently-asked-questions/personal-income-tax-faqs/
          https://www.newcastlede.gov/DocumentCenter/View/20147/City-of-Wilmington-Wage-Tax-Information-PDF
        
        Tennessee:
        -         
        - https://www.tn.gov/revenue/taxes.html
        """
        
        # Check if the parameters are set upstream in the instance
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        federal_agi = self.federal_agi if hasattr(self, 'federal_agi') else federal_agi
        income = self.income if hasattr(self, 'income') else income
        sources = self.sources if hasattr(self, 'sources') else sources
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Delaware', 'state_occupation': 'Tennessee', "capital_gains_long": capital_gains_long, "federal_agi": federal_agi, "income": income, "sources": sources, "capital_gains_short": capital_gains_short, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def delaware_and_texas(self, capital_gains_long=0, federal_agi=75000, income=100000, sources=False, capital_gains_short=0, credits_state=0):
        """
        --- Parameters ----
        
        Delaware:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        federal_agi : int, optional
            Adjusted Gross Income (AGI) for federal taxes. Default is 0. Sources - https://www.irs.gov/e-file-providers/definition-of-adjusted-gross-income, https://www.irs.gov/newsroom/taxpayers-should-know-the-difference-between-standard-and-itemized-deductions
        income : int, optional
            Gross income. Default is 0.
        
        Texas:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Delaware:
        - https://revenue.delaware.gov/software-developer/tax-rate-changes/
          https://revenuefiles.delaware.gov/2022/PIT-RES_TY22_2022-01_PaperInteractive.pdf
          https://revenuefiles.delaware.gov/2022/PIT-RES_TY22_2022-02_Instructions.pdf
          https://revenue.delaware.gov/frequently-asked-questions/personal-income-tax-faqs/
          https://www.newcastlede.gov/DocumentCenter/View/20147/City-of-Wilmington-Wage-Tax-Information-PDF
        
        Texas:
        -         
        - https://comptroller.texas.gov/taxes/
        """
        
        # Check if the parameters are set upstream in the instance
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        federal_agi = self.federal_agi if hasattr(self, 'federal_agi') else federal_agi
        income = self.income if hasattr(self, 'income') else income
        sources = self.sources if hasattr(self, 'sources') else sources
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Delaware', 'state_occupation': 'Texas', "capital_gains_long": capital_gains_long, "federal_agi": federal_agi, "income": income, "sources": sources, "capital_gains_short": capital_gains_short, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def delaware_and_utah(self, state_agi=75000, capital_gains_short=0, capital_gains_long=0, filing_status='single', federal_agi=75000, income=100000, dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Delaware:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        federal_agi : int, optional
            Adjusted Gross Income (AGI) for federal taxes. Default is 0. Sources - https://www.irs.gov/e-file-providers/definition-of-adjusted-gross-income, https://www.irs.gov/newsroom/taxpayers-should-know-the-difference-between-standard-and-itemized-deductions
        income : int, optional
            Gross income. Default is 0.
        
        Utah:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Delaware:
        - https://revenue.delaware.gov/software-developer/tax-rate-changes/
          https://revenuefiles.delaware.gov/2022/PIT-RES_TY22_2022-01_PaperInteractive.pdf
          https://revenuefiles.delaware.gov/2022/PIT-RES_TY22_2022-02_Instructions.pdf
          https://revenue.delaware.gov/frequently-asked-questions/personal-income-tax-faqs/
          https://www.newcastlede.gov/DocumentCenter/View/20147/City-of-Wilmington-Wage-Tax-Information-PDF
        
        Utah:
        -         
        - https://incometax.utah.gov/paying/tax-rates
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        federal_agi = self.federal_agi if hasattr(self, 'federal_agi') else federal_agi
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Delaware', 'state_occupation': 'Utah', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "federal_agi": federal_agi, "income": income, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def delaware_and_vermont(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, capital_gains_long=0, filing_status='single', federal_agi=75000, income=100000, dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Delaware:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        federal_agi : int, optional
            Adjusted Gross Income (AGI) for federal taxes. Default is 0. Sources - https://www.irs.gov/e-file-providers/definition-of-adjusted-gross-income, https://www.irs.gov/newsroom/taxpayers-should-know-the-difference-between-standard-and-itemized-deductions
        income : int, optional
            Gross income. Default is 0.
        
        Vermont:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Delaware:
        - https://revenue.delaware.gov/software-developer/tax-rate-changes/
          https://revenuefiles.delaware.gov/2022/PIT-RES_TY22_2022-01_PaperInteractive.pdf
          https://revenuefiles.delaware.gov/2022/PIT-RES_TY22_2022-02_Instructions.pdf
          https://revenue.delaware.gov/frequently-asked-questions/personal-income-tax-faqs/
          https://www.newcastlede.gov/DocumentCenter/View/20147/City-of-Wilmington-Wage-Tax-Information-PDF
        
        Vermont:
        -         
        - https://tax.vermont.gov/sites/tax/files/documents/GB-1210-2023.pdf
          https://tax.vermont.gov/sites/tax/files/documents/Income%20Booklet-2022.pdf
          https://tax.vermont.gov/sites/tax/files/documents/Income%20Booklet-2022.pdf
          https://tax.vermont.gov/sites/tax/files/documents/Income%20Booklet-2022.pdf
          https://tax.vermont.gov/individuals/personal-income-tax/tax-credits
          https://tax.vermont.gov/individuals/personal-income-tax/taxable-income
          https://tax.vermont.gov/individuals/personal-income-tax/local-option-tax
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        federal_agi = self.federal_agi if hasattr(self, 'federal_agi') else federal_agi
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Delaware', 'state_occupation': 'Vermont', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "federal_agi": federal_agi, "income": income, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def delaware_and_virginia(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, capital_gains_long=0, filing_status='single', federal_agi=75000, income=100000, dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Delaware:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        federal_agi : int, optional
            Adjusted Gross Income (AGI) for federal taxes. Default is 0. Sources - https://www.irs.gov/e-file-providers/definition-of-adjusted-gross-income, https://www.irs.gov/newsroom/taxpayers-should-know-the-difference-between-standard-and-itemized-deductions
        income : int, optional
            Gross income. Default is 0.
        
        Virginia:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Delaware:
        - https://revenue.delaware.gov/software-developer/tax-rate-changes/
          https://revenuefiles.delaware.gov/2022/PIT-RES_TY22_2022-01_PaperInteractive.pdf
          https://revenuefiles.delaware.gov/2022/PIT-RES_TY22_2022-02_Instructions.pdf
          https://revenue.delaware.gov/frequently-asked-questions/personal-income-tax-faqs/
          https://www.newcastlede.gov/DocumentCenter/View/20147/City-of-Wilmington-Wage-Tax-Information-PDF
        
        Virginia:
        -         
        - https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/reciprocity
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        federal_agi = self.federal_agi if hasattr(self, 'federal_agi') else federal_agi
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Delaware', 'state_occupation': 'Virginia', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "federal_agi": federal_agi, "income": income, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def delaware_and_washington(self, capital_gains_short=0, capital_gains_long=0, federal_agi=75000, income=100000, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Delaware:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        federal_agi : int, optional
            Adjusted Gross Income (AGI) for federal taxes. Default is 0. Sources - https://www.irs.gov/e-file-providers/definition-of-adjusted-gross-income, https://www.irs.gov/newsroom/taxpayers-should-know-the-difference-between-standard-and-itemized-deductions
        income : int, optional
            Gross income. Default is 0.
        
        Washington:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Delaware:
        - https://revenue.delaware.gov/software-developer/tax-rate-changes/
          https://revenuefiles.delaware.gov/2022/PIT-RES_TY22_2022-01_PaperInteractive.pdf
          https://revenuefiles.delaware.gov/2022/PIT-RES_TY22_2022-02_Instructions.pdf
          https://revenue.delaware.gov/frequently-asked-questions/personal-income-tax-faqs/
          https://www.newcastlede.gov/DocumentCenter/View/20147/City-of-Wilmington-Wage-Tax-Information-PDF
        
        Washington:
        -         
        - https://dor.wa.gov/taxes-rates/income-tax
          https://dor.wa.gov/taxes-rates/other-taxes/capital-gains-tax
          https://dor.wa.gov/taxes-rates/income-tax
        """
        
        # Check if the parameters are set upstream in the instance
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        federal_agi = self.federal_agi if hasattr(self, 'federal_agi') else federal_agi
        income = self.income if hasattr(self, 'income') else income
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Delaware', 'state_occupation': 'Washington', "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "federal_agi": federal_agi, "income": income, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def delaware_and_west_virginia(self, pay_periods=24, capital_gains_short=0, county_occupation='Other', dependents=0, sources=False, state_agi=75000, capital_gains_long=0, filing_status='single', federal_agi=75000, income=100000, credits_state=0):
        """
        --- Parameters ----
        
        Delaware:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        federal_agi : int, optional
            Adjusted Gross Income (AGI) for federal taxes. Default is 0. Sources - https://www.irs.gov/e-file-providers/definition-of-adjusted-gross-income, https://www.irs.gov/newsroom/taxpayers-should-know-the-difference-between-standard-and-itemized-deductions
        income : int, optional
            Gross income. Default is 0.
        
        West Virginia:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_occupation : str, optional
            County of occupation. Default is Other.
        pay_periods : int, optional
            Number of pay periods. Options are 13, 24, 26, 52. Default is 24.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Delaware:
        - https://revenue.delaware.gov/software-developer/tax-rate-changes/
          https://revenuefiles.delaware.gov/2022/PIT-RES_TY22_2022-01_PaperInteractive.pdf
          https://revenuefiles.delaware.gov/2022/PIT-RES_TY22_2022-02_Instructions.pdf
          https://revenue.delaware.gov/frequently-asked-questions/personal-income-tax-faqs/
          https://www.newcastlede.gov/DocumentCenter/View/20147/City-of-Wilmington-Wage-Tax-Information-PDF
        
        West Virginia:
        -         
        - https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://tax.wv.gov/Documents/TSD/tsd438.pdf
          https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2011/Taxes-11-35att.html
        """
        
        # Check if the parameters are set upstream in the instance
        pay_periods = self.pay_periods if hasattr(self, 'pay_periods') else pay_periods
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        county_occupation = self.county_occupation if hasattr(self, 'county_occupation') else county_occupation
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        federal_agi = self.federal_agi if hasattr(self, 'federal_agi') else federal_agi
        income = self.income if hasattr(self, 'income') else income
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Delaware', 'state_occupation': 'WestVirginia', "pay_periods": pay_periods, "capital_gains_short": capital_gains_short, "county_occupation": county_occupation, "dependents": dependents, "sources": sources, "state_agi": state_agi, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "federal_agi": federal_agi, "income": income, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def delaware_and_wisconsin(self, state_agi=75000, capital_gains_short=0, capital_gains_long=0, filing_status='single', federal_agi=75000, income=100000, dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Delaware:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        federal_agi : int, optional
            Adjusted Gross Income (AGI) for federal taxes. Default is 0. Sources - https://www.irs.gov/e-file-providers/definition-of-adjusted-gross-income, https://www.irs.gov/newsroom/taxpayers-should-know-the-difference-between-standard-and-itemized-deductions
        income : int, optional
            Gross income. Default is 0.
        
        Wisconsin:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Delaware:
        - https://revenue.delaware.gov/software-developer/tax-rate-changes/
          https://revenuefiles.delaware.gov/2022/PIT-RES_TY22_2022-01_PaperInteractive.pdf
          https://revenuefiles.delaware.gov/2022/PIT-RES_TY22_2022-02_Instructions.pdf
          https://revenue.delaware.gov/frequently-asked-questions/personal-income-tax-faqs/
          https://www.newcastlede.gov/DocumentCenter/View/20147/City-of-Wilmington-Wage-Tax-Information-PDF
        
        Wisconsin:
        -         
        - https://www.revenue.wi.gov/TaxForms2023/2023-Form1-ES-Inst.pdf
          https://www.revenue.wi.gov/TaxForms2023/2023-Form1-ES-Inst.pdf
          https://www.revenue.wi.gov/TaxForms2023/2023-Form1-ES-Inst.pdf
          https://www.revenue.wi.gov/Pages/FAQS/pcs-work.aspx
          https://www.revenue.wi.gov/Dor%20Publications/pb103.pdf
          https://www.revenue.wi.gov/Pages/FAQS/pcs-work.aspx
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        federal_agi = self.federal_agi if hasattr(self, 'federal_agi') else federal_agi
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Delaware', 'state_occupation': 'Wisconsin', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "federal_agi": federal_agi, "income": income, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def delaware_and_wyoming(self, capital_gains_long=0, federal_agi=75000, income=100000, sources=False, capital_gains_short=0, credits_state=0):
        """
        --- Parameters ----
        
        Delaware:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        federal_agi : int, optional
            Adjusted Gross Income (AGI) for federal taxes. Default is 0. Sources - https://www.irs.gov/e-file-providers/definition-of-adjusted-gross-income, https://www.irs.gov/newsroom/taxpayers-should-know-the-difference-between-standard-and-itemized-deductions
        income : int, optional
            Gross income. Default is 0.
        
        Wyoming:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Delaware:
        - https://revenue.delaware.gov/software-developer/tax-rate-changes/
          https://revenuefiles.delaware.gov/2022/PIT-RES_TY22_2022-01_PaperInteractive.pdf
          https://revenuefiles.delaware.gov/2022/PIT-RES_TY22_2022-02_Instructions.pdf
          https://revenue.delaware.gov/frequently-asked-questions/personal-income-tax-faqs/
          https://www.newcastlede.gov/DocumentCenter/View/20147/City-of-Wilmington-Wage-Tax-Information-PDF
        
        Wyoming:
        -         
        - https://revenue.wyo.gov/
        """
        
        # Check if the parameters are set upstream in the instance
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        federal_agi = self.federal_agi if hasattr(self, 'federal_agi') else federal_agi
        income = self.income if hasattr(self, 'income') else income
        sources = self.sources if hasattr(self, 'sources') else sources
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Delaware', 'state_occupation': 'Wyoming', "capital_gains_long": capital_gains_long, "federal_agi": federal_agi, "income": income, "sources": sources, "capital_gains_short": capital_gains_short, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def florida_and_alabama(self, county_residence='Other', state_agi=75000, capital_gains_short=0, capital_gains_long=0, filing_status='single', income=100000, dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Florida:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        Alabama:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Florida:
        - https://floridarevenue.com/taxes/taxesfees/Pages/tax_interest_rates.aspx
        
        Alabama:
        -         
        - https://www.revenue.alabama.gov/ultraviewer/viewer/basic_viewer/index.html?form=2023/02/22f40.pdf
          https://www.revenue.alabama.gov/faqs/what-is-alabamas-individual-income-tax-rate/
          https://www.revenue.alabama.gov/ultraviewer/viewer/basic_viewer/index.html?form=2023/02/22f40.pdf
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-24.htm?taxmap=true
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-24.htm?taxmap=true
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-24.htm?taxmap=true
          https://www.revenue.alabama.gov/individual-corporate/income-to-be-reported-on-the-alabama-income-tax-return/
          https://www.almonline.org/TaxRates.aspx#OccupationalTax
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Florida', 'state_occupation': 'Alabama', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def florida_and_alaska(self, sources=False):
        """
        --- Parameters ----
        
        Florida:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        Alaska:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Florida:
        - https://floridarevenue.com/taxes/taxesfees/Pages/tax_interest_rates.aspx
        
        Alaska:
        -         
        - https://www.commerce.alaska.gov/web/dcra/officeofthestateassessor/alaskataxfacts.aspx
        """
        
        # Check if the parameters are set upstream in the instance
        sources = self.sources if hasattr(self, 'sources') else sources
        
        # State-specific parameters
        data = {
            'state_residence': 'Florida', 'state_occupation': 'Alaska', "sources": sources
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def florida_and_arizona(self, state_agi=75000, capital_gains_short=0, capital_gains_long=0, filing_status='single', income=100000, dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Florida:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        Arizona:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        credits_state : int, optional
            Credits for state taxes. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Florida:
        - https://floridarevenue.com/taxes/taxesfees/Pages/tax_interest_rates.aspx
        
        Arizona:
        -         
        - https://azdor.gov/forms/individual/form-140-x-y-tables
          https://azdor.gov/forms/individual/form-140-x-y-tables
          https://azdor.gov/forms/individual/form-140-x-y-tables
          https://azdor.gov/businesses-arizona/withholding-tax/withholding-exceptions
          "https://azdor.gov/individuals/income-tax-filing-assistance/identifying-other-taxable-income#:~:text=In%20regards%20to%20capital%20gains
         the%20individuals%20regular%20tax%20rate.
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Florida', 'state_occupation': 'Arizona', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def florida_and_arkansas(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, capital_gains_long=0, filing_status='single', income=100000, dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Florida:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        Arkansas:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Florida:
        - https://floridarevenue.com/taxes/taxesfees/Pages/tax_interest_rates.aspx
        
        Arkansas:
        -         
        - https://www.arkansasedc.com/why-arkansas/business-climate/tax-structure/personal-income-tax
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_and_AR1000NR_Instructions.pdf
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_FullYearResidentIndividualIncomeTaxReturn.pdf
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_and_AR1000NR_Instructions.pdf
          https://www.arkansasedc.com/why-arkansas/business-climate/tax-structure/capital-gains-tax-reduction
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_and_AR1000NR_Instructions.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Florida', 'state_occupation': 'Arkansas', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def florida_and_california(self, state_agi=75000, capital_gains_short=0, capital_gains_long=0, filing_status='single', income=100000, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Florida:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        California:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Florida:
        - https://floridarevenue.com/taxes/taxesfees/Pages/tax_interest_rates.aspx
        
        California:
        -         
        - https://www.ftb.ca.gov/forms/2022/2022-540-tax-rate-schedules.pdf
          https://www.ftb.ca.gov/file/personal/deductions/index.html
          https://www.ftb.ca.gov/about-ftb/newsroom/caleitc/eligibility-and-credit-information.html
          https://www.ftb.ca.gov/file/personal/income-types/capital-gains-and-losses.html#:~:text=California%20does%20not%20have%20a
         are%20taxed%20as%20ordinary%20income
          https://www.ftb.ca.gov/file/personal/
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Florida', 'state_occupation': 'California', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def florida_and_colorado(self, county_residence='Other', state_agi=75000, capital_gains_short=0, capital_gains_long=0, income=100000, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Florida:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        Colorado:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Florida:
        - https://floridarevenue.com/taxes/taxesfees/Pages/tax_interest_rates.aspx
        
        Colorado:
        -         
        - https://leg.colorado.gov/agencies/legislative-council-staff/individual-income-tax%C2%A0
          https://leg.colorado.gov/agencies/legislative-council-staff/individual-income-tax%C2%A0
          https://tax.colorado.gov/sites/tax/files/documents/ITT_Capital_Gain_Subtraction_Feb_2022.pdf
          https://www.auroragov.org/business_services/taxes/occupational_privilege_tax
          https://www.denvergov.org/Government/Agencies-Departments-Offices/Agencies-Departments-Offices-Directory/Department-of-Finance/Our-Divisions/Treasury/Business-Tax-Information#:~:text=Occupational%20privilege%20tax
         -%28show%20below%29&text=Each%20taxable%20employee%20is%20liable
         month%20for%20each%20taxable%20employee
          https://www.glendale.co.us/355/Occupational-Privilege-Tax
          https://www.greenwoodvillage.com/1220/Occupational-Privilege-Tax-OPT
          https://ci.sheridan.co.us/288/Occupational-Privilege-Tax
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        income = self.income if hasattr(self, 'income') else income
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Florida', 'state_occupation': 'Colorado', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "income": income, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def florida_and_connecticut(self, state_agi=75000, capital_gains_short=0, capital_gains_long=0, filing_status='single', income=100000, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Florida:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        Connecticut:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Florida:
        - https://floridarevenue.com/taxes/taxesfees/Pages/tax_interest_rates.aspx
        
        Connecticut:
        -         
        - https://portal.ct.gov/-/media/DRS/Forms/2022/Income/2022-CT-1040-Instructions_1222.pdf
          https://www.cga.ct.gov/2022/rpt/pdf/2022-R-0108.pdf
          https://www.cga.ct.gov/current/pub/chap_229.htm#sec_12-703
          https://portal.ct.gov/-/media/DRS/Forms/2022/Income/2022-CT-1040-Instructions_1222.pdf
          "https://portal.ct.gov/DRS/Publications/TSSNs/TSSN-29#:~:text=An%20individuals%20net%20capital%20gains
         portion%20of%20Social%20Security%20benefits"
          "https://portal.ct.gov/DRS/Publications/TSSNs/TSSN-29#:~:text=An%20individuals%20net%20capital%20gains
         portion%20of%20Social%20Security%20benefits"
          https://portal.ct.gov/DRS/Individuals/Resident-Income-Tax/Tax-Information
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Florida', 'state_occupation': 'Connecticut', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def florida_and_delaware(self, capital_gains_long=0, federal_agi=75000, income=100000, sources=False, capital_gains_short=0, credits_state=0):
        """
        --- Parameters ----
        
        Florida:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        Delaware:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        federal_agi : int, optional
            Adjusted Gross Income (AGI) for federal taxes. Default is 0. Sources - https://www.irs.gov/e-file-providers/definition-of-adjusted-gross-income, https://www.irs.gov/newsroom/taxpayers-should-know-the-difference-between-standard-and-itemized-deductions
        income : int, optional
            Gross income. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Florida:
        - https://floridarevenue.com/taxes/taxesfees/Pages/tax_interest_rates.aspx
        
        Delaware:
        -         
        - https://revenue.delaware.gov/software-developer/tax-rate-changes/
          https://revenuefiles.delaware.gov/2022/PIT-RES_TY22_2022-01_PaperInteractive.pdf
          https://revenuefiles.delaware.gov/2022/PIT-RES_TY22_2022-02_Instructions.pdf
          https://revenue.delaware.gov/frequently-asked-questions/personal-income-tax-faqs/
          https://www.newcastlede.gov/DocumentCenter/View/20147/City-of-Wilmington-Wage-Tax-Information-PDF
        """
        
        # Check if the parameters are set upstream in the instance
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        federal_agi = self.federal_agi if hasattr(self, 'federal_agi') else federal_agi
        income = self.income if hasattr(self, 'income') else income
        sources = self.sources if hasattr(self, 'sources') else sources
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Florida', 'state_occupation': 'Delaware', "capital_gains_long": capital_gains_long, "federal_agi": federal_agi, "income": income, "sources": sources, "capital_gains_short": capital_gains_short, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def florida_and_georgia(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, capital_gains_long=0, filing_status='single', income=100000, dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Florida:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        Georgia:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Florida:
        - https://floridarevenue.com/taxes/taxesfees/Pages/tax_interest_rates.aspx
        
        Georgia:
        -         
        - https://dor.georgia.gov/tax-tables-georgia-tax-rate-schedule
          https://apps.dor.ga.gov/FillableForms/PDFViewer/Index?form=2022GA500
          https://apps.dor.ga.gov/FillableForms/PDFViewer/Index?form=2022GA500
          https://apps.dor.ga.gov/FillableForms/PDFViewer/Index?form=2022GA500
          https://dor.georgia.gov/years-individual-income-tax-forms
          https://dor.georgia.gov/years-individual-income-tax-forms
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Florida', 'state_occupation': 'Georgia', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def florida_and_hawaii(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, capital_gains_long=0, filing_status='single', income=100000, dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Florida:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        Hawaii:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Florida:
        - https://floridarevenue.com/taxes/taxesfees/Pages/tax_interest_rates.aspx
        
        Hawaii:
        -         
        - https://tax.hawaii.gov/forms/d_18table-on/d_18table-on_p13/
          https://files.hawaii.gov/tax/forms/2018/18table-on.pdf
          https://files.hawaii.gov/tax/news/pubs/22outline.pdf
          https://files.hawaii.gov/tax/news/pubs/FastTaxRefGuide_2017.pdf
          https://www.tfhawaii.org/wordpress/state-tax-resources/mini-tax-guides/individual-income-tax-chapter-235/#:~:text=Long%20term%20capital%20gains%20are%20taxed%20at%20a%20maximum%20of%207.25%25
          https://www.grassrootinstitute.org/2023/02/hb337-increasing-tax-on-capital-gains-a-bad-risk/
          https://files.hawaii.gov/tax/news/pubs/22outline.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Florida', 'state_occupation': 'Hawaii', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def florida_and_idaho(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, capital_gains_long=0, filing_status='single', income=100000, dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Florida:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        Idaho:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Florida:
        - https://floridarevenue.com/taxes/taxesfees/Pages/tax_interest_rates.aspx
        
        Idaho:
        -         
        - https://tax.idaho.gov/taxes/income-tax/individual-income/individual-income-tax-rate-schedule/
          https://tax.idaho.gov/wp-content/uploads/forms/EIN00046/EIN00046_03-01-2023.pdf
          https://tax.idaho.gov/taxes/income-tax/individual-income/idaho-source-income/
          https://tax.idaho.gov/taxes/income-tax/individual-income/idaho-source-income/
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Florida', 'state_occupation': 'Idaho', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def florida_and_illinois(self, state_agi=75000, capital_gains_short=0, capital_gains_long=0, filing_status='single', income=100000, dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Florida:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        Illinois:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Florida:
        - https://floridarevenue.com/taxes/taxesfees/Pages/tax_interest_rates.aspx
        
        Illinois:
        -         
        - https://tax.illinois.gov/research/taxrates/income.html#IndividualIncome
          https://tax.illinois.gov/content/dam/soi/en/web/tax/forms/incometax/documents/currentyear/individual/il-1040-instr.pdf
          https://tax.illinois.gov/questionsandanswers/12.html
          https://tax.illinois.gov/content/dam/soi/en/web/tax/forms/incometax/documents/currentyear/individual/il-1040-instr.pdf
          https://tax.illinois.gov/research/taxrates/personalproperty.html
          https://tax.illinois.gov/localgovernments/income.html
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Florida', 'state_occupation': 'Illinois', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def florida_and_indiana(self, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_short=0, capital_gains_long=0, filing_status='single', income=100000, dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Florida:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        Indiana:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Florida:
        - https://floridarevenue.com/taxes/taxesfees/Pages/tax_interest_rates.aspx
        
        Indiana:
        -         
        - https://www.in.gov/dor/tax-forms/2022-individual-income-tax-forms/
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-32.htm
          https://www.in.gov/dor/files/ib28.pdf
          https://www.in.gov/dor/tax-forms/2022-individual-income-tax-forms/
          https://www.in.gov/dor/files/2022-county-tax-rates-and-codes.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Florida', 'state_occupation': 'Indiana', "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def florida_and_iowa(self, county_residence='Other', state_agi=75000, capital_gains_short=0, capital_gains_long=0, filing_status='single', income=100000, dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Florida:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        Iowa:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Florida:
        - https://floridarevenue.com/taxes/taxesfees/Pages/tax_interest_rates.aspx
        
        Iowa:
        -         
        - https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
          https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
          https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
          https://tax.iowa.gov/iowa-illinois-reciprocal-agreement
          https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Florida', 'state_occupation': 'Iowa', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def florida_and_kansas(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, capital_gains_long=0, filing_status='single', income=100000, dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Florida:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        Kansas:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Florida:
        - https://floridarevenue.com/taxes/taxesfees/Pages/tax_interest_rates.aspx
        
        Kansas:
        -         
        - https://www.ksrevenue.gov/taxrates.html
          https://www.ksrevenue.gov/incomebook22.html#0
          https://www.ksrevenue.gov/incomebook22.html#0
          https://www.ksrevenue.gov/incomebook22.html#0
          https://www.ksrevenue.gov/incomebook22.html
          https://www.ksrevenue.gov/incomebook22.html
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Florida', 'state_occupation': 'Kansas', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def florida_and_kentucky(self, county_residence='Other', state_agi=75000, capital_gains_short=0, capital_gains_long=0, filing_status='single', income=100000, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Florida:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        Kentucky:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Florida:
        - https://floridarevenue.com/taxes/taxesfees/Pages/tax_interest_rates.aspx
        
        Kentucky:
        -         
        - https://revenue.ky.gov/News/Pages/Dor-Announces-Updates-to-Individual-Income-Tax-for-2023-Tax-Year.aspx
          https://revenue.ky.gov/Individual/Individual-Income-Tax/Pages/default.aspx
          https://revenue.ky.gov/News/Pages/Dor-Announces-Updates-to-Individual-Income-Tax-for-2023-Tax-Year.aspx
          https://revenue.ky.gov/Individual/Individual-Income-Tax/Pages/default.aspx
          https://revenue.ky.gov/Dor%20Training%20Materials/103%20KAR%2017.140.%20Individual%20income%20tax%20-%20reciprocity%20-%20nonresidents.pdf
          https://revenue.ky.gov/Individual/Individual-Income-Tax/Pages/default.aspx
          https://nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2021/Taxes-21-32.htm?taxmap=true
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Florida', 'state_occupation': 'Kentucky', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def florida_and_louisiana(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, capital_gains_long=0, filing_status='single', income=100000, dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Florida:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        Louisiana:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Florida:
        - https://floridarevenue.com/taxes/taxesfees/Pages/tax_interest_rates.aspx
        
        Louisiana:
        -         
        - https://revenue.louisiana.gov/individualincometax
          https://revenue.louisiana.gov/individualincometax
          https://www.revenue.louisiana.gov/taxforms/6935(11_02)F.pdf
          https://revenue.louisiana.gov/individualincometax
          https://revenue.louisiana.gov/individualincometax
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Florida', 'state_occupation': 'Louisiana', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def florida_and_maine(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, capital_gains_long=0, filing_status='single', income=100000, dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Florida:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        Maine:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Florida:
        - https://floridarevenue.com/taxes/taxesfees/Pages/tax_interest_rates.aspx
        
        Maine:
        -         
        - https://www.maine.gov/revenue/tax-return-forms/individual-income-tax-2022
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/ind_tax_rate_sched_2022.pdf
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/22_1040me_book_gen_instruc_revisedFeb23.pdf
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/22_1040me_book_gen_instruc_revisedFeb23.pdf
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/22_1040me_book_gen_instruc_revisedFeb23.pdf
          https://www.maine.gov/revenue/tax-return-forms/individual-income-tax-2022
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Florida', 'state_occupation': 'Maine', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def florida_and_maryland(self, county_residence='Other', state_agi=75000, capital_gains_short=0, capital_gains_long=0, filing_status='single', income=100000, dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Florida:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        Maryland:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Florida:
        - https://floridarevenue.com/taxes/taxesfees/Pages/tax_interest_rates.aspx
        
        Maryland:
        -         
        - https://www.marylandtaxes.gov/individual/income/tax-info/tax-rates.php
          https://www.marylandtaxes.gov/forms/22_forms/Resident_Booklet.pdf
          https://www.marylandtaxes.gov/forms/Tax_Publications/Administrative_Releases/Income_and_Estate_Tax_Releases/ar_it37.pdf
          https://www.marylandtaxes.gov/individual/index.php
          https://www.marylandtaxes.gov/individual/income/tax-info/tax-rates.php
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Florida', 'state_occupation': 'Maryland', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def florida_and_massachusetts(self, state_agi=75000, capital_gains_short=0, capital_gains_long=0, filing_status='single', income=100000, dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Florida:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        Massachusetts:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Florida:
        - https://floridarevenue.com/taxes/taxesfees/Pages/tax_interest_rates.aspx
        
        Massachusetts:
        -         
        - https://www.mass.gov/service-details/massachusetts-tax-rates
          https://www.mass.gov/service-details/view-massachusetts-personal-income-tax-exemptions
          https://www.mass.gov/service-details/view-massachusetts-personal-income-tax-exemptions
          https://www.mass.gov/service-details/view-child-and-dependent-related-credits
          https://www.mass.gov/service-details/massachusetts-tax-rates
          https://www.mass.gov/personal-income-tax
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Florida', 'state_occupation': 'Massachusetts', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def florida_and_michigan(self, county_residence='Other', state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', county_occupation='Other', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Florida:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        Michigan:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        county_occupation : str, optional
            County of occupation. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Florida:
        - https://floridarevenue.com/taxes/taxesfees/Pages/tax_interest_rates.aspx
        
        Michigan:
        -         
        - https://www.michigan.gov/taxes/iit/new-developments-for-tax-year-2022
          "https://www.michigan.gov/taxes/-/media/Project/Websites/taxes/Forms/2023/2023-SUW/446_Withholding-Guide_2023.pdf?rev=177ca3d8cb034392bf84e317ed78c320&hash=19C5F42F43A89A91863D229349F48883#:~:text=The%20exemption%20amount%20is%20%245%2C400
         employees%20Michigan%20income%20tax%20return"
          "https://www.michigan.gov/taxes/-/media/Project/Websites/taxes/Forms/2023/2023-SUW/446_Withholding-Guide_2023.pdf?rev=177ca3d8cb034392bf84e317ed78c320&hash=19C5F42F43A89A91863D229349F48883#:~:text=The%20exemption%20amount%20is%20%245%2C400
         employees%20Michigan%20income%20tax%20return"
          https://www.michigan.gov/taxes/questions/iit/accordion/residency/are-my-wages-earned-in-another-state-taxable-in-michigan-if-i-am-a-michigan-resident-1
          https://www.michigan.gov/taxes/-/media/Project/Websites/taxes/Forms/2022/2022-IIT-Forms/BOOK_MI-1040.pdf?rev=4524f1cbbc0243b194ab05a6680be75b&hash=40DFE71E651F7C1D65C31F6CD03D944F
          https://www.michigan.gov/taxes/questions/iit/accordion/general/what-cities-impose-an-income-tax
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        county_occupation = self.county_occupation if hasattr(self, 'county_occupation') else county_occupation
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Florida', 'state_occupation': 'Michigan', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "county_occupation": county_occupation, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def florida_and_minnesota(self, state_agi=75000, capital_gains_short=0, capital_gains_long=0, filing_status='single', income=100000, dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Florida:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        Minnesota:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Florida:
        - https://floridarevenue.com/taxes/taxesfees/Pages/tax_interest_rates.aspx
        
        Minnesota:
        -         
        - https://www.revenue.state.mn.us/minnesota-income-tax-rates-and-brackets
          https://www.revenue.state.mn.us/child-and-dependent-care-credit
          https://www.revenue.state.mn.us/dependent-exemptions
          https://www.revenue.state.mn.us/minnesota-standard-deduction
          https://mn.gov/mmb/assets/mwr_form_tcm1059-128581.pdf
          https://www.revenue.state.mn.us/individual-income-tax
          https://www.revenue.state.mn.us/individual-income-tax
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Florida', 'state_occupation': 'Minnesota', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def florida_and_mississippi(self, state_agi=75000, capital_gains_short=0, capital_gains_long=0, filing_status='single', income=100000, dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Florida:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        Mississippi:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Florida:
        - https://floridarevenue.com/taxes/taxesfees/Pages/tax_interest_rates.aspx
        
        Mississippi:
        -         
        - https://www.dor.ms.gov/individual/tax-rates
          https://www.dor.ms.gov/individual/tax-rates
          https://www.dor.ms.gov/individual/tax-rates
          https://www.dor.ms.gov/individual
          https://www.dor.ms.gov/individual
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Florida', 'state_occupation': 'Mississippi', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def florida_and_missouri(self, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_short=0, capital_gains_long=0, filing_status='single', income=100000, dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Florida:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        Missouri:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Florida:
        - https://floridarevenue.com/taxes/taxesfees/Pages/tax_interest_rates.aspx
        
        Missouri:
        -         
        - https://dor.mo.gov/taxation/individual/tax-types/income/year-changes/
          https://dor.mo.gov/taxation/individual/tax-types/income/year-changes/
          https://dor.mo.gov/taxation/individual/tax-types/income/year-changes/
          https://www.kcmo.gov/city-hall/departments/finance/earnings-tax
          https://www.stlouis-mo.gov/government/departments/collector/earnings-tax/individual-earnings-tax-info.cfm
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Florida', 'state_occupation': 'Missouri', "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def florida_and_montana(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, capital_gains_long=0, filing_status='single', income=100000, dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Florida:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        Montana:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Florida:
        - https://floridarevenue.com/taxes/taxesfees/Pages/tax_interest_rates.aspx
        
        Montana:
        -         
        - https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/2022/12/montana-employees-withholding-allowance-and-exemption-certificate-form-mw-4.pdf
          https://mtrevenue.gov/2021/11/29/simplification-of-montana-income-taxation/?utm_source=rss&utm_medium=rss&utm_campaign=simplification-of-montana-income-taxation
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Florida', 'state_occupation': 'Montana', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def florida_and_nebraska(self, state_agi=75000, capital_gains_short=0, capital_gains_long=0, filing_status='single', income=100000, dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Florida:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        Nebraska:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Florida:
        - https://floridarevenue.com/taxes/taxesfees/Pages/tax_interest_rates.aspx
        
        Nebraska:
        -         
        - https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2023_Circular_EN_8-429r2023_Whole_Book_11.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Florida', 'state_occupation': 'Nebraska', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def florida_and_nevada(self, sources=False):
        """
        --- Parameters ----
        
        Florida:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        Nevada:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Florida:
        - https://floridarevenue.com/taxes/taxesfees/Pages/tax_interest_rates.aspx
        
        Nevada:
        -         
        - https://www.nvsos.gov/sos/businesses/the-nevada-advantage
        """
        
        # Check if the parameters are set upstream in the instance
        sources = self.sources if hasattr(self, 'sources') else sources
        
        # State-specific parameters
        data = {
            'state_residence': 'Florida', 'state_occupation': 'Nevada', "sources": sources
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def florida_and_new_hampshire(self, sources=False):
        """
        --- Parameters ----
        
        Florida:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        New Hampshire:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Florida:
        - https://floridarevenue.com/taxes/taxesfees/Pages/tax_interest_rates.aspx
        
        New Hampshire:
        -         
        - https://www.revenue.nh.gov/assistance/tax-overview.htm
        """
        
        # Check if the parameters are set upstream in the instance
        sources = self.sources if hasattr(self, 'sources') else sources
        
        # State-specific parameters
        data = {
            'state_residence': 'Florida', 'state_occupation': 'NewHampshire', "sources": sources
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def florida_and_new_jersey(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, capital_gains_long=0, filing_status='single', income=100000, dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Florida:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        New Jersey:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Florida:
        - https://floridarevenue.com/taxes/taxesfees/Pages/tax_interest_rates.aspx
        
        New Jersey:
        -         
        - https://www.state.nj.us/treasury/taxation/taxtables.shtml
          https://www.state.nj.us/treasury/taxation/njit13.shtml
          https://www.nj.gov/treasury/taxation/njit25.shtml
          https://www.state.nj.us/treasury/taxation/individuals/freqqiti.shtml
          https://www.state.nj.us/treasury/taxation/localtax.shtml
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Florida', 'state_occupation': 'NewJersey', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def florida_and_new_mexico(self, state_agi=75000, capital_gains_short=0, capital_gains_long=0, filing_status='single', income=100000, dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Florida:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        New Mexico:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Florida:
        - https://floridarevenue.com/taxes/taxesfees/Pages/tax_interest_rates.aspx
        
        New Mexico:
        -         
        - https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/fdf3c548-8aba-4b9c-9eb4-bb564c716015/FYI-104.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/0bf943fd-652e-4400-bb1b-cd8397eb5a95/2022pit-adj.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/2f1a6781-9534-4436-b427-1557f9592099/2022pit-adj-ins.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/0bf943fd-652e-4400-bb1b-cd8397eb5a95/2022pit-adj.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/1afc56af-ea90-4d48-82e5-1f9aeb43255a/PITbook2022.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/fdf3c548-8aba-4b9c-9eb4-bb564c716015/FYI-104.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Florida', 'state_occupation': 'NewMexico', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def florida_and_new_york(self, county_residence='Other', state_agi=75000, capital_gains_short=0, capital_gains_long=0, filing_status='single', income=100000, dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Florida:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        New York:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Florida:
        - https://floridarevenue.com/taxes/taxesfees/Pages/tax_interest_rates.aspx
        
        New York:
        -         
        - https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nys-tax-rate-schedule
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nys-tax-rate-schedule
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nys-tax-rate-schedule
          https://www.tax.ny.gov/pit/
          https://www.tax.ny.gov/pit/file/nonresident-faqs.htm
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#yonkers-tax-rate-schedule
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nyc-tax-rate-schedule
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Florida', 'state_occupation': 'NewYork', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def florida_and_north_carolina(self, state_agi=75000, capital_gains_short=0, capital_gains_long=0, filing_status='single', income=100000, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Florida:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        North Carolina:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Florida:
        - https://floridarevenue.com/taxes/taxesfees/Pages/tax_interest_rates.aspx
        
        North Carolina:
        -         
        - https://www.ncdor.gov/taxes-forms/tax-rate-schedules
          https://www.ncdor.gov/taxes-forms/individual-income-tax/north-carolina-standard-deduction-or-north-carolina-itemized-deductions
          https://www.ncdor.gov/taxes-forms/individual-income-tax
          https://www.ncdor.gov/taxes-forms/individual-income-tax
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Florida', 'state_occupation': 'NorthCarolina', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def florida_and_north_dakota(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, capital_gains_long=0, filing_status='single', income=100000, dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Florida:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        North Dakota:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Florida:
        - https://floridarevenue.com/taxes/taxesfees/Pages/tax_interest_rates.aspx
        
        North Dakota:
        -         
        - https://www.tax.nd.gov/forms
          https://www.tax.nd.gov/sites/www/files/documents/forms/individual/2022-iit/2022-individual-income-tax-booklet.pdf
          https://www.tax.nd.gov/sites/www/files/documents/forms/2022-individual-income-tax-booklet.pdf
          https://www.tax.nd.gov/tax-exemptions-credits/income-tax-exemptions-credits
          https://www.tax.nd.gov/sites/www/files/documents/forms/business/ndwrfillable.pdf
          https://www.tax.nd.gov/news/tax-legislative-changes/significant-changes-law/individual-income-tax-history
          https://www.tax.nd.gov/forms
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Florida', 'state_occupation': 'NorthDakota', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def florida_and_ohio(self, county_residence='Other', state_agi=75000, capital_gains_short=0, capital_gains_long=0, filing_status='single', income=100000, dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Florida:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        Ohio:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Florida:
        - https://floridarevenue.com/taxes/taxesfees/Pages/tax_interest_rates.aspx
        
        Ohio:
        -         
        - https://tax.ohio.gov/individual/resources/annual-tax-rates
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://tax.ohio.gov/static/forms/employer_withholding/generic/wth_it4nr.pdf
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://www.ritaohio.com/TaxRatesTable
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Florida', 'state_occupation': 'Ohio', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def florida_and_oklahoma(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, capital_gains_long=0, filing_status='single', income=100000, dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Florida:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        Oklahoma:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Florida:
        - https://floridarevenue.com/taxes/taxesfees/Pages/tax_interest_rates.aspx
        
        Oklahoma:
        -         
        - https://oklahoma.gov/content/dam/ok/en/tax/documents/resources/publications/businesses/withholding-tables/WHTables-2022.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Florida', 'state_occupation': 'Oklahoma', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def florida_and_oregon(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, capital_gains_long=0, filing_status='single', income=100000, dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Florida:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        Oregon:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Florida:
        - https://floridarevenue.com/taxes/taxesfees/Pages/tax_interest_rates.aspx
        
        Oregon:
        -         
        - https://www.oregon.gov/dor/programs/individuals/Pages/PIT.aspx
          https://www.oregon.gov/dor/programs/individuals/Documents/Full%20year%20resident
         %20Form%20or-40%20filers.pdf
          https://www.oregon.gov/dor/programs/individuals/Documents/Part-year%20and%20nonresident
         %20Form%20or-40-P%20and%20or-40-N%20filers.pdf.pdf
          https://www.oregon.gov/dor/forms/FormsPubs/form-or-40_101-040_2022.pdf
          https://www.oregon.gov/dor/forms/FormsPubs/form-or-40_101-040_2022.pdf
          https://www.oregon.gov/dor/forms/FormsPubs/form-or-40_101-040_2022.pdf
          https://www.portland.gov/revenue/personal-tax
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Florida', 'state_occupation': 'Oregon', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def florida_and_pennsylvania(self, county_residence='Other', state_agi=75000, county_occupation='Other', capital_gains_short=0, capital_gains_long=0, income=100000, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Florida:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        Pennsylvania:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        county_residence : str, optional
            County of residence. Default is Other.
        county_occupation : str, optional
            County of occupation. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Florida:
        - https://floridarevenue.com/taxes/taxesfees/Pages/tax_interest_rates.aspx
        
        Pennsylvania:
        -         
        - https://www.revenue.pa.gov/Tax%20Rates/Pages/default.aspx
          https://www.revenue.pa.gov/FormsandPublications/FormsforIndividuals/PIT/Documents/2022/2022_pa-40in.pdf
          https://www.revenue.pa.gov/FormsandPublications/FormsforBusinesses/EmployerWithholding/Documents/rev-419.pdf
          https://www.revenue.pa.gov/FormsandPublications/FormsforIndividuals/PIT/Documents/2022/2022_pa-40in.pdf
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-25.htm
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        county_occupation = self.county_occupation if hasattr(self, 'county_occupation') else county_occupation
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        income = self.income if hasattr(self, 'income') else income
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Florida', 'state_occupation': 'Pennsylvania', "county_residence": county_residence, "state_agi": state_agi, "county_occupation": county_occupation, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "income": income, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def florida_and_rhode_island(self, state_agi=75000, capital_gains_short=0, capital_gains_long=0, filing_status='single', income=100000, dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Florida:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        Rhode Island:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Florida:
        - https://floridarevenue.com/taxes/taxesfees/Pages/tax_interest_rates.aspx
        
        Rhode Island:
        -         
        - https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2022_1040WE_w_0.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Florida', 'state_occupation': 'RhodeIsland', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def florida_and_south_carolina(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, capital_gains_long=0, filing_status='single', income=100000, dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Florida:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        South Carolina:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Florida:
        - https://floridarevenue.com/taxes/taxesfees/Pages/tax_interest_rates.aspx
        
        South Carolina:
        -         
        - https://dor.sc.gov/resources-site/lawandpolicy/Advisory%20Opinions/IL22-15.pdf
          https://dor.sc.gov/forms-site/Forms/IITPacket_2022.pdf
          https://www.scstatehouse.gov/code/t12c006.php
          https://dor.sc.gov/resources-site/lawandpolicy/Advisory%20Opinions/IL22-15.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Florida', 'state_occupation': 'SouthCarolina', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def florida_and_south_dakota(self, sources=False):
        """
        --- Parameters ----
        
        Florida:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        South Dakota:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Florida:
        - https://floridarevenue.com/taxes/taxesfees/Pages/tax_interest_rates.aspx
        
        South Dakota:
        -         
        - https://dor.sd.gov/individuals/taxes/
        """
        
        # Check if the parameters are set upstream in the instance
        sources = self.sources if hasattr(self, 'sources') else sources
        
        # State-specific parameters
        data = {
            'state_residence': 'Florida', 'state_occupation': 'SouthDakota', "sources": sources
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def florida_and_tennessee(self, sources=False):
        """
        --- Parameters ----
        
        Florida:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        Tennessee:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Florida:
        - https://floridarevenue.com/taxes/taxesfees/Pages/tax_interest_rates.aspx
        
        Tennessee:
        -         
        - https://www.tn.gov/revenue/taxes.html
        """
        
        # Check if the parameters are set upstream in the instance
        sources = self.sources if hasattr(self, 'sources') else sources
        
        # State-specific parameters
        data = {
            'state_residence': 'Florida', 'state_occupation': 'Tennessee', "sources": sources
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def florida_and_texas(self, sources=False):
        """
        --- Parameters ----
        
        Florida:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        Texas:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Florida:
        - https://floridarevenue.com/taxes/taxesfees/Pages/tax_interest_rates.aspx
        
        Texas:
        -         
        - https://comptroller.texas.gov/taxes/
        """
        
        # Check if the parameters are set upstream in the instance
        sources = self.sources if hasattr(self, 'sources') else sources
        
        # State-specific parameters
        data = {
            'state_residence': 'Florida', 'state_occupation': 'Texas', "sources": sources
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def florida_and_utah(self, state_agi=75000, capital_gains_short=0, capital_gains_long=0, filing_status='single', income=100000, dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Florida:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        Utah:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Florida:
        - https://floridarevenue.com/taxes/taxesfees/Pages/tax_interest_rates.aspx
        
        Utah:
        -         
        - https://incometax.utah.gov/paying/tax-rates
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Florida', 'state_occupation': 'Utah', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def florida_and_vermont(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, capital_gains_long=0, filing_status='single', income=100000, dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Florida:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        Vermont:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Florida:
        - https://floridarevenue.com/taxes/taxesfees/Pages/tax_interest_rates.aspx
        
        Vermont:
        -         
        - https://tax.vermont.gov/sites/tax/files/documents/GB-1210-2023.pdf
          https://tax.vermont.gov/sites/tax/files/documents/Income%20Booklet-2022.pdf
          https://tax.vermont.gov/sites/tax/files/documents/Income%20Booklet-2022.pdf
          https://tax.vermont.gov/sites/tax/files/documents/Income%20Booklet-2022.pdf
          https://tax.vermont.gov/individuals/personal-income-tax/tax-credits
          https://tax.vermont.gov/individuals/personal-income-tax/taxable-income
          https://tax.vermont.gov/individuals/personal-income-tax/local-option-tax
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Florida', 'state_occupation': 'Vermont', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def florida_and_virginia(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, capital_gains_long=0, filing_status='single', income=100000, dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Florida:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        Virginia:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Florida:
        - https://floridarevenue.com/taxes/taxesfees/Pages/tax_interest_rates.aspx
        
        Virginia:
        -         
        - https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/reciprocity
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Florida', 'state_occupation': 'Virginia', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def florida_and_washington(self, capital_gains_long=0, sources=False, capital_gains_short=0):
        """
        --- Parameters ----
        
        Florida:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        Washington:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Florida:
        - https://floridarevenue.com/taxes/taxesfees/Pages/tax_interest_rates.aspx
        
        Washington:
        -         
        - https://dor.wa.gov/taxes-rates/income-tax
          https://dor.wa.gov/taxes-rates/other-taxes/capital-gains-tax
          https://dor.wa.gov/taxes-rates/income-tax
        """
        
        # Check if the parameters are set upstream in the instance
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        sources = self.sources if hasattr(self, 'sources') else sources
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        
        # State-specific parameters
        data = {
            'state_residence': 'Florida', 'state_occupation': 'Washington', "capital_gains_long": capital_gains_long, "sources": sources, "capital_gains_short": capital_gains_short
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def florida_and_west_virginia(self, pay_periods=24, state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', county_occupation='Other', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Florida:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        West Virginia:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_occupation : str, optional
            County of occupation. Default is Other.
        pay_periods : int, optional
            Number of pay periods. Options are 13, 24, 26, 52. Default is 24.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Florida:
        - https://floridarevenue.com/taxes/taxesfees/Pages/tax_interest_rates.aspx
        
        West Virginia:
        -         
        - https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://tax.wv.gov/Documents/TSD/tsd438.pdf
          https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2011/Taxes-11-35att.html
        """
        
        # Check if the parameters are set upstream in the instance
        pay_periods = self.pay_periods if hasattr(self, 'pay_periods') else pay_periods
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        county_occupation = self.county_occupation if hasattr(self, 'county_occupation') else county_occupation
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Florida', 'state_occupation': 'WestVirginia', "pay_periods": pay_periods, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "county_occupation": county_occupation, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def florida_and_wisconsin(self, state_agi=75000, capital_gains_short=0, capital_gains_long=0, filing_status='single', income=100000, dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Florida:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        Wisconsin:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Florida:
        - https://floridarevenue.com/taxes/taxesfees/Pages/tax_interest_rates.aspx
        
        Wisconsin:
        -         
        - https://www.revenue.wi.gov/TaxForms2023/2023-Form1-ES-Inst.pdf
          https://www.revenue.wi.gov/TaxForms2023/2023-Form1-ES-Inst.pdf
          https://www.revenue.wi.gov/TaxForms2023/2023-Form1-ES-Inst.pdf
          https://www.revenue.wi.gov/Pages/FAQS/pcs-work.aspx
          https://www.revenue.wi.gov/Dor%20Publications/pb103.pdf
          https://www.revenue.wi.gov/Pages/FAQS/pcs-work.aspx
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Florida', 'state_occupation': 'Wisconsin', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def florida_and_wyoming(self, sources=False):
        """
        --- Parameters ----
        
        Florida:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        Wyoming:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Florida:
        - https://floridarevenue.com/taxes/taxesfees/Pages/tax_interest_rates.aspx
        
        Wyoming:
        -         
        - https://revenue.wyo.gov/
        """
        
        # Check if the parameters are set upstream in the instance
        sources = self.sources if hasattr(self, 'sources') else sources
        
        # State-specific parameters
        data = {
            'state_residence': 'Florida', 'state_occupation': 'Wyoming', "sources": sources
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def georgia_and_alabama(self, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Georgia:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Alabama:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Georgia:
        - https://dor.georgia.gov/tax-tables-georgia-tax-rate-schedule
          https://apps.dor.ga.gov/FillableForms/PDFViewer/Index?form=2022GA500
          https://apps.dor.ga.gov/FillableForms/PDFViewer/Index?form=2022GA500
          https://apps.dor.ga.gov/FillableForms/PDFViewer/Index?form=2022GA500
          https://dor.georgia.gov/years-individual-income-tax-forms
          https://dor.georgia.gov/years-individual-income-tax-forms
        
        Alabama:
        -         
        - https://www.revenue.alabama.gov/ultraviewer/viewer/basic_viewer/index.html?form=2023/02/22f40.pdf
          https://www.revenue.alabama.gov/faqs/what-is-alabamas-individual-income-tax-rate/
          https://www.revenue.alabama.gov/ultraviewer/viewer/basic_viewer/index.html?form=2023/02/22f40.pdf
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-24.htm?taxmap=true
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-24.htm?taxmap=true
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-24.htm?taxmap=true
          https://www.revenue.alabama.gov/individual-corporate/income-to-be-reported-on-the-alabama-income-tax-return/
          https://www.almonline.org/TaxRates.aspx#OccupationalTax
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Georgia', 'state_occupation': 'Alabama', "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def georgia_and_alaska(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Georgia:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Alaska:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Georgia:
        - https://dor.georgia.gov/tax-tables-georgia-tax-rate-schedule
          https://apps.dor.ga.gov/FillableForms/PDFViewer/Index?form=2022GA500
          https://apps.dor.ga.gov/FillableForms/PDFViewer/Index?form=2022GA500
          https://apps.dor.ga.gov/FillableForms/PDFViewer/Index?form=2022GA500
          https://dor.georgia.gov/years-individual-income-tax-forms
          https://dor.georgia.gov/years-individual-income-tax-forms
        
        Alaska:
        -         
        - https://www.commerce.alaska.gov/web/dcra/officeofthestateassessor/alaskataxfacts.aspx
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Georgia', 'state_occupation': 'Alaska', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def georgia_and_arizona(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Georgia:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Arizona:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        credits_state : int, optional
            Credits for state taxes. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Georgia:
        - https://dor.georgia.gov/tax-tables-georgia-tax-rate-schedule
          https://apps.dor.ga.gov/FillableForms/PDFViewer/Index?form=2022GA500
          https://apps.dor.ga.gov/FillableForms/PDFViewer/Index?form=2022GA500
          https://apps.dor.ga.gov/FillableForms/PDFViewer/Index?form=2022GA500
          https://dor.georgia.gov/years-individual-income-tax-forms
          https://dor.georgia.gov/years-individual-income-tax-forms
        
        Arizona:
        -         
        - https://azdor.gov/forms/individual/form-140-x-y-tables
          https://azdor.gov/forms/individual/form-140-x-y-tables
          https://azdor.gov/forms/individual/form-140-x-y-tables
          https://azdor.gov/businesses-arizona/withholding-tax/withholding-exceptions
          "https://azdor.gov/individuals/income-tax-filing-assistance/identifying-other-taxable-income#:~:text=In%20regards%20to%20capital%20gains
         the%20individuals%20regular%20tax%20rate.
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Georgia', 'state_occupation': 'Arizona', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def georgia_and_arkansas(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Georgia:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Arkansas:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Georgia:
        - https://dor.georgia.gov/tax-tables-georgia-tax-rate-schedule
          https://apps.dor.ga.gov/FillableForms/PDFViewer/Index?form=2022GA500
          https://apps.dor.ga.gov/FillableForms/PDFViewer/Index?form=2022GA500
          https://apps.dor.ga.gov/FillableForms/PDFViewer/Index?form=2022GA500
          https://dor.georgia.gov/years-individual-income-tax-forms
          https://dor.georgia.gov/years-individual-income-tax-forms
        
        Arkansas:
        -         
        - https://www.arkansasedc.com/why-arkansas/business-climate/tax-structure/personal-income-tax
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_and_AR1000NR_Instructions.pdf
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_FullYearResidentIndividualIncomeTaxReturn.pdf
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_and_AR1000NR_Instructions.pdf
          https://www.arkansasedc.com/why-arkansas/business-climate/tax-structure/capital-gains-tax-reduction
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_and_AR1000NR_Instructions.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Georgia', 'state_occupation': 'Arkansas', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def georgia_and_california(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Georgia:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        California:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Georgia:
        - https://dor.georgia.gov/tax-tables-georgia-tax-rate-schedule
          https://apps.dor.ga.gov/FillableForms/PDFViewer/Index?form=2022GA500
          https://apps.dor.ga.gov/FillableForms/PDFViewer/Index?form=2022GA500
          https://apps.dor.ga.gov/FillableForms/PDFViewer/Index?form=2022GA500
          https://dor.georgia.gov/years-individual-income-tax-forms
          https://dor.georgia.gov/years-individual-income-tax-forms
        
        California:
        -         
        - https://www.ftb.ca.gov/forms/2022/2022-540-tax-rate-schedules.pdf
          https://www.ftb.ca.gov/file/personal/deductions/index.html
          https://www.ftb.ca.gov/about-ftb/newsroom/caleitc/eligibility-and-credit-information.html
          https://www.ftb.ca.gov/file/personal/income-types/capital-gains-and-losses.html#:~:text=California%20does%20not%20have%20a
         are%20taxed%20as%20ordinary%20income
          https://www.ftb.ca.gov/file/personal/
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Georgia', 'state_occupation': 'California', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def georgia_and_colorado(self, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Georgia:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Colorado:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Georgia:
        - https://dor.georgia.gov/tax-tables-georgia-tax-rate-schedule
          https://apps.dor.ga.gov/FillableForms/PDFViewer/Index?form=2022GA500
          https://apps.dor.ga.gov/FillableForms/PDFViewer/Index?form=2022GA500
          https://apps.dor.ga.gov/FillableForms/PDFViewer/Index?form=2022GA500
          https://dor.georgia.gov/years-individual-income-tax-forms
          https://dor.georgia.gov/years-individual-income-tax-forms
        
        Colorado:
        -         
        - https://leg.colorado.gov/agencies/legislative-council-staff/individual-income-tax%C2%A0
          https://leg.colorado.gov/agencies/legislative-council-staff/individual-income-tax%C2%A0
          https://tax.colorado.gov/sites/tax/files/documents/ITT_Capital_Gain_Subtraction_Feb_2022.pdf
          https://www.auroragov.org/business_services/taxes/occupational_privilege_tax
          https://www.denvergov.org/Government/Agencies-Departments-Offices/Agencies-Departments-Offices-Directory/Department-of-Finance/Our-Divisions/Treasury/Business-Tax-Information#:~:text=Occupational%20privilege%20tax
         -%28show%20below%29&text=Each%20taxable%20employee%20is%20liable
         month%20for%20each%20taxable%20employee
          https://www.glendale.co.us/355/Occupational-Privilege-Tax
          https://www.greenwoodvillage.com/1220/Occupational-Privilege-Tax-OPT
          https://ci.sheridan.co.us/288/Occupational-Privilege-Tax
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Georgia', 'state_occupation': 'Colorado', "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def georgia_and_connecticut(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Georgia:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Connecticut:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Georgia:
        - https://dor.georgia.gov/tax-tables-georgia-tax-rate-schedule
          https://apps.dor.ga.gov/FillableForms/PDFViewer/Index?form=2022GA500
          https://apps.dor.ga.gov/FillableForms/PDFViewer/Index?form=2022GA500
          https://apps.dor.ga.gov/FillableForms/PDFViewer/Index?form=2022GA500
          https://dor.georgia.gov/years-individual-income-tax-forms
          https://dor.georgia.gov/years-individual-income-tax-forms
        
        Connecticut:
        -         
        - https://portal.ct.gov/-/media/DRS/Forms/2022/Income/2022-CT-1040-Instructions_1222.pdf
          https://www.cga.ct.gov/2022/rpt/pdf/2022-R-0108.pdf
          https://www.cga.ct.gov/current/pub/chap_229.htm#sec_12-703
          https://portal.ct.gov/-/media/DRS/Forms/2022/Income/2022-CT-1040-Instructions_1222.pdf
          "https://portal.ct.gov/DRS/Publications/TSSNs/TSSN-29#:~:text=An%20individuals%20net%20capital%20gains
         portion%20of%20Social%20Security%20benefits"
          "https://portal.ct.gov/DRS/Publications/TSSNs/TSSN-29#:~:text=An%20individuals%20net%20capital%20gains
         portion%20of%20Social%20Security%20benefits"
          https://portal.ct.gov/DRS/Individuals/Resident-Income-Tax/Tax-Information
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Georgia', 'state_occupation': 'Connecticut', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def georgia_and_delaware(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, federal_agi=75000, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Georgia:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Delaware:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        federal_agi : int, optional
            Adjusted Gross Income (AGI) for federal taxes. Default is 0. Sources - https://www.irs.gov/e-file-providers/definition-of-adjusted-gross-income, https://www.irs.gov/newsroom/taxpayers-should-know-the-difference-between-standard-and-itemized-deductions
        income : int, optional
            Gross income. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Georgia:
        - https://dor.georgia.gov/tax-tables-georgia-tax-rate-schedule
          https://apps.dor.ga.gov/FillableForms/PDFViewer/Index?form=2022GA500
          https://apps.dor.ga.gov/FillableForms/PDFViewer/Index?form=2022GA500
          https://apps.dor.ga.gov/FillableForms/PDFViewer/Index?form=2022GA500
          https://dor.georgia.gov/years-individual-income-tax-forms
          https://dor.georgia.gov/years-individual-income-tax-forms
        
        Delaware:
        -         
        - https://revenue.delaware.gov/software-developer/tax-rate-changes/
          https://revenuefiles.delaware.gov/2022/PIT-RES_TY22_2022-01_PaperInteractive.pdf
          https://revenuefiles.delaware.gov/2022/PIT-RES_TY22_2022-02_Instructions.pdf
          https://revenue.delaware.gov/frequently-asked-questions/personal-income-tax-faqs/
          https://www.newcastlede.gov/DocumentCenter/View/20147/City-of-Wilmington-Wage-Tax-Information-PDF
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        federal_agi = self.federal_agi if hasattr(self, 'federal_agi') else federal_agi
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Georgia', 'state_occupation': 'Delaware', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "federal_agi": federal_agi, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def georgia_and_florida(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Georgia:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Florida:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Georgia:
        - https://dor.georgia.gov/tax-tables-georgia-tax-rate-schedule
          https://apps.dor.ga.gov/FillableForms/PDFViewer/Index?form=2022GA500
          https://apps.dor.ga.gov/FillableForms/PDFViewer/Index?form=2022GA500
          https://apps.dor.ga.gov/FillableForms/PDFViewer/Index?form=2022GA500
          https://dor.georgia.gov/years-individual-income-tax-forms
          https://dor.georgia.gov/years-individual-income-tax-forms
        
        Florida:
        -         
        - https://floridarevenue.com/taxes/taxesfees/Pages/tax_interest_rates.aspx
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Georgia', 'state_occupation': 'Florida', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def georgia_and_hawaii(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Georgia:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Hawaii:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Georgia:
        - https://dor.georgia.gov/tax-tables-georgia-tax-rate-schedule
          https://apps.dor.ga.gov/FillableForms/PDFViewer/Index?form=2022GA500
          https://apps.dor.ga.gov/FillableForms/PDFViewer/Index?form=2022GA500
          https://apps.dor.ga.gov/FillableForms/PDFViewer/Index?form=2022GA500
          https://dor.georgia.gov/years-individual-income-tax-forms
          https://dor.georgia.gov/years-individual-income-tax-forms
        
        Hawaii:
        -         
        - https://tax.hawaii.gov/forms/d_18table-on/d_18table-on_p13/
          https://files.hawaii.gov/tax/forms/2018/18table-on.pdf
          https://files.hawaii.gov/tax/news/pubs/22outline.pdf
          https://files.hawaii.gov/tax/news/pubs/FastTaxRefGuide_2017.pdf
          https://www.tfhawaii.org/wordpress/state-tax-resources/mini-tax-guides/individual-income-tax-chapter-235/#:~:text=Long%20term%20capital%20gains%20are%20taxed%20at%20a%20maximum%20of%207.25%25
          https://www.grassrootinstitute.org/2023/02/hb337-increasing-tax-on-capital-gains-a-bad-risk/
          https://files.hawaii.gov/tax/news/pubs/22outline.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Georgia', 'state_occupation': 'Hawaii', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def georgia_and_idaho(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Georgia:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Idaho:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Georgia:
        - https://dor.georgia.gov/tax-tables-georgia-tax-rate-schedule
          https://apps.dor.ga.gov/FillableForms/PDFViewer/Index?form=2022GA500
          https://apps.dor.ga.gov/FillableForms/PDFViewer/Index?form=2022GA500
          https://apps.dor.ga.gov/FillableForms/PDFViewer/Index?form=2022GA500
          https://dor.georgia.gov/years-individual-income-tax-forms
          https://dor.georgia.gov/years-individual-income-tax-forms
        
        Idaho:
        -         
        - https://tax.idaho.gov/taxes/income-tax/individual-income/individual-income-tax-rate-schedule/
          https://tax.idaho.gov/wp-content/uploads/forms/EIN00046/EIN00046_03-01-2023.pdf
          https://tax.idaho.gov/taxes/income-tax/individual-income/idaho-source-income/
          https://tax.idaho.gov/taxes/income-tax/individual-income/idaho-source-income/
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Georgia', 'state_occupation': 'Idaho', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def georgia_and_illinois(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Georgia:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Illinois:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Georgia:
        - https://dor.georgia.gov/tax-tables-georgia-tax-rate-schedule
          https://apps.dor.ga.gov/FillableForms/PDFViewer/Index?form=2022GA500
          https://apps.dor.ga.gov/FillableForms/PDFViewer/Index?form=2022GA500
          https://apps.dor.ga.gov/FillableForms/PDFViewer/Index?form=2022GA500
          https://dor.georgia.gov/years-individual-income-tax-forms
          https://dor.georgia.gov/years-individual-income-tax-forms
        
        Illinois:
        -         
        - https://tax.illinois.gov/research/taxrates/income.html#IndividualIncome
          https://tax.illinois.gov/content/dam/soi/en/web/tax/forms/incometax/documents/currentyear/individual/il-1040-instr.pdf
          https://tax.illinois.gov/questionsandanswers/12.html
          https://tax.illinois.gov/content/dam/soi/en/web/tax/forms/incometax/documents/currentyear/individual/il-1040-instr.pdf
          https://tax.illinois.gov/research/taxrates/personalproperty.html
          https://tax.illinois.gov/localgovernments/income.html
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Georgia', 'state_occupation': 'Illinois', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def georgia_and_indiana(self, capital_gains_short=0, dependents=0, sources=False, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_long=0, filing_status='single', income=100000, credits_state=0):
        """
        --- Parameters ----
        
        Georgia:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Indiana:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Georgia:
        - https://dor.georgia.gov/tax-tables-georgia-tax-rate-schedule
          https://apps.dor.ga.gov/FillableForms/PDFViewer/Index?form=2022GA500
          https://apps.dor.ga.gov/FillableForms/PDFViewer/Index?form=2022GA500
          https://apps.dor.ga.gov/FillableForms/PDFViewer/Index?form=2022GA500
          https://dor.georgia.gov/years-individual-income-tax-forms
          https://dor.georgia.gov/years-individual-income-tax-forms
        
        Indiana:
        -         
        - https://www.in.gov/dor/tax-forms/2022-individual-income-tax-forms/
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-32.htm
          https://www.in.gov/dor/files/ib28.pdf
          https://www.in.gov/dor/tax-forms/2022-individual-income-tax-forms/
          https://www.in.gov/dor/files/2022-county-tax-rates-and-codes.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Georgia', 'state_occupation': 'Indiana', "capital_gains_short": capital_gains_short, "dependents": dependents, "sources": sources, "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def georgia_and_iowa(self, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Georgia:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Iowa:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Georgia:
        - https://dor.georgia.gov/tax-tables-georgia-tax-rate-schedule
          https://apps.dor.ga.gov/FillableForms/PDFViewer/Index?form=2022GA500
          https://apps.dor.ga.gov/FillableForms/PDFViewer/Index?form=2022GA500
          https://apps.dor.ga.gov/FillableForms/PDFViewer/Index?form=2022GA500
          https://dor.georgia.gov/years-individual-income-tax-forms
          https://dor.georgia.gov/years-individual-income-tax-forms
        
        Iowa:
        -         
        - https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
          https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
          https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
          https://tax.iowa.gov/iowa-illinois-reciprocal-agreement
          https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Georgia', 'state_occupation': 'Iowa', "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def georgia_and_kansas(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Georgia:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Kansas:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Georgia:
        - https://dor.georgia.gov/tax-tables-georgia-tax-rate-schedule
          https://apps.dor.ga.gov/FillableForms/PDFViewer/Index?form=2022GA500
          https://apps.dor.ga.gov/FillableForms/PDFViewer/Index?form=2022GA500
          https://apps.dor.ga.gov/FillableForms/PDFViewer/Index?form=2022GA500
          https://dor.georgia.gov/years-individual-income-tax-forms
          https://dor.georgia.gov/years-individual-income-tax-forms
        
        Kansas:
        -         
        - https://www.ksrevenue.gov/taxrates.html
          https://www.ksrevenue.gov/incomebook22.html#0
          https://www.ksrevenue.gov/incomebook22.html#0
          https://www.ksrevenue.gov/incomebook22.html#0
          https://www.ksrevenue.gov/incomebook22.html
          https://www.ksrevenue.gov/incomebook22.html
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Georgia', 'state_occupation': 'Kansas', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def georgia_and_kentucky(self, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Georgia:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Kentucky:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Georgia:
        - https://dor.georgia.gov/tax-tables-georgia-tax-rate-schedule
          https://apps.dor.ga.gov/FillableForms/PDFViewer/Index?form=2022GA500
          https://apps.dor.ga.gov/FillableForms/PDFViewer/Index?form=2022GA500
          https://apps.dor.ga.gov/FillableForms/PDFViewer/Index?form=2022GA500
          https://dor.georgia.gov/years-individual-income-tax-forms
          https://dor.georgia.gov/years-individual-income-tax-forms
        
        Kentucky:
        -         
        - https://revenue.ky.gov/News/Pages/Dor-Announces-Updates-to-Individual-Income-Tax-for-2023-Tax-Year.aspx
          https://revenue.ky.gov/Individual/Individual-Income-Tax/Pages/default.aspx
          https://revenue.ky.gov/News/Pages/Dor-Announces-Updates-to-Individual-Income-Tax-for-2023-Tax-Year.aspx
          https://revenue.ky.gov/Individual/Individual-Income-Tax/Pages/default.aspx
          https://revenue.ky.gov/Dor%20Training%20Materials/103%20KAR%2017.140.%20Individual%20income%20tax%20-%20reciprocity%20-%20nonresidents.pdf
          https://revenue.ky.gov/Individual/Individual-Income-Tax/Pages/default.aspx
          https://nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2021/Taxes-21-32.htm?taxmap=true
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Georgia', 'state_occupation': 'Kentucky', "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def georgia_and_louisiana(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Georgia:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Louisiana:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Georgia:
        - https://dor.georgia.gov/tax-tables-georgia-tax-rate-schedule
          https://apps.dor.ga.gov/FillableForms/PDFViewer/Index?form=2022GA500
          https://apps.dor.ga.gov/FillableForms/PDFViewer/Index?form=2022GA500
          https://apps.dor.ga.gov/FillableForms/PDFViewer/Index?form=2022GA500
          https://dor.georgia.gov/years-individual-income-tax-forms
          https://dor.georgia.gov/years-individual-income-tax-forms
        
        Louisiana:
        -         
        - https://revenue.louisiana.gov/individualincometax
          https://revenue.louisiana.gov/individualincometax
          https://www.revenue.louisiana.gov/taxforms/6935(11_02)F.pdf
          https://revenue.louisiana.gov/individualincometax
          https://revenue.louisiana.gov/individualincometax
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Georgia', 'state_occupation': 'Louisiana', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def georgia_and_maine(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Georgia:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Maine:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Georgia:
        - https://dor.georgia.gov/tax-tables-georgia-tax-rate-schedule
          https://apps.dor.ga.gov/FillableForms/PDFViewer/Index?form=2022GA500
          https://apps.dor.ga.gov/FillableForms/PDFViewer/Index?form=2022GA500
          https://apps.dor.ga.gov/FillableForms/PDFViewer/Index?form=2022GA500
          https://dor.georgia.gov/years-individual-income-tax-forms
          https://dor.georgia.gov/years-individual-income-tax-forms
        
        Maine:
        -         
        - https://www.maine.gov/revenue/tax-return-forms/individual-income-tax-2022
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/ind_tax_rate_sched_2022.pdf
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/22_1040me_book_gen_instruc_revisedFeb23.pdf
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/22_1040me_book_gen_instruc_revisedFeb23.pdf
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/22_1040me_book_gen_instruc_revisedFeb23.pdf
          https://www.maine.gov/revenue/tax-return-forms/individual-income-tax-2022
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Georgia', 'state_occupation': 'Maine', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def georgia_and_maryland(self, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Georgia:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Maryland:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Georgia:
        - https://dor.georgia.gov/tax-tables-georgia-tax-rate-schedule
          https://apps.dor.ga.gov/FillableForms/PDFViewer/Index?form=2022GA500
          https://apps.dor.ga.gov/FillableForms/PDFViewer/Index?form=2022GA500
          https://apps.dor.ga.gov/FillableForms/PDFViewer/Index?form=2022GA500
          https://dor.georgia.gov/years-individual-income-tax-forms
          https://dor.georgia.gov/years-individual-income-tax-forms
        
        Maryland:
        -         
        - https://www.marylandtaxes.gov/individual/income/tax-info/tax-rates.php
          https://www.marylandtaxes.gov/forms/22_forms/Resident_Booklet.pdf
          https://www.marylandtaxes.gov/forms/Tax_Publications/Administrative_Releases/Income_and_Estate_Tax_Releases/ar_it37.pdf
          https://www.marylandtaxes.gov/individual/index.php
          https://www.marylandtaxes.gov/individual/income/tax-info/tax-rates.php
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Georgia', 'state_occupation': 'Maryland', "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def georgia_and_massachusetts(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Georgia:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Massachusetts:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Georgia:
        - https://dor.georgia.gov/tax-tables-georgia-tax-rate-schedule
          https://apps.dor.ga.gov/FillableForms/PDFViewer/Index?form=2022GA500
          https://apps.dor.ga.gov/FillableForms/PDFViewer/Index?form=2022GA500
          https://apps.dor.ga.gov/FillableForms/PDFViewer/Index?form=2022GA500
          https://dor.georgia.gov/years-individual-income-tax-forms
          https://dor.georgia.gov/years-individual-income-tax-forms
        
        Massachusetts:
        -         
        - https://www.mass.gov/service-details/massachusetts-tax-rates
          https://www.mass.gov/service-details/view-massachusetts-personal-income-tax-exemptions
          https://www.mass.gov/service-details/view-massachusetts-personal-income-tax-exemptions
          https://www.mass.gov/service-details/view-child-and-dependent-related-credits
          https://www.mass.gov/service-details/massachusetts-tax-rates
          https://www.mass.gov/personal-income-tax
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Georgia', 'state_occupation': 'Massachusetts', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def georgia_and_michigan(self, capital_gains_short=0, county_occupation='Other', dependents=0, sources=False, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_long=0, filing_status='single', income=100000, credits_state=0):
        """
        --- Parameters ----
        
        Georgia:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Michigan:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        county_occupation : str, optional
            County of occupation. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Georgia:
        - https://dor.georgia.gov/tax-tables-georgia-tax-rate-schedule
          https://apps.dor.ga.gov/FillableForms/PDFViewer/Index?form=2022GA500
          https://apps.dor.ga.gov/FillableForms/PDFViewer/Index?form=2022GA500
          https://apps.dor.ga.gov/FillableForms/PDFViewer/Index?form=2022GA500
          https://dor.georgia.gov/years-individual-income-tax-forms
          https://dor.georgia.gov/years-individual-income-tax-forms
        
        Michigan:
        -         
        - https://www.michigan.gov/taxes/iit/new-developments-for-tax-year-2022
          "https://www.michigan.gov/taxes/-/media/Project/Websites/taxes/Forms/2023/2023-SUW/446_Withholding-Guide_2023.pdf?rev=177ca3d8cb034392bf84e317ed78c320&hash=19C5F42F43A89A91863D229349F48883#:~:text=The%20exemption%20amount%20is%20%245%2C400
         employees%20Michigan%20income%20tax%20return"
          "https://www.michigan.gov/taxes/-/media/Project/Websites/taxes/Forms/2023/2023-SUW/446_Withholding-Guide_2023.pdf?rev=177ca3d8cb034392bf84e317ed78c320&hash=19C5F42F43A89A91863D229349F48883#:~:text=The%20exemption%20amount%20is%20%245%2C400
         employees%20Michigan%20income%20tax%20return"
          https://www.michigan.gov/taxes/questions/iit/accordion/residency/are-my-wages-earned-in-another-state-taxable-in-michigan-if-i-am-a-michigan-resident-1
          https://www.michigan.gov/taxes/-/media/Project/Websites/taxes/Forms/2022/2022-IIT-Forms/BOOK_MI-1040.pdf?rev=4524f1cbbc0243b194ab05a6680be75b&hash=40DFE71E651F7C1D65C31F6CD03D944F
          https://www.michigan.gov/taxes/questions/iit/accordion/general/what-cities-impose-an-income-tax
        """
        
        # Check if the parameters are set upstream in the instance
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        county_occupation = self.county_occupation if hasattr(self, 'county_occupation') else county_occupation
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Georgia', 'state_occupation': 'Michigan', "capital_gains_short": capital_gains_short, "county_occupation": county_occupation, "dependents": dependents, "sources": sources, "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def georgia_and_minnesota(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Georgia:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Minnesota:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Georgia:
        - https://dor.georgia.gov/tax-tables-georgia-tax-rate-schedule
          https://apps.dor.ga.gov/FillableForms/PDFViewer/Index?form=2022GA500
          https://apps.dor.ga.gov/FillableForms/PDFViewer/Index?form=2022GA500
          https://apps.dor.ga.gov/FillableForms/PDFViewer/Index?form=2022GA500
          https://dor.georgia.gov/years-individual-income-tax-forms
          https://dor.georgia.gov/years-individual-income-tax-forms
        
        Minnesota:
        -         
        - https://www.revenue.state.mn.us/minnesota-income-tax-rates-and-brackets
          https://www.revenue.state.mn.us/child-and-dependent-care-credit
          https://www.revenue.state.mn.us/dependent-exemptions
          https://www.revenue.state.mn.us/minnesota-standard-deduction
          https://mn.gov/mmb/assets/mwr_form_tcm1059-128581.pdf
          https://www.revenue.state.mn.us/individual-income-tax
          https://www.revenue.state.mn.us/individual-income-tax
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Georgia', 'state_occupation': 'Minnesota', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def georgia_and_mississippi(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Georgia:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Mississippi:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Georgia:
        - https://dor.georgia.gov/tax-tables-georgia-tax-rate-schedule
          https://apps.dor.ga.gov/FillableForms/PDFViewer/Index?form=2022GA500
          https://apps.dor.ga.gov/FillableForms/PDFViewer/Index?form=2022GA500
          https://apps.dor.ga.gov/FillableForms/PDFViewer/Index?form=2022GA500
          https://dor.georgia.gov/years-individual-income-tax-forms
          https://dor.georgia.gov/years-individual-income-tax-forms
        
        Mississippi:
        -         
        - https://www.dor.ms.gov/individual/tax-rates
          https://www.dor.ms.gov/individual/tax-rates
          https://www.dor.ms.gov/individual/tax-rates
          https://www.dor.ms.gov/individual
          https://www.dor.ms.gov/individual
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Georgia', 'state_occupation': 'Mississippi', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def georgia_and_missouri(self, capital_gains_short=0, dependents=0, sources=False, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_long=0, filing_status='single', income=100000, credits_state=0):
        """
        --- Parameters ----
        
        Georgia:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Missouri:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Georgia:
        - https://dor.georgia.gov/tax-tables-georgia-tax-rate-schedule
          https://apps.dor.ga.gov/FillableForms/PDFViewer/Index?form=2022GA500
          https://apps.dor.ga.gov/FillableForms/PDFViewer/Index?form=2022GA500
          https://apps.dor.ga.gov/FillableForms/PDFViewer/Index?form=2022GA500
          https://dor.georgia.gov/years-individual-income-tax-forms
          https://dor.georgia.gov/years-individual-income-tax-forms
        
        Missouri:
        -         
        - https://dor.mo.gov/taxation/individual/tax-types/income/year-changes/
          https://dor.mo.gov/taxation/individual/tax-types/income/year-changes/
          https://dor.mo.gov/taxation/individual/tax-types/income/year-changes/
          https://www.kcmo.gov/city-hall/departments/finance/earnings-tax
          https://www.stlouis-mo.gov/government/departments/collector/earnings-tax/individual-earnings-tax-info.cfm
        """
        
        # Check if the parameters are set upstream in the instance
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Georgia', 'state_occupation': 'Missouri', "capital_gains_short": capital_gains_short, "dependents": dependents, "sources": sources, "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def georgia_and_montana(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Georgia:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Montana:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Georgia:
        - https://dor.georgia.gov/tax-tables-georgia-tax-rate-schedule
          https://apps.dor.ga.gov/FillableForms/PDFViewer/Index?form=2022GA500
          https://apps.dor.ga.gov/FillableForms/PDFViewer/Index?form=2022GA500
          https://apps.dor.ga.gov/FillableForms/PDFViewer/Index?form=2022GA500
          https://dor.georgia.gov/years-individual-income-tax-forms
          https://dor.georgia.gov/years-individual-income-tax-forms
        
        Montana:
        -         
        - https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/2022/12/montana-employees-withholding-allowance-and-exemption-certificate-form-mw-4.pdf
          https://mtrevenue.gov/2021/11/29/simplification-of-montana-income-taxation/?utm_source=rss&utm_medium=rss&utm_campaign=simplification-of-montana-income-taxation
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Georgia', 'state_occupation': 'Montana', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def georgia_and_nebraska(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Georgia:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Nebraska:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Georgia:
        - https://dor.georgia.gov/tax-tables-georgia-tax-rate-schedule
          https://apps.dor.ga.gov/FillableForms/PDFViewer/Index?form=2022GA500
          https://apps.dor.ga.gov/FillableForms/PDFViewer/Index?form=2022GA500
          https://apps.dor.ga.gov/FillableForms/PDFViewer/Index?form=2022GA500
          https://dor.georgia.gov/years-individual-income-tax-forms
          https://dor.georgia.gov/years-individual-income-tax-forms
        
        Nebraska:
        -         
        - https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2023_Circular_EN_8-429r2023_Whole_Book_11.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Georgia', 'state_occupation': 'Nebraska', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def georgia_and_nevada(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Georgia:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Nevada:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Georgia:
        - https://dor.georgia.gov/tax-tables-georgia-tax-rate-schedule
          https://apps.dor.ga.gov/FillableForms/PDFViewer/Index?form=2022GA500
          https://apps.dor.ga.gov/FillableForms/PDFViewer/Index?form=2022GA500
          https://apps.dor.ga.gov/FillableForms/PDFViewer/Index?form=2022GA500
          https://dor.georgia.gov/years-individual-income-tax-forms
          https://dor.georgia.gov/years-individual-income-tax-forms
        
        Nevada:
        -         
        - https://www.nvsos.gov/sos/businesses/the-nevada-advantage
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Georgia', 'state_occupation': 'Nevada', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def georgia_and_new_hampshire(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Georgia:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        New Hampshire:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Georgia:
        - https://dor.georgia.gov/tax-tables-georgia-tax-rate-schedule
          https://apps.dor.ga.gov/FillableForms/PDFViewer/Index?form=2022GA500
          https://apps.dor.ga.gov/FillableForms/PDFViewer/Index?form=2022GA500
          https://apps.dor.ga.gov/FillableForms/PDFViewer/Index?form=2022GA500
          https://dor.georgia.gov/years-individual-income-tax-forms
          https://dor.georgia.gov/years-individual-income-tax-forms
        
        New Hampshire:
        -         
        - https://www.revenue.nh.gov/assistance/tax-overview.htm
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Georgia', 'state_occupation': 'NewHampshire', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def georgia_and_new_jersey(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Georgia:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        New Jersey:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Georgia:
        - https://dor.georgia.gov/tax-tables-georgia-tax-rate-schedule
          https://apps.dor.ga.gov/FillableForms/PDFViewer/Index?form=2022GA500
          https://apps.dor.ga.gov/FillableForms/PDFViewer/Index?form=2022GA500
          https://apps.dor.ga.gov/FillableForms/PDFViewer/Index?form=2022GA500
          https://dor.georgia.gov/years-individual-income-tax-forms
          https://dor.georgia.gov/years-individual-income-tax-forms
        
        New Jersey:
        -         
        - https://www.state.nj.us/treasury/taxation/taxtables.shtml
          https://www.state.nj.us/treasury/taxation/njit13.shtml
          https://www.nj.gov/treasury/taxation/njit25.shtml
          https://www.state.nj.us/treasury/taxation/individuals/freqqiti.shtml
          https://www.state.nj.us/treasury/taxation/localtax.shtml
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Georgia', 'state_occupation': 'NewJersey', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def georgia_and_new_mexico(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Georgia:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        New Mexico:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Georgia:
        - https://dor.georgia.gov/tax-tables-georgia-tax-rate-schedule
          https://apps.dor.ga.gov/FillableForms/PDFViewer/Index?form=2022GA500
          https://apps.dor.ga.gov/FillableForms/PDFViewer/Index?form=2022GA500
          https://apps.dor.ga.gov/FillableForms/PDFViewer/Index?form=2022GA500
          https://dor.georgia.gov/years-individual-income-tax-forms
          https://dor.georgia.gov/years-individual-income-tax-forms
        
        New Mexico:
        -         
        - https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/fdf3c548-8aba-4b9c-9eb4-bb564c716015/FYI-104.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/0bf943fd-652e-4400-bb1b-cd8397eb5a95/2022pit-adj.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/2f1a6781-9534-4436-b427-1557f9592099/2022pit-adj-ins.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/0bf943fd-652e-4400-bb1b-cd8397eb5a95/2022pit-adj.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/1afc56af-ea90-4d48-82e5-1f9aeb43255a/PITbook2022.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/fdf3c548-8aba-4b9c-9eb4-bb564c716015/FYI-104.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Georgia', 'state_occupation': 'NewMexico', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def georgia_and_new_york(self, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Georgia:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        New York:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Georgia:
        - https://dor.georgia.gov/tax-tables-georgia-tax-rate-schedule
          https://apps.dor.ga.gov/FillableForms/PDFViewer/Index?form=2022GA500
          https://apps.dor.ga.gov/FillableForms/PDFViewer/Index?form=2022GA500
          https://apps.dor.ga.gov/FillableForms/PDFViewer/Index?form=2022GA500
          https://dor.georgia.gov/years-individual-income-tax-forms
          https://dor.georgia.gov/years-individual-income-tax-forms
        
        New York:
        -         
        - https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nys-tax-rate-schedule
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nys-tax-rate-schedule
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nys-tax-rate-schedule
          https://www.tax.ny.gov/pit/
          https://www.tax.ny.gov/pit/file/nonresident-faqs.htm
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#yonkers-tax-rate-schedule
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nyc-tax-rate-schedule
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Georgia', 'state_occupation': 'NewYork', "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def georgia_and_north_carolina(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Georgia:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        North Carolina:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Georgia:
        - https://dor.georgia.gov/tax-tables-georgia-tax-rate-schedule
          https://apps.dor.ga.gov/FillableForms/PDFViewer/Index?form=2022GA500
          https://apps.dor.ga.gov/FillableForms/PDFViewer/Index?form=2022GA500
          https://apps.dor.ga.gov/FillableForms/PDFViewer/Index?form=2022GA500
          https://dor.georgia.gov/years-individual-income-tax-forms
          https://dor.georgia.gov/years-individual-income-tax-forms
        
        North Carolina:
        -         
        - https://www.ncdor.gov/taxes-forms/tax-rate-schedules
          https://www.ncdor.gov/taxes-forms/individual-income-tax/north-carolina-standard-deduction-or-north-carolina-itemized-deductions
          https://www.ncdor.gov/taxes-forms/individual-income-tax
          https://www.ncdor.gov/taxes-forms/individual-income-tax
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Georgia', 'state_occupation': 'NorthCarolina', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def georgia_and_north_dakota(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Georgia:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        North Dakota:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Georgia:
        - https://dor.georgia.gov/tax-tables-georgia-tax-rate-schedule
          https://apps.dor.ga.gov/FillableForms/PDFViewer/Index?form=2022GA500
          https://apps.dor.ga.gov/FillableForms/PDFViewer/Index?form=2022GA500
          https://apps.dor.ga.gov/FillableForms/PDFViewer/Index?form=2022GA500
          https://dor.georgia.gov/years-individual-income-tax-forms
          https://dor.georgia.gov/years-individual-income-tax-forms
        
        North Dakota:
        -         
        - https://www.tax.nd.gov/forms
          https://www.tax.nd.gov/sites/www/files/documents/forms/individual/2022-iit/2022-individual-income-tax-booklet.pdf
          https://www.tax.nd.gov/sites/www/files/documents/forms/2022-individual-income-tax-booklet.pdf
          https://www.tax.nd.gov/tax-exemptions-credits/income-tax-exemptions-credits
          https://www.tax.nd.gov/sites/www/files/documents/forms/business/ndwrfillable.pdf
          https://www.tax.nd.gov/news/tax-legislative-changes/significant-changes-law/individual-income-tax-history
          https://www.tax.nd.gov/forms
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Georgia', 'state_occupation': 'NorthDakota', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def georgia_and_ohio(self, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Georgia:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Ohio:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Georgia:
        - https://dor.georgia.gov/tax-tables-georgia-tax-rate-schedule
          https://apps.dor.ga.gov/FillableForms/PDFViewer/Index?form=2022GA500
          https://apps.dor.ga.gov/FillableForms/PDFViewer/Index?form=2022GA500
          https://apps.dor.ga.gov/FillableForms/PDFViewer/Index?form=2022GA500
          https://dor.georgia.gov/years-individual-income-tax-forms
          https://dor.georgia.gov/years-individual-income-tax-forms
        
        Ohio:
        -         
        - https://tax.ohio.gov/individual/resources/annual-tax-rates
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://tax.ohio.gov/static/forms/employer_withholding/generic/wth_it4nr.pdf
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://www.ritaohio.com/TaxRatesTable
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Georgia', 'state_occupation': 'Ohio', "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def georgia_and_oklahoma(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Georgia:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Oklahoma:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Georgia:
        - https://dor.georgia.gov/tax-tables-georgia-tax-rate-schedule
          https://apps.dor.ga.gov/FillableForms/PDFViewer/Index?form=2022GA500
          https://apps.dor.ga.gov/FillableForms/PDFViewer/Index?form=2022GA500
          https://apps.dor.ga.gov/FillableForms/PDFViewer/Index?form=2022GA500
          https://dor.georgia.gov/years-individual-income-tax-forms
          https://dor.georgia.gov/years-individual-income-tax-forms
        
        Oklahoma:
        -         
        - https://oklahoma.gov/content/dam/ok/en/tax/documents/resources/publications/businesses/withholding-tables/WHTables-2022.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Georgia', 'state_occupation': 'Oklahoma', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def georgia_and_oregon(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Georgia:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Oregon:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Georgia:
        - https://dor.georgia.gov/tax-tables-georgia-tax-rate-schedule
          https://apps.dor.ga.gov/FillableForms/PDFViewer/Index?form=2022GA500
          https://apps.dor.ga.gov/FillableForms/PDFViewer/Index?form=2022GA500
          https://apps.dor.ga.gov/FillableForms/PDFViewer/Index?form=2022GA500
          https://dor.georgia.gov/years-individual-income-tax-forms
          https://dor.georgia.gov/years-individual-income-tax-forms
        
        Oregon:
        -         
        - https://www.oregon.gov/dor/programs/individuals/Pages/PIT.aspx
          https://www.oregon.gov/dor/programs/individuals/Documents/Full%20year%20resident
         %20Form%20or-40%20filers.pdf
          https://www.oregon.gov/dor/programs/individuals/Documents/Part-year%20and%20nonresident
         %20Form%20or-40-P%20and%20or-40-N%20filers.pdf.pdf
          https://www.oregon.gov/dor/forms/FormsPubs/form-or-40_101-040_2022.pdf
          https://www.oregon.gov/dor/forms/FormsPubs/form-or-40_101-040_2022.pdf
          https://www.oregon.gov/dor/forms/FormsPubs/form-or-40_101-040_2022.pdf
          https://www.portland.gov/revenue/personal-tax
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Georgia', 'state_occupation': 'Oregon', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def georgia_and_pennsylvania(self, county_residence='Other', state_agi=75000, county_occupation='Other', filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Georgia:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Pennsylvania:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        county_residence : str, optional
            County of residence. Default is Other.
        county_occupation : str, optional
            County of occupation. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Georgia:
        - https://dor.georgia.gov/tax-tables-georgia-tax-rate-schedule
          https://apps.dor.ga.gov/FillableForms/PDFViewer/Index?form=2022GA500
          https://apps.dor.ga.gov/FillableForms/PDFViewer/Index?form=2022GA500
          https://apps.dor.ga.gov/FillableForms/PDFViewer/Index?form=2022GA500
          https://dor.georgia.gov/years-individual-income-tax-forms
          https://dor.georgia.gov/years-individual-income-tax-forms
        
        Pennsylvania:
        -         
        - https://www.revenue.pa.gov/Tax%20Rates/Pages/default.aspx
          https://www.revenue.pa.gov/FormsandPublications/FormsforIndividuals/PIT/Documents/2022/2022_pa-40in.pdf
          https://www.revenue.pa.gov/FormsandPublications/FormsforBusinesses/EmployerWithholding/Documents/rev-419.pdf
          https://www.revenue.pa.gov/FormsandPublications/FormsforIndividuals/PIT/Documents/2022/2022_pa-40in.pdf
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-25.htm
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        county_occupation = self.county_occupation if hasattr(self, 'county_occupation') else county_occupation
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Georgia', 'state_occupation': 'Pennsylvania', "county_residence": county_residence, "state_agi": state_agi, "county_occupation": county_occupation, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def georgia_and_rhode_island(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Georgia:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Rhode Island:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Georgia:
        - https://dor.georgia.gov/tax-tables-georgia-tax-rate-schedule
          https://apps.dor.ga.gov/FillableForms/PDFViewer/Index?form=2022GA500
          https://apps.dor.ga.gov/FillableForms/PDFViewer/Index?form=2022GA500
          https://apps.dor.ga.gov/FillableForms/PDFViewer/Index?form=2022GA500
          https://dor.georgia.gov/years-individual-income-tax-forms
          https://dor.georgia.gov/years-individual-income-tax-forms
        
        Rhode Island:
        -         
        - https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2022_1040WE_w_0.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Georgia', 'state_occupation': 'RhodeIsland', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def georgia_and_south_carolina(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Georgia:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        South Carolina:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Georgia:
        - https://dor.georgia.gov/tax-tables-georgia-tax-rate-schedule
          https://apps.dor.ga.gov/FillableForms/PDFViewer/Index?form=2022GA500
          https://apps.dor.ga.gov/FillableForms/PDFViewer/Index?form=2022GA500
          https://apps.dor.ga.gov/FillableForms/PDFViewer/Index?form=2022GA500
          https://dor.georgia.gov/years-individual-income-tax-forms
          https://dor.georgia.gov/years-individual-income-tax-forms
        
        South Carolina:
        -         
        - https://dor.sc.gov/resources-site/lawandpolicy/Advisory%20Opinions/IL22-15.pdf
          https://dor.sc.gov/forms-site/Forms/IITPacket_2022.pdf
          https://www.scstatehouse.gov/code/t12c006.php
          https://dor.sc.gov/resources-site/lawandpolicy/Advisory%20Opinions/IL22-15.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Georgia', 'state_occupation': 'SouthCarolina', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def georgia_and_south_dakota(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Georgia:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        South Dakota:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Georgia:
        - https://dor.georgia.gov/tax-tables-georgia-tax-rate-schedule
          https://apps.dor.ga.gov/FillableForms/PDFViewer/Index?form=2022GA500
          https://apps.dor.ga.gov/FillableForms/PDFViewer/Index?form=2022GA500
          https://apps.dor.ga.gov/FillableForms/PDFViewer/Index?form=2022GA500
          https://dor.georgia.gov/years-individual-income-tax-forms
          https://dor.georgia.gov/years-individual-income-tax-forms
        
        South Dakota:
        -         
        - https://dor.sd.gov/individuals/taxes/
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Georgia', 'state_occupation': 'SouthDakota', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def georgia_and_tennessee(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Georgia:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Tennessee:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Georgia:
        - https://dor.georgia.gov/tax-tables-georgia-tax-rate-schedule
          https://apps.dor.ga.gov/FillableForms/PDFViewer/Index?form=2022GA500
          https://apps.dor.ga.gov/FillableForms/PDFViewer/Index?form=2022GA500
          https://apps.dor.ga.gov/FillableForms/PDFViewer/Index?form=2022GA500
          https://dor.georgia.gov/years-individual-income-tax-forms
          https://dor.georgia.gov/years-individual-income-tax-forms
        
        Tennessee:
        -         
        - https://www.tn.gov/revenue/taxes.html
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Georgia', 'state_occupation': 'Tennessee', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def georgia_and_texas(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Georgia:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Texas:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Georgia:
        - https://dor.georgia.gov/tax-tables-georgia-tax-rate-schedule
          https://apps.dor.ga.gov/FillableForms/PDFViewer/Index?form=2022GA500
          https://apps.dor.ga.gov/FillableForms/PDFViewer/Index?form=2022GA500
          https://apps.dor.ga.gov/FillableForms/PDFViewer/Index?form=2022GA500
          https://dor.georgia.gov/years-individual-income-tax-forms
          https://dor.georgia.gov/years-individual-income-tax-forms
        
        Texas:
        -         
        - https://comptroller.texas.gov/taxes/
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Georgia', 'state_occupation': 'Texas', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def georgia_and_utah(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Georgia:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Utah:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Georgia:
        - https://dor.georgia.gov/tax-tables-georgia-tax-rate-schedule
          https://apps.dor.ga.gov/FillableForms/PDFViewer/Index?form=2022GA500
          https://apps.dor.ga.gov/FillableForms/PDFViewer/Index?form=2022GA500
          https://apps.dor.ga.gov/FillableForms/PDFViewer/Index?form=2022GA500
          https://dor.georgia.gov/years-individual-income-tax-forms
          https://dor.georgia.gov/years-individual-income-tax-forms
        
        Utah:
        -         
        - https://incometax.utah.gov/paying/tax-rates
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Georgia', 'state_occupation': 'Utah', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def georgia_and_vermont(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Georgia:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Vermont:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Georgia:
        - https://dor.georgia.gov/tax-tables-georgia-tax-rate-schedule
          https://apps.dor.ga.gov/FillableForms/PDFViewer/Index?form=2022GA500
          https://apps.dor.ga.gov/FillableForms/PDFViewer/Index?form=2022GA500
          https://apps.dor.ga.gov/FillableForms/PDFViewer/Index?form=2022GA500
          https://dor.georgia.gov/years-individual-income-tax-forms
          https://dor.georgia.gov/years-individual-income-tax-forms
        
        Vermont:
        -         
        - https://tax.vermont.gov/sites/tax/files/documents/GB-1210-2023.pdf
          https://tax.vermont.gov/sites/tax/files/documents/Income%20Booklet-2022.pdf
          https://tax.vermont.gov/sites/tax/files/documents/Income%20Booklet-2022.pdf
          https://tax.vermont.gov/sites/tax/files/documents/Income%20Booklet-2022.pdf
          https://tax.vermont.gov/individuals/personal-income-tax/tax-credits
          https://tax.vermont.gov/individuals/personal-income-tax/taxable-income
          https://tax.vermont.gov/individuals/personal-income-tax/local-option-tax
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Georgia', 'state_occupation': 'Vermont', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def georgia_and_virginia(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Georgia:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Virginia:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Georgia:
        - https://dor.georgia.gov/tax-tables-georgia-tax-rate-schedule
          https://apps.dor.ga.gov/FillableForms/PDFViewer/Index?form=2022GA500
          https://apps.dor.ga.gov/FillableForms/PDFViewer/Index?form=2022GA500
          https://apps.dor.ga.gov/FillableForms/PDFViewer/Index?form=2022GA500
          https://dor.georgia.gov/years-individual-income-tax-forms
          https://dor.georgia.gov/years-individual-income-tax-forms
        
        Virginia:
        -         
        - https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/reciprocity
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Georgia', 'state_occupation': 'Virginia', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def georgia_and_washington(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Georgia:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Washington:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Georgia:
        - https://dor.georgia.gov/tax-tables-georgia-tax-rate-schedule
          https://apps.dor.ga.gov/FillableForms/PDFViewer/Index?form=2022GA500
          https://apps.dor.ga.gov/FillableForms/PDFViewer/Index?form=2022GA500
          https://apps.dor.ga.gov/FillableForms/PDFViewer/Index?form=2022GA500
          https://dor.georgia.gov/years-individual-income-tax-forms
          https://dor.georgia.gov/years-individual-income-tax-forms
        
        Washington:
        -         
        - https://dor.wa.gov/taxes-rates/income-tax
          https://dor.wa.gov/taxes-rates/other-taxes/capital-gains-tax
          https://dor.wa.gov/taxes-rates/income-tax
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Georgia', 'state_occupation': 'Washington', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def georgia_and_west_virginia(self, pay_periods=24, capital_gains_short=0, county_occupation='Other', dependents=0, sources=False, state_agi=75000, filers_over_65=0, capital_gains_long=0, filing_status='single', income=100000, credits_state=0):
        """
        --- Parameters ----
        
        Georgia:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        West Virginia:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_occupation : str, optional
            County of occupation. Default is Other.
        pay_periods : int, optional
            Number of pay periods. Options are 13, 24, 26, 52. Default is 24.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Georgia:
        - https://dor.georgia.gov/tax-tables-georgia-tax-rate-schedule
          https://apps.dor.ga.gov/FillableForms/PDFViewer/Index?form=2022GA500
          https://apps.dor.ga.gov/FillableForms/PDFViewer/Index?form=2022GA500
          https://apps.dor.ga.gov/FillableForms/PDFViewer/Index?form=2022GA500
          https://dor.georgia.gov/years-individual-income-tax-forms
          https://dor.georgia.gov/years-individual-income-tax-forms
        
        West Virginia:
        -         
        - https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://tax.wv.gov/Documents/TSD/tsd438.pdf
          https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2011/Taxes-11-35att.html
        """
        
        # Check if the parameters are set upstream in the instance
        pay_periods = self.pay_periods if hasattr(self, 'pay_periods') else pay_periods
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        county_occupation = self.county_occupation if hasattr(self, 'county_occupation') else county_occupation
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Georgia', 'state_occupation': 'WestVirginia', "pay_periods": pay_periods, "capital_gains_short": capital_gains_short, "county_occupation": county_occupation, "dependents": dependents, "sources": sources, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def georgia_and_wisconsin(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Georgia:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Wisconsin:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Georgia:
        - https://dor.georgia.gov/tax-tables-georgia-tax-rate-schedule
          https://apps.dor.ga.gov/FillableForms/PDFViewer/Index?form=2022GA500
          https://apps.dor.ga.gov/FillableForms/PDFViewer/Index?form=2022GA500
          https://apps.dor.ga.gov/FillableForms/PDFViewer/Index?form=2022GA500
          https://dor.georgia.gov/years-individual-income-tax-forms
          https://dor.georgia.gov/years-individual-income-tax-forms
        
        Wisconsin:
        -         
        - https://www.revenue.wi.gov/TaxForms2023/2023-Form1-ES-Inst.pdf
          https://www.revenue.wi.gov/TaxForms2023/2023-Form1-ES-Inst.pdf
          https://www.revenue.wi.gov/TaxForms2023/2023-Form1-ES-Inst.pdf
          https://www.revenue.wi.gov/Pages/FAQS/pcs-work.aspx
          https://www.revenue.wi.gov/Dor%20Publications/pb103.pdf
          https://www.revenue.wi.gov/Pages/FAQS/pcs-work.aspx
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Georgia', 'state_occupation': 'Wisconsin', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def georgia_and_wyoming(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Georgia:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Wyoming:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Georgia:
        - https://dor.georgia.gov/tax-tables-georgia-tax-rate-schedule
          https://apps.dor.ga.gov/FillableForms/PDFViewer/Index?form=2022GA500
          https://apps.dor.ga.gov/FillableForms/PDFViewer/Index?form=2022GA500
          https://apps.dor.ga.gov/FillableForms/PDFViewer/Index?form=2022GA500
          https://dor.georgia.gov/years-individual-income-tax-forms
          https://dor.georgia.gov/years-individual-income-tax-forms
        
        Wyoming:
        -         
        - https://revenue.wyo.gov/
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Georgia', 'state_occupation': 'Wyoming', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def hawaii_and_alabama(self, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Hawaii:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Alabama:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Hawaii:
        - https://tax.hawaii.gov/forms/d_18table-on/d_18table-on_p13/
          https://files.hawaii.gov/tax/forms/2018/18table-on.pdf
          https://files.hawaii.gov/tax/news/pubs/22outline.pdf
          https://files.hawaii.gov/tax/news/pubs/FastTaxRefGuide_2017.pdf
          https://www.tfhawaii.org/wordpress/state-tax-resources/mini-tax-guides/individual-income-tax-chapter-235/#:~:text=Long%20term%20capital%20gains%20are%20taxed%20at%20a%20maximum%20of%207.25%25
          https://www.grassrootinstitute.org/2023/02/hb337-increasing-tax-on-capital-gains-a-bad-risk/
          https://files.hawaii.gov/tax/news/pubs/22outline.pdf
        
        Alabama:
        -         
        - https://www.revenue.alabama.gov/ultraviewer/viewer/basic_viewer/index.html?form=2023/02/22f40.pdf
          https://www.revenue.alabama.gov/faqs/what-is-alabamas-individual-income-tax-rate/
          https://www.revenue.alabama.gov/ultraviewer/viewer/basic_viewer/index.html?form=2023/02/22f40.pdf
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-24.htm?taxmap=true
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-24.htm?taxmap=true
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-24.htm?taxmap=true
          https://www.revenue.alabama.gov/individual-corporate/income-to-be-reported-on-the-alabama-income-tax-return/
          https://www.almonline.org/TaxRates.aspx#OccupationalTax
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Hawaii', 'state_occupation': 'Alabama', "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def hawaii_and_alaska(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Hawaii:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Alaska:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Hawaii:
        - https://tax.hawaii.gov/forms/d_18table-on/d_18table-on_p13/
          https://files.hawaii.gov/tax/forms/2018/18table-on.pdf
          https://files.hawaii.gov/tax/news/pubs/22outline.pdf
          https://files.hawaii.gov/tax/news/pubs/FastTaxRefGuide_2017.pdf
          https://www.tfhawaii.org/wordpress/state-tax-resources/mini-tax-guides/individual-income-tax-chapter-235/#:~:text=Long%20term%20capital%20gains%20are%20taxed%20at%20a%20maximum%20of%207.25%25
          https://www.grassrootinstitute.org/2023/02/hb337-increasing-tax-on-capital-gains-a-bad-risk/
          https://files.hawaii.gov/tax/news/pubs/22outline.pdf
        
        Alaska:
        -         
        - https://www.commerce.alaska.gov/web/dcra/officeofthestateassessor/alaskataxfacts.aspx
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Hawaii', 'state_occupation': 'Alaska', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def hawaii_and_arizona(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Hawaii:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Arizona:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        credits_state : int, optional
            Credits for state taxes. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Hawaii:
        - https://tax.hawaii.gov/forms/d_18table-on/d_18table-on_p13/
          https://files.hawaii.gov/tax/forms/2018/18table-on.pdf
          https://files.hawaii.gov/tax/news/pubs/22outline.pdf
          https://files.hawaii.gov/tax/news/pubs/FastTaxRefGuide_2017.pdf
          https://www.tfhawaii.org/wordpress/state-tax-resources/mini-tax-guides/individual-income-tax-chapter-235/#:~:text=Long%20term%20capital%20gains%20are%20taxed%20at%20a%20maximum%20of%207.25%25
          https://www.grassrootinstitute.org/2023/02/hb337-increasing-tax-on-capital-gains-a-bad-risk/
          https://files.hawaii.gov/tax/news/pubs/22outline.pdf
        
        Arizona:
        -         
        - https://azdor.gov/forms/individual/form-140-x-y-tables
          https://azdor.gov/forms/individual/form-140-x-y-tables
          https://azdor.gov/forms/individual/form-140-x-y-tables
          https://azdor.gov/businesses-arizona/withholding-tax/withholding-exceptions
          "https://azdor.gov/individuals/income-tax-filing-assistance/identifying-other-taxable-income#:~:text=In%20regards%20to%20capital%20gains
         the%20individuals%20regular%20tax%20rate.
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Hawaii', 'state_occupation': 'Arizona', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def hawaii_and_arkansas(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Hawaii:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Arkansas:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Hawaii:
        - https://tax.hawaii.gov/forms/d_18table-on/d_18table-on_p13/
          https://files.hawaii.gov/tax/forms/2018/18table-on.pdf
          https://files.hawaii.gov/tax/news/pubs/22outline.pdf
          https://files.hawaii.gov/tax/news/pubs/FastTaxRefGuide_2017.pdf
          https://www.tfhawaii.org/wordpress/state-tax-resources/mini-tax-guides/individual-income-tax-chapter-235/#:~:text=Long%20term%20capital%20gains%20are%20taxed%20at%20a%20maximum%20of%207.25%25
          https://www.grassrootinstitute.org/2023/02/hb337-increasing-tax-on-capital-gains-a-bad-risk/
          https://files.hawaii.gov/tax/news/pubs/22outline.pdf
        
        Arkansas:
        -         
        - https://www.arkansasedc.com/why-arkansas/business-climate/tax-structure/personal-income-tax
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_and_AR1000NR_Instructions.pdf
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_FullYearResidentIndividualIncomeTaxReturn.pdf
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_and_AR1000NR_Instructions.pdf
          https://www.arkansasedc.com/why-arkansas/business-climate/tax-structure/capital-gains-tax-reduction
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_and_AR1000NR_Instructions.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Hawaii', 'state_occupation': 'Arkansas', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def hawaii_and_california(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Hawaii:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        California:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Hawaii:
        - https://tax.hawaii.gov/forms/d_18table-on/d_18table-on_p13/
          https://files.hawaii.gov/tax/forms/2018/18table-on.pdf
          https://files.hawaii.gov/tax/news/pubs/22outline.pdf
          https://files.hawaii.gov/tax/news/pubs/FastTaxRefGuide_2017.pdf
          https://www.tfhawaii.org/wordpress/state-tax-resources/mini-tax-guides/individual-income-tax-chapter-235/#:~:text=Long%20term%20capital%20gains%20are%20taxed%20at%20a%20maximum%20of%207.25%25
          https://www.grassrootinstitute.org/2023/02/hb337-increasing-tax-on-capital-gains-a-bad-risk/
          https://files.hawaii.gov/tax/news/pubs/22outline.pdf
        
        California:
        -         
        - https://www.ftb.ca.gov/forms/2022/2022-540-tax-rate-schedules.pdf
          https://www.ftb.ca.gov/file/personal/deductions/index.html
          https://www.ftb.ca.gov/about-ftb/newsroom/caleitc/eligibility-and-credit-information.html
          https://www.ftb.ca.gov/file/personal/income-types/capital-gains-and-losses.html#:~:text=California%20does%20not%20have%20a
         are%20taxed%20as%20ordinary%20income
          https://www.ftb.ca.gov/file/personal/
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Hawaii', 'state_occupation': 'California', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def hawaii_and_colorado(self, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Hawaii:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Colorado:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Hawaii:
        - https://tax.hawaii.gov/forms/d_18table-on/d_18table-on_p13/
          https://files.hawaii.gov/tax/forms/2018/18table-on.pdf
          https://files.hawaii.gov/tax/news/pubs/22outline.pdf
          https://files.hawaii.gov/tax/news/pubs/FastTaxRefGuide_2017.pdf
          https://www.tfhawaii.org/wordpress/state-tax-resources/mini-tax-guides/individual-income-tax-chapter-235/#:~:text=Long%20term%20capital%20gains%20are%20taxed%20at%20a%20maximum%20of%207.25%25
          https://www.grassrootinstitute.org/2023/02/hb337-increasing-tax-on-capital-gains-a-bad-risk/
          https://files.hawaii.gov/tax/news/pubs/22outline.pdf
        
        Colorado:
        -         
        - https://leg.colorado.gov/agencies/legislative-council-staff/individual-income-tax%C2%A0
          https://leg.colorado.gov/agencies/legislative-council-staff/individual-income-tax%C2%A0
          https://tax.colorado.gov/sites/tax/files/documents/ITT_Capital_Gain_Subtraction_Feb_2022.pdf
          https://www.auroragov.org/business_services/taxes/occupational_privilege_tax
          https://www.denvergov.org/Government/Agencies-Departments-Offices/Agencies-Departments-Offices-Directory/Department-of-Finance/Our-Divisions/Treasury/Business-Tax-Information#:~:text=Occupational%20privilege%20tax
         -%28show%20below%29&text=Each%20taxable%20employee%20is%20liable
         month%20for%20each%20taxable%20employee
          https://www.glendale.co.us/355/Occupational-Privilege-Tax
          https://www.greenwoodvillage.com/1220/Occupational-Privilege-Tax-OPT
          https://ci.sheridan.co.us/288/Occupational-Privilege-Tax
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Hawaii', 'state_occupation': 'Colorado', "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def hawaii_and_connecticut(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Hawaii:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Connecticut:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Hawaii:
        - https://tax.hawaii.gov/forms/d_18table-on/d_18table-on_p13/
          https://files.hawaii.gov/tax/forms/2018/18table-on.pdf
          https://files.hawaii.gov/tax/news/pubs/22outline.pdf
          https://files.hawaii.gov/tax/news/pubs/FastTaxRefGuide_2017.pdf
          https://www.tfhawaii.org/wordpress/state-tax-resources/mini-tax-guides/individual-income-tax-chapter-235/#:~:text=Long%20term%20capital%20gains%20are%20taxed%20at%20a%20maximum%20of%207.25%25
          https://www.grassrootinstitute.org/2023/02/hb337-increasing-tax-on-capital-gains-a-bad-risk/
          https://files.hawaii.gov/tax/news/pubs/22outline.pdf
        
        Connecticut:
        -         
        - https://portal.ct.gov/-/media/DRS/Forms/2022/Income/2022-CT-1040-Instructions_1222.pdf
          https://www.cga.ct.gov/2022/rpt/pdf/2022-R-0108.pdf
          https://www.cga.ct.gov/current/pub/chap_229.htm#sec_12-703
          https://portal.ct.gov/-/media/DRS/Forms/2022/Income/2022-CT-1040-Instructions_1222.pdf
          "https://portal.ct.gov/DRS/Publications/TSSNs/TSSN-29#:~:text=An%20individuals%20net%20capital%20gains
         portion%20of%20Social%20Security%20benefits"
          "https://portal.ct.gov/DRS/Publications/TSSNs/TSSN-29#:~:text=An%20individuals%20net%20capital%20gains
         portion%20of%20Social%20Security%20benefits"
          https://portal.ct.gov/DRS/Individuals/Resident-Income-Tax/Tax-Information
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Hawaii', 'state_occupation': 'Connecticut', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def hawaii_and_delaware(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, federal_agi=75000, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Hawaii:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Delaware:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        federal_agi : int, optional
            Adjusted Gross Income (AGI) for federal taxes. Default is 0. Sources - https://www.irs.gov/e-file-providers/definition-of-adjusted-gross-income, https://www.irs.gov/newsroom/taxpayers-should-know-the-difference-between-standard-and-itemized-deductions
        income : int, optional
            Gross income. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Hawaii:
        - https://tax.hawaii.gov/forms/d_18table-on/d_18table-on_p13/
          https://files.hawaii.gov/tax/forms/2018/18table-on.pdf
          https://files.hawaii.gov/tax/news/pubs/22outline.pdf
          https://files.hawaii.gov/tax/news/pubs/FastTaxRefGuide_2017.pdf
          https://www.tfhawaii.org/wordpress/state-tax-resources/mini-tax-guides/individual-income-tax-chapter-235/#:~:text=Long%20term%20capital%20gains%20are%20taxed%20at%20a%20maximum%20of%207.25%25
          https://www.grassrootinstitute.org/2023/02/hb337-increasing-tax-on-capital-gains-a-bad-risk/
          https://files.hawaii.gov/tax/news/pubs/22outline.pdf
        
        Delaware:
        -         
        - https://revenue.delaware.gov/software-developer/tax-rate-changes/
          https://revenuefiles.delaware.gov/2022/PIT-RES_TY22_2022-01_PaperInteractive.pdf
          https://revenuefiles.delaware.gov/2022/PIT-RES_TY22_2022-02_Instructions.pdf
          https://revenue.delaware.gov/frequently-asked-questions/personal-income-tax-faqs/
          https://www.newcastlede.gov/DocumentCenter/View/20147/City-of-Wilmington-Wage-Tax-Information-PDF
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        federal_agi = self.federal_agi if hasattr(self, 'federal_agi') else federal_agi
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Hawaii', 'state_occupation': 'Delaware', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "federal_agi": federal_agi, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def hawaii_and_florida(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Hawaii:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Florida:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Hawaii:
        - https://tax.hawaii.gov/forms/d_18table-on/d_18table-on_p13/
          https://files.hawaii.gov/tax/forms/2018/18table-on.pdf
          https://files.hawaii.gov/tax/news/pubs/22outline.pdf
          https://files.hawaii.gov/tax/news/pubs/FastTaxRefGuide_2017.pdf
          https://www.tfhawaii.org/wordpress/state-tax-resources/mini-tax-guides/individual-income-tax-chapter-235/#:~:text=Long%20term%20capital%20gains%20are%20taxed%20at%20a%20maximum%20of%207.25%25
          https://www.grassrootinstitute.org/2023/02/hb337-increasing-tax-on-capital-gains-a-bad-risk/
          https://files.hawaii.gov/tax/news/pubs/22outline.pdf
        
        Florida:
        -         
        - https://floridarevenue.com/taxes/taxesfees/Pages/tax_interest_rates.aspx
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Hawaii', 'state_occupation': 'Florida', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def hawaii_and_georgia(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Hawaii:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Georgia:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Hawaii:
        - https://tax.hawaii.gov/forms/d_18table-on/d_18table-on_p13/
          https://files.hawaii.gov/tax/forms/2018/18table-on.pdf
          https://files.hawaii.gov/tax/news/pubs/22outline.pdf
          https://files.hawaii.gov/tax/news/pubs/FastTaxRefGuide_2017.pdf
          https://www.tfhawaii.org/wordpress/state-tax-resources/mini-tax-guides/individual-income-tax-chapter-235/#:~:text=Long%20term%20capital%20gains%20are%20taxed%20at%20a%20maximum%20of%207.25%25
          https://www.grassrootinstitute.org/2023/02/hb337-increasing-tax-on-capital-gains-a-bad-risk/
          https://files.hawaii.gov/tax/news/pubs/22outline.pdf
        
        Georgia:
        -         
        - https://dor.georgia.gov/tax-tables-georgia-tax-rate-schedule
          https://apps.dor.ga.gov/FillableForms/PDFViewer/Index?form=2022GA500
          https://apps.dor.ga.gov/FillableForms/PDFViewer/Index?form=2022GA500
          https://apps.dor.ga.gov/FillableForms/PDFViewer/Index?form=2022GA500
          https://dor.georgia.gov/years-individual-income-tax-forms
          https://dor.georgia.gov/years-individual-income-tax-forms
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Hawaii', 'state_occupation': 'Georgia', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def hawaii_and_idaho(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Hawaii:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Idaho:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Hawaii:
        - https://tax.hawaii.gov/forms/d_18table-on/d_18table-on_p13/
          https://files.hawaii.gov/tax/forms/2018/18table-on.pdf
          https://files.hawaii.gov/tax/news/pubs/22outline.pdf
          https://files.hawaii.gov/tax/news/pubs/FastTaxRefGuide_2017.pdf
          https://www.tfhawaii.org/wordpress/state-tax-resources/mini-tax-guides/individual-income-tax-chapter-235/#:~:text=Long%20term%20capital%20gains%20are%20taxed%20at%20a%20maximum%20of%207.25%25
          https://www.grassrootinstitute.org/2023/02/hb337-increasing-tax-on-capital-gains-a-bad-risk/
          https://files.hawaii.gov/tax/news/pubs/22outline.pdf
        
        Idaho:
        -         
        - https://tax.idaho.gov/taxes/income-tax/individual-income/individual-income-tax-rate-schedule/
          https://tax.idaho.gov/wp-content/uploads/forms/EIN00046/EIN00046_03-01-2023.pdf
          https://tax.idaho.gov/taxes/income-tax/individual-income/idaho-source-income/
          https://tax.idaho.gov/taxes/income-tax/individual-income/idaho-source-income/
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Hawaii', 'state_occupation': 'Idaho', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def hawaii_and_illinois(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Hawaii:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Illinois:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Hawaii:
        - https://tax.hawaii.gov/forms/d_18table-on/d_18table-on_p13/
          https://files.hawaii.gov/tax/forms/2018/18table-on.pdf
          https://files.hawaii.gov/tax/news/pubs/22outline.pdf
          https://files.hawaii.gov/tax/news/pubs/FastTaxRefGuide_2017.pdf
          https://www.tfhawaii.org/wordpress/state-tax-resources/mini-tax-guides/individual-income-tax-chapter-235/#:~:text=Long%20term%20capital%20gains%20are%20taxed%20at%20a%20maximum%20of%207.25%25
          https://www.grassrootinstitute.org/2023/02/hb337-increasing-tax-on-capital-gains-a-bad-risk/
          https://files.hawaii.gov/tax/news/pubs/22outline.pdf
        
        Illinois:
        -         
        - https://tax.illinois.gov/research/taxrates/income.html#IndividualIncome
          https://tax.illinois.gov/content/dam/soi/en/web/tax/forms/incometax/documents/currentyear/individual/il-1040-instr.pdf
          https://tax.illinois.gov/questionsandanswers/12.html
          https://tax.illinois.gov/content/dam/soi/en/web/tax/forms/incometax/documents/currentyear/individual/il-1040-instr.pdf
          https://tax.illinois.gov/research/taxrates/personalproperty.html
          https://tax.illinois.gov/localgovernments/income.html
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Hawaii', 'state_occupation': 'Illinois', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def hawaii_and_indiana(self, capital_gains_short=0, dependents=0, sources=False, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_long=0, filing_status='single', income=100000, credits_state=0):
        """
        --- Parameters ----
        
        Hawaii:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Indiana:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Hawaii:
        - https://tax.hawaii.gov/forms/d_18table-on/d_18table-on_p13/
          https://files.hawaii.gov/tax/forms/2018/18table-on.pdf
          https://files.hawaii.gov/tax/news/pubs/22outline.pdf
          https://files.hawaii.gov/tax/news/pubs/FastTaxRefGuide_2017.pdf
          https://www.tfhawaii.org/wordpress/state-tax-resources/mini-tax-guides/individual-income-tax-chapter-235/#:~:text=Long%20term%20capital%20gains%20are%20taxed%20at%20a%20maximum%20of%207.25%25
          https://www.grassrootinstitute.org/2023/02/hb337-increasing-tax-on-capital-gains-a-bad-risk/
          https://files.hawaii.gov/tax/news/pubs/22outline.pdf
        
        Indiana:
        -         
        - https://www.in.gov/dor/tax-forms/2022-individual-income-tax-forms/
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-32.htm
          https://www.in.gov/dor/files/ib28.pdf
          https://www.in.gov/dor/tax-forms/2022-individual-income-tax-forms/
          https://www.in.gov/dor/files/2022-county-tax-rates-and-codes.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Hawaii', 'state_occupation': 'Indiana', "capital_gains_short": capital_gains_short, "dependents": dependents, "sources": sources, "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def hawaii_and_iowa(self, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Hawaii:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Iowa:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Hawaii:
        - https://tax.hawaii.gov/forms/d_18table-on/d_18table-on_p13/
          https://files.hawaii.gov/tax/forms/2018/18table-on.pdf
          https://files.hawaii.gov/tax/news/pubs/22outline.pdf
          https://files.hawaii.gov/tax/news/pubs/FastTaxRefGuide_2017.pdf
          https://www.tfhawaii.org/wordpress/state-tax-resources/mini-tax-guides/individual-income-tax-chapter-235/#:~:text=Long%20term%20capital%20gains%20are%20taxed%20at%20a%20maximum%20of%207.25%25
          https://www.grassrootinstitute.org/2023/02/hb337-increasing-tax-on-capital-gains-a-bad-risk/
          https://files.hawaii.gov/tax/news/pubs/22outline.pdf
        
        Iowa:
        -         
        - https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
          https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
          https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
          https://tax.iowa.gov/iowa-illinois-reciprocal-agreement
          https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Hawaii', 'state_occupation': 'Iowa', "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def hawaii_and_kansas(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Hawaii:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Kansas:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Hawaii:
        - https://tax.hawaii.gov/forms/d_18table-on/d_18table-on_p13/
          https://files.hawaii.gov/tax/forms/2018/18table-on.pdf
          https://files.hawaii.gov/tax/news/pubs/22outline.pdf
          https://files.hawaii.gov/tax/news/pubs/FastTaxRefGuide_2017.pdf
          https://www.tfhawaii.org/wordpress/state-tax-resources/mini-tax-guides/individual-income-tax-chapter-235/#:~:text=Long%20term%20capital%20gains%20are%20taxed%20at%20a%20maximum%20of%207.25%25
          https://www.grassrootinstitute.org/2023/02/hb337-increasing-tax-on-capital-gains-a-bad-risk/
          https://files.hawaii.gov/tax/news/pubs/22outline.pdf
        
        Kansas:
        -         
        - https://www.ksrevenue.gov/taxrates.html
          https://www.ksrevenue.gov/incomebook22.html#0
          https://www.ksrevenue.gov/incomebook22.html#0
          https://www.ksrevenue.gov/incomebook22.html#0
          https://www.ksrevenue.gov/incomebook22.html
          https://www.ksrevenue.gov/incomebook22.html
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Hawaii', 'state_occupation': 'Kansas', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def hawaii_and_kentucky(self, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Hawaii:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Kentucky:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Hawaii:
        - https://tax.hawaii.gov/forms/d_18table-on/d_18table-on_p13/
          https://files.hawaii.gov/tax/forms/2018/18table-on.pdf
          https://files.hawaii.gov/tax/news/pubs/22outline.pdf
          https://files.hawaii.gov/tax/news/pubs/FastTaxRefGuide_2017.pdf
          https://www.tfhawaii.org/wordpress/state-tax-resources/mini-tax-guides/individual-income-tax-chapter-235/#:~:text=Long%20term%20capital%20gains%20are%20taxed%20at%20a%20maximum%20of%207.25%25
          https://www.grassrootinstitute.org/2023/02/hb337-increasing-tax-on-capital-gains-a-bad-risk/
          https://files.hawaii.gov/tax/news/pubs/22outline.pdf
        
        Kentucky:
        -         
        - https://revenue.ky.gov/News/Pages/Dor-Announces-Updates-to-Individual-Income-Tax-for-2023-Tax-Year.aspx
          https://revenue.ky.gov/Individual/Individual-Income-Tax/Pages/default.aspx
          https://revenue.ky.gov/News/Pages/Dor-Announces-Updates-to-Individual-Income-Tax-for-2023-Tax-Year.aspx
          https://revenue.ky.gov/Individual/Individual-Income-Tax/Pages/default.aspx
          https://revenue.ky.gov/Dor%20Training%20Materials/103%20KAR%2017.140.%20Individual%20income%20tax%20-%20reciprocity%20-%20nonresidents.pdf
          https://revenue.ky.gov/Individual/Individual-Income-Tax/Pages/default.aspx
          https://nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2021/Taxes-21-32.htm?taxmap=true
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Hawaii', 'state_occupation': 'Kentucky', "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def hawaii_and_louisiana(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Hawaii:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Louisiana:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Hawaii:
        - https://tax.hawaii.gov/forms/d_18table-on/d_18table-on_p13/
          https://files.hawaii.gov/tax/forms/2018/18table-on.pdf
          https://files.hawaii.gov/tax/news/pubs/22outline.pdf
          https://files.hawaii.gov/tax/news/pubs/FastTaxRefGuide_2017.pdf
          https://www.tfhawaii.org/wordpress/state-tax-resources/mini-tax-guides/individual-income-tax-chapter-235/#:~:text=Long%20term%20capital%20gains%20are%20taxed%20at%20a%20maximum%20of%207.25%25
          https://www.grassrootinstitute.org/2023/02/hb337-increasing-tax-on-capital-gains-a-bad-risk/
          https://files.hawaii.gov/tax/news/pubs/22outline.pdf
        
        Louisiana:
        -         
        - https://revenue.louisiana.gov/individualincometax
          https://revenue.louisiana.gov/individualincometax
          https://www.revenue.louisiana.gov/taxforms/6935(11_02)F.pdf
          https://revenue.louisiana.gov/individualincometax
          https://revenue.louisiana.gov/individualincometax
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Hawaii', 'state_occupation': 'Louisiana', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def hawaii_and_maine(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Hawaii:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Maine:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Hawaii:
        - https://tax.hawaii.gov/forms/d_18table-on/d_18table-on_p13/
          https://files.hawaii.gov/tax/forms/2018/18table-on.pdf
          https://files.hawaii.gov/tax/news/pubs/22outline.pdf
          https://files.hawaii.gov/tax/news/pubs/FastTaxRefGuide_2017.pdf
          https://www.tfhawaii.org/wordpress/state-tax-resources/mini-tax-guides/individual-income-tax-chapter-235/#:~:text=Long%20term%20capital%20gains%20are%20taxed%20at%20a%20maximum%20of%207.25%25
          https://www.grassrootinstitute.org/2023/02/hb337-increasing-tax-on-capital-gains-a-bad-risk/
          https://files.hawaii.gov/tax/news/pubs/22outline.pdf
        
        Maine:
        -         
        - https://www.maine.gov/revenue/tax-return-forms/individual-income-tax-2022
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/ind_tax_rate_sched_2022.pdf
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/22_1040me_book_gen_instruc_revisedFeb23.pdf
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/22_1040me_book_gen_instruc_revisedFeb23.pdf
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/22_1040me_book_gen_instruc_revisedFeb23.pdf
          https://www.maine.gov/revenue/tax-return-forms/individual-income-tax-2022
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Hawaii', 'state_occupation': 'Maine', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def hawaii_and_maryland(self, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Hawaii:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Maryland:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Hawaii:
        - https://tax.hawaii.gov/forms/d_18table-on/d_18table-on_p13/
          https://files.hawaii.gov/tax/forms/2018/18table-on.pdf
          https://files.hawaii.gov/tax/news/pubs/22outline.pdf
          https://files.hawaii.gov/tax/news/pubs/FastTaxRefGuide_2017.pdf
          https://www.tfhawaii.org/wordpress/state-tax-resources/mini-tax-guides/individual-income-tax-chapter-235/#:~:text=Long%20term%20capital%20gains%20are%20taxed%20at%20a%20maximum%20of%207.25%25
          https://www.grassrootinstitute.org/2023/02/hb337-increasing-tax-on-capital-gains-a-bad-risk/
          https://files.hawaii.gov/tax/news/pubs/22outline.pdf
        
        Maryland:
        -         
        - https://www.marylandtaxes.gov/individual/income/tax-info/tax-rates.php
          https://www.marylandtaxes.gov/forms/22_forms/Resident_Booklet.pdf
          https://www.marylandtaxes.gov/forms/Tax_Publications/Administrative_Releases/Income_and_Estate_Tax_Releases/ar_it37.pdf
          https://www.marylandtaxes.gov/individual/index.php
          https://www.marylandtaxes.gov/individual/income/tax-info/tax-rates.php
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Hawaii', 'state_occupation': 'Maryland', "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def hawaii_and_massachusetts(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Hawaii:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Massachusetts:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Hawaii:
        - https://tax.hawaii.gov/forms/d_18table-on/d_18table-on_p13/
          https://files.hawaii.gov/tax/forms/2018/18table-on.pdf
          https://files.hawaii.gov/tax/news/pubs/22outline.pdf
          https://files.hawaii.gov/tax/news/pubs/FastTaxRefGuide_2017.pdf
          https://www.tfhawaii.org/wordpress/state-tax-resources/mini-tax-guides/individual-income-tax-chapter-235/#:~:text=Long%20term%20capital%20gains%20are%20taxed%20at%20a%20maximum%20of%207.25%25
          https://www.grassrootinstitute.org/2023/02/hb337-increasing-tax-on-capital-gains-a-bad-risk/
          https://files.hawaii.gov/tax/news/pubs/22outline.pdf
        
        Massachusetts:
        -         
        - https://www.mass.gov/service-details/massachusetts-tax-rates
          https://www.mass.gov/service-details/view-massachusetts-personal-income-tax-exemptions
          https://www.mass.gov/service-details/view-massachusetts-personal-income-tax-exemptions
          https://www.mass.gov/service-details/view-child-and-dependent-related-credits
          https://www.mass.gov/service-details/massachusetts-tax-rates
          https://www.mass.gov/personal-income-tax
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Hawaii', 'state_occupation': 'Massachusetts', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def hawaii_and_michigan(self, capital_gains_short=0, county_occupation='Other', dependents=0, sources=False, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_long=0, filing_status='single', income=100000, credits_state=0):
        """
        --- Parameters ----
        
        Hawaii:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Michigan:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        county_occupation : str, optional
            County of occupation. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Hawaii:
        - https://tax.hawaii.gov/forms/d_18table-on/d_18table-on_p13/
          https://files.hawaii.gov/tax/forms/2018/18table-on.pdf
          https://files.hawaii.gov/tax/news/pubs/22outline.pdf
          https://files.hawaii.gov/tax/news/pubs/FastTaxRefGuide_2017.pdf
          https://www.tfhawaii.org/wordpress/state-tax-resources/mini-tax-guides/individual-income-tax-chapter-235/#:~:text=Long%20term%20capital%20gains%20are%20taxed%20at%20a%20maximum%20of%207.25%25
          https://www.grassrootinstitute.org/2023/02/hb337-increasing-tax-on-capital-gains-a-bad-risk/
          https://files.hawaii.gov/tax/news/pubs/22outline.pdf
        
        Michigan:
        -         
        - https://www.michigan.gov/taxes/iit/new-developments-for-tax-year-2022
          "https://www.michigan.gov/taxes/-/media/Project/Websites/taxes/Forms/2023/2023-SUW/446_Withholding-Guide_2023.pdf?rev=177ca3d8cb034392bf84e317ed78c320&hash=19C5F42F43A89A91863D229349F48883#:~:text=The%20exemption%20amount%20is%20%245%2C400
         employees%20Michigan%20income%20tax%20return"
          "https://www.michigan.gov/taxes/-/media/Project/Websites/taxes/Forms/2023/2023-SUW/446_Withholding-Guide_2023.pdf?rev=177ca3d8cb034392bf84e317ed78c320&hash=19C5F42F43A89A91863D229349F48883#:~:text=The%20exemption%20amount%20is%20%245%2C400
         employees%20Michigan%20income%20tax%20return"
          https://www.michigan.gov/taxes/questions/iit/accordion/residency/are-my-wages-earned-in-another-state-taxable-in-michigan-if-i-am-a-michigan-resident-1
          https://www.michigan.gov/taxes/-/media/Project/Websites/taxes/Forms/2022/2022-IIT-Forms/BOOK_MI-1040.pdf?rev=4524f1cbbc0243b194ab05a6680be75b&hash=40DFE71E651F7C1D65C31F6CD03D944F
          https://www.michigan.gov/taxes/questions/iit/accordion/general/what-cities-impose-an-income-tax
        """
        
        # Check if the parameters are set upstream in the instance
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        county_occupation = self.county_occupation if hasattr(self, 'county_occupation') else county_occupation
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Hawaii', 'state_occupation': 'Michigan', "capital_gains_short": capital_gains_short, "county_occupation": county_occupation, "dependents": dependents, "sources": sources, "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def hawaii_and_minnesota(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Hawaii:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Minnesota:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Hawaii:
        - https://tax.hawaii.gov/forms/d_18table-on/d_18table-on_p13/
          https://files.hawaii.gov/tax/forms/2018/18table-on.pdf
          https://files.hawaii.gov/tax/news/pubs/22outline.pdf
          https://files.hawaii.gov/tax/news/pubs/FastTaxRefGuide_2017.pdf
          https://www.tfhawaii.org/wordpress/state-tax-resources/mini-tax-guides/individual-income-tax-chapter-235/#:~:text=Long%20term%20capital%20gains%20are%20taxed%20at%20a%20maximum%20of%207.25%25
          https://www.grassrootinstitute.org/2023/02/hb337-increasing-tax-on-capital-gains-a-bad-risk/
          https://files.hawaii.gov/tax/news/pubs/22outline.pdf
        
        Minnesota:
        -         
        - https://www.revenue.state.mn.us/minnesota-income-tax-rates-and-brackets
          https://www.revenue.state.mn.us/child-and-dependent-care-credit
          https://www.revenue.state.mn.us/dependent-exemptions
          https://www.revenue.state.mn.us/minnesota-standard-deduction
          https://mn.gov/mmb/assets/mwr_form_tcm1059-128581.pdf
          https://www.revenue.state.mn.us/individual-income-tax
          https://www.revenue.state.mn.us/individual-income-tax
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Hawaii', 'state_occupation': 'Minnesota', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def hawaii_and_mississippi(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Hawaii:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Mississippi:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Hawaii:
        - https://tax.hawaii.gov/forms/d_18table-on/d_18table-on_p13/
          https://files.hawaii.gov/tax/forms/2018/18table-on.pdf
          https://files.hawaii.gov/tax/news/pubs/22outline.pdf
          https://files.hawaii.gov/tax/news/pubs/FastTaxRefGuide_2017.pdf
          https://www.tfhawaii.org/wordpress/state-tax-resources/mini-tax-guides/individual-income-tax-chapter-235/#:~:text=Long%20term%20capital%20gains%20are%20taxed%20at%20a%20maximum%20of%207.25%25
          https://www.grassrootinstitute.org/2023/02/hb337-increasing-tax-on-capital-gains-a-bad-risk/
          https://files.hawaii.gov/tax/news/pubs/22outline.pdf
        
        Mississippi:
        -         
        - https://www.dor.ms.gov/individual/tax-rates
          https://www.dor.ms.gov/individual/tax-rates
          https://www.dor.ms.gov/individual/tax-rates
          https://www.dor.ms.gov/individual
          https://www.dor.ms.gov/individual
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Hawaii', 'state_occupation': 'Mississippi', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def hawaii_and_missouri(self, capital_gains_short=0, dependents=0, sources=False, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_long=0, filing_status='single', income=100000, credits_state=0):
        """
        --- Parameters ----
        
        Hawaii:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Missouri:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Hawaii:
        - https://tax.hawaii.gov/forms/d_18table-on/d_18table-on_p13/
          https://files.hawaii.gov/tax/forms/2018/18table-on.pdf
          https://files.hawaii.gov/tax/news/pubs/22outline.pdf
          https://files.hawaii.gov/tax/news/pubs/FastTaxRefGuide_2017.pdf
          https://www.tfhawaii.org/wordpress/state-tax-resources/mini-tax-guides/individual-income-tax-chapter-235/#:~:text=Long%20term%20capital%20gains%20are%20taxed%20at%20a%20maximum%20of%207.25%25
          https://www.grassrootinstitute.org/2023/02/hb337-increasing-tax-on-capital-gains-a-bad-risk/
          https://files.hawaii.gov/tax/news/pubs/22outline.pdf
        
        Missouri:
        -         
        - https://dor.mo.gov/taxation/individual/tax-types/income/year-changes/
          https://dor.mo.gov/taxation/individual/tax-types/income/year-changes/
          https://dor.mo.gov/taxation/individual/tax-types/income/year-changes/
          https://www.kcmo.gov/city-hall/departments/finance/earnings-tax
          https://www.stlouis-mo.gov/government/departments/collector/earnings-tax/individual-earnings-tax-info.cfm
        """
        
        # Check if the parameters are set upstream in the instance
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Hawaii', 'state_occupation': 'Missouri', "capital_gains_short": capital_gains_short, "dependents": dependents, "sources": sources, "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def hawaii_and_montana(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Hawaii:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Montana:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Hawaii:
        - https://tax.hawaii.gov/forms/d_18table-on/d_18table-on_p13/
          https://files.hawaii.gov/tax/forms/2018/18table-on.pdf
          https://files.hawaii.gov/tax/news/pubs/22outline.pdf
          https://files.hawaii.gov/tax/news/pubs/FastTaxRefGuide_2017.pdf
          https://www.tfhawaii.org/wordpress/state-tax-resources/mini-tax-guides/individual-income-tax-chapter-235/#:~:text=Long%20term%20capital%20gains%20are%20taxed%20at%20a%20maximum%20of%207.25%25
          https://www.grassrootinstitute.org/2023/02/hb337-increasing-tax-on-capital-gains-a-bad-risk/
          https://files.hawaii.gov/tax/news/pubs/22outline.pdf
        
        Montana:
        -         
        - https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/2022/12/montana-employees-withholding-allowance-and-exemption-certificate-form-mw-4.pdf
          https://mtrevenue.gov/2021/11/29/simplification-of-montana-income-taxation/?utm_source=rss&utm_medium=rss&utm_campaign=simplification-of-montana-income-taxation
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Hawaii', 'state_occupation': 'Montana', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def hawaii_and_nebraska(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Hawaii:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Nebraska:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Hawaii:
        - https://tax.hawaii.gov/forms/d_18table-on/d_18table-on_p13/
          https://files.hawaii.gov/tax/forms/2018/18table-on.pdf
          https://files.hawaii.gov/tax/news/pubs/22outline.pdf
          https://files.hawaii.gov/tax/news/pubs/FastTaxRefGuide_2017.pdf
          https://www.tfhawaii.org/wordpress/state-tax-resources/mini-tax-guides/individual-income-tax-chapter-235/#:~:text=Long%20term%20capital%20gains%20are%20taxed%20at%20a%20maximum%20of%207.25%25
          https://www.grassrootinstitute.org/2023/02/hb337-increasing-tax-on-capital-gains-a-bad-risk/
          https://files.hawaii.gov/tax/news/pubs/22outline.pdf
        
        Nebraska:
        -         
        - https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2023_Circular_EN_8-429r2023_Whole_Book_11.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Hawaii', 'state_occupation': 'Nebraska', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def hawaii_and_nevada(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Hawaii:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Nevada:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Hawaii:
        - https://tax.hawaii.gov/forms/d_18table-on/d_18table-on_p13/
          https://files.hawaii.gov/tax/forms/2018/18table-on.pdf
          https://files.hawaii.gov/tax/news/pubs/22outline.pdf
          https://files.hawaii.gov/tax/news/pubs/FastTaxRefGuide_2017.pdf
          https://www.tfhawaii.org/wordpress/state-tax-resources/mini-tax-guides/individual-income-tax-chapter-235/#:~:text=Long%20term%20capital%20gains%20are%20taxed%20at%20a%20maximum%20of%207.25%25
          https://www.grassrootinstitute.org/2023/02/hb337-increasing-tax-on-capital-gains-a-bad-risk/
          https://files.hawaii.gov/tax/news/pubs/22outline.pdf
        
        Nevada:
        -         
        - https://www.nvsos.gov/sos/businesses/the-nevada-advantage
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Hawaii', 'state_occupation': 'Nevada', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def hawaii_and_new_hampshire(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Hawaii:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        New Hampshire:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Hawaii:
        - https://tax.hawaii.gov/forms/d_18table-on/d_18table-on_p13/
          https://files.hawaii.gov/tax/forms/2018/18table-on.pdf
          https://files.hawaii.gov/tax/news/pubs/22outline.pdf
          https://files.hawaii.gov/tax/news/pubs/FastTaxRefGuide_2017.pdf
          https://www.tfhawaii.org/wordpress/state-tax-resources/mini-tax-guides/individual-income-tax-chapter-235/#:~:text=Long%20term%20capital%20gains%20are%20taxed%20at%20a%20maximum%20of%207.25%25
          https://www.grassrootinstitute.org/2023/02/hb337-increasing-tax-on-capital-gains-a-bad-risk/
          https://files.hawaii.gov/tax/news/pubs/22outline.pdf
        
        New Hampshire:
        -         
        - https://www.revenue.nh.gov/assistance/tax-overview.htm
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Hawaii', 'state_occupation': 'NewHampshire', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def hawaii_and_new_jersey(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Hawaii:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        New Jersey:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Hawaii:
        - https://tax.hawaii.gov/forms/d_18table-on/d_18table-on_p13/
          https://files.hawaii.gov/tax/forms/2018/18table-on.pdf
          https://files.hawaii.gov/tax/news/pubs/22outline.pdf
          https://files.hawaii.gov/tax/news/pubs/FastTaxRefGuide_2017.pdf
          https://www.tfhawaii.org/wordpress/state-tax-resources/mini-tax-guides/individual-income-tax-chapter-235/#:~:text=Long%20term%20capital%20gains%20are%20taxed%20at%20a%20maximum%20of%207.25%25
          https://www.grassrootinstitute.org/2023/02/hb337-increasing-tax-on-capital-gains-a-bad-risk/
          https://files.hawaii.gov/tax/news/pubs/22outline.pdf
        
        New Jersey:
        -         
        - https://www.state.nj.us/treasury/taxation/taxtables.shtml
          https://www.state.nj.us/treasury/taxation/njit13.shtml
          https://www.nj.gov/treasury/taxation/njit25.shtml
          https://www.state.nj.us/treasury/taxation/individuals/freqqiti.shtml
          https://www.state.nj.us/treasury/taxation/localtax.shtml
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Hawaii', 'state_occupation': 'NewJersey', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def hawaii_and_new_mexico(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Hawaii:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        New Mexico:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Hawaii:
        - https://tax.hawaii.gov/forms/d_18table-on/d_18table-on_p13/
          https://files.hawaii.gov/tax/forms/2018/18table-on.pdf
          https://files.hawaii.gov/tax/news/pubs/22outline.pdf
          https://files.hawaii.gov/tax/news/pubs/FastTaxRefGuide_2017.pdf
          https://www.tfhawaii.org/wordpress/state-tax-resources/mini-tax-guides/individual-income-tax-chapter-235/#:~:text=Long%20term%20capital%20gains%20are%20taxed%20at%20a%20maximum%20of%207.25%25
          https://www.grassrootinstitute.org/2023/02/hb337-increasing-tax-on-capital-gains-a-bad-risk/
          https://files.hawaii.gov/tax/news/pubs/22outline.pdf
        
        New Mexico:
        -         
        - https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/fdf3c548-8aba-4b9c-9eb4-bb564c716015/FYI-104.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/0bf943fd-652e-4400-bb1b-cd8397eb5a95/2022pit-adj.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/2f1a6781-9534-4436-b427-1557f9592099/2022pit-adj-ins.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/0bf943fd-652e-4400-bb1b-cd8397eb5a95/2022pit-adj.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/1afc56af-ea90-4d48-82e5-1f9aeb43255a/PITbook2022.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/fdf3c548-8aba-4b9c-9eb4-bb564c716015/FYI-104.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Hawaii', 'state_occupation': 'NewMexico', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def hawaii_and_new_york(self, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Hawaii:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        New York:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Hawaii:
        - https://tax.hawaii.gov/forms/d_18table-on/d_18table-on_p13/
          https://files.hawaii.gov/tax/forms/2018/18table-on.pdf
          https://files.hawaii.gov/tax/news/pubs/22outline.pdf
          https://files.hawaii.gov/tax/news/pubs/FastTaxRefGuide_2017.pdf
          https://www.tfhawaii.org/wordpress/state-tax-resources/mini-tax-guides/individual-income-tax-chapter-235/#:~:text=Long%20term%20capital%20gains%20are%20taxed%20at%20a%20maximum%20of%207.25%25
          https://www.grassrootinstitute.org/2023/02/hb337-increasing-tax-on-capital-gains-a-bad-risk/
          https://files.hawaii.gov/tax/news/pubs/22outline.pdf
        
        New York:
        -         
        - https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nys-tax-rate-schedule
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nys-tax-rate-schedule
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nys-tax-rate-schedule
          https://www.tax.ny.gov/pit/
          https://www.tax.ny.gov/pit/file/nonresident-faqs.htm
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#yonkers-tax-rate-schedule
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nyc-tax-rate-schedule
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Hawaii', 'state_occupation': 'NewYork', "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def hawaii_and_north_carolina(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Hawaii:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        North Carolina:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Hawaii:
        - https://tax.hawaii.gov/forms/d_18table-on/d_18table-on_p13/
          https://files.hawaii.gov/tax/forms/2018/18table-on.pdf
          https://files.hawaii.gov/tax/news/pubs/22outline.pdf
          https://files.hawaii.gov/tax/news/pubs/FastTaxRefGuide_2017.pdf
          https://www.tfhawaii.org/wordpress/state-tax-resources/mini-tax-guides/individual-income-tax-chapter-235/#:~:text=Long%20term%20capital%20gains%20are%20taxed%20at%20a%20maximum%20of%207.25%25
          https://www.grassrootinstitute.org/2023/02/hb337-increasing-tax-on-capital-gains-a-bad-risk/
          https://files.hawaii.gov/tax/news/pubs/22outline.pdf
        
        North Carolina:
        -         
        - https://www.ncdor.gov/taxes-forms/tax-rate-schedules
          https://www.ncdor.gov/taxes-forms/individual-income-tax/north-carolina-standard-deduction-or-north-carolina-itemized-deductions
          https://www.ncdor.gov/taxes-forms/individual-income-tax
          https://www.ncdor.gov/taxes-forms/individual-income-tax
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Hawaii', 'state_occupation': 'NorthCarolina', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def hawaii_and_north_dakota(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Hawaii:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        North Dakota:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Hawaii:
        - https://tax.hawaii.gov/forms/d_18table-on/d_18table-on_p13/
          https://files.hawaii.gov/tax/forms/2018/18table-on.pdf
          https://files.hawaii.gov/tax/news/pubs/22outline.pdf
          https://files.hawaii.gov/tax/news/pubs/FastTaxRefGuide_2017.pdf
          https://www.tfhawaii.org/wordpress/state-tax-resources/mini-tax-guides/individual-income-tax-chapter-235/#:~:text=Long%20term%20capital%20gains%20are%20taxed%20at%20a%20maximum%20of%207.25%25
          https://www.grassrootinstitute.org/2023/02/hb337-increasing-tax-on-capital-gains-a-bad-risk/
          https://files.hawaii.gov/tax/news/pubs/22outline.pdf
        
        North Dakota:
        -         
        - https://www.tax.nd.gov/forms
          https://www.tax.nd.gov/sites/www/files/documents/forms/individual/2022-iit/2022-individual-income-tax-booklet.pdf
          https://www.tax.nd.gov/sites/www/files/documents/forms/2022-individual-income-tax-booklet.pdf
          https://www.tax.nd.gov/tax-exemptions-credits/income-tax-exemptions-credits
          https://www.tax.nd.gov/sites/www/files/documents/forms/business/ndwrfillable.pdf
          https://www.tax.nd.gov/news/tax-legislative-changes/significant-changes-law/individual-income-tax-history
          https://www.tax.nd.gov/forms
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Hawaii', 'state_occupation': 'NorthDakota', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def hawaii_and_ohio(self, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Hawaii:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Ohio:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Hawaii:
        - https://tax.hawaii.gov/forms/d_18table-on/d_18table-on_p13/
          https://files.hawaii.gov/tax/forms/2018/18table-on.pdf
          https://files.hawaii.gov/tax/news/pubs/22outline.pdf
          https://files.hawaii.gov/tax/news/pubs/FastTaxRefGuide_2017.pdf
          https://www.tfhawaii.org/wordpress/state-tax-resources/mini-tax-guides/individual-income-tax-chapter-235/#:~:text=Long%20term%20capital%20gains%20are%20taxed%20at%20a%20maximum%20of%207.25%25
          https://www.grassrootinstitute.org/2023/02/hb337-increasing-tax-on-capital-gains-a-bad-risk/
          https://files.hawaii.gov/tax/news/pubs/22outline.pdf
        
        Ohio:
        -         
        - https://tax.ohio.gov/individual/resources/annual-tax-rates
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://tax.ohio.gov/static/forms/employer_withholding/generic/wth_it4nr.pdf
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://www.ritaohio.com/TaxRatesTable
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Hawaii', 'state_occupation': 'Ohio', "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def hawaii_and_oklahoma(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Hawaii:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Oklahoma:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Hawaii:
        - https://tax.hawaii.gov/forms/d_18table-on/d_18table-on_p13/
          https://files.hawaii.gov/tax/forms/2018/18table-on.pdf
          https://files.hawaii.gov/tax/news/pubs/22outline.pdf
          https://files.hawaii.gov/tax/news/pubs/FastTaxRefGuide_2017.pdf
          https://www.tfhawaii.org/wordpress/state-tax-resources/mini-tax-guides/individual-income-tax-chapter-235/#:~:text=Long%20term%20capital%20gains%20are%20taxed%20at%20a%20maximum%20of%207.25%25
          https://www.grassrootinstitute.org/2023/02/hb337-increasing-tax-on-capital-gains-a-bad-risk/
          https://files.hawaii.gov/tax/news/pubs/22outline.pdf
        
        Oklahoma:
        -         
        - https://oklahoma.gov/content/dam/ok/en/tax/documents/resources/publications/businesses/withholding-tables/WHTables-2022.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Hawaii', 'state_occupation': 'Oklahoma', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def hawaii_and_oregon(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Hawaii:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Oregon:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Hawaii:
        - https://tax.hawaii.gov/forms/d_18table-on/d_18table-on_p13/
          https://files.hawaii.gov/tax/forms/2018/18table-on.pdf
          https://files.hawaii.gov/tax/news/pubs/22outline.pdf
          https://files.hawaii.gov/tax/news/pubs/FastTaxRefGuide_2017.pdf
          https://www.tfhawaii.org/wordpress/state-tax-resources/mini-tax-guides/individual-income-tax-chapter-235/#:~:text=Long%20term%20capital%20gains%20are%20taxed%20at%20a%20maximum%20of%207.25%25
          https://www.grassrootinstitute.org/2023/02/hb337-increasing-tax-on-capital-gains-a-bad-risk/
          https://files.hawaii.gov/tax/news/pubs/22outline.pdf
        
        Oregon:
        -         
        - https://www.oregon.gov/dor/programs/individuals/Pages/PIT.aspx
          https://www.oregon.gov/dor/programs/individuals/Documents/Full%20year%20resident
         %20Form%20or-40%20filers.pdf
          https://www.oregon.gov/dor/programs/individuals/Documents/Part-year%20and%20nonresident
         %20Form%20or-40-P%20and%20or-40-N%20filers.pdf.pdf
          https://www.oregon.gov/dor/forms/FormsPubs/form-or-40_101-040_2022.pdf
          https://www.oregon.gov/dor/forms/FormsPubs/form-or-40_101-040_2022.pdf
          https://www.oregon.gov/dor/forms/FormsPubs/form-or-40_101-040_2022.pdf
          https://www.portland.gov/revenue/personal-tax
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Hawaii', 'state_occupation': 'Oregon', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def hawaii_and_pennsylvania(self, county_residence='Other', state_agi=75000, county_occupation='Other', filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Hawaii:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Pennsylvania:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        county_residence : str, optional
            County of residence. Default is Other.
        county_occupation : str, optional
            County of occupation. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Hawaii:
        - https://tax.hawaii.gov/forms/d_18table-on/d_18table-on_p13/
          https://files.hawaii.gov/tax/forms/2018/18table-on.pdf
          https://files.hawaii.gov/tax/news/pubs/22outline.pdf
          https://files.hawaii.gov/tax/news/pubs/FastTaxRefGuide_2017.pdf
          https://www.tfhawaii.org/wordpress/state-tax-resources/mini-tax-guides/individual-income-tax-chapter-235/#:~:text=Long%20term%20capital%20gains%20are%20taxed%20at%20a%20maximum%20of%207.25%25
          https://www.grassrootinstitute.org/2023/02/hb337-increasing-tax-on-capital-gains-a-bad-risk/
          https://files.hawaii.gov/tax/news/pubs/22outline.pdf
        
        Pennsylvania:
        -         
        - https://www.revenue.pa.gov/Tax%20Rates/Pages/default.aspx
          https://www.revenue.pa.gov/FormsandPublications/FormsforIndividuals/PIT/Documents/2022/2022_pa-40in.pdf
          https://www.revenue.pa.gov/FormsandPublications/FormsforBusinesses/EmployerWithholding/Documents/rev-419.pdf
          https://www.revenue.pa.gov/FormsandPublications/FormsforIndividuals/PIT/Documents/2022/2022_pa-40in.pdf
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-25.htm
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        county_occupation = self.county_occupation if hasattr(self, 'county_occupation') else county_occupation
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Hawaii', 'state_occupation': 'Pennsylvania', "county_residence": county_residence, "state_agi": state_agi, "county_occupation": county_occupation, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def hawaii_and_rhode_island(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Hawaii:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Rhode Island:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Hawaii:
        - https://tax.hawaii.gov/forms/d_18table-on/d_18table-on_p13/
          https://files.hawaii.gov/tax/forms/2018/18table-on.pdf
          https://files.hawaii.gov/tax/news/pubs/22outline.pdf
          https://files.hawaii.gov/tax/news/pubs/FastTaxRefGuide_2017.pdf
          https://www.tfhawaii.org/wordpress/state-tax-resources/mini-tax-guides/individual-income-tax-chapter-235/#:~:text=Long%20term%20capital%20gains%20are%20taxed%20at%20a%20maximum%20of%207.25%25
          https://www.grassrootinstitute.org/2023/02/hb337-increasing-tax-on-capital-gains-a-bad-risk/
          https://files.hawaii.gov/tax/news/pubs/22outline.pdf
        
        Rhode Island:
        -         
        - https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2022_1040WE_w_0.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Hawaii', 'state_occupation': 'RhodeIsland', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def hawaii_and_south_carolina(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Hawaii:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        South Carolina:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Hawaii:
        - https://tax.hawaii.gov/forms/d_18table-on/d_18table-on_p13/
          https://files.hawaii.gov/tax/forms/2018/18table-on.pdf
          https://files.hawaii.gov/tax/news/pubs/22outline.pdf
          https://files.hawaii.gov/tax/news/pubs/FastTaxRefGuide_2017.pdf
          https://www.tfhawaii.org/wordpress/state-tax-resources/mini-tax-guides/individual-income-tax-chapter-235/#:~:text=Long%20term%20capital%20gains%20are%20taxed%20at%20a%20maximum%20of%207.25%25
          https://www.grassrootinstitute.org/2023/02/hb337-increasing-tax-on-capital-gains-a-bad-risk/
          https://files.hawaii.gov/tax/news/pubs/22outline.pdf
        
        South Carolina:
        -         
        - https://dor.sc.gov/resources-site/lawandpolicy/Advisory%20Opinions/IL22-15.pdf
          https://dor.sc.gov/forms-site/Forms/IITPacket_2022.pdf
          https://www.scstatehouse.gov/code/t12c006.php
          https://dor.sc.gov/resources-site/lawandpolicy/Advisory%20Opinions/IL22-15.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Hawaii', 'state_occupation': 'SouthCarolina', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def hawaii_and_south_dakota(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Hawaii:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        South Dakota:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Hawaii:
        - https://tax.hawaii.gov/forms/d_18table-on/d_18table-on_p13/
          https://files.hawaii.gov/tax/forms/2018/18table-on.pdf
          https://files.hawaii.gov/tax/news/pubs/22outline.pdf
          https://files.hawaii.gov/tax/news/pubs/FastTaxRefGuide_2017.pdf
          https://www.tfhawaii.org/wordpress/state-tax-resources/mini-tax-guides/individual-income-tax-chapter-235/#:~:text=Long%20term%20capital%20gains%20are%20taxed%20at%20a%20maximum%20of%207.25%25
          https://www.grassrootinstitute.org/2023/02/hb337-increasing-tax-on-capital-gains-a-bad-risk/
          https://files.hawaii.gov/tax/news/pubs/22outline.pdf
        
        South Dakota:
        -         
        - https://dor.sd.gov/individuals/taxes/
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Hawaii', 'state_occupation': 'SouthDakota', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def hawaii_and_tennessee(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Hawaii:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Tennessee:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Hawaii:
        - https://tax.hawaii.gov/forms/d_18table-on/d_18table-on_p13/
          https://files.hawaii.gov/tax/forms/2018/18table-on.pdf
          https://files.hawaii.gov/tax/news/pubs/22outline.pdf
          https://files.hawaii.gov/tax/news/pubs/FastTaxRefGuide_2017.pdf
          https://www.tfhawaii.org/wordpress/state-tax-resources/mini-tax-guides/individual-income-tax-chapter-235/#:~:text=Long%20term%20capital%20gains%20are%20taxed%20at%20a%20maximum%20of%207.25%25
          https://www.grassrootinstitute.org/2023/02/hb337-increasing-tax-on-capital-gains-a-bad-risk/
          https://files.hawaii.gov/tax/news/pubs/22outline.pdf
        
        Tennessee:
        -         
        - https://www.tn.gov/revenue/taxes.html
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Hawaii', 'state_occupation': 'Tennessee', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def hawaii_and_texas(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Hawaii:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Texas:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Hawaii:
        - https://tax.hawaii.gov/forms/d_18table-on/d_18table-on_p13/
          https://files.hawaii.gov/tax/forms/2018/18table-on.pdf
          https://files.hawaii.gov/tax/news/pubs/22outline.pdf
          https://files.hawaii.gov/tax/news/pubs/FastTaxRefGuide_2017.pdf
          https://www.tfhawaii.org/wordpress/state-tax-resources/mini-tax-guides/individual-income-tax-chapter-235/#:~:text=Long%20term%20capital%20gains%20are%20taxed%20at%20a%20maximum%20of%207.25%25
          https://www.grassrootinstitute.org/2023/02/hb337-increasing-tax-on-capital-gains-a-bad-risk/
          https://files.hawaii.gov/tax/news/pubs/22outline.pdf
        
        Texas:
        -         
        - https://comptroller.texas.gov/taxes/
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Hawaii', 'state_occupation': 'Texas', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def hawaii_and_utah(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Hawaii:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Utah:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Hawaii:
        - https://tax.hawaii.gov/forms/d_18table-on/d_18table-on_p13/
          https://files.hawaii.gov/tax/forms/2018/18table-on.pdf
          https://files.hawaii.gov/tax/news/pubs/22outline.pdf
          https://files.hawaii.gov/tax/news/pubs/FastTaxRefGuide_2017.pdf
          https://www.tfhawaii.org/wordpress/state-tax-resources/mini-tax-guides/individual-income-tax-chapter-235/#:~:text=Long%20term%20capital%20gains%20are%20taxed%20at%20a%20maximum%20of%207.25%25
          https://www.grassrootinstitute.org/2023/02/hb337-increasing-tax-on-capital-gains-a-bad-risk/
          https://files.hawaii.gov/tax/news/pubs/22outline.pdf
        
        Utah:
        -         
        - https://incometax.utah.gov/paying/tax-rates
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Hawaii', 'state_occupation': 'Utah', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def hawaii_and_vermont(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Hawaii:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Vermont:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Hawaii:
        - https://tax.hawaii.gov/forms/d_18table-on/d_18table-on_p13/
          https://files.hawaii.gov/tax/forms/2018/18table-on.pdf
          https://files.hawaii.gov/tax/news/pubs/22outline.pdf
          https://files.hawaii.gov/tax/news/pubs/FastTaxRefGuide_2017.pdf
          https://www.tfhawaii.org/wordpress/state-tax-resources/mini-tax-guides/individual-income-tax-chapter-235/#:~:text=Long%20term%20capital%20gains%20are%20taxed%20at%20a%20maximum%20of%207.25%25
          https://www.grassrootinstitute.org/2023/02/hb337-increasing-tax-on-capital-gains-a-bad-risk/
          https://files.hawaii.gov/tax/news/pubs/22outline.pdf
        
        Vermont:
        -         
        - https://tax.vermont.gov/sites/tax/files/documents/GB-1210-2023.pdf
          https://tax.vermont.gov/sites/tax/files/documents/Income%20Booklet-2022.pdf
          https://tax.vermont.gov/sites/tax/files/documents/Income%20Booklet-2022.pdf
          https://tax.vermont.gov/sites/tax/files/documents/Income%20Booklet-2022.pdf
          https://tax.vermont.gov/individuals/personal-income-tax/tax-credits
          https://tax.vermont.gov/individuals/personal-income-tax/taxable-income
          https://tax.vermont.gov/individuals/personal-income-tax/local-option-tax
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Hawaii', 'state_occupation': 'Vermont', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def hawaii_and_virginia(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Hawaii:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Virginia:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Hawaii:
        - https://tax.hawaii.gov/forms/d_18table-on/d_18table-on_p13/
          https://files.hawaii.gov/tax/forms/2018/18table-on.pdf
          https://files.hawaii.gov/tax/news/pubs/22outline.pdf
          https://files.hawaii.gov/tax/news/pubs/FastTaxRefGuide_2017.pdf
          https://www.tfhawaii.org/wordpress/state-tax-resources/mini-tax-guides/individual-income-tax-chapter-235/#:~:text=Long%20term%20capital%20gains%20are%20taxed%20at%20a%20maximum%20of%207.25%25
          https://www.grassrootinstitute.org/2023/02/hb337-increasing-tax-on-capital-gains-a-bad-risk/
          https://files.hawaii.gov/tax/news/pubs/22outline.pdf
        
        Virginia:
        -         
        - https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/reciprocity
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Hawaii', 'state_occupation': 'Virginia', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def hawaii_and_washington(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Hawaii:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Washington:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Hawaii:
        - https://tax.hawaii.gov/forms/d_18table-on/d_18table-on_p13/
          https://files.hawaii.gov/tax/forms/2018/18table-on.pdf
          https://files.hawaii.gov/tax/news/pubs/22outline.pdf
          https://files.hawaii.gov/tax/news/pubs/FastTaxRefGuide_2017.pdf
          https://www.tfhawaii.org/wordpress/state-tax-resources/mini-tax-guides/individual-income-tax-chapter-235/#:~:text=Long%20term%20capital%20gains%20are%20taxed%20at%20a%20maximum%20of%207.25%25
          https://www.grassrootinstitute.org/2023/02/hb337-increasing-tax-on-capital-gains-a-bad-risk/
          https://files.hawaii.gov/tax/news/pubs/22outline.pdf
        
        Washington:
        -         
        - https://dor.wa.gov/taxes-rates/income-tax
          https://dor.wa.gov/taxes-rates/other-taxes/capital-gains-tax
          https://dor.wa.gov/taxes-rates/income-tax
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Hawaii', 'state_occupation': 'Washington', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def hawaii_and_west_virginia(self, pay_periods=24, capital_gains_short=0, county_occupation='Other', dependents=0, sources=False, state_agi=75000, filers_over_65=0, capital_gains_long=0, filing_status='single', income=100000, credits_state=0):
        """
        --- Parameters ----
        
        Hawaii:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        West Virginia:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_occupation : str, optional
            County of occupation. Default is Other.
        pay_periods : int, optional
            Number of pay periods. Options are 13, 24, 26, 52. Default is 24.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Hawaii:
        - https://tax.hawaii.gov/forms/d_18table-on/d_18table-on_p13/
          https://files.hawaii.gov/tax/forms/2018/18table-on.pdf
          https://files.hawaii.gov/tax/news/pubs/22outline.pdf
          https://files.hawaii.gov/tax/news/pubs/FastTaxRefGuide_2017.pdf
          https://www.tfhawaii.org/wordpress/state-tax-resources/mini-tax-guides/individual-income-tax-chapter-235/#:~:text=Long%20term%20capital%20gains%20are%20taxed%20at%20a%20maximum%20of%207.25%25
          https://www.grassrootinstitute.org/2023/02/hb337-increasing-tax-on-capital-gains-a-bad-risk/
          https://files.hawaii.gov/tax/news/pubs/22outline.pdf
        
        West Virginia:
        -         
        - https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://tax.wv.gov/Documents/TSD/tsd438.pdf
          https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2011/Taxes-11-35att.html
        """
        
        # Check if the parameters are set upstream in the instance
        pay_periods = self.pay_periods if hasattr(self, 'pay_periods') else pay_periods
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        county_occupation = self.county_occupation if hasattr(self, 'county_occupation') else county_occupation
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Hawaii', 'state_occupation': 'WestVirginia', "pay_periods": pay_periods, "capital_gains_short": capital_gains_short, "county_occupation": county_occupation, "dependents": dependents, "sources": sources, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def hawaii_and_wisconsin(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Hawaii:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Wisconsin:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Hawaii:
        - https://tax.hawaii.gov/forms/d_18table-on/d_18table-on_p13/
          https://files.hawaii.gov/tax/forms/2018/18table-on.pdf
          https://files.hawaii.gov/tax/news/pubs/22outline.pdf
          https://files.hawaii.gov/tax/news/pubs/FastTaxRefGuide_2017.pdf
          https://www.tfhawaii.org/wordpress/state-tax-resources/mini-tax-guides/individual-income-tax-chapter-235/#:~:text=Long%20term%20capital%20gains%20are%20taxed%20at%20a%20maximum%20of%207.25%25
          https://www.grassrootinstitute.org/2023/02/hb337-increasing-tax-on-capital-gains-a-bad-risk/
          https://files.hawaii.gov/tax/news/pubs/22outline.pdf
        
        Wisconsin:
        -         
        - https://www.revenue.wi.gov/TaxForms2023/2023-Form1-ES-Inst.pdf
          https://www.revenue.wi.gov/TaxForms2023/2023-Form1-ES-Inst.pdf
          https://www.revenue.wi.gov/TaxForms2023/2023-Form1-ES-Inst.pdf
          https://www.revenue.wi.gov/Pages/FAQS/pcs-work.aspx
          https://www.revenue.wi.gov/Dor%20Publications/pb103.pdf
          https://www.revenue.wi.gov/Pages/FAQS/pcs-work.aspx
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Hawaii', 'state_occupation': 'Wisconsin', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def hawaii_and_wyoming(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Hawaii:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Wyoming:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Hawaii:
        - https://tax.hawaii.gov/forms/d_18table-on/d_18table-on_p13/
          https://files.hawaii.gov/tax/forms/2018/18table-on.pdf
          https://files.hawaii.gov/tax/news/pubs/22outline.pdf
          https://files.hawaii.gov/tax/news/pubs/FastTaxRefGuide_2017.pdf
          https://www.tfhawaii.org/wordpress/state-tax-resources/mini-tax-guides/individual-income-tax-chapter-235/#:~:text=Long%20term%20capital%20gains%20are%20taxed%20at%20a%20maximum%20of%207.25%25
          https://www.grassrootinstitute.org/2023/02/hb337-increasing-tax-on-capital-gains-a-bad-risk/
          https://files.hawaii.gov/tax/news/pubs/22outline.pdf
        
        Wyoming:
        -         
        - https://revenue.wyo.gov/
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Hawaii', 'state_occupation': 'Wyoming', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def idaho_and_alabama(self, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Idaho:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Alabama:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Idaho:
        - https://tax.idaho.gov/taxes/income-tax/individual-income/individual-income-tax-rate-schedule/
          https://tax.idaho.gov/wp-content/uploads/forms/EIN00046/EIN00046_03-01-2023.pdf
          https://tax.idaho.gov/taxes/income-tax/individual-income/idaho-source-income/
          https://tax.idaho.gov/taxes/income-tax/individual-income/idaho-source-income/
        
        Alabama:
        -         
        - https://www.revenue.alabama.gov/ultraviewer/viewer/basic_viewer/index.html?form=2023/02/22f40.pdf
          https://www.revenue.alabama.gov/faqs/what-is-alabamas-individual-income-tax-rate/
          https://www.revenue.alabama.gov/ultraviewer/viewer/basic_viewer/index.html?form=2023/02/22f40.pdf
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-24.htm?taxmap=true
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-24.htm?taxmap=true
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-24.htm?taxmap=true
          https://www.revenue.alabama.gov/individual-corporate/income-to-be-reported-on-the-alabama-income-tax-return/
          https://www.almonline.org/TaxRates.aspx#OccupationalTax
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Idaho', 'state_occupation': 'Alabama', "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def idaho_and_alaska(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Idaho:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Alaska:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Idaho:
        - https://tax.idaho.gov/taxes/income-tax/individual-income/individual-income-tax-rate-schedule/
          https://tax.idaho.gov/wp-content/uploads/forms/EIN00046/EIN00046_03-01-2023.pdf
          https://tax.idaho.gov/taxes/income-tax/individual-income/idaho-source-income/
          https://tax.idaho.gov/taxes/income-tax/individual-income/idaho-source-income/
        
        Alaska:
        -         
        - https://www.commerce.alaska.gov/web/dcra/officeofthestateassessor/alaskataxfacts.aspx
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Idaho', 'state_occupation': 'Alaska', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def idaho_and_arizona(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Idaho:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Arizona:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        credits_state : int, optional
            Credits for state taxes. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Idaho:
        - https://tax.idaho.gov/taxes/income-tax/individual-income/individual-income-tax-rate-schedule/
          https://tax.idaho.gov/wp-content/uploads/forms/EIN00046/EIN00046_03-01-2023.pdf
          https://tax.idaho.gov/taxes/income-tax/individual-income/idaho-source-income/
          https://tax.idaho.gov/taxes/income-tax/individual-income/idaho-source-income/
        
        Arizona:
        -         
        - https://azdor.gov/forms/individual/form-140-x-y-tables
          https://azdor.gov/forms/individual/form-140-x-y-tables
          https://azdor.gov/forms/individual/form-140-x-y-tables
          https://azdor.gov/businesses-arizona/withholding-tax/withholding-exceptions
          "https://azdor.gov/individuals/income-tax-filing-assistance/identifying-other-taxable-income#:~:text=In%20regards%20to%20capital%20gains
         the%20individuals%20regular%20tax%20rate.
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Idaho', 'state_occupation': 'Arizona', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def idaho_and_arkansas(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Idaho:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Arkansas:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Idaho:
        - https://tax.idaho.gov/taxes/income-tax/individual-income/individual-income-tax-rate-schedule/
          https://tax.idaho.gov/wp-content/uploads/forms/EIN00046/EIN00046_03-01-2023.pdf
          https://tax.idaho.gov/taxes/income-tax/individual-income/idaho-source-income/
          https://tax.idaho.gov/taxes/income-tax/individual-income/idaho-source-income/
        
        Arkansas:
        -         
        - https://www.arkansasedc.com/why-arkansas/business-climate/tax-structure/personal-income-tax
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_and_AR1000NR_Instructions.pdf
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_FullYearResidentIndividualIncomeTaxReturn.pdf
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_and_AR1000NR_Instructions.pdf
          https://www.arkansasedc.com/why-arkansas/business-climate/tax-structure/capital-gains-tax-reduction
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_and_AR1000NR_Instructions.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Idaho', 'state_occupation': 'Arkansas', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def idaho_and_california(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Idaho:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        California:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Idaho:
        - https://tax.idaho.gov/taxes/income-tax/individual-income/individual-income-tax-rate-schedule/
          https://tax.idaho.gov/wp-content/uploads/forms/EIN00046/EIN00046_03-01-2023.pdf
          https://tax.idaho.gov/taxes/income-tax/individual-income/idaho-source-income/
          https://tax.idaho.gov/taxes/income-tax/individual-income/idaho-source-income/
        
        California:
        -         
        - https://www.ftb.ca.gov/forms/2022/2022-540-tax-rate-schedules.pdf
          https://www.ftb.ca.gov/file/personal/deductions/index.html
          https://www.ftb.ca.gov/about-ftb/newsroom/caleitc/eligibility-and-credit-information.html
          https://www.ftb.ca.gov/file/personal/income-types/capital-gains-and-losses.html#:~:text=California%20does%20not%20have%20a
         are%20taxed%20as%20ordinary%20income
          https://www.ftb.ca.gov/file/personal/
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Idaho', 'state_occupation': 'California', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def idaho_and_colorado(self, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Idaho:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Colorado:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Idaho:
        - https://tax.idaho.gov/taxes/income-tax/individual-income/individual-income-tax-rate-schedule/
          https://tax.idaho.gov/wp-content/uploads/forms/EIN00046/EIN00046_03-01-2023.pdf
          https://tax.idaho.gov/taxes/income-tax/individual-income/idaho-source-income/
          https://tax.idaho.gov/taxes/income-tax/individual-income/idaho-source-income/
        
        Colorado:
        -         
        - https://leg.colorado.gov/agencies/legislative-council-staff/individual-income-tax%C2%A0
          https://leg.colorado.gov/agencies/legislative-council-staff/individual-income-tax%C2%A0
          https://tax.colorado.gov/sites/tax/files/documents/ITT_Capital_Gain_Subtraction_Feb_2022.pdf
          https://www.auroragov.org/business_services/taxes/occupational_privilege_tax
          https://www.denvergov.org/Government/Agencies-Departments-Offices/Agencies-Departments-Offices-Directory/Department-of-Finance/Our-Divisions/Treasury/Business-Tax-Information#:~:text=Occupational%20privilege%20tax
         -%28show%20below%29&text=Each%20taxable%20employee%20is%20liable
         month%20for%20each%20taxable%20employee
          https://www.glendale.co.us/355/Occupational-Privilege-Tax
          https://www.greenwoodvillage.com/1220/Occupational-Privilege-Tax-OPT
          https://ci.sheridan.co.us/288/Occupational-Privilege-Tax
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Idaho', 'state_occupation': 'Colorado', "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def idaho_and_connecticut(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Idaho:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Connecticut:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Idaho:
        - https://tax.idaho.gov/taxes/income-tax/individual-income/individual-income-tax-rate-schedule/
          https://tax.idaho.gov/wp-content/uploads/forms/EIN00046/EIN00046_03-01-2023.pdf
          https://tax.idaho.gov/taxes/income-tax/individual-income/idaho-source-income/
          https://tax.idaho.gov/taxes/income-tax/individual-income/idaho-source-income/
        
        Connecticut:
        -         
        - https://portal.ct.gov/-/media/DRS/Forms/2022/Income/2022-CT-1040-Instructions_1222.pdf
          https://www.cga.ct.gov/2022/rpt/pdf/2022-R-0108.pdf
          https://www.cga.ct.gov/current/pub/chap_229.htm#sec_12-703
          https://portal.ct.gov/-/media/DRS/Forms/2022/Income/2022-CT-1040-Instructions_1222.pdf
          "https://portal.ct.gov/DRS/Publications/TSSNs/TSSN-29#:~:text=An%20individuals%20net%20capital%20gains
         portion%20of%20Social%20Security%20benefits"
          "https://portal.ct.gov/DRS/Publications/TSSNs/TSSN-29#:~:text=An%20individuals%20net%20capital%20gains
         portion%20of%20Social%20Security%20benefits"
          https://portal.ct.gov/DRS/Individuals/Resident-Income-Tax/Tax-Information
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Idaho', 'state_occupation': 'Connecticut', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def idaho_and_delaware(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, federal_agi=75000, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Idaho:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Delaware:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        federal_agi : int, optional
            Adjusted Gross Income (AGI) for federal taxes. Default is 0. Sources - https://www.irs.gov/e-file-providers/definition-of-adjusted-gross-income, https://www.irs.gov/newsroom/taxpayers-should-know-the-difference-between-standard-and-itemized-deductions
        income : int, optional
            Gross income. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Idaho:
        - https://tax.idaho.gov/taxes/income-tax/individual-income/individual-income-tax-rate-schedule/
          https://tax.idaho.gov/wp-content/uploads/forms/EIN00046/EIN00046_03-01-2023.pdf
          https://tax.idaho.gov/taxes/income-tax/individual-income/idaho-source-income/
          https://tax.idaho.gov/taxes/income-tax/individual-income/idaho-source-income/
        
        Delaware:
        -         
        - https://revenue.delaware.gov/software-developer/tax-rate-changes/
          https://revenuefiles.delaware.gov/2022/PIT-RES_TY22_2022-01_PaperInteractive.pdf
          https://revenuefiles.delaware.gov/2022/PIT-RES_TY22_2022-02_Instructions.pdf
          https://revenue.delaware.gov/frequently-asked-questions/personal-income-tax-faqs/
          https://www.newcastlede.gov/DocumentCenter/View/20147/City-of-Wilmington-Wage-Tax-Information-PDF
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        federal_agi = self.federal_agi if hasattr(self, 'federal_agi') else federal_agi
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Idaho', 'state_occupation': 'Delaware', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "federal_agi": federal_agi, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def idaho_and_florida(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Idaho:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Florida:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Idaho:
        - https://tax.idaho.gov/taxes/income-tax/individual-income/individual-income-tax-rate-schedule/
          https://tax.idaho.gov/wp-content/uploads/forms/EIN00046/EIN00046_03-01-2023.pdf
          https://tax.idaho.gov/taxes/income-tax/individual-income/idaho-source-income/
          https://tax.idaho.gov/taxes/income-tax/individual-income/idaho-source-income/
        
        Florida:
        -         
        - https://floridarevenue.com/taxes/taxesfees/Pages/tax_interest_rates.aspx
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Idaho', 'state_occupation': 'Florida', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def idaho_and_georgia(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Idaho:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Georgia:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Idaho:
        - https://tax.idaho.gov/taxes/income-tax/individual-income/individual-income-tax-rate-schedule/
          https://tax.idaho.gov/wp-content/uploads/forms/EIN00046/EIN00046_03-01-2023.pdf
          https://tax.idaho.gov/taxes/income-tax/individual-income/idaho-source-income/
          https://tax.idaho.gov/taxes/income-tax/individual-income/idaho-source-income/
        
        Georgia:
        -         
        - https://dor.georgia.gov/tax-tables-georgia-tax-rate-schedule
          https://apps.dor.ga.gov/FillableForms/PDFViewer/Index?form=2022GA500
          https://apps.dor.ga.gov/FillableForms/PDFViewer/Index?form=2022GA500
          https://apps.dor.ga.gov/FillableForms/PDFViewer/Index?form=2022GA500
          https://dor.georgia.gov/years-individual-income-tax-forms
          https://dor.georgia.gov/years-individual-income-tax-forms
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Idaho', 'state_occupation': 'Georgia', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def idaho_and_hawaii(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Idaho:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Hawaii:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Idaho:
        - https://tax.idaho.gov/taxes/income-tax/individual-income/individual-income-tax-rate-schedule/
          https://tax.idaho.gov/wp-content/uploads/forms/EIN00046/EIN00046_03-01-2023.pdf
          https://tax.idaho.gov/taxes/income-tax/individual-income/idaho-source-income/
          https://tax.idaho.gov/taxes/income-tax/individual-income/idaho-source-income/
        
        Hawaii:
        -         
        - https://tax.hawaii.gov/forms/d_18table-on/d_18table-on_p13/
          https://files.hawaii.gov/tax/forms/2018/18table-on.pdf
          https://files.hawaii.gov/tax/news/pubs/22outline.pdf
          https://files.hawaii.gov/tax/news/pubs/FastTaxRefGuide_2017.pdf
          https://www.tfhawaii.org/wordpress/state-tax-resources/mini-tax-guides/individual-income-tax-chapter-235/#:~:text=Long%20term%20capital%20gains%20are%20taxed%20at%20a%20maximum%20of%207.25%25
          https://www.grassrootinstitute.org/2023/02/hb337-increasing-tax-on-capital-gains-a-bad-risk/
          https://files.hawaii.gov/tax/news/pubs/22outline.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Idaho', 'state_occupation': 'Hawaii', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def idaho_and_illinois(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Idaho:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Illinois:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Idaho:
        - https://tax.idaho.gov/taxes/income-tax/individual-income/individual-income-tax-rate-schedule/
          https://tax.idaho.gov/wp-content/uploads/forms/EIN00046/EIN00046_03-01-2023.pdf
          https://tax.idaho.gov/taxes/income-tax/individual-income/idaho-source-income/
          https://tax.idaho.gov/taxes/income-tax/individual-income/idaho-source-income/
        
        Illinois:
        -         
        - https://tax.illinois.gov/research/taxrates/income.html#IndividualIncome
          https://tax.illinois.gov/content/dam/soi/en/web/tax/forms/incometax/documents/currentyear/individual/il-1040-instr.pdf
          https://tax.illinois.gov/questionsandanswers/12.html
          https://tax.illinois.gov/content/dam/soi/en/web/tax/forms/incometax/documents/currentyear/individual/il-1040-instr.pdf
          https://tax.illinois.gov/research/taxrates/personalproperty.html
          https://tax.illinois.gov/localgovernments/income.html
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Idaho', 'state_occupation': 'Illinois', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def idaho_and_indiana(self, capital_gains_short=0, dependents=0, sources=False, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_long=0, filing_status='single', income=100000, credits_state=0):
        """
        --- Parameters ----
        
        Idaho:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Indiana:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Idaho:
        - https://tax.idaho.gov/taxes/income-tax/individual-income/individual-income-tax-rate-schedule/
          https://tax.idaho.gov/wp-content/uploads/forms/EIN00046/EIN00046_03-01-2023.pdf
          https://tax.idaho.gov/taxes/income-tax/individual-income/idaho-source-income/
          https://tax.idaho.gov/taxes/income-tax/individual-income/idaho-source-income/
        
        Indiana:
        -         
        - https://www.in.gov/dor/tax-forms/2022-individual-income-tax-forms/
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-32.htm
          https://www.in.gov/dor/files/ib28.pdf
          https://www.in.gov/dor/tax-forms/2022-individual-income-tax-forms/
          https://www.in.gov/dor/files/2022-county-tax-rates-and-codes.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Idaho', 'state_occupation': 'Indiana', "capital_gains_short": capital_gains_short, "dependents": dependents, "sources": sources, "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def idaho_and_iowa(self, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Idaho:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Iowa:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Idaho:
        - https://tax.idaho.gov/taxes/income-tax/individual-income/individual-income-tax-rate-schedule/
          https://tax.idaho.gov/wp-content/uploads/forms/EIN00046/EIN00046_03-01-2023.pdf
          https://tax.idaho.gov/taxes/income-tax/individual-income/idaho-source-income/
          https://tax.idaho.gov/taxes/income-tax/individual-income/idaho-source-income/
        
        Iowa:
        -         
        - https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
          https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
          https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
          https://tax.iowa.gov/iowa-illinois-reciprocal-agreement
          https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Idaho', 'state_occupation': 'Iowa', "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def idaho_and_kansas(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Idaho:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Kansas:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Idaho:
        - https://tax.idaho.gov/taxes/income-tax/individual-income/individual-income-tax-rate-schedule/
          https://tax.idaho.gov/wp-content/uploads/forms/EIN00046/EIN00046_03-01-2023.pdf
          https://tax.idaho.gov/taxes/income-tax/individual-income/idaho-source-income/
          https://tax.idaho.gov/taxes/income-tax/individual-income/idaho-source-income/
        
        Kansas:
        -         
        - https://www.ksrevenue.gov/taxrates.html
          https://www.ksrevenue.gov/incomebook22.html#0
          https://www.ksrevenue.gov/incomebook22.html#0
          https://www.ksrevenue.gov/incomebook22.html#0
          https://www.ksrevenue.gov/incomebook22.html
          https://www.ksrevenue.gov/incomebook22.html
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Idaho', 'state_occupation': 'Kansas', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def idaho_and_kentucky(self, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Idaho:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Kentucky:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Idaho:
        - https://tax.idaho.gov/taxes/income-tax/individual-income/individual-income-tax-rate-schedule/
          https://tax.idaho.gov/wp-content/uploads/forms/EIN00046/EIN00046_03-01-2023.pdf
          https://tax.idaho.gov/taxes/income-tax/individual-income/idaho-source-income/
          https://tax.idaho.gov/taxes/income-tax/individual-income/idaho-source-income/
        
        Kentucky:
        -         
        - https://revenue.ky.gov/News/Pages/Dor-Announces-Updates-to-Individual-Income-Tax-for-2023-Tax-Year.aspx
          https://revenue.ky.gov/Individual/Individual-Income-Tax/Pages/default.aspx
          https://revenue.ky.gov/News/Pages/Dor-Announces-Updates-to-Individual-Income-Tax-for-2023-Tax-Year.aspx
          https://revenue.ky.gov/Individual/Individual-Income-Tax/Pages/default.aspx
          https://revenue.ky.gov/Dor%20Training%20Materials/103%20KAR%2017.140.%20Individual%20income%20tax%20-%20reciprocity%20-%20nonresidents.pdf
          https://revenue.ky.gov/Individual/Individual-Income-Tax/Pages/default.aspx
          https://nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2021/Taxes-21-32.htm?taxmap=true
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Idaho', 'state_occupation': 'Kentucky', "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def idaho_and_louisiana(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Idaho:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Louisiana:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Idaho:
        - https://tax.idaho.gov/taxes/income-tax/individual-income/individual-income-tax-rate-schedule/
          https://tax.idaho.gov/wp-content/uploads/forms/EIN00046/EIN00046_03-01-2023.pdf
          https://tax.idaho.gov/taxes/income-tax/individual-income/idaho-source-income/
          https://tax.idaho.gov/taxes/income-tax/individual-income/idaho-source-income/
        
        Louisiana:
        -         
        - https://revenue.louisiana.gov/individualincometax
          https://revenue.louisiana.gov/individualincometax
          https://www.revenue.louisiana.gov/taxforms/6935(11_02)F.pdf
          https://revenue.louisiana.gov/individualincometax
          https://revenue.louisiana.gov/individualincometax
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Idaho', 'state_occupation': 'Louisiana', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def idaho_and_maine(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Idaho:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Maine:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Idaho:
        - https://tax.idaho.gov/taxes/income-tax/individual-income/individual-income-tax-rate-schedule/
          https://tax.idaho.gov/wp-content/uploads/forms/EIN00046/EIN00046_03-01-2023.pdf
          https://tax.idaho.gov/taxes/income-tax/individual-income/idaho-source-income/
          https://tax.idaho.gov/taxes/income-tax/individual-income/idaho-source-income/
        
        Maine:
        -         
        - https://www.maine.gov/revenue/tax-return-forms/individual-income-tax-2022
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/ind_tax_rate_sched_2022.pdf
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/22_1040me_book_gen_instruc_revisedFeb23.pdf
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/22_1040me_book_gen_instruc_revisedFeb23.pdf
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/22_1040me_book_gen_instruc_revisedFeb23.pdf
          https://www.maine.gov/revenue/tax-return-forms/individual-income-tax-2022
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Idaho', 'state_occupation': 'Maine', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def idaho_and_maryland(self, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Idaho:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Maryland:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Idaho:
        - https://tax.idaho.gov/taxes/income-tax/individual-income/individual-income-tax-rate-schedule/
          https://tax.idaho.gov/wp-content/uploads/forms/EIN00046/EIN00046_03-01-2023.pdf
          https://tax.idaho.gov/taxes/income-tax/individual-income/idaho-source-income/
          https://tax.idaho.gov/taxes/income-tax/individual-income/idaho-source-income/
        
        Maryland:
        -         
        - https://www.marylandtaxes.gov/individual/income/tax-info/tax-rates.php
          https://www.marylandtaxes.gov/forms/22_forms/Resident_Booklet.pdf
          https://www.marylandtaxes.gov/forms/Tax_Publications/Administrative_Releases/Income_and_Estate_Tax_Releases/ar_it37.pdf
          https://www.marylandtaxes.gov/individual/index.php
          https://www.marylandtaxes.gov/individual/income/tax-info/tax-rates.php
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Idaho', 'state_occupation': 'Maryland', "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def idaho_and_massachusetts(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Idaho:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Massachusetts:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Idaho:
        - https://tax.idaho.gov/taxes/income-tax/individual-income/individual-income-tax-rate-schedule/
          https://tax.idaho.gov/wp-content/uploads/forms/EIN00046/EIN00046_03-01-2023.pdf
          https://tax.idaho.gov/taxes/income-tax/individual-income/idaho-source-income/
          https://tax.idaho.gov/taxes/income-tax/individual-income/idaho-source-income/
        
        Massachusetts:
        -         
        - https://www.mass.gov/service-details/massachusetts-tax-rates
          https://www.mass.gov/service-details/view-massachusetts-personal-income-tax-exemptions
          https://www.mass.gov/service-details/view-massachusetts-personal-income-tax-exemptions
          https://www.mass.gov/service-details/view-child-and-dependent-related-credits
          https://www.mass.gov/service-details/massachusetts-tax-rates
          https://www.mass.gov/personal-income-tax
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Idaho', 'state_occupation': 'Massachusetts', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def idaho_and_michigan(self, capital_gains_short=0, county_occupation='Other', dependents=0, sources=False, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_long=0, filing_status='single', income=100000, credits_state=0):
        """
        --- Parameters ----
        
        Idaho:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Michigan:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        county_occupation : str, optional
            County of occupation. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Idaho:
        - https://tax.idaho.gov/taxes/income-tax/individual-income/individual-income-tax-rate-schedule/
          https://tax.idaho.gov/wp-content/uploads/forms/EIN00046/EIN00046_03-01-2023.pdf
          https://tax.idaho.gov/taxes/income-tax/individual-income/idaho-source-income/
          https://tax.idaho.gov/taxes/income-tax/individual-income/idaho-source-income/
        
        Michigan:
        -         
        - https://www.michigan.gov/taxes/iit/new-developments-for-tax-year-2022
          "https://www.michigan.gov/taxes/-/media/Project/Websites/taxes/Forms/2023/2023-SUW/446_Withholding-Guide_2023.pdf?rev=177ca3d8cb034392bf84e317ed78c320&hash=19C5F42F43A89A91863D229349F48883#:~:text=The%20exemption%20amount%20is%20%245%2C400
         employees%20Michigan%20income%20tax%20return"
          "https://www.michigan.gov/taxes/-/media/Project/Websites/taxes/Forms/2023/2023-SUW/446_Withholding-Guide_2023.pdf?rev=177ca3d8cb034392bf84e317ed78c320&hash=19C5F42F43A89A91863D229349F48883#:~:text=The%20exemption%20amount%20is%20%245%2C400
         employees%20Michigan%20income%20tax%20return"
          https://www.michigan.gov/taxes/questions/iit/accordion/residency/are-my-wages-earned-in-another-state-taxable-in-michigan-if-i-am-a-michigan-resident-1
          https://www.michigan.gov/taxes/-/media/Project/Websites/taxes/Forms/2022/2022-IIT-Forms/BOOK_MI-1040.pdf?rev=4524f1cbbc0243b194ab05a6680be75b&hash=40DFE71E651F7C1D65C31F6CD03D944F
          https://www.michigan.gov/taxes/questions/iit/accordion/general/what-cities-impose-an-income-tax
        """
        
        # Check if the parameters are set upstream in the instance
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        county_occupation = self.county_occupation if hasattr(self, 'county_occupation') else county_occupation
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Idaho', 'state_occupation': 'Michigan', "capital_gains_short": capital_gains_short, "county_occupation": county_occupation, "dependents": dependents, "sources": sources, "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def idaho_and_minnesota(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Idaho:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Minnesota:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Idaho:
        - https://tax.idaho.gov/taxes/income-tax/individual-income/individual-income-tax-rate-schedule/
          https://tax.idaho.gov/wp-content/uploads/forms/EIN00046/EIN00046_03-01-2023.pdf
          https://tax.idaho.gov/taxes/income-tax/individual-income/idaho-source-income/
          https://tax.idaho.gov/taxes/income-tax/individual-income/idaho-source-income/
        
        Minnesota:
        -         
        - https://www.revenue.state.mn.us/minnesota-income-tax-rates-and-brackets
          https://www.revenue.state.mn.us/child-and-dependent-care-credit
          https://www.revenue.state.mn.us/dependent-exemptions
          https://www.revenue.state.mn.us/minnesota-standard-deduction
          https://mn.gov/mmb/assets/mwr_form_tcm1059-128581.pdf
          https://www.revenue.state.mn.us/individual-income-tax
          https://www.revenue.state.mn.us/individual-income-tax
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Idaho', 'state_occupation': 'Minnesota', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def idaho_and_mississippi(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Idaho:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Mississippi:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Idaho:
        - https://tax.idaho.gov/taxes/income-tax/individual-income/individual-income-tax-rate-schedule/
          https://tax.idaho.gov/wp-content/uploads/forms/EIN00046/EIN00046_03-01-2023.pdf
          https://tax.idaho.gov/taxes/income-tax/individual-income/idaho-source-income/
          https://tax.idaho.gov/taxes/income-tax/individual-income/idaho-source-income/
        
        Mississippi:
        -         
        - https://www.dor.ms.gov/individual/tax-rates
          https://www.dor.ms.gov/individual/tax-rates
          https://www.dor.ms.gov/individual/tax-rates
          https://www.dor.ms.gov/individual
          https://www.dor.ms.gov/individual
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Idaho', 'state_occupation': 'Mississippi', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def idaho_and_missouri(self, capital_gains_short=0, dependents=0, sources=False, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_long=0, filing_status='single', income=100000, credits_state=0):
        """
        --- Parameters ----
        
        Idaho:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Missouri:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Idaho:
        - https://tax.idaho.gov/taxes/income-tax/individual-income/individual-income-tax-rate-schedule/
          https://tax.idaho.gov/wp-content/uploads/forms/EIN00046/EIN00046_03-01-2023.pdf
          https://tax.idaho.gov/taxes/income-tax/individual-income/idaho-source-income/
          https://tax.idaho.gov/taxes/income-tax/individual-income/idaho-source-income/
        
        Missouri:
        -         
        - https://dor.mo.gov/taxation/individual/tax-types/income/year-changes/
          https://dor.mo.gov/taxation/individual/tax-types/income/year-changes/
          https://dor.mo.gov/taxation/individual/tax-types/income/year-changes/
          https://www.kcmo.gov/city-hall/departments/finance/earnings-tax
          https://www.stlouis-mo.gov/government/departments/collector/earnings-tax/individual-earnings-tax-info.cfm
        """
        
        # Check if the parameters are set upstream in the instance
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Idaho', 'state_occupation': 'Missouri', "capital_gains_short": capital_gains_short, "dependents": dependents, "sources": sources, "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def idaho_and_montana(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Idaho:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Montana:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Idaho:
        - https://tax.idaho.gov/taxes/income-tax/individual-income/individual-income-tax-rate-schedule/
          https://tax.idaho.gov/wp-content/uploads/forms/EIN00046/EIN00046_03-01-2023.pdf
          https://tax.idaho.gov/taxes/income-tax/individual-income/idaho-source-income/
          https://tax.idaho.gov/taxes/income-tax/individual-income/idaho-source-income/
        
        Montana:
        -         
        - https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/2022/12/montana-employees-withholding-allowance-and-exemption-certificate-form-mw-4.pdf
          https://mtrevenue.gov/2021/11/29/simplification-of-montana-income-taxation/?utm_source=rss&utm_medium=rss&utm_campaign=simplification-of-montana-income-taxation
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Idaho', 'state_occupation': 'Montana', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def idaho_and_nebraska(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Idaho:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Nebraska:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Idaho:
        - https://tax.idaho.gov/taxes/income-tax/individual-income/individual-income-tax-rate-schedule/
          https://tax.idaho.gov/wp-content/uploads/forms/EIN00046/EIN00046_03-01-2023.pdf
          https://tax.idaho.gov/taxes/income-tax/individual-income/idaho-source-income/
          https://tax.idaho.gov/taxes/income-tax/individual-income/idaho-source-income/
        
        Nebraska:
        -         
        - https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2023_Circular_EN_8-429r2023_Whole_Book_11.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Idaho', 'state_occupation': 'Nebraska', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def idaho_and_nevada(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Idaho:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Nevada:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Idaho:
        - https://tax.idaho.gov/taxes/income-tax/individual-income/individual-income-tax-rate-schedule/
          https://tax.idaho.gov/wp-content/uploads/forms/EIN00046/EIN00046_03-01-2023.pdf
          https://tax.idaho.gov/taxes/income-tax/individual-income/idaho-source-income/
          https://tax.idaho.gov/taxes/income-tax/individual-income/idaho-source-income/
        
        Nevada:
        -         
        - https://www.nvsos.gov/sos/businesses/the-nevada-advantage
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Idaho', 'state_occupation': 'Nevada', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def idaho_and_new_hampshire(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Idaho:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        New Hampshire:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Idaho:
        - https://tax.idaho.gov/taxes/income-tax/individual-income/individual-income-tax-rate-schedule/
          https://tax.idaho.gov/wp-content/uploads/forms/EIN00046/EIN00046_03-01-2023.pdf
          https://tax.idaho.gov/taxes/income-tax/individual-income/idaho-source-income/
          https://tax.idaho.gov/taxes/income-tax/individual-income/idaho-source-income/
        
        New Hampshire:
        -         
        - https://www.revenue.nh.gov/assistance/tax-overview.htm
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Idaho', 'state_occupation': 'NewHampshire', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def idaho_and_new_jersey(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Idaho:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        New Jersey:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Idaho:
        - https://tax.idaho.gov/taxes/income-tax/individual-income/individual-income-tax-rate-schedule/
          https://tax.idaho.gov/wp-content/uploads/forms/EIN00046/EIN00046_03-01-2023.pdf
          https://tax.idaho.gov/taxes/income-tax/individual-income/idaho-source-income/
          https://tax.idaho.gov/taxes/income-tax/individual-income/idaho-source-income/
        
        New Jersey:
        -         
        - https://www.state.nj.us/treasury/taxation/taxtables.shtml
          https://www.state.nj.us/treasury/taxation/njit13.shtml
          https://www.nj.gov/treasury/taxation/njit25.shtml
          https://www.state.nj.us/treasury/taxation/individuals/freqqiti.shtml
          https://www.state.nj.us/treasury/taxation/localtax.shtml
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Idaho', 'state_occupation': 'NewJersey', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def idaho_and_new_mexico(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Idaho:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        New Mexico:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Idaho:
        - https://tax.idaho.gov/taxes/income-tax/individual-income/individual-income-tax-rate-schedule/
          https://tax.idaho.gov/wp-content/uploads/forms/EIN00046/EIN00046_03-01-2023.pdf
          https://tax.idaho.gov/taxes/income-tax/individual-income/idaho-source-income/
          https://tax.idaho.gov/taxes/income-tax/individual-income/idaho-source-income/
        
        New Mexico:
        -         
        - https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/fdf3c548-8aba-4b9c-9eb4-bb564c716015/FYI-104.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/0bf943fd-652e-4400-bb1b-cd8397eb5a95/2022pit-adj.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/2f1a6781-9534-4436-b427-1557f9592099/2022pit-adj-ins.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/0bf943fd-652e-4400-bb1b-cd8397eb5a95/2022pit-adj.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/1afc56af-ea90-4d48-82e5-1f9aeb43255a/PITbook2022.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/fdf3c548-8aba-4b9c-9eb4-bb564c716015/FYI-104.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Idaho', 'state_occupation': 'NewMexico', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def idaho_and_new_york(self, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Idaho:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        New York:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Idaho:
        - https://tax.idaho.gov/taxes/income-tax/individual-income/individual-income-tax-rate-schedule/
          https://tax.idaho.gov/wp-content/uploads/forms/EIN00046/EIN00046_03-01-2023.pdf
          https://tax.idaho.gov/taxes/income-tax/individual-income/idaho-source-income/
          https://tax.idaho.gov/taxes/income-tax/individual-income/idaho-source-income/
        
        New York:
        -         
        - https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nys-tax-rate-schedule
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nys-tax-rate-schedule
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nys-tax-rate-schedule
          https://www.tax.ny.gov/pit/
          https://www.tax.ny.gov/pit/file/nonresident-faqs.htm
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#yonkers-tax-rate-schedule
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nyc-tax-rate-schedule
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Idaho', 'state_occupation': 'NewYork', "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def idaho_and_north_carolina(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Idaho:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        North Carolina:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Idaho:
        - https://tax.idaho.gov/taxes/income-tax/individual-income/individual-income-tax-rate-schedule/
          https://tax.idaho.gov/wp-content/uploads/forms/EIN00046/EIN00046_03-01-2023.pdf
          https://tax.idaho.gov/taxes/income-tax/individual-income/idaho-source-income/
          https://tax.idaho.gov/taxes/income-tax/individual-income/idaho-source-income/
        
        North Carolina:
        -         
        - https://www.ncdor.gov/taxes-forms/tax-rate-schedules
          https://www.ncdor.gov/taxes-forms/individual-income-tax/north-carolina-standard-deduction-or-north-carolina-itemized-deductions
          https://www.ncdor.gov/taxes-forms/individual-income-tax
          https://www.ncdor.gov/taxes-forms/individual-income-tax
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Idaho', 'state_occupation': 'NorthCarolina', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def idaho_and_north_dakota(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Idaho:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        North Dakota:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Idaho:
        - https://tax.idaho.gov/taxes/income-tax/individual-income/individual-income-tax-rate-schedule/
          https://tax.idaho.gov/wp-content/uploads/forms/EIN00046/EIN00046_03-01-2023.pdf
          https://tax.idaho.gov/taxes/income-tax/individual-income/idaho-source-income/
          https://tax.idaho.gov/taxes/income-tax/individual-income/idaho-source-income/
        
        North Dakota:
        -         
        - https://www.tax.nd.gov/forms
          https://www.tax.nd.gov/sites/www/files/documents/forms/individual/2022-iit/2022-individual-income-tax-booklet.pdf
          https://www.tax.nd.gov/sites/www/files/documents/forms/2022-individual-income-tax-booklet.pdf
          https://www.tax.nd.gov/tax-exemptions-credits/income-tax-exemptions-credits
          https://www.tax.nd.gov/sites/www/files/documents/forms/business/ndwrfillable.pdf
          https://www.tax.nd.gov/news/tax-legislative-changes/significant-changes-law/individual-income-tax-history
          https://www.tax.nd.gov/forms
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Idaho', 'state_occupation': 'NorthDakota', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def idaho_and_ohio(self, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Idaho:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Ohio:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Idaho:
        - https://tax.idaho.gov/taxes/income-tax/individual-income/individual-income-tax-rate-schedule/
          https://tax.idaho.gov/wp-content/uploads/forms/EIN00046/EIN00046_03-01-2023.pdf
          https://tax.idaho.gov/taxes/income-tax/individual-income/idaho-source-income/
          https://tax.idaho.gov/taxes/income-tax/individual-income/idaho-source-income/
        
        Ohio:
        -         
        - https://tax.ohio.gov/individual/resources/annual-tax-rates
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://tax.ohio.gov/static/forms/employer_withholding/generic/wth_it4nr.pdf
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://www.ritaohio.com/TaxRatesTable
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Idaho', 'state_occupation': 'Ohio', "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def idaho_and_oklahoma(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Idaho:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Oklahoma:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Idaho:
        - https://tax.idaho.gov/taxes/income-tax/individual-income/individual-income-tax-rate-schedule/
          https://tax.idaho.gov/wp-content/uploads/forms/EIN00046/EIN00046_03-01-2023.pdf
          https://tax.idaho.gov/taxes/income-tax/individual-income/idaho-source-income/
          https://tax.idaho.gov/taxes/income-tax/individual-income/idaho-source-income/
        
        Oklahoma:
        -         
        - https://oklahoma.gov/content/dam/ok/en/tax/documents/resources/publications/businesses/withholding-tables/WHTables-2022.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Idaho', 'state_occupation': 'Oklahoma', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def idaho_and_oregon(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Idaho:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Oregon:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Idaho:
        - https://tax.idaho.gov/taxes/income-tax/individual-income/individual-income-tax-rate-schedule/
          https://tax.idaho.gov/wp-content/uploads/forms/EIN00046/EIN00046_03-01-2023.pdf
          https://tax.idaho.gov/taxes/income-tax/individual-income/idaho-source-income/
          https://tax.idaho.gov/taxes/income-tax/individual-income/idaho-source-income/
        
        Oregon:
        -         
        - https://www.oregon.gov/dor/programs/individuals/Pages/PIT.aspx
          https://www.oregon.gov/dor/programs/individuals/Documents/Full%20year%20resident
         %20Form%20or-40%20filers.pdf
          https://www.oregon.gov/dor/programs/individuals/Documents/Part-year%20and%20nonresident
         %20Form%20or-40-P%20and%20or-40-N%20filers.pdf.pdf
          https://www.oregon.gov/dor/forms/FormsPubs/form-or-40_101-040_2022.pdf
          https://www.oregon.gov/dor/forms/FormsPubs/form-or-40_101-040_2022.pdf
          https://www.oregon.gov/dor/forms/FormsPubs/form-or-40_101-040_2022.pdf
          https://www.portland.gov/revenue/personal-tax
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Idaho', 'state_occupation': 'Oregon', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def idaho_and_pennsylvania(self, county_residence='Other', state_agi=75000, county_occupation='Other', filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Idaho:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Pennsylvania:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        county_residence : str, optional
            County of residence. Default is Other.
        county_occupation : str, optional
            County of occupation. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Idaho:
        - https://tax.idaho.gov/taxes/income-tax/individual-income/individual-income-tax-rate-schedule/
          https://tax.idaho.gov/wp-content/uploads/forms/EIN00046/EIN00046_03-01-2023.pdf
          https://tax.idaho.gov/taxes/income-tax/individual-income/idaho-source-income/
          https://tax.idaho.gov/taxes/income-tax/individual-income/idaho-source-income/
        
        Pennsylvania:
        -         
        - https://www.revenue.pa.gov/Tax%20Rates/Pages/default.aspx
          https://www.revenue.pa.gov/FormsandPublications/FormsforIndividuals/PIT/Documents/2022/2022_pa-40in.pdf
          https://www.revenue.pa.gov/FormsandPublications/FormsforBusinesses/EmployerWithholding/Documents/rev-419.pdf
          https://www.revenue.pa.gov/FormsandPublications/FormsforIndividuals/PIT/Documents/2022/2022_pa-40in.pdf
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-25.htm
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        county_occupation = self.county_occupation if hasattr(self, 'county_occupation') else county_occupation
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Idaho', 'state_occupation': 'Pennsylvania', "county_residence": county_residence, "state_agi": state_agi, "county_occupation": county_occupation, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def idaho_and_rhode_island(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Idaho:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Rhode Island:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Idaho:
        - https://tax.idaho.gov/taxes/income-tax/individual-income/individual-income-tax-rate-schedule/
          https://tax.idaho.gov/wp-content/uploads/forms/EIN00046/EIN00046_03-01-2023.pdf
          https://tax.idaho.gov/taxes/income-tax/individual-income/idaho-source-income/
          https://tax.idaho.gov/taxes/income-tax/individual-income/idaho-source-income/
        
        Rhode Island:
        -         
        - https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2022_1040WE_w_0.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Idaho', 'state_occupation': 'RhodeIsland', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def idaho_and_south_carolina(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Idaho:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        South Carolina:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Idaho:
        - https://tax.idaho.gov/taxes/income-tax/individual-income/individual-income-tax-rate-schedule/
          https://tax.idaho.gov/wp-content/uploads/forms/EIN00046/EIN00046_03-01-2023.pdf
          https://tax.idaho.gov/taxes/income-tax/individual-income/idaho-source-income/
          https://tax.idaho.gov/taxes/income-tax/individual-income/idaho-source-income/
        
        South Carolina:
        -         
        - https://dor.sc.gov/resources-site/lawandpolicy/Advisory%20Opinions/IL22-15.pdf
          https://dor.sc.gov/forms-site/Forms/IITPacket_2022.pdf
          https://www.scstatehouse.gov/code/t12c006.php
          https://dor.sc.gov/resources-site/lawandpolicy/Advisory%20Opinions/IL22-15.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Idaho', 'state_occupation': 'SouthCarolina', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def idaho_and_south_dakota(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Idaho:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        South Dakota:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Idaho:
        - https://tax.idaho.gov/taxes/income-tax/individual-income/individual-income-tax-rate-schedule/
          https://tax.idaho.gov/wp-content/uploads/forms/EIN00046/EIN00046_03-01-2023.pdf
          https://tax.idaho.gov/taxes/income-tax/individual-income/idaho-source-income/
          https://tax.idaho.gov/taxes/income-tax/individual-income/idaho-source-income/
        
        South Dakota:
        -         
        - https://dor.sd.gov/individuals/taxes/
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Idaho', 'state_occupation': 'SouthDakota', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def idaho_and_tennessee(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Idaho:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Tennessee:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Idaho:
        - https://tax.idaho.gov/taxes/income-tax/individual-income/individual-income-tax-rate-schedule/
          https://tax.idaho.gov/wp-content/uploads/forms/EIN00046/EIN00046_03-01-2023.pdf
          https://tax.idaho.gov/taxes/income-tax/individual-income/idaho-source-income/
          https://tax.idaho.gov/taxes/income-tax/individual-income/idaho-source-income/
        
        Tennessee:
        -         
        - https://www.tn.gov/revenue/taxes.html
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Idaho', 'state_occupation': 'Tennessee', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def idaho_and_texas(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Idaho:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Texas:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Idaho:
        - https://tax.idaho.gov/taxes/income-tax/individual-income/individual-income-tax-rate-schedule/
          https://tax.idaho.gov/wp-content/uploads/forms/EIN00046/EIN00046_03-01-2023.pdf
          https://tax.idaho.gov/taxes/income-tax/individual-income/idaho-source-income/
          https://tax.idaho.gov/taxes/income-tax/individual-income/idaho-source-income/
        
        Texas:
        -         
        - https://comptroller.texas.gov/taxes/
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Idaho', 'state_occupation': 'Texas', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def idaho_and_utah(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Idaho:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Utah:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Idaho:
        - https://tax.idaho.gov/taxes/income-tax/individual-income/individual-income-tax-rate-schedule/
          https://tax.idaho.gov/wp-content/uploads/forms/EIN00046/EIN00046_03-01-2023.pdf
          https://tax.idaho.gov/taxes/income-tax/individual-income/idaho-source-income/
          https://tax.idaho.gov/taxes/income-tax/individual-income/idaho-source-income/
        
        Utah:
        -         
        - https://incometax.utah.gov/paying/tax-rates
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Idaho', 'state_occupation': 'Utah', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def idaho_and_vermont(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Idaho:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Vermont:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Idaho:
        - https://tax.idaho.gov/taxes/income-tax/individual-income/individual-income-tax-rate-schedule/
          https://tax.idaho.gov/wp-content/uploads/forms/EIN00046/EIN00046_03-01-2023.pdf
          https://tax.idaho.gov/taxes/income-tax/individual-income/idaho-source-income/
          https://tax.idaho.gov/taxes/income-tax/individual-income/idaho-source-income/
        
        Vermont:
        -         
        - https://tax.vermont.gov/sites/tax/files/documents/GB-1210-2023.pdf
          https://tax.vermont.gov/sites/tax/files/documents/Income%20Booklet-2022.pdf
          https://tax.vermont.gov/sites/tax/files/documents/Income%20Booklet-2022.pdf
          https://tax.vermont.gov/sites/tax/files/documents/Income%20Booklet-2022.pdf
          https://tax.vermont.gov/individuals/personal-income-tax/tax-credits
          https://tax.vermont.gov/individuals/personal-income-tax/taxable-income
          https://tax.vermont.gov/individuals/personal-income-tax/local-option-tax
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Idaho', 'state_occupation': 'Vermont', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def idaho_and_virginia(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Idaho:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Virginia:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Idaho:
        - https://tax.idaho.gov/taxes/income-tax/individual-income/individual-income-tax-rate-schedule/
          https://tax.idaho.gov/wp-content/uploads/forms/EIN00046/EIN00046_03-01-2023.pdf
          https://tax.idaho.gov/taxes/income-tax/individual-income/idaho-source-income/
          https://tax.idaho.gov/taxes/income-tax/individual-income/idaho-source-income/
        
        Virginia:
        -         
        - https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/reciprocity
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Idaho', 'state_occupation': 'Virginia', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def idaho_and_washington(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Idaho:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Washington:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Idaho:
        - https://tax.idaho.gov/taxes/income-tax/individual-income/individual-income-tax-rate-schedule/
          https://tax.idaho.gov/wp-content/uploads/forms/EIN00046/EIN00046_03-01-2023.pdf
          https://tax.idaho.gov/taxes/income-tax/individual-income/idaho-source-income/
          https://tax.idaho.gov/taxes/income-tax/individual-income/idaho-source-income/
        
        Washington:
        -         
        - https://dor.wa.gov/taxes-rates/income-tax
          https://dor.wa.gov/taxes-rates/other-taxes/capital-gains-tax
          https://dor.wa.gov/taxes-rates/income-tax
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Idaho', 'state_occupation': 'Washington', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def idaho_and_west_virginia(self, pay_periods=24, capital_gains_short=0, county_occupation='Other', dependents=0, sources=False, state_agi=75000, filers_over_65=0, capital_gains_long=0, filing_status='single', income=100000, credits_state=0):
        """
        --- Parameters ----
        
        Idaho:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        West Virginia:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_occupation : str, optional
            County of occupation. Default is Other.
        pay_periods : int, optional
            Number of pay periods. Options are 13, 24, 26, 52. Default is 24.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Idaho:
        - https://tax.idaho.gov/taxes/income-tax/individual-income/individual-income-tax-rate-schedule/
          https://tax.idaho.gov/wp-content/uploads/forms/EIN00046/EIN00046_03-01-2023.pdf
          https://tax.idaho.gov/taxes/income-tax/individual-income/idaho-source-income/
          https://tax.idaho.gov/taxes/income-tax/individual-income/idaho-source-income/
        
        West Virginia:
        -         
        - https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://tax.wv.gov/Documents/TSD/tsd438.pdf
          https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2011/Taxes-11-35att.html
        """
        
        # Check if the parameters are set upstream in the instance
        pay_periods = self.pay_periods if hasattr(self, 'pay_periods') else pay_periods
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        county_occupation = self.county_occupation if hasattr(self, 'county_occupation') else county_occupation
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Idaho', 'state_occupation': 'WestVirginia', "pay_periods": pay_periods, "capital_gains_short": capital_gains_short, "county_occupation": county_occupation, "dependents": dependents, "sources": sources, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def idaho_and_wisconsin(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Idaho:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Wisconsin:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Idaho:
        - https://tax.idaho.gov/taxes/income-tax/individual-income/individual-income-tax-rate-schedule/
          https://tax.idaho.gov/wp-content/uploads/forms/EIN00046/EIN00046_03-01-2023.pdf
          https://tax.idaho.gov/taxes/income-tax/individual-income/idaho-source-income/
          https://tax.idaho.gov/taxes/income-tax/individual-income/idaho-source-income/
        
        Wisconsin:
        -         
        - https://www.revenue.wi.gov/TaxForms2023/2023-Form1-ES-Inst.pdf
          https://www.revenue.wi.gov/TaxForms2023/2023-Form1-ES-Inst.pdf
          https://www.revenue.wi.gov/TaxForms2023/2023-Form1-ES-Inst.pdf
          https://www.revenue.wi.gov/Pages/FAQS/pcs-work.aspx
          https://www.revenue.wi.gov/Dor%20Publications/pb103.pdf
          https://www.revenue.wi.gov/Pages/FAQS/pcs-work.aspx
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Idaho', 'state_occupation': 'Wisconsin', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def idaho_and_wyoming(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Idaho:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Wyoming:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Idaho:
        - https://tax.idaho.gov/taxes/income-tax/individual-income/individual-income-tax-rate-schedule/
          https://tax.idaho.gov/wp-content/uploads/forms/EIN00046/EIN00046_03-01-2023.pdf
          https://tax.idaho.gov/taxes/income-tax/individual-income/idaho-source-income/
          https://tax.idaho.gov/taxes/income-tax/individual-income/idaho-source-income/
        
        Wyoming:
        -         
        - https://revenue.wyo.gov/
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Idaho', 'state_occupation': 'Wyoming', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def illinois_and_alabama(self, county_residence='Other', state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Illinois:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Alabama:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Illinois:
        - https://tax.illinois.gov/research/taxrates/income.html#IndividualIncome
          https://tax.illinois.gov/content/dam/soi/en/web/tax/forms/incometax/documents/currentyear/individual/il-1040-instr.pdf
          https://tax.illinois.gov/questionsandanswers/12.html
          https://tax.illinois.gov/content/dam/soi/en/web/tax/forms/incometax/documents/currentyear/individual/il-1040-instr.pdf
          https://tax.illinois.gov/research/taxrates/personalproperty.html
          https://tax.illinois.gov/localgovernments/income.html
        
        Alabama:
        -         
        - https://www.revenue.alabama.gov/ultraviewer/viewer/basic_viewer/index.html?form=2023/02/22f40.pdf
          https://www.revenue.alabama.gov/faqs/what-is-alabamas-individual-income-tax-rate/
          https://www.revenue.alabama.gov/ultraviewer/viewer/basic_viewer/index.html?form=2023/02/22f40.pdf
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-24.htm?taxmap=true
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-24.htm?taxmap=true
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-24.htm?taxmap=true
          https://www.revenue.alabama.gov/individual-corporate/income-to-be-reported-on-the-alabama-income-tax-return/
          https://www.almonline.org/TaxRates.aspx#OccupationalTax
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Illinois', 'state_occupation': 'Alabama', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def illinois_and_alaska(self, state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Illinois:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Alaska:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Illinois:
        - https://tax.illinois.gov/research/taxrates/income.html#IndividualIncome
          https://tax.illinois.gov/content/dam/soi/en/web/tax/forms/incometax/documents/currentyear/individual/il-1040-instr.pdf
          https://tax.illinois.gov/questionsandanswers/12.html
          https://tax.illinois.gov/content/dam/soi/en/web/tax/forms/incometax/documents/currentyear/individual/il-1040-instr.pdf
          https://tax.illinois.gov/research/taxrates/personalproperty.html
          https://tax.illinois.gov/localgovernments/income.html
        
        Alaska:
        -         
        - https://www.commerce.alaska.gov/web/dcra/officeofthestateassessor/alaskataxfacts.aspx
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Illinois', 'state_occupation': 'Alaska', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def illinois_and_arizona(self, state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Illinois:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Arizona:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        credits_state : int, optional
            Credits for state taxes. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Illinois:
        - https://tax.illinois.gov/research/taxrates/income.html#IndividualIncome
          https://tax.illinois.gov/content/dam/soi/en/web/tax/forms/incometax/documents/currentyear/individual/il-1040-instr.pdf
          https://tax.illinois.gov/questionsandanswers/12.html
          https://tax.illinois.gov/content/dam/soi/en/web/tax/forms/incometax/documents/currentyear/individual/il-1040-instr.pdf
          https://tax.illinois.gov/research/taxrates/personalproperty.html
          https://tax.illinois.gov/localgovernments/income.html
        
        Arizona:
        -         
        - https://azdor.gov/forms/individual/form-140-x-y-tables
          https://azdor.gov/forms/individual/form-140-x-y-tables
          https://azdor.gov/forms/individual/form-140-x-y-tables
          https://azdor.gov/businesses-arizona/withholding-tax/withholding-exceptions
          "https://azdor.gov/individuals/income-tax-filing-assistance/identifying-other-taxable-income#:~:text=In%20regards%20to%20capital%20gains
         the%20individuals%20regular%20tax%20rate.
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Illinois', 'state_occupation': 'Arizona', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def illinois_and_arkansas(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Illinois:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Arkansas:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Illinois:
        - https://tax.illinois.gov/research/taxrates/income.html#IndividualIncome
          https://tax.illinois.gov/content/dam/soi/en/web/tax/forms/incometax/documents/currentyear/individual/il-1040-instr.pdf
          https://tax.illinois.gov/questionsandanswers/12.html
          https://tax.illinois.gov/content/dam/soi/en/web/tax/forms/incometax/documents/currentyear/individual/il-1040-instr.pdf
          https://tax.illinois.gov/research/taxrates/personalproperty.html
          https://tax.illinois.gov/localgovernments/income.html
        
        Arkansas:
        -         
        - https://www.arkansasedc.com/why-arkansas/business-climate/tax-structure/personal-income-tax
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_and_AR1000NR_Instructions.pdf
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_FullYearResidentIndividualIncomeTaxReturn.pdf
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_and_AR1000NR_Instructions.pdf
          https://www.arkansasedc.com/why-arkansas/business-climate/tax-structure/capital-gains-tax-reduction
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_and_AR1000NR_Instructions.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Illinois', 'state_occupation': 'Arkansas', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def illinois_and_california(self, state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Illinois:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        California:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Illinois:
        - https://tax.illinois.gov/research/taxrates/income.html#IndividualIncome
          https://tax.illinois.gov/content/dam/soi/en/web/tax/forms/incometax/documents/currentyear/individual/il-1040-instr.pdf
          https://tax.illinois.gov/questionsandanswers/12.html
          https://tax.illinois.gov/content/dam/soi/en/web/tax/forms/incometax/documents/currentyear/individual/il-1040-instr.pdf
          https://tax.illinois.gov/research/taxrates/personalproperty.html
          https://tax.illinois.gov/localgovernments/income.html
        
        California:
        -         
        - https://www.ftb.ca.gov/forms/2022/2022-540-tax-rate-schedules.pdf
          https://www.ftb.ca.gov/file/personal/deductions/index.html
          https://www.ftb.ca.gov/about-ftb/newsroom/caleitc/eligibility-and-credit-information.html
          https://www.ftb.ca.gov/file/personal/income-types/capital-gains-and-losses.html#:~:text=California%20does%20not%20have%20a
         are%20taxed%20as%20ordinary%20income
          https://www.ftb.ca.gov/file/personal/
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Illinois', 'state_occupation': 'California', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def illinois_and_colorado(self, county_residence='Other', state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Illinois:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Colorado:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Illinois:
        - https://tax.illinois.gov/research/taxrates/income.html#IndividualIncome
          https://tax.illinois.gov/content/dam/soi/en/web/tax/forms/incometax/documents/currentyear/individual/il-1040-instr.pdf
          https://tax.illinois.gov/questionsandanswers/12.html
          https://tax.illinois.gov/content/dam/soi/en/web/tax/forms/incometax/documents/currentyear/individual/il-1040-instr.pdf
          https://tax.illinois.gov/research/taxrates/personalproperty.html
          https://tax.illinois.gov/localgovernments/income.html
        
        Colorado:
        -         
        - https://leg.colorado.gov/agencies/legislative-council-staff/individual-income-tax%C2%A0
          https://leg.colorado.gov/agencies/legislative-council-staff/individual-income-tax%C2%A0
          https://tax.colorado.gov/sites/tax/files/documents/ITT_Capital_Gain_Subtraction_Feb_2022.pdf
          https://www.auroragov.org/business_services/taxes/occupational_privilege_tax
          https://www.denvergov.org/Government/Agencies-Departments-Offices/Agencies-Departments-Offices-Directory/Department-of-Finance/Our-Divisions/Treasury/Business-Tax-Information#:~:text=Occupational%20privilege%20tax
         -%28show%20below%29&text=Each%20taxable%20employee%20is%20liable
         month%20for%20each%20taxable%20employee
          https://www.glendale.co.us/355/Occupational-Privilege-Tax
          https://www.greenwoodvillage.com/1220/Occupational-Privilege-Tax-OPT
          https://ci.sheridan.co.us/288/Occupational-Privilege-Tax
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Illinois', 'state_occupation': 'Colorado', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def illinois_and_connecticut(self, state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Illinois:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Connecticut:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Illinois:
        - https://tax.illinois.gov/research/taxrates/income.html#IndividualIncome
          https://tax.illinois.gov/content/dam/soi/en/web/tax/forms/incometax/documents/currentyear/individual/il-1040-instr.pdf
          https://tax.illinois.gov/questionsandanswers/12.html
          https://tax.illinois.gov/content/dam/soi/en/web/tax/forms/incometax/documents/currentyear/individual/il-1040-instr.pdf
          https://tax.illinois.gov/research/taxrates/personalproperty.html
          https://tax.illinois.gov/localgovernments/income.html
        
        Connecticut:
        -         
        - https://portal.ct.gov/-/media/DRS/Forms/2022/Income/2022-CT-1040-Instructions_1222.pdf
          https://www.cga.ct.gov/2022/rpt/pdf/2022-R-0108.pdf
          https://www.cga.ct.gov/current/pub/chap_229.htm#sec_12-703
          https://portal.ct.gov/-/media/DRS/Forms/2022/Income/2022-CT-1040-Instructions_1222.pdf
          "https://portal.ct.gov/DRS/Publications/TSSNs/TSSN-29#:~:text=An%20individuals%20net%20capital%20gains
         portion%20of%20Social%20Security%20benefits"
          "https://portal.ct.gov/DRS/Publications/TSSNs/TSSN-29#:~:text=An%20individuals%20net%20capital%20gains
         portion%20of%20Social%20Security%20benefits"
          https://portal.ct.gov/DRS/Individuals/Resident-Income-Tax/Tax-Information
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Illinois', 'state_occupation': 'Connecticut', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def illinois_and_delaware(self, state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, federal_agi=75000, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Illinois:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Delaware:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        federal_agi : int, optional
            Adjusted Gross Income (AGI) for federal taxes. Default is 0. Sources - https://www.irs.gov/e-file-providers/definition-of-adjusted-gross-income, https://www.irs.gov/newsroom/taxpayers-should-know-the-difference-between-standard-and-itemized-deductions
        income : int, optional
            Gross income. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Illinois:
        - https://tax.illinois.gov/research/taxrates/income.html#IndividualIncome
          https://tax.illinois.gov/content/dam/soi/en/web/tax/forms/incometax/documents/currentyear/individual/il-1040-instr.pdf
          https://tax.illinois.gov/questionsandanswers/12.html
          https://tax.illinois.gov/content/dam/soi/en/web/tax/forms/incometax/documents/currentyear/individual/il-1040-instr.pdf
          https://tax.illinois.gov/research/taxrates/personalproperty.html
          https://tax.illinois.gov/localgovernments/income.html
        
        Delaware:
        -         
        - https://revenue.delaware.gov/software-developer/tax-rate-changes/
          https://revenuefiles.delaware.gov/2022/PIT-RES_TY22_2022-01_PaperInteractive.pdf
          https://revenuefiles.delaware.gov/2022/PIT-RES_TY22_2022-02_Instructions.pdf
          https://revenue.delaware.gov/frequently-asked-questions/personal-income-tax-faqs/
          https://www.newcastlede.gov/DocumentCenter/View/20147/City-of-Wilmington-Wage-Tax-Information-PDF
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        federal_agi = self.federal_agi if hasattr(self, 'federal_agi') else federal_agi
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Illinois', 'state_occupation': 'Delaware', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "federal_agi": federal_agi, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def illinois_and_florida(self, state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Illinois:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Florida:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Illinois:
        - https://tax.illinois.gov/research/taxrates/income.html#IndividualIncome
          https://tax.illinois.gov/content/dam/soi/en/web/tax/forms/incometax/documents/currentyear/individual/il-1040-instr.pdf
          https://tax.illinois.gov/questionsandanswers/12.html
          https://tax.illinois.gov/content/dam/soi/en/web/tax/forms/incometax/documents/currentyear/individual/il-1040-instr.pdf
          https://tax.illinois.gov/research/taxrates/personalproperty.html
          https://tax.illinois.gov/localgovernments/income.html
        
        Florida:
        -         
        - https://floridarevenue.com/taxes/taxesfees/Pages/tax_interest_rates.aspx
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Illinois', 'state_occupation': 'Florida', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def illinois_and_georgia(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Illinois:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Georgia:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Illinois:
        - https://tax.illinois.gov/research/taxrates/income.html#IndividualIncome
          https://tax.illinois.gov/content/dam/soi/en/web/tax/forms/incometax/documents/currentyear/individual/il-1040-instr.pdf
          https://tax.illinois.gov/questionsandanswers/12.html
          https://tax.illinois.gov/content/dam/soi/en/web/tax/forms/incometax/documents/currentyear/individual/il-1040-instr.pdf
          https://tax.illinois.gov/research/taxrates/personalproperty.html
          https://tax.illinois.gov/localgovernments/income.html
        
        Georgia:
        -         
        - https://dor.georgia.gov/tax-tables-georgia-tax-rate-schedule
          https://apps.dor.ga.gov/FillableForms/PDFViewer/Index?form=2022GA500
          https://apps.dor.ga.gov/FillableForms/PDFViewer/Index?form=2022GA500
          https://apps.dor.ga.gov/FillableForms/PDFViewer/Index?form=2022GA500
          https://dor.georgia.gov/years-individual-income-tax-forms
          https://dor.georgia.gov/years-individual-income-tax-forms
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Illinois', 'state_occupation': 'Georgia', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def illinois_and_hawaii(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Illinois:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Hawaii:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Illinois:
        - https://tax.illinois.gov/research/taxrates/income.html#IndividualIncome
          https://tax.illinois.gov/content/dam/soi/en/web/tax/forms/incometax/documents/currentyear/individual/il-1040-instr.pdf
          https://tax.illinois.gov/questionsandanswers/12.html
          https://tax.illinois.gov/content/dam/soi/en/web/tax/forms/incometax/documents/currentyear/individual/il-1040-instr.pdf
          https://tax.illinois.gov/research/taxrates/personalproperty.html
          https://tax.illinois.gov/localgovernments/income.html
        
        Hawaii:
        -         
        - https://tax.hawaii.gov/forms/d_18table-on/d_18table-on_p13/
          https://files.hawaii.gov/tax/forms/2018/18table-on.pdf
          https://files.hawaii.gov/tax/news/pubs/22outline.pdf
          https://files.hawaii.gov/tax/news/pubs/FastTaxRefGuide_2017.pdf
          https://www.tfhawaii.org/wordpress/state-tax-resources/mini-tax-guides/individual-income-tax-chapter-235/#:~:text=Long%20term%20capital%20gains%20are%20taxed%20at%20a%20maximum%20of%207.25%25
          https://www.grassrootinstitute.org/2023/02/hb337-increasing-tax-on-capital-gains-a-bad-risk/
          https://files.hawaii.gov/tax/news/pubs/22outline.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Illinois', 'state_occupation': 'Hawaii', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def illinois_and_idaho(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Illinois:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Idaho:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Illinois:
        - https://tax.illinois.gov/research/taxrates/income.html#IndividualIncome
          https://tax.illinois.gov/content/dam/soi/en/web/tax/forms/incometax/documents/currentyear/individual/il-1040-instr.pdf
          https://tax.illinois.gov/questionsandanswers/12.html
          https://tax.illinois.gov/content/dam/soi/en/web/tax/forms/incometax/documents/currentyear/individual/il-1040-instr.pdf
          https://tax.illinois.gov/research/taxrates/personalproperty.html
          https://tax.illinois.gov/localgovernments/income.html
        
        Idaho:
        -         
        - https://tax.idaho.gov/taxes/income-tax/individual-income/individual-income-tax-rate-schedule/
          https://tax.idaho.gov/wp-content/uploads/forms/EIN00046/EIN00046_03-01-2023.pdf
          https://tax.idaho.gov/taxes/income-tax/individual-income/idaho-source-income/
          https://tax.idaho.gov/taxes/income-tax/individual-income/idaho-source-income/
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Illinois', 'state_occupation': 'Idaho', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def illinois_and_indiana(self, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Illinois:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Indiana:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Illinois:
        - https://tax.illinois.gov/research/taxrates/income.html#IndividualIncome
          https://tax.illinois.gov/content/dam/soi/en/web/tax/forms/incometax/documents/currentyear/individual/il-1040-instr.pdf
          https://tax.illinois.gov/questionsandanswers/12.html
          https://tax.illinois.gov/content/dam/soi/en/web/tax/forms/incometax/documents/currentyear/individual/il-1040-instr.pdf
          https://tax.illinois.gov/research/taxrates/personalproperty.html
          https://tax.illinois.gov/localgovernments/income.html
        
        Indiana:
        -         
        - https://www.in.gov/dor/tax-forms/2022-individual-income-tax-forms/
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-32.htm
          https://www.in.gov/dor/files/ib28.pdf
          https://www.in.gov/dor/tax-forms/2022-individual-income-tax-forms/
          https://www.in.gov/dor/files/2022-county-tax-rates-and-codes.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Illinois', 'state_occupation': 'Indiana', "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def illinois_and_iowa(self, state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Illinois:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Iowa:
        


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Illinois:
        - https://tax.illinois.gov/research/taxrates/income.html#IndividualIncome
          https://tax.illinois.gov/content/dam/soi/en/web/tax/forms/incometax/documents/currentyear/individual/il-1040-instr.pdf
          https://tax.illinois.gov/questionsandanswers/12.html
          https://tax.illinois.gov/content/dam/soi/en/web/tax/forms/incometax/documents/currentyear/individual/il-1040-instr.pdf
          https://tax.illinois.gov/research/taxrates/personalproperty.html
          https://tax.illinois.gov/localgovernments/income.html
        
        Iowa:
        -         
        - https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
          https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
          https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
          https://tax.iowa.gov/iowa-illinois-reciprocal-agreement
          https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Illinois', 'state_occupation': 'Iowa', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def illinois_and_kansas(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Illinois:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Kansas:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Illinois:
        - https://tax.illinois.gov/research/taxrates/income.html#IndividualIncome
          https://tax.illinois.gov/content/dam/soi/en/web/tax/forms/incometax/documents/currentyear/individual/il-1040-instr.pdf
          https://tax.illinois.gov/questionsandanswers/12.html
          https://tax.illinois.gov/content/dam/soi/en/web/tax/forms/incometax/documents/currentyear/individual/il-1040-instr.pdf
          https://tax.illinois.gov/research/taxrates/personalproperty.html
          https://tax.illinois.gov/localgovernments/income.html
        
        Kansas:
        -         
        - https://www.ksrevenue.gov/taxrates.html
          https://www.ksrevenue.gov/incomebook22.html#0
          https://www.ksrevenue.gov/incomebook22.html#0
          https://www.ksrevenue.gov/incomebook22.html#0
          https://www.ksrevenue.gov/incomebook22.html
          https://www.ksrevenue.gov/incomebook22.html
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Illinois', 'state_occupation': 'Kansas', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def illinois_and_kentucky(self, state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Illinois:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Kentucky:
        


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Illinois:
        - https://tax.illinois.gov/research/taxrates/income.html#IndividualIncome
          https://tax.illinois.gov/content/dam/soi/en/web/tax/forms/incometax/documents/currentyear/individual/il-1040-instr.pdf
          https://tax.illinois.gov/questionsandanswers/12.html
          https://tax.illinois.gov/content/dam/soi/en/web/tax/forms/incometax/documents/currentyear/individual/il-1040-instr.pdf
          https://tax.illinois.gov/research/taxrates/personalproperty.html
          https://tax.illinois.gov/localgovernments/income.html
        
        Kentucky:
        -         
        - https://revenue.ky.gov/News/Pages/Dor-Announces-Updates-to-Individual-Income-Tax-for-2023-Tax-Year.aspx
          https://revenue.ky.gov/Individual/Individual-Income-Tax/Pages/default.aspx
          https://revenue.ky.gov/News/Pages/Dor-Announces-Updates-to-Individual-Income-Tax-for-2023-Tax-Year.aspx
          https://revenue.ky.gov/Individual/Individual-Income-Tax/Pages/default.aspx
          https://revenue.ky.gov/Dor%20Training%20Materials/103%20KAR%2017.140.%20Individual%20income%20tax%20-%20reciprocity%20-%20nonresidents.pdf
          https://revenue.ky.gov/Individual/Individual-Income-Tax/Pages/default.aspx
          https://nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2021/Taxes-21-32.htm?taxmap=true
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Illinois', 'state_occupation': 'Kentucky', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def illinois_and_louisiana(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Illinois:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Louisiana:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Illinois:
        - https://tax.illinois.gov/research/taxrates/income.html#IndividualIncome
          https://tax.illinois.gov/content/dam/soi/en/web/tax/forms/incometax/documents/currentyear/individual/il-1040-instr.pdf
          https://tax.illinois.gov/questionsandanswers/12.html
          https://tax.illinois.gov/content/dam/soi/en/web/tax/forms/incometax/documents/currentyear/individual/il-1040-instr.pdf
          https://tax.illinois.gov/research/taxrates/personalproperty.html
          https://tax.illinois.gov/localgovernments/income.html
        
        Louisiana:
        -         
        - https://revenue.louisiana.gov/individualincometax
          https://revenue.louisiana.gov/individualincometax
          https://www.revenue.louisiana.gov/taxforms/6935(11_02)F.pdf
          https://revenue.louisiana.gov/individualincometax
          https://revenue.louisiana.gov/individualincometax
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Illinois', 'state_occupation': 'Louisiana', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def illinois_and_maine(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Illinois:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Maine:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Illinois:
        - https://tax.illinois.gov/research/taxrates/income.html#IndividualIncome
          https://tax.illinois.gov/content/dam/soi/en/web/tax/forms/incometax/documents/currentyear/individual/il-1040-instr.pdf
          https://tax.illinois.gov/questionsandanswers/12.html
          https://tax.illinois.gov/content/dam/soi/en/web/tax/forms/incometax/documents/currentyear/individual/il-1040-instr.pdf
          https://tax.illinois.gov/research/taxrates/personalproperty.html
          https://tax.illinois.gov/localgovernments/income.html
        
        Maine:
        -         
        - https://www.maine.gov/revenue/tax-return-forms/individual-income-tax-2022
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/ind_tax_rate_sched_2022.pdf
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/22_1040me_book_gen_instruc_revisedFeb23.pdf
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/22_1040me_book_gen_instruc_revisedFeb23.pdf
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/22_1040me_book_gen_instruc_revisedFeb23.pdf
          https://www.maine.gov/revenue/tax-return-forms/individual-income-tax-2022
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Illinois', 'state_occupation': 'Maine', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def illinois_and_maryland(self, county_residence='Other', state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Illinois:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Maryland:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Illinois:
        - https://tax.illinois.gov/research/taxrates/income.html#IndividualIncome
          https://tax.illinois.gov/content/dam/soi/en/web/tax/forms/incometax/documents/currentyear/individual/il-1040-instr.pdf
          https://tax.illinois.gov/questionsandanswers/12.html
          https://tax.illinois.gov/content/dam/soi/en/web/tax/forms/incometax/documents/currentyear/individual/il-1040-instr.pdf
          https://tax.illinois.gov/research/taxrates/personalproperty.html
          https://tax.illinois.gov/localgovernments/income.html
        
        Maryland:
        -         
        - https://www.marylandtaxes.gov/individual/income/tax-info/tax-rates.php
          https://www.marylandtaxes.gov/forms/22_forms/Resident_Booklet.pdf
          https://www.marylandtaxes.gov/forms/Tax_Publications/Administrative_Releases/Income_and_Estate_Tax_Releases/ar_it37.pdf
          https://www.marylandtaxes.gov/individual/index.php
          https://www.marylandtaxes.gov/individual/income/tax-info/tax-rates.php
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Illinois', 'state_occupation': 'Maryland', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def illinois_and_massachusetts(self, state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Illinois:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Massachusetts:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Illinois:
        - https://tax.illinois.gov/research/taxrates/income.html#IndividualIncome
          https://tax.illinois.gov/content/dam/soi/en/web/tax/forms/incometax/documents/currentyear/individual/il-1040-instr.pdf
          https://tax.illinois.gov/questionsandanswers/12.html
          https://tax.illinois.gov/content/dam/soi/en/web/tax/forms/incometax/documents/currentyear/individual/il-1040-instr.pdf
          https://tax.illinois.gov/research/taxrates/personalproperty.html
          https://tax.illinois.gov/localgovernments/income.html
        
        Massachusetts:
        -         
        - https://www.mass.gov/service-details/massachusetts-tax-rates
          https://www.mass.gov/service-details/view-massachusetts-personal-income-tax-exemptions
          https://www.mass.gov/service-details/view-massachusetts-personal-income-tax-exemptions
          https://www.mass.gov/service-details/view-child-and-dependent-related-credits
          https://www.mass.gov/service-details/massachusetts-tax-rates
          https://www.mass.gov/personal-income-tax
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Illinois', 'state_occupation': 'Massachusetts', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def illinois_and_michigan(self, state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Illinois:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Michigan:
        


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Illinois:
        - https://tax.illinois.gov/research/taxrates/income.html#IndividualIncome
          https://tax.illinois.gov/content/dam/soi/en/web/tax/forms/incometax/documents/currentyear/individual/il-1040-instr.pdf
          https://tax.illinois.gov/questionsandanswers/12.html
          https://tax.illinois.gov/content/dam/soi/en/web/tax/forms/incometax/documents/currentyear/individual/il-1040-instr.pdf
          https://tax.illinois.gov/research/taxrates/personalproperty.html
          https://tax.illinois.gov/localgovernments/income.html
        
        Michigan:
        -         
        - https://www.michigan.gov/taxes/iit/new-developments-for-tax-year-2022
          "https://www.michigan.gov/taxes/-/media/Project/Websites/taxes/Forms/2023/2023-SUW/446_Withholding-Guide_2023.pdf?rev=177ca3d8cb034392bf84e317ed78c320&hash=19C5F42F43A89A91863D229349F48883#:~:text=The%20exemption%20amount%20is%20%245%2C400
         employees%20Michigan%20income%20tax%20return"
          "https://www.michigan.gov/taxes/-/media/Project/Websites/taxes/Forms/2023/2023-SUW/446_Withholding-Guide_2023.pdf?rev=177ca3d8cb034392bf84e317ed78c320&hash=19C5F42F43A89A91863D229349F48883#:~:text=The%20exemption%20amount%20is%20%245%2C400
         employees%20Michigan%20income%20tax%20return"
          https://www.michigan.gov/taxes/questions/iit/accordion/residency/are-my-wages-earned-in-another-state-taxable-in-michigan-if-i-am-a-michigan-resident-1
          https://www.michigan.gov/taxes/-/media/Project/Websites/taxes/Forms/2022/2022-IIT-Forms/BOOK_MI-1040.pdf?rev=4524f1cbbc0243b194ab05a6680be75b&hash=40DFE71E651F7C1D65C31F6CD03D944F
          https://www.michigan.gov/taxes/questions/iit/accordion/general/what-cities-impose-an-income-tax
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Illinois', 'state_occupation': 'Michigan', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def illinois_and_minnesota(self, state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Illinois:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Minnesota:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Illinois:
        - https://tax.illinois.gov/research/taxrates/income.html#IndividualIncome
          https://tax.illinois.gov/content/dam/soi/en/web/tax/forms/incometax/documents/currentyear/individual/il-1040-instr.pdf
          https://tax.illinois.gov/questionsandanswers/12.html
          https://tax.illinois.gov/content/dam/soi/en/web/tax/forms/incometax/documents/currentyear/individual/il-1040-instr.pdf
          https://tax.illinois.gov/research/taxrates/personalproperty.html
          https://tax.illinois.gov/localgovernments/income.html
        
        Minnesota:
        -         
        - https://www.revenue.state.mn.us/minnesota-income-tax-rates-and-brackets
          https://www.revenue.state.mn.us/child-and-dependent-care-credit
          https://www.revenue.state.mn.us/dependent-exemptions
          https://www.revenue.state.mn.us/minnesota-standard-deduction
          https://mn.gov/mmb/assets/mwr_form_tcm1059-128581.pdf
          https://www.revenue.state.mn.us/individual-income-tax
          https://www.revenue.state.mn.us/individual-income-tax
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Illinois', 'state_occupation': 'Minnesota', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def illinois_and_mississippi(self, state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Illinois:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Mississippi:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Illinois:
        - https://tax.illinois.gov/research/taxrates/income.html#IndividualIncome
          https://tax.illinois.gov/content/dam/soi/en/web/tax/forms/incometax/documents/currentyear/individual/il-1040-instr.pdf
          https://tax.illinois.gov/questionsandanswers/12.html
          https://tax.illinois.gov/content/dam/soi/en/web/tax/forms/incometax/documents/currentyear/individual/il-1040-instr.pdf
          https://tax.illinois.gov/research/taxrates/personalproperty.html
          https://tax.illinois.gov/localgovernments/income.html
        
        Mississippi:
        -         
        - https://www.dor.ms.gov/individual/tax-rates
          https://www.dor.ms.gov/individual/tax-rates
          https://www.dor.ms.gov/individual/tax-rates
          https://www.dor.ms.gov/individual
          https://www.dor.ms.gov/individual
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Illinois', 'state_occupation': 'Mississippi', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def illinois_and_missouri(self, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Illinois:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Missouri:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Illinois:
        - https://tax.illinois.gov/research/taxrates/income.html#IndividualIncome
          https://tax.illinois.gov/content/dam/soi/en/web/tax/forms/incometax/documents/currentyear/individual/il-1040-instr.pdf
          https://tax.illinois.gov/questionsandanswers/12.html
          https://tax.illinois.gov/content/dam/soi/en/web/tax/forms/incometax/documents/currentyear/individual/il-1040-instr.pdf
          https://tax.illinois.gov/research/taxrates/personalproperty.html
          https://tax.illinois.gov/localgovernments/income.html
        
        Missouri:
        -         
        - https://dor.mo.gov/taxation/individual/tax-types/income/year-changes/
          https://dor.mo.gov/taxation/individual/tax-types/income/year-changes/
          https://dor.mo.gov/taxation/individual/tax-types/income/year-changes/
          https://www.kcmo.gov/city-hall/departments/finance/earnings-tax
          https://www.stlouis-mo.gov/government/departments/collector/earnings-tax/individual-earnings-tax-info.cfm
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Illinois', 'state_occupation': 'Missouri', "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def illinois_and_montana(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Illinois:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Montana:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Illinois:
        - https://tax.illinois.gov/research/taxrates/income.html#IndividualIncome
          https://tax.illinois.gov/content/dam/soi/en/web/tax/forms/incometax/documents/currentyear/individual/il-1040-instr.pdf
          https://tax.illinois.gov/questionsandanswers/12.html
          https://tax.illinois.gov/content/dam/soi/en/web/tax/forms/incometax/documents/currentyear/individual/il-1040-instr.pdf
          https://tax.illinois.gov/research/taxrates/personalproperty.html
          https://tax.illinois.gov/localgovernments/income.html
        
        Montana:
        -         
        - https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/2022/12/montana-employees-withholding-allowance-and-exemption-certificate-form-mw-4.pdf
          https://mtrevenue.gov/2021/11/29/simplification-of-montana-income-taxation/?utm_source=rss&utm_medium=rss&utm_campaign=simplification-of-montana-income-taxation
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Illinois', 'state_occupation': 'Montana', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def illinois_and_nebraska(self, state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Illinois:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Nebraska:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Illinois:
        - https://tax.illinois.gov/research/taxrates/income.html#IndividualIncome
          https://tax.illinois.gov/content/dam/soi/en/web/tax/forms/incometax/documents/currentyear/individual/il-1040-instr.pdf
          https://tax.illinois.gov/questionsandanswers/12.html
          https://tax.illinois.gov/content/dam/soi/en/web/tax/forms/incometax/documents/currentyear/individual/il-1040-instr.pdf
          https://tax.illinois.gov/research/taxrates/personalproperty.html
          https://tax.illinois.gov/localgovernments/income.html
        
        Nebraska:
        -         
        - https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2023_Circular_EN_8-429r2023_Whole_Book_11.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Illinois', 'state_occupation': 'Nebraska', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def illinois_and_nevada(self, state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Illinois:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Nevada:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Illinois:
        - https://tax.illinois.gov/research/taxrates/income.html#IndividualIncome
          https://tax.illinois.gov/content/dam/soi/en/web/tax/forms/incometax/documents/currentyear/individual/il-1040-instr.pdf
          https://tax.illinois.gov/questionsandanswers/12.html
          https://tax.illinois.gov/content/dam/soi/en/web/tax/forms/incometax/documents/currentyear/individual/il-1040-instr.pdf
          https://tax.illinois.gov/research/taxrates/personalproperty.html
          https://tax.illinois.gov/localgovernments/income.html
        
        Nevada:
        -         
        - https://www.nvsos.gov/sos/businesses/the-nevada-advantage
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Illinois', 'state_occupation': 'Nevada', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def illinois_and_new_hampshire(self, state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Illinois:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        New Hampshire:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Illinois:
        - https://tax.illinois.gov/research/taxrates/income.html#IndividualIncome
          https://tax.illinois.gov/content/dam/soi/en/web/tax/forms/incometax/documents/currentyear/individual/il-1040-instr.pdf
          https://tax.illinois.gov/questionsandanswers/12.html
          https://tax.illinois.gov/content/dam/soi/en/web/tax/forms/incometax/documents/currentyear/individual/il-1040-instr.pdf
          https://tax.illinois.gov/research/taxrates/personalproperty.html
          https://tax.illinois.gov/localgovernments/income.html
        
        New Hampshire:
        -         
        - https://www.revenue.nh.gov/assistance/tax-overview.htm
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Illinois', 'state_occupation': 'NewHampshire', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def illinois_and_new_jersey(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Illinois:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        New Jersey:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Illinois:
        - https://tax.illinois.gov/research/taxrates/income.html#IndividualIncome
          https://tax.illinois.gov/content/dam/soi/en/web/tax/forms/incometax/documents/currentyear/individual/il-1040-instr.pdf
          https://tax.illinois.gov/questionsandanswers/12.html
          https://tax.illinois.gov/content/dam/soi/en/web/tax/forms/incometax/documents/currentyear/individual/il-1040-instr.pdf
          https://tax.illinois.gov/research/taxrates/personalproperty.html
          https://tax.illinois.gov/localgovernments/income.html
        
        New Jersey:
        -         
        - https://www.state.nj.us/treasury/taxation/taxtables.shtml
          https://www.state.nj.us/treasury/taxation/njit13.shtml
          https://www.nj.gov/treasury/taxation/njit25.shtml
          https://www.state.nj.us/treasury/taxation/individuals/freqqiti.shtml
          https://www.state.nj.us/treasury/taxation/localtax.shtml
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Illinois', 'state_occupation': 'NewJersey', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def illinois_and_new_mexico(self, state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Illinois:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        New Mexico:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Illinois:
        - https://tax.illinois.gov/research/taxrates/income.html#IndividualIncome
          https://tax.illinois.gov/content/dam/soi/en/web/tax/forms/incometax/documents/currentyear/individual/il-1040-instr.pdf
          https://tax.illinois.gov/questionsandanswers/12.html
          https://tax.illinois.gov/content/dam/soi/en/web/tax/forms/incometax/documents/currentyear/individual/il-1040-instr.pdf
          https://tax.illinois.gov/research/taxrates/personalproperty.html
          https://tax.illinois.gov/localgovernments/income.html
        
        New Mexico:
        -         
        - https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/fdf3c548-8aba-4b9c-9eb4-bb564c716015/FYI-104.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/0bf943fd-652e-4400-bb1b-cd8397eb5a95/2022pit-adj.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/2f1a6781-9534-4436-b427-1557f9592099/2022pit-adj-ins.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/0bf943fd-652e-4400-bb1b-cd8397eb5a95/2022pit-adj.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/1afc56af-ea90-4d48-82e5-1f9aeb43255a/PITbook2022.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/fdf3c548-8aba-4b9c-9eb4-bb564c716015/FYI-104.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Illinois', 'state_occupation': 'NewMexico', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def illinois_and_new_york(self, county_residence='Other', state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Illinois:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        New York:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Illinois:
        - https://tax.illinois.gov/research/taxrates/income.html#IndividualIncome
          https://tax.illinois.gov/content/dam/soi/en/web/tax/forms/incometax/documents/currentyear/individual/il-1040-instr.pdf
          https://tax.illinois.gov/questionsandanswers/12.html
          https://tax.illinois.gov/content/dam/soi/en/web/tax/forms/incometax/documents/currentyear/individual/il-1040-instr.pdf
          https://tax.illinois.gov/research/taxrates/personalproperty.html
          https://tax.illinois.gov/localgovernments/income.html
        
        New York:
        -         
        - https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nys-tax-rate-schedule
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nys-tax-rate-schedule
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nys-tax-rate-schedule
          https://www.tax.ny.gov/pit/
          https://www.tax.ny.gov/pit/file/nonresident-faqs.htm
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#yonkers-tax-rate-schedule
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nyc-tax-rate-schedule
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Illinois', 'state_occupation': 'NewYork', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def illinois_and_north_carolina(self, state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Illinois:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        North Carolina:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Illinois:
        - https://tax.illinois.gov/research/taxrates/income.html#IndividualIncome
          https://tax.illinois.gov/content/dam/soi/en/web/tax/forms/incometax/documents/currentyear/individual/il-1040-instr.pdf
          https://tax.illinois.gov/questionsandanswers/12.html
          https://tax.illinois.gov/content/dam/soi/en/web/tax/forms/incometax/documents/currentyear/individual/il-1040-instr.pdf
          https://tax.illinois.gov/research/taxrates/personalproperty.html
          https://tax.illinois.gov/localgovernments/income.html
        
        North Carolina:
        -         
        - https://www.ncdor.gov/taxes-forms/tax-rate-schedules
          https://www.ncdor.gov/taxes-forms/individual-income-tax/north-carolina-standard-deduction-or-north-carolina-itemized-deductions
          https://www.ncdor.gov/taxes-forms/individual-income-tax
          https://www.ncdor.gov/taxes-forms/individual-income-tax
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Illinois', 'state_occupation': 'NorthCarolina', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def illinois_and_north_dakota(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Illinois:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        North Dakota:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Illinois:
        - https://tax.illinois.gov/research/taxrates/income.html#IndividualIncome
          https://tax.illinois.gov/content/dam/soi/en/web/tax/forms/incometax/documents/currentyear/individual/il-1040-instr.pdf
          https://tax.illinois.gov/questionsandanswers/12.html
          https://tax.illinois.gov/content/dam/soi/en/web/tax/forms/incometax/documents/currentyear/individual/il-1040-instr.pdf
          https://tax.illinois.gov/research/taxrates/personalproperty.html
          https://tax.illinois.gov/localgovernments/income.html
        
        North Dakota:
        -         
        - https://www.tax.nd.gov/forms
          https://www.tax.nd.gov/sites/www/files/documents/forms/individual/2022-iit/2022-individual-income-tax-booklet.pdf
          https://www.tax.nd.gov/sites/www/files/documents/forms/2022-individual-income-tax-booklet.pdf
          https://www.tax.nd.gov/tax-exemptions-credits/income-tax-exemptions-credits
          https://www.tax.nd.gov/sites/www/files/documents/forms/business/ndwrfillable.pdf
          https://www.tax.nd.gov/news/tax-legislative-changes/significant-changes-law/individual-income-tax-history
          https://www.tax.nd.gov/forms
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Illinois', 'state_occupation': 'NorthDakota', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def illinois_and_ohio(self, county_residence='Other', state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Illinois:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Ohio:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Illinois:
        - https://tax.illinois.gov/research/taxrates/income.html#IndividualIncome
          https://tax.illinois.gov/content/dam/soi/en/web/tax/forms/incometax/documents/currentyear/individual/il-1040-instr.pdf
          https://tax.illinois.gov/questionsandanswers/12.html
          https://tax.illinois.gov/content/dam/soi/en/web/tax/forms/incometax/documents/currentyear/individual/il-1040-instr.pdf
          https://tax.illinois.gov/research/taxrates/personalproperty.html
          https://tax.illinois.gov/localgovernments/income.html
        
        Ohio:
        -         
        - https://tax.ohio.gov/individual/resources/annual-tax-rates
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://tax.ohio.gov/static/forms/employer_withholding/generic/wth_it4nr.pdf
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://www.ritaohio.com/TaxRatesTable
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Illinois', 'state_occupation': 'Ohio', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def illinois_and_oklahoma(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Illinois:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Oklahoma:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Illinois:
        - https://tax.illinois.gov/research/taxrates/income.html#IndividualIncome
          https://tax.illinois.gov/content/dam/soi/en/web/tax/forms/incometax/documents/currentyear/individual/il-1040-instr.pdf
          https://tax.illinois.gov/questionsandanswers/12.html
          https://tax.illinois.gov/content/dam/soi/en/web/tax/forms/incometax/documents/currentyear/individual/il-1040-instr.pdf
          https://tax.illinois.gov/research/taxrates/personalproperty.html
          https://tax.illinois.gov/localgovernments/income.html
        
        Oklahoma:
        -         
        - https://oklahoma.gov/content/dam/ok/en/tax/documents/resources/publications/businesses/withholding-tables/WHTables-2022.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Illinois', 'state_occupation': 'Oklahoma', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def illinois_and_oregon(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Illinois:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Oregon:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Illinois:
        - https://tax.illinois.gov/research/taxrates/income.html#IndividualIncome
          https://tax.illinois.gov/content/dam/soi/en/web/tax/forms/incometax/documents/currentyear/individual/il-1040-instr.pdf
          https://tax.illinois.gov/questionsandanswers/12.html
          https://tax.illinois.gov/content/dam/soi/en/web/tax/forms/incometax/documents/currentyear/individual/il-1040-instr.pdf
          https://tax.illinois.gov/research/taxrates/personalproperty.html
          https://tax.illinois.gov/localgovernments/income.html
        
        Oregon:
        -         
        - https://www.oregon.gov/dor/programs/individuals/Pages/PIT.aspx
          https://www.oregon.gov/dor/programs/individuals/Documents/Full%20year%20resident
         %20Form%20or-40%20filers.pdf
          https://www.oregon.gov/dor/programs/individuals/Documents/Part-year%20and%20nonresident
         %20Form%20or-40-P%20and%20or-40-N%20filers.pdf.pdf
          https://www.oregon.gov/dor/forms/FormsPubs/form-or-40_101-040_2022.pdf
          https://www.oregon.gov/dor/forms/FormsPubs/form-or-40_101-040_2022.pdf
          https://www.oregon.gov/dor/forms/FormsPubs/form-or-40_101-040_2022.pdf
          https://www.portland.gov/revenue/personal-tax
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Illinois', 'state_occupation': 'Oregon', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def illinois_and_pennsylvania(self, county_residence='Other', state_agi=75000, county_occupation='Other', capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Illinois:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Pennsylvania:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        county_residence : str, optional
            County of residence. Default is Other.
        county_occupation : str, optional
            County of occupation. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Illinois:
        - https://tax.illinois.gov/research/taxrates/income.html#IndividualIncome
          https://tax.illinois.gov/content/dam/soi/en/web/tax/forms/incometax/documents/currentyear/individual/il-1040-instr.pdf
          https://tax.illinois.gov/questionsandanswers/12.html
          https://tax.illinois.gov/content/dam/soi/en/web/tax/forms/incometax/documents/currentyear/individual/il-1040-instr.pdf
          https://tax.illinois.gov/research/taxrates/personalproperty.html
          https://tax.illinois.gov/localgovernments/income.html
        
        Pennsylvania:
        -         
        - https://www.revenue.pa.gov/Tax%20Rates/Pages/default.aspx
          https://www.revenue.pa.gov/FormsandPublications/FormsforIndividuals/PIT/Documents/2022/2022_pa-40in.pdf
          https://www.revenue.pa.gov/FormsandPublications/FormsforBusinesses/EmployerWithholding/Documents/rev-419.pdf
          https://www.revenue.pa.gov/FormsandPublications/FormsforIndividuals/PIT/Documents/2022/2022_pa-40in.pdf
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-25.htm
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        county_occupation = self.county_occupation if hasattr(self, 'county_occupation') else county_occupation
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Illinois', 'state_occupation': 'Pennsylvania', "county_residence": county_residence, "state_agi": state_agi, "county_occupation": county_occupation, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def illinois_and_rhode_island(self, state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Illinois:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Rhode Island:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Illinois:
        - https://tax.illinois.gov/research/taxrates/income.html#IndividualIncome
          https://tax.illinois.gov/content/dam/soi/en/web/tax/forms/incometax/documents/currentyear/individual/il-1040-instr.pdf
          https://tax.illinois.gov/questionsandanswers/12.html
          https://tax.illinois.gov/content/dam/soi/en/web/tax/forms/incometax/documents/currentyear/individual/il-1040-instr.pdf
          https://tax.illinois.gov/research/taxrates/personalproperty.html
          https://tax.illinois.gov/localgovernments/income.html
        
        Rhode Island:
        -         
        - https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2022_1040WE_w_0.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Illinois', 'state_occupation': 'RhodeIsland', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def illinois_and_south_carolina(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Illinois:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        South Carolina:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Illinois:
        - https://tax.illinois.gov/research/taxrates/income.html#IndividualIncome
          https://tax.illinois.gov/content/dam/soi/en/web/tax/forms/incometax/documents/currentyear/individual/il-1040-instr.pdf
          https://tax.illinois.gov/questionsandanswers/12.html
          https://tax.illinois.gov/content/dam/soi/en/web/tax/forms/incometax/documents/currentyear/individual/il-1040-instr.pdf
          https://tax.illinois.gov/research/taxrates/personalproperty.html
          https://tax.illinois.gov/localgovernments/income.html
        
        South Carolina:
        -         
        - https://dor.sc.gov/resources-site/lawandpolicy/Advisory%20Opinions/IL22-15.pdf
          https://dor.sc.gov/forms-site/Forms/IITPacket_2022.pdf
          https://www.scstatehouse.gov/code/t12c006.php
          https://dor.sc.gov/resources-site/lawandpolicy/Advisory%20Opinions/IL22-15.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Illinois', 'state_occupation': 'SouthCarolina', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def illinois_and_south_dakota(self, state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Illinois:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        South Dakota:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Illinois:
        - https://tax.illinois.gov/research/taxrates/income.html#IndividualIncome
          https://tax.illinois.gov/content/dam/soi/en/web/tax/forms/incometax/documents/currentyear/individual/il-1040-instr.pdf
          https://tax.illinois.gov/questionsandanswers/12.html
          https://tax.illinois.gov/content/dam/soi/en/web/tax/forms/incometax/documents/currentyear/individual/il-1040-instr.pdf
          https://tax.illinois.gov/research/taxrates/personalproperty.html
          https://tax.illinois.gov/localgovernments/income.html
        
        South Dakota:
        -         
        - https://dor.sd.gov/individuals/taxes/
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Illinois', 'state_occupation': 'SouthDakota', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def illinois_and_tennessee(self, state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Illinois:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Tennessee:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Illinois:
        - https://tax.illinois.gov/research/taxrates/income.html#IndividualIncome
          https://tax.illinois.gov/content/dam/soi/en/web/tax/forms/incometax/documents/currentyear/individual/il-1040-instr.pdf
          https://tax.illinois.gov/questionsandanswers/12.html
          https://tax.illinois.gov/content/dam/soi/en/web/tax/forms/incometax/documents/currentyear/individual/il-1040-instr.pdf
          https://tax.illinois.gov/research/taxrates/personalproperty.html
          https://tax.illinois.gov/localgovernments/income.html
        
        Tennessee:
        -         
        - https://www.tn.gov/revenue/taxes.html
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Illinois', 'state_occupation': 'Tennessee', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def illinois_and_texas(self, state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Illinois:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Texas:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Illinois:
        - https://tax.illinois.gov/research/taxrates/income.html#IndividualIncome
          https://tax.illinois.gov/content/dam/soi/en/web/tax/forms/incometax/documents/currentyear/individual/il-1040-instr.pdf
          https://tax.illinois.gov/questionsandanswers/12.html
          https://tax.illinois.gov/content/dam/soi/en/web/tax/forms/incometax/documents/currentyear/individual/il-1040-instr.pdf
          https://tax.illinois.gov/research/taxrates/personalproperty.html
          https://tax.illinois.gov/localgovernments/income.html
        
        Texas:
        -         
        - https://comptroller.texas.gov/taxes/
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Illinois', 'state_occupation': 'Texas', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def illinois_and_utah(self, state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Illinois:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Utah:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Illinois:
        - https://tax.illinois.gov/research/taxrates/income.html#IndividualIncome
          https://tax.illinois.gov/content/dam/soi/en/web/tax/forms/incometax/documents/currentyear/individual/il-1040-instr.pdf
          https://tax.illinois.gov/questionsandanswers/12.html
          https://tax.illinois.gov/content/dam/soi/en/web/tax/forms/incometax/documents/currentyear/individual/il-1040-instr.pdf
          https://tax.illinois.gov/research/taxrates/personalproperty.html
          https://tax.illinois.gov/localgovernments/income.html
        
        Utah:
        -         
        - https://incometax.utah.gov/paying/tax-rates
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Illinois', 'state_occupation': 'Utah', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def illinois_and_vermont(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Illinois:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Vermont:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Illinois:
        - https://tax.illinois.gov/research/taxrates/income.html#IndividualIncome
          https://tax.illinois.gov/content/dam/soi/en/web/tax/forms/incometax/documents/currentyear/individual/il-1040-instr.pdf
          https://tax.illinois.gov/questionsandanswers/12.html
          https://tax.illinois.gov/content/dam/soi/en/web/tax/forms/incometax/documents/currentyear/individual/il-1040-instr.pdf
          https://tax.illinois.gov/research/taxrates/personalproperty.html
          https://tax.illinois.gov/localgovernments/income.html
        
        Vermont:
        -         
        - https://tax.vermont.gov/sites/tax/files/documents/GB-1210-2023.pdf
          https://tax.vermont.gov/sites/tax/files/documents/Income%20Booklet-2022.pdf
          https://tax.vermont.gov/sites/tax/files/documents/Income%20Booklet-2022.pdf
          https://tax.vermont.gov/sites/tax/files/documents/Income%20Booklet-2022.pdf
          https://tax.vermont.gov/individuals/personal-income-tax/tax-credits
          https://tax.vermont.gov/individuals/personal-income-tax/taxable-income
          https://tax.vermont.gov/individuals/personal-income-tax/local-option-tax
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Illinois', 'state_occupation': 'Vermont', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def illinois_and_virginia(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Illinois:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Virginia:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Illinois:
        - https://tax.illinois.gov/research/taxrates/income.html#IndividualIncome
          https://tax.illinois.gov/content/dam/soi/en/web/tax/forms/incometax/documents/currentyear/individual/il-1040-instr.pdf
          https://tax.illinois.gov/questionsandanswers/12.html
          https://tax.illinois.gov/content/dam/soi/en/web/tax/forms/incometax/documents/currentyear/individual/il-1040-instr.pdf
          https://tax.illinois.gov/research/taxrates/personalproperty.html
          https://tax.illinois.gov/localgovernments/income.html
        
        Virginia:
        -         
        - https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/reciprocity
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Illinois', 'state_occupation': 'Virginia', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def illinois_and_washington(self, state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Illinois:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Washington:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Illinois:
        - https://tax.illinois.gov/research/taxrates/income.html#IndividualIncome
          https://tax.illinois.gov/content/dam/soi/en/web/tax/forms/incometax/documents/currentyear/individual/il-1040-instr.pdf
          https://tax.illinois.gov/questionsandanswers/12.html
          https://tax.illinois.gov/content/dam/soi/en/web/tax/forms/incometax/documents/currentyear/individual/il-1040-instr.pdf
          https://tax.illinois.gov/research/taxrates/personalproperty.html
          https://tax.illinois.gov/localgovernments/income.html
        
        Washington:
        -         
        - https://dor.wa.gov/taxes-rates/income-tax
          https://dor.wa.gov/taxes-rates/other-taxes/capital-gains-tax
          https://dor.wa.gov/taxes-rates/income-tax
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Illinois', 'state_occupation': 'Washington', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def illinois_and_west_virginia(self, county_occupation='Other', state_agi=75000, pay_periods=24, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Illinois:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        West Virginia:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_occupation : str, optional
            County of occupation. Default is Other.
        pay_periods : int, optional
            Number of pay periods. Options are 13, 24, 26, 52. Default is 24.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Illinois:
        - https://tax.illinois.gov/research/taxrates/income.html#IndividualIncome
          https://tax.illinois.gov/content/dam/soi/en/web/tax/forms/incometax/documents/currentyear/individual/il-1040-instr.pdf
          https://tax.illinois.gov/questionsandanswers/12.html
          https://tax.illinois.gov/content/dam/soi/en/web/tax/forms/incometax/documents/currentyear/individual/il-1040-instr.pdf
          https://tax.illinois.gov/research/taxrates/personalproperty.html
          https://tax.illinois.gov/localgovernments/income.html
        
        West Virginia:
        -         
        - https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://tax.wv.gov/Documents/TSD/tsd438.pdf
          https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2011/Taxes-11-35att.html
        """
        
        # Check if the parameters are set upstream in the instance
        county_occupation = self.county_occupation if hasattr(self, 'county_occupation') else county_occupation
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        pay_periods = self.pay_periods if hasattr(self, 'pay_periods') else pay_periods
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Illinois', 'state_occupation': 'WestVirginia', "county_occupation": county_occupation, "state_agi": state_agi, "pay_periods": pay_periods, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def illinois_and_wisconsin(self, state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Illinois:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Wisconsin:
        


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Illinois:
        - https://tax.illinois.gov/research/taxrates/income.html#IndividualIncome
          https://tax.illinois.gov/content/dam/soi/en/web/tax/forms/incometax/documents/currentyear/individual/il-1040-instr.pdf
          https://tax.illinois.gov/questionsandanswers/12.html
          https://tax.illinois.gov/content/dam/soi/en/web/tax/forms/incometax/documents/currentyear/individual/il-1040-instr.pdf
          https://tax.illinois.gov/research/taxrates/personalproperty.html
          https://tax.illinois.gov/localgovernments/income.html
        
        Wisconsin:
        -         
        - https://www.revenue.wi.gov/TaxForms2023/2023-Form1-ES-Inst.pdf
          https://www.revenue.wi.gov/TaxForms2023/2023-Form1-ES-Inst.pdf
          https://www.revenue.wi.gov/TaxForms2023/2023-Form1-ES-Inst.pdf
          https://www.revenue.wi.gov/Pages/FAQS/pcs-work.aspx
          https://www.revenue.wi.gov/Dor%20Publications/pb103.pdf
          https://www.revenue.wi.gov/Pages/FAQS/pcs-work.aspx
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Illinois', 'state_occupation': 'Wisconsin', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def illinois_and_wyoming(self, state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Illinois:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Wyoming:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Illinois:
        - https://tax.illinois.gov/research/taxrates/income.html#IndividualIncome
          https://tax.illinois.gov/content/dam/soi/en/web/tax/forms/incometax/documents/currentyear/individual/il-1040-instr.pdf
          https://tax.illinois.gov/questionsandanswers/12.html
          https://tax.illinois.gov/content/dam/soi/en/web/tax/forms/incometax/documents/currentyear/individual/il-1040-instr.pdf
          https://tax.illinois.gov/research/taxrates/personalproperty.html
          https://tax.illinois.gov/localgovernments/income.html
        
        Wyoming:
        -         
        - https://revenue.wyo.gov/
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Illinois', 'state_occupation': 'Wyoming', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def indiana_and_alabama(self, capital_gains_short=0, dependents=0, sources=False, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_long=0, filing_status='single', income=100000, credits_state=0):
        """
        --- Parameters ----
        
        Indiana:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Alabama:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Indiana:
        - https://www.in.gov/dor/tax-forms/2022-individual-income-tax-forms/
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-32.htm
          https://www.in.gov/dor/files/ib28.pdf
          https://www.in.gov/dor/tax-forms/2022-individual-income-tax-forms/
          https://www.in.gov/dor/files/2022-county-tax-rates-and-codes.pdf
        
        Alabama:
        -         
        - https://www.revenue.alabama.gov/ultraviewer/viewer/basic_viewer/index.html?form=2023/02/22f40.pdf
          https://www.revenue.alabama.gov/faqs/what-is-alabamas-individual-income-tax-rate/
          https://www.revenue.alabama.gov/ultraviewer/viewer/basic_viewer/index.html?form=2023/02/22f40.pdf
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-24.htm?taxmap=true
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-24.htm?taxmap=true
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-24.htm?taxmap=true
          https://www.revenue.alabama.gov/individual-corporate/income-to-be-reported-on-the-alabama-income-tax-return/
          https://www.almonline.org/TaxRates.aspx#OccupationalTax
        """
        
        # Check if the parameters are set upstream in the instance
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Indiana', 'state_occupation': 'Alabama', "capital_gains_short": capital_gains_short, "dependents": dependents, "sources": sources, "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def indiana_and_alaska(self, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Indiana:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Alaska:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Indiana:
        - https://www.in.gov/dor/tax-forms/2022-individual-income-tax-forms/
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-32.htm
          https://www.in.gov/dor/files/ib28.pdf
          https://www.in.gov/dor/tax-forms/2022-individual-income-tax-forms/
          https://www.in.gov/dor/files/2022-county-tax-rates-and-codes.pdf
        
        Alaska:
        -         
        - https://www.commerce.alaska.gov/web/dcra/officeofthestateassessor/alaskataxfacts.aspx
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Indiana', 'state_occupation': 'Alaska', "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def indiana_and_arizona(self, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Indiana:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Arizona:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        credits_state : int, optional
            Credits for state taxes. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Indiana:
        - https://www.in.gov/dor/tax-forms/2022-individual-income-tax-forms/
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-32.htm
          https://www.in.gov/dor/files/ib28.pdf
          https://www.in.gov/dor/tax-forms/2022-individual-income-tax-forms/
          https://www.in.gov/dor/files/2022-county-tax-rates-and-codes.pdf
        
        Arizona:
        -         
        - https://azdor.gov/forms/individual/form-140-x-y-tables
          https://azdor.gov/forms/individual/form-140-x-y-tables
          https://azdor.gov/forms/individual/form-140-x-y-tables
          https://azdor.gov/businesses-arizona/withholding-tax/withholding-exceptions
          "https://azdor.gov/individuals/income-tax-filing-assistance/identifying-other-taxable-income#:~:text=In%20regards%20to%20capital%20gains
         the%20individuals%20regular%20tax%20rate.
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Indiana', 'state_occupation': 'Arizona', "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def indiana_and_arkansas(self, capital_gains_short=0, dependents=0, sources=False, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_long=0, filing_status='single', income=100000, credits_state=0):
        """
        --- Parameters ----
        
        Indiana:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Arkansas:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Indiana:
        - https://www.in.gov/dor/tax-forms/2022-individual-income-tax-forms/
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-32.htm
          https://www.in.gov/dor/files/ib28.pdf
          https://www.in.gov/dor/tax-forms/2022-individual-income-tax-forms/
          https://www.in.gov/dor/files/2022-county-tax-rates-and-codes.pdf
        
        Arkansas:
        -         
        - https://www.arkansasedc.com/why-arkansas/business-climate/tax-structure/personal-income-tax
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_and_AR1000NR_Instructions.pdf
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_FullYearResidentIndividualIncomeTaxReturn.pdf
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_and_AR1000NR_Instructions.pdf
          https://www.arkansasedc.com/why-arkansas/business-climate/tax-structure/capital-gains-tax-reduction
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_and_AR1000NR_Instructions.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Indiana', 'state_occupation': 'Arkansas', "capital_gains_short": capital_gains_short, "dependents": dependents, "sources": sources, "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def indiana_and_california(self, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Indiana:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        California:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Indiana:
        - https://www.in.gov/dor/tax-forms/2022-individual-income-tax-forms/
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-32.htm
          https://www.in.gov/dor/files/ib28.pdf
          https://www.in.gov/dor/tax-forms/2022-individual-income-tax-forms/
          https://www.in.gov/dor/files/2022-county-tax-rates-and-codes.pdf
        
        California:
        -         
        - https://www.ftb.ca.gov/forms/2022/2022-540-tax-rate-schedules.pdf
          https://www.ftb.ca.gov/file/personal/deductions/index.html
          https://www.ftb.ca.gov/about-ftb/newsroom/caleitc/eligibility-and-credit-information.html
          https://www.ftb.ca.gov/file/personal/income-types/capital-gains-and-losses.html#:~:text=California%20does%20not%20have%20a
         are%20taxed%20as%20ordinary%20income
          https://www.ftb.ca.gov/file/personal/
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Indiana', 'state_occupation': 'California', "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def indiana_and_colorado(self, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Indiana:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Colorado:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Indiana:
        - https://www.in.gov/dor/tax-forms/2022-individual-income-tax-forms/
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-32.htm
          https://www.in.gov/dor/files/ib28.pdf
          https://www.in.gov/dor/tax-forms/2022-individual-income-tax-forms/
          https://www.in.gov/dor/files/2022-county-tax-rates-and-codes.pdf
        
        Colorado:
        -         
        - https://leg.colorado.gov/agencies/legislative-council-staff/individual-income-tax%C2%A0
          https://leg.colorado.gov/agencies/legislative-council-staff/individual-income-tax%C2%A0
          https://tax.colorado.gov/sites/tax/files/documents/ITT_Capital_Gain_Subtraction_Feb_2022.pdf
          https://www.auroragov.org/business_services/taxes/occupational_privilege_tax
          https://www.denvergov.org/Government/Agencies-Departments-Offices/Agencies-Departments-Offices-Directory/Department-of-Finance/Our-Divisions/Treasury/Business-Tax-Information#:~:text=Occupational%20privilege%20tax
         -%28show%20below%29&text=Each%20taxable%20employee%20is%20liable
         month%20for%20each%20taxable%20employee
          https://www.glendale.co.us/355/Occupational-Privilege-Tax
          https://www.greenwoodvillage.com/1220/Occupational-Privilege-Tax-OPT
          https://ci.sheridan.co.us/288/Occupational-Privilege-Tax
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Indiana', 'state_occupation': 'Colorado', "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def indiana_and_connecticut(self, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Indiana:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Connecticut:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Indiana:
        - https://www.in.gov/dor/tax-forms/2022-individual-income-tax-forms/
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-32.htm
          https://www.in.gov/dor/files/ib28.pdf
          https://www.in.gov/dor/tax-forms/2022-individual-income-tax-forms/
          https://www.in.gov/dor/files/2022-county-tax-rates-and-codes.pdf
        
        Connecticut:
        -         
        - https://portal.ct.gov/-/media/DRS/Forms/2022/Income/2022-CT-1040-Instructions_1222.pdf
          https://www.cga.ct.gov/2022/rpt/pdf/2022-R-0108.pdf
          https://www.cga.ct.gov/current/pub/chap_229.htm#sec_12-703
          https://portal.ct.gov/-/media/DRS/Forms/2022/Income/2022-CT-1040-Instructions_1222.pdf
          "https://portal.ct.gov/DRS/Publications/TSSNs/TSSN-29#:~:text=An%20individuals%20net%20capital%20gains
         portion%20of%20Social%20Security%20benefits"
          "https://portal.ct.gov/DRS/Publications/TSSNs/TSSN-29#:~:text=An%20individuals%20net%20capital%20gains
         portion%20of%20Social%20Security%20benefits"
          https://portal.ct.gov/DRS/Individuals/Resident-Income-Tax/Tax-Information
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Indiana', 'state_occupation': 'Connecticut', "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def indiana_and_delaware(self, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, federal_agi=75000, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Indiana:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Delaware:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        federal_agi : int, optional
            Adjusted Gross Income (AGI) for federal taxes. Default is 0. Sources - https://www.irs.gov/e-file-providers/definition-of-adjusted-gross-income, https://www.irs.gov/newsroom/taxpayers-should-know-the-difference-between-standard-and-itemized-deductions
        income : int, optional
            Gross income. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Indiana:
        - https://www.in.gov/dor/tax-forms/2022-individual-income-tax-forms/
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-32.htm
          https://www.in.gov/dor/files/ib28.pdf
          https://www.in.gov/dor/tax-forms/2022-individual-income-tax-forms/
          https://www.in.gov/dor/files/2022-county-tax-rates-and-codes.pdf
        
        Delaware:
        -         
        - https://revenue.delaware.gov/software-developer/tax-rate-changes/
          https://revenuefiles.delaware.gov/2022/PIT-RES_TY22_2022-01_PaperInteractive.pdf
          https://revenuefiles.delaware.gov/2022/PIT-RES_TY22_2022-02_Instructions.pdf
          https://revenue.delaware.gov/frequently-asked-questions/personal-income-tax-faqs/
          https://www.newcastlede.gov/DocumentCenter/View/20147/City-of-Wilmington-Wage-Tax-Information-PDF
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        federal_agi = self.federal_agi if hasattr(self, 'federal_agi') else federal_agi
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Indiana', 'state_occupation': 'Delaware', "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "federal_agi": federal_agi, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def indiana_and_florida(self, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Indiana:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Florida:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Indiana:
        - https://www.in.gov/dor/tax-forms/2022-individual-income-tax-forms/
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-32.htm
          https://www.in.gov/dor/files/ib28.pdf
          https://www.in.gov/dor/tax-forms/2022-individual-income-tax-forms/
          https://www.in.gov/dor/files/2022-county-tax-rates-and-codes.pdf
        
        Florida:
        -         
        - https://floridarevenue.com/taxes/taxesfees/Pages/tax_interest_rates.aspx
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Indiana', 'state_occupation': 'Florida', "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def indiana_and_georgia(self, capital_gains_short=0, dependents=0, sources=False, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_long=0, filing_status='single', income=100000, credits_state=0):
        """
        --- Parameters ----
        
        Indiana:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Georgia:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Indiana:
        - https://www.in.gov/dor/tax-forms/2022-individual-income-tax-forms/
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-32.htm
          https://www.in.gov/dor/files/ib28.pdf
          https://www.in.gov/dor/tax-forms/2022-individual-income-tax-forms/
          https://www.in.gov/dor/files/2022-county-tax-rates-and-codes.pdf
        
        Georgia:
        -         
        - https://dor.georgia.gov/tax-tables-georgia-tax-rate-schedule
          https://apps.dor.ga.gov/FillableForms/PDFViewer/Index?form=2022GA500
          https://apps.dor.ga.gov/FillableForms/PDFViewer/Index?form=2022GA500
          https://apps.dor.ga.gov/FillableForms/PDFViewer/Index?form=2022GA500
          https://dor.georgia.gov/years-individual-income-tax-forms
          https://dor.georgia.gov/years-individual-income-tax-forms
        """
        
        # Check if the parameters are set upstream in the instance
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Indiana', 'state_occupation': 'Georgia', "capital_gains_short": capital_gains_short, "dependents": dependents, "sources": sources, "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def indiana_and_hawaii(self, capital_gains_short=0, dependents=0, sources=False, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_long=0, filing_status='single', income=100000, credits_state=0):
        """
        --- Parameters ----
        
        Indiana:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Hawaii:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Indiana:
        - https://www.in.gov/dor/tax-forms/2022-individual-income-tax-forms/
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-32.htm
          https://www.in.gov/dor/files/ib28.pdf
          https://www.in.gov/dor/tax-forms/2022-individual-income-tax-forms/
          https://www.in.gov/dor/files/2022-county-tax-rates-and-codes.pdf
        
        Hawaii:
        -         
        - https://tax.hawaii.gov/forms/d_18table-on/d_18table-on_p13/
          https://files.hawaii.gov/tax/forms/2018/18table-on.pdf
          https://files.hawaii.gov/tax/news/pubs/22outline.pdf
          https://files.hawaii.gov/tax/news/pubs/FastTaxRefGuide_2017.pdf
          https://www.tfhawaii.org/wordpress/state-tax-resources/mini-tax-guides/individual-income-tax-chapter-235/#:~:text=Long%20term%20capital%20gains%20are%20taxed%20at%20a%20maximum%20of%207.25%25
          https://www.grassrootinstitute.org/2023/02/hb337-increasing-tax-on-capital-gains-a-bad-risk/
          https://files.hawaii.gov/tax/news/pubs/22outline.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Indiana', 'state_occupation': 'Hawaii', "capital_gains_short": capital_gains_short, "dependents": dependents, "sources": sources, "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def indiana_and_idaho(self, capital_gains_short=0, dependents=0, sources=False, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_long=0, filing_status='single', income=100000, credits_state=0):
        """
        --- Parameters ----
        
        Indiana:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Idaho:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Indiana:
        - https://www.in.gov/dor/tax-forms/2022-individual-income-tax-forms/
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-32.htm
          https://www.in.gov/dor/files/ib28.pdf
          https://www.in.gov/dor/tax-forms/2022-individual-income-tax-forms/
          https://www.in.gov/dor/files/2022-county-tax-rates-and-codes.pdf
        
        Idaho:
        -         
        - https://tax.idaho.gov/taxes/income-tax/individual-income/individual-income-tax-rate-schedule/
          https://tax.idaho.gov/wp-content/uploads/forms/EIN00046/EIN00046_03-01-2023.pdf
          https://tax.idaho.gov/taxes/income-tax/individual-income/idaho-source-income/
          https://tax.idaho.gov/taxes/income-tax/individual-income/idaho-source-income/
        """
        
        # Check if the parameters are set upstream in the instance
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Indiana', 'state_occupation': 'Idaho', "capital_gains_short": capital_gains_short, "dependents": dependents, "sources": sources, "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def indiana_and_illinois(self, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Indiana:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Illinois:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Indiana:
        - https://www.in.gov/dor/tax-forms/2022-individual-income-tax-forms/
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-32.htm
          https://www.in.gov/dor/files/ib28.pdf
          https://www.in.gov/dor/tax-forms/2022-individual-income-tax-forms/
          https://www.in.gov/dor/files/2022-county-tax-rates-and-codes.pdf
        
        Illinois:
        -         
        - https://tax.illinois.gov/research/taxrates/income.html#IndividualIncome
          https://tax.illinois.gov/content/dam/soi/en/web/tax/forms/incometax/documents/currentyear/individual/il-1040-instr.pdf
          https://tax.illinois.gov/questionsandanswers/12.html
          https://tax.illinois.gov/content/dam/soi/en/web/tax/forms/incometax/documents/currentyear/individual/il-1040-instr.pdf
          https://tax.illinois.gov/research/taxrates/personalproperty.html
          https://tax.illinois.gov/localgovernments/income.html
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Indiana', 'state_occupation': 'Illinois', "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def indiana_and_iowa(self, capital_gains_short=0, dependents=0, sources=False, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_long=0, filing_status='single', income=100000, credits_state=0):
        """
        --- Parameters ----
        
        Indiana:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Iowa:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Indiana:
        - https://www.in.gov/dor/tax-forms/2022-individual-income-tax-forms/
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-32.htm
          https://www.in.gov/dor/files/ib28.pdf
          https://www.in.gov/dor/tax-forms/2022-individual-income-tax-forms/
          https://www.in.gov/dor/files/2022-county-tax-rates-and-codes.pdf
        
        Iowa:
        -         
        - https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
          https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
          https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
          https://tax.iowa.gov/iowa-illinois-reciprocal-agreement
          https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
        """
        
        # Check if the parameters are set upstream in the instance
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Indiana', 'state_occupation': 'Iowa', "capital_gains_short": capital_gains_short, "dependents": dependents, "sources": sources, "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def indiana_and_kansas(self, capital_gains_short=0, dependents=0, sources=False, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_long=0, filing_status='single', income=100000, credits_state=0):
        """
        --- Parameters ----
        
        Indiana:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Kansas:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Indiana:
        - https://www.in.gov/dor/tax-forms/2022-individual-income-tax-forms/
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-32.htm
          https://www.in.gov/dor/files/ib28.pdf
          https://www.in.gov/dor/tax-forms/2022-individual-income-tax-forms/
          https://www.in.gov/dor/files/2022-county-tax-rates-and-codes.pdf
        
        Kansas:
        -         
        - https://www.ksrevenue.gov/taxrates.html
          https://www.ksrevenue.gov/incomebook22.html#0
          https://www.ksrevenue.gov/incomebook22.html#0
          https://www.ksrevenue.gov/incomebook22.html#0
          https://www.ksrevenue.gov/incomebook22.html
          https://www.ksrevenue.gov/incomebook22.html
        """
        
        # Check if the parameters are set upstream in the instance
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Indiana', 'state_occupation': 'Kansas', "capital_gains_short": capital_gains_short, "dependents": dependents, "sources": sources, "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def indiana_and_kentucky(self, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Indiana:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Kentucky:
        


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Indiana:
        - https://www.in.gov/dor/tax-forms/2022-individual-income-tax-forms/
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-32.htm
          https://www.in.gov/dor/files/ib28.pdf
          https://www.in.gov/dor/tax-forms/2022-individual-income-tax-forms/
          https://www.in.gov/dor/files/2022-county-tax-rates-and-codes.pdf
        
        Kentucky:
        -         
        - https://revenue.ky.gov/News/Pages/Dor-Announces-Updates-to-Individual-Income-Tax-for-2023-Tax-Year.aspx
          https://revenue.ky.gov/Individual/Individual-Income-Tax/Pages/default.aspx
          https://revenue.ky.gov/News/Pages/Dor-Announces-Updates-to-Individual-Income-Tax-for-2023-Tax-Year.aspx
          https://revenue.ky.gov/Individual/Individual-Income-Tax/Pages/default.aspx
          https://revenue.ky.gov/Dor%20Training%20Materials/103%20KAR%2017.140.%20Individual%20income%20tax%20-%20reciprocity%20-%20nonresidents.pdf
          https://revenue.ky.gov/Individual/Individual-Income-Tax/Pages/default.aspx
          https://nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2021/Taxes-21-32.htm?taxmap=true
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Indiana', 'state_occupation': 'Kentucky', "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def indiana_and_louisiana(self, capital_gains_short=0, dependents=0, sources=False, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_long=0, filing_status='single', income=100000, credits_state=0):
        """
        --- Parameters ----
        
        Indiana:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Louisiana:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Indiana:
        - https://www.in.gov/dor/tax-forms/2022-individual-income-tax-forms/
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-32.htm
          https://www.in.gov/dor/files/ib28.pdf
          https://www.in.gov/dor/tax-forms/2022-individual-income-tax-forms/
          https://www.in.gov/dor/files/2022-county-tax-rates-and-codes.pdf
        
        Louisiana:
        -         
        - https://revenue.louisiana.gov/individualincometax
          https://revenue.louisiana.gov/individualincometax
          https://www.revenue.louisiana.gov/taxforms/6935(11_02)F.pdf
          https://revenue.louisiana.gov/individualincometax
          https://revenue.louisiana.gov/individualincometax
        """
        
        # Check if the parameters are set upstream in the instance
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Indiana', 'state_occupation': 'Louisiana', "capital_gains_short": capital_gains_short, "dependents": dependents, "sources": sources, "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def indiana_and_maine(self, capital_gains_short=0, dependents=0, sources=False, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_long=0, filing_status='single', income=100000, credits_state=0):
        """
        --- Parameters ----
        
        Indiana:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Maine:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Indiana:
        - https://www.in.gov/dor/tax-forms/2022-individual-income-tax-forms/
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-32.htm
          https://www.in.gov/dor/files/ib28.pdf
          https://www.in.gov/dor/tax-forms/2022-individual-income-tax-forms/
          https://www.in.gov/dor/files/2022-county-tax-rates-and-codes.pdf
        
        Maine:
        -         
        - https://www.maine.gov/revenue/tax-return-forms/individual-income-tax-2022
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/ind_tax_rate_sched_2022.pdf
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/22_1040me_book_gen_instruc_revisedFeb23.pdf
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/22_1040me_book_gen_instruc_revisedFeb23.pdf
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/22_1040me_book_gen_instruc_revisedFeb23.pdf
          https://www.maine.gov/revenue/tax-return-forms/individual-income-tax-2022
        """
        
        # Check if the parameters are set upstream in the instance
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Indiana', 'state_occupation': 'Maine', "capital_gains_short": capital_gains_short, "dependents": dependents, "sources": sources, "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def indiana_and_maryland(self, capital_gains_short=0, dependents=0, sources=False, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_long=0, filing_status='single', income=100000, credits_state=0):
        """
        --- Parameters ----
        
        Indiana:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Maryland:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Indiana:
        - https://www.in.gov/dor/tax-forms/2022-individual-income-tax-forms/
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-32.htm
          https://www.in.gov/dor/files/ib28.pdf
          https://www.in.gov/dor/tax-forms/2022-individual-income-tax-forms/
          https://www.in.gov/dor/files/2022-county-tax-rates-and-codes.pdf
        
        Maryland:
        -         
        - https://www.marylandtaxes.gov/individual/income/tax-info/tax-rates.php
          https://www.marylandtaxes.gov/forms/22_forms/Resident_Booklet.pdf
          https://www.marylandtaxes.gov/forms/Tax_Publications/Administrative_Releases/Income_and_Estate_Tax_Releases/ar_it37.pdf
          https://www.marylandtaxes.gov/individual/index.php
          https://www.marylandtaxes.gov/individual/income/tax-info/tax-rates.php
        """
        
        # Check if the parameters are set upstream in the instance
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Indiana', 'state_occupation': 'Maryland', "capital_gains_short": capital_gains_short, "dependents": dependents, "sources": sources, "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def indiana_and_massachusetts(self, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Indiana:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Massachusetts:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Indiana:
        - https://www.in.gov/dor/tax-forms/2022-individual-income-tax-forms/
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-32.htm
          https://www.in.gov/dor/files/ib28.pdf
          https://www.in.gov/dor/tax-forms/2022-individual-income-tax-forms/
          https://www.in.gov/dor/files/2022-county-tax-rates-and-codes.pdf
        
        Massachusetts:
        -         
        - https://www.mass.gov/service-details/massachusetts-tax-rates
          https://www.mass.gov/service-details/view-massachusetts-personal-income-tax-exemptions
          https://www.mass.gov/service-details/view-massachusetts-personal-income-tax-exemptions
          https://www.mass.gov/service-details/view-child-and-dependent-related-credits
          https://www.mass.gov/service-details/massachusetts-tax-rates
          https://www.mass.gov/personal-income-tax
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Indiana', 'state_occupation': 'Massachusetts', "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def indiana_and_michigan(self, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Indiana:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Michigan:
        


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Indiana:
        - https://www.in.gov/dor/tax-forms/2022-individual-income-tax-forms/
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-32.htm
          https://www.in.gov/dor/files/ib28.pdf
          https://www.in.gov/dor/tax-forms/2022-individual-income-tax-forms/
          https://www.in.gov/dor/files/2022-county-tax-rates-and-codes.pdf
        
        Michigan:
        -         
        - https://www.michigan.gov/taxes/iit/new-developments-for-tax-year-2022
          "https://www.michigan.gov/taxes/-/media/Project/Websites/taxes/Forms/2023/2023-SUW/446_Withholding-Guide_2023.pdf?rev=177ca3d8cb034392bf84e317ed78c320&hash=19C5F42F43A89A91863D229349F48883#:~:text=The%20exemption%20amount%20is%20%245%2C400
         employees%20Michigan%20income%20tax%20return"
          "https://www.michigan.gov/taxes/-/media/Project/Websites/taxes/Forms/2023/2023-SUW/446_Withholding-Guide_2023.pdf?rev=177ca3d8cb034392bf84e317ed78c320&hash=19C5F42F43A89A91863D229349F48883#:~:text=The%20exemption%20amount%20is%20%245%2C400
         employees%20Michigan%20income%20tax%20return"
          https://www.michigan.gov/taxes/questions/iit/accordion/residency/are-my-wages-earned-in-another-state-taxable-in-michigan-if-i-am-a-michigan-resident-1
          https://www.michigan.gov/taxes/-/media/Project/Websites/taxes/Forms/2022/2022-IIT-Forms/BOOK_MI-1040.pdf?rev=4524f1cbbc0243b194ab05a6680be75b&hash=40DFE71E651F7C1D65C31F6CD03D944F
          https://www.michigan.gov/taxes/questions/iit/accordion/general/what-cities-impose-an-income-tax
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Indiana', 'state_occupation': 'Michigan', "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def indiana_and_minnesota(self, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Indiana:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Minnesota:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Indiana:
        - https://www.in.gov/dor/tax-forms/2022-individual-income-tax-forms/
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-32.htm
          https://www.in.gov/dor/files/ib28.pdf
          https://www.in.gov/dor/tax-forms/2022-individual-income-tax-forms/
          https://www.in.gov/dor/files/2022-county-tax-rates-and-codes.pdf
        
        Minnesota:
        -         
        - https://www.revenue.state.mn.us/minnesota-income-tax-rates-and-brackets
          https://www.revenue.state.mn.us/child-and-dependent-care-credit
          https://www.revenue.state.mn.us/dependent-exemptions
          https://www.revenue.state.mn.us/minnesota-standard-deduction
          https://mn.gov/mmb/assets/mwr_form_tcm1059-128581.pdf
          https://www.revenue.state.mn.us/individual-income-tax
          https://www.revenue.state.mn.us/individual-income-tax
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Indiana', 'state_occupation': 'Minnesota', "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def indiana_and_mississippi(self, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Indiana:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Mississippi:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Indiana:
        - https://www.in.gov/dor/tax-forms/2022-individual-income-tax-forms/
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-32.htm
          https://www.in.gov/dor/files/ib28.pdf
          https://www.in.gov/dor/tax-forms/2022-individual-income-tax-forms/
          https://www.in.gov/dor/files/2022-county-tax-rates-and-codes.pdf
        
        Mississippi:
        -         
        - https://www.dor.ms.gov/individual/tax-rates
          https://www.dor.ms.gov/individual/tax-rates
          https://www.dor.ms.gov/individual/tax-rates
          https://www.dor.ms.gov/individual
          https://www.dor.ms.gov/individual
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Indiana', 'state_occupation': 'Mississippi', "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def indiana_and_missouri(self, capital_gains_short=0, dependents=0, sources=False, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_long=0, filing_status='single', income=100000, credits_state=0):
        """
        --- Parameters ----
        
        Indiana:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Missouri:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Indiana:
        - https://www.in.gov/dor/tax-forms/2022-individual-income-tax-forms/
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-32.htm
          https://www.in.gov/dor/files/ib28.pdf
          https://www.in.gov/dor/tax-forms/2022-individual-income-tax-forms/
          https://www.in.gov/dor/files/2022-county-tax-rates-and-codes.pdf
        
        Missouri:
        -         
        - https://dor.mo.gov/taxation/individual/tax-types/income/year-changes/
          https://dor.mo.gov/taxation/individual/tax-types/income/year-changes/
          https://dor.mo.gov/taxation/individual/tax-types/income/year-changes/
          https://www.kcmo.gov/city-hall/departments/finance/earnings-tax
          https://www.stlouis-mo.gov/government/departments/collector/earnings-tax/individual-earnings-tax-info.cfm
        """
        
        # Check if the parameters are set upstream in the instance
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Indiana', 'state_occupation': 'Missouri', "capital_gains_short": capital_gains_short, "dependents": dependents, "sources": sources, "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def indiana_and_montana(self, capital_gains_short=0, dependents=0, sources=False, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_long=0, filing_status='single', income=100000, credits_state=0):
        """
        --- Parameters ----
        
        Indiana:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Montana:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Indiana:
        - https://www.in.gov/dor/tax-forms/2022-individual-income-tax-forms/
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-32.htm
          https://www.in.gov/dor/files/ib28.pdf
          https://www.in.gov/dor/tax-forms/2022-individual-income-tax-forms/
          https://www.in.gov/dor/files/2022-county-tax-rates-and-codes.pdf
        
        Montana:
        -         
        - https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/2022/12/montana-employees-withholding-allowance-and-exemption-certificate-form-mw-4.pdf
          https://mtrevenue.gov/2021/11/29/simplification-of-montana-income-taxation/?utm_source=rss&utm_medium=rss&utm_campaign=simplification-of-montana-income-taxation
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Indiana', 'state_occupation': 'Montana', "capital_gains_short": capital_gains_short, "dependents": dependents, "sources": sources, "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def indiana_and_nebraska(self, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Indiana:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Nebraska:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Indiana:
        - https://www.in.gov/dor/tax-forms/2022-individual-income-tax-forms/
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-32.htm
          https://www.in.gov/dor/files/ib28.pdf
          https://www.in.gov/dor/tax-forms/2022-individual-income-tax-forms/
          https://www.in.gov/dor/files/2022-county-tax-rates-and-codes.pdf
        
        Nebraska:
        -         
        - https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2023_Circular_EN_8-429r2023_Whole_Book_11.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Indiana', 'state_occupation': 'Nebraska', "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def indiana_and_nevada(self, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Indiana:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Nevada:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Indiana:
        - https://www.in.gov/dor/tax-forms/2022-individual-income-tax-forms/
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-32.htm
          https://www.in.gov/dor/files/ib28.pdf
          https://www.in.gov/dor/tax-forms/2022-individual-income-tax-forms/
          https://www.in.gov/dor/files/2022-county-tax-rates-and-codes.pdf
        
        Nevada:
        -         
        - https://www.nvsos.gov/sos/businesses/the-nevada-advantage
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Indiana', 'state_occupation': 'Nevada', "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def indiana_and_new_hampshire(self, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Indiana:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        New Hampshire:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Indiana:
        - https://www.in.gov/dor/tax-forms/2022-individual-income-tax-forms/
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-32.htm
          https://www.in.gov/dor/files/ib28.pdf
          https://www.in.gov/dor/tax-forms/2022-individual-income-tax-forms/
          https://www.in.gov/dor/files/2022-county-tax-rates-and-codes.pdf
        
        New Hampshire:
        -         
        - https://www.revenue.nh.gov/assistance/tax-overview.htm
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Indiana', 'state_occupation': 'NewHampshire', "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def indiana_and_new_jersey(self, capital_gains_short=0, dependents=0, sources=False, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_long=0, filing_status='single', income=100000, credits_state=0):
        """
        --- Parameters ----
        
        Indiana:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        New Jersey:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Indiana:
        - https://www.in.gov/dor/tax-forms/2022-individual-income-tax-forms/
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-32.htm
          https://www.in.gov/dor/files/ib28.pdf
          https://www.in.gov/dor/tax-forms/2022-individual-income-tax-forms/
          https://www.in.gov/dor/files/2022-county-tax-rates-and-codes.pdf
        
        New Jersey:
        -         
        - https://www.state.nj.us/treasury/taxation/taxtables.shtml
          https://www.state.nj.us/treasury/taxation/njit13.shtml
          https://www.nj.gov/treasury/taxation/njit25.shtml
          https://www.state.nj.us/treasury/taxation/individuals/freqqiti.shtml
          https://www.state.nj.us/treasury/taxation/localtax.shtml
        """
        
        # Check if the parameters are set upstream in the instance
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Indiana', 'state_occupation': 'NewJersey', "capital_gains_short": capital_gains_short, "dependents": dependents, "sources": sources, "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def indiana_and_new_mexico(self, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Indiana:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        New Mexico:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Indiana:
        - https://www.in.gov/dor/tax-forms/2022-individual-income-tax-forms/
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-32.htm
          https://www.in.gov/dor/files/ib28.pdf
          https://www.in.gov/dor/tax-forms/2022-individual-income-tax-forms/
          https://www.in.gov/dor/files/2022-county-tax-rates-and-codes.pdf
        
        New Mexico:
        -         
        - https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/fdf3c548-8aba-4b9c-9eb4-bb564c716015/FYI-104.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/0bf943fd-652e-4400-bb1b-cd8397eb5a95/2022pit-adj.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/2f1a6781-9534-4436-b427-1557f9592099/2022pit-adj-ins.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/0bf943fd-652e-4400-bb1b-cd8397eb5a95/2022pit-adj.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/1afc56af-ea90-4d48-82e5-1f9aeb43255a/PITbook2022.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/fdf3c548-8aba-4b9c-9eb4-bb564c716015/FYI-104.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Indiana', 'state_occupation': 'NewMexico', "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def indiana_and_new_york(self, capital_gains_short=0, dependents=0, sources=False, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_long=0, filing_status='single', income=100000, credits_state=0):
        """
        --- Parameters ----
        
        Indiana:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        New York:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Indiana:
        - https://www.in.gov/dor/tax-forms/2022-individual-income-tax-forms/
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-32.htm
          https://www.in.gov/dor/files/ib28.pdf
          https://www.in.gov/dor/tax-forms/2022-individual-income-tax-forms/
          https://www.in.gov/dor/files/2022-county-tax-rates-and-codes.pdf
        
        New York:
        -         
        - https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nys-tax-rate-schedule
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nys-tax-rate-schedule
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nys-tax-rate-schedule
          https://www.tax.ny.gov/pit/
          https://www.tax.ny.gov/pit/file/nonresident-faqs.htm
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#yonkers-tax-rate-schedule
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nyc-tax-rate-schedule
        """
        
        # Check if the parameters are set upstream in the instance
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Indiana', 'state_occupation': 'NewYork', "capital_gains_short": capital_gains_short, "dependents": dependents, "sources": sources, "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def indiana_and_north_carolina(self, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Indiana:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        North Carolina:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Indiana:
        - https://www.in.gov/dor/tax-forms/2022-individual-income-tax-forms/
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-32.htm
          https://www.in.gov/dor/files/ib28.pdf
          https://www.in.gov/dor/tax-forms/2022-individual-income-tax-forms/
          https://www.in.gov/dor/files/2022-county-tax-rates-and-codes.pdf
        
        North Carolina:
        -         
        - https://www.ncdor.gov/taxes-forms/tax-rate-schedules
          https://www.ncdor.gov/taxes-forms/individual-income-tax/north-carolina-standard-deduction-or-north-carolina-itemized-deductions
          https://www.ncdor.gov/taxes-forms/individual-income-tax
          https://www.ncdor.gov/taxes-forms/individual-income-tax
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Indiana', 'state_occupation': 'NorthCarolina', "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def indiana_and_north_dakota(self, capital_gains_short=0, dependents=0, sources=False, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_long=0, filing_status='single', income=100000, credits_state=0):
        """
        --- Parameters ----
        
        Indiana:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        North Dakota:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Indiana:
        - https://www.in.gov/dor/tax-forms/2022-individual-income-tax-forms/
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-32.htm
          https://www.in.gov/dor/files/ib28.pdf
          https://www.in.gov/dor/tax-forms/2022-individual-income-tax-forms/
          https://www.in.gov/dor/files/2022-county-tax-rates-and-codes.pdf
        
        North Dakota:
        -         
        - https://www.tax.nd.gov/forms
          https://www.tax.nd.gov/sites/www/files/documents/forms/individual/2022-iit/2022-individual-income-tax-booklet.pdf
          https://www.tax.nd.gov/sites/www/files/documents/forms/2022-individual-income-tax-booklet.pdf
          https://www.tax.nd.gov/tax-exemptions-credits/income-tax-exemptions-credits
          https://www.tax.nd.gov/sites/www/files/documents/forms/business/ndwrfillable.pdf
          https://www.tax.nd.gov/news/tax-legislative-changes/significant-changes-law/individual-income-tax-history
          https://www.tax.nd.gov/forms
        """
        
        # Check if the parameters are set upstream in the instance
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Indiana', 'state_occupation': 'NorthDakota', "capital_gains_short": capital_gains_short, "dependents": dependents, "sources": sources, "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def indiana_and_ohio(self, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Indiana:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Ohio:
        


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Indiana:
        - https://www.in.gov/dor/tax-forms/2022-individual-income-tax-forms/
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-32.htm
          https://www.in.gov/dor/files/ib28.pdf
          https://www.in.gov/dor/tax-forms/2022-individual-income-tax-forms/
          https://www.in.gov/dor/files/2022-county-tax-rates-and-codes.pdf
        
        Ohio:
        -         
        - https://tax.ohio.gov/individual/resources/annual-tax-rates
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://tax.ohio.gov/static/forms/employer_withholding/generic/wth_it4nr.pdf
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://www.ritaohio.com/TaxRatesTable
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Indiana', 'state_occupation': 'Ohio', "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def indiana_and_oklahoma(self, capital_gains_short=0, dependents=0, sources=False, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_long=0, filing_status='single', income=100000, credits_state=0):
        """
        --- Parameters ----
        
        Indiana:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Oklahoma:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Indiana:
        - https://www.in.gov/dor/tax-forms/2022-individual-income-tax-forms/
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-32.htm
          https://www.in.gov/dor/files/ib28.pdf
          https://www.in.gov/dor/tax-forms/2022-individual-income-tax-forms/
          https://www.in.gov/dor/files/2022-county-tax-rates-and-codes.pdf
        
        Oklahoma:
        -         
        - https://oklahoma.gov/content/dam/ok/en/tax/documents/resources/publications/businesses/withholding-tables/WHTables-2022.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Indiana', 'state_occupation': 'Oklahoma', "capital_gains_short": capital_gains_short, "dependents": dependents, "sources": sources, "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def indiana_and_oregon(self, capital_gains_short=0, dependents=0, sources=False, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_long=0, filing_status='single', income=100000, credits_state=0):
        """
        --- Parameters ----
        
        Indiana:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Oregon:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Indiana:
        - https://www.in.gov/dor/tax-forms/2022-individual-income-tax-forms/
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-32.htm
          https://www.in.gov/dor/files/ib28.pdf
          https://www.in.gov/dor/tax-forms/2022-individual-income-tax-forms/
          https://www.in.gov/dor/files/2022-county-tax-rates-and-codes.pdf
        
        Oregon:
        -         
        - https://www.oregon.gov/dor/programs/individuals/Pages/PIT.aspx
          https://www.oregon.gov/dor/programs/individuals/Documents/Full%20year%20resident
         %20Form%20or-40%20filers.pdf
          https://www.oregon.gov/dor/programs/individuals/Documents/Part-year%20and%20nonresident
         %20Form%20or-40-P%20and%20or-40-N%20filers.pdf.pdf
          https://www.oregon.gov/dor/forms/FormsPubs/form-or-40_101-040_2022.pdf
          https://www.oregon.gov/dor/forms/FormsPubs/form-or-40_101-040_2022.pdf
          https://www.oregon.gov/dor/forms/FormsPubs/form-or-40_101-040_2022.pdf
          https://www.portland.gov/revenue/personal-tax
        """
        
        # Check if the parameters are set upstream in the instance
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Indiana', 'state_occupation': 'Oregon', "capital_gains_short": capital_gains_short, "dependents": dependents, "sources": sources, "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def indiana_and_pennsylvania(self, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Indiana:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Pennsylvania:
        


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Indiana:
        - https://www.in.gov/dor/tax-forms/2022-individual-income-tax-forms/
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-32.htm
          https://www.in.gov/dor/files/ib28.pdf
          https://www.in.gov/dor/tax-forms/2022-individual-income-tax-forms/
          https://www.in.gov/dor/files/2022-county-tax-rates-and-codes.pdf
        
        Pennsylvania:
        -         
        - https://www.revenue.pa.gov/Tax%20Rates/Pages/default.aspx
          https://www.revenue.pa.gov/FormsandPublications/FormsforIndividuals/PIT/Documents/2022/2022_pa-40in.pdf
          https://www.revenue.pa.gov/FormsandPublications/FormsforBusinesses/EmployerWithholding/Documents/rev-419.pdf
          https://www.revenue.pa.gov/FormsandPublications/FormsforIndividuals/PIT/Documents/2022/2022_pa-40in.pdf
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-25.htm
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Indiana', 'state_occupation': 'Pennsylvania', "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def indiana_and_rhode_island(self, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Indiana:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Rhode Island:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Indiana:
        - https://www.in.gov/dor/tax-forms/2022-individual-income-tax-forms/
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-32.htm
          https://www.in.gov/dor/files/ib28.pdf
          https://www.in.gov/dor/tax-forms/2022-individual-income-tax-forms/
          https://www.in.gov/dor/files/2022-county-tax-rates-and-codes.pdf
        
        Rhode Island:
        -         
        - https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2022_1040WE_w_0.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Indiana', 'state_occupation': 'RhodeIsland', "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def indiana_and_south_carolina(self, capital_gains_short=0, dependents=0, sources=False, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_long=0, filing_status='single', income=100000, credits_state=0):
        """
        --- Parameters ----
        
        Indiana:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        South Carolina:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Indiana:
        - https://www.in.gov/dor/tax-forms/2022-individual-income-tax-forms/
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-32.htm
          https://www.in.gov/dor/files/ib28.pdf
          https://www.in.gov/dor/tax-forms/2022-individual-income-tax-forms/
          https://www.in.gov/dor/files/2022-county-tax-rates-and-codes.pdf
        
        South Carolina:
        -         
        - https://dor.sc.gov/resources-site/lawandpolicy/Advisory%20Opinions/IL22-15.pdf
          https://dor.sc.gov/forms-site/Forms/IITPacket_2022.pdf
          https://www.scstatehouse.gov/code/t12c006.php
          https://dor.sc.gov/resources-site/lawandpolicy/Advisory%20Opinions/IL22-15.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Indiana', 'state_occupation': 'SouthCarolina', "capital_gains_short": capital_gains_short, "dependents": dependents, "sources": sources, "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def indiana_and_south_dakota(self, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Indiana:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        South Dakota:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Indiana:
        - https://www.in.gov/dor/tax-forms/2022-individual-income-tax-forms/
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-32.htm
          https://www.in.gov/dor/files/ib28.pdf
          https://www.in.gov/dor/tax-forms/2022-individual-income-tax-forms/
          https://www.in.gov/dor/files/2022-county-tax-rates-and-codes.pdf
        
        South Dakota:
        -         
        - https://dor.sd.gov/individuals/taxes/
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Indiana', 'state_occupation': 'SouthDakota', "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def indiana_and_tennessee(self, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Indiana:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Tennessee:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Indiana:
        - https://www.in.gov/dor/tax-forms/2022-individual-income-tax-forms/
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-32.htm
          https://www.in.gov/dor/files/ib28.pdf
          https://www.in.gov/dor/tax-forms/2022-individual-income-tax-forms/
          https://www.in.gov/dor/files/2022-county-tax-rates-and-codes.pdf
        
        Tennessee:
        -         
        - https://www.tn.gov/revenue/taxes.html
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Indiana', 'state_occupation': 'Tennessee', "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def indiana_and_texas(self, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Indiana:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Texas:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Indiana:
        - https://www.in.gov/dor/tax-forms/2022-individual-income-tax-forms/
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-32.htm
          https://www.in.gov/dor/files/ib28.pdf
          https://www.in.gov/dor/tax-forms/2022-individual-income-tax-forms/
          https://www.in.gov/dor/files/2022-county-tax-rates-and-codes.pdf
        
        Texas:
        -         
        - https://comptroller.texas.gov/taxes/
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Indiana', 'state_occupation': 'Texas', "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def indiana_and_utah(self, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Indiana:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Utah:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Indiana:
        - https://www.in.gov/dor/tax-forms/2022-individual-income-tax-forms/
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-32.htm
          https://www.in.gov/dor/files/ib28.pdf
          https://www.in.gov/dor/tax-forms/2022-individual-income-tax-forms/
          https://www.in.gov/dor/files/2022-county-tax-rates-and-codes.pdf
        
        Utah:
        -         
        - https://incometax.utah.gov/paying/tax-rates
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Indiana', 'state_occupation': 'Utah', "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def indiana_and_vermont(self, capital_gains_short=0, dependents=0, sources=False, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_long=0, filing_status='single', income=100000, credits_state=0):
        """
        --- Parameters ----
        
        Indiana:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Vermont:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Indiana:
        - https://www.in.gov/dor/tax-forms/2022-individual-income-tax-forms/
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-32.htm
          https://www.in.gov/dor/files/ib28.pdf
          https://www.in.gov/dor/tax-forms/2022-individual-income-tax-forms/
          https://www.in.gov/dor/files/2022-county-tax-rates-and-codes.pdf
        
        Vermont:
        -         
        - https://tax.vermont.gov/sites/tax/files/documents/GB-1210-2023.pdf
          https://tax.vermont.gov/sites/tax/files/documents/Income%20Booklet-2022.pdf
          https://tax.vermont.gov/sites/tax/files/documents/Income%20Booklet-2022.pdf
          https://tax.vermont.gov/sites/tax/files/documents/Income%20Booklet-2022.pdf
          https://tax.vermont.gov/individuals/personal-income-tax/tax-credits
          https://tax.vermont.gov/individuals/personal-income-tax/taxable-income
          https://tax.vermont.gov/individuals/personal-income-tax/local-option-tax
        """
        
        # Check if the parameters are set upstream in the instance
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Indiana', 'state_occupation': 'Vermont', "capital_gains_short": capital_gains_short, "dependents": dependents, "sources": sources, "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def indiana_and_virginia(self, capital_gains_short=0, dependents=0, sources=False, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_long=0, filing_status='single', income=100000, credits_state=0):
        """
        --- Parameters ----
        
        Indiana:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Virginia:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Indiana:
        - https://www.in.gov/dor/tax-forms/2022-individual-income-tax-forms/
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-32.htm
          https://www.in.gov/dor/files/ib28.pdf
          https://www.in.gov/dor/tax-forms/2022-individual-income-tax-forms/
          https://www.in.gov/dor/files/2022-county-tax-rates-and-codes.pdf
        
        Virginia:
        -         
        - https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/reciprocity
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Indiana', 'state_occupation': 'Virginia', "capital_gains_short": capital_gains_short, "dependents": dependents, "sources": sources, "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def indiana_and_washington(self, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Indiana:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Washington:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Indiana:
        - https://www.in.gov/dor/tax-forms/2022-individual-income-tax-forms/
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-32.htm
          https://www.in.gov/dor/files/ib28.pdf
          https://www.in.gov/dor/tax-forms/2022-individual-income-tax-forms/
          https://www.in.gov/dor/files/2022-county-tax-rates-and-codes.pdf
        
        Washington:
        -         
        - https://dor.wa.gov/taxes-rates/income-tax
          https://dor.wa.gov/taxes-rates/other-taxes/capital-gains-tax
          https://dor.wa.gov/taxes-rates/income-tax
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Indiana', 'state_occupation': 'Washington', "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def indiana_and_west_virginia(self, pay_periods=24, capital_gains_short=0, county_occupation='Other', dependents=0, sources=False, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_long=0, filing_status='single', income=100000, credits_state=0):
        """
        --- Parameters ----
        
        Indiana:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        West Virginia:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_occupation : str, optional
            County of occupation. Default is Other.
        pay_periods : int, optional
            Number of pay periods. Options are 13, 24, 26, 52. Default is 24.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Indiana:
        - https://www.in.gov/dor/tax-forms/2022-individual-income-tax-forms/
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-32.htm
          https://www.in.gov/dor/files/ib28.pdf
          https://www.in.gov/dor/tax-forms/2022-individual-income-tax-forms/
          https://www.in.gov/dor/files/2022-county-tax-rates-and-codes.pdf
        
        West Virginia:
        -         
        - https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://tax.wv.gov/Documents/TSD/tsd438.pdf
          https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2011/Taxes-11-35att.html
        """
        
        # Check if the parameters are set upstream in the instance
        pay_periods = self.pay_periods if hasattr(self, 'pay_periods') else pay_periods
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        county_occupation = self.county_occupation if hasattr(self, 'county_occupation') else county_occupation
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Indiana', 'state_occupation': 'WestVirginia', "pay_periods": pay_periods, "capital_gains_short": capital_gains_short, "county_occupation": county_occupation, "dependents": dependents, "sources": sources, "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def indiana_and_wisconsin(self, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Indiana:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Wisconsin:
        


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Indiana:
        - https://www.in.gov/dor/tax-forms/2022-individual-income-tax-forms/
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-32.htm
          https://www.in.gov/dor/files/ib28.pdf
          https://www.in.gov/dor/tax-forms/2022-individual-income-tax-forms/
          https://www.in.gov/dor/files/2022-county-tax-rates-and-codes.pdf
        
        Wisconsin:
        -         
        - https://www.revenue.wi.gov/TaxForms2023/2023-Form1-ES-Inst.pdf
          https://www.revenue.wi.gov/TaxForms2023/2023-Form1-ES-Inst.pdf
          https://www.revenue.wi.gov/TaxForms2023/2023-Form1-ES-Inst.pdf
          https://www.revenue.wi.gov/Pages/FAQS/pcs-work.aspx
          https://www.revenue.wi.gov/Dor%20Publications/pb103.pdf
          https://www.revenue.wi.gov/Pages/FAQS/pcs-work.aspx
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Indiana', 'state_occupation': 'Wisconsin', "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def indiana_and_wyoming(self, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Indiana:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Wyoming:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Indiana:
        - https://www.in.gov/dor/tax-forms/2022-individual-income-tax-forms/
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-32.htm
          https://www.in.gov/dor/files/ib28.pdf
          https://www.in.gov/dor/tax-forms/2022-individual-income-tax-forms/
          https://www.in.gov/dor/files/2022-county-tax-rates-and-codes.pdf
        
        Wyoming:
        -         
        - https://revenue.wyo.gov/
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Indiana', 'state_occupation': 'Wyoming', "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def iowa_and_alabama(self, county_residence='Other', state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Iowa:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Alabama:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Iowa:
        - https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
          https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
          https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
          https://tax.iowa.gov/iowa-illinois-reciprocal-agreement
          https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
        
        Alabama:
        -         
        - https://www.revenue.alabama.gov/ultraviewer/viewer/basic_viewer/index.html?form=2023/02/22f40.pdf
          https://www.revenue.alabama.gov/faqs/what-is-alabamas-individual-income-tax-rate/
          https://www.revenue.alabama.gov/ultraviewer/viewer/basic_viewer/index.html?form=2023/02/22f40.pdf
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-24.htm?taxmap=true
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-24.htm?taxmap=true
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-24.htm?taxmap=true
          https://www.revenue.alabama.gov/individual-corporate/income-to-be-reported-on-the-alabama-income-tax-return/
          https://www.almonline.org/TaxRates.aspx#OccupationalTax
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Iowa', 'state_occupation': 'Alabama', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def iowa_and_alaska(self, county_residence='Other', state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Iowa:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Alaska:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Iowa:
        - https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
          https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
          https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
          https://tax.iowa.gov/iowa-illinois-reciprocal-agreement
          https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
        
        Alaska:
        -         
        - https://www.commerce.alaska.gov/web/dcra/officeofthestateassessor/alaskataxfacts.aspx
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Iowa', 'state_occupation': 'Alaska', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def iowa_and_arizona(self, county_residence='Other', state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Iowa:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Arizona:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        credits_state : int, optional
            Credits for state taxes. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Iowa:
        - https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
          https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
          https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
          https://tax.iowa.gov/iowa-illinois-reciprocal-agreement
          https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
        
        Arizona:
        -         
        - https://azdor.gov/forms/individual/form-140-x-y-tables
          https://azdor.gov/forms/individual/form-140-x-y-tables
          https://azdor.gov/forms/individual/form-140-x-y-tables
          https://azdor.gov/businesses-arizona/withholding-tax/withholding-exceptions
          "https://azdor.gov/individuals/income-tax-filing-assistance/identifying-other-taxable-income#:~:text=In%20regards%20to%20capital%20gains
         the%20individuals%20regular%20tax%20rate.
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Iowa', 'state_occupation': 'Arizona', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def iowa_and_arkansas(self, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Iowa:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Arkansas:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Iowa:
        - https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
          https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
          https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
          https://tax.iowa.gov/iowa-illinois-reciprocal-agreement
          https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
        
        Arkansas:
        -         
        - https://www.arkansasedc.com/why-arkansas/business-climate/tax-structure/personal-income-tax
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_and_AR1000NR_Instructions.pdf
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_FullYearResidentIndividualIncomeTaxReturn.pdf
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_and_AR1000NR_Instructions.pdf
          https://www.arkansasedc.com/why-arkansas/business-climate/tax-structure/capital-gains-tax-reduction
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_and_AR1000NR_Instructions.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Iowa', 'state_occupation': 'Arkansas', "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def iowa_and_california(self, county_residence='Other', state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Iowa:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        California:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Iowa:
        - https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
          https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
          https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
          https://tax.iowa.gov/iowa-illinois-reciprocal-agreement
          https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
        
        California:
        -         
        - https://www.ftb.ca.gov/forms/2022/2022-540-tax-rate-schedules.pdf
          https://www.ftb.ca.gov/file/personal/deductions/index.html
          https://www.ftb.ca.gov/about-ftb/newsroom/caleitc/eligibility-and-credit-information.html
          https://www.ftb.ca.gov/file/personal/income-types/capital-gains-and-losses.html#:~:text=California%20does%20not%20have%20a
         are%20taxed%20as%20ordinary%20income
          https://www.ftb.ca.gov/file/personal/
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Iowa', 'state_occupation': 'California', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def iowa_and_colorado(self, county_residence='Other', state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Iowa:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Colorado:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Iowa:
        - https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
          https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
          https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
          https://tax.iowa.gov/iowa-illinois-reciprocal-agreement
          https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
        
        Colorado:
        -         
        - https://leg.colorado.gov/agencies/legislative-council-staff/individual-income-tax%C2%A0
          https://leg.colorado.gov/agencies/legislative-council-staff/individual-income-tax%C2%A0
          https://tax.colorado.gov/sites/tax/files/documents/ITT_Capital_Gain_Subtraction_Feb_2022.pdf
          https://www.auroragov.org/business_services/taxes/occupational_privilege_tax
          https://www.denvergov.org/Government/Agencies-Departments-Offices/Agencies-Departments-Offices-Directory/Department-of-Finance/Our-Divisions/Treasury/Business-Tax-Information#:~:text=Occupational%20privilege%20tax
         -%28show%20below%29&text=Each%20taxable%20employee%20is%20liable
         month%20for%20each%20taxable%20employee
          https://www.glendale.co.us/355/Occupational-Privilege-Tax
          https://www.greenwoodvillage.com/1220/Occupational-Privilege-Tax-OPT
          https://ci.sheridan.co.us/288/Occupational-Privilege-Tax
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Iowa', 'state_occupation': 'Colorado', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def iowa_and_connecticut(self, county_residence='Other', state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Iowa:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Connecticut:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Iowa:
        - https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
          https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
          https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
          https://tax.iowa.gov/iowa-illinois-reciprocal-agreement
          https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
        
        Connecticut:
        -         
        - https://portal.ct.gov/-/media/DRS/Forms/2022/Income/2022-CT-1040-Instructions_1222.pdf
          https://www.cga.ct.gov/2022/rpt/pdf/2022-R-0108.pdf
          https://www.cga.ct.gov/current/pub/chap_229.htm#sec_12-703
          https://portal.ct.gov/-/media/DRS/Forms/2022/Income/2022-CT-1040-Instructions_1222.pdf
          "https://portal.ct.gov/DRS/Publications/TSSNs/TSSN-29#:~:text=An%20individuals%20net%20capital%20gains
         portion%20of%20Social%20Security%20benefits"
          "https://portal.ct.gov/DRS/Publications/TSSNs/TSSN-29#:~:text=An%20individuals%20net%20capital%20gains
         portion%20of%20Social%20Security%20benefits"
          https://portal.ct.gov/DRS/Individuals/Resident-Income-Tax/Tax-Information
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Iowa', 'state_occupation': 'Connecticut', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def iowa_and_delaware(self, county_residence='Other', state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, federal_agi=75000, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Iowa:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Delaware:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        federal_agi : int, optional
            Adjusted Gross Income (AGI) for federal taxes. Default is 0. Sources - https://www.irs.gov/e-file-providers/definition-of-adjusted-gross-income, https://www.irs.gov/newsroom/taxpayers-should-know-the-difference-between-standard-and-itemized-deductions
        income : int, optional
            Gross income. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Iowa:
        - https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
          https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
          https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
          https://tax.iowa.gov/iowa-illinois-reciprocal-agreement
          https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
        
        Delaware:
        -         
        - https://revenue.delaware.gov/software-developer/tax-rate-changes/
          https://revenuefiles.delaware.gov/2022/PIT-RES_TY22_2022-01_PaperInteractive.pdf
          https://revenuefiles.delaware.gov/2022/PIT-RES_TY22_2022-02_Instructions.pdf
          https://revenue.delaware.gov/frequently-asked-questions/personal-income-tax-faqs/
          https://www.newcastlede.gov/DocumentCenter/View/20147/City-of-Wilmington-Wage-Tax-Information-PDF
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        federal_agi = self.federal_agi if hasattr(self, 'federal_agi') else federal_agi
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Iowa', 'state_occupation': 'Delaware', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "federal_agi": federal_agi, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def iowa_and_florida(self, county_residence='Other', state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Iowa:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Florida:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Iowa:
        - https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
          https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
          https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
          https://tax.iowa.gov/iowa-illinois-reciprocal-agreement
          https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
        
        Florida:
        -         
        - https://floridarevenue.com/taxes/taxesfees/Pages/tax_interest_rates.aspx
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Iowa', 'state_occupation': 'Florida', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def iowa_and_georgia(self, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Iowa:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Georgia:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Iowa:
        - https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
          https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
          https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
          https://tax.iowa.gov/iowa-illinois-reciprocal-agreement
          https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
        
        Georgia:
        -         
        - https://dor.georgia.gov/tax-tables-georgia-tax-rate-schedule
          https://apps.dor.ga.gov/FillableForms/PDFViewer/Index?form=2022GA500
          https://apps.dor.ga.gov/FillableForms/PDFViewer/Index?form=2022GA500
          https://apps.dor.ga.gov/FillableForms/PDFViewer/Index?form=2022GA500
          https://dor.georgia.gov/years-individual-income-tax-forms
          https://dor.georgia.gov/years-individual-income-tax-forms
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Iowa', 'state_occupation': 'Georgia', "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def iowa_and_hawaii(self, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Iowa:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Hawaii:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Iowa:
        - https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
          https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
          https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
          https://tax.iowa.gov/iowa-illinois-reciprocal-agreement
          https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
        
        Hawaii:
        -         
        - https://tax.hawaii.gov/forms/d_18table-on/d_18table-on_p13/
          https://files.hawaii.gov/tax/forms/2018/18table-on.pdf
          https://files.hawaii.gov/tax/news/pubs/22outline.pdf
          https://files.hawaii.gov/tax/news/pubs/FastTaxRefGuide_2017.pdf
          https://www.tfhawaii.org/wordpress/state-tax-resources/mini-tax-guides/individual-income-tax-chapter-235/#:~:text=Long%20term%20capital%20gains%20are%20taxed%20at%20a%20maximum%20of%207.25%25
          https://www.grassrootinstitute.org/2023/02/hb337-increasing-tax-on-capital-gains-a-bad-risk/
          https://files.hawaii.gov/tax/news/pubs/22outline.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Iowa', 'state_occupation': 'Hawaii', "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def iowa_and_idaho(self, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Iowa:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Idaho:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Iowa:
        - https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
          https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
          https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
          https://tax.iowa.gov/iowa-illinois-reciprocal-agreement
          https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
        
        Idaho:
        -         
        - https://tax.idaho.gov/taxes/income-tax/individual-income/individual-income-tax-rate-schedule/
          https://tax.idaho.gov/wp-content/uploads/forms/EIN00046/EIN00046_03-01-2023.pdf
          https://tax.idaho.gov/taxes/income-tax/individual-income/idaho-source-income/
          https://tax.idaho.gov/taxes/income-tax/individual-income/idaho-source-income/
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Iowa', 'state_occupation': 'Idaho', "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def iowa_and_illinois(self, county_residence='Other', state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Iowa:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Illinois:
        


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Iowa:
        - https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
          https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
          https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
          https://tax.iowa.gov/iowa-illinois-reciprocal-agreement
          https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
        
        Illinois:
        -         
        - https://tax.illinois.gov/research/taxrates/income.html#IndividualIncome
          https://tax.illinois.gov/content/dam/soi/en/web/tax/forms/incometax/documents/currentyear/individual/il-1040-instr.pdf
          https://tax.illinois.gov/questionsandanswers/12.html
          https://tax.illinois.gov/content/dam/soi/en/web/tax/forms/incometax/documents/currentyear/individual/il-1040-instr.pdf
          https://tax.illinois.gov/research/taxrates/personalproperty.html
          https://tax.illinois.gov/localgovernments/income.html
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Iowa', 'state_occupation': 'Illinois', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def iowa_and_indiana(self, capital_gains_short=0, dependents=0, sources=False, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_long=0, filing_status='single', income=100000, credits_state=0):
        """
        --- Parameters ----
        
        Iowa:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Indiana:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Iowa:
        - https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
          https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
          https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
          https://tax.iowa.gov/iowa-illinois-reciprocal-agreement
          https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
        
        Indiana:
        -         
        - https://www.in.gov/dor/tax-forms/2022-individual-income-tax-forms/
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-32.htm
          https://www.in.gov/dor/files/ib28.pdf
          https://www.in.gov/dor/tax-forms/2022-individual-income-tax-forms/
          https://www.in.gov/dor/files/2022-county-tax-rates-and-codes.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Iowa', 'state_occupation': 'Indiana', "capital_gains_short": capital_gains_short, "dependents": dependents, "sources": sources, "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def iowa_and_kansas(self, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Iowa:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Kansas:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Iowa:
        - https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
          https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
          https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
          https://tax.iowa.gov/iowa-illinois-reciprocal-agreement
          https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
        
        Kansas:
        -         
        - https://www.ksrevenue.gov/taxrates.html
          https://www.ksrevenue.gov/incomebook22.html#0
          https://www.ksrevenue.gov/incomebook22.html#0
          https://www.ksrevenue.gov/incomebook22.html#0
          https://www.ksrevenue.gov/incomebook22.html
          https://www.ksrevenue.gov/incomebook22.html
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Iowa', 'state_occupation': 'Kansas', "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def iowa_and_kentucky(self, county_residence='Other', state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Iowa:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Kentucky:
        


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Iowa:
        - https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
          https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
          https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
          https://tax.iowa.gov/iowa-illinois-reciprocal-agreement
          https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
        
        Kentucky:
        -         
        - https://revenue.ky.gov/News/Pages/Dor-Announces-Updates-to-Individual-Income-Tax-for-2023-Tax-Year.aspx
          https://revenue.ky.gov/Individual/Individual-Income-Tax/Pages/default.aspx
          https://revenue.ky.gov/News/Pages/Dor-Announces-Updates-to-Individual-Income-Tax-for-2023-Tax-Year.aspx
          https://revenue.ky.gov/Individual/Individual-Income-Tax/Pages/default.aspx
          https://revenue.ky.gov/Dor%20Training%20Materials/103%20KAR%2017.140.%20Individual%20income%20tax%20-%20reciprocity%20-%20nonresidents.pdf
          https://revenue.ky.gov/Individual/Individual-Income-Tax/Pages/default.aspx
          https://nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2021/Taxes-21-32.htm?taxmap=true
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Iowa', 'state_occupation': 'Kentucky', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def iowa_and_louisiana(self, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Iowa:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Louisiana:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Iowa:
        - https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
          https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
          https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
          https://tax.iowa.gov/iowa-illinois-reciprocal-agreement
          https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
        
        Louisiana:
        -         
        - https://revenue.louisiana.gov/individualincometax
          https://revenue.louisiana.gov/individualincometax
          https://www.revenue.louisiana.gov/taxforms/6935(11_02)F.pdf
          https://revenue.louisiana.gov/individualincometax
          https://revenue.louisiana.gov/individualincometax
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Iowa', 'state_occupation': 'Louisiana', "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def iowa_and_maine(self, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Iowa:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Maine:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Iowa:
        - https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
          https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
          https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
          https://tax.iowa.gov/iowa-illinois-reciprocal-agreement
          https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
        
        Maine:
        -         
        - https://www.maine.gov/revenue/tax-return-forms/individual-income-tax-2022
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/ind_tax_rate_sched_2022.pdf
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/22_1040me_book_gen_instruc_revisedFeb23.pdf
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/22_1040me_book_gen_instruc_revisedFeb23.pdf
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/22_1040me_book_gen_instruc_revisedFeb23.pdf
          https://www.maine.gov/revenue/tax-return-forms/individual-income-tax-2022
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Iowa', 'state_occupation': 'Maine', "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def iowa_and_maryland(self, county_residence='Other', state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Iowa:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Maryland:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Iowa:
        - https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
          https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
          https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
          https://tax.iowa.gov/iowa-illinois-reciprocal-agreement
          https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
        
        Maryland:
        -         
        - https://www.marylandtaxes.gov/individual/income/tax-info/tax-rates.php
          https://www.marylandtaxes.gov/forms/22_forms/Resident_Booklet.pdf
          https://www.marylandtaxes.gov/forms/Tax_Publications/Administrative_Releases/Income_and_Estate_Tax_Releases/ar_it37.pdf
          https://www.marylandtaxes.gov/individual/index.php
          https://www.marylandtaxes.gov/individual/income/tax-info/tax-rates.php
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Iowa', 'state_occupation': 'Maryland', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def iowa_and_massachusetts(self, county_residence='Other', state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Iowa:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Massachusetts:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Iowa:
        - https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
          https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
          https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
          https://tax.iowa.gov/iowa-illinois-reciprocal-agreement
          https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
        
        Massachusetts:
        -         
        - https://www.mass.gov/service-details/massachusetts-tax-rates
          https://www.mass.gov/service-details/view-massachusetts-personal-income-tax-exemptions
          https://www.mass.gov/service-details/view-massachusetts-personal-income-tax-exemptions
          https://www.mass.gov/service-details/view-child-and-dependent-related-credits
          https://www.mass.gov/service-details/massachusetts-tax-rates
          https://www.mass.gov/personal-income-tax
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Iowa', 'state_occupation': 'Massachusetts', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def iowa_and_michigan(self, county_residence='Other', state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Iowa:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Michigan:
        


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Iowa:
        - https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
          https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
          https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
          https://tax.iowa.gov/iowa-illinois-reciprocal-agreement
          https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
        
        Michigan:
        -         
        - https://www.michigan.gov/taxes/iit/new-developments-for-tax-year-2022
          "https://www.michigan.gov/taxes/-/media/Project/Websites/taxes/Forms/2023/2023-SUW/446_Withholding-Guide_2023.pdf?rev=177ca3d8cb034392bf84e317ed78c320&hash=19C5F42F43A89A91863D229349F48883#:~:text=The%20exemption%20amount%20is%20%245%2C400
         employees%20Michigan%20income%20tax%20return"
          "https://www.michigan.gov/taxes/-/media/Project/Websites/taxes/Forms/2023/2023-SUW/446_Withholding-Guide_2023.pdf?rev=177ca3d8cb034392bf84e317ed78c320&hash=19C5F42F43A89A91863D229349F48883#:~:text=The%20exemption%20amount%20is%20%245%2C400
         employees%20Michigan%20income%20tax%20return"
          https://www.michigan.gov/taxes/questions/iit/accordion/residency/are-my-wages-earned-in-another-state-taxable-in-michigan-if-i-am-a-michigan-resident-1
          https://www.michigan.gov/taxes/-/media/Project/Websites/taxes/Forms/2022/2022-IIT-Forms/BOOK_MI-1040.pdf?rev=4524f1cbbc0243b194ab05a6680be75b&hash=40DFE71E651F7C1D65C31F6CD03D944F
          https://www.michigan.gov/taxes/questions/iit/accordion/general/what-cities-impose-an-income-tax
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Iowa', 'state_occupation': 'Michigan', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def iowa_and_minnesota(self, county_residence='Other', state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Iowa:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Minnesota:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Iowa:
        - https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
          https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
          https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
          https://tax.iowa.gov/iowa-illinois-reciprocal-agreement
          https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
        
        Minnesota:
        -         
        - https://www.revenue.state.mn.us/minnesota-income-tax-rates-and-brackets
          https://www.revenue.state.mn.us/child-and-dependent-care-credit
          https://www.revenue.state.mn.us/dependent-exemptions
          https://www.revenue.state.mn.us/minnesota-standard-deduction
          https://mn.gov/mmb/assets/mwr_form_tcm1059-128581.pdf
          https://www.revenue.state.mn.us/individual-income-tax
          https://www.revenue.state.mn.us/individual-income-tax
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Iowa', 'state_occupation': 'Minnesota', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def iowa_and_mississippi(self, county_residence='Other', state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Iowa:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Mississippi:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Iowa:
        - https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
          https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
          https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
          https://tax.iowa.gov/iowa-illinois-reciprocal-agreement
          https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
        
        Mississippi:
        -         
        - https://www.dor.ms.gov/individual/tax-rates
          https://www.dor.ms.gov/individual/tax-rates
          https://www.dor.ms.gov/individual/tax-rates
          https://www.dor.ms.gov/individual
          https://www.dor.ms.gov/individual
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Iowa', 'state_occupation': 'Mississippi', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def iowa_and_missouri(self, capital_gains_short=0, dependents=0, sources=False, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_long=0, filing_status='single', income=100000, credits_state=0):
        """
        --- Parameters ----
        
        Iowa:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Missouri:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Iowa:
        - https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
          https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
          https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
          https://tax.iowa.gov/iowa-illinois-reciprocal-agreement
          https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
        
        Missouri:
        -         
        - https://dor.mo.gov/taxation/individual/tax-types/income/year-changes/
          https://dor.mo.gov/taxation/individual/tax-types/income/year-changes/
          https://dor.mo.gov/taxation/individual/tax-types/income/year-changes/
          https://www.kcmo.gov/city-hall/departments/finance/earnings-tax
          https://www.stlouis-mo.gov/government/departments/collector/earnings-tax/individual-earnings-tax-info.cfm
        """
        
        # Check if the parameters are set upstream in the instance
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Iowa', 'state_occupation': 'Missouri', "capital_gains_short": capital_gains_short, "dependents": dependents, "sources": sources, "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def iowa_and_montana(self, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Iowa:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Montana:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Iowa:
        - https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
          https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
          https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
          https://tax.iowa.gov/iowa-illinois-reciprocal-agreement
          https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
        
        Montana:
        -         
        - https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/2022/12/montana-employees-withholding-allowance-and-exemption-certificate-form-mw-4.pdf
          https://mtrevenue.gov/2021/11/29/simplification-of-montana-income-taxation/?utm_source=rss&utm_medium=rss&utm_campaign=simplification-of-montana-income-taxation
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Iowa', 'state_occupation': 'Montana', "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def iowa_and_nebraska(self, county_residence='Other', state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Iowa:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Nebraska:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Iowa:
        - https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
          https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
          https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
          https://tax.iowa.gov/iowa-illinois-reciprocal-agreement
          https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
        
        Nebraska:
        -         
        - https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2023_Circular_EN_8-429r2023_Whole_Book_11.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Iowa', 'state_occupation': 'Nebraska', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def iowa_and_nevada(self, county_residence='Other', state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Iowa:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Nevada:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Iowa:
        - https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
          https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
          https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
          https://tax.iowa.gov/iowa-illinois-reciprocal-agreement
          https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
        
        Nevada:
        -         
        - https://www.nvsos.gov/sos/businesses/the-nevada-advantage
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Iowa', 'state_occupation': 'Nevada', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def iowa_and_new_hampshire(self, county_residence='Other', state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Iowa:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        New Hampshire:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Iowa:
        - https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
          https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
          https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
          https://tax.iowa.gov/iowa-illinois-reciprocal-agreement
          https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
        
        New Hampshire:
        -         
        - https://www.revenue.nh.gov/assistance/tax-overview.htm
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Iowa', 'state_occupation': 'NewHampshire', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def iowa_and_new_jersey(self, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Iowa:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        New Jersey:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Iowa:
        - https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
          https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
          https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
          https://tax.iowa.gov/iowa-illinois-reciprocal-agreement
          https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
        
        New Jersey:
        -         
        - https://www.state.nj.us/treasury/taxation/taxtables.shtml
          https://www.state.nj.us/treasury/taxation/njit13.shtml
          https://www.nj.gov/treasury/taxation/njit25.shtml
          https://www.state.nj.us/treasury/taxation/individuals/freqqiti.shtml
          https://www.state.nj.us/treasury/taxation/localtax.shtml
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Iowa', 'state_occupation': 'NewJersey', "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def iowa_and_new_mexico(self, county_residence='Other', state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Iowa:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        New Mexico:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Iowa:
        - https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
          https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
          https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
          https://tax.iowa.gov/iowa-illinois-reciprocal-agreement
          https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
        
        New Mexico:
        -         
        - https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/fdf3c548-8aba-4b9c-9eb4-bb564c716015/FYI-104.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/0bf943fd-652e-4400-bb1b-cd8397eb5a95/2022pit-adj.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/2f1a6781-9534-4436-b427-1557f9592099/2022pit-adj-ins.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/0bf943fd-652e-4400-bb1b-cd8397eb5a95/2022pit-adj.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/1afc56af-ea90-4d48-82e5-1f9aeb43255a/PITbook2022.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/fdf3c548-8aba-4b9c-9eb4-bb564c716015/FYI-104.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Iowa', 'state_occupation': 'NewMexico', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def iowa_and_new_york(self, county_residence='Other', state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Iowa:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        New York:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Iowa:
        - https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
          https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
          https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
          https://tax.iowa.gov/iowa-illinois-reciprocal-agreement
          https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
        
        New York:
        -         
        - https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nys-tax-rate-schedule
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nys-tax-rate-schedule
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nys-tax-rate-schedule
          https://www.tax.ny.gov/pit/
          https://www.tax.ny.gov/pit/file/nonresident-faqs.htm
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#yonkers-tax-rate-schedule
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nyc-tax-rate-schedule
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Iowa', 'state_occupation': 'NewYork', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def iowa_and_north_carolina(self, county_residence='Other', state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Iowa:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        North Carolina:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Iowa:
        - https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
          https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
          https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
          https://tax.iowa.gov/iowa-illinois-reciprocal-agreement
          https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
        
        North Carolina:
        -         
        - https://www.ncdor.gov/taxes-forms/tax-rate-schedules
          https://www.ncdor.gov/taxes-forms/individual-income-tax/north-carolina-standard-deduction-or-north-carolina-itemized-deductions
          https://www.ncdor.gov/taxes-forms/individual-income-tax
          https://www.ncdor.gov/taxes-forms/individual-income-tax
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Iowa', 'state_occupation': 'NorthCarolina', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def iowa_and_north_dakota(self, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Iowa:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        North Dakota:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Iowa:
        - https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
          https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
          https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
          https://tax.iowa.gov/iowa-illinois-reciprocal-agreement
          https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
        
        North Dakota:
        -         
        - https://www.tax.nd.gov/forms
          https://www.tax.nd.gov/sites/www/files/documents/forms/individual/2022-iit/2022-individual-income-tax-booklet.pdf
          https://www.tax.nd.gov/sites/www/files/documents/forms/2022-individual-income-tax-booklet.pdf
          https://www.tax.nd.gov/tax-exemptions-credits/income-tax-exemptions-credits
          https://www.tax.nd.gov/sites/www/files/documents/forms/business/ndwrfillable.pdf
          https://www.tax.nd.gov/news/tax-legislative-changes/significant-changes-law/individual-income-tax-history
          https://www.tax.nd.gov/forms
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Iowa', 'state_occupation': 'NorthDakota', "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def iowa_and_ohio(self, county_residence='Other', state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Iowa:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Ohio:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Iowa:
        - https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
          https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
          https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
          https://tax.iowa.gov/iowa-illinois-reciprocal-agreement
          https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
        
        Ohio:
        -         
        - https://tax.ohio.gov/individual/resources/annual-tax-rates
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://tax.ohio.gov/static/forms/employer_withholding/generic/wth_it4nr.pdf
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://www.ritaohio.com/TaxRatesTable
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Iowa', 'state_occupation': 'Ohio', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def iowa_and_oklahoma(self, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Iowa:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Oklahoma:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Iowa:
        - https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
          https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
          https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
          https://tax.iowa.gov/iowa-illinois-reciprocal-agreement
          https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
        
        Oklahoma:
        -         
        - https://oklahoma.gov/content/dam/ok/en/tax/documents/resources/publications/businesses/withholding-tables/WHTables-2022.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Iowa', 'state_occupation': 'Oklahoma', "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def iowa_and_oregon(self, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Iowa:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Oregon:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Iowa:
        - https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
          https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
          https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
          https://tax.iowa.gov/iowa-illinois-reciprocal-agreement
          https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
        
        Oregon:
        -         
        - https://www.oregon.gov/dor/programs/individuals/Pages/PIT.aspx
          https://www.oregon.gov/dor/programs/individuals/Documents/Full%20year%20resident
         %20Form%20or-40%20filers.pdf
          https://www.oregon.gov/dor/programs/individuals/Documents/Part-year%20and%20nonresident
         %20Form%20or-40-P%20and%20or-40-N%20filers.pdf.pdf
          https://www.oregon.gov/dor/forms/FormsPubs/form-or-40_101-040_2022.pdf
          https://www.oregon.gov/dor/forms/FormsPubs/form-or-40_101-040_2022.pdf
          https://www.oregon.gov/dor/forms/FormsPubs/form-or-40_101-040_2022.pdf
          https://www.portland.gov/revenue/personal-tax
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Iowa', 'state_occupation': 'Oregon', "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def iowa_and_pennsylvania(self, county_residence='Other', state_agi=75000, county_occupation='Other', capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Iowa:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Pennsylvania:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        county_residence : str, optional
            County of residence. Default is Other.
        county_occupation : str, optional
            County of occupation. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Iowa:
        - https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
          https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
          https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
          https://tax.iowa.gov/iowa-illinois-reciprocal-agreement
          https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
        
        Pennsylvania:
        -         
        - https://www.revenue.pa.gov/Tax%20Rates/Pages/default.aspx
          https://www.revenue.pa.gov/FormsandPublications/FormsforIndividuals/PIT/Documents/2022/2022_pa-40in.pdf
          https://www.revenue.pa.gov/FormsandPublications/FormsforBusinesses/EmployerWithholding/Documents/rev-419.pdf
          https://www.revenue.pa.gov/FormsandPublications/FormsforIndividuals/PIT/Documents/2022/2022_pa-40in.pdf
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-25.htm
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        county_occupation = self.county_occupation if hasattr(self, 'county_occupation') else county_occupation
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Iowa', 'state_occupation': 'Pennsylvania', "county_residence": county_residence, "state_agi": state_agi, "county_occupation": county_occupation, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def iowa_and_rhode_island(self, county_residence='Other', state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Iowa:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Rhode Island:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Iowa:
        - https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
          https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
          https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
          https://tax.iowa.gov/iowa-illinois-reciprocal-agreement
          https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
        
        Rhode Island:
        -         
        - https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2022_1040WE_w_0.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Iowa', 'state_occupation': 'RhodeIsland', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def iowa_and_south_carolina(self, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Iowa:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        South Carolina:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Iowa:
        - https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
          https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
          https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
          https://tax.iowa.gov/iowa-illinois-reciprocal-agreement
          https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
        
        South Carolina:
        -         
        - https://dor.sc.gov/resources-site/lawandpolicy/Advisory%20Opinions/IL22-15.pdf
          https://dor.sc.gov/forms-site/Forms/IITPacket_2022.pdf
          https://www.scstatehouse.gov/code/t12c006.php
          https://dor.sc.gov/resources-site/lawandpolicy/Advisory%20Opinions/IL22-15.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Iowa', 'state_occupation': 'SouthCarolina', "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def iowa_and_south_dakota(self, county_residence='Other', state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Iowa:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        South Dakota:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Iowa:
        - https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
          https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
          https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
          https://tax.iowa.gov/iowa-illinois-reciprocal-agreement
          https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
        
        South Dakota:
        -         
        - https://dor.sd.gov/individuals/taxes/
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Iowa', 'state_occupation': 'SouthDakota', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def iowa_and_tennessee(self, county_residence='Other', state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Iowa:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Tennessee:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Iowa:
        - https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
          https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
          https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
          https://tax.iowa.gov/iowa-illinois-reciprocal-agreement
          https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
        
        Tennessee:
        -         
        - https://www.tn.gov/revenue/taxes.html
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Iowa', 'state_occupation': 'Tennessee', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def iowa_and_texas(self, county_residence='Other', state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Iowa:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Texas:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Iowa:
        - https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
          https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
          https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
          https://tax.iowa.gov/iowa-illinois-reciprocal-agreement
          https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
        
        Texas:
        -         
        - https://comptroller.texas.gov/taxes/
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Iowa', 'state_occupation': 'Texas', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def iowa_and_utah(self, county_residence='Other', state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Iowa:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Utah:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Iowa:
        - https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
          https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
          https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
          https://tax.iowa.gov/iowa-illinois-reciprocal-agreement
          https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
        
        Utah:
        -         
        - https://incometax.utah.gov/paying/tax-rates
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Iowa', 'state_occupation': 'Utah', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def iowa_and_vermont(self, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Iowa:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Vermont:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Iowa:
        - https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
          https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
          https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
          https://tax.iowa.gov/iowa-illinois-reciprocal-agreement
          https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
        
        Vermont:
        -         
        - https://tax.vermont.gov/sites/tax/files/documents/GB-1210-2023.pdf
          https://tax.vermont.gov/sites/tax/files/documents/Income%20Booklet-2022.pdf
          https://tax.vermont.gov/sites/tax/files/documents/Income%20Booklet-2022.pdf
          https://tax.vermont.gov/sites/tax/files/documents/Income%20Booklet-2022.pdf
          https://tax.vermont.gov/individuals/personal-income-tax/tax-credits
          https://tax.vermont.gov/individuals/personal-income-tax/taxable-income
          https://tax.vermont.gov/individuals/personal-income-tax/local-option-tax
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Iowa', 'state_occupation': 'Vermont', "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def iowa_and_virginia(self, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Iowa:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Virginia:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Iowa:
        - https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
          https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
          https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
          https://tax.iowa.gov/iowa-illinois-reciprocal-agreement
          https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
        
        Virginia:
        -         
        - https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/reciprocity
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Iowa', 'state_occupation': 'Virginia', "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def iowa_and_washington(self, county_residence='Other', state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Iowa:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Washington:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Iowa:
        - https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
          https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
          https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
          https://tax.iowa.gov/iowa-illinois-reciprocal-agreement
          https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
        
        Washington:
        -         
        - https://dor.wa.gov/taxes-rates/income-tax
          https://dor.wa.gov/taxes-rates/other-taxes/capital-gains-tax
          https://dor.wa.gov/taxes-rates/income-tax
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Iowa', 'state_occupation': 'Washington', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def iowa_and_west_virginia(self, pay_periods=24, capital_gains_short=0, county_occupation='Other', dependents=0, sources=False, county_residence='Other', state_agi=75000, capital_gains_long=0, filing_status='single', income=100000, credits_state=0):
        """
        --- Parameters ----
        
        Iowa:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        West Virginia:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_occupation : str, optional
            County of occupation. Default is Other.
        pay_periods : int, optional
            Number of pay periods. Options are 13, 24, 26, 52. Default is 24.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Iowa:
        - https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
          https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
          https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
          https://tax.iowa.gov/iowa-illinois-reciprocal-agreement
          https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
        
        West Virginia:
        -         
        - https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://tax.wv.gov/Documents/TSD/tsd438.pdf
          https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2011/Taxes-11-35att.html
        """
        
        # Check if the parameters are set upstream in the instance
        pay_periods = self.pay_periods if hasattr(self, 'pay_periods') else pay_periods
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        county_occupation = self.county_occupation if hasattr(self, 'county_occupation') else county_occupation
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Iowa', 'state_occupation': 'WestVirginia', "pay_periods": pay_periods, "capital_gains_short": capital_gains_short, "county_occupation": county_occupation, "dependents": dependents, "sources": sources, "county_residence": county_residence, "state_agi": state_agi, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def iowa_and_wisconsin(self, county_residence='Other', state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Iowa:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Wisconsin:
        


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Iowa:
        - https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
          https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
          https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
          https://tax.iowa.gov/iowa-illinois-reciprocal-agreement
          https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
        
        Wisconsin:
        -         
        - https://www.revenue.wi.gov/TaxForms2023/2023-Form1-ES-Inst.pdf
          https://www.revenue.wi.gov/TaxForms2023/2023-Form1-ES-Inst.pdf
          https://www.revenue.wi.gov/TaxForms2023/2023-Form1-ES-Inst.pdf
          https://www.revenue.wi.gov/Pages/FAQS/pcs-work.aspx
          https://www.revenue.wi.gov/Dor%20Publications/pb103.pdf
          https://www.revenue.wi.gov/Pages/FAQS/pcs-work.aspx
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Iowa', 'state_occupation': 'Wisconsin', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def iowa_and_wyoming(self, county_residence='Other', state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Iowa:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Wyoming:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Iowa:
        - https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
          https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
          https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
          https://tax.iowa.gov/iowa-illinois-reciprocal-agreement
          https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
        
        Wyoming:
        -         
        - https://revenue.wyo.gov/
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Iowa', 'state_occupation': 'Wyoming', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def kansas_and_alabama(self, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Kansas:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Alabama:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Kansas:
        - https://www.ksrevenue.gov/taxrates.html
          https://www.ksrevenue.gov/incomebook22.html#0
          https://www.ksrevenue.gov/incomebook22.html#0
          https://www.ksrevenue.gov/incomebook22.html#0
          https://www.ksrevenue.gov/incomebook22.html
          https://www.ksrevenue.gov/incomebook22.html
        
        Alabama:
        -         
        - https://www.revenue.alabama.gov/ultraviewer/viewer/basic_viewer/index.html?form=2023/02/22f40.pdf
          https://www.revenue.alabama.gov/faqs/what-is-alabamas-individual-income-tax-rate/
          https://www.revenue.alabama.gov/ultraviewer/viewer/basic_viewer/index.html?form=2023/02/22f40.pdf
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-24.htm?taxmap=true
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-24.htm?taxmap=true
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-24.htm?taxmap=true
          https://www.revenue.alabama.gov/individual-corporate/income-to-be-reported-on-the-alabama-income-tax-return/
          https://www.almonline.org/TaxRates.aspx#OccupationalTax
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Kansas', 'state_occupation': 'Alabama', "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def kansas_and_alaska(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Kansas:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Alaska:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Kansas:
        - https://www.ksrevenue.gov/taxrates.html
          https://www.ksrevenue.gov/incomebook22.html#0
          https://www.ksrevenue.gov/incomebook22.html#0
          https://www.ksrevenue.gov/incomebook22.html#0
          https://www.ksrevenue.gov/incomebook22.html
          https://www.ksrevenue.gov/incomebook22.html
        
        Alaska:
        -         
        - https://www.commerce.alaska.gov/web/dcra/officeofthestateassessor/alaskataxfacts.aspx
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Kansas', 'state_occupation': 'Alaska', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def kansas_and_arizona(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Kansas:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Arizona:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        credits_state : int, optional
            Credits for state taxes. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Kansas:
        - https://www.ksrevenue.gov/taxrates.html
          https://www.ksrevenue.gov/incomebook22.html#0
          https://www.ksrevenue.gov/incomebook22.html#0
          https://www.ksrevenue.gov/incomebook22.html#0
          https://www.ksrevenue.gov/incomebook22.html
          https://www.ksrevenue.gov/incomebook22.html
        
        Arizona:
        -         
        - https://azdor.gov/forms/individual/form-140-x-y-tables
          https://azdor.gov/forms/individual/form-140-x-y-tables
          https://azdor.gov/forms/individual/form-140-x-y-tables
          https://azdor.gov/businesses-arizona/withholding-tax/withholding-exceptions
          "https://azdor.gov/individuals/income-tax-filing-assistance/identifying-other-taxable-income#:~:text=In%20regards%20to%20capital%20gains
         the%20individuals%20regular%20tax%20rate.
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Kansas', 'state_occupation': 'Arizona', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def kansas_and_arkansas(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Kansas:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Arkansas:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Kansas:
        - https://www.ksrevenue.gov/taxrates.html
          https://www.ksrevenue.gov/incomebook22.html#0
          https://www.ksrevenue.gov/incomebook22.html#0
          https://www.ksrevenue.gov/incomebook22.html#0
          https://www.ksrevenue.gov/incomebook22.html
          https://www.ksrevenue.gov/incomebook22.html
        
        Arkansas:
        -         
        - https://www.arkansasedc.com/why-arkansas/business-climate/tax-structure/personal-income-tax
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_and_AR1000NR_Instructions.pdf
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_FullYearResidentIndividualIncomeTaxReturn.pdf
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_and_AR1000NR_Instructions.pdf
          https://www.arkansasedc.com/why-arkansas/business-climate/tax-structure/capital-gains-tax-reduction
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_and_AR1000NR_Instructions.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Kansas', 'state_occupation': 'Arkansas', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def kansas_and_california(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Kansas:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        California:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Kansas:
        - https://www.ksrevenue.gov/taxrates.html
          https://www.ksrevenue.gov/incomebook22.html#0
          https://www.ksrevenue.gov/incomebook22.html#0
          https://www.ksrevenue.gov/incomebook22.html#0
          https://www.ksrevenue.gov/incomebook22.html
          https://www.ksrevenue.gov/incomebook22.html
        
        California:
        -         
        - https://www.ftb.ca.gov/forms/2022/2022-540-tax-rate-schedules.pdf
          https://www.ftb.ca.gov/file/personal/deductions/index.html
          https://www.ftb.ca.gov/about-ftb/newsroom/caleitc/eligibility-and-credit-information.html
          https://www.ftb.ca.gov/file/personal/income-types/capital-gains-and-losses.html#:~:text=California%20does%20not%20have%20a
         are%20taxed%20as%20ordinary%20income
          https://www.ftb.ca.gov/file/personal/
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Kansas', 'state_occupation': 'California', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def kansas_and_colorado(self, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Kansas:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Colorado:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Kansas:
        - https://www.ksrevenue.gov/taxrates.html
          https://www.ksrevenue.gov/incomebook22.html#0
          https://www.ksrevenue.gov/incomebook22.html#0
          https://www.ksrevenue.gov/incomebook22.html#0
          https://www.ksrevenue.gov/incomebook22.html
          https://www.ksrevenue.gov/incomebook22.html
        
        Colorado:
        -         
        - https://leg.colorado.gov/agencies/legislative-council-staff/individual-income-tax%C2%A0
          https://leg.colorado.gov/agencies/legislative-council-staff/individual-income-tax%C2%A0
          https://tax.colorado.gov/sites/tax/files/documents/ITT_Capital_Gain_Subtraction_Feb_2022.pdf
          https://www.auroragov.org/business_services/taxes/occupational_privilege_tax
          https://www.denvergov.org/Government/Agencies-Departments-Offices/Agencies-Departments-Offices-Directory/Department-of-Finance/Our-Divisions/Treasury/Business-Tax-Information#:~:text=Occupational%20privilege%20tax
         -%28show%20below%29&text=Each%20taxable%20employee%20is%20liable
         month%20for%20each%20taxable%20employee
          https://www.glendale.co.us/355/Occupational-Privilege-Tax
          https://www.greenwoodvillage.com/1220/Occupational-Privilege-Tax-OPT
          https://ci.sheridan.co.us/288/Occupational-Privilege-Tax
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Kansas', 'state_occupation': 'Colorado', "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def kansas_and_connecticut(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Kansas:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Connecticut:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Kansas:
        - https://www.ksrevenue.gov/taxrates.html
          https://www.ksrevenue.gov/incomebook22.html#0
          https://www.ksrevenue.gov/incomebook22.html#0
          https://www.ksrevenue.gov/incomebook22.html#0
          https://www.ksrevenue.gov/incomebook22.html
          https://www.ksrevenue.gov/incomebook22.html
        
        Connecticut:
        -         
        - https://portal.ct.gov/-/media/DRS/Forms/2022/Income/2022-CT-1040-Instructions_1222.pdf
          https://www.cga.ct.gov/2022/rpt/pdf/2022-R-0108.pdf
          https://www.cga.ct.gov/current/pub/chap_229.htm#sec_12-703
          https://portal.ct.gov/-/media/DRS/Forms/2022/Income/2022-CT-1040-Instructions_1222.pdf
          "https://portal.ct.gov/DRS/Publications/TSSNs/TSSN-29#:~:text=An%20individuals%20net%20capital%20gains
         portion%20of%20Social%20Security%20benefits"
          "https://portal.ct.gov/DRS/Publications/TSSNs/TSSN-29#:~:text=An%20individuals%20net%20capital%20gains
         portion%20of%20Social%20Security%20benefits"
          https://portal.ct.gov/DRS/Individuals/Resident-Income-Tax/Tax-Information
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Kansas', 'state_occupation': 'Connecticut', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def kansas_and_delaware(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, federal_agi=75000, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Kansas:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Delaware:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        federal_agi : int, optional
            Adjusted Gross Income (AGI) for federal taxes. Default is 0. Sources - https://www.irs.gov/e-file-providers/definition-of-adjusted-gross-income, https://www.irs.gov/newsroom/taxpayers-should-know-the-difference-between-standard-and-itemized-deductions
        income : int, optional
            Gross income. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Kansas:
        - https://www.ksrevenue.gov/taxrates.html
          https://www.ksrevenue.gov/incomebook22.html#0
          https://www.ksrevenue.gov/incomebook22.html#0
          https://www.ksrevenue.gov/incomebook22.html#0
          https://www.ksrevenue.gov/incomebook22.html
          https://www.ksrevenue.gov/incomebook22.html
        
        Delaware:
        -         
        - https://revenue.delaware.gov/software-developer/tax-rate-changes/
          https://revenuefiles.delaware.gov/2022/PIT-RES_TY22_2022-01_PaperInteractive.pdf
          https://revenuefiles.delaware.gov/2022/PIT-RES_TY22_2022-02_Instructions.pdf
          https://revenue.delaware.gov/frequently-asked-questions/personal-income-tax-faqs/
          https://www.newcastlede.gov/DocumentCenter/View/20147/City-of-Wilmington-Wage-Tax-Information-PDF
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        federal_agi = self.federal_agi if hasattr(self, 'federal_agi') else federal_agi
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Kansas', 'state_occupation': 'Delaware', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "federal_agi": federal_agi, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def kansas_and_florida(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Kansas:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Florida:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Kansas:
        - https://www.ksrevenue.gov/taxrates.html
          https://www.ksrevenue.gov/incomebook22.html#0
          https://www.ksrevenue.gov/incomebook22.html#0
          https://www.ksrevenue.gov/incomebook22.html#0
          https://www.ksrevenue.gov/incomebook22.html
          https://www.ksrevenue.gov/incomebook22.html
        
        Florida:
        -         
        - https://floridarevenue.com/taxes/taxesfees/Pages/tax_interest_rates.aspx
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Kansas', 'state_occupation': 'Florida', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def kansas_and_georgia(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Kansas:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Georgia:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Kansas:
        - https://www.ksrevenue.gov/taxrates.html
          https://www.ksrevenue.gov/incomebook22.html#0
          https://www.ksrevenue.gov/incomebook22.html#0
          https://www.ksrevenue.gov/incomebook22.html#0
          https://www.ksrevenue.gov/incomebook22.html
          https://www.ksrevenue.gov/incomebook22.html
        
        Georgia:
        -         
        - https://dor.georgia.gov/tax-tables-georgia-tax-rate-schedule
          https://apps.dor.ga.gov/FillableForms/PDFViewer/Index?form=2022GA500
          https://apps.dor.ga.gov/FillableForms/PDFViewer/Index?form=2022GA500
          https://apps.dor.ga.gov/FillableForms/PDFViewer/Index?form=2022GA500
          https://dor.georgia.gov/years-individual-income-tax-forms
          https://dor.georgia.gov/years-individual-income-tax-forms
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Kansas', 'state_occupation': 'Georgia', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def kansas_and_hawaii(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Kansas:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Hawaii:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Kansas:
        - https://www.ksrevenue.gov/taxrates.html
          https://www.ksrevenue.gov/incomebook22.html#0
          https://www.ksrevenue.gov/incomebook22.html#0
          https://www.ksrevenue.gov/incomebook22.html#0
          https://www.ksrevenue.gov/incomebook22.html
          https://www.ksrevenue.gov/incomebook22.html
        
        Hawaii:
        -         
        - https://tax.hawaii.gov/forms/d_18table-on/d_18table-on_p13/
          https://files.hawaii.gov/tax/forms/2018/18table-on.pdf
          https://files.hawaii.gov/tax/news/pubs/22outline.pdf
          https://files.hawaii.gov/tax/news/pubs/FastTaxRefGuide_2017.pdf
          https://www.tfhawaii.org/wordpress/state-tax-resources/mini-tax-guides/individual-income-tax-chapter-235/#:~:text=Long%20term%20capital%20gains%20are%20taxed%20at%20a%20maximum%20of%207.25%25
          https://www.grassrootinstitute.org/2023/02/hb337-increasing-tax-on-capital-gains-a-bad-risk/
          https://files.hawaii.gov/tax/news/pubs/22outline.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Kansas', 'state_occupation': 'Hawaii', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def kansas_and_idaho(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Kansas:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Idaho:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Kansas:
        - https://www.ksrevenue.gov/taxrates.html
          https://www.ksrevenue.gov/incomebook22.html#0
          https://www.ksrevenue.gov/incomebook22.html#0
          https://www.ksrevenue.gov/incomebook22.html#0
          https://www.ksrevenue.gov/incomebook22.html
          https://www.ksrevenue.gov/incomebook22.html
        
        Idaho:
        -         
        - https://tax.idaho.gov/taxes/income-tax/individual-income/individual-income-tax-rate-schedule/
          https://tax.idaho.gov/wp-content/uploads/forms/EIN00046/EIN00046_03-01-2023.pdf
          https://tax.idaho.gov/taxes/income-tax/individual-income/idaho-source-income/
          https://tax.idaho.gov/taxes/income-tax/individual-income/idaho-source-income/
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Kansas', 'state_occupation': 'Idaho', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def kansas_and_illinois(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Kansas:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Illinois:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Kansas:
        - https://www.ksrevenue.gov/taxrates.html
          https://www.ksrevenue.gov/incomebook22.html#0
          https://www.ksrevenue.gov/incomebook22.html#0
          https://www.ksrevenue.gov/incomebook22.html#0
          https://www.ksrevenue.gov/incomebook22.html
          https://www.ksrevenue.gov/incomebook22.html
        
        Illinois:
        -         
        - https://tax.illinois.gov/research/taxrates/income.html#IndividualIncome
          https://tax.illinois.gov/content/dam/soi/en/web/tax/forms/incometax/documents/currentyear/individual/il-1040-instr.pdf
          https://tax.illinois.gov/questionsandanswers/12.html
          https://tax.illinois.gov/content/dam/soi/en/web/tax/forms/incometax/documents/currentyear/individual/il-1040-instr.pdf
          https://tax.illinois.gov/research/taxrates/personalproperty.html
          https://tax.illinois.gov/localgovernments/income.html
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Kansas', 'state_occupation': 'Illinois', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def kansas_and_indiana(self, capital_gains_short=0, dependents=0, sources=False, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_long=0, filing_status='single', income=100000, credits_state=0):
        """
        --- Parameters ----
        
        Kansas:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Indiana:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Kansas:
        - https://www.ksrevenue.gov/taxrates.html
          https://www.ksrevenue.gov/incomebook22.html#0
          https://www.ksrevenue.gov/incomebook22.html#0
          https://www.ksrevenue.gov/incomebook22.html#0
          https://www.ksrevenue.gov/incomebook22.html
          https://www.ksrevenue.gov/incomebook22.html
        
        Indiana:
        -         
        - https://www.in.gov/dor/tax-forms/2022-individual-income-tax-forms/
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-32.htm
          https://www.in.gov/dor/files/ib28.pdf
          https://www.in.gov/dor/tax-forms/2022-individual-income-tax-forms/
          https://www.in.gov/dor/files/2022-county-tax-rates-and-codes.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Kansas', 'state_occupation': 'Indiana', "capital_gains_short": capital_gains_short, "dependents": dependents, "sources": sources, "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def kansas_and_iowa(self, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Kansas:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Iowa:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Kansas:
        - https://www.ksrevenue.gov/taxrates.html
          https://www.ksrevenue.gov/incomebook22.html#0
          https://www.ksrevenue.gov/incomebook22.html#0
          https://www.ksrevenue.gov/incomebook22.html#0
          https://www.ksrevenue.gov/incomebook22.html
          https://www.ksrevenue.gov/incomebook22.html
        
        Iowa:
        -         
        - https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
          https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
          https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
          https://tax.iowa.gov/iowa-illinois-reciprocal-agreement
          https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Kansas', 'state_occupation': 'Iowa', "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def kansas_and_kentucky(self, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Kansas:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Kentucky:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Kansas:
        - https://www.ksrevenue.gov/taxrates.html
          https://www.ksrevenue.gov/incomebook22.html#0
          https://www.ksrevenue.gov/incomebook22.html#0
          https://www.ksrevenue.gov/incomebook22.html#0
          https://www.ksrevenue.gov/incomebook22.html
          https://www.ksrevenue.gov/incomebook22.html
        
        Kentucky:
        -         
        - https://revenue.ky.gov/News/Pages/Dor-Announces-Updates-to-Individual-Income-Tax-for-2023-Tax-Year.aspx
          https://revenue.ky.gov/Individual/Individual-Income-Tax/Pages/default.aspx
          https://revenue.ky.gov/News/Pages/Dor-Announces-Updates-to-Individual-Income-Tax-for-2023-Tax-Year.aspx
          https://revenue.ky.gov/Individual/Individual-Income-Tax/Pages/default.aspx
          https://revenue.ky.gov/Dor%20Training%20Materials/103%20KAR%2017.140.%20Individual%20income%20tax%20-%20reciprocity%20-%20nonresidents.pdf
          https://revenue.ky.gov/Individual/Individual-Income-Tax/Pages/default.aspx
          https://nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2021/Taxes-21-32.htm?taxmap=true
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Kansas', 'state_occupation': 'Kentucky', "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def kansas_and_louisiana(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Kansas:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Louisiana:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Kansas:
        - https://www.ksrevenue.gov/taxrates.html
          https://www.ksrevenue.gov/incomebook22.html#0
          https://www.ksrevenue.gov/incomebook22.html#0
          https://www.ksrevenue.gov/incomebook22.html#0
          https://www.ksrevenue.gov/incomebook22.html
          https://www.ksrevenue.gov/incomebook22.html
        
        Louisiana:
        -         
        - https://revenue.louisiana.gov/individualincometax
          https://revenue.louisiana.gov/individualincometax
          https://www.revenue.louisiana.gov/taxforms/6935(11_02)F.pdf
          https://revenue.louisiana.gov/individualincometax
          https://revenue.louisiana.gov/individualincometax
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Kansas', 'state_occupation': 'Louisiana', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def kansas_and_maine(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Kansas:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Maine:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Kansas:
        - https://www.ksrevenue.gov/taxrates.html
          https://www.ksrevenue.gov/incomebook22.html#0
          https://www.ksrevenue.gov/incomebook22.html#0
          https://www.ksrevenue.gov/incomebook22.html#0
          https://www.ksrevenue.gov/incomebook22.html
          https://www.ksrevenue.gov/incomebook22.html
        
        Maine:
        -         
        - https://www.maine.gov/revenue/tax-return-forms/individual-income-tax-2022
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/ind_tax_rate_sched_2022.pdf
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/22_1040me_book_gen_instruc_revisedFeb23.pdf
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/22_1040me_book_gen_instruc_revisedFeb23.pdf
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/22_1040me_book_gen_instruc_revisedFeb23.pdf
          https://www.maine.gov/revenue/tax-return-forms/individual-income-tax-2022
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Kansas', 'state_occupation': 'Maine', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def kansas_and_maryland(self, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Kansas:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Maryland:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Kansas:
        - https://www.ksrevenue.gov/taxrates.html
          https://www.ksrevenue.gov/incomebook22.html#0
          https://www.ksrevenue.gov/incomebook22.html#0
          https://www.ksrevenue.gov/incomebook22.html#0
          https://www.ksrevenue.gov/incomebook22.html
          https://www.ksrevenue.gov/incomebook22.html
        
        Maryland:
        -         
        - https://www.marylandtaxes.gov/individual/income/tax-info/tax-rates.php
          https://www.marylandtaxes.gov/forms/22_forms/Resident_Booklet.pdf
          https://www.marylandtaxes.gov/forms/Tax_Publications/Administrative_Releases/Income_and_Estate_Tax_Releases/ar_it37.pdf
          https://www.marylandtaxes.gov/individual/index.php
          https://www.marylandtaxes.gov/individual/income/tax-info/tax-rates.php
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Kansas', 'state_occupation': 'Maryland', "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def kansas_and_massachusetts(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Kansas:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Massachusetts:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Kansas:
        - https://www.ksrevenue.gov/taxrates.html
          https://www.ksrevenue.gov/incomebook22.html#0
          https://www.ksrevenue.gov/incomebook22.html#0
          https://www.ksrevenue.gov/incomebook22.html#0
          https://www.ksrevenue.gov/incomebook22.html
          https://www.ksrevenue.gov/incomebook22.html
        
        Massachusetts:
        -         
        - https://www.mass.gov/service-details/massachusetts-tax-rates
          https://www.mass.gov/service-details/view-massachusetts-personal-income-tax-exemptions
          https://www.mass.gov/service-details/view-massachusetts-personal-income-tax-exemptions
          https://www.mass.gov/service-details/view-child-and-dependent-related-credits
          https://www.mass.gov/service-details/massachusetts-tax-rates
          https://www.mass.gov/personal-income-tax
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Kansas', 'state_occupation': 'Massachusetts', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def kansas_and_michigan(self, capital_gains_short=0, county_occupation='Other', dependents=0, sources=False, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_long=0, filing_status='single', income=100000, credits_state=0):
        """
        --- Parameters ----
        
        Kansas:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Michigan:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        county_occupation : str, optional
            County of occupation. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Kansas:
        - https://www.ksrevenue.gov/taxrates.html
          https://www.ksrevenue.gov/incomebook22.html#0
          https://www.ksrevenue.gov/incomebook22.html#0
          https://www.ksrevenue.gov/incomebook22.html#0
          https://www.ksrevenue.gov/incomebook22.html
          https://www.ksrevenue.gov/incomebook22.html
        
        Michigan:
        -         
        - https://www.michigan.gov/taxes/iit/new-developments-for-tax-year-2022
          "https://www.michigan.gov/taxes/-/media/Project/Websites/taxes/Forms/2023/2023-SUW/446_Withholding-Guide_2023.pdf?rev=177ca3d8cb034392bf84e317ed78c320&hash=19C5F42F43A89A91863D229349F48883#:~:text=The%20exemption%20amount%20is%20%245%2C400
         employees%20Michigan%20income%20tax%20return"
          "https://www.michigan.gov/taxes/-/media/Project/Websites/taxes/Forms/2023/2023-SUW/446_Withholding-Guide_2023.pdf?rev=177ca3d8cb034392bf84e317ed78c320&hash=19C5F42F43A89A91863D229349F48883#:~:text=The%20exemption%20amount%20is%20%245%2C400
         employees%20Michigan%20income%20tax%20return"
          https://www.michigan.gov/taxes/questions/iit/accordion/residency/are-my-wages-earned-in-another-state-taxable-in-michigan-if-i-am-a-michigan-resident-1
          https://www.michigan.gov/taxes/-/media/Project/Websites/taxes/Forms/2022/2022-IIT-Forms/BOOK_MI-1040.pdf?rev=4524f1cbbc0243b194ab05a6680be75b&hash=40DFE71E651F7C1D65C31F6CD03D944F
          https://www.michigan.gov/taxes/questions/iit/accordion/general/what-cities-impose-an-income-tax
        """
        
        # Check if the parameters are set upstream in the instance
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        county_occupation = self.county_occupation if hasattr(self, 'county_occupation') else county_occupation
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Kansas', 'state_occupation': 'Michigan', "capital_gains_short": capital_gains_short, "county_occupation": county_occupation, "dependents": dependents, "sources": sources, "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def kansas_and_minnesota(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Kansas:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Minnesota:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Kansas:
        - https://www.ksrevenue.gov/taxrates.html
          https://www.ksrevenue.gov/incomebook22.html#0
          https://www.ksrevenue.gov/incomebook22.html#0
          https://www.ksrevenue.gov/incomebook22.html#0
          https://www.ksrevenue.gov/incomebook22.html
          https://www.ksrevenue.gov/incomebook22.html
        
        Minnesota:
        -         
        - https://www.revenue.state.mn.us/minnesota-income-tax-rates-and-brackets
          https://www.revenue.state.mn.us/child-and-dependent-care-credit
          https://www.revenue.state.mn.us/dependent-exemptions
          https://www.revenue.state.mn.us/minnesota-standard-deduction
          https://mn.gov/mmb/assets/mwr_form_tcm1059-128581.pdf
          https://www.revenue.state.mn.us/individual-income-tax
          https://www.revenue.state.mn.us/individual-income-tax
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Kansas', 'state_occupation': 'Minnesota', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def kansas_and_mississippi(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Kansas:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Mississippi:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Kansas:
        - https://www.ksrevenue.gov/taxrates.html
          https://www.ksrevenue.gov/incomebook22.html#0
          https://www.ksrevenue.gov/incomebook22.html#0
          https://www.ksrevenue.gov/incomebook22.html#0
          https://www.ksrevenue.gov/incomebook22.html
          https://www.ksrevenue.gov/incomebook22.html
        
        Mississippi:
        -         
        - https://www.dor.ms.gov/individual/tax-rates
          https://www.dor.ms.gov/individual/tax-rates
          https://www.dor.ms.gov/individual/tax-rates
          https://www.dor.ms.gov/individual
          https://www.dor.ms.gov/individual
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Kansas', 'state_occupation': 'Mississippi', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def kansas_and_missouri(self, capital_gains_short=0, dependents=0, sources=False, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_long=0, filing_status='single', income=100000, credits_state=0):
        """
        --- Parameters ----
        
        Kansas:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Missouri:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Kansas:
        - https://www.ksrevenue.gov/taxrates.html
          https://www.ksrevenue.gov/incomebook22.html#0
          https://www.ksrevenue.gov/incomebook22.html#0
          https://www.ksrevenue.gov/incomebook22.html#0
          https://www.ksrevenue.gov/incomebook22.html
          https://www.ksrevenue.gov/incomebook22.html
        
        Missouri:
        -         
        - https://dor.mo.gov/taxation/individual/tax-types/income/year-changes/
          https://dor.mo.gov/taxation/individual/tax-types/income/year-changes/
          https://dor.mo.gov/taxation/individual/tax-types/income/year-changes/
          https://www.kcmo.gov/city-hall/departments/finance/earnings-tax
          https://www.stlouis-mo.gov/government/departments/collector/earnings-tax/individual-earnings-tax-info.cfm
        """
        
        # Check if the parameters are set upstream in the instance
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Kansas', 'state_occupation': 'Missouri', "capital_gains_short": capital_gains_short, "dependents": dependents, "sources": sources, "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def kansas_and_montana(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Kansas:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Montana:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Kansas:
        - https://www.ksrevenue.gov/taxrates.html
          https://www.ksrevenue.gov/incomebook22.html#0
          https://www.ksrevenue.gov/incomebook22.html#0
          https://www.ksrevenue.gov/incomebook22.html#0
          https://www.ksrevenue.gov/incomebook22.html
          https://www.ksrevenue.gov/incomebook22.html
        
        Montana:
        -         
        - https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/2022/12/montana-employees-withholding-allowance-and-exemption-certificate-form-mw-4.pdf
          https://mtrevenue.gov/2021/11/29/simplification-of-montana-income-taxation/?utm_source=rss&utm_medium=rss&utm_campaign=simplification-of-montana-income-taxation
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Kansas', 'state_occupation': 'Montana', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def kansas_and_nebraska(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Kansas:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Nebraska:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Kansas:
        - https://www.ksrevenue.gov/taxrates.html
          https://www.ksrevenue.gov/incomebook22.html#0
          https://www.ksrevenue.gov/incomebook22.html#0
          https://www.ksrevenue.gov/incomebook22.html#0
          https://www.ksrevenue.gov/incomebook22.html
          https://www.ksrevenue.gov/incomebook22.html
        
        Nebraska:
        -         
        - https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2023_Circular_EN_8-429r2023_Whole_Book_11.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Kansas', 'state_occupation': 'Nebraska', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def kansas_and_nevada(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Kansas:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Nevada:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Kansas:
        - https://www.ksrevenue.gov/taxrates.html
          https://www.ksrevenue.gov/incomebook22.html#0
          https://www.ksrevenue.gov/incomebook22.html#0
          https://www.ksrevenue.gov/incomebook22.html#0
          https://www.ksrevenue.gov/incomebook22.html
          https://www.ksrevenue.gov/incomebook22.html
        
        Nevada:
        -         
        - https://www.nvsos.gov/sos/businesses/the-nevada-advantage
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Kansas', 'state_occupation': 'Nevada', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def kansas_and_new_hampshire(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Kansas:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        New Hampshire:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Kansas:
        - https://www.ksrevenue.gov/taxrates.html
          https://www.ksrevenue.gov/incomebook22.html#0
          https://www.ksrevenue.gov/incomebook22.html#0
          https://www.ksrevenue.gov/incomebook22.html#0
          https://www.ksrevenue.gov/incomebook22.html
          https://www.ksrevenue.gov/incomebook22.html
        
        New Hampshire:
        -         
        - https://www.revenue.nh.gov/assistance/tax-overview.htm
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Kansas', 'state_occupation': 'NewHampshire', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def kansas_and_new_jersey(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Kansas:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        New Jersey:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Kansas:
        - https://www.ksrevenue.gov/taxrates.html
          https://www.ksrevenue.gov/incomebook22.html#0
          https://www.ksrevenue.gov/incomebook22.html#0
          https://www.ksrevenue.gov/incomebook22.html#0
          https://www.ksrevenue.gov/incomebook22.html
          https://www.ksrevenue.gov/incomebook22.html
        
        New Jersey:
        -         
        - https://www.state.nj.us/treasury/taxation/taxtables.shtml
          https://www.state.nj.us/treasury/taxation/njit13.shtml
          https://www.nj.gov/treasury/taxation/njit25.shtml
          https://www.state.nj.us/treasury/taxation/individuals/freqqiti.shtml
          https://www.state.nj.us/treasury/taxation/localtax.shtml
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Kansas', 'state_occupation': 'NewJersey', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def kansas_and_new_mexico(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Kansas:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        New Mexico:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Kansas:
        - https://www.ksrevenue.gov/taxrates.html
          https://www.ksrevenue.gov/incomebook22.html#0
          https://www.ksrevenue.gov/incomebook22.html#0
          https://www.ksrevenue.gov/incomebook22.html#0
          https://www.ksrevenue.gov/incomebook22.html
          https://www.ksrevenue.gov/incomebook22.html
        
        New Mexico:
        -         
        - https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/fdf3c548-8aba-4b9c-9eb4-bb564c716015/FYI-104.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/0bf943fd-652e-4400-bb1b-cd8397eb5a95/2022pit-adj.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/2f1a6781-9534-4436-b427-1557f9592099/2022pit-adj-ins.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/0bf943fd-652e-4400-bb1b-cd8397eb5a95/2022pit-adj.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/1afc56af-ea90-4d48-82e5-1f9aeb43255a/PITbook2022.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/fdf3c548-8aba-4b9c-9eb4-bb564c716015/FYI-104.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Kansas', 'state_occupation': 'NewMexico', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def kansas_and_new_york(self, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Kansas:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        New York:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Kansas:
        - https://www.ksrevenue.gov/taxrates.html
          https://www.ksrevenue.gov/incomebook22.html#0
          https://www.ksrevenue.gov/incomebook22.html#0
          https://www.ksrevenue.gov/incomebook22.html#0
          https://www.ksrevenue.gov/incomebook22.html
          https://www.ksrevenue.gov/incomebook22.html
        
        New York:
        -         
        - https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nys-tax-rate-schedule
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nys-tax-rate-schedule
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nys-tax-rate-schedule
          https://www.tax.ny.gov/pit/
          https://www.tax.ny.gov/pit/file/nonresident-faqs.htm
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#yonkers-tax-rate-schedule
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nyc-tax-rate-schedule
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Kansas', 'state_occupation': 'NewYork', "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def kansas_and_north_carolina(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Kansas:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        North Carolina:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Kansas:
        - https://www.ksrevenue.gov/taxrates.html
          https://www.ksrevenue.gov/incomebook22.html#0
          https://www.ksrevenue.gov/incomebook22.html#0
          https://www.ksrevenue.gov/incomebook22.html#0
          https://www.ksrevenue.gov/incomebook22.html
          https://www.ksrevenue.gov/incomebook22.html
        
        North Carolina:
        -         
        - https://www.ncdor.gov/taxes-forms/tax-rate-schedules
          https://www.ncdor.gov/taxes-forms/individual-income-tax/north-carolina-standard-deduction-or-north-carolina-itemized-deductions
          https://www.ncdor.gov/taxes-forms/individual-income-tax
          https://www.ncdor.gov/taxes-forms/individual-income-tax
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Kansas', 'state_occupation': 'NorthCarolina', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def kansas_and_north_dakota(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Kansas:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        North Dakota:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Kansas:
        - https://www.ksrevenue.gov/taxrates.html
          https://www.ksrevenue.gov/incomebook22.html#0
          https://www.ksrevenue.gov/incomebook22.html#0
          https://www.ksrevenue.gov/incomebook22.html#0
          https://www.ksrevenue.gov/incomebook22.html
          https://www.ksrevenue.gov/incomebook22.html
        
        North Dakota:
        -         
        - https://www.tax.nd.gov/forms
          https://www.tax.nd.gov/sites/www/files/documents/forms/individual/2022-iit/2022-individual-income-tax-booklet.pdf
          https://www.tax.nd.gov/sites/www/files/documents/forms/2022-individual-income-tax-booklet.pdf
          https://www.tax.nd.gov/tax-exemptions-credits/income-tax-exemptions-credits
          https://www.tax.nd.gov/sites/www/files/documents/forms/business/ndwrfillable.pdf
          https://www.tax.nd.gov/news/tax-legislative-changes/significant-changes-law/individual-income-tax-history
          https://www.tax.nd.gov/forms
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Kansas', 'state_occupation': 'NorthDakota', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def kansas_and_ohio(self, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Kansas:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Ohio:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Kansas:
        - https://www.ksrevenue.gov/taxrates.html
          https://www.ksrevenue.gov/incomebook22.html#0
          https://www.ksrevenue.gov/incomebook22.html#0
          https://www.ksrevenue.gov/incomebook22.html#0
          https://www.ksrevenue.gov/incomebook22.html
          https://www.ksrevenue.gov/incomebook22.html
        
        Ohio:
        -         
        - https://tax.ohio.gov/individual/resources/annual-tax-rates
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://tax.ohio.gov/static/forms/employer_withholding/generic/wth_it4nr.pdf
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://www.ritaohio.com/TaxRatesTable
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Kansas', 'state_occupation': 'Ohio', "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def kansas_and_oklahoma(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Kansas:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Oklahoma:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Kansas:
        - https://www.ksrevenue.gov/taxrates.html
          https://www.ksrevenue.gov/incomebook22.html#0
          https://www.ksrevenue.gov/incomebook22.html#0
          https://www.ksrevenue.gov/incomebook22.html#0
          https://www.ksrevenue.gov/incomebook22.html
          https://www.ksrevenue.gov/incomebook22.html
        
        Oklahoma:
        -         
        - https://oklahoma.gov/content/dam/ok/en/tax/documents/resources/publications/businesses/withholding-tables/WHTables-2022.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Kansas', 'state_occupation': 'Oklahoma', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def kansas_and_oregon(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Kansas:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Oregon:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Kansas:
        - https://www.ksrevenue.gov/taxrates.html
          https://www.ksrevenue.gov/incomebook22.html#0
          https://www.ksrevenue.gov/incomebook22.html#0
          https://www.ksrevenue.gov/incomebook22.html#0
          https://www.ksrevenue.gov/incomebook22.html
          https://www.ksrevenue.gov/incomebook22.html
        
        Oregon:
        -         
        - https://www.oregon.gov/dor/programs/individuals/Pages/PIT.aspx
          https://www.oregon.gov/dor/programs/individuals/Documents/Full%20year%20resident
         %20Form%20or-40%20filers.pdf
          https://www.oregon.gov/dor/programs/individuals/Documents/Part-year%20and%20nonresident
         %20Form%20or-40-P%20and%20or-40-N%20filers.pdf.pdf
          https://www.oregon.gov/dor/forms/FormsPubs/form-or-40_101-040_2022.pdf
          https://www.oregon.gov/dor/forms/FormsPubs/form-or-40_101-040_2022.pdf
          https://www.oregon.gov/dor/forms/FormsPubs/form-or-40_101-040_2022.pdf
          https://www.portland.gov/revenue/personal-tax
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Kansas', 'state_occupation': 'Oregon', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def kansas_and_pennsylvania(self, county_residence='Other', state_agi=75000, county_occupation='Other', filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Kansas:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Pennsylvania:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        county_residence : str, optional
            County of residence. Default is Other.
        county_occupation : str, optional
            County of occupation. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Kansas:
        - https://www.ksrevenue.gov/taxrates.html
          https://www.ksrevenue.gov/incomebook22.html#0
          https://www.ksrevenue.gov/incomebook22.html#0
          https://www.ksrevenue.gov/incomebook22.html#0
          https://www.ksrevenue.gov/incomebook22.html
          https://www.ksrevenue.gov/incomebook22.html
        
        Pennsylvania:
        -         
        - https://www.revenue.pa.gov/Tax%20Rates/Pages/default.aspx
          https://www.revenue.pa.gov/FormsandPublications/FormsforIndividuals/PIT/Documents/2022/2022_pa-40in.pdf
          https://www.revenue.pa.gov/FormsandPublications/FormsforBusinesses/EmployerWithholding/Documents/rev-419.pdf
          https://www.revenue.pa.gov/FormsandPublications/FormsforIndividuals/PIT/Documents/2022/2022_pa-40in.pdf
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-25.htm
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        county_occupation = self.county_occupation if hasattr(self, 'county_occupation') else county_occupation
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Kansas', 'state_occupation': 'Pennsylvania', "county_residence": county_residence, "state_agi": state_agi, "county_occupation": county_occupation, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def kansas_and_rhode_island(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Kansas:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Rhode Island:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Kansas:
        - https://www.ksrevenue.gov/taxrates.html
          https://www.ksrevenue.gov/incomebook22.html#0
          https://www.ksrevenue.gov/incomebook22.html#0
          https://www.ksrevenue.gov/incomebook22.html#0
          https://www.ksrevenue.gov/incomebook22.html
          https://www.ksrevenue.gov/incomebook22.html
        
        Rhode Island:
        -         
        - https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2022_1040WE_w_0.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Kansas', 'state_occupation': 'RhodeIsland', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def kansas_and_south_carolina(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Kansas:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        South Carolina:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Kansas:
        - https://www.ksrevenue.gov/taxrates.html
          https://www.ksrevenue.gov/incomebook22.html#0
          https://www.ksrevenue.gov/incomebook22.html#0
          https://www.ksrevenue.gov/incomebook22.html#0
          https://www.ksrevenue.gov/incomebook22.html
          https://www.ksrevenue.gov/incomebook22.html
        
        South Carolina:
        -         
        - https://dor.sc.gov/resources-site/lawandpolicy/Advisory%20Opinions/IL22-15.pdf
          https://dor.sc.gov/forms-site/Forms/IITPacket_2022.pdf
          https://www.scstatehouse.gov/code/t12c006.php
          https://dor.sc.gov/resources-site/lawandpolicy/Advisory%20Opinions/IL22-15.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Kansas', 'state_occupation': 'SouthCarolina', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def kansas_and_south_dakota(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Kansas:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        South Dakota:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Kansas:
        - https://www.ksrevenue.gov/taxrates.html
          https://www.ksrevenue.gov/incomebook22.html#0
          https://www.ksrevenue.gov/incomebook22.html#0
          https://www.ksrevenue.gov/incomebook22.html#0
          https://www.ksrevenue.gov/incomebook22.html
          https://www.ksrevenue.gov/incomebook22.html
        
        South Dakota:
        -         
        - https://dor.sd.gov/individuals/taxes/
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Kansas', 'state_occupation': 'SouthDakota', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def kansas_and_tennessee(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Kansas:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Tennessee:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Kansas:
        - https://www.ksrevenue.gov/taxrates.html
          https://www.ksrevenue.gov/incomebook22.html#0
          https://www.ksrevenue.gov/incomebook22.html#0
          https://www.ksrevenue.gov/incomebook22.html#0
          https://www.ksrevenue.gov/incomebook22.html
          https://www.ksrevenue.gov/incomebook22.html
        
        Tennessee:
        -         
        - https://www.tn.gov/revenue/taxes.html
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Kansas', 'state_occupation': 'Tennessee', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def kansas_and_texas(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Kansas:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Texas:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Kansas:
        - https://www.ksrevenue.gov/taxrates.html
          https://www.ksrevenue.gov/incomebook22.html#0
          https://www.ksrevenue.gov/incomebook22.html#0
          https://www.ksrevenue.gov/incomebook22.html#0
          https://www.ksrevenue.gov/incomebook22.html
          https://www.ksrevenue.gov/incomebook22.html
        
        Texas:
        -         
        - https://comptroller.texas.gov/taxes/
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Kansas', 'state_occupation': 'Texas', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def kansas_and_utah(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Kansas:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Utah:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Kansas:
        - https://www.ksrevenue.gov/taxrates.html
          https://www.ksrevenue.gov/incomebook22.html#0
          https://www.ksrevenue.gov/incomebook22.html#0
          https://www.ksrevenue.gov/incomebook22.html#0
          https://www.ksrevenue.gov/incomebook22.html
          https://www.ksrevenue.gov/incomebook22.html
        
        Utah:
        -         
        - https://incometax.utah.gov/paying/tax-rates
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Kansas', 'state_occupation': 'Utah', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def kansas_and_vermont(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Kansas:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Vermont:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Kansas:
        - https://www.ksrevenue.gov/taxrates.html
          https://www.ksrevenue.gov/incomebook22.html#0
          https://www.ksrevenue.gov/incomebook22.html#0
          https://www.ksrevenue.gov/incomebook22.html#0
          https://www.ksrevenue.gov/incomebook22.html
          https://www.ksrevenue.gov/incomebook22.html
        
        Vermont:
        -         
        - https://tax.vermont.gov/sites/tax/files/documents/GB-1210-2023.pdf
          https://tax.vermont.gov/sites/tax/files/documents/Income%20Booklet-2022.pdf
          https://tax.vermont.gov/sites/tax/files/documents/Income%20Booklet-2022.pdf
          https://tax.vermont.gov/sites/tax/files/documents/Income%20Booklet-2022.pdf
          https://tax.vermont.gov/individuals/personal-income-tax/tax-credits
          https://tax.vermont.gov/individuals/personal-income-tax/taxable-income
          https://tax.vermont.gov/individuals/personal-income-tax/local-option-tax
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Kansas', 'state_occupation': 'Vermont', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def kansas_and_virginia(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Kansas:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Virginia:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Kansas:
        - https://www.ksrevenue.gov/taxrates.html
          https://www.ksrevenue.gov/incomebook22.html#0
          https://www.ksrevenue.gov/incomebook22.html#0
          https://www.ksrevenue.gov/incomebook22.html#0
          https://www.ksrevenue.gov/incomebook22.html
          https://www.ksrevenue.gov/incomebook22.html
        
        Virginia:
        -         
        - https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/reciprocity
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Kansas', 'state_occupation': 'Virginia', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def kansas_and_washington(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Kansas:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Washington:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Kansas:
        - https://www.ksrevenue.gov/taxrates.html
          https://www.ksrevenue.gov/incomebook22.html#0
          https://www.ksrevenue.gov/incomebook22.html#0
          https://www.ksrevenue.gov/incomebook22.html#0
          https://www.ksrevenue.gov/incomebook22.html
          https://www.ksrevenue.gov/incomebook22.html
        
        Washington:
        -         
        - https://dor.wa.gov/taxes-rates/income-tax
          https://dor.wa.gov/taxes-rates/other-taxes/capital-gains-tax
          https://dor.wa.gov/taxes-rates/income-tax
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Kansas', 'state_occupation': 'Washington', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def kansas_and_west_virginia(self, pay_periods=24, capital_gains_short=0, county_occupation='Other', dependents=0, sources=False, state_agi=75000, filers_over_65=0, capital_gains_long=0, filing_status='single', income=100000, credits_state=0):
        """
        --- Parameters ----
        
        Kansas:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        West Virginia:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_occupation : str, optional
            County of occupation. Default is Other.
        pay_periods : int, optional
            Number of pay periods. Options are 13, 24, 26, 52. Default is 24.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Kansas:
        - https://www.ksrevenue.gov/taxrates.html
          https://www.ksrevenue.gov/incomebook22.html#0
          https://www.ksrevenue.gov/incomebook22.html#0
          https://www.ksrevenue.gov/incomebook22.html#0
          https://www.ksrevenue.gov/incomebook22.html
          https://www.ksrevenue.gov/incomebook22.html
        
        West Virginia:
        -         
        - https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://tax.wv.gov/Documents/TSD/tsd438.pdf
          https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2011/Taxes-11-35att.html
        """
        
        # Check if the parameters are set upstream in the instance
        pay_periods = self.pay_periods if hasattr(self, 'pay_periods') else pay_periods
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        county_occupation = self.county_occupation if hasattr(self, 'county_occupation') else county_occupation
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Kansas', 'state_occupation': 'WestVirginia', "pay_periods": pay_periods, "capital_gains_short": capital_gains_short, "county_occupation": county_occupation, "dependents": dependents, "sources": sources, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def kansas_and_wisconsin(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Kansas:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Wisconsin:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Kansas:
        - https://www.ksrevenue.gov/taxrates.html
          https://www.ksrevenue.gov/incomebook22.html#0
          https://www.ksrevenue.gov/incomebook22.html#0
          https://www.ksrevenue.gov/incomebook22.html#0
          https://www.ksrevenue.gov/incomebook22.html
          https://www.ksrevenue.gov/incomebook22.html
        
        Wisconsin:
        -         
        - https://www.revenue.wi.gov/TaxForms2023/2023-Form1-ES-Inst.pdf
          https://www.revenue.wi.gov/TaxForms2023/2023-Form1-ES-Inst.pdf
          https://www.revenue.wi.gov/TaxForms2023/2023-Form1-ES-Inst.pdf
          https://www.revenue.wi.gov/Pages/FAQS/pcs-work.aspx
          https://www.revenue.wi.gov/Dor%20Publications/pb103.pdf
          https://www.revenue.wi.gov/Pages/FAQS/pcs-work.aspx
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Kansas', 'state_occupation': 'Wisconsin', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def kansas_and_wyoming(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Kansas:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Wyoming:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Kansas:
        - https://www.ksrevenue.gov/taxrates.html
          https://www.ksrevenue.gov/incomebook22.html#0
          https://www.ksrevenue.gov/incomebook22.html#0
          https://www.ksrevenue.gov/incomebook22.html#0
          https://www.ksrevenue.gov/incomebook22.html
          https://www.ksrevenue.gov/incomebook22.html
        
        Wyoming:
        -         
        - https://revenue.wyo.gov/
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Kansas', 'state_occupation': 'Wyoming', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def kentucky_and_alabama(self, county_residence='Other', state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Kentucky:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        
        Alabama:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Kentucky:
        - https://revenue.ky.gov/News/Pages/Dor-Announces-Updates-to-Individual-Income-Tax-for-2023-Tax-Year.aspx
          https://revenue.ky.gov/Individual/Individual-Income-Tax/Pages/default.aspx
          https://revenue.ky.gov/News/Pages/Dor-Announces-Updates-to-Individual-Income-Tax-for-2023-Tax-Year.aspx
          https://revenue.ky.gov/Individual/Individual-Income-Tax/Pages/default.aspx
          https://revenue.ky.gov/Dor%20Training%20Materials/103%20KAR%2017.140.%20Individual%20income%20tax%20-%20reciprocity%20-%20nonresidents.pdf
          https://revenue.ky.gov/Individual/Individual-Income-Tax/Pages/default.aspx
          https://nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2021/Taxes-21-32.htm?taxmap=true
        
        Alabama:
        -         
        - https://www.revenue.alabama.gov/ultraviewer/viewer/basic_viewer/index.html?form=2023/02/22f40.pdf
          https://www.revenue.alabama.gov/faqs/what-is-alabamas-individual-income-tax-rate/
          https://www.revenue.alabama.gov/ultraviewer/viewer/basic_viewer/index.html?form=2023/02/22f40.pdf
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-24.htm?taxmap=true
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-24.htm?taxmap=true
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-24.htm?taxmap=true
          https://www.revenue.alabama.gov/individual-corporate/income-to-be-reported-on-the-alabama-income-tax-return/
          https://www.almonline.org/TaxRates.aspx#OccupationalTax
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Kentucky', 'state_occupation': 'Alabama', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def kentucky_and_alaska(self, county_residence='Other', state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Kentucky:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        
        Alaska:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Kentucky:
        - https://revenue.ky.gov/News/Pages/Dor-Announces-Updates-to-Individual-Income-Tax-for-2023-Tax-Year.aspx
          https://revenue.ky.gov/Individual/Individual-Income-Tax/Pages/default.aspx
          https://revenue.ky.gov/News/Pages/Dor-Announces-Updates-to-Individual-Income-Tax-for-2023-Tax-Year.aspx
          https://revenue.ky.gov/Individual/Individual-Income-Tax/Pages/default.aspx
          https://revenue.ky.gov/Dor%20Training%20Materials/103%20KAR%2017.140.%20Individual%20income%20tax%20-%20reciprocity%20-%20nonresidents.pdf
          https://revenue.ky.gov/Individual/Individual-Income-Tax/Pages/default.aspx
          https://nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2021/Taxes-21-32.htm?taxmap=true
        
        Alaska:
        -         
        - https://www.commerce.alaska.gov/web/dcra/officeofthestateassessor/alaskataxfacts.aspx
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Kentucky', 'state_occupation': 'Alaska', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def kentucky_and_arizona(self, county_residence='Other', state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Kentucky:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        
        Arizona:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        credits_state : int, optional
            Credits for state taxes. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Kentucky:
        - https://revenue.ky.gov/News/Pages/Dor-Announces-Updates-to-Individual-Income-Tax-for-2023-Tax-Year.aspx
          https://revenue.ky.gov/Individual/Individual-Income-Tax/Pages/default.aspx
          https://revenue.ky.gov/News/Pages/Dor-Announces-Updates-to-Individual-Income-Tax-for-2023-Tax-Year.aspx
          https://revenue.ky.gov/Individual/Individual-Income-Tax/Pages/default.aspx
          https://revenue.ky.gov/Dor%20Training%20Materials/103%20KAR%2017.140.%20Individual%20income%20tax%20-%20reciprocity%20-%20nonresidents.pdf
          https://revenue.ky.gov/Individual/Individual-Income-Tax/Pages/default.aspx
          https://nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2021/Taxes-21-32.htm?taxmap=true
        
        Arizona:
        -         
        - https://azdor.gov/forms/individual/form-140-x-y-tables
          https://azdor.gov/forms/individual/form-140-x-y-tables
          https://azdor.gov/forms/individual/form-140-x-y-tables
          https://azdor.gov/businesses-arizona/withholding-tax/withholding-exceptions
          "https://azdor.gov/individuals/income-tax-filing-assistance/identifying-other-taxable-income#:~:text=In%20regards%20to%20capital%20gains
         the%20individuals%20regular%20tax%20rate.
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Kentucky', 'state_occupation': 'Arizona', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def kentucky_and_arkansas(self, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Kentucky:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        
        Arkansas:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Kentucky:
        - https://revenue.ky.gov/News/Pages/Dor-Announces-Updates-to-Individual-Income-Tax-for-2023-Tax-Year.aspx
          https://revenue.ky.gov/Individual/Individual-Income-Tax/Pages/default.aspx
          https://revenue.ky.gov/News/Pages/Dor-Announces-Updates-to-Individual-Income-Tax-for-2023-Tax-Year.aspx
          https://revenue.ky.gov/Individual/Individual-Income-Tax/Pages/default.aspx
          https://revenue.ky.gov/Dor%20Training%20Materials/103%20KAR%2017.140.%20Individual%20income%20tax%20-%20reciprocity%20-%20nonresidents.pdf
          https://revenue.ky.gov/Individual/Individual-Income-Tax/Pages/default.aspx
          https://nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2021/Taxes-21-32.htm?taxmap=true
        
        Arkansas:
        -         
        - https://www.arkansasedc.com/why-arkansas/business-climate/tax-structure/personal-income-tax
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_and_AR1000NR_Instructions.pdf
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_FullYearResidentIndividualIncomeTaxReturn.pdf
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_and_AR1000NR_Instructions.pdf
          https://www.arkansasedc.com/why-arkansas/business-climate/tax-structure/capital-gains-tax-reduction
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_and_AR1000NR_Instructions.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Kentucky', 'state_occupation': 'Arkansas', "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def kentucky_and_california(self, county_residence='Other', state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Kentucky:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        
        California:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Kentucky:
        - https://revenue.ky.gov/News/Pages/Dor-Announces-Updates-to-Individual-Income-Tax-for-2023-Tax-Year.aspx
          https://revenue.ky.gov/Individual/Individual-Income-Tax/Pages/default.aspx
          https://revenue.ky.gov/News/Pages/Dor-Announces-Updates-to-Individual-Income-Tax-for-2023-Tax-Year.aspx
          https://revenue.ky.gov/Individual/Individual-Income-Tax/Pages/default.aspx
          https://revenue.ky.gov/Dor%20Training%20Materials/103%20KAR%2017.140.%20Individual%20income%20tax%20-%20reciprocity%20-%20nonresidents.pdf
          https://revenue.ky.gov/Individual/Individual-Income-Tax/Pages/default.aspx
          https://nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2021/Taxes-21-32.htm?taxmap=true
        
        California:
        -         
        - https://www.ftb.ca.gov/forms/2022/2022-540-tax-rate-schedules.pdf
          https://www.ftb.ca.gov/file/personal/deductions/index.html
          https://www.ftb.ca.gov/about-ftb/newsroom/caleitc/eligibility-and-credit-information.html
          https://www.ftb.ca.gov/file/personal/income-types/capital-gains-and-losses.html#:~:text=California%20does%20not%20have%20a
         are%20taxed%20as%20ordinary%20income
          https://www.ftb.ca.gov/file/personal/
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Kentucky', 'state_occupation': 'California', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def kentucky_and_colorado(self, county_residence='Other', state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Kentucky:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        
        Colorado:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Kentucky:
        - https://revenue.ky.gov/News/Pages/Dor-Announces-Updates-to-Individual-Income-Tax-for-2023-Tax-Year.aspx
          https://revenue.ky.gov/Individual/Individual-Income-Tax/Pages/default.aspx
          https://revenue.ky.gov/News/Pages/Dor-Announces-Updates-to-Individual-Income-Tax-for-2023-Tax-Year.aspx
          https://revenue.ky.gov/Individual/Individual-Income-Tax/Pages/default.aspx
          https://revenue.ky.gov/Dor%20Training%20Materials/103%20KAR%2017.140.%20Individual%20income%20tax%20-%20reciprocity%20-%20nonresidents.pdf
          https://revenue.ky.gov/Individual/Individual-Income-Tax/Pages/default.aspx
          https://nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2021/Taxes-21-32.htm?taxmap=true
        
        Colorado:
        -         
        - https://leg.colorado.gov/agencies/legislative-council-staff/individual-income-tax%C2%A0
          https://leg.colorado.gov/agencies/legislative-council-staff/individual-income-tax%C2%A0
          https://tax.colorado.gov/sites/tax/files/documents/ITT_Capital_Gain_Subtraction_Feb_2022.pdf
          https://www.auroragov.org/business_services/taxes/occupational_privilege_tax
          https://www.denvergov.org/Government/Agencies-Departments-Offices/Agencies-Departments-Offices-Directory/Department-of-Finance/Our-Divisions/Treasury/Business-Tax-Information#:~:text=Occupational%20privilege%20tax
         -%28show%20below%29&text=Each%20taxable%20employee%20is%20liable
         month%20for%20each%20taxable%20employee
          https://www.glendale.co.us/355/Occupational-Privilege-Tax
          https://www.greenwoodvillage.com/1220/Occupational-Privilege-Tax-OPT
          https://ci.sheridan.co.us/288/Occupational-Privilege-Tax
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Kentucky', 'state_occupation': 'Colorado', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def kentucky_and_connecticut(self, county_residence='Other', state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Kentucky:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        
        Connecticut:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Kentucky:
        - https://revenue.ky.gov/News/Pages/Dor-Announces-Updates-to-Individual-Income-Tax-for-2023-Tax-Year.aspx
          https://revenue.ky.gov/Individual/Individual-Income-Tax/Pages/default.aspx
          https://revenue.ky.gov/News/Pages/Dor-Announces-Updates-to-Individual-Income-Tax-for-2023-Tax-Year.aspx
          https://revenue.ky.gov/Individual/Individual-Income-Tax/Pages/default.aspx
          https://revenue.ky.gov/Dor%20Training%20Materials/103%20KAR%2017.140.%20Individual%20income%20tax%20-%20reciprocity%20-%20nonresidents.pdf
          https://revenue.ky.gov/Individual/Individual-Income-Tax/Pages/default.aspx
          https://nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2021/Taxes-21-32.htm?taxmap=true
        
        Connecticut:
        -         
        - https://portal.ct.gov/-/media/DRS/Forms/2022/Income/2022-CT-1040-Instructions_1222.pdf
          https://www.cga.ct.gov/2022/rpt/pdf/2022-R-0108.pdf
          https://www.cga.ct.gov/current/pub/chap_229.htm#sec_12-703
          https://portal.ct.gov/-/media/DRS/Forms/2022/Income/2022-CT-1040-Instructions_1222.pdf
          "https://portal.ct.gov/DRS/Publications/TSSNs/TSSN-29#:~:text=An%20individuals%20net%20capital%20gains
         portion%20of%20Social%20Security%20benefits"
          "https://portal.ct.gov/DRS/Publications/TSSNs/TSSN-29#:~:text=An%20individuals%20net%20capital%20gains
         portion%20of%20Social%20Security%20benefits"
          https://portal.ct.gov/DRS/Individuals/Resident-Income-Tax/Tax-Information
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Kentucky', 'state_occupation': 'Connecticut', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def kentucky_and_delaware(self, county_residence='Other', state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, federal_agi=75000, filing_status='single', sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Kentucky:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        
        Delaware:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        federal_agi : int, optional
            Adjusted Gross Income (AGI) for federal taxes. Default is 0. Sources - https://www.irs.gov/e-file-providers/definition-of-adjusted-gross-income, https://www.irs.gov/newsroom/taxpayers-should-know-the-difference-between-standard-and-itemized-deductions
        income : int, optional
            Gross income. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Kentucky:
        - https://revenue.ky.gov/News/Pages/Dor-Announces-Updates-to-Individual-Income-Tax-for-2023-Tax-Year.aspx
          https://revenue.ky.gov/Individual/Individual-Income-Tax/Pages/default.aspx
          https://revenue.ky.gov/News/Pages/Dor-Announces-Updates-to-Individual-Income-Tax-for-2023-Tax-Year.aspx
          https://revenue.ky.gov/Individual/Individual-Income-Tax/Pages/default.aspx
          https://revenue.ky.gov/Dor%20Training%20Materials/103%20KAR%2017.140.%20Individual%20income%20tax%20-%20reciprocity%20-%20nonresidents.pdf
          https://revenue.ky.gov/Individual/Individual-Income-Tax/Pages/default.aspx
          https://nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2021/Taxes-21-32.htm?taxmap=true
        
        Delaware:
        -         
        - https://revenue.delaware.gov/software-developer/tax-rate-changes/
          https://revenuefiles.delaware.gov/2022/PIT-RES_TY22_2022-01_PaperInteractive.pdf
          https://revenuefiles.delaware.gov/2022/PIT-RES_TY22_2022-02_Instructions.pdf
          https://revenue.delaware.gov/frequently-asked-questions/personal-income-tax-faqs/
          https://www.newcastlede.gov/DocumentCenter/View/20147/City-of-Wilmington-Wage-Tax-Information-PDF
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        federal_agi = self.federal_agi if hasattr(self, 'federal_agi') else federal_agi
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Kentucky', 'state_occupation': 'Delaware', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "federal_agi": federal_agi, "filing_status": filing_status, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def kentucky_and_florida(self, county_residence='Other', state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Kentucky:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        
        Florida:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Kentucky:
        - https://revenue.ky.gov/News/Pages/Dor-Announces-Updates-to-Individual-Income-Tax-for-2023-Tax-Year.aspx
          https://revenue.ky.gov/Individual/Individual-Income-Tax/Pages/default.aspx
          https://revenue.ky.gov/News/Pages/Dor-Announces-Updates-to-Individual-Income-Tax-for-2023-Tax-Year.aspx
          https://revenue.ky.gov/Individual/Individual-Income-Tax/Pages/default.aspx
          https://revenue.ky.gov/Dor%20Training%20Materials/103%20KAR%2017.140.%20Individual%20income%20tax%20-%20reciprocity%20-%20nonresidents.pdf
          https://revenue.ky.gov/Individual/Individual-Income-Tax/Pages/default.aspx
          https://nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2021/Taxes-21-32.htm?taxmap=true
        
        Florida:
        -         
        - https://floridarevenue.com/taxes/taxesfees/Pages/tax_interest_rates.aspx
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Kentucky', 'state_occupation': 'Florida', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def kentucky_and_georgia(self, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Kentucky:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        
        Georgia:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Kentucky:
        - https://revenue.ky.gov/News/Pages/Dor-Announces-Updates-to-Individual-Income-Tax-for-2023-Tax-Year.aspx
          https://revenue.ky.gov/Individual/Individual-Income-Tax/Pages/default.aspx
          https://revenue.ky.gov/News/Pages/Dor-Announces-Updates-to-Individual-Income-Tax-for-2023-Tax-Year.aspx
          https://revenue.ky.gov/Individual/Individual-Income-Tax/Pages/default.aspx
          https://revenue.ky.gov/Dor%20Training%20Materials/103%20KAR%2017.140.%20Individual%20income%20tax%20-%20reciprocity%20-%20nonresidents.pdf
          https://revenue.ky.gov/Individual/Individual-Income-Tax/Pages/default.aspx
          https://nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2021/Taxes-21-32.htm?taxmap=true
        
        Georgia:
        -         
        - https://dor.georgia.gov/tax-tables-georgia-tax-rate-schedule
          https://apps.dor.ga.gov/FillableForms/PDFViewer/Index?form=2022GA500
          https://apps.dor.ga.gov/FillableForms/PDFViewer/Index?form=2022GA500
          https://apps.dor.ga.gov/FillableForms/PDFViewer/Index?form=2022GA500
          https://dor.georgia.gov/years-individual-income-tax-forms
          https://dor.georgia.gov/years-individual-income-tax-forms
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Kentucky', 'state_occupation': 'Georgia', "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def kentucky_and_hawaii(self, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Kentucky:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        
        Hawaii:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Kentucky:
        - https://revenue.ky.gov/News/Pages/Dor-Announces-Updates-to-Individual-Income-Tax-for-2023-Tax-Year.aspx
          https://revenue.ky.gov/Individual/Individual-Income-Tax/Pages/default.aspx
          https://revenue.ky.gov/News/Pages/Dor-Announces-Updates-to-Individual-Income-Tax-for-2023-Tax-Year.aspx
          https://revenue.ky.gov/Individual/Individual-Income-Tax/Pages/default.aspx
          https://revenue.ky.gov/Dor%20Training%20Materials/103%20KAR%2017.140.%20Individual%20income%20tax%20-%20reciprocity%20-%20nonresidents.pdf
          https://revenue.ky.gov/Individual/Individual-Income-Tax/Pages/default.aspx
          https://nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2021/Taxes-21-32.htm?taxmap=true
        
        Hawaii:
        -         
        - https://tax.hawaii.gov/forms/d_18table-on/d_18table-on_p13/
          https://files.hawaii.gov/tax/forms/2018/18table-on.pdf
          https://files.hawaii.gov/tax/news/pubs/22outline.pdf
          https://files.hawaii.gov/tax/news/pubs/FastTaxRefGuide_2017.pdf
          https://www.tfhawaii.org/wordpress/state-tax-resources/mini-tax-guides/individual-income-tax-chapter-235/#:~:text=Long%20term%20capital%20gains%20are%20taxed%20at%20a%20maximum%20of%207.25%25
          https://www.grassrootinstitute.org/2023/02/hb337-increasing-tax-on-capital-gains-a-bad-risk/
          https://files.hawaii.gov/tax/news/pubs/22outline.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Kentucky', 'state_occupation': 'Hawaii', "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def kentucky_and_idaho(self, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Kentucky:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        
        Idaho:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Kentucky:
        - https://revenue.ky.gov/News/Pages/Dor-Announces-Updates-to-Individual-Income-Tax-for-2023-Tax-Year.aspx
          https://revenue.ky.gov/Individual/Individual-Income-Tax/Pages/default.aspx
          https://revenue.ky.gov/News/Pages/Dor-Announces-Updates-to-Individual-Income-Tax-for-2023-Tax-Year.aspx
          https://revenue.ky.gov/Individual/Individual-Income-Tax/Pages/default.aspx
          https://revenue.ky.gov/Dor%20Training%20Materials/103%20KAR%2017.140.%20Individual%20income%20tax%20-%20reciprocity%20-%20nonresidents.pdf
          https://revenue.ky.gov/Individual/Individual-Income-Tax/Pages/default.aspx
          https://nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2021/Taxes-21-32.htm?taxmap=true
        
        Idaho:
        -         
        - https://tax.idaho.gov/taxes/income-tax/individual-income/individual-income-tax-rate-schedule/
          https://tax.idaho.gov/wp-content/uploads/forms/EIN00046/EIN00046_03-01-2023.pdf
          https://tax.idaho.gov/taxes/income-tax/individual-income/idaho-source-income/
          https://tax.idaho.gov/taxes/income-tax/individual-income/idaho-source-income/
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Kentucky', 'state_occupation': 'Idaho', "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def kentucky_and_illinois(self, county_residence='Other', state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Kentucky:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        
        Illinois:
        


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Kentucky:
        - https://revenue.ky.gov/News/Pages/Dor-Announces-Updates-to-Individual-Income-Tax-for-2023-Tax-Year.aspx
          https://revenue.ky.gov/Individual/Individual-Income-Tax/Pages/default.aspx
          https://revenue.ky.gov/News/Pages/Dor-Announces-Updates-to-Individual-Income-Tax-for-2023-Tax-Year.aspx
          https://revenue.ky.gov/Individual/Individual-Income-Tax/Pages/default.aspx
          https://revenue.ky.gov/Dor%20Training%20Materials/103%20KAR%2017.140.%20Individual%20income%20tax%20-%20reciprocity%20-%20nonresidents.pdf
          https://revenue.ky.gov/Individual/Individual-Income-Tax/Pages/default.aspx
          https://nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2021/Taxes-21-32.htm?taxmap=true
        
        Illinois:
        -         
        - https://tax.illinois.gov/research/taxrates/income.html#IndividualIncome
          https://tax.illinois.gov/content/dam/soi/en/web/tax/forms/incometax/documents/currentyear/individual/il-1040-instr.pdf
          https://tax.illinois.gov/questionsandanswers/12.html
          https://tax.illinois.gov/content/dam/soi/en/web/tax/forms/incometax/documents/currentyear/individual/il-1040-instr.pdf
          https://tax.illinois.gov/research/taxrates/personalproperty.html
          https://tax.illinois.gov/localgovernments/income.html
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Kentucky', 'state_occupation': 'Illinois', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def kentucky_and_indiana(self, county_residence='Other', state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Kentucky:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        
        Indiana:
        


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Kentucky:
        - https://revenue.ky.gov/News/Pages/Dor-Announces-Updates-to-Individual-Income-Tax-for-2023-Tax-Year.aspx
          https://revenue.ky.gov/Individual/Individual-Income-Tax/Pages/default.aspx
          https://revenue.ky.gov/News/Pages/Dor-Announces-Updates-to-Individual-Income-Tax-for-2023-Tax-Year.aspx
          https://revenue.ky.gov/Individual/Individual-Income-Tax/Pages/default.aspx
          https://revenue.ky.gov/Dor%20Training%20Materials/103%20KAR%2017.140.%20Individual%20income%20tax%20-%20reciprocity%20-%20nonresidents.pdf
          https://revenue.ky.gov/Individual/Individual-Income-Tax/Pages/default.aspx
          https://nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2021/Taxes-21-32.htm?taxmap=true
        
        Indiana:
        -         
        - https://www.in.gov/dor/tax-forms/2022-individual-income-tax-forms/
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-32.htm
          https://www.in.gov/dor/files/ib28.pdf
          https://www.in.gov/dor/tax-forms/2022-individual-income-tax-forms/
          https://www.in.gov/dor/files/2022-county-tax-rates-and-codes.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Kentucky', 'state_occupation': 'Indiana', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def kentucky_and_iowa(self, county_residence='Other', state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Kentucky:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        
        Iowa:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Kentucky:
        - https://revenue.ky.gov/News/Pages/Dor-Announces-Updates-to-Individual-Income-Tax-for-2023-Tax-Year.aspx
          https://revenue.ky.gov/Individual/Individual-Income-Tax/Pages/default.aspx
          https://revenue.ky.gov/News/Pages/Dor-Announces-Updates-to-Individual-Income-Tax-for-2023-Tax-Year.aspx
          https://revenue.ky.gov/Individual/Individual-Income-Tax/Pages/default.aspx
          https://revenue.ky.gov/Dor%20Training%20Materials/103%20KAR%2017.140.%20Individual%20income%20tax%20-%20reciprocity%20-%20nonresidents.pdf
          https://revenue.ky.gov/Individual/Individual-Income-Tax/Pages/default.aspx
          https://nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2021/Taxes-21-32.htm?taxmap=true
        
        Iowa:
        -         
        - https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
          https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
          https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
          https://tax.iowa.gov/iowa-illinois-reciprocal-agreement
          https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Kentucky', 'state_occupation': 'Iowa', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def kentucky_and_kansas(self, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Kentucky:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        
        Kansas:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Kentucky:
        - https://revenue.ky.gov/News/Pages/Dor-Announces-Updates-to-Individual-Income-Tax-for-2023-Tax-Year.aspx
          https://revenue.ky.gov/Individual/Individual-Income-Tax/Pages/default.aspx
          https://revenue.ky.gov/News/Pages/Dor-Announces-Updates-to-Individual-Income-Tax-for-2023-Tax-Year.aspx
          https://revenue.ky.gov/Individual/Individual-Income-Tax/Pages/default.aspx
          https://revenue.ky.gov/Dor%20Training%20Materials/103%20KAR%2017.140.%20Individual%20income%20tax%20-%20reciprocity%20-%20nonresidents.pdf
          https://revenue.ky.gov/Individual/Individual-Income-Tax/Pages/default.aspx
          https://nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2021/Taxes-21-32.htm?taxmap=true
        
        Kansas:
        -         
        - https://www.ksrevenue.gov/taxrates.html
          https://www.ksrevenue.gov/incomebook22.html#0
          https://www.ksrevenue.gov/incomebook22.html#0
          https://www.ksrevenue.gov/incomebook22.html#0
          https://www.ksrevenue.gov/incomebook22.html
          https://www.ksrevenue.gov/incomebook22.html
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Kentucky', 'state_occupation': 'Kansas', "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def kentucky_and_louisiana(self, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Kentucky:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        
        Louisiana:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Kentucky:
        - https://revenue.ky.gov/News/Pages/Dor-Announces-Updates-to-Individual-Income-Tax-for-2023-Tax-Year.aspx
          https://revenue.ky.gov/Individual/Individual-Income-Tax/Pages/default.aspx
          https://revenue.ky.gov/News/Pages/Dor-Announces-Updates-to-Individual-Income-Tax-for-2023-Tax-Year.aspx
          https://revenue.ky.gov/Individual/Individual-Income-Tax/Pages/default.aspx
          https://revenue.ky.gov/Dor%20Training%20Materials/103%20KAR%2017.140.%20Individual%20income%20tax%20-%20reciprocity%20-%20nonresidents.pdf
          https://revenue.ky.gov/Individual/Individual-Income-Tax/Pages/default.aspx
          https://nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2021/Taxes-21-32.htm?taxmap=true
        
        Louisiana:
        -         
        - https://revenue.louisiana.gov/individualincometax
          https://revenue.louisiana.gov/individualincometax
          https://www.revenue.louisiana.gov/taxforms/6935(11_02)F.pdf
          https://revenue.louisiana.gov/individualincometax
          https://revenue.louisiana.gov/individualincometax
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Kentucky', 'state_occupation': 'Louisiana', "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def kentucky_and_maine(self, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Kentucky:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        
        Maine:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Kentucky:
        - https://revenue.ky.gov/News/Pages/Dor-Announces-Updates-to-Individual-Income-Tax-for-2023-Tax-Year.aspx
          https://revenue.ky.gov/Individual/Individual-Income-Tax/Pages/default.aspx
          https://revenue.ky.gov/News/Pages/Dor-Announces-Updates-to-Individual-Income-Tax-for-2023-Tax-Year.aspx
          https://revenue.ky.gov/Individual/Individual-Income-Tax/Pages/default.aspx
          https://revenue.ky.gov/Dor%20Training%20Materials/103%20KAR%2017.140.%20Individual%20income%20tax%20-%20reciprocity%20-%20nonresidents.pdf
          https://revenue.ky.gov/Individual/Individual-Income-Tax/Pages/default.aspx
          https://nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2021/Taxes-21-32.htm?taxmap=true
        
        Maine:
        -         
        - https://www.maine.gov/revenue/tax-return-forms/individual-income-tax-2022
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/ind_tax_rate_sched_2022.pdf
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/22_1040me_book_gen_instruc_revisedFeb23.pdf
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/22_1040me_book_gen_instruc_revisedFeb23.pdf
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/22_1040me_book_gen_instruc_revisedFeb23.pdf
          https://www.maine.gov/revenue/tax-return-forms/individual-income-tax-2022
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Kentucky', 'state_occupation': 'Maine', "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def kentucky_and_maryland(self, county_residence='Other', state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Kentucky:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        
        Maryland:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Kentucky:
        - https://revenue.ky.gov/News/Pages/Dor-Announces-Updates-to-Individual-Income-Tax-for-2023-Tax-Year.aspx
          https://revenue.ky.gov/Individual/Individual-Income-Tax/Pages/default.aspx
          https://revenue.ky.gov/News/Pages/Dor-Announces-Updates-to-Individual-Income-Tax-for-2023-Tax-Year.aspx
          https://revenue.ky.gov/Individual/Individual-Income-Tax/Pages/default.aspx
          https://revenue.ky.gov/Dor%20Training%20Materials/103%20KAR%2017.140.%20Individual%20income%20tax%20-%20reciprocity%20-%20nonresidents.pdf
          https://revenue.ky.gov/Individual/Individual-Income-Tax/Pages/default.aspx
          https://nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2021/Taxes-21-32.htm?taxmap=true
        
        Maryland:
        -         
        - https://www.marylandtaxes.gov/individual/income/tax-info/tax-rates.php
          https://www.marylandtaxes.gov/forms/22_forms/Resident_Booklet.pdf
          https://www.marylandtaxes.gov/forms/Tax_Publications/Administrative_Releases/Income_and_Estate_Tax_Releases/ar_it37.pdf
          https://www.marylandtaxes.gov/individual/index.php
          https://www.marylandtaxes.gov/individual/income/tax-info/tax-rates.php
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Kentucky', 'state_occupation': 'Maryland', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def kentucky_and_massachusetts(self, county_residence='Other', state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Kentucky:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        
        Massachusetts:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Kentucky:
        - https://revenue.ky.gov/News/Pages/Dor-Announces-Updates-to-Individual-Income-Tax-for-2023-Tax-Year.aspx
          https://revenue.ky.gov/Individual/Individual-Income-Tax/Pages/default.aspx
          https://revenue.ky.gov/News/Pages/Dor-Announces-Updates-to-Individual-Income-Tax-for-2023-Tax-Year.aspx
          https://revenue.ky.gov/Individual/Individual-Income-Tax/Pages/default.aspx
          https://revenue.ky.gov/Dor%20Training%20Materials/103%20KAR%2017.140.%20Individual%20income%20tax%20-%20reciprocity%20-%20nonresidents.pdf
          https://revenue.ky.gov/Individual/Individual-Income-Tax/Pages/default.aspx
          https://nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2021/Taxes-21-32.htm?taxmap=true
        
        Massachusetts:
        -         
        - https://www.mass.gov/service-details/massachusetts-tax-rates
          https://www.mass.gov/service-details/view-massachusetts-personal-income-tax-exemptions
          https://www.mass.gov/service-details/view-massachusetts-personal-income-tax-exemptions
          https://www.mass.gov/service-details/view-child-and-dependent-related-credits
          https://www.mass.gov/service-details/massachusetts-tax-rates
          https://www.mass.gov/personal-income-tax
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Kentucky', 'state_occupation': 'Massachusetts', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def kentucky_and_michigan(self, county_residence='Other', state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Kentucky:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        
        Michigan:
        


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Kentucky:
        - https://revenue.ky.gov/News/Pages/Dor-Announces-Updates-to-Individual-Income-Tax-for-2023-Tax-Year.aspx
          https://revenue.ky.gov/Individual/Individual-Income-Tax/Pages/default.aspx
          https://revenue.ky.gov/News/Pages/Dor-Announces-Updates-to-Individual-Income-Tax-for-2023-Tax-Year.aspx
          https://revenue.ky.gov/Individual/Individual-Income-Tax/Pages/default.aspx
          https://revenue.ky.gov/Dor%20Training%20Materials/103%20KAR%2017.140.%20Individual%20income%20tax%20-%20reciprocity%20-%20nonresidents.pdf
          https://revenue.ky.gov/Individual/Individual-Income-Tax/Pages/default.aspx
          https://nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2021/Taxes-21-32.htm?taxmap=true
        
        Michigan:
        -         
        - https://www.michigan.gov/taxes/iit/new-developments-for-tax-year-2022
          "https://www.michigan.gov/taxes/-/media/Project/Websites/taxes/Forms/2023/2023-SUW/446_Withholding-Guide_2023.pdf?rev=177ca3d8cb034392bf84e317ed78c320&hash=19C5F42F43A89A91863D229349F48883#:~:text=The%20exemption%20amount%20is%20%245%2C400
         employees%20Michigan%20income%20tax%20return"
          "https://www.michigan.gov/taxes/-/media/Project/Websites/taxes/Forms/2023/2023-SUW/446_Withholding-Guide_2023.pdf?rev=177ca3d8cb034392bf84e317ed78c320&hash=19C5F42F43A89A91863D229349F48883#:~:text=The%20exemption%20amount%20is%20%245%2C400
         employees%20Michigan%20income%20tax%20return"
          https://www.michigan.gov/taxes/questions/iit/accordion/residency/are-my-wages-earned-in-another-state-taxable-in-michigan-if-i-am-a-michigan-resident-1
          https://www.michigan.gov/taxes/-/media/Project/Websites/taxes/Forms/2022/2022-IIT-Forms/BOOK_MI-1040.pdf?rev=4524f1cbbc0243b194ab05a6680be75b&hash=40DFE71E651F7C1D65C31F6CD03D944F
          https://www.michigan.gov/taxes/questions/iit/accordion/general/what-cities-impose-an-income-tax
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Kentucky', 'state_occupation': 'Michigan', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def kentucky_and_minnesota(self, county_residence='Other', state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Kentucky:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        
        Minnesota:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Kentucky:
        - https://revenue.ky.gov/News/Pages/Dor-Announces-Updates-to-Individual-Income-Tax-for-2023-Tax-Year.aspx
          https://revenue.ky.gov/Individual/Individual-Income-Tax/Pages/default.aspx
          https://revenue.ky.gov/News/Pages/Dor-Announces-Updates-to-Individual-Income-Tax-for-2023-Tax-Year.aspx
          https://revenue.ky.gov/Individual/Individual-Income-Tax/Pages/default.aspx
          https://revenue.ky.gov/Dor%20Training%20Materials/103%20KAR%2017.140.%20Individual%20income%20tax%20-%20reciprocity%20-%20nonresidents.pdf
          https://revenue.ky.gov/Individual/Individual-Income-Tax/Pages/default.aspx
          https://nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2021/Taxes-21-32.htm?taxmap=true
        
        Minnesota:
        -         
        - https://www.revenue.state.mn.us/minnesota-income-tax-rates-and-brackets
          https://www.revenue.state.mn.us/child-and-dependent-care-credit
          https://www.revenue.state.mn.us/dependent-exemptions
          https://www.revenue.state.mn.us/minnesota-standard-deduction
          https://mn.gov/mmb/assets/mwr_form_tcm1059-128581.pdf
          https://www.revenue.state.mn.us/individual-income-tax
          https://www.revenue.state.mn.us/individual-income-tax
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Kentucky', 'state_occupation': 'Minnesota', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def kentucky_and_mississippi(self, county_residence='Other', state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Kentucky:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        
        Mississippi:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Kentucky:
        - https://revenue.ky.gov/News/Pages/Dor-Announces-Updates-to-Individual-Income-Tax-for-2023-Tax-Year.aspx
          https://revenue.ky.gov/Individual/Individual-Income-Tax/Pages/default.aspx
          https://revenue.ky.gov/News/Pages/Dor-Announces-Updates-to-Individual-Income-Tax-for-2023-Tax-Year.aspx
          https://revenue.ky.gov/Individual/Individual-Income-Tax/Pages/default.aspx
          https://revenue.ky.gov/Dor%20Training%20Materials/103%20KAR%2017.140.%20Individual%20income%20tax%20-%20reciprocity%20-%20nonresidents.pdf
          https://revenue.ky.gov/Individual/Individual-Income-Tax/Pages/default.aspx
          https://nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2021/Taxes-21-32.htm?taxmap=true
        
        Mississippi:
        -         
        - https://www.dor.ms.gov/individual/tax-rates
          https://www.dor.ms.gov/individual/tax-rates
          https://www.dor.ms.gov/individual/tax-rates
          https://www.dor.ms.gov/individual
          https://www.dor.ms.gov/individual
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Kentucky', 'state_occupation': 'Mississippi', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def kentucky_and_missouri(self, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Kentucky:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        
        Missouri:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Kentucky:
        - https://revenue.ky.gov/News/Pages/Dor-Announces-Updates-to-Individual-Income-Tax-for-2023-Tax-Year.aspx
          https://revenue.ky.gov/Individual/Individual-Income-Tax/Pages/default.aspx
          https://revenue.ky.gov/News/Pages/Dor-Announces-Updates-to-Individual-Income-Tax-for-2023-Tax-Year.aspx
          https://revenue.ky.gov/Individual/Individual-Income-Tax/Pages/default.aspx
          https://revenue.ky.gov/Dor%20Training%20Materials/103%20KAR%2017.140.%20Individual%20income%20tax%20-%20reciprocity%20-%20nonresidents.pdf
          https://revenue.ky.gov/Individual/Individual-Income-Tax/Pages/default.aspx
          https://nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2021/Taxes-21-32.htm?taxmap=true
        
        Missouri:
        -         
        - https://dor.mo.gov/taxation/individual/tax-types/income/year-changes/
          https://dor.mo.gov/taxation/individual/tax-types/income/year-changes/
          https://dor.mo.gov/taxation/individual/tax-types/income/year-changes/
          https://www.kcmo.gov/city-hall/departments/finance/earnings-tax
          https://www.stlouis-mo.gov/government/departments/collector/earnings-tax/individual-earnings-tax-info.cfm
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Kentucky', 'state_occupation': 'Missouri', "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def kentucky_and_montana(self, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Kentucky:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        
        Montana:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Kentucky:
        - https://revenue.ky.gov/News/Pages/Dor-Announces-Updates-to-Individual-Income-Tax-for-2023-Tax-Year.aspx
          https://revenue.ky.gov/Individual/Individual-Income-Tax/Pages/default.aspx
          https://revenue.ky.gov/News/Pages/Dor-Announces-Updates-to-Individual-Income-Tax-for-2023-Tax-Year.aspx
          https://revenue.ky.gov/Individual/Individual-Income-Tax/Pages/default.aspx
          https://revenue.ky.gov/Dor%20Training%20Materials/103%20KAR%2017.140.%20Individual%20income%20tax%20-%20reciprocity%20-%20nonresidents.pdf
          https://revenue.ky.gov/Individual/Individual-Income-Tax/Pages/default.aspx
          https://nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2021/Taxes-21-32.htm?taxmap=true
        
        Montana:
        -         
        - https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/2022/12/montana-employees-withholding-allowance-and-exemption-certificate-form-mw-4.pdf
          https://mtrevenue.gov/2021/11/29/simplification-of-montana-income-taxation/?utm_source=rss&utm_medium=rss&utm_campaign=simplification-of-montana-income-taxation
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Kentucky', 'state_occupation': 'Montana', "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def kentucky_and_nebraska(self, county_residence='Other', state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Kentucky:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        
        Nebraska:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Kentucky:
        - https://revenue.ky.gov/News/Pages/Dor-Announces-Updates-to-Individual-Income-Tax-for-2023-Tax-Year.aspx
          https://revenue.ky.gov/Individual/Individual-Income-Tax/Pages/default.aspx
          https://revenue.ky.gov/News/Pages/Dor-Announces-Updates-to-Individual-Income-Tax-for-2023-Tax-Year.aspx
          https://revenue.ky.gov/Individual/Individual-Income-Tax/Pages/default.aspx
          https://revenue.ky.gov/Dor%20Training%20Materials/103%20KAR%2017.140.%20Individual%20income%20tax%20-%20reciprocity%20-%20nonresidents.pdf
          https://revenue.ky.gov/Individual/Individual-Income-Tax/Pages/default.aspx
          https://nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2021/Taxes-21-32.htm?taxmap=true
        
        Nebraska:
        -         
        - https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2023_Circular_EN_8-429r2023_Whole_Book_11.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Kentucky', 'state_occupation': 'Nebraska', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def kentucky_and_nevada(self, county_residence='Other', state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Kentucky:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        
        Nevada:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Kentucky:
        - https://revenue.ky.gov/News/Pages/Dor-Announces-Updates-to-Individual-Income-Tax-for-2023-Tax-Year.aspx
          https://revenue.ky.gov/Individual/Individual-Income-Tax/Pages/default.aspx
          https://revenue.ky.gov/News/Pages/Dor-Announces-Updates-to-Individual-Income-Tax-for-2023-Tax-Year.aspx
          https://revenue.ky.gov/Individual/Individual-Income-Tax/Pages/default.aspx
          https://revenue.ky.gov/Dor%20Training%20Materials/103%20KAR%2017.140.%20Individual%20income%20tax%20-%20reciprocity%20-%20nonresidents.pdf
          https://revenue.ky.gov/Individual/Individual-Income-Tax/Pages/default.aspx
          https://nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2021/Taxes-21-32.htm?taxmap=true
        
        Nevada:
        -         
        - https://www.nvsos.gov/sos/businesses/the-nevada-advantage
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Kentucky', 'state_occupation': 'Nevada', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def kentucky_and_new_hampshire(self, county_residence='Other', state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Kentucky:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        
        New Hampshire:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Kentucky:
        - https://revenue.ky.gov/News/Pages/Dor-Announces-Updates-to-Individual-Income-Tax-for-2023-Tax-Year.aspx
          https://revenue.ky.gov/Individual/Individual-Income-Tax/Pages/default.aspx
          https://revenue.ky.gov/News/Pages/Dor-Announces-Updates-to-Individual-Income-Tax-for-2023-Tax-Year.aspx
          https://revenue.ky.gov/Individual/Individual-Income-Tax/Pages/default.aspx
          https://revenue.ky.gov/Dor%20Training%20Materials/103%20KAR%2017.140.%20Individual%20income%20tax%20-%20reciprocity%20-%20nonresidents.pdf
          https://revenue.ky.gov/Individual/Individual-Income-Tax/Pages/default.aspx
          https://nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2021/Taxes-21-32.htm?taxmap=true
        
        New Hampshire:
        -         
        - https://www.revenue.nh.gov/assistance/tax-overview.htm
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Kentucky', 'state_occupation': 'NewHampshire', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def kentucky_and_new_jersey(self, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Kentucky:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        
        New Jersey:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Kentucky:
        - https://revenue.ky.gov/News/Pages/Dor-Announces-Updates-to-Individual-Income-Tax-for-2023-Tax-Year.aspx
          https://revenue.ky.gov/Individual/Individual-Income-Tax/Pages/default.aspx
          https://revenue.ky.gov/News/Pages/Dor-Announces-Updates-to-Individual-Income-Tax-for-2023-Tax-Year.aspx
          https://revenue.ky.gov/Individual/Individual-Income-Tax/Pages/default.aspx
          https://revenue.ky.gov/Dor%20Training%20Materials/103%20KAR%2017.140.%20Individual%20income%20tax%20-%20reciprocity%20-%20nonresidents.pdf
          https://revenue.ky.gov/Individual/Individual-Income-Tax/Pages/default.aspx
          https://nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2021/Taxes-21-32.htm?taxmap=true
        
        New Jersey:
        -         
        - https://www.state.nj.us/treasury/taxation/taxtables.shtml
          https://www.state.nj.us/treasury/taxation/njit13.shtml
          https://www.nj.gov/treasury/taxation/njit25.shtml
          https://www.state.nj.us/treasury/taxation/individuals/freqqiti.shtml
          https://www.state.nj.us/treasury/taxation/localtax.shtml
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Kentucky', 'state_occupation': 'NewJersey', "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def kentucky_and_new_mexico(self, county_residence='Other', state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Kentucky:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        
        New Mexico:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Kentucky:
        - https://revenue.ky.gov/News/Pages/Dor-Announces-Updates-to-Individual-Income-Tax-for-2023-Tax-Year.aspx
          https://revenue.ky.gov/Individual/Individual-Income-Tax/Pages/default.aspx
          https://revenue.ky.gov/News/Pages/Dor-Announces-Updates-to-Individual-Income-Tax-for-2023-Tax-Year.aspx
          https://revenue.ky.gov/Individual/Individual-Income-Tax/Pages/default.aspx
          https://revenue.ky.gov/Dor%20Training%20Materials/103%20KAR%2017.140.%20Individual%20income%20tax%20-%20reciprocity%20-%20nonresidents.pdf
          https://revenue.ky.gov/Individual/Individual-Income-Tax/Pages/default.aspx
          https://nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2021/Taxes-21-32.htm?taxmap=true
        
        New Mexico:
        -         
        - https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/fdf3c548-8aba-4b9c-9eb4-bb564c716015/FYI-104.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/0bf943fd-652e-4400-bb1b-cd8397eb5a95/2022pit-adj.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/2f1a6781-9534-4436-b427-1557f9592099/2022pit-adj-ins.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/0bf943fd-652e-4400-bb1b-cd8397eb5a95/2022pit-adj.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/1afc56af-ea90-4d48-82e5-1f9aeb43255a/PITbook2022.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/fdf3c548-8aba-4b9c-9eb4-bb564c716015/FYI-104.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Kentucky', 'state_occupation': 'NewMexico', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def kentucky_and_new_york(self, county_residence='Other', state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Kentucky:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        
        New York:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Kentucky:
        - https://revenue.ky.gov/News/Pages/Dor-Announces-Updates-to-Individual-Income-Tax-for-2023-Tax-Year.aspx
          https://revenue.ky.gov/Individual/Individual-Income-Tax/Pages/default.aspx
          https://revenue.ky.gov/News/Pages/Dor-Announces-Updates-to-Individual-Income-Tax-for-2023-Tax-Year.aspx
          https://revenue.ky.gov/Individual/Individual-Income-Tax/Pages/default.aspx
          https://revenue.ky.gov/Dor%20Training%20Materials/103%20KAR%2017.140.%20Individual%20income%20tax%20-%20reciprocity%20-%20nonresidents.pdf
          https://revenue.ky.gov/Individual/Individual-Income-Tax/Pages/default.aspx
          https://nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2021/Taxes-21-32.htm?taxmap=true
        
        New York:
        -         
        - https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nys-tax-rate-schedule
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nys-tax-rate-schedule
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nys-tax-rate-schedule
          https://www.tax.ny.gov/pit/
          https://www.tax.ny.gov/pit/file/nonresident-faqs.htm
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#yonkers-tax-rate-schedule
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nyc-tax-rate-schedule
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Kentucky', 'state_occupation': 'NewYork', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def kentucky_and_north_carolina(self, county_residence='Other', state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Kentucky:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        
        North Carolina:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Kentucky:
        - https://revenue.ky.gov/News/Pages/Dor-Announces-Updates-to-Individual-Income-Tax-for-2023-Tax-Year.aspx
          https://revenue.ky.gov/Individual/Individual-Income-Tax/Pages/default.aspx
          https://revenue.ky.gov/News/Pages/Dor-Announces-Updates-to-Individual-Income-Tax-for-2023-Tax-Year.aspx
          https://revenue.ky.gov/Individual/Individual-Income-Tax/Pages/default.aspx
          https://revenue.ky.gov/Dor%20Training%20Materials/103%20KAR%2017.140.%20Individual%20income%20tax%20-%20reciprocity%20-%20nonresidents.pdf
          https://revenue.ky.gov/Individual/Individual-Income-Tax/Pages/default.aspx
          https://nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2021/Taxes-21-32.htm?taxmap=true
        
        North Carolina:
        -         
        - https://www.ncdor.gov/taxes-forms/tax-rate-schedules
          https://www.ncdor.gov/taxes-forms/individual-income-tax/north-carolina-standard-deduction-or-north-carolina-itemized-deductions
          https://www.ncdor.gov/taxes-forms/individual-income-tax
          https://www.ncdor.gov/taxes-forms/individual-income-tax
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Kentucky', 'state_occupation': 'NorthCarolina', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def kentucky_and_north_dakota(self, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Kentucky:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        
        North Dakota:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Kentucky:
        - https://revenue.ky.gov/News/Pages/Dor-Announces-Updates-to-Individual-Income-Tax-for-2023-Tax-Year.aspx
          https://revenue.ky.gov/Individual/Individual-Income-Tax/Pages/default.aspx
          https://revenue.ky.gov/News/Pages/Dor-Announces-Updates-to-Individual-Income-Tax-for-2023-Tax-Year.aspx
          https://revenue.ky.gov/Individual/Individual-Income-Tax/Pages/default.aspx
          https://revenue.ky.gov/Dor%20Training%20Materials/103%20KAR%2017.140.%20Individual%20income%20tax%20-%20reciprocity%20-%20nonresidents.pdf
          https://revenue.ky.gov/Individual/Individual-Income-Tax/Pages/default.aspx
          https://nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2021/Taxes-21-32.htm?taxmap=true
        
        North Dakota:
        -         
        - https://www.tax.nd.gov/forms
          https://www.tax.nd.gov/sites/www/files/documents/forms/individual/2022-iit/2022-individual-income-tax-booklet.pdf
          https://www.tax.nd.gov/sites/www/files/documents/forms/2022-individual-income-tax-booklet.pdf
          https://www.tax.nd.gov/tax-exemptions-credits/income-tax-exemptions-credits
          https://www.tax.nd.gov/sites/www/files/documents/forms/business/ndwrfillable.pdf
          https://www.tax.nd.gov/news/tax-legislative-changes/significant-changes-law/individual-income-tax-history
          https://www.tax.nd.gov/forms
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Kentucky', 'state_occupation': 'NorthDakota', "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def kentucky_and_ohio(self, county_residence='Other', state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Kentucky:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        
        Ohio:
        


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Kentucky:
        - https://revenue.ky.gov/News/Pages/Dor-Announces-Updates-to-Individual-Income-Tax-for-2023-Tax-Year.aspx
          https://revenue.ky.gov/Individual/Individual-Income-Tax/Pages/default.aspx
          https://revenue.ky.gov/News/Pages/Dor-Announces-Updates-to-Individual-Income-Tax-for-2023-Tax-Year.aspx
          https://revenue.ky.gov/Individual/Individual-Income-Tax/Pages/default.aspx
          https://revenue.ky.gov/Dor%20Training%20Materials/103%20KAR%2017.140.%20Individual%20income%20tax%20-%20reciprocity%20-%20nonresidents.pdf
          https://revenue.ky.gov/Individual/Individual-Income-Tax/Pages/default.aspx
          https://nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2021/Taxes-21-32.htm?taxmap=true
        
        Ohio:
        -         
        - https://tax.ohio.gov/individual/resources/annual-tax-rates
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://tax.ohio.gov/static/forms/employer_withholding/generic/wth_it4nr.pdf
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://www.ritaohio.com/TaxRatesTable
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Kentucky', 'state_occupation': 'Ohio', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def kentucky_and_oklahoma(self, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Kentucky:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        
        Oklahoma:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Kentucky:
        - https://revenue.ky.gov/News/Pages/Dor-Announces-Updates-to-Individual-Income-Tax-for-2023-Tax-Year.aspx
          https://revenue.ky.gov/Individual/Individual-Income-Tax/Pages/default.aspx
          https://revenue.ky.gov/News/Pages/Dor-Announces-Updates-to-Individual-Income-Tax-for-2023-Tax-Year.aspx
          https://revenue.ky.gov/Individual/Individual-Income-Tax/Pages/default.aspx
          https://revenue.ky.gov/Dor%20Training%20Materials/103%20KAR%2017.140.%20Individual%20income%20tax%20-%20reciprocity%20-%20nonresidents.pdf
          https://revenue.ky.gov/Individual/Individual-Income-Tax/Pages/default.aspx
          https://nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2021/Taxes-21-32.htm?taxmap=true
        
        Oklahoma:
        -         
        - https://oklahoma.gov/content/dam/ok/en/tax/documents/resources/publications/businesses/withholding-tables/WHTables-2022.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Kentucky', 'state_occupation': 'Oklahoma', "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def kentucky_and_oregon(self, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Kentucky:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        
        Oregon:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Kentucky:
        - https://revenue.ky.gov/News/Pages/Dor-Announces-Updates-to-Individual-Income-Tax-for-2023-Tax-Year.aspx
          https://revenue.ky.gov/Individual/Individual-Income-Tax/Pages/default.aspx
          https://revenue.ky.gov/News/Pages/Dor-Announces-Updates-to-Individual-Income-Tax-for-2023-Tax-Year.aspx
          https://revenue.ky.gov/Individual/Individual-Income-Tax/Pages/default.aspx
          https://revenue.ky.gov/Dor%20Training%20Materials/103%20KAR%2017.140.%20Individual%20income%20tax%20-%20reciprocity%20-%20nonresidents.pdf
          https://revenue.ky.gov/Individual/Individual-Income-Tax/Pages/default.aspx
          https://nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2021/Taxes-21-32.htm?taxmap=true
        
        Oregon:
        -         
        - https://www.oregon.gov/dor/programs/individuals/Pages/PIT.aspx
          https://www.oregon.gov/dor/programs/individuals/Documents/Full%20year%20resident
         %20Form%20or-40%20filers.pdf
          https://www.oregon.gov/dor/programs/individuals/Documents/Part-year%20and%20nonresident
         %20Form%20or-40-P%20and%20or-40-N%20filers.pdf.pdf
          https://www.oregon.gov/dor/forms/FormsPubs/form-or-40_101-040_2022.pdf
          https://www.oregon.gov/dor/forms/FormsPubs/form-or-40_101-040_2022.pdf
          https://www.oregon.gov/dor/forms/FormsPubs/form-or-40_101-040_2022.pdf
          https://www.portland.gov/revenue/personal-tax
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Kentucky', 'state_occupation': 'Oregon', "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def kentucky_and_pennsylvania(self, county_residence='Other', state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Kentucky:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        
        Pennsylvania:
        


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Kentucky:
        - https://revenue.ky.gov/News/Pages/Dor-Announces-Updates-to-Individual-Income-Tax-for-2023-Tax-Year.aspx
          https://revenue.ky.gov/Individual/Individual-Income-Tax/Pages/default.aspx
          https://revenue.ky.gov/News/Pages/Dor-Announces-Updates-to-Individual-Income-Tax-for-2023-Tax-Year.aspx
          https://revenue.ky.gov/Individual/Individual-Income-Tax/Pages/default.aspx
          https://revenue.ky.gov/Dor%20Training%20Materials/103%20KAR%2017.140.%20Individual%20income%20tax%20-%20reciprocity%20-%20nonresidents.pdf
          https://revenue.ky.gov/Individual/Individual-Income-Tax/Pages/default.aspx
          https://nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2021/Taxes-21-32.htm?taxmap=true
        
        Pennsylvania:
        -         
        - https://www.revenue.pa.gov/Tax%20Rates/Pages/default.aspx
          https://www.revenue.pa.gov/FormsandPublications/FormsforIndividuals/PIT/Documents/2022/2022_pa-40in.pdf
          https://www.revenue.pa.gov/FormsandPublications/FormsforBusinesses/EmployerWithholding/Documents/rev-419.pdf
          https://www.revenue.pa.gov/FormsandPublications/FormsforIndividuals/PIT/Documents/2022/2022_pa-40in.pdf
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-25.htm
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Kentucky', 'state_occupation': 'Pennsylvania', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def kentucky_and_rhode_island(self, county_residence='Other', state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Kentucky:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        
        Rhode Island:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Kentucky:
        - https://revenue.ky.gov/News/Pages/Dor-Announces-Updates-to-Individual-Income-Tax-for-2023-Tax-Year.aspx
          https://revenue.ky.gov/Individual/Individual-Income-Tax/Pages/default.aspx
          https://revenue.ky.gov/News/Pages/Dor-Announces-Updates-to-Individual-Income-Tax-for-2023-Tax-Year.aspx
          https://revenue.ky.gov/Individual/Individual-Income-Tax/Pages/default.aspx
          https://revenue.ky.gov/Dor%20Training%20Materials/103%20KAR%2017.140.%20Individual%20income%20tax%20-%20reciprocity%20-%20nonresidents.pdf
          https://revenue.ky.gov/Individual/Individual-Income-Tax/Pages/default.aspx
          https://nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2021/Taxes-21-32.htm?taxmap=true
        
        Rhode Island:
        -         
        - https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2022_1040WE_w_0.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Kentucky', 'state_occupation': 'RhodeIsland', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def kentucky_and_south_carolina(self, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Kentucky:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        
        South Carolina:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Kentucky:
        - https://revenue.ky.gov/News/Pages/Dor-Announces-Updates-to-Individual-Income-Tax-for-2023-Tax-Year.aspx
          https://revenue.ky.gov/Individual/Individual-Income-Tax/Pages/default.aspx
          https://revenue.ky.gov/News/Pages/Dor-Announces-Updates-to-Individual-Income-Tax-for-2023-Tax-Year.aspx
          https://revenue.ky.gov/Individual/Individual-Income-Tax/Pages/default.aspx
          https://revenue.ky.gov/Dor%20Training%20Materials/103%20KAR%2017.140.%20Individual%20income%20tax%20-%20reciprocity%20-%20nonresidents.pdf
          https://revenue.ky.gov/Individual/Individual-Income-Tax/Pages/default.aspx
          https://nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2021/Taxes-21-32.htm?taxmap=true
        
        South Carolina:
        -         
        - https://dor.sc.gov/resources-site/lawandpolicy/Advisory%20Opinions/IL22-15.pdf
          https://dor.sc.gov/forms-site/Forms/IITPacket_2022.pdf
          https://www.scstatehouse.gov/code/t12c006.php
          https://dor.sc.gov/resources-site/lawandpolicy/Advisory%20Opinions/IL22-15.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Kentucky', 'state_occupation': 'SouthCarolina', "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def kentucky_and_south_dakota(self, county_residence='Other', state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Kentucky:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        
        South Dakota:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Kentucky:
        - https://revenue.ky.gov/News/Pages/Dor-Announces-Updates-to-Individual-Income-Tax-for-2023-Tax-Year.aspx
          https://revenue.ky.gov/Individual/Individual-Income-Tax/Pages/default.aspx
          https://revenue.ky.gov/News/Pages/Dor-Announces-Updates-to-Individual-Income-Tax-for-2023-Tax-Year.aspx
          https://revenue.ky.gov/Individual/Individual-Income-Tax/Pages/default.aspx
          https://revenue.ky.gov/Dor%20Training%20Materials/103%20KAR%2017.140.%20Individual%20income%20tax%20-%20reciprocity%20-%20nonresidents.pdf
          https://revenue.ky.gov/Individual/Individual-Income-Tax/Pages/default.aspx
          https://nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2021/Taxes-21-32.htm?taxmap=true
        
        South Dakota:
        -         
        - https://dor.sd.gov/individuals/taxes/
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Kentucky', 'state_occupation': 'SouthDakota', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def kentucky_and_tennessee(self, county_residence='Other', state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Kentucky:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        
        Tennessee:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Kentucky:
        - https://revenue.ky.gov/News/Pages/Dor-Announces-Updates-to-Individual-Income-Tax-for-2023-Tax-Year.aspx
          https://revenue.ky.gov/Individual/Individual-Income-Tax/Pages/default.aspx
          https://revenue.ky.gov/News/Pages/Dor-Announces-Updates-to-Individual-Income-Tax-for-2023-Tax-Year.aspx
          https://revenue.ky.gov/Individual/Individual-Income-Tax/Pages/default.aspx
          https://revenue.ky.gov/Dor%20Training%20Materials/103%20KAR%2017.140.%20Individual%20income%20tax%20-%20reciprocity%20-%20nonresidents.pdf
          https://revenue.ky.gov/Individual/Individual-Income-Tax/Pages/default.aspx
          https://nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2021/Taxes-21-32.htm?taxmap=true
        
        Tennessee:
        -         
        - https://www.tn.gov/revenue/taxes.html
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Kentucky', 'state_occupation': 'Tennessee', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def kentucky_and_texas(self, county_residence='Other', state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Kentucky:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        
        Texas:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Kentucky:
        - https://revenue.ky.gov/News/Pages/Dor-Announces-Updates-to-Individual-Income-Tax-for-2023-Tax-Year.aspx
          https://revenue.ky.gov/Individual/Individual-Income-Tax/Pages/default.aspx
          https://revenue.ky.gov/News/Pages/Dor-Announces-Updates-to-Individual-Income-Tax-for-2023-Tax-Year.aspx
          https://revenue.ky.gov/Individual/Individual-Income-Tax/Pages/default.aspx
          https://revenue.ky.gov/Dor%20Training%20Materials/103%20KAR%2017.140.%20Individual%20income%20tax%20-%20reciprocity%20-%20nonresidents.pdf
          https://revenue.ky.gov/Individual/Individual-Income-Tax/Pages/default.aspx
          https://nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2021/Taxes-21-32.htm?taxmap=true
        
        Texas:
        -         
        - https://comptroller.texas.gov/taxes/
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Kentucky', 'state_occupation': 'Texas', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def kentucky_and_utah(self, county_residence='Other', state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Kentucky:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        
        Utah:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Kentucky:
        - https://revenue.ky.gov/News/Pages/Dor-Announces-Updates-to-Individual-Income-Tax-for-2023-Tax-Year.aspx
          https://revenue.ky.gov/Individual/Individual-Income-Tax/Pages/default.aspx
          https://revenue.ky.gov/News/Pages/Dor-Announces-Updates-to-Individual-Income-Tax-for-2023-Tax-Year.aspx
          https://revenue.ky.gov/Individual/Individual-Income-Tax/Pages/default.aspx
          https://revenue.ky.gov/Dor%20Training%20Materials/103%20KAR%2017.140.%20Individual%20income%20tax%20-%20reciprocity%20-%20nonresidents.pdf
          https://revenue.ky.gov/Individual/Individual-Income-Tax/Pages/default.aspx
          https://nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2021/Taxes-21-32.htm?taxmap=true
        
        Utah:
        -         
        - https://incometax.utah.gov/paying/tax-rates
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Kentucky', 'state_occupation': 'Utah', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def kentucky_and_vermont(self, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Kentucky:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        
        Vermont:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Kentucky:
        - https://revenue.ky.gov/News/Pages/Dor-Announces-Updates-to-Individual-Income-Tax-for-2023-Tax-Year.aspx
          https://revenue.ky.gov/Individual/Individual-Income-Tax/Pages/default.aspx
          https://revenue.ky.gov/News/Pages/Dor-Announces-Updates-to-Individual-Income-Tax-for-2023-Tax-Year.aspx
          https://revenue.ky.gov/Individual/Individual-Income-Tax/Pages/default.aspx
          https://revenue.ky.gov/Dor%20Training%20Materials/103%20KAR%2017.140.%20Individual%20income%20tax%20-%20reciprocity%20-%20nonresidents.pdf
          https://revenue.ky.gov/Individual/Individual-Income-Tax/Pages/default.aspx
          https://nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2021/Taxes-21-32.htm?taxmap=true
        
        Vermont:
        -         
        - https://tax.vermont.gov/sites/tax/files/documents/GB-1210-2023.pdf
          https://tax.vermont.gov/sites/tax/files/documents/Income%20Booklet-2022.pdf
          https://tax.vermont.gov/sites/tax/files/documents/Income%20Booklet-2022.pdf
          https://tax.vermont.gov/sites/tax/files/documents/Income%20Booklet-2022.pdf
          https://tax.vermont.gov/individuals/personal-income-tax/tax-credits
          https://tax.vermont.gov/individuals/personal-income-tax/taxable-income
          https://tax.vermont.gov/individuals/personal-income-tax/local-option-tax
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Kentucky', 'state_occupation': 'Vermont', "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def kentucky_and_virginia(self, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Kentucky:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        
        Virginia:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Kentucky:
        - https://revenue.ky.gov/News/Pages/Dor-Announces-Updates-to-Individual-Income-Tax-for-2023-Tax-Year.aspx
          https://revenue.ky.gov/Individual/Individual-Income-Tax/Pages/default.aspx
          https://revenue.ky.gov/News/Pages/Dor-Announces-Updates-to-Individual-Income-Tax-for-2023-Tax-Year.aspx
          https://revenue.ky.gov/Individual/Individual-Income-Tax/Pages/default.aspx
          https://revenue.ky.gov/Dor%20Training%20Materials/103%20KAR%2017.140.%20Individual%20income%20tax%20-%20reciprocity%20-%20nonresidents.pdf
          https://revenue.ky.gov/Individual/Individual-Income-Tax/Pages/default.aspx
          https://nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2021/Taxes-21-32.htm?taxmap=true
        
        Virginia:
        -         
        - https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/reciprocity
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Kentucky', 'state_occupation': 'Virginia', "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def kentucky_and_washington(self, county_residence='Other', state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Kentucky:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        
        Washington:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Kentucky:
        - https://revenue.ky.gov/News/Pages/Dor-Announces-Updates-to-Individual-Income-Tax-for-2023-Tax-Year.aspx
          https://revenue.ky.gov/Individual/Individual-Income-Tax/Pages/default.aspx
          https://revenue.ky.gov/News/Pages/Dor-Announces-Updates-to-Individual-Income-Tax-for-2023-Tax-Year.aspx
          https://revenue.ky.gov/Individual/Individual-Income-Tax/Pages/default.aspx
          https://revenue.ky.gov/Dor%20Training%20Materials/103%20KAR%2017.140.%20Individual%20income%20tax%20-%20reciprocity%20-%20nonresidents.pdf
          https://revenue.ky.gov/Individual/Individual-Income-Tax/Pages/default.aspx
          https://nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2021/Taxes-21-32.htm?taxmap=true
        
        Washington:
        -         
        - https://dor.wa.gov/taxes-rates/income-tax
          https://dor.wa.gov/taxes-rates/other-taxes/capital-gains-tax
          https://dor.wa.gov/taxes-rates/income-tax
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Kentucky', 'state_occupation': 'Washington', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def kentucky_and_west_virginia(self, county_residence='Other', state_agi=75000, county_occupation='Other', pay_periods=24, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Kentucky:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        
        West Virginia:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_occupation : str, optional
            County of occupation. Default is Other.
        pay_periods : int, optional
            Number of pay periods. Options are 13, 24, 26, 52. Default is 24.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Kentucky:
        - https://revenue.ky.gov/News/Pages/Dor-Announces-Updates-to-Individual-Income-Tax-for-2023-Tax-Year.aspx
          https://revenue.ky.gov/Individual/Individual-Income-Tax/Pages/default.aspx
          https://revenue.ky.gov/News/Pages/Dor-Announces-Updates-to-Individual-Income-Tax-for-2023-Tax-Year.aspx
          https://revenue.ky.gov/Individual/Individual-Income-Tax/Pages/default.aspx
          https://revenue.ky.gov/Dor%20Training%20Materials/103%20KAR%2017.140.%20Individual%20income%20tax%20-%20reciprocity%20-%20nonresidents.pdf
          https://revenue.ky.gov/Individual/Individual-Income-Tax/Pages/default.aspx
          https://nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2021/Taxes-21-32.htm?taxmap=true
        
        West Virginia:
        -         
        - https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://tax.wv.gov/Documents/TSD/tsd438.pdf
          https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2011/Taxes-11-35att.html
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        county_occupation = self.county_occupation if hasattr(self, 'county_occupation') else county_occupation
        pay_periods = self.pay_periods if hasattr(self, 'pay_periods') else pay_periods
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Kentucky', 'state_occupation': 'WestVirginia', "county_residence": county_residence, "state_agi": state_agi, "county_occupation": county_occupation, "pay_periods": pay_periods, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def kentucky_and_wisconsin(self, county_residence='Other', state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Kentucky:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        
        Wisconsin:
        


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Kentucky:
        - https://revenue.ky.gov/News/Pages/Dor-Announces-Updates-to-Individual-Income-Tax-for-2023-Tax-Year.aspx
          https://revenue.ky.gov/Individual/Individual-Income-Tax/Pages/default.aspx
          https://revenue.ky.gov/News/Pages/Dor-Announces-Updates-to-Individual-Income-Tax-for-2023-Tax-Year.aspx
          https://revenue.ky.gov/Individual/Individual-Income-Tax/Pages/default.aspx
          https://revenue.ky.gov/Dor%20Training%20Materials/103%20KAR%2017.140.%20Individual%20income%20tax%20-%20reciprocity%20-%20nonresidents.pdf
          https://revenue.ky.gov/Individual/Individual-Income-Tax/Pages/default.aspx
          https://nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2021/Taxes-21-32.htm?taxmap=true
        
        Wisconsin:
        -         
        - https://www.revenue.wi.gov/TaxForms2023/2023-Form1-ES-Inst.pdf
          https://www.revenue.wi.gov/TaxForms2023/2023-Form1-ES-Inst.pdf
          https://www.revenue.wi.gov/TaxForms2023/2023-Form1-ES-Inst.pdf
          https://www.revenue.wi.gov/Pages/FAQS/pcs-work.aspx
          https://www.revenue.wi.gov/Dor%20Publications/pb103.pdf
          https://www.revenue.wi.gov/Pages/FAQS/pcs-work.aspx
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Kentucky', 'state_occupation': 'Wisconsin', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def kentucky_and_wyoming(self, county_residence='Other', state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Kentucky:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        
        Wyoming:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Kentucky:
        - https://revenue.ky.gov/News/Pages/Dor-Announces-Updates-to-Individual-Income-Tax-for-2023-Tax-Year.aspx
          https://revenue.ky.gov/Individual/Individual-Income-Tax/Pages/default.aspx
          https://revenue.ky.gov/News/Pages/Dor-Announces-Updates-to-Individual-Income-Tax-for-2023-Tax-Year.aspx
          https://revenue.ky.gov/Individual/Individual-Income-Tax/Pages/default.aspx
          https://revenue.ky.gov/Dor%20Training%20Materials/103%20KAR%2017.140.%20Individual%20income%20tax%20-%20reciprocity%20-%20nonresidents.pdf
          https://revenue.ky.gov/Individual/Individual-Income-Tax/Pages/default.aspx
          https://nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2021/Taxes-21-32.htm?taxmap=true
        
        Wyoming:
        -         
        - https://revenue.wyo.gov/
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Kentucky', 'state_occupation': 'Wyoming', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def louisiana_and_alabama(self, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Louisiana:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Alabama:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Louisiana:
        - https://revenue.louisiana.gov/individualincometax
          https://revenue.louisiana.gov/individualincometax
          https://www.revenue.louisiana.gov/taxforms/6935(11_02)F.pdf
          https://revenue.louisiana.gov/individualincometax
          https://revenue.louisiana.gov/individualincometax
        
        Alabama:
        -         
        - https://www.revenue.alabama.gov/ultraviewer/viewer/basic_viewer/index.html?form=2023/02/22f40.pdf
          https://www.revenue.alabama.gov/faqs/what-is-alabamas-individual-income-tax-rate/
          https://www.revenue.alabama.gov/ultraviewer/viewer/basic_viewer/index.html?form=2023/02/22f40.pdf
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-24.htm?taxmap=true
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-24.htm?taxmap=true
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-24.htm?taxmap=true
          https://www.revenue.alabama.gov/individual-corporate/income-to-be-reported-on-the-alabama-income-tax-return/
          https://www.almonline.org/TaxRates.aspx#OccupationalTax
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Louisiana', 'state_occupation': 'Alabama', "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def louisiana_and_alaska(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Louisiana:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Alaska:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Louisiana:
        - https://revenue.louisiana.gov/individualincometax
          https://revenue.louisiana.gov/individualincometax
          https://www.revenue.louisiana.gov/taxforms/6935(11_02)F.pdf
          https://revenue.louisiana.gov/individualincometax
          https://revenue.louisiana.gov/individualincometax
        
        Alaska:
        -         
        - https://www.commerce.alaska.gov/web/dcra/officeofthestateassessor/alaskataxfacts.aspx
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Louisiana', 'state_occupation': 'Alaska', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def louisiana_and_arizona(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Louisiana:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Arizona:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        credits_state : int, optional
            Credits for state taxes. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Louisiana:
        - https://revenue.louisiana.gov/individualincometax
          https://revenue.louisiana.gov/individualincometax
          https://www.revenue.louisiana.gov/taxforms/6935(11_02)F.pdf
          https://revenue.louisiana.gov/individualincometax
          https://revenue.louisiana.gov/individualincometax
        
        Arizona:
        -         
        - https://azdor.gov/forms/individual/form-140-x-y-tables
          https://azdor.gov/forms/individual/form-140-x-y-tables
          https://azdor.gov/forms/individual/form-140-x-y-tables
          https://azdor.gov/businesses-arizona/withholding-tax/withholding-exceptions
          "https://azdor.gov/individuals/income-tax-filing-assistance/identifying-other-taxable-income#:~:text=In%20regards%20to%20capital%20gains
         the%20individuals%20regular%20tax%20rate.
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Louisiana', 'state_occupation': 'Arizona', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def louisiana_and_arkansas(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Louisiana:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Arkansas:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Louisiana:
        - https://revenue.louisiana.gov/individualincometax
          https://revenue.louisiana.gov/individualincometax
          https://www.revenue.louisiana.gov/taxforms/6935(11_02)F.pdf
          https://revenue.louisiana.gov/individualincometax
          https://revenue.louisiana.gov/individualincometax
        
        Arkansas:
        -         
        - https://www.arkansasedc.com/why-arkansas/business-climate/tax-structure/personal-income-tax
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_and_AR1000NR_Instructions.pdf
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_FullYearResidentIndividualIncomeTaxReturn.pdf
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_and_AR1000NR_Instructions.pdf
          https://www.arkansasedc.com/why-arkansas/business-climate/tax-structure/capital-gains-tax-reduction
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_and_AR1000NR_Instructions.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Louisiana', 'state_occupation': 'Arkansas', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def louisiana_and_california(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Louisiana:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        California:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Louisiana:
        - https://revenue.louisiana.gov/individualincometax
          https://revenue.louisiana.gov/individualincometax
          https://www.revenue.louisiana.gov/taxforms/6935(11_02)F.pdf
          https://revenue.louisiana.gov/individualincometax
          https://revenue.louisiana.gov/individualincometax
        
        California:
        -         
        - https://www.ftb.ca.gov/forms/2022/2022-540-tax-rate-schedules.pdf
          https://www.ftb.ca.gov/file/personal/deductions/index.html
          https://www.ftb.ca.gov/about-ftb/newsroom/caleitc/eligibility-and-credit-information.html
          https://www.ftb.ca.gov/file/personal/income-types/capital-gains-and-losses.html#:~:text=California%20does%20not%20have%20a
         are%20taxed%20as%20ordinary%20income
          https://www.ftb.ca.gov/file/personal/
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Louisiana', 'state_occupation': 'California', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def louisiana_and_colorado(self, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Louisiana:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Colorado:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Louisiana:
        - https://revenue.louisiana.gov/individualincometax
          https://revenue.louisiana.gov/individualincometax
          https://www.revenue.louisiana.gov/taxforms/6935(11_02)F.pdf
          https://revenue.louisiana.gov/individualincometax
          https://revenue.louisiana.gov/individualincometax
        
        Colorado:
        -         
        - https://leg.colorado.gov/agencies/legislative-council-staff/individual-income-tax%C2%A0
          https://leg.colorado.gov/agencies/legislative-council-staff/individual-income-tax%C2%A0
          https://tax.colorado.gov/sites/tax/files/documents/ITT_Capital_Gain_Subtraction_Feb_2022.pdf
          https://www.auroragov.org/business_services/taxes/occupational_privilege_tax
          https://www.denvergov.org/Government/Agencies-Departments-Offices/Agencies-Departments-Offices-Directory/Department-of-Finance/Our-Divisions/Treasury/Business-Tax-Information#:~:text=Occupational%20privilege%20tax
         -%28show%20below%29&text=Each%20taxable%20employee%20is%20liable
         month%20for%20each%20taxable%20employee
          https://www.glendale.co.us/355/Occupational-Privilege-Tax
          https://www.greenwoodvillage.com/1220/Occupational-Privilege-Tax-OPT
          https://ci.sheridan.co.us/288/Occupational-Privilege-Tax
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Louisiana', 'state_occupation': 'Colorado', "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def louisiana_and_connecticut(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Louisiana:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Connecticut:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Louisiana:
        - https://revenue.louisiana.gov/individualincometax
          https://revenue.louisiana.gov/individualincometax
          https://www.revenue.louisiana.gov/taxforms/6935(11_02)F.pdf
          https://revenue.louisiana.gov/individualincometax
          https://revenue.louisiana.gov/individualincometax
        
        Connecticut:
        -         
        - https://portal.ct.gov/-/media/DRS/Forms/2022/Income/2022-CT-1040-Instructions_1222.pdf
          https://www.cga.ct.gov/2022/rpt/pdf/2022-R-0108.pdf
          https://www.cga.ct.gov/current/pub/chap_229.htm#sec_12-703
          https://portal.ct.gov/-/media/DRS/Forms/2022/Income/2022-CT-1040-Instructions_1222.pdf
          "https://portal.ct.gov/DRS/Publications/TSSNs/TSSN-29#:~:text=An%20individuals%20net%20capital%20gains
         portion%20of%20Social%20Security%20benefits"
          "https://portal.ct.gov/DRS/Publications/TSSNs/TSSN-29#:~:text=An%20individuals%20net%20capital%20gains
         portion%20of%20Social%20Security%20benefits"
          https://portal.ct.gov/DRS/Individuals/Resident-Income-Tax/Tax-Information
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Louisiana', 'state_occupation': 'Connecticut', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def louisiana_and_delaware(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, federal_agi=75000, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Louisiana:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Delaware:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        federal_agi : int, optional
            Adjusted Gross Income (AGI) for federal taxes. Default is 0. Sources - https://www.irs.gov/e-file-providers/definition-of-adjusted-gross-income, https://www.irs.gov/newsroom/taxpayers-should-know-the-difference-between-standard-and-itemized-deductions
        income : int, optional
            Gross income. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Louisiana:
        - https://revenue.louisiana.gov/individualincometax
          https://revenue.louisiana.gov/individualincometax
          https://www.revenue.louisiana.gov/taxforms/6935(11_02)F.pdf
          https://revenue.louisiana.gov/individualincometax
          https://revenue.louisiana.gov/individualincometax
        
        Delaware:
        -         
        - https://revenue.delaware.gov/software-developer/tax-rate-changes/
          https://revenuefiles.delaware.gov/2022/PIT-RES_TY22_2022-01_PaperInteractive.pdf
          https://revenuefiles.delaware.gov/2022/PIT-RES_TY22_2022-02_Instructions.pdf
          https://revenue.delaware.gov/frequently-asked-questions/personal-income-tax-faqs/
          https://www.newcastlede.gov/DocumentCenter/View/20147/City-of-Wilmington-Wage-Tax-Information-PDF
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        federal_agi = self.federal_agi if hasattr(self, 'federal_agi') else federal_agi
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Louisiana', 'state_occupation': 'Delaware', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "federal_agi": federal_agi, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def louisiana_and_florida(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Louisiana:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Florida:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Louisiana:
        - https://revenue.louisiana.gov/individualincometax
          https://revenue.louisiana.gov/individualincometax
          https://www.revenue.louisiana.gov/taxforms/6935(11_02)F.pdf
          https://revenue.louisiana.gov/individualincometax
          https://revenue.louisiana.gov/individualincometax
        
        Florida:
        -         
        - https://floridarevenue.com/taxes/taxesfees/Pages/tax_interest_rates.aspx
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Louisiana', 'state_occupation': 'Florida', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def louisiana_and_georgia(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Louisiana:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Georgia:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Louisiana:
        - https://revenue.louisiana.gov/individualincometax
          https://revenue.louisiana.gov/individualincometax
          https://www.revenue.louisiana.gov/taxforms/6935(11_02)F.pdf
          https://revenue.louisiana.gov/individualincometax
          https://revenue.louisiana.gov/individualincometax
        
        Georgia:
        -         
        - https://dor.georgia.gov/tax-tables-georgia-tax-rate-schedule
          https://apps.dor.ga.gov/FillableForms/PDFViewer/Index?form=2022GA500
          https://apps.dor.ga.gov/FillableForms/PDFViewer/Index?form=2022GA500
          https://apps.dor.ga.gov/FillableForms/PDFViewer/Index?form=2022GA500
          https://dor.georgia.gov/years-individual-income-tax-forms
          https://dor.georgia.gov/years-individual-income-tax-forms
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Louisiana', 'state_occupation': 'Georgia', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def louisiana_and_hawaii(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Louisiana:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Hawaii:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Louisiana:
        - https://revenue.louisiana.gov/individualincometax
          https://revenue.louisiana.gov/individualincometax
          https://www.revenue.louisiana.gov/taxforms/6935(11_02)F.pdf
          https://revenue.louisiana.gov/individualincometax
          https://revenue.louisiana.gov/individualincometax
        
        Hawaii:
        -         
        - https://tax.hawaii.gov/forms/d_18table-on/d_18table-on_p13/
          https://files.hawaii.gov/tax/forms/2018/18table-on.pdf
          https://files.hawaii.gov/tax/news/pubs/22outline.pdf
          https://files.hawaii.gov/tax/news/pubs/FastTaxRefGuide_2017.pdf
          https://www.tfhawaii.org/wordpress/state-tax-resources/mini-tax-guides/individual-income-tax-chapter-235/#:~:text=Long%20term%20capital%20gains%20are%20taxed%20at%20a%20maximum%20of%207.25%25
          https://www.grassrootinstitute.org/2023/02/hb337-increasing-tax-on-capital-gains-a-bad-risk/
          https://files.hawaii.gov/tax/news/pubs/22outline.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Louisiana', 'state_occupation': 'Hawaii', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def louisiana_and_idaho(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Louisiana:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Idaho:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Louisiana:
        - https://revenue.louisiana.gov/individualincometax
          https://revenue.louisiana.gov/individualincometax
          https://www.revenue.louisiana.gov/taxforms/6935(11_02)F.pdf
          https://revenue.louisiana.gov/individualincometax
          https://revenue.louisiana.gov/individualincometax
        
        Idaho:
        -         
        - https://tax.idaho.gov/taxes/income-tax/individual-income/individual-income-tax-rate-schedule/
          https://tax.idaho.gov/wp-content/uploads/forms/EIN00046/EIN00046_03-01-2023.pdf
          https://tax.idaho.gov/taxes/income-tax/individual-income/idaho-source-income/
          https://tax.idaho.gov/taxes/income-tax/individual-income/idaho-source-income/
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Louisiana', 'state_occupation': 'Idaho', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def louisiana_and_illinois(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Louisiana:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Illinois:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Louisiana:
        - https://revenue.louisiana.gov/individualincometax
          https://revenue.louisiana.gov/individualincometax
          https://www.revenue.louisiana.gov/taxforms/6935(11_02)F.pdf
          https://revenue.louisiana.gov/individualincometax
          https://revenue.louisiana.gov/individualincometax
        
        Illinois:
        -         
        - https://tax.illinois.gov/research/taxrates/income.html#IndividualIncome
          https://tax.illinois.gov/content/dam/soi/en/web/tax/forms/incometax/documents/currentyear/individual/il-1040-instr.pdf
          https://tax.illinois.gov/questionsandanswers/12.html
          https://tax.illinois.gov/content/dam/soi/en/web/tax/forms/incometax/documents/currentyear/individual/il-1040-instr.pdf
          https://tax.illinois.gov/research/taxrates/personalproperty.html
          https://tax.illinois.gov/localgovernments/income.html
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Louisiana', 'state_occupation': 'Illinois', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def louisiana_and_indiana(self, capital_gains_short=0, dependents=0, sources=False, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_long=0, filing_status='single', income=100000, credits_state=0):
        """
        --- Parameters ----
        
        Louisiana:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Indiana:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Louisiana:
        - https://revenue.louisiana.gov/individualincometax
          https://revenue.louisiana.gov/individualincometax
          https://www.revenue.louisiana.gov/taxforms/6935(11_02)F.pdf
          https://revenue.louisiana.gov/individualincometax
          https://revenue.louisiana.gov/individualincometax
        
        Indiana:
        -         
        - https://www.in.gov/dor/tax-forms/2022-individual-income-tax-forms/
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-32.htm
          https://www.in.gov/dor/files/ib28.pdf
          https://www.in.gov/dor/tax-forms/2022-individual-income-tax-forms/
          https://www.in.gov/dor/files/2022-county-tax-rates-and-codes.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Louisiana', 'state_occupation': 'Indiana', "capital_gains_short": capital_gains_short, "dependents": dependents, "sources": sources, "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def louisiana_and_iowa(self, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Louisiana:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Iowa:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Louisiana:
        - https://revenue.louisiana.gov/individualincometax
          https://revenue.louisiana.gov/individualincometax
          https://www.revenue.louisiana.gov/taxforms/6935(11_02)F.pdf
          https://revenue.louisiana.gov/individualincometax
          https://revenue.louisiana.gov/individualincometax
        
        Iowa:
        -         
        - https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
          https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
          https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
          https://tax.iowa.gov/iowa-illinois-reciprocal-agreement
          https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Louisiana', 'state_occupation': 'Iowa', "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def louisiana_and_kansas(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Louisiana:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Kansas:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Louisiana:
        - https://revenue.louisiana.gov/individualincometax
          https://revenue.louisiana.gov/individualincometax
          https://www.revenue.louisiana.gov/taxforms/6935(11_02)F.pdf
          https://revenue.louisiana.gov/individualincometax
          https://revenue.louisiana.gov/individualincometax
        
        Kansas:
        -         
        - https://www.ksrevenue.gov/taxrates.html
          https://www.ksrevenue.gov/incomebook22.html#0
          https://www.ksrevenue.gov/incomebook22.html#0
          https://www.ksrevenue.gov/incomebook22.html#0
          https://www.ksrevenue.gov/incomebook22.html
          https://www.ksrevenue.gov/incomebook22.html
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Louisiana', 'state_occupation': 'Kansas', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def louisiana_and_kentucky(self, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Louisiana:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Kentucky:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Louisiana:
        - https://revenue.louisiana.gov/individualincometax
          https://revenue.louisiana.gov/individualincometax
          https://www.revenue.louisiana.gov/taxforms/6935(11_02)F.pdf
          https://revenue.louisiana.gov/individualincometax
          https://revenue.louisiana.gov/individualincometax
        
        Kentucky:
        -         
        - https://revenue.ky.gov/News/Pages/Dor-Announces-Updates-to-Individual-Income-Tax-for-2023-Tax-Year.aspx
          https://revenue.ky.gov/Individual/Individual-Income-Tax/Pages/default.aspx
          https://revenue.ky.gov/News/Pages/Dor-Announces-Updates-to-Individual-Income-Tax-for-2023-Tax-Year.aspx
          https://revenue.ky.gov/Individual/Individual-Income-Tax/Pages/default.aspx
          https://revenue.ky.gov/Dor%20Training%20Materials/103%20KAR%2017.140.%20Individual%20income%20tax%20-%20reciprocity%20-%20nonresidents.pdf
          https://revenue.ky.gov/Individual/Individual-Income-Tax/Pages/default.aspx
          https://nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2021/Taxes-21-32.htm?taxmap=true
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Louisiana', 'state_occupation': 'Kentucky', "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def louisiana_and_maine(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Louisiana:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Maine:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Louisiana:
        - https://revenue.louisiana.gov/individualincometax
          https://revenue.louisiana.gov/individualincometax
          https://www.revenue.louisiana.gov/taxforms/6935(11_02)F.pdf
          https://revenue.louisiana.gov/individualincometax
          https://revenue.louisiana.gov/individualincometax
        
        Maine:
        -         
        - https://www.maine.gov/revenue/tax-return-forms/individual-income-tax-2022
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/ind_tax_rate_sched_2022.pdf
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/22_1040me_book_gen_instruc_revisedFeb23.pdf
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/22_1040me_book_gen_instruc_revisedFeb23.pdf
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/22_1040me_book_gen_instruc_revisedFeb23.pdf
          https://www.maine.gov/revenue/tax-return-forms/individual-income-tax-2022
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Louisiana', 'state_occupation': 'Maine', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def louisiana_and_maryland(self, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Louisiana:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Maryland:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Louisiana:
        - https://revenue.louisiana.gov/individualincometax
          https://revenue.louisiana.gov/individualincometax
          https://www.revenue.louisiana.gov/taxforms/6935(11_02)F.pdf
          https://revenue.louisiana.gov/individualincometax
          https://revenue.louisiana.gov/individualincometax
        
        Maryland:
        -         
        - https://www.marylandtaxes.gov/individual/income/tax-info/tax-rates.php
          https://www.marylandtaxes.gov/forms/22_forms/Resident_Booklet.pdf
          https://www.marylandtaxes.gov/forms/Tax_Publications/Administrative_Releases/Income_and_Estate_Tax_Releases/ar_it37.pdf
          https://www.marylandtaxes.gov/individual/index.php
          https://www.marylandtaxes.gov/individual/income/tax-info/tax-rates.php
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Louisiana', 'state_occupation': 'Maryland', "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def louisiana_and_massachusetts(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Louisiana:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Massachusetts:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Louisiana:
        - https://revenue.louisiana.gov/individualincometax
          https://revenue.louisiana.gov/individualincometax
          https://www.revenue.louisiana.gov/taxforms/6935(11_02)F.pdf
          https://revenue.louisiana.gov/individualincometax
          https://revenue.louisiana.gov/individualincometax
        
        Massachusetts:
        -         
        - https://www.mass.gov/service-details/massachusetts-tax-rates
          https://www.mass.gov/service-details/view-massachusetts-personal-income-tax-exemptions
          https://www.mass.gov/service-details/view-massachusetts-personal-income-tax-exemptions
          https://www.mass.gov/service-details/view-child-and-dependent-related-credits
          https://www.mass.gov/service-details/massachusetts-tax-rates
          https://www.mass.gov/personal-income-tax
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Louisiana', 'state_occupation': 'Massachusetts', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def louisiana_and_michigan(self, capital_gains_short=0, county_occupation='Other', dependents=0, sources=False, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_long=0, filing_status='single', income=100000, credits_state=0):
        """
        --- Parameters ----
        
        Louisiana:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Michigan:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        county_occupation : str, optional
            County of occupation. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Louisiana:
        - https://revenue.louisiana.gov/individualincometax
          https://revenue.louisiana.gov/individualincometax
          https://www.revenue.louisiana.gov/taxforms/6935(11_02)F.pdf
          https://revenue.louisiana.gov/individualincometax
          https://revenue.louisiana.gov/individualincometax
        
        Michigan:
        -         
        - https://www.michigan.gov/taxes/iit/new-developments-for-tax-year-2022
          "https://www.michigan.gov/taxes/-/media/Project/Websites/taxes/Forms/2023/2023-SUW/446_Withholding-Guide_2023.pdf?rev=177ca3d8cb034392bf84e317ed78c320&hash=19C5F42F43A89A91863D229349F48883#:~:text=The%20exemption%20amount%20is%20%245%2C400
         employees%20Michigan%20income%20tax%20return"
          "https://www.michigan.gov/taxes/-/media/Project/Websites/taxes/Forms/2023/2023-SUW/446_Withholding-Guide_2023.pdf?rev=177ca3d8cb034392bf84e317ed78c320&hash=19C5F42F43A89A91863D229349F48883#:~:text=The%20exemption%20amount%20is%20%245%2C400
         employees%20Michigan%20income%20tax%20return"
          https://www.michigan.gov/taxes/questions/iit/accordion/residency/are-my-wages-earned-in-another-state-taxable-in-michigan-if-i-am-a-michigan-resident-1
          https://www.michigan.gov/taxes/-/media/Project/Websites/taxes/Forms/2022/2022-IIT-Forms/BOOK_MI-1040.pdf?rev=4524f1cbbc0243b194ab05a6680be75b&hash=40DFE71E651F7C1D65C31F6CD03D944F
          https://www.michigan.gov/taxes/questions/iit/accordion/general/what-cities-impose-an-income-tax
        """
        
        # Check if the parameters are set upstream in the instance
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        county_occupation = self.county_occupation if hasattr(self, 'county_occupation') else county_occupation
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Louisiana', 'state_occupation': 'Michigan', "capital_gains_short": capital_gains_short, "county_occupation": county_occupation, "dependents": dependents, "sources": sources, "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def louisiana_and_minnesota(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Louisiana:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Minnesota:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Louisiana:
        - https://revenue.louisiana.gov/individualincometax
          https://revenue.louisiana.gov/individualincometax
          https://www.revenue.louisiana.gov/taxforms/6935(11_02)F.pdf
          https://revenue.louisiana.gov/individualincometax
          https://revenue.louisiana.gov/individualincometax
        
        Minnesota:
        -         
        - https://www.revenue.state.mn.us/minnesota-income-tax-rates-and-brackets
          https://www.revenue.state.mn.us/child-and-dependent-care-credit
          https://www.revenue.state.mn.us/dependent-exemptions
          https://www.revenue.state.mn.us/minnesota-standard-deduction
          https://mn.gov/mmb/assets/mwr_form_tcm1059-128581.pdf
          https://www.revenue.state.mn.us/individual-income-tax
          https://www.revenue.state.mn.us/individual-income-tax
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Louisiana', 'state_occupation': 'Minnesota', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def louisiana_and_mississippi(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Louisiana:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Mississippi:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Louisiana:
        - https://revenue.louisiana.gov/individualincometax
          https://revenue.louisiana.gov/individualincometax
          https://www.revenue.louisiana.gov/taxforms/6935(11_02)F.pdf
          https://revenue.louisiana.gov/individualincometax
          https://revenue.louisiana.gov/individualincometax
        
        Mississippi:
        -         
        - https://www.dor.ms.gov/individual/tax-rates
          https://www.dor.ms.gov/individual/tax-rates
          https://www.dor.ms.gov/individual/tax-rates
          https://www.dor.ms.gov/individual
          https://www.dor.ms.gov/individual
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Louisiana', 'state_occupation': 'Mississippi', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def louisiana_and_missouri(self, capital_gains_short=0, dependents=0, sources=False, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_long=0, filing_status='single', income=100000, credits_state=0):
        """
        --- Parameters ----
        
        Louisiana:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Missouri:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Louisiana:
        - https://revenue.louisiana.gov/individualincometax
          https://revenue.louisiana.gov/individualincometax
          https://www.revenue.louisiana.gov/taxforms/6935(11_02)F.pdf
          https://revenue.louisiana.gov/individualincometax
          https://revenue.louisiana.gov/individualincometax
        
        Missouri:
        -         
        - https://dor.mo.gov/taxation/individual/tax-types/income/year-changes/
          https://dor.mo.gov/taxation/individual/tax-types/income/year-changes/
          https://dor.mo.gov/taxation/individual/tax-types/income/year-changes/
          https://www.kcmo.gov/city-hall/departments/finance/earnings-tax
          https://www.stlouis-mo.gov/government/departments/collector/earnings-tax/individual-earnings-tax-info.cfm
        """
        
        # Check if the parameters are set upstream in the instance
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Louisiana', 'state_occupation': 'Missouri', "capital_gains_short": capital_gains_short, "dependents": dependents, "sources": sources, "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def louisiana_and_montana(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Louisiana:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Montana:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Louisiana:
        - https://revenue.louisiana.gov/individualincometax
          https://revenue.louisiana.gov/individualincometax
          https://www.revenue.louisiana.gov/taxforms/6935(11_02)F.pdf
          https://revenue.louisiana.gov/individualincometax
          https://revenue.louisiana.gov/individualincometax
        
        Montana:
        -         
        - https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/2022/12/montana-employees-withholding-allowance-and-exemption-certificate-form-mw-4.pdf
          https://mtrevenue.gov/2021/11/29/simplification-of-montana-income-taxation/?utm_source=rss&utm_medium=rss&utm_campaign=simplification-of-montana-income-taxation
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Louisiana', 'state_occupation': 'Montana', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def louisiana_and_nebraska(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Louisiana:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Nebraska:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Louisiana:
        - https://revenue.louisiana.gov/individualincometax
          https://revenue.louisiana.gov/individualincometax
          https://www.revenue.louisiana.gov/taxforms/6935(11_02)F.pdf
          https://revenue.louisiana.gov/individualincometax
          https://revenue.louisiana.gov/individualincometax
        
        Nebraska:
        -         
        - https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2023_Circular_EN_8-429r2023_Whole_Book_11.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Louisiana', 'state_occupation': 'Nebraska', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def louisiana_and_nevada(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Louisiana:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Nevada:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Louisiana:
        - https://revenue.louisiana.gov/individualincometax
          https://revenue.louisiana.gov/individualincometax
          https://www.revenue.louisiana.gov/taxforms/6935(11_02)F.pdf
          https://revenue.louisiana.gov/individualincometax
          https://revenue.louisiana.gov/individualincometax
        
        Nevada:
        -         
        - https://www.nvsos.gov/sos/businesses/the-nevada-advantage
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Louisiana', 'state_occupation': 'Nevada', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def louisiana_and_new_hampshire(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Louisiana:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        New Hampshire:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Louisiana:
        - https://revenue.louisiana.gov/individualincometax
          https://revenue.louisiana.gov/individualincometax
          https://www.revenue.louisiana.gov/taxforms/6935(11_02)F.pdf
          https://revenue.louisiana.gov/individualincometax
          https://revenue.louisiana.gov/individualincometax
        
        New Hampshire:
        -         
        - https://www.revenue.nh.gov/assistance/tax-overview.htm
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Louisiana', 'state_occupation': 'NewHampshire', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def louisiana_and_new_jersey(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Louisiana:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        New Jersey:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Louisiana:
        - https://revenue.louisiana.gov/individualincometax
          https://revenue.louisiana.gov/individualincometax
          https://www.revenue.louisiana.gov/taxforms/6935(11_02)F.pdf
          https://revenue.louisiana.gov/individualincometax
          https://revenue.louisiana.gov/individualincometax
        
        New Jersey:
        -         
        - https://www.state.nj.us/treasury/taxation/taxtables.shtml
          https://www.state.nj.us/treasury/taxation/njit13.shtml
          https://www.nj.gov/treasury/taxation/njit25.shtml
          https://www.state.nj.us/treasury/taxation/individuals/freqqiti.shtml
          https://www.state.nj.us/treasury/taxation/localtax.shtml
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Louisiana', 'state_occupation': 'NewJersey', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def louisiana_and_new_mexico(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Louisiana:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        New Mexico:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Louisiana:
        - https://revenue.louisiana.gov/individualincometax
          https://revenue.louisiana.gov/individualincometax
          https://www.revenue.louisiana.gov/taxforms/6935(11_02)F.pdf
          https://revenue.louisiana.gov/individualincometax
          https://revenue.louisiana.gov/individualincometax
        
        New Mexico:
        -         
        - https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/fdf3c548-8aba-4b9c-9eb4-bb564c716015/FYI-104.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/0bf943fd-652e-4400-bb1b-cd8397eb5a95/2022pit-adj.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/2f1a6781-9534-4436-b427-1557f9592099/2022pit-adj-ins.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/0bf943fd-652e-4400-bb1b-cd8397eb5a95/2022pit-adj.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/1afc56af-ea90-4d48-82e5-1f9aeb43255a/PITbook2022.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/fdf3c548-8aba-4b9c-9eb4-bb564c716015/FYI-104.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Louisiana', 'state_occupation': 'NewMexico', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def louisiana_and_new_york(self, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Louisiana:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        New York:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Louisiana:
        - https://revenue.louisiana.gov/individualincometax
          https://revenue.louisiana.gov/individualincometax
          https://www.revenue.louisiana.gov/taxforms/6935(11_02)F.pdf
          https://revenue.louisiana.gov/individualincometax
          https://revenue.louisiana.gov/individualincometax
        
        New York:
        -         
        - https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nys-tax-rate-schedule
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nys-tax-rate-schedule
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nys-tax-rate-schedule
          https://www.tax.ny.gov/pit/
          https://www.tax.ny.gov/pit/file/nonresident-faqs.htm
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#yonkers-tax-rate-schedule
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nyc-tax-rate-schedule
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Louisiana', 'state_occupation': 'NewYork', "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def louisiana_and_north_carolina(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Louisiana:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        North Carolina:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Louisiana:
        - https://revenue.louisiana.gov/individualincometax
          https://revenue.louisiana.gov/individualincometax
          https://www.revenue.louisiana.gov/taxforms/6935(11_02)F.pdf
          https://revenue.louisiana.gov/individualincometax
          https://revenue.louisiana.gov/individualincometax
        
        North Carolina:
        -         
        - https://www.ncdor.gov/taxes-forms/tax-rate-schedules
          https://www.ncdor.gov/taxes-forms/individual-income-tax/north-carolina-standard-deduction-or-north-carolina-itemized-deductions
          https://www.ncdor.gov/taxes-forms/individual-income-tax
          https://www.ncdor.gov/taxes-forms/individual-income-tax
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Louisiana', 'state_occupation': 'NorthCarolina', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def louisiana_and_north_dakota(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Louisiana:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        North Dakota:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Louisiana:
        - https://revenue.louisiana.gov/individualincometax
          https://revenue.louisiana.gov/individualincometax
          https://www.revenue.louisiana.gov/taxforms/6935(11_02)F.pdf
          https://revenue.louisiana.gov/individualincometax
          https://revenue.louisiana.gov/individualincometax
        
        North Dakota:
        -         
        - https://www.tax.nd.gov/forms
          https://www.tax.nd.gov/sites/www/files/documents/forms/individual/2022-iit/2022-individual-income-tax-booklet.pdf
          https://www.tax.nd.gov/sites/www/files/documents/forms/2022-individual-income-tax-booklet.pdf
          https://www.tax.nd.gov/tax-exemptions-credits/income-tax-exemptions-credits
          https://www.tax.nd.gov/sites/www/files/documents/forms/business/ndwrfillable.pdf
          https://www.tax.nd.gov/news/tax-legislative-changes/significant-changes-law/individual-income-tax-history
          https://www.tax.nd.gov/forms
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Louisiana', 'state_occupation': 'NorthDakota', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def louisiana_and_ohio(self, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Louisiana:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Ohio:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Louisiana:
        - https://revenue.louisiana.gov/individualincometax
          https://revenue.louisiana.gov/individualincometax
          https://www.revenue.louisiana.gov/taxforms/6935(11_02)F.pdf
          https://revenue.louisiana.gov/individualincometax
          https://revenue.louisiana.gov/individualincometax
        
        Ohio:
        -         
        - https://tax.ohio.gov/individual/resources/annual-tax-rates
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://tax.ohio.gov/static/forms/employer_withholding/generic/wth_it4nr.pdf
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://www.ritaohio.com/TaxRatesTable
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Louisiana', 'state_occupation': 'Ohio', "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def louisiana_and_oklahoma(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Louisiana:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Oklahoma:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Louisiana:
        - https://revenue.louisiana.gov/individualincometax
          https://revenue.louisiana.gov/individualincometax
          https://www.revenue.louisiana.gov/taxforms/6935(11_02)F.pdf
          https://revenue.louisiana.gov/individualincometax
          https://revenue.louisiana.gov/individualincometax
        
        Oklahoma:
        -         
        - https://oklahoma.gov/content/dam/ok/en/tax/documents/resources/publications/businesses/withholding-tables/WHTables-2022.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Louisiana', 'state_occupation': 'Oklahoma', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def louisiana_and_oregon(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Louisiana:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Oregon:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Louisiana:
        - https://revenue.louisiana.gov/individualincometax
          https://revenue.louisiana.gov/individualincometax
          https://www.revenue.louisiana.gov/taxforms/6935(11_02)F.pdf
          https://revenue.louisiana.gov/individualincometax
          https://revenue.louisiana.gov/individualincometax
        
        Oregon:
        -         
        - https://www.oregon.gov/dor/programs/individuals/Pages/PIT.aspx
          https://www.oregon.gov/dor/programs/individuals/Documents/Full%20year%20resident
         %20Form%20or-40%20filers.pdf
          https://www.oregon.gov/dor/programs/individuals/Documents/Part-year%20and%20nonresident
         %20Form%20or-40-P%20and%20or-40-N%20filers.pdf.pdf
          https://www.oregon.gov/dor/forms/FormsPubs/form-or-40_101-040_2022.pdf
          https://www.oregon.gov/dor/forms/FormsPubs/form-or-40_101-040_2022.pdf
          https://www.oregon.gov/dor/forms/FormsPubs/form-or-40_101-040_2022.pdf
          https://www.portland.gov/revenue/personal-tax
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Louisiana', 'state_occupation': 'Oregon', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def louisiana_and_pennsylvania(self, county_residence='Other', state_agi=75000, county_occupation='Other', filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Louisiana:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Pennsylvania:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        county_residence : str, optional
            County of residence. Default is Other.
        county_occupation : str, optional
            County of occupation. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Louisiana:
        - https://revenue.louisiana.gov/individualincometax
          https://revenue.louisiana.gov/individualincometax
          https://www.revenue.louisiana.gov/taxforms/6935(11_02)F.pdf
          https://revenue.louisiana.gov/individualincometax
          https://revenue.louisiana.gov/individualincometax
        
        Pennsylvania:
        -         
        - https://www.revenue.pa.gov/Tax%20Rates/Pages/default.aspx
          https://www.revenue.pa.gov/FormsandPublications/FormsforIndividuals/PIT/Documents/2022/2022_pa-40in.pdf
          https://www.revenue.pa.gov/FormsandPublications/FormsforBusinesses/EmployerWithholding/Documents/rev-419.pdf
          https://www.revenue.pa.gov/FormsandPublications/FormsforIndividuals/PIT/Documents/2022/2022_pa-40in.pdf
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-25.htm
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        county_occupation = self.county_occupation if hasattr(self, 'county_occupation') else county_occupation
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Louisiana', 'state_occupation': 'Pennsylvania', "county_residence": county_residence, "state_agi": state_agi, "county_occupation": county_occupation, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def louisiana_and_rhode_island(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Louisiana:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Rhode Island:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Louisiana:
        - https://revenue.louisiana.gov/individualincometax
          https://revenue.louisiana.gov/individualincometax
          https://www.revenue.louisiana.gov/taxforms/6935(11_02)F.pdf
          https://revenue.louisiana.gov/individualincometax
          https://revenue.louisiana.gov/individualincometax
        
        Rhode Island:
        -         
        - https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2022_1040WE_w_0.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Louisiana', 'state_occupation': 'RhodeIsland', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def louisiana_and_south_carolina(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Louisiana:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        South Carolina:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Louisiana:
        - https://revenue.louisiana.gov/individualincometax
          https://revenue.louisiana.gov/individualincometax
          https://www.revenue.louisiana.gov/taxforms/6935(11_02)F.pdf
          https://revenue.louisiana.gov/individualincometax
          https://revenue.louisiana.gov/individualincometax
        
        South Carolina:
        -         
        - https://dor.sc.gov/resources-site/lawandpolicy/Advisory%20Opinions/IL22-15.pdf
          https://dor.sc.gov/forms-site/Forms/IITPacket_2022.pdf
          https://www.scstatehouse.gov/code/t12c006.php
          https://dor.sc.gov/resources-site/lawandpolicy/Advisory%20Opinions/IL22-15.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Louisiana', 'state_occupation': 'SouthCarolina', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def louisiana_and_south_dakota(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Louisiana:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        South Dakota:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Louisiana:
        - https://revenue.louisiana.gov/individualincometax
          https://revenue.louisiana.gov/individualincometax
          https://www.revenue.louisiana.gov/taxforms/6935(11_02)F.pdf
          https://revenue.louisiana.gov/individualincometax
          https://revenue.louisiana.gov/individualincometax
        
        South Dakota:
        -         
        - https://dor.sd.gov/individuals/taxes/
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Louisiana', 'state_occupation': 'SouthDakota', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def louisiana_and_tennessee(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Louisiana:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Tennessee:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Louisiana:
        - https://revenue.louisiana.gov/individualincometax
          https://revenue.louisiana.gov/individualincometax
          https://www.revenue.louisiana.gov/taxforms/6935(11_02)F.pdf
          https://revenue.louisiana.gov/individualincometax
          https://revenue.louisiana.gov/individualincometax
        
        Tennessee:
        -         
        - https://www.tn.gov/revenue/taxes.html
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Louisiana', 'state_occupation': 'Tennessee', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def louisiana_and_texas(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Louisiana:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Texas:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Louisiana:
        - https://revenue.louisiana.gov/individualincometax
          https://revenue.louisiana.gov/individualincometax
          https://www.revenue.louisiana.gov/taxforms/6935(11_02)F.pdf
          https://revenue.louisiana.gov/individualincometax
          https://revenue.louisiana.gov/individualincometax
        
        Texas:
        -         
        - https://comptroller.texas.gov/taxes/
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Louisiana', 'state_occupation': 'Texas', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def louisiana_and_utah(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Louisiana:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Utah:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Louisiana:
        - https://revenue.louisiana.gov/individualincometax
          https://revenue.louisiana.gov/individualincometax
          https://www.revenue.louisiana.gov/taxforms/6935(11_02)F.pdf
          https://revenue.louisiana.gov/individualincometax
          https://revenue.louisiana.gov/individualincometax
        
        Utah:
        -         
        - https://incometax.utah.gov/paying/tax-rates
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Louisiana', 'state_occupation': 'Utah', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def louisiana_and_vermont(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Louisiana:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Vermont:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Louisiana:
        - https://revenue.louisiana.gov/individualincometax
          https://revenue.louisiana.gov/individualincometax
          https://www.revenue.louisiana.gov/taxforms/6935(11_02)F.pdf
          https://revenue.louisiana.gov/individualincometax
          https://revenue.louisiana.gov/individualincometax
        
        Vermont:
        -         
        - https://tax.vermont.gov/sites/tax/files/documents/GB-1210-2023.pdf
          https://tax.vermont.gov/sites/tax/files/documents/Income%20Booklet-2022.pdf
          https://tax.vermont.gov/sites/tax/files/documents/Income%20Booklet-2022.pdf
          https://tax.vermont.gov/sites/tax/files/documents/Income%20Booklet-2022.pdf
          https://tax.vermont.gov/individuals/personal-income-tax/tax-credits
          https://tax.vermont.gov/individuals/personal-income-tax/taxable-income
          https://tax.vermont.gov/individuals/personal-income-tax/local-option-tax
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Louisiana', 'state_occupation': 'Vermont', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def louisiana_and_virginia(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Louisiana:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Virginia:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Louisiana:
        - https://revenue.louisiana.gov/individualincometax
          https://revenue.louisiana.gov/individualincometax
          https://www.revenue.louisiana.gov/taxforms/6935(11_02)F.pdf
          https://revenue.louisiana.gov/individualincometax
          https://revenue.louisiana.gov/individualincometax
        
        Virginia:
        -         
        - https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/reciprocity
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Louisiana', 'state_occupation': 'Virginia', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def louisiana_and_washington(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Louisiana:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Washington:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Louisiana:
        - https://revenue.louisiana.gov/individualincometax
          https://revenue.louisiana.gov/individualincometax
          https://www.revenue.louisiana.gov/taxforms/6935(11_02)F.pdf
          https://revenue.louisiana.gov/individualincometax
          https://revenue.louisiana.gov/individualincometax
        
        Washington:
        -         
        - https://dor.wa.gov/taxes-rates/income-tax
          https://dor.wa.gov/taxes-rates/other-taxes/capital-gains-tax
          https://dor.wa.gov/taxes-rates/income-tax
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Louisiana', 'state_occupation': 'Washington', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def louisiana_and_west_virginia(self, pay_periods=24, capital_gains_short=0, county_occupation='Other', dependents=0, sources=False, state_agi=75000, filers_over_65=0, capital_gains_long=0, filing_status='single', income=100000, credits_state=0):
        """
        --- Parameters ----
        
        Louisiana:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        West Virginia:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_occupation : str, optional
            County of occupation. Default is Other.
        pay_periods : int, optional
            Number of pay periods. Options are 13, 24, 26, 52. Default is 24.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Louisiana:
        - https://revenue.louisiana.gov/individualincometax
          https://revenue.louisiana.gov/individualincometax
          https://www.revenue.louisiana.gov/taxforms/6935(11_02)F.pdf
          https://revenue.louisiana.gov/individualincometax
          https://revenue.louisiana.gov/individualincometax
        
        West Virginia:
        -         
        - https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://tax.wv.gov/Documents/TSD/tsd438.pdf
          https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2011/Taxes-11-35att.html
        """
        
        # Check if the parameters are set upstream in the instance
        pay_periods = self.pay_periods if hasattr(self, 'pay_periods') else pay_periods
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        county_occupation = self.county_occupation if hasattr(self, 'county_occupation') else county_occupation
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Louisiana', 'state_occupation': 'WestVirginia', "pay_periods": pay_periods, "capital_gains_short": capital_gains_short, "county_occupation": county_occupation, "dependents": dependents, "sources": sources, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def louisiana_and_wisconsin(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Louisiana:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Wisconsin:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Louisiana:
        - https://revenue.louisiana.gov/individualincometax
          https://revenue.louisiana.gov/individualincometax
          https://www.revenue.louisiana.gov/taxforms/6935(11_02)F.pdf
          https://revenue.louisiana.gov/individualincometax
          https://revenue.louisiana.gov/individualincometax
        
        Wisconsin:
        -         
        - https://www.revenue.wi.gov/TaxForms2023/2023-Form1-ES-Inst.pdf
          https://www.revenue.wi.gov/TaxForms2023/2023-Form1-ES-Inst.pdf
          https://www.revenue.wi.gov/TaxForms2023/2023-Form1-ES-Inst.pdf
          https://www.revenue.wi.gov/Pages/FAQS/pcs-work.aspx
          https://www.revenue.wi.gov/Dor%20Publications/pb103.pdf
          https://www.revenue.wi.gov/Pages/FAQS/pcs-work.aspx
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Louisiana', 'state_occupation': 'Wisconsin', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def louisiana_and_wyoming(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Louisiana:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Wyoming:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Louisiana:
        - https://revenue.louisiana.gov/individualincometax
          https://revenue.louisiana.gov/individualincometax
          https://www.revenue.louisiana.gov/taxforms/6935(11_02)F.pdf
          https://revenue.louisiana.gov/individualincometax
          https://revenue.louisiana.gov/individualincometax
        
        Wyoming:
        -         
        - https://revenue.wyo.gov/
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Louisiana', 'state_occupation': 'Wyoming', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def maine_and_alabama(self, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Maine:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Alabama:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Maine:
        - https://www.maine.gov/revenue/tax-return-forms/individual-income-tax-2022
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/ind_tax_rate_sched_2022.pdf
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/22_1040me_book_gen_instruc_revisedFeb23.pdf
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/22_1040me_book_gen_instruc_revisedFeb23.pdf
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/22_1040me_book_gen_instruc_revisedFeb23.pdf
          https://www.maine.gov/revenue/tax-return-forms/individual-income-tax-2022
        
        Alabama:
        -         
        - https://www.revenue.alabama.gov/ultraviewer/viewer/basic_viewer/index.html?form=2023/02/22f40.pdf
          https://www.revenue.alabama.gov/faqs/what-is-alabamas-individual-income-tax-rate/
          https://www.revenue.alabama.gov/ultraviewer/viewer/basic_viewer/index.html?form=2023/02/22f40.pdf
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-24.htm?taxmap=true
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-24.htm?taxmap=true
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-24.htm?taxmap=true
          https://www.revenue.alabama.gov/individual-corporate/income-to-be-reported-on-the-alabama-income-tax-return/
          https://www.almonline.org/TaxRates.aspx#OccupationalTax
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Maine', 'state_occupation': 'Alabama', "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def maine_and_alaska(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Maine:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Alaska:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Maine:
        - https://www.maine.gov/revenue/tax-return-forms/individual-income-tax-2022
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/ind_tax_rate_sched_2022.pdf
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/22_1040me_book_gen_instruc_revisedFeb23.pdf
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/22_1040me_book_gen_instruc_revisedFeb23.pdf
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/22_1040me_book_gen_instruc_revisedFeb23.pdf
          https://www.maine.gov/revenue/tax-return-forms/individual-income-tax-2022
        
        Alaska:
        -         
        - https://www.commerce.alaska.gov/web/dcra/officeofthestateassessor/alaskataxfacts.aspx
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Maine', 'state_occupation': 'Alaska', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def maine_and_arizona(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Maine:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Arizona:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        credits_state : int, optional
            Credits for state taxes. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Maine:
        - https://www.maine.gov/revenue/tax-return-forms/individual-income-tax-2022
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/ind_tax_rate_sched_2022.pdf
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/22_1040me_book_gen_instruc_revisedFeb23.pdf
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/22_1040me_book_gen_instruc_revisedFeb23.pdf
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/22_1040me_book_gen_instruc_revisedFeb23.pdf
          https://www.maine.gov/revenue/tax-return-forms/individual-income-tax-2022
        
        Arizona:
        -         
        - https://azdor.gov/forms/individual/form-140-x-y-tables
          https://azdor.gov/forms/individual/form-140-x-y-tables
          https://azdor.gov/forms/individual/form-140-x-y-tables
          https://azdor.gov/businesses-arizona/withholding-tax/withholding-exceptions
          "https://azdor.gov/individuals/income-tax-filing-assistance/identifying-other-taxable-income#:~:text=In%20regards%20to%20capital%20gains
         the%20individuals%20regular%20tax%20rate.
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Maine', 'state_occupation': 'Arizona', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def maine_and_arkansas(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Maine:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Arkansas:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Maine:
        - https://www.maine.gov/revenue/tax-return-forms/individual-income-tax-2022
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/ind_tax_rate_sched_2022.pdf
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/22_1040me_book_gen_instruc_revisedFeb23.pdf
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/22_1040me_book_gen_instruc_revisedFeb23.pdf
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/22_1040me_book_gen_instruc_revisedFeb23.pdf
          https://www.maine.gov/revenue/tax-return-forms/individual-income-tax-2022
        
        Arkansas:
        -         
        - https://www.arkansasedc.com/why-arkansas/business-climate/tax-structure/personal-income-tax
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_and_AR1000NR_Instructions.pdf
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_FullYearResidentIndividualIncomeTaxReturn.pdf
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_and_AR1000NR_Instructions.pdf
          https://www.arkansasedc.com/why-arkansas/business-climate/tax-structure/capital-gains-tax-reduction
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_and_AR1000NR_Instructions.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Maine', 'state_occupation': 'Arkansas', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def maine_and_california(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Maine:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        California:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Maine:
        - https://www.maine.gov/revenue/tax-return-forms/individual-income-tax-2022
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/ind_tax_rate_sched_2022.pdf
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/22_1040me_book_gen_instruc_revisedFeb23.pdf
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/22_1040me_book_gen_instruc_revisedFeb23.pdf
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/22_1040me_book_gen_instruc_revisedFeb23.pdf
          https://www.maine.gov/revenue/tax-return-forms/individual-income-tax-2022
        
        California:
        -         
        - https://www.ftb.ca.gov/forms/2022/2022-540-tax-rate-schedules.pdf
          https://www.ftb.ca.gov/file/personal/deductions/index.html
          https://www.ftb.ca.gov/about-ftb/newsroom/caleitc/eligibility-and-credit-information.html
          https://www.ftb.ca.gov/file/personal/income-types/capital-gains-and-losses.html#:~:text=California%20does%20not%20have%20a
         are%20taxed%20as%20ordinary%20income
          https://www.ftb.ca.gov/file/personal/
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Maine', 'state_occupation': 'California', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def maine_and_colorado(self, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Maine:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Colorado:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Maine:
        - https://www.maine.gov/revenue/tax-return-forms/individual-income-tax-2022
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/ind_tax_rate_sched_2022.pdf
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/22_1040me_book_gen_instruc_revisedFeb23.pdf
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/22_1040me_book_gen_instruc_revisedFeb23.pdf
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/22_1040me_book_gen_instruc_revisedFeb23.pdf
          https://www.maine.gov/revenue/tax-return-forms/individual-income-tax-2022
        
        Colorado:
        -         
        - https://leg.colorado.gov/agencies/legislative-council-staff/individual-income-tax%C2%A0
          https://leg.colorado.gov/agencies/legislative-council-staff/individual-income-tax%C2%A0
          https://tax.colorado.gov/sites/tax/files/documents/ITT_Capital_Gain_Subtraction_Feb_2022.pdf
          https://www.auroragov.org/business_services/taxes/occupational_privilege_tax
          https://www.denvergov.org/Government/Agencies-Departments-Offices/Agencies-Departments-Offices-Directory/Department-of-Finance/Our-Divisions/Treasury/Business-Tax-Information#:~:text=Occupational%20privilege%20tax
         -%28show%20below%29&text=Each%20taxable%20employee%20is%20liable
         month%20for%20each%20taxable%20employee
          https://www.glendale.co.us/355/Occupational-Privilege-Tax
          https://www.greenwoodvillage.com/1220/Occupational-Privilege-Tax-OPT
          https://ci.sheridan.co.us/288/Occupational-Privilege-Tax
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Maine', 'state_occupation': 'Colorado', "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def maine_and_connecticut(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Maine:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Connecticut:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Maine:
        - https://www.maine.gov/revenue/tax-return-forms/individual-income-tax-2022
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/ind_tax_rate_sched_2022.pdf
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/22_1040me_book_gen_instruc_revisedFeb23.pdf
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/22_1040me_book_gen_instruc_revisedFeb23.pdf
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/22_1040me_book_gen_instruc_revisedFeb23.pdf
          https://www.maine.gov/revenue/tax-return-forms/individual-income-tax-2022
        
        Connecticut:
        -         
        - https://portal.ct.gov/-/media/DRS/Forms/2022/Income/2022-CT-1040-Instructions_1222.pdf
          https://www.cga.ct.gov/2022/rpt/pdf/2022-R-0108.pdf
          https://www.cga.ct.gov/current/pub/chap_229.htm#sec_12-703
          https://portal.ct.gov/-/media/DRS/Forms/2022/Income/2022-CT-1040-Instructions_1222.pdf
          "https://portal.ct.gov/DRS/Publications/TSSNs/TSSN-29#:~:text=An%20individuals%20net%20capital%20gains
         portion%20of%20Social%20Security%20benefits"
          "https://portal.ct.gov/DRS/Publications/TSSNs/TSSN-29#:~:text=An%20individuals%20net%20capital%20gains
         portion%20of%20Social%20Security%20benefits"
          https://portal.ct.gov/DRS/Individuals/Resident-Income-Tax/Tax-Information
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Maine', 'state_occupation': 'Connecticut', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def maine_and_delaware(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, federal_agi=75000, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Maine:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Delaware:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        federal_agi : int, optional
            Adjusted Gross Income (AGI) for federal taxes. Default is 0. Sources - https://www.irs.gov/e-file-providers/definition-of-adjusted-gross-income, https://www.irs.gov/newsroom/taxpayers-should-know-the-difference-between-standard-and-itemized-deductions
        income : int, optional
            Gross income. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Maine:
        - https://www.maine.gov/revenue/tax-return-forms/individual-income-tax-2022
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/ind_tax_rate_sched_2022.pdf
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/22_1040me_book_gen_instruc_revisedFeb23.pdf
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/22_1040me_book_gen_instruc_revisedFeb23.pdf
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/22_1040me_book_gen_instruc_revisedFeb23.pdf
          https://www.maine.gov/revenue/tax-return-forms/individual-income-tax-2022
        
        Delaware:
        -         
        - https://revenue.delaware.gov/software-developer/tax-rate-changes/
          https://revenuefiles.delaware.gov/2022/PIT-RES_TY22_2022-01_PaperInteractive.pdf
          https://revenuefiles.delaware.gov/2022/PIT-RES_TY22_2022-02_Instructions.pdf
          https://revenue.delaware.gov/frequently-asked-questions/personal-income-tax-faqs/
          https://www.newcastlede.gov/DocumentCenter/View/20147/City-of-Wilmington-Wage-Tax-Information-PDF
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        federal_agi = self.federal_agi if hasattr(self, 'federal_agi') else federal_agi
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Maine', 'state_occupation': 'Delaware', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "federal_agi": federal_agi, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def maine_and_florida(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Maine:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Florida:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Maine:
        - https://www.maine.gov/revenue/tax-return-forms/individual-income-tax-2022
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/ind_tax_rate_sched_2022.pdf
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/22_1040me_book_gen_instruc_revisedFeb23.pdf
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/22_1040me_book_gen_instruc_revisedFeb23.pdf
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/22_1040me_book_gen_instruc_revisedFeb23.pdf
          https://www.maine.gov/revenue/tax-return-forms/individual-income-tax-2022
        
        Florida:
        -         
        - https://floridarevenue.com/taxes/taxesfees/Pages/tax_interest_rates.aspx
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Maine', 'state_occupation': 'Florida', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def maine_and_georgia(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Maine:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Georgia:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Maine:
        - https://www.maine.gov/revenue/tax-return-forms/individual-income-tax-2022
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/ind_tax_rate_sched_2022.pdf
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/22_1040me_book_gen_instruc_revisedFeb23.pdf
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/22_1040me_book_gen_instruc_revisedFeb23.pdf
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/22_1040me_book_gen_instruc_revisedFeb23.pdf
          https://www.maine.gov/revenue/tax-return-forms/individual-income-tax-2022
        
        Georgia:
        -         
        - https://dor.georgia.gov/tax-tables-georgia-tax-rate-schedule
          https://apps.dor.ga.gov/FillableForms/PDFViewer/Index?form=2022GA500
          https://apps.dor.ga.gov/FillableForms/PDFViewer/Index?form=2022GA500
          https://apps.dor.ga.gov/FillableForms/PDFViewer/Index?form=2022GA500
          https://dor.georgia.gov/years-individual-income-tax-forms
          https://dor.georgia.gov/years-individual-income-tax-forms
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Maine', 'state_occupation': 'Georgia', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def maine_and_hawaii(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Maine:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Hawaii:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Maine:
        - https://www.maine.gov/revenue/tax-return-forms/individual-income-tax-2022
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/ind_tax_rate_sched_2022.pdf
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/22_1040me_book_gen_instruc_revisedFeb23.pdf
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/22_1040me_book_gen_instruc_revisedFeb23.pdf
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/22_1040me_book_gen_instruc_revisedFeb23.pdf
          https://www.maine.gov/revenue/tax-return-forms/individual-income-tax-2022
        
        Hawaii:
        -         
        - https://tax.hawaii.gov/forms/d_18table-on/d_18table-on_p13/
          https://files.hawaii.gov/tax/forms/2018/18table-on.pdf
          https://files.hawaii.gov/tax/news/pubs/22outline.pdf
          https://files.hawaii.gov/tax/news/pubs/FastTaxRefGuide_2017.pdf
          https://www.tfhawaii.org/wordpress/state-tax-resources/mini-tax-guides/individual-income-tax-chapter-235/#:~:text=Long%20term%20capital%20gains%20are%20taxed%20at%20a%20maximum%20of%207.25%25
          https://www.grassrootinstitute.org/2023/02/hb337-increasing-tax-on-capital-gains-a-bad-risk/
          https://files.hawaii.gov/tax/news/pubs/22outline.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Maine', 'state_occupation': 'Hawaii', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def maine_and_idaho(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Maine:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Idaho:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Maine:
        - https://www.maine.gov/revenue/tax-return-forms/individual-income-tax-2022
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/ind_tax_rate_sched_2022.pdf
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/22_1040me_book_gen_instruc_revisedFeb23.pdf
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/22_1040me_book_gen_instruc_revisedFeb23.pdf
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/22_1040me_book_gen_instruc_revisedFeb23.pdf
          https://www.maine.gov/revenue/tax-return-forms/individual-income-tax-2022
        
        Idaho:
        -         
        - https://tax.idaho.gov/taxes/income-tax/individual-income/individual-income-tax-rate-schedule/
          https://tax.idaho.gov/wp-content/uploads/forms/EIN00046/EIN00046_03-01-2023.pdf
          https://tax.idaho.gov/taxes/income-tax/individual-income/idaho-source-income/
          https://tax.idaho.gov/taxes/income-tax/individual-income/idaho-source-income/
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Maine', 'state_occupation': 'Idaho', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def maine_and_illinois(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Maine:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Illinois:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Maine:
        - https://www.maine.gov/revenue/tax-return-forms/individual-income-tax-2022
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/ind_tax_rate_sched_2022.pdf
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/22_1040me_book_gen_instruc_revisedFeb23.pdf
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/22_1040me_book_gen_instruc_revisedFeb23.pdf
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/22_1040me_book_gen_instruc_revisedFeb23.pdf
          https://www.maine.gov/revenue/tax-return-forms/individual-income-tax-2022
        
        Illinois:
        -         
        - https://tax.illinois.gov/research/taxrates/income.html#IndividualIncome
          https://tax.illinois.gov/content/dam/soi/en/web/tax/forms/incometax/documents/currentyear/individual/il-1040-instr.pdf
          https://tax.illinois.gov/questionsandanswers/12.html
          https://tax.illinois.gov/content/dam/soi/en/web/tax/forms/incometax/documents/currentyear/individual/il-1040-instr.pdf
          https://tax.illinois.gov/research/taxrates/personalproperty.html
          https://tax.illinois.gov/localgovernments/income.html
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Maine', 'state_occupation': 'Illinois', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def maine_and_indiana(self, capital_gains_short=0, dependents=0, sources=False, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_long=0, filing_status='single', income=100000, credits_state=0):
        """
        --- Parameters ----
        
        Maine:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Indiana:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Maine:
        - https://www.maine.gov/revenue/tax-return-forms/individual-income-tax-2022
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/ind_tax_rate_sched_2022.pdf
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/22_1040me_book_gen_instruc_revisedFeb23.pdf
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/22_1040me_book_gen_instruc_revisedFeb23.pdf
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/22_1040me_book_gen_instruc_revisedFeb23.pdf
          https://www.maine.gov/revenue/tax-return-forms/individual-income-tax-2022
        
        Indiana:
        -         
        - https://www.in.gov/dor/tax-forms/2022-individual-income-tax-forms/
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-32.htm
          https://www.in.gov/dor/files/ib28.pdf
          https://www.in.gov/dor/tax-forms/2022-individual-income-tax-forms/
          https://www.in.gov/dor/files/2022-county-tax-rates-and-codes.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Maine', 'state_occupation': 'Indiana', "capital_gains_short": capital_gains_short, "dependents": dependents, "sources": sources, "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def maine_and_iowa(self, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Maine:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Iowa:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Maine:
        - https://www.maine.gov/revenue/tax-return-forms/individual-income-tax-2022
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/ind_tax_rate_sched_2022.pdf
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/22_1040me_book_gen_instruc_revisedFeb23.pdf
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/22_1040me_book_gen_instruc_revisedFeb23.pdf
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/22_1040me_book_gen_instruc_revisedFeb23.pdf
          https://www.maine.gov/revenue/tax-return-forms/individual-income-tax-2022
        
        Iowa:
        -         
        - https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
          https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
          https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
          https://tax.iowa.gov/iowa-illinois-reciprocal-agreement
          https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Maine', 'state_occupation': 'Iowa', "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def maine_and_kansas(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Maine:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Kansas:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Maine:
        - https://www.maine.gov/revenue/tax-return-forms/individual-income-tax-2022
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/ind_tax_rate_sched_2022.pdf
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/22_1040me_book_gen_instruc_revisedFeb23.pdf
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/22_1040me_book_gen_instruc_revisedFeb23.pdf
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/22_1040me_book_gen_instruc_revisedFeb23.pdf
          https://www.maine.gov/revenue/tax-return-forms/individual-income-tax-2022
        
        Kansas:
        -         
        - https://www.ksrevenue.gov/taxrates.html
          https://www.ksrevenue.gov/incomebook22.html#0
          https://www.ksrevenue.gov/incomebook22.html#0
          https://www.ksrevenue.gov/incomebook22.html#0
          https://www.ksrevenue.gov/incomebook22.html
          https://www.ksrevenue.gov/incomebook22.html
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Maine', 'state_occupation': 'Kansas', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def maine_and_kentucky(self, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Maine:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Kentucky:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Maine:
        - https://www.maine.gov/revenue/tax-return-forms/individual-income-tax-2022
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/ind_tax_rate_sched_2022.pdf
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/22_1040me_book_gen_instruc_revisedFeb23.pdf
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/22_1040me_book_gen_instruc_revisedFeb23.pdf
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/22_1040me_book_gen_instruc_revisedFeb23.pdf
          https://www.maine.gov/revenue/tax-return-forms/individual-income-tax-2022
        
        Kentucky:
        -         
        - https://revenue.ky.gov/News/Pages/Dor-Announces-Updates-to-Individual-Income-Tax-for-2023-Tax-Year.aspx
          https://revenue.ky.gov/Individual/Individual-Income-Tax/Pages/default.aspx
          https://revenue.ky.gov/News/Pages/Dor-Announces-Updates-to-Individual-Income-Tax-for-2023-Tax-Year.aspx
          https://revenue.ky.gov/Individual/Individual-Income-Tax/Pages/default.aspx
          https://revenue.ky.gov/Dor%20Training%20Materials/103%20KAR%2017.140.%20Individual%20income%20tax%20-%20reciprocity%20-%20nonresidents.pdf
          https://revenue.ky.gov/Individual/Individual-Income-Tax/Pages/default.aspx
          https://nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2021/Taxes-21-32.htm?taxmap=true
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Maine', 'state_occupation': 'Kentucky', "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def maine_and_louisiana(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Maine:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Louisiana:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Maine:
        - https://www.maine.gov/revenue/tax-return-forms/individual-income-tax-2022
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/ind_tax_rate_sched_2022.pdf
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/22_1040me_book_gen_instruc_revisedFeb23.pdf
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/22_1040me_book_gen_instruc_revisedFeb23.pdf
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/22_1040me_book_gen_instruc_revisedFeb23.pdf
          https://www.maine.gov/revenue/tax-return-forms/individual-income-tax-2022
        
        Louisiana:
        -         
        - https://revenue.louisiana.gov/individualincometax
          https://revenue.louisiana.gov/individualincometax
          https://www.revenue.louisiana.gov/taxforms/6935(11_02)F.pdf
          https://revenue.louisiana.gov/individualincometax
          https://revenue.louisiana.gov/individualincometax
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Maine', 'state_occupation': 'Louisiana', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def maine_and_maryland(self, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Maine:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Maryland:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Maine:
        - https://www.maine.gov/revenue/tax-return-forms/individual-income-tax-2022
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/ind_tax_rate_sched_2022.pdf
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/22_1040me_book_gen_instruc_revisedFeb23.pdf
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/22_1040me_book_gen_instruc_revisedFeb23.pdf
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/22_1040me_book_gen_instruc_revisedFeb23.pdf
          https://www.maine.gov/revenue/tax-return-forms/individual-income-tax-2022
        
        Maryland:
        -         
        - https://www.marylandtaxes.gov/individual/income/tax-info/tax-rates.php
          https://www.marylandtaxes.gov/forms/22_forms/Resident_Booklet.pdf
          https://www.marylandtaxes.gov/forms/Tax_Publications/Administrative_Releases/Income_and_Estate_Tax_Releases/ar_it37.pdf
          https://www.marylandtaxes.gov/individual/index.php
          https://www.marylandtaxes.gov/individual/income/tax-info/tax-rates.php
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Maine', 'state_occupation': 'Maryland', "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def maine_and_massachusetts(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Maine:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Massachusetts:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Maine:
        - https://www.maine.gov/revenue/tax-return-forms/individual-income-tax-2022
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/ind_tax_rate_sched_2022.pdf
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/22_1040me_book_gen_instruc_revisedFeb23.pdf
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/22_1040me_book_gen_instruc_revisedFeb23.pdf
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/22_1040me_book_gen_instruc_revisedFeb23.pdf
          https://www.maine.gov/revenue/tax-return-forms/individual-income-tax-2022
        
        Massachusetts:
        -         
        - https://www.mass.gov/service-details/massachusetts-tax-rates
          https://www.mass.gov/service-details/view-massachusetts-personal-income-tax-exemptions
          https://www.mass.gov/service-details/view-massachusetts-personal-income-tax-exemptions
          https://www.mass.gov/service-details/view-child-and-dependent-related-credits
          https://www.mass.gov/service-details/massachusetts-tax-rates
          https://www.mass.gov/personal-income-tax
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Maine', 'state_occupation': 'Massachusetts', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def maine_and_michigan(self, capital_gains_short=0, county_occupation='Other', dependents=0, sources=False, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_long=0, filing_status='single', income=100000, credits_state=0):
        """
        --- Parameters ----
        
        Maine:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Michigan:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        county_occupation : str, optional
            County of occupation. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Maine:
        - https://www.maine.gov/revenue/tax-return-forms/individual-income-tax-2022
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/ind_tax_rate_sched_2022.pdf
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/22_1040me_book_gen_instruc_revisedFeb23.pdf
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/22_1040me_book_gen_instruc_revisedFeb23.pdf
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/22_1040me_book_gen_instruc_revisedFeb23.pdf
          https://www.maine.gov/revenue/tax-return-forms/individual-income-tax-2022
        
        Michigan:
        -         
        - https://www.michigan.gov/taxes/iit/new-developments-for-tax-year-2022
          "https://www.michigan.gov/taxes/-/media/Project/Websites/taxes/Forms/2023/2023-SUW/446_Withholding-Guide_2023.pdf?rev=177ca3d8cb034392bf84e317ed78c320&hash=19C5F42F43A89A91863D229349F48883#:~:text=The%20exemption%20amount%20is%20%245%2C400
         employees%20Michigan%20income%20tax%20return"
          "https://www.michigan.gov/taxes/-/media/Project/Websites/taxes/Forms/2023/2023-SUW/446_Withholding-Guide_2023.pdf?rev=177ca3d8cb034392bf84e317ed78c320&hash=19C5F42F43A89A91863D229349F48883#:~:text=The%20exemption%20amount%20is%20%245%2C400
         employees%20Michigan%20income%20tax%20return"
          https://www.michigan.gov/taxes/questions/iit/accordion/residency/are-my-wages-earned-in-another-state-taxable-in-michigan-if-i-am-a-michigan-resident-1
          https://www.michigan.gov/taxes/-/media/Project/Websites/taxes/Forms/2022/2022-IIT-Forms/BOOK_MI-1040.pdf?rev=4524f1cbbc0243b194ab05a6680be75b&hash=40DFE71E651F7C1D65C31F6CD03D944F
          https://www.michigan.gov/taxes/questions/iit/accordion/general/what-cities-impose-an-income-tax
        """
        
        # Check if the parameters are set upstream in the instance
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        county_occupation = self.county_occupation if hasattr(self, 'county_occupation') else county_occupation
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Maine', 'state_occupation': 'Michigan', "capital_gains_short": capital_gains_short, "county_occupation": county_occupation, "dependents": dependents, "sources": sources, "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def maine_and_minnesota(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Maine:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Minnesota:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Maine:
        - https://www.maine.gov/revenue/tax-return-forms/individual-income-tax-2022
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/ind_tax_rate_sched_2022.pdf
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/22_1040me_book_gen_instruc_revisedFeb23.pdf
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/22_1040me_book_gen_instruc_revisedFeb23.pdf
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/22_1040me_book_gen_instruc_revisedFeb23.pdf
          https://www.maine.gov/revenue/tax-return-forms/individual-income-tax-2022
        
        Minnesota:
        -         
        - https://www.revenue.state.mn.us/minnesota-income-tax-rates-and-brackets
          https://www.revenue.state.mn.us/child-and-dependent-care-credit
          https://www.revenue.state.mn.us/dependent-exemptions
          https://www.revenue.state.mn.us/minnesota-standard-deduction
          https://mn.gov/mmb/assets/mwr_form_tcm1059-128581.pdf
          https://www.revenue.state.mn.us/individual-income-tax
          https://www.revenue.state.mn.us/individual-income-tax
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Maine', 'state_occupation': 'Minnesota', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def maine_and_mississippi(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Maine:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Mississippi:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Maine:
        - https://www.maine.gov/revenue/tax-return-forms/individual-income-tax-2022
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/ind_tax_rate_sched_2022.pdf
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/22_1040me_book_gen_instruc_revisedFeb23.pdf
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/22_1040me_book_gen_instruc_revisedFeb23.pdf
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/22_1040me_book_gen_instruc_revisedFeb23.pdf
          https://www.maine.gov/revenue/tax-return-forms/individual-income-tax-2022
        
        Mississippi:
        -         
        - https://www.dor.ms.gov/individual/tax-rates
          https://www.dor.ms.gov/individual/tax-rates
          https://www.dor.ms.gov/individual/tax-rates
          https://www.dor.ms.gov/individual
          https://www.dor.ms.gov/individual
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Maine', 'state_occupation': 'Mississippi', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def maine_and_missouri(self, capital_gains_short=0, dependents=0, sources=False, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_long=0, filing_status='single', income=100000, credits_state=0):
        """
        --- Parameters ----
        
        Maine:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Missouri:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Maine:
        - https://www.maine.gov/revenue/tax-return-forms/individual-income-tax-2022
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/ind_tax_rate_sched_2022.pdf
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/22_1040me_book_gen_instruc_revisedFeb23.pdf
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/22_1040me_book_gen_instruc_revisedFeb23.pdf
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/22_1040me_book_gen_instruc_revisedFeb23.pdf
          https://www.maine.gov/revenue/tax-return-forms/individual-income-tax-2022
        
        Missouri:
        -         
        - https://dor.mo.gov/taxation/individual/tax-types/income/year-changes/
          https://dor.mo.gov/taxation/individual/tax-types/income/year-changes/
          https://dor.mo.gov/taxation/individual/tax-types/income/year-changes/
          https://www.kcmo.gov/city-hall/departments/finance/earnings-tax
          https://www.stlouis-mo.gov/government/departments/collector/earnings-tax/individual-earnings-tax-info.cfm
        """
        
        # Check if the parameters are set upstream in the instance
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Maine', 'state_occupation': 'Missouri', "capital_gains_short": capital_gains_short, "dependents": dependents, "sources": sources, "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def maine_and_montana(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Maine:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Montana:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Maine:
        - https://www.maine.gov/revenue/tax-return-forms/individual-income-tax-2022
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/ind_tax_rate_sched_2022.pdf
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/22_1040me_book_gen_instruc_revisedFeb23.pdf
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/22_1040me_book_gen_instruc_revisedFeb23.pdf
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/22_1040me_book_gen_instruc_revisedFeb23.pdf
          https://www.maine.gov/revenue/tax-return-forms/individual-income-tax-2022
        
        Montana:
        -         
        - https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/2022/12/montana-employees-withholding-allowance-and-exemption-certificate-form-mw-4.pdf
          https://mtrevenue.gov/2021/11/29/simplification-of-montana-income-taxation/?utm_source=rss&utm_medium=rss&utm_campaign=simplification-of-montana-income-taxation
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Maine', 'state_occupation': 'Montana', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def maine_and_nebraska(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Maine:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Nebraska:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Maine:
        - https://www.maine.gov/revenue/tax-return-forms/individual-income-tax-2022
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/ind_tax_rate_sched_2022.pdf
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/22_1040me_book_gen_instruc_revisedFeb23.pdf
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/22_1040me_book_gen_instruc_revisedFeb23.pdf
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/22_1040me_book_gen_instruc_revisedFeb23.pdf
          https://www.maine.gov/revenue/tax-return-forms/individual-income-tax-2022
        
        Nebraska:
        -         
        - https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2023_Circular_EN_8-429r2023_Whole_Book_11.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Maine', 'state_occupation': 'Nebraska', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def maine_and_nevada(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Maine:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Nevada:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Maine:
        - https://www.maine.gov/revenue/tax-return-forms/individual-income-tax-2022
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/ind_tax_rate_sched_2022.pdf
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/22_1040me_book_gen_instruc_revisedFeb23.pdf
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/22_1040me_book_gen_instruc_revisedFeb23.pdf
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/22_1040me_book_gen_instruc_revisedFeb23.pdf
          https://www.maine.gov/revenue/tax-return-forms/individual-income-tax-2022
        
        Nevada:
        -         
        - https://www.nvsos.gov/sos/businesses/the-nevada-advantage
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Maine', 'state_occupation': 'Nevada', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def maine_and_new_hampshire(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Maine:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        New Hampshire:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Maine:
        - https://www.maine.gov/revenue/tax-return-forms/individual-income-tax-2022
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/ind_tax_rate_sched_2022.pdf
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/22_1040me_book_gen_instruc_revisedFeb23.pdf
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/22_1040me_book_gen_instruc_revisedFeb23.pdf
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/22_1040me_book_gen_instruc_revisedFeb23.pdf
          https://www.maine.gov/revenue/tax-return-forms/individual-income-tax-2022
        
        New Hampshire:
        -         
        - https://www.revenue.nh.gov/assistance/tax-overview.htm
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Maine', 'state_occupation': 'NewHampshire', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def maine_and_new_jersey(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Maine:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        New Jersey:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Maine:
        - https://www.maine.gov/revenue/tax-return-forms/individual-income-tax-2022
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/ind_tax_rate_sched_2022.pdf
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/22_1040me_book_gen_instruc_revisedFeb23.pdf
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/22_1040me_book_gen_instruc_revisedFeb23.pdf
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/22_1040me_book_gen_instruc_revisedFeb23.pdf
          https://www.maine.gov/revenue/tax-return-forms/individual-income-tax-2022
        
        New Jersey:
        -         
        - https://www.state.nj.us/treasury/taxation/taxtables.shtml
          https://www.state.nj.us/treasury/taxation/njit13.shtml
          https://www.nj.gov/treasury/taxation/njit25.shtml
          https://www.state.nj.us/treasury/taxation/individuals/freqqiti.shtml
          https://www.state.nj.us/treasury/taxation/localtax.shtml
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Maine', 'state_occupation': 'NewJersey', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def maine_and_new_mexico(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Maine:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        New Mexico:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Maine:
        - https://www.maine.gov/revenue/tax-return-forms/individual-income-tax-2022
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/ind_tax_rate_sched_2022.pdf
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/22_1040me_book_gen_instruc_revisedFeb23.pdf
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/22_1040me_book_gen_instruc_revisedFeb23.pdf
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/22_1040me_book_gen_instruc_revisedFeb23.pdf
          https://www.maine.gov/revenue/tax-return-forms/individual-income-tax-2022
        
        New Mexico:
        -         
        - https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/fdf3c548-8aba-4b9c-9eb4-bb564c716015/FYI-104.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/0bf943fd-652e-4400-bb1b-cd8397eb5a95/2022pit-adj.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/2f1a6781-9534-4436-b427-1557f9592099/2022pit-adj-ins.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/0bf943fd-652e-4400-bb1b-cd8397eb5a95/2022pit-adj.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/1afc56af-ea90-4d48-82e5-1f9aeb43255a/PITbook2022.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/fdf3c548-8aba-4b9c-9eb4-bb564c716015/FYI-104.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Maine', 'state_occupation': 'NewMexico', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def maine_and_new_york(self, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Maine:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        New York:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Maine:
        - https://www.maine.gov/revenue/tax-return-forms/individual-income-tax-2022
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/ind_tax_rate_sched_2022.pdf
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/22_1040me_book_gen_instruc_revisedFeb23.pdf
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/22_1040me_book_gen_instruc_revisedFeb23.pdf
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/22_1040me_book_gen_instruc_revisedFeb23.pdf
          https://www.maine.gov/revenue/tax-return-forms/individual-income-tax-2022
        
        New York:
        -         
        - https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nys-tax-rate-schedule
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nys-tax-rate-schedule
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nys-tax-rate-schedule
          https://www.tax.ny.gov/pit/
          https://www.tax.ny.gov/pit/file/nonresident-faqs.htm
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#yonkers-tax-rate-schedule
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nyc-tax-rate-schedule
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Maine', 'state_occupation': 'NewYork', "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def maine_and_north_carolina(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Maine:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        North Carolina:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Maine:
        - https://www.maine.gov/revenue/tax-return-forms/individual-income-tax-2022
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/ind_tax_rate_sched_2022.pdf
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/22_1040me_book_gen_instruc_revisedFeb23.pdf
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/22_1040me_book_gen_instruc_revisedFeb23.pdf
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/22_1040me_book_gen_instruc_revisedFeb23.pdf
          https://www.maine.gov/revenue/tax-return-forms/individual-income-tax-2022
        
        North Carolina:
        -         
        - https://www.ncdor.gov/taxes-forms/tax-rate-schedules
          https://www.ncdor.gov/taxes-forms/individual-income-tax/north-carolina-standard-deduction-or-north-carolina-itemized-deductions
          https://www.ncdor.gov/taxes-forms/individual-income-tax
          https://www.ncdor.gov/taxes-forms/individual-income-tax
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Maine', 'state_occupation': 'NorthCarolina', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def maine_and_north_dakota(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Maine:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        North Dakota:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Maine:
        - https://www.maine.gov/revenue/tax-return-forms/individual-income-tax-2022
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/ind_tax_rate_sched_2022.pdf
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/22_1040me_book_gen_instruc_revisedFeb23.pdf
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/22_1040me_book_gen_instruc_revisedFeb23.pdf
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/22_1040me_book_gen_instruc_revisedFeb23.pdf
          https://www.maine.gov/revenue/tax-return-forms/individual-income-tax-2022
        
        North Dakota:
        -         
        - https://www.tax.nd.gov/forms
          https://www.tax.nd.gov/sites/www/files/documents/forms/individual/2022-iit/2022-individual-income-tax-booklet.pdf
          https://www.tax.nd.gov/sites/www/files/documents/forms/2022-individual-income-tax-booklet.pdf
          https://www.tax.nd.gov/tax-exemptions-credits/income-tax-exemptions-credits
          https://www.tax.nd.gov/sites/www/files/documents/forms/business/ndwrfillable.pdf
          https://www.tax.nd.gov/news/tax-legislative-changes/significant-changes-law/individual-income-tax-history
          https://www.tax.nd.gov/forms
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Maine', 'state_occupation': 'NorthDakota', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def maine_and_ohio(self, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Maine:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Ohio:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Maine:
        - https://www.maine.gov/revenue/tax-return-forms/individual-income-tax-2022
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/ind_tax_rate_sched_2022.pdf
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/22_1040me_book_gen_instruc_revisedFeb23.pdf
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/22_1040me_book_gen_instruc_revisedFeb23.pdf
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/22_1040me_book_gen_instruc_revisedFeb23.pdf
          https://www.maine.gov/revenue/tax-return-forms/individual-income-tax-2022
        
        Ohio:
        -         
        - https://tax.ohio.gov/individual/resources/annual-tax-rates
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://tax.ohio.gov/static/forms/employer_withholding/generic/wth_it4nr.pdf
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://www.ritaohio.com/TaxRatesTable
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Maine', 'state_occupation': 'Ohio', "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def maine_and_oklahoma(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Maine:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Oklahoma:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Maine:
        - https://www.maine.gov/revenue/tax-return-forms/individual-income-tax-2022
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/ind_tax_rate_sched_2022.pdf
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/22_1040me_book_gen_instruc_revisedFeb23.pdf
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/22_1040me_book_gen_instruc_revisedFeb23.pdf
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/22_1040me_book_gen_instruc_revisedFeb23.pdf
          https://www.maine.gov/revenue/tax-return-forms/individual-income-tax-2022
        
        Oklahoma:
        -         
        - https://oklahoma.gov/content/dam/ok/en/tax/documents/resources/publications/businesses/withholding-tables/WHTables-2022.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Maine', 'state_occupation': 'Oklahoma', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def maine_and_oregon(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Maine:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Oregon:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Maine:
        - https://www.maine.gov/revenue/tax-return-forms/individual-income-tax-2022
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/ind_tax_rate_sched_2022.pdf
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/22_1040me_book_gen_instruc_revisedFeb23.pdf
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/22_1040me_book_gen_instruc_revisedFeb23.pdf
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/22_1040me_book_gen_instruc_revisedFeb23.pdf
          https://www.maine.gov/revenue/tax-return-forms/individual-income-tax-2022
        
        Oregon:
        -         
        - https://www.oregon.gov/dor/programs/individuals/Pages/PIT.aspx
          https://www.oregon.gov/dor/programs/individuals/Documents/Full%20year%20resident
         %20Form%20or-40%20filers.pdf
          https://www.oregon.gov/dor/programs/individuals/Documents/Part-year%20and%20nonresident
         %20Form%20or-40-P%20and%20or-40-N%20filers.pdf.pdf
          https://www.oregon.gov/dor/forms/FormsPubs/form-or-40_101-040_2022.pdf
          https://www.oregon.gov/dor/forms/FormsPubs/form-or-40_101-040_2022.pdf
          https://www.oregon.gov/dor/forms/FormsPubs/form-or-40_101-040_2022.pdf
          https://www.portland.gov/revenue/personal-tax
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Maine', 'state_occupation': 'Oregon', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def maine_and_pennsylvania(self, county_residence='Other', state_agi=75000, county_occupation='Other', filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Maine:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Pennsylvania:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        county_residence : str, optional
            County of residence. Default is Other.
        county_occupation : str, optional
            County of occupation. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Maine:
        - https://www.maine.gov/revenue/tax-return-forms/individual-income-tax-2022
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/ind_tax_rate_sched_2022.pdf
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/22_1040me_book_gen_instruc_revisedFeb23.pdf
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/22_1040me_book_gen_instruc_revisedFeb23.pdf
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/22_1040me_book_gen_instruc_revisedFeb23.pdf
          https://www.maine.gov/revenue/tax-return-forms/individual-income-tax-2022
        
        Pennsylvania:
        -         
        - https://www.revenue.pa.gov/Tax%20Rates/Pages/default.aspx
          https://www.revenue.pa.gov/FormsandPublications/FormsforIndividuals/PIT/Documents/2022/2022_pa-40in.pdf
          https://www.revenue.pa.gov/FormsandPublications/FormsforBusinesses/EmployerWithholding/Documents/rev-419.pdf
          https://www.revenue.pa.gov/FormsandPublications/FormsforIndividuals/PIT/Documents/2022/2022_pa-40in.pdf
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-25.htm
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        county_occupation = self.county_occupation if hasattr(self, 'county_occupation') else county_occupation
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Maine', 'state_occupation': 'Pennsylvania', "county_residence": county_residence, "state_agi": state_agi, "county_occupation": county_occupation, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def maine_and_rhode_island(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Maine:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Rhode Island:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Maine:
        - https://www.maine.gov/revenue/tax-return-forms/individual-income-tax-2022
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/ind_tax_rate_sched_2022.pdf
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/22_1040me_book_gen_instruc_revisedFeb23.pdf
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/22_1040me_book_gen_instruc_revisedFeb23.pdf
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/22_1040me_book_gen_instruc_revisedFeb23.pdf
          https://www.maine.gov/revenue/tax-return-forms/individual-income-tax-2022
        
        Rhode Island:
        -         
        - https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2022_1040WE_w_0.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Maine', 'state_occupation': 'RhodeIsland', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def maine_and_south_carolina(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Maine:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        South Carolina:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Maine:
        - https://www.maine.gov/revenue/tax-return-forms/individual-income-tax-2022
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/ind_tax_rate_sched_2022.pdf
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/22_1040me_book_gen_instruc_revisedFeb23.pdf
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/22_1040me_book_gen_instruc_revisedFeb23.pdf
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/22_1040me_book_gen_instruc_revisedFeb23.pdf
          https://www.maine.gov/revenue/tax-return-forms/individual-income-tax-2022
        
        South Carolina:
        -         
        - https://dor.sc.gov/resources-site/lawandpolicy/Advisory%20Opinions/IL22-15.pdf
          https://dor.sc.gov/forms-site/Forms/IITPacket_2022.pdf
          https://www.scstatehouse.gov/code/t12c006.php
          https://dor.sc.gov/resources-site/lawandpolicy/Advisory%20Opinions/IL22-15.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Maine', 'state_occupation': 'SouthCarolina', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def maine_and_south_dakota(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Maine:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        South Dakota:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Maine:
        - https://www.maine.gov/revenue/tax-return-forms/individual-income-tax-2022
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/ind_tax_rate_sched_2022.pdf
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/22_1040me_book_gen_instruc_revisedFeb23.pdf
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/22_1040me_book_gen_instruc_revisedFeb23.pdf
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/22_1040me_book_gen_instruc_revisedFeb23.pdf
          https://www.maine.gov/revenue/tax-return-forms/individual-income-tax-2022
        
        South Dakota:
        -         
        - https://dor.sd.gov/individuals/taxes/
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Maine', 'state_occupation': 'SouthDakota', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def maine_and_tennessee(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Maine:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Tennessee:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Maine:
        - https://www.maine.gov/revenue/tax-return-forms/individual-income-tax-2022
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/ind_tax_rate_sched_2022.pdf
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/22_1040me_book_gen_instruc_revisedFeb23.pdf
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/22_1040me_book_gen_instruc_revisedFeb23.pdf
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/22_1040me_book_gen_instruc_revisedFeb23.pdf
          https://www.maine.gov/revenue/tax-return-forms/individual-income-tax-2022
        
        Tennessee:
        -         
        - https://www.tn.gov/revenue/taxes.html
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Maine', 'state_occupation': 'Tennessee', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def maine_and_texas(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Maine:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Texas:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Maine:
        - https://www.maine.gov/revenue/tax-return-forms/individual-income-tax-2022
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/ind_tax_rate_sched_2022.pdf
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/22_1040me_book_gen_instruc_revisedFeb23.pdf
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/22_1040me_book_gen_instruc_revisedFeb23.pdf
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/22_1040me_book_gen_instruc_revisedFeb23.pdf
          https://www.maine.gov/revenue/tax-return-forms/individual-income-tax-2022
        
        Texas:
        -         
        - https://comptroller.texas.gov/taxes/
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Maine', 'state_occupation': 'Texas', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def maine_and_utah(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Maine:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Utah:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Maine:
        - https://www.maine.gov/revenue/tax-return-forms/individual-income-tax-2022
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/ind_tax_rate_sched_2022.pdf
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/22_1040me_book_gen_instruc_revisedFeb23.pdf
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/22_1040me_book_gen_instruc_revisedFeb23.pdf
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/22_1040me_book_gen_instruc_revisedFeb23.pdf
          https://www.maine.gov/revenue/tax-return-forms/individual-income-tax-2022
        
        Utah:
        -         
        - https://incometax.utah.gov/paying/tax-rates
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Maine', 'state_occupation': 'Utah', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def maine_and_vermont(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Maine:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Vermont:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Maine:
        - https://www.maine.gov/revenue/tax-return-forms/individual-income-tax-2022
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/ind_tax_rate_sched_2022.pdf
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/22_1040me_book_gen_instruc_revisedFeb23.pdf
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/22_1040me_book_gen_instruc_revisedFeb23.pdf
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/22_1040me_book_gen_instruc_revisedFeb23.pdf
          https://www.maine.gov/revenue/tax-return-forms/individual-income-tax-2022
        
        Vermont:
        -         
        - https://tax.vermont.gov/sites/tax/files/documents/GB-1210-2023.pdf
          https://tax.vermont.gov/sites/tax/files/documents/Income%20Booklet-2022.pdf
          https://tax.vermont.gov/sites/tax/files/documents/Income%20Booklet-2022.pdf
          https://tax.vermont.gov/sites/tax/files/documents/Income%20Booklet-2022.pdf
          https://tax.vermont.gov/individuals/personal-income-tax/tax-credits
          https://tax.vermont.gov/individuals/personal-income-tax/taxable-income
          https://tax.vermont.gov/individuals/personal-income-tax/local-option-tax
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Maine', 'state_occupation': 'Vermont', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def maine_and_virginia(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Maine:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Virginia:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Maine:
        - https://www.maine.gov/revenue/tax-return-forms/individual-income-tax-2022
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/ind_tax_rate_sched_2022.pdf
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/22_1040me_book_gen_instruc_revisedFeb23.pdf
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/22_1040me_book_gen_instruc_revisedFeb23.pdf
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/22_1040me_book_gen_instruc_revisedFeb23.pdf
          https://www.maine.gov/revenue/tax-return-forms/individual-income-tax-2022
        
        Virginia:
        -         
        - https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/reciprocity
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Maine', 'state_occupation': 'Virginia', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def maine_and_washington(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Maine:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Washington:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Maine:
        - https://www.maine.gov/revenue/tax-return-forms/individual-income-tax-2022
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/ind_tax_rate_sched_2022.pdf
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/22_1040me_book_gen_instruc_revisedFeb23.pdf
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/22_1040me_book_gen_instruc_revisedFeb23.pdf
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/22_1040me_book_gen_instruc_revisedFeb23.pdf
          https://www.maine.gov/revenue/tax-return-forms/individual-income-tax-2022
        
        Washington:
        -         
        - https://dor.wa.gov/taxes-rates/income-tax
          https://dor.wa.gov/taxes-rates/other-taxes/capital-gains-tax
          https://dor.wa.gov/taxes-rates/income-tax
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Maine', 'state_occupation': 'Washington', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def maine_and_west_virginia(self, pay_periods=24, capital_gains_short=0, county_occupation='Other', dependents=0, sources=False, state_agi=75000, filers_over_65=0, capital_gains_long=0, filing_status='single', income=100000, credits_state=0):
        """
        --- Parameters ----
        
        Maine:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        West Virginia:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_occupation : str, optional
            County of occupation. Default is Other.
        pay_periods : int, optional
            Number of pay periods. Options are 13, 24, 26, 52. Default is 24.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Maine:
        - https://www.maine.gov/revenue/tax-return-forms/individual-income-tax-2022
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/ind_tax_rate_sched_2022.pdf
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/22_1040me_book_gen_instruc_revisedFeb23.pdf
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/22_1040me_book_gen_instruc_revisedFeb23.pdf
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/22_1040me_book_gen_instruc_revisedFeb23.pdf
          https://www.maine.gov/revenue/tax-return-forms/individual-income-tax-2022
        
        West Virginia:
        -         
        - https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://tax.wv.gov/Documents/TSD/tsd438.pdf
          https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2011/Taxes-11-35att.html
        """
        
        # Check if the parameters are set upstream in the instance
        pay_periods = self.pay_periods if hasattr(self, 'pay_periods') else pay_periods
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        county_occupation = self.county_occupation if hasattr(self, 'county_occupation') else county_occupation
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Maine', 'state_occupation': 'WestVirginia', "pay_periods": pay_periods, "capital_gains_short": capital_gains_short, "county_occupation": county_occupation, "dependents": dependents, "sources": sources, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def maine_and_wisconsin(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Maine:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Wisconsin:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Maine:
        - https://www.maine.gov/revenue/tax-return-forms/individual-income-tax-2022
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/ind_tax_rate_sched_2022.pdf
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/22_1040me_book_gen_instruc_revisedFeb23.pdf
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/22_1040me_book_gen_instruc_revisedFeb23.pdf
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/22_1040me_book_gen_instruc_revisedFeb23.pdf
          https://www.maine.gov/revenue/tax-return-forms/individual-income-tax-2022
        
        Wisconsin:
        -         
        - https://www.revenue.wi.gov/TaxForms2023/2023-Form1-ES-Inst.pdf
          https://www.revenue.wi.gov/TaxForms2023/2023-Form1-ES-Inst.pdf
          https://www.revenue.wi.gov/TaxForms2023/2023-Form1-ES-Inst.pdf
          https://www.revenue.wi.gov/Pages/FAQS/pcs-work.aspx
          https://www.revenue.wi.gov/Dor%20Publications/pb103.pdf
          https://www.revenue.wi.gov/Pages/FAQS/pcs-work.aspx
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Maine', 'state_occupation': 'Wisconsin', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def maine_and_wyoming(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Maine:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Wyoming:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Maine:
        - https://www.maine.gov/revenue/tax-return-forms/individual-income-tax-2022
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/ind_tax_rate_sched_2022.pdf
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/22_1040me_book_gen_instruc_revisedFeb23.pdf
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/22_1040me_book_gen_instruc_revisedFeb23.pdf
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/22_1040me_book_gen_instruc_revisedFeb23.pdf
          https://www.maine.gov/revenue/tax-return-forms/individual-income-tax-2022
        
        Wyoming:
        -         
        - https://revenue.wyo.gov/
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Maine', 'state_occupation': 'Wyoming', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def maryland_and_alabama(self, county_residence='Other', state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Maryland:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Alabama:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Maryland:
        - https://www.marylandtaxes.gov/individual/income/tax-info/tax-rates.php
          https://www.marylandtaxes.gov/forms/22_forms/Resident_Booklet.pdf
          https://www.marylandtaxes.gov/forms/Tax_Publications/Administrative_Releases/Income_and_Estate_Tax_Releases/ar_it37.pdf
          https://www.marylandtaxes.gov/individual/index.php
          https://www.marylandtaxes.gov/individual/income/tax-info/tax-rates.php
        
        Alabama:
        -         
        - https://www.revenue.alabama.gov/ultraviewer/viewer/basic_viewer/index.html?form=2023/02/22f40.pdf
          https://www.revenue.alabama.gov/faqs/what-is-alabamas-individual-income-tax-rate/
          https://www.revenue.alabama.gov/ultraviewer/viewer/basic_viewer/index.html?form=2023/02/22f40.pdf
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-24.htm?taxmap=true
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-24.htm?taxmap=true
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-24.htm?taxmap=true
          https://www.revenue.alabama.gov/individual-corporate/income-to-be-reported-on-the-alabama-income-tax-return/
          https://www.almonline.org/TaxRates.aspx#OccupationalTax
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Maryland', 'state_occupation': 'Alabama', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def maryland_and_alaska(self, county_residence='Other', state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Maryland:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Alaska:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Maryland:
        - https://www.marylandtaxes.gov/individual/income/tax-info/tax-rates.php
          https://www.marylandtaxes.gov/forms/22_forms/Resident_Booklet.pdf
          https://www.marylandtaxes.gov/forms/Tax_Publications/Administrative_Releases/Income_and_Estate_Tax_Releases/ar_it37.pdf
          https://www.marylandtaxes.gov/individual/index.php
          https://www.marylandtaxes.gov/individual/income/tax-info/tax-rates.php
        
        Alaska:
        -         
        - https://www.commerce.alaska.gov/web/dcra/officeofthestateassessor/alaskataxfacts.aspx
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Maryland', 'state_occupation': 'Alaska', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def maryland_and_arizona(self, county_residence='Other', state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Maryland:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Arizona:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        credits_state : int, optional
            Credits for state taxes. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Maryland:
        - https://www.marylandtaxes.gov/individual/income/tax-info/tax-rates.php
          https://www.marylandtaxes.gov/forms/22_forms/Resident_Booklet.pdf
          https://www.marylandtaxes.gov/forms/Tax_Publications/Administrative_Releases/Income_and_Estate_Tax_Releases/ar_it37.pdf
          https://www.marylandtaxes.gov/individual/index.php
          https://www.marylandtaxes.gov/individual/income/tax-info/tax-rates.php
        
        Arizona:
        -         
        - https://azdor.gov/forms/individual/form-140-x-y-tables
          https://azdor.gov/forms/individual/form-140-x-y-tables
          https://azdor.gov/forms/individual/form-140-x-y-tables
          https://azdor.gov/businesses-arizona/withholding-tax/withholding-exceptions
          "https://azdor.gov/individuals/income-tax-filing-assistance/identifying-other-taxable-income#:~:text=In%20regards%20to%20capital%20gains
         the%20individuals%20regular%20tax%20rate.
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Maryland', 'state_occupation': 'Arizona', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def maryland_and_arkansas(self, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Maryland:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Arkansas:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Maryland:
        - https://www.marylandtaxes.gov/individual/income/tax-info/tax-rates.php
          https://www.marylandtaxes.gov/forms/22_forms/Resident_Booklet.pdf
          https://www.marylandtaxes.gov/forms/Tax_Publications/Administrative_Releases/Income_and_Estate_Tax_Releases/ar_it37.pdf
          https://www.marylandtaxes.gov/individual/index.php
          https://www.marylandtaxes.gov/individual/income/tax-info/tax-rates.php
        
        Arkansas:
        -         
        - https://www.arkansasedc.com/why-arkansas/business-climate/tax-structure/personal-income-tax
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_and_AR1000NR_Instructions.pdf
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_FullYearResidentIndividualIncomeTaxReturn.pdf
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_and_AR1000NR_Instructions.pdf
          https://www.arkansasedc.com/why-arkansas/business-climate/tax-structure/capital-gains-tax-reduction
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_and_AR1000NR_Instructions.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Maryland', 'state_occupation': 'Arkansas', "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def maryland_and_california(self, county_residence='Other', state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Maryland:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        California:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Maryland:
        - https://www.marylandtaxes.gov/individual/income/tax-info/tax-rates.php
          https://www.marylandtaxes.gov/forms/22_forms/Resident_Booklet.pdf
          https://www.marylandtaxes.gov/forms/Tax_Publications/Administrative_Releases/Income_and_Estate_Tax_Releases/ar_it37.pdf
          https://www.marylandtaxes.gov/individual/index.php
          https://www.marylandtaxes.gov/individual/income/tax-info/tax-rates.php
        
        California:
        -         
        - https://www.ftb.ca.gov/forms/2022/2022-540-tax-rate-schedules.pdf
          https://www.ftb.ca.gov/file/personal/deductions/index.html
          https://www.ftb.ca.gov/about-ftb/newsroom/caleitc/eligibility-and-credit-information.html
          https://www.ftb.ca.gov/file/personal/income-types/capital-gains-and-losses.html#:~:text=California%20does%20not%20have%20a
         are%20taxed%20as%20ordinary%20income
          https://www.ftb.ca.gov/file/personal/
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Maryland', 'state_occupation': 'California', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def maryland_and_colorado(self, county_residence='Other', state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Maryland:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Colorado:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Maryland:
        - https://www.marylandtaxes.gov/individual/income/tax-info/tax-rates.php
          https://www.marylandtaxes.gov/forms/22_forms/Resident_Booklet.pdf
          https://www.marylandtaxes.gov/forms/Tax_Publications/Administrative_Releases/Income_and_Estate_Tax_Releases/ar_it37.pdf
          https://www.marylandtaxes.gov/individual/index.php
          https://www.marylandtaxes.gov/individual/income/tax-info/tax-rates.php
        
        Colorado:
        -         
        - https://leg.colorado.gov/agencies/legislative-council-staff/individual-income-tax%C2%A0
          https://leg.colorado.gov/agencies/legislative-council-staff/individual-income-tax%C2%A0
          https://tax.colorado.gov/sites/tax/files/documents/ITT_Capital_Gain_Subtraction_Feb_2022.pdf
          https://www.auroragov.org/business_services/taxes/occupational_privilege_tax
          https://www.denvergov.org/Government/Agencies-Departments-Offices/Agencies-Departments-Offices-Directory/Department-of-Finance/Our-Divisions/Treasury/Business-Tax-Information#:~:text=Occupational%20privilege%20tax
         -%28show%20below%29&text=Each%20taxable%20employee%20is%20liable
         month%20for%20each%20taxable%20employee
          https://www.glendale.co.us/355/Occupational-Privilege-Tax
          https://www.greenwoodvillage.com/1220/Occupational-Privilege-Tax-OPT
          https://ci.sheridan.co.us/288/Occupational-Privilege-Tax
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Maryland', 'state_occupation': 'Colorado', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def maryland_and_connecticut(self, county_residence='Other', state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Maryland:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Connecticut:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Maryland:
        - https://www.marylandtaxes.gov/individual/income/tax-info/tax-rates.php
          https://www.marylandtaxes.gov/forms/22_forms/Resident_Booklet.pdf
          https://www.marylandtaxes.gov/forms/Tax_Publications/Administrative_Releases/Income_and_Estate_Tax_Releases/ar_it37.pdf
          https://www.marylandtaxes.gov/individual/index.php
          https://www.marylandtaxes.gov/individual/income/tax-info/tax-rates.php
        
        Connecticut:
        -         
        - https://portal.ct.gov/-/media/DRS/Forms/2022/Income/2022-CT-1040-Instructions_1222.pdf
          https://www.cga.ct.gov/2022/rpt/pdf/2022-R-0108.pdf
          https://www.cga.ct.gov/current/pub/chap_229.htm#sec_12-703
          https://portal.ct.gov/-/media/DRS/Forms/2022/Income/2022-CT-1040-Instructions_1222.pdf
          "https://portal.ct.gov/DRS/Publications/TSSNs/TSSN-29#:~:text=An%20individuals%20net%20capital%20gains
         portion%20of%20Social%20Security%20benefits"
          "https://portal.ct.gov/DRS/Publications/TSSNs/TSSN-29#:~:text=An%20individuals%20net%20capital%20gains
         portion%20of%20Social%20Security%20benefits"
          https://portal.ct.gov/DRS/Individuals/Resident-Income-Tax/Tax-Information
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Maryland', 'state_occupation': 'Connecticut', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def maryland_and_delaware(self, county_residence='Other', state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, federal_agi=75000, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Maryland:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Delaware:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        federal_agi : int, optional
            Adjusted Gross Income (AGI) for federal taxes. Default is 0. Sources - https://www.irs.gov/e-file-providers/definition-of-adjusted-gross-income, https://www.irs.gov/newsroom/taxpayers-should-know-the-difference-between-standard-and-itemized-deductions
        income : int, optional
            Gross income. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Maryland:
        - https://www.marylandtaxes.gov/individual/income/tax-info/tax-rates.php
          https://www.marylandtaxes.gov/forms/22_forms/Resident_Booklet.pdf
          https://www.marylandtaxes.gov/forms/Tax_Publications/Administrative_Releases/Income_and_Estate_Tax_Releases/ar_it37.pdf
          https://www.marylandtaxes.gov/individual/index.php
          https://www.marylandtaxes.gov/individual/income/tax-info/tax-rates.php
        
        Delaware:
        -         
        - https://revenue.delaware.gov/software-developer/tax-rate-changes/
          https://revenuefiles.delaware.gov/2022/PIT-RES_TY22_2022-01_PaperInteractive.pdf
          https://revenuefiles.delaware.gov/2022/PIT-RES_TY22_2022-02_Instructions.pdf
          https://revenue.delaware.gov/frequently-asked-questions/personal-income-tax-faqs/
          https://www.newcastlede.gov/DocumentCenter/View/20147/City-of-Wilmington-Wage-Tax-Information-PDF
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        federal_agi = self.federal_agi if hasattr(self, 'federal_agi') else federal_agi
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Maryland', 'state_occupation': 'Delaware', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "federal_agi": federal_agi, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def maryland_and_florida(self, county_residence='Other', state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Maryland:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Florida:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Maryland:
        - https://www.marylandtaxes.gov/individual/income/tax-info/tax-rates.php
          https://www.marylandtaxes.gov/forms/22_forms/Resident_Booklet.pdf
          https://www.marylandtaxes.gov/forms/Tax_Publications/Administrative_Releases/Income_and_Estate_Tax_Releases/ar_it37.pdf
          https://www.marylandtaxes.gov/individual/index.php
          https://www.marylandtaxes.gov/individual/income/tax-info/tax-rates.php
        
        Florida:
        -         
        - https://floridarevenue.com/taxes/taxesfees/Pages/tax_interest_rates.aspx
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Maryland', 'state_occupation': 'Florida', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def maryland_and_georgia(self, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Maryland:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Georgia:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Maryland:
        - https://www.marylandtaxes.gov/individual/income/tax-info/tax-rates.php
          https://www.marylandtaxes.gov/forms/22_forms/Resident_Booklet.pdf
          https://www.marylandtaxes.gov/forms/Tax_Publications/Administrative_Releases/Income_and_Estate_Tax_Releases/ar_it37.pdf
          https://www.marylandtaxes.gov/individual/index.php
          https://www.marylandtaxes.gov/individual/income/tax-info/tax-rates.php
        
        Georgia:
        -         
        - https://dor.georgia.gov/tax-tables-georgia-tax-rate-schedule
          https://apps.dor.ga.gov/FillableForms/PDFViewer/Index?form=2022GA500
          https://apps.dor.ga.gov/FillableForms/PDFViewer/Index?form=2022GA500
          https://apps.dor.ga.gov/FillableForms/PDFViewer/Index?form=2022GA500
          https://dor.georgia.gov/years-individual-income-tax-forms
          https://dor.georgia.gov/years-individual-income-tax-forms
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Maryland', 'state_occupation': 'Georgia', "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def maryland_and_hawaii(self, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Maryland:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Hawaii:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Maryland:
        - https://www.marylandtaxes.gov/individual/income/tax-info/tax-rates.php
          https://www.marylandtaxes.gov/forms/22_forms/Resident_Booklet.pdf
          https://www.marylandtaxes.gov/forms/Tax_Publications/Administrative_Releases/Income_and_Estate_Tax_Releases/ar_it37.pdf
          https://www.marylandtaxes.gov/individual/index.php
          https://www.marylandtaxes.gov/individual/income/tax-info/tax-rates.php
        
        Hawaii:
        -         
        - https://tax.hawaii.gov/forms/d_18table-on/d_18table-on_p13/
          https://files.hawaii.gov/tax/forms/2018/18table-on.pdf
          https://files.hawaii.gov/tax/news/pubs/22outline.pdf
          https://files.hawaii.gov/tax/news/pubs/FastTaxRefGuide_2017.pdf
          https://www.tfhawaii.org/wordpress/state-tax-resources/mini-tax-guides/individual-income-tax-chapter-235/#:~:text=Long%20term%20capital%20gains%20are%20taxed%20at%20a%20maximum%20of%207.25%25
          https://www.grassrootinstitute.org/2023/02/hb337-increasing-tax-on-capital-gains-a-bad-risk/
          https://files.hawaii.gov/tax/news/pubs/22outline.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Maryland', 'state_occupation': 'Hawaii', "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def maryland_and_idaho(self, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Maryland:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Idaho:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Maryland:
        - https://www.marylandtaxes.gov/individual/income/tax-info/tax-rates.php
          https://www.marylandtaxes.gov/forms/22_forms/Resident_Booklet.pdf
          https://www.marylandtaxes.gov/forms/Tax_Publications/Administrative_Releases/Income_and_Estate_Tax_Releases/ar_it37.pdf
          https://www.marylandtaxes.gov/individual/index.php
          https://www.marylandtaxes.gov/individual/income/tax-info/tax-rates.php
        
        Idaho:
        -         
        - https://tax.idaho.gov/taxes/income-tax/individual-income/individual-income-tax-rate-schedule/
          https://tax.idaho.gov/wp-content/uploads/forms/EIN00046/EIN00046_03-01-2023.pdf
          https://tax.idaho.gov/taxes/income-tax/individual-income/idaho-source-income/
          https://tax.idaho.gov/taxes/income-tax/individual-income/idaho-source-income/
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Maryland', 'state_occupation': 'Idaho', "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def maryland_and_illinois(self, county_residence='Other', state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Maryland:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Illinois:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Maryland:
        - https://www.marylandtaxes.gov/individual/income/tax-info/tax-rates.php
          https://www.marylandtaxes.gov/forms/22_forms/Resident_Booklet.pdf
          https://www.marylandtaxes.gov/forms/Tax_Publications/Administrative_Releases/Income_and_Estate_Tax_Releases/ar_it37.pdf
          https://www.marylandtaxes.gov/individual/index.php
          https://www.marylandtaxes.gov/individual/income/tax-info/tax-rates.php
        
        Illinois:
        -         
        - https://tax.illinois.gov/research/taxrates/income.html#IndividualIncome
          https://tax.illinois.gov/content/dam/soi/en/web/tax/forms/incometax/documents/currentyear/individual/il-1040-instr.pdf
          https://tax.illinois.gov/questionsandanswers/12.html
          https://tax.illinois.gov/content/dam/soi/en/web/tax/forms/incometax/documents/currentyear/individual/il-1040-instr.pdf
          https://tax.illinois.gov/research/taxrates/personalproperty.html
          https://tax.illinois.gov/localgovernments/income.html
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Maryland', 'state_occupation': 'Illinois', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def maryland_and_indiana(self, capital_gains_short=0, dependents=0, sources=False, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_long=0, filing_status='single', income=100000, credits_state=0):
        """
        --- Parameters ----
        
        Maryland:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Indiana:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Maryland:
        - https://www.marylandtaxes.gov/individual/income/tax-info/tax-rates.php
          https://www.marylandtaxes.gov/forms/22_forms/Resident_Booklet.pdf
          https://www.marylandtaxes.gov/forms/Tax_Publications/Administrative_Releases/Income_and_Estate_Tax_Releases/ar_it37.pdf
          https://www.marylandtaxes.gov/individual/index.php
          https://www.marylandtaxes.gov/individual/income/tax-info/tax-rates.php
        
        Indiana:
        -         
        - https://www.in.gov/dor/tax-forms/2022-individual-income-tax-forms/
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-32.htm
          https://www.in.gov/dor/files/ib28.pdf
          https://www.in.gov/dor/tax-forms/2022-individual-income-tax-forms/
          https://www.in.gov/dor/files/2022-county-tax-rates-and-codes.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Maryland', 'state_occupation': 'Indiana', "capital_gains_short": capital_gains_short, "dependents": dependents, "sources": sources, "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def maryland_and_iowa(self, county_residence='Other', state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Maryland:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Iowa:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Maryland:
        - https://www.marylandtaxes.gov/individual/income/tax-info/tax-rates.php
          https://www.marylandtaxes.gov/forms/22_forms/Resident_Booklet.pdf
          https://www.marylandtaxes.gov/forms/Tax_Publications/Administrative_Releases/Income_and_Estate_Tax_Releases/ar_it37.pdf
          https://www.marylandtaxes.gov/individual/index.php
          https://www.marylandtaxes.gov/individual/income/tax-info/tax-rates.php
        
        Iowa:
        -         
        - https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
          https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
          https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
          https://tax.iowa.gov/iowa-illinois-reciprocal-agreement
          https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Maryland', 'state_occupation': 'Iowa', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def maryland_and_kansas(self, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Maryland:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Kansas:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Maryland:
        - https://www.marylandtaxes.gov/individual/income/tax-info/tax-rates.php
          https://www.marylandtaxes.gov/forms/22_forms/Resident_Booklet.pdf
          https://www.marylandtaxes.gov/forms/Tax_Publications/Administrative_Releases/Income_and_Estate_Tax_Releases/ar_it37.pdf
          https://www.marylandtaxes.gov/individual/index.php
          https://www.marylandtaxes.gov/individual/income/tax-info/tax-rates.php
        
        Kansas:
        -         
        - https://www.ksrevenue.gov/taxrates.html
          https://www.ksrevenue.gov/incomebook22.html#0
          https://www.ksrevenue.gov/incomebook22.html#0
          https://www.ksrevenue.gov/incomebook22.html#0
          https://www.ksrevenue.gov/incomebook22.html
          https://www.ksrevenue.gov/incomebook22.html
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Maryland', 'state_occupation': 'Kansas', "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def maryland_and_kentucky(self, county_residence='Other', state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Maryland:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Kentucky:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Maryland:
        - https://www.marylandtaxes.gov/individual/income/tax-info/tax-rates.php
          https://www.marylandtaxes.gov/forms/22_forms/Resident_Booklet.pdf
          https://www.marylandtaxes.gov/forms/Tax_Publications/Administrative_Releases/Income_and_Estate_Tax_Releases/ar_it37.pdf
          https://www.marylandtaxes.gov/individual/index.php
          https://www.marylandtaxes.gov/individual/income/tax-info/tax-rates.php
        
        Kentucky:
        -         
        - https://revenue.ky.gov/News/Pages/Dor-Announces-Updates-to-Individual-Income-Tax-for-2023-Tax-Year.aspx
          https://revenue.ky.gov/Individual/Individual-Income-Tax/Pages/default.aspx
          https://revenue.ky.gov/News/Pages/Dor-Announces-Updates-to-Individual-Income-Tax-for-2023-Tax-Year.aspx
          https://revenue.ky.gov/Individual/Individual-Income-Tax/Pages/default.aspx
          https://revenue.ky.gov/Dor%20Training%20Materials/103%20KAR%2017.140.%20Individual%20income%20tax%20-%20reciprocity%20-%20nonresidents.pdf
          https://revenue.ky.gov/Individual/Individual-Income-Tax/Pages/default.aspx
          https://nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2021/Taxes-21-32.htm?taxmap=true
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Maryland', 'state_occupation': 'Kentucky', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def maryland_and_louisiana(self, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Maryland:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Louisiana:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Maryland:
        - https://www.marylandtaxes.gov/individual/income/tax-info/tax-rates.php
          https://www.marylandtaxes.gov/forms/22_forms/Resident_Booklet.pdf
          https://www.marylandtaxes.gov/forms/Tax_Publications/Administrative_Releases/Income_and_Estate_Tax_Releases/ar_it37.pdf
          https://www.marylandtaxes.gov/individual/index.php
          https://www.marylandtaxes.gov/individual/income/tax-info/tax-rates.php
        
        Louisiana:
        -         
        - https://revenue.louisiana.gov/individualincometax
          https://revenue.louisiana.gov/individualincometax
          https://www.revenue.louisiana.gov/taxforms/6935(11_02)F.pdf
          https://revenue.louisiana.gov/individualincometax
          https://revenue.louisiana.gov/individualincometax
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Maryland', 'state_occupation': 'Louisiana', "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def maryland_and_maine(self, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Maryland:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Maine:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Maryland:
        - https://www.marylandtaxes.gov/individual/income/tax-info/tax-rates.php
          https://www.marylandtaxes.gov/forms/22_forms/Resident_Booklet.pdf
          https://www.marylandtaxes.gov/forms/Tax_Publications/Administrative_Releases/Income_and_Estate_Tax_Releases/ar_it37.pdf
          https://www.marylandtaxes.gov/individual/index.php
          https://www.marylandtaxes.gov/individual/income/tax-info/tax-rates.php
        
        Maine:
        -         
        - https://www.maine.gov/revenue/tax-return-forms/individual-income-tax-2022
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/ind_tax_rate_sched_2022.pdf
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/22_1040me_book_gen_instruc_revisedFeb23.pdf
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/22_1040me_book_gen_instruc_revisedFeb23.pdf
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/22_1040me_book_gen_instruc_revisedFeb23.pdf
          https://www.maine.gov/revenue/tax-return-forms/individual-income-tax-2022
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Maryland', 'state_occupation': 'Maine', "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def maryland_and_massachusetts(self, county_residence='Other', state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Maryland:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Massachusetts:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Maryland:
        - https://www.marylandtaxes.gov/individual/income/tax-info/tax-rates.php
          https://www.marylandtaxes.gov/forms/22_forms/Resident_Booklet.pdf
          https://www.marylandtaxes.gov/forms/Tax_Publications/Administrative_Releases/Income_and_Estate_Tax_Releases/ar_it37.pdf
          https://www.marylandtaxes.gov/individual/index.php
          https://www.marylandtaxes.gov/individual/income/tax-info/tax-rates.php
        
        Massachusetts:
        -         
        - https://www.mass.gov/service-details/massachusetts-tax-rates
          https://www.mass.gov/service-details/view-massachusetts-personal-income-tax-exemptions
          https://www.mass.gov/service-details/view-massachusetts-personal-income-tax-exemptions
          https://www.mass.gov/service-details/view-child-and-dependent-related-credits
          https://www.mass.gov/service-details/massachusetts-tax-rates
          https://www.mass.gov/personal-income-tax
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Maryland', 'state_occupation': 'Massachusetts', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def maryland_and_michigan(self, capital_gains_short=0, county_occupation='Other', dependents=0, sources=False, county_residence='Other', state_agi=75000, capital_gains_long=0, filing_status='single', income=100000, credits_state=0):
        """
        --- Parameters ----
        
        Maryland:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Michigan:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        county_occupation : str, optional
            County of occupation. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Maryland:
        - https://www.marylandtaxes.gov/individual/income/tax-info/tax-rates.php
          https://www.marylandtaxes.gov/forms/22_forms/Resident_Booklet.pdf
          https://www.marylandtaxes.gov/forms/Tax_Publications/Administrative_Releases/Income_and_Estate_Tax_Releases/ar_it37.pdf
          https://www.marylandtaxes.gov/individual/index.php
          https://www.marylandtaxes.gov/individual/income/tax-info/tax-rates.php
        
        Michigan:
        -         
        - https://www.michigan.gov/taxes/iit/new-developments-for-tax-year-2022
          "https://www.michigan.gov/taxes/-/media/Project/Websites/taxes/Forms/2023/2023-SUW/446_Withholding-Guide_2023.pdf?rev=177ca3d8cb034392bf84e317ed78c320&hash=19C5F42F43A89A91863D229349F48883#:~:text=The%20exemption%20amount%20is%20%245%2C400
         employees%20Michigan%20income%20tax%20return"
          "https://www.michigan.gov/taxes/-/media/Project/Websites/taxes/Forms/2023/2023-SUW/446_Withholding-Guide_2023.pdf?rev=177ca3d8cb034392bf84e317ed78c320&hash=19C5F42F43A89A91863D229349F48883#:~:text=The%20exemption%20amount%20is%20%245%2C400
         employees%20Michigan%20income%20tax%20return"
          https://www.michigan.gov/taxes/questions/iit/accordion/residency/are-my-wages-earned-in-another-state-taxable-in-michigan-if-i-am-a-michigan-resident-1
          https://www.michigan.gov/taxes/-/media/Project/Websites/taxes/Forms/2022/2022-IIT-Forms/BOOK_MI-1040.pdf?rev=4524f1cbbc0243b194ab05a6680be75b&hash=40DFE71E651F7C1D65C31F6CD03D944F
          https://www.michigan.gov/taxes/questions/iit/accordion/general/what-cities-impose-an-income-tax
        """
        
        # Check if the parameters are set upstream in the instance
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        county_occupation = self.county_occupation if hasattr(self, 'county_occupation') else county_occupation
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Maryland', 'state_occupation': 'Michigan', "capital_gains_short": capital_gains_short, "county_occupation": county_occupation, "dependents": dependents, "sources": sources, "county_residence": county_residence, "state_agi": state_agi, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def maryland_and_minnesota(self, county_residence='Other', state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Maryland:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Minnesota:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Maryland:
        - https://www.marylandtaxes.gov/individual/income/tax-info/tax-rates.php
          https://www.marylandtaxes.gov/forms/22_forms/Resident_Booklet.pdf
          https://www.marylandtaxes.gov/forms/Tax_Publications/Administrative_Releases/Income_and_Estate_Tax_Releases/ar_it37.pdf
          https://www.marylandtaxes.gov/individual/index.php
          https://www.marylandtaxes.gov/individual/income/tax-info/tax-rates.php
        
        Minnesota:
        -         
        - https://www.revenue.state.mn.us/minnesota-income-tax-rates-and-brackets
          https://www.revenue.state.mn.us/child-and-dependent-care-credit
          https://www.revenue.state.mn.us/dependent-exemptions
          https://www.revenue.state.mn.us/minnesota-standard-deduction
          https://mn.gov/mmb/assets/mwr_form_tcm1059-128581.pdf
          https://www.revenue.state.mn.us/individual-income-tax
          https://www.revenue.state.mn.us/individual-income-tax
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Maryland', 'state_occupation': 'Minnesota', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def maryland_and_mississippi(self, county_residence='Other', state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Maryland:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Mississippi:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Maryland:
        - https://www.marylandtaxes.gov/individual/income/tax-info/tax-rates.php
          https://www.marylandtaxes.gov/forms/22_forms/Resident_Booklet.pdf
          https://www.marylandtaxes.gov/forms/Tax_Publications/Administrative_Releases/Income_and_Estate_Tax_Releases/ar_it37.pdf
          https://www.marylandtaxes.gov/individual/index.php
          https://www.marylandtaxes.gov/individual/income/tax-info/tax-rates.php
        
        Mississippi:
        -         
        - https://www.dor.ms.gov/individual/tax-rates
          https://www.dor.ms.gov/individual/tax-rates
          https://www.dor.ms.gov/individual/tax-rates
          https://www.dor.ms.gov/individual
          https://www.dor.ms.gov/individual
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Maryland', 'state_occupation': 'Mississippi', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def maryland_and_missouri(self, capital_gains_short=0, dependents=0, sources=False, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_long=0, filing_status='single', income=100000, credits_state=0):
        """
        --- Parameters ----
        
        Maryland:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Missouri:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Maryland:
        - https://www.marylandtaxes.gov/individual/income/tax-info/tax-rates.php
          https://www.marylandtaxes.gov/forms/22_forms/Resident_Booklet.pdf
          https://www.marylandtaxes.gov/forms/Tax_Publications/Administrative_Releases/Income_and_Estate_Tax_Releases/ar_it37.pdf
          https://www.marylandtaxes.gov/individual/index.php
          https://www.marylandtaxes.gov/individual/income/tax-info/tax-rates.php
        
        Missouri:
        -         
        - https://dor.mo.gov/taxation/individual/tax-types/income/year-changes/
          https://dor.mo.gov/taxation/individual/tax-types/income/year-changes/
          https://dor.mo.gov/taxation/individual/tax-types/income/year-changes/
          https://www.kcmo.gov/city-hall/departments/finance/earnings-tax
          https://www.stlouis-mo.gov/government/departments/collector/earnings-tax/individual-earnings-tax-info.cfm
        """
        
        # Check if the parameters are set upstream in the instance
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Maryland', 'state_occupation': 'Missouri', "capital_gains_short": capital_gains_short, "dependents": dependents, "sources": sources, "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def maryland_and_montana(self, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Maryland:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Montana:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Maryland:
        - https://www.marylandtaxes.gov/individual/income/tax-info/tax-rates.php
          https://www.marylandtaxes.gov/forms/22_forms/Resident_Booklet.pdf
          https://www.marylandtaxes.gov/forms/Tax_Publications/Administrative_Releases/Income_and_Estate_Tax_Releases/ar_it37.pdf
          https://www.marylandtaxes.gov/individual/index.php
          https://www.marylandtaxes.gov/individual/income/tax-info/tax-rates.php
        
        Montana:
        -         
        - https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/2022/12/montana-employees-withholding-allowance-and-exemption-certificate-form-mw-4.pdf
          https://mtrevenue.gov/2021/11/29/simplification-of-montana-income-taxation/?utm_source=rss&utm_medium=rss&utm_campaign=simplification-of-montana-income-taxation
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Maryland', 'state_occupation': 'Montana', "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def maryland_and_nebraska(self, county_residence='Other', state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Maryland:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Nebraska:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Maryland:
        - https://www.marylandtaxes.gov/individual/income/tax-info/tax-rates.php
          https://www.marylandtaxes.gov/forms/22_forms/Resident_Booklet.pdf
          https://www.marylandtaxes.gov/forms/Tax_Publications/Administrative_Releases/Income_and_Estate_Tax_Releases/ar_it37.pdf
          https://www.marylandtaxes.gov/individual/index.php
          https://www.marylandtaxes.gov/individual/income/tax-info/tax-rates.php
        
        Nebraska:
        -         
        - https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2023_Circular_EN_8-429r2023_Whole_Book_11.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Maryland', 'state_occupation': 'Nebraska', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def maryland_and_nevada(self, county_residence='Other', state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Maryland:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Nevada:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Maryland:
        - https://www.marylandtaxes.gov/individual/income/tax-info/tax-rates.php
          https://www.marylandtaxes.gov/forms/22_forms/Resident_Booklet.pdf
          https://www.marylandtaxes.gov/forms/Tax_Publications/Administrative_Releases/Income_and_Estate_Tax_Releases/ar_it37.pdf
          https://www.marylandtaxes.gov/individual/index.php
          https://www.marylandtaxes.gov/individual/income/tax-info/tax-rates.php
        
        Nevada:
        -         
        - https://www.nvsos.gov/sos/businesses/the-nevada-advantage
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Maryland', 'state_occupation': 'Nevada', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def maryland_and_new_hampshire(self, county_residence='Other', state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Maryland:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        New Hampshire:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Maryland:
        - https://www.marylandtaxes.gov/individual/income/tax-info/tax-rates.php
          https://www.marylandtaxes.gov/forms/22_forms/Resident_Booklet.pdf
          https://www.marylandtaxes.gov/forms/Tax_Publications/Administrative_Releases/Income_and_Estate_Tax_Releases/ar_it37.pdf
          https://www.marylandtaxes.gov/individual/index.php
          https://www.marylandtaxes.gov/individual/income/tax-info/tax-rates.php
        
        New Hampshire:
        -         
        - https://www.revenue.nh.gov/assistance/tax-overview.htm
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Maryland', 'state_occupation': 'NewHampshire', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def maryland_and_new_jersey(self, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Maryland:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        New Jersey:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Maryland:
        - https://www.marylandtaxes.gov/individual/income/tax-info/tax-rates.php
          https://www.marylandtaxes.gov/forms/22_forms/Resident_Booklet.pdf
          https://www.marylandtaxes.gov/forms/Tax_Publications/Administrative_Releases/Income_and_Estate_Tax_Releases/ar_it37.pdf
          https://www.marylandtaxes.gov/individual/index.php
          https://www.marylandtaxes.gov/individual/income/tax-info/tax-rates.php
        
        New Jersey:
        -         
        - https://www.state.nj.us/treasury/taxation/taxtables.shtml
          https://www.state.nj.us/treasury/taxation/njit13.shtml
          https://www.nj.gov/treasury/taxation/njit25.shtml
          https://www.state.nj.us/treasury/taxation/individuals/freqqiti.shtml
          https://www.state.nj.us/treasury/taxation/localtax.shtml
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Maryland', 'state_occupation': 'NewJersey', "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def maryland_and_new_mexico(self, county_residence='Other', state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Maryland:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        New Mexico:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Maryland:
        - https://www.marylandtaxes.gov/individual/income/tax-info/tax-rates.php
          https://www.marylandtaxes.gov/forms/22_forms/Resident_Booklet.pdf
          https://www.marylandtaxes.gov/forms/Tax_Publications/Administrative_Releases/Income_and_Estate_Tax_Releases/ar_it37.pdf
          https://www.marylandtaxes.gov/individual/index.php
          https://www.marylandtaxes.gov/individual/income/tax-info/tax-rates.php
        
        New Mexico:
        -         
        - https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/fdf3c548-8aba-4b9c-9eb4-bb564c716015/FYI-104.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/0bf943fd-652e-4400-bb1b-cd8397eb5a95/2022pit-adj.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/2f1a6781-9534-4436-b427-1557f9592099/2022pit-adj-ins.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/0bf943fd-652e-4400-bb1b-cd8397eb5a95/2022pit-adj.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/1afc56af-ea90-4d48-82e5-1f9aeb43255a/PITbook2022.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/fdf3c548-8aba-4b9c-9eb4-bb564c716015/FYI-104.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Maryland', 'state_occupation': 'NewMexico', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def maryland_and_new_york(self, county_residence='Other', state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Maryland:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        New York:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Maryland:
        - https://www.marylandtaxes.gov/individual/income/tax-info/tax-rates.php
          https://www.marylandtaxes.gov/forms/22_forms/Resident_Booklet.pdf
          https://www.marylandtaxes.gov/forms/Tax_Publications/Administrative_Releases/Income_and_Estate_Tax_Releases/ar_it37.pdf
          https://www.marylandtaxes.gov/individual/index.php
          https://www.marylandtaxes.gov/individual/income/tax-info/tax-rates.php
        
        New York:
        -         
        - https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nys-tax-rate-schedule
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nys-tax-rate-schedule
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nys-tax-rate-schedule
          https://www.tax.ny.gov/pit/
          https://www.tax.ny.gov/pit/file/nonresident-faqs.htm
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#yonkers-tax-rate-schedule
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nyc-tax-rate-schedule
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Maryland', 'state_occupation': 'NewYork', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def maryland_and_north_carolina(self, county_residence='Other', state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Maryland:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        North Carolina:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Maryland:
        - https://www.marylandtaxes.gov/individual/income/tax-info/tax-rates.php
          https://www.marylandtaxes.gov/forms/22_forms/Resident_Booklet.pdf
          https://www.marylandtaxes.gov/forms/Tax_Publications/Administrative_Releases/Income_and_Estate_Tax_Releases/ar_it37.pdf
          https://www.marylandtaxes.gov/individual/index.php
          https://www.marylandtaxes.gov/individual/income/tax-info/tax-rates.php
        
        North Carolina:
        -         
        - https://www.ncdor.gov/taxes-forms/tax-rate-schedules
          https://www.ncdor.gov/taxes-forms/individual-income-tax/north-carolina-standard-deduction-or-north-carolina-itemized-deductions
          https://www.ncdor.gov/taxes-forms/individual-income-tax
          https://www.ncdor.gov/taxes-forms/individual-income-tax
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Maryland', 'state_occupation': 'NorthCarolina', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def maryland_and_north_dakota(self, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Maryland:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        North Dakota:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Maryland:
        - https://www.marylandtaxes.gov/individual/income/tax-info/tax-rates.php
          https://www.marylandtaxes.gov/forms/22_forms/Resident_Booklet.pdf
          https://www.marylandtaxes.gov/forms/Tax_Publications/Administrative_Releases/Income_and_Estate_Tax_Releases/ar_it37.pdf
          https://www.marylandtaxes.gov/individual/index.php
          https://www.marylandtaxes.gov/individual/income/tax-info/tax-rates.php
        
        North Dakota:
        -         
        - https://www.tax.nd.gov/forms
          https://www.tax.nd.gov/sites/www/files/documents/forms/individual/2022-iit/2022-individual-income-tax-booklet.pdf
          https://www.tax.nd.gov/sites/www/files/documents/forms/2022-individual-income-tax-booklet.pdf
          https://www.tax.nd.gov/tax-exemptions-credits/income-tax-exemptions-credits
          https://www.tax.nd.gov/sites/www/files/documents/forms/business/ndwrfillable.pdf
          https://www.tax.nd.gov/news/tax-legislative-changes/significant-changes-law/individual-income-tax-history
          https://www.tax.nd.gov/forms
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Maryland', 'state_occupation': 'NorthDakota', "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def maryland_and_ohio(self, county_residence='Other', state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Maryland:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Ohio:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Maryland:
        - https://www.marylandtaxes.gov/individual/income/tax-info/tax-rates.php
          https://www.marylandtaxes.gov/forms/22_forms/Resident_Booklet.pdf
          https://www.marylandtaxes.gov/forms/Tax_Publications/Administrative_Releases/Income_and_Estate_Tax_Releases/ar_it37.pdf
          https://www.marylandtaxes.gov/individual/index.php
          https://www.marylandtaxes.gov/individual/income/tax-info/tax-rates.php
        
        Ohio:
        -         
        - https://tax.ohio.gov/individual/resources/annual-tax-rates
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://tax.ohio.gov/static/forms/employer_withholding/generic/wth_it4nr.pdf
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://www.ritaohio.com/TaxRatesTable
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Maryland', 'state_occupation': 'Ohio', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def maryland_and_oklahoma(self, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Maryland:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Oklahoma:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Maryland:
        - https://www.marylandtaxes.gov/individual/income/tax-info/tax-rates.php
          https://www.marylandtaxes.gov/forms/22_forms/Resident_Booklet.pdf
          https://www.marylandtaxes.gov/forms/Tax_Publications/Administrative_Releases/Income_and_Estate_Tax_Releases/ar_it37.pdf
          https://www.marylandtaxes.gov/individual/index.php
          https://www.marylandtaxes.gov/individual/income/tax-info/tax-rates.php
        
        Oklahoma:
        -         
        - https://oklahoma.gov/content/dam/ok/en/tax/documents/resources/publications/businesses/withholding-tables/WHTables-2022.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Maryland', 'state_occupation': 'Oklahoma', "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def maryland_and_oregon(self, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Maryland:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Oregon:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Maryland:
        - https://www.marylandtaxes.gov/individual/income/tax-info/tax-rates.php
          https://www.marylandtaxes.gov/forms/22_forms/Resident_Booklet.pdf
          https://www.marylandtaxes.gov/forms/Tax_Publications/Administrative_Releases/Income_and_Estate_Tax_Releases/ar_it37.pdf
          https://www.marylandtaxes.gov/individual/index.php
          https://www.marylandtaxes.gov/individual/income/tax-info/tax-rates.php
        
        Oregon:
        -         
        - https://www.oregon.gov/dor/programs/individuals/Pages/PIT.aspx
          https://www.oregon.gov/dor/programs/individuals/Documents/Full%20year%20resident
         %20Form%20or-40%20filers.pdf
          https://www.oregon.gov/dor/programs/individuals/Documents/Part-year%20and%20nonresident
         %20Form%20or-40-P%20and%20or-40-N%20filers.pdf.pdf
          https://www.oregon.gov/dor/forms/FormsPubs/form-or-40_101-040_2022.pdf
          https://www.oregon.gov/dor/forms/FormsPubs/form-or-40_101-040_2022.pdf
          https://www.oregon.gov/dor/forms/FormsPubs/form-or-40_101-040_2022.pdf
          https://www.portland.gov/revenue/personal-tax
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Maryland', 'state_occupation': 'Oregon', "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def maryland_and_pennsylvania(self, county_residence='Other', state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Maryland:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Pennsylvania:
        


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Maryland:
        - https://www.marylandtaxes.gov/individual/income/tax-info/tax-rates.php
          https://www.marylandtaxes.gov/forms/22_forms/Resident_Booklet.pdf
          https://www.marylandtaxes.gov/forms/Tax_Publications/Administrative_Releases/Income_and_Estate_Tax_Releases/ar_it37.pdf
          https://www.marylandtaxes.gov/individual/index.php
          https://www.marylandtaxes.gov/individual/income/tax-info/tax-rates.php
        
        Pennsylvania:
        -         
        - https://www.revenue.pa.gov/Tax%20Rates/Pages/default.aspx
          https://www.revenue.pa.gov/FormsandPublications/FormsforIndividuals/PIT/Documents/2022/2022_pa-40in.pdf
          https://www.revenue.pa.gov/FormsandPublications/FormsforBusinesses/EmployerWithholding/Documents/rev-419.pdf
          https://www.revenue.pa.gov/FormsandPublications/FormsforIndividuals/PIT/Documents/2022/2022_pa-40in.pdf
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-25.htm
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Maryland', 'state_occupation': 'Pennsylvania', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def maryland_and_rhode_island(self, county_residence='Other', state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Maryland:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Rhode Island:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Maryland:
        - https://www.marylandtaxes.gov/individual/income/tax-info/tax-rates.php
          https://www.marylandtaxes.gov/forms/22_forms/Resident_Booklet.pdf
          https://www.marylandtaxes.gov/forms/Tax_Publications/Administrative_Releases/Income_and_Estate_Tax_Releases/ar_it37.pdf
          https://www.marylandtaxes.gov/individual/index.php
          https://www.marylandtaxes.gov/individual/income/tax-info/tax-rates.php
        
        Rhode Island:
        -         
        - https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2022_1040WE_w_0.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Maryland', 'state_occupation': 'RhodeIsland', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def maryland_and_south_carolina(self, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Maryland:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        South Carolina:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Maryland:
        - https://www.marylandtaxes.gov/individual/income/tax-info/tax-rates.php
          https://www.marylandtaxes.gov/forms/22_forms/Resident_Booklet.pdf
          https://www.marylandtaxes.gov/forms/Tax_Publications/Administrative_Releases/Income_and_Estate_Tax_Releases/ar_it37.pdf
          https://www.marylandtaxes.gov/individual/index.php
          https://www.marylandtaxes.gov/individual/income/tax-info/tax-rates.php
        
        South Carolina:
        -         
        - https://dor.sc.gov/resources-site/lawandpolicy/Advisory%20Opinions/IL22-15.pdf
          https://dor.sc.gov/forms-site/Forms/IITPacket_2022.pdf
          https://www.scstatehouse.gov/code/t12c006.php
          https://dor.sc.gov/resources-site/lawandpolicy/Advisory%20Opinions/IL22-15.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Maryland', 'state_occupation': 'SouthCarolina', "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def maryland_and_south_dakota(self, county_residence='Other', state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Maryland:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        South Dakota:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Maryland:
        - https://www.marylandtaxes.gov/individual/income/tax-info/tax-rates.php
          https://www.marylandtaxes.gov/forms/22_forms/Resident_Booklet.pdf
          https://www.marylandtaxes.gov/forms/Tax_Publications/Administrative_Releases/Income_and_Estate_Tax_Releases/ar_it37.pdf
          https://www.marylandtaxes.gov/individual/index.php
          https://www.marylandtaxes.gov/individual/income/tax-info/tax-rates.php
        
        South Dakota:
        -         
        - https://dor.sd.gov/individuals/taxes/
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Maryland', 'state_occupation': 'SouthDakota', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def maryland_and_tennessee(self, county_residence='Other', state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Maryland:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Tennessee:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Maryland:
        - https://www.marylandtaxes.gov/individual/income/tax-info/tax-rates.php
          https://www.marylandtaxes.gov/forms/22_forms/Resident_Booklet.pdf
          https://www.marylandtaxes.gov/forms/Tax_Publications/Administrative_Releases/Income_and_Estate_Tax_Releases/ar_it37.pdf
          https://www.marylandtaxes.gov/individual/index.php
          https://www.marylandtaxes.gov/individual/income/tax-info/tax-rates.php
        
        Tennessee:
        -         
        - https://www.tn.gov/revenue/taxes.html
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Maryland', 'state_occupation': 'Tennessee', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def maryland_and_texas(self, county_residence='Other', state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Maryland:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Texas:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Maryland:
        - https://www.marylandtaxes.gov/individual/income/tax-info/tax-rates.php
          https://www.marylandtaxes.gov/forms/22_forms/Resident_Booklet.pdf
          https://www.marylandtaxes.gov/forms/Tax_Publications/Administrative_Releases/Income_and_Estate_Tax_Releases/ar_it37.pdf
          https://www.marylandtaxes.gov/individual/index.php
          https://www.marylandtaxes.gov/individual/income/tax-info/tax-rates.php
        
        Texas:
        -         
        - https://comptroller.texas.gov/taxes/
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Maryland', 'state_occupation': 'Texas', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def maryland_and_utah(self, county_residence='Other', state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Maryland:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Utah:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Maryland:
        - https://www.marylandtaxes.gov/individual/income/tax-info/tax-rates.php
          https://www.marylandtaxes.gov/forms/22_forms/Resident_Booklet.pdf
          https://www.marylandtaxes.gov/forms/Tax_Publications/Administrative_Releases/Income_and_Estate_Tax_Releases/ar_it37.pdf
          https://www.marylandtaxes.gov/individual/index.php
          https://www.marylandtaxes.gov/individual/income/tax-info/tax-rates.php
        
        Utah:
        -         
        - https://incometax.utah.gov/paying/tax-rates
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Maryland', 'state_occupation': 'Utah', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def maryland_and_vermont(self, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Maryland:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Vermont:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Maryland:
        - https://www.marylandtaxes.gov/individual/income/tax-info/tax-rates.php
          https://www.marylandtaxes.gov/forms/22_forms/Resident_Booklet.pdf
          https://www.marylandtaxes.gov/forms/Tax_Publications/Administrative_Releases/Income_and_Estate_Tax_Releases/ar_it37.pdf
          https://www.marylandtaxes.gov/individual/index.php
          https://www.marylandtaxes.gov/individual/income/tax-info/tax-rates.php
        
        Vermont:
        -         
        - https://tax.vermont.gov/sites/tax/files/documents/GB-1210-2023.pdf
          https://tax.vermont.gov/sites/tax/files/documents/Income%20Booklet-2022.pdf
          https://tax.vermont.gov/sites/tax/files/documents/Income%20Booklet-2022.pdf
          https://tax.vermont.gov/sites/tax/files/documents/Income%20Booklet-2022.pdf
          https://tax.vermont.gov/individuals/personal-income-tax/tax-credits
          https://tax.vermont.gov/individuals/personal-income-tax/taxable-income
          https://tax.vermont.gov/individuals/personal-income-tax/local-option-tax
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Maryland', 'state_occupation': 'Vermont', "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def maryland_and_virginia(self, county_residence='Other', state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Maryland:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Virginia:
        


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Maryland:
        - https://www.marylandtaxes.gov/individual/income/tax-info/tax-rates.php
          https://www.marylandtaxes.gov/forms/22_forms/Resident_Booklet.pdf
          https://www.marylandtaxes.gov/forms/Tax_Publications/Administrative_Releases/Income_and_Estate_Tax_Releases/ar_it37.pdf
          https://www.marylandtaxes.gov/individual/index.php
          https://www.marylandtaxes.gov/individual/income/tax-info/tax-rates.php
        
        Virginia:
        -         
        - https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/reciprocity
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Maryland', 'state_occupation': 'Virginia', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def maryland_and_washington(self, county_residence='Other', state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Maryland:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Washington:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Maryland:
        - https://www.marylandtaxes.gov/individual/income/tax-info/tax-rates.php
          https://www.marylandtaxes.gov/forms/22_forms/Resident_Booklet.pdf
          https://www.marylandtaxes.gov/forms/Tax_Publications/Administrative_Releases/Income_and_Estate_Tax_Releases/ar_it37.pdf
          https://www.marylandtaxes.gov/individual/index.php
          https://www.marylandtaxes.gov/individual/income/tax-info/tax-rates.php
        
        Washington:
        -         
        - https://dor.wa.gov/taxes-rates/income-tax
          https://dor.wa.gov/taxes-rates/other-taxes/capital-gains-tax
          https://dor.wa.gov/taxes-rates/income-tax
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Maryland', 'state_occupation': 'Washington', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def maryland_and_west_virginia(self, pay_periods=24, capital_gains_short=0, county_occupation='Other', dependents=0, sources=False, county_residence='Other', state_agi=75000, capital_gains_long=0, filing_status='single', income=100000, credits_state=0):
        """
        --- Parameters ----
        
        Maryland:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        West Virginia:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_occupation : str, optional
            County of occupation. Default is Other.
        pay_periods : int, optional
            Number of pay periods. Options are 13, 24, 26, 52. Default is 24.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Maryland:
        - https://www.marylandtaxes.gov/individual/income/tax-info/tax-rates.php
          https://www.marylandtaxes.gov/forms/22_forms/Resident_Booklet.pdf
          https://www.marylandtaxes.gov/forms/Tax_Publications/Administrative_Releases/Income_and_Estate_Tax_Releases/ar_it37.pdf
          https://www.marylandtaxes.gov/individual/index.php
          https://www.marylandtaxes.gov/individual/income/tax-info/tax-rates.php
        
        West Virginia:
        -         
        - https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://tax.wv.gov/Documents/TSD/tsd438.pdf
          https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2011/Taxes-11-35att.html
        """
        
        # Check if the parameters are set upstream in the instance
        pay_periods = self.pay_periods if hasattr(self, 'pay_periods') else pay_periods
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        county_occupation = self.county_occupation if hasattr(self, 'county_occupation') else county_occupation
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Maryland', 'state_occupation': 'WestVirginia', "pay_periods": pay_periods, "capital_gains_short": capital_gains_short, "county_occupation": county_occupation, "dependents": dependents, "sources": sources, "county_residence": county_residence, "state_agi": state_agi, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def maryland_and_wisconsin(self, county_residence='Other', state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Maryland:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Wisconsin:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Maryland:
        - https://www.marylandtaxes.gov/individual/income/tax-info/tax-rates.php
          https://www.marylandtaxes.gov/forms/22_forms/Resident_Booklet.pdf
          https://www.marylandtaxes.gov/forms/Tax_Publications/Administrative_Releases/Income_and_Estate_Tax_Releases/ar_it37.pdf
          https://www.marylandtaxes.gov/individual/index.php
          https://www.marylandtaxes.gov/individual/income/tax-info/tax-rates.php
        
        Wisconsin:
        -         
        - https://www.revenue.wi.gov/TaxForms2023/2023-Form1-ES-Inst.pdf
          https://www.revenue.wi.gov/TaxForms2023/2023-Form1-ES-Inst.pdf
          https://www.revenue.wi.gov/TaxForms2023/2023-Form1-ES-Inst.pdf
          https://www.revenue.wi.gov/Pages/FAQS/pcs-work.aspx
          https://www.revenue.wi.gov/Dor%20Publications/pb103.pdf
          https://www.revenue.wi.gov/Pages/FAQS/pcs-work.aspx
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Maryland', 'state_occupation': 'Wisconsin', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def maryland_and_wyoming(self, county_residence='Other', state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Maryland:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Wyoming:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Maryland:
        - https://www.marylandtaxes.gov/individual/income/tax-info/tax-rates.php
          https://www.marylandtaxes.gov/forms/22_forms/Resident_Booklet.pdf
          https://www.marylandtaxes.gov/forms/Tax_Publications/Administrative_Releases/Income_and_Estate_Tax_Releases/ar_it37.pdf
          https://www.marylandtaxes.gov/individual/index.php
          https://www.marylandtaxes.gov/individual/income/tax-info/tax-rates.php
        
        Wyoming:
        -         
        - https://revenue.wyo.gov/
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Maryland', 'state_occupation': 'Wyoming', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def massachusetts_and_alabama(self, county_residence='Other', state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Massachusetts:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Alabama:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Massachusetts:
        - https://www.mass.gov/service-details/massachusetts-tax-rates
          https://www.mass.gov/service-details/view-massachusetts-personal-income-tax-exemptions
          https://www.mass.gov/service-details/view-massachusetts-personal-income-tax-exemptions
          https://www.mass.gov/service-details/view-child-and-dependent-related-credits
          https://www.mass.gov/service-details/massachusetts-tax-rates
          https://www.mass.gov/personal-income-tax
        
        Alabama:
        -         
        - https://www.revenue.alabama.gov/ultraviewer/viewer/basic_viewer/index.html?form=2023/02/22f40.pdf
          https://www.revenue.alabama.gov/faqs/what-is-alabamas-individual-income-tax-rate/
          https://www.revenue.alabama.gov/ultraviewer/viewer/basic_viewer/index.html?form=2023/02/22f40.pdf
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-24.htm?taxmap=true
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-24.htm?taxmap=true
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-24.htm?taxmap=true
          https://www.revenue.alabama.gov/individual-corporate/income-to-be-reported-on-the-alabama-income-tax-return/
          https://www.almonline.org/TaxRates.aspx#OccupationalTax
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Massachusetts', 'state_occupation': 'Alabama', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def massachusetts_and_alaska(self, state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Massachusetts:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Alaska:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Massachusetts:
        - https://www.mass.gov/service-details/massachusetts-tax-rates
          https://www.mass.gov/service-details/view-massachusetts-personal-income-tax-exemptions
          https://www.mass.gov/service-details/view-massachusetts-personal-income-tax-exemptions
          https://www.mass.gov/service-details/view-child-and-dependent-related-credits
          https://www.mass.gov/service-details/massachusetts-tax-rates
          https://www.mass.gov/personal-income-tax
        
        Alaska:
        -         
        - https://www.commerce.alaska.gov/web/dcra/officeofthestateassessor/alaskataxfacts.aspx
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Massachusetts', 'state_occupation': 'Alaska', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def massachusetts_and_arizona(self, state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Massachusetts:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Arizona:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        credits_state : int, optional
            Credits for state taxes. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Massachusetts:
        - https://www.mass.gov/service-details/massachusetts-tax-rates
          https://www.mass.gov/service-details/view-massachusetts-personal-income-tax-exemptions
          https://www.mass.gov/service-details/view-massachusetts-personal-income-tax-exemptions
          https://www.mass.gov/service-details/view-child-and-dependent-related-credits
          https://www.mass.gov/service-details/massachusetts-tax-rates
          https://www.mass.gov/personal-income-tax
        
        Arizona:
        -         
        - https://azdor.gov/forms/individual/form-140-x-y-tables
          https://azdor.gov/forms/individual/form-140-x-y-tables
          https://azdor.gov/forms/individual/form-140-x-y-tables
          https://azdor.gov/businesses-arizona/withholding-tax/withholding-exceptions
          "https://azdor.gov/individuals/income-tax-filing-assistance/identifying-other-taxable-income#:~:text=In%20regards%20to%20capital%20gains
         the%20individuals%20regular%20tax%20rate.
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Massachusetts', 'state_occupation': 'Arizona', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def massachusetts_and_arkansas(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Massachusetts:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Arkansas:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Massachusetts:
        - https://www.mass.gov/service-details/massachusetts-tax-rates
          https://www.mass.gov/service-details/view-massachusetts-personal-income-tax-exemptions
          https://www.mass.gov/service-details/view-massachusetts-personal-income-tax-exemptions
          https://www.mass.gov/service-details/view-child-and-dependent-related-credits
          https://www.mass.gov/service-details/massachusetts-tax-rates
          https://www.mass.gov/personal-income-tax
        
        Arkansas:
        -         
        - https://www.arkansasedc.com/why-arkansas/business-climate/tax-structure/personal-income-tax
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_and_AR1000NR_Instructions.pdf
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_FullYearResidentIndividualIncomeTaxReturn.pdf
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_and_AR1000NR_Instructions.pdf
          https://www.arkansasedc.com/why-arkansas/business-climate/tax-structure/capital-gains-tax-reduction
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_and_AR1000NR_Instructions.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Massachusetts', 'state_occupation': 'Arkansas', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def massachusetts_and_california(self, state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Massachusetts:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        California:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Massachusetts:
        - https://www.mass.gov/service-details/massachusetts-tax-rates
          https://www.mass.gov/service-details/view-massachusetts-personal-income-tax-exemptions
          https://www.mass.gov/service-details/view-massachusetts-personal-income-tax-exemptions
          https://www.mass.gov/service-details/view-child-and-dependent-related-credits
          https://www.mass.gov/service-details/massachusetts-tax-rates
          https://www.mass.gov/personal-income-tax
        
        California:
        -         
        - https://www.ftb.ca.gov/forms/2022/2022-540-tax-rate-schedules.pdf
          https://www.ftb.ca.gov/file/personal/deductions/index.html
          https://www.ftb.ca.gov/about-ftb/newsroom/caleitc/eligibility-and-credit-information.html
          https://www.ftb.ca.gov/file/personal/income-types/capital-gains-and-losses.html#:~:text=California%20does%20not%20have%20a
         are%20taxed%20as%20ordinary%20income
          https://www.ftb.ca.gov/file/personal/
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Massachusetts', 'state_occupation': 'California', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def massachusetts_and_colorado(self, county_residence='Other', state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Massachusetts:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Colorado:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Massachusetts:
        - https://www.mass.gov/service-details/massachusetts-tax-rates
          https://www.mass.gov/service-details/view-massachusetts-personal-income-tax-exemptions
          https://www.mass.gov/service-details/view-massachusetts-personal-income-tax-exemptions
          https://www.mass.gov/service-details/view-child-and-dependent-related-credits
          https://www.mass.gov/service-details/massachusetts-tax-rates
          https://www.mass.gov/personal-income-tax
        
        Colorado:
        -         
        - https://leg.colorado.gov/agencies/legislative-council-staff/individual-income-tax%C2%A0
          https://leg.colorado.gov/agencies/legislative-council-staff/individual-income-tax%C2%A0
          https://tax.colorado.gov/sites/tax/files/documents/ITT_Capital_Gain_Subtraction_Feb_2022.pdf
          https://www.auroragov.org/business_services/taxes/occupational_privilege_tax
          https://www.denvergov.org/Government/Agencies-Departments-Offices/Agencies-Departments-Offices-Directory/Department-of-Finance/Our-Divisions/Treasury/Business-Tax-Information#:~:text=Occupational%20privilege%20tax
         -%28show%20below%29&text=Each%20taxable%20employee%20is%20liable
         month%20for%20each%20taxable%20employee
          https://www.glendale.co.us/355/Occupational-Privilege-Tax
          https://www.greenwoodvillage.com/1220/Occupational-Privilege-Tax-OPT
          https://ci.sheridan.co.us/288/Occupational-Privilege-Tax
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Massachusetts', 'state_occupation': 'Colorado', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def massachusetts_and_connecticut(self, state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Massachusetts:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Connecticut:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Massachusetts:
        - https://www.mass.gov/service-details/massachusetts-tax-rates
          https://www.mass.gov/service-details/view-massachusetts-personal-income-tax-exemptions
          https://www.mass.gov/service-details/view-massachusetts-personal-income-tax-exemptions
          https://www.mass.gov/service-details/view-child-and-dependent-related-credits
          https://www.mass.gov/service-details/massachusetts-tax-rates
          https://www.mass.gov/personal-income-tax
        
        Connecticut:
        -         
        - https://portal.ct.gov/-/media/DRS/Forms/2022/Income/2022-CT-1040-Instructions_1222.pdf
          https://www.cga.ct.gov/2022/rpt/pdf/2022-R-0108.pdf
          https://www.cga.ct.gov/current/pub/chap_229.htm#sec_12-703
          https://portal.ct.gov/-/media/DRS/Forms/2022/Income/2022-CT-1040-Instructions_1222.pdf
          "https://portal.ct.gov/DRS/Publications/TSSNs/TSSN-29#:~:text=An%20individuals%20net%20capital%20gains
         portion%20of%20Social%20Security%20benefits"
          "https://portal.ct.gov/DRS/Publications/TSSNs/TSSN-29#:~:text=An%20individuals%20net%20capital%20gains
         portion%20of%20Social%20Security%20benefits"
          https://portal.ct.gov/DRS/Individuals/Resident-Income-Tax/Tax-Information
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Massachusetts', 'state_occupation': 'Connecticut', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def massachusetts_and_delaware(self, state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, federal_agi=75000, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Massachusetts:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Delaware:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        federal_agi : int, optional
            Adjusted Gross Income (AGI) for federal taxes. Default is 0. Sources - https://www.irs.gov/e-file-providers/definition-of-adjusted-gross-income, https://www.irs.gov/newsroom/taxpayers-should-know-the-difference-between-standard-and-itemized-deductions
        income : int, optional
            Gross income. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Massachusetts:
        - https://www.mass.gov/service-details/massachusetts-tax-rates
          https://www.mass.gov/service-details/view-massachusetts-personal-income-tax-exemptions
          https://www.mass.gov/service-details/view-massachusetts-personal-income-tax-exemptions
          https://www.mass.gov/service-details/view-child-and-dependent-related-credits
          https://www.mass.gov/service-details/massachusetts-tax-rates
          https://www.mass.gov/personal-income-tax
        
        Delaware:
        -         
        - https://revenue.delaware.gov/software-developer/tax-rate-changes/
          https://revenuefiles.delaware.gov/2022/PIT-RES_TY22_2022-01_PaperInteractive.pdf
          https://revenuefiles.delaware.gov/2022/PIT-RES_TY22_2022-02_Instructions.pdf
          https://revenue.delaware.gov/frequently-asked-questions/personal-income-tax-faqs/
          https://www.newcastlede.gov/DocumentCenter/View/20147/City-of-Wilmington-Wage-Tax-Information-PDF
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        federal_agi = self.federal_agi if hasattr(self, 'federal_agi') else federal_agi
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Massachusetts', 'state_occupation': 'Delaware', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "federal_agi": federal_agi, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def massachusetts_and_florida(self, state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Massachusetts:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Florida:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Massachusetts:
        - https://www.mass.gov/service-details/massachusetts-tax-rates
          https://www.mass.gov/service-details/view-massachusetts-personal-income-tax-exemptions
          https://www.mass.gov/service-details/view-massachusetts-personal-income-tax-exemptions
          https://www.mass.gov/service-details/view-child-and-dependent-related-credits
          https://www.mass.gov/service-details/massachusetts-tax-rates
          https://www.mass.gov/personal-income-tax
        
        Florida:
        -         
        - https://floridarevenue.com/taxes/taxesfees/Pages/tax_interest_rates.aspx
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Massachusetts', 'state_occupation': 'Florida', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def massachusetts_and_georgia(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Massachusetts:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Georgia:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Massachusetts:
        - https://www.mass.gov/service-details/massachusetts-tax-rates
          https://www.mass.gov/service-details/view-massachusetts-personal-income-tax-exemptions
          https://www.mass.gov/service-details/view-massachusetts-personal-income-tax-exemptions
          https://www.mass.gov/service-details/view-child-and-dependent-related-credits
          https://www.mass.gov/service-details/massachusetts-tax-rates
          https://www.mass.gov/personal-income-tax
        
        Georgia:
        -         
        - https://dor.georgia.gov/tax-tables-georgia-tax-rate-schedule
          https://apps.dor.ga.gov/FillableForms/PDFViewer/Index?form=2022GA500
          https://apps.dor.ga.gov/FillableForms/PDFViewer/Index?form=2022GA500
          https://apps.dor.ga.gov/FillableForms/PDFViewer/Index?form=2022GA500
          https://dor.georgia.gov/years-individual-income-tax-forms
          https://dor.georgia.gov/years-individual-income-tax-forms
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Massachusetts', 'state_occupation': 'Georgia', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def massachusetts_and_hawaii(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Massachusetts:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Hawaii:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Massachusetts:
        - https://www.mass.gov/service-details/massachusetts-tax-rates
          https://www.mass.gov/service-details/view-massachusetts-personal-income-tax-exemptions
          https://www.mass.gov/service-details/view-massachusetts-personal-income-tax-exemptions
          https://www.mass.gov/service-details/view-child-and-dependent-related-credits
          https://www.mass.gov/service-details/massachusetts-tax-rates
          https://www.mass.gov/personal-income-tax
        
        Hawaii:
        -         
        - https://tax.hawaii.gov/forms/d_18table-on/d_18table-on_p13/
          https://files.hawaii.gov/tax/forms/2018/18table-on.pdf
          https://files.hawaii.gov/tax/news/pubs/22outline.pdf
          https://files.hawaii.gov/tax/news/pubs/FastTaxRefGuide_2017.pdf
          https://www.tfhawaii.org/wordpress/state-tax-resources/mini-tax-guides/individual-income-tax-chapter-235/#:~:text=Long%20term%20capital%20gains%20are%20taxed%20at%20a%20maximum%20of%207.25%25
          https://www.grassrootinstitute.org/2023/02/hb337-increasing-tax-on-capital-gains-a-bad-risk/
          https://files.hawaii.gov/tax/news/pubs/22outline.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Massachusetts', 'state_occupation': 'Hawaii', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def massachusetts_and_idaho(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Massachusetts:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Idaho:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Massachusetts:
        - https://www.mass.gov/service-details/massachusetts-tax-rates
          https://www.mass.gov/service-details/view-massachusetts-personal-income-tax-exemptions
          https://www.mass.gov/service-details/view-massachusetts-personal-income-tax-exemptions
          https://www.mass.gov/service-details/view-child-and-dependent-related-credits
          https://www.mass.gov/service-details/massachusetts-tax-rates
          https://www.mass.gov/personal-income-tax
        
        Idaho:
        -         
        - https://tax.idaho.gov/taxes/income-tax/individual-income/individual-income-tax-rate-schedule/
          https://tax.idaho.gov/wp-content/uploads/forms/EIN00046/EIN00046_03-01-2023.pdf
          https://tax.idaho.gov/taxes/income-tax/individual-income/idaho-source-income/
          https://tax.idaho.gov/taxes/income-tax/individual-income/idaho-source-income/
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Massachusetts', 'state_occupation': 'Idaho', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def massachusetts_and_illinois(self, state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Massachusetts:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Illinois:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Massachusetts:
        - https://www.mass.gov/service-details/massachusetts-tax-rates
          https://www.mass.gov/service-details/view-massachusetts-personal-income-tax-exemptions
          https://www.mass.gov/service-details/view-massachusetts-personal-income-tax-exemptions
          https://www.mass.gov/service-details/view-child-and-dependent-related-credits
          https://www.mass.gov/service-details/massachusetts-tax-rates
          https://www.mass.gov/personal-income-tax
        
        Illinois:
        -         
        - https://tax.illinois.gov/research/taxrates/income.html#IndividualIncome
          https://tax.illinois.gov/content/dam/soi/en/web/tax/forms/incometax/documents/currentyear/individual/il-1040-instr.pdf
          https://tax.illinois.gov/questionsandanswers/12.html
          https://tax.illinois.gov/content/dam/soi/en/web/tax/forms/incometax/documents/currentyear/individual/il-1040-instr.pdf
          https://tax.illinois.gov/research/taxrates/personalproperty.html
          https://tax.illinois.gov/localgovernments/income.html
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Massachusetts', 'state_occupation': 'Illinois', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def massachusetts_and_indiana(self, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Massachusetts:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Indiana:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Massachusetts:
        - https://www.mass.gov/service-details/massachusetts-tax-rates
          https://www.mass.gov/service-details/view-massachusetts-personal-income-tax-exemptions
          https://www.mass.gov/service-details/view-massachusetts-personal-income-tax-exemptions
          https://www.mass.gov/service-details/view-child-and-dependent-related-credits
          https://www.mass.gov/service-details/massachusetts-tax-rates
          https://www.mass.gov/personal-income-tax
        
        Indiana:
        -         
        - https://www.in.gov/dor/tax-forms/2022-individual-income-tax-forms/
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-32.htm
          https://www.in.gov/dor/files/ib28.pdf
          https://www.in.gov/dor/tax-forms/2022-individual-income-tax-forms/
          https://www.in.gov/dor/files/2022-county-tax-rates-and-codes.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Massachusetts', 'state_occupation': 'Indiana', "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def massachusetts_and_iowa(self, county_residence='Other', state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Massachusetts:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Iowa:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Massachusetts:
        - https://www.mass.gov/service-details/massachusetts-tax-rates
          https://www.mass.gov/service-details/view-massachusetts-personal-income-tax-exemptions
          https://www.mass.gov/service-details/view-massachusetts-personal-income-tax-exemptions
          https://www.mass.gov/service-details/view-child-and-dependent-related-credits
          https://www.mass.gov/service-details/massachusetts-tax-rates
          https://www.mass.gov/personal-income-tax
        
        Iowa:
        -         
        - https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
          https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
          https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
          https://tax.iowa.gov/iowa-illinois-reciprocal-agreement
          https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Massachusetts', 'state_occupation': 'Iowa', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def massachusetts_and_kansas(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Massachusetts:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Kansas:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Massachusetts:
        - https://www.mass.gov/service-details/massachusetts-tax-rates
          https://www.mass.gov/service-details/view-massachusetts-personal-income-tax-exemptions
          https://www.mass.gov/service-details/view-massachusetts-personal-income-tax-exemptions
          https://www.mass.gov/service-details/view-child-and-dependent-related-credits
          https://www.mass.gov/service-details/massachusetts-tax-rates
          https://www.mass.gov/personal-income-tax
        
        Kansas:
        -         
        - https://www.ksrevenue.gov/taxrates.html
          https://www.ksrevenue.gov/incomebook22.html#0
          https://www.ksrevenue.gov/incomebook22.html#0
          https://www.ksrevenue.gov/incomebook22.html#0
          https://www.ksrevenue.gov/incomebook22.html
          https://www.ksrevenue.gov/incomebook22.html
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Massachusetts', 'state_occupation': 'Kansas', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def massachusetts_and_kentucky(self, county_residence='Other', state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Massachusetts:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Kentucky:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Massachusetts:
        - https://www.mass.gov/service-details/massachusetts-tax-rates
          https://www.mass.gov/service-details/view-massachusetts-personal-income-tax-exemptions
          https://www.mass.gov/service-details/view-massachusetts-personal-income-tax-exemptions
          https://www.mass.gov/service-details/view-child-and-dependent-related-credits
          https://www.mass.gov/service-details/massachusetts-tax-rates
          https://www.mass.gov/personal-income-tax
        
        Kentucky:
        -         
        - https://revenue.ky.gov/News/Pages/Dor-Announces-Updates-to-Individual-Income-Tax-for-2023-Tax-Year.aspx
          https://revenue.ky.gov/Individual/Individual-Income-Tax/Pages/default.aspx
          https://revenue.ky.gov/News/Pages/Dor-Announces-Updates-to-Individual-Income-Tax-for-2023-Tax-Year.aspx
          https://revenue.ky.gov/Individual/Individual-Income-Tax/Pages/default.aspx
          https://revenue.ky.gov/Dor%20Training%20Materials/103%20KAR%2017.140.%20Individual%20income%20tax%20-%20reciprocity%20-%20nonresidents.pdf
          https://revenue.ky.gov/Individual/Individual-Income-Tax/Pages/default.aspx
          https://nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2021/Taxes-21-32.htm?taxmap=true
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Massachusetts', 'state_occupation': 'Kentucky', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def massachusetts_and_louisiana(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Massachusetts:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Louisiana:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Massachusetts:
        - https://www.mass.gov/service-details/massachusetts-tax-rates
          https://www.mass.gov/service-details/view-massachusetts-personal-income-tax-exemptions
          https://www.mass.gov/service-details/view-massachusetts-personal-income-tax-exemptions
          https://www.mass.gov/service-details/view-child-and-dependent-related-credits
          https://www.mass.gov/service-details/massachusetts-tax-rates
          https://www.mass.gov/personal-income-tax
        
        Louisiana:
        -         
        - https://revenue.louisiana.gov/individualincometax
          https://revenue.louisiana.gov/individualincometax
          https://www.revenue.louisiana.gov/taxforms/6935(11_02)F.pdf
          https://revenue.louisiana.gov/individualincometax
          https://revenue.louisiana.gov/individualincometax
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Massachusetts', 'state_occupation': 'Louisiana', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def massachusetts_and_maine(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Massachusetts:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Maine:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Massachusetts:
        - https://www.mass.gov/service-details/massachusetts-tax-rates
          https://www.mass.gov/service-details/view-massachusetts-personal-income-tax-exemptions
          https://www.mass.gov/service-details/view-massachusetts-personal-income-tax-exemptions
          https://www.mass.gov/service-details/view-child-and-dependent-related-credits
          https://www.mass.gov/service-details/massachusetts-tax-rates
          https://www.mass.gov/personal-income-tax
        
        Maine:
        -         
        - https://www.maine.gov/revenue/tax-return-forms/individual-income-tax-2022
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/ind_tax_rate_sched_2022.pdf
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/22_1040me_book_gen_instruc_revisedFeb23.pdf
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/22_1040me_book_gen_instruc_revisedFeb23.pdf
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/22_1040me_book_gen_instruc_revisedFeb23.pdf
          https://www.maine.gov/revenue/tax-return-forms/individual-income-tax-2022
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Massachusetts', 'state_occupation': 'Maine', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def massachusetts_and_maryland(self, county_residence='Other', state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Massachusetts:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Maryland:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Massachusetts:
        - https://www.mass.gov/service-details/massachusetts-tax-rates
          https://www.mass.gov/service-details/view-massachusetts-personal-income-tax-exemptions
          https://www.mass.gov/service-details/view-massachusetts-personal-income-tax-exemptions
          https://www.mass.gov/service-details/view-child-and-dependent-related-credits
          https://www.mass.gov/service-details/massachusetts-tax-rates
          https://www.mass.gov/personal-income-tax
        
        Maryland:
        -         
        - https://www.marylandtaxes.gov/individual/income/tax-info/tax-rates.php
          https://www.marylandtaxes.gov/forms/22_forms/Resident_Booklet.pdf
          https://www.marylandtaxes.gov/forms/Tax_Publications/Administrative_Releases/Income_and_Estate_Tax_Releases/ar_it37.pdf
          https://www.marylandtaxes.gov/individual/index.php
          https://www.marylandtaxes.gov/individual/income/tax-info/tax-rates.php
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Massachusetts', 'state_occupation': 'Maryland', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def massachusetts_and_michigan(self, county_occupation='Other', state_agi=75000, county_residence='Other', capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Massachusetts:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Michigan:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        county_occupation : str, optional
            County of occupation. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Massachusetts:
        - https://www.mass.gov/service-details/massachusetts-tax-rates
          https://www.mass.gov/service-details/view-massachusetts-personal-income-tax-exemptions
          https://www.mass.gov/service-details/view-massachusetts-personal-income-tax-exemptions
          https://www.mass.gov/service-details/view-child-and-dependent-related-credits
          https://www.mass.gov/service-details/massachusetts-tax-rates
          https://www.mass.gov/personal-income-tax
        
        Michigan:
        -         
        - https://www.michigan.gov/taxes/iit/new-developments-for-tax-year-2022
          "https://www.michigan.gov/taxes/-/media/Project/Websites/taxes/Forms/2023/2023-SUW/446_Withholding-Guide_2023.pdf?rev=177ca3d8cb034392bf84e317ed78c320&hash=19C5F42F43A89A91863D229349F48883#:~:text=The%20exemption%20amount%20is%20%245%2C400
         employees%20Michigan%20income%20tax%20return"
          "https://www.michigan.gov/taxes/-/media/Project/Websites/taxes/Forms/2023/2023-SUW/446_Withholding-Guide_2023.pdf?rev=177ca3d8cb034392bf84e317ed78c320&hash=19C5F42F43A89A91863D229349F48883#:~:text=The%20exemption%20amount%20is%20%245%2C400
         employees%20Michigan%20income%20tax%20return"
          https://www.michigan.gov/taxes/questions/iit/accordion/residency/are-my-wages-earned-in-another-state-taxable-in-michigan-if-i-am-a-michigan-resident-1
          https://www.michigan.gov/taxes/-/media/Project/Websites/taxes/Forms/2022/2022-IIT-Forms/BOOK_MI-1040.pdf?rev=4524f1cbbc0243b194ab05a6680be75b&hash=40DFE71E651F7C1D65C31F6CD03D944F
          https://www.michigan.gov/taxes/questions/iit/accordion/general/what-cities-impose-an-income-tax
        """
        
        # Check if the parameters are set upstream in the instance
        county_occupation = self.county_occupation if hasattr(self, 'county_occupation') else county_occupation
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Massachusetts', 'state_occupation': 'Michigan', "county_occupation": county_occupation, "state_agi": state_agi, "county_residence": county_residence, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def massachusetts_and_minnesota(self, state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Massachusetts:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Minnesota:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Massachusetts:
        - https://www.mass.gov/service-details/massachusetts-tax-rates
          https://www.mass.gov/service-details/view-massachusetts-personal-income-tax-exemptions
          https://www.mass.gov/service-details/view-massachusetts-personal-income-tax-exemptions
          https://www.mass.gov/service-details/view-child-and-dependent-related-credits
          https://www.mass.gov/service-details/massachusetts-tax-rates
          https://www.mass.gov/personal-income-tax
        
        Minnesota:
        -         
        - https://www.revenue.state.mn.us/minnesota-income-tax-rates-and-brackets
          https://www.revenue.state.mn.us/child-and-dependent-care-credit
          https://www.revenue.state.mn.us/dependent-exemptions
          https://www.revenue.state.mn.us/minnesota-standard-deduction
          https://mn.gov/mmb/assets/mwr_form_tcm1059-128581.pdf
          https://www.revenue.state.mn.us/individual-income-tax
          https://www.revenue.state.mn.us/individual-income-tax
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Massachusetts', 'state_occupation': 'Minnesota', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def massachusetts_and_mississippi(self, state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Massachusetts:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Mississippi:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Massachusetts:
        - https://www.mass.gov/service-details/massachusetts-tax-rates
          https://www.mass.gov/service-details/view-massachusetts-personal-income-tax-exemptions
          https://www.mass.gov/service-details/view-massachusetts-personal-income-tax-exemptions
          https://www.mass.gov/service-details/view-child-and-dependent-related-credits
          https://www.mass.gov/service-details/massachusetts-tax-rates
          https://www.mass.gov/personal-income-tax
        
        Mississippi:
        -         
        - https://www.dor.ms.gov/individual/tax-rates
          https://www.dor.ms.gov/individual/tax-rates
          https://www.dor.ms.gov/individual/tax-rates
          https://www.dor.ms.gov/individual
          https://www.dor.ms.gov/individual
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Massachusetts', 'state_occupation': 'Mississippi', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def massachusetts_and_missouri(self, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Massachusetts:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Missouri:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Massachusetts:
        - https://www.mass.gov/service-details/massachusetts-tax-rates
          https://www.mass.gov/service-details/view-massachusetts-personal-income-tax-exemptions
          https://www.mass.gov/service-details/view-massachusetts-personal-income-tax-exemptions
          https://www.mass.gov/service-details/view-child-and-dependent-related-credits
          https://www.mass.gov/service-details/massachusetts-tax-rates
          https://www.mass.gov/personal-income-tax
        
        Missouri:
        -         
        - https://dor.mo.gov/taxation/individual/tax-types/income/year-changes/
          https://dor.mo.gov/taxation/individual/tax-types/income/year-changes/
          https://dor.mo.gov/taxation/individual/tax-types/income/year-changes/
          https://www.kcmo.gov/city-hall/departments/finance/earnings-tax
          https://www.stlouis-mo.gov/government/departments/collector/earnings-tax/individual-earnings-tax-info.cfm
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Massachusetts', 'state_occupation': 'Missouri', "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def massachusetts_and_montana(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Massachusetts:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Montana:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Massachusetts:
        - https://www.mass.gov/service-details/massachusetts-tax-rates
          https://www.mass.gov/service-details/view-massachusetts-personal-income-tax-exemptions
          https://www.mass.gov/service-details/view-massachusetts-personal-income-tax-exemptions
          https://www.mass.gov/service-details/view-child-and-dependent-related-credits
          https://www.mass.gov/service-details/massachusetts-tax-rates
          https://www.mass.gov/personal-income-tax
        
        Montana:
        -         
        - https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/2022/12/montana-employees-withholding-allowance-and-exemption-certificate-form-mw-4.pdf
          https://mtrevenue.gov/2021/11/29/simplification-of-montana-income-taxation/?utm_source=rss&utm_medium=rss&utm_campaign=simplification-of-montana-income-taxation
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Massachusetts', 'state_occupation': 'Montana', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def massachusetts_and_nebraska(self, state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Massachusetts:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Nebraska:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Massachusetts:
        - https://www.mass.gov/service-details/massachusetts-tax-rates
          https://www.mass.gov/service-details/view-massachusetts-personal-income-tax-exemptions
          https://www.mass.gov/service-details/view-massachusetts-personal-income-tax-exemptions
          https://www.mass.gov/service-details/view-child-and-dependent-related-credits
          https://www.mass.gov/service-details/massachusetts-tax-rates
          https://www.mass.gov/personal-income-tax
        
        Nebraska:
        -         
        - https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2023_Circular_EN_8-429r2023_Whole_Book_11.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Massachusetts', 'state_occupation': 'Nebraska', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def massachusetts_and_nevada(self, state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Massachusetts:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Nevada:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Massachusetts:
        - https://www.mass.gov/service-details/massachusetts-tax-rates
          https://www.mass.gov/service-details/view-massachusetts-personal-income-tax-exemptions
          https://www.mass.gov/service-details/view-massachusetts-personal-income-tax-exemptions
          https://www.mass.gov/service-details/view-child-and-dependent-related-credits
          https://www.mass.gov/service-details/massachusetts-tax-rates
          https://www.mass.gov/personal-income-tax
        
        Nevada:
        -         
        - https://www.nvsos.gov/sos/businesses/the-nevada-advantage
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Massachusetts', 'state_occupation': 'Nevada', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def massachusetts_and_new_hampshire(self, state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Massachusetts:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        New Hampshire:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Massachusetts:
        - https://www.mass.gov/service-details/massachusetts-tax-rates
          https://www.mass.gov/service-details/view-massachusetts-personal-income-tax-exemptions
          https://www.mass.gov/service-details/view-massachusetts-personal-income-tax-exemptions
          https://www.mass.gov/service-details/view-child-and-dependent-related-credits
          https://www.mass.gov/service-details/massachusetts-tax-rates
          https://www.mass.gov/personal-income-tax
        
        New Hampshire:
        -         
        - https://www.revenue.nh.gov/assistance/tax-overview.htm
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Massachusetts', 'state_occupation': 'NewHampshire', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def massachusetts_and_new_jersey(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Massachusetts:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        New Jersey:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Massachusetts:
        - https://www.mass.gov/service-details/massachusetts-tax-rates
          https://www.mass.gov/service-details/view-massachusetts-personal-income-tax-exemptions
          https://www.mass.gov/service-details/view-massachusetts-personal-income-tax-exemptions
          https://www.mass.gov/service-details/view-child-and-dependent-related-credits
          https://www.mass.gov/service-details/massachusetts-tax-rates
          https://www.mass.gov/personal-income-tax
        
        New Jersey:
        -         
        - https://www.state.nj.us/treasury/taxation/taxtables.shtml
          https://www.state.nj.us/treasury/taxation/njit13.shtml
          https://www.nj.gov/treasury/taxation/njit25.shtml
          https://www.state.nj.us/treasury/taxation/individuals/freqqiti.shtml
          https://www.state.nj.us/treasury/taxation/localtax.shtml
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Massachusetts', 'state_occupation': 'NewJersey', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def massachusetts_and_new_mexico(self, state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Massachusetts:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        New Mexico:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Massachusetts:
        - https://www.mass.gov/service-details/massachusetts-tax-rates
          https://www.mass.gov/service-details/view-massachusetts-personal-income-tax-exemptions
          https://www.mass.gov/service-details/view-massachusetts-personal-income-tax-exemptions
          https://www.mass.gov/service-details/view-child-and-dependent-related-credits
          https://www.mass.gov/service-details/massachusetts-tax-rates
          https://www.mass.gov/personal-income-tax
        
        New Mexico:
        -         
        - https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/fdf3c548-8aba-4b9c-9eb4-bb564c716015/FYI-104.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/0bf943fd-652e-4400-bb1b-cd8397eb5a95/2022pit-adj.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/2f1a6781-9534-4436-b427-1557f9592099/2022pit-adj-ins.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/0bf943fd-652e-4400-bb1b-cd8397eb5a95/2022pit-adj.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/1afc56af-ea90-4d48-82e5-1f9aeb43255a/PITbook2022.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/fdf3c548-8aba-4b9c-9eb4-bb564c716015/FYI-104.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Massachusetts', 'state_occupation': 'NewMexico', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def massachusetts_and_new_york(self, county_residence='Other', state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Massachusetts:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        New York:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Massachusetts:
        - https://www.mass.gov/service-details/massachusetts-tax-rates
          https://www.mass.gov/service-details/view-massachusetts-personal-income-tax-exemptions
          https://www.mass.gov/service-details/view-massachusetts-personal-income-tax-exemptions
          https://www.mass.gov/service-details/view-child-and-dependent-related-credits
          https://www.mass.gov/service-details/massachusetts-tax-rates
          https://www.mass.gov/personal-income-tax
        
        New York:
        -         
        - https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nys-tax-rate-schedule
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nys-tax-rate-schedule
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nys-tax-rate-schedule
          https://www.tax.ny.gov/pit/
          https://www.tax.ny.gov/pit/file/nonresident-faqs.htm
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#yonkers-tax-rate-schedule
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nyc-tax-rate-schedule
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Massachusetts', 'state_occupation': 'NewYork', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def massachusetts_and_north_carolina(self, state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Massachusetts:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        North Carolina:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Massachusetts:
        - https://www.mass.gov/service-details/massachusetts-tax-rates
          https://www.mass.gov/service-details/view-massachusetts-personal-income-tax-exemptions
          https://www.mass.gov/service-details/view-massachusetts-personal-income-tax-exemptions
          https://www.mass.gov/service-details/view-child-and-dependent-related-credits
          https://www.mass.gov/service-details/massachusetts-tax-rates
          https://www.mass.gov/personal-income-tax
        
        North Carolina:
        -         
        - https://www.ncdor.gov/taxes-forms/tax-rate-schedules
          https://www.ncdor.gov/taxes-forms/individual-income-tax/north-carolina-standard-deduction-or-north-carolina-itemized-deductions
          https://www.ncdor.gov/taxes-forms/individual-income-tax
          https://www.ncdor.gov/taxes-forms/individual-income-tax
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Massachusetts', 'state_occupation': 'NorthCarolina', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def massachusetts_and_north_dakota(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Massachusetts:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        North Dakota:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Massachusetts:
        - https://www.mass.gov/service-details/massachusetts-tax-rates
          https://www.mass.gov/service-details/view-massachusetts-personal-income-tax-exemptions
          https://www.mass.gov/service-details/view-massachusetts-personal-income-tax-exemptions
          https://www.mass.gov/service-details/view-child-and-dependent-related-credits
          https://www.mass.gov/service-details/massachusetts-tax-rates
          https://www.mass.gov/personal-income-tax
        
        North Dakota:
        -         
        - https://www.tax.nd.gov/forms
          https://www.tax.nd.gov/sites/www/files/documents/forms/individual/2022-iit/2022-individual-income-tax-booklet.pdf
          https://www.tax.nd.gov/sites/www/files/documents/forms/2022-individual-income-tax-booklet.pdf
          https://www.tax.nd.gov/tax-exemptions-credits/income-tax-exemptions-credits
          https://www.tax.nd.gov/sites/www/files/documents/forms/business/ndwrfillable.pdf
          https://www.tax.nd.gov/news/tax-legislative-changes/significant-changes-law/individual-income-tax-history
          https://www.tax.nd.gov/forms
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Massachusetts', 'state_occupation': 'NorthDakota', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def massachusetts_and_ohio(self, county_residence='Other', state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Massachusetts:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Ohio:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Massachusetts:
        - https://www.mass.gov/service-details/massachusetts-tax-rates
          https://www.mass.gov/service-details/view-massachusetts-personal-income-tax-exemptions
          https://www.mass.gov/service-details/view-massachusetts-personal-income-tax-exemptions
          https://www.mass.gov/service-details/view-child-and-dependent-related-credits
          https://www.mass.gov/service-details/massachusetts-tax-rates
          https://www.mass.gov/personal-income-tax
        
        Ohio:
        -         
        - https://tax.ohio.gov/individual/resources/annual-tax-rates
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://tax.ohio.gov/static/forms/employer_withholding/generic/wth_it4nr.pdf
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://www.ritaohio.com/TaxRatesTable
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Massachusetts', 'state_occupation': 'Ohio', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def massachusetts_and_oklahoma(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Massachusetts:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Oklahoma:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Massachusetts:
        - https://www.mass.gov/service-details/massachusetts-tax-rates
          https://www.mass.gov/service-details/view-massachusetts-personal-income-tax-exemptions
          https://www.mass.gov/service-details/view-massachusetts-personal-income-tax-exemptions
          https://www.mass.gov/service-details/view-child-and-dependent-related-credits
          https://www.mass.gov/service-details/massachusetts-tax-rates
          https://www.mass.gov/personal-income-tax
        
        Oklahoma:
        -         
        - https://oklahoma.gov/content/dam/ok/en/tax/documents/resources/publications/businesses/withholding-tables/WHTables-2022.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Massachusetts', 'state_occupation': 'Oklahoma', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def massachusetts_and_oregon(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Massachusetts:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Oregon:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Massachusetts:
        - https://www.mass.gov/service-details/massachusetts-tax-rates
          https://www.mass.gov/service-details/view-massachusetts-personal-income-tax-exemptions
          https://www.mass.gov/service-details/view-massachusetts-personal-income-tax-exemptions
          https://www.mass.gov/service-details/view-child-and-dependent-related-credits
          https://www.mass.gov/service-details/massachusetts-tax-rates
          https://www.mass.gov/personal-income-tax
        
        Oregon:
        -         
        - https://www.oregon.gov/dor/programs/individuals/Pages/PIT.aspx
          https://www.oregon.gov/dor/programs/individuals/Documents/Full%20year%20resident
         %20Form%20or-40%20filers.pdf
          https://www.oregon.gov/dor/programs/individuals/Documents/Part-year%20and%20nonresident
         %20Form%20or-40-P%20and%20or-40-N%20filers.pdf.pdf
          https://www.oregon.gov/dor/forms/FormsPubs/form-or-40_101-040_2022.pdf
          https://www.oregon.gov/dor/forms/FormsPubs/form-or-40_101-040_2022.pdf
          https://www.oregon.gov/dor/forms/FormsPubs/form-or-40_101-040_2022.pdf
          https://www.portland.gov/revenue/personal-tax
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Massachusetts', 'state_occupation': 'Oregon', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def massachusetts_and_pennsylvania(self, county_residence='Other', state_agi=75000, county_occupation='Other', capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Massachusetts:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Pennsylvania:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        county_residence : str, optional
            County of residence. Default is Other.
        county_occupation : str, optional
            County of occupation. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Massachusetts:
        - https://www.mass.gov/service-details/massachusetts-tax-rates
          https://www.mass.gov/service-details/view-massachusetts-personal-income-tax-exemptions
          https://www.mass.gov/service-details/view-massachusetts-personal-income-tax-exemptions
          https://www.mass.gov/service-details/view-child-and-dependent-related-credits
          https://www.mass.gov/service-details/massachusetts-tax-rates
          https://www.mass.gov/personal-income-tax
        
        Pennsylvania:
        -         
        - https://www.revenue.pa.gov/Tax%20Rates/Pages/default.aspx
          https://www.revenue.pa.gov/FormsandPublications/FormsforIndividuals/PIT/Documents/2022/2022_pa-40in.pdf
          https://www.revenue.pa.gov/FormsandPublications/FormsforBusinesses/EmployerWithholding/Documents/rev-419.pdf
          https://www.revenue.pa.gov/FormsandPublications/FormsforIndividuals/PIT/Documents/2022/2022_pa-40in.pdf
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-25.htm
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        county_occupation = self.county_occupation if hasattr(self, 'county_occupation') else county_occupation
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Massachusetts', 'state_occupation': 'Pennsylvania', "county_residence": county_residence, "state_agi": state_agi, "county_occupation": county_occupation, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def massachusetts_and_rhode_island(self, state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Massachusetts:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Rhode Island:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Massachusetts:
        - https://www.mass.gov/service-details/massachusetts-tax-rates
          https://www.mass.gov/service-details/view-massachusetts-personal-income-tax-exemptions
          https://www.mass.gov/service-details/view-massachusetts-personal-income-tax-exemptions
          https://www.mass.gov/service-details/view-child-and-dependent-related-credits
          https://www.mass.gov/service-details/massachusetts-tax-rates
          https://www.mass.gov/personal-income-tax
        
        Rhode Island:
        -         
        - https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2022_1040WE_w_0.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Massachusetts', 'state_occupation': 'RhodeIsland', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def massachusetts_and_south_carolina(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Massachusetts:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        South Carolina:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Massachusetts:
        - https://www.mass.gov/service-details/massachusetts-tax-rates
          https://www.mass.gov/service-details/view-massachusetts-personal-income-tax-exemptions
          https://www.mass.gov/service-details/view-massachusetts-personal-income-tax-exemptions
          https://www.mass.gov/service-details/view-child-and-dependent-related-credits
          https://www.mass.gov/service-details/massachusetts-tax-rates
          https://www.mass.gov/personal-income-tax
        
        South Carolina:
        -         
        - https://dor.sc.gov/resources-site/lawandpolicy/Advisory%20Opinions/IL22-15.pdf
          https://dor.sc.gov/forms-site/Forms/IITPacket_2022.pdf
          https://www.scstatehouse.gov/code/t12c006.php
          https://dor.sc.gov/resources-site/lawandpolicy/Advisory%20Opinions/IL22-15.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Massachusetts', 'state_occupation': 'SouthCarolina', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def massachusetts_and_south_dakota(self, state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Massachusetts:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        South Dakota:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Massachusetts:
        - https://www.mass.gov/service-details/massachusetts-tax-rates
          https://www.mass.gov/service-details/view-massachusetts-personal-income-tax-exemptions
          https://www.mass.gov/service-details/view-massachusetts-personal-income-tax-exemptions
          https://www.mass.gov/service-details/view-child-and-dependent-related-credits
          https://www.mass.gov/service-details/massachusetts-tax-rates
          https://www.mass.gov/personal-income-tax
        
        South Dakota:
        -         
        - https://dor.sd.gov/individuals/taxes/
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Massachusetts', 'state_occupation': 'SouthDakota', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def massachusetts_and_tennessee(self, state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Massachusetts:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Tennessee:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Massachusetts:
        - https://www.mass.gov/service-details/massachusetts-tax-rates
          https://www.mass.gov/service-details/view-massachusetts-personal-income-tax-exemptions
          https://www.mass.gov/service-details/view-massachusetts-personal-income-tax-exemptions
          https://www.mass.gov/service-details/view-child-and-dependent-related-credits
          https://www.mass.gov/service-details/massachusetts-tax-rates
          https://www.mass.gov/personal-income-tax
        
        Tennessee:
        -         
        - https://www.tn.gov/revenue/taxes.html
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Massachusetts', 'state_occupation': 'Tennessee', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def massachusetts_and_texas(self, state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Massachusetts:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Texas:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Massachusetts:
        - https://www.mass.gov/service-details/massachusetts-tax-rates
          https://www.mass.gov/service-details/view-massachusetts-personal-income-tax-exemptions
          https://www.mass.gov/service-details/view-massachusetts-personal-income-tax-exemptions
          https://www.mass.gov/service-details/view-child-and-dependent-related-credits
          https://www.mass.gov/service-details/massachusetts-tax-rates
          https://www.mass.gov/personal-income-tax
        
        Texas:
        -         
        - https://comptroller.texas.gov/taxes/
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Massachusetts', 'state_occupation': 'Texas', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def massachusetts_and_utah(self, state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Massachusetts:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Utah:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Massachusetts:
        - https://www.mass.gov/service-details/massachusetts-tax-rates
          https://www.mass.gov/service-details/view-massachusetts-personal-income-tax-exemptions
          https://www.mass.gov/service-details/view-massachusetts-personal-income-tax-exemptions
          https://www.mass.gov/service-details/view-child-and-dependent-related-credits
          https://www.mass.gov/service-details/massachusetts-tax-rates
          https://www.mass.gov/personal-income-tax
        
        Utah:
        -         
        - https://incometax.utah.gov/paying/tax-rates
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Massachusetts', 'state_occupation': 'Utah', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def massachusetts_and_vermont(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Massachusetts:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Vermont:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Massachusetts:
        - https://www.mass.gov/service-details/massachusetts-tax-rates
          https://www.mass.gov/service-details/view-massachusetts-personal-income-tax-exemptions
          https://www.mass.gov/service-details/view-massachusetts-personal-income-tax-exemptions
          https://www.mass.gov/service-details/view-child-and-dependent-related-credits
          https://www.mass.gov/service-details/massachusetts-tax-rates
          https://www.mass.gov/personal-income-tax
        
        Vermont:
        -         
        - https://tax.vermont.gov/sites/tax/files/documents/GB-1210-2023.pdf
          https://tax.vermont.gov/sites/tax/files/documents/Income%20Booklet-2022.pdf
          https://tax.vermont.gov/sites/tax/files/documents/Income%20Booklet-2022.pdf
          https://tax.vermont.gov/sites/tax/files/documents/Income%20Booklet-2022.pdf
          https://tax.vermont.gov/individuals/personal-income-tax/tax-credits
          https://tax.vermont.gov/individuals/personal-income-tax/taxable-income
          https://tax.vermont.gov/individuals/personal-income-tax/local-option-tax
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Massachusetts', 'state_occupation': 'Vermont', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def massachusetts_and_virginia(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Massachusetts:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Virginia:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Massachusetts:
        - https://www.mass.gov/service-details/massachusetts-tax-rates
          https://www.mass.gov/service-details/view-massachusetts-personal-income-tax-exemptions
          https://www.mass.gov/service-details/view-massachusetts-personal-income-tax-exemptions
          https://www.mass.gov/service-details/view-child-and-dependent-related-credits
          https://www.mass.gov/service-details/massachusetts-tax-rates
          https://www.mass.gov/personal-income-tax
        
        Virginia:
        -         
        - https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/reciprocity
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Massachusetts', 'state_occupation': 'Virginia', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def massachusetts_and_washington(self, state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Massachusetts:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Washington:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Massachusetts:
        - https://www.mass.gov/service-details/massachusetts-tax-rates
          https://www.mass.gov/service-details/view-massachusetts-personal-income-tax-exemptions
          https://www.mass.gov/service-details/view-massachusetts-personal-income-tax-exemptions
          https://www.mass.gov/service-details/view-child-and-dependent-related-credits
          https://www.mass.gov/service-details/massachusetts-tax-rates
          https://www.mass.gov/personal-income-tax
        
        Washington:
        -         
        - https://dor.wa.gov/taxes-rates/income-tax
          https://dor.wa.gov/taxes-rates/other-taxes/capital-gains-tax
          https://dor.wa.gov/taxes-rates/income-tax
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Massachusetts', 'state_occupation': 'Washington', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def massachusetts_and_west_virginia(self, county_occupation='Other', state_agi=75000, pay_periods=24, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Massachusetts:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        West Virginia:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_occupation : str, optional
            County of occupation. Default is Other.
        pay_periods : int, optional
            Number of pay periods. Options are 13, 24, 26, 52. Default is 24.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Massachusetts:
        - https://www.mass.gov/service-details/massachusetts-tax-rates
          https://www.mass.gov/service-details/view-massachusetts-personal-income-tax-exemptions
          https://www.mass.gov/service-details/view-massachusetts-personal-income-tax-exemptions
          https://www.mass.gov/service-details/view-child-and-dependent-related-credits
          https://www.mass.gov/service-details/massachusetts-tax-rates
          https://www.mass.gov/personal-income-tax
        
        West Virginia:
        -         
        - https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://tax.wv.gov/Documents/TSD/tsd438.pdf
          https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2011/Taxes-11-35att.html
        """
        
        # Check if the parameters are set upstream in the instance
        county_occupation = self.county_occupation if hasattr(self, 'county_occupation') else county_occupation
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        pay_periods = self.pay_periods if hasattr(self, 'pay_periods') else pay_periods
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Massachusetts', 'state_occupation': 'WestVirginia', "county_occupation": county_occupation, "state_agi": state_agi, "pay_periods": pay_periods, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def massachusetts_and_wisconsin(self, state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Massachusetts:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Wisconsin:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Massachusetts:
        - https://www.mass.gov/service-details/massachusetts-tax-rates
          https://www.mass.gov/service-details/view-massachusetts-personal-income-tax-exemptions
          https://www.mass.gov/service-details/view-massachusetts-personal-income-tax-exemptions
          https://www.mass.gov/service-details/view-child-and-dependent-related-credits
          https://www.mass.gov/service-details/massachusetts-tax-rates
          https://www.mass.gov/personal-income-tax
        
        Wisconsin:
        -         
        - https://www.revenue.wi.gov/TaxForms2023/2023-Form1-ES-Inst.pdf
          https://www.revenue.wi.gov/TaxForms2023/2023-Form1-ES-Inst.pdf
          https://www.revenue.wi.gov/TaxForms2023/2023-Form1-ES-Inst.pdf
          https://www.revenue.wi.gov/Pages/FAQS/pcs-work.aspx
          https://www.revenue.wi.gov/Dor%20Publications/pb103.pdf
          https://www.revenue.wi.gov/Pages/FAQS/pcs-work.aspx
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Massachusetts', 'state_occupation': 'Wisconsin', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def massachusetts_and_wyoming(self, state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Massachusetts:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Wyoming:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Massachusetts:
        - https://www.mass.gov/service-details/massachusetts-tax-rates
          https://www.mass.gov/service-details/view-massachusetts-personal-income-tax-exemptions
          https://www.mass.gov/service-details/view-massachusetts-personal-income-tax-exemptions
          https://www.mass.gov/service-details/view-child-and-dependent-related-credits
          https://www.mass.gov/service-details/massachusetts-tax-rates
          https://www.mass.gov/personal-income-tax
        
        Wyoming:
        -         
        - https://revenue.wyo.gov/
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Massachusetts', 'state_occupation': 'Wyoming', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def michigan_and_alabama(self, capital_gains_short=0, county_occupation='Other', dependents=0, sources=False, county_residence='Other', state_agi=75000, capital_gains_long=0, filing_status='single', income=100000, credits_state=0):
        """
        --- Parameters ----
        
        Michigan:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        county_occupation : str, optional
            County of occupation. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Alabama:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Michigan:
        - https://www.michigan.gov/taxes/iit/new-developments-for-tax-year-2022
          "https://www.michigan.gov/taxes/-/media/Project/Websites/taxes/Forms/2023/2023-SUW/446_Withholding-Guide_2023.pdf?rev=177ca3d8cb034392bf84e317ed78c320&hash=19C5F42F43A89A91863D229349F48883#:~:text=The%20exemption%20amount%20is%20%245%2C400
         employees%20Michigan%20income%20tax%20return"
          "https://www.michigan.gov/taxes/-/media/Project/Websites/taxes/Forms/2023/2023-SUW/446_Withholding-Guide_2023.pdf?rev=177ca3d8cb034392bf84e317ed78c320&hash=19C5F42F43A89A91863D229349F48883#:~:text=The%20exemption%20amount%20is%20%245%2C400
         employees%20Michigan%20income%20tax%20return"
          https://www.michigan.gov/taxes/questions/iit/accordion/residency/are-my-wages-earned-in-another-state-taxable-in-michigan-if-i-am-a-michigan-resident-1
          https://www.michigan.gov/taxes/-/media/Project/Websites/taxes/Forms/2022/2022-IIT-Forms/BOOK_MI-1040.pdf?rev=4524f1cbbc0243b194ab05a6680be75b&hash=40DFE71E651F7C1D65C31F6CD03D944F
          https://www.michigan.gov/taxes/questions/iit/accordion/general/what-cities-impose-an-income-tax
        
        Alabama:
        -         
        - https://www.revenue.alabama.gov/ultraviewer/viewer/basic_viewer/index.html?form=2023/02/22f40.pdf
          https://www.revenue.alabama.gov/faqs/what-is-alabamas-individual-income-tax-rate/
          https://www.revenue.alabama.gov/ultraviewer/viewer/basic_viewer/index.html?form=2023/02/22f40.pdf
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-24.htm?taxmap=true
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-24.htm?taxmap=true
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-24.htm?taxmap=true
          https://www.revenue.alabama.gov/individual-corporate/income-to-be-reported-on-the-alabama-income-tax-return/
          https://www.almonline.org/TaxRates.aspx#OccupationalTax
        """
        
        # Check if the parameters are set upstream in the instance
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        county_occupation = self.county_occupation if hasattr(self, 'county_occupation') else county_occupation
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Michigan', 'state_occupation': 'Alabama', "capital_gains_short": capital_gains_short, "county_occupation": county_occupation, "dependents": dependents, "sources": sources, "county_residence": county_residence, "state_agi": state_agi, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def michigan_and_alaska(self, county_occupation='Other', county_residence='Other', state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Michigan:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        county_occupation : str, optional
            County of occupation. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Alaska:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Michigan:
        - https://www.michigan.gov/taxes/iit/new-developments-for-tax-year-2022
          "https://www.michigan.gov/taxes/-/media/Project/Websites/taxes/Forms/2023/2023-SUW/446_Withholding-Guide_2023.pdf?rev=177ca3d8cb034392bf84e317ed78c320&hash=19C5F42F43A89A91863D229349F48883#:~:text=The%20exemption%20amount%20is%20%245%2C400
         employees%20Michigan%20income%20tax%20return"
          "https://www.michigan.gov/taxes/-/media/Project/Websites/taxes/Forms/2023/2023-SUW/446_Withholding-Guide_2023.pdf?rev=177ca3d8cb034392bf84e317ed78c320&hash=19C5F42F43A89A91863D229349F48883#:~:text=The%20exemption%20amount%20is%20%245%2C400
         employees%20Michigan%20income%20tax%20return"
          https://www.michigan.gov/taxes/questions/iit/accordion/residency/are-my-wages-earned-in-another-state-taxable-in-michigan-if-i-am-a-michigan-resident-1
          https://www.michigan.gov/taxes/-/media/Project/Websites/taxes/Forms/2022/2022-IIT-Forms/BOOK_MI-1040.pdf?rev=4524f1cbbc0243b194ab05a6680be75b&hash=40DFE71E651F7C1D65C31F6CD03D944F
          https://www.michigan.gov/taxes/questions/iit/accordion/general/what-cities-impose-an-income-tax
        
        Alaska:
        -         
        - https://www.commerce.alaska.gov/web/dcra/officeofthestateassessor/alaskataxfacts.aspx
        """
        
        # Check if the parameters are set upstream in the instance
        county_occupation = self.county_occupation if hasattr(self, 'county_occupation') else county_occupation
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Michigan', 'state_occupation': 'Alaska', "county_occupation": county_occupation, "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def michigan_and_arizona(self, county_occupation='Other', county_residence='Other', state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Michigan:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        county_occupation : str, optional
            County of occupation. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Arizona:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        credits_state : int, optional
            Credits for state taxes. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Michigan:
        - https://www.michigan.gov/taxes/iit/new-developments-for-tax-year-2022
          "https://www.michigan.gov/taxes/-/media/Project/Websites/taxes/Forms/2023/2023-SUW/446_Withholding-Guide_2023.pdf?rev=177ca3d8cb034392bf84e317ed78c320&hash=19C5F42F43A89A91863D229349F48883#:~:text=The%20exemption%20amount%20is%20%245%2C400
         employees%20Michigan%20income%20tax%20return"
          "https://www.michigan.gov/taxes/-/media/Project/Websites/taxes/Forms/2023/2023-SUW/446_Withholding-Guide_2023.pdf?rev=177ca3d8cb034392bf84e317ed78c320&hash=19C5F42F43A89A91863D229349F48883#:~:text=The%20exemption%20amount%20is%20%245%2C400
         employees%20Michigan%20income%20tax%20return"
          https://www.michigan.gov/taxes/questions/iit/accordion/residency/are-my-wages-earned-in-another-state-taxable-in-michigan-if-i-am-a-michigan-resident-1
          https://www.michigan.gov/taxes/-/media/Project/Websites/taxes/Forms/2022/2022-IIT-Forms/BOOK_MI-1040.pdf?rev=4524f1cbbc0243b194ab05a6680be75b&hash=40DFE71E651F7C1D65C31F6CD03D944F
          https://www.michigan.gov/taxes/questions/iit/accordion/general/what-cities-impose-an-income-tax
        
        Arizona:
        -         
        - https://azdor.gov/forms/individual/form-140-x-y-tables
          https://azdor.gov/forms/individual/form-140-x-y-tables
          https://azdor.gov/forms/individual/form-140-x-y-tables
          https://azdor.gov/businesses-arizona/withholding-tax/withholding-exceptions
          "https://azdor.gov/individuals/income-tax-filing-assistance/identifying-other-taxable-income#:~:text=In%20regards%20to%20capital%20gains
         the%20individuals%20regular%20tax%20rate.
        """
        
        # Check if the parameters are set upstream in the instance
        county_occupation = self.county_occupation if hasattr(self, 'county_occupation') else county_occupation
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Michigan', 'state_occupation': 'Arizona', "county_occupation": county_occupation, "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def michigan_and_arkansas(self, capital_gains_short=0, county_occupation='Other', dependents=0, sources=False, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_long=0, filing_status='single', income=100000, credits_state=0):
        """
        --- Parameters ----
        
        Michigan:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        county_occupation : str, optional
            County of occupation. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Arkansas:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Michigan:
        - https://www.michigan.gov/taxes/iit/new-developments-for-tax-year-2022
          "https://www.michigan.gov/taxes/-/media/Project/Websites/taxes/Forms/2023/2023-SUW/446_Withholding-Guide_2023.pdf?rev=177ca3d8cb034392bf84e317ed78c320&hash=19C5F42F43A89A91863D229349F48883#:~:text=The%20exemption%20amount%20is%20%245%2C400
         employees%20Michigan%20income%20tax%20return"
          "https://www.michigan.gov/taxes/-/media/Project/Websites/taxes/Forms/2023/2023-SUW/446_Withholding-Guide_2023.pdf?rev=177ca3d8cb034392bf84e317ed78c320&hash=19C5F42F43A89A91863D229349F48883#:~:text=The%20exemption%20amount%20is%20%245%2C400
         employees%20Michigan%20income%20tax%20return"
          https://www.michigan.gov/taxes/questions/iit/accordion/residency/are-my-wages-earned-in-another-state-taxable-in-michigan-if-i-am-a-michigan-resident-1
          https://www.michigan.gov/taxes/-/media/Project/Websites/taxes/Forms/2022/2022-IIT-Forms/BOOK_MI-1040.pdf?rev=4524f1cbbc0243b194ab05a6680be75b&hash=40DFE71E651F7C1D65C31F6CD03D944F
          https://www.michigan.gov/taxes/questions/iit/accordion/general/what-cities-impose-an-income-tax
        
        Arkansas:
        -         
        - https://www.arkansasedc.com/why-arkansas/business-climate/tax-structure/personal-income-tax
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_and_AR1000NR_Instructions.pdf
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_FullYearResidentIndividualIncomeTaxReturn.pdf
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_and_AR1000NR_Instructions.pdf
          https://www.arkansasedc.com/why-arkansas/business-climate/tax-structure/capital-gains-tax-reduction
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_and_AR1000NR_Instructions.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        county_occupation = self.county_occupation if hasattr(self, 'county_occupation') else county_occupation
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Michigan', 'state_occupation': 'Arkansas', "capital_gains_short": capital_gains_short, "county_occupation": county_occupation, "dependents": dependents, "sources": sources, "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def michigan_and_california(self, county_occupation='Other', county_residence='Other', state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Michigan:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        county_occupation : str, optional
            County of occupation. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        California:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Michigan:
        - https://www.michigan.gov/taxes/iit/new-developments-for-tax-year-2022
          "https://www.michigan.gov/taxes/-/media/Project/Websites/taxes/Forms/2023/2023-SUW/446_Withholding-Guide_2023.pdf?rev=177ca3d8cb034392bf84e317ed78c320&hash=19C5F42F43A89A91863D229349F48883#:~:text=The%20exemption%20amount%20is%20%245%2C400
         employees%20Michigan%20income%20tax%20return"
          "https://www.michigan.gov/taxes/-/media/Project/Websites/taxes/Forms/2023/2023-SUW/446_Withholding-Guide_2023.pdf?rev=177ca3d8cb034392bf84e317ed78c320&hash=19C5F42F43A89A91863D229349F48883#:~:text=The%20exemption%20amount%20is%20%245%2C400
         employees%20Michigan%20income%20tax%20return"
          https://www.michigan.gov/taxes/questions/iit/accordion/residency/are-my-wages-earned-in-another-state-taxable-in-michigan-if-i-am-a-michigan-resident-1
          https://www.michigan.gov/taxes/-/media/Project/Websites/taxes/Forms/2022/2022-IIT-Forms/BOOK_MI-1040.pdf?rev=4524f1cbbc0243b194ab05a6680be75b&hash=40DFE71E651F7C1D65C31F6CD03D944F
          https://www.michigan.gov/taxes/questions/iit/accordion/general/what-cities-impose-an-income-tax
        
        California:
        -         
        - https://www.ftb.ca.gov/forms/2022/2022-540-tax-rate-schedules.pdf
          https://www.ftb.ca.gov/file/personal/deductions/index.html
          https://www.ftb.ca.gov/about-ftb/newsroom/caleitc/eligibility-and-credit-information.html
          https://www.ftb.ca.gov/file/personal/income-types/capital-gains-and-losses.html#:~:text=California%20does%20not%20have%20a
         are%20taxed%20as%20ordinary%20income
          https://www.ftb.ca.gov/file/personal/
        """
        
        # Check if the parameters are set upstream in the instance
        county_occupation = self.county_occupation if hasattr(self, 'county_occupation') else county_occupation
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Michigan', 'state_occupation': 'California', "county_occupation": county_occupation, "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def michigan_and_colorado(self, county_occupation='Other', county_residence='Other', state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Michigan:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        county_occupation : str, optional
            County of occupation. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Colorado:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Michigan:
        - https://www.michigan.gov/taxes/iit/new-developments-for-tax-year-2022
          "https://www.michigan.gov/taxes/-/media/Project/Websites/taxes/Forms/2023/2023-SUW/446_Withholding-Guide_2023.pdf?rev=177ca3d8cb034392bf84e317ed78c320&hash=19C5F42F43A89A91863D229349F48883#:~:text=The%20exemption%20amount%20is%20%245%2C400
         employees%20Michigan%20income%20tax%20return"
          "https://www.michigan.gov/taxes/-/media/Project/Websites/taxes/Forms/2023/2023-SUW/446_Withholding-Guide_2023.pdf?rev=177ca3d8cb034392bf84e317ed78c320&hash=19C5F42F43A89A91863D229349F48883#:~:text=The%20exemption%20amount%20is%20%245%2C400
         employees%20Michigan%20income%20tax%20return"
          https://www.michigan.gov/taxes/questions/iit/accordion/residency/are-my-wages-earned-in-another-state-taxable-in-michigan-if-i-am-a-michigan-resident-1
          https://www.michigan.gov/taxes/-/media/Project/Websites/taxes/Forms/2022/2022-IIT-Forms/BOOK_MI-1040.pdf?rev=4524f1cbbc0243b194ab05a6680be75b&hash=40DFE71E651F7C1D65C31F6CD03D944F
          https://www.michigan.gov/taxes/questions/iit/accordion/general/what-cities-impose-an-income-tax
        
        Colorado:
        -         
        - https://leg.colorado.gov/agencies/legislative-council-staff/individual-income-tax%C2%A0
          https://leg.colorado.gov/agencies/legislative-council-staff/individual-income-tax%C2%A0
          https://tax.colorado.gov/sites/tax/files/documents/ITT_Capital_Gain_Subtraction_Feb_2022.pdf
          https://www.auroragov.org/business_services/taxes/occupational_privilege_tax
          https://www.denvergov.org/Government/Agencies-Departments-Offices/Agencies-Departments-Offices-Directory/Department-of-Finance/Our-Divisions/Treasury/Business-Tax-Information#:~:text=Occupational%20privilege%20tax
         -%28show%20below%29&text=Each%20taxable%20employee%20is%20liable
         month%20for%20each%20taxable%20employee
          https://www.glendale.co.us/355/Occupational-Privilege-Tax
          https://www.greenwoodvillage.com/1220/Occupational-Privilege-Tax-OPT
          https://ci.sheridan.co.us/288/Occupational-Privilege-Tax
        """
        
        # Check if the parameters are set upstream in the instance
        county_occupation = self.county_occupation if hasattr(self, 'county_occupation') else county_occupation
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Michigan', 'state_occupation': 'Colorado', "county_occupation": county_occupation, "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def michigan_and_connecticut(self, county_occupation='Other', county_residence='Other', state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Michigan:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        county_occupation : str, optional
            County of occupation. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Connecticut:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Michigan:
        - https://www.michigan.gov/taxes/iit/new-developments-for-tax-year-2022
          "https://www.michigan.gov/taxes/-/media/Project/Websites/taxes/Forms/2023/2023-SUW/446_Withholding-Guide_2023.pdf?rev=177ca3d8cb034392bf84e317ed78c320&hash=19C5F42F43A89A91863D229349F48883#:~:text=The%20exemption%20amount%20is%20%245%2C400
         employees%20Michigan%20income%20tax%20return"
          "https://www.michigan.gov/taxes/-/media/Project/Websites/taxes/Forms/2023/2023-SUW/446_Withholding-Guide_2023.pdf?rev=177ca3d8cb034392bf84e317ed78c320&hash=19C5F42F43A89A91863D229349F48883#:~:text=The%20exemption%20amount%20is%20%245%2C400
         employees%20Michigan%20income%20tax%20return"
          https://www.michigan.gov/taxes/questions/iit/accordion/residency/are-my-wages-earned-in-another-state-taxable-in-michigan-if-i-am-a-michigan-resident-1
          https://www.michigan.gov/taxes/-/media/Project/Websites/taxes/Forms/2022/2022-IIT-Forms/BOOK_MI-1040.pdf?rev=4524f1cbbc0243b194ab05a6680be75b&hash=40DFE71E651F7C1D65C31F6CD03D944F
          https://www.michigan.gov/taxes/questions/iit/accordion/general/what-cities-impose-an-income-tax
        
        Connecticut:
        -         
        - https://portal.ct.gov/-/media/DRS/Forms/2022/Income/2022-CT-1040-Instructions_1222.pdf
          https://www.cga.ct.gov/2022/rpt/pdf/2022-R-0108.pdf
          https://www.cga.ct.gov/current/pub/chap_229.htm#sec_12-703
          https://portal.ct.gov/-/media/DRS/Forms/2022/Income/2022-CT-1040-Instructions_1222.pdf
          "https://portal.ct.gov/DRS/Publications/TSSNs/TSSN-29#:~:text=An%20individuals%20net%20capital%20gains
         portion%20of%20Social%20Security%20benefits"
          "https://portal.ct.gov/DRS/Publications/TSSNs/TSSN-29#:~:text=An%20individuals%20net%20capital%20gains
         portion%20of%20Social%20Security%20benefits"
          https://portal.ct.gov/DRS/Individuals/Resident-Income-Tax/Tax-Information
        """
        
        # Check if the parameters are set upstream in the instance
        county_occupation = self.county_occupation if hasattr(self, 'county_occupation') else county_occupation
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Michigan', 'state_occupation': 'Connecticut', "county_occupation": county_occupation, "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def michigan_and_delaware(self, county_occupation='Other', county_residence='Other', state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, federal_agi=75000, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Michigan:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        county_occupation : str, optional
            County of occupation. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Delaware:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        federal_agi : int, optional
            Adjusted Gross Income (AGI) for federal taxes. Default is 0. Sources - https://www.irs.gov/e-file-providers/definition-of-adjusted-gross-income, https://www.irs.gov/newsroom/taxpayers-should-know-the-difference-between-standard-and-itemized-deductions
        income : int, optional
            Gross income. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Michigan:
        - https://www.michigan.gov/taxes/iit/new-developments-for-tax-year-2022
          "https://www.michigan.gov/taxes/-/media/Project/Websites/taxes/Forms/2023/2023-SUW/446_Withholding-Guide_2023.pdf?rev=177ca3d8cb034392bf84e317ed78c320&hash=19C5F42F43A89A91863D229349F48883#:~:text=The%20exemption%20amount%20is%20%245%2C400
         employees%20Michigan%20income%20tax%20return"
          "https://www.michigan.gov/taxes/-/media/Project/Websites/taxes/Forms/2023/2023-SUW/446_Withholding-Guide_2023.pdf?rev=177ca3d8cb034392bf84e317ed78c320&hash=19C5F42F43A89A91863D229349F48883#:~:text=The%20exemption%20amount%20is%20%245%2C400
         employees%20Michigan%20income%20tax%20return"
          https://www.michigan.gov/taxes/questions/iit/accordion/residency/are-my-wages-earned-in-another-state-taxable-in-michigan-if-i-am-a-michigan-resident-1
          https://www.michigan.gov/taxes/-/media/Project/Websites/taxes/Forms/2022/2022-IIT-Forms/BOOK_MI-1040.pdf?rev=4524f1cbbc0243b194ab05a6680be75b&hash=40DFE71E651F7C1D65C31F6CD03D944F
          https://www.michigan.gov/taxes/questions/iit/accordion/general/what-cities-impose-an-income-tax
        
        Delaware:
        -         
        - https://revenue.delaware.gov/software-developer/tax-rate-changes/
          https://revenuefiles.delaware.gov/2022/PIT-RES_TY22_2022-01_PaperInteractive.pdf
          https://revenuefiles.delaware.gov/2022/PIT-RES_TY22_2022-02_Instructions.pdf
          https://revenue.delaware.gov/frequently-asked-questions/personal-income-tax-faqs/
          https://www.newcastlede.gov/DocumentCenter/View/20147/City-of-Wilmington-Wage-Tax-Information-PDF
        """
        
        # Check if the parameters are set upstream in the instance
        county_occupation = self.county_occupation if hasattr(self, 'county_occupation') else county_occupation
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        federal_agi = self.federal_agi if hasattr(self, 'federal_agi') else federal_agi
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Michigan', 'state_occupation': 'Delaware', "county_occupation": county_occupation, "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "federal_agi": federal_agi, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def michigan_and_florida(self, county_occupation='Other', county_residence='Other', state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Michigan:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        county_occupation : str, optional
            County of occupation. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Florida:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Michigan:
        - https://www.michigan.gov/taxes/iit/new-developments-for-tax-year-2022
          "https://www.michigan.gov/taxes/-/media/Project/Websites/taxes/Forms/2023/2023-SUW/446_Withholding-Guide_2023.pdf?rev=177ca3d8cb034392bf84e317ed78c320&hash=19C5F42F43A89A91863D229349F48883#:~:text=The%20exemption%20amount%20is%20%245%2C400
         employees%20Michigan%20income%20tax%20return"
          "https://www.michigan.gov/taxes/-/media/Project/Websites/taxes/Forms/2023/2023-SUW/446_Withholding-Guide_2023.pdf?rev=177ca3d8cb034392bf84e317ed78c320&hash=19C5F42F43A89A91863D229349F48883#:~:text=The%20exemption%20amount%20is%20%245%2C400
         employees%20Michigan%20income%20tax%20return"
          https://www.michigan.gov/taxes/questions/iit/accordion/residency/are-my-wages-earned-in-another-state-taxable-in-michigan-if-i-am-a-michigan-resident-1
          https://www.michigan.gov/taxes/-/media/Project/Websites/taxes/Forms/2022/2022-IIT-Forms/BOOK_MI-1040.pdf?rev=4524f1cbbc0243b194ab05a6680be75b&hash=40DFE71E651F7C1D65C31F6CD03D944F
          https://www.michigan.gov/taxes/questions/iit/accordion/general/what-cities-impose-an-income-tax
        
        Florida:
        -         
        - https://floridarevenue.com/taxes/taxesfees/Pages/tax_interest_rates.aspx
        """
        
        # Check if the parameters are set upstream in the instance
        county_occupation = self.county_occupation if hasattr(self, 'county_occupation') else county_occupation
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Michigan', 'state_occupation': 'Florida', "county_occupation": county_occupation, "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def michigan_and_georgia(self, capital_gains_short=0, county_occupation='Other', dependents=0, sources=False, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_long=0, filing_status='single', income=100000, credits_state=0):
        """
        --- Parameters ----
        
        Michigan:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        county_occupation : str, optional
            County of occupation. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Georgia:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Michigan:
        - https://www.michigan.gov/taxes/iit/new-developments-for-tax-year-2022
          "https://www.michigan.gov/taxes/-/media/Project/Websites/taxes/Forms/2023/2023-SUW/446_Withholding-Guide_2023.pdf?rev=177ca3d8cb034392bf84e317ed78c320&hash=19C5F42F43A89A91863D229349F48883#:~:text=The%20exemption%20amount%20is%20%245%2C400
         employees%20Michigan%20income%20tax%20return"
          "https://www.michigan.gov/taxes/-/media/Project/Websites/taxes/Forms/2023/2023-SUW/446_Withholding-Guide_2023.pdf?rev=177ca3d8cb034392bf84e317ed78c320&hash=19C5F42F43A89A91863D229349F48883#:~:text=The%20exemption%20amount%20is%20%245%2C400
         employees%20Michigan%20income%20tax%20return"
          https://www.michigan.gov/taxes/questions/iit/accordion/residency/are-my-wages-earned-in-another-state-taxable-in-michigan-if-i-am-a-michigan-resident-1
          https://www.michigan.gov/taxes/-/media/Project/Websites/taxes/Forms/2022/2022-IIT-Forms/BOOK_MI-1040.pdf?rev=4524f1cbbc0243b194ab05a6680be75b&hash=40DFE71E651F7C1D65C31F6CD03D944F
          https://www.michigan.gov/taxes/questions/iit/accordion/general/what-cities-impose-an-income-tax
        
        Georgia:
        -         
        - https://dor.georgia.gov/tax-tables-georgia-tax-rate-schedule
          https://apps.dor.ga.gov/FillableForms/PDFViewer/Index?form=2022GA500
          https://apps.dor.ga.gov/FillableForms/PDFViewer/Index?form=2022GA500
          https://apps.dor.ga.gov/FillableForms/PDFViewer/Index?form=2022GA500
          https://dor.georgia.gov/years-individual-income-tax-forms
          https://dor.georgia.gov/years-individual-income-tax-forms
        """
        
        # Check if the parameters are set upstream in the instance
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        county_occupation = self.county_occupation if hasattr(self, 'county_occupation') else county_occupation
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Michigan', 'state_occupation': 'Georgia', "capital_gains_short": capital_gains_short, "county_occupation": county_occupation, "dependents": dependents, "sources": sources, "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def michigan_and_hawaii(self, capital_gains_short=0, county_occupation='Other', dependents=0, sources=False, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_long=0, filing_status='single', income=100000, credits_state=0):
        """
        --- Parameters ----
        
        Michigan:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        county_occupation : str, optional
            County of occupation. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Hawaii:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Michigan:
        - https://www.michigan.gov/taxes/iit/new-developments-for-tax-year-2022
          "https://www.michigan.gov/taxes/-/media/Project/Websites/taxes/Forms/2023/2023-SUW/446_Withholding-Guide_2023.pdf?rev=177ca3d8cb034392bf84e317ed78c320&hash=19C5F42F43A89A91863D229349F48883#:~:text=The%20exemption%20amount%20is%20%245%2C400
         employees%20Michigan%20income%20tax%20return"
          "https://www.michigan.gov/taxes/-/media/Project/Websites/taxes/Forms/2023/2023-SUW/446_Withholding-Guide_2023.pdf?rev=177ca3d8cb034392bf84e317ed78c320&hash=19C5F42F43A89A91863D229349F48883#:~:text=The%20exemption%20amount%20is%20%245%2C400
         employees%20Michigan%20income%20tax%20return"
          https://www.michigan.gov/taxes/questions/iit/accordion/residency/are-my-wages-earned-in-another-state-taxable-in-michigan-if-i-am-a-michigan-resident-1
          https://www.michigan.gov/taxes/-/media/Project/Websites/taxes/Forms/2022/2022-IIT-Forms/BOOK_MI-1040.pdf?rev=4524f1cbbc0243b194ab05a6680be75b&hash=40DFE71E651F7C1D65C31F6CD03D944F
          https://www.michigan.gov/taxes/questions/iit/accordion/general/what-cities-impose-an-income-tax
        
        Hawaii:
        -         
        - https://tax.hawaii.gov/forms/d_18table-on/d_18table-on_p13/
          https://files.hawaii.gov/tax/forms/2018/18table-on.pdf
          https://files.hawaii.gov/tax/news/pubs/22outline.pdf
          https://files.hawaii.gov/tax/news/pubs/FastTaxRefGuide_2017.pdf
          https://www.tfhawaii.org/wordpress/state-tax-resources/mini-tax-guides/individual-income-tax-chapter-235/#:~:text=Long%20term%20capital%20gains%20are%20taxed%20at%20a%20maximum%20of%207.25%25
          https://www.grassrootinstitute.org/2023/02/hb337-increasing-tax-on-capital-gains-a-bad-risk/
          https://files.hawaii.gov/tax/news/pubs/22outline.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        county_occupation = self.county_occupation if hasattr(self, 'county_occupation') else county_occupation
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Michigan', 'state_occupation': 'Hawaii', "capital_gains_short": capital_gains_short, "county_occupation": county_occupation, "dependents": dependents, "sources": sources, "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def michigan_and_idaho(self, capital_gains_short=0, county_occupation='Other', dependents=0, sources=False, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_long=0, filing_status='single', income=100000, credits_state=0):
        """
        --- Parameters ----
        
        Michigan:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        county_occupation : str, optional
            County of occupation. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Idaho:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Michigan:
        - https://www.michigan.gov/taxes/iit/new-developments-for-tax-year-2022
          "https://www.michigan.gov/taxes/-/media/Project/Websites/taxes/Forms/2023/2023-SUW/446_Withholding-Guide_2023.pdf?rev=177ca3d8cb034392bf84e317ed78c320&hash=19C5F42F43A89A91863D229349F48883#:~:text=The%20exemption%20amount%20is%20%245%2C400
         employees%20Michigan%20income%20tax%20return"
          "https://www.michigan.gov/taxes/-/media/Project/Websites/taxes/Forms/2023/2023-SUW/446_Withholding-Guide_2023.pdf?rev=177ca3d8cb034392bf84e317ed78c320&hash=19C5F42F43A89A91863D229349F48883#:~:text=The%20exemption%20amount%20is%20%245%2C400
         employees%20Michigan%20income%20tax%20return"
          https://www.michigan.gov/taxes/questions/iit/accordion/residency/are-my-wages-earned-in-another-state-taxable-in-michigan-if-i-am-a-michigan-resident-1
          https://www.michigan.gov/taxes/-/media/Project/Websites/taxes/Forms/2022/2022-IIT-Forms/BOOK_MI-1040.pdf?rev=4524f1cbbc0243b194ab05a6680be75b&hash=40DFE71E651F7C1D65C31F6CD03D944F
          https://www.michigan.gov/taxes/questions/iit/accordion/general/what-cities-impose-an-income-tax
        
        Idaho:
        -         
        - https://tax.idaho.gov/taxes/income-tax/individual-income/individual-income-tax-rate-schedule/
          https://tax.idaho.gov/wp-content/uploads/forms/EIN00046/EIN00046_03-01-2023.pdf
          https://tax.idaho.gov/taxes/income-tax/individual-income/idaho-source-income/
          https://tax.idaho.gov/taxes/income-tax/individual-income/idaho-source-income/
        """
        
        # Check if the parameters are set upstream in the instance
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        county_occupation = self.county_occupation if hasattr(self, 'county_occupation') else county_occupation
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Michigan', 'state_occupation': 'Idaho', "capital_gains_short": capital_gains_short, "county_occupation": county_occupation, "dependents": dependents, "sources": sources, "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def michigan_and_illinois(self, county_occupation='Other', county_residence='Other', state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Michigan:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        county_occupation : str, optional
            County of occupation. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Illinois:
        


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Michigan:
        - https://www.michigan.gov/taxes/iit/new-developments-for-tax-year-2022
          "https://www.michigan.gov/taxes/-/media/Project/Websites/taxes/Forms/2023/2023-SUW/446_Withholding-Guide_2023.pdf?rev=177ca3d8cb034392bf84e317ed78c320&hash=19C5F42F43A89A91863D229349F48883#:~:text=The%20exemption%20amount%20is%20%245%2C400
         employees%20Michigan%20income%20tax%20return"
          "https://www.michigan.gov/taxes/-/media/Project/Websites/taxes/Forms/2023/2023-SUW/446_Withholding-Guide_2023.pdf?rev=177ca3d8cb034392bf84e317ed78c320&hash=19C5F42F43A89A91863D229349F48883#:~:text=The%20exemption%20amount%20is%20%245%2C400
         employees%20Michigan%20income%20tax%20return"
          https://www.michigan.gov/taxes/questions/iit/accordion/residency/are-my-wages-earned-in-another-state-taxable-in-michigan-if-i-am-a-michigan-resident-1
          https://www.michigan.gov/taxes/-/media/Project/Websites/taxes/Forms/2022/2022-IIT-Forms/BOOK_MI-1040.pdf?rev=4524f1cbbc0243b194ab05a6680be75b&hash=40DFE71E651F7C1D65C31F6CD03D944F
          https://www.michigan.gov/taxes/questions/iit/accordion/general/what-cities-impose-an-income-tax
        
        Illinois:
        -         
        - https://tax.illinois.gov/research/taxrates/income.html#IndividualIncome
          https://tax.illinois.gov/content/dam/soi/en/web/tax/forms/incometax/documents/currentyear/individual/il-1040-instr.pdf
          https://tax.illinois.gov/questionsandanswers/12.html
          https://tax.illinois.gov/content/dam/soi/en/web/tax/forms/incometax/documents/currentyear/individual/il-1040-instr.pdf
          https://tax.illinois.gov/research/taxrates/personalproperty.html
          https://tax.illinois.gov/localgovernments/income.html
        """
        
        # Check if the parameters are set upstream in the instance
        county_occupation = self.county_occupation if hasattr(self, 'county_occupation') else county_occupation
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Michigan', 'state_occupation': 'Illinois', "county_occupation": county_occupation, "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def michigan_and_indiana(self, county_occupation='Other', county_residence='Other', state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Michigan:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        county_occupation : str, optional
            County of occupation. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Indiana:
        


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Michigan:
        - https://www.michigan.gov/taxes/iit/new-developments-for-tax-year-2022
          "https://www.michigan.gov/taxes/-/media/Project/Websites/taxes/Forms/2023/2023-SUW/446_Withholding-Guide_2023.pdf?rev=177ca3d8cb034392bf84e317ed78c320&hash=19C5F42F43A89A91863D229349F48883#:~:text=The%20exemption%20amount%20is%20%245%2C400
         employees%20Michigan%20income%20tax%20return"
          "https://www.michigan.gov/taxes/-/media/Project/Websites/taxes/Forms/2023/2023-SUW/446_Withholding-Guide_2023.pdf?rev=177ca3d8cb034392bf84e317ed78c320&hash=19C5F42F43A89A91863D229349F48883#:~:text=The%20exemption%20amount%20is%20%245%2C400
         employees%20Michigan%20income%20tax%20return"
          https://www.michigan.gov/taxes/questions/iit/accordion/residency/are-my-wages-earned-in-another-state-taxable-in-michigan-if-i-am-a-michigan-resident-1
          https://www.michigan.gov/taxes/-/media/Project/Websites/taxes/Forms/2022/2022-IIT-Forms/BOOK_MI-1040.pdf?rev=4524f1cbbc0243b194ab05a6680be75b&hash=40DFE71E651F7C1D65C31F6CD03D944F
          https://www.michigan.gov/taxes/questions/iit/accordion/general/what-cities-impose-an-income-tax
        
        Indiana:
        -         
        - https://www.in.gov/dor/tax-forms/2022-individual-income-tax-forms/
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-32.htm
          https://www.in.gov/dor/files/ib28.pdf
          https://www.in.gov/dor/tax-forms/2022-individual-income-tax-forms/
          https://www.in.gov/dor/files/2022-county-tax-rates-and-codes.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        county_occupation = self.county_occupation if hasattr(self, 'county_occupation') else county_occupation
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Michigan', 'state_occupation': 'Indiana', "county_occupation": county_occupation, "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def michigan_and_iowa(self, capital_gains_short=0, county_occupation='Other', dependents=0, sources=False, county_residence='Other', state_agi=75000, capital_gains_long=0, filing_status='single', income=100000, credits_state=0):
        """
        --- Parameters ----
        
        Michigan:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        county_occupation : str, optional
            County of occupation. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Iowa:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Michigan:
        - https://www.michigan.gov/taxes/iit/new-developments-for-tax-year-2022
          "https://www.michigan.gov/taxes/-/media/Project/Websites/taxes/Forms/2023/2023-SUW/446_Withholding-Guide_2023.pdf?rev=177ca3d8cb034392bf84e317ed78c320&hash=19C5F42F43A89A91863D229349F48883#:~:text=The%20exemption%20amount%20is%20%245%2C400
         employees%20Michigan%20income%20tax%20return"
          "https://www.michigan.gov/taxes/-/media/Project/Websites/taxes/Forms/2023/2023-SUW/446_Withholding-Guide_2023.pdf?rev=177ca3d8cb034392bf84e317ed78c320&hash=19C5F42F43A89A91863D229349F48883#:~:text=The%20exemption%20amount%20is%20%245%2C400
         employees%20Michigan%20income%20tax%20return"
          https://www.michigan.gov/taxes/questions/iit/accordion/residency/are-my-wages-earned-in-another-state-taxable-in-michigan-if-i-am-a-michigan-resident-1
          https://www.michigan.gov/taxes/-/media/Project/Websites/taxes/Forms/2022/2022-IIT-Forms/BOOK_MI-1040.pdf?rev=4524f1cbbc0243b194ab05a6680be75b&hash=40DFE71E651F7C1D65C31F6CD03D944F
          https://www.michigan.gov/taxes/questions/iit/accordion/general/what-cities-impose-an-income-tax
        
        Iowa:
        -         
        - https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
          https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
          https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
          https://tax.iowa.gov/iowa-illinois-reciprocal-agreement
          https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
        """
        
        # Check if the parameters are set upstream in the instance
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        county_occupation = self.county_occupation if hasattr(self, 'county_occupation') else county_occupation
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Michigan', 'state_occupation': 'Iowa', "capital_gains_short": capital_gains_short, "county_occupation": county_occupation, "dependents": dependents, "sources": sources, "county_residence": county_residence, "state_agi": state_agi, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def michigan_and_kansas(self, capital_gains_short=0, county_occupation='Other', dependents=0, sources=False, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_long=0, filing_status='single', income=100000, credits_state=0):
        """
        --- Parameters ----
        
        Michigan:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        county_occupation : str, optional
            County of occupation. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Kansas:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Michigan:
        - https://www.michigan.gov/taxes/iit/new-developments-for-tax-year-2022
          "https://www.michigan.gov/taxes/-/media/Project/Websites/taxes/Forms/2023/2023-SUW/446_Withholding-Guide_2023.pdf?rev=177ca3d8cb034392bf84e317ed78c320&hash=19C5F42F43A89A91863D229349F48883#:~:text=The%20exemption%20amount%20is%20%245%2C400
         employees%20Michigan%20income%20tax%20return"
          "https://www.michigan.gov/taxes/-/media/Project/Websites/taxes/Forms/2023/2023-SUW/446_Withholding-Guide_2023.pdf?rev=177ca3d8cb034392bf84e317ed78c320&hash=19C5F42F43A89A91863D229349F48883#:~:text=The%20exemption%20amount%20is%20%245%2C400
         employees%20Michigan%20income%20tax%20return"
          https://www.michigan.gov/taxes/questions/iit/accordion/residency/are-my-wages-earned-in-another-state-taxable-in-michigan-if-i-am-a-michigan-resident-1
          https://www.michigan.gov/taxes/-/media/Project/Websites/taxes/Forms/2022/2022-IIT-Forms/BOOK_MI-1040.pdf?rev=4524f1cbbc0243b194ab05a6680be75b&hash=40DFE71E651F7C1D65C31F6CD03D944F
          https://www.michigan.gov/taxes/questions/iit/accordion/general/what-cities-impose-an-income-tax
        
        Kansas:
        -         
        - https://www.ksrevenue.gov/taxrates.html
          https://www.ksrevenue.gov/incomebook22.html#0
          https://www.ksrevenue.gov/incomebook22.html#0
          https://www.ksrevenue.gov/incomebook22.html#0
          https://www.ksrevenue.gov/incomebook22.html
          https://www.ksrevenue.gov/incomebook22.html
        """
        
        # Check if the parameters are set upstream in the instance
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        county_occupation = self.county_occupation if hasattr(self, 'county_occupation') else county_occupation
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Michigan', 'state_occupation': 'Kansas', "capital_gains_short": capital_gains_short, "county_occupation": county_occupation, "dependents": dependents, "sources": sources, "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def michigan_and_kentucky(self, county_occupation='Other', county_residence='Other', state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Michigan:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        county_occupation : str, optional
            County of occupation. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Kentucky:
        


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Michigan:
        - https://www.michigan.gov/taxes/iit/new-developments-for-tax-year-2022
          "https://www.michigan.gov/taxes/-/media/Project/Websites/taxes/Forms/2023/2023-SUW/446_Withholding-Guide_2023.pdf?rev=177ca3d8cb034392bf84e317ed78c320&hash=19C5F42F43A89A91863D229349F48883#:~:text=The%20exemption%20amount%20is%20%245%2C400
         employees%20Michigan%20income%20tax%20return"
          "https://www.michigan.gov/taxes/-/media/Project/Websites/taxes/Forms/2023/2023-SUW/446_Withholding-Guide_2023.pdf?rev=177ca3d8cb034392bf84e317ed78c320&hash=19C5F42F43A89A91863D229349F48883#:~:text=The%20exemption%20amount%20is%20%245%2C400
         employees%20Michigan%20income%20tax%20return"
          https://www.michigan.gov/taxes/questions/iit/accordion/residency/are-my-wages-earned-in-another-state-taxable-in-michigan-if-i-am-a-michigan-resident-1
          https://www.michigan.gov/taxes/-/media/Project/Websites/taxes/Forms/2022/2022-IIT-Forms/BOOK_MI-1040.pdf?rev=4524f1cbbc0243b194ab05a6680be75b&hash=40DFE71E651F7C1D65C31F6CD03D944F
          https://www.michigan.gov/taxes/questions/iit/accordion/general/what-cities-impose-an-income-tax
        
        Kentucky:
        -         
        - https://revenue.ky.gov/News/Pages/Dor-Announces-Updates-to-Individual-Income-Tax-for-2023-Tax-Year.aspx
          https://revenue.ky.gov/Individual/Individual-Income-Tax/Pages/default.aspx
          https://revenue.ky.gov/News/Pages/Dor-Announces-Updates-to-Individual-Income-Tax-for-2023-Tax-Year.aspx
          https://revenue.ky.gov/Individual/Individual-Income-Tax/Pages/default.aspx
          https://revenue.ky.gov/Dor%20Training%20Materials/103%20KAR%2017.140.%20Individual%20income%20tax%20-%20reciprocity%20-%20nonresidents.pdf
          https://revenue.ky.gov/Individual/Individual-Income-Tax/Pages/default.aspx
          https://nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2021/Taxes-21-32.htm?taxmap=true
        """
        
        # Check if the parameters are set upstream in the instance
        county_occupation = self.county_occupation if hasattr(self, 'county_occupation') else county_occupation
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Michigan', 'state_occupation': 'Kentucky', "county_occupation": county_occupation, "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def michigan_and_louisiana(self, capital_gains_short=0, county_occupation='Other', dependents=0, sources=False, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_long=0, filing_status='single', income=100000, credits_state=0):
        """
        --- Parameters ----
        
        Michigan:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        county_occupation : str, optional
            County of occupation. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Louisiana:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Michigan:
        - https://www.michigan.gov/taxes/iit/new-developments-for-tax-year-2022
          "https://www.michigan.gov/taxes/-/media/Project/Websites/taxes/Forms/2023/2023-SUW/446_Withholding-Guide_2023.pdf?rev=177ca3d8cb034392bf84e317ed78c320&hash=19C5F42F43A89A91863D229349F48883#:~:text=The%20exemption%20amount%20is%20%245%2C400
         employees%20Michigan%20income%20tax%20return"
          "https://www.michigan.gov/taxes/-/media/Project/Websites/taxes/Forms/2023/2023-SUW/446_Withholding-Guide_2023.pdf?rev=177ca3d8cb034392bf84e317ed78c320&hash=19C5F42F43A89A91863D229349F48883#:~:text=The%20exemption%20amount%20is%20%245%2C400
         employees%20Michigan%20income%20tax%20return"
          https://www.michigan.gov/taxes/questions/iit/accordion/residency/are-my-wages-earned-in-another-state-taxable-in-michigan-if-i-am-a-michigan-resident-1
          https://www.michigan.gov/taxes/-/media/Project/Websites/taxes/Forms/2022/2022-IIT-Forms/BOOK_MI-1040.pdf?rev=4524f1cbbc0243b194ab05a6680be75b&hash=40DFE71E651F7C1D65C31F6CD03D944F
          https://www.michigan.gov/taxes/questions/iit/accordion/general/what-cities-impose-an-income-tax
        
        Louisiana:
        -         
        - https://revenue.louisiana.gov/individualincometax
          https://revenue.louisiana.gov/individualincometax
          https://www.revenue.louisiana.gov/taxforms/6935(11_02)F.pdf
          https://revenue.louisiana.gov/individualincometax
          https://revenue.louisiana.gov/individualincometax
        """
        
        # Check if the parameters are set upstream in the instance
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        county_occupation = self.county_occupation if hasattr(self, 'county_occupation') else county_occupation
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Michigan', 'state_occupation': 'Louisiana', "capital_gains_short": capital_gains_short, "county_occupation": county_occupation, "dependents": dependents, "sources": sources, "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def michigan_and_maine(self, capital_gains_short=0, county_occupation='Other', dependents=0, sources=False, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_long=0, filing_status='single', income=100000, credits_state=0):
        """
        --- Parameters ----
        
        Michigan:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        county_occupation : str, optional
            County of occupation. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Maine:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Michigan:
        - https://www.michigan.gov/taxes/iit/new-developments-for-tax-year-2022
          "https://www.michigan.gov/taxes/-/media/Project/Websites/taxes/Forms/2023/2023-SUW/446_Withholding-Guide_2023.pdf?rev=177ca3d8cb034392bf84e317ed78c320&hash=19C5F42F43A89A91863D229349F48883#:~:text=The%20exemption%20amount%20is%20%245%2C400
         employees%20Michigan%20income%20tax%20return"
          "https://www.michigan.gov/taxes/-/media/Project/Websites/taxes/Forms/2023/2023-SUW/446_Withholding-Guide_2023.pdf?rev=177ca3d8cb034392bf84e317ed78c320&hash=19C5F42F43A89A91863D229349F48883#:~:text=The%20exemption%20amount%20is%20%245%2C400
         employees%20Michigan%20income%20tax%20return"
          https://www.michigan.gov/taxes/questions/iit/accordion/residency/are-my-wages-earned-in-another-state-taxable-in-michigan-if-i-am-a-michigan-resident-1
          https://www.michigan.gov/taxes/-/media/Project/Websites/taxes/Forms/2022/2022-IIT-Forms/BOOK_MI-1040.pdf?rev=4524f1cbbc0243b194ab05a6680be75b&hash=40DFE71E651F7C1D65C31F6CD03D944F
          https://www.michigan.gov/taxes/questions/iit/accordion/general/what-cities-impose-an-income-tax
        
        Maine:
        -         
        - https://www.maine.gov/revenue/tax-return-forms/individual-income-tax-2022
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/ind_tax_rate_sched_2022.pdf
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/22_1040me_book_gen_instruc_revisedFeb23.pdf
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/22_1040me_book_gen_instruc_revisedFeb23.pdf
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/22_1040me_book_gen_instruc_revisedFeb23.pdf
          https://www.maine.gov/revenue/tax-return-forms/individual-income-tax-2022
        """
        
        # Check if the parameters are set upstream in the instance
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        county_occupation = self.county_occupation if hasattr(self, 'county_occupation') else county_occupation
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Michigan', 'state_occupation': 'Maine', "capital_gains_short": capital_gains_short, "county_occupation": county_occupation, "dependents": dependents, "sources": sources, "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def michigan_and_maryland(self, capital_gains_short=0, county_occupation='Other', dependents=0, sources=False, county_residence='Other', state_agi=75000, capital_gains_long=0, filing_status='single', income=100000, credits_state=0):
        """
        --- Parameters ----
        
        Michigan:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        county_occupation : str, optional
            County of occupation. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Maryland:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Michigan:
        - https://www.michigan.gov/taxes/iit/new-developments-for-tax-year-2022
          "https://www.michigan.gov/taxes/-/media/Project/Websites/taxes/Forms/2023/2023-SUW/446_Withholding-Guide_2023.pdf?rev=177ca3d8cb034392bf84e317ed78c320&hash=19C5F42F43A89A91863D229349F48883#:~:text=The%20exemption%20amount%20is%20%245%2C400
         employees%20Michigan%20income%20tax%20return"
          "https://www.michigan.gov/taxes/-/media/Project/Websites/taxes/Forms/2023/2023-SUW/446_Withholding-Guide_2023.pdf?rev=177ca3d8cb034392bf84e317ed78c320&hash=19C5F42F43A89A91863D229349F48883#:~:text=The%20exemption%20amount%20is%20%245%2C400
         employees%20Michigan%20income%20tax%20return"
          https://www.michigan.gov/taxes/questions/iit/accordion/residency/are-my-wages-earned-in-another-state-taxable-in-michigan-if-i-am-a-michigan-resident-1
          https://www.michigan.gov/taxes/-/media/Project/Websites/taxes/Forms/2022/2022-IIT-Forms/BOOK_MI-1040.pdf?rev=4524f1cbbc0243b194ab05a6680be75b&hash=40DFE71E651F7C1D65C31F6CD03D944F
          https://www.michigan.gov/taxes/questions/iit/accordion/general/what-cities-impose-an-income-tax
        
        Maryland:
        -         
        - https://www.marylandtaxes.gov/individual/income/tax-info/tax-rates.php
          https://www.marylandtaxes.gov/forms/22_forms/Resident_Booklet.pdf
          https://www.marylandtaxes.gov/forms/Tax_Publications/Administrative_Releases/Income_and_Estate_Tax_Releases/ar_it37.pdf
          https://www.marylandtaxes.gov/individual/index.php
          https://www.marylandtaxes.gov/individual/income/tax-info/tax-rates.php
        """
        
        # Check if the parameters are set upstream in the instance
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        county_occupation = self.county_occupation if hasattr(self, 'county_occupation') else county_occupation
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Michigan', 'state_occupation': 'Maryland', "capital_gains_short": capital_gains_short, "county_occupation": county_occupation, "dependents": dependents, "sources": sources, "county_residence": county_residence, "state_agi": state_agi, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def michigan_and_massachusetts(self, county_occupation='Other', county_residence='Other', state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Michigan:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        county_occupation : str, optional
            County of occupation. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Massachusetts:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Michigan:
        - https://www.michigan.gov/taxes/iit/new-developments-for-tax-year-2022
          "https://www.michigan.gov/taxes/-/media/Project/Websites/taxes/Forms/2023/2023-SUW/446_Withholding-Guide_2023.pdf?rev=177ca3d8cb034392bf84e317ed78c320&hash=19C5F42F43A89A91863D229349F48883#:~:text=The%20exemption%20amount%20is%20%245%2C400
         employees%20Michigan%20income%20tax%20return"
          "https://www.michigan.gov/taxes/-/media/Project/Websites/taxes/Forms/2023/2023-SUW/446_Withholding-Guide_2023.pdf?rev=177ca3d8cb034392bf84e317ed78c320&hash=19C5F42F43A89A91863D229349F48883#:~:text=The%20exemption%20amount%20is%20%245%2C400
         employees%20Michigan%20income%20tax%20return"
          https://www.michigan.gov/taxes/questions/iit/accordion/residency/are-my-wages-earned-in-another-state-taxable-in-michigan-if-i-am-a-michigan-resident-1
          https://www.michigan.gov/taxes/-/media/Project/Websites/taxes/Forms/2022/2022-IIT-Forms/BOOK_MI-1040.pdf?rev=4524f1cbbc0243b194ab05a6680be75b&hash=40DFE71E651F7C1D65C31F6CD03D944F
          https://www.michigan.gov/taxes/questions/iit/accordion/general/what-cities-impose-an-income-tax
        
        Massachusetts:
        -         
        - https://www.mass.gov/service-details/massachusetts-tax-rates
          https://www.mass.gov/service-details/view-massachusetts-personal-income-tax-exemptions
          https://www.mass.gov/service-details/view-massachusetts-personal-income-tax-exemptions
          https://www.mass.gov/service-details/view-child-and-dependent-related-credits
          https://www.mass.gov/service-details/massachusetts-tax-rates
          https://www.mass.gov/personal-income-tax
        """
        
        # Check if the parameters are set upstream in the instance
        county_occupation = self.county_occupation if hasattr(self, 'county_occupation') else county_occupation
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Michigan', 'state_occupation': 'Massachusetts', "county_occupation": county_occupation, "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def michigan_and_minnesota(self, county_occupation='Other', county_residence='Other', state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Michigan:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        county_occupation : str, optional
            County of occupation. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Minnesota:
        


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Michigan:
        - https://www.michigan.gov/taxes/iit/new-developments-for-tax-year-2022
          "https://www.michigan.gov/taxes/-/media/Project/Websites/taxes/Forms/2023/2023-SUW/446_Withholding-Guide_2023.pdf?rev=177ca3d8cb034392bf84e317ed78c320&hash=19C5F42F43A89A91863D229349F48883#:~:text=The%20exemption%20amount%20is%20%245%2C400
         employees%20Michigan%20income%20tax%20return"
          "https://www.michigan.gov/taxes/-/media/Project/Websites/taxes/Forms/2023/2023-SUW/446_Withholding-Guide_2023.pdf?rev=177ca3d8cb034392bf84e317ed78c320&hash=19C5F42F43A89A91863D229349F48883#:~:text=The%20exemption%20amount%20is%20%245%2C400
         employees%20Michigan%20income%20tax%20return"
          https://www.michigan.gov/taxes/questions/iit/accordion/residency/are-my-wages-earned-in-another-state-taxable-in-michigan-if-i-am-a-michigan-resident-1
          https://www.michigan.gov/taxes/-/media/Project/Websites/taxes/Forms/2022/2022-IIT-Forms/BOOK_MI-1040.pdf?rev=4524f1cbbc0243b194ab05a6680be75b&hash=40DFE71E651F7C1D65C31F6CD03D944F
          https://www.michigan.gov/taxes/questions/iit/accordion/general/what-cities-impose-an-income-tax
        
        Minnesota:
        -         
        - https://www.revenue.state.mn.us/minnesota-income-tax-rates-and-brackets
          https://www.revenue.state.mn.us/child-and-dependent-care-credit
          https://www.revenue.state.mn.us/dependent-exemptions
          https://www.revenue.state.mn.us/minnesota-standard-deduction
          https://mn.gov/mmb/assets/mwr_form_tcm1059-128581.pdf
          https://www.revenue.state.mn.us/individual-income-tax
          https://www.revenue.state.mn.us/individual-income-tax
        """
        
        # Check if the parameters are set upstream in the instance
        county_occupation = self.county_occupation if hasattr(self, 'county_occupation') else county_occupation
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Michigan', 'state_occupation': 'Minnesota', "county_occupation": county_occupation, "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def michigan_and_mississippi(self, county_occupation='Other', county_residence='Other', state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Michigan:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        county_occupation : str, optional
            County of occupation. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Mississippi:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Michigan:
        - https://www.michigan.gov/taxes/iit/new-developments-for-tax-year-2022
          "https://www.michigan.gov/taxes/-/media/Project/Websites/taxes/Forms/2023/2023-SUW/446_Withholding-Guide_2023.pdf?rev=177ca3d8cb034392bf84e317ed78c320&hash=19C5F42F43A89A91863D229349F48883#:~:text=The%20exemption%20amount%20is%20%245%2C400
         employees%20Michigan%20income%20tax%20return"
          "https://www.michigan.gov/taxes/-/media/Project/Websites/taxes/Forms/2023/2023-SUW/446_Withholding-Guide_2023.pdf?rev=177ca3d8cb034392bf84e317ed78c320&hash=19C5F42F43A89A91863D229349F48883#:~:text=The%20exemption%20amount%20is%20%245%2C400
         employees%20Michigan%20income%20tax%20return"
          https://www.michigan.gov/taxes/questions/iit/accordion/residency/are-my-wages-earned-in-another-state-taxable-in-michigan-if-i-am-a-michigan-resident-1
          https://www.michigan.gov/taxes/-/media/Project/Websites/taxes/Forms/2022/2022-IIT-Forms/BOOK_MI-1040.pdf?rev=4524f1cbbc0243b194ab05a6680be75b&hash=40DFE71E651F7C1D65C31F6CD03D944F
          https://www.michigan.gov/taxes/questions/iit/accordion/general/what-cities-impose-an-income-tax
        
        Mississippi:
        -         
        - https://www.dor.ms.gov/individual/tax-rates
          https://www.dor.ms.gov/individual/tax-rates
          https://www.dor.ms.gov/individual/tax-rates
          https://www.dor.ms.gov/individual
          https://www.dor.ms.gov/individual
        """
        
        # Check if the parameters are set upstream in the instance
        county_occupation = self.county_occupation if hasattr(self, 'county_occupation') else county_occupation
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Michigan', 'state_occupation': 'Mississippi', "county_occupation": county_occupation, "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def michigan_and_missouri(self, capital_gains_short=0, county_occupation='Other', dependents=0, sources=False, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_long=0, filing_status='single', income=100000, credits_state=0):
        """
        --- Parameters ----
        
        Michigan:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        county_occupation : str, optional
            County of occupation. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Missouri:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Michigan:
        - https://www.michigan.gov/taxes/iit/new-developments-for-tax-year-2022
          "https://www.michigan.gov/taxes/-/media/Project/Websites/taxes/Forms/2023/2023-SUW/446_Withholding-Guide_2023.pdf?rev=177ca3d8cb034392bf84e317ed78c320&hash=19C5F42F43A89A91863D229349F48883#:~:text=The%20exemption%20amount%20is%20%245%2C400
         employees%20Michigan%20income%20tax%20return"
          "https://www.michigan.gov/taxes/-/media/Project/Websites/taxes/Forms/2023/2023-SUW/446_Withholding-Guide_2023.pdf?rev=177ca3d8cb034392bf84e317ed78c320&hash=19C5F42F43A89A91863D229349F48883#:~:text=The%20exemption%20amount%20is%20%245%2C400
         employees%20Michigan%20income%20tax%20return"
          https://www.michigan.gov/taxes/questions/iit/accordion/residency/are-my-wages-earned-in-another-state-taxable-in-michigan-if-i-am-a-michigan-resident-1
          https://www.michigan.gov/taxes/-/media/Project/Websites/taxes/Forms/2022/2022-IIT-Forms/BOOK_MI-1040.pdf?rev=4524f1cbbc0243b194ab05a6680be75b&hash=40DFE71E651F7C1D65C31F6CD03D944F
          https://www.michigan.gov/taxes/questions/iit/accordion/general/what-cities-impose-an-income-tax
        
        Missouri:
        -         
        - https://dor.mo.gov/taxation/individual/tax-types/income/year-changes/
          https://dor.mo.gov/taxation/individual/tax-types/income/year-changes/
          https://dor.mo.gov/taxation/individual/tax-types/income/year-changes/
          https://www.kcmo.gov/city-hall/departments/finance/earnings-tax
          https://www.stlouis-mo.gov/government/departments/collector/earnings-tax/individual-earnings-tax-info.cfm
        """
        
        # Check if the parameters are set upstream in the instance
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        county_occupation = self.county_occupation if hasattr(self, 'county_occupation') else county_occupation
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Michigan', 'state_occupation': 'Missouri', "capital_gains_short": capital_gains_short, "county_occupation": county_occupation, "dependents": dependents, "sources": sources, "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def michigan_and_montana(self, capital_gains_short=0, county_occupation='Other', dependents=0, sources=False, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_long=0, filing_status='single', income=100000, credits_state=0):
        """
        --- Parameters ----
        
        Michigan:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        county_occupation : str, optional
            County of occupation. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Montana:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Michigan:
        - https://www.michigan.gov/taxes/iit/new-developments-for-tax-year-2022
          "https://www.michigan.gov/taxes/-/media/Project/Websites/taxes/Forms/2023/2023-SUW/446_Withholding-Guide_2023.pdf?rev=177ca3d8cb034392bf84e317ed78c320&hash=19C5F42F43A89A91863D229349F48883#:~:text=The%20exemption%20amount%20is%20%245%2C400
         employees%20Michigan%20income%20tax%20return"
          "https://www.michigan.gov/taxes/-/media/Project/Websites/taxes/Forms/2023/2023-SUW/446_Withholding-Guide_2023.pdf?rev=177ca3d8cb034392bf84e317ed78c320&hash=19C5F42F43A89A91863D229349F48883#:~:text=The%20exemption%20amount%20is%20%245%2C400
         employees%20Michigan%20income%20tax%20return"
          https://www.michigan.gov/taxes/questions/iit/accordion/residency/are-my-wages-earned-in-another-state-taxable-in-michigan-if-i-am-a-michigan-resident-1
          https://www.michigan.gov/taxes/-/media/Project/Websites/taxes/Forms/2022/2022-IIT-Forms/BOOK_MI-1040.pdf?rev=4524f1cbbc0243b194ab05a6680be75b&hash=40DFE71E651F7C1D65C31F6CD03D944F
          https://www.michigan.gov/taxes/questions/iit/accordion/general/what-cities-impose-an-income-tax
        
        Montana:
        -         
        - https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/2022/12/montana-employees-withholding-allowance-and-exemption-certificate-form-mw-4.pdf
          https://mtrevenue.gov/2021/11/29/simplification-of-montana-income-taxation/?utm_source=rss&utm_medium=rss&utm_campaign=simplification-of-montana-income-taxation
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        county_occupation = self.county_occupation if hasattr(self, 'county_occupation') else county_occupation
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Michigan', 'state_occupation': 'Montana', "capital_gains_short": capital_gains_short, "county_occupation": county_occupation, "dependents": dependents, "sources": sources, "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def michigan_and_nebraska(self, county_occupation='Other', county_residence='Other', state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Michigan:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        county_occupation : str, optional
            County of occupation. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Nebraska:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Michigan:
        - https://www.michigan.gov/taxes/iit/new-developments-for-tax-year-2022
          "https://www.michigan.gov/taxes/-/media/Project/Websites/taxes/Forms/2023/2023-SUW/446_Withholding-Guide_2023.pdf?rev=177ca3d8cb034392bf84e317ed78c320&hash=19C5F42F43A89A91863D229349F48883#:~:text=The%20exemption%20amount%20is%20%245%2C400
         employees%20Michigan%20income%20tax%20return"
          "https://www.michigan.gov/taxes/-/media/Project/Websites/taxes/Forms/2023/2023-SUW/446_Withholding-Guide_2023.pdf?rev=177ca3d8cb034392bf84e317ed78c320&hash=19C5F42F43A89A91863D229349F48883#:~:text=The%20exemption%20amount%20is%20%245%2C400
         employees%20Michigan%20income%20tax%20return"
          https://www.michigan.gov/taxes/questions/iit/accordion/residency/are-my-wages-earned-in-another-state-taxable-in-michigan-if-i-am-a-michigan-resident-1
          https://www.michigan.gov/taxes/-/media/Project/Websites/taxes/Forms/2022/2022-IIT-Forms/BOOK_MI-1040.pdf?rev=4524f1cbbc0243b194ab05a6680be75b&hash=40DFE71E651F7C1D65C31F6CD03D944F
          https://www.michigan.gov/taxes/questions/iit/accordion/general/what-cities-impose-an-income-tax
        
        Nebraska:
        -         
        - https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2023_Circular_EN_8-429r2023_Whole_Book_11.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        county_occupation = self.county_occupation if hasattr(self, 'county_occupation') else county_occupation
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Michigan', 'state_occupation': 'Nebraska', "county_occupation": county_occupation, "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def michigan_and_nevada(self, county_occupation='Other', county_residence='Other', state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Michigan:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        county_occupation : str, optional
            County of occupation. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Nevada:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Michigan:
        - https://www.michigan.gov/taxes/iit/new-developments-for-tax-year-2022
          "https://www.michigan.gov/taxes/-/media/Project/Websites/taxes/Forms/2023/2023-SUW/446_Withholding-Guide_2023.pdf?rev=177ca3d8cb034392bf84e317ed78c320&hash=19C5F42F43A89A91863D229349F48883#:~:text=The%20exemption%20amount%20is%20%245%2C400
         employees%20Michigan%20income%20tax%20return"
          "https://www.michigan.gov/taxes/-/media/Project/Websites/taxes/Forms/2023/2023-SUW/446_Withholding-Guide_2023.pdf?rev=177ca3d8cb034392bf84e317ed78c320&hash=19C5F42F43A89A91863D229349F48883#:~:text=The%20exemption%20amount%20is%20%245%2C400
         employees%20Michigan%20income%20tax%20return"
          https://www.michigan.gov/taxes/questions/iit/accordion/residency/are-my-wages-earned-in-another-state-taxable-in-michigan-if-i-am-a-michigan-resident-1
          https://www.michigan.gov/taxes/-/media/Project/Websites/taxes/Forms/2022/2022-IIT-Forms/BOOK_MI-1040.pdf?rev=4524f1cbbc0243b194ab05a6680be75b&hash=40DFE71E651F7C1D65C31F6CD03D944F
          https://www.michigan.gov/taxes/questions/iit/accordion/general/what-cities-impose-an-income-tax
        
        Nevada:
        -         
        - https://www.nvsos.gov/sos/businesses/the-nevada-advantage
        """
        
        # Check if the parameters are set upstream in the instance
        county_occupation = self.county_occupation if hasattr(self, 'county_occupation') else county_occupation
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Michigan', 'state_occupation': 'Nevada', "county_occupation": county_occupation, "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def michigan_and_new_hampshire(self, county_occupation='Other', county_residence='Other', state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Michigan:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        county_occupation : str, optional
            County of occupation. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        New Hampshire:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Michigan:
        - https://www.michigan.gov/taxes/iit/new-developments-for-tax-year-2022
          "https://www.michigan.gov/taxes/-/media/Project/Websites/taxes/Forms/2023/2023-SUW/446_Withholding-Guide_2023.pdf?rev=177ca3d8cb034392bf84e317ed78c320&hash=19C5F42F43A89A91863D229349F48883#:~:text=The%20exemption%20amount%20is%20%245%2C400
         employees%20Michigan%20income%20tax%20return"
          "https://www.michigan.gov/taxes/-/media/Project/Websites/taxes/Forms/2023/2023-SUW/446_Withholding-Guide_2023.pdf?rev=177ca3d8cb034392bf84e317ed78c320&hash=19C5F42F43A89A91863D229349F48883#:~:text=The%20exemption%20amount%20is%20%245%2C400
         employees%20Michigan%20income%20tax%20return"
          https://www.michigan.gov/taxes/questions/iit/accordion/residency/are-my-wages-earned-in-another-state-taxable-in-michigan-if-i-am-a-michigan-resident-1
          https://www.michigan.gov/taxes/-/media/Project/Websites/taxes/Forms/2022/2022-IIT-Forms/BOOK_MI-1040.pdf?rev=4524f1cbbc0243b194ab05a6680be75b&hash=40DFE71E651F7C1D65C31F6CD03D944F
          https://www.michigan.gov/taxes/questions/iit/accordion/general/what-cities-impose-an-income-tax
        
        New Hampshire:
        -         
        - https://www.revenue.nh.gov/assistance/tax-overview.htm
        """
        
        # Check if the parameters are set upstream in the instance
        county_occupation = self.county_occupation if hasattr(self, 'county_occupation') else county_occupation
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Michigan', 'state_occupation': 'NewHampshire', "county_occupation": county_occupation, "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def michigan_and_new_jersey(self, capital_gains_short=0, county_occupation='Other', dependents=0, sources=False, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_long=0, filing_status='single', income=100000, credits_state=0):
        """
        --- Parameters ----
        
        Michigan:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        county_occupation : str, optional
            County of occupation. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        New Jersey:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Michigan:
        - https://www.michigan.gov/taxes/iit/new-developments-for-tax-year-2022
          "https://www.michigan.gov/taxes/-/media/Project/Websites/taxes/Forms/2023/2023-SUW/446_Withholding-Guide_2023.pdf?rev=177ca3d8cb034392bf84e317ed78c320&hash=19C5F42F43A89A91863D229349F48883#:~:text=The%20exemption%20amount%20is%20%245%2C400
         employees%20Michigan%20income%20tax%20return"
          "https://www.michigan.gov/taxes/-/media/Project/Websites/taxes/Forms/2023/2023-SUW/446_Withholding-Guide_2023.pdf?rev=177ca3d8cb034392bf84e317ed78c320&hash=19C5F42F43A89A91863D229349F48883#:~:text=The%20exemption%20amount%20is%20%245%2C400
         employees%20Michigan%20income%20tax%20return"
          https://www.michigan.gov/taxes/questions/iit/accordion/residency/are-my-wages-earned-in-another-state-taxable-in-michigan-if-i-am-a-michigan-resident-1
          https://www.michigan.gov/taxes/-/media/Project/Websites/taxes/Forms/2022/2022-IIT-Forms/BOOK_MI-1040.pdf?rev=4524f1cbbc0243b194ab05a6680be75b&hash=40DFE71E651F7C1D65C31F6CD03D944F
          https://www.michigan.gov/taxes/questions/iit/accordion/general/what-cities-impose-an-income-tax
        
        New Jersey:
        -         
        - https://www.state.nj.us/treasury/taxation/taxtables.shtml
          https://www.state.nj.us/treasury/taxation/njit13.shtml
          https://www.nj.gov/treasury/taxation/njit25.shtml
          https://www.state.nj.us/treasury/taxation/individuals/freqqiti.shtml
          https://www.state.nj.us/treasury/taxation/localtax.shtml
        """
        
        # Check if the parameters are set upstream in the instance
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        county_occupation = self.county_occupation if hasattr(self, 'county_occupation') else county_occupation
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Michigan', 'state_occupation': 'NewJersey', "capital_gains_short": capital_gains_short, "county_occupation": county_occupation, "dependents": dependents, "sources": sources, "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def michigan_and_new_mexico(self, county_occupation='Other', county_residence='Other', state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Michigan:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        county_occupation : str, optional
            County of occupation. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        New Mexico:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Michigan:
        - https://www.michigan.gov/taxes/iit/new-developments-for-tax-year-2022
          "https://www.michigan.gov/taxes/-/media/Project/Websites/taxes/Forms/2023/2023-SUW/446_Withholding-Guide_2023.pdf?rev=177ca3d8cb034392bf84e317ed78c320&hash=19C5F42F43A89A91863D229349F48883#:~:text=The%20exemption%20amount%20is%20%245%2C400
         employees%20Michigan%20income%20tax%20return"
          "https://www.michigan.gov/taxes/-/media/Project/Websites/taxes/Forms/2023/2023-SUW/446_Withholding-Guide_2023.pdf?rev=177ca3d8cb034392bf84e317ed78c320&hash=19C5F42F43A89A91863D229349F48883#:~:text=The%20exemption%20amount%20is%20%245%2C400
         employees%20Michigan%20income%20tax%20return"
          https://www.michigan.gov/taxes/questions/iit/accordion/residency/are-my-wages-earned-in-another-state-taxable-in-michigan-if-i-am-a-michigan-resident-1
          https://www.michigan.gov/taxes/-/media/Project/Websites/taxes/Forms/2022/2022-IIT-Forms/BOOK_MI-1040.pdf?rev=4524f1cbbc0243b194ab05a6680be75b&hash=40DFE71E651F7C1D65C31F6CD03D944F
          https://www.michigan.gov/taxes/questions/iit/accordion/general/what-cities-impose-an-income-tax
        
        New Mexico:
        -         
        - https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/fdf3c548-8aba-4b9c-9eb4-bb564c716015/FYI-104.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/0bf943fd-652e-4400-bb1b-cd8397eb5a95/2022pit-adj.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/2f1a6781-9534-4436-b427-1557f9592099/2022pit-adj-ins.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/0bf943fd-652e-4400-bb1b-cd8397eb5a95/2022pit-adj.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/1afc56af-ea90-4d48-82e5-1f9aeb43255a/PITbook2022.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/fdf3c548-8aba-4b9c-9eb4-bb564c716015/FYI-104.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        county_occupation = self.county_occupation if hasattr(self, 'county_occupation') else county_occupation
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Michigan', 'state_occupation': 'NewMexico', "county_occupation": county_occupation, "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def michigan_and_new_york(self, capital_gains_short=0, county_occupation='Other', dependents=0, sources=False, county_residence='Other', state_agi=75000, capital_gains_long=0, filing_status='single', income=100000, credits_state=0):
        """
        --- Parameters ----
        
        Michigan:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        county_occupation : str, optional
            County of occupation. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        New York:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Michigan:
        - https://www.michigan.gov/taxes/iit/new-developments-for-tax-year-2022
          "https://www.michigan.gov/taxes/-/media/Project/Websites/taxes/Forms/2023/2023-SUW/446_Withholding-Guide_2023.pdf?rev=177ca3d8cb034392bf84e317ed78c320&hash=19C5F42F43A89A91863D229349F48883#:~:text=The%20exemption%20amount%20is%20%245%2C400
         employees%20Michigan%20income%20tax%20return"
          "https://www.michigan.gov/taxes/-/media/Project/Websites/taxes/Forms/2023/2023-SUW/446_Withholding-Guide_2023.pdf?rev=177ca3d8cb034392bf84e317ed78c320&hash=19C5F42F43A89A91863D229349F48883#:~:text=The%20exemption%20amount%20is%20%245%2C400
         employees%20Michigan%20income%20tax%20return"
          https://www.michigan.gov/taxes/questions/iit/accordion/residency/are-my-wages-earned-in-another-state-taxable-in-michigan-if-i-am-a-michigan-resident-1
          https://www.michigan.gov/taxes/-/media/Project/Websites/taxes/Forms/2022/2022-IIT-Forms/BOOK_MI-1040.pdf?rev=4524f1cbbc0243b194ab05a6680be75b&hash=40DFE71E651F7C1D65C31F6CD03D944F
          https://www.michigan.gov/taxes/questions/iit/accordion/general/what-cities-impose-an-income-tax
        
        New York:
        -         
        - https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nys-tax-rate-schedule
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nys-tax-rate-schedule
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nys-tax-rate-schedule
          https://www.tax.ny.gov/pit/
          https://www.tax.ny.gov/pit/file/nonresident-faqs.htm
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#yonkers-tax-rate-schedule
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nyc-tax-rate-schedule
        """
        
        # Check if the parameters are set upstream in the instance
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        county_occupation = self.county_occupation if hasattr(self, 'county_occupation') else county_occupation
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Michigan', 'state_occupation': 'NewYork', "capital_gains_short": capital_gains_short, "county_occupation": county_occupation, "dependents": dependents, "sources": sources, "county_residence": county_residence, "state_agi": state_agi, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def michigan_and_north_carolina(self, county_occupation='Other', county_residence='Other', state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Michigan:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        county_occupation : str, optional
            County of occupation. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        North Carolina:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Michigan:
        - https://www.michigan.gov/taxes/iit/new-developments-for-tax-year-2022
          "https://www.michigan.gov/taxes/-/media/Project/Websites/taxes/Forms/2023/2023-SUW/446_Withholding-Guide_2023.pdf?rev=177ca3d8cb034392bf84e317ed78c320&hash=19C5F42F43A89A91863D229349F48883#:~:text=The%20exemption%20amount%20is%20%245%2C400
         employees%20Michigan%20income%20tax%20return"
          "https://www.michigan.gov/taxes/-/media/Project/Websites/taxes/Forms/2023/2023-SUW/446_Withholding-Guide_2023.pdf?rev=177ca3d8cb034392bf84e317ed78c320&hash=19C5F42F43A89A91863D229349F48883#:~:text=The%20exemption%20amount%20is%20%245%2C400
         employees%20Michigan%20income%20tax%20return"
          https://www.michigan.gov/taxes/questions/iit/accordion/residency/are-my-wages-earned-in-another-state-taxable-in-michigan-if-i-am-a-michigan-resident-1
          https://www.michigan.gov/taxes/-/media/Project/Websites/taxes/Forms/2022/2022-IIT-Forms/BOOK_MI-1040.pdf?rev=4524f1cbbc0243b194ab05a6680be75b&hash=40DFE71E651F7C1D65C31F6CD03D944F
          https://www.michigan.gov/taxes/questions/iit/accordion/general/what-cities-impose-an-income-tax
        
        North Carolina:
        -         
        - https://www.ncdor.gov/taxes-forms/tax-rate-schedules
          https://www.ncdor.gov/taxes-forms/individual-income-tax/north-carolina-standard-deduction-or-north-carolina-itemized-deductions
          https://www.ncdor.gov/taxes-forms/individual-income-tax
          https://www.ncdor.gov/taxes-forms/individual-income-tax
        """
        
        # Check if the parameters are set upstream in the instance
        county_occupation = self.county_occupation if hasattr(self, 'county_occupation') else county_occupation
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Michigan', 'state_occupation': 'NorthCarolina', "county_occupation": county_occupation, "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def michigan_and_north_dakota(self, capital_gains_short=0, county_occupation='Other', dependents=0, sources=False, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_long=0, filing_status='single', income=100000, credits_state=0):
        """
        --- Parameters ----
        
        Michigan:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        county_occupation : str, optional
            County of occupation. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        North Dakota:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Michigan:
        - https://www.michigan.gov/taxes/iit/new-developments-for-tax-year-2022
          "https://www.michigan.gov/taxes/-/media/Project/Websites/taxes/Forms/2023/2023-SUW/446_Withholding-Guide_2023.pdf?rev=177ca3d8cb034392bf84e317ed78c320&hash=19C5F42F43A89A91863D229349F48883#:~:text=The%20exemption%20amount%20is%20%245%2C400
         employees%20Michigan%20income%20tax%20return"
          "https://www.michigan.gov/taxes/-/media/Project/Websites/taxes/Forms/2023/2023-SUW/446_Withholding-Guide_2023.pdf?rev=177ca3d8cb034392bf84e317ed78c320&hash=19C5F42F43A89A91863D229349F48883#:~:text=The%20exemption%20amount%20is%20%245%2C400
         employees%20Michigan%20income%20tax%20return"
          https://www.michigan.gov/taxes/questions/iit/accordion/residency/are-my-wages-earned-in-another-state-taxable-in-michigan-if-i-am-a-michigan-resident-1
          https://www.michigan.gov/taxes/-/media/Project/Websites/taxes/Forms/2022/2022-IIT-Forms/BOOK_MI-1040.pdf?rev=4524f1cbbc0243b194ab05a6680be75b&hash=40DFE71E651F7C1D65C31F6CD03D944F
          https://www.michigan.gov/taxes/questions/iit/accordion/general/what-cities-impose-an-income-tax
        
        North Dakota:
        -         
        - https://www.tax.nd.gov/forms
          https://www.tax.nd.gov/sites/www/files/documents/forms/individual/2022-iit/2022-individual-income-tax-booklet.pdf
          https://www.tax.nd.gov/sites/www/files/documents/forms/2022-individual-income-tax-booklet.pdf
          https://www.tax.nd.gov/tax-exemptions-credits/income-tax-exemptions-credits
          https://www.tax.nd.gov/sites/www/files/documents/forms/business/ndwrfillable.pdf
          https://www.tax.nd.gov/news/tax-legislative-changes/significant-changes-law/individual-income-tax-history
          https://www.tax.nd.gov/forms
        """
        
        # Check if the parameters are set upstream in the instance
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        county_occupation = self.county_occupation if hasattr(self, 'county_occupation') else county_occupation
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Michigan', 'state_occupation': 'NorthDakota', "capital_gains_short": capital_gains_short, "county_occupation": county_occupation, "dependents": dependents, "sources": sources, "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def michigan_and_ohio(self, county_occupation='Other', county_residence='Other', state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Michigan:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        county_occupation : str, optional
            County of occupation. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Ohio:
        


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Michigan:
        - https://www.michigan.gov/taxes/iit/new-developments-for-tax-year-2022
          "https://www.michigan.gov/taxes/-/media/Project/Websites/taxes/Forms/2023/2023-SUW/446_Withholding-Guide_2023.pdf?rev=177ca3d8cb034392bf84e317ed78c320&hash=19C5F42F43A89A91863D229349F48883#:~:text=The%20exemption%20amount%20is%20%245%2C400
         employees%20Michigan%20income%20tax%20return"
          "https://www.michigan.gov/taxes/-/media/Project/Websites/taxes/Forms/2023/2023-SUW/446_Withholding-Guide_2023.pdf?rev=177ca3d8cb034392bf84e317ed78c320&hash=19C5F42F43A89A91863D229349F48883#:~:text=The%20exemption%20amount%20is%20%245%2C400
         employees%20Michigan%20income%20tax%20return"
          https://www.michigan.gov/taxes/questions/iit/accordion/residency/are-my-wages-earned-in-another-state-taxable-in-michigan-if-i-am-a-michigan-resident-1
          https://www.michigan.gov/taxes/-/media/Project/Websites/taxes/Forms/2022/2022-IIT-Forms/BOOK_MI-1040.pdf?rev=4524f1cbbc0243b194ab05a6680be75b&hash=40DFE71E651F7C1D65C31F6CD03D944F
          https://www.michigan.gov/taxes/questions/iit/accordion/general/what-cities-impose-an-income-tax
        
        Ohio:
        -         
        - https://tax.ohio.gov/individual/resources/annual-tax-rates
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://tax.ohio.gov/static/forms/employer_withholding/generic/wth_it4nr.pdf
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://www.ritaohio.com/TaxRatesTable
        """
        
        # Check if the parameters are set upstream in the instance
        county_occupation = self.county_occupation if hasattr(self, 'county_occupation') else county_occupation
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Michigan', 'state_occupation': 'Ohio', "county_occupation": county_occupation, "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def michigan_and_oklahoma(self, capital_gains_short=0, county_occupation='Other', dependents=0, sources=False, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_long=0, filing_status='single', income=100000, credits_state=0):
        """
        --- Parameters ----
        
        Michigan:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        county_occupation : str, optional
            County of occupation. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Oklahoma:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Michigan:
        - https://www.michigan.gov/taxes/iit/new-developments-for-tax-year-2022
          "https://www.michigan.gov/taxes/-/media/Project/Websites/taxes/Forms/2023/2023-SUW/446_Withholding-Guide_2023.pdf?rev=177ca3d8cb034392bf84e317ed78c320&hash=19C5F42F43A89A91863D229349F48883#:~:text=The%20exemption%20amount%20is%20%245%2C400
         employees%20Michigan%20income%20tax%20return"
          "https://www.michigan.gov/taxes/-/media/Project/Websites/taxes/Forms/2023/2023-SUW/446_Withholding-Guide_2023.pdf?rev=177ca3d8cb034392bf84e317ed78c320&hash=19C5F42F43A89A91863D229349F48883#:~:text=The%20exemption%20amount%20is%20%245%2C400
         employees%20Michigan%20income%20tax%20return"
          https://www.michigan.gov/taxes/questions/iit/accordion/residency/are-my-wages-earned-in-another-state-taxable-in-michigan-if-i-am-a-michigan-resident-1
          https://www.michigan.gov/taxes/-/media/Project/Websites/taxes/Forms/2022/2022-IIT-Forms/BOOK_MI-1040.pdf?rev=4524f1cbbc0243b194ab05a6680be75b&hash=40DFE71E651F7C1D65C31F6CD03D944F
          https://www.michigan.gov/taxes/questions/iit/accordion/general/what-cities-impose-an-income-tax
        
        Oklahoma:
        -         
        - https://oklahoma.gov/content/dam/ok/en/tax/documents/resources/publications/businesses/withholding-tables/WHTables-2022.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        county_occupation = self.county_occupation if hasattr(self, 'county_occupation') else county_occupation
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Michigan', 'state_occupation': 'Oklahoma', "capital_gains_short": capital_gains_short, "county_occupation": county_occupation, "dependents": dependents, "sources": sources, "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def michigan_and_oregon(self, capital_gains_short=0, county_occupation='Other', dependents=0, sources=False, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_long=0, filing_status='single', income=100000, credits_state=0):
        """
        --- Parameters ----
        
        Michigan:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        county_occupation : str, optional
            County of occupation. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Oregon:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Michigan:
        - https://www.michigan.gov/taxes/iit/new-developments-for-tax-year-2022
          "https://www.michigan.gov/taxes/-/media/Project/Websites/taxes/Forms/2023/2023-SUW/446_Withholding-Guide_2023.pdf?rev=177ca3d8cb034392bf84e317ed78c320&hash=19C5F42F43A89A91863D229349F48883#:~:text=The%20exemption%20amount%20is%20%245%2C400
         employees%20Michigan%20income%20tax%20return"
          "https://www.michigan.gov/taxes/-/media/Project/Websites/taxes/Forms/2023/2023-SUW/446_Withholding-Guide_2023.pdf?rev=177ca3d8cb034392bf84e317ed78c320&hash=19C5F42F43A89A91863D229349F48883#:~:text=The%20exemption%20amount%20is%20%245%2C400
         employees%20Michigan%20income%20tax%20return"
          https://www.michigan.gov/taxes/questions/iit/accordion/residency/are-my-wages-earned-in-another-state-taxable-in-michigan-if-i-am-a-michigan-resident-1
          https://www.michigan.gov/taxes/-/media/Project/Websites/taxes/Forms/2022/2022-IIT-Forms/BOOK_MI-1040.pdf?rev=4524f1cbbc0243b194ab05a6680be75b&hash=40DFE71E651F7C1D65C31F6CD03D944F
          https://www.michigan.gov/taxes/questions/iit/accordion/general/what-cities-impose-an-income-tax
        
        Oregon:
        -         
        - https://www.oregon.gov/dor/programs/individuals/Pages/PIT.aspx
          https://www.oregon.gov/dor/programs/individuals/Documents/Full%20year%20resident
         %20Form%20or-40%20filers.pdf
          https://www.oregon.gov/dor/programs/individuals/Documents/Part-year%20and%20nonresident
         %20Form%20or-40-P%20and%20or-40-N%20filers.pdf.pdf
          https://www.oregon.gov/dor/forms/FormsPubs/form-or-40_101-040_2022.pdf
          https://www.oregon.gov/dor/forms/FormsPubs/form-or-40_101-040_2022.pdf
          https://www.oregon.gov/dor/forms/FormsPubs/form-or-40_101-040_2022.pdf
          https://www.portland.gov/revenue/personal-tax
        """
        
        # Check if the parameters are set upstream in the instance
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        county_occupation = self.county_occupation if hasattr(self, 'county_occupation') else county_occupation
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Michigan', 'state_occupation': 'Oregon', "capital_gains_short": capital_gains_short, "county_occupation": county_occupation, "dependents": dependents, "sources": sources, "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def michigan_and_pennsylvania(self, county_occupation='Other', county_residence='Other', state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Michigan:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        county_occupation : str, optional
            County of occupation. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Pennsylvania:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        county_residence : str, optional
            County of residence. Default is Other.
        county_occupation : str, optional
            County of occupation. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Michigan:
        - https://www.michigan.gov/taxes/iit/new-developments-for-tax-year-2022
          "https://www.michigan.gov/taxes/-/media/Project/Websites/taxes/Forms/2023/2023-SUW/446_Withholding-Guide_2023.pdf?rev=177ca3d8cb034392bf84e317ed78c320&hash=19C5F42F43A89A91863D229349F48883#:~:text=The%20exemption%20amount%20is%20%245%2C400
         employees%20Michigan%20income%20tax%20return"
          "https://www.michigan.gov/taxes/-/media/Project/Websites/taxes/Forms/2023/2023-SUW/446_Withholding-Guide_2023.pdf?rev=177ca3d8cb034392bf84e317ed78c320&hash=19C5F42F43A89A91863D229349F48883#:~:text=The%20exemption%20amount%20is%20%245%2C400
         employees%20Michigan%20income%20tax%20return"
          https://www.michigan.gov/taxes/questions/iit/accordion/residency/are-my-wages-earned-in-another-state-taxable-in-michigan-if-i-am-a-michigan-resident-1
          https://www.michigan.gov/taxes/-/media/Project/Websites/taxes/Forms/2022/2022-IIT-Forms/BOOK_MI-1040.pdf?rev=4524f1cbbc0243b194ab05a6680be75b&hash=40DFE71E651F7C1D65C31F6CD03D944F
          https://www.michigan.gov/taxes/questions/iit/accordion/general/what-cities-impose-an-income-tax
        
        Pennsylvania:
        -         
        - https://www.revenue.pa.gov/Tax%20Rates/Pages/default.aspx
          https://www.revenue.pa.gov/FormsandPublications/FormsforIndividuals/PIT/Documents/2022/2022_pa-40in.pdf
          https://www.revenue.pa.gov/FormsandPublications/FormsforBusinesses/EmployerWithholding/Documents/rev-419.pdf
          https://www.revenue.pa.gov/FormsandPublications/FormsforIndividuals/PIT/Documents/2022/2022_pa-40in.pdf
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-25.htm
        """
        
        # Check if the parameters are set upstream in the instance
        county_occupation = self.county_occupation if hasattr(self, 'county_occupation') else county_occupation
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Michigan', 'state_occupation': 'Pennsylvania', "county_occupation": county_occupation, "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def michigan_and_rhode_island(self, county_occupation='Other', county_residence='Other', state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Michigan:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        county_occupation : str, optional
            County of occupation. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Rhode Island:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Michigan:
        - https://www.michigan.gov/taxes/iit/new-developments-for-tax-year-2022
          "https://www.michigan.gov/taxes/-/media/Project/Websites/taxes/Forms/2023/2023-SUW/446_Withholding-Guide_2023.pdf?rev=177ca3d8cb034392bf84e317ed78c320&hash=19C5F42F43A89A91863D229349F48883#:~:text=The%20exemption%20amount%20is%20%245%2C400
         employees%20Michigan%20income%20tax%20return"
          "https://www.michigan.gov/taxes/-/media/Project/Websites/taxes/Forms/2023/2023-SUW/446_Withholding-Guide_2023.pdf?rev=177ca3d8cb034392bf84e317ed78c320&hash=19C5F42F43A89A91863D229349F48883#:~:text=The%20exemption%20amount%20is%20%245%2C400
         employees%20Michigan%20income%20tax%20return"
          https://www.michigan.gov/taxes/questions/iit/accordion/residency/are-my-wages-earned-in-another-state-taxable-in-michigan-if-i-am-a-michigan-resident-1
          https://www.michigan.gov/taxes/-/media/Project/Websites/taxes/Forms/2022/2022-IIT-Forms/BOOK_MI-1040.pdf?rev=4524f1cbbc0243b194ab05a6680be75b&hash=40DFE71E651F7C1D65C31F6CD03D944F
          https://www.michigan.gov/taxes/questions/iit/accordion/general/what-cities-impose-an-income-tax
        
        Rhode Island:
        -         
        - https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2022_1040WE_w_0.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        county_occupation = self.county_occupation if hasattr(self, 'county_occupation') else county_occupation
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Michigan', 'state_occupation': 'RhodeIsland', "county_occupation": county_occupation, "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def michigan_and_south_carolina(self, capital_gains_short=0, county_occupation='Other', dependents=0, sources=False, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_long=0, filing_status='single', income=100000, credits_state=0):
        """
        --- Parameters ----
        
        Michigan:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        county_occupation : str, optional
            County of occupation. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        South Carolina:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Michigan:
        - https://www.michigan.gov/taxes/iit/new-developments-for-tax-year-2022
          "https://www.michigan.gov/taxes/-/media/Project/Websites/taxes/Forms/2023/2023-SUW/446_Withholding-Guide_2023.pdf?rev=177ca3d8cb034392bf84e317ed78c320&hash=19C5F42F43A89A91863D229349F48883#:~:text=The%20exemption%20amount%20is%20%245%2C400
         employees%20Michigan%20income%20tax%20return"
          "https://www.michigan.gov/taxes/-/media/Project/Websites/taxes/Forms/2023/2023-SUW/446_Withholding-Guide_2023.pdf?rev=177ca3d8cb034392bf84e317ed78c320&hash=19C5F42F43A89A91863D229349F48883#:~:text=The%20exemption%20amount%20is%20%245%2C400
         employees%20Michigan%20income%20tax%20return"
          https://www.michigan.gov/taxes/questions/iit/accordion/residency/are-my-wages-earned-in-another-state-taxable-in-michigan-if-i-am-a-michigan-resident-1
          https://www.michigan.gov/taxes/-/media/Project/Websites/taxes/Forms/2022/2022-IIT-Forms/BOOK_MI-1040.pdf?rev=4524f1cbbc0243b194ab05a6680be75b&hash=40DFE71E651F7C1D65C31F6CD03D944F
          https://www.michigan.gov/taxes/questions/iit/accordion/general/what-cities-impose-an-income-tax
        
        South Carolina:
        -         
        - https://dor.sc.gov/resources-site/lawandpolicy/Advisory%20Opinions/IL22-15.pdf
          https://dor.sc.gov/forms-site/Forms/IITPacket_2022.pdf
          https://www.scstatehouse.gov/code/t12c006.php
          https://dor.sc.gov/resources-site/lawandpolicy/Advisory%20Opinions/IL22-15.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        county_occupation = self.county_occupation if hasattr(self, 'county_occupation') else county_occupation
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Michigan', 'state_occupation': 'SouthCarolina', "capital_gains_short": capital_gains_short, "county_occupation": county_occupation, "dependents": dependents, "sources": sources, "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def michigan_and_south_dakota(self, county_occupation='Other', county_residence='Other', state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Michigan:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        county_occupation : str, optional
            County of occupation. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        South Dakota:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Michigan:
        - https://www.michigan.gov/taxes/iit/new-developments-for-tax-year-2022
          "https://www.michigan.gov/taxes/-/media/Project/Websites/taxes/Forms/2023/2023-SUW/446_Withholding-Guide_2023.pdf?rev=177ca3d8cb034392bf84e317ed78c320&hash=19C5F42F43A89A91863D229349F48883#:~:text=The%20exemption%20amount%20is%20%245%2C400
         employees%20Michigan%20income%20tax%20return"
          "https://www.michigan.gov/taxes/-/media/Project/Websites/taxes/Forms/2023/2023-SUW/446_Withholding-Guide_2023.pdf?rev=177ca3d8cb034392bf84e317ed78c320&hash=19C5F42F43A89A91863D229349F48883#:~:text=The%20exemption%20amount%20is%20%245%2C400
         employees%20Michigan%20income%20tax%20return"
          https://www.michigan.gov/taxes/questions/iit/accordion/residency/are-my-wages-earned-in-another-state-taxable-in-michigan-if-i-am-a-michigan-resident-1
          https://www.michigan.gov/taxes/-/media/Project/Websites/taxes/Forms/2022/2022-IIT-Forms/BOOK_MI-1040.pdf?rev=4524f1cbbc0243b194ab05a6680be75b&hash=40DFE71E651F7C1D65C31F6CD03D944F
          https://www.michigan.gov/taxes/questions/iit/accordion/general/what-cities-impose-an-income-tax
        
        South Dakota:
        -         
        - https://dor.sd.gov/individuals/taxes/
        """
        
        # Check if the parameters are set upstream in the instance
        county_occupation = self.county_occupation if hasattr(self, 'county_occupation') else county_occupation
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Michigan', 'state_occupation': 'SouthDakota', "county_occupation": county_occupation, "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def michigan_and_tennessee(self, county_occupation='Other', county_residence='Other', state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Michigan:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        county_occupation : str, optional
            County of occupation. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Tennessee:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Michigan:
        - https://www.michigan.gov/taxes/iit/new-developments-for-tax-year-2022
          "https://www.michigan.gov/taxes/-/media/Project/Websites/taxes/Forms/2023/2023-SUW/446_Withholding-Guide_2023.pdf?rev=177ca3d8cb034392bf84e317ed78c320&hash=19C5F42F43A89A91863D229349F48883#:~:text=The%20exemption%20amount%20is%20%245%2C400
         employees%20Michigan%20income%20tax%20return"
          "https://www.michigan.gov/taxes/-/media/Project/Websites/taxes/Forms/2023/2023-SUW/446_Withholding-Guide_2023.pdf?rev=177ca3d8cb034392bf84e317ed78c320&hash=19C5F42F43A89A91863D229349F48883#:~:text=The%20exemption%20amount%20is%20%245%2C400
         employees%20Michigan%20income%20tax%20return"
          https://www.michigan.gov/taxes/questions/iit/accordion/residency/are-my-wages-earned-in-another-state-taxable-in-michigan-if-i-am-a-michigan-resident-1
          https://www.michigan.gov/taxes/-/media/Project/Websites/taxes/Forms/2022/2022-IIT-Forms/BOOK_MI-1040.pdf?rev=4524f1cbbc0243b194ab05a6680be75b&hash=40DFE71E651F7C1D65C31F6CD03D944F
          https://www.michigan.gov/taxes/questions/iit/accordion/general/what-cities-impose-an-income-tax
        
        Tennessee:
        -         
        - https://www.tn.gov/revenue/taxes.html
        """
        
        # Check if the parameters are set upstream in the instance
        county_occupation = self.county_occupation if hasattr(self, 'county_occupation') else county_occupation
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Michigan', 'state_occupation': 'Tennessee', "county_occupation": county_occupation, "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def michigan_and_texas(self, county_occupation='Other', county_residence='Other', state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Michigan:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        county_occupation : str, optional
            County of occupation. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Texas:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Michigan:
        - https://www.michigan.gov/taxes/iit/new-developments-for-tax-year-2022
          "https://www.michigan.gov/taxes/-/media/Project/Websites/taxes/Forms/2023/2023-SUW/446_Withholding-Guide_2023.pdf?rev=177ca3d8cb034392bf84e317ed78c320&hash=19C5F42F43A89A91863D229349F48883#:~:text=The%20exemption%20amount%20is%20%245%2C400
         employees%20Michigan%20income%20tax%20return"
          "https://www.michigan.gov/taxes/-/media/Project/Websites/taxes/Forms/2023/2023-SUW/446_Withholding-Guide_2023.pdf?rev=177ca3d8cb034392bf84e317ed78c320&hash=19C5F42F43A89A91863D229349F48883#:~:text=The%20exemption%20amount%20is%20%245%2C400
         employees%20Michigan%20income%20tax%20return"
          https://www.michigan.gov/taxes/questions/iit/accordion/residency/are-my-wages-earned-in-another-state-taxable-in-michigan-if-i-am-a-michigan-resident-1
          https://www.michigan.gov/taxes/-/media/Project/Websites/taxes/Forms/2022/2022-IIT-Forms/BOOK_MI-1040.pdf?rev=4524f1cbbc0243b194ab05a6680be75b&hash=40DFE71E651F7C1D65C31F6CD03D944F
          https://www.michigan.gov/taxes/questions/iit/accordion/general/what-cities-impose-an-income-tax
        
        Texas:
        -         
        - https://comptroller.texas.gov/taxes/
        """
        
        # Check if the parameters are set upstream in the instance
        county_occupation = self.county_occupation if hasattr(self, 'county_occupation') else county_occupation
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Michigan', 'state_occupation': 'Texas', "county_occupation": county_occupation, "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def michigan_and_utah(self, county_occupation='Other', county_residence='Other', state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Michigan:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        county_occupation : str, optional
            County of occupation. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Utah:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Michigan:
        - https://www.michigan.gov/taxes/iit/new-developments-for-tax-year-2022
          "https://www.michigan.gov/taxes/-/media/Project/Websites/taxes/Forms/2023/2023-SUW/446_Withholding-Guide_2023.pdf?rev=177ca3d8cb034392bf84e317ed78c320&hash=19C5F42F43A89A91863D229349F48883#:~:text=The%20exemption%20amount%20is%20%245%2C400
         employees%20Michigan%20income%20tax%20return"
          "https://www.michigan.gov/taxes/-/media/Project/Websites/taxes/Forms/2023/2023-SUW/446_Withholding-Guide_2023.pdf?rev=177ca3d8cb034392bf84e317ed78c320&hash=19C5F42F43A89A91863D229349F48883#:~:text=The%20exemption%20amount%20is%20%245%2C400
         employees%20Michigan%20income%20tax%20return"
          https://www.michigan.gov/taxes/questions/iit/accordion/residency/are-my-wages-earned-in-another-state-taxable-in-michigan-if-i-am-a-michigan-resident-1
          https://www.michigan.gov/taxes/-/media/Project/Websites/taxes/Forms/2022/2022-IIT-Forms/BOOK_MI-1040.pdf?rev=4524f1cbbc0243b194ab05a6680be75b&hash=40DFE71E651F7C1D65C31F6CD03D944F
          https://www.michigan.gov/taxes/questions/iit/accordion/general/what-cities-impose-an-income-tax
        
        Utah:
        -         
        - https://incometax.utah.gov/paying/tax-rates
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        county_occupation = self.county_occupation if hasattr(self, 'county_occupation') else county_occupation
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Michigan', 'state_occupation': 'Utah', "county_occupation": county_occupation, "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def michigan_and_vermont(self, capital_gains_short=0, county_occupation='Other', dependents=0, sources=False, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_long=0, filing_status='single', income=100000, credits_state=0):
        """
        --- Parameters ----
        
        Michigan:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        county_occupation : str, optional
            County of occupation. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Vermont:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Michigan:
        - https://www.michigan.gov/taxes/iit/new-developments-for-tax-year-2022
          "https://www.michigan.gov/taxes/-/media/Project/Websites/taxes/Forms/2023/2023-SUW/446_Withholding-Guide_2023.pdf?rev=177ca3d8cb034392bf84e317ed78c320&hash=19C5F42F43A89A91863D229349F48883#:~:text=The%20exemption%20amount%20is%20%245%2C400
         employees%20Michigan%20income%20tax%20return"
          "https://www.michigan.gov/taxes/-/media/Project/Websites/taxes/Forms/2023/2023-SUW/446_Withholding-Guide_2023.pdf?rev=177ca3d8cb034392bf84e317ed78c320&hash=19C5F42F43A89A91863D229349F48883#:~:text=The%20exemption%20amount%20is%20%245%2C400
         employees%20Michigan%20income%20tax%20return"
          https://www.michigan.gov/taxes/questions/iit/accordion/residency/are-my-wages-earned-in-another-state-taxable-in-michigan-if-i-am-a-michigan-resident-1
          https://www.michigan.gov/taxes/-/media/Project/Websites/taxes/Forms/2022/2022-IIT-Forms/BOOK_MI-1040.pdf?rev=4524f1cbbc0243b194ab05a6680be75b&hash=40DFE71E651F7C1D65C31F6CD03D944F
          https://www.michigan.gov/taxes/questions/iit/accordion/general/what-cities-impose-an-income-tax
        
        Vermont:
        -         
        - https://tax.vermont.gov/sites/tax/files/documents/GB-1210-2023.pdf
          https://tax.vermont.gov/sites/tax/files/documents/Income%20Booklet-2022.pdf
          https://tax.vermont.gov/sites/tax/files/documents/Income%20Booklet-2022.pdf
          https://tax.vermont.gov/sites/tax/files/documents/Income%20Booklet-2022.pdf
          https://tax.vermont.gov/individuals/personal-income-tax/tax-credits
          https://tax.vermont.gov/individuals/personal-income-tax/taxable-income
          https://tax.vermont.gov/individuals/personal-income-tax/local-option-tax
        """
        
        # Check if the parameters are set upstream in the instance
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        county_occupation = self.county_occupation if hasattr(self, 'county_occupation') else county_occupation
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Michigan', 'state_occupation': 'Vermont', "capital_gains_short": capital_gains_short, "county_occupation": county_occupation, "dependents": dependents, "sources": sources, "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def michigan_and_virginia(self, capital_gains_short=0, county_occupation='Other', dependents=0, sources=False, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_long=0, filing_status='single', income=100000, credits_state=0):
        """
        --- Parameters ----
        
        Michigan:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        county_occupation : str, optional
            County of occupation. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Virginia:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Michigan:
        - https://www.michigan.gov/taxes/iit/new-developments-for-tax-year-2022
          "https://www.michigan.gov/taxes/-/media/Project/Websites/taxes/Forms/2023/2023-SUW/446_Withholding-Guide_2023.pdf?rev=177ca3d8cb034392bf84e317ed78c320&hash=19C5F42F43A89A91863D229349F48883#:~:text=The%20exemption%20amount%20is%20%245%2C400
         employees%20Michigan%20income%20tax%20return"
          "https://www.michigan.gov/taxes/-/media/Project/Websites/taxes/Forms/2023/2023-SUW/446_Withholding-Guide_2023.pdf?rev=177ca3d8cb034392bf84e317ed78c320&hash=19C5F42F43A89A91863D229349F48883#:~:text=The%20exemption%20amount%20is%20%245%2C400
         employees%20Michigan%20income%20tax%20return"
          https://www.michigan.gov/taxes/questions/iit/accordion/residency/are-my-wages-earned-in-another-state-taxable-in-michigan-if-i-am-a-michigan-resident-1
          https://www.michigan.gov/taxes/-/media/Project/Websites/taxes/Forms/2022/2022-IIT-Forms/BOOK_MI-1040.pdf?rev=4524f1cbbc0243b194ab05a6680be75b&hash=40DFE71E651F7C1D65C31F6CD03D944F
          https://www.michigan.gov/taxes/questions/iit/accordion/general/what-cities-impose-an-income-tax
        
        Virginia:
        -         
        - https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/reciprocity
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        county_occupation = self.county_occupation if hasattr(self, 'county_occupation') else county_occupation
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Michigan', 'state_occupation': 'Virginia', "capital_gains_short": capital_gains_short, "county_occupation": county_occupation, "dependents": dependents, "sources": sources, "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def michigan_and_washington(self, county_occupation='Other', county_residence='Other', state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Michigan:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        county_occupation : str, optional
            County of occupation. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Washington:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Michigan:
        - https://www.michigan.gov/taxes/iit/new-developments-for-tax-year-2022
          "https://www.michigan.gov/taxes/-/media/Project/Websites/taxes/Forms/2023/2023-SUW/446_Withholding-Guide_2023.pdf?rev=177ca3d8cb034392bf84e317ed78c320&hash=19C5F42F43A89A91863D229349F48883#:~:text=The%20exemption%20amount%20is%20%245%2C400
         employees%20Michigan%20income%20tax%20return"
          "https://www.michigan.gov/taxes/-/media/Project/Websites/taxes/Forms/2023/2023-SUW/446_Withholding-Guide_2023.pdf?rev=177ca3d8cb034392bf84e317ed78c320&hash=19C5F42F43A89A91863D229349F48883#:~:text=The%20exemption%20amount%20is%20%245%2C400
         employees%20Michigan%20income%20tax%20return"
          https://www.michigan.gov/taxes/questions/iit/accordion/residency/are-my-wages-earned-in-another-state-taxable-in-michigan-if-i-am-a-michigan-resident-1
          https://www.michigan.gov/taxes/-/media/Project/Websites/taxes/Forms/2022/2022-IIT-Forms/BOOK_MI-1040.pdf?rev=4524f1cbbc0243b194ab05a6680be75b&hash=40DFE71E651F7C1D65C31F6CD03D944F
          https://www.michigan.gov/taxes/questions/iit/accordion/general/what-cities-impose-an-income-tax
        
        Washington:
        -         
        - https://dor.wa.gov/taxes-rates/income-tax
          https://dor.wa.gov/taxes-rates/other-taxes/capital-gains-tax
          https://dor.wa.gov/taxes-rates/income-tax
        """
        
        # Check if the parameters are set upstream in the instance
        county_occupation = self.county_occupation if hasattr(self, 'county_occupation') else county_occupation
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Michigan', 'state_occupation': 'Washington', "county_occupation": county_occupation, "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def michigan_and_west_virginia(self, pay_periods=24, capital_gains_short=0, county_occupation='Other', dependents=0, sources=False, county_residence='Other', state_agi=75000, capital_gains_long=0, filing_status='single', income=100000, credits_state=0):
        """
        --- Parameters ----
        
        Michigan:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        county_occupation : str, optional
            County of occupation. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        West Virginia:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_occupation : str, optional
            County of occupation. Default is Other.
        pay_periods : int, optional
            Number of pay periods. Options are 13, 24, 26, 52. Default is 24.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Michigan:
        - https://www.michigan.gov/taxes/iit/new-developments-for-tax-year-2022
          "https://www.michigan.gov/taxes/-/media/Project/Websites/taxes/Forms/2023/2023-SUW/446_Withholding-Guide_2023.pdf?rev=177ca3d8cb034392bf84e317ed78c320&hash=19C5F42F43A89A91863D229349F48883#:~:text=The%20exemption%20amount%20is%20%245%2C400
         employees%20Michigan%20income%20tax%20return"
          "https://www.michigan.gov/taxes/-/media/Project/Websites/taxes/Forms/2023/2023-SUW/446_Withholding-Guide_2023.pdf?rev=177ca3d8cb034392bf84e317ed78c320&hash=19C5F42F43A89A91863D229349F48883#:~:text=The%20exemption%20amount%20is%20%245%2C400
         employees%20Michigan%20income%20tax%20return"
          https://www.michigan.gov/taxes/questions/iit/accordion/residency/are-my-wages-earned-in-another-state-taxable-in-michigan-if-i-am-a-michigan-resident-1
          https://www.michigan.gov/taxes/-/media/Project/Websites/taxes/Forms/2022/2022-IIT-Forms/BOOK_MI-1040.pdf?rev=4524f1cbbc0243b194ab05a6680be75b&hash=40DFE71E651F7C1D65C31F6CD03D944F
          https://www.michigan.gov/taxes/questions/iit/accordion/general/what-cities-impose-an-income-tax
        
        West Virginia:
        -         
        - https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://tax.wv.gov/Documents/TSD/tsd438.pdf
          https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2011/Taxes-11-35att.html
        """
        
        # Check if the parameters are set upstream in the instance
        pay_periods = self.pay_periods if hasattr(self, 'pay_periods') else pay_periods
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        county_occupation = self.county_occupation if hasattr(self, 'county_occupation') else county_occupation
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Michigan', 'state_occupation': 'WestVirginia', "pay_periods": pay_periods, "capital_gains_short": capital_gains_short, "county_occupation": county_occupation, "dependents": dependents, "sources": sources, "county_residence": county_residence, "state_agi": state_agi, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def michigan_and_wisconsin(self, county_occupation='Other', county_residence='Other', state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Michigan:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        county_occupation : str, optional
            County of occupation. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Wisconsin:
        


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Michigan:
        - https://www.michigan.gov/taxes/iit/new-developments-for-tax-year-2022
          "https://www.michigan.gov/taxes/-/media/Project/Websites/taxes/Forms/2023/2023-SUW/446_Withholding-Guide_2023.pdf?rev=177ca3d8cb034392bf84e317ed78c320&hash=19C5F42F43A89A91863D229349F48883#:~:text=The%20exemption%20amount%20is%20%245%2C400
         employees%20Michigan%20income%20tax%20return"
          "https://www.michigan.gov/taxes/-/media/Project/Websites/taxes/Forms/2023/2023-SUW/446_Withholding-Guide_2023.pdf?rev=177ca3d8cb034392bf84e317ed78c320&hash=19C5F42F43A89A91863D229349F48883#:~:text=The%20exemption%20amount%20is%20%245%2C400
         employees%20Michigan%20income%20tax%20return"
          https://www.michigan.gov/taxes/questions/iit/accordion/residency/are-my-wages-earned-in-another-state-taxable-in-michigan-if-i-am-a-michigan-resident-1
          https://www.michigan.gov/taxes/-/media/Project/Websites/taxes/Forms/2022/2022-IIT-Forms/BOOK_MI-1040.pdf?rev=4524f1cbbc0243b194ab05a6680be75b&hash=40DFE71E651F7C1D65C31F6CD03D944F
          https://www.michigan.gov/taxes/questions/iit/accordion/general/what-cities-impose-an-income-tax
        
        Wisconsin:
        -         
        - https://www.revenue.wi.gov/TaxForms2023/2023-Form1-ES-Inst.pdf
          https://www.revenue.wi.gov/TaxForms2023/2023-Form1-ES-Inst.pdf
          https://www.revenue.wi.gov/TaxForms2023/2023-Form1-ES-Inst.pdf
          https://www.revenue.wi.gov/Pages/FAQS/pcs-work.aspx
          https://www.revenue.wi.gov/Dor%20Publications/pb103.pdf
          https://www.revenue.wi.gov/Pages/FAQS/pcs-work.aspx
        """
        
        # Check if the parameters are set upstream in the instance
        county_occupation = self.county_occupation if hasattr(self, 'county_occupation') else county_occupation
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Michigan', 'state_occupation': 'Wisconsin', "county_occupation": county_occupation, "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def michigan_and_wyoming(self, county_occupation='Other', county_residence='Other', state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Michigan:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        county_occupation : str, optional
            County of occupation. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Wyoming:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Michigan:
        - https://www.michigan.gov/taxes/iit/new-developments-for-tax-year-2022
          "https://www.michigan.gov/taxes/-/media/Project/Websites/taxes/Forms/2023/2023-SUW/446_Withholding-Guide_2023.pdf?rev=177ca3d8cb034392bf84e317ed78c320&hash=19C5F42F43A89A91863D229349F48883#:~:text=The%20exemption%20amount%20is%20%245%2C400
         employees%20Michigan%20income%20tax%20return"
          "https://www.michigan.gov/taxes/-/media/Project/Websites/taxes/Forms/2023/2023-SUW/446_Withholding-Guide_2023.pdf?rev=177ca3d8cb034392bf84e317ed78c320&hash=19C5F42F43A89A91863D229349F48883#:~:text=The%20exemption%20amount%20is%20%245%2C400
         employees%20Michigan%20income%20tax%20return"
          https://www.michigan.gov/taxes/questions/iit/accordion/residency/are-my-wages-earned-in-another-state-taxable-in-michigan-if-i-am-a-michigan-resident-1
          https://www.michigan.gov/taxes/-/media/Project/Websites/taxes/Forms/2022/2022-IIT-Forms/BOOK_MI-1040.pdf?rev=4524f1cbbc0243b194ab05a6680be75b&hash=40DFE71E651F7C1D65C31F6CD03D944F
          https://www.michigan.gov/taxes/questions/iit/accordion/general/what-cities-impose-an-income-tax
        
        Wyoming:
        -         
        - https://revenue.wyo.gov/
        """
        
        # Check if the parameters are set upstream in the instance
        county_occupation = self.county_occupation if hasattr(self, 'county_occupation') else county_occupation
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Michigan', 'state_occupation': 'Wyoming', "county_occupation": county_occupation, "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def minnesota_and_alabama(self, county_residence='Other', state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Minnesota:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Alabama:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Minnesota:
        - https://www.revenue.state.mn.us/minnesota-income-tax-rates-and-brackets
          https://www.revenue.state.mn.us/child-and-dependent-care-credit
          https://www.revenue.state.mn.us/dependent-exemptions
          https://www.revenue.state.mn.us/minnesota-standard-deduction
          https://mn.gov/mmb/assets/mwr_form_tcm1059-128581.pdf
          https://www.revenue.state.mn.us/individual-income-tax
          https://www.revenue.state.mn.us/individual-income-tax
        
        Alabama:
        -         
        - https://www.revenue.alabama.gov/ultraviewer/viewer/basic_viewer/index.html?form=2023/02/22f40.pdf
          https://www.revenue.alabama.gov/faqs/what-is-alabamas-individual-income-tax-rate/
          https://www.revenue.alabama.gov/ultraviewer/viewer/basic_viewer/index.html?form=2023/02/22f40.pdf
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-24.htm?taxmap=true
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-24.htm?taxmap=true
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-24.htm?taxmap=true
          https://www.revenue.alabama.gov/individual-corporate/income-to-be-reported-on-the-alabama-income-tax-return/
          https://www.almonline.org/TaxRates.aspx#OccupationalTax
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Minnesota', 'state_occupation': 'Alabama', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def minnesota_and_alaska(self, state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Minnesota:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Alaska:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Minnesota:
        - https://www.revenue.state.mn.us/minnesota-income-tax-rates-and-brackets
          https://www.revenue.state.mn.us/child-and-dependent-care-credit
          https://www.revenue.state.mn.us/dependent-exemptions
          https://www.revenue.state.mn.us/minnesota-standard-deduction
          https://mn.gov/mmb/assets/mwr_form_tcm1059-128581.pdf
          https://www.revenue.state.mn.us/individual-income-tax
          https://www.revenue.state.mn.us/individual-income-tax
        
        Alaska:
        -         
        - https://www.commerce.alaska.gov/web/dcra/officeofthestateassessor/alaskataxfacts.aspx
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Minnesota', 'state_occupation': 'Alaska', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def minnesota_and_arizona(self, state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Minnesota:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Arizona:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        credits_state : int, optional
            Credits for state taxes. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Minnesota:
        - https://www.revenue.state.mn.us/minnesota-income-tax-rates-and-brackets
          https://www.revenue.state.mn.us/child-and-dependent-care-credit
          https://www.revenue.state.mn.us/dependent-exemptions
          https://www.revenue.state.mn.us/minnesota-standard-deduction
          https://mn.gov/mmb/assets/mwr_form_tcm1059-128581.pdf
          https://www.revenue.state.mn.us/individual-income-tax
          https://www.revenue.state.mn.us/individual-income-tax
        
        Arizona:
        -         
        - https://azdor.gov/forms/individual/form-140-x-y-tables
          https://azdor.gov/forms/individual/form-140-x-y-tables
          https://azdor.gov/forms/individual/form-140-x-y-tables
          https://azdor.gov/businesses-arizona/withholding-tax/withholding-exceptions
          "https://azdor.gov/individuals/income-tax-filing-assistance/identifying-other-taxable-income#:~:text=In%20regards%20to%20capital%20gains
         the%20individuals%20regular%20tax%20rate.
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Minnesota', 'state_occupation': 'Arizona', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def minnesota_and_arkansas(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Minnesota:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Arkansas:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Minnesota:
        - https://www.revenue.state.mn.us/minnesota-income-tax-rates-and-brackets
          https://www.revenue.state.mn.us/child-and-dependent-care-credit
          https://www.revenue.state.mn.us/dependent-exemptions
          https://www.revenue.state.mn.us/minnesota-standard-deduction
          https://mn.gov/mmb/assets/mwr_form_tcm1059-128581.pdf
          https://www.revenue.state.mn.us/individual-income-tax
          https://www.revenue.state.mn.us/individual-income-tax
        
        Arkansas:
        -         
        - https://www.arkansasedc.com/why-arkansas/business-climate/tax-structure/personal-income-tax
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_and_AR1000NR_Instructions.pdf
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_FullYearResidentIndividualIncomeTaxReturn.pdf
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_and_AR1000NR_Instructions.pdf
          https://www.arkansasedc.com/why-arkansas/business-climate/tax-structure/capital-gains-tax-reduction
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_and_AR1000NR_Instructions.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Minnesota', 'state_occupation': 'Arkansas', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def minnesota_and_california(self, state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Minnesota:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        California:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Minnesota:
        - https://www.revenue.state.mn.us/minnesota-income-tax-rates-and-brackets
          https://www.revenue.state.mn.us/child-and-dependent-care-credit
          https://www.revenue.state.mn.us/dependent-exemptions
          https://www.revenue.state.mn.us/minnesota-standard-deduction
          https://mn.gov/mmb/assets/mwr_form_tcm1059-128581.pdf
          https://www.revenue.state.mn.us/individual-income-tax
          https://www.revenue.state.mn.us/individual-income-tax
        
        California:
        -         
        - https://www.ftb.ca.gov/forms/2022/2022-540-tax-rate-schedules.pdf
          https://www.ftb.ca.gov/file/personal/deductions/index.html
          https://www.ftb.ca.gov/about-ftb/newsroom/caleitc/eligibility-and-credit-information.html
          https://www.ftb.ca.gov/file/personal/income-types/capital-gains-and-losses.html#:~:text=California%20does%20not%20have%20a
         are%20taxed%20as%20ordinary%20income
          https://www.ftb.ca.gov/file/personal/
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Minnesota', 'state_occupation': 'California', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def minnesota_and_colorado(self, county_residence='Other', state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Minnesota:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Colorado:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Minnesota:
        - https://www.revenue.state.mn.us/minnesota-income-tax-rates-and-brackets
          https://www.revenue.state.mn.us/child-and-dependent-care-credit
          https://www.revenue.state.mn.us/dependent-exemptions
          https://www.revenue.state.mn.us/minnesota-standard-deduction
          https://mn.gov/mmb/assets/mwr_form_tcm1059-128581.pdf
          https://www.revenue.state.mn.us/individual-income-tax
          https://www.revenue.state.mn.us/individual-income-tax
        
        Colorado:
        -         
        - https://leg.colorado.gov/agencies/legislative-council-staff/individual-income-tax%C2%A0
          https://leg.colorado.gov/agencies/legislative-council-staff/individual-income-tax%C2%A0
          https://tax.colorado.gov/sites/tax/files/documents/ITT_Capital_Gain_Subtraction_Feb_2022.pdf
          https://www.auroragov.org/business_services/taxes/occupational_privilege_tax
          https://www.denvergov.org/Government/Agencies-Departments-Offices/Agencies-Departments-Offices-Directory/Department-of-Finance/Our-Divisions/Treasury/Business-Tax-Information#:~:text=Occupational%20privilege%20tax
         -%28show%20below%29&text=Each%20taxable%20employee%20is%20liable
         month%20for%20each%20taxable%20employee
          https://www.glendale.co.us/355/Occupational-Privilege-Tax
          https://www.greenwoodvillage.com/1220/Occupational-Privilege-Tax-OPT
          https://ci.sheridan.co.us/288/Occupational-Privilege-Tax
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Minnesota', 'state_occupation': 'Colorado', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def minnesota_and_connecticut(self, state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Minnesota:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Connecticut:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Minnesota:
        - https://www.revenue.state.mn.us/minnesota-income-tax-rates-and-brackets
          https://www.revenue.state.mn.us/child-and-dependent-care-credit
          https://www.revenue.state.mn.us/dependent-exemptions
          https://www.revenue.state.mn.us/minnesota-standard-deduction
          https://mn.gov/mmb/assets/mwr_form_tcm1059-128581.pdf
          https://www.revenue.state.mn.us/individual-income-tax
          https://www.revenue.state.mn.us/individual-income-tax
        
        Connecticut:
        -         
        - https://portal.ct.gov/-/media/DRS/Forms/2022/Income/2022-CT-1040-Instructions_1222.pdf
          https://www.cga.ct.gov/2022/rpt/pdf/2022-R-0108.pdf
          https://www.cga.ct.gov/current/pub/chap_229.htm#sec_12-703
          https://portal.ct.gov/-/media/DRS/Forms/2022/Income/2022-CT-1040-Instructions_1222.pdf
          "https://portal.ct.gov/DRS/Publications/TSSNs/TSSN-29#:~:text=An%20individuals%20net%20capital%20gains
         portion%20of%20Social%20Security%20benefits"
          "https://portal.ct.gov/DRS/Publications/TSSNs/TSSN-29#:~:text=An%20individuals%20net%20capital%20gains
         portion%20of%20Social%20Security%20benefits"
          https://portal.ct.gov/DRS/Individuals/Resident-Income-Tax/Tax-Information
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Minnesota', 'state_occupation': 'Connecticut', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def minnesota_and_delaware(self, state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, federal_agi=75000, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Minnesota:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Delaware:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        federal_agi : int, optional
            Adjusted Gross Income (AGI) for federal taxes. Default is 0. Sources - https://www.irs.gov/e-file-providers/definition-of-adjusted-gross-income, https://www.irs.gov/newsroom/taxpayers-should-know-the-difference-between-standard-and-itemized-deductions
        income : int, optional
            Gross income. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Minnesota:
        - https://www.revenue.state.mn.us/minnesota-income-tax-rates-and-brackets
          https://www.revenue.state.mn.us/child-and-dependent-care-credit
          https://www.revenue.state.mn.us/dependent-exemptions
          https://www.revenue.state.mn.us/minnesota-standard-deduction
          https://mn.gov/mmb/assets/mwr_form_tcm1059-128581.pdf
          https://www.revenue.state.mn.us/individual-income-tax
          https://www.revenue.state.mn.us/individual-income-tax
        
        Delaware:
        -         
        - https://revenue.delaware.gov/software-developer/tax-rate-changes/
          https://revenuefiles.delaware.gov/2022/PIT-RES_TY22_2022-01_PaperInteractive.pdf
          https://revenuefiles.delaware.gov/2022/PIT-RES_TY22_2022-02_Instructions.pdf
          https://revenue.delaware.gov/frequently-asked-questions/personal-income-tax-faqs/
          https://www.newcastlede.gov/DocumentCenter/View/20147/City-of-Wilmington-Wage-Tax-Information-PDF
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        federal_agi = self.federal_agi if hasattr(self, 'federal_agi') else federal_agi
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Minnesota', 'state_occupation': 'Delaware', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "federal_agi": federal_agi, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def minnesota_and_florida(self, state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Minnesota:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Florida:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Minnesota:
        - https://www.revenue.state.mn.us/minnesota-income-tax-rates-and-brackets
          https://www.revenue.state.mn.us/child-and-dependent-care-credit
          https://www.revenue.state.mn.us/dependent-exemptions
          https://www.revenue.state.mn.us/minnesota-standard-deduction
          https://mn.gov/mmb/assets/mwr_form_tcm1059-128581.pdf
          https://www.revenue.state.mn.us/individual-income-tax
          https://www.revenue.state.mn.us/individual-income-tax
        
        Florida:
        -         
        - https://floridarevenue.com/taxes/taxesfees/Pages/tax_interest_rates.aspx
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Minnesota', 'state_occupation': 'Florida', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def minnesota_and_georgia(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Minnesota:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Georgia:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Minnesota:
        - https://www.revenue.state.mn.us/minnesota-income-tax-rates-and-brackets
          https://www.revenue.state.mn.us/child-and-dependent-care-credit
          https://www.revenue.state.mn.us/dependent-exemptions
          https://www.revenue.state.mn.us/minnesota-standard-deduction
          https://mn.gov/mmb/assets/mwr_form_tcm1059-128581.pdf
          https://www.revenue.state.mn.us/individual-income-tax
          https://www.revenue.state.mn.us/individual-income-tax
        
        Georgia:
        -         
        - https://dor.georgia.gov/tax-tables-georgia-tax-rate-schedule
          https://apps.dor.ga.gov/FillableForms/PDFViewer/Index?form=2022GA500
          https://apps.dor.ga.gov/FillableForms/PDFViewer/Index?form=2022GA500
          https://apps.dor.ga.gov/FillableForms/PDFViewer/Index?form=2022GA500
          https://dor.georgia.gov/years-individual-income-tax-forms
          https://dor.georgia.gov/years-individual-income-tax-forms
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Minnesota', 'state_occupation': 'Georgia', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def minnesota_and_hawaii(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Minnesota:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Hawaii:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Minnesota:
        - https://www.revenue.state.mn.us/minnesota-income-tax-rates-and-brackets
          https://www.revenue.state.mn.us/child-and-dependent-care-credit
          https://www.revenue.state.mn.us/dependent-exemptions
          https://www.revenue.state.mn.us/minnesota-standard-deduction
          https://mn.gov/mmb/assets/mwr_form_tcm1059-128581.pdf
          https://www.revenue.state.mn.us/individual-income-tax
          https://www.revenue.state.mn.us/individual-income-tax
        
        Hawaii:
        -         
        - https://tax.hawaii.gov/forms/d_18table-on/d_18table-on_p13/
          https://files.hawaii.gov/tax/forms/2018/18table-on.pdf
          https://files.hawaii.gov/tax/news/pubs/22outline.pdf
          https://files.hawaii.gov/tax/news/pubs/FastTaxRefGuide_2017.pdf
          https://www.tfhawaii.org/wordpress/state-tax-resources/mini-tax-guides/individual-income-tax-chapter-235/#:~:text=Long%20term%20capital%20gains%20are%20taxed%20at%20a%20maximum%20of%207.25%25
          https://www.grassrootinstitute.org/2023/02/hb337-increasing-tax-on-capital-gains-a-bad-risk/
          https://files.hawaii.gov/tax/news/pubs/22outline.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Minnesota', 'state_occupation': 'Hawaii', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def minnesota_and_idaho(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Minnesota:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Idaho:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Minnesota:
        - https://www.revenue.state.mn.us/minnesota-income-tax-rates-and-brackets
          https://www.revenue.state.mn.us/child-and-dependent-care-credit
          https://www.revenue.state.mn.us/dependent-exemptions
          https://www.revenue.state.mn.us/minnesota-standard-deduction
          https://mn.gov/mmb/assets/mwr_form_tcm1059-128581.pdf
          https://www.revenue.state.mn.us/individual-income-tax
          https://www.revenue.state.mn.us/individual-income-tax
        
        Idaho:
        -         
        - https://tax.idaho.gov/taxes/income-tax/individual-income/individual-income-tax-rate-schedule/
          https://tax.idaho.gov/wp-content/uploads/forms/EIN00046/EIN00046_03-01-2023.pdf
          https://tax.idaho.gov/taxes/income-tax/individual-income/idaho-source-income/
          https://tax.idaho.gov/taxes/income-tax/individual-income/idaho-source-income/
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Minnesota', 'state_occupation': 'Idaho', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def minnesota_and_illinois(self, state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Minnesota:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Illinois:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Minnesota:
        - https://www.revenue.state.mn.us/minnesota-income-tax-rates-and-brackets
          https://www.revenue.state.mn.us/child-and-dependent-care-credit
          https://www.revenue.state.mn.us/dependent-exemptions
          https://www.revenue.state.mn.us/minnesota-standard-deduction
          https://mn.gov/mmb/assets/mwr_form_tcm1059-128581.pdf
          https://www.revenue.state.mn.us/individual-income-tax
          https://www.revenue.state.mn.us/individual-income-tax
        
        Illinois:
        -         
        - https://tax.illinois.gov/research/taxrates/income.html#IndividualIncome
          https://tax.illinois.gov/content/dam/soi/en/web/tax/forms/incometax/documents/currentyear/individual/il-1040-instr.pdf
          https://tax.illinois.gov/questionsandanswers/12.html
          https://tax.illinois.gov/content/dam/soi/en/web/tax/forms/incometax/documents/currentyear/individual/il-1040-instr.pdf
          https://tax.illinois.gov/research/taxrates/personalproperty.html
          https://tax.illinois.gov/localgovernments/income.html
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Minnesota', 'state_occupation': 'Illinois', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def minnesota_and_indiana(self, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Minnesota:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Indiana:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Minnesota:
        - https://www.revenue.state.mn.us/minnesota-income-tax-rates-and-brackets
          https://www.revenue.state.mn.us/child-and-dependent-care-credit
          https://www.revenue.state.mn.us/dependent-exemptions
          https://www.revenue.state.mn.us/minnesota-standard-deduction
          https://mn.gov/mmb/assets/mwr_form_tcm1059-128581.pdf
          https://www.revenue.state.mn.us/individual-income-tax
          https://www.revenue.state.mn.us/individual-income-tax
        
        Indiana:
        -         
        - https://www.in.gov/dor/tax-forms/2022-individual-income-tax-forms/
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-32.htm
          https://www.in.gov/dor/files/ib28.pdf
          https://www.in.gov/dor/tax-forms/2022-individual-income-tax-forms/
          https://www.in.gov/dor/files/2022-county-tax-rates-and-codes.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Minnesota', 'state_occupation': 'Indiana', "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def minnesota_and_iowa(self, county_residence='Other', state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Minnesota:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Iowa:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Minnesota:
        - https://www.revenue.state.mn.us/minnesota-income-tax-rates-and-brackets
          https://www.revenue.state.mn.us/child-and-dependent-care-credit
          https://www.revenue.state.mn.us/dependent-exemptions
          https://www.revenue.state.mn.us/minnesota-standard-deduction
          https://mn.gov/mmb/assets/mwr_form_tcm1059-128581.pdf
          https://www.revenue.state.mn.us/individual-income-tax
          https://www.revenue.state.mn.us/individual-income-tax
        
        Iowa:
        -         
        - https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
          https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
          https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
          https://tax.iowa.gov/iowa-illinois-reciprocal-agreement
          https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Minnesota', 'state_occupation': 'Iowa', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def minnesota_and_kansas(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Minnesota:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Kansas:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Minnesota:
        - https://www.revenue.state.mn.us/minnesota-income-tax-rates-and-brackets
          https://www.revenue.state.mn.us/child-and-dependent-care-credit
          https://www.revenue.state.mn.us/dependent-exemptions
          https://www.revenue.state.mn.us/minnesota-standard-deduction
          https://mn.gov/mmb/assets/mwr_form_tcm1059-128581.pdf
          https://www.revenue.state.mn.us/individual-income-tax
          https://www.revenue.state.mn.us/individual-income-tax
        
        Kansas:
        -         
        - https://www.ksrevenue.gov/taxrates.html
          https://www.ksrevenue.gov/incomebook22.html#0
          https://www.ksrevenue.gov/incomebook22.html#0
          https://www.ksrevenue.gov/incomebook22.html#0
          https://www.ksrevenue.gov/incomebook22.html
          https://www.ksrevenue.gov/incomebook22.html
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Minnesota', 'state_occupation': 'Kansas', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def minnesota_and_kentucky(self, county_residence='Other', state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Minnesota:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Kentucky:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Minnesota:
        - https://www.revenue.state.mn.us/minnesota-income-tax-rates-and-brackets
          https://www.revenue.state.mn.us/child-and-dependent-care-credit
          https://www.revenue.state.mn.us/dependent-exemptions
          https://www.revenue.state.mn.us/minnesota-standard-deduction
          https://mn.gov/mmb/assets/mwr_form_tcm1059-128581.pdf
          https://www.revenue.state.mn.us/individual-income-tax
          https://www.revenue.state.mn.us/individual-income-tax
        
        Kentucky:
        -         
        - https://revenue.ky.gov/News/Pages/Dor-Announces-Updates-to-Individual-Income-Tax-for-2023-Tax-Year.aspx
          https://revenue.ky.gov/Individual/Individual-Income-Tax/Pages/default.aspx
          https://revenue.ky.gov/News/Pages/Dor-Announces-Updates-to-Individual-Income-Tax-for-2023-Tax-Year.aspx
          https://revenue.ky.gov/Individual/Individual-Income-Tax/Pages/default.aspx
          https://revenue.ky.gov/Dor%20Training%20Materials/103%20KAR%2017.140.%20Individual%20income%20tax%20-%20reciprocity%20-%20nonresidents.pdf
          https://revenue.ky.gov/Individual/Individual-Income-Tax/Pages/default.aspx
          https://nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2021/Taxes-21-32.htm?taxmap=true
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Minnesota', 'state_occupation': 'Kentucky', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def minnesota_and_louisiana(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Minnesota:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Louisiana:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Minnesota:
        - https://www.revenue.state.mn.us/minnesota-income-tax-rates-and-brackets
          https://www.revenue.state.mn.us/child-and-dependent-care-credit
          https://www.revenue.state.mn.us/dependent-exemptions
          https://www.revenue.state.mn.us/minnesota-standard-deduction
          https://mn.gov/mmb/assets/mwr_form_tcm1059-128581.pdf
          https://www.revenue.state.mn.us/individual-income-tax
          https://www.revenue.state.mn.us/individual-income-tax
        
        Louisiana:
        -         
        - https://revenue.louisiana.gov/individualincometax
          https://revenue.louisiana.gov/individualincometax
          https://www.revenue.louisiana.gov/taxforms/6935(11_02)F.pdf
          https://revenue.louisiana.gov/individualincometax
          https://revenue.louisiana.gov/individualincometax
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Minnesota', 'state_occupation': 'Louisiana', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def minnesota_and_maine(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Minnesota:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Maine:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Minnesota:
        - https://www.revenue.state.mn.us/minnesota-income-tax-rates-and-brackets
          https://www.revenue.state.mn.us/child-and-dependent-care-credit
          https://www.revenue.state.mn.us/dependent-exemptions
          https://www.revenue.state.mn.us/minnesota-standard-deduction
          https://mn.gov/mmb/assets/mwr_form_tcm1059-128581.pdf
          https://www.revenue.state.mn.us/individual-income-tax
          https://www.revenue.state.mn.us/individual-income-tax
        
        Maine:
        -         
        - https://www.maine.gov/revenue/tax-return-forms/individual-income-tax-2022
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/ind_tax_rate_sched_2022.pdf
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/22_1040me_book_gen_instruc_revisedFeb23.pdf
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/22_1040me_book_gen_instruc_revisedFeb23.pdf
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/22_1040me_book_gen_instruc_revisedFeb23.pdf
          https://www.maine.gov/revenue/tax-return-forms/individual-income-tax-2022
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Minnesota', 'state_occupation': 'Maine', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def minnesota_and_maryland(self, county_residence='Other', state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Minnesota:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Maryland:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Minnesota:
        - https://www.revenue.state.mn.us/minnesota-income-tax-rates-and-brackets
          https://www.revenue.state.mn.us/child-and-dependent-care-credit
          https://www.revenue.state.mn.us/dependent-exemptions
          https://www.revenue.state.mn.us/minnesota-standard-deduction
          https://mn.gov/mmb/assets/mwr_form_tcm1059-128581.pdf
          https://www.revenue.state.mn.us/individual-income-tax
          https://www.revenue.state.mn.us/individual-income-tax
        
        Maryland:
        -         
        - https://www.marylandtaxes.gov/individual/income/tax-info/tax-rates.php
          https://www.marylandtaxes.gov/forms/22_forms/Resident_Booklet.pdf
          https://www.marylandtaxes.gov/forms/Tax_Publications/Administrative_Releases/Income_and_Estate_Tax_Releases/ar_it37.pdf
          https://www.marylandtaxes.gov/individual/index.php
          https://www.marylandtaxes.gov/individual/income/tax-info/tax-rates.php
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Minnesota', 'state_occupation': 'Maryland', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def minnesota_and_massachusetts(self, state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Minnesota:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Massachusetts:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Minnesota:
        - https://www.revenue.state.mn.us/minnesota-income-tax-rates-and-brackets
          https://www.revenue.state.mn.us/child-and-dependent-care-credit
          https://www.revenue.state.mn.us/dependent-exemptions
          https://www.revenue.state.mn.us/minnesota-standard-deduction
          https://mn.gov/mmb/assets/mwr_form_tcm1059-128581.pdf
          https://www.revenue.state.mn.us/individual-income-tax
          https://www.revenue.state.mn.us/individual-income-tax
        
        Massachusetts:
        -         
        - https://www.mass.gov/service-details/massachusetts-tax-rates
          https://www.mass.gov/service-details/view-massachusetts-personal-income-tax-exemptions
          https://www.mass.gov/service-details/view-massachusetts-personal-income-tax-exemptions
          https://www.mass.gov/service-details/view-child-and-dependent-related-credits
          https://www.mass.gov/service-details/massachusetts-tax-rates
          https://www.mass.gov/personal-income-tax
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Minnesota', 'state_occupation': 'Massachusetts', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def minnesota_and_michigan(self, state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Minnesota:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Michigan:
        


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Minnesota:
        - https://www.revenue.state.mn.us/minnesota-income-tax-rates-and-brackets
          https://www.revenue.state.mn.us/child-and-dependent-care-credit
          https://www.revenue.state.mn.us/dependent-exemptions
          https://www.revenue.state.mn.us/minnesota-standard-deduction
          https://mn.gov/mmb/assets/mwr_form_tcm1059-128581.pdf
          https://www.revenue.state.mn.us/individual-income-tax
          https://www.revenue.state.mn.us/individual-income-tax
        
        Michigan:
        -         
        - https://www.michigan.gov/taxes/iit/new-developments-for-tax-year-2022
          "https://www.michigan.gov/taxes/-/media/Project/Websites/taxes/Forms/2023/2023-SUW/446_Withholding-Guide_2023.pdf?rev=177ca3d8cb034392bf84e317ed78c320&hash=19C5F42F43A89A91863D229349F48883#:~:text=The%20exemption%20amount%20is%20%245%2C400
         employees%20Michigan%20income%20tax%20return"
          "https://www.michigan.gov/taxes/-/media/Project/Websites/taxes/Forms/2023/2023-SUW/446_Withholding-Guide_2023.pdf?rev=177ca3d8cb034392bf84e317ed78c320&hash=19C5F42F43A89A91863D229349F48883#:~:text=The%20exemption%20amount%20is%20%245%2C400
         employees%20Michigan%20income%20tax%20return"
          https://www.michigan.gov/taxes/questions/iit/accordion/residency/are-my-wages-earned-in-another-state-taxable-in-michigan-if-i-am-a-michigan-resident-1
          https://www.michigan.gov/taxes/-/media/Project/Websites/taxes/Forms/2022/2022-IIT-Forms/BOOK_MI-1040.pdf?rev=4524f1cbbc0243b194ab05a6680be75b&hash=40DFE71E651F7C1D65C31F6CD03D944F
          https://www.michigan.gov/taxes/questions/iit/accordion/general/what-cities-impose-an-income-tax
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Minnesota', 'state_occupation': 'Michigan', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def minnesota_and_mississippi(self, state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Minnesota:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Mississippi:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Minnesota:
        - https://www.revenue.state.mn.us/minnesota-income-tax-rates-and-brackets
          https://www.revenue.state.mn.us/child-and-dependent-care-credit
          https://www.revenue.state.mn.us/dependent-exemptions
          https://www.revenue.state.mn.us/minnesota-standard-deduction
          https://mn.gov/mmb/assets/mwr_form_tcm1059-128581.pdf
          https://www.revenue.state.mn.us/individual-income-tax
          https://www.revenue.state.mn.us/individual-income-tax
        
        Mississippi:
        -         
        - https://www.dor.ms.gov/individual/tax-rates
          https://www.dor.ms.gov/individual/tax-rates
          https://www.dor.ms.gov/individual/tax-rates
          https://www.dor.ms.gov/individual
          https://www.dor.ms.gov/individual
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Minnesota', 'state_occupation': 'Mississippi', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def minnesota_and_missouri(self, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Minnesota:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Missouri:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Minnesota:
        - https://www.revenue.state.mn.us/minnesota-income-tax-rates-and-brackets
          https://www.revenue.state.mn.us/child-and-dependent-care-credit
          https://www.revenue.state.mn.us/dependent-exemptions
          https://www.revenue.state.mn.us/minnesota-standard-deduction
          https://mn.gov/mmb/assets/mwr_form_tcm1059-128581.pdf
          https://www.revenue.state.mn.us/individual-income-tax
          https://www.revenue.state.mn.us/individual-income-tax
        
        Missouri:
        -         
        - https://dor.mo.gov/taxation/individual/tax-types/income/year-changes/
          https://dor.mo.gov/taxation/individual/tax-types/income/year-changes/
          https://dor.mo.gov/taxation/individual/tax-types/income/year-changes/
          https://www.kcmo.gov/city-hall/departments/finance/earnings-tax
          https://www.stlouis-mo.gov/government/departments/collector/earnings-tax/individual-earnings-tax-info.cfm
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Minnesota', 'state_occupation': 'Missouri', "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def minnesota_and_montana(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Minnesota:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Montana:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Minnesota:
        - https://www.revenue.state.mn.us/minnesota-income-tax-rates-and-brackets
          https://www.revenue.state.mn.us/child-and-dependent-care-credit
          https://www.revenue.state.mn.us/dependent-exemptions
          https://www.revenue.state.mn.us/minnesota-standard-deduction
          https://mn.gov/mmb/assets/mwr_form_tcm1059-128581.pdf
          https://www.revenue.state.mn.us/individual-income-tax
          https://www.revenue.state.mn.us/individual-income-tax
        
        Montana:
        -         
        - https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/2022/12/montana-employees-withholding-allowance-and-exemption-certificate-form-mw-4.pdf
          https://mtrevenue.gov/2021/11/29/simplification-of-montana-income-taxation/?utm_source=rss&utm_medium=rss&utm_campaign=simplification-of-montana-income-taxation
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Minnesota', 'state_occupation': 'Montana', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def minnesota_and_nebraska(self, state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Minnesota:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Nebraska:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Minnesota:
        - https://www.revenue.state.mn.us/minnesota-income-tax-rates-and-brackets
          https://www.revenue.state.mn.us/child-and-dependent-care-credit
          https://www.revenue.state.mn.us/dependent-exemptions
          https://www.revenue.state.mn.us/minnesota-standard-deduction
          https://mn.gov/mmb/assets/mwr_form_tcm1059-128581.pdf
          https://www.revenue.state.mn.us/individual-income-tax
          https://www.revenue.state.mn.us/individual-income-tax
        
        Nebraska:
        -         
        - https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2023_Circular_EN_8-429r2023_Whole_Book_11.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Minnesota', 'state_occupation': 'Nebraska', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def minnesota_and_nevada(self, state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Minnesota:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Nevada:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Minnesota:
        - https://www.revenue.state.mn.us/minnesota-income-tax-rates-and-brackets
          https://www.revenue.state.mn.us/child-and-dependent-care-credit
          https://www.revenue.state.mn.us/dependent-exemptions
          https://www.revenue.state.mn.us/minnesota-standard-deduction
          https://mn.gov/mmb/assets/mwr_form_tcm1059-128581.pdf
          https://www.revenue.state.mn.us/individual-income-tax
          https://www.revenue.state.mn.us/individual-income-tax
        
        Nevada:
        -         
        - https://www.nvsos.gov/sos/businesses/the-nevada-advantage
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Minnesota', 'state_occupation': 'Nevada', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def minnesota_and_new_hampshire(self, state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Minnesota:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        New Hampshire:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Minnesota:
        - https://www.revenue.state.mn.us/minnesota-income-tax-rates-and-brackets
          https://www.revenue.state.mn.us/child-and-dependent-care-credit
          https://www.revenue.state.mn.us/dependent-exemptions
          https://www.revenue.state.mn.us/minnesota-standard-deduction
          https://mn.gov/mmb/assets/mwr_form_tcm1059-128581.pdf
          https://www.revenue.state.mn.us/individual-income-tax
          https://www.revenue.state.mn.us/individual-income-tax
        
        New Hampshire:
        -         
        - https://www.revenue.nh.gov/assistance/tax-overview.htm
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Minnesota', 'state_occupation': 'NewHampshire', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def minnesota_and_new_jersey(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Minnesota:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        New Jersey:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Minnesota:
        - https://www.revenue.state.mn.us/minnesota-income-tax-rates-and-brackets
          https://www.revenue.state.mn.us/child-and-dependent-care-credit
          https://www.revenue.state.mn.us/dependent-exemptions
          https://www.revenue.state.mn.us/minnesota-standard-deduction
          https://mn.gov/mmb/assets/mwr_form_tcm1059-128581.pdf
          https://www.revenue.state.mn.us/individual-income-tax
          https://www.revenue.state.mn.us/individual-income-tax
        
        New Jersey:
        -         
        - https://www.state.nj.us/treasury/taxation/taxtables.shtml
          https://www.state.nj.us/treasury/taxation/njit13.shtml
          https://www.nj.gov/treasury/taxation/njit25.shtml
          https://www.state.nj.us/treasury/taxation/individuals/freqqiti.shtml
          https://www.state.nj.us/treasury/taxation/localtax.shtml
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Minnesota', 'state_occupation': 'NewJersey', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def minnesota_and_new_mexico(self, state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Minnesota:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        New Mexico:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Minnesota:
        - https://www.revenue.state.mn.us/minnesota-income-tax-rates-and-brackets
          https://www.revenue.state.mn.us/child-and-dependent-care-credit
          https://www.revenue.state.mn.us/dependent-exemptions
          https://www.revenue.state.mn.us/minnesota-standard-deduction
          https://mn.gov/mmb/assets/mwr_form_tcm1059-128581.pdf
          https://www.revenue.state.mn.us/individual-income-tax
          https://www.revenue.state.mn.us/individual-income-tax
        
        New Mexico:
        -         
        - https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/fdf3c548-8aba-4b9c-9eb4-bb564c716015/FYI-104.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/0bf943fd-652e-4400-bb1b-cd8397eb5a95/2022pit-adj.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/2f1a6781-9534-4436-b427-1557f9592099/2022pit-adj-ins.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/0bf943fd-652e-4400-bb1b-cd8397eb5a95/2022pit-adj.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/1afc56af-ea90-4d48-82e5-1f9aeb43255a/PITbook2022.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/fdf3c548-8aba-4b9c-9eb4-bb564c716015/FYI-104.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Minnesota', 'state_occupation': 'NewMexico', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def minnesota_and_new_york(self, county_residence='Other', state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Minnesota:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        New York:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Minnesota:
        - https://www.revenue.state.mn.us/minnesota-income-tax-rates-and-brackets
          https://www.revenue.state.mn.us/child-and-dependent-care-credit
          https://www.revenue.state.mn.us/dependent-exemptions
          https://www.revenue.state.mn.us/minnesota-standard-deduction
          https://mn.gov/mmb/assets/mwr_form_tcm1059-128581.pdf
          https://www.revenue.state.mn.us/individual-income-tax
          https://www.revenue.state.mn.us/individual-income-tax
        
        New York:
        -         
        - https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nys-tax-rate-schedule
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nys-tax-rate-schedule
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nys-tax-rate-schedule
          https://www.tax.ny.gov/pit/
          https://www.tax.ny.gov/pit/file/nonresident-faqs.htm
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#yonkers-tax-rate-schedule
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nyc-tax-rate-schedule
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Minnesota', 'state_occupation': 'NewYork', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def minnesota_and_north_carolina(self, state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Minnesota:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        North Carolina:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Minnesota:
        - https://www.revenue.state.mn.us/minnesota-income-tax-rates-and-brackets
          https://www.revenue.state.mn.us/child-and-dependent-care-credit
          https://www.revenue.state.mn.us/dependent-exemptions
          https://www.revenue.state.mn.us/minnesota-standard-deduction
          https://mn.gov/mmb/assets/mwr_form_tcm1059-128581.pdf
          https://www.revenue.state.mn.us/individual-income-tax
          https://www.revenue.state.mn.us/individual-income-tax
        
        North Carolina:
        -         
        - https://www.ncdor.gov/taxes-forms/tax-rate-schedules
          https://www.ncdor.gov/taxes-forms/individual-income-tax/north-carolina-standard-deduction-or-north-carolina-itemized-deductions
          https://www.ncdor.gov/taxes-forms/individual-income-tax
          https://www.ncdor.gov/taxes-forms/individual-income-tax
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Minnesota', 'state_occupation': 'NorthCarolina', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def minnesota_and_north_dakota(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Minnesota:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        North Dakota:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Minnesota:
        - https://www.revenue.state.mn.us/minnesota-income-tax-rates-and-brackets
          https://www.revenue.state.mn.us/child-and-dependent-care-credit
          https://www.revenue.state.mn.us/dependent-exemptions
          https://www.revenue.state.mn.us/minnesota-standard-deduction
          https://mn.gov/mmb/assets/mwr_form_tcm1059-128581.pdf
          https://www.revenue.state.mn.us/individual-income-tax
          https://www.revenue.state.mn.us/individual-income-tax
        
        North Dakota:
        -         
        - https://www.tax.nd.gov/forms
          https://www.tax.nd.gov/sites/www/files/documents/forms/individual/2022-iit/2022-individual-income-tax-booklet.pdf
          https://www.tax.nd.gov/sites/www/files/documents/forms/2022-individual-income-tax-booklet.pdf
          https://www.tax.nd.gov/tax-exemptions-credits/income-tax-exemptions-credits
          https://www.tax.nd.gov/sites/www/files/documents/forms/business/ndwrfillable.pdf
          https://www.tax.nd.gov/news/tax-legislative-changes/significant-changes-law/individual-income-tax-history
          https://www.tax.nd.gov/forms
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Minnesota', 'state_occupation': 'NorthDakota', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def minnesota_and_ohio(self, county_residence='Other', state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Minnesota:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Ohio:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Minnesota:
        - https://www.revenue.state.mn.us/minnesota-income-tax-rates-and-brackets
          https://www.revenue.state.mn.us/child-and-dependent-care-credit
          https://www.revenue.state.mn.us/dependent-exemptions
          https://www.revenue.state.mn.us/minnesota-standard-deduction
          https://mn.gov/mmb/assets/mwr_form_tcm1059-128581.pdf
          https://www.revenue.state.mn.us/individual-income-tax
          https://www.revenue.state.mn.us/individual-income-tax
        
        Ohio:
        -         
        - https://tax.ohio.gov/individual/resources/annual-tax-rates
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://tax.ohio.gov/static/forms/employer_withholding/generic/wth_it4nr.pdf
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://www.ritaohio.com/TaxRatesTable
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Minnesota', 'state_occupation': 'Ohio', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def minnesota_and_oklahoma(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Minnesota:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Oklahoma:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Minnesota:
        - https://www.revenue.state.mn.us/minnesota-income-tax-rates-and-brackets
          https://www.revenue.state.mn.us/child-and-dependent-care-credit
          https://www.revenue.state.mn.us/dependent-exemptions
          https://www.revenue.state.mn.us/minnesota-standard-deduction
          https://mn.gov/mmb/assets/mwr_form_tcm1059-128581.pdf
          https://www.revenue.state.mn.us/individual-income-tax
          https://www.revenue.state.mn.us/individual-income-tax
        
        Oklahoma:
        -         
        - https://oklahoma.gov/content/dam/ok/en/tax/documents/resources/publications/businesses/withholding-tables/WHTables-2022.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Minnesota', 'state_occupation': 'Oklahoma', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def minnesota_and_oregon(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Minnesota:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Oregon:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Minnesota:
        - https://www.revenue.state.mn.us/minnesota-income-tax-rates-and-brackets
          https://www.revenue.state.mn.us/child-and-dependent-care-credit
          https://www.revenue.state.mn.us/dependent-exemptions
          https://www.revenue.state.mn.us/minnesota-standard-deduction
          https://mn.gov/mmb/assets/mwr_form_tcm1059-128581.pdf
          https://www.revenue.state.mn.us/individual-income-tax
          https://www.revenue.state.mn.us/individual-income-tax
        
        Oregon:
        -         
        - https://www.oregon.gov/dor/programs/individuals/Pages/PIT.aspx
          https://www.oregon.gov/dor/programs/individuals/Documents/Full%20year%20resident
         %20Form%20or-40%20filers.pdf
          https://www.oregon.gov/dor/programs/individuals/Documents/Part-year%20and%20nonresident
         %20Form%20or-40-P%20and%20or-40-N%20filers.pdf.pdf
          https://www.oregon.gov/dor/forms/FormsPubs/form-or-40_101-040_2022.pdf
          https://www.oregon.gov/dor/forms/FormsPubs/form-or-40_101-040_2022.pdf
          https://www.oregon.gov/dor/forms/FormsPubs/form-or-40_101-040_2022.pdf
          https://www.portland.gov/revenue/personal-tax
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Minnesota', 'state_occupation': 'Oregon', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def minnesota_and_pennsylvania(self, county_residence='Other', state_agi=75000, county_occupation='Other', capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Minnesota:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Pennsylvania:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        county_residence : str, optional
            County of residence. Default is Other.
        county_occupation : str, optional
            County of occupation. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Minnesota:
        - https://www.revenue.state.mn.us/minnesota-income-tax-rates-and-brackets
          https://www.revenue.state.mn.us/child-and-dependent-care-credit
          https://www.revenue.state.mn.us/dependent-exemptions
          https://www.revenue.state.mn.us/minnesota-standard-deduction
          https://mn.gov/mmb/assets/mwr_form_tcm1059-128581.pdf
          https://www.revenue.state.mn.us/individual-income-tax
          https://www.revenue.state.mn.us/individual-income-tax
        
        Pennsylvania:
        -         
        - https://www.revenue.pa.gov/Tax%20Rates/Pages/default.aspx
          https://www.revenue.pa.gov/FormsandPublications/FormsforIndividuals/PIT/Documents/2022/2022_pa-40in.pdf
          https://www.revenue.pa.gov/FormsandPublications/FormsforBusinesses/EmployerWithholding/Documents/rev-419.pdf
          https://www.revenue.pa.gov/FormsandPublications/FormsforIndividuals/PIT/Documents/2022/2022_pa-40in.pdf
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-25.htm
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        county_occupation = self.county_occupation if hasattr(self, 'county_occupation') else county_occupation
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Minnesota', 'state_occupation': 'Pennsylvania', "county_residence": county_residence, "state_agi": state_agi, "county_occupation": county_occupation, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def minnesota_and_rhode_island(self, state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Minnesota:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Rhode Island:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Minnesota:
        - https://www.revenue.state.mn.us/minnesota-income-tax-rates-and-brackets
          https://www.revenue.state.mn.us/child-and-dependent-care-credit
          https://www.revenue.state.mn.us/dependent-exemptions
          https://www.revenue.state.mn.us/minnesota-standard-deduction
          https://mn.gov/mmb/assets/mwr_form_tcm1059-128581.pdf
          https://www.revenue.state.mn.us/individual-income-tax
          https://www.revenue.state.mn.us/individual-income-tax
        
        Rhode Island:
        -         
        - https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2022_1040WE_w_0.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Minnesota', 'state_occupation': 'RhodeIsland', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def minnesota_and_south_carolina(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Minnesota:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        South Carolina:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Minnesota:
        - https://www.revenue.state.mn.us/minnesota-income-tax-rates-and-brackets
          https://www.revenue.state.mn.us/child-and-dependent-care-credit
          https://www.revenue.state.mn.us/dependent-exemptions
          https://www.revenue.state.mn.us/minnesota-standard-deduction
          https://mn.gov/mmb/assets/mwr_form_tcm1059-128581.pdf
          https://www.revenue.state.mn.us/individual-income-tax
          https://www.revenue.state.mn.us/individual-income-tax
        
        South Carolina:
        -         
        - https://dor.sc.gov/resources-site/lawandpolicy/Advisory%20Opinions/IL22-15.pdf
          https://dor.sc.gov/forms-site/Forms/IITPacket_2022.pdf
          https://www.scstatehouse.gov/code/t12c006.php
          https://dor.sc.gov/resources-site/lawandpolicy/Advisory%20Opinions/IL22-15.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Minnesota', 'state_occupation': 'SouthCarolina', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def minnesota_and_south_dakota(self, state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Minnesota:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        South Dakota:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Minnesota:
        - https://www.revenue.state.mn.us/minnesota-income-tax-rates-and-brackets
          https://www.revenue.state.mn.us/child-and-dependent-care-credit
          https://www.revenue.state.mn.us/dependent-exemptions
          https://www.revenue.state.mn.us/minnesota-standard-deduction
          https://mn.gov/mmb/assets/mwr_form_tcm1059-128581.pdf
          https://www.revenue.state.mn.us/individual-income-tax
          https://www.revenue.state.mn.us/individual-income-tax
        
        South Dakota:
        -         
        - https://dor.sd.gov/individuals/taxes/
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Minnesota', 'state_occupation': 'SouthDakota', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def minnesota_and_tennessee(self, state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Minnesota:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Tennessee:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Minnesota:
        - https://www.revenue.state.mn.us/minnesota-income-tax-rates-and-brackets
          https://www.revenue.state.mn.us/child-and-dependent-care-credit
          https://www.revenue.state.mn.us/dependent-exemptions
          https://www.revenue.state.mn.us/minnesota-standard-deduction
          https://mn.gov/mmb/assets/mwr_form_tcm1059-128581.pdf
          https://www.revenue.state.mn.us/individual-income-tax
          https://www.revenue.state.mn.us/individual-income-tax
        
        Tennessee:
        -         
        - https://www.tn.gov/revenue/taxes.html
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Minnesota', 'state_occupation': 'Tennessee', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def minnesota_and_texas(self, state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Minnesota:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Texas:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Minnesota:
        - https://www.revenue.state.mn.us/minnesota-income-tax-rates-and-brackets
          https://www.revenue.state.mn.us/child-and-dependent-care-credit
          https://www.revenue.state.mn.us/dependent-exemptions
          https://www.revenue.state.mn.us/minnesota-standard-deduction
          https://mn.gov/mmb/assets/mwr_form_tcm1059-128581.pdf
          https://www.revenue.state.mn.us/individual-income-tax
          https://www.revenue.state.mn.us/individual-income-tax
        
        Texas:
        -         
        - https://comptroller.texas.gov/taxes/
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Minnesota', 'state_occupation': 'Texas', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def minnesota_and_utah(self, state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Minnesota:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Utah:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Minnesota:
        - https://www.revenue.state.mn.us/minnesota-income-tax-rates-and-brackets
          https://www.revenue.state.mn.us/child-and-dependent-care-credit
          https://www.revenue.state.mn.us/dependent-exemptions
          https://www.revenue.state.mn.us/minnesota-standard-deduction
          https://mn.gov/mmb/assets/mwr_form_tcm1059-128581.pdf
          https://www.revenue.state.mn.us/individual-income-tax
          https://www.revenue.state.mn.us/individual-income-tax
        
        Utah:
        -         
        - https://incometax.utah.gov/paying/tax-rates
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Minnesota', 'state_occupation': 'Utah', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def minnesota_and_vermont(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Minnesota:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Vermont:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Minnesota:
        - https://www.revenue.state.mn.us/minnesota-income-tax-rates-and-brackets
          https://www.revenue.state.mn.us/child-and-dependent-care-credit
          https://www.revenue.state.mn.us/dependent-exemptions
          https://www.revenue.state.mn.us/minnesota-standard-deduction
          https://mn.gov/mmb/assets/mwr_form_tcm1059-128581.pdf
          https://www.revenue.state.mn.us/individual-income-tax
          https://www.revenue.state.mn.us/individual-income-tax
        
        Vermont:
        -         
        - https://tax.vermont.gov/sites/tax/files/documents/GB-1210-2023.pdf
          https://tax.vermont.gov/sites/tax/files/documents/Income%20Booklet-2022.pdf
          https://tax.vermont.gov/sites/tax/files/documents/Income%20Booklet-2022.pdf
          https://tax.vermont.gov/sites/tax/files/documents/Income%20Booklet-2022.pdf
          https://tax.vermont.gov/individuals/personal-income-tax/tax-credits
          https://tax.vermont.gov/individuals/personal-income-tax/taxable-income
          https://tax.vermont.gov/individuals/personal-income-tax/local-option-tax
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Minnesota', 'state_occupation': 'Vermont', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def minnesota_and_virginia(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Minnesota:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Virginia:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Minnesota:
        - https://www.revenue.state.mn.us/minnesota-income-tax-rates-and-brackets
          https://www.revenue.state.mn.us/child-and-dependent-care-credit
          https://www.revenue.state.mn.us/dependent-exemptions
          https://www.revenue.state.mn.us/minnesota-standard-deduction
          https://mn.gov/mmb/assets/mwr_form_tcm1059-128581.pdf
          https://www.revenue.state.mn.us/individual-income-tax
          https://www.revenue.state.mn.us/individual-income-tax
        
        Virginia:
        -         
        - https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/reciprocity
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Minnesota', 'state_occupation': 'Virginia', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def minnesota_and_washington(self, state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Minnesota:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Washington:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Minnesota:
        - https://www.revenue.state.mn.us/minnesota-income-tax-rates-and-brackets
          https://www.revenue.state.mn.us/child-and-dependent-care-credit
          https://www.revenue.state.mn.us/dependent-exemptions
          https://www.revenue.state.mn.us/minnesota-standard-deduction
          https://mn.gov/mmb/assets/mwr_form_tcm1059-128581.pdf
          https://www.revenue.state.mn.us/individual-income-tax
          https://www.revenue.state.mn.us/individual-income-tax
        
        Washington:
        -         
        - https://dor.wa.gov/taxes-rates/income-tax
          https://dor.wa.gov/taxes-rates/other-taxes/capital-gains-tax
          https://dor.wa.gov/taxes-rates/income-tax
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Minnesota', 'state_occupation': 'Washington', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def minnesota_and_west_virginia(self, county_occupation='Other', state_agi=75000, pay_periods=24, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Minnesota:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        West Virginia:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_occupation : str, optional
            County of occupation. Default is Other.
        pay_periods : int, optional
            Number of pay periods. Options are 13, 24, 26, 52. Default is 24.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Minnesota:
        - https://www.revenue.state.mn.us/minnesota-income-tax-rates-and-brackets
          https://www.revenue.state.mn.us/child-and-dependent-care-credit
          https://www.revenue.state.mn.us/dependent-exemptions
          https://www.revenue.state.mn.us/minnesota-standard-deduction
          https://mn.gov/mmb/assets/mwr_form_tcm1059-128581.pdf
          https://www.revenue.state.mn.us/individual-income-tax
          https://www.revenue.state.mn.us/individual-income-tax
        
        West Virginia:
        -         
        - https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://tax.wv.gov/Documents/TSD/tsd438.pdf
          https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2011/Taxes-11-35att.html
        """
        
        # Check if the parameters are set upstream in the instance
        county_occupation = self.county_occupation if hasattr(self, 'county_occupation') else county_occupation
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        pay_periods = self.pay_periods if hasattr(self, 'pay_periods') else pay_periods
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Minnesota', 'state_occupation': 'WestVirginia', "county_occupation": county_occupation, "state_agi": state_agi, "pay_periods": pay_periods, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def minnesota_and_wisconsin(self, state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Minnesota:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Wisconsin:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Minnesota:
        - https://www.revenue.state.mn.us/minnesota-income-tax-rates-and-brackets
          https://www.revenue.state.mn.us/child-and-dependent-care-credit
          https://www.revenue.state.mn.us/dependent-exemptions
          https://www.revenue.state.mn.us/minnesota-standard-deduction
          https://mn.gov/mmb/assets/mwr_form_tcm1059-128581.pdf
          https://www.revenue.state.mn.us/individual-income-tax
          https://www.revenue.state.mn.us/individual-income-tax
        
        Wisconsin:
        -         
        - https://www.revenue.wi.gov/TaxForms2023/2023-Form1-ES-Inst.pdf
          https://www.revenue.wi.gov/TaxForms2023/2023-Form1-ES-Inst.pdf
          https://www.revenue.wi.gov/TaxForms2023/2023-Form1-ES-Inst.pdf
          https://www.revenue.wi.gov/Pages/FAQS/pcs-work.aspx
          https://www.revenue.wi.gov/Dor%20Publications/pb103.pdf
          https://www.revenue.wi.gov/Pages/FAQS/pcs-work.aspx
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Minnesota', 'state_occupation': 'Wisconsin', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def minnesota_and_wyoming(self, state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Minnesota:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Wyoming:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Minnesota:
        - https://www.revenue.state.mn.us/minnesota-income-tax-rates-and-brackets
          https://www.revenue.state.mn.us/child-and-dependent-care-credit
          https://www.revenue.state.mn.us/dependent-exemptions
          https://www.revenue.state.mn.us/minnesota-standard-deduction
          https://mn.gov/mmb/assets/mwr_form_tcm1059-128581.pdf
          https://www.revenue.state.mn.us/individual-income-tax
          https://www.revenue.state.mn.us/individual-income-tax
        
        Wyoming:
        -         
        - https://revenue.wyo.gov/
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Minnesota', 'state_occupation': 'Wyoming', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def mississippi_and_alabama(self, county_residence='Other', state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Mississippi:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Alabama:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Mississippi:
        - https://www.dor.ms.gov/individual/tax-rates
          https://www.dor.ms.gov/individual/tax-rates
          https://www.dor.ms.gov/individual/tax-rates
          https://www.dor.ms.gov/individual
          https://www.dor.ms.gov/individual
        
        Alabama:
        -         
        - https://www.revenue.alabama.gov/ultraviewer/viewer/basic_viewer/index.html?form=2023/02/22f40.pdf
          https://www.revenue.alabama.gov/faqs/what-is-alabamas-individual-income-tax-rate/
          https://www.revenue.alabama.gov/ultraviewer/viewer/basic_viewer/index.html?form=2023/02/22f40.pdf
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-24.htm?taxmap=true
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-24.htm?taxmap=true
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-24.htm?taxmap=true
          https://www.revenue.alabama.gov/individual-corporate/income-to-be-reported-on-the-alabama-income-tax-return/
          https://www.almonline.org/TaxRates.aspx#OccupationalTax
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Mississippi', 'state_occupation': 'Alabama', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def mississippi_and_alaska(self, state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Mississippi:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Alaska:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Mississippi:
        - https://www.dor.ms.gov/individual/tax-rates
          https://www.dor.ms.gov/individual/tax-rates
          https://www.dor.ms.gov/individual/tax-rates
          https://www.dor.ms.gov/individual
          https://www.dor.ms.gov/individual
        
        Alaska:
        -         
        - https://www.commerce.alaska.gov/web/dcra/officeofthestateassessor/alaskataxfacts.aspx
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Mississippi', 'state_occupation': 'Alaska', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def mississippi_and_arizona(self, state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Mississippi:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Arizona:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        credits_state : int, optional
            Credits for state taxes. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Mississippi:
        - https://www.dor.ms.gov/individual/tax-rates
          https://www.dor.ms.gov/individual/tax-rates
          https://www.dor.ms.gov/individual/tax-rates
          https://www.dor.ms.gov/individual
          https://www.dor.ms.gov/individual
        
        Arizona:
        -         
        - https://azdor.gov/forms/individual/form-140-x-y-tables
          https://azdor.gov/forms/individual/form-140-x-y-tables
          https://azdor.gov/forms/individual/form-140-x-y-tables
          https://azdor.gov/businesses-arizona/withholding-tax/withholding-exceptions
          "https://azdor.gov/individuals/income-tax-filing-assistance/identifying-other-taxable-income#:~:text=In%20regards%20to%20capital%20gains
         the%20individuals%20regular%20tax%20rate.
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Mississippi', 'state_occupation': 'Arizona', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def mississippi_and_arkansas(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Mississippi:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Arkansas:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Mississippi:
        - https://www.dor.ms.gov/individual/tax-rates
          https://www.dor.ms.gov/individual/tax-rates
          https://www.dor.ms.gov/individual/tax-rates
          https://www.dor.ms.gov/individual
          https://www.dor.ms.gov/individual
        
        Arkansas:
        -         
        - https://www.arkansasedc.com/why-arkansas/business-climate/tax-structure/personal-income-tax
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_and_AR1000NR_Instructions.pdf
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_FullYearResidentIndividualIncomeTaxReturn.pdf
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_and_AR1000NR_Instructions.pdf
          https://www.arkansasedc.com/why-arkansas/business-climate/tax-structure/capital-gains-tax-reduction
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_and_AR1000NR_Instructions.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Mississippi', 'state_occupation': 'Arkansas', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def mississippi_and_california(self, state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Mississippi:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        California:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Mississippi:
        - https://www.dor.ms.gov/individual/tax-rates
          https://www.dor.ms.gov/individual/tax-rates
          https://www.dor.ms.gov/individual/tax-rates
          https://www.dor.ms.gov/individual
          https://www.dor.ms.gov/individual
        
        California:
        -         
        - https://www.ftb.ca.gov/forms/2022/2022-540-tax-rate-schedules.pdf
          https://www.ftb.ca.gov/file/personal/deductions/index.html
          https://www.ftb.ca.gov/about-ftb/newsroom/caleitc/eligibility-and-credit-information.html
          https://www.ftb.ca.gov/file/personal/income-types/capital-gains-and-losses.html#:~:text=California%20does%20not%20have%20a
         are%20taxed%20as%20ordinary%20income
          https://www.ftb.ca.gov/file/personal/
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Mississippi', 'state_occupation': 'California', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def mississippi_and_colorado(self, county_residence='Other', state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Mississippi:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Colorado:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Mississippi:
        - https://www.dor.ms.gov/individual/tax-rates
          https://www.dor.ms.gov/individual/tax-rates
          https://www.dor.ms.gov/individual/tax-rates
          https://www.dor.ms.gov/individual
          https://www.dor.ms.gov/individual
        
        Colorado:
        -         
        - https://leg.colorado.gov/agencies/legislative-council-staff/individual-income-tax%C2%A0
          https://leg.colorado.gov/agencies/legislative-council-staff/individual-income-tax%C2%A0
          https://tax.colorado.gov/sites/tax/files/documents/ITT_Capital_Gain_Subtraction_Feb_2022.pdf
          https://www.auroragov.org/business_services/taxes/occupational_privilege_tax
          https://www.denvergov.org/Government/Agencies-Departments-Offices/Agencies-Departments-Offices-Directory/Department-of-Finance/Our-Divisions/Treasury/Business-Tax-Information#:~:text=Occupational%20privilege%20tax
         -%28show%20below%29&text=Each%20taxable%20employee%20is%20liable
         month%20for%20each%20taxable%20employee
          https://www.glendale.co.us/355/Occupational-Privilege-Tax
          https://www.greenwoodvillage.com/1220/Occupational-Privilege-Tax-OPT
          https://ci.sheridan.co.us/288/Occupational-Privilege-Tax
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Mississippi', 'state_occupation': 'Colorado', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def mississippi_and_connecticut(self, state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Mississippi:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Connecticut:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Mississippi:
        - https://www.dor.ms.gov/individual/tax-rates
          https://www.dor.ms.gov/individual/tax-rates
          https://www.dor.ms.gov/individual/tax-rates
          https://www.dor.ms.gov/individual
          https://www.dor.ms.gov/individual
        
        Connecticut:
        -         
        - https://portal.ct.gov/-/media/DRS/Forms/2022/Income/2022-CT-1040-Instructions_1222.pdf
          https://www.cga.ct.gov/2022/rpt/pdf/2022-R-0108.pdf
          https://www.cga.ct.gov/current/pub/chap_229.htm#sec_12-703
          https://portal.ct.gov/-/media/DRS/Forms/2022/Income/2022-CT-1040-Instructions_1222.pdf
          "https://portal.ct.gov/DRS/Publications/TSSNs/TSSN-29#:~:text=An%20individuals%20net%20capital%20gains
         portion%20of%20Social%20Security%20benefits"
          "https://portal.ct.gov/DRS/Publications/TSSNs/TSSN-29#:~:text=An%20individuals%20net%20capital%20gains
         portion%20of%20Social%20Security%20benefits"
          https://portal.ct.gov/DRS/Individuals/Resident-Income-Tax/Tax-Information
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Mississippi', 'state_occupation': 'Connecticut', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def mississippi_and_delaware(self, state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, federal_agi=75000, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Mississippi:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Delaware:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        federal_agi : int, optional
            Adjusted Gross Income (AGI) for federal taxes. Default is 0. Sources - https://www.irs.gov/e-file-providers/definition-of-adjusted-gross-income, https://www.irs.gov/newsroom/taxpayers-should-know-the-difference-between-standard-and-itemized-deductions
        income : int, optional
            Gross income. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Mississippi:
        - https://www.dor.ms.gov/individual/tax-rates
          https://www.dor.ms.gov/individual/tax-rates
          https://www.dor.ms.gov/individual/tax-rates
          https://www.dor.ms.gov/individual
          https://www.dor.ms.gov/individual
        
        Delaware:
        -         
        - https://revenue.delaware.gov/software-developer/tax-rate-changes/
          https://revenuefiles.delaware.gov/2022/PIT-RES_TY22_2022-01_PaperInteractive.pdf
          https://revenuefiles.delaware.gov/2022/PIT-RES_TY22_2022-02_Instructions.pdf
          https://revenue.delaware.gov/frequently-asked-questions/personal-income-tax-faqs/
          https://www.newcastlede.gov/DocumentCenter/View/20147/City-of-Wilmington-Wage-Tax-Information-PDF
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        federal_agi = self.federal_agi if hasattr(self, 'federal_agi') else federal_agi
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Mississippi', 'state_occupation': 'Delaware', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "federal_agi": federal_agi, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def mississippi_and_florida(self, state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Mississippi:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Florida:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Mississippi:
        - https://www.dor.ms.gov/individual/tax-rates
          https://www.dor.ms.gov/individual/tax-rates
          https://www.dor.ms.gov/individual/tax-rates
          https://www.dor.ms.gov/individual
          https://www.dor.ms.gov/individual
        
        Florida:
        -         
        - https://floridarevenue.com/taxes/taxesfees/Pages/tax_interest_rates.aspx
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Mississippi', 'state_occupation': 'Florida', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def mississippi_and_georgia(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Mississippi:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Georgia:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Mississippi:
        - https://www.dor.ms.gov/individual/tax-rates
          https://www.dor.ms.gov/individual/tax-rates
          https://www.dor.ms.gov/individual/tax-rates
          https://www.dor.ms.gov/individual
          https://www.dor.ms.gov/individual
        
        Georgia:
        -         
        - https://dor.georgia.gov/tax-tables-georgia-tax-rate-schedule
          https://apps.dor.ga.gov/FillableForms/PDFViewer/Index?form=2022GA500
          https://apps.dor.ga.gov/FillableForms/PDFViewer/Index?form=2022GA500
          https://apps.dor.ga.gov/FillableForms/PDFViewer/Index?form=2022GA500
          https://dor.georgia.gov/years-individual-income-tax-forms
          https://dor.georgia.gov/years-individual-income-tax-forms
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Mississippi', 'state_occupation': 'Georgia', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def mississippi_and_hawaii(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Mississippi:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Hawaii:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Mississippi:
        - https://www.dor.ms.gov/individual/tax-rates
          https://www.dor.ms.gov/individual/tax-rates
          https://www.dor.ms.gov/individual/tax-rates
          https://www.dor.ms.gov/individual
          https://www.dor.ms.gov/individual
        
        Hawaii:
        -         
        - https://tax.hawaii.gov/forms/d_18table-on/d_18table-on_p13/
          https://files.hawaii.gov/tax/forms/2018/18table-on.pdf
          https://files.hawaii.gov/tax/news/pubs/22outline.pdf
          https://files.hawaii.gov/tax/news/pubs/FastTaxRefGuide_2017.pdf
          https://www.tfhawaii.org/wordpress/state-tax-resources/mini-tax-guides/individual-income-tax-chapter-235/#:~:text=Long%20term%20capital%20gains%20are%20taxed%20at%20a%20maximum%20of%207.25%25
          https://www.grassrootinstitute.org/2023/02/hb337-increasing-tax-on-capital-gains-a-bad-risk/
          https://files.hawaii.gov/tax/news/pubs/22outline.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Mississippi', 'state_occupation': 'Hawaii', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def mississippi_and_idaho(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Mississippi:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Idaho:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Mississippi:
        - https://www.dor.ms.gov/individual/tax-rates
          https://www.dor.ms.gov/individual/tax-rates
          https://www.dor.ms.gov/individual/tax-rates
          https://www.dor.ms.gov/individual
          https://www.dor.ms.gov/individual
        
        Idaho:
        -         
        - https://tax.idaho.gov/taxes/income-tax/individual-income/individual-income-tax-rate-schedule/
          https://tax.idaho.gov/wp-content/uploads/forms/EIN00046/EIN00046_03-01-2023.pdf
          https://tax.idaho.gov/taxes/income-tax/individual-income/idaho-source-income/
          https://tax.idaho.gov/taxes/income-tax/individual-income/idaho-source-income/
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Mississippi', 'state_occupation': 'Idaho', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def mississippi_and_illinois(self, state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Mississippi:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Illinois:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Mississippi:
        - https://www.dor.ms.gov/individual/tax-rates
          https://www.dor.ms.gov/individual/tax-rates
          https://www.dor.ms.gov/individual/tax-rates
          https://www.dor.ms.gov/individual
          https://www.dor.ms.gov/individual
        
        Illinois:
        -         
        - https://tax.illinois.gov/research/taxrates/income.html#IndividualIncome
          https://tax.illinois.gov/content/dam/soi/en/web/tax/forms/incometax/documents/currentyear/individual/il-1040-instr.pdf
          https://tax.illinois.gov/questionsandanswers/12.html
          https://tax.illinois.gov/content/dam/soi/en/web/tax/forms/incometax/documents/currentyear/individual/il-1040-instr.pdf
          https://tax.illinois.gov/research/taxrates/personalproperty.html
          https://tax.illinois.gov/localgovernments/income.html
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Mississippi', 'state_occupation': 'Illinois', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def mississippi_and_indiana(self, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Mississippi:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Indiana:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Mississippi:
        - https://www.dor.ms.gov/individual/tax-rates
          https://www.dor.ms.gov/individual/tax-rates
          https://www.dor.ms.gov/individual/tax-rates
          https://www.dor.ms.gov/individual
          https://www.dor.ms.gov/individual
        
        Indiana:
        -         
        - https://www.in.gov/dor/tax-forms/2022-individual-income-tax-forms/
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-32.htm
          https://www.in.gov/dor/files/ib28.pdf
          https://www.in.gov/dor/tax-forms/2022-individual-income-tax-forms/
          https://www.in.gov/dor/files/2022-county-tax-rates-and-codes.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Mississippi', 'state_occupation': 'Indiana', "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def mississippi_and_iowa(self, county_residence='Other', state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Mississippi:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Iowa:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Mississippi:
        - https://www.dor.ms.gov/individual/tax-rates
          https://www.dor.ms.gov/individual/tax-rates
          https://www.dor.ms.gov/individual/tax-rates
          https://www.dor.ms.gov/individual
          https://www.dor.ms.gov/individual
        
        Iowa:
        -         
        - https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
          https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
          https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
          https://tax.iowa.gov/iowa-illinois-reciprocal-agreement
          https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Mississippi', 'state_occupation': 'Iowa', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def mississippi_and_kansas(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Mississippi:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Kansas:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Mississippi:
        - https://www.dor.ms.gov/individual/tax-rates
          https://www.dor.ms.gov/individual/tax-rates
          https://www.dor.ms.gov/individual/tax-rates
          https://www.dor.ms.gov/individual
          https://www.dor.ms.gov/individual
        
        Kansas:
        -         
        - https://www.ksrevenue.gov/taxrates.html
          https://www.ksrevenue.gov/incomebook22.html#0
          https://www.ksrevenue.gov/incomebook22.html#0
          https://www.ksrevenue.gov/incomebook22.html#0
          https://www.ksrevenue.gov/incomebook22.html
          https://www.ksrevenue.gov/incomebook22.html
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Mississippi', 'state_occupation': 'Kansas', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def mississippi_and_kentucky(self, county_residence='Other', state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Mississippi:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Kentucky:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Mississippi:
        - https://www.dor.ms.gov/individual/tax-rates
          https://www.dor.ms.gov/individual/tax-rates
          https://www.dor.ms.gov/individual/tax-rates
          https://www.dor.ms.gov/individual
          https://www.dor.ms.gov/individual
        
        Kentucky:
        -         
        - https://revenue.ky.gov/News/Pages/Dor-Announces-Updates-to-Individual-Income-Tax-for-2023-Tax-Year.aspx
          https://revenue.ky.gov/Individual/Individual-Income-Tax/Pages/default.aspx
          https://revenue.ky.gov/News/Pages/Dor-Announces-Updates-to-Individual-Income-Tax-for-2023-Tax-Year.aspx
          https://revenue.ky.gov/Individual/Individual-Income-Tax/Pages/default.aspx
          https://revenue.ky.gov/Dor%20Training%20Materials/103%20KAR%2017.140.%20Individual%20income%20tax%20-%20reciprocity%20-%20nonresidents.pdf
          https://revenue.ky.gov/Individual/Individual-Income-Tax/Pages/default.aspx
          https://nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2021/Taxes-21-32.htm?taxmap=true
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Mississippi', 'state_occupation': 'Kentucky', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def mississippi_and_louisiana(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Mississippi:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Louisiana:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Mississippi:
        - https://www.dor.ms.gov/individual/tax-rates
          https://www.dor.ms.gov/individual/tax-rates
          https://www.dor.ms.gov/individual/tax-rates
          https://www.dor.ms.gov/individual
          https://www.dor.ms.gov/individual
        
        Louisiana:
        -         
        - https://revenue.louisiana.gov/individualincometax
          https://revenue.louisiana.gov/individualincometax
          https://www.revenue.louisiana.gov/taxforms/6935(11_02)F.pdf
          https://revenue.louisiana.gov/individualincometax
          https://revenue.louisiana.gov/individualincometax
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Mississippi', 'state_occupation': 'Louisiana', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def mississippi_and_maine(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Mississippi:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Maine:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Mississippi:
        - https://www.dor.ms.gov/individual/tax-rates
          https://www.dor.ms.gov/individual/tax-rates
          https://www.dor.ms.gov/individual/tax-rates
          https://www.dor.ms.gov/individual
          https://www.dor.ms.gov/individual
        
        Maine:
        -         
        - https://www.maine.gov/revenue/tax-return-forms/individual-income-tax-2022
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/ind_tax_rate_sched_2022.pdf
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/22_1040me_book_gen_instruc_revisedFeb23.pdf
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/22_1040me_book_gen_instruc_revisedFeb23.pdf
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/22_1040me_book_gen_instruc_revisedFeb23.pdf
          https://www.maine.gov/revenue/tax-return-forms/individual-income-tax-2022
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Mississippi', 'state_occupation': 'Maine', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def mississippi_and_maryland(self, county_residence='Other', state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Mississippi:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Maryland:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Mississippi:
        - https://www.dor.ms.gov/individual/tax-rates
          https://www.dor.ms.gov/individual/tax-rates
          https://www.dor.ms.gov/individual/tax-rates
          https://www.dor.ms.gov/individual
          https://www.dor.ms.gov/individual
        
        Maryland:
        -         
        - https://www.marylandtaxes.gov/individual/income/tax-info/tax-rates.php
          https://www.marylandtaxes.gov/forms/22_forms/Resident_Booklet.pdf
          https://www.marylandtaxes.gov/forms/Tax_Publications/Administrative_Releases/Income_and_Estate_Tax_Releases/ar_it37.pdf
          https://www.marylandtaxes.gov/individual/index.php
          https://www.marylandtaxes.gov/individual/income/tax-info/tax-rates.php
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Mississippi', 'state_occupation': 'Maryland', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def mississippi_and_massachusetts(self, state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Mississippi:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Massachusetts:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Mississippi:
        - https://www.dor.ms.gov/individual/tax-rates
          https://www.dor.ms.gov/individual/tax-rates
          https://www.dor.ms.gov/individual/tax-rates
          https://www.dor.ms.gov/individual
          https://www.dor.ms.gov/individual
        
        Massachusetts:
        -         
        - https://www.mass.gov/service-details/massachusetts-tax-rates
          https://www.mass.gov/service-details/view-massachusetts-personal-income-tax-exemptions
          https://www.mass.gov/service-details/view-massachusetts-personal-income-tax-exemptions
          https://www.mass.gov/service-details/view-child-and-dependent-related-credits
          https://www.mass.gov/service-details/massachusetts-tax-rates
          https://www.mass.gov/personal-income-tax
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Mississippi', 'state_occupation': 'Massachusetts', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def mississippi_and_michigan(self, county_occupation='Other', state_agi=75000, county_residence='Other', capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Mississippi:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Michigan:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        county_occupation : str, optional
            County of occupation. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Mississippi:
        - https://www.dor.ms.gov/individual/tax-rates
          https://www.dor.ms.gov/individual/tax-rates
          https://www.dor.ms.gov/individual/tax-rates
          https://www.dor.ms.gov/individual
          https://www.dor.ms.gov/individual
        
        Michigan:
        -         
        - https://www.michigan.gov/taxes/iit/new-developments-for-tax-year-2022
          "https://www.michigan.gov/taxes/-/media/Project/Websites/taxes/Forms/2023/2023-SUW/446_Withholding-Guide_2023.pdf?rev=177ca3d8cb034392bf84e317ed78c320&hash=19C5F42F43A89A91863D229349F48883#:~:text=The%20exemption%20amount%20is%20%245%2C400
         employees%20Michigan%20income%20tax%20return"
          "https://www.michigan.gov/taxes/-/media/Project/Websites/taxes/Forms/2023/2023-SUW/446_Withholding-Guide_2023.pdf?rev=177ca3d8cb034392bf84e317ed78c320&hash=19C5F42F43A89A91863D229349F48883#:~:text=The%20exemption%20amount%20is%20%245%2C400
         employees%20Michigan%20income%20tax%20return"
          https://www.michigan.gov/taxes/questions/iit/accordion/residency/are-my-wages-earned-in-another-state-taxable-in-michigan-if-i-am-a-michigan-resident-1
          https://www.michigan.gov/taxes/-/media/Project/Websites/taxes/Forms/2022/2022-IIT-Forms/BOOK_MI-1040.pdf?rev=4524f1cbbc0243b194ab05a6680be75b&hash=40DFE71E651F7C1D65C31F6CD03D944F
          https://www.michigan.gov/taxes/questions/iit/accordion/general/what-cities-impose-an-income-tax
        """
        
        # Check if the parameters are set upstream in the instance
        county_occupation = self.county_occupation if hasattr(self, 'county_occupation') else county_occupation
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Mississippi', 'state_occupation': 'Michigan', "county_occupation": county_occupation, "state_agi": state_agi, "county_residence": county_residence, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def mississippi_and_minnesota(self, state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Mississippi:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Minnesota:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Mississippi:
        - https://www.dor.ms.gov/individual/tax-rates
          https://www.dor.ms.gov/individual/tax-rates
          https://www.dor.ms.gov/individual/tax-rates
          https://www.dor.ms.gov/individual
          https://www.dor.ms.gov/individual
        
        Minnesota:
        -         
        - https://www.revenue.state.mn.us/minnesota-income-tax-rates-and-brackets
          https://www.revenue.state.mn.us/child-and-dependent-care-credit
          https://www.revenue.state.mn.us/dependent-exemptions
          https://www.revenue.state.mn.us/minnesota-standard-deduction
          https://mn.gov/mmb/assets/mwr_form_tcm1059-128581.pdf
          https://www.revenue.state.mn.us/individual-income-tax
          https://www.revenue.state.mn.us/individual-income-tax
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Mississippi', 'state_occupation': 'Minnesota', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def mississippi_and_missouri(self, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Mississippi:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Missouri:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Mississippi:
        - https://www.dor.ms.gov/individual/tax-rates
          https://www.dor.ms.gov/individual/tax-rates
          https://www.dor.ms.gov/individual/tax-rates
          https://www.dor.ms.gov/individual
          https://www.dor.ms.gov/individual
        
        Missouri:
        -         
        - https://dor.mo.gov/taxation/individual/tax-types/income/year-changes/
          https://dor.mo.gov/taxation/individual/tax-types/income/year-changes/
          https://dor.mo.gov/taxation/individual/tax-types/income/year-changes/
          https://www.kcmo.gov/city-hall/departments/finance/earnings-tax
          https://www.stlouis-mo.gov/government/departments/collector/earnings-tax/individual-earnings-tax-info.cfm
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Mississippi', 'state_occupation': 'Missouri', "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def mississippi_and_montana(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Mississippi:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Montana:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Mississippi:
        - https://www.dor.ms.gov/individual/tax-rates
          https://www.dor.ms.gov/individual/tax-rates
          https://www.dor.ms.gov/individual/tax-rates
          https://www.dor.ms.gov/individual
          https://www.dor.ms.gov/individual
        
        Montana:
        -         
        - https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/2022/12/montana-employees-withholding-allowance-and-exemption-certificate-form-mw-4.pdf
          https://mtrevenue.gov/2021/11/29/simplification-of-montana-income-taxation/?utm_source=rss&utm_medium=rss&utm_campaign=simplification-of-montana-income-taxation
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Mississippi', 'state_occupation': 'Montana', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def mississippi_and_nebraska(self, state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Mississippi:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Nebraska:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Mississippi:
        - https://www.dor.ms.gov/individual/tax-rates
          https://www.dor.ms.gov/individual/tax-rates
          https://www.dor.ms.gov/individual/tax-rates
          https://www.dor.ms.gov/individual
          https://www.dor.ms.gov/individual
        
        Nebraska:
        -         
        - https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2023_Circular_EN_8-429r2023_Whole_Book_11.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Mississippi', 'state_occupation': 'Nebraska', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def mississippi_and_nevada(self, state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Mississippi:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Nevada:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Mississippi:
        - https://www.dor.ms.gov/individual/tax-rates
          https://www.dor.ms.gov/individual/tax-rates
          https://www.dor.ms.gov/individual/tax-rates
          https://www.dor.ms.gov/individual
          https://www.dor.ms.gov/individual
        
        Nevada:
        -         
        - https://www.nvsos.gov/sos/businesses/the-nevada-advantage
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Mississippi', 'state_occupation': 'Nevada', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def mississippi_and_new_hampshire(self, state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Mississippi:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        New Hampshire:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Mississippi:
        - https://www.dor.ms.gov/individual/tax-rates
          https://www.dor.ms.gov/individual/tax-rates
          https://www.dor.ms.gov/individual/tax-rates
          https://www.dor.ms.gov/individual
          https://www.dor.ms.gov/individual
        
        New Hampshire:
        -         
        - https://www.revenue.nh.gov/assistance/tax-overview.htm
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Mississippi', 'state_occupation': 'NewHampshire', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def mississippi_and_new_jersey(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Mississippi:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        New Jersey:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Mississippi:
        - https://www.dor.ms.gov/individual/tax-rates
          https://www.dor.ms.gov/individual/tax-rates
          https://www.dor.ms.gov/individual/tax-rates
          https://www.dor.ms.gov/individual
          https://www.dor.ms.gov/individual
        
        New Jersey:
        -         
        - https://www.state.nj.us/treasury/taxation/taxtables.shtml
          https://www.state.nj.us/treasury/taxation/njit13.shtml
          https://www.nj.gov/treasury/taxation/njit25.shtml
          https://www.state.nj.us/treasury/taxation/individuals/freqqiti.shtml
          https://www.state.nj.us/treasury/taxation/localtax.shtml
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Mississippi', 'state_occupation': 'NewJersey', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def mississippi_and_new_mexico(self, state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Mississippi:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        New Mexico:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Mississippi:
        - https://www.dor.ms.gov/individual/tax-rates
          https://www.dor.ms.gov/individual/tax-rates
          https://www.dor.ms.gov/individual/tax-rates
          https://www.dor.ms.gov/individual
          https://www.dor.ms.gov/individual
        
        New Mexico:
        -         
        - https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/fdf3c548-8aba-4b9c-9eb4-bb564c716015/FYI-104.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/0bf943fd-652e-4400-bb1b-cd8397eb5a95/2022pit-adj.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/2f1a6781-9534-4436-b427-1557f9592099/2022pit-adj-ins.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/0bf943fd-652e-4400-bb1b-cd8397eb5a95/2022pit-adj.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/1afc56af-ea90-4d48-82e5-1f9aeb43255a/PITbook2022.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/fdf3c548-8aba-4b9c-9eb4-bb564c716015/FYI-104.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Mississippi', 'state_occupation': 'NewMexico', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def mississippi_and_new_york(self, county_residence='Other', state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Mississippi:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        New York:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Mississippi:
        - https://www.dor.ms.gov/individual/tax-rates
          https://www.dor.ms.gov/individual/tax-rates
          https://www.dor.ms.gov/individual/tax-rates
          https://www.dor.ms.gov/individual
          https://www.dor.ms.gov/individual
        
        New York:
        -         
        - https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nys-tax-rate-schedule
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nys-tax-rate-schedule
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nys-tax-rate-schedule
          https://www.tax.ny.gov/pit/
          https://www.tax.ny.gov/pit/file/nonresident-faqs.htm
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#yonkers-tax-rate-schedule
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nyc-tax-rate-schedule
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Mississippi', 'state_occupation': 'NewYork', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def mississippi_and_north_carolina(self, state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Mississippi:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        North Carolina:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Mississippi:
        - https://www.dor.ms.gov/individual/tax-rates
          https://www.dor.ms.gov/individual/tax-rates
          https://www.dor.ms.gov/individual/tax-rates
          https://www.dor.ms.gov/individual
          https://www.dor.ms.gov/individual
        
        North Carolina:
        -         
        - https://www.ncdor.gov/taxes-forms/tax-rate-schedules
          https://www.ncdor.gov/taxes-forms/individual-income-tax/north-carolina-standard-deduction-or-north-carolina-itemized-deductions
          https://www.ncdor.gov/taxes-forms/individual-income-tax
          https://www.ncdor.gov/taxes-forms/individual-income-tax
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Mississippi', 'state_occupation': 'NorthCarolina', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def mississippi_and_north_dakota(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Mississippi:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        North Dakota:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Mississippi:
        - https://www.dor.ms.gov/individual/tax-rates
          https://www.dor.ms.gov/individual/tax-rates
          https://www.dor.ms.gov/individual/tax-rates
          https://www.dor.ms.gov/individual
          https://www.dor.ms.gov/individual
        
        North Dakota:
        -         
        - https://www.tax.nd.gov/forms
          https://www.tax.nd.gov/sites/www/files/documents/forms/individual/2022-iit/2022-individual-income-tax-booklet.pdf
          https://www.tax.nd.gov/sites/www/files/documents/forms/2022-individual-income-tax-booklet.pdf
          https://www.tax.nd.gov/tax-exemptions-credits/income-tax-exemptions-credits
          https://www.tax.nd.gov/sites/www/files/documents/forms/business/ndwrfillable.pdf
          https://www.tax.nd.gov/news/tax-legislative-changes/significant-changes-law/individual-income-tax-history
          https://www.tax.nd.gov/forms
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Mississippi', 'state_occupation': 'NorthDakota', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def mississippi_and_ohio(self, county_residence='Other', state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Mississippi:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Ohio:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Mississippi:
        - https://www.dor.ms.gov/individual/tax-rates
          https://www.dor.ms.gov/individual/tax-rates
          https://www.dor.ms.gov/individual/tax-rates
          https://www.dor.ms.gov/individual
          https://www.dor.ms.gov/individual
        
        Ohio:
        -         
        - https://tax.ohio.gov/individual/resources/annual-tax-rates
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://tax.ohio.gov/static/forms/employer_withholding/generic/wth_it4nr.pdf
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://www.ritaohio.com/TaxRatesTable
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Mississippi', 'state_occupation': 'Ohio', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def mississippi_and_oklahoma(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Mississippi:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Oklahoma:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Mississippi:
        - https://www.dor.ms.gov/individual/tax-rates
          https://www.dor.ms.gov/individual/tax-rates
          https://www.dor.ms.gov/individual/tax-rates
          https://www.dor.ms.gov/individual
          https://www.dor.ms.gov/individual
        
        Oklahoma:
        -         
        - https://oklahoma.gov/content/dam/ok/en/tax/documents/resources/publications/businesses/withholding-tables/WHTables-2022.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Mississippi', 'state_occupation': 'Oklahoma', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def mississippi_and_oregon(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Mississippi:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Oregon:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Mississippi:
        - https://www.dor.ms.gov/individual/tax-rates
          https://www.dor.ms.gov/individual/tax-rates
          https://www.dor.ms.gov/individual/tax-rates
          https://www.dor.ms.gov/individual
          https://www.dor.ms.gov/individual
        
        Oregon:
        -         
        - https://www.oregon.gov/dor/programs/individuals/Pages/PIT.aspx
          https://www.oregon.gov/dor/programs/individuals/Documents/Full%20year%20resident
         %20Form%20or-40%20filers.pdf
          https://www.oregon.gov/dor/programs/individuals/Documents/Part-year%20and%20nonresident
         %20Form%20or-40-P%20and%20or-40-N%20filers.pdf.pdf
          https://www.oregon.gov/dor/forms/FormsPubs/form-or-40_101-040_2022.pdf
          https://www.oregon.gov/dor/forms/FormsPubs/form-or-40_101-040_2022.pdf
          https://www.oregon.gov/dor/forms/FormsPubs/form-or-40_101-040_2022.pdf
          https://www.portland.gov/revenue/personal-tax
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Mississippi', 'state_occupation': 'Oregon', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def mississippi_and_pennsylvania(self, county_residence='Other', state_agi=75000, county_occupation='Other', capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Mississippi:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Pennsylvania:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        county_residence : str, optional
            County of residence. Default is Other.
        county_occupation : str, optional
            County of occupation. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Mississippi:
        - https://www.dor.ms.gov/individual/tax-rates
          https://www.dor.ms.gov/individual/tax-rates
          https://www.dor.ms.gov/individual/tax-rates
          https://www.dor.ms.gov/individual
          https://www.dor.ms.gov/individual
        
        Pennsylvania:
        -         
        - https://www.revenue.pa.gov/Tax%20Rates/Pages/default.aspx
          https://www.revenue.pa.gov/FormsandPublications/FormsforIndividuals/PIT/Documents/2022/2022_pa-40in.pdf
          https://www.revenue.pa.gov/FormsandPublications/FormsforBusinesses/EmployerWithholding/Documents/rev-419.pdf
          https://www.revenue.pa.gov/FormsandPublications/FormsforIndividuals/PIT/Documents/2022/2022_pa-40in.pdf
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-25.htm
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        county_occupation = self.county_occupation if hasattr(self, 'county_occupation') else county_occupation
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Mississippi', 'state_occupation': 'Pennsylvania', "county_residence": county_residence, "state_agi": state_agi, "county_occupation": county_occupation, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def mississippi_and_rhode_island(self, state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Mississippi:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Rhode Island:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Mississippi:
        - https://www.dor.ms.gov/individual/tax-rates
          https://www.dor.ms.gov/individual/tax-rates
          https://www.dor.ms.gov/individual/tax-rates
          https://www.dor.ms.gov/individual
          https://www.dor.ms.gov/individual
        
        Rhode Island:
        -         
        - https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2022_1040WE_w_0.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Mississippi', 'state_occupation': 'RhodeIsland', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def mississippi_and_south_carolina(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Mississippi:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        South Carolina:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Mississippi:
        - https://www.dor.ms.gov/individual/tax-rates
          https://www.dor.ms.gov/individual/tax-rates
          https://www.dor.ms.gov/individual/tax-rates
          https://www.dor.ms.gov/individual
          https://www.dor.ms.gov/individual
        
        South Carolina:
        -         
        - https://dor.sc.gov/resources-site/lawandpolicy/Advisory%20Opinions/IL22-15.pdf
          https://dor.sc.gov/forms-site/Forms/IITPacket_2022.pdf
          https://www.scstatehouse.gov/code/t12c006.php
          https://dor.sc.gov/resources-site/lawandpolicy/Advisory%20Opinions/IL22-15.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Mississippi', 'state_occupation': 'SouthCarolina', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def mississippi_and_south_dakota(self, state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Mississippi:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        South Dakota:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Mississippi:
        - https://www.dor.ms.gov/individual/tax-rates
          https://www.dor.ms.gov/individual/tax-rates
          https://www.dor.ms.gov/individual/tax-rates
          https://www.dor.ms.gov/individual
          https://www.dor.ms.gov/individual
        
        South Dakota:
        -         
        - https://dor.sd.gov/individuals/taxes/
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Mississippi', 'state_occupation': 'SouthDakota', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def mississippi_and_tennessee(self, state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Mississippi:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Tennessee:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Mississippi:
        - https://www.dor.ms.gov/individual/tax-rates
          https://www.dor.ms.gov/individual/tax-rates
          https://www.dor.ms.gov/individual/tax-rates
          https://www.dor.ms.gov/individual
          https://www.dor.ms.gov/individual
        
        Tennessee:
        -         
        - https://www.tn.gov/revenue/taxes.html
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Mississippi', 'state_occupation': 'Tennessee', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def mississippi_and_texas(self, state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Mississippi:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Texas:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Mississippi:
        - https://www.dor.ms.gov/individual/tax-rates
          https://www.dor.ms.gov/individual/tax-rates
          https://www.dor.ms.gov/individual/tax-rates
          https://www.dor.ms.gov/individual
          https://www.dor.ms.gov/individual
        
        Texas:
        -         
        - https://comptroller.texas.gov/taxes/
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Mississippi', 'state_occupation': 'Texas', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def mississippi_and_utah(self, state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Mississippi:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Utah:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Mississippi:
        - https://www.dor.ms.gov/individual/tax-rates
          https://www.dor.ms.gov/individual/tax-rates
          https://www.dor.ms.gov/individual/tax-rates
          https://www.dor.ms.gov/individual
          https://www.dor.ms.gov/individual
        
        Utah:
        -         
        - https://incometax.utah.gov/paying/tax-rates
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Mississippi', 'state_occupation': 'Utah', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def mississippi_and_vermont(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Mississippi:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Vermont:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Mississippi:
        - https://www.dor.ms.gov/individual/tax-rates
          https://www.dor.ms.gov/individual/tax-rates
          https://www.dor.ms.gov/individual/tax-rates
          https://www.dor.ms.gov/individual
          https://www.dor.ms.gov/individual
        
        Vermont:
        -         
        - https://tax.vermont.gov/sites/tax/files/documents/GB-1210-2023.pdf
          https://tax.vermont.gov/sites/tax/files/documents/Income%20Booklet-2022.pdf
          https://tax.vermont.gov/sites/tax/files/documents/Income%20Booklet-2022.pdf
          https://tax.vermont.gov/sites/tax/files/documents/Income%20Booklet-2022.pdf
          https://tax.vermont.gov/individuals/personal-income-tax/tax-credits
          https://tax.vermont.gov/individuals/personal-income-tax/taxable-income
          https://tax.vermont.gov/individuals/personal-income-tax/local-option-tax
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Mississippi', 'state_occupation': 'Vermont', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def mississippi_and_virginia(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Mississippi:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Virginia:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Mississippi:
        - https://www.dor.ms.gov/individual/tax-rates
          https://www.dor.ms.gov/individual/tax-rates
          https://www.dor.ms.gov/individual/tax-rates
          https://www.dor.ms.gov/individual
          https://www.dor.ms.gov/individual
        
        Virginia:
        -         
        - https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/reciprocity
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Mississippi', 'state_occupation': 'Virginia', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def mississippi_and_washington(self, state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Mississippi:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Washington:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Mississippi:
        - https://www.dor.ms.gov/individual/tax-rates
          https://www.dor.ms.gov/individual/tax-rates
          https://www.dor.ms.gov/individual/tax-rates
          https://www.dor.ms.gov/individual
          https://www.dor.ms.gov/individual
        
        Washington:
        -         
        - https://dor.wa.gov/taxes-rates/income-tax
          https://dor.wa.gov/taxes-rates/other-taxes/capital-gains-tax
          https://dor.wa.gov/taxes-rates/income-tax
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Mississippi', 'state_occupation': 'Washington', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def mississippi_and_west_virginia(self, county_occupation='Other', state_agi=75000, pay_periods=24, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Mississippi:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        West Virginia:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_occupation : str, optional
            County of occupation. Default is Other.
        pay_periods : int, optional
            Number of pay periods. Options are 13, 24, 26, 52. Default is 24.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Mississippi:
        - https://www.dor.ms.gov/individual/tax-rates
          https://www.dor.ms.gov/individual/tax-rates
          https://www.dor.ms.gov/individual/tax-rates
          https://www.dor.ms.gov/individual
          https://www.dor.ms.gov/individual
        
        West Virginia:
        -         
        - https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://tax.wv.gov/Documents/TSD/tsd438.pdf
          https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2011/Taxes-11-35att.html
        """
        
        # Check if the parameters are set upstream in the instance
        county_occupation = self.county_occupation if hasattr(self, 'county_occupation') else county_occupation
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        pay_periods = self.pay_periods if hasattr(self, 'pay_periods') else pay_periods
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Mississippi', 'state_occupation': 'WestVirginia', "county_occupation": county_occupation, "state_agi": state_agi, "pay_periods": pay_periods, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def mississippi_and_wisconsin(self, state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Mississippi:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Wisconsin:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Mississippi:
        - https://www.dor.ms.gov/individual/tax-rates
          https://www.dor.ms.gov/individual/tax-rates
          https://www.dor.ms.gov/individual/tax-rates
          https://www.dor.ms.gov/individual
          https://www.dor.ms.gov/individual
        
        Wisconsin:
        -         
        - https://www.revenue.wi.gov/TaxForms2023/2023-Form1-ES-Inst.pdf
          https://www.revenue.wi.gov/TaxForms2023/2023-Form1-ES-Inst.pdf
          https://www.revenue.wi.gov/TaxForms2023/2023-Form1-ES-Inst.pdf
          https://www.revenue.wi.gov/Pages/FAQS/pcs-work.aspx
          https://www.revenue.wi.gov/Dor%20Publications/pb103.pdf
          https://www.revenue.wi.gov/Pages/FAQS/pcs-work.aspx
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Mississippi', 'state_occupation': 'Wisconsin', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def mississippi_and_wyoming(self, state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Mississippi:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Wyoming:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Mississippi:
        - https://www.dor.ms.gov/individual/tax-rates
          https://www.dor.ms.gov/individual/tax-rates
          https://www.dor.ms.gov/individual/tax-rates
          https://www.dor.ms.gov/individual
          https://www.dor.ms.gov/individual
        
        Wyoming:
        -         
        - https://revenue.wyo.gov/
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Mississippi', 'state_occupation': 'Wyoming', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def missouri_and_alabama(self, capital_gains_short=0, dependents=0, sources=False, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_long=0, filing_status='single', income=100000, credits_state=0):
        """
        --- Parameters ----
        
        Missouri:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Alabama:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Missouri:
        - https://dor.mo.gov/taxation/individual/tax-types/income/year-changes/
          https://dor.mo.gov/taxation/individual/tax-types/income/year-changes/
          https://dor.mo.gov/taxation/individual/tax-types/income/year-changes/
          https://www.kcmo.gov/city-hall/departments/finance/earnings-tax
          https://www.stlouis-mo.gov/government/departments/collector/earnings-tax/individual-earnings-tax-info.cfm
        
        Alabama:
        -         
        - https://www.revenue.alabama.gov/ultraviewer/viewer/basic_viewer/index.html?form=2023/02/22f40.pdf
          https://www.revenue.alabama.gov/faqs/what-is-alabamas-individual-income-tax-rate/
          https://www.revenue.alabama.gov/ultraviewer/viewer/basic_viewer/index.html?form=2023/02/22f40.pdf
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-24.htm?taxmap=true
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-24.htm?taxmap=true
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-24.htm?taxmap=true
          https://www.revenue.alabama.gov/individual-corporate/income-to-be-reported-on-the-alabama-income-tax-return/
          https://www.almonline.org/TaxRates.aspx#OccupationalTax
        """
        
        # Check if the parameters are set upstream in the instance
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Missouri', 'state_occupation': 'Alabama', "capital_gains_short": capital_gains_short, "dependents": dependents, "sources": sources, "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def missouri_and_alaska(self, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Missouri:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Alaska:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Missouri:
        - https://dor.mo.gov/taxation/individual/tax-types/income/year-changes/
          https://dor.mo.gov/taxation/individual/tax-types/income/year-changes/
          https://dor.mo.gov/taxation/individual/tax-types/income/year-changes/
          https://www.kcmo.gov/city-hall/departments/finance/earnings-tax
          https://www.stlouis-mo.gov/government/departments/collector/earnings-tax/individual-earnings-tax-info.cfm
        
        Alaska:
        -         
        - https://www.commerce.alaska.gov/web/dcra/officeofthestateassessor/alaskataxfacts.aspx
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Missouri', 'state_occupation': 'Alaska', "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def missouri_and_arizona(self, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Missouri:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Arizona:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        credits_state : int, optional
            Credits for state taxes. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Missouri:
        - https://dor.mo.gov/taxation/individual/tax-types/income/year-changes/
          https://dor.mo.gov/taxation/individual/tax-types/income/year-changes/
          https://dor.mo.gov/taxation/individual/tax-types/income/year-changes/
          https://www.kcmo.gov/city-hall/departments/finance/earnings-tax
          https://www.stlouis-mo.gov/government/departments/collector/earnings-tax/individual-earnings-tax-info.cfm
        
        Arizona:
        -         
        - https://azdor.gov/forms/individual/form-140-x-y-tables
          https://azdor.gov/forms/individual/form-140-x-y-tables
          https://azdor.gov/forms/individual/form-140-x-y-tables
          https://azdor.gov/businesses-arizona/withholding-tax/withholding-exceptions
          "https://azdor.gov/individuals/income-tax-filing-assistance/identifying-other-taxable-income#:~:text=In%20regards%20to%20capital%20gains
         the%20individuals%20regular%20tax%20rate.
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Missouri', 'state_occupation': 'Arizona', "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def missouri_and_arkansas(self, capital_gains_short=0, dependents=0, sources=False, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_long=0, filing_status='single', income=100000, credits_state=0):
        """
        --- Parameters ----
        
        Missouri:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Arkansas:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Missouri:
        - https://dor.mo.gov/taxation/individual/tax-types/income/year-changes/
          https://dor.mo.gov/taxation/individual/tax-types/income/year-changes/
          https://dor.mo.gov/taxation/individual/tax-types/income/year-changes/
          https://www.kcmo.gov/city-hall/departments/finance/earnings-tax
          https://www.stlouis-mo.gov/government/departments/collector/earnings-tax/individual-earnings-tax-info.cfm
        
        Arkansas:
        -         
        - https://www.arkansasedc.com/why-arkansas/business-climate/tax-structure/personal-income-tax
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_and_AR1000NR_Instructions.pdf
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_FullYearResidentIndividualIncomeTaxReturn.pdf
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_and_AR1000NR_Instructions.pdf
          https://www.arkansasedc.com/why-arkansas/business-climate/tax-structure/capital-gains-tax-reduction
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_and_AR1000NR_Instructions.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Missouri', 'state_occupation': 'Arkansas', "capital_gains_short": capital_gains_short, "dependents": dependents, "sources": sources, "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def missouri_and_california(self, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Missouri:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        California:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Missouri:
        - https://dor.mo.gov/taxation/individual/tax-types/income/year-changes/
          https://dor.mo.gov/taxation/individual/tax-types/income/year-changes/
          https://dor.mo.gov/taxation/individual/tax-types/income/year-changes/
          https://www.kcmo.gov/city-hall/departments/finance/earnings-tax
          https://www.stlouis-mo.gov/government/departments/collector/earnings-tax/individual-earnings-tax-info.cfm
        
        California:
        -         
        - https://www.ftb.ca.gov/forms/2022/2022-540-tax-rate-schedules.pdf
          https://www.ftb.ca.gov/file/personal/deductions/index.html
          https://www.ftb.ca.gov/about-ftb/newsroom/caleitc/eligibility-and-credit-information.html
          https://www.ftb.ca.gov/file/personal/income-types/capital-gains-and-losses.html#:~:text=California%20does%20not%20have%20a
         are%20taxed%20as%20ordinary%20income
          https://www.ftb.ca.gov/file/personal/
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Missouri', 'state_occupation': 'California', "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def missouri_and_colorado(self, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Missouri:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Colorado:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Missouri:
        - https://dor.mo.gov/taxation/individual/tax-types/income/year-changes/
          https://dor.mo.gov/taxation/individual/tax-types/income/year-changes/
          https://dor.mo.gov/taxation/individual/tax-types/income/year-changes/
          https://www.kcmo.gov/city-hall/departments/finance/earnings-tax
          https://www.stlouis-mo.gov/government/departments/collector/earnings-tax/individual-earnings-tax-info.cfm
        
        Colorado:
        -         
        - https://leg.colorado.gov/agencies/legislative-council-staff/individual-income-tax%C2%A0
          https://leg.colorado.gov/agencies/legislative-council-staff/individual-income-tax%C2%A0
          https://tax.colorado.gov/sites/tax/files/documents/ITT_Capital_Gain_Subtraction_Feb_2022.pdf
          https://www.auroragov.org/business_services/taxes/occupational_privilege_tax
          https://www.denvergov.org/Government/Agencies-Departments-Offices/Agencies-Departments-Offices-Directory/Department-of-Finance/Our-Divisions/Treasury/Business-Tax-Information#:~:text=Occupational%20privilege%20tax
         -%28show%20below%29&text=Each%20taxable%20employee%20is%20liable
         month%20for%20each%20taxable%20employee
          https://www.glendale.co.us/355/Occupational-Privilege-Tax
          https://www.greenwoodvillage.com/1220/Occupational-Privilege-Tax-OPT
          https://ci.sheridan.co.us/288/Occupational-Privilege-Tax
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Missouri', 'state_occupation': 'Colorado', "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def missouri_and_connecticut(self, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Missouri:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Connecticut:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Missouri:
        - https://dor.mo.gov/taxation/individual/tax-types/income/year-changes/
          https://dor.mo.gov/taxation/individual/tax-types/income/year-changes/
          https://dor.mo.gov/taxation/individual/tax-types/income/year-changes/
          https://www.kcmo.gov/city-hall/departments/finance/earnings-tax
          https://www.stlouis-mo.gov/government/departments/collector/earnings-tax/individual-earnings-tax-info.cfm
        
        Connecticut:
        -         
        - https://portal.ct.gov/-/media/DRS/Forms/2022/Income/2022-CT-1040-Instructions_1222.pdf
          https://www.cga.ct.gov/2022/rpt/pdf/2022-R-0108.pdf
          https://www.cga.ct.gov/current/pub/chap_229.htm#sec_12-703
          https://portal.ct.gov/-/media/DRS/Forms/2022/Income/2022-CT-1040-Instructions_1222.pdf
          "https://portal.ct.gov/DRS/Publications/TSSNs/TSSN-29#:~:text=An%20individuals%20net%20capital%20gains
         portion%20of%20Social%20Security%20benefits"
          "https://portal.ct.gov/DRS/Publications/TSSNs/TSSN-29#:~:text=An%20individuals%20net%20capital%20gains
         portion%20of%20Social%20Security%20benefits"
          https://portal.ct.gov/DRS/Individuals/Resident-Income-Tax/Tax-Information
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Missouri', 'state_occupation': 'Connecticut', "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def missouri_and_delaware(self, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, federal_agi=75000, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Missouri:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Delaware:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        federal_agi : int, optional
            Adjusted Gross Income (AGI) for federal taxes. Default is 0. Sources - https://www.irs.gov/e-file-providers/definition-of-adjusted-gross-income, https://www.irs.gov/newsroom/taxpayers-should-know-the-difference-between-standard-and-itemized-deductions
        income : int, optional
            Gross income. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Missouri:
        - https://dor.mo.gov/taxation/individual/tax-types/income/year-changes/
          https://dor.mo.gov/taxation/individual/tax-types/income/year-changes/
          https://dor.mo.gov/taxation/individual/tax-types/income/year-changes/
          https://www.kcmo.gov/city-hall/departments/finance/earnings-tax
          https://www.stlouis-mo.gov/government/departments/collector/earnings-tax/individual-earnings-tax-info.cfm
        
        Delaware:
        -         
        - https://revenue.delaware.gov/software-developer/tax-rate-changes/
          https://revenuefiles.delaware.gov/2022/PIT-RES_TY22_2022-01_PaperInteractive.pdf
          https://revenuefiles.delaware.gov/2022/PIT-RES_TY22_2022-02_Instructions.pdf
          https://revenue.delaware.gov/frequently-asked-questions/personal-income-tax-faqs/
          https://www.newcastlede.gov/DocumentCenter/View/20147/City-of-Wilmington-Wage-Tax-Information-PDF
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        federal_agi = self.federal_agi if hasattr(self, 'federal_agi') else federal_agi
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Missouri', 'state_occupation': 'Delaware', "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "federal_agi": federal_agi, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def missouri_and_florida(self, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Missouri:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Florida:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Missouri:
        - https://dor.mo.gov/taxation/individual/tax-types/income/year-changes/
          https://dor.mo.gov/taxation/individual/tax-types/income/year-changes/
          https://dor.mo.gov/taxation/individual/tax-types/income/year-changes/
          https://www.kcmo.gov/city-hall/departments/finance/earnings-tax
          https://www.stlouis-mo.gov/government/departments/collector/earnings-tax/individual-earnings-tax-info.cfm
        
        Florida:
        -         
        - https://floridarevenue.com/taxes/taxesfees/Pages/tax_interest_rates.aspx
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Missouri', 'state_occupation': 'Florida', "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def missouri_and_georgia(self, capital_gains_short=0, dependents=0, sources=False, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_long=0, filing_status='single', income=100000, credits_state=0):
        """
        --- Parameters ----
        
        Missouri:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Georgia:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Missouri:
        - https://dor.mo.gov/taxation/individual/tax-types/income/year-changes/
          https://dor.mo.gov/taxation/individual/tax-types/income/year-changes/
          https://dor.mo.gov/taxation/individual/tax-types/income/year-changes/
          https://www.kcmo.gov/city-hall/departments/finance/earnings-tax
          https://www.stlouis-mo.gov/government/departments/collector/earnings-tax/individual-earnings-tax-info.cfm
        
        Georgia:
        -         
        - https://dor.georgia.gov/tax-tables-georgia-tax-rate-schedule
          https://apps.dor.ga.gov/FillableForms/PDFViewer/Index?form=2022GA500
          https://apps.dor.ga.gov/FillableForms/PDFViewer/Index?form=2022GA500
          https://apps.dor.ga.gov/FillableForms/PDFViewer/Index?form=2022GA500
          https://dor.georgia.gov/years-individual-income-tax-forms
          https://dor.georgia.gov/years-individual-income-tax-forms
        """
        
        # Check if the parameters are set upstream in the instance
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Missouri', 'state_occupation': 'Georgia', "capital_gains_short": capital_gains_short, "dependents": dependents, "sources": sources, "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def missouri_and_hawaii(self, capital_gains_short=0, dependents=0, sources=False, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_long=0, filing_status='single', income=100000, credits_state=0):
        """
        --- Parameters ----
        
        Missouri:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Hawaii:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Missouri:
        - https://dor.mo.gov/taxation/individual/tax-types/income/year-changes/
          https://dor.mo.gov/taxation/individual/tax-types/income/year-changes/
          https://dor.mo.gov/taxation/individual/tax-types/income/year-changes/
          https://www.kcmo.gov/city-hall/departments/finance/earnings-tax
          https://www.stlouis-mo.gov/government/departments/collector/earnings-tax/individual-earnings-tax-info.cfm
        
        Hawaii:
        -         
        - https://tax.hawaii.gov/forms/d_18table-on/d_18table-on_p13/
          https://files.hawaii.gov/tax/forms/2018/18table-on.pdf
          https://files.hawaii.gov/tax/news/pubs/22outline.pdf
          https://files.hawaii.gov/tax/news/pubs/FastTaxRefGuide_2017.pdf
          https://www.tfhawaii.org/wordpress/state-tax-resources/mini-tax-guides/individual-income-tax-chapter-235/#:~:text=Long%20term%20capital%20gains%20are%20taxed%20at%20a%20maximum%20of%207.25%25
          https://www.grassrootinstitute.org/2023/02/hb337-increasing-tax-on-capital-gains-a-bad-risk/
          https://files.hawaii.gov/tax/news/pubs/22outline.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Missouri', 'state_occupation': 'Hawaii', "capital_gains_short": capital_gains_short, "dependents": dependents, "sources": sources, "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def missouri_and_idaho(self, capital_gains_short=0, dependents=0, sources=False, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_long=0, filing_status='single', income=100000, credits_state=0):
        """
        --- Parameters ----
        
        Missouri:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Idaho:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Missouri:
        - https://dor.mo.gov/taxation/individual/tax-types/income/year-changes/
          https://dor.mo.gov/taxation/individual/tax-types/income/year-changes/
          https://dor.mo.gov/taxation/individual/tax-types/income/year-changes/
          https://www.kcmo.gov/city-hall/departments/finance/earnings-tax
          https://www.stlouis-mo.gov/government/departments/collector/earnings-tax/individual-earnings-tax-info.cfm
        
        Idaho:
        -         
        - https://tax.idaho.gov/taxes/income-tax/individual-income/individual-income-tax-rate-schedule/
          https://tax.idaho.gov/wp-content/uploads/forms/EIN00046/EIN00046_03-01-2023.pdf
          https://tax.idaho.gov/taxes/income-tax/individual-income/idaho-source-income/
          https://tax.idaho.gov/taxes/income-tax/individual-income/idaho-source-income/
        """
        
        # Check if the parameters are set upstream in the instance
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Missouri', 'state_occupation': 'Idaho', "capital_gains_short": capital_gains_short, "dependents": dependents, "sources": sources, "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def missouri_and_illinois(self, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Missouri:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Illinois:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Missouri:
        - https://dor.mo.gov/taxation/individual/tax-types/income/year-changes/
          https://dor.mo.gov/taxation/individual/tax-types/income/year-changes/
          https://dor.mo.gov/taxation/individual/tax-types/income/year-changes/
          https://www.kcmo.gov/city-hall/departments/finance/earnings-tax
          https://www.stlouis-mo.gov/government/departments/collector/earnings-tax/individual-earnings-tax-info.cfm
        
        Illinois:
        -         
        - https://tax.illinois.gov/research/taxrates/income.html#IndividualIncome
          https://tax.illinois.gov/content/dam/soi/en/web/tax/forms/incometax/documents/currentyear/individual/il-1040-instr.pdf
          https://tax.illinois.gov/questionsandanswers/12.html
          https://tax.illinois.gov/content/dam/soi/en/web/tax/forms/incometax/documents/currentyear/individual/il-1040-instr.pdf
          https://tax.illinois.gov/research/taxrates/personalproperty.html
          https://tax.illinois.gov/localgovernments/income.html
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Missouri', 'state_occupation': 'Illinois', "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def missouri_and_indiana(self, capital_gains_short=0, dependents=0, sources=False, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_long=0, filing_status='single', income=100000, credits_state=0):
        """
        --- Parameters ----
        
        Missouri:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Indiana:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Missouri:
        - https://dor.mo.gov/taxation/individual/tax-types/income/year-changes/
          https://dor.mo.gov/taxation/individual/tax-types/income/year-changes/
          https://dor.mo.gov/taxation/individual/tax-types/income/year-changes/
          https://www.kcmo.gov/city-hall/departments/finance/earnings-tax
          https://www.stlouis-mo.gov/government/departments/collector/earnings-tax/individual-earnings-tax-info.cfm
        
        Indiana:
        -         
        - https://www.in.gov/dor/tax-forms/2022-individual-income-tax-forms/
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-32.htm
          https://www.in.gov/dor/files/ib28.pdf
          https://www.in.gov/dor/tax-forms/2022-individual-income-tax-forms/
          https://www.in.gov/dor/files/2022-county-tax-rates-and-codes.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Missouri', 'state_occupation': 'Indiana', "capital_gains_short": capital_gains_short, "dependents": dependents, "sources": sources, "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def missouri_and_iowa(self, capital_gains_short=0, dependents=0, sources=False, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_long=0, filing_status='single', income=100000, credits_state=0):
        """
        --- Parameters ----
        
        Missouri:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Iowa:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Missouri:
        - https://dor.mo.gov/taxation/individual/tax-types/income/year-changes/
          https://dor.mo.gov/taxation/individual/tax-types/income/year-changes/
          https://dor.mo.gov/taxation/individual/tax-types/income/year-changes/
          https://www.kcmo.gov/city-hall/departments/finance/earnings-tax
          https://www.stlouis-mo.gov/government/departments/collector/earnings-tax/individual-earnings-tax-info.cfm
        
        Iowa:
        -         
        - https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
          https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
          https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
          https://tax.iowa.gov/iowa-illinois-reciprocal-agreement
          https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
        """
        
        # Check if the parameters are set upstream in the instance
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Missouri', 'state_occupation': 'Iowa', "capital_gains_short": capital_gains_short, "dependents": dependents, "sources": sources, "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def missouri_and_kansas(self, capital_gains_short=0, dependents=0, sources=False, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_long=0, filing_status='single', income=100000, credits_state=0):
        """
        --- Parameters ----
        
        Missouri:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Kansas:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Missouri:
        - https://dor.mo.gov/taxation/individual/tax-types/income/year-changes/
          https://dor.mo.gov/taxation/individual/tax-types/income/year-changes/
          https://dor.mo.gov/taxation/individual/tax-types/income/year-changes/
          https://www.kcmo.gov/city-hall/departments/finance/earnings-tax
          https://www.stlouis-mo.gov/government/departments/collector/earnings-tax/individual-earnings-tax-info.cfm
        
        Kansas:
        -         
        - https://www.ksrevenue.gov/taxrates.html
          https://www.ksrevenue.gov/incomebook22.html#0
          https://www.ksrevenue.gov/incomebook22.html#0
          https://www.ksrevenue.gov/incomebook22.html#0
          https://www.ksrevenue.gov/incomebook22.html
          https://www.ksrevenue.gov/incomebook22.html
        """
        
        # Check if the parameters are set upstream in the instance
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Missouri', 'state_occupation': 'Kansas', "capital_gains_short": capital_gains_short, "dependents": dependents, "sources": sources, "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def missouri_and_kentucky(self, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Missouri:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Kentucky:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Missouri:
        - https://dor.mo.gov/taxation/individual/tax-types/income/year-changes/
          https://dor.mo.gov/taxation/individual/tax-types/income/year-changes/
          https://dor.mo.gov/taxation/individual/tax-types/income/year-changes/
          https://www.kcmo.gov/city-hall/departments/finance/earnings-tax
          https://www.stlouis-mo.gov/government/departments/collector/earnings-tax/individual-earnings-tax-info.cfm
        
        Kentucky:
        -         
        - https://revenue.ky.gov/News/Pages/Dor-Announces-Updates-to-Individual-Income-Tax-for-2023-Tax-Year.aspx
          https://revenue.ky.gov/Individual/Individual-Income-Tax/Pages/default.aspx
          https://revenue.ky.gov/News/Pages/Dor-Announces-Updates-to-Individual-Income-Tax-for-2023-Tax-Year.aspx
          https://revenue.ky.gov/Individual/Individual-Income-Tax/Pages/default.aspx
          https://revenue.ky.gov/Dor%20Training%20Materials/103%20KAR%2017.140.%20Individual%20income%20tax%20-%20reciprocity%20-%20nonresidents.pdf
          https://revenue.ky.gov/Individual/Individual-Income-Tax/Pages/default.aspx
          https://nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2021/Taxes-21-32.htm?taxmap=true
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Missouri', 'state_occupation': 'Kentucky', "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def missouri_and_louisiana(self, capital_gains_short=0, dependents=0, sources=False, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_long=0, filing_status='single', income=100000, credits_state=0):
        """
        --- Parameters ----
        
        Missouri:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Louisiana:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Missouri:
        - https://dor.mo.gov/taxation/individual/tax-types/income/year-changes/
          https://dor.mo.gov/taxation/individual/tax-types/income/year-changes/
          https://dor.mo.gov/taxation/individual/tax-types/income/year-changes/
          https://www.kcmo.gov/city-hall/departments/finance/earnings-tax
          https://www.stlouis-mo.gov/government/departments/collector/earnings-tax/individual-earnings-tax-info.cfm
        
        Louisiana:
        -         
        - https://revenue.louisiana.gov/individualincometax
          https://revenue.louisiana.gov/individualincometax
          https://www.revenue.louisiana.gov/taxforms/6935(11_02)F.pdf
          https://revenue.louisiana.gov/individualincometax
          https://revenue.louisiana.gov/individualincometax
        """
        
        # Check if the parameters are set upstream in the instance
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Missouri', 'state_occupation': 'Louisiana', "capital_gains_short": capital_gains_short, "dependents": dependents, "sources": sources, "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def missouri_and_maine(self, capital_gains_short=0, dependents=0, sources=False, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_long=0, filing_status='single', income=100000, credits_state=0):
        """
        --- Parameters ----
        
        Missouri:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Maine:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Missouri:
        - https://dor.mo.gov/taxation/individual/tax-types/income/year-changes/
          https://dor.mo.gov/taxation/individual/tax-types/income/year-changes/
          https://dor.mo.gov/taxation/individual/tax-types/income/year-changes/
          https://www.kcmo.gov/city-hall/departments/finance/earnings-tax
          https://www.stlouis-mo.gov/government/departments/collector/earnings-tax/individual-earnings-tax-info.cfm
        
        Maine:
        -         
        - https://www.maine.gov/revenue/tax-return-forms/individual-income-tax-2022
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/ind_tax_rate_sched_2022.pdf
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/22_1040me_book_gen_instruc_revisedFeb23.pdf
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/22_1040me_book_gen_instruc_revisedFeb23.pdf
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/22_1040me_book_gen_instruc_revisedFeb23.pdf
          https://www.maine.gov/revenue/tax-return-forms/individual-income-tax-2022
        """
        
        # Check if the parameters are set upstream in the instance
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Missouri', 'state_occupation': 'Maine', "capital_gains_short": capital_gains_short, "dependents": dependents, "sources": sources, "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def missouri_and_maryland(self, capital_gains_short=0, dependents=0, sources=False, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_long=0, filing_status='single', income=100000, credits_state=0):
        """
        --- Parameters ----
        
        Missouri:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Maryland:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Missouri:
        - https://dor.mo.gov/taxation/individual/tax-types/income/year-changes/
          https://dor.mo.gov/taxation/individual/tax-types/income/year-changes/
          https://dor.mo.gov/taxation/individual/tax-types/income/year-changes/
          https://www.kcmo.gov/city-hall/departments/finance/earnings-tax
          https://www.stlouis-mo.gov/government/departments/collector/earnings-tax/individual-earnings-tax-info.cfm
        
        Maryland:
        -         
        - https://www.marylandtaxes.gov/individual/income/tax-info/tax-rates.php
          https://www.marylandtaxes.gov/forms/22_forms/Resident_Booklet.pdf
          https://www.marylandtaxes.gov/forms/Tax_Publications/Administrative_Releases/Income_and_Estate_Tax_Releases/ar_it37.pdf
          https://www.marylandtaxes.gov/individual/index.php
          https://www.marylandtaxes.gov/individual/income/tax-info/tax-rates.php
        """
        
        # Check if the parameters are set upstream in the instance
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Missouri', 'state_occupation': 'Maryland', "capital_gains_short": capital_gains_short, "dependents": dependents, "sources": sources, "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def missouri_and_massachusetts(self, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Missouri:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Massachusetts:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Missouri:
        - https://dor.mo.gov/taxation/individual/tax-types/income/year-changes/
          https://dor.mo.gov/taxation/individual/tax-types/income/year-changes/
          https://dor.mo.gov/taxation/individual/tax-types/income/year-changes/
          https://www.kcmo.gov/city-hall/departments/finance/earnings-tax
          https://www.stlouis-mo.gov/government/departments/collector/earnings-tax/individual-earnings-tax-info.cfm
        
        Massachusetts:
        -         
        - https://www.mass.gov/service-details/massachusetts-tax-rates
          https://www.mass.gov/service-details/view-massachusetts-personal-income-tax-exemptions
          https://www.mass.gov/service-details/view-massachusetts-personal-income-tax-exemptions
          https://www.mass.gov/service-details/view-child-and-dependent-related-credits
          https://www.mass.gov/service-details/massachusetts-tax-rates
          https://www.mass.gov/personal-income-tax
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Missouri', 'state_occupation': 'Massachusetts', "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def missouri_and_michigan(self, capital_gains_short=0, county_occupation='Other', dependents=0, sources=False, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_long=0, filing_status='single', income=100000, credits_state=0):
        """
        --- Parameters ----
        
        Missouri:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Michigan:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        county_occupation : str, optional
            County of occupation. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Missouri:
        - https://dor.mo.gov/taxation/individual/tax-types/income/year-changes/
          https://dor.mo.gov/taxation/individual/tax-types/income/year-changes/
          https://dor.mo.gov/taxation/individual/tax-types/income/year-changes/
          https://www.kcmo.gov/city-hall/departments/finance/earnings-tax
          https://www.stlouis-mo.gov/government/departments/collector/earnings-tax/individual-earnings-tax-info.cfm
        
        Michigan:
        -         
        - https://www.michigan.gov/taxes/iit/new-developments-for-tax-year-2022
          "https://www.michigan.gov/taxes/-/media/Project/Websites/taxes/Forms/2023/2023-SUW/446_Withholding-Guide_2023.pdf?rev=177ca3d8cb034392bf84e317ed78c320&hash=19C5F42F43A89A91863D229349F48883#:~:text=The%20exemption%20amount%20is%20%245%2C400
         employees%20Michigan%20income%20tax%20return"
          "https://www.michigan.gov/taxes/-/media/Project/Websites/taxes/Forms/2023/2023-SUW/446_Withholding-Guide_2023.pdf?rev=177ca3d8cb034392bf84e317ed78c320&hash=19C5F42F43A89A91863D229349F48883#:~:text=The%20exemption%20amount%20is%20%245%2C400
         employees%20Michigan%20income%20tax%20return"
          https://www.michigan.gov/taxes/questions/iit/accordion/residency/are-my-wages-earned-in-another-state-taxable-in-michigan-if-i-am-a-michigan-resident-1
          https://www.michigan.gov/taxes/-/media/Project/Websites/taxes/Forms/2022/2022-IIT-Forms/BOOK_MI-1040.pdf?rev=4524f1cbbc0243b194ab05a6680be75b&hash=40DFE71E651F7C1D65C31F6CD03D944F
          https://www.michigan.gov/taxes/questions/iit/accordion/general/what-cities-impose-an-income-tax
        """
        
        # Check if the parameters are set upstream in the instance
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        county_occupation = self.county_occupation if hasattr(self, 'county_occupation') else county_occupation
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Missouri', 'state_occupation': 'Michigan', "capital_gains_short": capital_gains_short, "county_occupation": county_occupation, "dependents": dependents, "sources": sources, "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def missouri_and_minnesota(self, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Missouri:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Minnesota:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Missouri:
        - https://dor.mo.gov/taxation/individual/tax-types/income/year-changes/
          https://dor.mo.gov/taxation/individual/tax-types/income/year-changes/
          https://dor.mo.gov/taxation/individual/tax-types/income/year-changes/
          https://www.kcmo.gov/city-hall/departments/finance/earnings-tax
          https://www.stlouis-mo.gov/government/departments/collector/earnings-tax/individual-earnings-tax-info.cfm
        
        Minnesota:
        -         
        - https://www.revenue.state.mn.us/minnesota-income-tax-rates-and-brackets
          https://www.revenue.state.mn.us/child-and-dependent-care-credit
          https://www.revenue.state.mn.us/dependent-exemptions
          https://www.revenue.state.mn.us/minnesota-standard-deduction
          https://mn.gov/mmb/assets/mwr_form_tcm1059-128581.pdf
          https://www.revenue.state.mn.us/individual-income-tax
          https://www.revenue.state.mn.us/individual-income-tax
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Missouri', 'state_occupation': 'Minnesota', "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def missouri_and_mississippi(self, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Missouri:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Mississippi:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Missouri:
        - https://dor.mo.gov/taxation/individual/tax-types/income/year-changes/
          https://dor.mo.gov/taxation/individual/tax-types/income/year-changes/
          https://dor.mo.gov/taxation/individual/tax-types/income/year-changes/
          https://www.kcmo.gov/city-hall/departments/finance/earnings-tax
          https://www.stlouis-mo.gov/government/departments/collector/earnings-tax/individual-earnings-tax-info.cfm
        
        Mississippi:
        -         
        - https://www.dor.ms.gov/individual/tax-rates
          https://www.dor.ms.gov/individual/tax-rates
          https://www.dor.ms.gov/individual/tax-rates
          https://www.dor.ms.gov/individual
          https://www.dor.ms.gov/individual
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Missouri', 'state_occupation': 'Mississippi', "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def missouri_and_montana(self, capital_gains_short=0, dependents=0, sources=False, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_long=0, filing_status='single', income=100000, credits_state=0):
        """
        --- Parameters ----
        
        Missouri:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Montana:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Missouri:
        - https://dor.mo.gov/taxation/individual/tax-types/income/year-changes/
          https://dor.mo.gov/taxation/individual/tax-types/income/year-changes/
          https://dor.mo.gov/taxation/individual/tax-types/income/year-changes/
          https://www.kcmo.gov/city-hall/departments/finance/earnings-tax
          https://www.stlouis-mo.gov/government/departments/collector/earnings-tax/individual-earnings-tax-info.cfm
        
        Montana:
        -         
        - https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/2022/12/montana-employees-withholding-allowance-and-exemption-certificate-form-mw-4.pdf
          https://mtrevenue.gov/2021/11/29/simplification-of-montana-income-taxation/?utm_source=rss&utm_medium=rss&utm_campaign=simplification-of-montana-income-taxation
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Missouri', 'state_occupation': 'Montana', "capital_gains_short": capital_gains_short, "dependents": dependents, "sources": sources, "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def missouri_and_nebraska(self, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Missouri:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Nebraska:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Missouri:
        - https://dor.mo.gov/taxation/individual/tax-types/income/year-changes/
          https://dor.mo.gov/taxation/individual/tax-types/income/year-changes/
          https://dor.mo.gov/taxation/individual/tax-types/income/year-changes/
          https://www.kcmo.gov/city-hall/departments/finance/earnings-tax
          https://www.stlouis-mo.gov/government/departments/collector/earnings-tax/individual-earnings-tax-info.cfm
        
        Nebraska:
        -         
        - https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2023_Circular_EN_8-429r2023_Whole_Book_11.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Missouri', 'state_occupation': 'Nebraska', "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def missouri_and_nevada(self, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Missouri:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Nevada:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Missouri:
        - https://dor.mo.gov/taxation/individual/tax-types/income/year-changes/
          https://dor.mo.gov/taxation/individual/tax-types/income/year-changes/
          https://dor.mo.gov/taxation/individual/tax-types/income/year-changes/
          https://www.kcmo.gov/city-hall/departments/finance/earnings-tax
          https://www.stlouis-mo.gov/government/departments/collector/earnings-tax/individual-earnings-tax-info.cfm
        
        Nevada:
        -         
        - https://www.nvsos.gov/sos/businesses/the-nevada-advantage
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Missouri', 'state_occupation': 'Nevada', "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def missouri_and_new_hampshire(self, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Missouri:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        New Hampshire:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Missouri:
        - https://dor.mo.gov/taxation/individual/tax-types/income/year-changes/
          https://dor.mo.gov/taxation/individual/tax-types/income/year-changes/
          https://dor.mo.gov/taxation/individual/tax-types/income/year-changes/
          https://www.kcmo.gov/city-hall/departments/finance/earnings-tax
          https://www.stlouis-mo.gov/government/departments/collector/earnings-tax/individual-earnings-tax-info.cfm
        
        New Hampshire:
        -         
        - https://www.revenue.nh.gov/assistance/tax-overview.htm
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Missouri', 'state_occupation': 'NewHampshire', "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def missouri_and_new_jersey(self, capital_gains_short=0, dependents=0, sources=False, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_long=0, filing_status='single', income=100000, credits_state=0):
        """
        --- Parameters ----
        
        Missouri:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        New Jersey:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Missouri:
        - https://dor.mo.gov/taxation/individual/tax-types/income/year-changes/
          https://dor.mo.gov/taxation/individual/tax-types/income/year-changes/
          https://dor.mo.gov/taxation/individual/tax-types/income/year-changes/
          https://www.kcmo.gov/city-hall/departments/finance/earnings-tax
          https://www.stlouis-mo.gov/government/departments/collector/earnings-tax/individual-earnings-tax-info.cfm
        
        New Jersey:
        -         
        - https://www.state.nj.us/treasury/taxation/taxtables.shtml
          https://www.state.nj.us/treasury/taxation/njit13.shtml
          https://www.nj.gov/treasury/taxation/njit25.shtml
          https://www.state.nj.us/treasury/taxation/individuals/freqqiti.shtml
          https://www.state.nj.us/treasury/taxation/localtax.shtml
        """
        
        # Check if the parameters are set upstream in the instance
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Missouri', 'state_occupation': 'NewJersey', "capital_gains_short": capital_gains_short, "dependents": dependents, "sources": sources, "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def missouri_and_new_mexico(self, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Missouri:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        New Mexico:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Missouri:
        - https://dor.mo.gov/taxation/individual/tax-types/income/year-changes/
          https://dor.mo.gov/taxation/individual/tax-types/income/year-changes/
          https://dor.mo.gov/taxation/individual/tax-types/income/year-changes/
          https://www.kcmo.gov/city-hall/departments/finance/earnings-tax
          https://www.stlouis-mo.gov/government/departments/collector/earnings-tax/individual-earnings-tax-info.cfm
        
        New Mexico:
        -         
        - https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/fdf3c548-8aba-4b9c-9eb4-bb564c716015/FYI-104.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/0bf943fd-652e-4400-bb1b-cd8397eb5a95/2022pit-adj.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/2f1a6781-9534-4436-b427-1557f9592099/2022pit-adj-ins.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/0bf943fd-652e-4400-bb1b-cd8397eb5a95/2022pit-adj.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/1afc56af-ea90-4d48-82e5-1f9aeb43255a/PITbook2022.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/fdf3c548-8aba-4b9c-9eb4-bb564c716015/FYI-104.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Missouri', 'state_occupation': 'NewMexico', "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def missouri_and_new_york(self, capital_gains_short=0, dependents=0, sources=False, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_long=0, filing_status='single', income=100000, credits_state=0):
        """
        --- Parameters ----
        
        Missouri:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        New York:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Missouri:
        - https://dor.mo.gov/taxation/individual/tax-types/income/year-changes/
          https://dor.mo.gov/taxation/individual/tax-types/income/year-changes/
          https://dor.mo.gov/taxation/individual/tax-types/income/year-changes/
          https://www.kcmo.gov/city-hall/departments/finance/earnings-tax
          https://www.stlouis-mo.gov/government/departments/collector/earnings-tax/individual-earnings-tax-info.cfm
        
        New York:
        -         
        - https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nys-tax-rate-schedule
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nys-tax-rate-schedule
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nys-tax-rate-schedule
          https://www.tax.ny.gov/pit/
          https://www.tax.ny.gov/pit/file/nonresident-faqs.htm
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#yonkers-tax-rate-schedule
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nyc-tax-rate-schedule
        """
        
        # Check if the parameters are set upstream in the instance
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Missouri', 'state_occupation': 'NewYork', "capital_gains_short": capital_gains_short, "dependents": dependents, "sources": sources, "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def missouri_and_north_carolina(self, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Missouri:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        North Carolina:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Missouri:
        - https://dor.mo.gov/taxation/individual/tax-types/income/year-changes/
          https://dor.mo.gov/taxation/individual/tax-types/income/year-changes/
          https://dor.mo.gov/taxation/individual/tax-types/income/year-changes/
          https://www.kcmo.gov/city-hall/departments/finance/earnings-tax
          https://www.stlouis-mo.gov/government/departments/collector/earnings-tax/individual-earnings-tax-info.cfm
        
        North Carolina:
        -         
        - https://www.ncdor.gov/taxes-forms/tax-rate-schedules
          https://www.ncdor.gov/taxes-forms/individual-income-tax/north-carolina-standard-deduction-or-north-carolina-itemized-deductions
          https://www.ncdor.gov/taxes-forms/individual-income-tax
          https://www.ncdor.gov/taxes-forms/individual-income-tax
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Missouri', 'state_occupation': 'NorthCarolina', "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def missouri_and_north_dakota(self, capital_gains_short=0, dependents=0, sources=False, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_long=0, filing_status='single', income=100000, credits_state=0):
        """
        --- Parameters ----
        
        Missouri:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        North Dakota:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Missouri:
        - https://dor.mo.gov/taxation/individual/tax-types/income/year-changes/
          https://dor.mo.gov/taxation/individual/tax-types/income/year-changes/
          https://dor.mo.gov/taxation/individual/tax-types/income/year-changes/
          https://www.kcmo.gov/city-hall/departments/finance/earnings-tax
          https://www.stlouis-mo.gov/government/departments/collector/earnings-tax/individual-earnings-tax-info.cfm
        
        North Dakota:
        -         
        - https://www.tax.nd.gov/forms
          https://www.tax.nd.gov/sites/www/files/documents/forms/individual/2022-iit/2022-individual-income-tax-booklet.pdf
          https://www.tax.nd.gov/sites/www/files/documents/forms/2022-individual-income-tax-booklet.pdf
          https://www.tax.nd.gov/tax-exemptions-credits/income-tax-exemptions-credits
          https://www.tax.nd.gov/sites/www/files/documents/forms/business/ndwrfillable.pdf
          https://www.tax.nd.gov/news/tax-legislative-changes/significant-changes-law/individual-income-tax-history
          https://www.tax.nd.gov/forms
        """
        
        # Check if the parameters are set upstream in the instance
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Missouri', 'state_occupation': 'NorthDakota', "capital_gains_short": capital_gains_short, "dependents": dependents, "sources": sources, "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def missouri_and_ohio(self, capital_gains_short=0, dependents=0, sources=False, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_long=0, filing_status='single', income=100000, credits_state=0):
        """
        --- Parameters ----
        
        Missouri:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Ohio:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Missouri:
        - https://dor.mo.gov/taxation/individual/tax-types/income/year-changes/
          https://dor.mo.gov/taxation/individual/tax-types/income/year-changes/
          https://dor.mo.gov/taxation/individual/tax-types/income/year-changes/
          https://www.kcmo.gov/city-hall/departments/finance/earnings-tax
          https://www.stlouis-mo.gov/government/departments/collector/earnings-tax/individual-earnings-tax-info.cfm
        
        Ohio:
        -         
        - https://tax.ohio.gov/individual/resources/annual-tax-rates
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://tax.ohio.gov/static/forms/employer_withholding/generic/wth_it4nr.pdf
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://www.ritaohio.com/TaxRatesTable
        """
        
        # Check if the parameters are set upstream in the instance
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Missouri', 'state_occupation': 'Ohio', "capital_gains_short": capital_gains_short, "dependents": dependents, "sources": sources, "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def missouri_and_oklahoma(self, capital_gains_short=0, dependents=0, sources=False, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_long=0, filing_status='single', income=100000, credits_state=0):
        """
        --- Parameters ----
        
        Missouri:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Oklahoma:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Missouri:
        - https://dor.mo.gov/taxation/individual/tax-types/income/year-changes/
          https://dor.mo.gov/taxation/individual/tax-types/income/year-changes/
          https://dor.mo.gov/taxation/individual/tax-types/income/year-changes/
          https://www.kcmo.gov/city-hall/departments/finance/earnings-tax
          https://www.stlouis-mo.gov/government/departments/collector/earnings-tax/individual-earnings-tax-info.cfm
        
        Oklahoma:
        -         
        - https://oklahoma.gov/content/dam/ok/en/tax/documents/resources/publications/businesses/withholding-tables/WHTables-2022.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Missouri', 'state_occupation': 'Oklahoma', "capital_gains_short": capital_gains_short, "dependents": dependents, "sources": sources, "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def missouri_and_oregon(self, capital_gains_short=0, dependents=0, sources=False, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_long=0, filing_status='single', income=100000, credits_state=0):
        """
        --- Parameters ----
        
        Missouri:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Oregon:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Missouri:
        - https://dor.mo.gov/taxation/individual/tax-types/income/year-changes/
          https://dor.mo.gov/taxation/individual/tax-types/income/year-changes/
          https://dor.mo.gov/taxation/individual/tax-types/income/year-changes/
          https://www.kcmo.gov/city-hall/departments/finance/earnings-tax
          https://www.stlouis-mo.gov/government/departments/collector/earnings-tax/individual-earnings-tax-info.cfm
        
        Oregon:
        -         
        - https://www.oregon.gov/dor/programs/individuals/Pages/PIT.aspx
          https://www.oregon.gov/dor/programs/individuals/Documents/Full%20year%20resident
         %20Form%20or-40%20filers.pdf
          https://www.oregon.gov/dor/programs/individuals/Documents/Part-year%20and%20nonresident
         %20Form%20or-40-P%20and%20or-40-N%20filers.pdf.pdf
          https://www.oregon.gov/dor/forms/FormsPubs/form-or-40_101-040_2022.pdf
          https://www.oregon.gov/dor/forms/FormsPubs/form-or-40_101-040_2022.pdf
          https://www.oregon.gov/dor/forms/FormsPubs/form-or-40_101-040_2022.pdf
          https://www.portland.gov/revenue/personal-tax
        """
        
        # Check if the parameters are set upstream in the instance
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Missouri', 'state_occupation': 'Oregon', "capital_gains_short": capital_gains_short, "dependents": dependents, "sources": sources, "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def missouri_and_pennsylvania(self, county_residence='Other', state_agi=75000, county_occupation='Other', filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Missouri:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Pennsylvania:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        county_residence : str, optional
            County of residence. Default is Other.
        county_occupation : str, optional
            County of occupation. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Missouri:
        - https://dor.mo.gov/taxation/individual/tax-types/income/year-changes/
          https://dor.mo.gov/taxation/individual/tax-types/income/year-changes/
          https://dor.mo.gov/taxation/individual/tax-types/income/year-changes/
          https://www.kcmo.gov/city-hall/departments/finance/earnings-tax
          https://www.stlouis-mo.gov/government/departments/collector/earnings-tax/individual-earnings-tax-info.cfm
        
        Pennsylvania:
        -         
        - https://www.revenue.pa.gov/Tax%20Rates/Pages/default.aspx
          https://www.revenue.pa.gov/FormsandPublications/FormsforIndividuals/PIT/Documents/2022/2022_pa-40in.pdf
          https://www.revenue.pa.gov/FormsandPublications/FormsforBusinesses/EmployerWithholding/Documents/rev-419.pdf
          https://www.revenue.pa.gov/FormsandPublications/FormsforIndividuals/PIT/Documents/2022/2022_pa-40in.pdf
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-25.htm
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        county_occupation = self.county_occupation if hasattr(self, 'county_occupation') else county_occupation
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Missouri', 'state_occupation': 'Pennsylvania', "county_residence": county_residence, "state_agi": state_agi, "county_occupation": county_occupation, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def missouri_and_rhode_island(self, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Missouri:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Rhode Island:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Missouri:
        - https://dor.mo.gov/taxation/individual/tax-types/income/year-changes/
          https://dor.mo.gov/taxation/individual/tax-types/income/year-changes/
          https://dor.mo.gov/taxation/individual/tax-types/income/year-changes/
          https://www.kcmo.gov/city-hall/departments/finance/earnings-tax
          https://www.stlouis-mo.gov/government/departments/collector/earnings-tax/individual-earnings-tax-info.cfm
        
        Rhode Island:
        -         
        - https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2022_1040WE_w_0.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Missouri', 'state_occupation': 'RhodeIsland', "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def missouri_and_south_carolina(self, capital_gains_short=0, dependents=0, sources=False, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_long=0, filing_status='single', income=100000, credits_state=0):
        """
        --- Parameters ----
        
        Missouri:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        South Carolina:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Missouri:
        - https://dor.mo.gov/taxation/individual/tax-types/income/year-changes/
          https://dor.mo.gov/taxation/individual/tax-types/income/year-changes/
          https://dor.mo.gov/taxation/individual/tax-types/income/year-changes/
          https://www.kcmo.gov/city-hall/departments/finance/earnings-tax
          https://www.stlouis-mo.gov/government/departments/collector/earnings-tax/individual-earnings-tax-info.cfm
        
        South Carolina:
        -         
        - https://dor.sc.gov/resources-site/lawandpolicy/Advisory%20Opinions/IL22-15.pdf
          https://dor.sc.gov/forms-site/Forms/IITPacket_2022.pdf
          https://www.scstatehouse.gov/code/t12c006.php
          https://dor.sc.gov/resources-site/lawandpolicy/Advisory%20Opinions/IL22-15.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Missouri', 'state_occupation': 'SouthCarolina', "capital_gains_short": capital_gains_short, "dependents": dependents, "sources": sources, "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def missouri_and_south_dakota(self, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Missouri:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        South Dakota:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Missouri:
        - https://dor.mo.gov/taxation/individual/tax-types/income/year-changes/
          https://dor.mo.gov/taxation/individual/tax-types/income/year-changes/
          https://dor.mo.gov/taxation/individual/tax-types/income/year-changes/
          https://www.kcmo.gov/city-hall/departments/finance/earnings-tax
          https://www.stlouis-mo.gov/government/departments/collector/earnings-tax/individual-earnings-tax-info.cfm
        
        South Dakota:
        -         
        - https://dor.sd.gov/individuals/taxes/
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Missouri', 'state_occupation': 'SouthDakota', "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def missouri_and_tennessee(self, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Missouri:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Tennessee:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Missouri:
        - https://dor.mo.gov/taxation/individual/tax-types/income/year-changes/
          https://dor.mo.gov/taxation/individual/tax-types/income/year-changes/
          https://dor.mo.gov/taxation/individual/tax-types/income/year-changes/
          https://www.kcmo.gov/city-hall/departments/finance/earnings-tax
          https://www.stlouis-mo.gov/government/departments/collector/earnings-tax/individual-earnings-tax-info.cfm
        
        Tennessee:
        -         
        - https://www.tn.gov/revenue/taxes.html
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Missouri', 'state_occupation': 'Tennessee', "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def missouri_and_texas(self, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Missouri:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Texas:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Missouri:
        - https://dor.mo.gov/taxation/individual/tax-types/income/year-changes/
          https://dor.mo.gov/taxation/individual/tax-types/income/year-changes/
          https://dor.mo.gov/taxation/individual/tax-types/income/year-changes/
          https://www.kcmo.gov/city-hall/departments/finance/earnings-tax
          https://www.stlouis-mo.gov/government/departments/collector/earnings-tax/individual-earnings-tax-info.cfm
        
        Texas:
        -         
        - https://comptroller.texas.gov/taxes/
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Missouri', 'state_occupation': 'Texas', "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def missouri_and_utah(self, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Missouri:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Utah:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Missouri:
        - https://dor.mo.gov/taxation/individual/tax-types/income/year-changes/
          https://dor.mo.gov/taxation/individual/tax-types/income/year-changes/
          https://dor.mo.gov/taxation/individual/tax-types/income/year-changes/
          https://www.kcmo.gov/city-hall/departments/finance/earnings-tax
          https://www.stlouis-mo.gov/government/departments/collector/earnings-tax/individual-earnings-tax-info.cfm
        
        Utah:
        -         
        - https://incometax.utah.gov/paying/tax-rates
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Missouri', 'state_occupation': 'Utah', "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def missouri_and_vermont(self, capital_gains_short=0, dependents=0, sources=False, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_long=0, filing_status='single', income=100000, credits_state=0):
        """
        --- Parameters ----
        
        Missouri:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Vermont:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Missouri:
        - https://dor.mo.gov/taxation/individual/tax-types/income/year-changes/
          https://dor.mo.gov/taxation/individual/tax-types/income/year-changes/
          https://dor.mo.gov/taxation/individual/tax-types/income/year-changes/
          https://www.kcmo.gov/city-hall/departments/finance/earnings-tax
          https://www.stlouis-mo.gov/government/departments/collector/earnings-tax/individual-earnings-tax-info.cfm
        
        Vermont:
        -         
        - https://tax.vermont.gov/sites/tax/files/documents/GB-1210-2023.pdf
          https://tax.vermont.gov/sites/tax/files/documents/Income%20Booklet-2022.pdf
          https://tax.vermont.gov/sites/tax/files/documents/Income%20Booklet-2022.pdf
          https://tax.vermont.gov/sites/tax/files/documents/Income%20Booklet-2022.pdf
          https://tax.vermont.gov/individuals/personal-income-tax/tax-credits
          https://tax.vermont.gov/individuals/personal-income-tax/taxable-income
          https://tax.vermont.gov/individuals/personal-income-tax/local-option-tax
        """
        
        # Check if the parameters are set upstream in the instance
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Missouri', 'state_occupation': 'Vermont', "capital_gains_short": capital_gains_short, "dependents": dependents, "sources": sources, "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def missouri_and_virginia(self, capital_gains_short=0, dependents=0, sources=False, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_long=0, filing_status='single', income=100000, credits_state=0):
        """
        --- Parameters ----
        
        Missouri:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Virginia:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Missouri:
        - https://dor.mo.gov/taxation/individual/tax-types/income/year-changes/
          https://dor.mo.gov/taxation/individual/tax-types/income/year-changes/
          https://dor.mo.gov/taxation/individual/tax-types/income/year-changes/
          https://www.kcmo.gov/city-hall/departments/finance/earnings-tax
          https://www.stlouis-mo.gov/government/departments/collector/earnings-tax/individual-earnings-tax-info.cfm
        
        Virginia:
        -         
        - https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/reciprocity
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Missouri', 'state_occupation': 'Virginia', "capital_gains_short": capital_gains_short, "dependents": dependents, "sources": sources, "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def missouri_and_washington(self, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Missouri:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Washington:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Missouri:
        - https://dor.mo.gov/taxation/individual/tax-types/income/year-changes/
          https://dor.mo.gov/taxation/individual/tax-types/income/year-changes/
          https://dor.mo.gov/taxation/individual/tax-types/income/year-changes/
          https://www.kcmo.gov/city-hall/departments/finance/earnings-tax
          https://www.stlouis-mo.gov/government/departments/collector/earnings-tax/individual-earnings-tax-info.cfm
        
        Washington:
        -         
        - https://dor.wa.gov/taxes-rates/income-tax
          https://dor.wa.gov/taxes-rates/other-taxes/capital-gains-tax
          https://dor.wa.gov/taxes-rates/income-tax
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Missouri', 'state_occupation': 'Washington', "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def missouri_and_west_virginia(self, pay_periods=24, capital_gains_short=0, county_occupation='Other', dependents=0, sources=False, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_long=0, filing_status='single', income=100000, credits_state=0):
        """
        --- Parameters ----
        
        Missouri:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        West Virginia:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_occupation : str, optional
            County of occupation. Default is Other.
        pay_periods : int, optional
            Number of pay periods. Options are 13, 24, 26, 52. Default is 24.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Missouri:
        - https://dor.mo.gov/taxation/individual/tax-types/income/year-changes/
          https://dor.mo.gov/taxation/individual/tax-types/income/year-changes/
          https://dor.mo.gov/taxation/individual/tax-types/income/year-changes/
          https://www.kcmo.gov/city-hall/departments/finance/earnings-tax
          https://www.stlouis-mo.gov/government/departments/collector/earnings-tax/individual-earnings-tax-info.cfm
        
        West Virginia:
        -         
        - https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://tax.wv.gov/Documents/TSD/tsd438.pdf
          https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2011/Taxes-11-35att.html
        """
        
        # Check if the parameters are set upstream in the instance
        pay_periods = self.pay_periods if hasattr(self, 'pay_periods') else pay_periods
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        county_occupation = self.county_occupation if hasattr(self, 'county_occupation') else county_occupation
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Missouri', 'state_occupation': 'WestVirginia', "pay_periods": pay_periods, "capital_gains_short": capital_gains_short, "county_occupation": county_occupation, "dependents": dependents, "sources": sources, "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def missouri_and_wisconsin(self, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Missouri:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Wisconsin:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Missouri:
        - https://dor.mo.gov/taxation/individual/tax-types/income/year-changes/
          https://dor.mo.gov/taxation/individual/tax-types/income/year-changes/
          https://dor.mo.gov/taxation/individual/tax-types/income/year-changes/
          https://www.kcmo.gov/city-hall/departments/finance/earnings-tax
          https://www.stlouis-mo.gov/government/departments/collector/earnings-tax/individual-earnings-tax-info.cfm
        
        Wisconsin:
        -         
        - https://www.revenue.wi.gov/TaxForms2023/2023-Form1-ES-Inst.pdf
          https://www.revenue.wi.gov/TaxForms2023/2023-Form1-ES-Inst.pdf
          https://www.revenue.wi.gov/TaxForms2023/2023-Form1-ES-Inst.pdf
          https://www.revenue.wi.gov/Pages/FAQS/pcs-work.aspx
          https://www.revenue.wi.gov/Dor%20Publications/pb103.pdf
          https://www.revenue.wi.gov/Pages/FAQS/pcs-work.aspx
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Missouri', 'state_occupation': 'Wisconsin', "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def missouri_and_wyoming(self, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Missouri:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Wyoming:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Missouri:
        - https://dor.mo.gov/taxation/individual/tax-types/income/year-changes/
          https://dor.mo.gov/taxation/individual/tax-types/income/year-changes/
          https://dor.mo.gov/taxation/individual/tax-types/income/year-changes/
          https://www.kcmo.gov/city-hall/departments/finance/earnings-tax
          https://www.stlouis-mo.gov/government/departments/collector/earnings-tax/individual-earnings-tax-info.cfm
        
        Wyoming:
        -         
        - https://revenue.wyo.gov/
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Missouri', 'state_occupation': 'Wyoming', "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def montana_and_alabama(self, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Montana:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Alabama:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Montana:
        - https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/2022/12/montana-employees-withholding-allowance-and-exemption-certificate-form-mw-4.pdf
          https://mtrevenue.gov/2021/11/29/simplification-of-montana-income-taxation/?utm_source=rss&utm_medium=rss&utm_campaign=simplification-of-montana-income-taxation
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
        
        Alabama:
        -         
        - https://www.revenue.alabama.gov/ultraviewer/viewer/basic_viewer/index.html?form=2023/02/22f40.pdf
          https://www.revenue.alabama.gov/faqs/what-is-alabamas-individual-income-tax-rate/
          https://www.revenue.alabama.gov/ultraviewer/viewer/basic_viewer/index.html?form=2023/02/22f40.pdf
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-24.htm?taxmap=true
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-24.htm?taxmap=true
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-24.htm?taxmap=true
          https://www.revenue.alabama.gov/individual-corporate/income-to-be-reported-on-the-alabama-income-tax-return/
          https://www.almonline.org/TaxRates.aspx#OccupationalTax
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Montana', 'state_occupation': 'Alabama', "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def montana_and_alaska(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Montana:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Alaska:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Montana:
        - https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/2022/12/montana-employees-withholding-allowance-and-exemption-certificate-form-mw-4.pdf
          https://mtrevenue.gov/2021/11/29/simplification-of-montana-income-taxation/?utm_source=rss&utm_medium=rss&utm_campaign=simplification-of-montana-income-taxation
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
        
        Alaska:
        -         
        - https://www.commerce.alaska.gov/web/dcra/officeofthestateassessor/alaskataxfacts.aspx
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Montana', 'state_occupation': 'Alaska', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def montana_and_arizona(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Montana:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Arizona:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        credits_state : int, optional
            Credits for state taxes. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Montana:
        - https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/2022/12/montana-employees-withholding-allowance-and-exemption-certificate-form-mw-4.pdf
          https://mtrevenue.gov/2021/11/29/simplification-of-montana-income-taxation/?utm_source=rss&utm_medium=rss&utm_campaign=simplification-of-montana-income-taxation
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
        
        Arizona:
        -         
        - https://azdor.gov/forms/individual/form-140-x-y-tables
          https://azdor.gov/forms/individual/form-140-x-y-tables
          https://azdor.gov/forms/individual/form-140-x-y-tables
          https://azdor.gov/businesses-arizona/withholding-tax/withholding-exceptions
          "https://azdor.gov/individuals/income-tax-filing-assistance/identifying-other-taxable-income#:~:text=In%20regards%20to%20capital%20gains
         the%20individuals%20regular%20tax%20rate.
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Montana', 'state_occupation': 'Arizona', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def montana_and_arkansas(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Montana:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Arkansas:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Montana:
        - https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/2022/12/montana-employees-withholding-allowance-and-exemption-certificate-form-mw-4.pdf
          https://mtrevenue.gov/2021/11/29/simplification-of-montana-income-taxation/?utm_source=rss&utm_medium=rss&utm_campaign=simplification-of-montana-income-taxation
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
        
        Arkansas:
        -         
        - https://www.arkansasedc.com/why-arkansas/business-climate/tax-structure/personal-income-tax
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_and_AR1000NR_Instructions.pdf
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_FullYearResidentIndividualIncomeTaxReturn.pdf
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_and_AR1000NR_Instructions.pdf
          https://www.arkansasedc.com/why-arkansas/business-climate/tax-structure/capital-gains-tax-reduction
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_and_AR1000NR_Instructions.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Montana', 'state_occupation': 'Arkansas', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def montana_and_california(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Montana:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        California:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Montana:
        - https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/2022/12/montana-employees-withholding-allowance-and-exemption-certificate-form-mw-4.pdf
          https://mtrevenue.gov/2021/11/29/simplification-of-montana-income-taxation/?utm_source=rss&utm_medium=rss&utm_campaign=simplification-of-montana-income-taxation
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
        
        California:
        -         
        - https://www.ftb.ca.gov/forms/2022/2022-540-tax-rate-schedules.pdf
          https://www.ftb.ca.gov/file/personal/deductions/index.html
          https://www.ftb.ca.gov/about-ftb/newsroom/caleitc/eligibility-and-credit-information.html
          https://www.ftb.ca.gov/file/personal/income-types/capital-gains-and-losses.html#:~:text=California%20does%20not%20have%20a
         are%20taxed%20as%20ordinary%20income
          https://www.ftb.ca.gov/file/personal/
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Montana', 'state_occupation': 'California', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def montana_and_colorado(self, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Montana:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Colorado:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Montana:
        - https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/2022/12/montana-employees-withholding-allowance-and-exemption-certificate-form-mw-4.pdf
          https://mtrevenue.gov/2021/11/29/simplification-of-montana-income-taxation/?utm_source=rss&utm_medium=rss&utm_campaign=simplification-of-montana-income-taxation
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
        
        Colorado:
        -         
        - https://leg.colorado.gov/agencies/legislative-council-staff/individual-income-tax%C2%A0
          https://leg.colorado.gov/agencies/legislative-council-staff/individual-income-tax%C2%A0
          https://tax.colorado.gov/sites/tax/files/documents/ITT_Capital_Gain_Subtraction_Feb_2022.pdf
          https://www.auroragov.org/business_services/taxes/occupational_privilege_tax
          https://www.denvergov.org/Government/Agencies-Departments-Offices/Agencies-Departments-Offices-Directory/Department-of-Finance/Our-Divisions/Treasury/Business-Tax-Information#:~:text=Occupational%20privilege%20tax
         -%28show%20below%29&text=Each%20taxable%20employee%20is%20liable
         month%20for%20each%20taxable%20employee
          https://www.glendale.co.us/355/Occupational-Privilege-Tax
          https://www.greenwoodvillage.com/1220/Occupational-Privilege-Tax-OPT
          https://ci.sheridan.co.us/288/Occupational-Privilege-Tax
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Montana', 'state_occupation': 'Colorado', "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def montana_and_connecticut(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Montana:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Connecticut:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Montana:
        - https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/2022/12/montana-employees-withholding-allowance-and-exemption-certificate-form-mw-4.pdf
          https://mtrevenue.gov/2021/11/29/simplification-of-montana-income-taxation/?utm_source=rss&utm_medium=rss&utm_campaign=simplification-of-montana-income-taxation
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
        
        Connecticut:
        -         
        - https://portal.ct.gov/-/media/DRS/Forms/2022/Income/2022-CT-1040-Instructions_1222.pdf
          https://www.cga.ct.gov/2022/rpt/pdf/2022-R-0108.pdf
          https://www.cga.ct.gov/current/pub/chap_229.htm#sec_12-703
          https://portal.ct.gov/-/media/DRS/Forms/2022/Income/2022-CT-1040-Instructions_1222.pdf
          "https://portal.ct.gov/DRS/Publications/TSSNs/TSSN-29#:~:text=An%20individuals%20net%20capital%20gains
         portion%20of%20Social%20Security%20benefits"
          "https://portal.ct.gov/DRS/Publications/TSSNs/TSSN-29#:~:text=An%20individuals%20net%20capital%20gains
         portion%20of%20Social%20Security%20benefits"
          https://portal.ct.gov/DRS/Individuals/Resident-Income-Tax/Tax-Information
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Montana', 'state_occupation': 'Connecticut', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def montana_and_delaware(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, federal_agi=75000, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Montana:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Delaware:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        federal_agi : int, optional
            Adjusted Gross Income (AGI) for federal taxes. Default is 0. Sources - https://www.irs.gov/e-file-providers/definition-of-adjusted-gross-income, https://www.irs.gov/newsroom/taxpayers-should-know-the-difference-between-standard-and-itemized-deductions
        income : int, optional
            Gross income. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Montana:
        - https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/2022/12/montana-employees-withholding-allowance-and-exemption-certificate-form-mw-4.pdf
          https://mtrevenue.gov/2021/11/29/simplification-of-montana-income-taxation/?utm_source=rss&utm_medium=rss&utm_campaign=simplification-of-montana-income-taxation
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
        
        Delaware:
        -         
        - https://revenue.delaware.gov/software-developer/tax-rate-changes/
          https://revenuefiles.delaware.gov/2022/PIT-RES_TY22_2022-01_PaperInteractive.pdf
          https://revenuefiles.delaware.gov/2022/PIT-RES_TY22_2022-02_Instructions.pdf
          https://revenue.delaware.gov/frequently-asked-questions/personal-income-tax-faqs/
          https://www.newcastlede.gov/DocumentCenter/View/20147/City-of-Wilmington-Wage-Tax-Information-PDF
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        federal_agi = self.federal_agi if hasattr(self, 'federal_agi') else federal_agi
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Montana', 'state_occupation': 'Delaware', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "federal_agi": federal_agi, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def montana_and_florida(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Montana:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Florida:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Montana:
        - https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/2022/12/montana-employees-withholding-allowance-and-exemption-certificate-form-mw-4.pdf
          https://mtrevenue.gov/2021/11/29/simplification-of-montana-income-taxation/?utm_source=rss&utm_medium=rss&utm_campaign=simplification-of-montana-income-taxation
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
        
        Florida:
        -         
        - https://floridarevenue.com/taxes/taxesfees/Pages/tax_interest_rates.aspx
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Montana', 'state_occupation': 'Florida', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def montana_and_georgia(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Montana:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Georgia:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Montana:
        - https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/2022/12/montana-employees-withholding-allowance-and-exemption-certificate-form-mw-4.pdf
          https://mtrevenue.gov/2021/11/29/simplification-of-montana-income-taxation/?utm_source=rss&utm_medium=rss&utm_campaign=simplification-of-montana-income-taxation
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
        
        Georgia:
        -         
        - https://dor.georgia.gov/tax-tables-georgia-tax-rate-schedule
          https://apps.dor.ga.gov/FillableForms/PDFViewer/Index?form=2022GA500
          https://apps.dor.ga.gov/FillableForms/PDFViewer/Index?form=2022GA500
          https://apps.dor.ga.gov/FillableForms/PDFViewer/Index?form=2022GA500
          https://dor.georgia.gov/years-individual-income-tax-forms
          https://dor.georgia.gov/years-individual-income-tax-forms
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Montana', 'state_occupation': 'Georgia', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def montana_and_hawaii(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Montana:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Hawaii:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Montana:
        - https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/2022/12/montana-employees-withholding-allowance-and-exemption-certificate-form-mw-4.pdf
          https://mtrevenue.gov/2021/11/29/simplification-of-montana-income-taxation/?utm_source=rss&utm_medium=rss&utm_campaign=simplification-of-montana-income-taxation
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
        
        Hawaii:
        -         
        - https://tax.hawaii.gov/forms/d_18table-on/d_18table-on_p13/
          https://files.hawaii.gov/tax/forms/2018/18table-on.pdf
          https://files.hawaii.gov/tax/news/pubs/22outline.pdf
          https://files.hawaii.gov/tax/news/pubs/FastTaxRefGuide_2017.pdf
          https://www.tfhawaii.org/wordpress/state-tax-resources/mini-tax-guides/individual-income-tax-chapter-235/#:~:text=Long%20term%20capital%20gains%20are%20taxed%20at%20a%20maximum%20of%207.25%25
          https://www.grassrootinstitute.org/2023/02/hb337-increasing-tax-on-capital-gains-a-bad-risk/
          https://files.hawaii.gov/tax/news/pubs/22outline.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Montana', 'state_occupation': 'Hawaii', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def montana_and_idaho(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Montana:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Idaho:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Montana:
        - https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/2022/12/montana-employees-withholding-allowance-and-exemption-certificate-form-mw-4.pdf
          https://mtrevenue.gov/2021/11/29/simplification-of-montana-income-taxation/?utm_source=rss&utm_medium=rss&utm_campaign=simplification-of-montana-income-taxation
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
        
        Idaho:
        -         
        - https://tax.idaho.gov/taxes/income-tax/individual-income/individual-income-tax-rate-schedule/
          https://tax.idaho.gov/wp-content/uploads/forms/EIN00046/EIN00046_03-01-2023.pdf
          https://tax.idaho.gov/taxes/income-tax/individual-income/idaho-source-income/
          https://tax.idaho.gov/taxes/income-tax/individual-income/idaho-source-income/
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Montana', 'state_occupation': 'Idaho', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def montana_and_illinois(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Montana:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Illinois:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Montana:
        - https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/2022/12/montana-employees-withholding-allowance-and-exemption-certificate-form-mw-4.pdf
          https://mtrevenue.gov/2021/11/29/simplification-of-montana-income-taxation/?utm_source=rss&utm_medium=rss&utm_campaign=simplification-of-montana-income-taxation
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
        
        Illinois:
        -         
        - https://tax.illinois.gov/research/taxrates/income.html#IndividualIncome
          https://tax.illinois.gov/content/dam/soi/en/web/tax/forms/incometax/documents/currentyear/individual/il-1040-instr.pdf
          https://tax.illinois.gov/questionsandanswers/12.html
          https://tax.illinois.gov/content/dam/soi/en/web/tax/forms/incometax/documents/currentyear/individual/il-1040-instr.pdf
          https://tax.illinois.gov/research/taxrates/personalproperty.html
          https://tax.illinois.gov/localgovernments/income.html
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Montana', 'state_occupation': 'Illinois', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def montana_and_indiana(self, capital_gains_short=0, dependents=0, sources=False, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_long=0, filing_status='single', income=100000, credits_state=0):
        """
        --- Parameters ----
        
        Montana:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Indiana:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Montana:
        - https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/2022/12/montana-employees-withholding-allowance-and-exemption-certificate-form-mw-4.pdf
          https://mtrevenue.gov/2021/11/29/simplification-of-montana-income-taxation/?utm_source=rss&utm_medium=rss&utm_campaign=simplification-of-montana-income-taxation
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
        
        Indiana:
        -         
        - https://www.in.gov/dor/tax-forms/2022-individual-income-tax-forms/
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-32.htm
          https://www.in.gov/dor/files/ib28.pdf
          https://www.in.gov/dor/tax-forms/2022-individual-income-tax-forms/
          https://www.in.gov/dor/files/2022-county-tax-rates-and-codes.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Montana', 'state_occupation': 'Indiana', "capital_gains_short": capital_gains_short, "dependents": dependents, "sources": sources, "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def montana_and_iowa(self, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Montana:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Iowa:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Montana:
        - https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/2022/12/montana-employees-withholding-allowance-and-exemption-certificate-form-mw-4.pdf
          https://mtrevenue.gov/2021/11/29/simplification-of-montana-income-taxation/?utm_source=rss&utm_medium=rss&utm_campaign=simplification-of-montana-income-taxation
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
        
        Iowa:
        -         
        - https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
          https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
          https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
          https://tax.iowa.gov/iowa-illinois-reciprocal-agreement
          https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Montana', 'state_occupation': 'Iowa', "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def montana_and_kansas(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Montana:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Kansas:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Montana:
        - https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/2022/12/montana-employees-withholding-allowance-and-exemption-certificate-form-mw-4.pdf
          https://mtrevenue.gov/2021/11/29/simplification-of-montana-income-taxation/?utm_source=rss&utm_medium=rss&utm_campaign=simplification-of-montana-income-taxation
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
        
        Kansas:
        -         
        - https://www.ksrevenue.gov/taxrates.html
          https://www.ksrevenue.gov/incomebook22.html#0
          https://www.ksrevenue.gov/incomebook22.html#0
          https://www.ksrevenue.gov/incomebook22.html#0
          https://www.ksrevenue.gov/incomebook22.html
          https://www.ksrevenue.gov/incomebook22.html
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Montana', 'state_occupation': 'Kansas', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def montana_and_kentucky(self, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Montana:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Kentucky:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Montana:
        - https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/2022/12/montana-employees-withholding-allowance-and-exemption-certificate-form-mw-4.pdf
          https://mtrevenue.gov/2021/11/29/simplification-of-montana-income-taxation/?utm_source=rss&utm_medium=rss&utm_campaign=simplification-of-montana-income-taxation
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
        
        Kentucky:
        -         
        - https://revenue.ky.gov/News/Pages/Dor-Announces-Updates-to-Individual-Income-Tax-for-2023-Tax-Year.aspx
          https://revenue.ky.gov/Individual/Individual-Income-Tax/Pages/default.aspx
          https://revenue.ky.gov/News/Pages/Dor-Announces-Updates-to-Individual-Income-Tax-for-2023-Tax-Year.aspx
          https://revenue.ky.gov/Individual/Individual-Income-Tax/Pages/default.aspx
          https://revenue.ky.gov/Dor%20Training%20Materials/103%20KAR%2017.140.%20Individual%20income%20tax%20-%20reciprocity%20-%20nonresidents.pdf
          https://revenue.ky.gov/Individual/Individual-Income-Tax/Pages/default.aspx
          https://nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2021/Taxes-21-32.htm?taxmap=true
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Montana', 'state_occupation': 'Kentucky', "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def montana_and_louisiana(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Montana:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Louisiana:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Montana:
        - https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/2022/12/montana-employees-withholding-allowance-and-exemption-certificate-form-mw-4.pdf
          https://mtrevenue.gov/2021/11/29/simplification-of-montana-income-taxation/?utm_source=rss&utm_medium=rss&utm_campaign=simplification-of-montana-income-taxation
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
        
        Louisiana:
        -         
        - https://revenue.louisiana.gov/individualincometax
          https://revenue.louisiana.gov/individualincometax
          https://www.revenue.louisiana.gov/taxforms/6935(11_02)F.pdf
          https://revenue.louisiana.gov/individualincometax
          https://revenue.louisiana.gov/individualincometax
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Montana', 'state_occupation': 'Louisiana', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def montana_and_maine(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Montana:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Maine:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Montana:
        - https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/2022/12/montana-employees-withholding-allowance-and-exemption-certificate-form-mw-4.pdf
          https://mtrevenue.gov/2021/11/29/simplification-of-montana-income-taxation/?utm_source=rss&utm_medium=rss&utm_campaign=simplification-of-montana-income-taxation
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
        
        Maine:
        -         
        - https://www.maine.gov/revenue/tax-return-forms/individual-income-tax-2022
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/ind_tax_rate_sched_2022.pdf
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/22_1040me_book_gen_instruc_revisedFeb23.pdf
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/22_1040me_book_gen_instruc_revisedFeb23.pdf
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/22_1040me_book_gen_instruc_revisedFeb23.pdf
          https://www.maine.gov/revenue/tax-return-forms/individual-income-tax-2022
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Montana', 'state_occupation': 'Maine', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def montana_and_maryland(self, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Montana:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Maryland:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Montana:
        - https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/2022/12/montana-employees-withholding-allowance-and-exemption-certificate-form-mw-4.pdf
          https://mtrevenue.gov/2021/11/29/simplification-of-montana-income-taxation/?utm_source=rss&utm_medium=rss&utm_campaign=simplification-of-montana-income-taxation
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
        
        Maryland:
        -         
        - https://www.marylandtaxes.gov/individual/income/tax-info/tax-rates.php
          https://www.marylandtaxes.gov/forms/22_forms/Resident_Booklet.pdf
          https://www.marylandtaxes.gov/forms/Tax_Publications/Administrative_Releases/Income_and_Estate_Tax_Releases/ar_it37.pdf
          https://www.marylandtaxes.gov/individual/index.php
          https://www.marylandtaxes.gov/individual/income/tax-info/tax-rates.php
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Montana', 'state_occupation': 'Maryland', "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def montana_and_massachusetts(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Montana:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Massachusetts:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Montana:
        - https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/2022/12/montana-employees-withholding-allowance-and-exemption-certificate-form-mw-4.pdf
          https://mtrevenue.gov/2021/11/29/simplification-of-montana-income-taxation/?utm_source=rss&utm_medium=rss&utm_campaign=simplification-of-montana-income-taxation
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
        
        Massachusetts:
        -         
        - https://www.mass.gov/service-details/massachusetts-tax-rates
          https://www.mass.gov/service-details/view-massachusetts-personal-income-tax-exemptions
          https://www.mass.gov/service-details/view-massachusetts-personal-income-tax-exemptions
          https://www.mass.gov/service-details/view-child-and-dependent-related-credits
          https://www.mass.gov/service-details/massachusetts-tax-rates
          https://www.mass.gov/personal-income-tax
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Montana', 'state_occupation': 'Massachusetts', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def montana_and_michigan(self, capital_gains_short=0, county_occupation='Other', dependents=0, sources=False, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_long=0, filing_status='single', income=100000, credits_state=0):
        """
        --- Parameters ----
        
        Montana:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Michigan:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        county_occupation : str, optional
            County of occupation. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Montana:
        - https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/2022/12/montana-employees-withholding-allowance-and-exemption-certificate-form-mw-4.pdf
          https://mtrevenue.gov/2021/11/29/simplification-of-montana-income-taxation/?utm_source=rss&utm_medium=rss&utm_campaign=simplification-of-montana-income-taxation
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
        
        Michigan:
        -         
        - https://www.michigan.gov/taxes/iit/new-developments-for-tax-year-2022
          "https://www.michigan.gov/taxes/-/media/Project/Websites/taxes/Forms/2023/2023-SUW/446_Withholding-Guide_2023.pdf?rev=177ca3d8cb034392bf84e317ed78c320&hash=19C5F42F43A89A91863D229349F48883#:~:text=The%20exemption%20amount%20is%20%245%2C400
         employees%20Michigan%20income%20tax%20return"
          "https://www.michigan.gov/taxes/-/media/Project/Websites/taxes/Forms/2023/2023-SUW/446_Withholding-Guide_2023.pdf?rev=177ca3d8cb034392bf84e317ed78c320&hash=19C5F42F43A89A91863D229349F48883#:~:text=The%20exemption%20amount%20is%20%245%2C400
         employees%20Michigan%20income%20tax%20return"
          https://www.michigan.gov/taxes/questions/iit/accordion/residency/are-my-wages-earned-in-another-state-taxable-in-michigan-if-i-am-a-michigan-resident-1
          https://www.michigan.gov/taxes/-/media/Project/Websites/taxes/Forms/2022/2022-IIT-Forms/BOOK_MI-1040.pdf?rev=4524f1cbbc0243b194ab05a6680be75b&hash=40DFE71E651F7C1D65C31F6CD03D944F
          https://www.michigan.gov/taxes/questions/iit/accordion/general/what-cities-impose-an-income-tax
        """
        
        # Check if the parameters are set upstream in the instance
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        county_occupation = self.county_occupation if hasattr(self, 'county_occupation') else county_occupation
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Montana', 'state_occupation': 'Michigan', "capital_gains_short": capital_gains_short, "county_occupation": county_occupation, "dependents": dependents, "sources": sources, "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def montana_and_minnesota(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Montana:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Minnesota:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Montana:
        - https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/2022/12/montana-employees-withholding-allowance-and-exemption-certificate-form-mw-4.pdf
          https://mtrevenue.gov/2021/11/29/simplification-of-montana-income-taxation/?utm_source=rss&utm_medium=rss&utm_campaign=simplification-of-montana-income-taxation
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
        
        Minnesota:
        -         
        - https://www.revenue.state.mn.us/minnesota-income-tax-rates-and-brackets
          https://www.revenue.state.mn.us/child-and-dependent-care-credit
          https://www.revenue.state.mn.us/dependent-exemptions
          https://www.revenue.state.mn.us/minnesota-standard-deduction
          https://mn.gov/mmb/assets/mwr_form_tcm1059-128581.pdf
          https://www.revenue.state.mn.us/individual-income-tax
          https://www.revenue.state.mn.us/individual-income-tax
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Montana', 'state_occupation': 'Minnesota', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def montana_and_mississippi(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Montana:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Mississippi:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Montana:
        - https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/2022/12/montana-employees-withholding-allowance-and-exemption-certificate-form-mw-4.pdf
          https://mtrevenue.gov/2021/11/29/simplification-of-montana-income-taxation/?utm_source=rss&utm_medium=rss&utm_campaign=simplification-of-montana-income-taxation
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
        
        Mississippi:
        -         
        - https://www.dor.ms.gov/individual/tax-rates
          https://www.dor.ms.gov/individual/tax-rates
          https://www.dor.ms.gov/individual/tax-rates
          https://www.dor.ms.gov/individual
          https://www.dor.ms.gov/individual
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Montana', 'state_occupation': 'Mississippi', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def montana_and_missouri(self, capital_gains_short=0, dependents=0, sources=False, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_long=0, filing_status='single', income=100000, credits_state=0):
        """
        --- Parameters ----
        
        Montana:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Missouri:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Montana:
        - https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/2022/12/montana-employees-withholding-allowance-and-exemption-certificate-form-mw-4.pdf
          https://mtrevenue.gov/2021/11/29/simplification-of-montana-income-taxation/?utm_source=rss&utm_medium=rss&utm_campaign=simplification-of-montana-income-taxation
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
        
        Missouri:
        -         
        - https://dor.mo.gov/taxation/individual/tax-types/income/year-changes/
          https://dor.mo.gov/taxation/individual/tax-types/income/year-changes/
          https://dor.mo.gov/taxation/individual/tax-types/income/year-changes/
          https://www.kcmo.gov/city-hall/departments/finance/earnings-tax
          https://www.stlouis-mo.gov/government/departments/collector/earnings-tax/individual-earnings-tax-info.cfm
        """
        
        # Check if the parameters are set upstream in the instance
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Montana', 'state_occupation': 'Missouri', "capital_gains_short": capital_gains_short, "dependents": dependents, "sources": sources, "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def montana_and_nebraska(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Montana:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Nebraska:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Montana:
        - https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/2022/12/montana-employees-withholding-allowance-and-exemption-certificate-form-mw-4.pdf
          https://mtrevenue.gov/2021/11/29/simplification-of-montana-income-taxation/?utm_source=rss&utm_medium=rss&utm_campaign=simplification-of-montana-income-taxation
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
        
        Nebraska:
        -         
        - https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2023_Circular_EN_8-429r2023_Whole_Book_11.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Montana', 'state_occupation': 'Nebraska', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def montana_and_nevada(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Montana:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Nevada:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Montana:
        - https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/2022/12/montana-employees-withholding-allowance-and-exemption-certificate-form-mw-4.pdf
          https://mtrevenue.gov/2021/11/29/simplification-of-montana-income-taxation/?utm_source=rss&utm_medium=rss&utm_campaign=simplification-of-montana-income-taxation
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
        
        Nevada:
        -         
        - https://www.nvsos.gov/sos/businesses/the-nevada-advantage
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Montana', 'state_occupation': 'Nevada', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def montana_and_new_hampshire(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Montana:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        New Hampshire:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Montana:
        - https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/2022/12/montana-employees-withholding-allowance-and-exemption-certificate-form-mw-4.pdf
          https://mtrevenue.gov/2021/11/29/simplification-of-montana-income-taxation/?utm_source=rss&utm_medium=rss&utm_campaign=simplification-of-montana-income-taxation
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
        
        New Hampshire:
        -         
        - https://www.revenue.nh.gov/assistance/tax-overview.htm
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Montana', 'state_occupation': 'NewHampshire', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def montana_and_new_jersey(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Montana:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        New Jersey:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Montana:
        - https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/2022/12/montana-employees-withholding-allowance-and-exemption-certificate-form-mw-4.pdf
          https://mtrevenue.gov/2021/11/29/simplification-of-montana-income-taxation/?utm_source=rss&utm_medium=rss&utm_campaign=simplification-of-montana-income-taxation
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
        
        New Jersey:
        -         
        - https://www.state.nj.us/treasury/taxation/taxtables.shtml
          https://www.state.nj.us/treasury/taxation/njit13.shtml
          https://www.nj.gov/treasury/taxation/njit25.shtml
          https://www.state.nj.us/treasury/taxation/individuals/freqqiti.shtml
          https://www.state.nj.us/treasury/taxation/localtax.shtml
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Montana', 'state_occupation': 'NewJersey', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def montana_and_new_mexico(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Montana:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        New Mexico:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Montana:
        - https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/2022/12/montana-employees-withholding-allowance-and-exemption-certificate-form-mw-4.pdf
          https://mtrevenue.gov/2021/11/29/simplification-of-montana-income-taxation/?utm_source=rss&utm_medium=rss&utm_campaign=simplification-of-montana-income-taxation
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
        
        New Mexico:
        -         
        - https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/fdf3c548-8aba-4b9c-9eb4-bb564c716015/FYI-104.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/0bf943fd-652e-4400-bb1b-cd8397eb5a95/2022pit-adj.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/2f1a6781-9534-4436-b427-1557f9592099/2022pit-adj-ins.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/0bf943fd-652e-4400-bb1b-cd8397eb5a95/2022pit-adj.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/1afc56af-ea90-4d48-82e5-1f9aeb43255a/PITbook2022.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/fdf3c548-8aba-4b9c-9eb4-bb564c716015/FYI-104.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Montana', 'state_occupation': 'NewMexico', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def montana_and_new_york(self, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Montana:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        New York:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Montana:
        - https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/2022/12/montana-employees-withholding-allowance-and-exemption-certificate-form-mw-4.pdf
          https://mtrevenue.gov/2021/11/29/simplification-of-montana-income-taxation/?utm_source=rss&utm_medium=rss&utm_campaign=simplification-of-montana-income-taxation
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
        
        New York:
        -         
        - https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nys-tax-rate-schedule
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nys-tax-rate-schedule
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nys-tax-rate-schedule
          https://www.tax.ny.gov/pit/
          https://www.tax.ny.gov/pit/file/nonresident-faqs.htm
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#yonkers-tax-rate-schedule
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nyc-tax-rate-schedule
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Montana', 'state_occupation': 'NewYork', "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def montana_and_north_carolina(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Montana:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        North Carolina:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Montana:
        - https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/2022/12/montana-employees-withholding-allowance-and-exemption-certificate-form-mw-4.pdf
          https://mtrevenue.gov/2021/11/29/simplification-of-montana-income-taxation/?utm_source=rss&utm_medium=rss&utm_campaign=simplification-of-montana-income-taxation
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
        
        North Carolina:
        -         
        - https://www.ncdor.gov/taxes-forms/tax-rate-schedules
          https://www.ncdor.gov/taxes-forms/individual-income-tax/north-carolina-standard-deduction-or-north-carolina-itemized-deductions
          https://www.ncdor.gov/taxes-forms/individual-income-tax
          https://www.ncdor.gov/taxes-forms/individual-income-tax
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Montana', 'state_occupation': 'NorthCarolina', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def montana_and_north_dakota(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Montana:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        North Dakota:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Montana:
        - https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/2022/12/montana-employees-withholding-allowance-and-exemption-certificate-form-mw-4.pdf
          https://mtrevenue.gov/2021/11/29/simplification-of-montana-income-taxation/?utm_source=rss&utm_medium=rss&utm_campaign=simplification-of-montana-income-taxation
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
        
        North Dakota:
        -         
        - https://www.tax.nd.gov/forms
          https://www.tax.nd.gov/sites/www/files/documents/forms/individual/2022-iit/2022-individual-income-tax-booklet.pdf
          https://www.tax.nd.gov/sites/www/files/documents/forms/2022-individual-income-tax-booklet.pdf
          https://www.tax.nd.gov/tax-exemptions-credits/income-tax-exemptions-credits
          https://www.tax.nd.gov/sites/www/files/documents/forms/business/ndwrfillable.pdf
          https://www.tax.nd.gov/news/tax-legislative-changes/significant-changes-law/individual-income-tax-history
          https://www.tax.nd.gov/forms
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Montana', 'state_occupation': 'NorthDakota', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def montana_and_ohio(self, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Montana:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Ohio:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Montana:
        - https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/2022/12/montana-employees-withholding-allowance-and-exemption-certificate-form-mw-4.pdf
          https://mtrevenue.gov/2021/11/29/simplification-of-montana-income-taxation/?utm_source=rss&utm_medium=rss&utm_campaign=simplification-of-montana-income-taxation
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
        
        Ohio:
        -         
        - https://tax.ohio.gov/individual/resources/annual-tax-rates
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://tax.ohio.gov/static/forms/employer_withholding/generic/wth_it4nr.pdf
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://www.ritaohio.com/TaxRatesTable
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Montana', 'state_occupation': 'Ohio', "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def montana_and_oklahoma(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Montana:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Oklahoma:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Montana:
        - https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/2022/12/montana-employees-withholding-allowance-and-exemption-certificate-form-mw-4.pdf
          https://mtrevenue.gov/2021/11/29/simplification-of-montana-income-taxation/?utm_source=rss&utm_medium=rss&utm_campaign=simplification-of-montana-income-taxation
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
        
        Oklahoma:
        -         
        - https://oklahoma.gov/content/dam/ok/en/tax/documents/resources/publications/businesses/withholding-tables/WHTables-2022.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Montana', 'state_occupation': 'Oklahoma', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def montana_and_oregon(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Montana:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Oregon:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Montana:
        - https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/2022/12/montana-employees-withholding-allowance-and-exemption-certificate-form-mw-4.pdf
          https://mtrevenue.gov/2021/11/29/simplification-of-montana-income-taxation/?utm_source=rss&utm_medium=rss&utm_campaign=simplification-of-montana-income-taxation
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
        
        Oregon:
        -         
        - https://www.oregon.gov/dor/programs/individuals/Pages/PIT.aspx
          https://www.oregon.gov/dor/programs/individuals/Documents/Full%20year%20resident
         %20Form%20or-40%20filers.pdf
          https://www.oregon.gov/dor/programs/individuals/Documents/Part-year%20and%20nonresident
         %20Form%20or-40-P%20and%20or-40-N%20filers.pdf.pdf
          https://www.oregon.gov/dor/forms/FormsPubs/form-or-40_101-040_2022.pdf
          https://www.oregon.gov/dor/forms/FormsPubs/form-or-40_101-040_2022.pdf
          https://www.oregon.gov/dor/forms/FormsPubs/form-or-40_101-040_2022.pdf
          https://www.portland.gov/revenue/personal-tax
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Montana', 'state_occupation': 'Oregon', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def montana_and_pennsylvania(self, county_residence='Other', state_agi=75000, county_occupation='Other', filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Montana:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Pennsylvania:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        county_residence : str, optional
            County of residence. Default is Other.
        county_occupation : str, optional
            County of occupation. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Montana:
        - https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/2022/12/montana-employees-withholding-allowance-and-exemption-certificate-form-mw-4.pdf
          https://mtrevenue.gov/2021/11/29/simplification-of-montana-income-taxation/?utm_source=rss&utm_medium=rss&utm_campaign=simplification-of-montana-income-taxation
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
        
        Pennsylvania:
        -         
        - https://www.revenue.pa.gov/Tax%20Rates/Pages/default.aspx
          https://www.revenue.pa.gov/FormsandPublications/FormsforIndividuals/PIT/Documents/2022/2022_pa-40in.pdf
          https://www.revenue.pa.gov/FormsandPublications/FormsforBusinesses/EmployerWithholding/Documents/rev-419.pdf
          https://www.revenue.pa.gov/FormsandPublications/FormsforIndividuals/PIT/Documents/2022/2022_pa-40in.pdf
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-25.htm
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        county_occupation = self.county_occupation if hasattr(self, 'county_occupation') else county_occupation
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Montana', 'state_occupation': 'Pennsylvania', "county_residence": county_residence, "state_agi": state_agi, "county_occupation": county_occupation, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def montana_and_rhode_island(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Montana:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Rhode Island:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Montana:
        - https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/2022/12/montana-employees-withholding-allowance-and-exemption-certificate-form-mw-4.pdf
          https://mtrevenue.gov/2021/11/29/simplification-of-montana-income-taxation/?utm_source=rss&utm_medium=rss&utm_campaign=simplification-of-montana-income-taxation
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
        
        Rhode Island:
        -         
        - https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2022_1040WE_w_0.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Montana', 'state_occupation': 'RhodeIsland', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def montana_and_south_carolina(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Montana:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        South Carolina:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Montana:
        - https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/2022/12/montana-employees-withholding-allowance-and-exemption-certificate-form-mw-4.pdf
          https://mtrevenue.gov/2021/11/29/simplification-of-montana-income-taxation/?utm_source=rss&utm_medium=rss&utm_campaign=simplification-of-montana-income-taxation
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
        
        South Carolina:
        -         
        - https://dor.sc.gov/resources-site/lawandpolicy/Advisory%20Opinions/IL22-15.pdf
          https://dor.sc.gov/forms-site/Forms/IITPacket_2022.pdf
          https://www.scstatehouse.gov/code/t12c006.php
          https://dor.sc.gov/resources-site/lawandpolicy/Advisory%20Opinions/IL22-15.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Montana', 'state_occupation': 'SouthCarolina', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def montana_and_south_dakota(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Montana:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        South Dakota:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Montana:
        - https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/2022/12/montana-employees-withholding-allowance-and-exemption-certificate-form-mw-4.pdf
          https://mtrevenue.gov/2021/11/29/simplification-of-montana-income-taxation/?utm_source=rss&utm_medium=rss&utm_campaign=simplification-of-montana-income-taxation
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
        
        South Dakota:
        -         
        - https://dor.sd.gov/individuals/taxes/
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Montana', 'state_occupation': 'SouthDakota', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def montana_and_tennessee(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Montana:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Tennessee:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Montana:
        - https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/2022/12/montana-employees-withholding-allowance-and-exemption-certificate-form-mw-4.pdf
          https://mtrevenue.gov/2021/11/29/simplification-of-montana-income-taxation/?utm_source=rss&utm_medium=rss&utm_campaign=simplification-of-montana-income-taxation
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
        
        Tennessee:
        -         
        - https://www.tn.gov/revenue/taxes.html
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Montana', 'state_occupation': 'Tennessee', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def montana_and_texas(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Montana:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Texas:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Montana:
        - https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/2022/12/montana-employees-withholding-allowance-and-exemption-certificate-form-mw-4.pdf
          https://mtrevenue.gov/2021/11/29/simplification-of-montana-income-taxation/?utm_source=rss&utm_medium=rss&utm_campaign=simplification-of-montana-income-taxation
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
        
        Texas:
        -         
        - https://comptroller.texas.gov/taxes/
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Montana', 'state_occupation': 'Texas', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def montana_and_utah(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Montana:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Utah:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Montana:
        - https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/2022/12/montana-employees-withholding-allowance-and-exemption-certificate-form-mw-4.pdf
          https://mtrevenue.gov/2021/11/29/simplification-of-montana-income-taxation/?utm_source=rss&utm_medium=rss&utm_campaign=simplification-of-montana-income-taxation
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
        
        Utah:
        -         
        - https://incometax.utah.gov/paying/tax-rates
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Montana', 'state_occupation': 'Utah', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def montana_and_vermont(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Montana:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Vermont:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Montana:
        - https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/2022/12/montana-employees-withholding-allowance-and-exemption-certificate-form-mw-4.pdf
          https://mtrevenue.gov/2021/11/29/simplification-of-montana-income-taxation/?utm_source=rss&utm_medium=rss&utm_campaign=simplification-of-montana-income-taxation
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
        
        Vermont:
        -         
        - https://tax.vermont.gov/sites/tax/files/documents/GB-1210-2023.pdf
          https://tax.vermont.gov/sites/tax/files/documents/Income%20Booklet-2022.pdf
          https://tax.vermont.gov/sites/tax/files/documents/Income%20Booklet-2022.pdf
          https://tax.vermont.gov/sites/tax/files/documents/Income%20Booklet-2022.pdf
          https://tax.vermont.gov/individuals/personal-income-tax/tax-credits
          https://tax.vermont.gov/individuals/personal-income-tax/taxable-income
          https://tax.vermont.gov/individuals/personal-income-tax/local-option-tax
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Montana', 'state_occupation': 'Vermont', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def montana_and_virginia(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Montana:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Virginia:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Montana:
        - https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/2022/12/montana-employees-withholding-allowance-and-exemption-certificate-form-mw-4.pdf
          https://mtrevenue.gov/2021/11/29/simplification-of-montana-income-taxation/?utm_source=rss&utm_medium=rss&utm_campaign=simplification-of-montana-income-taxation
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
        
        Virginia:
        -         
        - https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/reciprocity
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Montana', 'state_occupation': 'Virginia', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def montana_and_washington(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Montana:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Washington:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Montana:
        - https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/2022/12/montana-employees-withholding-allowance-and-exemption-certificate-form-mw-4.pdf
          https://mtrevenue.gov/2021/11/29/simplification-of-montana-income-taxation/?utm_source=rss&utm_medium=rss&utm_campaign=simplification-of-montana-income-taxation
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
        
        Washington:
        -         
        - https://dor.wa.gov/taxes-rates/income-tax
          https://dor.wa.gov/taxes-rates/other-taxes/capital-gains-tax
          https://dor.wa.gov/taxes-rates/income-tax
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Montana', 'state_occupation': 'Washington', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def montana_and_west_virginia(self, pay_periods=24, capital_gains_short=0, county_occupation='Other', dependents=0, sources=False, state_agi=75000, filers_over_65=0, capital_gains_long=0, filing_status='single', income=100000, credits_state=0):
        """
        --- Parameters ----
        
        Montana:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        West Virginia:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_occupation : str, optional
            County of occupation. Default is Other.
        pay_periods : int, optional
            Number of pay periods. Options are 13, 24, 26, 52. Default is 24.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Montana:
        - https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/2022/12/montana-employees-withholding-allowance-and-exemption-certificate-form-mw-4.pdf
          https://mtrevenue.gov/2021/11/29/simplification-of-montana-income-taxation/?utm_source=rss&utm_medium=rss&utm_campaign=simplification-of-montana-income-taxation
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
        
        West Virginia:
        -         
        - https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://tax.wv.gov/Documents/TSD/tsd438.pdf
          https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2011/Taxes-11-35att.html
        """
        
        # Check if the parameters are set upstream in the instance
        pay_periods = self.pay_periods if hasattr(self, 'pay_periods') else pay_periods
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        county_occupation = self.county_occupation if hasattr(self, 'county_occupation') else county_occupation
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Montana', 'state_occupation': 'WestVirginia', "pay_periods": pay_periods, "capital_gains_short": capital_gains_short, "county_occupation": county_occupation, "dependents": dependents, "sources": sources, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def montana_and_wisconsin(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Montana:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Wisconsin:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Montana:
        - https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/2022/12/montana-employees-withholding-allowance-and-exemption-certificate-form-mw-4.pdf
          https://mtrevenue.gov/2021/11/29/simplification-of-montana-income-taxation/?utm_source=rss&utm_medium=rss&utm_campaign=simplification-of-montana-income-taxation
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
        
        Wisconsin:
        -         
        - https://www.revenue.wi.gov/TaxForms2023/2023-Form1-ES-Inst.pdf
          https://www.revenue.wi.gov/TaxForms2023/2023-Form1-ES-Inst.pdf
          https://www.revenue.wi.gov/TaxForms2023/2023-Form1-ES-Inst.pdf
          https://www.revenue.wi.gov/Pages/FAQS/pcs-work.aspx
          https://www.revenue.wi.gov/Dor%20Publications/pb103.pdf
          https://www.revenue.wi.gov/Pages/FAQS/pcs-work.aspx
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Montana', 'state_occupation': 'Wisconsin', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def montana_and_wyoming(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Montana:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Wyoming:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Montana:
        - https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/2022/12/montana-employees-withholding-allowance-and-exemption-certificate-form-mw-4.pdf
          https://mtrevenue.gov/2021/11/29/simplification-of-montana-income-taxation/?utm_source=rss&utm_medium=rss&utm_campaign=simplification-of-montana-income-taxation
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
        
        Wyoming:
        -         
        - https://revenue.wyo.gov/
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Montana', 'state_occupation': 'Wyoming', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def nebraska_and_alabama(self, county_residence='Other', state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Nebraska:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Alabama:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Nebraska:
        - https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2023_Circular_EN_8-429r2023_Whole_Book_11.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
        
        Alabama:
        -         
        - https://www.revenue.alabama.gov/ultraviewer/viewer/basic_viewer/index.html?form=2023/02/22f40.pdf
          https://www.revenue.alabama.gov/faqs/what-is-alabamas-individual-income-tax-rate/
          https://www.revenue.alabama.gov/ultraviewer/viewer/basic_viewer/index.html?form=2023/02/22f40.pdf
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-24.htm?taxmap=true
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-24.htm?taxmap=true
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-24.htm?taxmap=true
          https://www.revenue.alabama.gov/individual-corporate/income-to-be-reported-on-the-alabama-income-tax-return/
          https://www.almonline.org/TaxRates.aspx#OccupationalTax
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Nebraska', 'state_occupation': 'Alabama', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def nebraska_and_alaska(self, state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Nebraska:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Alaska:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Nebraska:
        - https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2023_Circular_EN_8-429r2023_Whole_Book_11.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
        
        Alaska:
        -         
        - https://www.commerce.alaska.gov/web/dcra/officeofthestateassessor/alaskataxfacts.aspx
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Nebraska', 'state_occupation': 'Alaska', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def nebraska_and_arizona(self, state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Nebraska:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Arizona:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        credits_state : int, optional
            Credits for state taxes. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Nebraska:
        - https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2023_Circular_EN_8-429r2023_Whole_Book_11.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
        
        Arizona:
        -         
        - https://azdor.gov/forms/individual/form-140-x-y-tables
          https://azdor.gov/forms/individual/form-140-x-y-tables
          https://azdor.gov/forms/individual/form-140-x-y-tables
          https://azdor.gov/businesses-arizona/withholding-tax/withholding-exceptions
          "https://azdor.gov/individuals/income-tax-filing-assistance/identifying-other-taxable-income#:~:text=In%20regards%20to%20capital%20gains
         the%20individuals%20regular%20tax%20rate.
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Nebraska', 'state_occupation': 'Arizona', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def nebraska_and_arkansas(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Nebraska:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Arkansas:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Nebraska:
        - https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2023_Circular_EN_8-429r2023_Whole_Book_11.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
        
        Arkansas:
        -         
        - https://www.arkansasedc.com/why-arkansas/business-climate/tax-structure/personal-income-tax
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_and_AR1000NR_Instructions.pdf
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_FullYearResidentIndividualIncomeTaxReturn.pdf
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_and_AR1000NR_Instructions.pdf
          https://www.arkansasedc.com/why-arkansas/business-climate/tax-structure/capital-gains-tax-reduction
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_and_AR1000NR_Instructions.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Nebraska', 'state_occupation': 'Arkansas', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def nebraska_and_california(self, state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Nebraska:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        California:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Nebraska:
        - https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2023_Circular_EN_8-429r2023_Whole_Book_11.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
        
        California:
        -         
        - https://www.ftb.ca.gov/forms/2022/2022-540-tax-rate-schedules.pdf
          https://www.ftb.ca.gov/file/personal/deductions/index.html
          https://www.ftb.ca.gov/about-ftb/newsroom/caleitc/eligibility-and-credit-information.html
          https://www.ftb.ca.gov/file/personal/income-types/capital-gains-and-losses.html#:~:text=California%20does%20not%20have%20a
         are%20taxed%20as%20ordinary%20income
          https://www.ftb.ca.gov/file/personal/
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Nebraska', 'state_occupation': 'California', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def nebraska_and_colorado(self, county_residence='Other', state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Nebraska:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Colorado:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Nebraska:
        - https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2023_Circular_EN_8-429r2023_Whole_Book_11.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
        
        Colorado:
        -         
        - https://leg.colorado.gov/agencies/legislative-council-staff/individual-income-tax%C2%A0
          https://leg.colorado.gov/agencies/legislative-council-staff/individual-income-tax%C2%A0
          https://tax.colorado.gov/sites/tax/files/documents/ITT_Capital_Gain_Subtraction_Feb_2022.pdf
          https://www.auroragov.org/business_services/taxes/occupational_privilege_tax
          https://www.denvergov.org/Government/Agencies-Departments-Offices/Agencies-Departments-Offices-Directory/Department-of-Finance/Our-Divisions/Treasury/Business-Tax-Information#:~:text=Occupational%20privilege%20tax
         -%28show%20below%29&text=Each%20taxable%20employee%20is%20liable
         month%20for%20each%20taxable%20employee
          https://www.glendale.co.us/355/Occupational-Privilege-Tax
          https://www.greenwoodvillage.com/1220/Occupational-Privilege-Tax-OPT
          https://ci.sheridan.co.us/288/Occupational-Privilege-Tax
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Nebraska', 'state_occupation': 'Colorado', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def nebraska_and_connecticut(self, state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Nebraska:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Connecticut:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Nebraska:
        - https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2023_Circular_EN_8-429r2023_Whole_Book_11.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
        
        Connecticut:
        -         
        - https://portal.ct.gov/-/media/DRS/Forms/2022/Income/2022-CT-1040-Instructions_1222.pdf
          https://www.cga.ct.gov/2022/rpt/pdf/2022-R-0108.pdf
          https://www.cga.ct.gov/current/pub/chap_229.htm#sec_12-703
          https://portal.ct.gov/-/media/DRS/Forms/2022/Income/2022-CT-1040-Instructions_1222.pdf
          "https://portal.ct.gov/DRS/Publications/TSSNs/TSSN-29#:~:text=An%20individuals%20net%20capital%20gains
         portion%20of%20Social%20Security%20benefits"
          "https://portal.ct.gov/DRS/Publications/TSSNs/TSSN-29#:~:text=An%20individuals%20net%20capital%20gains
         portion%20of%20Social%20Security%20benefits"
          https://portal.ct.gov/DRS/Individuals/Resident-Income-Tax/Tax-Information
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Nebraska', 'state_occupation': 'Connecticut', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def nebraska_and_delaware(self, state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, federal_agi=75000, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Nebraska:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Delaware:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        federal_agi : int, optional
            Adjusted Gross Income (AGI) for federal taxes. Default is 0. Sources - https://www.irs.gov/e-file-providers/definition-of-adjusted-gross-income, https://www.irs.gov/newsroom/taxpayers-should-know-the-difference-between-standard-and-itemized-deductions
        income : int, optional
            Gross income. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Nebraska:
        - https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2023_Circular_EN_8-429r2023_Whole_Book_11.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
        
        Delaware:
        -         
        - https://revenue.delaware.gov/software-developer/tax-rate-changes/
          https://revenuefiles.delaware.gov/2022/PIT-RES_TY22_2022-01_PaperInteractive.pdf
          https://revenuefiles.delaware.gov/2022/PIT-RES_TY22_2022-02_Instructions.pdf
          https://revenue.delaware.gov/frequently-asked-questions/personal-income-tax-faqs/
          https://www.newcastlede.gov/DocumentCenter/View/20147/City-of-Wilmington-Wage-Tax-Information-PDF
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        federal_agi = self.federal_agi if hasattr(self, 'federal_agi') else federal_agi
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Nebraska', 'state_occupation': 'Delaware', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "federal_agi": federal_agi, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def nebraska_and_florida(self, state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Nebraska:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Florida:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Nebraska:
        - https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2023_Circular_EN_8-429r2023_Whole_Book_11.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
        
        Florida:
        -         
        - https://floridarevenue.com/taxes/taxesfees/Pages/tax_interest_rates.aspx
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Nebraska', 'state_occupation': 'Florida', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def nebraska_and_georgia(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Nebraska:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Georgia:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Nebraska:
        - https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2023_Circular_EN_8-429r2023_Whole_Book_11.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
        
        Georgia:
        -         
        - https://dor.georgia.gov/tax-tables-georgia-tax-rate-schedule
          https://apps.dor.ga.gov/FillableForms/PDFViewer/Index?form=2022GA500
          https://apps.dor.ga.gov/FillableForms/PDFViewer/Index?form=2022GA500
          https://apps.dor.ga.gov/FillableForms/PDFViewer/Index?form=2022GA500
          https://dor.georgia.gov/years-individual-income-tax-forms
          https://dor.georgia.gov/years-individual-income-tax-forms
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Nebraska', 'state_occupation': 'Georgia', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def nebraska_and_hawaii(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Nebraska:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Hawaii:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Nebraska:
        - https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2023_Circular_EN_8-429r2023_Whole_Book_11.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
        
        Hawaii:
        -         
        - https://tax.hawaii.gov/forms/d_18table-on/d_18table-on_p13/
          https://files.hawaii.gov/tax/forms/2018/18table-on.pdf
          https://files.hawaii.gov/tax/news/pubs/22outline.pdf
          https://files.hawaii.gov/tax/news/pubs/FastTaxRefGuide_2017.pdf
          https://www.tfhawaii.org/wordpress/state-tax-resources/mini-tax-guides/individual-income-tax-chapter-235/#:~:text=Long%20term%20capital%20gains%20are%20taxed%20at%20a%20maximum%20of%207.25%25
          https://www.grassrootinstitute.org/2023/02/hb337-increasing-tax-on-capital-gains-a-bad-risk/
          https://files.hawaii.gov/tax/news/pubs/22outline.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Nebraska', 'state_occupation': 'Hawaii', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def nebraska_and_idaho(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Nebraska:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Idaho:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Nebraska:
        - https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2023_Circular_EN_8-429r2023_Whole_Book_11.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
        
        Idaho:
        -         
        - https://tax.idaho.gov/taxes/income-tax/individual-income/individual-income-tax-rate-schedule/
          https://tax.idaho.gov/wp-content/uploads/forms/EIN00046/EIN00046_03-01-2023.pdf
          https://tax.idaho.gov/taxes/income-tax/individual-income/idaho-source-income/
          https://tax.idaho.gov/taxes/income-tax/individual-income/idaho-source-income/
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Nebraska', 'state_occupation': 'Idaho', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def nebraska_and_illinois(self, state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Nebraska:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Illinois:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Nebraska:
        - https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2023_Circular_EN_8-429r2023_Whole_Book_11.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
        
        Illinois:
        -         
        - https://tax.illinois.gov/research/taxrates/income.html#IndividualIncome
          https://tax.illinois.gov/content/dam/soi/en/web/tax/forms/incometax/documents/currentyear/individual/il-1040-instr.pdf
          https://tax.illinois.gov/questionsandanswers/12.html
          https://tax.illinois.gov/content/dam/soi/en/web/tax/forms/incometax/documents/currentyear/individual/il-1040-instr.pdf
          https://tax.illinois.gov/research/taxrates/personalproperty.html
          https://tax.illinois.gov/localgovernments/income.html
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Nebraska', 'state_occupation': 'Illinois', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def nebraska_and_indiana(self, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Nebraska:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Indiana:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Nebraska:
        - https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2023_Circular_EN_8-429r2023_Whole_Book_11.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
        
        Indiana:
        -         
        - https://www.in.gov/dor/tax-forms/2022-individual-income-tax-forms/
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-32.htm
          https://www.in.gov/dor/files/ib28.pdf
          https://www.in.gov/dor/tax-forms/2022-individual-income-tax-forms/
          https://www.in.gov/dor/files/2022-county-tax-rates-and-codes.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Nebraska', 'state_occupation': 'Indiana', "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def nebraska_and_iowa(self, county_residence='Other', state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Nebraska:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Iowa:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Nebraska:
        - https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2023_Circular_EN_8-429r2023_Whole_Book_11.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
        
        Iowa:
        -         
        - https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
          https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
          https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
          https://tax.iowa.gov/iowa-illinois-reciprocal-agreement
          https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Nebraska', 'state_occupation': 'Iowa', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def nebraska_and_kansas(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Nebraska:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Kansas:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Nebraska:
        - https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2023_Circular_EN_8-429r2023_Whole_Book_11.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
        
        Kansas:
        -         
        - https://www.ksrevenue.gov/taxrates.html
          https://www.ksrevenue.gov/incomebook22.html#0
          https://www.ksrevenue.gov/incomebook22.html#0
          https://www.ksrevenue.gov/incomebook22.html#0
          https://www.ksrevenue.gov/incomebook22.html
          https://www.ksrevenue.gov/incomebook22.html
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Nebraska', 'state_occupation': 'Kansas', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def nebraska_and_kentucky(self, county_residence='Other', state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Nebraska:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Kentucky:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Nebraska:
        - https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2023_Circular_EN_8-429r2023_Whole_Book_11.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
        
        Kentucky:
        -         
        - https://revenue.ky.gov/News/Pages/Dor-Announces-Updates-to-Individual-Income-Tax-for-2023-Tax-Year.aspx
          https://revenue.ky.gov/Individual/Individual-Income-Tax/Pages/default.aspx
          https://revenue.ky.gov/News/Pages/Dor-Announces-Updates-to-Individual-Income-Tax-for-2023-Tax-Year.aspx
          https://revenue.ky.gov/Individual/Individual-Income-Tax/Pages/default.aspx
          https://revenue.ky.gov/Dor%20Training%20Materials/103%20KAR%2017.140.%20Individual%20income%20tax%20-%20reciprocity%20-%20nonresidents.pdf
          https://revenue.ky.gov/Individual/Individual-Income-Tax/Pages/default.aspx
          https://nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2021/Taxes-21-32.htm?taxmap=true
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Nebraska', 'state_occupation': 'Kentucky', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def nebraska_and_louisiana(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Nebraska:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Louisiana:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Nebraska:
        - https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2023_Circular_EN_8-429r2023_Whole_Book_11.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
        
        Louisiana:
        -         
        - https://revenue.louisiana.gov/individualincometax
          https://revenue.louisiana.gov/individualincometax
          https://www.revenue.louisiana.gov/taxforms/6935(11_02)F.pdf
          https://revenue.louisiana.gov/individualincometax
          https://revenue.louisiana.gov/individualincometax
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Nebraska', 'state_occupation': 'Louisiana', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def nebraska_and_maine(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Nebraska:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Maine:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Nebraska:
        - https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2023_Circular_EN_8-429r2023_Whole_Book_11.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
        
        Maine:
        -         
        - https://www.maine.gov/revenue/tax-return-forms/individual-income-tax-2022
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/ind_tax_rate_sched_2022.pdf
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/22_1040me_book_gen_instruc_revisedFeb23.pdf
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/22_1040me_book_gen_instruc_revisedFeb23.pdf
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/22_1040me_book_gen_instruc_revisedFeb23.pdf
          https://www.maine.gov/revenue/tax-return-forms/individual-income-tax-2022
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Nebraska', 'state_occupation': 'Maine', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def nebraska_and_maryland(self, county_residence='Other', state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Nebraska:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Maryland:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Nebraska:
        - https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2023_Circular_EN_8-429r2023_Whole_Book_11.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
        
        Maryland:
        -         
        - https://www.marylandtaxes.gov/individual/income/tax-info/tax-rates.php
          https://www.marylandtaxes.gov/forms/22_forms/Resident_Booklet.pdf
          https://www.marylandtaxes.gov/forms/Tax_Publications/Administrative_Releases/Income_and_Estate_Tax_Releases/ar_it37.pdf
          https://www.marylandtaxes.gov/individual/index.php
          https://www.marylandtaxes.gov/individual/income/tax-info/tax-rates.php
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Nebraska', 'state_occupation': 'Maryland', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def nebraska_and_massachusetts(self, state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Nebraska:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Massachusetts:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Nebraska:
        - https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2023_Circular_EN_8-429r2023_Whole_Book_11.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
        
        Massachusetts:
        -         
        - https://www.mass.gov/service-details/massachusetts-tax-rates
          https://www.mass.gov/service-details/view-massachusetts-personal-income-tax-exemptions
          https://www.mass.gov/service-details/view-massachusetts-personal-income-tax-exemptions
          https://www.mass.gov/service-details/view-child-and-dependent-related-credits
          https://www.mass.gov/service-details/massachusetts-tax-rates
          https://www.mass.gov/personal-income-tax
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Nebraska', 'state_occupation': 'Massachusetts', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def nebraska_and_michigan(self, county_occupation='Other', state_agi=75000, county_residence='Other', capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Nebraska:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Michigan:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        county_occupation : str, optional
            County of occupation. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Nebraska:
        - https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2023_Circular_EN_8-429r2023_Whole_Book_11.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
        
        Michigan:
        -         
        - https://www.michigan.gov/taxes/iit/new-developments-for-tax-year-2022
          "https://www.michigan.gov/taxes/-/media/Project/Websites/taxes/Forms/2023/2023-SUW/446_Withholding-Guide_2023.pdf?rev=177ca3d8cb034392bf84e317ed78c320&hash=19C5F42F43A89A91863D229349F48883#:~:text=The%20exemption%20amount%20is%20%245%2C400
         employees%20Michigan%20income%20tax%20return"
          "https://www.michigan.gov/taxes/-/media/Project/Websites/taxes/Forms/2023/2023-SUW/446_Withholding-Guide_2023.pdf?rev=177ca3d8cb034392bf84e317ed78c320&hash=19C5F42F43A89A91863D229349F48883#:~:text=The%20exemption%20amount%20is%20%245%2C400
         employees%20Michigan%20income%20tax%20return"
          https://www.michigan.gov/taxes/questions/iit/accordion/residency/are-my-wages-earned-in-another-state-taxable-in-michigan-if-i-am-a-michigan-resident-1
          https://www.michigan.gov/taxes/-/media/Project/Websites/taxes/Forms/2022/2022-IIT-Forms/BOOK_MI-1040.pdf?rev=4524f1cbbc0243b194ab05a6680be75b&hash=40DFE71E651F7C1D65C31F6CD03D944F
          https://www.michigan.gov/taxes/questions/iit/accordion/general/what-cities-impose-an-income-tax
        """
        
        # Check if the parameters are set upstream in the instance
        county_occupation = self.county_occupation if hasattr(self, 'county_occupation') else county_occupation
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Nebraska', 'state_occupation': 'Michigan', "county_occupation": county_occupation, "state_agi": state_agi, "county_residence": county_residence, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def nebraska_and_minnesota(self, state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Nebraska:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Minnesota:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Nebraska:
        - https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2023_Circular_EN_8-429r2023_Whole_Book_11.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
        
        Minnesota:
        -         
        - https://www.revenue.state.mn.us/minnesota-income-tax-rates-and-brackets
          https://www.revenue.state.mn.us/child-and-dependent-care-credit
          https://www.revenue.state.mn.us/dependent-exemptions
          https://www.revenue.state.mn.us/minnesota-standard-deduction
          https://mn.gov/mmb/assets/mwr_form_tcm1059-128581.pdf
          https://www.revenue.state.mn.us/individual-income-tax
          https://www.revenue.state.mn.us/individual-income-tax
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Nebraska', 'state_occupation': 'Minnesota', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def nebraska_and_mississippi(self, state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Nebraska:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Mississippi:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Nebraska:
        - https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2023_Circular_EN_8-429r2023_Whole_Book_11.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
        
        Mississippi:
        -         
        - https://www.dor.ms.gov/individual/tax-rates
          https://www.dor.ms.gov/individual/tax-rates
          https://www.dor.ms.gov/individual/tax-rates
          https://www.dor.ms.gov/individual
          https://www.dor.ms.gov/individual
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Nebraska', 'state_occupation': 'Mississippi', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def nebraska_and_missouri(self, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Nebraska:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Missouri:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Nebraska:
        - https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2023_Circular_EN_8-429r2023_Whole_Book_11.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
        
        Missouri:
        -         
        - https://dor.mo.gov/taxation/individual/tax-types/income/year-changes/
          https://dor.mo.gov/taxation/individual/tax-types/income/year-changes/
          https://dor.mo.gov/taxation/individual/tax-types/income/year-changes/
          https://www.kcmo.gov/city-hall/departments/finance/earnings-tax
          https://www.stlouis-mo.gov/government/departments/collector/earnings-tax/individual-earnings-tax-info.cfm
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Nebraska', 'state_occupation': 'Missouri', "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def nebraska_and_montana(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Nebraska:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Montana:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Nebraska:
        - https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2023_Circular_EN_8-429r2023_Whole_Book_11.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
        
        Montana:
        -         
        - https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/2022/12/montana-employees-withholding-allowance-and-exemption-certificate-form-mw-4.pdf
          https://mtrevenue.gov/2021/11/29/simplification-of-montana-income-taxation/?utm_source=rss&utm_medium=rss&utm_campaign=simplification-of-montana-income-taxation
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Nebraska', 'state_occupation': 'Montana', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def nebraska_and_nevada(self, state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Nebraska:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Nevada:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Nebraska:
        - https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2023_Circular_EN_8-429r2023_Whole_Book_11.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
        
        Nevada:
        -         
        - https://www.nvsos.gov/sos/businesses/the-nevada-advantage
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Nebraska', 'state_occupation': 'Nevada', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def nebraska_and_new_hampshire(self, state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Nebraska:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        New Hampshire:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Nebraska:
        - https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2023_Circular_EN_8-429r2023_Whole_Book_11.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
        
        New Hampshire:
        -         
        - https://www.revenue.nh.gov/assistance/tax-overview.htm
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Nebraska', 'state_occupation': 'NewHampshire', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def nebraska_and_new_jersey(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Nebraska:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        New Jersey:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Nebraska:
        - https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2023_Circular_EN_8-429r2023_Whole_Book_11.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
        
        New Jersey:
        -         
        - https://www.state.nj.us/treasury/taxation/taxtables.shtml
          https://www.state.nj.us/treasury/taxation/njit13.shtml
          https://www.nj.gov/treasury/taxation/njit25.shtml
          https://www.state.nj.us/treasury/taxation/individuals/freqqiti.shtml
          https://www.state.nj.us/treasury/taxation/localtax.shtml
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Nebraska', 'state_occupation': 'NewJersey', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def nebraska_and_new_mexico(self, state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Nebraska:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        New Mexico:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Nebraska:
        - https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2023_Circular_EN_8-429r2023_Whole_Book_11.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
        
        New Mexico:
        -         
        - https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/fdf3c548-8aba-4b9c-9eb4-bb564c716015/FYI-104.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/0bf943fd-652e-4400-bb1b-cd8397eb5a95/2022pit-adj.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/2f1a6781-9534-4436-b427-1557f9592099/2022pit-adj-ins.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/0bf943fd-652e-4400-bb1b-cd8397eb5a95/2022pit-adj.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/1afc56af-ea90-4d48-82e5-1f9aeb43255a/PITbook2022.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/fdf3c548-8aba-4b9c-9eb4-bb564c716015/FYI-104.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Nebraska', 'state_occupation': 'NewMexico', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def nebraska_and_new_york(self, county_residence='Other', state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Nebraska:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        New York:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Nebraska:
        - https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2023_Circular_EN_8-429r2023_Whole_Book_11.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
        
        New York:
        -         
        - https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nys-tax-rate-schedule
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nys-tax-rate-schedule
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nys-tax-rate-schedule
          https://www.tax.ny.gov/pit/
          https://www.tax.ny.gov/pit/file/nonresident-faqs.htm
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#yonkers-tax-rate-schedule
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nyc-tax-rate-schedule
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Nebraska', 'state_occupation': 'NewYork', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def nebraska_and_north_carolina(self, state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Nebraska:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        North Carolina:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Nebraska:
        - https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2023_Circular_EN_8-429r2023_Whole_Book_11.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
        
        North Carolina:
        -         
        - https://www.ncdor.gov/taxes-forms/tax-rate-schedules
          https://www.ncdor.gov/taxes-forms/individual-income-tax/north-carolina-standard-deduction-or-north-carolina-itemized-deductions
          https://www.ncdor.gov/taxes-forms/individual-income-tax
          https://www.ncdor.gov/taxes-forms/individual-income-tax
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Nebraska', 'state_occupation': 'NorthCarolina', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def nebraska_and_north_dakota(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Nebraska:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        North Dakota:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Nebraska:
        - https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2023_Circular_EN_8-429r2023_Whole_Book_11.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
        
        North Dakota:
        -         
        - https://www.tax.nd.gov/forms
          https://www.tax.nd.gov/sites/www/files/documents/forms/individual/2022-iit/2022-individual-income-tax-booklet.pdf
          https://www.tax.nd.gov/sites/www/files/documents/forms/2022-individual-income-tax-booklet.pdf
          https://www.tax.nd.gov/tax-exemptions-credits/income-tax-exemptions-credits
          https://www.tax.nd.gov/sites/www/files/documents/forms/business/ndwrfillable.pdf
          https://www.tax.nd.gov/news/tax-legislative-changes/significant-changes-law/individual-income-tax-history
          https://www.tax.nd.gov/forms
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Nebraska', 'state_occupation': 'NorthDakota', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def nebraska_and_ohio(self, county_residence='Other', state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Nebraska:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Ohio:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Nebraska:
        - https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2023_Circular_EN_8-429r2023_Whole_Book_11.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
        
        Ohio:
        -         
        - https://tax.ohio.gov/individual/resources/annual-tax-rates
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://tax.ohio.gov/static/forms/employer_withholding/generic/wth_it4nr.pdf
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://www.ritaohio.com/TaxRatesTable
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Nebraska', 'state_occupation': 'Ohio', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def nebraska_and_oklahoma(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Nebraska:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Oklahoma:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Nebraska:
        - https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2023_Circular_EN_8-429r2023_Whole_Book_11.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
        
        Oklahoma:
        -         
        - https://oklahoma.gov/content/dam/ok/en/tax/documents/resources/publications/businesses/withholding-tables/WHTables-2022.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Nebraska', 'state_occupation': 'Oklahoma', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def nebraska_and_oregon(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Nebraska:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Oregon:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Nebraska:
        - https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2023_Circular_EN_8-429r2023_Whole_Book_11.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
        
        Oregon:
        -         
        - https://www.oregon.gov/dor/programs/individuals/Pages/PIT.aspx
          https://www.oregon.gov/dor/programs/individuals/Documents/Full%20year%20resident
         %20Form%20or-40%20filers.pdf
          https://www.oregon.gov/dor/programs/individuals/Documents/Part-year%20and%20nonresident
         %20Form%20or-40-P%20and%20or-40-N%20filers.pdf.pdf
          https://www.oregon.gov/dor/forms/FormsPubs/form-or-40_101-040_2022.pdf
          https://www.oregon.gov/dor/forms/FormsPubs/form-or-40_101-040_2022.pdf
          https://www.oregon.gov/dor/forms/FormsPubs/form-or-40_101-040_2022.pdf
          https://www.portland.gov/revenue/personal-tax
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Nebraska', 'state_occupation': 'Oregon', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def nebraska_and_pennsylvania(self, county_residence='Other', state_agi=75000, county_occupation='Other', capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Nebraska:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Pennsylvania:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        county_residence : str, optional
            County of residence. Default is Other.
        county_occupation : str, optional
            County of occupation. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Nebraska:
        - https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2023_Circular_EN_8-429r2023_Whole_Book_11.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
        
        Pennsylvania:
        -         
        - https://www.revenue.pa.gov/Tax%20Rates/Pages/default.aspx
          https://www.revenue.pa.gov/FormsandPublications/FormsforIndividuals/PIT/Documents/2022/2022_pa-40in.pdf
          https://www.revenue.pa.gov/FormsandPublications/FormsforBusinesses/EmployerWithholding/Documents/rev-419.pdf
          https://www.revenue.pa.gov/FormsandPublications/FormsforIndividuals/PIT/Documents/2022/2022_pa-40in.pdf
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-25.htm
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        county_occupation = self.county_occupation if hasattr(self, 'county_occupation') else county_occupation
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Nebraska', 'state_occupation': 'Pennsylvania', "county_residence": county_residence, "state_agi": state_agi, "county_occupation": county_occupation, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def nebraska_and_rhode_island(self, state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Nebraska:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Rhode Island:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Nebraska:
        - https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2023_Circular_EN_8-429r2023_Whole_Book_11.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
        
        Rhode Island:
        -         
        - https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2022_1040WE_w_0.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Nebraska', 'state_occupation': 'RhodeIsland', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def nebraska_and_south_carolina(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Nebraska:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        South Carolina:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Nebraska:
        - https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2023_Circular_EN_8-429r2023_Whole_Book_11.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
        
        South Carolina:
        -         
        - https://dor.sc.gov/resources-site/lawandpolicy/Advisory%20Opinions/IL22-15.pdf
          https://dor.sc.gov/forms-site/Forms/IITPacket_2022.pdf
          https://www.scstatehouse.gov/code/t12c006.php
          https://dor.sc.gov/resources-site/lawandpolicy/Advisory%20Opinions/IL22-15.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Nebraska', 'state_occupation': 'SouthCarolina', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def nebraska_and_south_dakota(self, state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Nebraska:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        South Dakota:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Nebraska:
        - https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2023_Circular_EN_8-429r2023_Whole_Book_11.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
        
        South Dakota:
        -         
        - https://dor.sd.gov/individuals/taxes/
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Nebraska', 'state_occupation': 'SouthDakota', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def nebraska_and_tennessee(self, state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Nebraska:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Tennessee:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Nebraska:
        - https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2023_Circular_EN_8-429r2023_Whole_Book_11.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
        
        Tennessee:
        -         
        - https://www.tn.gov/revenue/taxes.html
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Nebraska', 'state_occupation': 'Tennessee', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def nebraska_and_texas(self, state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Nebraska:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Texas:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Nebraska:
        - https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2023_Circular_EN_8-429r2023_Whole_Book_11.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
        
        Texas:
        -         
        - https://comptroller.texas.gov/taxes/
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Nebraska', 'state_occupation': 'Texas', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def nebraska_and_utah(self, state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Nebraska:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Utah:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Nebraska:
        - https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2023_Circular_EN_8-429r2023_Whole_Book_11.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
        
        Utah:
        -         
        - https://incometax.utah.gov/paying/tax-rates
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Nebraska', 'state_occupation': 'Utah', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def nebraska_and_vermont(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Nebraska:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Vermont:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Nebraska:
        - https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2023_Circular_EN_8-429r2023_Whole_Book_11.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
        
        Vermont:
        -         
        - https://tax.vermont.gov/sites/tax/files/documents/GB-1210-2023.pdf
          https://tax.vermont.gov/sites/tax/files/documents/Income%20Booklet-2022.pdf
          https://tax.vermont.gov/sites/tax/files/documents/Income%20Booklet-2022.pdf
          https://tax.vermont.gov/sites/tax/files/documents/Income%20Booklet-2022.pdf
          https://tax.vermont.gov/individuals/personal-income-tax/tax-credits
          https://tax.vermont.gov/individuals/personal-income-tax/taxable-income
          https://tax.vermont.gov/individuals/personal-income-tax/local-option-tax
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Nebraska', 'state_occupation': 'Vermont', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def nebraska_and_virginia(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Nebraska:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Virginia:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Nebraska:
        - https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2023_Circular_EN_8-429r2023_Whole_Book_11.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
        
        Virginia:
        -         
        - https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/reciprocity
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Nebraska', 'state_occupation': 'Virginia', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def nebraska_and_washington(self, state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Nebraska:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Washington:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Nebraska:
        - https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2023_Circular_EN_8-429r2023_Whole_Book_11.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
        
        Washington:
        -         
        - https://dor.wa.gov/taxes-rates/income-tax
          https://dor.wa.gov/taxes-rates/other-taxes/capital-gains-tax
          https://dor.wa.gov/taxes-rates/income-tax
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Nebraska', 'state_occupation': 'Washington', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def nebraska_and_west_virginia(self, county_occupation='Other', state_agi=75000, pay_periods=24, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Nebraska:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        West Virginia:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_occupation : str, optional
            County of occupation. Default is Other.
        pay_periods : int, optional
            Number of pay periods. Options are 13, 24, 26, 52. Default is 24.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Nebraska:
        - https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2023_Circular_EN_8-429r2023_Whole_Book_11.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
        
        West Virginia:
        -         
        - https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://tax.wv.gov/Documents/TSD/tsd438.pdf
          https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2011/Taxes-11-35att.html
        """
        
        # Check if the parameters are set upstream in the instance
        county_occupation = self.county_occupation if hasattr(self, 'county_occupation') else county_occupation
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        pay_periods = self.pay_periods if hasattr(self, 'pay_periods') else pay_periods
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Nebraska', 'state_occupation': 'WestVirginia', "county_occupation": county_occupation, "state_agi": state_agi, "pay_periods": pay_periods, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def nebraska_and_wisconsin(self, state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Nebraska:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Wisconsin:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Nebraska:
        - https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2023_Circular_EN_8-429r2023_Whole_Book_11.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
        
        Wisconsin:
        -         
        - https://www.revenue.wi.gov/TaxForms2023/2023-Form1-ES-Inst.pdf
          https://www.revenue.wi.gov/TaxForms2023/2023-Form1-ES-Inst.pdf
          https://www.revenue.wi.gov/TaxForms2023/2023-Form1-ES-Inst.pdf
          https://www.revenue.wi.gov/Pages/FAQS/pcs-work.aspx
          https://www.revenue.wi.gov/Dor%20Publications/pb103.pdf
          https://www.revenue.wi.gov/Pages/FAQS/pcs-work.aspx
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Nebraska', 'state_occupation': 'Wisconsin', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def nebraska_and_wyoming(self, state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Nebraska:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Wyoming:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Nebraska:
        - https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2023_Circular_EN_8-429r2023_Whole_Book_11.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
        
        Wyoming:
        -         
        - https://revenue.wyo.gov/
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Nebraska', 'state_occupation': 'Wyoming', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def nevada_and_alabama(self, county_residence='Other', state_agi=75000, capital_gains_short=0, capital_gains_long=0, filing_status='single', income=100000, dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Nevada:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        Alabama:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Nevada:
        - https://www.nvsos.gov/sos/businesses/the-nevada-advantage
        
        Alabama:
        -         
        - https://www.revenue.alabama.gov/ultraviewer/viewer/basic_viewer/index.html?form=2023/02/22f40.pdf
          https://www.revenue.alabama.gov/faqs/what-is-alabamas-individual-income-tax-rate/
          https://www.revenue.alabama.gov/ultraviewer/viewer/basic_viewer/index.html?form=2023/02/22f40.pdf
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-24.htm?taxmap=true
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-24.htm?taxmap=true
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-24.htm?taxmap=true
          https://www.revenue.alabama.gov/individual-corporate/income-to-be-reported-on-the-alabama-income-tax-return/
          https://www.almonline.org/TaxRates.aspx#OccupationalTax
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Nevada', 'state_occupation': 'Alabama', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def nevada_and_alaska(self, sources=False):
        """
        --- Parameters ----
        
        Nevada:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        Alaska:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Nevada:
        - https://www.nvsos.gov/sos/businesses/the-nevada-advantage
        
        Alaska:
        -         
        - https://www.commerce.alaska.gov/web/dcra/officeofthestateassessor/alaskataxfacts.aspx
        """
        
        # Check if the parameters are set upstream in the instance
        sources = self.sources if hasattr(self, 'sources') else sources
        
        # State-specific parameters
        data = {
            'state_residence': 'Nevada', 'state_occupation': 'Alaska', "sources": sources
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def nevada_and_arizona(self, state_agi=75000, capital_gains_short=0, capital_gains_long=0, filing_status='single', income=100000, dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Nevada:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        Arizona:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        credits_state : int, optional
            Credits for state taxes. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Nevada:
        - https://www.nvsos.gov/sos/businesses/the-nevada-advantage
        
        Arizona:
        -         
        - https://azdor.gov/forms/individual/form-140-x-y-tables
          https://azdor.gov/forms/individual/form-140-x-y-tables
          https://azdor.gov/forms/individual/form-140-x-y-tables
          https://azdor.gov/businesses-arizona/withholding-tax/withholding-exceptions
          "https://azdor.gov/individuals/income-tax-filing-assistance/identifying-other-taxable-income#:~:text=In%20regards%20to%20capital%20gains
         the%20individuals%20regular%20tax%20rate.
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Nevada', 'state_occupation': 'Arizona', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def nevada_and_arkansas(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, capital_gains_long=0, filing_status='single', income=100000, dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Nevada:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        Arkansas:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Nevada:
        - https://www.nvsos.gov/sos/businesses/the-nevada-advantage
        
        Arkansas:
        -         
        - https://www.arkansasedc.com/why-arkansas/business-climate/tax-structure/personal-income-tax
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_and_AR1000NR_Instructions.pdf
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_FullYearResidentIndividualIncomeTaxReturn.pdf
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_and_AR1000NR_Instructions.pdf
          https://www.arkansasedc.com/why-arkansas/business-climate/tax-structure/capital-gains-tax-reduction
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_and_AR1000NR_Instructions.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Nevada', 'state_occupation': 'Arkansas', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def nevada_and_california(self, state_agi=75000, capital_gains_short=0, capital_gains_long=0, filing_status='single', income=100000, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Nevada:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        California:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Nevada:
        - https://www.nvsos.gov/sos/businesses/the-nevada-advantage
        
        California:
        -         
        - https://www.ftb.ca.gov/forms/2022/2022-540-tax-rate-schedules.pdf
          https://www.ftb.ca.gov/file/personal/deductions/index.html
          https://www.ftb.ca.gov/about-ftb/newsroom/caleitc/eligibility-and-credit-information.html
          https://www.ftb.ca.gov/file/personal/income-types/capital-gains-and-losses.html#:~:text=California%20does%20not%20have%20a
         are%20taxed%20as%20ordinary%20income
          https://www.ftb.ca.gov/file/personal/
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Nevada', 'state_occupation': 'California', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def nevada_and_colorado(self, county_residence='Other', state_agi=75000, capital_gains_short=0, capital_gains_long=0, income=100000, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Nevada:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        Colorado:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Nevada:
        - https://www.nvsos.gov/sos/businesses/the-nevada-advantage
        
        Colorado:
        -         
        - https://leg.colorado.gov/agencies/legislative-council-staff/individual-income-tax%C2%A0
          https://leg.colorado.gov/agencies/legislative-council-staff/individual-income-tax%C2%A0
          https://tax.colorado.gov/sites/tax/files/documents/ITT_Capital_Gain_Subtraction_Feb_2022.pdf
          https://www.auroragov.org/business_services/taxes/occupational_privilege_tax
          https://www.denvergov.org/Government/Agencies-Departments-Offices/Agencies-Departments-Offices-Directory/Department-of-Finance/Our-Divisions/Treasury/Business-Tax-Information#:~:text=Occupational%20privilege%20tax
         -%28show%20below%29&text=Each%20taxable%20employee%20is%20liable
         month%20for%20each%20taxable%20employee
          https://www.glendale.co.us/355/Occupational-Privilege-Tax
          https://www.greenwoodvillage.com/1220/Occupational-Privilege-Tax-OPT
          https://ci.sheridan.co.us/288/Occupational-Privilege-Tax
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        income = self.income if hasattr(self, 'income') else income
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Nevada', 'state_occupation': 'Colorado', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "income": income, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def nevada_and_connecticut(self, state_agi=75000, capital_gains_short=0, capital_gains_long=0, filing_status='single', income=100000, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Nevada:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        Connecticut:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Nevada:
        - https://www.nvsos.gov/sos/businesses/the-nevada-advantage
        
        Connecticut:
        -         
        - https://portal.ct.gov/-/media/DRS/Forms/2022/Income/2022-CT-1040-Instructions_1222.pdf
          https://www.cga.ct.gov/2022/rpt/pdf/2022-R-0108.pdf
          https://www.cga.ct.gov/current/pub/chap_229.htm#sec_12-703
          https://portal.ct.gov/-/media/DRS/Forms/2022/Income/2022-CT-1040-Instructions_1222.pdf
          "https://portal.ct.gov/DRS/Publications/TSSNs/TSSN-29#:~:text=An%20individuals%20net%20capital%20gains
         portion%20of%20Social%20Security%20benefits"
          "https://portal.ct.gov/DRS/Publications/TSSNs/TSSN-29#:~:text=An%20individuals%20net%20capital%20gains
         portion%20of%20Social%20Security%20benefits"
          https://portal.ct.gov/DRS/Individuals/Resident-Income-Tax/Tax-Information
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Nevada', 'state_occupation': 'Connecticut', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def nevada_and_delaware(self, capital_gains_long=0, federal_agi=75000, income=100000, sources=False, capital_gains_short=0, credits_state=0):
        """
        --- Parameters ----
        
        Nevada:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        Delaware:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        federal_agi : int, optional
            Adjusted Gross Income (AGI) for federal taxes. Default is 0. Sources - https://www.irs.gov/e-file-providers/definition-of-adjusted-gross-income, https://www.irs.gov/newsroom/taxpayers-should-know-the-difference-between-standard-and-itemized-deductions
        income : int, optional
            Gross income. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Nevada:
        - https://www.nvsos.gov/sos/businesses/the-nevada-advantage
        
        Delaware:
        -         
        - https://revenue.delaware.gov/software-developer/tax-rate-changes/
          https://revenuefiles.delaware.gov/2022/PIT-RES_TY22_2022-01_PaperInteractive.pdf
          https://revenuefiles.delaware.gov/2022/PIT-RES_TY22_2022-02_Instructions.pdf
          https://revenue.delaware.gov/frequently-asked-questions/personal-income-tax-faqs/
          https://www.newcastlede.gov/DocumentCenter/View/20147/City-of-Wilmington-Wage-Tax-Information-PDF
        """
        
        # Check if the parameters are set upstream in the instance
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        federal_agi = self.federal_agi if hasattr(self, 'federal_agi') else federal_agi
        income = self.income if hasattr(self, 'income') else income
        sources = self.sources if hasattr(self, 'sources') else sources
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Nevada', 'state_occupation': 'Delaware', "capital_gains_long": capital_gains_long, "federal_agi": federal_agi, "income": income, "sources": sources, "capital_gains_short": capital_gains_short, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def nevada_and_florida(self, sources=False):
        """
        --- Parameters ----
        
        Nevada:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        Florida:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Nevada:
        - https://www.nvsos.gov/sos/businesses/the-nevada-advantage
        
        Florida:
        -         
        - https://floridarevenue.com/taxes/taxesfees/Pages/tax_interest_rates.aspx
        """
        
        # Check if the parameters are set upstream in the instance
        sources = self.sources if hasattr(self, 'sources') else sources
        
        # State-specific parameters
        data = {
            'state_residence': 'Nevada', 'state_occupation': 'Florida', "sources": sources
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def nevada_and_georgia(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, capital_gains_long=0, filing_status='single', income=100000, dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Nevada:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        Georgia:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Nevada:
        - https://www.nvsos.gov/sos/businesses/the-nevada-advantage
        
        Georgia:
        -         
        - https://dor.georgia.gov/tax-tables-georgia-tax-rate-schedule
          https://apps.dor.ga.gov/FillableForms/PDFViewer/Index?form=2022GA500
          https://apps.dor.ga.gov/FillableForms/PDFViewer/Index?form=2022GA500
          https://apps.dor.ga.gov/FillableForms/PDFViewer/Index?form=2022GA500
          https://dor.georgia.gov/years-individual-income-tax-forms
          https://dor.georgia.gov/years-individual-income-tax-forms
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Nevada', 'state_occupation': 'Georgia', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def nevada_and_hawaii(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, capital_gains_long=0, filing_status='single', income=100000, dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Nevada:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        Hawaii:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Nevada:
        - https://www.nvsos.gov/sos/businesses/the-nevada-advantage
        
        Hawaii:
        -         
        - https://tax.hawaii.gov/forms/d_18table-on/d_18table-on_p13/
          https://files.hawaii.gov/tax/forms/2018/18table-on.pdf
          https://files.hawaii.gov/tax/news/pubs/22outline.pdf
          https://files.hawaii.gov/tax/news/pubs/FastTaxRefGuide_2017.pdf
          https://www.tfhawaii.org/wordpress/state-tax-resources/mini-tax-guides/individual-income-tax-chapter-235/#:~:text=Long%20term%20capital%20gains%20are%20taxed%20at%20a%20maximum%20of%207.25%25
          https://www.grassrootinstitute.org/2023/02/hb337-increasing-tax-on-capital-gains-a-bad-risk/
          https://files.hawaii.gov/tax/news/pubs/22outline.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Nevada', 'state_occupation': 'Hawaii', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def nevada_and_idaho(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, capital_gains_long=0, filing_status='single', income=100000, dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Nevada:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        Idaho:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Nevada:
        - https://www.nvsos.gov/sos/businesses/the-nevada-advantage
        
        Idaho:
        -         
        - https://tax.idaho.gov/taxes/income-tax/individual-income/individual-income-tax-rate-schedule/
          https://tax.idaho.gov/wp-content/uploads/forms/EIN00046/EIN00046_03-01-2023.pdf
          https://tax.idaho.gov/taxes/income-tax/individual-income/idaho-source-income/
          https://tax.idaho.gov/taxes/income-tax/individual-income/idaho-source-income/
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Nevada', 'state_occupation': 'Idaho', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def nevada_and_illinois(self, state_agi=75000, capital_gains_short=0, capital_gains_long=0, filing_status='single', income=100000, dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Nevada:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        Illinois:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Nevada:
        - https://www.nvsos.gov/sos/businesses/the-nevada-advantage
        
        Illinois:
        -         
        - https://tax.illinois.gov/research/taxrates/income.html#IndividualIncome
          https://tax.illinois.gov/content/dam/soi/en/web/tax/forms/incometax/documents/currentyear/individual/il-1040-instr.pdf
          https://tax.illinois.gov/questionsandanswers/12.html
          https://tax.illinois.gov/content/dam/soi/en/web/tax/forms/incometax/documents/currentyear/individual/il-1040-instr.pdf
          https://tax.illinois.gov/research/taxrates/personalproperty.html
          https://tax.illinois.gov/localgovernments/income.html
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Nevada', 'state_occupation': 'Illinois', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def nevada_and_indiana(self, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_short=0, capital_gains_long=0, filing_status='single', income=100000, dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Nevada:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        Indiana:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Nevada:
        - https://www.nvsos.gov/sos/businesses/the-nevada-advantage
        
        Indiana:
        -         
        - https://www.in.gov/dor/tax-forms/2022-individual-income-tax-forms/
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-32.htm
          https://www.in.gov/dor/files/ib28.pdf
          https://www.in.gov/dor/tax-forms/2022-individual-income-tax-forms/
          https://www.in.gov/dor/files/2022-county-tax-rates-and-codes.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Nevada', 'state_occupation': 'Indiana', "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def nevada_and_iowa(self, county_residence='Other', state_agi=75000, capital_gains_short=0, capital_gains_long=0, filing_status='single', income=100000, dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Nevada:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        Iowa:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Nevada:
        - https://www.nvsos.gov/sos/businesses/the-nevada-advantage
        
        Iowa:
        -         
        - https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
          https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
          https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
          https://tax.iowa.gov/iowa-illinois-reciprocal-agreement
          https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Nevada', 'state_occupation': 'Iowa', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def nevada_and_kansas(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, capital_gains_long=0, filing_status='single', income=100000, dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Nevada:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        Kansas:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Nevada:
        - https://www.nvsos.gov/sos/businesses/the-nevada-advantage
        
        Kansas:
        -         
        - https://www.ksrevenue.gov/taxrates.html
          https://www.ksrevenue.gov/incomebook22.html#0
          https://www.ksrevenue.gov/incomebook22.html#0
          https://www.ksrevenue.gov/incomebook22.html#0
          https://www.ksrevenue.gov/incomebook22.html
          https://www.ksrevenue.gov/incomebook22.html
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Nevada', 'state_occupation': 'Kansas', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def nevada_and_kentucky(self, county_residence='Other', state_agi=75000, capital_gains_short=0, capital_gains_long=0, filing_status='single', income=100000, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Nevada:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        Kentucky:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Nevada:
        - https://www.nvsos.gov/sos/businesses/the-nevada-advantage
        
        Kentucky:
        -         
        - https://revenue.ky.gov/News/Pages/Dor-Announces-Updates-to-Individual-Income-Tax-for-2023-Tax-Year.aspx
          https://revenue.ky.gov/Individual/Individual-Income-Tax/Pages/default.aspx
          https://revenue.ky.gov/News/Pages/Dor-Announces-Updates-to-Individual-Income-Tax-for-2023-Tax-Year.aspx
          https://revenue.ky.gov/Individual/Individual-Income-Tax/Pages/default.aspx
          https://revenue.ky.gov/Dor%20Training%20Materials/103%20KAR%2017.140.%20Individual%20income%20tax%20-%20reciprocity%20-%20nonresidents.pdf
          https://revenue.ky.gov/Individual/Individual-Income-Tax/Pages/default.aspx
          https://nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2021/Taxes-21-32.htm?taxmap=true
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Nevada', 'state_occupation': 'Kentucky', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def nevada_and_louisiana(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, capital_gains_long=0, filing_status='single', income=100000, dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Nevada:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        Louisiana:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Nevada:
        - https://www.nvsos.gov/sos/businesses/the-nevada-advantage
        
        Louisiana:
        -         
        - https://revenue.louisiana.gov/individualincometax
          https://revenue.louisiana.gov/individualincometax
          https://www.revenue.louisiana.gov/taxforms/6935(11_02)F.pdf
          https://revenue.louisiana.gov/individualincometax
          https://revenue.louisiana.gov/individualincometax
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Nevada', 'state_occupation': 'Louisiana', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def nevada_and_maine(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, capital_gains_long=0, filing_status='single', income=100000, dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Nevada:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        Maine:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Nevada:
        - https://www.nvsos.gov/sos/businesses/the-nevada-advantage
        
        Maine:
        -         
        - https://www.maine.gov/revenue/tax-return-forms/individual-income-tax-2022
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/ind_tax_rate_sched_2022.pdf
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/22_1040me_book_gen_instruc_revisedFeb23.pdf
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/22_1040me_book_gen_instruc_revisedFeb23.pdf
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/22_1040me_book_gen_instruc_revisedFeb23.pdf
          https://www.maine.gov/revenue/tax-return-forms/individual-income-tax-2022
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Nevada', 'state_occupation': 'Maine', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def nevada_and_maryland(self, county_residence='Other', state_agi=75000, capital_gains_short=0, capital_gains_long=0, filing_status='single', income=100000, dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Nevada:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        Maryland:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Nevada:
        - https://www.nvsos.gov/sos/businesses/the-nevada-advantage
        
        Maryland:
        -         
        - https://www.marylandtaxes.gov/individual/income/tax-info/tax-rates.php
          https://www.marylandtaxes.gov/forms/22_forms/Resident_Booklet.pdf
          https://www.marylandtaxes.gov/forms/Tax_Publications/Administrative_Releases/Income_and_Estate_Tax_Releases/ar_it37.pdf
          https://www.marylandtaxes.gov/individual/index.php
          https://www.marylandtaxes.gov/individual/income/tax-info/tax-rates.php
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Nevada', 'state_occupation': 'Maryland', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def nevada_and_massachusetts(self, state_agi=75000, capital_gains_short=0, capital_gains_long=0, filing_status='single', income=100000, dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Nevada:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        Massachusetts:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Nevada:
        - https://www.nvsos.gov/sos/businesses/the-nevada-advantage
        
        Massachusetts:
        -         
        - https://www.mass.gov/service-details/massachusetts-tax-rates
          https://www.mass.gov/service-details/view-massachusetts-personal-income-tax-exemptions
          https://www.mass.gov/service-details/view-massachusetts-personal-income-tax-exemptions
          https://www.mass.gov/service-details/view-child-and-dependent-related-credits
          https://www.mass.gov/service-details/massachusetts-tax-rates
          https://www.mass.gov/personal-income-tax
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Nevada', 'state_occupation': 'Massachusetts', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def nevada_and_michigan(self, county_residence='Other', state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', county_occupation='Other', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Nevada:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        Michigan:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        county_occupation : str, optional
            County of occupation. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Nevada:
        - https://www.nvsos.gov/sos/businesses/the-nevada-advantage
        
        Michigan:
        -         
        - https://www.michigan.gov/taxes/iit/new-developments-for-tax-year-2022
          "https://www.michigan.gov/taxes/-/media/Project/Websites/taxes/Forms/2023/2023-SUW/446_Withholding-Guide_2023.pdf?rev=177ca3d8cb034392bf84e317ed78c320&hash=19C5F42F43A89A91863D229349F48883#:~:text=The%20exemption%20amount%20is%20%245%2C400
         employees%20Michigan%20income%20tax%20return"
          "https://www.michigan.gov/taxes/-/media/Project/Websites/taxes/Forms/2023/2023-SUW/446_Withholding-Guide_2023.pdf?rev=177ca3d8cb034392bf84e317ed78c320&hash=19C5F42F43A89A91863D229349F48883#:~:text=The%20exemption%20amount%20is%20%245%2C400
         employees%20Michigan%20income%20tax%20return"
          https://www.michigan.gov/taxes/questions/iit/accordion/residency/are-my-wages-earned-in-another-state-taxable-in-michigan-if-i-am-a-michigan-resident-1
          https://www.michigan.gov/taxes/-/media/Project/Websites/taxes/Forms/2022/2022-IIT-Forms/BOOK_MI-1040.pdf?rev=4524f1cbbc0243b194ab05a6680be75b&hash=40DFE71E651F7C1D65C31F6CD03D944F
          https://www.michigan.gov/taxes/questions/iit/accordion/general/what-cities-impose-an-income-tax
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        county_occupation = self.county_occupation if hasattr(self, 'county_occupation') else county_occupation
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Nevada', 'state_occupation': 'Michigan', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "county_occupation": county_occupation, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def nevada_and_minnesota(self, state_agi=75000, capital_gains_short=0, capital_gains_long=0, filing_status='single', income=100000, dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Nevada:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        Minnesota:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Nevada:
        - https://www.nvsos.gov/sos/businesses/the-nevada-advantage
        
        Minnesota:
        -         
        - https://www.revenue.state.mn.us/minnesota-income-tax-rates-and-brackets
          https://www.revenue.state.mn.us/child-and-dependent-care-credit
          https://www.revenue.state.mn.us/dependent-exemptions
          https://www.revenue.state.mn.us/minnesota-standard-deduction
          https://mn.gov/mmb/assets/mwr_form_tcm1059-128581.pdf
          https://www.revenue.state.mn.us/individual-income-tax
          https://www.revenue.state.mn.us/individual-income-tax
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Nevada', 'state_occupation': 'Minnesota', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def nevada_and_mississippi(self, state_agi=75000, capital_gains_short=0, capital_gains_long=0, filing_status='single', income=100000, dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Nevada:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        Mississippi:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Nevada:
        - https://www.nvsos.gov/sos/businesses/the-nevada-advantage
        
        Mississippi:
        -         
        - https://www.dor.ms.gov/individual/tax-rates
          https://www.dor.ms.gov/individual/tax-rates
          https://www.dor.ms.gov/individual/tax-rates
          https://www.dor.ms.gov/individual
          https://www.dor.ms.gov/individual
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Nevada', 'state_occupation': 'Mississippi', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def nevada_and_missouri(self, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_short=0, capital_gains_long=0, filing_status='single', income=100000, dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Nevada:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        Missouri:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Nevada:
        - https://www.nvsos.gov/sos/businesses/the-nevada-advantage
        
        Missouri:
        -         
        - https://dor.mo.gov/taxation/individual/tax-types/income/year-changes/
          https://dor.mo.gov/taxation/individual/tax-types/income/year-changes/
          https://dor.mo.gov/taxation/individual/tax-types/income/year-changes/
          https://www.kcmo.gov/city-hall/departments/finance/earnings-tax
          https://www.stlouis-mo.gov/government/departments/collector/earnings-tax/individual-earnings-tax-info.cfm
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Nevada', 'state_occupation': 'Missouri', "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def nevada_and_montana(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, capital_gains_long=0, filing_status='single', income=100000, dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Nevada:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        Montana:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Nevada:
        - https://www.nvsos.gov/sos/businesses/the-nevada-advantage
        
        Montana:
        -         
        - https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/2022/12/montana-employees-withholding-allowance-and-exemption-certificate-form-mw-4.pdf
          https://mtrevenue.gov/2021/11/29/simplification-of-montana-income-taxation/?utm_source=rss&utm_medium=rss&utm_campaign=simplification-of-montana-income-taxation
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Nevada', 'state_occupation': 'Montana', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def nevada_and_nebraska(self, state_agi=75000, capital_gains_short=0, capital_gains_long=0, filing_status='single', income=100000, dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Nevada:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        Nebraska:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Nevada:
        - https://www.nvsos.gov/sos/businesses/the-nevada-advantage
        
        Nebraska:
        -         
        - https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2023_Circular_EN_8-429r2023_Whole_Book_11.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Nevada', 'state_occupation': 'Nebraska', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def nevada_and_new_hampshire(self, sources=False):
        """
        --- Parameters ----
        
        Nevada:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        New Hampshire:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Nevada:
        - https://www.nvsos.gov/sos/businesses/the-nevada-advantage
        
        New Hampshire:
        -         
        - https://www.revenue.nh.gov/assistance/tax-overview.htm
        """
        
        # Check if the parameters are set upstream in the instance
        sources = self.sources if hasattr(self, 'sources') else sources
        
        # State-specific parameters
        data = {
            'state_residence': 'Nevada', 'state_occupation': 'NewHampshire', "sources": sources
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def nevada_and_new_jersey(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, capital_gains_long=0, filing_status='single', income=100000, dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Nevada:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        New Jersey:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Nevada:
        - https://www.nvsos.gov/sos/businesses/the-nevada-advantage
        
        New Jersey:
        -         
        - https://www.state.nj.us/treasury/taxation/taxtables.shtml
          https://www.state.nj.us/treasury/taxation/njit13.shtml
          https://www.nj.gov/treasury/taxation/njit25.shtml
          https://www.state.nj.us/treasury/taxation/individuals/freqqiti.shtml
          https://www.state.nj.us/treasury/taxation/localtax.shtml
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Nevada', 'state_occupation': 'NewJersey', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def nevada_and_new_mexico(self, state_agi=75000, capital_gains_short=0, capital_gains_long=0, filing_status='single', income=100000, dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Nevada:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        New Mexico:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Nevada:
        - https://www.nvsos.gov/sos/businesses/the-nevada-advantage
        
        New Mexico:
        -         
        - https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/fdf3c548-8aba-4b9c-9eb4-bb564c716015/FYI-104.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/0bf943fd-652e-4400-bb1b-cd8397eb5a95/2022pit-adj.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/2f1a6781-9534-4436-b427-1557f9592099/2022pit-adj-ins.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/0bf943fd-652e-4400-bb1b-cd8397eb5a95/2022pit-adj.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/1afc56af-ea90-4d48-82e5-1f9aeb43255a/PITbook2022.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/fdf3c548-8aba-4b9c-9eb4-bb564c716015/FYI-104.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Nevada', 'state_occupation': 'NewMexico', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def nevada_and_new_york(self, county_residence='Other', state_agi=75000, capital_gains_short=0, capital_gains_long=0, filing_status='single', income=100000, dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Nevada:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        New York:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Nevada:
        - https://www.nvsos.gov/sos/businesses/the-nevada-advantage
        
        New York:
        -         
        - https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nys-tax-rate-schedule
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nys-tax-rate-schedule
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nys-tax-rate-schedule
          https://www.tax.ny.gov/pit/
          https://www.tax.ny.gov/pit/file/nonresident-faqs.htm
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#yonkers-tax-rate-schedule
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nyc-tax-rate-schedule
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Nevada', 'state_occupation': 'NewYork', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def nevada_and_north_carolina(self, state_agi=75000, capital_gains_short=0, capital_gains_long=0, filing_status='single', income=100000, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Nevada:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        North Carolina:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Nevada:
        - https://www.nvsos.gov/sos/businesses/the-nevada-advantage
        
        North Carolina:
        -         
        - https://www.ncdor.gov/taxes-forms/tax-rate-schedules
          https://www.ncdor.gov/taxes-forms/individual-income-tax/north-carolina-standard-deduction-or-north-carolina-itemized-deductions
          https://www.ncdor.gov/taxes-forms/individual-income-tax
          https://www.ncdor.gov/taxes-forms/individual-income-tax
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Nevada', 'state_occupation': 'NorthCarolina', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def nevada_and_north_dakota(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, capital_gains_long=0, filing_status='single', income=100000, dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Nevada:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        North Dakota:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Nevada:
        - https://www.nvsos.gov/sos/businesses/the-nevada-advantage
        
        North Dakota:
        -         
        - https://www.tax.nd.gov/forms
          https://www.tax.nd.gov/sites/www/files/documents/forms/individual/2022-iit/2022-individual-income-tax-booklet.pdf
          https://www.tax.nd.gov/sites/www/files/documents/forms/2022-individual-income-tax-booklet.pdf
          https://www.tax.nd.gov/tax-exemptions-credits/income-tax-exemptions-credits
          https://www.tax.nd.gov/sites/www/files/documents/forms/business/ndwrfillable.pdf
          https://www.tax.nd.gov/news/tax-legislative-changes/significant-changes-law/individual-income-tax-history
          https://www.tax.nd.gov/forms
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Nevada', 'state_occupation': 'NorthDakota', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def nevada_and_ohio(self, county_residence='Other', state_agi=75000, capital_gains_short=0, capital_gains_long=0, filing_status='single', income=100000, dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Nevada:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        Ohio:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Nevada:
        - https://www.nvsos.gov/sos/businesses/the-nevada-advantage
        
        Ohio:
        -         
        - https://tax.ohio.gov/individual/resources/annual-tax-rates
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://tax.ohio.gov/static/forms/employer_withholding/generic/wth_it4nr.pdf
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://www.ritaohio.com/TaxRatesTable
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Nevada', 'state_occupation': 'Ohio', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def nevada_and_oklahoma(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, capital_gains_long=0, filing_status='single', income=100000, dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Nevada:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        Oklahoma:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Nevada:
        - https://www.nvsos.gov/sos/businesses/the-nevada-advantage
        
        Oklahoma:
        -         
        - https://oklahoma.gov/content/dam/ok/en/tax/documents/resources/publications/businesses/withholding-tables/WHTables-2022.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Nevada', 'state_occupation': 'Oklahoma', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def nevada_and_oregon(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, capital_gains_long=0, filing_status='single', income=100000, dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Nevada:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        Oregon:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Nevada:
        - https://www.nvsos.gov/sos/businesses/the-nevada-advantage
        
        Oregon:
        -         
        - https://www.oregon.gov/dor/programs/individuals/Pages/PIT.aspx
          https://www.oregon.gov/dor/programs/individuals/Documents/Full%20year%20resident
         %20Form%20or-40%20filers.pdf
          https://www.oregon.gov/dor/programs/individuals/Documents/Part-year%20and%20nonresident
         %20Form%20or-40-P%20and%20or-40-N%20filers.pdf.pdf
          https://www.oregon.gov/dor/forms/FormsPubs/form-or-40_101-040_2022.pdf
          https://www.oregon.gov/dor/forms/FormsPubs/form-or-40_101-040_2022.pdf
          https://www.oregon.gov/dor/forms/FormsPubs/form-or-40_101-040_2022.pdf
          https://www.portland.gov/revenue/personal-tax
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Nevada', 'state_occupation': 'Oregon', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def nevada_and_pennsylvania(self, county_residence='Other', state_agi=75000, county_occupation='Other', capital_gains_short=0, capital_gains_long=0, income=100000, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Nevada:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        Pennsylvania:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        county_residence : str, optional
            County of residence. Default is Other.
        county_occupation : str, optional
            County of occupation. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Nevada:
        - https://www.nvsos.gov/sos/businesses/the-nevada-advantage
        
        Pennsylvania:
        -         
        - https://www.revenue.pa.gov/Tax%20Rates/Pages/default.aspx
          https://www.revenue.pa.gov/FormsandPublications/FormsforIndividuals/PIT/Documents/2022/2022_pa-40in.pdf
          https://www.revenue.pa.gov/FormsandPublications/FormsforBusinesses/EmployerWithholding/Documents/rev-419.pdf
          https://www.revenue.pa.gov/FormsandPublications/FormsforIndividuals/PIT/Documents/2022/2022_pa-40in.pdf
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-25.htm
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        county_occupation = self.county_occupation if hasattr(self, 'county_occupation') else county_occupation
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        income = self.income if hasattr(self, 'income') else income
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Nevada', 'state_occupation': 'Pennsylvania', "county_residence": county_residence, "state_agi": state_agi, "county_occupation": county_occupation, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "income": income, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def nevada_and_rhode_island(self, state_agi=75000, capital_gains_short=0, capital_gains_long=0, filing_status='single', income=100000, dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Nevada:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        Rhode Island:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Nevada:
        - https://www.nvsos.gov/sos/businesses/the-nevada-advantage
        
        Rhode Island:
        -         
        - https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2022_1040WE_w_0.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Nevada', 'state_occupation': 'RhodeIsland', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def nevada_and_south_carolina(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, capital_gains_long=0, filing_status='single', income=100000, dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Nevada:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        South Carolina:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Nevada:
        - https://www.nvsos.gov/sos/businesses/the-nevada-advantage
        
        South Carolina:
        -         
        - https://dor.sc.gov/resources-site/lawandpolicy/Advisory%20Opinions/IL22-15.pdf
          https://dor.sc.gov/forms-site/Forms/IITPacket_2022.pdf
          https://www.scstatehouse.gov/code/t12c006.php
          https://dor.sc.gov/resources-site/lawandpolicy/Advisory%20Opinions/IL22-15.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Nevada', 'state_occupation': 'SouthCarolina', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def nevada_and_south_dakota(self, sources=False):
        """
        --- Parameters ----
        
        Nevada:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        South Dakota:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Nevada:
        - https://www.nvsos.gov/sos/businesses/the-nevada-advantage
        
        South Dakota:
        -         
        - https://dor.sd.gov/individuals/taxes/
        """
        
        # Check if the parameters are set upstream in the instance
        sources = self.sources if hasattr(self, 'sources') else sources
        
        # State-specific parameters
        data = {
            'state_residence': 'Nevada', 'state_occupation': 'SouthDakota', "sources": sources
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def nevada_and_tennessee(self, sources=False):
        """
        --- Parameters ----
        
        Nevada:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        Tennessee:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Nevada:
        - https://www.nvsos.gov/sos/businesses/the-nevada-advantage
        
        Tennessee:
        -         
        - https://www.tn.gov/revenue/taxes.html
        """
        
        # Check if the parameters are set upstream in the instance
        sources = self.sources if hasattr(self, 'sources') else sources
        
        # State-specific parameters
        data = {
            'state_residence': 'Nevada', 'state_occupation': 'Tennessee', "sources": sources
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def nevada_and_texas(self, sources=False):
        """
        --- Parameters ----
        
        Nevada:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        Texas:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Nevada:
        - https://www.nvsos.gov/sos/businesses/the-nevada-advantage
        
        Texas:
        -         
        - https://comptroller.texas.gov/taxes/
        """
        
        # Check if the parameters are set upstream in the instance
        sources = self.sources if hasattr(self, 'sources') else sources
        
        # State-specific parameters
        data = {
            'state_residence': 'Nevada', 'state_occupation': 'Texas', "sources": sources
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def nevada_and_utah(self, state_agi=75000, capital_gains_short=0, capital_gains_long=0, filing_status='single', income=100000, dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Nevada:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        Utah:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Nevada:
        - https://www.nvsos.gov/sos/businesses/the-nevada-advantage
        
        Utah:
        -         
        - https://incometax.utah.gov/paying/tax-rates
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Nevada', 'state_occupation': 'Utah', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def nevada_and_vermont(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, capital_gains_long=0, filing_status='single', income=100000, dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Nevada:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        Vermont:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Nevada:
        - https://www.nvsos.gov/sos/businesses/the-nevada-advantage
        
        Vermont:
        -         
        - https://tax.vermont.gov/sites/tax/files/documents/GB-1210-2023.pdf
          https://tax.vermont.gov/sites/tax/files/documents/Income%20Booklet-2022.pdf
          https://tax.vermont.gov/sites/tax/files/documents/Income%20Booklet-2022.pdf
          https://tax.vermont.gov/sites/tax/files/documents/Income%20Booklet-2022.pdf
          https://tax.vermont.gov/individuals/personal-income-tax/tax-credits
          https://tax.vermont.gov/individuals/personal-income-tax/taxable-income
          https://tax.vermont.gov/individuals/personal-income-tax/local-option-tax
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Nevada', 'state_occupation': 'Vermont', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def nevada_and_virginia(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, capital_gains_long=0, filing_status='single', income=100000, dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Nevada:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        Virginia:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Nevada:
        - https://www.nvsos.gov/sos/businesses/the-nevada-advantage
        
        Virginia:
        -         
        - https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/reciprocity
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Nevada', 'state_occupation': 'Virginia', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def nevada_and_washington(self, capital_gains_long=0, sources=False, capital_gains_short=0):
        """
        --- Parameters ----
        
        Nevada:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        Washington:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Nevada:
        - https://www.nvsos.gov/sos/businesses/the-nevada-advantage
        
        Washington:
        -         
        - https://dor.wa.gov/taxes-rates/income-tax
          https://dor.wa.gov/taxes-rates/other-taxes/capital-gains-tax
          https://dor.wa.gov/taxes-rates/income-tax
        """
        
        # Check if the parameters are set upstream in the instance
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        sources = self.sources if hasattr(self, 'sources') else sources
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        
        # State-specific parameters
        data = {
            'state_residence': 'Nevada', 'state_occupation': 'Washington', "capital_gains_long": capital_gains_long, "sources": sources, "capital_gains_short": capital_gains_short
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def nevada_and_west_virginia(self, pay_periods=24, state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', county_occupation='Other', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Nevada:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        West Virginia:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_occupation : str, optional
            County of occupation. Default is Other.
        pay_periods : int, optional
            Number of pay periods. Options are 13, 24, 26, 52. Default is 24.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Nevada:
        - https://www.nvsos.gov/sos/businesses/the-nevada-advantage
        
        West Virginia:
        -         
        - https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://tax.wv.gov/Documents/TSD/tsd438.pdf
          https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2011/Taxes-11-35att.html
        """
        
        # Check if the parameters are set upstream in the instance
        pay_periods = self.pay_periods if hasattr(self, 'pay_periods') else pay_periods
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        county_occupation = self.county_occupation if hasattr(self, 'county_occupation') else county_occupation
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Nevada', 'state_occupation': 'WestVirginia', "pay_periods": pay_periods, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "county_occupation": county_occupation, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def nevada_and_wisconsin(self, state_agi=75000, capital_gains_short=0, capital_gains_long=0, filing_status='single', income=100000, dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Nevada:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        Wisconsin:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Nevada:
        - https://www.nvsos.gov/sos/businesses/the-nevada-advantage
        
        Wisconsin:
        -         
        - https://www.revenue.wi.gov/TaxForms2023/2023-Form1-ES-Inst.pdf
          https://www.revenue.wi.gov/TaxForms2023/2023-Form1-ES-Inst.pdf
          https://www.revenue.wi.gov/TaxForms2023/2023-Form1-ES-Inst.pdf
          https://www.revenue.wi.gov/Pages/FAQS/pcs-work.aspx
          https://www.revenue.wi.gov/Dor%20Publications/pb103.pdf
          https://www.revenue.wi.gov/Pages/FAQS/pcs-work.aspx
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Nevada', 'state_occupation': 'Wisconsin', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def nevada_and_wyoming(self, sources=False):
        """
        --- Parameters ----
        
        Nevada:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        Wyoming:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Nevada:
        - https://www.nvsos.gov/sos/businesses/the-nevada-advantage
        
        Wyoming:
        -         
        - https://revenue.wyo.gov/
        """
        
        # Check if the parameters are set upstream in the instance
        sources = self.sources if hasattr(self, 'sources') else sources
        
        # State-specific parameters
        data = {
            'state_residence': 'Nevada', 'state_occupation': 'Wyoming', "sources": sources
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def new_hampshire_and_alabama(self, county_residence='Other', state_agi=75000, capital_gains_short=0, capital_gains_long=0, filing_status='single', income=100000, dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        New Hampshire:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        Alabama:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        New Hampshire:
        - https://www.revenue.nh.gov/assistance/tax-overview.htm
        
        Alabama:
        -         
        - https://www.revenue.alabama.gov/ultraviewer/viewer/basic_viewer/index.html?form=2023/02/22f40.pdf
          https://www.revenue.alabama.gov/faqs/what-is-alabamas-individual-income-tax-rate/
          https://www.revenue.alabama.gov/ultraviewer/viewer/basic_viewer/index.html?form=2023/02/22f40.pdf
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-24.htm?taxmap=true
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-24.htm?taxmap=true
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-24.htm?taxmap=true
          https://www.revenue.alabama.gov/individual-corporate/income-to-be-reported-on-the-alabama-income-tax-return/
          https://www.almonline.org/TaxRates.aspx#OccupationalTax
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'NewHampshire', 'state_occupation': 'Alabama', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def new_hampshire_and_alaska(self, sources=False):
        """
        --- Parameters ----
        
        New Hampshire:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        Alaska:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        New Hampshire:
        - https://www.revenue.nh.gov/assistance/tax-overview.htm
        
        Alaska:
        -         
        - https://www.commerce.alaska.gov/web/dcra/officeofthestateassessor/alaskataxfacts.aspx
        """
        
        # Check if the parameters are set upstream in the instance
        sources = self.sources if hasattr(self, 'sources') else sources
        
        # State-specific parameters
        data = {
            'state_residence': 'NewHampshire', 'state_occupation': 'Alaska', "sources": sources
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def new_hampshire_and_arizona(self, state_agi=75000, capital_gains_short=0, capital_gains_long=0, filing_status='single', income=100000, dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        New Hampshire:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        Arizona:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        credits_state : int, optional
            Credits for state taxes. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        New Hampshire:
        - https://www.revenue.nh.gov/assistance/tax-overview.htm
        
        Arizona:
        -         
        - https://azdor.gov/forms/individual/form-140-x-y-tables
          https://azdor.gov/forms/individual/form-140-x-y-tables
          https://azdor.gov/forms/individual/form-140-x-y-tables
          https://azdor.gov/businesses-arizona/withholding-tax/withholding-exceptions
          "https://azdor.gov/individuals/income-tax-filing-assistance/identifying-other-taxable-income#:~:text=In%20regards%20to%20capital%20gains
         the%20individuals%20regular%20tax%20rate.
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'NewHampshire', 'state_occupation': 'Arizona', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def new_hampshire_and_arkansas(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, capital_gains_long=0, filing_status='single', income=100000, dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        New Hampshire:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        Arkansas:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        New Hampshire:
        - https://www.revenue.nh.gov/assistance/tax-overview.htm
        
        Arkansas:
        -         
        - https://www.arkansasedc.com/why-arkansas/business-climate/tax-structure/personal-income-tax
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_and_AR1000NR_Instructions.pdf
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_FullYearResidentIndividualIncomeTaxReturn.pdf
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_and_AR1000NR_Instructions.pdf
          https://www.arkansasedc.com/why-arkansas/business-climate/tax-structure/capital-gains-tax-reduction
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_and_AR1000NR_Instructions.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'NewHampshire', 'state_occupation': 'Arkansas', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def new_hampshire_and_california(self, state_agi=75000, capital_gains_short=0, capital_gains_long=0, filing_status='single', income=100000, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        New Hampshire:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        California:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        New Hampshire:
        - https://www.revenue.nh.gov/assistance/tax-overview.htm
        
        California:
        -         
        - https://www.ftb.ca.gov/forms/2022/2022-540-tax-rate-schedules.pdf
          https://www.ftb.ca.gov/file/personal/deductions/index.html
          https://www.ftb.ca.gov/about-ftb/newsroom/caleitc/eligibility-and-credit-information.html
          https://www.ftb.ca.gov/file/personal/income-types/capital-gains-and-losses.html#:~:text=California%20does%20not%20have%20a
         are%20taxed%20as%20ordinary%20income
          https://www.ftb.ca.gov/file/personal/
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'NewHampshire', 'state_occupation': 'California', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def new_hampshire_and_colorado(self, county_residence='Other', state_agi=75000, capital_gains_short=0, capital_gains_long=0, income=100000, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        New Hampshire:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        Colorado:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        New Hampshire:
        - https://www.revenue.nh.gov/assistance/tax-overview.htm
        
        Colorado:
        -         
        - https://leg.colorado.gov/agencies/legislative-council-staff/individual-income-tax%C2%A0
          https://leg.colorado.gov/agencies/legislative-council-staff/individual-income-tax%C2%A0
          https://tax.colorado.gov/sites/tax/files/documents/ITT_Capital_Gain_Subtraction_Feb_2022.pdf
          https://www.auroragov.org/business_services/taxes/occupational_privilege_tax
          https://www.denvergov.org/Government/Agencies-Departments-Offices/Agencies-Departments-Offices-Directory/Department-of-Finance/Our-Divisions/Treasury/Business-Tax-Information#:~:text=Occupational%20privilege%20tax
         -%28show%20below%29&text=Each%20taxable%20employee%20is%20liable
         month%20for%20each%20taxable%20employee
          https://www.glendale.co.us/355/Occupational-Privilege-Tax
          https://www.greenwoodvillage.com/1220/Occupational-Privilege-Tax-OPT
          https://ci.sheridan.co.us/288/Occupational-Privilege-Tax
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        income = self.income if hasattr(self, 'income') else income
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'NewHampshire', 'state_occupation': 'Colorado', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "income": income, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def new_hampshire_and_connecticut(self, state_agi=75000, capital_gains_short=0, capital_gains_long=0, filing_status='single', income=100000, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        New Hampshire:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        Connecticut:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        New Hampshire:
        - https://www.revenue.nh.gov/assistance/tax-overview.htm
        
        Connecticut:
        -         
        - https://portal.ct.gov/-/media/DRS/Forms/2022/Income/2022-CT-1040-Instructions_1222.pdf
          https://www.cga.ct.gov/2022/rpt/pdf/2022-R-0108.pdf
          https://www.cga.ct.gov/current/pub/chap_229.htm#sec_12-703
          https://portal.ct.gov/-/media/DRS/Forms/2022/Income/2022-CT-1040-Instructions_1222.pdf
          "https://portal.ct.gov/DRS/Publications/TSSNs/TSSN-29#:~:text=An%20individuals%20net%20capital%20gains
         portion%20of%20Social%20Security%20benefits"
          "https://portal.ct.gov/DRS/Publications/TSSNs/TSSN-29#:~:text=An%20individuals%20net%20capital%20gains
         portion%20of%20Social%20Security%20benefits"
          https://portal.ct.gov/DRS/Individuals/Resident-Income-Tax/Tax-Information
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'NewHampshire', 'state_occupation': 'Connecticut', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def new_hampshire_and_delaware(self, capital_gains_long=0, federal_agi=75000, income=100000, sources=False, capital_gains_short=0, credits_state=0):
        """
        --- Parameters ----
        
        New Hampshire:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        Delaware:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        federal_agi : int, optional
            Adjusted Gross Income (AGI) for federal taxes. Default is 0. Sources - https://www.irs.gov/e-file-providers/definition-of-adjusted-gross-income, https://www.irs.gov/newsroom/taxpayers-should-know-the-difference-between-standard-and-itemized-deductions
        income : int, optional
            Gross income. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        New Hampshire:
        - https://www.revenue.nh.gov/assistance/tax-overview.htm
        
        Delaware:
        -         
        - https://revenue.delaware.gov/software-developer/tax-rate-changes/
          https://revenuefiles.delaware.gov/2022/PIT-RES_TY22_2022-01_PaperInteractive.pdf
          https://revenuefiles.delaware.gov/2022/PIT-RES_TY22_2022-02_Instructions.pdf
          https://revenue.delaware.gov/frequently-asked-questions/personal-income-tax-faqs/
          https://www.newcastlede.gov/DocumentCenter/View/20147/City-of-Wilmington-Wage-Tax-Information-PDF
        """
        
        # Check if the parameters are set upstream in the instance
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        federal_agi = self.federal_agi if hasattr(self, 'federal_agi') else federal_agi
        income = self.income if hasattr(self, 'income') else income
        sources = self.sources if hasattr(self, 'sources') else sources
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'NewHampshire', 'state_occupation': 'Delaware', "capital_gains_long": capital_gains_long, "federal_agi": federal_agi, "income": income, "sources": sources, "capital_gains_short": capital_gains_short, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def new_hampshire_and_florida(self, sources=False):
        """
        --- Parameters ----
        
        New Hampshire:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        Florida:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        New Hampshire:
        - https://www.revenue.nh.gov/assistance/tax-overview.htm
        
        Florida:
        -         
        - https://floridarevenue.com/taxes/taxesfees/Pages/tax_interest_rates.aspx
        """
        
        # Check if the parameters are set upstream in the instance
        sources = self.sources if hasattr(self, 'sources') else sources
        
        # State-specific parameters
        data = {
            'state_residence': 'NewHampshire', 'state_occupation': 'Florida', "sources": sources
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def new_hampshire_and_georgia(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, capital_gains_long=0, filing_status='single', income=100000, dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        New Hampshire:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        Georgia:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        New Hampshire:
        - https://www.revenue.nh.gov/assistance/tax-overview.htm
        
        Georgia:
        -         
        - https://dor.georgia.gov/tax-tables-georgia-tax-rate-schedule
          https://apps.dor.ga.gov/FillableForms/PDFViewer/Index?form=2022GA500
          https://apps.dor.ga.gov/FillableForms/PDFViewer/Index?form=2022GA500
          https://apps.dor.ga.gov/FillableForms/PDFViewer/Index?form=2022GA500
          https://dor.georgia.gov/years-individual-income-tax-forms
          https://dor.georgia.gov/years-individual-income-tax-forms
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'NewHampshire', 'state_occupation': 'Georgia', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def new_hampshire_and_hawaii(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, capital_gains_long=0, filing_status='single', income=100000, dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        New Hampshire:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        Hawaii:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        New Hampshire:
        - https://www.revenue.nh.gov/assistance/tax-overview.htm
        
        Hawaii:
        -         
        - https://tax.hawaii.gov/forms/d_18table-on/d_18table-on_p13/
          https://files.hawaii.gov/tax/forms/2018/18table-on.pdf
          https://files.hawaii.gov/tax/news/pubs/22outline.pdf
          https://files.hawaii.gov/tax/news/pubs/FastTaxRefGuide_2017.pdf
          https://www.tfhawaii.org/wordpress/state-tax-resources/mini-tax-guides/individual-income-tax-chapter-235/#:~:text=Long%20term%20capital%20gains%20are%20taxed%20at%20a%20maximum%20of%207.25%25
          https://www.grassrootinstitute.org/2023/02/hb337-increasing-tax-on-capital-gains-a-bad-risk/
          https://files.hawaii.gov/tax/news/pubs/22outline.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'NewHampshire', 'state_occupation': 'Hawaii', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def new_hampshire_and_idaho(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, capital_gains_long=0, filing_status='single', income=100000, dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        New Hampshire:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        Idaho:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        New Hampshire:
        - https://www.revenue.nh.gov/assistance/tax-overview.htm
        
        Idaho:
        -         
        - https://tax.idaho.gov/taxes/income-tax/individual-income/individual-income-tax-rate-schedule/
          https://tax.idaho.gov/wp-content/uploads/forms/EIN00046/EIN00046_03-01-2023.pdf
          https://tax.idaho.gov/taxes/income-tax/individual-income/idaho-source-income/
          https://tax.idaho.gov/taxes/income-tax/individual-income/idaho-source-income/
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'NewHampshire', 'state_occupation': 'Idaho', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def new_hampshire_and_illinois(self, state_agi=75000, capital_gains_short=0, capital_gains_long=0, filing_status='single', income=100000, dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        New Hampshire:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        Illinois:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        New Hampshire:
        - https://www.revenue.nh.gov/assistance/tax-overview.htm
        
        Illinois:
        -         
        - https://tax.illinois.gov/research/taxrates/income.html#IndividualIncome
          https://tax.illinois.gov/content/dam/soi/en/web/tax/forms/incometax/documents/currentyear/individual/il-1040-instr.pdf
          https://tax.illinois.gov/questionsandanswers/12.html
          https://tax.illinois.gov/content/dam/soi/en/web/tax/forms/incometax/documents/currentyear/individual/il-1040-instr.pdf
          https://tax.illinois.gov/research/taxrates/personalproperty.html
          https://tax.illinois.gov/localgovernments/income.html
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'NewHampshire', 'state_occupation': 'Illinois', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def new_hampshire_and_indiana(self, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_short=0, capital_gains_long=0, filing_status='single', income=100000, dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        New Hampshire:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        Indiana:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        New Hampshire:
        - https://www.revenue.nh.gov/assistance/tax-overview.htm
        
        Indiana:
        -         
        - https://www.in.gov/dor/tax-forms/2022-individual-income-tax-forms/
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-32.htm
          https://www.in.gov/dor/files/ib28.pdf
          https://www.in.gov/dor/tax-forms/2022-individual-income-tax-forms/
          https://www.in.gov/dor/files/2022-county-tax-rates-and-codes.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'NewHampshire', 'state_occupation': 'Indiana', "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def new_hampshire_and_iowa(self, county_residence='Other', state_agi=75000, capital_gains_short=0, capital_gains_long=0, filing_status='single', income=100000, dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        New Hampshire:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        Iowa:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        New Hampshire:
        - https://www.revenue.nh.gov/assistance/tax-overview.htm
        
        Iowa:
        -         
        - https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
          https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
          https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
          https://tax.iowa.gov/iowa-illinois-reciprocal-agreement
          https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'NewHampshire', 'state_occupation': 'Iowa', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def new_hampshire_and_kansas(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, capital_gains_long=0, filing_status='single', income=100000, dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        New Hampshire:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        Kansas:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        New Hampshire:
        - https://www.revenue.nh.gov/assistance/tax-overview.htm
        
        Kansas:
        -         
        - https://www.ksrevenue.gov/taxrates.html
          https://www.ksrevenue.gov/incomebook22.html#0
          https://www.ksrevenue.gov/incomebook22.html#0
          https://www.ksrevenue.gov/incomebook22.html#0
          https://www.ksrevenue.gov/incomebook22.html
          https://www.ksrevenue.gov/incomebook22.html
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'NewHampshire', 'state_occupation': 'Kansas', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def new_hampshire_and_kentucky(self, county_residence='Other', state_agi=75000, capital_gains_short=0, capital_gains_long=0, filing_status='single', income=100000, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        New Hampshire:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        Kentucky:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        New Hampshire:
        - https://www.revenue.nh.gov/assistance/tax-overview.htm
        
        Kentucky:
        -         
        - https://revenue.ky.gov/News/Pages/Dor-Announces-Updates-to-Individual-Income-Tax-for-2023-Tax-Year.aspx
          https://revenue.ky.gov/Individual/Individual-Income-Tax/Pages/default.aspx
          https://revenue.ky.gov/News/Pages/Dor-Announces-Updates-to-Individual-Income-Tax-for-2023-Tax-Year.aspx
          https://revenue.ky.gov/Individual/Individual-Income-Tax/Pages/default.aspx
          https://revenue.ky.gov/Dor%20Training%20Materials/103%20KAR%2017.140.%20Individual%20income%20tax%20-%20reciprocity%20-%20nonresidents.pdf
          https://revenue.ky.gov/Individual/Individual-Income-Tax/Pages/default.aspx
          https://nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2021/Taxes-21-32.htm?taxmap=true
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'NewHampshire', 'state_occupation': 'Kentucky', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def new_hampshire_and_louisiana(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, capital_gains_long=0, filing_status='single', income=100000, dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        New Hampshire:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        Louisiana:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        New Hampshire:
        - https://www.revenue.nh.gov/assistance/tax-overview.htm
        
        Louisiana:
        -         
        - https://revenue.louisiana.gov/individualincometax
          https://revenue.louisiana.gov/individualincometax
          https://www.revenue.louisiana.gov/taxforms/6935(11_02)F.pdf
          https://revenue.louisiana.gov/individualincometax
          https://revenue.louisiana.gov/individualincometax
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'NewHampshire', 'state_occupation': 'Louisiana', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def new_hampshire_and_maine(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, capital_gains_long=0, filing_status='single', income=100000, dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        New Hampshire:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        Maine:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        New Hampshire:
        - https://www.revenue.nh.gov/assistance/tax-overview.htm
        
        Maine:
        -         
        - https://www.maine.gov/revenue/tax-return-forms/individual-income-tax-2022
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/ind_tax_rate_sched_2022.pdf
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/22_1040me_book_gen_instruc_revisedFeb23.pdf
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/22_1040me_book_gen_instruc_revisedFeb23.pdf
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/22_1040me_book_gen_instruc_revisedFeb23.pdf
          https://www.maine.gov/revenue/tax-return-forms/individual-income-tax-2022
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'NewHampshire', 'state_occupation': 'Maine', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def new_hampshire_and_maryland(self, county_residence='Other', state_agi=75000, capital_gains_short=0, capital_gains_long=0, filing_status='single', income=100000, dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        New Hampshire:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        Maryland:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        New Hampshire:
        - https://www.revenue.nh.gov/assistance/tax-overview.htm
        
        Maryland:
        -         
        - https://www.marylandtaxes.gov/individual/income/tax-info/tax-rates.php
          https://www.marylandtaxes.gov/forms/22_forms/Resident_Booklet.pdf
          https://www.marylandtaxes.gov/forms/Tax_Publications/Administrative_Releases/Income_and_Estate_Tax_Releases/ar_it37.pdf
          https://www.marylandtaxes.gov/individual/index.php
          https://www.marylandtaxes.gov/individual/income/tax-info/tax-rates.php
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'NewHampshire', 'state_occupation': 'Maryland', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def new_hampshire_and_massachusetts(self, state_agi=75000, capital_gains_short=0, capital_gains_long=0, filing_status='single', income=100000, dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        New Hampshire:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        Massachusetts:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        New Hampshire:
        - https://www.revenue.nh.gov/assistance/tax-overview.htm
        
        Massachusetts:
        -         
        - https://www.mass.gov/service-details/massachusetts-tax-rates
          https://www.mass.gov/service-details/view-massachusetts-personal-income-tax-exemptions
          https://www.mass.gov/service-details/view-massachusetts-personal-income-tax-exemptions
          https://www.mass.gov/service-details/view-child-and-dependent-related-credits
          https://www.mass.gov/service-details/massachusetts-tax-rates
          https://www.mass.gov/personal-income-tax
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'NewHampshire', 'state_occupation': 'Massachusetts', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def new_hampshire_and_michigan(self, county_residence='Other', state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', county_occupation='Other', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        New Hampshire:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        Michigan:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        county_occupation : str, optional
            County of occupation. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        New Hampshire:
        - https://www.revenue.nh.gov/assistance/tax-overview.htm
        
        Michigan:
        -         
        - https://www.michigan.gov/taxes/iit/new-developments-for-tax-year-2022
          "https://www.michigan.gov/taxes/-/media/Project/Websites/taxes/Forms/2023/2023-SUW/446_Withholding-Guide_2023.pdf?rev=177ca3d8cb034392bf84e317ed78c320&hash=19C5F42F43A89A91863D229349F48883#:~:text=The%20exemption%20amount%20is%20%245%2C400
         employees%20Michigan%20income%20tax%20return"
          "https://www.michigan.gov/taxes/-/media/Project/Websites/taxes/Forms/2023/2023-SUW/446_Withholding-Guide_2023.pdf?rev=177ca3d8cb034392bf84e317ed78c320&hash=19C5F42F43A89A91863D229349F48883#:~:text=The%20exemption%20amount%20is%20%245%2C400
         employees%20Michigan%20income%20tax%20return"
          https://www.michigan.gov/taxes/questions/iit/accordion/residency/are-my-wages-earned-in-another-state-taxable-in-michigan-if-i-am-a-michigan-resident-1
          https://www.michigan.gov/taxes/-/media/Project/Websites/taxes/Forms/2022/2022-IIT-Forms/BOOK_MI-1040.pdf?rev=4524f1cbbc0243b194ab05a6680be75b&hash=40DFE71E651F7C1D65C31F6CD03D944F
          https://www.michigan.gov/taxes/questions/iit/accordion/general/what-cities-impose-an-income-tax
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        county_occupation = self.county_occupation if hasattr(self, 'county_occupation') else county_occupation
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'NewHampshire', 'state_occupation': 'Michigan', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "county_occupation": county_occupation, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def new_hampshire_and_minnesota(self, state_agi=75000, capital_gains_short=0, capital_gains_long=0, filing_status='single', income=100000, dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        New Hampshire:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        Minnesota:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        New Hampshire:
        - https://www.revenue.nh.gov/assistance/tax-overview.htm
        
        Minnesota:
        -         
        - https://www.revenue.state.mn.us/minnesota-income-tax-rates-and-brackets
          https://www.revenue.state.mn.us/child-and-dependent-care-credit
          https://www.revenue.state.mn.us/dependent-exemptions
          https://www.revenue.state.mn.us/minnesota-standard-deduction
          https://mn.gov/mmb/assets/mwr_form_tcm1059-128581.pdf
          https://www.revenue.state.mn.us/individual-income-tax
          https://www.revenue.state.mn.us/individual-income-tax
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'NewHampshire', 'state_occupation': 'Minnesota', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def new_hampshire_and_mississippi(self, state_agi=75000, capital_gains_short=0, capital_gains_long=0, filing_status='single', income=100000, dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        New Hampshire:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        Mississippi:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        New Hampshire:
        - https://www.revenue.nh.gov/assistance/tax-overview.htm
        
        Mississippi:
        -         
        - https://www.dor.ms.gov/individual/tax-rates
          https://www.dor.ms.gov/individual/tax-rates
          https://www.dor.ms.gov/individual/tax-rates
          https://www.dor.ms.gov/individual
          https://www.dor.ms.gov/individual
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'NewHampshire', 'state_occupation': 'Mississippi', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def new_hampshire_and_missouri(self, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_short=0, capital_gains_long=0, filing_status='single', income=100000, dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        New Hampshire:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        Missouri:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        New Hampshire:
        - https://www.revenue.nh.gov/assistance/tax-overview.htm
        
        Missouri:
        -         
        - https://dor.mo.gov/taxation/individual/tax-types/income/year-changes/
          https://dor.mo.gov/taxation/individual/tax-types/income/year-changes/
          https://dor.mo.gov/taxation/individual/tax-types/income/year-changes/
          https://www.kcmo.gov/city-hall/departments/finance/earnings-tax
          https://www.stlouis-mo.gov/government/departments/collector/earnings-tax/individual-earnings-tax-info.cfm
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'NewHampshire', 'state_occupation': 'Missouri', "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def new_hampshire_and_montana(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, capital_gains_long=0, filing_status='single', income=100000, dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        New Hampshire:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        Montana:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        New Hampshire:
        - https://www.revenue.nh.gov/assistance/tax-overview.htm
        
        Montana:
        -         
        - https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/2022/12/montana-employees-withholding-allowance-and-exemption-certificate-form-mw-4.pdf
          https://mtrevenue.gov/2021/11/29/simplification-of-montana-income-taxation/?utm_source=rss&utm_medium=rss&utm_campaign=simplification-of-montana-income-taxation
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'NewHampshire', 'state_occupation': 'Montana', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def new_hampshire_and_nebraska(self, state_agi=75000, capital_gains_short=0, capital_gains_long=0, filing_status='single', income=100000, dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        New Hampshire:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        Nebraska:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        New Hampshire:
        - https://www.revenue.nh.gov/assistance/tax-overview.htm
        
        Nebraska:
        -         
        - https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2023_Circular_EN_8-429r2023_Whole_Book_11.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'NewHampshire', 'state_occupation': 'Nebraska', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def new_hampshire_and_nevada(self, sources=False):
        """
        --- Parameters ----
        
        New Hampshire:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        Nevada:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        New Hampshire:
        - https://www.revenue.nh.gov/assistance/tax-overview.htm
        
        Nevada:
        -         
        - https://www.nvsos.gov/sos/businesses/the-nevada-advantage
        """
        
        # Check if the parameters are set upstream in the instance
        sources = self.sources if hasattr(self, 'sources') else sources
        
        # State-specific parameters
        data = {
            'state_residence': 'NewHampshire', 'state_occupation': 'Nevada', "sources": sources
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def new_hampshire_and_new_jersey(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, capital_gains_long=0, filing_status='single', income=100000, dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        New Hampshire:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        New Jersey:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        New Hampshire:
        - https://www.revenue.nh.gov/assistance/tax-overview.htm
        
        New Jersey:
        -         
        - https://www.state.nj.us/treasury/taxation/taxtables.shtml
          https://www.state.nj.us/treasury/taxation/njit13.shtml
          https://www.nj.gov/treasury/taxation/njit25.shtml
          https://www.state.nj.us/treasury/taxation/individuals/freqqiti.shtml
          https://www.state.nj.us/treasury/taxation/localtax.shtml
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'NewHampshire', 'state_occupation': 'NewJersey', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def new_hampshire_and_new_mexico(self, state_agi=75000, capital_gains_short=0, capital_gains_long=0, filing_status='single', income=100000, dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        New Hampshire:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        New Mexico:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        New Hampshire:
        - https://www.revenue.nh.gov/assistance/tax-overview.htm
        
        New Mexico:
        -         
        - https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/fdf3c548-8aba-4b9c-9eb4-bb564c716015/FYI-104.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/0bf943fd-652e-4400-bb1b-cd8397eb5a95/2022pit-adj.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/2f1a6781-9534-4436-b427-1557f9592099/2022pit-adj-ins.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/0bf943fd-652e-4400-bb1b-cd8397eb5a95/2022pit-adj.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/1afc56af-ea90-4d48-82e5-1f9aeb43255a/PITbook2022.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/fdf3c548-8aba-4b9c-9eb4-bb564c716015/FYI-104.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'NewHampshire', 'state_occupation': 'NewMexico', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def new_hampshire_and_new_york(self, county_residence='Other', state_agi=75000, capital_gains_short=0, capital_gains_long=0, filing_status='single', income=100000, dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        New Hampshire:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        New York:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        New Hampshire:
        - https://www.revenue.nh.gov/assistance/tax-overview.htm
        
        New York:
        -         
        - https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nys-tax-rate-schedule
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nys-tax-rate-schedule
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nys-tax-rate-schedule
          https://www.tax.ny.gov/pit/
          https://www.tax.ny.gov/pit/file/nonresident-faqs.htm
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#yonkers-tax-rate-schedule
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nyc-tax-rate-schedule
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'NewHampshire', 'state_occupation': 'NewYork', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def new_hampshire_and_north_carolina(self, state_agi=75000, capital_gains_short=0, capital_gains_long=0, filing_status='single', income=100000, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        New Hampshire:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        North Carolina:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        New Hampshire:
        - https://www.revenue.nh.gov/assistance/tax-overview.htm
        
        North Carolina:
        -         
        - https://www.ncdor.gov/taxes-forms/tax-rate-schedules
          https://www.ncdor.gov/taxes-forms/individual-income-tax/north-carolina-standard-deduction-or-north-carolina-itemized-deductions
          https://www.ncdor.gov/taxes-forms/individual-income-tax
          https://www.ncdor.gov/taxes-forms/individual-income-tax
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'NewHampshire', 'state_occupation': 'NorthCarolina', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def new_hampshire_and_north_dakota(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, capital_gains_long=0, filing_status='single', income=100000, dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        New Hampshire:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        North Dakota:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        New Hampshire:
        - https://www.revenue.nh.gov/assistance/tax-overview.htm
        
        North Dakota:
        -         
        - https://www.tax.nd.gov/forms
          https://www.tax.nd.gov/sites/www/files/documents/forms/individual/2022-iit/2022-individual-income-tax-booklet.pdf
          https://www.tax.nd.gov/sites/www/files/documents/forms/2022-individual-income-tax-booklet.pdf
          https://www.tax.nd.gov/tax-exemptions-credits/income-tax-exemptions-credits
          https://www.tax.nd.gov/sites/www/files/documents/forms/business/ndwrfillable.pdf
          https://www.tax.nd.gov/news/tax-legislative-changes/significant-changes-law/individual-income-tax-history
          https://www.tax.nd.gov/forms
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'NewHampshire', 'state_occupation': 'NorthDakota', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def new_hampshire_and_ohio(self, county_residence='Other', state_agi=75000, capital_gains_short=0, capital_gains_long=0, filing_status='single', income=100000, dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        New Hampshire:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        Ohio:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        New Hampshire:
        - https://www.revenue.nh.gov/assistance/tax-overview.htm
        
        Ohio:
        -         
        - https://tax.ohio.gov/individual/resources/annual-tax-rates
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://tax.ohio.gov/static/forms/employer_withholding/generic/wth_it4nr.pdf
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://www.ritaohio.com/TaxRatesTable
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'NewHampshire', 'state_occupation': 'Ohio', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def new_hampshire_and_oklahoma(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, capital_gains_long=0, filing_status='single', income=100000, dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        New Hampshire:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        Oklahoma:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        New Hampshire:
        - https://www.revenue.nh.gov/assistance/tax-overview.htm
        
        Oklahoma:
        -         
        - https://oklahoma.gov/content/dam/ok/en/tax/documents/resources/publications/businesses/withholding-tables/WHTables-2022.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'NewHampshire', 'state_occupation': 'Oklahoma', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def new_hampshire_and_oregon(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, capital_gains_long=0, filing_status='single', income=100000, dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        New Hampshire:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        Oregon:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        New Hampshire:
        - https://www.revenue.nh.gov/assistance/tax-overview.htm
        
        Oregon:
        -         
        - https://www.oregon.gov/dor/programs/individuals/Pages/PIT.aspx
          https://www.oregon.gov/dor/programs/individuals/Documents/Full%20year%20resident
         %20Form%20or-40%20filers.pdf
          https://www.oregon.gov/dor/programs/individuals/Documents/Part-year%20and%20nonresident
         %20Form%20or-40-P%20and%20or-40-N%20filers.pdf.pdf
          https://www.oregon.gov/dor/forms/FormsPubs/form-or-40_101-040_2022.pdf
          https://www.oregon.gov/dor/forms/FormsPubs/form-or-40_101-040_2022.pdf
          https://www.oregon.gov/dor/forms/FormsPubs/form-or-40_101-040_2022.pdf
          https://www.portland.gov/revenue/personal-tax
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'NewHampshire', 'state_occupation': 'Oregon', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def new_hampshire_and_pennsylvania(self, county_residence='Other', state_agi=75000, county_occupation='Other', capital_gains_short=0, capital_gains_long=0, income=100000, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        New Hampshire:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        Pennsylvania:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        county_residence : str, optional
            County of residence. Default is Other.
        county_occupation : str, optional
            County of occupation. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        New Hampshire:
        - https://www.revenue.nh.gov/assistance/tax-overview.htm
        
        Pennsylvania:
        -         
        - https://www.revenue.pa.gov/Tax%20Rates/Pages/default.aspx
          https://www.revenue.pa.gov/FormsandPublications/FormsforIndividuals/PIT/Documents/2022/2022_pa-40in.pdf
          https://www.revenue.pa.gov/FormsandPublications/FormsforBusinesses/EmployerWithholding/Documents/rev-419.pdf
          https://www.revenue.pa.gov/FormsandPublications/FormsforIndividuals/PIT/Documents/2022/2022_pa-40in.pdf
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-25.htm
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        county_occupation = self.county_occupation if hasattr(self, 'county_occupation') else county_occupation
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        income = self.income if hasattr(self, 'income') else income
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'NewHampshire', 'state_occupation': 'Pennsylvania', "county_residence": county_residence, "state_agi": state_agi, "county_occupation": county_occupation, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "income": income, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def new_hampshire_and_rhode_island(self, state_agi=75000, capital_gains_short=0, capital_gains_long=0, filing_status='single', income=100000, dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        New Hampshire:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        Rhode Island:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        New Hampshire:
        - https://www.revenue.nh.gov/assistance/tax-overview.htm
        
        Rhode Island:
        -         
        - https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2022_1040WE_w_0.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'NewHampshire', 'state_occupation': 'RhodeIsland', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def new_hampshire_and_south_carolina(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, capital_gains_long=0, filing_status='single', income=100000, dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        New Hampshire:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        South Carolina:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        New Hampshire:
        - https://www.revenue.nh.gov/assistance/tax-overview.htm
        
        South Carolina:
        -         
        - https://dor.sc.gov/resources-site/lawandpolicy/Advisory%20Opinions/IL22-15.pdf
          https://dor.sc.gov/forms-site/Forms/IITPacket_2022.pdf
          https://www.scstatehouse.gov/code/t12c006.php
          https://dor.sc.gov/resources-site/lawandpolicy/Advisory%20Opinions/IL22-15.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'NewHampshire', 'state_occupation': 'SouthCarolina', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def new_hampshire_and_south_dakota(self, sources=False):
        """
        --- Parameters ----
        
        New Hampshire:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        South Dakota:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        New Hampshire:
        - https://www.revenue.nh.gov/assistance/tax-overview.htm
        
        South Dakota:
        -         
        - https://dor.sd.gov/individuals/taxes/
        """
        
        # Check if the parameters are set upstream in the instance
        sources = self.sources if hasattr(self, 'sources') else sources
        
        # State-specific parameters
        data = {
            'state_residence': 'NewHampshire', 'state_occupation': 'SouthDakota', "sources": sources
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def new_hampshire_and_tennessee(self, sources=False):
        """
        --- Parameters ----
        
        New Hampshire:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        Tennessee:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        New Hampshire:
        - https://www.revenue.nh.gov/assistance/tax-overview.htm
        
        Tennessee:
        -         
        - https://www.tn.gov/revenue/taxes.html
        """
        
        # Check if the parameters are set upstream in the instance
        sources = self.sources if hasattr(self, 'sources') else sources
        
        # State-specific parameters
        data = {
            'state_residence': 'NewHampshire', 'state_occupation': 'Tennessee', "sources": sources
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def new_hampshire_and_texas(self, sources=False):
        """
        --- Parameters ----
        
        New Hampshire:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        Texas:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        New Hampshire:
        - https://www.revenue.nh.gov/assistance/tax-overview.htm
        
        Texas:
        -         
        - https://comptroller.texas.gov/taxes/
        """
        
        # Check if the parameters are set upstream in the instance
        sources = self.sources if hasattr(self, 'sources') else sources
        
        # State-specific parameters
        data = {
            'state_residence': 'NewHampshire', 'state_occupation': 'Texas', "sources": sources
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def new_hampshire_and_utah(self, state_agi=75000, capital_gains_short=0, capital_gains_long=0, filing_status='single', income=100000, dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        New Hampshire:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        Utah:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        New Hampshire:
        - https://www.revenue.nh.gov/assistance/tax-overview.htm
        
        Utah:
        -         
        - https://incometax.utah.gov/paying/tax-rates
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'NewHampshire', 'state_occupation': 'Utah', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def new_hampshire_and_vermont(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, capital_gains_long=0, filing_status='single', income=100000, dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        New Hampshire:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        Vermont:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        New Hampshire:
        - https://www.revenue.nh.gov/assistance/tax-overview.htm
        
        Vermont:
        -         
        - https://tax.vermont.gov/sites/tax/files/documents/GB-1210-2023.pdf
          https://tax.vermont.gov/sites/tax/files/documents/Income%20Booklet-2022.pdf
          https://tax.vermont.gov/sites/tax/files/documents/Income%20Booklet-2022.pdf
          https://tax.vermont.gov/sites/tax/files/documents/Income%20Booklet-2022.pdf
          https://tax.vermont.gov/individuals/personal-income-tax/tax-credits
          https://tax.vermont.gov/individuals/personal-income-tax/taxable-income
          https://tax.vermont.gov/individuals/personal-income-tax/local-option-tax
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'NewHampshire', 'state_occupation': 'Vermont', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def new_hampshire_and_virginia(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, capital_gains_long=0, filing_status='single', income=100000, dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        New Hampshire:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        Virginia:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        New Hampshire:
        - https://www.revenue.nh.gov/assistance/tax-overview.htm
        
        Virginia:
        -         
        - https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/reciprocity
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'NewHampshire', 'state_occupation': 'Virginia', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def new_hampshire_and_washington(self, capital_gains_long=0, sources=False, capital_gains_short=0):
        """
        --- Parameters ----
        
        New Hampshire:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        Washington:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        New Hampshire:
        - https://www.revenue.nh.gov/assistance/tax-overview.htm
        
        Washington:
        -         
        - https://dor.wa.gov/taxes-rates/income-tax
          https://dor.wa.gov/taxes-rates/other-taxes/capital-gains-tax
          https://dor.wa.gov/taxes-rates/income-tax
        """
        
        # Check if the parameters are set upstream in the instance
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        sources = self.sources if hasattr(self, 'sources') else sources
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        
        # State-specific parameters
        data = {
            'state_residence': 'NewHampshire', 'state_occupation': 'Washington', "capital_gains_long": capital_gains_long, "sources": sources, "capital_gains_short": capital_gains_short
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def new_hampshire_and_west_virginia(self, pay_periods=24, state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', county_occupation='Other', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        New Hampshire:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        West Virginia:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_occupation : str, optional
            County of occupation. Default is Other.
        pay_periods : int, optional
            Number of pay periods. Options are 13, 24, 26, 52. Default is 24.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        New Hampshire:
        - https://www.revenue.nh.gov/assistance/tax-overview.htm
        
        West Virginia:
        -         
        - https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://tax.wv.gov/Documents/TSD/tsd438.pdf
          https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2011/Taxes-11-35att.html
        """
        
        # Check if the parameters are set upstream in the instance
        pay_periods = self.pay_periods if hasattr(self, 'pay_periods') else pay_periods
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        county_occupation = self.county_occupation if hasattr(self, 'county_occupation') else county_occupation
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'NewHampshire', 'state_occupation': 'WestVirginia', "pay_periods": pay_periods, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "county_occupation": county_occupation, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def new_hampshire_and_wisconsin(self, state_agi=75000, capital_gains_short=0, capital_gains_long=0, filing_status='single', income=100000, dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        New Hampshire:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        Wisconsin:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        New Hampshire:
        - https://www.revenue.nh.gov/assistance/tax-overview.htm
        
        Wisconsin:
        -         
        - https://www.revenue.wi.gov/TaxForms2023/2023-Form1-ES-Inst.pdf
          https://www.revenue.wi.gov/TaxForms2023/2023-Form1-ES-Inst.pdf
          https://www.revenue.wi.gov/TaxForms2023/2023-Form1-ES-Inst.pdf
          https://www.revenue.wi.gov/Pages/FAQS/pcs-work.aspx
          https://www.revenue.wi.gov/Dor%20Publications/pb103.pdf
          https://www.revenue.wi.gov/Pages/FAQS/pcs-work.aspx
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'NewHampshire', 'state_occupation': 'Wisconsin', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def new_hampshire_and_wyoming(self, sources=False):
        """
        --- Parameters ----
        
        New Hampshire:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        Wyoming:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        New Hampshire:
        - https://www.revenue.nh.gov/assistance/tax-overview.htm
        
        Wyoming:
        -         
        - https://revenue.wyo.gov/
        """
        
        # Check if the parameters are set upstream in the instance
        sources = self.sources if hasattr(self, 'sources') else sources
        
        # State-specific parameters
        data = {
            'state_residence': 'NewHampshire', 'state_occupation': 'Wyoming', "sources": sources
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def new_jersey_and_alabama(self, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        New Jersey:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Alabama:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        New Jersey:
        - https://www.state.nj.us/treasury/taxation/taxtables.shtml
          https://www.state.nj.us/treasury/taxation/njit13.shtml
          https://www.nj.gov/treasury/taxation/njit25.shtml
          https://www.state.nj.us/treasury/taxation/individuals/freqqiti.shtml
          https://www.state.nj.us/treasury/taxation/localtax.shtml
        
        Alabama:
        -         
        - https://www.revenue.alabama.gov/ultraviewer/viewer/basic_viewer/index.html?form=2023/02/22f40.pdf
          https://www.revenue.alabama.gov/faqs/what-is-alabamas-individual-income-tax-rate/
          https://www.revenue.alabama.gov/ultraviewer/viewer/basic_viewer/index.html?form=2023/02/22f40.pdf
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-24.htm?taxmap=true
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-24.htm?taxmap=true
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-24.htm?taxmap=true
          https://www.revenue.alabama.gov/individual-corporate/income-to-be-reported-on-the-alabama-income-tax-return/
          https://www.almonline.org/TaxRates.aspx#OccupationalTax
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'NewJersey', 'state_occupation': 'Alabama', "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def new_jersey_and_alaska(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        New Jersey:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Alaska:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        New Jersey:
        - https://www.state.nj.us/treasury/taxation/taxtables.shtml
          https://www.state.nj.us/treasury/taxation/njit13.shtml
          https://www.nj.gov/treasury/taxation/njit25.shtml
          https://www.state.nj.us/treasury/taxation/individuals/freqqiti.shtml
          https://www.state.nj.us/treasury/taxation/localtax.shtml
        
        Alaska:
        -         
        - https://www.commerce.alaska.gov/web/dcra/officeofthestateassessor/alaskataxfacts.aspx
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'NewJersey', 'state_occupation': 'Alaska', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def new_jersey_and_arizona(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        New Jersey:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Arizona:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        credits_state : int, optional
            Credits for state taxes. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        New Jersey:
        - https://www.state.nj.us/treasury/taxation/taxtables.shtml
          https://www.state.nj.us/treasury/taxation/njit13.shtml
          https://www.nj.gov/treasury/taxation/njit25.shtml
          https://www.state.nj.us/treasury/taxation/individuals/freqqiti.shtml
          https://www.state.nj.us/treasury/taxation/localtax.shtml
        
        Arizona:
        -         
        - https://azdor.gov/forms/individual/form-140-x-y-tables
          https://azdor.gov/forms/individual/form-140-x-y-tables
          https://azdor.gov/forms/individual/form-140-x-y-tables
          https://azdor.gov/businesses-arizona/withholding-tax/withholding-exceptions
          "https://azdor.gov/individuals/income-tax-filing-assistance/identifying-other-taxable-income#:~:text=In%20regards%20to%20capital%20gains
         the%20individuals%20regular%20tax%20rate.
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'NewJersey', 'state_occupation': 'Arizona', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def new_jersey_and_arkansas(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        New Jersey:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Arkansas:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        New Jersey:
        - https://www.state.nj.us/treasury/taxation/taxtables.shtml
          https://www.state.nj.us/treasury/taxation/njit13.shtml
          https://www.nj.gov/treasury/taxation/njit25.shtml
          https://www.state.nj.us/treasury/taxation/individuals/freqqiti.shtml
          https://www.state.nj.us/treasury/taxation/localtax.shtml
        
        Arkansas:
        -         
        - https://www.arkansasedc.com/why-arkansas/business-climate/tax-structure/personal-income-tax
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_and_AR1000NR_Instructions.pdf
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_FullYearResidentIndividualIncomeTaxReturn.pdf
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_and_AR1000NR_Instructions.pdf
          https://www.arkansasedc.com/why-arkansas/business-climate/tax-structure/capital-gains-tax-reduction
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_and_AR1000NR_Instructions.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'NewJersey', 'state_occupation': 'Arkansas', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def new_jersey_and_california(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        New Jersey:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        California:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        New Jersey:
        - https://www.state.nj.us/treasury/taxation/taxtables.shtml
          https://www.state.nj.us/treasury/taxation/njit13.shtml
          https://www.nj.gov/treasury/taxation/njit25.shtml
          https://www.state.nj.us/treasury/taxation/individuals/freqqiti.shtml
          https://www.state.nj.us/treasury/taxation/localtax.shtml
        
        California:
        -         
        - https://www.ftb.ca.gov/forms/2022/2022-540-tax-rate-schedules.pdf
          https://www.ftb.ca.gov/file/personal/deductions/index.html
          https://www.ftb.ca.gov/about-ftb/newsroom/caleitc/eligibility-and-credit-information.html
          https://www.ftb.ca.gov/file/personal/income-types/capital-gains-and-losses.html#:~:text=California%20does%20not%20have%20a
         are%20taxed%20as%20ordinary%20income
          https://www.ftb.ca.gov/file/personal/
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'NewJersey', 'state_occupation': 'California', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def new_jersey_and_colorado(self, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        New Jersey:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Colorado:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        New Jersey:
        - https://www.state.nj.us/treasury/taxation/taxtables.shtml
          https://www.state.nj.us/treasury/taxation/njit13.shtml
          https://www.nj.gov/treasury/taxation/njit25.shtml
          https://www.state.nj.us/treasury/taxation/individuals/freqqiti.shtml
          https://www.state.nj.us/treasury/taxation/localtax.shtml
        
        Colorado:
        -         
        - https://leg.colorado.gov/agencies/legislative-council-staff/individual-income-tax%C2%A0
          https://leg.colorado.gov/agencies/legislative-council-staff/individual-income-tax%C2%A0
          https://tax.colorado.gov/sites/tax/files/documents/ITT_Capital_Gain_Subtraction_Feb_2022.pdf
          https://www.auroragov.org/business_services/taxes/occupational_privilege_tax
          https://www.denvergov.org/Government/Agencies-Departments-Offices/Agencies-Departments-Offices-Directory/Department-of-Finance/Our-Divisions/Treasury/Business-Tax-Information#:~:text=Occupational%20privilege%20tax
         -%28show%20below%29&text=Each%20taxable%20employee%20is%20liable
         month%20for%20each%20taxable%20employee
          https://www.glendale.co.us/355/Occupational-Privilege-Tax
          https://www.greenwoodvillage.com/1220/Occupational-Privilege-Tax-OPT
          https://ci.sheridan.co.us/288/Occupational-Privilege-Tax
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'NewJersey', 'state_occupation': 'Colorado', "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def new_jersey_and_connecticut(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        New Jersey:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Connecticut:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        New Jersey:
        - https://www.state.nj.us/treasury/taxation/taxtables.shtml
          https://www.state.nj.us/treasury/taxation/njit13.shtml
          https://www.nj.gov/treasury/taxation/njit25.shtml
          https://www.state.nj.us/treasury/taxation/individuals/freqqiti.shtml
          https://www.state.nj.us/treasury/taxation/localtax.shtml
        
        Connecticut:
        -         
        - https://portal.ct.gov/-/media/DRS/Forms/2022/Income/2022-CT-1040-Instructions_1222.pdf
          https://www.cga.ct.gov/2022/rpt/pdf/2022-R-0108.pdf
          https://www.cga.ct.gov/current/pub/chap_229.htm#sec_12-703
          https://portal.ct.gov/-/media/DRS/Forms/2022/Income/2022-CT-1040-Instructions_1222.pdf
          "https://portal.ct.gov/DRS/Publications/TSSNs/TSSN-29#:~:text=An%20individuals%20net%20capital%20gains
         portion%20of%20Social%20Security%20benefits"
          "https://portal.ct.gov/DRS/Publications/TSSNs/TSSN-29#:~:text=An%20individuals%20net%20capital%20gains
         portion%20of%20Social%20Security%20benefits"
          https://portal.ct.gov/DRS/Individuals/Resident-Income-Tax/Tax-Information
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'NewJersey', 'state_occupation': 'Connecticut', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def new_jersey_and_delaware(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, federal_agi=75000, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        New Jersey:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Delaware:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        federal_agi : int, optional
            Adjusted Gross Income (AGI) for federal taxes. Default is 0. Sources - https://www.irs.gov/e-file-providers/definition-of-adjusted-gross-income, https://www.irs.gov/newsroom/taxpayers-should-know-the-difference-between-standard-and-itemized-deductions
        income : int, optional
            Gross income. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        New Jersey:
        - https://www.state.nj.us/treasury/taxation/taxtables.shtml
          https://www.state.nj.us/treasury/taxation/njit13.shtml
          https://www.nj.gov/treasury/taxation/njit25.shtml
          https://www.state.nj.us/treasury/taxation/individuals/freqqiti.shtml
          https://www.state.nj.us/treasury/taxation/localtax.shtml
        
        Delaware:
        -         
        - https://revenue.delaware.gov/software-developer/tax-rate-changes/
          https://revenuefiles.delaware.gov/2022/PIT-RES_TY22_2022-01_PaperInteractive.pdf
          https://revenuefiles.delaware.gov/2022/PIT-RES_TY22_2022-02_Instructions.pdf
          https://revenue.delaware.gov/frequently-asked-questions/personal-income-tax-faqs/
          https://www.newcastlede.gov/DocumentCenter/View/20147/City-of-Wilmington-Wage-Tax-Information-PDF
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        federal_agi = self.federal_agi if hasattr(self, 'federal_agi') else federal_agi
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'NewJersey', 'state_occupation': 'Delaware', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "federal_agi": federal_agi, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def new_jersey_and_florida(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        New Jersey:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Florida:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        New Jersey:
        - https://www.state.nj.us/treasury/taxation/taxtables.shtml
          https://www.state.nj.us/treasury/taxation/njit13.shtml
          https://www.nj.gov/treasury/taxation/njit25.shtml
          https://www.state.nj.us/treasury/taxation/individuals/freqqiti.shtml
          https://www.state.nj.us/treasury/taxation/localtax.shtml
        
        Florida:
        -         
        - https://floridarevenue.com/taxes/taxesfees/Pages/tax_interest_rates.aspx
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'NewJersey', 'state_occupation': 'Florida', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def new_jersey_and_georgia(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        New Jersey:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Georgia:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        New Jersey:
        - https://www.state.nj.us/treasury/taxation/taxtables.shtml
          https://www.state.nj.us/treasury/taxation/njit13.shtml
          https://www.nj.gov/treasury/taxation/njit25.shtml
          https://www.state.nj.us/treasury/taxation/individuals/freqqiti.shtml
          https://www.state.nj.us/treasury/taxation/localtax.shtml
        
        Georgia:
        -         
        - https://dor.georgia.gov/tax-tables-georgia-tax-rate-schedule
          https://apps.dor.ga.gov/FillableForms/PDFViewer/Index?form=2022GA500
          https://apps.dor.ga.gov/FillableForms/PDFViewer/Index?form=2022GA500
          https://apps.dor.ga.gov/FillableForms/PDFViewer/Index?form=2022GA500
          https://dor.georgia.gov/years-individual-income-tax-forms
          https://dor.georgia.gov/years-individual-income-tax-forms
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'NewJersey', 'state_occupation': 'Georgia', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def new_jersey_and_hawaii(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        New Jersey:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Hawaii:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        New Jersey:
        - https://www.state.nj.us/treasury/taxation/taxtables.shtml
          https://www.state.nj.us/treasury/taxation/njit13.shtml
          https://www.nj.gov/treasury/taxation/njit25.shtml
          https://www.state.nj.us/treasury/taxation/individuals/freqqiti.shtml
          https://www.state.nj.us/treasury/taxation/localtax.shtml
        
        Hawaii:
        -         
        - https://tax.hawaii.gov/forms/d_18table-on/d_18table-on_p13/
          https://files.hawaii.gov/tax/forms/2018/18table-on.pdf
          https://files.hawaii.gov/tax/news/pubs/22outline.pdf
          https://files.hawaii.gov/tax/news/pubs/FastTaxRefGuide_2017.pdf
          https://www.tfhawaii.org/wordpress/state-tax-resources/mini-tax-guides/individual-income-tax-chapter-235/#:~:text=Long%20term%20capital%20gains%20are%20taxed%20at%20a%20maximum%20of%207.25%25
          https://www.grassrootinstitute.org/2023/02/hb337-increasing-tax-on-capital-gains-a-bad-risk/
          https://files.hawaii.gov/tax/news/pubs/22outline.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'NewJersey', 'state_occupation': 'Hawaii', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def new_jersey_and_idaho(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        New Jersey:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Idaho:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        New Jersey:
        - https://www.state.nj.us/treasury/taxation/taxtables.shtml
          https://www.state.nj.us/treasury/taxation/njit13.shtml
          https://www.nj.gov/treasury/taxation/njit25.shtml
          https://www.state.nj.us/treasury/taxation/individuals/freqqiti.shtml
          https://www.state.nj.us/treasury/taxation/localtax.shtml
        
        Idaho:
        -         
        - https://tax.idaho.gov/taxes/income-tax/individual-income/individual-income-tax-rate-schedule/
          https://tax.idaho.gov/wp-content/uploads/forms/EIN00046/EIN00046_03-01-2023.pdf
          https://tax.idaho.gov/taxes/income-tax/individual-income/idaho-source-income/
          https://tax.idaho.gov/taxes/income-tax/individual-income/idaho-source-income/
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'NewJersey', 'state_occupation': 'Idaho', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def new_jersey_and_illinois(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        New Jersey:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Illinois:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        New Jersey:
        - https://www.state.nj.us/treasury/taxation/taxtables.shtml
          https://www.state.nj.us/treasury/taxation/njit13.shtml
          https://www.nj.gov/treasury/taxation/njit25.shtml
          https://www.state.nj.us/treasury/taxation/individuals/freqqiti.shtml
          https://www.state.nj.us/treasury/taxation/localtax.shtml
        
        Illinois:
        -         
        - https://tax.illinois.gov/research/taxrates/income.html#IndividualIncome
          https://tax.illinois.gov/content/dam/soi/en/web/tax/forms/incometax/documents/currentyear/individual/il-1040-instr.pdf
          https://tax.illinois.gov/questionsandanswers/12.html
          https://tax.illinois.gov/content/dam/soi/en/web/tax/forms/incometax/documents/currentyear/individual/il-1040-instr.pdf
          https://tax.illinois.gov/research/taxrates/personalproperty.html
          https://tax.illinois.gov/localgovernments/income.html
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'NewJersey', 'state_occupation': 'Illinois', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def new_jersey_and_indiana(self, capital_gains_short=0, dependents=0, sources=False, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_long=0, filing_status='single', income=100000, credits_state=0):
        """
        --- Parameters ----
        
        New Jersey:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Indiana:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        New Jersey:
        - https://www.state.nj.us/treasury/taxation/taxtables.shtml
          https://www.state.nj.us/treasury/taxation/njit13.shtml
          https://www.nj.gov/treasury/taxation/njit25.shtml
          https://www.state.nj.us/treasury/taxation/individuals/freqqiti.shtml
          https://www.state.nj.us/treasury/taxation/localtax.shtml
        
        Indiana:
        -         
        - https://www.in.gov/dor/tax-forms/2022-individual-income-tax-forms/
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-32.htm
          https://www.in.gov/dor/files/ib28.pdf
          https://www.in.gov/dor/tax-forms/2022-individual-income-tax-forms/
          https://www.in.gov/dor/files/2022-county-tax-rates-and-codes.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'NewJersey', 'state_occupation': 'Indiana', "capital_gains_short": capital_gains_short, "dependents": dependents, "sources": sources, "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def new_jersey_and_iowa(self, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        New Jersey:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Iowa:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        New Jersey:
        - https://www.state.nj.us/treasury/taxation/taxtables.shtml
          https://www.state.nj.us/treasury/taxation/njit13.shtml
          https://www.nj.gov/treasury/taxation/njit25.shtml
          https://www.state.nj.us/treasury/taxation/individuals/freqqiti.shtml
          https://www.state.nj.us/treasury/taxation/localtax.shtml
        
        Iowa:
        -         
        - https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
          https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
          https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
          https://tax.iowa.gov/iowa-illinois-reciprocal-agreement
          https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'NewJersey', 'state_occupation': 'Iowa', "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def new_jersey_and_kansas(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        New Jersey:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Kansas:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        New Jersey:
        - https://www.state.nj.us/treasury/taxation/taxtables.shtml
          https://www.state.nj.us/treasury/taxation/njit13.shtml
          https://www.nj.gov/treasury/taxation/njit25.shtml
          https://www.state.nj.us/treasury/taxation/individuals/freqqiti.shtml
          https://www.state.nj.us/treasury/taxation/localtax.shtml
        
        Kansas:
        -         
        - https://www.ksrevenue.gov/taxrates.html
          https://www.ksrevenue.gov/incomebook22.html#0
          https://www.ksrevenue.gov/incomebook22.html#0
          https://www.ksrevenue.gov/incomebook22.html#0
          https://www.ksrevenue.gov/incomebook22.html
          https://www.ksrevenue.gov/incomebook22.html
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'NewJersey', 'state_occupation': 'Kansas', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def new_jersey_and_kentucky(self, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        New Jersey:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Kentucky:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        New Jersey:
        - https://www.state.nj.us/treasury/taxation/taxtables.shtml
          https://www.state.nj.us/treasury/taxation/njit13.shtml
          https://www.nj.gov/treasury/taxation/njit25.shtml
          https://www.state.nj.us/treasury/taxation/individuals/freqqiti.shtml
          https://www.state.nj.us/treasury/taxation/localtax.shtml
        
        Kentucky:
        -         
        - https://revenue.ky.gov/News/Pages/Dor-Announces-Updates-to-Individual-Income-Tax-for-2023-Tax-Year.aspx
          https://revenue.ky.gov/Individual/Individual-Income-Tax/Pages/default.aspx
          https://revenue.ky.gov/News/Pages/Dor-Announces-Updates-to-Individual-Income-Tax-for-2023-Tax-Year.aspx
          https://revenue.ky.gov/Individual/Individual-Income-Tax/Pages/default.aspx
          https://revenue.ky.gov/Dor%20Training%20Materials/103%20KAR%2017.140.%20Individual%20income%20tax%20-%20reciprocity%20-%20nonresidents.pdf
          https://revenue.ky.gov/Individual/Individual-Income-Tax/Pages/default.aspx
          https://nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2021/Taxes-21-32.htm?taxmap=true
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'NewJersey', 'state_occupation': 'Kentucky', "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def new_jersey_and_louisiana(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        New Jersey:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Louisiana:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        New Jersey:
        - https://www.state.nj.us/treasury/taxation/taxtables.shtml
          https://www.state.nj.us/treasury/taxation/njit13.shtml
          https://www.nj.gov/treasury/taxation/njit25.shtml
          https://www.state.nj.us/treasury/taxation/individuals/freqqiti.shtml
          https://www.state.nj.us/treasury/taxation/localtax.shtml
        
        Louisiana:
        -         
        - https://revenue.louisiana.gov/individualincometax
          https://revenue.louisiana.gov/individualincometax
          https://www.revenue.louisiana.gov/taxforms/6935(11_02)F.pdf
          https://revenue.louisiana.gov/individualincometax
          https://revenue.louisiana.gov/individualincometax
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'NewJersey', 'state_occupation': 'Louisiana', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def new_jersey_and_maine(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        New Jersey:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Maine:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        New Jersey:
        - https://www.state.nj.us/treasury/taxation/taxtables.shtml
          https://www.state.nj.us/treasury/taxation/njit13.shtml
          https://www.nj.gov/treasury/taxation/njit25.shtml
          https://www.state.nj.us/treasury/taxation/individuals/freqqiti.shtml
          https://www.state.nj.us/treasury/taxation/localtax.shtml
        
        Maine:
        -         
        - https://www.maine.gov/revenue/tax-return-forms/individual-income-tax-2022
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/ind_tax_rate_sched_2022.pdf
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/22_1040me_book_gen_instruc_revisedFeb23.pdf
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/22_1040me_book_gen_instruc_revisedFeb23.pdf
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/22_1040me_book_gen_instruc_revisedFeb23.pdf
          https://www.maine.gov/revenue/tax-return-forms/individual-income-tax-2022
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'NewJersey', 'state_occupation': 'Maine', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def new_jersey_and_maryland(self, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        New Jersey:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Maryland:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        New Jersey:
        - https://www.state.nj.us/treasury/taxation/taxtables.shtml
          https://www.state.nj.us/treasury/taxation/njit13.shtml
          https://www.nj.gov/treasury/taxation/njit25.shtml
          https://www.state.nj.us/treasury/taxation/individuals/freqqiti.shtml
          https://www.state.nj.us/treasury/taxation/localtax.shtml
        
        Maryland:
        -         
        - https://www.marylandtaxes.gov/individual/income/tax-info/tax-rates.php
          https://www.marylandtaxes.gov/forms/22_forms/Resident_Booklet.pdf
          https://www.marylandtaxes.gov/forms/Tax_Publications/Administrative_Releases/Income_and_Estate_Tax_Releases/ar_it37.pdf
          https://www.marylandtaxes.gov/individual/index.php
          https://www.marylandtaxes.gov/individual/income/tax-info/tax-rates.php
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'NewJersey', 'state_occupation': 'Maryland', "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def new_jersey_and_massachusetts(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        New Jersey:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Massachusetts:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        New Jersey:
        - https://www.state.nj.us/treasury/taxation/taxtables.shtml
          https://www.state.nj.us/treasury/taxation/njit13.shtml
          https://www.nj.gov/treasury/taxation/njit25.shtml
          https://www.state.nj.us/treasury/taxation/individuals/freqqiti.shtml
          https://www.state.nj.us/treasury/taxation/localtax.shtml
        
        Massachusetts:
        -         
        - https://www.mass.gov/service-details/massachusetts-tax-rates
          https://www.mass.gov/service-details/view-massachusetts-personal-income-tax-exemptions
          https://www.mass.gov/service-details/view-massachusetts-personal-income-tax-exemptions
          https://www.mass.gov/service-details/view-child-and-dependent-related-credits
          https://www.mass.gov/service-details/massachusetts-tax-rates
          https://www.mass.gov/personal-income-tax
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'NewJersey', 'state_occupation': 'Massachusetts', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def new_jersey_and_michigan(self, capital_gains_short=0, county_occupation='Other', dependents=0, sources=False, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_long=0, filing_status='single', income=100000, credits_state=0):
        """
        --- Parameters ----
        
        New Jersey:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Michigan:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        county_occupation : str, optional
            County of occupation. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        New Jersey:
        - https://www.state.nj.us/treasury/taxation/taxtables.shtml
          https://www.state.nj.us/treasury/taxation/njit13.shtml
          https://www.nj.gov/treasury/taxation/njit25.shtml
          https://www.state.nj.us/treasury/taxation/individuals/freqqiti.shtml
          https://www.state.nj.us/treasury/taxation/localtax.shtml
        
        Michigan:
        -         
        - https://www.michigan.gov/taxes/iit/new-developments-for-tax-year-2022
          "https://www.michigan.gov/taxes/-/media/Project/Websites/taxes/Forms/2023/2023-SUW/446_Withholding-Guide_2023.pdf?rev=177ca3d8cb034392bf84e317ed78c320&hash=19C5F42F43A89A91863D229349F48883#:~:text=The%20exemption%20amount%20is%20%245%2C400
         employees%20Michigan%20income%20tax%20return"
          "https://www.michigan.gov/taxes/-/media/Project/Websites/taxes/Forms/2023/2023-SUW/446_Withholding-Guide_2023.pdf?rev=177ca3d8cb034392bf84e317ed78c320&hash=19C5F42F43A89A91863D229349F48883#:~:text=The%20exemption%20amount%20is%20%245%2C400
         employees%20Michigan%20income%20tax%20return"
          https://www.michigan.gov/taxes/questions/iit/accordion/residency/are-my-wages-earned-in-another-state-taxable-in-michigan-if-i-am-a-michigan-resident-1
          https://www.michigan.gov/taxes/-/media/Project/Websites/taxes/Forms/2022/2022-IIT-Forms/BOOK_MI-1040.pdf?rev=4524f1cbbc0243b194ab05a6680be75b&hash=40DFE71E651F7C1D65C31F6CD03D944F
          https://www.michigan.gov/taxes/questions/iit/accordion/general/what-cities-impose-an-income-tax
        """
        
        # Check if the parameters are set upstream in the instance
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        county_occupation = self.county_occupation if hasattr(self, 'county_occupation') else county_occupation
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'NewJersey', 'state_occupation': 'Michigan', "capital_gains_short": capital_gains_short, "county_occupation": county_occupation, "dependents": dependents, "sources": sources, "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def new_jersey_and_minnesota(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        New Jersey:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Minnesota:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        New Jersey:
        - https://www.state.nj.us/treasury/taxation/taxtables.shtml
          https://www.state.nj.us/treasury/taxation/njit13.shtml
          https://www.nj.gov/treasury/taxation/njit25.shtml
          https://www.state.nj.us/treasury/taxation/individuals/freqqiti.shtml
          https://www.state.nj.us/treasury/taxation/localtax.shtml
        
        Minnesota:
        -         
        - https://www.revenue.state.mn.us/minnesota-income-tax-rates-and-brackets
          https://www.revenue.state.mn.us/child-and-dependent-care-credit
          https://www.revenue.state.mn.us/dependent-exemptions
          https://www.revenue.state.mn.us/minnesota-standard-deduction
          https://mn.gov/mmb/assets/mwr_form_tcm1059-128581.pdf
          https://www.revenue.state.mn.us/individual-income-tax
          https://www.revenue.state.mn.us/individual-income-tax
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'NewJersey', 'state_occupation': 'Minnesota', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def new_jersey_and_mississippi(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        New Jersey:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Mississippi:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        New Jersey:
        - https://www.state.nj.us/treasury/taxation/taxtables.shtml
          https://www.state.nj.us/treasury/taxation/njit13.shtml
          https://www.nj.gov/treasury/taxation/njit25.shtml
          https://www.state.nj.us/treasury/taxation/individuals/freqqiti.shtml
          https://www.state.nj.us/treasury/taxation/localtax.shtml
        
        Mississippi:
        -         
        - https://www.dor.ms.gov/individual/tax-rates
          https://www.dor.ms.gov/individual/tax-rates
          https://www.dor.ms.gov/individual/tax-rates
          https://www.dor.ms.gov/individual
          https://www.dor.ms.gov/individual
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'NewJersey', 'state_occupation': 'Mississippi', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def new_jersey_and_missouri(self, capital_gains_short=0, dependents=0, sources=False, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_long=0, filing_status='single', income=100000, credits_state=0):
        """
        --- Parameters ----
        
        New Jersey:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Missouri:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        New Jersey:
        - https://www.state.nj.us/treasury/taxation/taxtables.shtml
          https://www.state.nj.us/treasury/taxation/njit13.shtml
          https://www.nj.gov/treasury/taxation/njit25.shtml
          https://www.state.nj.us/treasury/taxation/individuals/freqqiti.shtml
          https://www.state.nj.us/treasury/taxation/localtax.shtml
        
        Missouri:
        -         
        - https://dor.mo.gov/taxation/individual/tax-types/income/year-changes/
          https://dor.mo.gov/taxation/individual/tax-types/income/year-changes/
          https://dor.mo.gov/taxation/individual/tax-types/income/year-changes/
          https://www.kcmo.gov/city-hall/departments/finance/earnings-tax
          https://www.stlouis-mo.gov/government/departments/collector/earnings-tax/individual-earnings-tax-info.cfm
        """
        
        # Check if the parameters are set upstream in the instance
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'NewJersey', 'state_occupation': 'Missouri', "capital_gains_short": capital_gains_short, "dependents": dependents, "sources": sources, "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def new_jersey_and_montana(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        New Jersey:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Montana:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        New Jersey:
        - https://www.state.nj.us/treasury/taxation/taxtables.shtml
          https://www.state.nj.us/treasury/taxation/njit13.shtml
          https://www.nj.gov/treasury/taxation/njit25.shtml
          https://www.state.nj.us/treasury/taxation/individuals/freqqiti.shtml
          https://www.state.nj.us/treasury/taxation/localtax.shtml
        
        Montana:
        -         
        - https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/2022/12/montana-employees-withholding-allowance-and-exemption-certificate-form-mw-4.pdf
          https://mtrevenue.gov/2021/11/29/simplification-of-montana-income-taxation/?utm_source=rss&utm_medium=rss&utm_campaign=simplification-of-montana-income-taxation
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'NewJersey', 'state_occupation': 'Montana', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def new_jersey_and_nebraska(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        New Jersey:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Nebraska:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        New Jersey:
        - https://www.state.nj.us/treasury/taxation/taxtables.shtml
          https://www.state.nj.us/treasury/taxation/njit13.shtml
          https://www.nj.gov/treasury/taxation/njit25.shtml
          https://www.state.nj.us/treasury/taxation/individuals/freqqiti.shtml
          https://www.state.nj.us/treasury/taxation/localtax.shtml
        
        Nebraska:
        -         
        - https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2023_Circular_EN_8-429r2023_Whole_Book_11.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'NewJersey', 'state_occupation': 'Nebraska', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def new_jersey_and_nevada(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        New Jersey:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Nevada:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        New Jersey:
        - https://www.state.nj.us/treasury/taxation/taxtables.shtml
          https://www.state.nj.us/treasury/taxation/njit13.shtml
          https://www.nj.gov/treasury/taxation/njit25.shtml
          https://www.state.nj.us/treasury/taxation/individuals/freqqiti.shtml
          https://www.state.nj.us/treasury/taxation/localtax.shtml
        
        Nevada:
        -         
        - https://www.nvsos.gov/sos/businesses/the-nevada-advantage
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'NewJersey', 'state_occupation': 'Nevada', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def new_jersey_and_new_hampshire(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        New Jersey:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        New Hampshire:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        New Jersey:
        - https://www.state.nj.us/treasury/taxation/taxtables.shtml
          https://www.state.nj.us/treasury/taxation/njit13.shtml
          https://www.nj.gov/treasury/taxation/njit25.shtml
          https://www.state.nj.us/treasury/taxation/individuals/freqqiti.shtml
          https://www.state.nj.us/treasury/taxation/localtax.shtml
        
        New Hampshire:
        -         
        - https://www.revenue.nh.gov/assistance/tax-overview.htm
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'NewJersey', 'state_occupation': 'NewHampshire', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def new_jersey_and_new_mexico(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        New Jersey:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        New Mexico:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        New Jersey:
        - https://www.state.nj.us/treasury/taxation/taxtables.shtml
          https://www.state.nj.us/treasury/taxation/njit13.shtml
          https://www.nj.gov/treasury/taxation/njit25.shtml
          https://www.state.nj.us/treasury/taxation/individuals/freqqiti.shtml
          https://www.state.nj.us/treasury/taxation/localtax.shtml
        
        New Mexico:
        -         
        - https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/fdf3c548-8aba-4b9c-9eb4-bb564c716015/FYI-104.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/0bf943fd-652e-4400-bb1b-cd8397eb5a95/2022pit-adj.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/2f1a6781-9534-4436-b427-1557f9592099/2022pit-adj-ins.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/0bf943fd-652e-4400-bb1b-cd8397eb5a95/2022pit-adj.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/1afc56af-ea90-4d48-82e5-1f9aeb43255a/PITbook2022.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/fdf3c548-8aba-4b9c-9eb4-bb564c716015/FYI-104.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'NewJersey', 'state_occupation': 'NewMexico', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def new_jersey_and_new_york(self, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        New Jersey:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        New York:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        New Jersey:
        - https://www.state.nj.us/treasury/taxation/taxtables.shtml
          https://www.state.nj.us/treasury/taxation/njit13.shtml
          https://www.nj.gov/treasury/taxation/njit25.shtml
          https://www.state.nj.us/treasury/taxation/individuals/freqqiti.shtml
          https://www.state.nj.us/treasury/taxation/localtax.shtml
        
        New York:
        -         
        - https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nys-tax-rate-schedule
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nys-tax-rate-schedule
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nys-tax-rate-schedule
          https://www.tax.ny.gov/pit/
          https://www.tax.ny.gov/pit/file/nonresident-faqs.htm
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#yonkers-tax-rate-schedule
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nyc-tax-rate-schedule
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'NewJersey', 'state_occupation': 'NewYork', "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def new_jersey_and_north_carolina(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        New Jersey:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        North Carolina:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        New Jersey:
        - https://www.state.nj.us/treasury/taxation/taxtables.shtml
          https://www.state.nj.us/treasury/taxation/njit13.shtml
          https://www.nj.gov/treasury/taxation/njit25.shtml
          https://www.state.nj.us/treasury/taxation/individuals/freqqiti.shtml
          https://www.state.nj.us/treasury/taxation/localtax.shtml
        
        North Carolina:
        -         
        - https://www.ncdor.gov/taxes-forms/tax-rate-schedules
          https://www.ncdor.gov/taxes-forms/individual-income-tax/north-carolina-standard-deduction-or-north-carolina-itemized-deductions
          https://www.ncdor.gov/taxes-forms/individual-income-tax
          https://www.ncdor.gov/taxes-forms/individual-income-tax
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'NewJersey', 'state_occupation': 'NorthCarolina', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def new_jersey_and_north_dakota(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        New Jersey:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        North Dakota:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        New Jersey:
        - https://www.state.nj.us/treasury/taxation/taxtables.shtml
          https://www.state.nj.us/treasury/taxation/njit13.shtml
          https://www.nj.gov/treasury/taxation/njit25.shtml
          https://www.state.nj.us/treasury/taxation/individuals/freqqiti.shtml
          https://www.state.nj.us/treasury/taxation/localtax.shtml
        
        North Dakota:
        -         
        - https://www.tax.nd.gov/forms
          https://www.tax.nd.gov/sites/www/files/documents/forms/individual/2022-iit/2022-individual-income-tax-booklet.pdf
          https://www.tax.nd.gov/sites/www/files/documents/forms/2022-individual-income-tax-booklet.pdf
          https://www.tax.nd.gov/tax-exemptions-credits/income-tax-exemptions-credits
          https://www.tax.nd.gov/sites/www/files/documents/forms/business/ndwrfillable.pdf
          https://www.tax.nd.gov/news/tax-legislative-changes/significant-changes-law/individual-income-tax-history
          https://www.tax.nd.gov/forms
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'NewJersey', 'state_occupation': 'NorthDakota', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def new_jersey_and_ohio(self, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        New Jersey:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Ohio:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        New Jersey:
        - https://www.state.nj.us/treasury/taxation/taxtables.shtml
          https://www.state.nj.us/treasury/taxation/njit13.shtml
          https://www.nj.gov/treasury/taxation/njit25.shtml
          https://www.state.nj.us/treasury/taxation/individuals/freqqiti.shtml
          https://www.state.nj.us/treasury/taxation/localtax.shtml
        
        Ohio:
        -         
        - https://tax.ohio.gov/individual/resources/annual-tax-rates
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://tax.ohio.gov/static/forms/employer_withholding/generic/wth_it4nr.pdf
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://www.ritaohio.com/TaxRatesTable
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'NewJersey', 'state_occupation': 'Ohio', "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def new_jersey_and_oklahoma(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        New Jersey:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Oklahoma:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        New Jersey:
        - https://www.state.nj.us/treasury/taxation/taxtables.shtml
          https://www.state.nj.us/treasury/taxation/njit13.shtml
          https://www.nj.gov/treasury/taxation/njit25.shtml
          https://www.state.nj.us/treasury/taxation/individuals/freqqiti.shtml
          https://www.state.nj.us/treasury/taxation/localtax.shtml
        
        Oklahoma:
        -         
        - https://oklahoma.gov/content/dam/ok/en/tax/documents/resources/publications/businesses/withholding-tables/WHTables-2022.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'NewJersey', 'state_occupation': 'Oklahoma', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def new_jersey_and_oregon(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        New Jersey:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Oregon:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        New Jersey:
        - https://www.state.nj.us/treasury/taxation/taxtables.shtml
          https://www.state.nj.us/treasury/taxation/njit13.shtml
          https://www.nj.gov/treasury/taxation/njit25.shtml
          https://www.state.nj.us/treasury/taxation/individuals/freqqiti.shtml
          https://www.state.nj.us/treasury/taxation/localtax.shtml
        
        Oregon:
        -         
        - https://www.oregon.gov/dor/programs/individuals/Pages/PIT.aspx
          https://www.oregon.gov/dor/programs/individuals/Documents/Full%20year%20resident
         %20Form%20or-40%20filers.pdf
          https://www.oregon.gov/dor/programs/individuals/Documents/Part-year%20and%20nonresident
         %20Form%20or-40-P%20and%20or-40-N%20filers.pdf.pdf
          https://www.oregon.gov/dor/forms/FormsPubs/form-or-40_101-040_2022.pdf
          https://www.oregon.gov/dor/forms/FormsPubs/form-or-40_101-040_2022.pdf
          https://www.oregon.gov/dor/forms/FormsPubs/form-or-40_101-040_2022.pdf
          https://www.portland.gov/revenue/personal-tax
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'NewJersey', 'state_occupation': 'Oregon', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def new_jersey_and_pennsylvania(self, county_residence='Other', state_agi=75000, county_occupation='Other', filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        New Jersey:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Pennsylvania:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        county_residence : str, optional
            County of residence. Default is Other.
        county_occupation : str, optional
            County of occupation. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        New Jersey:
        - https://www.state.nj.us/treasury/taxation/taxtables.shtml
          https://www.state.nj.us/treasury/taxation/njit13.shtml
          https://www.nj.gov/treasury/taxation/njit25.shtml
          https://www.state.nj.us/treasury/taxation/individuals/freqqiti.shtml
          https://www.state.nj.us/treasury/taxation/localtax.shtml
        
        Pennsylvania:
        -         
        - https://www.revenue.pa.gov/Tax%20Rates/Pages/default.aspx
          https://www.revenue.pa.gov/FormsandPublications/FormsforIndividuals/PIT/Documents/2022/2022_pa-40in.pdf
          https://www.revenue.pa.gov/FormsandPublications/FormsforBusinesses/EmployerWithholding/Documents/rev-419.pdf
          https://www.revenue.pa.gov/FormsandPublications/FormsforIndividuals/PIT/Documents/2022/2022_pa-40in.pdf
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-25.htm
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        county_occupation = self.county_occupation if hasattr(self, 'county_occupation') else county_occupation
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'NewJersey', 'state_occupation': 'Pennsylvania', "county_residence": county_residence, "state_agi": state_agi, "county_occupation": county_occupation, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def new_jersey_and_rhode_island(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        New Jersey:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Rhode Island:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        New Jersey:
        - https://www.state.nj.us/treasury/taxation/taxtables.shtml
          https://www.state.nj.us/treasury/taxation/njit13.shtml
          https://www.nj.gov/treasury/taxation/njit25.shtml
          https://www.state.nj.us/treasury/taxation/individuals/freqqiti.shtml
          https://www.state.nj.us/treasury/taxation/localtax.shtml
        
        Rhode Island:
        -         
        - https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2022_1040WE_w_0.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'NewJersey', 'state_occupation': 'RhodeIsland', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def new_jersey_and_south_carolina(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        New Jersey:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        South Carolina:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        New Jersey:
        - https://www.state.nj.us/treasury/taxation/taxtables.shtml
          https://www.state.nj.us/treasury/taxation/njit13.shtml
          https://www.nj.gov/treasury/taxation/njit25.shtml
          https://www.state.nj.us/treasury/taxation/individuals/freqqiti.shtml
          https://www.state.nj.us/treasury/taxation/localtax.shtml
        
        South Carolina:
        -         
        - https://dor.sc.gov/resources-site/lawandpolicy/Advisory%20Opinions/IL22-15.pdf
          https://dor.sc.gov/forms-site/Forms/IITPacket_2022.pdf
          https://www.scstatehouse.gov/code/t12c006.php
          https://dor.sc.gov/resources-site/lawandpolicy/Advisory%20Opinions/IL22-15.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'NewJersey', 'state_occupation': 'SouthCarolina', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def new_jersey_and_south_dakota(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        New Jersey:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        South Dakota:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        New Jersey:
        - https://www.state.nj.us/treasury/taxation/taxtables.shtml
          https://www.state.nj.us/treasury/taxation/njit13.shtml
          https://www.nj.gov/treasury/taxation/njit25.shtml
          https://www.state.nj.us/treasury/taxation/individuals/freqqiti.shtml
          https://www.state.nj.us/treasury/taxation/localtax.shtml
        
        South Dakota:
        -         
        - https://dor.sd.gov/individuals/taxes/
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'NewJersey', 'state_occupation': 'SouthDakota', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def new_jersey_and_tennessee(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        New Jersey:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Tennessee:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        New Jersey:
        - https://www.state.nj.us/treasury/taxation/taxtables.shtml
          https://www.state.nj.us/treasury/taxation/njit13.shtml
          https://www.nj.gov/treasury/taxation/njit25.shtml
          https://www.state.nj.us/treasury/taxation/individuals/freqqiti.shtml
          https://www.state.nj.us/treasury/taxation/localtax.shtml
        
        Tennessee:
        -         
        - https://www.tn.gov/revenue/taxes.html
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'NewJersey', 'state_occupation': 'Tennessee', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def new_jersey_and_texas(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        New Jersey:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Texas:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        New Jersey:
        - https://www.state.nj.us/treasury/taxation/taxtables.shtml
          https://www.state.nj.us/treasury/taxation/njit13.shtml
          https://www.nj.gov/treasury/taxation/njit25.shtml
          https://www.state.nj.us/treasury/taxation/individuals/freqqiti.shtml
          https://www.state.nj.us/treasury/taxation/localtax.shtml
        
        Texas:
        -         
        - https://comptroller.texas.gov/taxes/
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'NewJersey', 'state_occupation': 'Texas', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def new_jersey_and_utah(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        New Jersey:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Utah:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        New Jersey:
        - https://www.state.nj.us/treasury/taxation/taxtables.shtml
          https://www.state.nj.us/treasury/taxation/njit13.shtml
          https://www.nj.gov/treasury/taxation/njit25.shtml
          https://www.state.nj.us/treasury/taxation/individuals/freqqiti.shtml
          https://www.state.nj.us/treasury/taxation/localtax.shtml
        
        Utah:
        -         
        - https://incometax.utah.gov/paying/tax-rates
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'NewJersey', 'state_occupation': 'Utah', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def new_jersey_and_vermont(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        New Jersey:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Vermont:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        New Jersey:
        - https://www.state.nj.us/treasury/taxation/taxtables.shtml
          https://www.state.nj.us/treasury/taxation/njit13.shtml
          https://www.nj.gov/treasury/taxation/njit25.shtml
          https://www.state.nj.us/treasury/taxation/individuals/freqqiti.shtml
          https://www.state.nj.us/treasury/taxation/localtax.shtml
        
        Vermont:
        -         
        - https://tax.vermont.gov/sites/tax/files/documents/GB-1210-2023.pdf
          https://tax.vermont.gov/sites/tax/files/documents/Income%20Booklet-2022.pdf
          https://tax.vermont.gov/sites/tax/files/documents/Income%20Booklet-2022.pdf
          https://tax.vermont.gov/sites/tax/files/documents/Income%20Booklet-2022.pdf
          https://tax.vermont.gov/individuals/personal-income-tax/tax-credits
          https://tax.vermont.gov/individuals/personal-income-tax/taxable-income
          https://tax.vermont.gov/individuals/personal-income-tax/local-option-tax
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'NewJersey', 'state_occupation': 'Vermont', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def new_jersey_and_virginia(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        New Jersey:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Virginia:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        New Jersey:
        - https://www.state.nj.us/treasury/taxation/taxtables.shtml
          https://www.state.nj.us/treasury/taxation/njit13.shtml
          https://www.nj.gov/treasury/taxation/njit25.shtml
          https://www.state.nj.us/treasury/taxation/individuals/freqqiti.shtml
          https://www.state.nj.us/treasury/taxation/localtax.shtml
        
        Virginia:
        -         
        - https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/reciprocity
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'NewJersey', 'state_occupation': 'Virginia', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def new_jersey_and_washington(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        New Jersey:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Washington:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        New Jersey:
        - https://www.state.nj.us/treasury/taxation/taxtables.shtml
          https://www.state.nj.us/treasury/taxation/njit13.shtml
          https://www.nj.gov/treasury/taxation/njit25.shtml
          https://www.state.nj.us/treasury/taxation/individuals/freqqiti.shtml
          https://www.state.nj.us/treasury/taxation/localtax.shtml
        
        Washington:
        -         
        - https://dor.wa.gov/taxes-rates/income-tax
          https://dor.wa.gov/taxes-rates/other-taxes/capital-gains-tax
          https://dor.wa.gov/taxes-rates/income-tax
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'NewJersey', 'state_occupation': 'Washington', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def new_jersey_and_west_virginia(self, pay_periods=24, capital_gains_short=0, county_occupation='Other', dependents=0, sources=False, state_agi=75000, filers_over_65=0, capital_gains_long=0, filing_status='single', income=100000, credits_state=0):
        """
        --- Parameters ----
        
        New Jersey:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        West Virginia:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_occupation : str, optional
            County of occupation. Default is Other.
        pay_periods : int, optional
            Number of pay periods. Options are 13, 24, 26, 52. Default is 24.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        New Jersey:
        - https://www.state.nj.us/treasury/taxation/taxtables.shtml
          https://www.state.nj.us/treasury/taxation/njit13.shtml
          https://www.nj.gov/treasury/taxation/njit25.shtml
          https://www.state.nj.us/treasury/taxation/individuals/freqqiti.shtml
          https://www.state.nj.us/treasury/taxation/localtax.shtml
        
        West Virginia:
        -         
        - https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://tax.wv.gov/Documents/TSD/tsd438.pdf
          https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2011/Taxes-11-35att.html
        """
        
        # Check if the parameters are set upstream in the instance
        pay_periods = self.pay_periods if hasattr(self, 'pay_periods') else pay_periods
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        county_occupation = self.county_occupation if hasattr(self, 'county_occupation') else county_occupation
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'NewJersey', 'state_occupation': 'WestVirginia', "pay_periods": pay_periods, "capital_gains_short": capital_gains_short, "county_occupation": county_occupation, "dependents": dependents, "sources": sources, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def new_jersey_and_wisconsin(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        New Jersey:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Wisconsin:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        New Jersey:
        - https://www.state.nj.us/treasury/taxation/taxtables.shtml
          https://www.state.nj.us/treasury/taxation/njit13.shtml
          https://www.nj.gov/treasury/taxation/njit25.shtml
          https://www.state.nj.us/treasury/taxation/individuals/freqqiti.shtml
          https://www.state.nj.us/treasury/taxation/localtax.shtml
        
        Wisconsin:
        -         
        - https://www.revenue.wi.gov/TaxForms2023/2023-Form1-ES-Inst.pdf
          https://www.revenue.wi.gov/TaxForms2023/2023-Form1-ES-Inst.pdf
          https://www.revenue.wi.gov/TaxForms2023/2023-Form1-ES-Inst.pdf
          https://www.revenue.wi.gov/Pages/FAQS/pcs-work.aspx
          https://www.revenue.wi.gov/Dor%20Publications/pb103.pdf
          https://www.revenue.wi.gov/Pages/FAQS/pcs-work.aspx
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'NewJersey', 'state_occupation': 'Wisconsin', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def new_jersey_and_wyoming(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        New Jersey:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Wyoming:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        New Jersey:
        - https://www.state.nj.us/treasury/taxation/taxtables.shtml
          https://www.state.nj.us/treasury/taxation/njit13.shtml
          https://www.nj.gov/treasury/taxation/njit25.shtml
          https://www.state.nj.us/treasury/taxation/individuals/freqqiti.shtml
          https://www.state.nj.us/treasury/taxation/localtax.shtml
        
        Wyoming:
        -         
        - https://revenue.wyo.gov/
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'NewJersey', 'state_occupation': 'Wyoming', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def new_mexico_and_alabama(self, county_residence='Other', state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        New Mexico:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Alabama:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        New Mexico:
        - https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/fdf3c548-8aba-4b9c-9eb4-bb564c716015/FYI-104.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/0bf943fd-652e-4400-bb1b-cd8397eb5a95/2022pit-adj.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/2f1a6781-9534-4436-b427-1557f9592099/2022pit-adj-ins.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/0bf943fd-652e-4400-bb1b-cd8397eb5a95/2022pit-adj.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/1afc56af-ea90-4d48-82e5-1f9aeb43255a/PITbook2022.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/fdf3c548-8aba-4b9c-9eb4-bb564c716015/FYI-104.pdf
        
        Alabama:
        -         
        - https://www.revenue.alabama.gov/ultraviewer/viewer/basic_viewer/index.html?form=2023/02/22f40.pdf
          https://www.revenue.alabama.gov/faqs/what-is-alabamas-individual-income-tax-rate/
          https://www.revenue.alabama.gov/ultraviewer/viewer/basic_viewer/index.html?form=2023/02/22f40.pdf
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-24.htm?taxmap=true
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-24.htm?taxmap=true
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-24.htm?taxmap=true
          https://www.revenue.alabama.gov/individual-corporate/income-to-be-reported-on-the-alabama-income-tax-return/
          https://www.almonline.org/TaxRates.aspx#OccupationalTax
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'NewMexico', 'state_occupation': 'Alabama', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def new_mexico_and_alaska(self, state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        New Mexico:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Alaska:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        New Mexico:
        - https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/fdf3c548-8aba-4b9c-9eb4-bb564c716015/FYI-104.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/0bf943fd-652e-4400-bb1b-cd8397eb5a95/2022pit-adj.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/2f1a6781-9534-4436-b427-1557f9592099/2022pit-adj-ins.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/0bf943fd-652e-4400-bb1b-cd8397eb5a95/2022pit-adj.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/1afc56af-ea90-4d48-82e5-1f9aeb43255a/PITbook2022.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/fdf3c548-8aba-4b9c-9eb4-bb564c716015/FYI-104.pdf
        
        Alaska:
        -         
        - https://www.commerce.alaska.gov/web/dcra/officeofthestateassessor/alaskataxfacts.aspx
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'NewMexico', 'state_occupation': 'Alaska', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def new_mexico_and_arizona(self, state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        New Mexico:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Arizona:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        credits_state : int, optional
            Credits for state taxes. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        New Mexico:
        - https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/fdf3c548-8aba-4b9c-9eb4-bb564c716015/FYI-104.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/0bf943fd-652e-4400-bb1b-cd8397eb5a95/2022pit-adj.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/2f1a6781-9534-4436-b427-1557f9592099/2022pit-adj-ins.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/0bf943fd-652e-4400-bb1b-cd8397eb5a95/2022pit-adj.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/1afc56af-ea90-4d48-82e5-1f9aeb43255a/PITbook2022.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/fdf3c548-8aba-4b9c-9eb4-bb564c716015/FYI-104.pdf
        
        Arizona:
        -         
        - https://azdor.gov/forms/individual/form-140-x-y-tables
          https://azdor.gov/forms/individual/form-140-x-y-tables
          https://azdor.gov/forms/individual/form-140-x-y-tables
          https://azdor.gov/businesses-arizona/withholding-tax/withholding-exceptions
          "https://azdor.gov/individuals/income-tax-filing-assistance/identifying-other-taxable-income#:~:text=In%20regards%20to%20capital%20gains
         the%20individuals%20regular%20tax%20rate.
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'NewMexico', 'state_occupation': 'Arizona', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def new_mexico_and_arkansas(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        New Mexico:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Arkansas:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        New Mexico:
        - https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/fdf3c548-8aba-4b9c-9eb4-bb564c716015/FYI-104.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/0bf943fd-652e-4400-bb1b-cd8397eb5a95/2022pit-adj.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/2f1a6781-9534-4436-b427-1557f9592099/2022pit-adj-ins.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/0bf943fd-652e-4400-bb1b-cd8397eb5a95/2022pit-adj.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/1afc56af-ea90-4d48-82e5-1f9aeb43255a/PITbook2022.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/fdf3c548-8aba-4b9c-9eb4-bb564c716015/FYI-104.pdf
        
        Arkansas:
        -         
        - https://www.arkansasedc.com/why-arkansas/business-climate/tax-structure/personal-income-tax
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_and_AR1000NR_Instructions.pdf
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_FullYearResidentIndividualIncomeTaxReturn.pdf
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_and_AR1000NR_Instructions.pdf
          https://www.arkansasedc.com/why-arkansas/business-climate/tax-structure/capital-gains-tax-reduction
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_and_AR1000NR_Instructions.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'NewMexico', 'state_occupation': 'Arkansas', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def new_mexico_and_california(self, state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        New Mexico:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        California:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        New Mexico:
        - https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/fdf3c548-8aba-4b9c-9eb4-bb564c716015/FYI-104.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/0bf943fd-652e-4400-bb1b-cd8397eb5a95/2022pit-adj.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/2f1a6781-9534-4436-b427-1557f9592099/2022pit-adj-ins.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/0bf943fd-652e-4400-bb1b-cd8397eb5a95/2022pit-adj.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/1afc56af-ea90-4d48-82e5-1f9aeb43255a/PITbook2022.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/fdf3c548-8aba-4b9c-9eb4-bb564c716015/FYI-104.pdf
        
        California:
        -         
        - https://www.ftb.ca.gov/forms/2022/2022-540-tax-rate-schedules.pdf
          https://www.ftb.ca.gov/file/personal/deductions/index.html
          https://www.ftb.ca.gov/about-ftb/newsroom/caleitc/eligibility-and-credit-information.html
          https://www.ftb.ca.gov/file/personal/income-types/capital-gains-and-losses.html#:~:text=California%20does%20not%20have%20a
         are%20taxed%20as%20ordinary%20income
          https://www.ftb.ca.gov/file/personal/
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'NewMexico', 'state_occupation': 'California', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def new_mexico_and_colorado(self, county_residence='Other', state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        New Mexico:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Colorado:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        New Mexico:
        - https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/fdf3c548-8aba-4b9c-9eb4-bb564c716015/FYI-104.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/0bf943fd-652e-4400-bb1b-cd8397eb5a95/2022pit-adj.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/2f1a6781-9534-4436-b427-1557f9592099/2022pit-adj-ins.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/0bf943fd-652e-4400-bb1b-cd8397eb5a95/2022pit-adj.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/1afc56af-ea90-4d48-82e5-1f9aeb43255a/PITbook2022.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/fdf3c548-8aba-4b9c-9eb4-bb564c716015/FYI-104.pdf
        
        Colorado:
        -         
        - https://leg.colorado.gov/agencies/legislative-council-staff/individual-income-tax%C2%A0
          https://leg.colorado.gov/agencies/legislative-council-staff/individual-income-tax%C2%A0
          https://tax.colorado.gov/sites/tax/files/documents/ITT_Capital_Gain_Subtraction_Feb_2022.pdf
          https://www.auroragov.org/business_services/taxes/occupational_privilege_tax
          https://www.denvergov.org/Government/Agencies-Departments-Offices/Agencies-Departments-Offices-Directory/Department-of-Finance/Our-Divisions/Treasury/Business-Tax-Information#:~:text=Occupational%20privilege%20tax
         -%28show%20below%29&text=Each%20taxable%20employee%20is%20liable
         month%20for%20each%20taxable%20employee
          https://www.glendale.co.us/355/Occupational-Privilege-Tax
          https://www.greenwoodvillage.com/1220/Occupational-Privilege-Tax-OPT
          https://ci.sheridan.co.us/288/Occupational-Privilege-Tax
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'NewMexico', 'state_occupation': 'Colorado', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def new_mexico_and_connecticut(self, state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        New Mexico:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Connecticut:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        New Mexico:
        - https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/fdf3c548-8aba-4b9c-9eb4-bb564c716015/FYI-104.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/0bf943fd-652e-4400-bb1b-cd8397eb5a95/2022pit-adj.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/2f1a6781-9534-4436-b427-1557f9592099/2022pit-adj-ins.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/0bf943fd-652e-4400-bb1b-cd8397eb5a95/2022pit-adj.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/1afc56af-ea90-4d48-82e5-1f9aeb43255a/PITbook2022.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/fdf3c548-8aba-4b9c-9eb4-bb564c716015/FYI-104.pdf
        
        Connecticut:
        -         
        - https://portal.ct.gov/-/media/DRS/Forms/2022/Income/2022-CT-1040-Instructions_1222.pdf
          https://www.cga.ct.gov/2022/rpt/pdf/2022-R-0108.pdf
          https://www.cga.ct.gov/current/pub/chap_229.htm#sec_12-703
          https://portal.ct.gov/-/media/DRS/Forms/2022/Income/2022-CT-1040-Instructions_1222.pdf
          "https://portal.ct.gov/DRS/Publications/TSSNs/TSSN-29#:~:text=An%20individuals%20net%20capital%20gains
         portion%20of%20Social%20Security%20benefits"
          "https://portal.ct.gov/DRS/Publications/TSSNs/TSSN-29#:~:text=An%20individuals%20net%20capital%20gains
         portion%20of%20Social%20Security%20benefits"
          https://portal.ct.gov/DRS/Individuals/Resident-Income-Tax/Tax-Information
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'NewMexico', 'state_occupation': 'Connecticut', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def new_mexico_and_delaware(self, state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, federal_agi=75000, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        New Mexico:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Delaware:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        federal_agi : int, optional
            Adjusted Gross Income (AGI) for federal taxes. Default is 0. Sources - https://www.irs.gov/e-file-providers/definition-of-adjusted-gross-income, https://www.irs.gov/newsroom/taxpayers-should-know-the-difference-between-standard-and-itemized-deductions
        income : int, optional
            Gross income. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        New Mexico:
        - https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/fdf3c548-8aba-4b9c-9eb4-bb564c716015/FYI-104.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/0bf943fd-652e-4400-bb1b-cd8397eb5a95/2022pit-adj.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/2f1a6781-9534-4436-b427-1557f9592099/2022pit-adj-ins.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/0bf943fd-652e-4400-bb1b-cd8397eb5a95/2022pit-adj.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/1afc56af-ea90-4d48-82e5-1f9aeb43255a/PITbook2022.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/fdf3c548-8aba-4b9c-9eb4-bb564c716015/FYI-104.pdf
        
        Delaware:
        -         
        - https://revenue.delaware.gov/software-developer/tax-rate-changes/
          https://revenuefiles.delaware.gov/2022/PIT-RES_TY22_2022-01_PaperInteractive.pdf
          https://revenuefiles.delaware.gov/2022/PIT-RES_TY22_2022-02_Instructions.pdf
          https://revenue.delaware.gov/frequently-asked-questions/personal-income-tax-faqs/
          https://www.newcastlede.gov/DocumentCenter/View/20147/City-of-Wilmington-Wage-Tax-Information-PDF
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        federal_agi = self.federal_agi if hasattr(self, 'federal_agi') else federal_agi
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'NewMexico', 'state_occupation': 'Delaware', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "federal_agi": federal_agi, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def new_mexico_and_florida(self, state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        New Mexico:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Florida:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        New Mexico:
        - https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/fdf3c548-8aba-4b9c-9eb4-bb564c716015/FYI-104.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/0bf943fd-652e-4400-bb1b-cd8397eb5a95/2022pit-adj.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/2f1a6781-9534-4436-b427-1557f9592099/2022pit-adj-ins.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/0bf943fd-652e-4400-bb1b-cd8397eb5a95/2022pit-adj.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/1afc56af-ea90-4d48-82e5-1f9aeb43255a/PITbook2022.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/fdf3c548-8aba-4b9c-9eb4-bb564c716015/FYI-104.pdf
        
        Florida:
        -         
        - https://floridarevenue.com/taxes/taxesfees/Pages/tax_interest_rates.aspx
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'NewMexico', 'state_occupation': 'Florida', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def new_mexico_and_georgia(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        New Mexico:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Georgia:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        New Mexico:
        - https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/fdf3c548-8aba-4b9c-9eb4-bb564c716015/FYI-104.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/0bf943fd-652e-4400-bb1b-cd8397eb5a95/2022pit-adj.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/2f1a6781-9534-4436-b427-1557f9592099/2022pit-adj-ins.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/0bf943fd-652e-4400-bb1b-cd8397eb5a95/2022pit-adj.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/1afc56af-ea90-4d48-82e5-1f9aeb43255a/PITbook2022.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/fdf3c548-8aba-4b9c-9eb4-bb564c716015/FYI-104.pdf
        
        Georgia:
        -         
        - https://dor.georgia.gov/tax-tables-georgia-tax-rate-schedule
          https://apps.dor.ga.gov/FillableForms/PDFViewer/Index?form=2022GA500
          https://apps.dor.ga.gov/FillableForms/PDFViewer/Index?form=2022GA500
          https://apps.dor.ga.gov/FillableForms/PDFViewer/Index?form=2022GA500
          https://dor.georgia.gov/years-individual-income-tax-forms
          https://dor.georgia.gov/years-individual-income-tax-forms
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'NewMexico', 'state_occupation': 'Georgia', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def new_mexico_and_hawaii(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        New Mexico:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Hawaii:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        New Mexico:
        - https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/fdf3c548-8aba-4b9c-9eb4-bb564c716015/FYI-104.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/0bf943fd-652e-4400-bb1b-cd8397eb5a95/2022pit-adj.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/2f1a6781-9534-4436-b427-1557f9592099/2022pit-adj-ins.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/0bf943fd-652e-4400-bb1b-cd8397eb5a95/2022pit-adj.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/1afc56af-ea90-4d48-82e5-1f9aeb43255a/PITbook2022.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/fdf3c548-8aba-4b9c-9eb4-bb564c716015/FYI-104.pdf
        
        Hawaii:
        -         
        - https://tax.hawaii.gov/forms/d_18table-on/d_18table-on_p13/
          https://files.hawaii.gov/tax/forms/2018/18table-on.pdf
          https://files.hawaii.gov/tax/news/pubs/22outline.pdf
          https://files.hawaii.gov/tax/news/pubs/FastTaxRefGuide_2017.pdf
          https://www.tfhawaii.org/wordpress/state-tax-resources/mini-tax-guides/individual-income-tax-chapter-235/#:~:text=Long%20term%20capital%20gains%20are%20taxed%20at%20a%20maximum%20of%207.25%25
          https://www.grassrootinstitute.org/2023/02/hb337-increasing-tax-on-capital-gains-a-bad-risk/
          https://files.hawaii.gov/tax/news/pubs/22outline.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'NewMexico', 'state_occupation': 'Hawaii', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def new_mexico_and_idaho(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        New Mexico:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Idaho:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        New Mexico:
        - https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/fdf3c548-8aba-4b9c-9eb4-bb564c716015/FYI-104.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/0bf943fd-652e-4400-bb1b-cd8397eb5a95/2022pit-adj.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/2f1a6781-9534-4436-b427-1557f9592099/2022pit-adj-ins.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/0bf943fd-652e-4400-bb1b-cd8397eb5a95/2022pit-adj.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/1afc56af-ea90-4d48-82e5-1f9aeb43255a/PITbook2022.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/fdf3c548-8aba-4b9c-9eb4-bb564c716015/FYI-104.pdf
        
        Idaho:
        -         
        - https://tax.idaho.gov/taxes/income-tax/individual-income/individual-income-tax-rate-schedule/
          https://tax.idaho.gov/wp-content/uploads/forms/EIN00046/EIN00046_03-01-2023.pdf
          https://tax.idaho.gov/taxes/income-tax/individual-income/idaho-source-income/
          https://tax.idaho.gov/taxes/income-tax/individual-income/idaho-source-income/
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'NewMexico', 'state_occupation': 'Idaho', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def new_mexico_and_illinois(self, state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        New Mexico:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Illinois:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        New Mexico:
        - https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/fdf3c548-8aba-4b9c-9eb4-bb564c716015/FYI-104.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/0bf943fd-652e-4400-bb1b-cd8397eb5a95/2022pit-adj.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/2f1a6781-9534-4436-b427-1557f9592099/2022pit-adj-ins.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/0bf943fd-652e-4400-bb1b-cd8397eb5a95/2022pit-adj.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/1afc56af-ea90-4d48-82e5-1f9aeb43255a/PITbook2022.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/fdf3c548-8aba-4b9c-9eb4-bb564c716015/FYI-104.pdf
        
        Illinois:
        -         
        - https://tax.illinois.gov/research/taxrates/income.html#IndividualIncome
          https://tax.illinois.gov/content/dam/soi/en/web/tax/forms/incometax/documents/currentyear/individual/il-1040-instr.pdf
          https://tax.illinois.gov/questionsandanswers/12.html
          https://tax.illinois.gov/content/dam/soi/en/web/tax/forms/incometax/documents/currentyear/individual/il-1040-instr.pdf
          https://tax.illinois.gov/research/taxrates/personalproperty.html
          https://tax.illinois.gov/localgovernments/income.html
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'NewMexico', 'state_occupation': 'Illinois', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def new_mexico_and_indiana(self, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        New Mexico:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Indiana:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        New Mexico:
        - https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/fdf3c548-8aba-4b9c-9eb4-bb564c716015/FYI-104.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/0bf943fd-652e-4400-bb1b-cd8397eb5a95/2022pit-adj.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/2f1a6781-9534-4436-b427-1557f9592099/2022pit-adj-ins.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/0bf943fd-652e-4400-bb1b-cd8397eb5a95/2022pit-adj.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/1afc56af-ea90-4d48-82e5-1f9aeb43255a/PITbook2022.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/fdf3c548-8aba-4b9c-9eb4-bb564c716015/FYI-104.pdf
        
        Indiana:
        -         
        - https://www.in.gov/dor/tax-forms/2022-individual-income-tax-forms/
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-32.htm
          https://www.in.gov/dor/files/ib28.pdf
          https://www.in.gov/dor/tax-forms/2022-individual-income-tax-forms/
          https://www.in.gov/dor/files/2022-county-tax-rates-and-codes.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'NewMexico', 'state_occupation': 'Indiana', "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def new_mexico_and_iowa(self, county_residence='Other', state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        New Mexico:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Iowa:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        New Mexico:
        - https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/fdf3c548-8aba-4b9c-9eb4-bb564c716015/FYI-104.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/0bf943fd-652e-4400-bb1b-cd8397eb5a95/2022pit-adj.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/2f1a6781-9534-4436-b427-1557f9592099/2022pit-adj-ins.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/0bf943fd-652e-4400-bb1b-cd8397eb5a95/2022pit-adj.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/1afc56af-ea90-4d48-82e5-1f9aeb43255a/PITbook2022.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/fdf3c548-8aba-4b9c-9eb4-bb564c716015/FYI-104.pdf
        
        Iowa:
        -         
        - https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
          https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
          https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
          https://tax.iowa.gov/iowa-illinois-reciprocal-agreement
          https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'NewMexico', 'state_occupation': 'Iowa', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def new_mexico_and_kansas(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        New Mexico:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Kansas:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        New Mexico:
        - https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/fdf3c548-8aba-4b9c-9eb4-bb564c716015/FYI-104.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/0bf943fd-652e-4400-bb1b-cd8397eb5a95/2022pit-adj.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/2f1a6781-9534-4436-b427-1557f9592099/2022pit-adj-ins.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/0bf943fd-652e-4400-bb1b-cd8397eb5a95/2022pit-adj.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/1afc56af-ea90-4d48-82e5-1f9aeb43255a/PITbook2022.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/fdf3c548-8aba-4b9c-9eb4-bb564c716015/FYI-104.pdf
        
        Kansas:
        -         
        - https://www.ksrevenue.gov/taxrates.html
          https://www.ksrevenue.gov/incomebook22.html#0
          https://www.ksrevenue.gov/incomebook22.html#0
          https://www.ksrevenue.gov/incomebook22.html#0
          https://www.ksrevenue.gov/incomebook22.html
          https://www.ksrevenue.gov/incomebook22.html
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'NewMexico', 'state_occupation': 'Kansas', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def new_mexico_and_kentucky(self, county_residence='Other', state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        New Mexico:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Kentucky:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        New Mexico:
        - https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/fdf3c548-8aba-4b9c-9eb4-bb564c716015/FYI-104.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/0bf943fd-652e-4400-bb1b-cd8397eb5a95/2022pit-adj.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/2f1a6781-9534-4436-b427-1557f9592099/2022pit-adj-ins.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/0bf943fd-652e-4400-bb1b-cd8397eb5a95/2022pit-adj.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/1afc56af-ea90-4d48-82e5-1f9aeb43255a/PITbook2022.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/fdf3c548-8aba-4b9c-9eb4-bb564c716015/FYI-104.pdf
        
        Kentucky:
        -         
        - https://revenue.ky.gov/News/Pages/Dor-Announces-Updates-to-Individual-Income-Tax-for-2023-Tax-Year.aspx
          https://revenue.ky.gov/Individual/Individual-Income-Tax/Pages/default.aspx
          https://revenue.ky.gov/News/Pages/Dor-Announces-Updates-to-Individual-Income-Tax-for-2023-Tax-Year.aspx
          https://revenue.ky.gov/Individual/Individual-Income-Tax/Pages/default.aspx
          https://revenue.ky.gov/Dor%20Training%20Materials/103%20KAR%2017.140.%20Individual%20income%20tax%20-%20reciprocity%20-%20nonresidents.pdf
          https://revenue.ky.gov/Individual/Individual-Income-Tax/Pages/default.aspx
          https://nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2021/Taxes-21-32.htm?taxmap=true
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'NewMexico', 'state_occupation': 'Kentucky', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def new_mexico_and_louisiana(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        New Mexico:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Louisiana:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        New Mexico:
        - https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/fdf3c548-8aba-4b9c-9eb4-bb564c716015/FYI-104.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/0bf943fd-652e-4400-bb1b-cd8397eb5a95/2022pit-adj.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/2f1a6781-9534-4436-b427-1557f9592099/2022pit-adj-ins.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/0bf943fd-652e-4400-bb1b-cd8397eb5a95/2022pit-adj.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/1afc56af-ea90-4d48-82e5-1f9aeb43255a/PITbook2022.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/fdf3c548-8aba-4b9c-9eb4-bb564c716015/FYI-104.pdf
        
        Louisiana:
        -         
        - https://revenue.louisiana.gov/individualincometax
          https://revenue.louisiana.gov/individualincometax
          https://www.revenue.louisiana.gov/taxforms/6935(11_02)F.pdf
          https://revenue.louisiana.gov/individualincometax
          https://revenue.louisiana.gov/individualincometax
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'NewMexico', 'state_occupation': 'Louisiana', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def new_mexico_and_maine(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        New Mexico:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Maine:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        New Mexico:
        - https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/fdf3c548-8aba-4b9c-9eb4-bb564c716015/FYI-104.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/0bf943fd-652e-4400-bb1b-cd8397eb5a95/2022pit-adj.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/2f1a6781-9534-4436-b427-1557f9592099/2022pit-adj-ins.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/0bf943fd-652e-4400-bb1b-cd8397eb5a95/2022pit-adj.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/1afc56af-ea90-4d48-82e5-1f9aeb43255a/PITbook2022.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/fdf3c548-8aba-4b9c-9eb4-bb564c716015/FYI-104.pdf
        
        Maine:
        -         
        - https://www.maine.gov/revenue/tax-return-forms/individual-income-tax-2022
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/ind_tax_rate_sched_2022.pdf
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/22_1040me_book_gen_instruc_revisedFeb23.pdf
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/22_1040me_book_gen_instruc_revisedFeb23.pdf
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/22_1040me_book_gen_instruc_revisedFeb23.pdf
          https://www.maine.gov/revenue/tax-return-forms/individual-income-tax-2022
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'NewMexico', 'state_occupation': 'Maine', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def new_mexico_and_maryland(self, county_residence='Other', state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        New Mexico:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Maryland:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        New Mexico:
        - https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/fdf3c548-8aba-4b9c-9eb4-bb564c716015/FYI-104.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/0bf943fd-652e-4400-bb1b-cd8397eb5a95/2022pit-adj.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/2f1a6781-9534-4436-b427-1557f9592099/2022pit-adj-ins.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/0bf943fd-652e-4400-bb1b-cd8397eb5a95/2022pit-adj.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/1afc56af-ea90-4d48-82e5-1f9aeb43255a/PITbook2022.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/fdf3c548-8aba-4b9c-9eb4-bb564c716015/FYI-104.pdf
        
        Maryland:
        -         
        - https://www.marylandtaxes.gov/individual/income/tax-info/tax-rates.php
          https://www.marylandtaxes.gov/forms/22_forms/Resident_Booklet.pdf
          https://www.marylandtaxes.gov/forms/Tax_Publications/Administrative_Releases/Income_and_Estate_Tax_Releases/ar_it37.pdf
          https://www.marylandtaxes.gov/individual/index.php
          https://www.marylandtaxes.gov/individual/income/tax-info/tax-rates.php
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'NewMexico', 'state_occupation': 'Maryland', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def new_mexico_and_massachusetts(self, state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        New Mexico:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Massachusetts:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        New Mexico:
        - https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/fdf3c548-8aba-4b9c-9eb4-bb564c716015/FYI-104.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/0bf943fd-652e-4400-bb1b-cd8397eb5a95/2022pit-adj.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/2f1a6781-9534-4436-b427-1557f9592099/2022pit-adj-ins.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/0bf943fd-652e-4400-bb1b-cd8397eb5a95/2022pit-adj.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/1afc56af-ea90-4d48-82e5-1f9aeb43255a/PITbook2022.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/fdf3c548-8aba-4b9c-9eb4-bb564c716015/FYI-104.pdf
        
        Massachusetts:
        -         
        - https://www.mass.gov/service-details/massachusetts-tax-rates
          https://www.mass.gov/service-details/view-massachusetts-personal-income-tax-exemptions
          https://www.mass.gov/service-details/view-massachusetts-personal-income-tax-exemptions
          https://www.mass.gov/service-details/view-child-and-dependent-related-credits
          https://www.mass.gov/service-details/massachusetts-tax-rates
          https://www.mass.gov/personal-income-tax
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'NewMexico', 'state_occupation': 'Massachusetts', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def new_mexico_and_michigan(self, county_occupation='Other', state_agi=75000, county_residence='Other', capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        New Mexico:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Michigan:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        county_occupation : str, optional
            County of occupation. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        New Mexico:
        - https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/fdf3c548-8aba-4b9c-9eb4-bb564c716015/FYI-104.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/0bf943fd-652e-4400-bb1b-cd8397eb5a95/2022pit-adj.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/2f1a6781-9534-4436-b427-1557f9592099/2022pit-adj-ins.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/0bf943fd-652e-4400-bb1b-cd8397eb5a95/2022pit-adj.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/1afc56af-ea90-4d48-82e5-1f9aeb43255a/PITbook2022.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/fdf3c548-8aba-4b9c-9eb4-bb564c716015/FYI-104.pdf
        
        Michigan:
        -         
        - https://www.michigan.gov/taxes/iit/new-developments-for-tax-year-2022
          "https://www.michigan.gov/taxes/-/media/Project/Websites/taxes/Forms/2023/2023-SUW/446_Withholding-Guide_2023.pdf?rev=177ca3d8cb034392bf84e317ed78c320&hash=19C5F42F43A89A91863D229349F48883#:~:text=The%20exemption%20amount%20is%20%245%2C400
         employees%20Michigan%20income%20tax%20return"
          "https://www.michigan.gov/taxes/-/media/Project/Websites/taxes/Forms/2023/2023-SUW/446_Withholding-Guide_2023.pdf?rev=177ca3d8cb034392bf84e317ed78c320&hash=19C5F42F43A89A91863D229349F48883#:~:text=The%20exemption%20amount%20is%20%245%2C400
         employees%20Michigan%20income%20tax%20return"
          https://www.michigan.gov/taxes/questions/iit/accordion/residency/are-my-wages-earned-in-another-state-taxable-in-michigan-if-i-am-a-michigan-resident-1
          https://www.michigan.gov/taxes/-/media/Project/Websites/taxes/Forms/2022/2022-IIT-Forms/BOOK_MI-1040.pdf?rev=4524f1cbbc0243b194ab05a6680be75b&hash=40DFE71E651F7C1D65C31F6CD03D944F
          https://www.michigan.gov/taxes/questions/iit/accordion/general/what-cities-impose-an-income-tax
        """
        
        # Check if the parameters are set upstream in the instance
        county_occupation = self.county_occupation if hasattr(self, 'county_occupation') else county_occupation
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'NewMexico', 'state_occupation': 'Michigan', "county_occupation": county_occupation, "state_agi": state_agi, "county_residence": county_residence, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def new_mexico_and_minnesota(self, state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        New Mexico:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Minnesota:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        New Mexico:
        - https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/fdf3c548-8aba-4b9c-9eb4-bb564c716015/FYI-104.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/0bf943fd-652e-4400-bb1b-cd8397eb5a95/2022pit-adj.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/2f1a6781-9534-4436-b427-1557f9592099/2022pit-adj-ins.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/0bf943fd-652e-4400-bb1b-cd8397eb5a95/2022pit-adj.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/1afc56af-ea90-4d48-82e5-1f9aeb43255a/PITbook2022.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/fdf3c548-8aba-4b9c-9eb4-bb564c716015/FYI-104.pdf
        
        Minnesota:
        -         
        - https://www.revenue.state.mn.us/minnesota-income-tax-rates-and-brackets
          https://www.revenue.state.mn.us/child-and-dependent-care-credit
          https://www.revenue.state.mn.us/dependent-exemptions
          https://www.revenue.state.mn.us/minnesota-standard-deduction
          https://mn.gov/mmb/assets/mwr_form_tcm1059-128581.pdf
          https://www.revenue.state.mn.us/individual-income-tax
          https://www.revenue.state.mn.us/individual-income-tax
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'NewMexico', 'state_occupation': 'Minnesota', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def new_mexico_and_mississippi(self, state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        New Mexico:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Mississippi:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        New Mexico:
        - https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/fdf3c548-8aba-4b9c-9eb4-bb564c716015/FYI-104.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/0bf943fd-652e-4400-bb1b-cd8397eb5a95/2022pit-adj.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/2f1a6781-9534-4436-b427-1557f9592099/2022pit-adj-ins.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/0bf943fd-652e-4400-bb1b-cd8397eb5a95/2022pit-adj.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/1afc56af-ea90-4d48-82e5-1f9aeb43255a/PITbook2022.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/fdf3c548-8aba-4b9c-9eb4-bb564c716015/FYI-104.pdf
        
        Mississippi:
        -         
        - https://www.dor.ms.gov/individual/tax-rates
          https://www.dor.ms.gov/individual/tax-rates
          https://www.dor.ms.gov/individual/tax-rates
          https://www.dor.ms.gov/individual
          https://www.dor.ms.gov/individual
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'NewMexico', 'state_occupation': 'Mississippi', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def new_mexico_and_missouri(self, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        New Mexico:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Missouri:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        New Mexico:
        - https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/fdf3c548-8aba-4b9c-9eb4-bb564c716015/FYI-104.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/0bf943fd-652e-4400-bb1b-cd8397eb5a95/2022pit-adj.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/2f1a6781-9534-4436-b427-1557f9592099/2022pit-adj-ins.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/0bf943fd-652e-4400-bb1b-cd8397eb5a95/2022pit-adj.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/1afc56af-ea90-4d48-82e5-1f9aeb43255a/PITbook2022.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/fdf3c548-8aba-4b9c-9eb4-bb564c716015/FYI-104.pdf
        
        Missouri:
        -         
        - https://dor.mo.gov/taxation/individual/tax-types/income/year-changes/
          https://dor.mo.gov/taxation/individual/tax-types/income/year-changes/
          https://dor.mo.gov/taxation/individual/tax-types/income/year-changes/
          https://www.kcmo.gov/city-hall/departments/finance/earnings-tax
          https://www.stlouis-mo.gov/government/departments/collector/earnings-tax/individual-earnings-tax-info.cfm
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'NewMexico', 'state_occupation': 'Missouri', "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def new_mexico_and_montana(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        New Mexico:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Montana:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        New Mexico:
        - https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/fdf3c548-8aba-4b9c-9eb4-bb564c716015/FYI-104.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/0bf943fd-652e-4400-bb1b-cd8397eb5a95/2022pit-adj.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/2f1a6781-9534-4436-b427-1557f9592099/2022pit-adj-ins.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/0bf943fd-652e-4400-bb1b-cd8397eb5a95/2022pit-adj.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/1afc56af-ea90-4d48-82e5-1f9aeb43255a/PITbook2022.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/fdf3c548-8aba-4b9c-9eb4-bb564c716015/FYI-104.pdf
        
        Montana:
        -         
        - https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/2022/12/montana-employees-withholding-allowance-and-exemption-certificate-form-mw-4.pdf
          https://mtrevenue.gov/2021/11/29/simplification-of-montana-income-taxation/?utm_source=rss&utm_medium=rss&utm_campaign=simplification-of-montana-income-taxation
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'NewMexico', 'state_occupation': 'Montana', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def new_mexico_and_nebraska(self, state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        New Mexico:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Nebraska:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        New Mexico:
        - https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/fdf3c548-8aba-4b9c-9eb4-bb564c716015/FYI-104.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/0bf943fd-652e-4400-bb1b-cd8397eb5a95/2022pit-adj.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/2f1a6781-9534-4436-b427-1557f9592099/2022pit-adj-ins.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/0bf943fd-652e-4400-bb1b-cd8397eb5a95/2022pit-adj.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/1afc56af-ea90-4d48-82e5-1f9aeb43255a/PITbook2022.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/fdf3c548-8aba-4b9c-9eb4-bb564c716015/FYI-104.pdf
        
        Nebraska:
        -         
        - https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2023_Circular_EN_8-429r2023_Whole_Book_11.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'NewMexico', 'state_occupation': 'Nebraska', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def new_mexico_and_nevada(self, state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        New Mexico:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Nevada:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        New Mexico:
        - https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/fdf3c548-8aba-4b9c-9eb4-bb564c716015/FYI-104.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/0bf943fd-652e-4400-bb1b-cd8397eb5a95/2022pit-adj.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/2f1a6781-9534-4436-b427-1557f9592099/2022pit-adj-ins.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/0bf943fd-652e-4400-bb1b-cd8397eb5a95/2022pit-adj.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/1afc56af-ea90-4d48-82e5-1f9aeb43255a/PITbook2022.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/fdf3c548-8aba-4b9c-9eb4-bb564c716015/FYI-104.pdf
        
        Nevada:
        -         
        - https://www.nvsos.gov/sos/businesses/the-nevada-advantage
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'NewMexico', 'state_occupation': 'Nevada', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def new_mexico_and_new_hampshire(self, state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        New Mexico:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        New Hampshire:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        New Mexico:
        - https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/fdf3c548-8aba-4b9c-9eb4-bb564c716015/FYI-104.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/0bf943fd-652e-4400-bb1b-cd8397eb5a95/2022pit-adj.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/2f1a6781-9534-4436-b427-1557f9592099/2022pit-adj-ins.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/0bf943fd-652e-4400-bb1b-cd8397eb5a95/2022pit-adj.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/1afc56af-ea90-4d48-82e5-1f9aeb43255a/PITbook2022.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/fdf3c548-8aba-4b9c-9eb4-bb564c716015/FYI-104.pdf
        
        New Hampshire:
        -         
        - https://www.revenue.nh.gov/assistance/tax-overview.htm
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'NewMexico', 'state_occupation': 'NewHampshire', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def new_mexico_and_new_jersey(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        New Mexico:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        New Jersey:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        New Mexico:
        - https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/fdf3c548-8aba-4b9c-9eb4-bb564c716015/FYI-104.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/0bf943fd-652e-4400-bb1b-cd8397eb5a95/2022pit-adj.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/2f1a6781-9534-4436-b427-1557f9592099/2022pit-adj-ins.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/0bf943fd-652e-4400-bb1b-cd8397eb5a95/2022pit-adj.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/1afc56af-ea90-4d48-82e5-1f9aeb43255a/PITbook2022.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/fdf3c548-8aba-4b9c-9eb4-bb564c716015/FYI-104.pdf
        
        New Jersey:
        -         
        - https://www.state.nj.us/treasury/taxation/taxtables.shtml
          https://www.state.nj.us/treasury/taxation/njit13.shtml
          https://www.nj.gov/treasury/taxation/njit25.shtml
          https://www.state.nj.us/treasury/taxation/individuals/freqqiti.shtml
          https://www.state.nj.us/treasury/taxation/localtax.shtml
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'NewMexico', 'state_occupation': 'NewJersey', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def new_mexico_and_new_york(self, county_residence='Other', state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        New Mexico:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        New York:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        New Mexico:
        - https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/fdf3c548-8aba-4b9c-9eb4-bb564c716015/FYI-104.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/0bf943fd-652e-4400-bb1b-cd8397eb5a95/2022pit-adj.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/2f1a6781-9534-4436-b427-1557f9592099/2022pit-adj-ins.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/0bf943fd-652e-4400-bb1b-cd8397eb5a95/2022pit-adj.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/1afc56af-ea90-4d48-82e5-1f9aeb43255a/PITbook2022.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/fdf3c548-8aba-4b9c-9eb4-bb564c716015/FYI-104.pdf
        
        New York:
        -         
        - https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nys-tax-rate-schedule
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nys-tax-rate-schedule
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nys-tax-rate-schedule
          https://www.tax.ny.gov/pit/
          https://www.tax.ny.gov/pit/file/nonresident-faqs.htm
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#yonkers-tax-rate-schedule
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nyc-tax-rate-schedule
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'NewMexico', 'state_occupation': 'NewYork', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def new_mexico_and_north_carolina(self, state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        New Mexico:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        North Carolina:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        New Mexico:
        - https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/fdf3c548-8aba-4b9c-9eb4-bb564c716015/FYI-104.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/0bf943fd-652e-4400-bb1b-cd8397eb5a95/2022pit-adj.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/2f1a6781-9534-4436-b427-1557f9592099/2022pit-adj-ins.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/0bf943fd-652e-4400-bb1b-cd8397eb5a95/2022pit-adj.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/1afc56af-ea90-4d48-82e5-1f9aeb43255a/PITbook2022.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/fdf3c548-8aba-4b9c-9eb4-bb564c716015/FYI-104.pdf
        
        North Carolina:
        -         
        - https://www.ncdor.gov/taxes-forms/tax-rate-schedules
          https://www.ncdor.gov/taxes-forms/individual-income-tax/north-carolina-standard-deduction-or-north-carolina-itemized-deductions
          https://www.ncdor.gov/taxes-forms/individual-income-tax
          https://www.ncdor.gov/taxes-forms/individual-income-tax
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'NewMexico', 'state_occupation': 'NorthCarolina', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def new_mexico_and_north_dakota(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        New Mexico:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        North Dakota:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        New Mexico:
        - https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/fdf3c548-8aba-4b9c-9eb4-bb564c716015/FYI-104.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/0bf943fd-652e-4400-bb1b-cd8397eb5a95/2022pit-adj.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/2f1a6781-9534-4436-b427-1557f9592099/2022pit-adj-ins.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/0bf943fd-652e-4400-bb1b-cd8397eb5a95/2022pit-adj.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/1afc56af-ea90-4d48-82e5-1f9aeb43255a/PITbook2022.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/fdf3c548-8aba-4b9c-9eb4-bb564c716015/FYI-104.pdf
        
        North Dakota:
        -         
        - https://www.tax.nd.gov/forms
          https://www.tax.nd.gov/sites/www/files/documents/forms/individual/2022-iit/2022-individual-income-tax-booklet.pdf
          https://www.tax.nd.gov/sites/www/files/documents/forms/2022-individual-income-tax-booklet.pdf
          https://www.tax.nd.gov/tax-exemptions-credits/income-tax-exemptions-credits
          https://www.tax.nd.gov/sites/www/files/documents/forms/business/ndwrfillable.pdf
          https://www.tax.nd.gov/news/tax-legislative-changes/significant-changes-law/individual-income-tax-history
          https://www.tax.nd.gov/forms
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'NewMexico', 'state_occupation': 'NorthDakota', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def new_mexico_and_ohio(self, county_residence='Other', state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        New Mexico:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Ohio:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        New Mexico:
        - https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/fdf3c548-8aba-4b9c-9eb4-bb564c716015/FYI-104.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/0bf943fd-652e-4400-bb1b-cd8397eb5a95/2022pit-adj.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/2f1a6781-9534-4436-b427-1557f9592099/2022pit-adj-ins.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/0bf943fd-652e-4400-bb1b-cd8397eb5a95/2022pit-adj.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/1afc56af-ea90-4d48-82e5-1f9aeb43255a/PITbook2022.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/fdf3c548-8aba-4b9c-9eb4-bb564c716015/FYI-104.pdf
        
        Ohio:
        -         
        - https://tax.ohio.gov/individual/resources/annual-tax-rates
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://tax.ohio.gov/static/forms/employer_withholding/generic/wth_it4nr.pdf
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://www.ritaohio.com/TaxRatesTable
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'NewMexico', 'state_occupation': 'Ohio', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def new_mexico_and_oklahoma(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        New Mexico:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Oklahoma:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        New Mexico:
        - https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/fdf3c548-8aba-4b9c-9eb4-bb564c716015/FYI-104.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/0bf943fd-652e-4400-bb1b-cd8397eb5a95/2022pit-adj.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/2f1a6781-9534-4436-b427-1557f9592099/2022pit-adj-ins.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/0bf943fd-652e-4400-bb1b-cd8397eb5a95/2022pit-adj.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/1afc56af-ea90-4d48-82e5-1f9aeb43255a/PITbook2022.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/fdf3c548-8aba-4b9c-9eb4-bb564c716015/FYI-104.pdf
        
        Oklahoma:
        -         
        - https://oklahoma.gov/content/dam/ok/en/tax/documents/resources/publications/businesses/withholding-tables/WHTables-2022.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'NewMexico', 'state_occupation': 'Oklahoma', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def new_mexico_and_oregon(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        New Mexico:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Oregon:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        New Mexico:
        - https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/fdf3c548-8aba-4b9c-9eb4-bb564c716015/FYI-104.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/0bf943fd-652e-4400-bb1b-cd8397eb5a95/2022pit-adj.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/2f1a6781-9534-4436-b427-1557f9592099/2022pit-adj-ins.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/0bf943fd-652e-4400-bb1b-cd8397eb5a95/2022pit-adj.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/1afc56af-ea90-4d48-82e5-1f9aeb43255a/PITbook2022.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/fdf3c548-8aba-4b9c-9eb4-bb564c716015/FYI-104.pdf
        
        Oregon:
        -         
        - https://www.oregon.gov/dor/programs/individuals/Pages/PIT.aspx
          https://www.oregon.gov/dor/programs/individuals/Documents/Full%20year%20resident
         %20Form%20or-40%20filers.pdf
          https://www.oregon.gov/dor/programs/individuals/Documents/Part-year%20and%20nonresident
         %20Form%20or-40-P%20and%20or-40-N%20filers.pdf.pdf
          https://www.oregon.gov/dor/forms/FormsPubs/form-or-40_101-040_2022.pdf
          https://www.oregon.gov/dor/forms/FormsPubs/form-or-40_101-040_2022.pdf
          https://www.oregon.gov/dor/forms/FormsPubs/form-or-40_101-040_2022.pdf
          https://www.portland.gov/revenue/personal-tax
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'NewMexico', 'state_occupation': 'Oregon', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def new_mexico_and_pennsylvania(self, county_residence='Other', state_agi=75000, county_occupation='Other', capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        New Mexico:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Pennsylvania:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        county_residence : str, optional
            County of residence. Default is Other.
        county_occupation : str, optional
            County of occupation. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        New Mexico:
        - https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/fdf3c548-8aba-4b9c-9eb4-bb564c716015/FYI-104.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/0bf943fd-652e-4400-bb1b-cd8397eb5a95/2022pit-adj.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/2f1a6781-9534-4436-b427-1557f9592099/2022pit-adj-ins.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/0bf943fd-652e-4400-bb1b-cd8397eb5a95/2022pit-adj.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/1afc56af-ea90-4d48-82e5-1f9aeb43255a/PITbook2022.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/fdf3c548-8aba-4b9c-9eb4-bb564c716015/FYI-104.pdf
        
        Pennsylvania:
        -         
        - https://www.revenue.pa.gov/Tax%20Rates/Pages/default.aspx
          https://www.revenue.pa.gov/FormsandPublications/FormsforIndividuals/PIT/Documents/2022/2022_pa-40in.pdf
          https://www.revenue.pa.gov/FormsandPublications/FormsforBusinesses/EmployerWithholding/Documents/rev-419.pdf
          https://www.revenue.pa.gov/FormsandPublications/FormsforIndividuals/PIT/Documents/2022/2022_pa-40in.pdf
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-25.htm
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        county_occupation = self.county_occupation if hasattr(self, 'county_occupation') else county_occupation
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'NewMexico', 'state_occupation': 'Pennsylvania', "county_residence": county_residence, "state_agi": state_agi, "county_occupation": county_occupation, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def new_mexico_and_rhode_island(self, state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        New Mexico:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Rhode Island:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        New Mexico:
        - https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/fdf3c548-8aba-4b9c-9eb4-bb564c716015/FYI-104.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/0bf943fd-652e-4400-bb1b-cd8397eb5a95/2022pit-adj.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/2f1a6781-9534-4436-b427-1557f9592099/2022pit-adj-ins.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/0bf943fd-652e-4400-bb1b-cd8397eb5a95/2022pit-adj.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/1afc56af-ea90-4d48-82e5-1f9aeb43255a/PITbook2022.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/fdf3c548-8aba-4b9c-9eb4-bb564c716015/FYI-104.pdf
        
        Rhode Island:
        -         
        - https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2022_1040WE_w_0.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'NewMexico', 'state_occupation': 'RhodeIsland', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def new_mexico_and_south_carolina(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        New Mexico:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        South Carolina:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        New Mexico:
        - https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/fdf3c548-8aba-4b9c-9eb4-bb564c716015/FYI-104.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/0bf943fd-652e-4400-bb1b-cd8397eb5a95/2022pit-adj.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/2f1a6781-9534-4436-b427-1557f9592099/2022pit-adj-ins.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/0bf943fd-652e-4400-bb1b-cd8397eb5a95/2022pit-adj.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/1afc56af-ea90-4d48-82e5-1f9aeb43255a/PITbook2022.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/fdf3c548-8aba-4b9c-9eb4-bb564c716015/FYI-104.pdf
        
        South Carolina:
        -         
        - https://dor.sc.gov/resources-site/lawandpolicy/Advisory%20Opinions/IL22-15.pdf
          https://dor.sc.gov/forms-site/Forms/IITPacket_2022.pdf
          https://www.scstatehouse.gov/code/t12c006.php
          https://dor.sc.gov/resources-site/lawandpolicy/Advisory%20Opinions/IL22-15.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'NewMexico', 'state_occupation': 'SouthCarolina', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def new_mexico_and_south_dakota(self, state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        New Mexico:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        South Dakota:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        New Mexico:
        - https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/fdf3c548-8aba-4b9c-9eb4-bb564c716015/FYI-104.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/0bf943fd-652e-4400-bb1b-cd8397eb5a95/2022pit-adj.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/2f1a6781-9534-4436-b427-1557f9592099/2022pit-adj-ins.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/0bf943fd-652e-4400-bb1b-cd8397eb5a95/2022pit-adj.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/1afc56af-ea90-4d48-82e5-1f9aeb43255a/PITbook2022.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/fdf3c548-8aba-4b9c-9eb4-bb564c716015/FYI-104.pdf
        
        South Dakota:
        -         
        - https://dor.sd.gov/individuals/taxes/
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'NewMexico', 'state_occupation': 'SouthDakota', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def new_mexico_and_tennessee(self, state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        New Mexico:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Tennessee:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        New Mexico:
        - https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/fdf3c548-8aba-4b9c-9eb4-bb564c716015/FYI-104.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/0bf943fd-652e-4400-bb1b-cd8397eb5a95/2022pit-adj.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/2f1a6781-9534-4436-b427-1557f9592099/2022pit-adj-ins.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/0bf943fd-652e-4400-bb1b-cd8397eb5a95/2022pit-adj.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/1afc56af-ea90-4d48-82e5-1f9aeb43255a/PITbook2022.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/fdf3c548-8aba-4b9c-9eb4-bb564c716015/FYI-104.pdf
        
        Tennessee:
        -         
        - https://www.tn.gov/revenue/taxes.html
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'NewMexico', 'state_occupation': 'Tennessee', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def new_mexico_and_texas(self, state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        New Mexico:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Texas:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        New Mexico:
        - https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/fdf3c548-8aba-4b9c-9eb4-bb564c716015/FYI-104.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/0bf943fd-652e-4400-bb1b-cd8397eb5a95/2022pit-adj.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/2f1a6781-9534-4436-b427-1557f9592099/2022pit-adj-ins.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/0bf943fd-652e-4400-bb1b-cd8397eb5a95/2022pit-adj.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/1afc56af-ea90-4d48-82e5-1f9aeb43255a/PITbook2022.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/fdf3c548-8aba-4b9c-9eb4-bb564c716015/FYI-104.pdf
        
        Texas:
        -         
        - https://comptroller.texas.gov/taxes/
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'NewMexico', 'state_occupation': 'Texas', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def new_mexico_and_utah(self, state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        New Mexico:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Utah:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        New Mexico:
        - https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/fdf3c548-8aba-4b9c-9eb4-bb564c716015/FYI-104.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/0bf943fd-652e-4400-bb1b-cd8397eb5a95/2022pit-adj.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/2f1a6781-9534-4436-b427-1557f9592099/2022pit-adj-ins.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/0bf943fd-652e-4400-bb1b-cd8397eb5a95/2022pit-adj.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/1afc56af-ea90-4d48-82e5-1f9aeb43255a/PITbook2022.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/fdf3c548-8aba-4b9c-9eb4-bb564c716015/FYI-104.pdf
        
        Utah:
        -         
        - https://incometax.utah.gov/paying/tax-rates
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'NewMexico', 'state_occupation': 'Utah', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def new_mexico_and_vermont(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        New Mexico:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Vermont:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        New Mexico:
        - https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/fdf3c548-8aba-4b9c-9eb4-bb564c716015/FYI-104.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/0bf943fd-652e-4400-bb1b-cd8397eb5a95/2022pit-adj.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/2f1a6781-9534-4436-b427-1557f9592099/2022pit-adj-ins.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/0bf943fd-652e-4400-bb1b-cd8397eb5a95/2022pit-adj.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/1afc56af-ea90-4d48-82e5-1f9aeb43255a/PITbook2022.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/fdf3c548-8aba-4b9c-9eb4-bb564c716015/FYI-104.pdf
        
        Vermont:
        -         
        - https://tax.vermont.gov/sites/tax/files/documents/GB-1210-2023.pdf
          https://tax.vermont.gov/sites/tax/files/documents/Income%20Booklet-2022.pdf
          https://tax.vermont.gov/sites/tax/files/documents/Income%20Booklet-2022.pdf
          https://tax.vermont.gov/sites/tax/files/documents/Income%20Booklet-2022.pdf
          https://tax.vermont.gov/individuals/personal-income-tax/tax-credits
          https://tax.vermont.gov/individuals/personal-income-tax/taxable-income
          https://tax.vermont.gov/individuals/personal-income-tax/local-option-tax
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'NewMexico', 'state_occupation': 'Vermont', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def new_mexico_and_virginia(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        New Mexico:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Virginia:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        New Mexico:
        - https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/fdf3c548-8aba-4b9c-9eb4-bb564c716015/FYI-104.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/0bf943fd-652e-4400-bb1b-cd8397eb5a95/2022pit-adj.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/2f1a6781-9534-4436-b427-1557f9592099/2022pit-adj-ins.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/0bf943fd-652e-4400-bb1b-cd8397eb5a95/2022pit-adj.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/1afc56af-ea90-4d48-82e5-1f9aeb43255a/PITbook2022.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/fdf3c548-8aba-4b9c-9eb4-bb564c716015/FYI-104.pdf
        
        Virginia:
        -         
        - https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/reciprocity
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'NewMexico', 'state_occupation': 'Virginia', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def new_mexico_and_washington(self, state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        New Mexico:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Washington:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        New Mexico:
        - https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/fdf3c548-8aba-4b9c-9eb4-bb564c716015/FYI-104.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/0bf943fd-652e-4400-bb1b-cd8397eb5a95/2022pit-adj.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/2f1a6781-9534-4436-b427-1557f9592099/2022pit-adj-ins.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/0bf943fd-652e-4400-bb1b-cd8397eb5a95/2022pit-adj.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/1afc56af-ea90-4d48-82e5-1f9aeb43255a/PITbook2022.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/fdf3c548-8aba-4b9c-9eb4-bb564c716015/FYI-104.pdf
        
        Washington:
        -         
        - https://dor.wa.gov/taxes-rates/income-tax
          https://dor.wa.gov/taxes-rates/other-taxes/capital-gains-tax
          https://dor.wa.gov/taxes-rates/income-tax
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'NewMexico', 'state_occupation': 'Washington', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def new_mexico_and_west_virginia(self, county_occupation='Other', state_agi=75000, pay_periods=24, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        New Mexico:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        West Virginia:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_occupation : str, optional
            County of occupation. Default is Other.
        pay_periods : int, optional
            Number of pay periods. Options are 13, 24, 26, 52. Default is 24.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        New Mexico:
        - https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/fdf3c548-8aba-4b9c-9eb4-bb564c716015/FYI-104.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/0bf943fd-652e-4400-bb1b-cd8397eb5a95/2022pit-adj.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/2f1a6781-9534-4436-b427-1557f9592099/2022pit-adj-ins.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/0bf943fd-652e-4400-bb1b-cd8397eb5a95/2022pit-adj.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/1afc56af-ea90-4d48-82e5-1f9aeb43255a/PITbook2022.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/fdf3c548-8aba-4b9c-9eb4-bb564c716015/FYI-104.pdf
        
        West Virginia:
        -         
        - https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://tax.wv.gov/Documents/TSD/tsd438.pdf
          https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2011/Taxes-11-35att.html
        """
        
        # Check if the parameters are set upstream in the instance
        county_occupation = self.county_occupation if hasattr(self, 'county_occupation') else county_occupation
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        pay_periods = self.pay_periods if hasattr(self, 'pay_periods') else pay_periods
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'NewMexico', 'state_occupation': 'WestVirginia', "county_occupation": county_occupation, "state_agi": state_agi, "pay_periods": pay_periods, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def new_mexico_and_wisconsin(self, state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        New Mexico:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Wisconsin:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        New Mexico:
        - https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/fdf3c548-8aba-4b9c-9eb4-bb564c716015/FYI-104.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/0bf943fd-652e-4400-bb1b-cd8397eb5a95/2022pit-adj.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/2f1a6781-9534-4436-b427-1557f9592099/2022pit-adj-ins.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/0bf943fd-652e-4400-bb1b-cd8397eb5a95/2022pit-adj.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/1afc56af-ea90-4d48-82e5-1f9aeb43255a/PITbook2022.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/fdf3c548-8aba-4b9c-9eb4-bb564c716015/FYI-104.pdf
        
        Wisconsin:
        -         
        - https://www.revenue.wi.gov/TaxForms2023/2023-Form1-ES-Inst.pdf
          https://www.revenue.wi.gov/TaxForms2023/2023-Form1-ES-Inst.pdf
          https://www.revenue.wi.gov/TaxForms2023/2023-Form1-ES-Inst.pdf
          https://www.revenue.wi.gov/Pages/FAQS/pcs-work.aspx
          https://www.revenue.wi.gov/Dor%20Publications/pb103.pdf
          https://www.revenue.wi.gov/Pages/FAQS/pcs-work.aspx
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'NewMexico', 'state_occupation': 'Wisconsin', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def new_mexico_and_wyoming(self, state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        New Mexico:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Wyoming:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        New Mexico:
        - https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/fdf3c548-8aba-4b9c-9eb4-bb564c716015/FYI-104.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/0bf943fd-652e-4400-bb1b-cd8397eb5a95/2022pit-adj.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/2f1a6781-9534-4436-b427-1557f9592099/2022pit-adj-ins.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/0bf943fd-652e-4400-bb1b-cd8397eb5a95/2022pit-adj.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/1afc56af-ea90-4d48-82e5-1f9aeb43255a/PITbook2022.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/fdf3c548-8aba-4b9c-9eb4-bb564c716015/FYI-104.pdf
        
        Wyoming:
        -         
        - https://revenue.wyo.gov/
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'NewMexico', 'state_occupation': 'Wyoming', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def new_york_and_alabama(self, county_residence='Other', state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        New York:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Alabama:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        New York:
        - https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nys-tax-rate-schedule
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nys-tax-rate-schedule
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nys-tax-rate-schedule
          https://www.tax.ny.gov/pit/
          https://www.tax.ny.gov/pit/file/nonresident-faqs.htm
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#yonkers-tax-rate-schedule
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nyc-tax-rate-schedule
        
        Alabama:
        -         
        - https://www.revenue.alabama.gov/ultraviewer/viewer/basic_viewer/index.html?form=2023/02/22f40.pdf
          https://www.revenue.alabama.gov/faqs/what-is-alabamas-individual-income-tax-rate/
          https://www.revenue.alabama.gov/ultraviewer/viewer/basic_viewer/index.html?form=2023/02/22f40.pdf
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-24.htm?taxmap=true
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-24.htm?taxmap=true
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-24.htm?taxmap=true
          https://www.revenue.alabama.gov/individual-corporate/income-to-be-reported-on-the-alabama-income-tax-return/
          https://www.almonline.org/TaxRates.aspx#OccupationalTax
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'NewYork', 'state_occupation': 'Alabama', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def new_york_and_alaska(self, county_residence='Other', state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        New York:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Alaska:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        New York:
        - https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nys-tax-rate-schedule
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nys-tax-rate-schedule
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nys-tax-rate-schedule
          https://www.tax.ny.gov/pit/
          https://www.tax.ny.gov/pit/file/nonresident-faqs.htm
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#yonkers-tax-rate-schedule
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nyc-tax-rate-schedule
        
        Alaska:
        -         
        - https://www.commerce.alaska.gov/web/dcra/officeofthestateassessor/alaskataxfacts.aspx
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'NewYork', 'state_occupation': 'Alaska', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def new_york_and_arizona(self, county_residence='Other', state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        New York:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Arizona:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        credits_state : int, optional
            Credits for state taxes. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        New York:
        - https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nys-tax-rate-schedule
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nys-tax-rate-schedule
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nys-tax-rate-schedule
          https://www.tax.ny.gov/pit/
          https://www.tax.ny.gov/pit/file/nonresident-faqs.htm
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#yonkers-tax-rate-schedule
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nyc-tax-rate-schedule
        
        Arizona:
        -         
        - https://azdor.gov/forms/individual/form-140-x-y-tables
          https://azdor.gov/forms/individual/form-140-x-y-tables
          https://azdor.gov/forms/individual/form-140-x-y-tables
          https://azdor.gov/businesses-arizona/withholding-tax/withholding-exceptions
          "https://azdor.gov/individuals/income-tax-filing-assistance/identifying-other-taxable-income#:~:text=In%20regards%20to%20capital%20gains
         the%20individuals%20regular%20tax%20rate.
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'NewYork', 'state_occupation': 'Arizona', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def new_york_and_arkansas(self, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        New York:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Arkansas:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        New York:
        - https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nys-tax-rate-schedule
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nys-tax-rate-schedule
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nys-tax-rate-schedule
          https://www.tax.ny.gov/pit/
          https://www.tax.ny.gov/pit/file/nonresident-faqs.htm
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#yonkers-tax-rate-schedule
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nyc-tax-rate-schedule
        
        Arkansas:
        -         
        - https://www.arkansasedc.com/why-arkansas/business-climate/tax-structure/personal-income-tax
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_and_AR1000NR_Instructions.pdf
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_FullYearResidentIndividualIncomeTaxReturn.pdf
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_and_AR1000NR_Instructions.pdf
          https://www.arkansasedc.com/why-arkansas/business-climate/tax-structure/capital-gains-tax-reduction
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_and_AR1000NR_Instructions.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'NewYork', 'state_occupation': 'Arkansas', "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def new_york_and_california(self, county_residence='Other', state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        New York:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        California:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        New York:
        - https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nys-tax-rate-schedule
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nys-tax-rate-schedule
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nys-tax-rate-schedule
          https://www.tax.ny.gov/pit/
          https://www.tax.ny.gov/pit/file/nonresident-faqs.htm
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#yonkers-tax-rate-schedule
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nyc-tax-rate-schedule
        
        California:
        -         
        - https://www.ftb.ca.gov/forms/2022/2022-540-tax-rate-schedules.pdf
          https://www.ftb.ca.gov/file/personal/deductions/index.html
          https://www.ftb.ca.gov/about-ftb/newsroom/caleitc/eligibility-and-credit-information.html
          https://www.ftb.ca.gov/file/personal/income-types/capital-gains-and-losses.html#:~:text=California%20does%20not%20have%20a
         are%20taxed%20as%20ordinary%20income
          https://www.ftb.ca.gov/file/personal/
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'NewYork', 'state_occupation': 'California', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def new_york_and_colorado(self, county_residence='Other', state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        New York:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Colorado:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        New York:
        - https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nys-tax-rate-schedule
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nys-tax-rate-schedule
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nys-tax-rate-schedule
          https://www.tax.ny.gov/pit/
          https://www.tax.ny.gov/pit/file/nonresident-faqs.htm
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#yonkers-tax-rate-schedule
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nyc-tax-rate-schedule
        
        Colorado:
        -         
        - https://leg.colorado.gov/agencies/legislative-council-staff/individual-income-tax%C2%A0
          https://leg.colorado.gov/agencies/legislative-council-staff/individual-income-tax%C2%A0
          https://tax.colorado.gov/sites/tax/files/documents/ITT_Capital_Gain_Subtraction_Feb_2022.pdf
          https://www.auroragov.org/business_services/taxes/occupational_privilege_tax
          https://www.denvergov.org/Government/Agencies-Departments-Offices/Agencies-Departments-Offices-Directory/Department-of-Finance/Our-Divisions/Treasury/Business-Tax-Information#:~:text=Occupational%20privilege%20tax
         -%28show%20below%29&text=Each%20taxable%20employee%20is%20liable
         month%20for%20each%20taxable%20employee
          https://www.glendale.co.us/355/Occupational-Privilege-Tax
          https://www.greenwoodvillage.com/1220/Occupational-Privilege-Tax-OPT
          https://ci.sheridan.co.us/288/Occupational-Privilege-Tax
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'NewYork', 'state_occupation': 'Colorado', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def new_york_and_connecticut(self, county_residence='Other', state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        New York:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Connecticut:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        New York:
        - https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nys-tax-rate-schedule
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nys-tax-rate-schedule
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nys-tax-rate-schedule
          https://www.tax.ny.gov/pit/
          https://www.tax.ny.gov/pit/file/nonresident-faqs.htm
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#yonkers-tax-rate-schedule
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nyc-tax-rate-schedule
        
        Connecticut:
        -         
        - https://portal.ct.gov/-/media/DRS/Forms/2022/Income/2022-CT-1040-Instructions_1222.pdf
          https://www.cga.ct.gov/2022/rpt/pdf/2022-R-0108.pdf
          https://www.cga.ct.gov/current/pub/chap_229.htm#sec_12-703
          https://portal.ct.gov/-/media/DRS/Forms/2022/Income/2022-CT-1040-Instructions_1222.pdf
          "https://portal.ct.gov/DRS/Publications/TSSNs/TSSN-29#:~:text=An%20individuals%20net%20capital%20gains
         portion%20of%20Social%20Security%20benefits"
          "https://portal.ct.gov/DRS/Publications/TSSNs/TSSN-29#:~:text=An%20individuals%20net%20capital%20gains
         portion%20of%20Social%20Security%20benefits"
          https://portal.ct.gov/DRS/Individuals/Resident-Income-Tax/Tax-Information
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'NewYork', 'state_occupation': 'Connecticut', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def new_york_and_delaware(self, county_residence='Other', state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, federal_agi=75000, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        New York:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Delaware:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        federal_agi : int, optional
            Adjusted Gross Income (AGI) for federal taxes. Default is 0. Sources - https://www.irs.gov/e-file-providers/definition-of-adjusted-gross-income, https://www.irs.gov/newsroom/taxpayers-should-know-the-difference-between-standard-and-itemized-deductions
        income : int, optional
            Gross income. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        New York:
        - https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nys-tax-rate-schedule
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nys-tax-rate-schedule
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nys-tax-rate-schedule
          https://www.tax.ny.gov/pit/
          https://www.tax.ny.gov/pit/file/nonresident-faqs.htm
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#yonkers-tax-rate-schedule
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nyc-tax-rate-schedule
        
        Delaware:
        -         
        - https://revenue.delaware.gov/software-developer/tax-rate-changes/
          https://revenuefiles.delaware.gov/2022/PIT-RES_TY22_2022-01_PaperInteractive.pdf
          https://revenuefiles.delaware.gov/2022/PIT-RES_TY22_2022-02_Instructions.pdf
          https://revenue.delaware.gov/frequently-asked-questions/personal-income-tax-faqs/
          https://www.newcastlede.gov/DocumentCenter/View/20147/City-of-Wilmington-Wage-Tax-Information-PDF
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        federal_agi = self.federal_agi if hasattr(self, 'federal_agi') else federal_agi
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'NewYork', 'state_occupation': 'Delaware', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "federal_agi": federal_agi, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def new_york_and_florida(self, county_residence='Other', state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        New York:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Florida:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        New York:
        - https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nys-tax-rate-schedule
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nys-tax-rate-schedule
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nys-tax-rate-schedule
          https://www.tax.ny.gov/pit/
          https://www.tax.ny.gov/pit/file/nonresident-faqs.htm
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#yonkers-tax-rate-schedule
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nyc-tax-rate-schedule
        
        Florida:
        -         
        - https://floridarevenue.com/taxes/taxesfees/Pages/tax_interest_rates.aspx
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'NewYork', 'state_occupation': 'Florida', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def new_york_and_georgia(self, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        New York:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Georgia:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        New York:
        - https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nys-tax-rate-schedule
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nys-tax-rate-schedule
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nys-tax-rate-schedule
          https://www.tax.ny.gov/pit/
          https://www.tax.ny.gov/pit/file/nonresident-faqs.htm
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#yonkers-tax-rate-schedule
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nyc-tax-rate-schedule
        
        Georgia:
        -         
        - https://dor.georgia.gov/tax-tables-georgia-tax-rate-schedule
          https://apps.dor.ga.gov/FillableForms/PDFViewer/Index?form=2022GA500
          https://apps.dor.ga.gov/FillableForms/PDFViewer/Index?form=2022GA500
          https://apps.dor.ga.gov/FillableForms/PDFViewer/Index?form=2022GA500
          https://dor.georgia.gov/years-individual-income-tax-forms
          https://dor.georgia.gov/years-individual-income-tax-forms
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'NewYork', 'state_occupation': 'Georgia', "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def new_york_and_hawaii(self, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        New York:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Hawaii:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        New York:
        - https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nys-tax-rate-schedule
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nys-tax-rate-schedule
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nys-tax-rate-schedule
          https://www.tax.ny.gov/pit/
          https://www.tax.ny.gov/pit/file/nonresident-faqs.htm
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#yonkers-tax-rate-schedule
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nyc-tax-rate-schedule
        
        Hawaii:
        -         
        - https://tax.hawaii.gov/forms/d_18table-on/d_18table-on_p13/
          https://files.hawaii.gov/tax/forms/2018/18table-on.pdf
          https://files.hawaii.gov/tax/news/pubs/22outline.pdf
          https://files.hawaii.gov/tax/news/pubs/FastTaxRefGuide_2017.pdf
          https://www.tfhawaii.org/wordpress/state-tax-resources/mini-tax-guides/individual-income-tax-chapter-235/#:~:text=Long%20term%20capital%20gains%20are%20taxed%20at%20a%20maximum%20of%207.25%25
          https://www.grassrootinstitute.org/2023/02/hb337-increasing-tax-on-capital-gains-a-bad-risk/
          https://files.hawaii.gov/tax/news/pubs/22outline.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'NewYork', 'state_occupation': 'Hawaii', "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def new_york_and_idaho(self, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        New York:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Idaho:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        New York:
        - https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nys-tax-rate-schedule
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nys-tax-rate-schedule
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nys-tax-rate-schedule
          https://www.tax.ny.gov/pit/
          https://www.tax.ny.gov/pit/file/nonresident-faqs.htm
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#yonkers-tax-rate-schedule
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nyc-tax-rate-schedule
        
        Idaho:
        -         
        - https://tax.idaho.gov/taxes/income-tax/individual-income/individual-income-tax-rate-schedule/
          https://tax.idaho.gov/wp-content/uploads/forms/EIN00046/EIN00046_03-01-2023.pdf
          https://tax.idaho.gov/taxes/income-tax/individual-income/idaho-source-income/
          https://tax.idaho.gov/taxes/income-tax/individual-income/idaho-source-income/
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'NewYork', 'state_occupation': 'Idaho', "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def new_york_and_illinois(self, county_residence='Other', state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        New York:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Illinois:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        New York:
        - https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nys-tax-rate-schedule
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nys-tax-rate-schedule
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nys-tax-rate-schedule
          https://www.tax.ny.gov/pit/
          https://www.tax.ny.gov/pit/file/nonresident-faqs.htm
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#yonkers-tax-rate-schedule
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nyc-tax-rate-schedule
        
        Illinois:
        -         
        - https://tax.illinois.gov/research/taxrates/income.html#IndividualIncome
          https://tax.illinois.gov/content/dam/soi/en/web/tax/forms/incometax/documents/currentyear/individual/il-1040-instr.pdf
          https://tax.illinois.gov/questionsandanswers/12.html
          https://tax.illinois.gov/content/dam/soi/en/web/tax/forms/incometax/documents/currentyear/individual/il-1040-instr.pdf
          https://tax.illinois.gov/research/taxrates/personalproperty.html
          https://tax.illinois.gov/localgovernments/income.html
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'NewYork', 'state_occupation': 'Illinois', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def new_york_and_indiana(self, capital_gains_short=0, dependents=0, sources=False, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_long=0, filing_status='single', income=100000, credits_state=0):
        """
        --- Parameters ----
        
        New York:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Indiana:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        New York:
        - https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nys-tax-rate-schedule
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nys-tax-rate-schedule
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nys-tax-rate-schedule
          https://www.tax.ny.gov/pit/
          https://www.tax.ny.gov/pit/file/nonresident-faqs.htm
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#yonkers-tax-rate-schedule
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nyc-tax-rate-schedule
        
        Indiana:
        -         
        - https://www.in.gov/dor/tax-forms/2022-individual-income-tax-forms/
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-32.htm
          https://www.in.gov/dor/files/ib28.pdf
          https://www.in.gov/dor/tax-forms/2022-individual-income-tax-forms/
          https://www.in.gov/dor/files/2022-county-tax-rates-and-codes.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'NewYork', 'state_occupation': 'Indiana', "capital_gains_short": capital_gains_short, "dependents": dependents, "sources": sources, "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def new_york_and_iowa(self, county_residence='Other', state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        New York:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Iowa:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        New York:
        - https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nys-tax-rate-schedule
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nys-tax-rate-schedule
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nys-tax-rate-schedule
          https://www.tax.ny.gov/pit/
          https://www.tax.ny.gov/pit/file/nonresident-faqs.htm
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#yonkers-tax-rate-schedule
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nyc-tax-rate-schedule
        
        Iowa:
        -         
        - https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
          https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
          https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
          https://tax.iowa.gov/iowa-illinois-reciprocal-agreement
          https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'NewYork', 'state_occupation': 'Iowa', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def new_york_and_kansas(self, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        New York:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Kansas:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        New York:
        - https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nys-tax-rate-schedule
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nys-tax-rate-schedule
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nys-tax-rate-schedule
          https://www.tax.ny.gov/pit/
          https://www.tax.ny.gov/pit/file/nonresident-faqs.htm
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#yonkers-tax-rate-schedule
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nyc-tax-rate-schedule
        
        Kansas:
        -         
        - https://www.ksrevenue.gov/taxrates.html
          https://www.ksrevenue.gov/incomebook22.html#0
          https://www.ksrevenue.gov/incomebook22.html#0
          https://www.ksrevenue.gov/incomebook22.html#0
          https://www.ksrevenue.gov/incomebook22.html
          https://www.ksrevenue.gov/incomebook22.html
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'NewYork', 'state_occupation': 'Kansas', "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def new_york_and_kentucky(self, county_residence='Other', state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        New York:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Kentucky:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        New York:
        - https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nys-tax-rate-schedule
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nys-tax-rate-schedule
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nys-tax-rate-schedule
          https://www.tax.ny.gov/pit/
          https://www.tax.ny.gov/pit/file/nonresident-faqs.htm
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#yonkers-tax-rate-schedule
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nyc-tax-rate-schedule
        
        Kentucky:
        -         
        - https://revenue.ky.gov/News/Pages/Dor-Announces-Updates-to-Individual-Income-Tax-for-2023-Tax-Year.aspx
          https://revenue.ky.gov/Individual/Individual-Income-Tax/Pages/default.aspx
          https://revenue.ky.gov/News/Pages/Dor-Announces-Updates-to-Individual-Income-Tax-for-2023-Tax-Year.aspx
          https://revenue.ky.gov/Individual/Individual-Income-Tax/Pages/default.aspx
          https://revenue.ky.gov/Dor%20Training%20Materials/103%20KAR%2017.140.%20Individual%20income%20tax%20-%20reciprocity%20-%20nonresidents.pdf
          https://revenue.ky.gov/Individual/Individual-Income-Tax/Pages/default.aspx
          https://nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2021/Taxes-21-32.htm?taxmap=true
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'NewYork', 'state_occupation': 'Kentucky', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def new_york_and_louisiana(self, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        New York:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Louisiana:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        New York:
        - https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nys-tax-rate-schedule
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nys-tax-rate-schedule
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nys-tax-rate-schedule
          https://www.tax.ny.gov/pit/
          https://www.tax.ny.gov/pit/file/nonresident-faqs.htm
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#yonkers-tax-rate-schedule
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nyc-tax-rate-schedule
        
        Louisiana:
        -         
        - https://revenue.louisiana.gov/individualincometax
          https://revenue.louisiana.gov/individualincometax
          https://www.revenue.louisiana.gov/taxforms/6935(11_02)F.pdf
          https://revenue.louisiana.gov/individualincometax
          https://revenue.louisiana.gov/individualincometax
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'NewYork', 'state_occupation': 'Louisiana', "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def new_york_and_maine(self, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        New York:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Maine:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        New York:
        - https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nys-tax-rate-schedule
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nys-tax-rate-schedule
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nys-tax-rate-schedule
          https://www.tax.ny.gov/pit/
          https://www.tax.ny.gov/pit/file/nonresident-faqs.htm
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#yonkers-tax-rate-schedule
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nyc-tax-rate-schedule
        
        Maine:
        -         
        - https://www.maine.gov/revenue/tax-return-forms/individual-income-tax-2022
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/ind_tax_rate_sched_2022.pdf
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/22_1040me_book_gen_instruc_revisedFeb23.pdf
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/22_1040me_book_gen_instruc_revisedFeb23.pdf
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/22_1040me_book_gen_instruc_revisedFeb23.pdf
          https://www.maine.gov/revenue/tax-return-forms/individual-income-tax-2022
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'NewYork', 'state_occupation': 'Maine', "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def new_york_and_maryland(self, county_residence='Other', state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        New York:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Maryland:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        New York:
        - https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nys-tax-rate-schedule
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nys-tax-rate-schedule
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nys-tax-rate-schedule
          https://www.tax.ny.gov/pit/
          https://www.tax.ny.gov/pit/file/nonresident-faqs.htm
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#yonkers-tax-rate-schedule
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nyc-tax-rate-schedule
        
        Maryland:
        -         
        - https://www.marylandtaxes.gov/individual/income/tax-info/tax-rates.php
          https://www.marylandtaxes.gov/forms/22_forms/Resident_Booklet.pdf
          https://www.marylandtaxes.gov/forms/Tax_Publications/Administrative_Releases/Income_and_Estate_Tax_Releases/ar_it37.pdf
          https://www.marylandtaxes.gov/individual/index.php
          https://www.marylandtaxes.gov/individual/income/tax-info/tax-rates.php
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'NewYork', 'state_occupation': 'Maryland', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def new_york_and_massachusetts(self, county_residence='Other', state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        New York:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Massachusetts:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        New York:
        - https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nys-tax-rate-schedule
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nys-tax-rate-schedule
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nys-tax-rate-schedule
          https://www.tax.ny.gov/pit/
          https://www.tax.ny.gov/pit/file/nonresident-faqs.htm
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#yonkers-tax-rate-schedule
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nyc-tax-rate-schedule
        
        Massachusetts:
        -         
        - https://www.mass.gov/service-details/massachusetts-tax-rates
          https://www.mass.gov/service-details/view-massachusetts-personal-income-tax-exemptions
          https://www.mass.gov/service-details/view-massachusetts-personal-income-tax-exemptions
          https://www.mass.gov/service-details/view-child-and-dependent-related-credits
          https://www.mass.gov/service-details/massachusetts-tax-rates
          https://www.mass.gov/personal-income-tax
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'NewYork', 'state_occupation': 'Massachusetts', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def new_york_and_michigan(self, capital_gains_short=0, county_occupation='Other', dependents=0, sources=False, county_residence='Other', state_agi=75000, capital_gains_long=0, filing_status='single', income=100000, credits_state=0):
        """
        --- Parameters ----
        
        New York:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Michigan:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        county_occupation : str, optional
            County of occupation. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        New York:
        - https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nys-tax-rate-schedule
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nys-tax-rate-schedule
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nys-tax-rate-schedule
          https://www.tax.ny.gov/pit/
          https://www.tax.ny.gov/pit/file/nonresident-faqs.htm
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#yonkers-tax-rate-schedule
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nyc-tax-rate-schedule
        
        Michigan:
        -         
        - https://www.michigan.gov/taxes/iit/new-developments-for-tax-year-2022
          "https://www.michigan.gov/taxes/-/media/Project/Websites/taxes/Forms/2023/2023-SUW/446_Withholding-Guide_2023.pdf?rev=177ca3d8cb034392bf84e317ed78c320&hash=19C5F42F43A89A91863D229349F48883#:~:text=The%20exemption%20amount%20is%20%245%2C400
         employees%20Michigan%20income%20tax%20return"
          "https://www.michigan.gov/taxes/-/media/Project/Websites/taxes/Forms/2023/2023-SUW/446_Withholding-Guide_2023.pdf?rev=177ca3d8cb034392bf84e317ed78c320&hash=19C5F42F43A89A91863D229349F48883#:~:text=The%20exemption%20amount%20is%20%245%2C400
         employees%20Michigan%20income%20tax%20return"
          https://www.michigan.gov/taxes/questions/iit/accordion/residency/are-my-wages-earned-in-another-state-taxable-in-michigan-if-i-am-a-michigan-resident-1
          https://www.michigan.gov/taxes/-/media/Project/Websites/taxes/Forms/2022/2022-IIT-Forms/BOOK_MI-1040.pdf?rev=4524f1cbbc0243b194ab05a6680be75b&hash=40DFE71E651F7C1D65C31F6CD03D944F
          https://www.michigan.gov/taxes/questions/iit/accordion/general/what-cities-impose-an-income-tax
        """
        
        # Check if the parameters are set upstream in the instance
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        county_occupation = self.county_occupation if hasattr(self, 'county_occupation') else county_occupation
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'NewYork', 'state_occupation': 'Michigan', "capital_gains_short": capital_gains_short, "county_occupation": county_occupation, "dependents": dependents, "sources": sources, "county_residence": county_residence, "state_agi": state_agi, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def new_york_and_minnesota(self, county_residence='Other', state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        New York:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Minnesota:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        New York:
        - https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nys-tax-rate-schedule
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nys-tax-rate-schedule
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nys-tax-rate-schedule
          https://www.tax.ny.gov/pit/
          https://www.tax.ny.gov/pit/file/nonresident-faqs.htm
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#yonkers-tax-rate-schedule
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nyc-tax-rate-schedule
        
        Minnesota:
        -         
        - https://www.revenue.state.mn.us/minnesota-income-tax-rates-and-brackets
          https://www.revenue.state.mn.us/child-and-dependent-care-credit
          https://www.revenue.state.mn.us/dependent-exemptions
          https://www.revenue.state.mn.us/minnesota-standard-deduction
          https://mn.gov/mmb/assets/mwr_form_tcm1059-128581.pdf
          https://www.revenue.state.mn.us/individual-income-tax
          https://www.revenue.state.mn.us/individual-income-tax
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'NewYork', 'state_occupation': 'Minnesota', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def new_york_and_mississippi(self, county_residence='Other', state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        New York:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Mississippi:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        New York:
        - https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nys-tax-rate-schedule
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nys-tax-rate-schedule
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nys-tax-rate-schedule
          https://www.tax.ny.gov/pit/
          https://www.tax.ny.gov/pit/file/nonresident-faqs.htm
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#yonkers-tax-rate-schedule
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nyc-tax-rate-schedule
        
        Mississippi:
        -         
        - https://www.dor.ms.gov/individual/tax-rates
          https://www.dor.ms.gov/individual/tax-rates
          https://www.dor.ms.gov/individual/tax-rates
          https://www.dor.ms.gov/individual
          https://www.dor.ms.gov/individual
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'NewYork', 'state_occupation': 'Mississippi', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def new_york_and_missouri(self, capital_gains_short=0, dependents=0, sources=False, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_long=0, filing_status='single', income=100000, credits_state=0):
        """
        --- Parameters ----
        
        New York:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Missouri:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        New York:
        - https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nys-tax-rate-schedule
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nys-tax-rate-schedule
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nys-tax-rate-schedule
          https://www.tax.ny.gov/pit/
          https://www.tax.ny.gov/pit/file/nonresident-faqs.htm
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#yonkers-tax-rate-schedule
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nyc-tax-rate-schedule
        
        Missouri:
        -         
        - https://dor.mo.gov/taxation/individual/tax-types/income/year-changes/
          https://dor.mo.gov/taxation/individual/tax-types/income/year-changes/
          https://dor.mo.gov/taxation/individual/tax-types/income/year-changes/
          https://www.kcmo.gov/city-hall/departments/finance/earnings-tax
          https://www.stlouis-mo.gov/government/departments/collector/earnings-tax/individual-earnings-tax-info.cfm
        """
        
        # Check if the parameters are set upstream in the instance
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'NewYork', 'state_occupation': 'Missouri', "capital_gains_short": capital_gains_short, "dependents": dependents, "sources": sources, "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def new_york_and_montana(self, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        New York:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Montana:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        New York:
        - https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nys-tax-rate-schedule
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nys-tax-rate-schedule
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nys-tax-rate-schedule
          https://www.tax.ny.gov/pit/
          https://www.tax.ny.gov/pit/file/nonresident-faqs.htm
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#yonkers-tax-rate-schedule
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nyc-tax-rate-schedule
        
        Montana:
        -         
        - https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/2022/12/montana-employees-withholding-allowance-and-exemption-certificate-form-mw-4.pdf
          https://mtrevenue.gov/2021/11/29/simplification-of-montana-income-taxation/?utm_source=rss&utm_medium=rss&utm_campaign=simplification-of-montana-income-taxation
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'NewYork', 'state_occupation': 'Montana', "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def new_york_and_nebraska(self, county_residence='Other', state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        New York:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Nebraska:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        New York:
        - https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nys-tax-rate-schedule
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nys-tax-rate-schedule
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nys-tax-rate-schedule
          https://www.tax.ny.gov/pit/
          https://www.tax.ny.gov/pit/file/nonresident-faqs.htm
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#yonkers-tax-rate-schedule
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nyc-tax-rate-schedule
        
        Nebraska:
        -         
        - https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2023_Circular_EN_8-429r2023_Whole_Book_11.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'NewYork', 'state_occupation': 'Nebraska', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def new_york_and_nevada(self, county_residence='Other', state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        New York:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Nevada:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        New York:
        - https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nys-tax-rate-schedule
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nys-tax-rate-schedule
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nys-tax-rate-schedule
          https://www.tax.ny.gov/pit/
          https://www.tax.ny.gov/pit/file/nonresident-faqs.htm
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#yonkers-tax-rate-schedule
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nyc-tax-rate-schedule
        
        Nevada:
        -         
        - https://www.nvsos.gov/sos/businesses/the-nevada-advantage
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'NewYork', 'state_occupation': 'Nevada', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def new_york_and_new_hampshire(self, county_residence='Other', state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        New York:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        New Hampshire:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        New York:
        - https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nys-tax-rate-schedule
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nys-tax-rate-schedule
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nys-tax-rate-schedule
          https://www.tax.ny.gov/pit/
          https://www.tax.ny.gov/pit/file/nonresident-faqs.htm
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#yonkers-tax-rate-schedule
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nyc-tax-rate-schedule
        
        New Hampshire:
        -         
        - https://www.revenue.nh.gov/assistance/tax-overview.htm
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'NewYork', 'state_occupation': 'NewHampshire', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def new_york_and_new_jersey(self, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        New York:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        New Jersey:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        New York:
        - https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nys-tax-rate-schedule
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nys-tax-rate-schedule
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nys-tax-rate-schedule
          https://www.tax.ny.gov/pit/
          https://www.tax.ny.gov/pit/file/nonresident-faqs.htm
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#yonkers-tax-rate-schedule
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nyc-tax-rate-schedule
        
        New Jersey:
        -         
        - https://www.state.nj.us/treasury/taxation/taxtables.shtml
          https://www.state.nj.us/treasury/taxation/njit13.shtml
          https://www.nj.gov/treasury/taxation/njit25.shtml
          https://www.state.nj.us/treasury/taxation/individuals/freqqiti.shtml
          https://www.state.nj.us/treasury/taxation/localtax.shtml
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'NewYork', 'state_occupation': 'NewJersey', "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def new_york_and_new_mexico(self, county_residence='Other', state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        New York:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        New Mexico:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        New York:
        - https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nys-tax-rate-schedule
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nys-tax-rate-schedule
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nys-tax-rate-schedule
          https://www.tax.ny.gov/pit/
          https://www.tax.ny.gov/pit/file/nonresident-faqs.htm
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#yonkers-tax-rate-schedule
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nyc-tax-rate-schedule
        
        New Mexico:
        -         
        - https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/fdf3c548-8aba-4b9c-9eb4-bb564c716015/FYI-104.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/0bf943fd-652e-4400-bb1b-cd8397eb5a95/2022pit-adj.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/2f1a6781-9534-4436-b427-1557f9592099/2022pit-adj-ins.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/0bf943fd-652e-4400-bb1b-cd8397eb5a95/2022pit-adj.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/1afc56af-ea90-4d48-82e5-1f9aeb43255a/PITbook2022.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/fdf3c548-8aba-4b9c-9eb4-bb564c716015/FYI-104.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'NewYork', 'state_occupation': 'NewMexico', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def new_york_and_north_carolina(self, county_residence='Other', state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        New York:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        North Carolina:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        New York:
        - https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nys-tax-rate-schedule
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nys-tax-rate-schedule
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nys-tax-rate-schedule
          https://www.tax.ny.gov/pit/
          https://www.tax.ny.gov/pit/file/nonresident-faqs.htm
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#yonkers-tax-rate-schedule
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nyc-tax-rate-schedule
        
        North Carolina:
        -         
        - https://www.ncdor.gov/taxes-forms/tax-rate-schedules
          https://www.ncdor.gov/taxes-forms/individual-income-tax/north-carolina-standard-deduction-or-north-carolina-itemized-deductions
          https://www.ncdor.gov/taxes-forms/individual-income-tax
          https://www.ncdor.gov/taxes-forms/individual-income-tax
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'NewYork', 'state_occupation': 'NorthCarolina', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def new_york_and_north_dakota(self, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        New York:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        North Dakota:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        New York:
        - https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nys-tax-rate-schedule
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nys-tax-rate-schedule
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nys-tax-rate-schedule
          https://www.tax.ny.gov/pit/
          https://www.tax.ny.gov/pit/file/nonresident-faqs.htm
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#yonkers-tax-rate-schedule
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nyc-tax-rate-schedule
        
        North Dakota:
        -         
        - https://www.tax.nd.gov/forms
          https://www.tax.nd.gov/sites/www/files/documents/forms/individual/2022-iit/2022-individual-income-tax-booklet.pdf
          https://www.tax.nd.gov/sites/www/files/documents/forms/2022-individual-income-tax-booklet.pdf
          https://www.tax.nd.gov/tax-exemptions-credits/income-tax-exemptions-credits
          https://www.tax.nd.gov/sites/www/files/documents/forms/business/ndwrfillable.pdf
          https://www.tax.nd.gov/news/tax-legislative-changes/significant-changes-law/individual-income-tax-history
          https://www.tax.nd.gov/forms
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'NewYork', 'state_occupation': 'NorthDakota', "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def new_york_and_ohio(self, county_residence='Other', state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        New York:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Ohio:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        New York:
        - https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nys-tax-rate-schedule
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nys-tax-rate-schedule
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nys-tax-rate-schedule
          https://www.tax.ny.gov/pit/
          https://www.tax.ny.gov/pit/file/nonresident-faqs.htm
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#yonkers-tax-rate-schedule
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nyc-tax-rate-schedule
        
        Ohio:
        -         
        - https://tax.ohio.gov/individual/resources/annual-tax-rates
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://tax.ohio.gov/static/forms/employer_withholding/generic/wth_it4nr.pdf
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://www.ritaohio.com/TaxRatesTable
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'NewYork', 'state_occupation': 'Ohio', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def new_york_and_oklahoma(self, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        New York:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Oklahoma:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        New York:
        - https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nys-tax-rate-schedule
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nys-tax-rate-schedule
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nys-tax-rate-schedule
          https://www.tax.ny.gov/pit/
          https://www.tax.ny.gov/pit/file/nonresident-faqs.htm
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#yonkers-tax-rate-schedule
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nyc-tax-rate-schedule
        
        Oklahoma:
        -         
        - https://oklahoma.gov/content/dam/ok/en/tax/documents/resources/publications/businesses/withholding-tables/WHTables-2022.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'NewYork', 'state_occupation': 'Oklahoma', "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def new_york_and_oregon(self, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        New York:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Oregon:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        New York:
        - https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nys-tax-rate-schedule
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nys-tax-rate-schedule
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nys-tax-rate-schedule
          https://www.tax.ny.gov/pit/
          https://www.tax.ny.gov/pit/file/nonresident-faqs.htm
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#yonkers-tax-rate-schedule
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nyc-tax-rate-schedule
        
        Oregon:
        -         
        - https://www.oregon.gov/dor/programs/individuals/Pages/PIT.aspx
          https://www.oregon.gov/dor/programs/individuals/Documents/Full%20year%20resident
         %20Form%20or-40%20filers.pdf
          https://www.oregon.gov/dor/programs/individuals/Documents/Part-year%20and%20nonresident
         %20Form%20or-40-P%20and%20or-40-N%20filers.pdf.pdf
          https://www.oregon.gov/dor/forms/FormsPubs/form-or-40_101-040_2022.pdf
          https://www.oregon.gov/dor/forms/FormsPubs/form-or-40_101-040_2022.pdf
          https://www.oregon.gov/dor/forms/FormsPubs/form-or-40_101-040_2022.pdf
          https://www.portland.gov/revenue/personal-tax
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'NewYork', 'state_occupation': 'Oregon', "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def new_york_and_pennsylvania(self, county_residence='Other', state_agi=75000, county_occupation='Other', capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        New York:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Pennsylvania:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        county_residence : str, optional
            County of residence. Default is Other.
        county_occupation : str, optional
            County of occupation. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        New York:
        - https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nys-tax-rate-schedule
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nys-tax-rate-schedule
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nys-tax-rate-schedule
          https://www.tax.ny.gov/pit/
          https://www.tax.ny.gov/pit/file/nonresident-faqs.htm
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#yonkers-tax-rate-schedule
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nyc-tax-rate-schedule
        
        Pennsylvania:
        -         
        - https://www.revenue.pa.gov/Tax%20Rates/Pages/default.aspx
          https://www.revenue.pa.gov/FormsandPublications/FormsforIndividuals/PIT/Documents/2022/2022_pa-40in.pdf
          https://www.revenue.pa.gov/FormsandPublications/FormsforBusinesses/EmployerWithholding/Documents/rev-419.pdf
          https://www.revenue.pa.gov/FormsandPublications/FormsforIndividuals/PIT/Documents/2022/2022_pa-40in.pdf
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-25.htm
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        county_occupation = self.county_occupation if hasattr(self, 'county_occupation') else county_occupation
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'NewYork', 'state_occupation': 'Pennsylvania', "county_residence": county_residence, "state_agi": state_agi, "county_occupation": county_occupation, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def new_york_and_rhode_island(self, county_residence='Other', state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        New York:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Rhode Island:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        New York:
        - https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nys-tax-rate-schedule
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nys-tax-rate-schedule
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nys-tax-rate-schedule
          https://www.tax.ny.gov/pit/
          https://www.tax.ny.gov/pit/file/nonresident-faqs.htm
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#yonkers-tax-rate-schedule
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nyc-tax-rate-schedule
        
        Rhode Island:
        -         
        - https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2022_1040WE_w_0.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'NewYork', 'state_occupation': 'RhodeIsland', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def new_york_and_south_carolina(self, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        New York:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        South Carolina:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        New York:
        - https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nys-tax-rate-schedule
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nys-tax-rate-schedule
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nys-tax-rate-schedule
          https://www.tax.ny.gov/pit/
          https://www.tax.ny.gov/pit/file/nonresident-faqs.htm
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#yonkers-tax-rate-schedule
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nyc-tax-rate-schedule
        
        South Carolina:
        -         
        - https://dor.sc.gov/resources-site/lawandpolicy/Advisory%20Opinions/IL22-15.pdf
          https://dor.sc.gov/forms-site/Forms/IITPacket_2022.pdf
          https://www.scstatehouse.gov/code/t12c006.php
          https://dor.sc.gov/resources-site/lawandpolicy/Advisory%20Opinions/IL22-15.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'NewYork', 'state_occupation': 'SouthCarolina', "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def new_york_and_south_dakota(self, county_residence='Other', state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        New York:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        South Dakota:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        New York:
        - https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nys-tax-rate-schedule
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nys-tax-rate-schedule
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nys-tax-rate-schedule
          https://www.tax.ny.gov/pit/
          https://www.tax.ny.gov/pit/file/nonresident-faqs.htm
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#yonkers-tax-rate-schedule
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nyc-tax-rate-schedule
        
        South Dakota:
        -         
        - https://dor.sd.gov/individuals/taxes/
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'NewYork', 'state_occupation': 'SouthDakota', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def new_york_and_tennessee(self, county_residence='Other', state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        New York:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Tennessee:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        New York:
        - https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nys-tax-rate-schedule
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nys-tax-rate-schedule
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nys-tax-rate-schedule
          https://www.tax.ny.gov/pit/
          https://www.tax.ny.gov/pit/file/nonresident-faqs.htm
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#yonkers-tax-rate-schedule
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nyc-tax-rate-schedule
        
        Tennessee:
        -         
        - https://www.tn.gov/revenue/taxes.html
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'NewYork', 'state_occupation': 'Tennessee', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def new_york_and_texas(self, county_residence='Other', state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        New York:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Texas:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        New York:
        - https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nys-tax-rate-schedule
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nys-tax-rate-schedule
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nys-tax-rate-schedule
          https://www.tax.ny.gov/pit/
          https://www.tax.ny.gov/pit/file/nonresident-faqs.htm
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#yonkers-tax-rate-schedule
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nyc-tax-rate-schedule
        
        Texas:
        -         
        - https://comptroller.texas.gov/taxes/
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'NewYork', 'state_occupation': 'Texas', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def new_york_and_utah(self, county_residence='Other', state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        New York:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Utah:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        New York:
        - https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nys-tax-rate-schedule
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nys-tax-rate-schedule
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nys-tax-rate-schedule
          https://www.tax.ny.gov/pit/
          https://www.tax.ny.gov/pit/file/nonresident-faqs.htm
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#yonkers-tax-rate-schedule
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nyc-tax-rate-schedule
        
        Utah:
        -         
        - https://incometax.utah.gov/paying/tax-rates
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'NewYork', 'state_occupation': 'Utah', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def new_york_and_vermont(self, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        New York:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Vermont:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        New York:
        - https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nys-tax-rate-schedule
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nys-tax-rate-schedule
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nys-tax-rate-schedule
          https://www.tax.ny.gov/pit/
          https://www.tax.ny.gov/pit/file/nonresident-faqs.htm
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#yonkers-tax-rate-schedule
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nyc-tax-rate-schedule
        
        Vermont:
        -         
        - https://tax.vermont.gov/sites/tax/files/documents/GB-1210-2023.pdf
          https://tax.vermont.gov/sites/tax/files/documents/Income%20Booklet-2022.pdf
          https://tax.vermont.gov/sites/tax/files/documents/Income%20Booklet-2022.pdf
          https://tax.vermont.gov/sites/tax/files/documents/Income%20Booklet-2022.pdf
          https://tax.vermont.gov/individuals/personal-income-tax/tax-credits
          https://tax.vermont.gov/individuals/personal-income-tax/taxable-income
          https://tax.vermont.gov/individuals/personal-income-tax/local-option-tax
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'NewYork', 'state_occupation': 'Vermont', "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def new_york_and_virginia(self, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        New York:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Virginia:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        New York:
        - https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nys-tax-rate-schedule
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nys-tax-rate-schedule
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nys-tax-rate-schedule
          https://www.tax.ny.gov/pit/
          https://www.tax.ny.gov/pit/file/nonresident-faqs.htm
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#yonkers-tax-rate-schedule
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nyc-tax-rate-schedule
        
        Virginia:
        -         
        - https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/reciprocity
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'NewYork', 'state_occupation': 'Virginia', "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def new_york_and_washington(self, county_residence='Other', state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        New York:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Washington:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        New York:
        - https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nys-tax-rate-schedule
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nys-tax-rate-schedule
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nys-tax-rate-schedule
          https://www.tax.ny.gov/pit/
          https://www.tax.ny.gov/pit/file/nonresident-faqs.htm
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#yonkers-tax-rate-schedule
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nyc-tax-rate-schedule
        
        Washington:
        -         
        - https://dor.wa.gov/taxes-rates/income-tax
          https://dor.wa.gov/taxes-rates/other-taxes/capital-gains-tax
          https://dor.wa.gov/taxes-rates/income-tax
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'NewYork', 'state_occupation': 'Washington', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def new_york_and_west_virginia(self, pay_periods=24, capital_gains_short=0, county_occupation='Other', dependents=0, sources=False, county_residence='Other', state_agi=75000, capital_gains_long=0, filing_status='single', income=100000, credits_state=0):
        """
        --- Parameters ----
        
        New York:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        West Virginia:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_occupation : str, optional
            County of occupation. Default is Other.
        pay_periods : int, optional
            Number of pay periods. Options are 13, 24, 26, 52. Default is 24.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        New York:
        - https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nys-tax-rate-schedule
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nys-tax-rate-schedule
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nys-tax-rate-schedule
          https://www.tax.ny.gov/pit/
          https://www.tax.ny.gov/pit/file/nonresident-faqs.htm
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#yonkers-tax-rate-schedule
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nyc-tax-rate-schedule
        
        West Virginia:
        -         
        - https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://tax.wv.gov/Documents/TSD/tsd438.pdf
          https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2011/Taxes-11-35att.html
        """
        
        # Check if the parameters are set upstream in the instance
        pay_periods = self.pay_periods if hasattr(self, 'pay_periods') else pay_periods
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        county_occupation = self.county_occupation if hasattr(self, 'county_occupation') else county_occupation
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'NewYork', 'state_occupation': 'WestVirginia', "pay_periods": pay_periods, "capital_gains_short": capital_gains_short, "county_occupation": county_occupation, "dependents": dependents, "sources": sources, "county_residence": county_residence, "state_agi": state_agi, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def new_york_and_wisconsin(self, county_residence='Other', state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        New York:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Wisconsin:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        New York:
        - https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nys-tax-rate-schedule
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nys-tax-rate-schedule
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nys-tax-rate-schedule
          https://www.tax.ny.gov/pit/
          https://www.tax.ny.gov/pit/file/nonresident-faqs.htm
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#yonkers-tax-rate-schedule
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nyc-tax-rate-schedule
        
        Wisconsin:
        -         
        - https://www.revenue.wi.gov/TaxForms2023/2023-Form1-ES-Inst.pdf
          https://www.revenue.wi.gov/TaxForms2023/2023-Form1-ES-Inst.pdf
          https://www.revenue.wi.gov/TaxForms2023/2023-Form1-ES-Inst.pdf
          https://www.revenue.wi.gov/Pages/FAQS/pcs-work.aspx
          https://www.revenue.wi.gov/Dor%20Publications/pb103.pdf
          https://www.revenue.wi.gov/Pages/FAQS/pcs-work.aspx
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'NewYork', 'state_occupation': 'Wisconsin', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def new_york_and_wyoming(self, county_residence='Other', state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        New York:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Wyoming:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        New York:
        - https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nys-tax-rate-schedule
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nys-tax-rate-schedule
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nys-tax-rate-schedule
          https://www.tax.ny.gov/pit/
          https://www.tax.ny.gov/pit/file/nonresident-faqs.htm
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#yonkers-tax-rate-schedule
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nyc-tax-rate-schedule
        
        Wyoming:
        -         
        - https://revenue.wyo.gov/
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'NewYork', 'state_occupation': 'Wyoming', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def north_carolina_and_alabama(self, capital_gains_short=0, income=100000, dependents=0, sources=False, county_residence='Other', state_agi=75000, capital_gains_long=0, filing_status='single', credits_state=0):
        """
        --- Parameters ----
        
        North Carolina:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        
        Alabama:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        North Carolina:
        - https://www.ncdor.gov/taxes-forms/tax-rate-schedules
          https://www.ncdor.gov/taxes-forms/individual-income-tax/north-carolina-standard-deduction-or-north-carolina-itemized-deductions
          https://www.ncdor.gov/taxes-forms/individual-income-tax
          https://www.ncdor.gov/taxes-forms/individual-income-tax
        
        Alabama:
        -         
        - https://www.revenue.alabama.gov/ultraviewer/viewer/basic_viewer/index.html?form=2023/02/22f40.pdf
          https://www.revenue.alabama.gov/faqs/what-is-alabamas-individual-income-tax-rate/
          https://www.revenue.alabama.gov/ultraviewer/viewer/basic_viewer/index.html?form=2023/02/22f40.pdf
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-24.htm?taxmap=true
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-24.htm?taxmap=true
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-24.htm?taxmap=true
          https://www.revenue.alabama.gov/individual-corporate/income-to-be-reported-on-the-alabama-income-tax-return/
          https://www.almonline.org/TaxRates.aspx#OccupationalTax
        """
        
        # Check if the parameters are set upstream in the instance
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'NorthCarolina', 'state_occupation': 'Alabama', "capital_gains_short": capital_gains_short, "income": income, "dependents": dependents, "sources": sources, "county_residence": county_residence, "state_agi": state_agi, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def north_carolina_and_alaska(self, capital_gains_long=0, state_agi=75000, filing_status='single', income=100000, sources=False, capital_gains_short=0, credits_state=0):
        """
        --- Parameters ----
        
        North Carolina:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        
        Alaska:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        North Carolina:
        - https://www.ncdor.gov/taxes-forms/tax-rate-schedules
          https://www.ncdor.gov/taxes-forms/individual-income-tax/north-carolina-standard-deduction-or-north-carolina-itemized-deductions
          https://www.ncdor.gov/taxes-forms/individual-income-tax
          https://www.ncdor.gov/taxes-forms/individual-income-tax
        
        Alaska:
        -         
        - https://www.commerce.alaska.gov/web/dcra/officeofthestateassessor/alaskataxfacts.aspx
        """
        
        # Check if the parameters are set upstream in the instance
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        sources = self.sources if hasattr(self, 'sources') else sources
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'NorthCarolina', 'state_occupation': 'Alaska', "capital_gains_long": capital_gains_long, "state_agi": state_agi, "filing_status": filing_status, "income": income, "sources": sources, "capital_gains_short": capital_gains_short, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def north_carolina_and_arizona(self, state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        North Carolina:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        
        Arizona:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        credits_state : int, optional
            Credits for state taxes. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        North Carolina:
        - https://www.ncdor.gov/taxes-forms/tax-rate-schedules
          https://www.ncdor.gov/taxes-forms/individual-income-tax/north-carolina-standard-deduction-or-north-carolina-itemized-deductions
          https://www.ncdor.gov/taxes-forms/individual-income-tax
          https://www.ncdor.gov/taxes-forms/individual-income-tax
        
        Arizona:
        -         
        - https://azdor.gov/forms/individual/form-140-x-y-tables
          https://azdor.gov/forms/individual/form-140-x-y-tables
          https://azdor.gov/forms/individual/form-140-x-y-tables
          https://azdor.gov/businesses-arizona/withholding-tax/withholding-exceptions
          "https://azdor.gov/individuals/income-tax-filing-assistance/identifying-other-taxable-income#:~:text=In%20regards%20to%20capital%20gains
         the%20individuals%20regular%20tax%20rate.
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'NorthCarolina', 'state_occupation': 'Arizona', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def north_carolina_and_arkansas(self, capital_gains_short=0, income=100000, dependents=0, sources=False, state_agi=75000, filers_over_65=0, capital_gains_long=0, filing_status='single', credits_state=0):
        """
        --- Parameters ----
        
        North Carolina:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        
        Arkansas:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        North Carolina:
        - https://www.ncdor.gov/taxes-forms/tax-rate-schedules
          https://www.ncdor.gov/taxes-forms/individual-income-tax/north-carolina-standard-deduction-or-north-carolina-itemized-deductions
          https://www.ncdor.gov/taxes-forms/individual-income-tax
          https://www.ncdor.gov/taxes-forms/individual-income-tax
        
        Arkansas:
        -         
        - https://www.arkansasedc.com/why-arkansas/business-climate/tax-structure/personal-income-tax
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_and_AR1000NR_Instructions.pdf
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_FullYearResidentIndividualIncomeTaxReturn.pdf
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_and_AR1000NR_Instructions.pdf
          https://www.arkansasedc.com/why-arkansas/business-climate/tax-structure/capital-gains-tax-reduction
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_and_AR1000NR_Instructions.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'NorthCarolina', 'state_occupation': 'Arkansas', "capital_gains_short": capital_gains_short, "income": income, "dependents": dependents, "sources": sources, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def north_carolina_and_california(self, state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', sources=False, credits_state=0):
        """
        --- Parameters ----
        
        North Carolina:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        
        California:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        North Carolina:
        - https://www.ncdor.gov/taxes-forms/tax-rate-schedules
          https://www.ncdor.gov/taxes-forms/individual-income-tax/north-carolina-standard-deduction-or-north-carolina-itemized-deductions
          https://www.ncdor.gov/taxes-forms/individual-income-tax
          https://www.ncdor.gov/taxes-forms/individual-income-tax
        
        California:
        -         
        - https://www.ftb.ca.gov/forms/2022/2022-540-tax-rate-schedules.pdf
          https://www.ftb.ca.gov/file/personal/deductions/index.html
          https://www.ftb.ca.gov/about-ftb/newsroom/caleitc/eligibility-and-credit-information.html
          https://www.ftb.ca.gov/file/personal/income-types/capital-gains-and-losses.html#:~:text=California%20does%20not%20have%20a
         are%20taxed%20as%20ordinary%20income
          https://www.ftb.ca.gov/file/personal/
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'NorthCarolina', 'state_occupation': 'California', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def north_carolina_and_colorado(self, county_residence='Other', state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', sources=False, credits_state=0):
        """
        --- Parameters ----
        
        North Carolina:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        
        Colorado:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        North Carolina:
        - https://www.ncdor.gov/taxes-forms/tax-rate-schedules
          https://www.ncdor.gov/taxes-forms/individual-income-tax/north-carolina-standard-deduction-or-north-carolina-itemized-deductions
          https://www.ncdor.gov/taxes-forms/individual-income-tax
          https://www.ncdor.gov/taxes-forms/individual-income-tax
        
        Colorado:
        -         
        - https://leg.colorado.gov/agencies/legislative-council-staff/individual-income-tax%C2%A0
          https://leg.colorado.gov/agencies/legislative-council-staff/individual-income-tax%C2%A0
          https://tax.colorado.gov/sites/tax/files/documents/ITT_Capital_Gain_Subtraction_Feb_2022.pdf
          https://www.auroragov.org/business_services/taxes/occupational_privilege_tax
          https://www.denvergov.org/Government/Agencies-Departments-Offices/Agencies-Departments-Offices-Directory/Department-of-Finance/Our-Divisions/Treasury/Business-Tax-Information#:~:text=Occupational%20privilege%20tax
         -%28show%20below%29&text=Each%20taxable%20employee%20is%20liable
         month%20for%20each%20taxable%20employee
          https://www.glendale.co.us/355/Occupational-Privilege-Tax
          https://www.greenwoodvillage.com/1220/Occupational-Privilege-Tax-OPT
          https://ci.sheridan.co.us/288/Occupational-Privilege-Tax
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'NorthCarolina', 'state_occupation': 'Colorado', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def north_carolina_and_connecticut(self, state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', sources=False, credits_state=0):
        """
        --- Parameters ----
        
        North Carolina:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        
        Connecticut:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        North Carolina:
        - https://www.ncdor.gov/taxes-forms/tax-rate-schedules
          https://www.ncdor.gov/taxes-forms/individual-income-tax/north-carolina-standard-deduction-or-north-carolina-itemized-deductions
          https://www.ncdor.gov/taxes-forms/individual-income-tax
          https://www.ncdor.gov/taxes-forms/individual-income-tax
        
        Connecticut:
        -         
        - https://portal.ct.gov/-/media/DRS/Forms/2022/Income/2022-CT-1040-Instructions_1222.pdf
          https://www.cga.ct.gov/2022/rpt/pdf/2022-R-0108.pdf
          https://www.cga.ct.gov/current/pub/chap_229.htm#sec_12-703
          https://portal.ct.gov/-/media/DRS/Forms/2022/Income/2022-CT-1040-Instructions_1222.pdf
          "https://portal.ct.gov/DRS/Publications/TSSNs/TSSN-29#:~:text=An%20individuals%20net%20capital%20gains
         portion%20of%20Social%20Security%20benefits"
          "https://portal.ct.gov/DRS/Publications/TSSNs/TSSN-29#:~:text=An%20individuals%20net%20capital%20gains
         portion%20of%20Social%20Security%20benefits"
          https://portal.ct.gov/DRS/Individuals/Resident-Income-Tax/Tax-Information
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'NorthCarolina', 'state_occupation': 'Connecticut', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def north_carolina_and_delaware(self, state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, federal_agi=75000, filing_status='single', sources=False, credits_state=0):
        """
        --- Parameters ----
        
        North Carolina:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        
        Delaware:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        federal_agi : int, optional
            Adjusted Gross Income (AGI) for federal taxes. Default is 0. Sources - https://www.irs.gov/e-file-providers/definition-of-adjusted-gross-income, https://www.irs.gov/newsroom/taxpayers-should-know-the-difference-between-standard-and-itemized-deductions
        income : int, optional
            Gross income. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        North Carolina:
        - https://www.ncdor.gov/taxes-forms/tax-rate-schedules
          https://www.ncdor.gov/taxes-forms/individual-income-tax/north-carolina-standard-deduction-or-north-carolina-itemized-deductions
          https://www.ncdor.gov/taxes-forms/individual-income-tax
          https://www.ncdor.gov/taxes-forms/individual-income-tax
        
        Delaware:
        -         
        - https://revenue.delaware.gov/software-developer/tax-rate-changes/
          https://revenuefiles.delaware.gov/2022/PIT-RES_TY22_2022-01_PaperInteractive.pdf
          https://revenuefiles.delaware.gov/2022/PIT-RES_TY22_2022-02_Instructions.pdf
          https://revenue.delaware.gov/frequently-asked-questions/personal-income-tax-faqs/
          https://www.newcastlede.gov/DocumentCenter/View/20147/City-of-Wilmington-Wage-Tax-Information-PDF
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        federal_agi = self.federal_agi if hasattr(self, 'federal_agi') else federal_agi
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'NorthCarolina', 'state_occupation': 'Delaware', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "federal_agi": federal_agi, "filing_status": filing_status, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def north_carolina_and_florida(self, capital_gains_long=0, state_agi=75000, filing_status='single', income=100000, sources=False, capital_gains_short=0, credits_state=0):
        """
        --- Parameters ----
        
        North Carolina:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        
        Florida:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        North Carolina:
        - https://www.ncdor.gov/taxes-forms/tax-rate-schedules
          https://www.ncdor.gov/taxes-forms/individual-income-tax/north-carolina-standard-deduction-or-north-carolina-itemized-deductions
          https://www.ncdor.gov/taxes-forms/individual-income-tax
          https://www.ncdor.gov/taxes-forms/individual-income-tax
        
        Florida:
        -         
        - https://floridarevenue.com/taxes/taxesfees/Pages/tax_interest_rates.aspx
        """
        
        # Check if the parameters are set upstream in the instance
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        sources = self.sources if hasattr(self, 'sources') else sources
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'NorthCarolina', 'state_occupation': 'Florida', "capital_gains_long": capital_gains_long, "state_agi": state_agi, "filing_status": filing_status, "income": income, "sources": sources, "capital_gains_short": capital_gains_short, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def north_carolina_and_georgia(self, capital_gains_short=0, income=100000, dependents=0, sources=False, state_agi=75000, filers_over_65=0, capital_gains_long=0, filing_status='single', credits_state=0):
        """
        --- Parameters ----
        
        North Carolina:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        
        Georgia:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        North Carolina:
        - https://www.ncdor.gov/taxes-forms/tax-rate-schedules
          https://www.ncdor.gov/taxes-forms/individual-income-tax/north-carolina-standard-deduction-or-north-carolina-itemized-deductions
          https://www.ncdor.gov/taxes-forms/individual-income-tax
          https://www.ncdor.gov/taxes-forms/individual-income-tax
        
        Georgia:
        -         
        - https://dor.georgia.gov/tax-tables-georgia-tax-rate-schedule
          https://apps.dor.ga.gov/FillableForms/PDFViewer/Index?form=2022GA500
          https://apps.dor.ga.gov/FillableForms/PDFViewer/Index?form=2022GA500
          https://apps.dor.ga.gov/FillableForms/PDFViewer/Index?form=2022GA500
          https://dor.georgia.gov/years-individual-income-tax-forms
          https://dor.georgia.gov/years-individual-income-tax-forms
        """
        
        # Check if the parameters are set upstream in the instance
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'NorthCarolina', 'state_occupation': 'Georgia', "capital_gains_short": capital_gains_short, "income": income, "dependents": dependents, "sources": sources, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def north_carolina_and_hawaii(self, capital_gains_short=0, income=100000, dependents=0, sources=False, state_agi=75000, filers_over_65=0, capital_gains_long=0, filing_status='single', credits_state=0):
        """
        --- Parameters ----
        
        North Carolina:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        
        Hawaii:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        North Carolina:
        - https://www.ncdor.gov/taxes-forms/tax-rate-schedules
          https://www.ncdor.gov/taxes-forms/individual-income-tax/north-carolina-standard-deduction-or-north-carolina-itemized-deductions
          https://www.ncdor.gov/taxes-forms/individual-income-tax
          https://www.ncdor.gov/taxes-forms/individual-income-tax
        
        Hawaii:
        -         
        - https://tax.hawaii.gov/forms/d_18table-on/d_18table-on_p13/
          https://files.hawaii.gov/tax/forms/2018/18table-on.pdf
          https://files.hawaii.gov/tax/news/pubs/22outline.pdf
          https://files.hawaii.gov/tax/news/pubs/FastTaxRefGuide_2017.pdf
          https://www.tfhawaii.org/wordpress/state-tax-resources/mini-tax-guides/individual-income-tax-chapter-235/#:~:text=Long%20term%20capital%20gains%20are%20taxed%20at%20a%20maximum%20of%207.25%25
          https://www.grassrootinstitute.org/2023/02/hb337-increasing-tax-on-capital-gains-a-bad-risk/
          https://files.hawaii.gov/tax/news/pubs/22outline.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'NorthCarolina', 'state_occupation': 'Hawaii', "capital_gains_short": capital_gains_short, "income": income, "dependents": dependents, "sources": sources, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def north_carolina_and_idaho(self, capital_gains_short=0, income=100000, dependents=0, sources=False, state_agi=75000, filers_over_65=0, capital_gains_long=0, filing_status='single', credits_state=0):
        """
        --- Parameters ----
        
        North Carolina:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        
        Idaho:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        North Carolina:
        - https://www.ncdor.gov/taxes-forms/tax-rate-schedules
          https://www.ncdor.gov/taxes-forms/individual-income-tax/north-carolina-standard-deduction-or-north-carolina-itemized-deductions
          https://www.ncdor.gov/taxes-forms/individual-income-tax
          https://www.ncdor.gov/taxes-forms/individual-income-tax
        
        Idaho:
        -         
        - https://tax.idaho.gov/taxes/income-tax/individual-income/individual-income-tax-rate-schedule/
          https://tax.idaho.gov/wp-content/uploads/forms/EIN00046/EIN00046_03-01-2023.pdf
          https://tax.idaho.gov/taxes/income-tax/individual-income/idaho-source-income/
          https://tax.idaho.gov/taxes/income-tax/individual-income/idaho-source-income/
        """
        
        # Check if the parameters are set upstream in the instance
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'NorthCarolina', 'state_occupation': 'Idaho', "capital_gains_short": capital_gains_short, "income": income, "dependents": dependents, "sources": sources, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def north_carolina_and_illinois(self, state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        North Carolina:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        
        Illinois:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        North Carolina:
        - https://www.ncdor.gov/taxes-forms/tax-rate-schedules
          https://www.ncdor.gov/taxes-forms/individual-income-tax/north-carolina-standard-deduction-or-north-carolina-itemized-deductions
          https://www.ncdor.gov/taxes-forms/individual-income-tax
          https://www.ncdor.gov/taxes-forms/individual-income-tax
        
        Illinois:
        -         
        - https://tax.illinois.gov/research/taxrates/income.html#IndividualIncome
          https://tax.illinois.gov/content/dam/soi/en/web/tax/forms/incometax/documents/currentyear/individual/il-1040-instr.pdf
          https://tax.illinois.gov/questionsandanswers/12.html
          https://tax.illinois.gov/content/dam/soi/en/web/tax/forms/incometax/documents/currentyear/individual/il-1040-instr.pdf
          https://tax.illinois.gov/research/taxrates/personalproperty.html
          https://tax.illinois.gov/localgovernments/income.html
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'NorthCarolina', 'state_occupation': 'Illinois', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def north_carolina_and_indiana(self, capital_gains_short=0, income=100000, dependents=0, sources=False, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_long=0, filing_status='single', credits_state=0):
        """
        --- Parameters ----
        
        North Carolina:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        
        Indiana:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        North Carolina:
        - https://www.ncdor.gov/taxes-forms/tax-rate-schedules
          https://www.ncdor.gov/taxes-forms/individual-income-tax/north-carolina-standard-deduction-or-north-carolina-itemized-deductions
          https://www.ncdor.gov/taxes-forms/individual-income-tax
          https://www.ncdor.gov/taxes-forms/individual-income-tax
        
        Indiana:
        -         
        - https://www.in.gov/dor/tax-forms/2022-individual-income-tax-forms/
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-32.htm
          https://www.in.gov/dor/files/ib28.pdf
          https://www.in.gov/dor/tax-forms/2022-individual-income-tax-forms/
          https://www.in.gov/dor/files/2022-county-tax-rates-and-codes.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'NorthCarolina', 'state_occupation': 'Indiana', "capital_gains_short": capital_gains_short, "income": income, "dependents": dependents, "sources": sources, "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def north_carolina_and_iowa(self, capital_gains_short=0, income=100000, dependents=0, sources=False, county_residence='Other', state_agi=75000, capital_gains_long=0, filing_status='single', credits_state=0):
        """
        --- Parameters ----
        
        North Carolina:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        
        Iowa:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        North Carolina:
        - https://www.ncdor.gov/taxes-forms/tax-rate-schedules
          https://www.ncdor.gov/taxes-forms/individual-income-tax/north-carolina-standard-deduction-or-north-carolina-itemized-deductions
          https://www.ncdor.gov/taxes-forms/individual-income-tax
          https://www.ncdor.gov/taxes-forms/individual-income-tax
        
        Iowa:
        -         
        - https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
          https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
          https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
          https://tax.iowa.gov/iowa-illinois-reciprocal-agreement
          https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
        """
        
        # Check if the parameters are set upstream in the instance
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'NorthCarolina', 'state_occupation': 'Iowa', "capital_gains_short": capital_gains_short, "income": income, "dependents": dependents, "sources": sources, "county_residence": county_residence, "state_agi": state_agi, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def north_carolina_and_kansas(self, capital_gains_short=0, income=100000, dependents=0, sources=False, state_agi=75000, filers_over_65=0, capital_gains_long=0, filing_status='single', credits_state=0):
        """
        --- Parameters ----
        
        North Carolina:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        
        Kansas:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        North Carolina:
        - https://www.ncdor.gov/taxes-forms/tax-rate-schedules
          https://www.ncdor.gov/taxes-forms/individual-income-tax/north-carolina-standard-deduction-or-north-carolina-itemized-deductions
          https://www.ncdor.gov/taxes-forms/individual-income-tax
          https://www.ncdor.gov/taxes-forms/individual-income-tax
        
        Kansas:
        -         
        - https://www.ksrevenue.gov/taxrates.html
          https://www.ksrevenue.gov/incomebook22.html#0
          https://www.ksrevenue.gov/incomebook22.html#0
          https://www.ksrevenue.gov/incomebook22.html#0
          https://www.ksrevenue.gov/incomebook22.html
          https://www.ksrevenue.gov/incomebook22.html
        """
        
        # Check if the parameters are set upstream in the instance
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'NorthCarolina', 'state_occupation': 'Kansas', "capital_gains_short": capital_gains_short, "income": income, "dependents": dependents, "sources": sources, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def north_carolina_and_kentucky(self, county_residence='Other', state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', sources=False, credits_state=0):
        """
        --- Parameters ----
        
        North Carolina:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        
        Kentucky:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        North Carolina:
        - https://www.ncdor.gov/taxes-forms/tax-rate-schedules
          https://www.ncdor.gov/taxes-forms/individual-income-tax/north-carolina-standard-deduction-or-north-carolina-itemized-deductions
          https://www.ncdor.gov/taxes-forms/individual-income-tax
          https://www.ncdor.gov/taxes-forms/individual-income-tax
        
        Kentucky:
        -         
        - https://revenue.ky.gov/News/Pages/Dor-Announces-Updates-to-Individual-Income-Tax-for-2023-Tax-Year.aspx
          https://revenue.ky.gov/Individual/Individual-Income-Tax/Pages/default.aspx
          https://revenue.ky.gov/News/Pages/Dor-Announces-Updates-to-Individual-Income-Tax-for-2023-Tax-Year.aspx
          https://revenue.ky.gov/Individual/Individual-Income-Tax/Pages/default.aspx
          https://revenue.ky.gov/Dor%20Training%20Materials/103%20KAR%2017.140.%20Individual%20income%20tax%20-%20reciprocity%20-%20nonresidents.pdf
          https://revenue.ky.gov/Individual/Individual-Income-Tax/Pages/default.aspx
          https://nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2021/Taxes-21-32.htm?taxmap=true
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'NorthCarolina', 'state_occupation': 'Kentucky', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def north_carolina_and_louisiana(self, capital_gains_short=0, income=100000, dependents=0, sources=False, state_agi=75000, filers_over_65=0, capital_gains_long=0, filing_status='single', credits_state=0):
        """
        --- Parameters ----
        
        North Carolina:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        
        Louisiana:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        North Carolina:
        - https://www.ncdor.gov/taxes-forms/tax-rate-schedules
          https://www.ncdor.gov/taxes-forms/individual-income-tax/north-carolina-standard-deduction-or-north-carolina-itemized-deductions
          https://www.ncdor.gov/taxes-forms/individual-income-tax
          https://www.ncdor.gov/taxes-forms/individual-income-tax
        
        Louisiana:
        -         
        - https://revenue.louisiana.gov/individualincometax
          https://revenue.louisiana.gov/individualincometax
          https://www.revenue.louisiana.gov/taxforms/6935(11_02)F.pdf
          https://revenue.louisiana.gov/individualincometax
          https://revenue.louisiana.gov/individualincometax
        """
        
        # Check if the parameters are set upstream in the instance
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'NorthCarolina', 'state_occupation': 'Louisiana', "capital_gains_short": capital_gains_short, "income": income, "dependents": dependents, "sources": sources, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def north_carolina_and_maine(self, capital_gains_short=0, income=100000, dependents=0, sources=False, state_agi=75000, filers_over_65=0, capital_gains_long=0, filing_status='single', credits_state=0):
        """
        --- Parameters ----
        
        North Carolina:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        
        Maine:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        North Carolina:
        - https://www.ncdor.gov/taxes-forms/tax-rate-schedules
          https://www.ncdor.gov/taxes-forms/individual-income-tax/north-carolina-standard-deduction-or-north-carolina-itemized-deductions
          https://www.ncdor.gov/taxes-forms/individual-income-tax
          https://www.ncdor.gov/taxes-forms/individual-income-tax
        
        Maine:
        -         
        - https://www.maine.gov/revenue/tax-return-forms/individual-income-tax-2022
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/ind_tax_rate_sched_2022.pdf
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/22_1040me_book_gen_instruc_revisedFeb23.pdf
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/22_1040me_book_gen_instruc_revisedFeb23.pdf
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/22_1040me_book_gen_instruc_revisedFeb23.pdf
          https://www.maine.gov/revenue/tax-return-forms/individual-income-tax-2022
        """
        
        # Check if the parameters are set upstream in the instance
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'NorthCarolina', 'state_occupation': 'Maine', "capital_gains_short": capital_gains_short, "income": income, "dependents": dependents, "sources": sources, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def north_carolina_and_maryland(self, capital_gains_short=0, income=100000, dependents=0, sources=False, county_residence='Other', state_agi=75000, capital_gains_long=0, filing_status='single', credits_state=0):
        """
        --- Parameters ----
        
        North Carolina:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        
        Maryland:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        North Carolina:
        - https://www.ncdor.gov/taxes-forms/tax-rate-schedules
          https://www.ncdor.gov/taxes-forms/individual-income-tax/north-carolina-standard-deduction-or-north-carolina-itemized-deductions
          https://www.ncdor.gov/taxes-forms/individual-income-tax
          https://www.ncdor.gov/taxes-forms/individual-income-tax
        
        Maryland:
        -         
        - https://www.marylandtaxes.gov/individual/income/tax-info/tax-rates.php
          https://www.marylandtaxes.gov/forms/22_forms/Resident_Booklet.pdf
          https://www.marylandtaxes.gov/forms/Tax_Publications/Administrative_Releases/Income_and_Estate_Tax_Releases/ar_it37.pdf
          https://www.marylandtaxes.gov/individual/index.php
          https://www.marylandtaxes.gov/individual/income/tax-info/tax-rates.php
        """
        
        # Check if the parameters are set upstream in the instance
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'NorthCarolina', 'state_occupation': 'Maryland', "capital_gains_short": capital_gains_short, "income": income, "dependents": dependents, "sources": sources, "county_residence": county_residence, "state_agi": state_agi, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def north_carolina_and_massachusetts(self, state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        North Carolina:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        
        Massachusetts:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        North Carolina:
        - https://www.ncdor.gov/taxes-forms/tax-rate-schedules
          https://www.ncdor.gov/taxes-forms/individual-income-tax/north-carolina-standard-deduction-or-north-carolina-itemized-deductions
          https://www.ncdor.gov/taxes-forms/individual-income-tax
          https://www.ncdor.gov/taxes-forms/individual-income-tax
        
        Massachusetts:
        -         
        - https://www.mass.gov/service-details/massachusetts-tax-rates
          https://www.mass.gov/service-details/view-massachusetts-personal-income-tax-exemptions
          https://www.mass.gov/service-details/view-massachusetts-personal-income-tax-exemptions
          https://www.mass.gov/service-details/view-child-and-dependent-related-credits
          https://www.mass.gov/service-details/massachusetts-tax-rates
          https://www.mass.gov/personal-income-tax
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'NorthCarolina', 'state_occupation': 'Massachusetts', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def north_carolina_and_michigan(self, capital_gains_short=0, income=100000, county_occupation='Other', dependents=0, sources=False, county_residence='Other', state_agi=75000, capital_gains_long=0, filing_status='single', credits_state=0):
        """
        --- Parameters ----
        
        North Carolina:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        
        Michigan:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        county_occupation : str, optional
            County of occupation. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        North Carolina:
        - https://www.ncdor.gov/taxes-forms/tax-rate-schedules
          https://www.ncdor.gov/taxes-forms/individual-income-tax/north-carolina-standard-deduction-or-north-carolina-itemized-deductions
          https://www.ncdor.gov/taxes-forms/individual-income-tax
          https://www.ncdor.gov/taxes-forms/individual-income-tax
        
        Michigan:
        -         
        - https://www.michigan.gov/taxes/iit/new-developments-for-tax-year-2022
          "https://www.michigan.gov/taxes/-/media/Project/Websites/taxes/Forms/2023/2023-SUW/446_Withholding-Guide_2023.pdf?rev=177ca3d8cb034392bf84e317ed78c320&hash=19C5F42F43A89A91863D229349F48883#:~:text=The%20exemption%20amount%20is%20%245%2C400
         employees%20Michigan%20income%20tax%20return"
          "https://www.michigan.gov/taxes/-/media/Project/Websites/taxes/Forms/2023/2023-SUW/446_Withholding-Guide_2023.pdf?rev=177ca3d8cb034392bf84e317ed78c320&hash=19C5F42F43A89A91863D229349F48883#:~:text=The%20exemption%20amount%20is%20%245%2C400
         employees%20Michigan%20income%20tax%20return"
          https://www.michigan.gov/taxes/questions/iit/accordion/residency/are-my-wages-earned-in-another-state-taxable-in-michigan-if-i-am-a-michigan-resident-1
          https://www.michigan.gov/taxes/-/media/Project/Websites/taxes/Forms/2022/2022-IIT-Forms/BOOK_MI-1040.pdf?rev=4524f1cbbc0243b194ab05a6680be75b&hash=40DFE71E651F7C1D65C31F6CD03D944F
          https://www.michigan.gov/taxes/questions/iit/accordion/general/what-cities-impose-an-income-tax
        """
        
        # Check if the parameters are set upstream in the instance
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        county_occupation = self.county_occupation if hasattr(self, 'county_occupation') else county_occupation
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'NorthCarolina', 'state_occupation': 'Michigan', "capital_gains_short": capital_gains_short, "income": income, "county_occupation": county_occupation, "dependents": dependents, "sources": sources, "county_residence": county_residence, "state_agi": state_agi, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def north_carolina_and_minnesota(self, state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        North Carolina:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        
        Minnesota:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        North Carolina:
        - https://www.ncdor.gov/taxes-forms/tax-rate-schedules
          https://www.ncdor.gov/taxes-forms/individual-income-tax/north-carolina-standard-deduction-or-north-carolina-itemized-deductions
          https://www.ncdor.gov/taxes-forms/individual-income-tax
          https://www.ncdor.gov/taxes-forms/individual-income-tax
        
        Minnesota:
        -         
        - https://www.revenue.state.mn.us/minnesota-income-tax-rates-and-brackets
          https://www.revenue.state.mn.us/child-and-dependent-care-credit
          https://www.revenue.state.mn.us/dependent-exemptions
          https://www.revenue.state.mn.us/minnesota-standard-deduction
          https://mn.gov/mmb/assets/mwr_form_tcm1059-128581.pdf
          https://www.revenue.state.mn.us/individual-income-tax
          https://www.revenue.state.mn.us/individual-income-tax
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'NorthCarolina', 'state_occupation': 'Minnesota', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def north_carolina_and_mississippi(self, state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        North Carolina:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        
        Mississippi:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        North Carolina:
        - https://www.ncdor.gov/taxes-forms/tax-rate-schedules
          https://www.ncdor.gov/taxes-forms/individual-income-tax/north-carolina-standard-deduction-or-north-carolina-itemized-deductions
          https://www.ncdor.gov/taxes-forms/individual-income-tax
          https://www.ncdor.gov/taxes-forms/individual-income-tax
        
        Mississippi:
        -         
        - https://www.dor.ms.gov/individual/tax-rates
          https://www.dor.ms.gov/individual/tax-rates
          https://www.dor.ms.gov/individual/tax-rates
          https://www.dor.ms.gov/individual
          https://www.dor.ms.gov/individual
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'NorthCarolina', 'state_occupation': 'Mississippi', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def north_carolina_and_missouri(self, capital_gains_short=0, income=100000, dependents=0, sources=False, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_long=0, filing_status='single', credits_state=0):
        """
        --- Parameters ----
        
        North Carolina:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        
        Missouri:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        North Carolina:
        - https://www.ncdor.gov/taxes-forms/tax-rate-schedules
          https://www.ncdor.gov/taxes-forms/individual-income-tax/north-carolina-standard-deduction-or-north-carolina-itemized-deductions
          https://www.ncdor.gov/taxes-forms/individual-income-tax
          https://www.ncdor.gov/taxes-forms/individual-income-tax
        
        Missouri:
        -         
        - https://dor.mo.gov/taxation/individual/tax-types/income/year-changes/
          https://dor.mo.gov/taxation/individual/tax-types/income/year-changes/
          https://dor.mo.gov/taxation/individual/tax-types/income/year-changes/
          https://www.kcmo.gov/city-hall/departments/finance/earnings-tax
          https://www.stlouis-mo.gov/government/departments/collector/earnings-tax/individual-earnings-tax-info.cfm
        """
        
        # Check if the parameters are set upstream in the instance
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'NorthCarolina', 'state_occupation': 'Missouri', "capital_gains_short": capital_gains_short, "income": income, "dependents": dependents, "sources": sources, "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def north_carolina_and_montana(self, capital_gains_short=0, income=100000, dependents=0, sources=False, state_agi=75000, filers_over_65=0, capital_gains_long=0, filing_status='single', credits_state=0):
        """
        --- Parameters ----
        
        North Carolina:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        
        Montana:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        North Carolina:
        - https://www.ncdor.gov/taxes-forms/tax-rate-schedules
          https://www.ncdor.gov/taxes-forms/individual-income-tax/north-carolina-standard-deduction-or-north-carolina-itemized-deductions
          https://www.ncdor.gov/taxes-forms/individual-income-tax
          https://www.ncdor.gov/taxes-forms/individual-income-tax
        
        Montana:
        -         
        - https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/2022/12/montana-employees-withholding-allowance-and-exemption-certificate-form-mw-4.pdf
          https://mtrevenue.gov/2021/11/29/simplification-of-montana-income-taxation/?utm_source=rss&utm_medium=rss&utm_campaign=simplification-of-montana-income-taxation
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'NorthCarolina', 'state_occupation': 'Montana', "capital_gains_short": capital_gains_short, "income": income, "dependents": dependents, "sources": sources, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def north_carolina_and_nebraska(self, state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        North Carolina:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        
        Nebraska:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        North Carolina:
        - https://www.ncdor.gov/taxes-forms/tax-rate-schedules
          https://www.ncdor.gov/taxes-forms/individual-income-tax/north-carolina-standard-deduction-or-north-carolina-itemized-deductions
          https://www.ncdor.gov/taxes-forms/individual-income-tax
          https://www.ncdor.gov/taxes-forms/individual-income-tax
        
        Nebraska:
        -         
        - https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2023_Circular_EN_8-429r2023_Whole_Book_11.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'NorthCarolina', 'state_occupation': 'Nebraska', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def north_carolina_and_nevada(self, capital_gains_long=0, state_agi=75000, filing_status='single', income=100000, sources=False, capital_gains_short=0, credits_state=0):
        """
        --- Parameters ----
        
        North Carolina:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        
        Nevada:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        North Carolina:
        - https://www.ncdor.gov/taxes-forms/tax-rate-schedules
          https://www.ncdor.gov/taxes-forms/individual-income-tax/north-carolina-standard-deduction-or-north-carolina-itemized-deductions
          https://www.ncdor.gov/taxes-forms/individual-income-tax
          https://www.ncdor.gov/taxes-forms/individual-income-tax
        
        Nevada:
        -         
        - https://www.nvsos.gov/sos/businesses/the-nevada-advantage
        """
        
        # Check if the parameters are set upstream in the instance
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        sources = self.sources if hasattr(self, 'sources') else sources
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'NorthCarolina', 'state_occupation': 'Nevada', "capital_gains_long": capital_gains_long, "state_agi": state_agi, "filing_status": filing_status, "income": income, "sources": sources, "capital_gains_short": capital_gains_short, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def north_carolina_and_new_hampshire(self, capital_gains_long=0, state_agi=75000, filing_status='single', income=100000, sources=False, capital_gains_short=0, credits_state=0):
        """
        --- Parameters ----
        
        North Carolina:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        
        New Hampshire:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        North Carolina:
        - https://www.ncdor.gov/taxes-forms/tax-rate-schedules
          https://www.ncdor.gov/taxes-forms/individual-income-tax/north-carolina-standard-deduction-or-north-carolina-itemized-deductions
          https://www.ncdor.gov/taxes-forms/individual-income-tax
          https://www.ncdor.gov/taxes-forms/individual-income-tax
        
        New Hampshire:
        -         
        - https://www.revenue.nh.gov/assistance/tax-overview.htm
        """
        
        # Check if the parameters are set upstream in the instance
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        sources = self.sources if hasattr(self, 'sources') else sources
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'NorthCarolina', 'state_occupation': 'NewHampshire', "capital_gains_long": capital_gains_long, "state_agi": state_agi, "filing_status": filing_status, "income": income, "sources": sources, "capital_gains_short": capital_gains_short, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def north_carolina_and_new_jersey(self, capital_gains_short=0, income=100000, dependents=0, sources=False, state_agi=75000, filers_over_65=0, capital_gains_long=0, filing_status='single', credits_state=0):
        """
        --- Parameters ----
        
        North Carolina:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        
        New Jersey:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        North Carolina:
        - https://www.ncdor.gov/taxes-forms/tax-rate-schedules
          https://www.ncdor.gov/taxes-forms/individual-income-tax/north-carolina-standard-deduction-or-north-carolina-itemized-deductions
          https://www.ncdor.gov/taxes-forms/individual-income-tax
          https://www.ncdor.gov/taxes-forms/individual-income-tax
        
        New Jersey:
        -         
        - https://www.state.nj.us/treasury/taxation/taxtables.shtml
          https://www.state.nj.us/treasury/taxation/njit13.shtml
          https://www.nj.gov/treasury/taxation/njit25.shtml
          https://www.state.nj.us/treasury/taxation/individuals/freqqiti.shtml
          https://www.state.nj.us/treasury/taxation/localtax.shtml
        """
        
        # Check if the parameters are set upstream in the instance
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'NorthCarolina', 'state_occupation': 'NewJersey', "capital_gains_short": capital_gains_short, "income": income, "dependents": dependents, "sources": sources, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def north_carolina_and_new_mexico(self, state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        North Carolina:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        
        New Mexico:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        North Carolina:
        - https://www.ncdor.gov/taxes-forms/tax-rate-schedules
          https://www.ncdor.gov/taxes-forms/individual-income-tax/north-carolina-standard-deduction-or-north-carolina-itemized-deductions
          https://www.ncdor.gov/taxes-forms/individual-income-tax
          https://www.ncdor.gov/taxes-forms/individual-income-tax
        
        New Mexico:
        -         
        - https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/fdf3c548-8aba-4b9c-9eb4-bb564c716015/FYI-104.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/0bf943fd-652e-4400-bb1b-cd8397eb5a95/2022pit-adj.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/2f1a6781-9534-4436-b427-1557f9592099/2022pit-adj-ins.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/0bf943fd-652e-4400-bb1b-cd8397eb5a95/2022pit-adj.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/1afc56af-ea90-4d48-82e5-1f9aeb43255a/PITbook2022.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/fdf3c548-8aba-4b9c-9eb4-bb564c716015/FYI-104.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'NorthCarolina', 'state_occupation': 'NewMexico', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def north_carolina_and_new_york(self, capital_gains_short=0, income=100000, dependents=0, sources=False, county_residence='Other', state_agi=75000, capital_gains_long=0, filing_status='single', credits_state=0):
        """
        --- Parameters ----
        
        North Carolina:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        
        New York:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        North Carolina:
        - https://www.ncdor.gov/taxes-forms/tax-rate-schedules
          https://www.ncdor.gov/taxes-forms/individual-income-tax/north-carolina-standard-deduction-or-north-carolina-itemized-deductions
          https://www.ncdor.gov/taxes-forms/individual-income-tax
          https://www.ncdor.gov/taxes-forms/individual-income-tax
        
        New York:
        -         
        - https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nys-tax-rate-schedule
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nys-tax-rate-schedule
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nys-tax-rate-schedule
          https://www.tax.ny.gov/pit/
          https://www.tax.ny.gov/pit/file/nonresident-faqs.htm
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#yonkers-tax-rate-schedule
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nyc-tax-rate-schedule
        """
        
        # Check if the parameters are set upstream in the instance
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'NorthCarolina', 'state_occupation': 'NewYork', "capital_gains_short": capital_gains_short, "income": income, "dependents": dependents, "sources": sources, "county_residence": county_residence, "state_agi": state_agi, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def north_carolina_and_north_dakota(self, capital_gains_short=0, income=100000, dependents=0, sources=False, state_agi=75000, filers_over_65=0, capital_gains_long=0, filing_status='single', credits_state=0):
        """
        --- Parameters ----
        
        North Carolina:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        
        North Dakota:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        North Carolina:
        - https://www.ncdor.gov/taxes-forms/tax-rate-schedules
          https://www.ncdor.gov/taxes-forms/individual-income-tax/north-carolina-standard-deduction-or-north-carolina-itemized-deductions
          https://www.ncdor.gov/taxes-forms/individual-income-tax
          https://www.ncdor.gov/taxes-forms/individual-income-tax
        
        North Dakota:
        -         
        - https://www.tax.nd.gov/forms
          https://www.tax.nd.gov/sites/www/files/documents/forms/individual/2022-iit/2022-individual-income-tax-booklet.pdf
          https://www.tax.nd.gov/sites/www/files/documents/forms/2022-individual-income-tax-booklet.pdf
          https://www.tax.nd.gov/tax-exemptions-credits/income-tax-exemptions-credits
          https://www.tax.nd.gov/sites/www/files/documents/forms/business/ndwrfillable.pdf
          https://www.tax.nd.gov/news/tax-legislative-changes/significant-changes-law/individual-income-tax-history
          https://www.tax.nd.gov/forms
        """
        
        # Check if the parameters are set upstream in the instance
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'NorthCarolina', 'state_occupation': 'NorthDakota', "capital_gains_short": capital_gains_short, "income": income, "dependents": dependents, "sources": sources, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def north_carolina_and_ohio(self, capital_gains_short=0, income=100000, dependents=0, sources=False, county_residence='Other', state_agi=75000, capital_gains_long=0, filing_status='single', credits_state=0):
        """
        --- Parameters ----
        
        North Carolina:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        
        Ohio:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        North Carolina:
        - https://www.ncdor.gov/taxes-forms/tax-rate-schedules
          https://www.ncdor.gov/taxes-forms/individual-income-tax/north-carolina-standard-deduction-or-north-carolina-itemized-deductions
          https://www.ncdor.gov/taxes-forms/individual-income-tax
          https://www.ncdor.gov/taxes-forms/individual-income-tax
        
        Ohio:
        -         
        - https://tax.ohio.gov/individual/resources/annual-tax-rates
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://tax.ohio.gov/static/forms/employer_withholding/generic/wth_it4nr.pdf
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://www.ritaohio.com/TaxRatesTable
        """
        
        # Check if the parameters are set upstream in the instance
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'NorthCarolina', 'state_occupation': 'Ohio', "capital_gains_short": capital_gains_short, "income": income, "dependents": dependents, "sources": sources, "county_residence": county_residence, "state_agi": state_agi, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def north_carolina_and_oklahoma(self, capital_gains_short=0, income=100000, dependents=0, sources=False, state_agi=75000, filers_over_65=0, capital_gains_long=0, filing_status='single', credits_state=0):
        """
        --- Parameters ----
        
        North Carolina:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        
        Oklahoma:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        North Carolina:
        - https://www.ncdor.gov/taxes-forms/tax-rate-schedules
          https://www.ncdor.gov/taxes-forms/individual-income-tax/north-carolina-standard-deduction-or-north-carolina-itemized-deductions
          https://www.ncdor.gov/taxes-forms/individual-income-tax
          https://www.ncdor.gov/taxes-forms/individual-income-tax
        
        Oklahoma:
        -         
        - https://oklahoma.gov/content/dam/ok/en/tax/documents/resources/publications/businesses/withholding-tables/WHTables-2022.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'NorthCarolina', 'state_occupation': 'Oklahoma', "capital_gains_short": capital_gains_short, "income": income, "dependents": dependents, "sources": sources, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def north_carolina_and_oregon(self, capital_gains_short=0, income=100000, dependents=0, sources=False, state_agi=75000, filers_over_65=0, capital_gains_long=0, filing_status='single', credits_state=0):
        """
        --- Parameters ----
        
        North Carolina:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        
        Oregon:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        North Carolina:
        - https://www.ncdor.gov/taxes-forms/tax-rate-schedules
          https://www.ncdor.gov/taxes-forms/individual-income-tax/north-carolina-standard-deduction-or-north-carolina-itemized-deductions
          https://www.ncdor.gov/taxes-forms/individual-income-tax
          https://www.ncdor.gov/taxes-forms/individual-income-tax
        
        Oregon:
        -         
        - https://www.oregon.gov/dor/programs/individuals/Pages/PIT.aspx
          https://www.oregon.gov/dor/programs/individuals/Documents/Full%20year%20resident
         %20Form%20or-40%20filers.pdf
          https://www.oregon.gov/dor/programs/individuals/Documents/Part-year%20and%20nonresident
         %20Form%20or-40-P%20and%20or-40-N%20filers.pdf.pdf
          https://www.oregon.gov/dor/forms/FormsPubs/form-or-40_101-040_2022.pdf
          https://www.oregon.gov/dor/forms/FormsPubs/form-or-40_101-040_2022.pdf
          https://www.oregon.gov/dor/forms/FormsPubs/form-or-40_101-040_2022.pdf
          https://www.portland.gov/revenue/personal-tax
        """
        
        # Check if the parameters are set upstream in the instance
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'NorthCarolina', 'state_occupation': 'Oregon', "capital_gains_short": capital_gains_short, "income": income, "dependents": dependents, "sources": sources, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def north_carolina_and_pennsylvania(self, county_residence='Other', state_agi=75000, county_occupation='Other', capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', sources=False, credits_state=0):
        """
        --- Parameters ----
        
        North Carolina:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        
        Pennsylvania:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        county_residence : str, optional
            County of residence. Default is Other.
        county_occupation : str, optional
            County of occupation. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        North Carolina:
        - https://www.ncdor.gov/taxes-forms/tax-rate-schedules
          https://www.ncdor.gov/taxes-forms/individual-income-tax/north-carolina-standard-deduction-or-north-carolina-itemized-deductions
          https://www.ncdor.gov/taxes-forms/individual-income-tax
          https://www.ncdor.gov/taxes-forms/individual-income-tax
        
        Pennsylvania:
        -         
        - https://www.revenue.pa.gov/Tax%20Rates/Pages/default.aspx
          https://www.revenue.pa.gov/FormsandPublications/FormsforIndividuals/PIT/Documents/2022/2022_pa-40in.pdf
          https://www.revenue.pa.gov/FormsandPublications/FormsforBusinesses/EmployerWithholding/Documents/rev-419.pdf
          https://www.revenue.pa.gov/FormsandPublications/FormsforIndividuals/PIT/Documents/2022/2022_pa-40in.pdf
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-25.htm
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        county_occupation = self.county_occupation if hasattr(self, 'county_occupation') else county_occupation
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'NorthCarolina', 'state_occupation': 'Pennsylvania', "county_residence": county_residence, "state_agi": state_agi, "county_occupation": county_occupation, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def north_carolina_and_rhode_island(self, state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        North Carolina:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        
        Rhode Island:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        North Carolina:
        - https://www.ncdor.gov/taxes-forms/tax-rate-schedules
          https://www.ncdor.gov/taxes-forms/individual-income-tax/north-carolina-standard-deduction-or-north-carolina-itemized-deductions
          https://www.ncdor.gov/taxes-forms/individual-income-tax
          https://www.ncdor.gov/taxes-forms/individual-income-tax
        
        Rhode Island:
        -         
        - https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2022_1040WE_w_0.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'NorthCarolina', 'state_occupation': 'RhodeIsland', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def north_carolina_and_south_carolina(self, capital_gains_short=0, income=100000, dependents=0, sources=False, state_agi=75000, filers_over_65=0, capital_gains_long=0, filing_status='single', credits_state=0):
        """
        --- Parameters ----
        
        North Carolina:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        
        South Carolina:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        North Carolina:
        - https://www.ncdor.gov/taxes-forms/tax-rate-schedules
          https://www.ncdor.gov/taxes-forms/individual-income-tax/north-carolina-standard-deduction-or-north-carolina-itemized-deductions
          https://www.ncdor.gov/taxes-forms/individual-income-tax
          https://www.ncdor.gov/taxes-forms/individual-income-tax
        
        South Carolina:
        -         
        - https://dor.sc.gov/resources-site/lawandpolicy/Advisory%20Opinions/IL22-15.pdf
          https://dor.sc.gov/forms-site/Forms/IITPacket_2022.pdf
          https://www.scstatehouse.gov/code/t12c006.php
          https://dor.sc.gov/resources-site/lawandpolicy/Advisory%20Opinions/IL22-15.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'NorthCarolina', 'state_occupation': 'SouthCarolina', "capital_gains_short": capital_gains_short, "income": income, "dependents": dependents, "sources": sources, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def north_carolina_and_south_dakota(self, capital_gains_long=0, state_agi=75000, filing_status='single', income=100000, sources=False, capital_gains_short=0, credits_state=0):
        """
        --- Parameters ----
        
        North Carolina:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        
        South Dakota:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        North Carolina:
        - https://www.ncdor.gov/taxes-forms/tax-rate-schedules
          https://www.ncdor.gov/taxes-forms/individual-income-tax/north-carolina-standard-deduction-or-north-carolina-itemized-deductions
          https://www.ncdor.gov/taxes-forms/individual-income-tax
          https://www.ncdor.gov/taxes-forms/individual-income-tax
        
        South Dakota:
        -         
        - https://dor.sd.gov/individuals/taxes/
        """
        
        # Check if the parameters are set upstream in the instance
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        sources = self.sources if hasattr(self, 'sources') else sources
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'NorthCarolina', 'state_occupation': 'SouthDakota', "capital_gains_long": capital_gains_long, "state_agi": state_agi, "filing_status": filing_status, "income": income, "sources": sources, "capital_gains_short": capital_gains_short, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def north_carolina_and_tennessee(self, capital_gains_long=0, state_agi=75000, filing_status='single', income=100000, sources=False, capital_gains_short=0, credits_state=0):
        """
        --- Parameters ----
        
        North Carolina:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        
        Tennessee:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        North Carolina:
        - https://www.ncdor.gov/taxes-forms/tax-rate-schedules
          https://www.ncdor.gov/taxes-forms/individual-income-tax/north-carolina-standard-deduction-or-north-carolina-itemized-deductions
          https://www.ncdor.gov/taxes-forms/individual-income-tax
          https://www.ncdor.gov/taxes-forms/individual-income-tax
        
        Tennessee:
        -         
        - https://www.tn.gov/revenue/taxes.html
        """
        
        # Check if the parameters are set upstream in the instance
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        sources = self.sources if hasattr(self, 'sources') else sources
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'NorthCarolina', 'state_occupation': 'Tennessee', "capital_gains_long": capital_gains_long, "state_agi": state_agi, "filing_status": filing_status, "income": income, "sources": sources, "capital_gains_short": capital_gains_short, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def north_carolina_and_texas(self, capital_gains_long=0, state_agi=75000, filing_status='single', income=100000, sources=False, capital_gains_short=0, credits_state=0):
        """
        --- Parameters ----
        
        North Carolina:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        
        Texas:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        North Carolina:
        - https://www.ncdor.gov/taxes-forms/tax-rate-schedules
          https://www.ncdor.gov/taxes-forms/individual-income-tax/north-carolina-standard-deduction-or-north-carolina-itemized-deductions
          https://www.ncdor.gov/taxes-forms/individual-income-tax
          https://www.ncdor.gov/taxes-forms/individual-income-tax
        
        Texas:
        -         
        - https://comptroller.texas.gov/taxes/
        """
        
        # Check if the parameters are set upstream in the instance
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        sources = self.sources if hasattr(self, 'sources') else sources
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'NorthCarolina', 'state_occupation': 'Texas', "capital_gains_long": capital_gains_long, "state_agi": state_agi, "filing_status": filing_status, "income": income, "sources": sources, "capital_gains_short": capital_gains_short, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def north_carolina_and_utah(self, state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        North Carolina:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        
        Utah:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        North Carolina:
        - https://www.ncdor.gov/taxes-forms/tax-rate-schedules
          https://www.ncdor.gov/taxes-forms/individual-income-tax/north-carolina-standard-deduction-or-north-carolina-itemized-deductions
          https://www.ncdor.gov/taxes-forms/individual-income-tax
          https://www.ncdor.gov/taxes-forms/individual-income-tax
        
        Utah:
        -         
        - https://incometax.utah.gov/paying/tax-rates
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'NorthCarolina', 'state_occupation': 'Utah', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def north_carolina_and_vermont(self, capital_gains_short=0, income=100000, dependents=0, sources=False, state_agi=75000, filers_over_65=0, capital_gains_long=0, filing_status='single', credits_state=0):
        """
        --- Parameters ----
        
        North Carolina:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        
        Vermont:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        North Carolina:
        - https://www.ncdor.gov/taxes-forms/tax-rate-schedules
          https://www.ncdor.gov/taxes-forms/individual-income-tax/north-carolina-standard-deduction-or-north-carolina-itemized-deductions
          https://www.ncdor.gov/taxes-forms/individual-income-tax
          https://www.ncdor.gov/taxes-forms/individual-income-tax
        
        Vermont:
        -         
        - https://tax.vermont.gov/sites/tax/files/documents/GB-1210-2023.pdf
          https://tax.vermont.gov/sites/tax/files/documents/Income%20Booklet-2022.pdf
          https://tax.vermont.gov/sites/tax/files/documents/Income%20Booklet-2022.pdf
          https://tax.vermont.gov/sites/tax/files/documents/Income%20Booklet-2022.pdf
          https://tax.vermont.gov/individuals/personal-income-tax/tax-credits
          https://tax.vermont.gov/individuals/personal-income-tax/taxable-income
          https://tax.vermont.gov/individuals/personal-income-tax/local-option-tax
        """
        
        # Check if the parameters are set upstream in the instance
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'NorthCarolina', 'state_occupation': 'Vermont', "capital_gains_short": capital_gains_short, "income": income, "dependents": dependents, "sources": sources, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def north_carolina_and_virginia(self, capital_gains_short=0, income=100000, dependents=0, sources=False, state_agi=75000, filers_over_65=0, capital_gains_long=0, filing_status='single', credits_state=0):
        """
        --- Parameters ----
        
        North Carolina:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        
        Virginia:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        North Carolina:
        - https://www.ncdor.gov/taxes-forms/tax-rate-schedules
          https://www.ncdor.gov/taxes-forms/individual-income-tax/north-carolina-standard-deduction-or-north-carolina-itemized-deductions
          https://www.ncdor.gov/taxes-forms/individual-income-tax
          https://www.ncdor.gov/taxes-forms/individual-income-tax
        
        Virginia:
        -         
        - https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/reciprocity
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'NorthCarolina', 'state_occupation': 'Virginia', "capital_gains_short": capital_gains_short, "income": income, "dependents": dependents, "sources": sources, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def north_carolina_and_washington(self, state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', sources=False, credits_state=0):
        """
        --- Parameters ----
        
        North Carolina:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        
        Washington:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        North Carolina:
        - https://www.ncdor.gov/taxes-forms/tax-rate-schedules
          https://www.ncdor.gov/taxes-forms/individual-income-tax/north-carolina-standard-deduction-or-north-carolina-itemized-deductions
          https://www.ncdor.gov/taxes-forms/individual-income-tax
          https://www.ncdor.gov/taxes-forms/individual-income-tax
        
        Washington:
        -         
        - https://dor.wa.gov/taxes-rates/income-tax
          https://dor.wa.gov/taxes-rates/other-taxes/capital-gains-tax
          https://dor.wa.gov/taxes-rates/income-tax
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'NorthCarolina', 'state_occupation': 'Washington', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def north_carolina_and_west_virginia(self, pay_periods=24, capital_gains_short=0, income=100000, county_occupation='Other', dependents=0, sources=False, state_agi=75000, capital_gains_long=0, filing_status='single', credits_state=0):
        """
        --- Parameters ----
        
        North Carolina:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        
        West Virginia:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_occupation : str, optional
            County of occupation. Default is Other.
        pay_periods : int, optional
            Number of pay periods. Options are 13, 24, 26, 52. Default is 24.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        North Carolina:
        - https://www.ncdor.gov/taxes-forms/tax-rate-schedules
          https://www.ncdor.gov/taxes-forms/individual-income-tax/north-carolina-standard-deduction-or-north-carolina-itemized-deductions
          https://www.ncdor.gov/taxes-forms/individual-income-tax
          https://www.ncdor.gov/taxes-forms/individual-income-tax
        
        West Virginia:
        -         
        - https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://tax.wv.gov/Documents/TSD/tsd438.pdf
          https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2011/Taxes-11-35att.html
        """
        
        # Check if the parameters are set upstream in the instance
        pay_periods = self.pay_periods if hasattr(self, 'pay_periods') else pay_periods
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        county_occupation = self.county_occupation if hasattr(self, 'county_occupation') else county_occupation
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'NorthCarolina', 'state_occupation': 'WestVirginia', "pay_periods": pay_periods, "capital_gains_short": capital_gains_short, "income": income, "county_occupation": county_occupation, "dependents": dependents, "sources": sources, "state_agi": state_agi, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def north_carolina_and_wisconsin(self, state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        North Carolina:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        
        Wisconsin:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        North Carolina:
        - https://www.ncdor.gov/taxes-forms/tax-rate-schedules
          https://www.ncdor.gov/taxes-forms/individual-income-tax/north-carolina-standard-deduction-or-north-carolina-itemized-deductions
          https://www.ncdor.gov/taxes-forms/individual-income-tax
          https://www.ncdor.gov/taxes-forms/individual-income-tax
        
        Wisconsin:
        -         
        - https://www.revenue.wi.gov/TaxForms2023/2023-Form1-ES-Inst.pdf
          https://www.revenue.wi.gov/TaxForms2023/2023-Form1-ES-Inst.pdf
          https://www.revenue.wi.gov/TaxForms2023/2023-Form1-ES-Inst.pdf
          https://www.revenue.wi.gov/Pages/FAQS/pcs-work.aspx
          https://www.revenue.wi.gov/Dor%20Publications/pb103.pdf
          https://www.revenue.wi.gov/Pages/FAQS/pcs-work.aspx
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'NorthCarolina', 'state_occupation': 'Wisconsin', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def north_carolina_and_wyoming(self, capital_gains_long=0, state_agi=75000, filing_status='single', income=100000, sources=False, capital_gains_short=0, credits_state=0):
        """
        --- Parameters ----
        
        North Carolina:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        
        Wyoming:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        North Carolina:
        - https://www.ncdor.gov/taxes-forms/tax-rate-schedules
          https://www.ncdor.gov/taxes-forms/individual-income-tax/north-carolina-standard-deduction-or-north-carolina-itemized-deductions
          https://www.ncdor.gov/taxes-forms/individual-income-tax
          https://www.ncdor.gov/taxes-forms/individual-income-tax
        
        Wyoming:
        -         
        - https://revenue.wyo.gov/
        """
        
        # Check if the parameters are set upstream in the instance
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        sources = self.sources if hasattr(self, 'sources') else sources
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'NorthCarolina', 'state_occupation': 'Wyoming', "capital_gains_long": capital_gains_long, "state_agi": state_agi, "filing_status": filing_status, "income": income, "sources": sources, "capital_gains_short": capital_gains_short, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def north_dakota_and_alabama(self, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        North Dakota:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Alabama:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        North Dakota:
        - https://www.tax.nd.gov/forms
          https://www.tax.nd.gov/sites/www/files/documents/forms/individual/2022-iit/2022-individual-income-tax-booklet.pdf
          https://www.tax.nd.gov/sites/www/files/documents/forms/2022-individual-income-tax-booklet.pdf
          https://www.tax.nd.gov/tax-exemptions-credits/income-tax-exemptions-credits
          https://www.tax.nd.gov/sites/www/files/documents/forms/business/ndwrfillable.pdf
          https://www.tax.nd.gov/news/tax-legislative-changes/significant-changes-law/individual-income-tax-history
          https://www.tax.nd.gov/forms
        
        Alabama:
        -         
        - https://www.revenue.alabama.gov/ultraviewer/viewer/basic_viewer/index.html?form=2023/02/22f40.pdf
          https://www.revenue.alabama.gov/faqs/what-is-alabamas-individual-income-tax-rate/
          https://www.revenue.alabama.gov/ultraviewer/viewer/basic_viewer/index.html?form=2023/02/22f40.pdf
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-24.htm?taxmap=true
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-24.htm?taxmap=true
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-24.htm?taxmap=true
          https://www.revenue.alabama.gov/individual-corporate/income-to-be-reported-on-the-alabama-income-tax-return/
          https://www.almonline.org/TaxRates.aspx#OccupationalTax
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'NorthDakota', 'state_occupation': 'Alabama', "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def north_dakota_and_alaska(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        North Dakota:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Alaska:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        North Dakota:
        - https://www.tax.nd.gov/forms
          https://www.tax.nd.gov/sites/www/files/documents/forms/individual/2022-iit/2022-individual-income-tax-booklet.pdf
          https://www.tax.nd.gov/sites/www/files/documents/forms/2022-individual-income-tax-booklet.pdf
          https://www.tax.nd.gov/tax-exemptions-credits/income-tax-exemptions-credits
          https://www.tax.nd.gov/sites/www/files/documents/forms/business/ndwrfillable.pdf
          https://www.tax.nd.gov/news/tax-legislative-changes/significant-changes-law/individual-income-tax-history
          https://www.tax.nd.gov/forms
        
        Alaska:
        -         
        - https://www.commerce.alaska.gov/web/dcra/officeofthestateassessor/alaskataxfacts.aspx
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'NorthDakota', 'state_occupation': 'Alaska', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def north_dakota_and_arizona(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        North Dakota:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Arizona:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        credits_state : int, optional
            Credits for state taxes. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        North Dakota:
        - https://www.tax.nd.gov/forms
          https://www.tax.nd.gov/sites/www/files/documents/forms/individual/2022-iit/2022-individual-income-tax-booklet.pdf
          https://www.tax.nd.gov/sites/www/files/documents/forms/2022-individual-income-tax-booklet.pdf
          https://www.tax.nd.gov/tax-exemptions-credits/income-tax-exemptions-credits
          https://www.tax.nd.gov/sites/www/files/documents/forms/business/ndwrfillable.pdf
          https://www.tax.nd.gov/news/tax-legislative-changes/significant-changes-law/individual-income-tax-history
          https://www.tax.nd.gov/forms
        
        Arizona:
        -         
        - https://azdor.gov/forms/individual/form-140-x-y-tables
          https://azdor.gov/forms/individual/form-140-x-y-tables
          https://azdor.gov/forms/individual/form-140-x-y-tables
          https://azdor.gov/businesses-arizona/withholding-tax/withholding-exceptions
          "https://azdor.gov/individuals/income-tax-filing-assistance/identifying-other-taxable-income#:~:text=In%20regards%20to%20capital%20gains
         the%20individuals%20regular%20tax%20rate.
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'NorthDakota', 'state_occupation': 'Arizona', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def north_dakota_and_arkansas(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        North Dakota:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Arkansas:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        North Dakota:
        - https://www.tax.nd.gov/forms
          https://www.tax.nd.gov/sites/www/files/documents/forms/individual/2022-iit/2022-individual-income-tax-booklet.pdf
          https://www.tax.nd.gov/sites/www/files/documents/forms/2022-individual-income-tax-booklet.pdf
          https://www.tax.nd.gov/tax-exemptions-credits/income-tax-exemptions-credits
          https://www.tax.nd.gov/sites/www/files/documents/forms/business/ndwrfillable.pdf
          https://www.tax.nd.gov/news/tax-legislative-changes/significant-changes-law/individual-income-tax-history
          https://www.tax.nd.gov/forms
        
        Arkansas:
        -         
        - https://www.arkansasedc.com/why-arkansas/business-climate/tax-structure/personal-income-tax
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_and_AR1000NR_Instructions.pdf
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_FullYearResidentIndividualIncomeTaxReturn.pdf
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_and_AR1000NR_Instructions.pdf
          https://www.arkansasedc.com/why-arkansas/business-climate/tax-structure/capital-gains-tax-reduction
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_and_AR1000NR_Instructions.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'NorthDakota', 'state_occupation': 'Arkansas', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def north_dakota_and_california(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        North Dakota:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        California:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        North Dakota:
        - https://www.tax.nd.gov/forms
          https://www.tax.nd.gov/sites/www/files/documents/forms/individual/2022-iit/2022-individual-income-tax-booklet.pdf
          https://www.tax.nd.gov/sites/www/files/documents/forms/2022-individual-income-tax-booklet.pdf
          https://www.tax.nd.gov/tax-exemptions-credits/income-tax-exemptions-credits
          https://www.tax.nd.gov/sites/www/files/documents/forms/business/ndwrfillable.pdf
          https://www.tax.nd.gov/news/tax-legislative-changes/significant-changes-law/individual-income-tax-history
          https://www.tax.nd.gov/forms
        
        California:
        -         
        - https://www.ftb.ca.gov/forms/2022/2022-540-tax-rate-schedules.pdf
          https://www.ftb.ca.gov/file/personal/deductions/index.html
          https://www.ftb.ca.gov/about-ftb/newsroom/caleitc/eligibility-and-credit-information.html
          https://www.ftb.ca.gov/file/personal/income-types/capital-gains-and-losses.html#:~:text=California%20does%20not%20have%20a
         are%20taxed%20as%20ordinary%20income
          https://www.ftb.ca.gov/file/personal/
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'NorthDakota', 'state_occupation': 'California', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def north_dakota_and_colorado(self, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        North Dakota:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Colorado:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        North Dakota:
        - https://www.tax.nd.gov/forms
          https://www.tax.nd.gov/sites/www/files/documents/forms/individual/2022-iit/2022-individual-income-tax-booklet.pdf
          https://www.tax.nd.gov/sites/www/files/documents/forms/2022-individual-income-tax-booklet.pdf
          https://www.tax.nd.gov/tax-exemptions-credits/income-tax-exemptions-credits
          https://www.tax.nd.gov/sites/www/files/documents/forms/business/ndwrfillable.pdf
          https://www.tax.nd.gov/news/tax-legislative-changes/significant-changes-law/individual-income-tax-history
          https://www.tax.nd.gov/forms
        
        Colorado:
        -         
        - https://leg.colorado.gov/agencies/legislative-council-staff/individual-income-tax%C2%A0
          https://leg.colorado.gov/agencies/legislative-council-staff/individual-income-tax%C2%A0
          https://tax.colorado.gov/sites/tax/files/documents/ITT_Capital_Gain_Subtraction_Feb_2022.pdf
          https://www.auroragov.org/business_services/taxes/occupational_privilege_tax
          https://www.denvergov.org/Government/Agencies-Departments-Offices/Agencies-Departments-Offices-Directory/Department-of-Finance/Our-Divisions/Treasury/Business-Tax-Information#:~:text=Occupational%20privilege%20tax
         -%28show%20below%29&text=Each%20taxable%20employee%20is%20liable
         month%20for%20each%20taxable%20employee
          https://www.glendale.co.us/355/Occupational-Privilege-Tax
          https://www.greenwoodvillage.com/1220/Occupational-Privilege-Tax-OPT
          https://ci.sheridan.co.us/288/Occupational-Privilege-Tax
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'NorthDakota', 'state_occupation': 'Colorado', "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def north_dakota_and_connecticut(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        North Dakota:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Connecticut:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        North Dakota:
        - https://www.tax.nd.gov/forms
          https://www.tax.nd.gov/sites/www/files/documents/forms/individual/2022-iit/2022-individual-income-tax-booklet.pdf
          https://www.tax.nd.gov/sites/www/files/documents/forms/2022-individual-income-tax-booklet.pdf
          https://www.tax.nd.gov/tax-exemptions-credits/income-tax-exemptions-credits
          https://www.tax.nd.gov/sites/www/files/documents/forms/business/ndwrfillable.pdf
          https://www.tax.nd.gov/news/tax-legislative-changes/significant-changes-law/individual-income-tax-history
          https://www.tax.nd.gov/forms
        
        Connecticut:
        -         
        - https://portal.ct.gov/-/media/DRS/Forms/2022/Income/2022-CT-1040-Instructions_1222.pdf
          https://www.cga.ct.gov/2022/rpt/pdf/2022-R-0108.pdf
          https://www.cga.ct.gov/current/pub/chap_229.htm#sec_12-703
          https://portal.ct.gov/-/media/DRS/Forms/2022/Income/2022-CT-1040-Instructions_1222.pdf
          "https://portal.ct.gov/DRS/Publications/TSSNs/TSSN-29#:~:text=An%20individuals%20net%20capital%20gains
         portion%20of%20Social%20Security%20benefits"
          "https://portal.ct.gov/DRS/Publications/TSSNs/TSSN-29#:~:text=An%20individuals%20net%20capital%20gains
         portion%20of%20Social%20Security%20benefits"
          https://portal.ct.gov/DRS/Individuals/Resident-Income-Tax/Tax-Information
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'NorthDakota', 'state_occupation': 'Connecticut', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def north_dakota_and_delaware(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, federal_agi=75000, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        North Dakota:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Delaware:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        federal_agi : int, optional
            Adjusted Gross Income (AGI) for federal taxes. Default is 0. Sources - https://www.irs.gov/e-file-providers/definition-of-adjusted-gross-income, https://www.irs.gov/newsroom/taxpayers-should-know-the-difference-between-standard-and-itemized-deductions
        income : int, optional
            Gross income. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        North Dakota:
        - https://www.tax.nd.gov/forms
          https://www.tax.nd.gov/sites/www/files/documents/forms/individual/2022-iit/2022-individual-income-tax-booklet.pdf
          https://www.tax.nd.gov/sites/www/files/documents/forms/2022-individual-income-tax-booklet.pdf
          https://www.tax.nd.gov/tax-exemptions-credits/income-tax-exemptions-credits
          https://www.tax.nd.gov/sites/www/files/documents/forms/business/ndwrfillable.pdf
          https://www.tax.nd.gov/news/tax-legislative-changes/significant-changes-law/individual-income-tax-history
          https://www.tax.nd.gov/forms
        
        Delaware:
        -         
        - https://revenue.delaware.gov/software-developer/tax-rate-changes/
          https://revenuefiles.delaware.gov/2022/PIT-RES_TY22_2022-01_PaperInteractive.pdf
          https://revenuefiles.delaware.gov/2022/PIT-RES_TY22_2022-02_Instructions.pdf
          https://revenue.delaware.gov/frequently-asked-questions/personal-income-tax-faqs/
          https://www.newcastlede.gov/DocumentCenter/View/20147/City-of-Wilmington-Wage-Tax-Information-PDF
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        federal_agi = self.federal_agi if hasattr(self, 'federal_agi') else federal_agi
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'NorthDakota', 'state_occupation': 'Delaware', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "federal_agi": federal_agi, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def north_dakota_and_florida(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        North Dakota:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Florida:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        North Dakota:
        - https://www.tax.nd.gov/forms
          https://www.tax.nd.gov/sites/www/files/documents/forms/individual/2022-iit/2022-individual-income-tax-booklet.pdf
          https://www.tax.nd.gov/sites/www/files/documents/forms/2022-individual-income-tax-booklet.pdf
          https://www.tax.nd.gov/tax-exemptions-credits/income-tax-exemptions-credits
          https://www.tax.nd.gov/sites/www/files/documents/forms/business/ndwrfillable.pdf
          https://www.tax.nd.gov/news/tax-legislative-changes/significant-changes-law/individual-income-tax-history
          https://www.tax.nd.gov/forms
        
        Florida:
        -         
        - https://floridarevenue.com/taxes/taxesfees/Pages/tax_interest_rates.aspx
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'NorthDakota', 'state_occupation': 'Florida', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def north_dakota_and_georgia(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        North Dakota:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Georgia:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        North Dakota:
        - https://www.tax.nd.gov/forms
          https://www.tax.nd.gov/sites/www/files/documents/forms/individual/2022-iit/2022-individual-income-tax-booklet.pdf
          https://www.tax.nd.gov/sites/www/files/documents/forms/2022-individual-income-tax-booklet.pdf
          https://www.tax.nd.gov/tax-exemptions-credits/income-tax-exemptions-credits
          https://www.tax.nd.gov/sites/www/files/documents/forms/business/ndwrfillable.pdf
          https://www.tax.nd.gov/news/tax-legislative-changes/significant-changes-law/individual-income-tax-history
          https://www.tax.nd.gov/forms
        
        Georgia:
        -         
        - https://dor.georgia.gov/tax-tables-georgia-tax-rate-schedule
          https://apps.dor.ga.gov/FillableForms/PDFViewer/Index?form=2022GA500
          https://apps.dor.ga.gov/FillableForms/PDFViewer/Index?form=2022GA500
          https://apps.dor.ga.gov/FillableForms/PDFViewer/Index?form=2022GA500
          https://dor.georgia.gov/years-individual-income-tax-forms
          https://dor.georgia.gov/years-individual-income-tax-forms
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'NorthDakota', 'state_occupation': 'Georgia', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def north_dakota_and_hawaii(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        North Dakota:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Hawaii:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        North Dakota:
        - https://www.tax.nd.gov/forms
          https://www.tax.nd.gov/sites/www/files/documents/forms/individual/2022-iit/2022-individual-income-tax-booklet.pdf
          https://www.tax.nd.gov/sites/www/files/documents/forms/2022-individual-income-tax-booklet.pdf
          https://www.tax.nd.gov/tax-exemptions-credits/income-tax-exemptions-credits
          https://www.tax.nd.gov/sites/www/files/documents/forms/business/ndwrfillable.pdf
          https://www.tax.nd.gov/news/tax-legislative-changes/significant-changes-law/individual-income-tax-history
          https://www.tax.nd.gov/forms
        
        Hawaii:
        -         
        - https://tax.hawaii.gov/forms/d_18table-on/d_18table-on_p13/
          https://files.hawaii.gov/tax/forms/2018/18table-on.pdf
          https://files.hawaii.gov/tax/news/pubs/22outline.pdf
          https://files.hawaii.gov/tax/news/pubs/FastTaxRefGuide_2017.pdf
          https://www.tfhawaii.org/wordpress/state-tax-resources/mini-tax-guides/individual-income-tax-chapter-235/#:~:text=Long%20term%20capital%20gains%20are%20taxed%20at%20a%20maximum%20of%207.25%25
          https://www.grassrootinstitute.org/2023/02/hb337-increasing-tax-on-capital-gains-a-bad-risk/
          https://files.hawaii.gov/tax/news/pubs/22outline.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'NorthDakota', 'state_occupation': 'Hawaii', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def north_dakota_and_idaho(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        North Dakota:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Idaho:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        North Dakota:
        - https://www.tax.nd.gov/forms
          https://www.tax.nd.gov/sites/www/files/documents/forms/individual/2022-iit/2022-individual-income-tax-booklet.pdf
          https://www.tax.nd.gov/sites/www/files/documents/forms/2022-individual-income-tax-booklet.pdf
          https://www.tax.nd.gov/tax-exemptions-credits/income-tax-exemptions-credits
          https://www.tax.nd.gov/sites/www/files/documents/forms/business/ndwrfillable.pdf
          https://www.tax.nd.gov/news/tax-legislative-changes/significant-changes-law/individual-income-tax-history
          https://www.tax.nd.gov/forms
        
        Idaho:
        -         
        - https://tax.idaho.gov/taxes/income-tax/individual-income/individual-income-tax-rate-schedule/
          https://tax.idaho.gov/wp-content/uploads/forms/EIN00046/EIN00046_03-01-2023.pdf
          https://tax.idaho.gov/taxes/income-tax/individual-income/idaho-source-income/
          https://tax.idaho.gov/taxes/income-tax/individual-income/idaho-source-income/
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'NorthDakota', 'state_occupation': 'Idaho', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def north_dakota_and_illinois(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        North Dakota:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Illinois:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        North Dakota:
        - https://www.tax.nd.gov/forms
          https://www.tax.nd.gov/sites/www/files/documents/forms/individual/2022-iit/2022-individual-income-tax-booklet.pdf
          https://www.tax.nd.gov/sites/www/files/documents/forms/2022-individual-income-tax-booklet.pdf
          https://www.tax.nd.gov/tax-exemptions-credits/income-tax-exemptions-credits
          https://www.tax.nd.gov/sites/www/files/documents/forms/business/ndwrfillable.pdf
          https://www.tax.nd.gov/news/tax-legislative-changes/significant-changes-law/individual-income-tax-history
          https://www.tax.nd.gov/forms
        
        Illinois:
        -         
        - https://tax.illinois.gov/research/taxrates/income.html#IndividualIncome
          https://tax.illinois.gov/content/dam/soi/en/web/tax/forms/incometax/documents/currentyear/individual/il-1040-instr.pdf
          https://tax.illinois.gov/questionsandanswers/12.html
          https://tax.illinois.gov/content/dam/soi/en/web/tax/forms/incometax/documents/currentyear/individual/il-1040-instr.pdf
          https://tax.illinois.gov/research/taxrates/personalproperty.html
          https://tax.illinois.gov/localgovernments/income.html
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'NorthDakota', 'state_occupation': 'Illinois', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def north_dakota_and_indiana(self, capital_gains_short=0, dependents=0, sources=False, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_long=0, filing_status='single', income=100000, credits_state=0):
        """
        --- Parameters ----
        
        North Dakota:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Indiana:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        North Dakota:
        - https://www.tax.nd.gov/forms
          https://www.tax.nd.gov/sites/www/files/documents/forms/individual/2022-iit/2022-individual-income-tax-booklet.pdf
          https://www.tax.nd.gov/sites/www/files/documents/forms/2022-individual-income-tax-booklet.pdf
          https://www.tax.nd.gov/tax-exemptions-credits/income-tax-exemptions-credits
          https://www.tax.nd.gov/sites/www/files/documents/forms/business/ndwrfillable.pdf
          https://www.tax.nd.gov/news/tax-legislative-changes/significant-changes-law/individual-income-tax-history
          https://www.tax.nd.gov/forms
        
        Indiana:
        -         
        - https://www.in.gov/dor/tax-forms/2022-individual-income-tax-forms/
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-32.htm
          https://www.in.gov/dor/files/ib28.pdf
          https://www.in.gov/dor/tax-forms/2022-individual-income-tax-forms/
          https://www.in.gov/dor/files/2022-county-tax-rates-and-codes.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'NorthDakota', 'state_occupation': 'Indiana', "capital_gains_short": capital_gains_short, "dependents": dependents, "sources": sources, "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def north_dakota_and_iowa(self, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        North Dakota:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Iowa:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        North Dakota:
        - https://www.tax.nd.gov/forms
          https://www.tax.nd.gov/sites/www/files/documents/forms/individual/2022-iit/2022-individual-income-tax-booklet.pdf
          https://www.tax.nd.gov/sites/www/files/documents/forms/2022-individual-income-tax-booklet.pdf
          https://www.tax.nd.gov/tax-exemptions-credits/income-tax-exemptions-credits
          https://www.tax.nd.gov/sites/www/files/documents/forms/business/ndwrfillable.pdf
          https://www.tax.nd.gov/news/tax-legislative-changes/significant-changes-law/individual-income-tax-history
          https://www.tax.nd.gov/forms
        
        Iowa:
        -         
        - https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
          https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
          https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
          https://tax.iowa.gov/iowa-illinois-reciprocal-agreement
          https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'NorthDakota', 'state_occupation': 'Iowa', "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def north_dakota_and_kansas(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        North Dakota:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Kansas:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        North Dakota:
        - https://www.tax.nd.gov/forms
          https://www.tax.nd.gov/sites/www/files/documents/forms/individual/2022-iit/2022-individual-income-tax-booklet.pdf
          https://www.tax.nd.gov/sites/www/files/documents/forms/2022-individual-income-tax-booklet.pdf
          https://www.tax.nd.gov/tax-exemptions-credits/income-tax-exemptions-credits
          https://www.tax.nd.gov/sites/www/files/documents/forms/business/ndwrfillable.pdf
          https://www.tax.nd.gov/news/tax-legislative-changes/significant-changes-law/individual-income-tax-history
          https://www.tax.nd.gov/forms
        
        Kansas:
        -         
        - https://www.ksrevenue.gov/taxrates.html
          https://www.ksrevenue.gov/incomebook22.html#0
          https://www.ksrevenue.gov/incomebook22.html#0
          https://www.ksrevenue.gov/incomebook22.html#0
          https://www.ksrevenue.gov/incomebook22.html
          https://www.ksrevenue.gov/incomebook22.html
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'NorthDakota', 'state_occupation': 'Kansas', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def north_dakota_and_kentucky(self, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        North Dakota:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Kentucky:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        North Dakota:
        - https://www.tax.nd.gov/forms
          https://www.tax.nd.gov/sites/www/files/documents/forms/individual/2022-iit/2022-individual-income-tax-booklet.pdf
          https://www.tax.nd.gov/sites/www/files/documents/forms/2022-individual-income-tax-booklet.pdf
          https://www.tax.nd.gov/tax-exemptions-credits/income-tax-exemptions-credits
          https://www.tax.nd.gov/sites/www/files/documents/forms/business/ndwrfillable.pdf
          https://www.tax.nd.gov/news/tax-legislative-changes/significant-changes-law/individual-income-tax-history
          https://www.tax.nd.gov/forms
        
        Kentucky:
        -         
        - https://revenue.ky.gov/News/Pages/Dor-Announces-Updates-to-Individual-Income-Tax-for-2023-Tax-Year.aspx
          https://revenue.ky.gov/Individual/Individual-Income-Tax/Pages/default.aspx
          https://revenue.ky.gov/News/Pages/Dor-Announces-Updates-to-Individual-Income-Tax-for-2023-Tax-Year.aspx
          https://revenue.ky.gov/Individual/Individual-Income-Tax/Pages/default.aspx
          https://revenue.ky.gov/Dor%20Training%20Materials/103%20KAR%2017.140.%20Individual%20income%20tax%20-%20reciprocity%20-%20nonresidents.pdf
          https://revenue.ky.gov/Individual/Individual-Income-Tax/Pages/default.aspx
          https://nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2021/Taxes-21-32.htm?taxmap=true
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'NorthDakota', 'state_occupation': 'Kentucky', "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def north_dakota_and_louisiana(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        North Dakota:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Louisiana:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        North Dakota:
        - https://www.tax.nd.gov/forms
          https://www.tax.nd.gov/sites/www/files/documents/forms/individual/2022-iit/2022-individual-income-tax-booklet.pdf
          https://www.tax.nd.gov/sites/www/files/documents/forms/2022-individual-income-tax-booklet.pdf
          https://www.tax.nd.gov/tax-exemptions-credits/income-tax-exemptions-credits
          https://www.tax.nd.gov/sites/www/files/documents/forms/business/ndwrfillable.pdf
          https://www.tax.nd.gov/news/tax-legislative-changes/significant-changes-law/individual-income-tax-history
          https://www.tax.nd.gov/forms
        
        Louisiana:
        -         
        - https://revenue.louisiana.gov/individualincometax
          https://revenue.louisiana.gov/individualincometax
          https://www.revenue.louisiana.gov/taxforms/6935(11_02)F.pdf
          https://revenue.louisiana.gov/individualincometax
          https://revenue.louisiana.gov/individualincometax
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'NorthDakota', 'state_occupation': 'Louisiana', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def north_dakota_and_maine(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        North Dakota:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Maine:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        North Dakota:
        - https://www.tax.nd.gov/forms
          https://www.tax.nd.gov/sites/www/files/documents/forms/individual/2022-iit/2022-individual-income-tax-booklet.pdf
          https://www.tax.nd.gov/sites/www/files/documents/forms/2022-individual-income-tax-booklet.pdf
          https://www.tax.nd.gov/tax-exemptions-credits/income-tax-exemptions-credits
          https://www.tax.nd.gov/sites/www/files/documents/forms/business/ndwrfillable.pdf
          https://www.tax.nd.gov/news/tax-legislative-changes/significant-changes-law/individual-income-tax-history
          https://www.tax.nd.gov/forms
        
        Maine:
        -         
        - https://www.maine.gov/revenue/tax-return-forms/individual-income-tax-2022
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/ind_tax_rate_sched_2022.pdf
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/22_1040me_book_gen_instruc_revisedFeb23.pdf
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/22_1040me_book_gen_instruc_revisedFeb23.pdf
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/22_1040me_book_gen_instruc_revisedFeb23.pdf
          https://www.maine.gov/revenue/tax-return-forms/individual-income-tax-2022
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'NorthDakota', 'state_occupation': 'Maine', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def north_dakota_and_maryland(self, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        North Dakota:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Maryland:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        North Dakota:
        - https://www.tax.nd.gov/forms
          https://www.tax.nd.gov/sites/www/files/documents/forms/individual/2022-iit/2022-individual-income-tax-booklet.pdf
          https://www.tax.nd.gov/sites/www/files/documents/forms/2022-individual-income-tax-booklet.pdf
          https://www.tax.nd.gov/tax-exemptions-credits/income-tax-exemptions-credits
          https://www.tax.nd.gov/sites/www/files/documents/forms/business/ndwrfillable.pdf
          https://www.tax.nd.gov/news/tax-legislative-changes/significant-changes-law/individual-income-tax-history
          https://www.tax.nd.gov/forms
        
        Maryland:
        -         
        - https://www.marylandtaxes.gov/individual/income/tax-info/tax-rates.php
          https://www.marylandtaxes.gov/forms/22_forms/Resident_Booklet.pdf
          https://www.marylandtaxes.gov/forms/Tax_Publications/Administrative_Releases/Income_and_Estate_Tax_Releases/ar_it37.pdf
          https://www.marylandtaxes.gov/individual/index.php
          https://www.marylandtaxes.gov/individual/income/tax-info/tax-rates.php
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'NorthDakota', 'state_occupation': 'Maryland', "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def north_dakota_and_massachusetts(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        North Dakota:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Massachusetts:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        North Dakota:
        - https://www.tax.nd.gov/forms
          https://www.tax.nd.gov/sites/www/files/documents/forms/individual/2022-iit/2022-individual-income-tax-booklet.pdf
          https://www.tax.nd.gov/sites/www/files/documents/forms/2022-individual-income-tax-booklet.pdf
          https://www.tax.nd.gov/tax-exemptions-credits/income-tax-exemptions-credits
          https://www.tax.nd.gov/sites/www/files/documents/forms/business/ndwrfillable.pdf
          https://www.tax.nd.gov/news/tax-legislative-changes/significant-changes-law/individual-income-tax-history
          https://www.tax.nd.gov/forms
        
        Massachusetts:
        -         
        - https://www.mass.gov/service-details/massachusetts-tax-rates
          https://www.mass.gov/service-details/view-massachusetts-personal-income-tax-exemptions
          https://www.mass.gov/service-details/view-massachusetts-personal-income-tax-exemptions
          https://www.mass.gov/service-details/view-child-and-dependent-related-credits
          https://www.mass.gov/service-details/massachusetts-tax-rates
          https://www.mass.gov/personal-income-tax
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'NorthDakota', 'state_occupation': 'Massachusetts', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def north_dakota_and_michigan(self, capital_gains_short=0, county_occupation='Other', dependents=0, sources=False, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_long=0, filing_status='single', income=100000, credits_state=0):
        """
        --- Parameters ----
        
        North Dakota:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Michigan:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        county_occupation : str, optional
            County of occupation. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        North Dakota:
        - https://www.tax.nd.gov/forms
          https://www.tax.nd.gov/sites/www/files/documents/forms/individual/2022-iit/2022-individual-income-tax-booklet.pdf
          https://www.tax.nd.gov/sites/www/files/documents/forms/2022-individual-income-tax-booklet.pdf
          https://www.tax.nd.gov/tax-exemptions-credits/income-tax-exemptions-credits
          https://www.tax.nd.gov/sites/www/files/documents/forms/business/ndwrfillable.pdf
          https://www.tax.nd.gov/news/tax-legislative-changes/significant-changes-law/individual-income-tax-history
          https://www.tax.nd.gov/forms
        
        Michigan:
        -         
        - https://www.michigan.gov/taxes/iit/new-developments-for-tax-year-2022
          "https://www.michigan.gov/taxes/-/media/Project/Websites/taxes/Forms/2023/2023-SUW/446_Withholding-Guide_2023.pdf?rev=177ca3d8cb034392bf84e317ed78c320&hash=19C5F42F43A89A91863D229349F48883#:~:text=The%20exemption%20amount%20is%20%245%2C400
         employees%20Michigan%20income%20tax%20return"
          "https://www.michigan.gov/taxes/-/media/Project/Websites/taxes/Forms/2023/2023-SUW/446_Withholding-Guide_2023.pdf?rev=177ca3d8cb034392bf84e317ed78c320&hash=19C5F42F43A89A91863D229349F48883#:~:text=The%20exemption%20amount%20is%20%245%2C400
         employees%20Michigan%20income%20tax%20return"
          https://www.michigan.gov/taxes/questions/iit/accordion/residency/are-my-wages-earned-in-another-state-taxable-in-michigan-if-i-am-a-michigan-resident-1
          https://www.michigan.gov/taxes/-/media/Project/Websites/taxes/Forms/2022/2022-IIT-Forms/BOOK_MI-1040.pdf?rev=4524f1cbbc0243b194ab05a6680be75b&hash=40DFE71E651F7C1D65C31F6CD03D944F
          https://www.michigan.gov/taxes/questions/iit/accordion/general/what-cities-impose-an-income-tax
        """
        
        # Check if the parameters are set upstream in the instance
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        county_occupation = self.county_occupation if hasattr(self, 'county_occupation') else county_occupation
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'NorthDakota', 'state_occupation': 'Michigan', "capital_gains_short": capital_gains_short, "county_occupation": county_occupation, "dependents": dependents, "sources": sources, "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def north_dakota_and_minnesota(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        North Dakota:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Minnesota:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        North Dakota:
        - https://www.tax.nd.gov/forms
          https://www.tax.nd.gov/sites/www/files/documents/forms/individual/2022-iit/2022-individual-income-tax-booklet.pdf
          https://www.tax.nd.gov/sites/www/files/documents/forms/2022-individual-income-tax-booklet.pdf
          https://www.tax.nd.gov/tax-exemptions-credits/income-tax-exemptions-credits
          https://www.tax.nd.gov/sites/www/files/documents/forms/business/ndwrfillable.pdf
          https://www.tax.nd.gov/news/tax-legislative-changes/significant-changes-law/individual-income-tax-history
          https://www.tax.nd.gov/forms
        
        Minnesota:
        -         
        - https://www.revenue.state.mn.us/minnesota-income-tax-rates-and-brackets
          https://www.revenue.state.mn.us/child-and-dependent-care-credit
          https://www.revenue.state.mn.us/dependent-exemptions
          https://www.revenue.state.mn.us/minnesota-standard-deduction
          https://mn.gov/mmb/assets/mwr_form_tcm1059-128581.pdf
          https://www.revenue.state.mn.us/individual-income-tax
          https://www.revenue.state.mn.us/individual-income-tax
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'NorthDakota', 'state_occupation': 'Minnesota', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def north_dakota_and_mississippi(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        North Dakota:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Mississippi:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        North Dakota:
        - https://www.tax.nd.gov/forms
          https://www.tax.nd.gov/sites/www/files/documents/forms/individual/2022-iit/2022-individual-income-tax-booklet.pdf
          https://www.tax.nd.gov/sites/www/files/documents/forms/2022-individual-income-tax-booklet.pdf
          https://www.tax.nd.gov/tax-exemptions-credits/income-tax-exemptions-credits
          https://www.tax.nd.gov/sites/www/files/documents/forms/business/ndwrfillable.pdf
          https://www.tax.nd.gov/news/tax-legislative-changes/significant-changes-law/individual-income-tax-history
          https://www.tax.nd.gov/forms
        
        Mississippi:
        -         
        - https://www.dor.ms.gov/individual/tax-rates
          https://www.dor.ms.gov/individual/tax-rates
          https://www.dor.ms.gov/individual/tax-rates
          https://www.dor.ms.gov/individual
          https://www.dor.ms.gov/individual
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'NorthDakota', 'state_occupation': 'Mississippi', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def north_dakota_and_missouri(self, capital_gains_short=0, dependents=0, sources=False, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_long=0, filing_status='single', income=100000, credits_state=0):
        """
        --- Parameters ----
        
        North Dakota:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Missouri:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        North Dakota:
        - https://www.tax.nd.gov/forms
          https://www.tax.nd.gov/sites/www/files/documents/forms/individual/2022-iit/2022-individual-income-tax-booklet.pdf
          https://www.tax.nd.gov/sites/www/files/documents/forms/2022-individual-income-tax-booklet.pdf
          https://www.tax.nd.gov/tax-exemptions-credits/income-tax-exemptions-credits
          https://www.tax.nd.gov/sites/www/files/documents/forms/business/ndwrfillable.pdf
          https://www.tax.nd.gov/news/tax-legislative-changes/significant-changes-law/individual-income-tax-history
          https://www.tax.nd.gov/forms
        
        Missouri:
        -         
        - https://dor.mo.gov/taxation/individual/tax-types/income/year-changes/
          https://dor.mo.gov/taxation/individual/tax-types/income/year-changes/
          https://dor.mo.gov/taxation/individual/tax-types/income/year-changes/
          https://www.kcmo.gov/city-hall/departments/finance/earnings-tax
          https://www.stlouis-mo.gov/government/departments/collector/earnings-tax/individual-earnings-tax-info.cfm
        """
        
        # Check if the parameters are set upstream in the instance
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'NorthDakota', 'state_occupation': 'Missouri', "capital_gains_short": capital_gains_short, "dependents": dependents, "sources": sources, "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def north_dakota_and_montana(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        North Dakota:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Montana:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        North Dakota:
        - https://www.tax.nd.gov/forms
          https://www.tax.nd.gov/sites/www/files/documents/forms/individual/2022-iit/2022-individual-income-tax-booklet.pdf
          https://www.tax.nd.gov/sites/www/files/documents/forms/2022-individual-income-tax-booklet.pdf
          https://www.tax.nd.gov/tax-exemptions-credits/income-tax-exemptions-credits
          https://www.tax.nd.gov/sites/www/files/documents/forms/business/ndwrfillable.pdf
          https://www.tax.nd.gov/news/tax-legislative-changes/significant-changes-law/individual-income-tax-history
          https://www.tax.nd.gov/forms
        
        Montana:
        -         
        - https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/2022/12/montana-employees-withholding-allowance-and-exemption-certificate-form-mw-4.pdf
          https://mtrevenue.gov/2021/11/29/simplification-of-montana-income-taxation/?utm_source=rss&utm_medium=rss&utm_campaign=simplification-of-montana-income-taxation
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'NorthDakota', 'state_occupation': 'Montana', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def north_dakota_and_nebraska(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        North Dakota:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Nebraska:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        North Dakota:
        - https://www.tax.nd.gov/forms
          https://www.tax.nd.gov/sites/www/files/documents/forms/individual/2022-iit/2022-individual-income-tax-booklet.pdf
          https://www.tax.nd.gov/sites/www/files/documents/forms/2022-individual-income-tax-booklet.pdf
          https://www.tax.nd.gov/tax-exemptions-credits/income-tax-exemptions-credits
          https://www.tax.nd.gov/sites/www/files/documents/forms/business/ndwrfillable.pdf
          https://www.tax.nd.gov/news/tax-legislative-changes/significant-changes-law/individual-income-tax-history
          https://www.tax.nd.gov/forms
        
        Nebraska:
        -         
        - https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2023_Circular_EN_8-429r2023_Whole_Book_11.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'NorthDakota', 'state_occupation': 'Nebraska', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def north_dakota_and_nevada(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        North Dakota:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Nevada:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        North Dakota:
        - https://www.tax.nd.gov/forms
          https://www.tax.nd.gov/sites/www/files/documents/forms/individual/2022-iit/2022-individual-income-tax-booklet.pdf
          https://www.tax.nd.gov/sites/www/files/documents/forms/2022-individual-income-tax-booklet.pdf
          https://www.tax.nd.gov/tax-exemptions-credits/income-tax-exemptions-credits
          https://www.tax.nd.gov/sites/www/files/documents/forms/business/ndwrfillable.pdf
          https://www.tax.nd.gov/news/tax-legislative-changes/significant-changes-law/individual-income-tax-history
          https://www.tax.nd.gov/forms
        
        Nevada:
        -         
        - https://www.nvsos.gov/sos/businesses/the-nevada-advantage
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'NorthDakota', 'state_occupation': 'Nevada', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def north_dakota_and_new_hampshire(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        North Dakota:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        New Hampshire:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        North Dakota:
        - https://www.tax.nd.gov/forms
          https://www.tax.nd.gov/sites/www/files/documents/forms/individual/2022-iit/2022-individual-income-tax-booklet.pdf
          https://www.tax.nd.gov/sites/www/files/documents/forms/2022-individual-income-tax-booklet.pdf
          https://www.tax.nd.gov/tax-exemptions-credits/income-tax-exemptions-credits
          https://www.tax.nd.gov/sites/www/files/documents/forms/business/ndwrfillable.pdf
          https://www.tax.nd.gov/news/tax-legislative-changes/significant-changes-law/individual-income-tax-history
          https://www.tax.nd.gov/forms
        
        New Hampshire:
        -         
        - https://www.revenue.nh.gov/assistance/tax-overview.htm
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'NorthDakota', 'state_occupation': 'NewHampshire', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def north_dakota_and_new_jersey(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        North Dakota:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        New Jersey:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        North Dakota:
        - https://www.tax.nd.gov/forms
          https://www.tax.nd.gov/sites/www/files/documents/forms/individual/2022-iit/2022-individual-income-tax-booklet.pdf
          https://www.tax.nd.gov/sites/www/files/documents/forms/2022-individual-income-tax-booklet.pdf
          https://www.tax.nd.gov/tax-exemptions-credits/income-tax-exemptions-credits
          https://www.tax.nd.gov/sites/www/files/documents/forms/business/ndwrfillable.pdf
          https://www.tax.nd.gov/news/tax-legislative-changes/significant-changes-law/individual-income-tax-history
          https://www.tax.nd.gov/forms
        
        New Jersey:
        -         
        - https://www.state.nj.us/treasury/taxation/taxtables.shtml
          https://www.state.nj.us/treasury/taxation/njit13.shtml
          https://www.nj.gov/treasury/taxation/njit25.shtml
          https://www.state.nj.us/treasury/taxation/individuals/freqqiti.shtml
          https://www.state.nj.us/treasury/taxation/localtax.shtml
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'NorthDakota', 'state_occupation': 'NewJersey', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def north_dakota_and_new_mexico(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        North Dakota:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        New Mexico:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        North Dakota:
        - https://www.tax.nd.gov/forms
          https://www.tax.nd.gov/sites/www/files/documents/forms/individual/2022-iit/2022-individual-income-tax-booklet.pdf
          https://www.tax.nd.gov/sites/www/files/documents/forms/2022-individual-income-tax-booklet.pdf
          https://www.tax.nd.gov/tax-exemptions-credits/income-tax-exemptions-credits
          https://www.tax.nd.gov/sites/www/files/documents/forms/business/ndwrfillable.pdf
          https://www.tax.nd.gov/news/tax-legislative-changes/significant-changes-law/individual-income-tax-history
          https://www.tax.nd.gov/forms
        
        New Mexico:
        -         
        - https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/fdf3c548-8aba-4b9c-9eb4-bb564c716015/FYI-104.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/0bf943fd-652e-4400-bb1b-cd8397eb5a95/2022pit-adj.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/2f1a6781-9534-4436-b427-1557f9592099/2022pit-adj-ins.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/0bf943fd-652e-4400-bb1b-cd8397eb5a95/2022pit-adj.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/1afc56af-ea90-4d48-82e5-1f9aeb43255a/PITbook2022.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/fdf3c548-8aba-4b9c-9eb4-bb564c716015/FYI-104.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'NorthDakota', 'state_occupation': 'NewMexico', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def north_dakota_and_new_york(self, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        North Dakota:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        New York:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        North Dakota:
        - https://www.tax.nd.gov/forms
          https://www.tax.nd.gov/sites/www/files/documents/forms/individual/2022-iit/2022-individual-income-tax-booklet.pdf
          https://www.tax.nd.gov/sites/www/files/documents/forms/2022-individual-income-tax-booklet.pdf
          https://www.tax.nd.gov/tax-exemptions-credits/income-tax-exemptions-credits
          https://www.tax.nd.gov/sites/www/files/documents/forms/business/ndwrfillable.pdf
          https://www.tax.nd.gov/news/tax-legislative-changes/significant-changes-law/individual-income-tax-history
          https://www.tax.nd.gov/forms
        
        New York:
        -         
        - https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nys-tax-rate-schedule
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nys-tax-rate-schedule
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nys-tax-rate-schedule
          https://www.tax.ny.gov/pit/
          https://www.tax.ny.gov/pit/file/nonresident-faqs.htm
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#yonkers-tax-rate-schedule
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nyc-tax-rate-schedule
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'NorthDakota', 'state_occupation': 'NewYork', "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def north_dakota_and_north_carolina(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        North Dakota:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        North Carolina:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        North Dakota:
        - https://www.tax.nd.gov/forms
          https://www.tax.nd.gov/sites/www/files/documents/forms/individual/2022-iit/2022-individual-income-tax-booklet.pdf
          https://www.tax.nd.gov/sites/www/files/documents/forms/2022-individual-income-tax-booklet.pdf
          https://www.tax.nd.gov/tax-exemptions-credits/income-tax-exemptions-credits
          https://www.tax.nd.gov/sites/www/files/documents/forms/business/ndwrfillable.pdf
          https://www.tax.nd.gov/news/tax-legislative-changes/significant-changes-law/individual-income-tax-history
          https://www.tax.nd.gov/forms
        
        North Carolina:
        -         
        - https://www.ncdor.gov/taxes-forms/tax-rate-schedules
          https://www.ncdor.gov/taxes-forms/individual-income-tax/north-carolina-standard-deduction-or-north-carolina-itemized-deductions
          https://www.ncdor.gov/taxes-forms/individual-income-tax
          https://www.ncdor.gov/taxes-forms/individual-income-tax
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'NorthDakota', 'state_occupation': 'NorthCarolina', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def north_dakota_and_ohio(self, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        North Dakota:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Ohio:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        North Dakota:
        - https://www.tax.nd.gov/forms
          https://www.tax.nd.gov/sites/www/files/documents/forms/individual/2022-iit/2022-individual-income-tax-booklet.pdf
          https://www.tax.nd.gov/sites/www/files/documents/forms/2022-individual-income-tax-booklet.pdf
          https://www.tax.nd.gov/tax-exemptions-credits/income-tax-exemptions-credits
          https://www.tax.nd.gov/sites/www/files/documents/forms/business/ndwrfillable.pdf
          https://www.tax.nd.gov/news/tax-legislative-changes/significant-changes-law/individual-income-tax-history
          https://www.tax.nd.gov/forms
        
        Ohio:
        -         
        - https://tax.ohio.gov/individual/resources/annual-tax-rates
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://tax.ohio.gov/static/forms/employer_withholding/generic/wth_it4nr.pdf
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://www.ritaohio.com/TaxRatesTable
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'NorthDakota', 'state_occupation': 'Ohio', "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def north_dakota_and_oklahoma(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        North Dakota:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Oklahoma:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        North Dakota:
        - https://www.tax.nd.gov/forms
          https://www.tax.nd.gov/sites/www/files/documents/forms/individual/2022-iit/2022-individual-income-tax-booklet.pdf
          https://www.tax.nd.gov/sites/www/files/documents/forms/2022-individual-income-tax-booklet.pdf
          https://www.tax.nd.gov/tax-exemptions-credits/income-tax-exemptions-credits
          https://www.tax.nd.gov/sites/www/files/documents/forms/business/ndwrfillable.pdf
          https://www.tax.nd.gov/news/tax-legislative-changes/significant-changes-law/individual-income-tax-history
          https://www.tax.nd.gov/forms
        
        Oklahoma:
        -         
        - https://oklahoma.gov/content/dam/ok/en/tax/documents/resources/publications/businesses/withholding-tables/WHTables-2022.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'NorthDakota', 'state_occupation': 'Oklahoma', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def north_dakota_and_oregon(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        North Dakota:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Oregon:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        North Dakota:
        - https://www.tax.nd.gov/forms
          https://www.tax.nd.gov/sites/www/files/documents/forms/individual/2022-iit/2022-individual-income-tax-booklet.pdf
          https://www.tax.nd.gov/sites/www/files/documents/forms/2022-individual-income-tax-booklet.pdf
          https://www.tax.nd.gov/tax-exemptions-credits/income-tax-exemptions-credits
          https://www.tax.nd.gov/sites/www/files/documents/forms/business/ndwrfillable.pdf
          https://www.tax.nd.gov/news/tax-legislative-changes/significant-changes-law/individual-income-tax-history
          https://www.tax.nd.gov/forms
        
        Oregon:
        -         
        - https://www.oregon.gov/dor/programs/individuals/Pages/PIT.aspx
          https://www.oregon.gov/dor/programs/individuals/Documents/Full%20year%20resident
         %20Form%20or-40%20filers.pdf
          https://www.oregon.gov/dor/programs/individuals/Documents/Part-year%20and%20nonresident
         %20Form%20or-40-P%20and%20or-40-N%20filers.pdf.pdf
          https://www.oregon.gov/dor/forms/FormsPubs/form-or-40_101-040_2022.pdf
          https://www.oregon.gov/dor/forms/FormsPubs/form-or-40_101-040_2022.pdf
          https://www.oregon.gov/dor/forms/FormsPubs/form-or-40_101-040_2022.pdf
          https://www.portland.gov/revenue/personal-tax
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'NorthDakota', 'state_occupation': 'Oregon', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def north_dakota_and_pennsylvania(self, county_residence='Other', state_agi=75000, county_occupation='Other', filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        North Dakota:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Pennsylvania:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        county_residence : str, optional
            County of residence. Default is Other.
        county_occupation : str, optional
            County of occupation. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        North Dakota:
        - https://www.tax.nd.gov/forms
          https://www.tax.nd.gov/sites/www/files/documents/forms/individual/2022-iit/2022-individual-income-tax-booklet.pdf
          https://www.tax.nd.gov/sites/www/files/documents/forms/2022-individual-income-tax-booklet.pdf
          https://www.tax.nd.gov/tax-exemptions-credits/income-tax-exemptions-credits
          https://www.tax.nd.gov/sites/www/files/documents/forms/business/ndwrfillable.pdf
          https://www.tax.nd.gov/news/tax-legislative-changes/significant-changes-law/individual-income-tax-history
          https://www.tax.nd.gov/forms
        
        Pennsylvania:
        -         
        - https://www.revenue.pa.gov/Tax%20Rates/Pages/default.aspx
          https://www.revenue.pa.gov/FormsandPublications/FormsforIndividuals/PIT/Documents/2022/2022_pa-40in.pdf
          https://www.revenue.pa.gov/FormsandPublications/FormsforBusinesses/EmployerWithholding/Documents/rev-419.pdf
          https://www.revenue.pa.gov/FormsandPublications/FormsforIndividuals/PIT/Documents/2022/2022_pa-40in.pdf
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-25.htm
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        county_occupation = self.county_occupation if hasattr(self, 'county_occupation') else county_occupation
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'NorthDakota', 'state_occupation': 'Pennsylvania', "county_residence": county_residence, "state_agi": state_agi, "county_occupation": county_occupation, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def north_dakota_and_rhode_island(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        North Dakota:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Rhode Island:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        North Dakota:
        - https://www.tax.nd.gov/forms
          https://www.tax.nd.gov/sites/www/files/documents/forms/individual/2022-iit/2022-individual-income-tax-booklet.pdf
          https://www.tax.nd.gov/sites/www/files/documents/forms/2022-individual-income-tax-booklet.pdf
          https://www.tax.nd.gov/tax-exemptions-credits/income-tax-exemptions-credits
          https://www.tax.nd.gov/sites/www/files/documents/forms/business/ndwrfillable.pdf
          https://www.tax.nd.gov/news/tax-legislative-changes/significant-changes-law/individual-income-tax-history
          https://www.tax.nd.gov/forms
        
        Rhode Island:
        -         
        - https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2022_1040WE_w_0.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'NorthDakota', 'state_occupation': 'RhodeIsland', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def north_dakota_and_south_carolina(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        North Dakota:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        South Carolina:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        North Dakota:
        - https://www.tax.nd.gov/forms
          https://www.tax.nd.gov/sites/www/files/documents/forms/individual/2022-iit/2022-individual-income-tax-booklet.pdf
          https://www.tax.nd.gov/sites/www/files/documents/forms/2022-individual-income-tax-booklet.pdf
          https://www.tax.nd.gov/tax-exemptions-credits/income-tax-exemptions-credits
          https://www.tax.nd.gov/sites/www/files/documents/forms/business/ndwrfillable.pdf
          https://www.tax.nd.gov/news/tax-legislative-changes/significant-changes-law/individual-income-tax-history
          https://www.tax.nd.gov/forms
        
        South Carolina:
        -         
        - https://dor.sc.gov/resources-site/lawandpolicy/Advisory%20Opinions/IL22-15.pdf
          https://dor.sc.gov/forms-site/Forms/IITPacket_2022.pdf
          https://www.scstatehouse.gov/code/t12c006.php
          https://dor.sc.gov/resources-site/lawandpolicy/Advisory%20Opinions/IL22-15.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'NorthDakota', 'state_occupation': 'SouthCarolina', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def north_dakota_and_south_dakota(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        North Dakota:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        South Dakota:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        North Dakota:
        - https://www.tax.nd.gov/forms
          https://www.tax.nd.gov/sites/www/files/documents/forms/individual/2022-iit/2022-individual-income-tax-booklet.pdf
          https://www.tax.nd.gov/sites/www/files/documents/forms/2022-individual-income-tax-booklet.pdf
          https://www.tax.nd.gov/tax-exemptions-credits/income-tax-exemptions-credits
          https://www.tax.nd.gov/sites/www/files/documents/forms/business/ndwrfillable.pdf
          https://www.tax.nd.gov/news/tax-legislative-changes/significant-changes-law/individual-income-tax-history
          https://www.tax.nd.gov/forms
        
        South Dakota:
        -         
        - https://dor.sd.gov/individuals/taxes/
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'NorthDakota', 'state_occupation': 'SouthDakota', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def north_dakota_and_tennessee(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        North Dakota:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Tennessee:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        North Dakota:
        - https://www.tax.nd.gov/forms
          https://www.tax.nd.gov/sites/www/files/documents/forms/individual/2022-iit/2022-individual-income-tax-booklet.pdf
          https://www.tax.nd.gov/sites/www/files/documents/forms/2022-individual-income-tax-booklet.pdf
          https://www.tax.nd.gov/tax-exemptions-credits/income-tax-exemptions-credits
          https://www.tax.nd.gov/sites/www/files/documents/forms/business/ndwrfillable.pdf
          https://www.tax.nd.gov/news/tax-legislative-changes/significant-changes-law/individual-income-tax-history
          https://www.tax.nd.gov/forms
        
        Tennessee:
        -         
        - https://www.tn.gov/revenue/taxes.html
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'NorthDakota', 'state_occupation': 'Tennessee', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def north_dakota_and_texas(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        North Dakota:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Texas:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        North Dakota:
        - https://www.tax.nd.gov/forms
          https://www.tax.nd.gov/sites/www/files/documents/forms/individual/2022-iit/2022-individual-income-tax-booklet.pdf
          https://www.tax.nd.gov/sites/www/files/documents/forms/2022-individual-income-tax-booklet.pdf
          https://www.tax.nd.gov/tax-exemptions-credits/income-tax-exemptions-credits
          https://www.tax.nd.gov/sites/www/files/documents/forms/business/ndwrfillable.pdf
          https://www.tax.nd.gov/news/tax-legislative-changes/significant-changes-law/individual-income-tax-history
          https://www.tax.nd.gov/forms
        
        Texas:
        -         
        - https://comptroller.texas.gov/taxes/
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'NorthDakota', 'state_occupation': 'Texas', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def north_dakota_and_utah(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        North Dakota:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Utah:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        North Dakota:
        - https://www.tax.nd.gov/forms
          https://www.tax.nd.gov/sites/www/files/documents/forms/individual/2022-iit/2022-individual-income-tax-booklet.pdf
          https://www.tax.nd.gov/sites/www/files/documents/forms/2022-individual-income-tax-booklet.pdf
          https://www.tax.nd.gov/tax-exemptions-credits/income-tax-exemptions-credits
          https://www.tax.nd.gov/sites/www/files/documents/forms/business/ndwrfillable.pdf
          https://www.tax.nd.gov/news/tax-legislative-changes/significant-changes-law/individual-income-tax-history
          https://www.tax.nd.gov/forms
        
        Utah:
        -         
        - https://incometax.utah.gov/paying/tax-rates
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'NorthDakota', 'state_occupation': 'Utah', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def north_dakota_and_vermont(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        North Dakota:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Vermont:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        North Dakota:
        - https://www.tax.nd.gov/forms
          https://www.tax.nd.gov/sites/www/files/documents/forms/individual/2022-iit/2022-individual-income-tax-booklet.pdf
          https://www.tax.nd.gov/sites/www/files/documents/forms/2022-individual-income-tax-booklet.pdf
          https://www.tax.nd.gov/tax-exemptions-credits/income-tax-exemptions-credits
          https://www.tax.nd.gov/sites/www/files/documents/forms/business/ndwrfillable.pdf
          https://www.tax.nd.gov/news/tax-legislative-changes/significant-changes-law/individual-income-tax-history
          https://www.tax.nd.gov/forms
        
        Vermont:
        -         
        - https://tax.vermont.gov/sites/tax/files/documents/GB-1210-2023.pdf
          https://tax.vermont.gov/sites/tax/files/documents/Income%20Booklet-2022.pdf
          https://tax.vermont.gov/sites/tax/files/documents/Income%20Booklet-2022.pdf
          https://tax.vermont.gov/sites/tax/files/documents/Income%20Booklet-2022.pdf
          https://tax.vermont.gov/individuals/personal-income-tax/tax-credits
          https://tax.vermont.gov/individuals/personal-income-tax/taxable-income
          https://tax.vermont.gov/individuals/personal-income-tax/local-option-tax
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'NorthDakota', 'state_occupation': 'Vermont', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def north_dakota_and_virginia(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        North Dakota:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Virginia:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        North Dakota:
        - https://www.tax.nd.gov/forms
          https://www.tax.nd.gov/sites/www/files/documents/forms/individual/2022-iit/2022-individual-income-tax-booklet.pdf
          https://www.tax.nd.gov/sites/www/files/documents/forms/2022-individual-income-tax-booklet.pdf
          https://www.tax.nd.gov/tax-exemptions-credits/income-tax-exemptions-credits
          https://www.tax.nd.gov/sites/www/files/documents/forms/business/ndwrfillable.pdf
          https://www.tax.nd.gov/news/tax-legislative-changes/significant-changes-law/individual-income-tax-history
          https://www.tax.nd.gov/forms
        
        Virginia:
        -         
        - https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/reciprocity
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'NorthDakota', 'state_occupation': 'Virginia', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def north_dakota_and_washington(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        North Dakota:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Washington:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        North Dakota:
        - https://www.tax.nd.gov/forms
          https://www.tax.nd.gov/sites/www/files/documents/forms/individual/2022-iit/2022-individual-income-tax-booklet.pdf
          https://www.tax.nd.gov/sites/www/files/documents/forms/2022-individual-income-tax-booklet.pdf
          https://www.tax.nd.gov/tax-exemptions-credits/income-tax-exemptions-credits
          https://www.tax.nd.gov/sites/www/files/documents/forms/business/ndwrfillable.pdf
          https://www.tax.nd.gov/news/tax-legislative-changes/significant-changes-law/individual-income-tax-history
          https://www.tax.nd.gov/forms
        
        Washington:
        -         
        - https://dor.wa.gov/taxes-rates/income-tax
          https://dor.wa.gov/taxes-rates/other-taxes/capital-gains-tax
          https://dor.wa.gov/taxes-rates/income-tax
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'NorthDakota', 'state_occupation': 'Washington', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def north_dakota_and_west_virginia(self, pay_periods=24, capital_gains_short=0, county_occupation='Other', dependents=0, sources=False, state_agi=75000, filers_over_65=0, capital_gains_long=0, filing_status='single', income=100000, credits_state=0):
        """
        --- Parameters ----
        
        North Dakota:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        West Virginia:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_occupation : str, optional
            County of occupation. Default is Other.
        pay_periods : int, optional
            Number of pay periods. Options are 13, 24, 26, 52. Default is 24.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        North Dakota:
        - https://www.tax.nd.gov/forms
          https://www.tax.nd.gov/sites/www/files/documents/forms/individual/2022-iit/2022-individual-income-tax-booklet.pdf
          https://www.tax.nd.gov/sites/www/files/documents/forms/2022-individual-income-tax-booklet.pdf
          https://www.tax.nd.gov/tax-exemptions-credits/income-tax-exemptions-credits
          https://www.tax.nd.gov/sites/www/files/documents/forms/business/ndwrfillable.pdf
          https://www.tax.nd.gov/news/tax-legislative-changes/significant-changes-law/individual-income-tax-history
          https://www.tax.nd.gov/forms
        
        West Virginia:
        -         
        - https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://tax.wv.gov/Documents/TSD/tsd438.pdf
          https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2011/Taxes-11-35att.html
        """
        
        # Check if the parameters are set upstream in the instance
        pay_periods = self.pay_periods if hasattr(self, 'pay_periods') else pay_periods
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        county_occupation = self.county_occupation if hasattr(self, 'county_occupation') else county_occupation
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'NorthDakota', 'state_occupation': 'WestVirginia', "pay_periods": pay_periods, "capital_gains_short": capital_gains_short, "county_occupation": county_occupation, "dependents": dependents, "sources": sources, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def north_dakota_and_wisconsin(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        North Dakota:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Wisconsin:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        North Dakota:
        - https://www.tax.nd.gov/forms
          https://www.tax.nd.gov/sites/www/files/documents/forms/individual/2022-iit/2022-individual-income-tax-booklet.pdf
          https://www.tax.nd.gov/sites/www/files/documents/forms/2022-individual-income-tax-booklet.pdf
          https://www.tax.nd.gov/tax-exemptions-credits/income-tax-exemptions-credits
          https://www.tax.nd.gov/sites/www/files/documents/forms/business/ndwrfillable.pdf
          https://www.tax.nd.gov/news/tax-legislative-changes/significant-changes-law/individual-income-tax-history
          https://www.tax.nd.gov/forms
        
        Wisconsin:
        -         
        - https://www.revenue.wi.gov/TaxForms2023/2023-Form1-ES-Inst.pdf
          https://www.revenue.wi.gov/TaxForms2023/2023-Form1-ES-Inst.pdf
          https://www.revenue.wi.gov/TaxForms2023/2023-Form1-ES-Inst.pdf
          https://www.revenue.wi.gov/Pages/FAQS/pcs-work.aspx
          https://www.revenue.wi.gov/Dor%20Publications/pb103.pdf
          https://www.revenue.wi.gov/Pages/FAQS/pcs-work.aspx
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'NorthDakota', 'state_occupation': 'Wisconsin', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def north_dakota_and_wyoming(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        North Dakota:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Wyoming:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        North Dakota:
        - https://www.tax.nd.gov/forms
          https://www.tax.nd.gov/sites/www/files/documents/forms/individual/2022-iit/2022-individual-income-tax-booklet.pdf
          https://www.tax.nd.gov/sites/www/files/documents/forms/2022-individual-income-tax-booklet.pdf
          https://www.tax.nd.gov/tax-exemptions-credits/income-tax-exemptions-credits
          https://www.tax.nd.gov/sites/www/files/documents/forms/business/ndwrfillable.pdf
          https://www.tax.nd.gov/news/tax-legislative-changes/significant-changes-law/individual-income-tax-history
          https://www.tax.nd.gov/forms
        
        Wyoming:
        -         
        - https://revenue.wyo.gov/
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'NorthDakota', 'state_occupation': 'Wyoming', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def ohio_and_alabama(self, county_residence='Other', state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Ohio:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Alabama:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Ohio:
        - https://tax.ohio.gov/individual/resources/annual-tax-rates
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://tax.ohio.gov/static/forms/employer_withholding/generic/wth_it4nr.pdf
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://www.ritaohio.com/TaxRatesTable
        
        Alabama:
        -         
        - https://www.revenue.alabama.gov/ultraviewer/viewer/basic_viewer/index.html?form=2023/02/22f40.pdf
          https://www.revenue.alabama.gov/faqs/what-is-alabamas-individual-income-tax-rate/
          https://www.revenue.alabama.gov/ultraviewer/viewer/basic_viewer/index.html?form=2023/02/22f40.pdf
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-24.htm?taxmap=true
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-24.htm?taxmap=true
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-24.htm?taxmap=true
          https://www.revenue.alabama.gov/individual-corporate/income-to-be-reported-on-the-alabama-income-tax-return/
          https://www.almonline.org/TaxRates.aspx#OccupationalTax
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Ohio', 'state_occupation': 'Alabama', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def ohio_and_alaska(self, county_residence='Other', state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Ohio:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Alaska:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Ohio:
        - https://tax.ohio.gov/individual/resources/annual-tax-rates
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://tax.ohio.gov/static/forms/employer_withholding/generic/wth_it4nr.pdf
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://www.ritaohio.com/TaxRatesTable
        
        Alaska:
        -         
        - https://www.commerce.alaska.gov/web/dcra/officeofthestateassessor/alaskataxfacts.aspx
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Ohio', 'state_occupation': 'Alaska', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def ohio_and_arizona(self, county_residence='Other', state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Ohio:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Arizona:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        credits_state : int, optional
            Credits for state taxes. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Ohio:
        - https://tax.ohio.gov/individual/resources/annual-tax-rates
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://tax.ohio.gov/static/forms/employer_withholding/generic/wth_it4nr.pdf
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://www.ritaohio.com/TaxRatesTable
        
        Arizona:
        -         
        - https://azdor.gov/forms/individual/form-140-x-y-tables
          https://azdor.gov/forms/individual/form-140-x-y-tables
          https://azdor.gov/forms/individual/form-140-x-y-tables
          https://azdor.gov/businesses-arizona/withholding-tax/withholding-exceptions
          "https://azdor.gov/individuals/income-tax-filing-assistance/identifying-other-taxable-income#:~:text=In%20regards%20to%20capital%20gains
         the%20individuals%20regular%20tax%20rate.
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Ohio', 'state_occupation': 'Arizona', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def ohio_and_arkansas(self, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Ohio:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Arkansas:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Ohio:
        - https://tax.ohio.gov/individual/resources/annual-tax-rates
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://tax.ohio.gov/static/forms/employer_withholding/generic/wth_it4nr.pdf
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://www.ritaohio.com/TaxRatesTable
        
        Arkansas:
        -         
        - https://www.arkansasedc.com/why-arkansas/business-climate/tax-structure/personal-income-tax
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_and_AR1000NR_Instructions.pdf
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_FullYearResidentIndividualIncomeTaxReturn.pdf
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_and_AR1000NR_Instructions.pdf
          https://www.arkansasedc.com/why-arkansas/business-climate/tax-structure/capital-gains-tax-reduction
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_and_AR1000NR_Instructions.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Ohio', 'state_occupation': 'Arkansas', "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def ohio_and_california(self, county_residence='Other', state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Ohio:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        California:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Ohio:
        - https://tax.ohio.gov/individual/resources/annual-tax-rates
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://tax.ohio.gov/static/forms/employer_withholding/generic/wth_it4nr.pdf
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://www.ritaohio.com/TaxRatesTable
        
        California:
        -         
        - https://www.ftb.ca.gov/forms/2022/2022-540-tax-rate-schedules.pdf
          https://www.ftb.ca.gov/file/personal/deductions/index.html
          https://www.ftb.ca.gov/about-ftb/newsroom/caleitc/eligibility-and-credit-information.html
          https://www.ftb.ca.gov/file/personal/income-types/capital-gains-and-losses.html#:~:text=California%20does%20not%20have%20a
         are%20taxed%20as%20ordinary%20income
          https://www.ftb.ca.gov/file/personal/
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Ohio', 'state_occupation': 'California', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def ohio_and_colorado(self, county_residence='Other', state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Ohio:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Colorado:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Ohio:
        - https://tax.ohio.gov/individual/resources/annual-tax-rates
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://tax.ohio.gov/static/forms/employer_withholding/generic/wth_it4nr.pdf
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://www.ritaohio.com/TaxRatesTable
        
        Colorado:
        -         
        - https://leg.colorado.gov/agencies/legislative-council-staff/individual-income-tax%C2%A0
          https://leg.colorado.gov/agencies/legislative-council-staff/individual-income-tax%C2%A0
          https://tax.colorado.gov/sites/tax/files/documents/ITT_Capital_Gain_Subtraction_Feb_2022.pdf
          https://www.auroragov.org/business_services/taxes/occupational_privilege_tax
          https://www.denvergov.org/Government/Agencies-Departments-Offices/Agencies-Departments-Offices-Directory/Department-of-Finance/Our-Divisions/Treasury/Business-Tax-Information#:~:text=Occupational%20privilege%20tax
         -%28show%20below%29&text=Each%20taxable%20employee%20is%20liable
         month%20for%20each%20taxable%20employee
          https://www.glendale.co.us/355/Occupational-Privilege-Tax
          https://www.greenwoodvillage.com/1220/Occupational-Privilege-Tax-OPT
          https://ci.sheridan.co.us/288/Occupational-Privilege-Tax
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Ohio', 'state_occupation': 'Colorado', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def ohio_and_connecticut(self, county_residence='Other', state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Ohio:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Connecticut:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Ohio:
        - https://tax.ohio.gov/individual/resources/annual-tax-rates
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://tax.ohio.gov/static/forms/employer_withholding/generic/wth_it4nr.pdf
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://www.ritaohio.com/TaxRatesTable
        
        Connecticut:
        -         
        - https://portal.ct.gov/-/media/DRS/Forms/2022/Income/2022-CT-1040-Instructions_1222.pdf
          https://www.cga.ct.gov/2022/rpt/pdf/2022-R-0108.pdf
          https://www.cga.ct.gov/current/pub/chap_229.htm#sec_12-703
          https://portal.ct.gov/-/media/DRS/Forms/2022/Income/2022-CT-1040-Instructions_1222.pdf
          "https://portal.ct.gov/DRS/Publications/TSSNs/TSSN-29#:~:text=An%20individuals%20net%20capital%20gains
         portion%20of%20Social%20Security%20benefits"
          "https://portal.ct.gov/DRS/Publications/TSSNs/TSSN-29#:~:text=An%20individuals%20net%20capital%20gains
         portion%20of%20Social%20Security%20benefits"
          https://portal.ct.gov/DRS/Individuals/Resident-Income-Tax/Tax-Information
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Ohio', 'state_occupation': 'Connecticut', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def ohio_and_delaware(self, county_residence='Other', state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, federal_agi=75000, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Ohio:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Delaware:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        federal_agi : int, optional
            Adjusted Gross Income (AGI) for federal taxes. Default is 0. Sources - https://www.irs.gov/e-file-providers/definition-of-adjusted-gross-income, https://www.irs.gov/newsroom/taxpayers-should-know-the-difference-between-standard-and-itemized-deductions
        income : int, optional
            Gross income. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Ohio:
        - https://tax.ohio.gov/individual/resources/annual-tax-rates
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://tax.ohio.gov/static/forms/employer_withholding/generic/wth_it4nr.pdf
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://www.ritaohio.com/TaxRatesTable
        
        Delaware:
        -         
        - https://revenue.delaware.gov/software-developer/tax-rate-changes/
          https://revenuefiles.delaware.gov/2022/PIT-RES_TY22_2022-01_PaperInteractive.pdf
          https://revenuefiles.delaware.gov/2022/PIT-RES_TY22_2022-02_Instructions.pdf
          https://revenue.delaware.gov/frequently-asked-questions/personal-income-tax-faqs/
          https://www.newcastlede.gov/DocumentCenter/View/20147/City-of-Wilmington-Wage-Tax-Information-PDF
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        federal_agi = self.federal_agi if hasattr(self, 'federal_agi') else federal_agi
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Ohio', 'state_occupation': 'Delaware', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "federal_agi": federal_agi, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def ohio_and_florida(self, county_residence='Other', state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Ohio:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Florida:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Ohio:
        - https://tax.ohio.gov/individual/resources/annual-tax-rates
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://tax.ohio.gov/static/forms/employer_withholding/generic/wth_it4nr.pdf
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://www.ritaohio.com/TaxRatesTable
        
        Florida:
        -         
        - https://floridarevenue.com/taxes/taxesfees/Pages/tax_interest_rates.aspx
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Ohio', 'state_occupation': 'Florida', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def ohio_and_georgia(self, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Ohio:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Georgia:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Ohio:
        - https://tax.ohio.gov/individual/resources/annual-tax-rates
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://tax.ohio.gov/static/forms/employer_withholding/generic/wth_it4nr.pdf
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://www.ritaohio.com/TaxRatesTable
        
        Georgia:
        -         
        - https://dor.georgia.gov/tax-tables-georgia-tax-rate-schedule
          https://apps.dor.ga.gov/FillableForms/PDFViewer/Index?form=2022GA500
          https://apps.dor.ga.gov/FillableForms/PDFViewer/Index?form=2022GA500
          https://apps.dor.ga.gov/FillableForms/PDFViewer/Index?form=2022GA500
          https://dor.georgia.gov/years-individual-income-tax-forms
          https://dor.georgia.gov/years-individual-income-tax-forms
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Ohio', 'state_occupation': 'Georgia', "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def ohio_and_hawaii(self, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Ohio:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Hawaii:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Ohio:
        - https://tax.ohio.gov/individual/resources/annual-tax-rates
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://tax.ohio.gov/static/forms/employer_withholding/generic/wth_it4nr.pdf
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://www.ritaohio.com/TaxRatesTable
        
        Hawaii:
        -         
        - https://tax.hawaii.gov/forms/d_18table-on/d_18table-on_p13/
          https://files.hawaii.gov/tax/forms/2018/18table-on.pdf
          https://files.hawaii.gov/tax/news/pubs/22outline.pdf
          https://files.hawaii.gov/tax/news/pubs/FastTaxRefGuide_2017.pdf
          https://www.tfhawaii.org/wordpress/state-tax-resources/mini-tax-guides/individual-income-tax-chapter-235/#:~:text=Long%20term%20capital%20gains%20are%20taxed%20at%20a%20maximum%20of%207.25%25
          https://www.grassrootinstitute.org/2023/02/hb337-increasing-tax-on-capital-gains-a-bad-risk/
          https://files.hawaii.gov/tax/news/pubs/22outline.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Ohio', 'state_occupation': 'Hawaii', "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def ohio_and_idaho(self, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Ohio:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Idaho:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Ohio:
        - https://tax.ohio.gov/individual/resources/annual-tax-rates
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://tax.ohio.gov/static/forms/employer_withholding/generic/wth_it4nr.pdf
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://www.ritaohio.com/TaxRatesTable
        
        Idaho:
        -         
        - https://tax.idaho.gov/taxes/income-tax/individual-income/individual-income-tax-rate-schedule/
          https://tax.idaho.gov/wp-content/uploads/forms/EIN00046/EIN00046_03-01-2023.pdf
          https://tax.idaho.gov/taxes/income-tax/individual-income/idaho-source-income/
          https://tax.idaho.gov/taxes/income-tax/individual-income/idaho-source-income/
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Ohio', 'state_occupation': 'Idaho', "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def ohio_and_illinois(self, county_residence='Other', state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Ohio:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Illinois:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Ohio:
        - https://tax.ohio.gov/individual/resources/annual-tax-rates
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://tax.ohio.gov/static/forms/employer_withholding/generic/wth_it4nr.pdf
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://www.ritaohio.com/TaxRatesTable
        
        Illinois:
        -         
        - https://tax.illinois.gov/research/taxrates/income.html#IndividualIncome
          https://tax.illinois.gov/content/dam/soi/en/web/tax/forms/incometax/documents/currentyear/individual/il-1040-instr.pdf
          https://tax.illinois.gov/questionsandanswers/12.html
          https://tax.illinois.gov/content/dam/soi/en/web/tax/forms/incometax/documents/currentyear/individual/il-1040-instr.pdf
          https://tax.illinois.gov/research/taxrates/personalproperty.html
          https://tax.illinois.gov/localgovernments/income.html
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Ohio', 'state_occupation': 'Illinois', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def ohio_and_indiana(self, county_residence='Other', state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Ohio:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Indiana:
        


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Ohio:
        - https://tax.ohio.gov/individual/resources/annual-tax-rates
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://tax.ohio.gov/static/forms/employer_withholding/generic/wth_it4nr.pdf
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://www.ritaohio.com/TaxRatesTable
        
        Indiana:
        -         
        - https://www.in.gov/dor/tax-forms/2022-individual-income-tax-forms/
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-32.htm
          https://www.in.gov/dor/files/ib28.pdf
          https://www.in.gov/dor/tax-forms/2022-individual-income-tax-forms/
          https://www.in.gov/dor/files/2022-county-tax-rates-and-codes.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Ohio', 'state_occupation': 'Indiana', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def ohio_and_iowa(self, county_residence='Other', state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Ohio:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Iowa:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Ohio:
        - https://tax.ohio.gov/individual/resources/annual-tax-rates
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://tax.ohio.gov/static/forms/employer_withholding/generic/wth_it4nr.pdf
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://www.ritaohio.com/TaxRatesTable
        
        Iowa:
        -         
        - https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
          https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
          https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
          https://tax.iowa.gov/iowa-illinois-reciprocal-agreement
          https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Ohio', 'state_occupation': 'Iowa', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def ohio_and_kansas(self, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Ohio:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Kansas:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Ohio:
        - https://tax.ohio.gov/individual/resources/annual-tax-rates
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://tax.ohio.gov/static/forms/employer_withholding/generic/wth_it4nr.pdf
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://www.ritaohio.com/TaxRatesTable
        
        Kansas:
        -         
        - https://www.ksrevenue.gov/taxrates.html
          https://www.ksrevenue.gov/incomebook22.html#0
          https://www.ksrevenue.gov/incomebook22.html#0
          https://www.ksrevenue.gov/incomebook22.html#0
          https://www.ksrevenue.gov/incomebook22.html
          https://www.ksrevenue.gov/incomebook22.html
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Ohio', 'state_occupation': 'Kansas', "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def ohio_and_kentucky(self, county_residence='Other', state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Ohio:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Kentucky:
        


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Ohio:
        - https://tax.ohio.gov/individual/resources/annual-tax-rates
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://tax.ohio.gov/static/forms/employer_withholding/generic/wth_it4nr.pdf
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://www.ritaohio.com/TaxRatesTable
        
        Kentucky:
        -         
        - https://revenue.ky.gov/News/Pages/Dor-Announces-Updates-to-Individual-Income-Tax-for-2023-Tax-Year.aspx
          https://revenue.ky.gov/Individual/Individual-Income-Tax/Pages/default.aspx
          https://revenue.ky.gov/News/Pages/Dor-Announces-Updates-to-Individual-Income-Tax-for-2023-Tax-Year.aspx
          https://revenue.ky.gov/Individual/Individual-Income-Tax/Pages/default.aspx
          https://revenue.ky.gov/Dor%20Training%20Materials/103%20KAR%2017.140.%20Individual%20income%20tax%20-%20reciprocity%20-%20nonresidents.pdf
          https://revenue.ky.gov/Individual/Individual-Income-Tax/Pages/default.aspx
          https://nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2021/Taxes-21-32.htm?taxmap=true
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Ohio', 'state_occupation': 'Kentucky', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def ohio_and_louisiana(self, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Ohio:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Louisiana:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Ohio:
        - https://tax.ohio.gov/individual/resources/annual-tax-rates
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://tax.ohio.gov/static/forms/employer_withholding/generic/wth_it4nr.pdf
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://www.ritaohio.com/TaxRatesTable
        
        Louisiana:
        -         
        - https://revenue.louisiana.gov/individualincometax
          https://revenue.louisiana.gov/individualincometax
          https://www.revenue.louisiana.gov/taxforms/6935(11_02)F.pdf
          https://revenue.louisiana.gov/individualincometax
          https://revenue.louisiana.gov/individualincometax
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Ohio', 'state_occupation': 'Louisiana', "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def ohio_and_maine(self, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Ohio:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Maine:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Ohio:
        - https://tax.ohio.gov/individual/resources/annual-tax-rates
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://tax.ohio.gov/static/forms/employer_withholding/generic/wth_it4nr.pdf
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://www.ritaohio.com/TaxRatesTable
        
        Maine:
        -         
        - https://www.maine.gov/revenue/tax-return-forms/individual-income-tax-2022
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/ind_tax_rate_sched_2022.pdf
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/22_1040me_book_gen_instruc_revisedFeb23.pdf
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/22_1040me_book_gen_instruc_revisedFeb23.pdf
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/22_1040me_book_gen_instruc_revisedFeb23.pdf
          https://www.maine.gov/revenue/tax-return-forms/individual-income-tax-2022
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Ohio', 'state_occupation': 'Maine', "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def ohio_and_maryland(self, county_residence='Other', state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Ohio:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Maryland:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Ohio:
        - https://tax.ohio.gov/individual/resources/annual-tax-rates
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://tax.ohio.gov/static/forms/employer_withholding/generic/wth_it4nr.pdf
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://www.ritaohio.com/TaxRatesTable
        
        Maryland:
        -         
        - https://www.marylandtaxes.gov/individual/income/tax-info/tax-rates.php
          https://www.marylandtaxes.gov/forms/22_forms/Resident_Booklet.pdf
          https://www.marylandtaxes.gov/forms/Tax_Publications/Administrative_Releases/Income_and_Estate_Tax_Releases/ar_it37.pdf
          https://www.marylandtaxes.gov/individual/index.php
          https://www.marylandtaxes.gov/individual/income/tax-info/tax-rates.php
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Ohio', 'state_occupation': 'Maryland', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def ohio_and_massachusetts(self, county_residence='Other', state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Ohio:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Massachusetts:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Ohio:
        - https://tax.ohio.gov/individual/resources/annual-tax-rates
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://tax.ohio.gov/static/forms/employer_withholding/generic/wth_it4nr.pdf
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://www.ritaohio.com/TaxRatesTable
        
        Massachusetts:
        -         
        - https://www.mass.gov/service-details/massachusetts-tax-rates
          https://www.mass.gov/service-details/view-massachusetts-personal-income-tax-exemptions
          https://www.mass.gov/service-details/view-massachusetts-personal-income-tax-exemptions
          https://www.mass.gov/service-details/view-child-and-dependent-related-credits
          https://www.mass.gov/service-details/massachusetts-tax-rates
          https://www.mass.gov/personal-income-tax
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Ohio', 'state_occupation': 'Massachusetts', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def ohio_and_michigan(self, county_residence='Other', state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Ohio:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Michigan:
        


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Ohio:
        - https://tax.ohio.gov/individual/resources/annual-tax-rates
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://tax.ohio.gov/static/forms/employer_withholding/generic/wth_it4nr.pdf
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://www.ritaohio.com/TaxRatesTable
        
        Michigan:
        -         
        - https://www.michigan.gov/taxes/iit/new-developments-for-tax-year-2022
          "https://www.michigan.gov/taxes/-/media/Project/Websites/taxes/Forms/2023/2023-SUW/446_Withholding-Guide_2023.pdf?rev=177ca3d8cb034392bf84e317ed78c320&hash=19C5F42F43A89A91863D229349F48883#:~:text=The%20exemption%20amount%20is%20%245%2C400
         employees%20Michigan%20income%20tax%20return"
          "https://www.michigan.gov/taxes/-/media/Project/Websites/taxes/Forms/2023/2023-SUW/446_Withholding-Guide_2023.pdf?rev=177ca3d8cb034392bf84e317ed78c320&hash=19C5F42F43A89A91863D229349F48883#:~:text=The%20exemption%20amount%20is%20%245%2C400
         employees%20Michigan%20income%20tax%20return"
          https://www.michigan.gov/taxes/questions/iit/accordion/residency/are-my-wages-earned-in-another-state-taxable-in-michigan-if-i-am-a-michigan-resident-1
          https://www.michigan.gov/taxes/-/media/Project/Websites/taxes/Forms/2022/2022-IIT-Forms/BOOK_MI-1040.pdf?rev=4524f1cbbc0243b194ab05a6680be75b&hash=40DFE71E651F7C1D65C31F6CD03D944F
          https://www.michigan.gov/taxes/questions/iit/accordion/general/what-cities-impose-an-income-tax
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Ohio', 'state_occupation': 'Michigan', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def ohio_and_minnesota(self, county_residence='Other', state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Ohio:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Minnesota:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Ohio:
        - https://tax.ohio.gov/individual/resources/annual-tax-rates
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://tax.ohio.gov/static/forms/employer_withholding/generic/wth_it4nr.pdf
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://www.ritaohio.com/TaxRatesTable
        
        Minnesota:
        -         
        - https://www.revenue.state.mn.us/minnesota-income-tax-rates-and-brackets
          https://www.revenue.state.mn.us/child-and-dependent-care-credit
          https://www.revenue.state.mn.us/dependent-exemptions
          https://www.revenue.state.mn.us/minnesota-standard-deduction
          https://mn.gov/mmb/assets/mwr_form_tcm1059-128581.pdf
          https://www.revenue.state.mn.us/individual-income-tax
          https://www.revenue.state.mn.us/individual-income-tax
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Ohio', 'state_occupation': 'Minnesota', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def ohio_and_mississippi(self, county_residence='Other', state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Ohio:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Mississippi:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Ohio:
        - https://tax.ohio.gov/individual/resources/annual-tax-rates
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://tax.ohio.gov/static/forms/employer_withholding/generic/wth_it4nr.pdf
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://www.ritaohio.com/TaxRatesTable
        
        Mississippi:
        -         
        - https://www.dor.ms.gov/individual/tax-rates
          https://www.dor.ms.gov/individual/tax-rates
          https://www.dor.ms.gov/individual/tax-rates
          https://www.dor.ms.gov/individual
          https://www.dor.ms.gov/individual
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Ohio', 'state_occupation': 'Mississippi', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def ohio_and_missouri(self, capital_gains_short=0, dependents=0, sources=False, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_long=0, filing_status='single', income=100000, credits_state=0):
        """
        --- Parameters ----
        
        Ohio:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Missouri:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Ohio:
        - https://tax.ohio.gov/individual/resources/annual-tax-rates
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://tax.ohio.gov/static/forms/employer_withholding/generic/wth_it4nr.pdf
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://www.ritaohio.com/TaxRatesTable
        
        Missouri:
        -         
        - https://dor.mo.gov/taxation/individual/tax-types/income/year-changes/
          https://dor.mo.gov/taxation/individual/tax-types/income/year-changes/
          https://dor.mo.gov/taxation/individual/tax-types/income/year-changes/
          https://www.kcmo.gov/city-hall/departments/finance/earnings-tax
          https://www.stlouis-mo.gov/government/departments/collector/earnings-tax/individual-earnings-tax-info.cfm
        """
        
        # Check if the parameters are set upstream in the instance
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Ohio', 'state_occupation': 'Missouri', "capital_gains_short": capital_gains_short, "dependents": dependents, "sources": sources, "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def ohio_and_montana(self, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Ohio:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Montana:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Ohio:
        - https://tax.ohio.gov/individual/resources/annual-tax-rates
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://tax.ohio.gov/static/forms/employer_withholding/generic/wth_it4nr.pdf
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://www.ritaohio.com/TaxRatesTable
        
        Montana:
        -         
        - https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/2022/12/montana-employees-withholding-allowance-and-exemption-certificate-form-mw-4.pdf
          https://mtrevenue.gov/2021/11/29/simplification-of-montana-income-taxation/?utm_source=rss&utm_medium=rss&utm_campaign=simplification-of-montana-income-taxation
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Ohio', 'state_occupation': 'Montana', "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def ohio_and_nebraska(self, county_residence='Other', state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Ohio:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Nebraska:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Ohio:
        - https://tax.ohio.gov/individual/resources/annual-tax-rates
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://tax.ohio.gov/static/forms/employer_withholding/generic/wth_it4nr.pdf
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://www.ritaohio.com/TaxRatesTable
        
        Nebraska:
        -         
        - https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2023_Circular_EN_8-429r2023_Whole_Book_11.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Ohio', 'state_occupation': 'Nebraska', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def ohio_and_nevada(self, county_residence='Other', state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Ohio:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Nevada:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Ohio:
        - https://tax.ohio.gov/individual/resources/annual-tax-rates
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://tax.ohio.gov/static/forms/employer_withholding/generic/wth_it4nr.pdf
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://www.ritaohio.com/TaxRatesTable
        
        Nevada:
        -         
        - https://www.nvsos.gov/sos/businesses/the-nevada-advantage
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Ohio', 'state_occupation': 'Nevada', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def ohio_and_new_hampshire(self, county_residence='Other', state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Ohio:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        New Hampshire:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Ohio:
        - https://tax.ohio.gov/individual/resources/annual-tax-rates
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://tax.ohio.gov/static/forms/employer_withholding/generic/wth_it4nr.pdf
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://www.ritaohio.com/TaxRatesTable
        
        New Hampshire:
        -         
        - https://www.revenue.nh.gov/assistance/tax-overview.htm
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Ohio', 'state_occupation': 'NewHampshire', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def ohio_and_new_jersey(self, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Ohio:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        New Jersey:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Ohio:
        - https://tax.ohio.gov/individual/resources/annual-tax-rates
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://tax.ohio.gov/static/forms/employer_withholding/generic/wth_it4nr.pdf
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://www.ritaohio.com/TaxRatesTable
        
        New Jersey:
        -         
        - https://www.state.nj.us/treasury/taxation/taxtables.shtml
          https://www.state.nj.us/treasury/taxation/njit13.shtml
          https://www.nj.gov/treasury/taxation/njit25.shtml
          https://www.state.nj.us/treasury/taxation/individuals/freqqiti.shtml
          https://www.state.nj.us/treasury/taxation/localtax.shtml
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Ohio', 'state_occupation': 'NewJersey', "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def ohio_and_new_mexico(self, county_residence='Other', state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Ohio:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        New Mexico:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Ohio:
        - https://tax.ohio.gov/individual/resources/annual-tax-rates
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://tax.ohio.gov/static/forms/employer_withholding/generic/wth_it4nr.pdf
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://www.ritaohio.com/TaxRatesTable
        
        New Mexico:
        -         
        - https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/fdf3c548-8aba-4b9c-9eb4-bb564c716015/FYI-104.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/0bf943fd-652e-4400-bb1b-cd8397eb5a95/2022pit-adj.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/2f1a6781-9534-4436-b427-1557f9592099/2022pit-adj-ins.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/0bf943fd-652e-4400-bb1b-cd8397eb5a95/2022pit-adj.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/1afc56af-ea90-4d48-82e5-1f9aeb43255a/PITbook2022.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/fdf3c548-8aba-4b9c-9eb4-bb564c716015/FYI-104.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Ohio', 'state_occupation': 'NewMexico', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def ohio_and_new_york(self, county_residence='Other', state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Ohio:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        New York:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Ohio:
        - https://tax.ohio.gov/individual/resources/annual-tax-rates
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://tax.ohio.gov/static/forms/employer_withholding/generic/wth_it4nr.pdf
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://www.ritaohio.com/TaxRatesTable
        
        New York:
        -         
        - https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nys-tax-rate-schedule
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nys-tax-rate-schedule
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nys-tax-rate-schedule
          https://www.tax.ny.gov/pit/
          https://www.tax.ny.gov/pit/file/nonresident-faqs.htm
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#yonkers-tax-rate-schedule
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nyc-tax-rate-schedule
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Ohio', 'state_occupation': 'NewYork', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def ohio_and_north_carolina(self, county_residence='Other', state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Ohio:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        North Carolina:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Ohio:
        - https://tax.ohio.gov/individual/resources/annual-tax-rates
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://tax.ohio.gov/static/forms/employer_withholding/generic/wth_it4nr.pdf
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://www.ritaohio.com/TaxRatesTable
        
        North Carolina:
        -         
        - https://www.ncdor.gov/taxes-forms/tax-rate-schedules
          https://www.ncdor.gov/taxes-forms/individual-income-tax/north-carolina-standard-deduction-or-north-carolina-itemized-deductions
          https://www.ncdor.gov/taxes-forms/individual-income-tax
          https://www.ncdor.gov/taxes-forms/individual-income-tax
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Ohio', 'state_occupation': 'NorthCarolina', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def ohio_and_north_dakota(self, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Ohio:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        North Dakota:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Ohio:
        - https://tax.ohio.gov/individual/resources/annual-tax-rates
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://tax.ohio.gov/static/forms/employer_withholding/generic/wth_it4nr.pdf
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://www.ritaohio.com/TaxRatesTable
        
        North Dakota:
        -         
        - https://www.tax.nd.gov/forms
          https://www.tax.nd.gov/sites/www/files/documents/forms/individual/2022-iit/2022-individual-income-tax-booklet.pdf
          https://www.tax.nd.gov/sites/www/files/documents/forms/2022-individual-income-tax-booklet.pdf
          https://www.tax.nd.gov/tax-exemptions-credits/income-tax-exemptions-credits
          https://www.tax.nd.gov/sites/www/files/documents/forms/business/ndwrfillable.pdf
          https://www.tax.nd.gov/news/tax-legislative-changes/significant-changes-law/individual-income-tax-history
          https://www.tax.nd.gov/forms
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Ohio', 'state_occupation': 'NorthDakota', "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def ohio_and_oklahoma(self, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Ohio:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Oklahoma:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Ohio:
        - https://tax.ohio.gov/individual/resources/annual-tax-rates
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://tax.ohio.gov/static/forms/employer_withholding/generic/wth_it4nr.pdf
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://www.ritaohio.com/TaxRatesTable
        
        Oklahoma:
        -         
        - https://oklahoma.gov/content/dam/ok/en/tax/documents/resources/publications/businesses/withholding-tables/WHTables-2022.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Ohio', 'state_occupation': 'Oklahoma', "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def ohio_and_oregon(self, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Ohio:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Oregon:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Ohio:
        - https://tax.ohio.gov/individual/resources/annual-tax-rates
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://tax.ohio.gov/static/forms/employer_withholding/generic/wth_it4nr.pdf
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://www.ritaohio.com/TaxRatesTable
        
        Oregon:
        -         
        - https://www.oregon.gov/dor/programs/individuals/Pages/PIT.aspx
          https://www.oregon.gov/dor/programs/individuals/Documents/Full%20year%20resident
         %20Form%20or-40%20filers.pdf
          https://www.oregon.gov/dor/programs/individuals/Documents/Part-year%20and%20nonresident
         %20Form%20or-40-P%20and%20or-40-N%20filers.pdf.pdf
          https://www.oregon.gov/dor/forms/FormsPubs/form-or-40_101-040_2022.pdf
          https://www.oregon.gov/dor/forms/FormsPubs/form-or-40_101-040_2022.pdf
          https://www.oregon.gov/dor/forms/FormsPubs/form-or-40_101-040_2022.pdf
          https://www.portland.gov/revenue/personal-tax
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Ohio', 'state_occupation': 'Oregon', "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def ohio_and_pennsylvania(self, county_residence='Other', state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Ohio:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Pennsylvania:
        


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Ohio:
        - https://tax.ohio.gov/individual/resources/annual-tax-rates
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://tax.ohio.gov/static/forms/employer_withholding/generic/wth_it4nr.pdf
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://www.ritaohio.com/TaxRatesTable
        
        Pennsylvania:
        -         
        - https://www.revenue.pa.gov/Tax%20Rates/Pages/default.aspx
          https://www.revenue.pa.gov/FormsandPublications/FormsforIndividuals/PIT/Documents/2022/2022_pa-40in.pdf
          https://www.revenue.pa.gov/FormsandPublications/FormsforBusinesses/EmployerWithholding/Documents/rev-419.pdf
          https://www.revenue.pa.gov/FormsandPublications/FormsforIndividuals/PIT/Documents/2022/2022_pa-40in.pdf
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-25.htm
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Ohio', 'state_occupation': 'Pennsylvania', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def ohio_and_rhode_island(self, county_residence='Other', state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Ohio:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Rhode Island:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Ohio:
        - https://tax.ohio.gov/individual/resources/annual-tax-rates
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://tax.ohio.gov/static/forms/employer_withholding/generic/wth_it4nr.pdf
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://www.ritaohio.com/TaxRatesTable
        
        Rhode Island:
        -         
        - https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2022_1040WE_w_0.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Ohio', 'state_occupation': 'RhodeIsland', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def ohio_and_south_carolina(self, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Ohio:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        South Carolina:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Ohio:
        - https://tax.ohio.gov/individual/resources/annual-tax-rates
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://tax.ohio.gov/static/forms/employer_withholding/generic/wth_it4nr.pdf
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://www.ritaohio.com/TaxRatesTable
        
        South Carolina:
        -         
        - https://dor.sc.gov/resources-site/lawandpolicy/Advisory%20Opinions/IL22-15.pdf
          https://dor.sc.gov/forms-site/Forms/IITPacket_2022.pdf
          https://www.scstatehouse.gov/code/t12c006.php
          https://dor.sc.gov/resources-site/lawandpolicy/Advisory%20Opinions/IL22-15.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Ohio', 'state_occupation': 'SouthCarolina', "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def ohio_and_south_dakota(self, county_residence='Other', state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Ohio:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        South Dakota:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Ohio:
        - https://tax.ohio.gov/individual/resources/annual-tax-rates
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://tax.ohio.gov/static/forms/employer_withholding/generic/wth_it4nr.pdf
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://www.ritaohio.com/TaxRatesTable
        
        South Dakota:
        -         
        - https://dor.sd.gov/individuals/taxes/
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Ohio', 'state_occupation': 'SouthDakota', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def ohio_and_tennessee(self, county_residence='Other', state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Ohio:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Tennessee:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Ohio:
        - https://tax.ohio.gov/individual/resources/annual-tax-rates
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://tax.ohio.gov/static/forms/employer_withholding/generic/wth_it4nr.pdf
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://www.ritaohio.com/TaxRatesTable
        
        Tennessee:
        -         
        - https://www.tn.gov/revenue/taxes.html
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Ohio', 'state_occupation': 'Tennessee', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def ohio_and_texas(self, county_residence='Other', state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Ohio:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Texas:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Ohio:
        - https://tax.ohio.gov/individual/resources/annual-tax-rates
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://tax.ohio.gov/static/forms/employer_withholding/generic/wth_it4nr.pdf
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://www.ritaohio.com/TaxRatesTable
        
        Texas:
        -         
        - https://comptroller.texas.gov/taxes/
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Ohio', 'state_occupation': 'Texas', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def ohio_and_utah(self, county_residence='Other', state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Ohio:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Utah:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Ohio:
        - https://tax.ohio.gov/individual/resources/annual-tax-rates
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://tax.ohio.gov/static/forms/employer_withholding/generic/wth_it4nr.pdf
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://www.ritaohio.com/TaxRatesTable
        
        Utah:
        -         
        - https://incometax.utah.gov/paying/tax-rates
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Ohio', 'state_occupation': 'Utah', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def ohio_and_vermont(self, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Ohio:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Vermont:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Ohio:
        - https://tax.ohio.gov/individual/resources/annual-tax-rates
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://tax.ohio.gov/static/forms/employer_withholding/generic/wth_it4nr.pdf
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://www.ritaohio.com/TaxRatesTable
        
        Vermont:
        -         
        - https://tax.vermont.gov/sites/tax/files/documents/GB-1210-2023.pdf
          https://tax.vermont.gov/sites/tax/files/documents/Income%20Booklet-2022.pdf
          https://tax.vermont.gov/sites/tax/files/documents/Income%20Booklet-2022.pdf
          https://tax.vermont.gov/sites/tax/files/documents/Income%20Booklet-2022.pdf
          https://tax.vermont.gov/individuals/personal-income-tax/tax-credits
          https://tax.vermont.gov/individuals/personal-income-tax/taxable-income
          https://tax.vermont.gov/individuals/personal-income-tax/local-option-tax
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Ohio', 'state_occupation': 'Vermont', "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def ohio_and_virginia(self, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Ohio:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Virginia:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Ohio:
        - https://tax.ohio.gov/individual/resources/annual-tax-rates
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://tax.ohio.gov/static/forms/employer_withholding/generic/wth_it4nr.pdf
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://www.ritaohio.com/TaxRatesTable
        
        Virginia:
        -         
        - https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/reciprocity
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Ohio', 'state_occupation': 'Virginia', "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def ohio_and_washington(self, county_residence='Other', state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Ohio:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Washington:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Ohio:
        - https://tax.ohio.gov/individual/resources/annual-tax-rates
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://tax.ohio.gov/static/forms/employer_withholding/generic/wth_it4nr.pdf
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://www.ritaohio.com/TaxRatesTable
        
        Washington:
        -         
        - https://dor.wa.gov/taxes-rates/income-tax
          https://dor.wa.gov/taxes-rates/other-taxes/capital-gains-tax
          https://dor.wa.gov/taxes-rates/income-tax
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Ohio', 'state_occupation': 'Washington', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def ohio_and_west_virginia(self, pay_periods=24, capital_gains_short=0, county_occupation='Other', dependents=0, sources=False, county_residence='Other', state_agi=75000, capital_gains_long=0, filing_status='single', income=100000, credits_state=0):
        """
        --- Parameters ----
        
        Ohio:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        West Virginia:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_occupation : str, optional
            County of occupation. Default is Other.
        pay_periods : int, optional
            Number of pay periods. Options are 13, 24, 26, 52. Default is 24.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Ohio:
        - https://tax.ohio.gov/individual/resources/annual-tax-rates
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://tax.ohio.gov/static/forms/employer_withholding/generic/wth_it4nr.pdf
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://www.ritaohio.com/TaxRatesTable
        
        West Virginia:
        -         
        - https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://tax.wv.gov/Documents/TSD/tsd438.pdf
          https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2011/Taxes-11-35att.html
        """
        
        # Check if the parameters are set upstream in the instance
        pay_periods = self.pay_periods if hasattr(self, 'pay_periods') else pay_periods
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        county_occupation = self.county_occupation if hasattr(self, 'county_occupation') else county_occupation
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Ohio', 'state_occupation': 'WestVirginia', "pay_periods": pay_periods, "capital_gains_short": capital_gains_short, "county_occupation": county_occupation, "dependents": dependents, "sources": sources, "county_residence": county_residence, "state_agi": state_agi, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def ohio_and_wisconsin(self, county_residence='Other', state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Ohio:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Wisconsin:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Ohio:
        - https://tax.ohio.gov/individual/resources/annual-tax-rates
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://tax.ohio.gov/static/forms/employer_withholding/generic/wth_it4nr.pdf
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://www.ritaohio.com/TaxRatesTable
        
        Wisconsin:
        -         
        - https://www.revenue.wi.gov/TaxForms2023/2023-Form1-ES-Inst.pdf
          https://www.revenue.wi.gov/TaxForms2023/2023-Form1-ES-Inst.pdf
          https://www.revenue.wi.gov/TaxForms2023/2023-Form1-ES-Inst.pdf
          https://www.revenue.wi.gov/Pages/FAQS/pcs-work.aspx
          https://www.revenue.wi.gov/Dor%20Publications/pb103.pdf
          https://www.revenue.wi.gov/Pages/FAQS/pcs-work.aspx
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Ohio', 'state_occupation': 'Wisconsin', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def ohio_and_wyoming(self, county_residence='Other', state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Ohio:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Wyoming:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Ohio:
        - https://tax.ohio.gov/individual/resources/annual-tax-rates
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://tax.ohio.gov/static/forms/employer_withholding/generic/wth_it4nr.pdf
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://www.ritaohio.com/TaxRatesTable
        
        Wyoming:
        -         
        - https://revenue.wyo.gov/
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Ohio', 'state_occupation': 'Wyoming', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def oklahoma_and_alabama(self, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Oklahoma:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Alabama:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Oklahoma:
        - https://oklahoma.gov/content/dam/ok/en/tax/documents/resources/publications/businesses/withholding-tables/WHTables-2022.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
        
        Alabama:
        -         
        - https://www.revenue.alabama.gov/ultraviewer/viewer/basic_viewer/index.html?form=2023/02/22f40.pdf
          https://www.revenue.alabama.gov/faqs/what-is-alabamas-individual-income-tax-rate/
          https://www.revenue.alabama.gov/ultraviewer/viewer/basic_viewer/index.html?form=2023/02/22f40.pdf
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-24.htm?taxmap=true
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-24.htm?taxmap=true
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-24.htm?taxmap=true
          https://www.revenue.alabama.gov/individual-corporate/income-to-be-reported-on-the-alabama-income-tax-return/
          https://www.almonline.org/TaxRates.aspx#OccupationalTax
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Oklahoma', 'state_occupation': 'Alabama', "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def oklahoma_and_alaska(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Oklahoma:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Alaska:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Oklahoma:
        - https://oklahoma.gov/content/dam/ok/en/tax/documents/resources/publications/businesses/withholding-tables/WHTables-2022.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
        
        Alaska:
        -         
        - https://www.commerce.alaska.gov/web/dcra/officeofthestateassessor/alaskataxfacts.aspx
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Oklahoma', 'state_occupation': 'Alaska', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def oklahoma_and_arizona(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Oklahoma:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Arizona:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        credits_state : int, optional
            Credits for state taxes. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Oklahoma:
        - https://oklahoma.gov/content/dam/ok/en/tax/documents/resources/publications/businesses/withholding-tables/WHTables-2022.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
        
        Arizona:
        -         
        - https://azdor.gov/forms/individual/form-140-x-y-tables
          https://azdor.gov/forms/individual/form-140-x-y-tables
          https://azdor.gov/forms/individual/form-140-x-y-tables
          https://azdor.gov/businesses-arizona/withholding-tax/withholding-exceptions
          "https://azdor.gov/individuals/income-tax-filing-assistance/identifying-other-taxable-income#:~:text=In%20regards%20to%20capital%20gains
         the%20individuals%20regular%20tax%20rate.
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Oklahoma', 'state_occupation': 'Arizona', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def oklahoma_and_arkansas(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Oklahoma:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Arkansas:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Oklahoma:
        - https://oklahoma.gov/content/dam/ok/en/tax/documents/resources/publications/businesses/withholding-tables/WHTables-2022.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
        
        Arkansas:
        -         
        - https://www.arkansasedc.com/why-arkansas/business-climate/tax-structure/personal-income-tax
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_and_AR1000NR_Instructions.pdf
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_FullYearResidentIndividualIncomeTaxReturn.pdf
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_and_AR1000NR_Instructions.pdf
          https://www.arkansasedc.com/why-arkansas/business-climate/tax-structure/capital-gains-tax-reduction
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_and_AR1000NR_Instructions.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Oklahoma', 'state_occupation': 'Arkansas', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def oklahoma_and_california(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Oklahoma:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        California:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Oklahoma:
        - https://oklahoma.gov/content/dam/ok/en/tax/documents/resources/publications/businesses/withholding-tables/WHTables-2022.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
        
        California:
        -         
        - https://www.ftb.ca.gov/forms/2022/2022-540-tax-rate-schedules.pdf
          https://www.ftb.ca.gov/file/personal/deductions/index.html
          https://www.ftb.ca.gov/about-ftb/newsroom/caleitc/eligibility-and-credit-information.html
          https://www.ftb.ca.gov/file/personal/income-types/capital-gains-and-losses.html#:~:text=California%20does%20not%20have%20a
         are%20taxed%20as%20ordinary%20income
          https://www.ftb.ca.gov/file/personal/
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Oklahoma', 'state_occupation': 'California', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def oklahoma_and_colorado(self, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Oklahoma:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Colorado:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Oklahoma:
        - https://oklahoma.gov/content/dam/ok/en/tax/documents/resources/publications/businesses/withholding-tables/WHTables-2022.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
        
        Colorado:
        -         
        - https://leg.colorado.gov/agencies/legislative-council-staff/individual-income-tax%C2%A0
          https://leg.colorado.gov/agencies/legislative-council-staff/individual-income-tax%C2%A0
          https://tax.colorado.gov/sites/tax/files/documents/ITT_Capital_Gain_Subtraction_Feb_2022.pdf
          https://www.auroragov.org/business_services/taxes/occupational_privilege_tax
          https://www.denvergov.org/Government/Agencies-Departments-Offices/Agencies-Departments-Offices-Directory/Department-of-Finance/Our-Divisions/Treasury/Business-Tax-Information#:~:text=Occupational%20privilege%20tax
         -%28show%20below%29&text=Each%20taxable%20employee%20is%20liable
         month%20for%20each%20taxable%20employee
          https://www.glendale.co.us/355/Occupational-Privilege-Tax
          https://www.greenwoodvillage.com/1220/Occupational-Privilege-Tax-OPT
          https://ci.sheridan.co.us/288/Occupational-Privilege-Tax
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Oklahoma', 'state_occupation': 'Colorado', "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def oklahoma_and_connecticut(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Oklahoma:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Connecticut:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Oklahoma:
        - https://oklahoma.gov/content/dam/ok/en/tax/documents/resources/publications/businesses/withholding-tables/WHTables-2022.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
        
        Connecticut:
        -         
        - https://portal.ct.gov/-/media/DRS/Forms/2022/Income/2022-CT-1040-Instructions_1222.pdf
          https://www.cga.ct.gov/2022/rpt/pdf/2022-R-0108.pdf
          https://www.cga.ct.gov/current/pub/chap_229.htm#sec_12-703
          https://portal.ct.gov/-/media/DRS/Forms/2022/Income/2022-CT-1040-Instructions_1222.pdf
          "https://portal.ct.gov/DRS/Publications/TSSNs/TSSN-29#:~:text=An%20individuals%20net%20capital%20gains
         portion%20of%20Social%20Security%20benefits"
          "https://portal.ct.gov/DRS/Publications/TSSNs/TSSN-29#:~:text=An%20individuals%20net%20capital%20gains
         portion%20of%20Social%20Security%20benefits"
          https://portal.ct.gov/DRS/Individuals/Resident-Income-Tax/Tax-Information
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Oklahoma', 'state_occupation': 'Connecticut', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def oklahoma_and_delaware(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, federal_agi=75000, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Oklahoma:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Delaware:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        federal_agi : int, optional
            Adjusted Gross Income (AGI) for federal taxes. Default is 0. Sources - https://www.irs.gov/e-file-providers/definition-of-adjusted-gross-income, https://www.irs.gov/newsroom/taxpayers-should-know-the-difference-between-standard-and-itemized-deductions
        income : int, optional
            Gross income. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Oklahoma:
        - https://oklahoma.gov/content/dam/ok/en/tax/documents/resources/publications/businesses/withholding-tables/WHTables-2022.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
        
        Delaware:
        -         
        - https://revenue.delaware.gov/software-developer/tax-rate-changes/
          https://revenuefiles.delaware.gov/2022/PIT-RES_TY22_2022-01_PaperInteractive.pdf
          https://revenuefiles.delaware.gov/2022/PIT-RES_TY22_2022-02_Instructions.pdf
          https://revenue.delaware.gov/frequently-asked-questions/personal-income-tax-faqs/
          https://www.newcastlede.gov/DocumentCenter/View/20147/City-of-Wilmington-Wage-Tax-Information-PDF
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        federal_agi = self.federal_agi if hasattr(self, 'federal_agi') else federal_agi
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Oklahoma', 'state_occupation': 'Delaware', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "federal_agi": federal_agi, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def oklahoma_and_florida(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Oklahoma:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Florida:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Oklahoma:
        - https://oklahoma.gov/content/dam/ok/en/tax/documents/resources/publications/businesses/withholding-tables/WHTables-2022.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
        
        Florida:
        -         
        - https://floridarevenue.com/taxes/taxesfees/Pages/tax_interest_rates.aspx
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Oklahoma', 'state_occupation': 'Florida', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def oklahoma_and_georgia(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Oklahoma:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Georgia:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Oklahoma:
        - https://oklahoma.gov/content/dam/ok/en/tax/documents/resources/publications/businesses/withholding-tables/WHTables-2022.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
        
        Georgia:
        -         
        - https://dor.georgia.gov/tax-tables-georgia-tax-rate-schedule
          https://apps.dor.ga.gov/FillableForms/PDFViewer/Index?form=2022GA500
          https://apps.dor.ga.gov/FillableForms/PDFViewer/Index?form=2022GA500
          https://apps.dor.ga.gov/FillableForms/PDFViewer/Index?form=2022GA500
          https://dor.georgia.gov/years-individual-income-tax-forms
          https://dor.georgia.gov/years-individual-income-tax-forms
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Oklahoma', 'state_occupation': 'Georgia', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def oklahoma_and_hawaii(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Oklahoma:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Hawaii:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Oklahoma:
        - https://oklahoma.gov/content/dam/ok/en/tax/documents/resources/publications/businesses/withholding-tables/WHTables-2022.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
        
        Hawaii:
        -         
        - https://tax.hawaii.gov/forms/d_18table-on/d_18table-on_p13/
          https://files.hawaii.gov/tax/forms/2018/18table-on.pdf
          https://files.hawaii.gov/tax/news/pubs/22outline.pdf
          https://files.hawaii.gov/tax/news/pubs/FastTaxRefGuide_2017.pdf
          https://www.tfhawaii.org/wordpress/state-tax-resources/mini-tax-guides/individual-income-tax-chapter-235/#:~:text=Long%20term%20capital%20gains%20are%20taxed%20at%20a%20maximum%20of%207.25%25
          https://www.grassrootinstitute.org/2023/02/hb337-increasing-tax-on-capital-gains-a-bad-risk/
          https://files.hawaii.gov/tax/news/pubs/22outline.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Oklahoma', 'state_occupation': 'Hawaii', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def oklahoma_and_idaho(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Oklahoma:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Idaho:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Oklahoma:
        - https://oklahoma.gov/content/dam/ok/en/tax/documents/resources/publications/businesses/withholding-tables/WHTables-2022.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
        
        Idaho:
        -         
        - https://tax.idaho.gov/taxes/income-tax/individual-income/individual-income-tax-rate-schedule/
          https://tax.idaho.gov/wp-content/uploads/forms/EIN00046/EIN00046_03-01-2023.pdf
          https://tax.idaho.gov/taxes/income-tax/individual-income/idaho-source-income/
          https://tax.idaho.gov/taxes/income-tax/individual-income/idaho-source-income/
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Oklahoma', 'state_occupation': 'Idaho', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def oklahoma_and_illinois(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Oklahoma:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Illinois:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Oklahoma:
        - https://oklahoma.gov/content/dam/ok/en/tax/documents/resources/publications/businesses/withholding-tables/WHTables-2022.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
        
        Illinois:
        -         
        - https://tax.illinois.gov/research/taxrates/income.html#IndividualIncome
          https://tax.illinois.gov/content/dam/soi/en/web/tax/forms/incometax/documents/currentyear/individual/il-1040-instr.pdf
          https://tax.illinois.gov/questionsandanswers/12.html
          https://tax.illinois.gov/content/dam/soi/en/web/tax/forms/incometax/documents/currentyear/individual/il-1040-instr.pdf
          https://tax.illinois.gov/research/taxrates/personalproperty.html
          https://tax.illinois.gov/localgovernments/income.html
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Oklahoma', 'state_occupation': 'Illinois', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def oklahoma_and_indiana(self, capital_gains_short=0, dependents=0, sources=False, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_long=0, filing_status='single', income=100000, credits_state=0):
        """
        --- Parameters ----
        
        Oklahoma:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Indiana:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Oklahoma:
        - https://oklahoma.gov/content/dam/ok/en/tax/documents/resources/publications/businesses/withholding-tables/WHTables-2022.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
        
        Indiana:
        -         
        - https://www.in.gov/dor/tax-forms/2022-individual-income-tax-forms/
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-32.htm
          https://www.in.gov/dor/files/ib28.pdf
          https://www.in.gov/dor/tax-forms/2022-individual-income-tax-forms/
          https://www.in.gov/dor/files/2022-county-tax-rates-and-codes.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Oklahoma', 'state_occupation': 'Indiana', "capital_gains_short": capital_gains_short, "dependents": dependents, "sources": sources, "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def oklahoma_and_iowa(self, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Oklahoma:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Iowa:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Oklahoma:
        - https://oklahoma.gov/content/dam/ok/en/tax/documents/resources/publications/businesses/withholding-tables/WHTables-2022.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
        
        Iowa:
        -         
        - https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
          https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
          https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
          https://tax.iowa.gov/iowa-illinois-reciprocal-agreement
          https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Oklahoma', 'state_occupation': 'Iowa', "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def oklahoma_and_kansas(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Oklahoma:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Kansas:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Oklahoma:
        - https://oklahoma.gov/content/dam/ok/en/tax/documents/resources/publications/businesses/withholding-tables/WHTables-2022.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
        
        Kansas:
        -         
        - https://www.ksrevenue.gov/taxrates.html
          https://www.ksrevenue.gov/incomebook22.html#0
          https://www.ksrevenue.gov/incomebook22.html#0
          https://www.ksrevenue.gov/incomebook22.html#0
          https://www.ksrevenue.gov/incomebook22.html
          https://www.ksrevenue.gov/incomebook22.html
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Oklahoma', 'state_occupation': 'Kansas', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def oklahoma_and_kentucky(self, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Oklahoma:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Kentucky:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Oklahoma:
        - https://oklahoma.gov/content/dam/ok/en/tax/documents/resources/publications/businesses/withholding-tables/WHTables-2022.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
        
        Kentucky:
        -         
        - https://revenue.ky.gov/News/Pages/Dor-Announces-Updates-to-Individual-Income-Tax-for-2023-Tax-Year.aspx
          https://revenue.ky.gov/Individual/Individual-Income-Tax/Pages/default.aspx
          https://revenue.ky.gov/News/Pages/Dor-Announces-Updates-to-Individual-Income-Tax-for-2023-Tax-Year.aspx
          https://revenue.ky.gov/Individual/Individual-Income-Tax/Pages/default.aspx
          https://revenue.ky.gov/Dor%20Training%20Materials/103%20KAR%2017.140.%20Individual%20income%20tax%20-%20reciprocity%20-%20nonresidents.pdf
          https://revenue.ky.gov/Individual/Individual-Income-Tax/Pages/default.aspx
          https://nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2021/Taxes-21-32.htm?taxmap=true
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Oklahoma', 'state_occupation': 'Kentucky', "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def oklahoma_and_louisiana(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Oklahoma:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Louisiana:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Oklahoma:
        - https://oklahoma.gov/content/dam/ok/en/tax/documents/resources/publications/businesses/withholding-tables/WHTables-2022.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
        
        Louisiana:
        -         
        - https://revenue.louisiana.gov/individualincometax
          https://revenue.louisiana.gov/individualincometax
          https://www.revenue.louisiana.gov/taxforms/6935(11_02)F.pdf
          https://revenue.louisiana.gov/individualincometax
          https://revenue.louisiana.gov/individualincometax
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Oklahoma', 'state_occupation': 'Louisiana', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def oklahoma_and_maine(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Oklahoma:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Maine:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Oklahoma:
        - https://oklahoma.gov/content/dam/ok/en/tax/documents/resources/publications/businesses/withholding-tables/WHTables-2022.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
        
        Maine:
        -         
        - https://www.maine.gov/revenue/tax-return-forms/individual-income-tax-2022
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/ind_tax_rate_sched_2022.pdf
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/22_1040me_book_gen_instruc_revisedFeb23.pdf
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/22_1040me_book_gen_instruc_revisedFeb23.pdf
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/22_1040me_book_gen_instruc_revisedFeb23.pdf
          https://www.maine.gov/revenue/tax-return-forms/individual-income-tax-2022
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Oklahoma', 'state_occupation': 'Maine', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def oklahoma_and_maryland(self, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Oklahoma:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Maryland:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Oklahoma:
        - https://oklahoma.gov/content/dam/ok/en/tax/documents/resources/publications/businesses/withholding-tables/WHTables-2022.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
        
        Maryland:
        -         
        - https://www.marylandtaxes.gov/individual/income/tax-info/tax-rates.php
          https://www.marylandtaxes.gov/forms/22_forms/Resident_Booklet.pdf
          https://www.marylandtaxes.gov/forms/Tax_Publications/Administrative_Releases/Income_and_Estate_Tax_Releases/ar_it37.pdf
          https://www.marylandtaxes.gov/individual/index.php
          https://www.marylandtaxes.gov/individual/income/tax-info/tax-rates.php
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Oklahoma', 'state_occupation': 'Maryland', "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def oklahoma_and_massachusetts(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Oklahoma:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Massachusetts:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Oklahoma:
        - https://oklahoma.gov/content/dam/ok/en/tax/documents/resources/publications/businesses/withholding-tables/WHTables-2022.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
        
        Massachusetts:
        -         
        - https://www.mass.gov/service-details/massachusetts-tax-rates
          https://www.mass.gov/service-details/view-massachusetts-personal-income-tax-exemptions
          https://www.mass.gov/service-details/view-massachusetts-personal-income-tax-exemptions
          https://www.mass.gov/service-details/view-child-and-dependent-related-credits
          https://www.mass.gov/service-details/massachusetts-tax-rates
          https://www.mass.gov/personal-income-tax
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Oklahoma', 'state_occupation': 'Massachusetts', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def oklahoma_and_michigan(self, capital_gains_short=0, county_occupation='Other', dependents=0, sources=False, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_long=0, filing_status='single', income=100000, credits_state=0):
        """
        --- Parameters ----
        
        Oklahoma:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Michigan:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        county_occupation : str, optional
            County of occupation. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Oklahoma:
        - https://oklahoma.gov/content/dam/ok/en/tax/documents/resources/publications/businesses/withholding-tables/WHTables-2022.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
        
        Michigan:
        -         
        - https://www.michigan.gov/taxes/iit/new-developments-for-tax-year-2022
          "https://www.michigan.gov/taxes/-/media/Project/Websites/taxes/Forms/2023/2023-SUW/446_Withholding-Guide_2023.pdf?rev=177ca3d8cb034392bf84e317ed78c320&hash=19C5F42F43A89A91863D229349F48883#:~:text=The%20exemption%20amount%20is%20%245%2C400
         employees%20Michigan%20income%20tax%20return"
          "https://www.michigan.gov/taxes/-/media/Project/Websites/taxes/Forms/2023/2023-SUW/446_Withholding-Guide_2023.pdf?rev=177ca3d8cb034392bf84e317ed78c320&hash=19C5F42F43A89A91863D229349F48883#:~:text=The%20exemption%20amount%20is%20%245%2C400
         employees%20Michigan%20income%20tax%20return"
          https://www.michigan.gov/taxes/questions/iit/accordion/residency/are-my-wages-earned-in-another-state-taxable-in-michigan-if-i-am-a-michigan-resident-1
          https://www.michigan.gov/taxes/-/media/Project/Websites/taxes/Forms/2022/2022-IIT-Forms/BOOK_MI-1040.pdf?rev=4524f1cbbc0243b194ab05a6680be75b&hash=40DFE71E651F7C1D65C31F6CD03D944F
          https://www.michigan.gov/taxes/questions/iit/accordion/general/what-cities-impose-an-income-tax
        """
        
        # Check if the parameters are set upstream in the instance
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        county_occupation = self.county_occupation if hasattr(self, 'county_occupation') else county_occupation
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Oklahoma', 'state_occupation': 'Michigan', "capital_gains_short": capital_gains_short, "county_occupation": county_occupation, "dependents": dependents, "sources": sources, "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def oklahoma_and_minnesota(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Oklahoma:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Minnesota:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Oklahoma:
        - https://oklahoma.gov/content/dam/ok/en/tax/documents/resources/publications/businesses/withholding-tables/WHTables-2022.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
        
        Minnesota:
        -         
        - https://www.revenue.state.mn.us/minnesota-income-tax-rates-and-brackets
          https://www.revenue.state.mn.us/child-and-dependent-care-credit
          https://www.revenue.state.mn.us/dependent-exemptions
          https://www.revenue.state.mn.us/minnesota-standard-deduction
          https://mn.gov/mmb/assets/mwr_form_tcm1059-128581.pdf
          https://www.revenue.state.mn.us/individual-income-tax
          https://www.revenue.state.mn.us/individual-income-tax
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Oklahoma', 'state_occupation': 'Minnesota', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def oklahoma_and_mississippi(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Oklahoma:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Mississippi:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Oklahoma:
        - https://oklahoma.gov/content/dam/ok/en/tax/documents/resources/publications/businesses/withholding-tables/WHTables-2022.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
        
        Mississippi:
        -         
        - https://www.dor.ms.gov/individual/tax-rates
          https://www.dor.ms.gov/individual/tax-rates
          https://www.dor.ms.gov/individual/tax-rates
          https://www.dor.ms.gov/individual
          https://www.dor.ms.gov/individual
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Oklahoma', 'state_occupation': 'Mississippi', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def oklahoma_and_missouri(self, capital_gains_short=0, dependents=0, sources=False, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_long=0, filing_status='single', income=100000, credits_state=0):
        """
        --- Parameters ----
        
        Oklahoma:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Missouri:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Oklahoma:
        - https://oklahoma.gov/content/dam/ok/en/tax/documents/resources/publications/businesses/withholding-tables/WHTables-2022.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
        
        Missouri:
        -         
        - https://dor.mo.gov/taxation/individual/tax-types/income/year-changes/
          https://dor.mo.gov/taxation/individual/tax-types/income/year-changes/
          https://dor.mo.gov/taxation/individual/tax-types/income/year-changes/
          https://www.kcmo.gov/city-hall/departments/finance/earnings-tax
          https://www.stlouis-mo.gov/government/departments/collector/earnings-tax/individual-earnings-tax-info.cfm
        """
        
        # Check if the parameters are set upstream in the instance
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Oklahoma', 'state_occupation': 'Missouri', "capital_gains_short": capital_gains_short, "dependents": dependents, "sources": sources, "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def oklahoma_and_montana(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Oklahoma:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Montana:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Oklahoma:
        - https://oklahoma.gov/content/dam/ok/en/tax/documents/resources/publications/businesses/withholding-tables/WHTables-2022.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
        
        Montana:
        -         
        - https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/2022/12/montana-employees-withholding-allowance-and-exemption-certificate-form-mw-4.pdf
          https://mtrevenue.gov/2021/11/29/simplification-of-montana-income-taxation/?utm_source=rss&utm_medium=rss&utm_campaign=simplification-of-montana-income-taxation
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Oklahoma', 'state_occupation': 'Montana', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def oklahoma_and_nebraska(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Oklahoma:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Nebraska:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Oklahoma:
        - https://oklahoma.gov/content/dam/ok/en/tax/documents/resources/publications/businesses/withholding-tables/WHTables-2022.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
        
        Nebraska:
        -         
        - https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2023_Circular_EN_8-429r2023_Whole_Book_11.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Oklahoma', 'state_occupation': 'Nebraska', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def oklahoma_and_nevada(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Oklahoma:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Nevada:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Oklahoma:
        - https://oklahoma.gov/content/dam/ok/en/tax/documents/resources/publications/businesses/withholding-tables/WHTables-2022.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
        
        Nevada:
        -         
        - https://www.nvsos.gov/sos/businesses/the-nevada-advantage
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Oklahoma', 'state_occupation': 'Nevada', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def oklahoma_and_new_hampshire(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Oklahoma:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        New Hampshire:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Oklahoma:
        - https://oklahoma.gov/content/dam/ok/en/tax/documents/resources/publications/businesses/withholding-tables/WHTables-2022.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
        
        New Hampshire:
        -         
        - https://www.revenue.nh.gov/assistance/tax-overview.htm
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Oklahoma', 'state_occupation': 'NewHampshire', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def oklahoma_and_new_jersey(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Oklahoma:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        New Jersey:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Oklahoma:
        - https://oklahoma.gov/content/dam/ok/en/tax/documents/resources/publications/businesses/withholding-tables/WHTables-2022.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
        
        New Jersey:
        -         
        - https://www.state.nj.us/treasury/taxation/taxtables.shtml
          https://www.state.nj.us/treasury/taxation/njit13.shtml
          https://www.nj.gov/treasury/taxation/njit25.shtml
          https://www.state.nj.us/treasury/taxation/individuals/freqqiti.shtml
          https://www.state.nj.us/treasury/taxation/localtax.shtml
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Oklahoma', 'state_occupation': 'NewJersey', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def oklahoma_and_new_mexico(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Oklahoma:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        New Mexico:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Oklahoma:
        - https://oklahoma.gov/content/dam/ok/en/tax/documents/resources/publications/businesses/withholding-tables/WHTables-2022.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
        
        New Mexico:
        -         
        - https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/fdf3c548-8aba-4b9c-9eb4-bb564c716015/FYI-104.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/0bf943fd-652e-4400-bb1b-cd8397eb5a95/2022pit-adj.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/2f1a6781-9534-4436-b427-1557f9592099/2022pit-adj-ins.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/0bf943fd-652e-4400-bb1b-cd8397eb5a95/2022pit-adj.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/1afc56af-ea90-4d48-82e5-1f9aeb43255a/PITbook2022.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/fdf3c548-8aba-4b9c-9eb4-bb564c716015/FYI-104.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Oklahoma', 'state_occupation': 'NewMexico', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def oklahoma_and_new_york(self, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Oklahoma:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        New York:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Oklahoma:
        - https://oklahoma.gov/content/dam/ok/en/tax/documents/resources/publications/businesses/withholding-tables/WHTables-2022.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
        
        New York:
        -         
        - https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nys-tax-rate-schedule
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nys-tax-rate-schedule
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nys-tax-rate-schedule
          https://www.tax.ny.gov/pit/
          https://www.tax.ny.gov/pit/file/nonresident-faqs.htm
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#yonkers-tax-rate-schedule
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nyc-tax-rate-schedule
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Oklahoma', 'state_occupation': 'NewYork', "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def oklahoma_and_north_carolina(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Oklahoma:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        North Carolina:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Oklahoma:
        - https://oklahoma.gov/content/dam/ok/en/tax/documents/resources/publications/businesses/withholding-tables/WHTables-2022.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
        
        North Carolina:
        -         
        - https://www.ncdor.gov/taxes-forms/tax-rate-schedules
          https://www.ncdor.gov/taxes-forms/individual-income-tax/north-carolina-standard-deduction-or-north-carolina-itemized-deductions
          https://www.ncdor.gov/taxes-forms/individual-income-tax
          https://www.ncdor.gov/taxes-forms/individual-income-tax
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Oklahoma', 'state_occupation': 'NorthCarolina', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def oklahoma_and_north_dakota(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Oklahoma:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        North Dakota:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Oklahoma:
        - https://oklahoma.gov/content/dam/ok/en/tax/documents/resources/publications/businesses/withholding-tables/WHTables-2022.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
        
        North Dakota:
        -         
        - https://www.tax.nd.gov/forms
          https://www.tax.nd.gov/sites/www/files/documents/forms/individual/2022-iit/2022-individual-income-tax-booklet.pdf
          https://www.tax.nd.gov/sites/www/files/documents/forms/2022-individual-income-tax-booklet.pdf
          https://www.tax.nd.gov/tax-exemptions-credits/income-tax-exemptions-credits
          https://www.tax.nd.gov/sites/www/files/documents/forms/business/ndwrfillable.pdf
          https://www.tax.nd.gov/news/tax-legislative-changes/significant-changes-law/individual-income-tax-history
          https://www.tax.nd.gov/forms
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Oklahoma', 'state_occupation': 'NorthDakota', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def oklahoma_and_ohio(self, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Oklahoma:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Ohio:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Oklahoma:
        - https://oklahoma.gov/content/dam/ok/en/tax/documents/resources/publications/businesses/withholding-tables/WHTables-2022.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
        
        Ohio:
        -         
        - https://tax.ohio.gov/individual/resources/annual-tax-rates
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://tax.ohio.gov/static/forms/employer_withholding/generic/wth_it4nr.pdf
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://www.ritaohio.com/TaxRatesTable
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Oklahoma', 'state_occupation': 'Ohio', "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def oklahoma_and_oregon(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Oklahoma:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Oregon:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Oklahoma:
        - https://oklahoma.gov/content/dam/ok/en/tax/documents/resources/publications/businesses/withholding-tables/WHTables-2022.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
        
        Oregon:
        -         
        - https://www.oregon.gov/dor/programs/individuals/Pages/PIT.aspx
          https://www.oregon.gov/dor/programs/individuals/Documents/Full%20year%20resident
         %20Form%20or-40%20filers.pdf
          https://www.oregon.gov/dor/programs/individuals/Documents/Part-year%20and%20nonresident
         %20Form%20or-40-P%20and%20or-40-N%20filers.pdf.pdf
          https://www.oregon.gov/dor/forms/FormsPubs/form-or-40_101-040_2022.pdf
          https://www.oregon.gov/dor/forms/FormsPubs/form-or-40_101-040_2022.pdf
          https://www.oregon.gov/dor/forms/FormsPubs/form-or-40_101-040_2022.pdf
          https://www.portland.gov/revenue/personal-tax
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Oklahoma', 'state_occupation': 'Oregon', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def oklahoma_and_pennsylvania(self, county_residence='Other', state_agi=75000, county_occupation='Other', filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Oklahoma:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Pennsylvania:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        county_residence : str, optional
            County of residence. Default is Other.
        county_occupation : str, optional
            County of occupation. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Oklahoma:
        - https://oklahoma.gov/content/dam/ok/en/tax/documents/resources/publications/businesses/withholding-tables/WHTables-2022.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
        
        Pennsylvania:
        -         
        - https://www.revenue.pa.gov/Tax%20Rates/Pages/default.aspx
          https://www.revenue.pa.gov/FormsandPublications/FormsforIndividuals/PIT/Documents/2022/2022_pa-40in.pdf
          https://www.revenue.pa.gov/FormsandPublications/FormsforBusinesses/EmployerWithholding/Documents/rev-419.pdf
          https://www.revenue.pa.gov/FormsandPublications/FormsforIndividuals/PIT/Documents/2022/2022_pa-40in.pdf
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-25.htm
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        county_occupation = self.county_occupation if hasattr(self, 'county_occupation') else county_occupation
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Oklahoma', 'state_occupation': 'Pennsylvania', "county_residence": county_residence, "state_agi": state_agi, "county_occupation": county_occupation, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def oklahoma_and_rhode_island(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Oklahoma:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Rhode Island:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Oklahoma:
        - https://oklahoma.gov/content/dam/ok/en/tax/documents/resources/publications/businesses/withholding-tables/WHTables-2022.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
        
        Rhode Island:
        -         
        - https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2022_1040WE_w_0.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Oklahoma', 'state_occupation': 'RhodeIsland', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def oklahoma_and_south_carolina(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Oklahoma:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        South Carolina:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Oklahoma:
        - https://oklahoma.gov/content/dam/ok/en/tax/documents/resources/publications/businesses/withholding-tables/WHTables-2022.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
        
        South Carolina:
        -         
        - https://dor.sc.gov/resources-site/lawandpolicy/Advisory%20Opinions/IL22-15.pdf
          https://dor.sc.gov/forms-site/Forms/IITPacket_2022.pdf
          https://www.scstatehouse.gov/code/t12c006.php
          https://dor.sc.gov/resources-site/lawandpolicy/Advisory%20Opinions/IL22-15.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Oklahoma', 'state_occupation': 'SouthCarolina', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def oklahoma_and_south_dakota(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Oklahoma:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        South Dakota:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Oklahoma:
        - https://oklahoma.gov/content/dam/ok/en/tax/documents/resources/publications/businesses/withholding-tables/WHTables-2022.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
        
        South Dakota:
        -         
        - https://dor.sd.gov/individuals/taxes/
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Oklahoma', 'state_occupation': 'SouthDakota', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def oklahoma_and_tennessee(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Oklahoma:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Tennessee:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Oklahoma:
        - https://oklahoma.gov/content/dam/ok/en/tax/documents/resources/publications/businesses/withholding-tables/WHTables-2022.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
        
        Tennessee:
        -         
        - https://www.tn.gov/revenue/taxes.html
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Oklahoma', 'state_occupation': 'Tennessee', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def oklahoma_and_texas(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Oklahoma:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Texas:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Oklahoma:
        - https://oklahoma.gov/content/dam/ok/en/tax/documents/resources/publications/businesses/withholding-tables/WHTables-2022.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
        
        Texas:
        -         
        - https://comptroller.texas.gov/taxes/
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Oklahoma', 'state_occupation': 'Texas', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def oklahoma_and_utah(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Oklahoma:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Utah:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Oklahoma:
        - https://oklahoma.gov/content/dam/ok/en/tax/documents/resources/publications/businesses/withholding-tables/WHTables-2022.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
        
        Utah:
        -         
        - https://incometax.utah.gov/paying/tax-rates
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Oklahoma', 'state_occupation': 'Utah', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def oklahoma_and_vermont(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Oklahoma:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Vermont:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Oklahoma:
        - https://oklahoma.gov/content/dam/ok/en/tax/documents/resources/publications/businesses/withholding-tables/WHTables-2022.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
        
        Vermont:
        -         
        - https://tax.vermont.gov/sites/tax/files/documents/GB-1210-2023.pdf
          https://tax.vermont.gov/sites/tax/files/documents/Income%20Booklet-2022.pdf
          https://tax.vermont.gov/sites/tax/files/documents/Income%20Booklet-2022.pdf
          https://tax.vermont.gov/sites/tax/files/documents/Income%20Booklet-2022.pdf
          https://tax.vermont.gov/individuals/personal-income-tax/tax-credits
          https://tax.vermont.gov/individuals/personal-income-tax/taxable-income
          https://tax.vermont.gov/individuals/personal-income-tax/local-option-tax
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Oklahoma', 'state_occupation': 'Vermont', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def oklahoma_and_virginia(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Oklahoma:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Virginia:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Oklahoma:
        - https://oklahoma.gov/content/dam/ok/en/tax/documents/resources/publications/businesses/withholding-tables/WHTables-2022.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
        
        Virginia:
        -         
        - https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/reciprocity
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Oklahoma', 'state_occupation': 'Virginia', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def oklahoma_and_washington(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Oklahoma:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Washington:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Oklahoma:
        - https://oklahoma.gov/content/dam/ok/en/tax/documents/resources/publications/businesses/withholding-tables/WHTables-2022.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
        
        Washington:
        -         
        - https://dor.wa.gov/taxes-rates/income-tax
          https://dor.wa.gov/taxes-rates/other-taxes/capital-gains-tax
          https://dor.wa.gov/taxes-rates/income-tax
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Oklahoma', 'state_occupation': 'Washington', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def oklahoma_and_west_virginia(self, pay_periods=24, capital_gains_short=0, county_occupation='Other', dependents=0, sources=False, state_agi=75000, filers_over_65=0, capital_gains_long=0, filing_status='single', income=100000, credits_state=0):
        """
        --- Parameters ----
        
        Oklahoma:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        West Virginia:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_occupation : str, optional
            County of occupation. Default is Other.
        pay_periods : int, optional
            Number of pay periods. Options are 13, 24, 26, 52. Default is 24.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Oklahoma:
        - https://oklahoma.gov/content/dam/ok/en/tax/documents/resources/publications/businesses/withholding-tables/WHTables-2022.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
        
        West Virginia:
        -         
        - https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://tax.wv.gov/Documents/TSD/tsd438.pdf
          https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2011/Taxes-11-35att.html
        """
        
        # Check if the parameters are set upstream in the instance
        pay_periods = self.pay_periods if hasattr(self, 'pay_periods') else pay_periods
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        county_occupation = self.county_occupation if hasattr(self, 'county_occupation') else county_occupation
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Oklahoma', 'state_occupation': 'WestVirginia', "pay_periods": pay_periods, "capital_gains_short": capital_gains_short, "county_occupation": county_occupation, "dependents": dependents, "sources": sources, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def oklahoma_and_wisconsin(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Oklahoma:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Wisconsin:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Oklahoma:
        - https://oklahoma.gov/content/dam/ok/en/tax/documents/resources/publications/businesses/withholding-tables/WHTables-2022.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
        
        Wisconsin:
        -         
        - https://www.revenue.wi.gov/TaxForms2023/2023-Form1-ES-Inst.pdf
          https://www.revenue.wi.gov/TaxForms2023/2023-Form1-ES-Inst.pdf
          https://www.revenue.wi.gov/TaxForms2023/2023-Form1-ES-Inst.pdf
          https://www.revenue.wi.gov/Pages/FAQS/pcs-work.aspx
          https://www.revenue.wi.gov/Dor%20Publications/pb103.pdf
          https://www.revenue.wi.gov/Pages/FAQS/pcs-work.aspx
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Oklahoma', 'state_occupation': 'Wisconsin', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def oklahoma_and_wyoming(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Oklahoma:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Wyoming:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Oklahoma:
        - https://oklahoma.gov/content/dam/ok/en/tax/documents/resources/publications/businesses/withholding-tables/WHTables-2022.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
        
        Wyoming:
        -         
        - https://revenue.wyo.gov/
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Oklahoma', 'state_occupation': 'Wyoming', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def oregon_and_alabama(self, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Oregon:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Alabama:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Oregon:
        - https://www.oregon.gov/dor/programs/individuals/Pages/PIT.aspx
          https://www.oregon.gov/dor/programs/individuals/Documents/Full%20year%20resident
         %20Form%20or-40%20filers.pdf
          https://www.oregon.gov/dor/programs/individuals/Documents/Part-year%20and%20nonresident
         %20Form%20or-40-P%20and%20or-40-N%20filers.pdf.pdf
          https://www.oregon.gov/dor/forms/FormsPubs/form-or-40_101-040_2022.pdf
          https://www.oregon.gov/dor/forms/FormsPubs/form-or-40_101-040_2022.pdf
          https://www.oregon.gov/dor/forms/FormsPubs/form-or-40_101-040_2022.pdf
          https://www.portland.gov/revenue/personal-tax
        
        Alabama:
        -         
        - https://www.revenue.alabama.gov/ultraviewer/viewer/basic_viewer/index.html?form=2023/02/22f40.pdf
          https://www.revenue.alabama.gov/faqs/what-is-alabamas-individual-income-tax-rate/
          https://www.revenue.alabama.gov/ultraviewer/viewer/basic_viewer/index.html?form=2023/02/22f40.pdf
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-24.htm?taxmap=true
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-24.htm?taxmap=true
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-24.htm?taxmap=true
          https://www.revenue.alabama.gov/individual-corporate/income-to-be-reported-on-the-alabama-income-tax-return/
          https://www.almonline.org/TaxRates.aspx#OccupationalTax
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Oregon', 'state_occupation': 'Alabama', "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def oregon_and_alaska(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Oregon:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Alaska:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Oregon:
        - https://www.oregon.gov/dor/programs/individuals/Pages/PIT.aspx
          https://www.oregon.gov/dor/programs/individuals/Documents/Full%20year%20resident
         %20Form%20or-40%20filers.pdf
          https://www.oregon.gov/dor/programs/individuals/Documents/Part-year%20and%20nonresident
         %20Form%20or-40-P%20and%20or-40-N%20filers.pdf.pdf
          https://www.oregon.gov/dor/forms/FormsPubs/form-or-40_101-040_2022.pdf
          https://www.oregon.gov/dor/forms/FormsPubs/form-or-40_101-040_2022.pdf
          https://www.oregon.gov/dor/forms/FormsPubs/form-or-40_101-040_2022.pdf
          https://www.portland.gov/revenue/personal-tax
        
        Alaska:
        -         
        - https://www.commerce.alaska.gov/web/dcra/officeofthestateassessor/alaskataxfacts.aspx
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Oregon', 'state_occupation': 'Alaska', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def oregon_and_arizona(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Oregon:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Arizona:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        credits_state : int, optional
            Credits for state taxes. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Oregon:
        - https://www.oregon.gov/dor/programs/individuals/Pages/PIT.aspx
          https://www.oregon.gov/dor/programs/individuals/Documents/Full%20year%20resident
         %20Form%20or-40%20filers.pdf
          https://www.oregon.gov/dor/programs/individuals/Documents/Part-year%20and%20nonresident
         %20Form%20or-40-P%20and%20or-40-N%20filers.pdf.pdf
          https://www.oregon.gov/dor/forms/FormsPubs/form-or-40_101-040_2022.pdf
          https://www.oregon.gov/dor/forms/FormsPubs/form-or-40_101-040_2022.pdf
          https://www.oregon.gov/dor/forms/FormsPubs/form-or-40_101-040_2022.pdf
          https://www.portland.gov/revenue/personal-tax
        
        Arizona:
        -         
        - https://azdor.gov/forms/individual/form-140-x-y-tables
          https://azdor.gov/forms/individual/form-140-x-y-tables
          https://azdor.gov/forms/individual/form-140-x-y-tables
          https://azdor.gov/businesses-arizona/withholding-tax/withholding-exceptions
          "https://azdor.gov/individuals/income-tax-filing-assistance/identifying-other-taxable-income#:~:text=In%20regards%20to%20capital%20gains
         the%20individuals%20regular%20tax%20rate.
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Oregon', 'state_occupation': 'Arizona', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def oregon_and_arkansas(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Oregon:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Arkansas:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Oregon:
        - https://www.oregon.gov/dor/programs/individuals/Pages/PIT.aspx
          https://www.oregon.gov/dor/programs/individuals/Documents/Full%20year%20resident
         %20Form%20or-40%20filers.pdf
          https://www.oregon.gov/dor/programs/individuals/Documents/Part-year%20and%20nonresident
         %20Form%20or-40-P%20and%20or-40-N%20filers.pdf.pdf
          https://www.oregon.gov/dor/forms/FormsPubs/form-or-40_101-040_2022.pdf
          https://www.oregon.gov/dor/forms/FormsPubs/form-or-40_101-040_2022.pdf
          https://www.oregon.gov/dor/forms/FormsPubs/form-or-40_101-040_2022.pdf
          https://www.portland.gov/revenue/personal-tax
        
        Arkansas:
        -         
        - https://www.arkansasedc.com/why-arkansas/business-climate/tax-structure/personal-income-tax
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_and_AR1000NR_Instructions.pdf
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_FullYearResidentIndividualIncomeTaxReturn.pdf
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_and_AR1000NR_Instructions.pdf
          https://www.arkansasedc.com/why-arkansas/business-climate/tax-structure/capital-gains-tax-reduction
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_and_AR1000NR_Instructions.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Oregon', 'state_occupation': 'Arkansas', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def oregon_and_california(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Oregon:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        California:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Oregon:
        - https://www.oregon.gov/dor/programs/individuals/Pages/PIT.aspx
          https://www.oregon.gov/dor/programs/individuals/Documents/Full%20year%20resident
         %20Form%20or-40%20filers.pdf
          https://www.oregon.gov/dor/programs/individuals/Documents/Part-year%20and%20nonresident
         %20Form%20or-40-P%20and%20or-40-N%20filers.pdf.pdf
          https://www.oregon.gov/dor/forms/FormsPubs/form-or-40_101-040_2022.pdf
          https://www.oregon.gov/dor/forms/FormsPubs/form-or-40_101-040_2022.pdf
          https://www.oregon.gov/dor/forms/FormsPubs/form-or-40_101-040_2022.pdf
          https://www.portland.gov/revenue/personal-tax
        
        California:
        -         
        - https://www.ftb.ca.gov/forms/2022/2022-540-tax-rate-schedules.pdf
          https://www.ftb.ca.gov/file/personal/deductions/index.html
          https://www.ftb.ca.gov/about-ftb/newsroom/caleitc/eligibility-and-credit-information.html
          https://www.ftb.ca.gov/file/personal/income-types/capital-gains-and-losses.html#:~:text=California%20does%20not%20have%20a
         are%20taxed%20as%20ordinary%20income
          https://www.ftb.ca.gov/file/personal/
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Oregon', 'state_occupation': 'California', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def oregon_and_colorado(self, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Oregon:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Colorado:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Oregon:
        - https://www.oregon.gov/dor/programs/individuals/Pages/PIT.aspx
          https://www.oregon.gov/dor/programs/individuals/Documents/Full%20year%20resident
         %20Form%20or-40%20filers.pdf
          https://www.oregon.gov/dor/programs/individuals/Documents/Part-year%20and%20nonresident
         %20Form%20or-40-P%20and%20or-40-N%20filers.pdf.pdf
          https://www.oregon.gov/dor/forms/FormsPubs/form-or-40_101-040_2022.pdf
          https://www.oregon.gov/dor/forms/FormsPubs/form-or-40_101-040_2022.pdf
          https://www.oregon.gov/dor/forms/FormsPubs/form-or-40_101-040_2022.pdf
          https://www.portland.gov/revenue/personal-tax
        
        Colorado:
        -         
        - https://leg.colorado.gov/agencies/legislative-council-staff/individual-income-tax%C2%A0
          https://leg.colorado.gov/agencies/legislative-council-staff/individual-income-tax%C2%A0
          https://tax.colorado.gov/sites/tax/files/documents/ITT_Capital_Gain_Subtraction_Feb_2022.pdf
          https://www.auroragov.org/business_services/taxes/occupational_privilege_tax
          https://www.denvergov.org/Government/Agencies-Departments-Offices/Agencies-Departments-Offices-Directory/Department-of-Finance/Our-Divisions/Treasury/Business-Tax-Information#:~:text=Occupational%20privilege%20tax
         -%28show%20below%29&text=Each%20taxable%20employee%20is%20liable
         month%20for%20each%20taxable%20employee
          https://www.glendale.co.us/355/Occupational-Privilege-Tax
          https://www.greenwoodvillage.com/1220/Occupational-Privilege-Tax-OPT
          https://ci.sheridan.co.us/288/Occupational-Privilege-Tax
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Oregon', 'state_occupation': 'Colorado', "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def oregon_and_connecticut(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Oregon:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Connecticut:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Oregon:
        - https://www.oregon.gov/dor/programs/individuals/Pages/PIT.aspx
          https://www.oregon.gov/dor/programs/individuals/Documents/Full%20year%20resident
         %20Form%20or-40%20filers.pdf
          https://www.oregon.gov/dor/programs/individuals/Documents/Part-year%20and%20nonresident
         %20Form%20or-40-P%20and%20or-40-N%20filers.pdf.pdf
          https://www.oregon.gov/dor/forms/FormsPubs/form-or-40_101-040_2022.pdf
          https://www.oregon.gov/dor/forms/FormsPubs/form-or-40_101-040_2022.pdf
          https://www.oregon.gov/dor/forms/FormsPubs/form-or-40_101-040_2022.pdf
          https://www.portland.gov/revenue/personal-tax
        
        Connecticut:
        -         
        - https://portal.ct.gov/-/media/DRS/Forms/2022/Income/2022-CT-1040-Instructions_1222.pdf
          https://www.cga.ct.gov/2022/rpt/pdf/2022-R-0108.pdf
          https://www.cga.ct.gov/current/pub/chap_229.htm#sec_12-703
          https://portal.ct.gov/-/media/DRS/Forms/2022/Income/2022-CT-1040-Instructions_1222.pdf
          "https://portal.ct.gov/DRS/Publications/TSSNs/TSSN-29#:~:text=An%20individuals%20net%20capital%20gains
         portion%20of%20Social%20Security%20benefits"
          "https://portal.ct.gov/DRS/Publications/TSSNs/TSSN-29#:~:text=An%20individuals%20net%20capital%20gains
         portion%20of%20Social%20Security%20benefits"
          https://portal.ct.gov/DRS/Individuals/Resident-Income-Tax/Tax-Information
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Oregon', 'state_occupation': 'Connecticut', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def oregon_and_delaware(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, federal_agi=75000, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Oregon:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Delaware:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        federal_agi : int, optional
            Adjusted Gross Income (AGI) for federal taxes. Default is 0. Sources - https://www.irs.gov/e-file-providers/definition-of-adjusted-gross-income, https://www.irs.gov/newsroom/taxpayers-should-know-the-difference-between-standard-and-itemized-deductions
        income : int, optional
            Gross income. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Oregon:
        - https://www.oregon.gov/dor/programs/individuals/Pages/PIT.aspx
          https://www.oregon.gov/dor/programs/individuals/Documents/Full%20year%20resident
         %20Form%20or-40%20filers.pdf
          https://www.oregon.gov/dor/programs/individuals/Documents/Part-year%20and%20nonresident
         %20Form%20or-40-P%20and%20or-40-N%20filers.pdf.pdf
          https://www.oregon.gov/dor/forms/FormsPubs/form-or-40_101-040_2022.pdf
          https://www.oregon.gov/dor/forms/FormsPubs/form-or-40_101-040_2022.pdf
          https://www.oregon.gov/dor/forms/FormsPubs/form-or-40_101-040_2022.pdf
          https://www.portland.gov/revenue/personal-tax
        
        Delaware:
        -         
        - https://revenue.delaware.gov/software-developer/tax-rate-changes/
          https://revenuefiles.delaware.gov/2022/PIT-RES_TY22_2022-01_PaperInteractive.pdf
          https://revenuefiles.delaware.gov/2022/PIT-RES_TY22_2022-02_Instructions.pdf
          https://revenue.delaware.gov/frequently-asked-questions/personal-income-tax-faqs/
          https://www.newcastlede.gov/DocumentCenter/View/20147/City-of-Wilmington-Wage-Tax-Information-PDF
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        federal_agi = self.federal_agi if hasattr(self, 'federal_agi') else federal_agi
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Oregon', 'state_occupation': 'Delaware', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "federal_agi": federal_agi, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def oregon_and_florida(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Oregon:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Florida:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Oregon:
        - https://www.oregon.gov/dor/programs/individuals/Pages/PIT.aspx
          https://www.oregon.gov/dor/programs/individuals/Documents/Full%20year%20resident
         %20Form%20or-40%20filers.pdf
          https://www.oregon.gov/dor/programs/individuals/Documents/Part-year%20and%20nonresident
         %20Form%20or-40-P%20and%20or-40-N%20filers.pdf.pdf
          https://www.oregon.gov/dor/forms/FormsPubs/form-or-40_101-040_2022.pdf
          https://www.oregon.gov/dor/forms/FormsPubs/form-or-40_101-040_2022.pdf
          https://www.oregon.gov/dor/forms/FormsPubs/form-or-40_101-040_2022.pdf
          https://www.portland.gov/revenue/personal-tax
        
        Florida:
        -         
        - https://floridarevenue.com/taxes/taxesfees/Pages/tax_interest_rates.aspx
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Oregon', 'state_occupation': 'Florida', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def oregon_and_georgia(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Oregon:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Georgia:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Oregon:
        - https://www.oregon.gov/dor/programs/individuals/Pages/PIT.aspx
          https://www.oregon.gov/dor/programs/individuals/Documents/Full%20year%20resident
         %20Form%20or-40%20filers.pdf
          https://www.oregon.gov/dor/programs/individuals/Documents/Part-year%20and%20nonresident
         %20Form%20or-40-P%20and%20or-40-N%20filers.pdf.pdf
          https://www.oregon.gov/dor/forms/FormsPubs/form-or-40_101-040_2022.pdf
          https://www.oregon.gov/dor/forms/FormsPubs/form-or-40_101-040_2022.pdf
          https://www.oregon.gov/dor/forms/FormsPubs/form-or-40_101-040_2022.pdf
          https://www.portland.gov/revenue/personal-tax
        
        Georgia:
        -         
        - https://dor.georgia.gov/tax-tables-georgia-tax-rate-schedule
          https://apps.dor.ga.gov/FillableForms/PDFViewer/Index?form=2022GA500
          https://apps.dor.ga.gov/FillableForms/PDFViewer/Index?form=2022GA500
          https://apps.dor.ga.gov/FillableForms/PDFViewer/Index?form=2022GA500
          https://dor.georgia.gov/years-individual-income-tax-forms
          https://dor.georgia.gov/years-individual-income-tax-forms
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Oregon', 'state_occupation': 'Georgia', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def oregon_and_hawaii(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Oregon:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Hawaii:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Oregon:
        - https://www.oregon.gov/dor/programs/individuals/Pages/PIT.aspx
          https://www.oregon.gov/dor/programs/individuals/Documents/Full%20year%20resident
         %20Form%20or-40%20filers.pdf
          https://www.oregon.gov/dor/programs/individuals/Documents/Part-year%20and%20nonresident
         %20Form%20or-40-P%20and%20or-40-N%20filers.pdf.pdf
          https://www.oregon.gov/dor/forms/FormsPubs/form-or-40_101-040_2022.pdf
          https://www.oregon.gov/dor/forms/FormsPubs/form-or-40_101-040_2022.pdf
          https://www.oregon.gov/dor/forms/FormsPubs/form-or-40_101-040_2022.pdf
          https://www.portland.gov/revenue/personal-tax
        
        Hawaii:
        -         
        - https://tax.hawaii.gov/forms/d_18table-on/d_18table-on_p13/
          https://files.hawaii.gov/tax/forms/2018/18table-on.pdf
          https://files.hawaii.gov/tax/news/pubs/22outline.pdf
          https://files.hawaii.gov/tax/news/pubs/FastTaxRefGuide_2017.pdf
          https://www.tfhawaii.org/wordpress/state-tax-resources/mini-tax-guides/individual-income-tax-chapter-235/#:~:text=Long%20term%20capital%20gains%20are%20taxed%20at%20a%20maximum%20of%207.25%25
          https://www.grassrootinstitute.org/2023/02/hb337-increasing-tax-on-capital-gains-a-bad-risk/
          https://files.hawaii.gov/tax/news/pubs/22outline.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Oregon', 'state_occupation': 'Hawaii', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def oregon_and_idaho(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Oregon:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Idaho:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Oregon:
        - https://www.oregon.gov/dor/programs/individuals/Pages/PIT.aspx
          https://www.oregon.gov/dor/programs/individuals/Documents/Full%20year%20resident
         %20Form%20or-40%20filers.pdf
          https://www.oregon.gov/dor/programs/individuals/Documents/Part-year%20and%20nonresident
         %20Form%20or-40-P%20and%20or-40-N%20filers.pdf.pdf
          https://www.oregon.gov/dor/forms/FormsPubs/form-or-40_101-040_2022.pdf
          https://www.oregon.gov/dor/forms/FormsPubs/form-or-40_101-040_2022.pdf
          https://www.oregon.gov/dor/forms/FormsPubs/form-or-40_101-040_2022.pdf
          https://www.portland.gov/revenue/personal-tax
        
        Idaho:
        -         
        - https://tax.idaho.gov/taxes/income-tax/individual-income/individual-income-tax-rate-schedule/
          https://tax.idaho.gov/wp-content/uploads/forms/EIN00046/EIN00046_03-01-2023.pdf
          https://tax.idaho.gov/taxes/income-tax/individual-income/idaho-source-income/
          https://tax.idaho.gov/taxes/income-tax/individual-income/idaho-source-income/
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Oregon', 'state_occupation': 'Idaho', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def oregon_and_illinois(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Oregon:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Illinois:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Oregon:
        - https://www.oregon.gov/dor/programs/individuals/Pages/PIT.aspx
          https://www.oregon.gov/dor/programs/individuals/Documents/Full%20year%20resident
         %20Form%20or-40%20filers.pdf
          https://www.oregon.gov/dor/programs/individuals/Documents/Part-year%20and%20nonresident
         %20Form%20or-40-P%20and%20or-40-N%20filers.pdf.pdf
          https://www.oregon.gov/dor/forms/FormsPubs/form-or-40_101-040_2022.pdf
          https://www.oregon.gov/dor/forms/FormsPubs/form-or-40_101-040_2022.pdf
          https://www.oregon.gov/dor/forms/FormsPubs/form-or-40_101-040_2022.pdf
          https://www.portland.gov/revenue/personal-tax
        
        Illinois:
        -         
        - https://tax.illinois.gov/research/taxrates/income.html#IndividualIncome
          https://tax.illinois.gov/content/dam/soi/en/web/tax/forms/incometax/documents/currentyear/individual/il-1040-instr.pdf
          https://tax.illinois.gov/questionsandanswers/12.html
          https://tax.illinois.gov/content/dam/soi/en/web/tax/forms/incometax/documents/currentyear/individual/il-1040-instr.pdf
          https://tax.illinois.gov/research/taxrates/personalproperty.html
          https://tax.illinois.gov/localgovernments/income.html
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Oregon', 'state_occupation': 'Illinois', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def oregon_and_indiana(self, capital_gains_short=0, dependents=0, sources=False, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_long=0, filing_status='single', income=100000, credits_state=0):
        """
        --- Parameters ----
        
        Oregon:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Indiana:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Oregon:
        - https://www.oregon.gov/dor/programs/individuals/Pages/PIT.aspx
          https://www.oregon.gov/dor/programs/individuals/Documents/Full%20year%20resident
         %20Form%20or-40%20filers.pdf
          https://www.oregon.gov/dor/programs/individuals/Documents/Part-year%20and%20nonresident
         %20Form%20or-40-P%20and%20or-40-N%20filers.pdf.pdf
          https://www.oregon.gov/dor/forms/FormsPubs/form-or-40_101-040_2022.pdf
          https://www.oregon.gov/dor/forms/FormsPubs/form-or-40_101-040_2022.pdf
          https://www.oregon.gov/dor/forms/FormsPubs/form-or-40_101-040_2022.pdf
          https://www.portland.gov/revenue/personal-tax
        
        Indiana:
        -         
        - https://www.in.gov/dor/tax-forms/2022-individual-income-tax-forms/
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-32.htm
          https://www.in.gov/dor/files/ib28.pdf
          https://www.in.gov/dor/tax-forms/2022-individual-income-tax-forms/
          https://www.in.gov/dor/files/2022-county-tax-rates-and-codes.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Oregon', 'state_occupation': 'Indiana', "capital_gains_short": capital_gains_short, "dependents": dependents, "sources": sources, "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def oregon_and_iowa(self, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Oregon:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Iowa:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Oregon:
        - https://www.oregon.gov/dor/programs/individuals/Pages/PIT.aspx
          https://www.oregon.gov/dor/programs/individuals/Documents/Full%20year%20resident
         %20Form%20or-40%20filers.pdf
          https://www.oregon.gov/dor/programs/individuals/Documents/Part-year%20and%20nonresident
         %20Form%20or-40-P%20and%20or-40-N%20filers.pdf.pdf
          https://www.oregon.gov/dor/forms/FormsPubs/form-or-40_101-040_2022.pdf
          https://www.oregon.gov/dor/forms/FormsPubs/form-or-40_101-040_2022.pdf
          https://www.oregon.gov/dor/forms/FormsPubs/form-or-40_101-040_2022.pdf
          https://www.portland.gov/revenue/personal-tax
        
        Iowa:
        -         
        - https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
          https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
          https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
          https://tax.iowa.gov/iowa-illinois-reciprocal-agreement
          https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Oregon', 'state_occupation': 'Iowa', "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def oregon_and_kansas(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Oregon:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Kansas:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Oregon:
        - https://www.oregon.gov/dor/programs/individuals/Pages/PIT.aspx
          https://www.oregon.gov/dor/programs/individuals/Documents/Full%20year%20resident
         %20Form%20or-40%20filers.pdf
          https://www.oregon.gov/dor/programs/individuals/Documents/Part-year%20and%20nonresident
         %20Form%20or-40-P%20and%20or-40-N%20filers.pdf.pdf
          https://www.oregon.gov/dor/forms/FormsPubs/form-or-40_101-040_2022.pdf
          https://www.oregon.gov/dor/forms/FormsPubs/form-or-40_101-040_2022.pdf
          https://www.oregon.gov/dor/forms/FormsPubs/form-or-40_101-040_2022.pdf
          https://www.portland.gov/revenue/personal-tax
        
        Kansas:
        -         
        - https://www.ksrevenue.gov/taxrates.html
          https://www.ksrevenue.gov/incomebook22.html#0
          https://www.ksrevenue.gov/incomebook22.html#0
          https://www.ksrevenue.gov/incomebook22.html#0
          https://www.ksrevenue.gov/incomebook22.html
          https://www.ksrevenue.gov/incomebook22.html
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Oregon', 'state_occupation': 'Kansas', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def oregon_and_kentucky(self, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Oregon:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Kentucky:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Oregon:
        - https://www.oregon.gov/dor/programs/individuals/Pages/PIT.aspx
          https://www.oregon.gov/dor/programs/individuals/Documents/Full%20year%20resident
         %20Form%20or-40%20filers.pdf
          https://www.oregon.gov/dor/programs/individuals/Documents/Part-year%20and%20nonresident
         %20Form%20or-40-P%20and%20or-40-N%20filers.pdf.pdf
          https://www.oregon.gov/dor/forms/FormsPubs/form-or-40_101-040_2022.pdf
          https://www.oregon.gov/dor/forms/FormsPubs/form-or-40_101-040_2022.pdf
          https://www.oregon.gov/dor/forms/FormsPubs/form-or-40_101-040_2022.pdf
          https://www.portland.gov/revenue/personal-tax
        
        Kentucky:
        -         
        - https://revenue.ky.gov/News/Pages/Dor-Announces-Updates-to-Individual-Income-Tax-for-2023-Tax-Year.aspx
          https://revenue.ky.gov/Individual/Individual-Income-Tax/Pages/default.aspx
          https://revenue.ky.gov/News/Pages/Dor-Announces-Updates-to-Individual-Income-Tax-for-2023-Tax-Year.aspx
          https://revenue.ky.gov/Individual/Individual-Income-Tax/Pages/default.aspx
          https://revenue.ky.gov/Dor%20Training%20Materials/103%20KAR%2017.140.%20Individual%20income%20tax%20-%20reciprocity%20-%20nonresidents.pdf
          https://revenue.ky.gov/Individual/Individual-Income-Tax/Pages/default.aspx
          https://nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2021/Taxes-21-32.htm?taxmap=true
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Oregon', 'state_occupation': 'Kentucky', "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def oregon_and_louisiana(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Oregon:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Louisiana:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Oregon:
        - https://www.oregon.gov/dor/programs/individuals/Pages/PIT.aspx
          https://www.oregon.gov/dor/programs/individuals/Documents/Full%20year%20resident
         %20Form%20or-40%20filers.pdf
          https://www.oregon.gov/dor/programs/individuals/Documents/Part-year%20and%20nonresident
         %20Form%20or-40-P%20and%20or-40-N%20filers.pdf.pdf
          https://www.oregon.gov/dor/forms/FormsPubs/form-or-40_101-040_2022.pdf
          https://www.oregon.gov/dor/forms/FormsPubs/form-or-40_101-040_2022.pdf
          https://www.oregon.gov/dor/forms/FormsPubs/form-or-40_101-040_2022.pdf
          https://www.portland.gov/revenue/personal-tax
        
        Louisiana:
        -         
        - https://revenue.louisiana.gov/individualincometax
          https://revenue.louisiana.gov/individualincometax
          https://www.revenue.louisiana.gov/taxforms/6935(11_02)F.pdf
          https://revenue.louisiana.gov/individualincometax
          https://revenue.louisiana.gov/individualincometax
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Oregon', 'state_occupation': 'Louisiana', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def oregon_and_maine(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Oregon:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Maine:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Oregon:
        - https://www.oregon.gov/dor/programs/individuals/Pages/PIT.aspx
          https://www.oregon.gov/dor/programs/individuals/Documents/Full%20year%20resident
         %20Form%20or-40%20filers.pdf
          https://www.oregon.gov/dor/programs/individuals/Documents/Part-year%20and%20nonresident
         %20Form%20or-40-P%20and%20or-40-N%20filers.pdf.pdf
          https://www.oregon.gov/dor/forms/FormsPubs/form-or-40_101-040_2022.pdf
          https://www.oregon.gov/dor/forms/FormsPubs/form-or-40_101-040_2022.pdf
          https://www.oregon.gov/dor/forms/FormsPubs/form-or-40_101-040_2022.pdf
          https://www.portland.gov/revenue/personal-tax
        
        Maine:
        -         
        - https://www.maine.gov/revenue/tax-return-forms/individual-income-tax-2022
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/ind_tax_rate_sched_2022.pdf
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/22_1040me_book_gen_instruc_revisedFeb23.pdf
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/22_1040me_book_gen_instruc_revisedFeb23.pdf
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/22_1040me_book_gen_instruc_revisedFeb23.pdf
          https://www.maine.gov/revenue/tax-return-forms/individual-income-tax-2022
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Oregon', 'state_occupation': 'Maine', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def oregon_and_maryland(self, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Oregon:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Maryland:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Oregon:
        - https://www.oregon.gov/dor/programs/individuals/Pages/PIT.aspx
          https://www.oregon.gov/dor/programs/individuals/Documents/Full%20year%20resident
         %20Form%20or-40%20filers.pdf
          https://www.oregon.gov/dor/programs/individuals/Documents/Part-year%20and%20nonresident
         %20Form%20or-40-P%20and%20or-40-N%20filers.pdf.pdf
          https://www.oregon.gov/dor/forms/FormsPubs/form-or-40_101-040_2022.pdf
          https://www.oregon.gov/dor/forms/FormsPubs/form-or-40_101-040_2022.pdf
          https://www.oregon.gov/dor/forms/FormsPubs/form-or-40_101-040_2022.pdf
          https://www.portland.gov/revenue/personal-tax
        
        Maryland:
        -         
        - https://www.marylandtaxes.gov/individual/income/tax-info/tax-rates.php
          https://www.marylandtaxes.gov/forms/22_forms/Resident_Booklet.pdf
          https://www.marylandtaxes.gov/forms/Tax_Publications/Administrative_Releases/Income_and_Estate_Tax_Releases/ar_it37.pdf
          https://www.marylandtaxes.gov/individual/index.php
          https://www.marylandtaxes.gov/individual/income/tax-info/tax-rates.php
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Oregon', 'state_occupation': 'Maryland', "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def oregon_and_massachusetts(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Oregon:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Massachusetts:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Oregon:
        - https://www.oregon.gov/dor/programs/individuals/Pages/PIT.aspx
          https://www.oregon.gov/dor/programs/individuals/Documents/Full%20year%20resident
         %20Form%20or-40%20filers.pdf
          https://www.oregon.gov/dor/programs/individuals/Documents/Part-year%20and%20nonresident
         %20Form%20or-40-P%20and%20or-40-N%20filers.pdf.pdf
          https://www.oregon.gov/dor/forms/FormsPubs/form-or-40_101-040_2022.pdf
          https://www.oregon.gov/dor/forms/FormsPubs/form-or-40_101-040_2022.pdf
          https://www.oregon.gov/dor/forms/FormsPubs/form-or-40_101-040_2022.pdf
          https://www.portland.gov/revenue/personal-tax
        
        Massachusetts:
        -         
        - https://www.mass.gov/service-details/massachusetts-tax-rates
          https://www.mass.gov/service-details/view-massachusetts-personal-income-tax-exemptions
          https://www.mass.gov/service-details/view-massachusetts-personal-income-tax-exemptions
          https://www.mass.gov/service-details/view-child-and-dependent-related-credits
          https://www.mass.gov/service-details/massachusetts-tax-rates
          https://www.mass.gov/personal-income-tax
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Oregon', 'state_occupation': 'Massachusetts', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def oregon_and_michigan(self, capital_gains_short=0, county_occupation='Other', dependents=0, sources=False, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_long=0, filing_status='single', income=100000, credits_state=0):
        """
        --- Parameters ----
        
        Oregon:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Michigan:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        county_occupation : str, optional
            County of occupation. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Oregon:
        - https://www.oregon.gov/dor/programs/individuals/Pages/PIT.aspx
          https://www.oregon.gov/dor/programs/individuals/Documents/Full%20year%20resident
         %20Form%20or-40%20filers.pdf
          https://www.oregon.gov/dor/programs/individuals/Documents/Part-year%20and%20nonresident
         %20Form%20or-40-P%20and%20or-40-N%20filers.pdf.pdf
          https://www.oregon.gov/dor/forms/FormsPubs/form-or-40_101-040_2022.pdf
          https://www.oregon.gov/dor/forms/FormsPubs/form-or-40_101-040_2022.pdf
          https://www.oregon.gov/dor/forms/FormsPubs/form-or-40_101-040_2022.pdf
          https://www.portland.gov/revenue/personal-tax
        
        Michigan:
        -         
        - https://www.michigan.gov/taxes/iit/new-developments-for-tax-year-2022
          "https://www.michigan.gov/taxes/-/media/Project/Websites/taxes/Forms/2023/2023-SUW/446_Withholding-Guide_2023.pdf?rev=177ca3d8cb034392bf84e317ed78c320&hash=19C5F42F43A89A91863D229349F48883#:~:text=The%20exemption%20amount%20is%20%245%2C400
         employees%20Michigan%20income%20tax%20return"
          "https://www.michigan.gov/taxes/-/media/Project/Websites/taxes/Forms/2023/2023-SUW/446_Withholding-Guide_2023.pdf?rev=177ca3d8cb034392bf84e317ed78c320&hash=19C5F42F43A89A91863D229349F48883#:~:text=The%20exemption%20amount%20is%20%245%2C400
         employees%20Michigan%20income%20tax%20return"
          https://www.michigan.gov/taxes/questions/iit/accordion/residency/are-my-wages-earned-in-another-state-taxable-in-michigan-if-i-am-a-michigan-resident-1
          https://www.michigan.gov/taxes/-/media/Project/Websites/taxes/Forms/2022/2022-IIT-Forms/BOOK_MI-1040.pdf?rev=4524f1cbbc0243b194ab05a6680be75b&hash=40DFE71E651F7C1D65C31F6CD03D944F
          https://www.michigan.gov/taxes/questions/iit/accordion/general/what-cities-impose-an-income-tax
        """
        
        # Check if the parameters are set upstream in the instance
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        county_occupation = self.county_occupation if hasattr(self, 'county_occupation') else county_occupation
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Oregon', 'state_occupation': 'Michigan', "capital_gains_short": capital_gains_short, "county_occupation": county_occupation, "dependents": dependents, "sources": sources, "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def oregon_and_minnesota(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Oregon:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Minnesota:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Oregon:
        - https://www.oregon.gov/dor/programs/individuals/Pages/PIT.aspx
          https://www.oregon.gov/dor/programs/individuals/Documents/Full%20year%20resident
         %20Form%20or-40%20filers.pdf
          https://www.oregon.gov/dor/programs/individuals/Documents/Part-year%20and%20nonresident
         %20Form%20or-40-P%20and%20or-40-N%20filers.pdf.pdf
          https://www.oregon.gov/dor/forms/FormsPubs/form-or-40_101-040_2022.pdf
          https://www.oregon.gov/dor/forms/FormsPubs/form-or-40_101-040_2022.pdf
          https://www.oregon.gov/dor/forms/FormsPubs/form-or-40_101-040_2022.pdf
          https://www.portland.gov/revenue/personal-tax
        
        Minnesota:
        -         
        - https://www.revenue.state.mn.us/minnesota-income-tax-rates-and-brackets
          https://www.revenue.state.mn.us/child-and-dependent-care-credit
          https://www.revenue.state.mn.us/dependent-exemptions
          https://www.revenue.state.mn.us/minnesota-standard-deduction
          https://mn.gov/mmb/assets/mwr_form_tcm1059-128581.pdf
          https://www.revenue.state.mn.us/individual-income-tax
          https://www.revenue.state.mn.us/individual-income-tax
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Oregon', 'state_occupation': 'Minnesota', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def oregon_and_mississippi(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Oregon:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Mississippi:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Oregon:
        - https://www.oregon.gov/dor/programs/individuals/Pages/PIT.aspx
          https://www.oregon.gov/dor/programs/individuals/Documents/Full%20year%20resident
         %20Form%20or-40%20filers.pdf
          https://www.oregon.gov/dor/programs/individuals/Documents/Part-year%20and%20nonresident
         %20Form%20or-40-P%20and%20or-40-N%20filers.pdf.pdf
          https://www.oregon.gov/dor/forms/FormsPubs/form-or-40_101-040_2022.pdf
          https://www.oregon.gov/dor/forms/FormsPubs/form-or-40_101-040_2022.pdf
          https://www.oregon.gov/dor/forms/FormsPubs/form-or-40_101-040_2022.pdf
          https://www.portland.gov/revenue/personal-tax
        
        Mississippi:
        -         
        - https://www.dor.ms.gov/individual/tax-rates
          https://www.dor.ms.gov/individual/tax-rates
          https://www.dor.ms.gov/individual/tax-rates
          https://www.dor.ms.gov/individual
          https://www.dor.ms.gov/individual
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Oregon', 'state_occupation': 'Mississippi', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def oregon_and_missouri(self, capital_gains_short=0, dependents=0, sources=False, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_long=0, filing_status='single', income=100000, credits_state=0):
        """
        --- Parameters ----
        
        Oregon:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Missouri:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Oregon:
        - https://www.oregon.gov/dor/programs/individuals/Pages/PIT.aspx
          https://www.oregon.gov/dor/programs/individuals/Documents/Full%20year%20resident
         %20Form%20or-40%20filers.pdf
          https://www.oregon.gov/dor/programs/individuals/Documents/Part-year%20and%20nonresident
         %20Form%20or-40-P%20and%20or-40-N%20filers.pdf.pdf
          https://www.oregon.gov/dor/forms/FormsPubs/form-or-40_101-040_2022.pdf
          https://www.oregon.gov/dor/forms/FormsPubs/form-or-40_101-040_2022.pdf
          https://www.oregon.gov/dor/forms/FormsPubs/form-or-40_101-040_2022.pdf
          https://www.portland.gov/revenue/personal-tax
        
        Missouri:
        -         
        - https://dor.mo.gov/taxation/individual/tax-types/income/year-changes/
          https://dor.mo.gov/taxation/individual/tax-types/income/year-changes/
          https://dor.mo.gov/taxation/individual/tax-types/income/year-changes/
          https://www.kcmo.gov/city-hall/departments/finance/earnings-tax
          https://www.stlouis-mo.gov/government/departments/collector/earnings-tax/individual-earnings-tax-info.cfm
        """
        
        # Check if the parameters are set upstream in the instance
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Oregon', 'state_occupation': 'Missouri', "capital_gains_short": capital_gains_short, "dependents": dependents, "sources": sources, "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def oregon_and_montana(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Oregon:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Montana:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Oregon:
        - https://www.oregon.gov/dor/programs/individuals/Pages/PIT.aspx
          https://www.oregon.gov/dor/programs/individuals/Documents/Full%20year%20resident
         %20Form%20or-40%20filers.pdf
          https://www.oregon.gov/dor/programs/individuals/Documents/Part-year%20and%20nonresident
         %20Form%20or-40-P%20and%20or-40-N%20filers.pdf.pdf
          https://www.oregon.gov/dor/forms/FormsPubs/form-or-40_101-040_2022.pdf
          https://www.oregon.gov/dor/forms/FormsPubs/form-or-40_101-040_2022.pdf
          https://www.oregon.gov/dor/forms/FormsPubs/form-or-40_101-040_2022.pdf
          https://www.portland.gov/revenue/personal-tax
        
        Montana:
        -         
        - https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/2022/12/montana-employees-withholding-allowance-and-exemption-certificate-form-mw-4.pdf
          https://mtrevenue.gov/2021/11/29/simplification-of-montana-income-taxation/?utm_source=rss&utm_medium=rss&utm_campaign=simplification-of-montana-income-taxation
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Oregon', 'state_occupation': 'Montana', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def oregon_and_nebraska(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Oregon:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Nebraska:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Oregon:
        - https://www.oregon.gov/dor/programs/individuals/Pages/PIT.aspx
          https://www.oregon.gov/dor/programs/individuals/Documents/Full%20year%20resident
         %20Form%20or-40%20filers.pdf
          https://www.oregon.gov/dor/programs/individuals/Documents/Part-year%20and%20nonresident
         %20Form%20or-40-P%20and%20or-40-N%20filers.pdf.pdf
          https://www.oregon.gov/dor/forms/FormsPubs/form-or-40_101-040_2022.pdf
          https://www.oregon.gov/dor/forms/FormsPubs/form-or-40_101-040_2022.pdf
          https://www.oregon.gov/dor/forms/FormsPubs/form-or-40_101-040_2022.pdf
          https://www.portland.gov/revenue/personal-tax
        
        Nebraska:
        -         
        - https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2023_Circular_EN_8-429r2023_Whole_Book_11.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Oregon', 'state_occupation': 'Nebraska', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def oregon_and_nevada(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Oregon:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Nevada:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Oregon:
        - https://www.oregon.gov/dor/programs/individuals/Pages/PIT.aspx
          https://www.oregon.gov/dor/programs/individuals/Documents/Full%20year%20resident
         %20Form%20or-40%20filers.pdf
          https://www.oregon.gov/dor/programs/individuals/Documents/Part-year%20and%20nonresident
         %20Form%20or-40-P%20and%20or-40-N%20filers.pdf.pdf
          https://www.oregon.gov/dor/forms/FormsPubs/form-or-40_101-040_2022.pdf
          https://www.oregon.gov/dor/forms/FormsPubs/form-or-40_101-040_2022.pdf
          https://www.oregon.gov/dor/forms/FormsPubs/form-or-40_101-040_2022.pdf
          https://www.portland.gov/revenue/personal-tax
        
        Nevada:
        -         
        - https://www.nvsos.gov/sos/businesses/the-nevada-advantage
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Oregon', 'state_occupation': 'Nevada', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def oregon_and_new_hampshire(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Oregon:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        New Hampshire:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Oregon:
        - https://www.oregon.gov/dor/programs/individuals/Pages/PIT.aspx
          https://www.oregon.gov/dor/programs/individuals/Documents/Full%20year%20resident
         %20Form%20or-40%20filers.pdf
          https://www.oregon.gov/dor/programs/individuals/Documents/Part-year%20and%20nonresident
         %20Form%20or-40-P%20and%20or-40-N%20filers.pdf.pdf
          https://www.oregon.gov/dor/forms/FormsPubs/form-or-40_101-040_2022.pdf
          https://www.oregon.gov/dor/forms/FormsPubs/form-or-40_101-040_2022.pdf
          https://www.oregon.gov/dor/forms/FormsPubs/form-or-40_101-040_2022.pdf
          https://www.portland.gov/revenue/personal-tax
        
        New Hampshire:
        -         
        - https://www.revenue.nh.gov/assistance/tax-overview.htm
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Oregon', 'state_occupation': 'NewHampshire', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def oregon_and_new_jersey(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Oregon:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        New Jersey:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Oregon:
        - https://www.oregon.gov/dor/programs/individuals/Pages/PIT.aspx
          https://www.oregon.gov/dor/programs/individuals/Documents/Full%20year%20resident
         %20Form%20or-40%20filers.pdf
          https://www.oregon.gov/dor/programs/individuals/Documents/Part-year%20and%20nonresident
         %20Form%20or-40-P%20and%20or-40-N%20filers.pdf.pdf
          https://www.oregon.gov/dor/forms/FormsPubs/form-or-40_101-040_2022.pdf
          https://www.oregon.gov/dor/forms/FormsPubs/form-or-40_101-040_2022.pdf
          https://www.oregon.gov/dor/forms/FormsPubs/form-or-40_101-040_2022.pdf
          https://www.portland.gov/revenue/personal-tax
        
        New Jersey:
        -         
        - https://www.state.nj.us/treasury/taxation/taxtables.shtml
          https://www.state.nj.us/treasury/taxation/njit13.shtml
          https://www.nj.gov/treasury/taxation/njit25.shtml
          https://www.state.nj.us/treasury/taxation/individuals/freqqiti.shtml
          https://www.state.nj.us/treasury/taxation/localtax.shtml
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Oregon', 'state_occupation': 'NewJersey', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def oregon_and_new_mexico(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Oregon:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        New Mexico:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Oregon:
        - https://www.oregon.gov/dor/programs/individuals/Pages/PIT.aspx
          https://www.oregon.gov/dor/programs/individuals/Documents/Full%20year%20resident
         %20Form%20or-40%20filers.pdf
          https://www.oregon.gov/dor/programs/individuals/Documents/Part-year%20and%20nonresident
         %20Form%20or-40-P%20and%20or-40-N%20filers.pdf.pdf
          https://www.oregon.gov/dor/forms/FormsPubs/form-or-40_101-040_2022.pdf
          https://www.oregon.gov/dor/forms/FormsPubs/form-or-40_101-040_2022.pdf
          https://www.oregon.gov/dor/forms/FormsPubs/form-or-40_101-040_2022.pdf
          https://www.portland.gov/revenue/personal-tax
        
        New Mexico:
        -         
        - https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/fdf3c548-8aba-4b9c-9eb4-bb564c716015/FYI-104.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/0bf943fd-652e-4400-bb1b-cd8397eb5a95/2022pit-adj.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/2f1a6781-9534-4436-b427-1557f9592099/2022pit-adj-ins.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/0bf943fd-652e-4400-bb1b-cd8397eb5a95/2022pit-adj.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/1afc56af-ea90-4d48-82e5-1f9aeb43255a/PITbook2022.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/fdf3c548-8aba-4b9c-9eb4-bb564c716015/FYI-104.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Oregon', 'state_occupation': 'NewMexico', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def oregon_and_new_york(self, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Oregon:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        New York:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Oregon:
        - https://www.oregon.gov/dor/programs/individuals/Pages/PIT.aspx
          https://www.oregon.gov/dor/programs/individuals/Documents/Full%20year%20resident
         %20Form%20or-40%20filers.pdf
          https://www.oregon.gov/dor/programs/individuals/Documents/Part-year%20and%20nonresident
         %20Form%20or-40-P%20and%20or-40-N%20filers.pdf.pdf
          https://www.oregon.gov/dor/forms/FormsPubs/form-or-40_101-040_2022.pdf
          https://www.oregon.gov/dor/forms/FormsPubs/form-or-40_101-040_2022.pdf
          https://www.oregon.gov/dor/forms/FormsPubs/form-or-40_101-040_2022.pdf
          https://www.portland.gov/revenue/personal-tax
        
        New York:
        -         
        - https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nys-tax-rate-schedule
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nys-tax-rate-schedule
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nys-tax-rate-schedule
          https://www.tax.ny.gov/pit/
          https://www.tax.ny.gov/pit/file/nonresident-faqs.htm
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#yonkers-tax-rate-schedule
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nyc-tax-rate-schedule
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Oregon', 'state_occupation': 'NewYork', "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def oregon_and_north_carolina(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Oregon:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        North Carolina:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Oregon:
        - https://www.oregon.gov/dor/programs/individuals/Pages/PIT.aspx
          https://www.oregon.gov/dor/programs/individuals/Documents/Full%20year%20resident
         %20Form%20or-40%20filers.pdf
          https://www.oregon.gov/dor/programs/individuals/Documents/Part-year%20and%20nonresident
         %20Form%20or-40-P%20and%20or-40-N%20filers.pdf.pdf
          https://www.oregon.gov/dor/forms/FormsPubs/form-or-40_101-040_2022.pdf
          https://www.oregon.gov/dor/forms/FormsPubs/form-or-40_101-040_2022.pdf
          https://www.oregon.gov/dor/forms/FormsPubs/form-or-40_101-040_2022.pdf
          https://www.portland.gov/revenue/personal-tax
        
        North Carolina:
        -         
        - https://www.ncdor.gov/taxes-forms/tax-rate-schedules
          https://www.ncdor.gov/taxes-forms/individual-income-tax/north-carolina-standard-deduction-or-north-carolina-itemized-deductions
          https://www.ncdor.gov/taxes-forms/individual-income-tax
          https://www.ncdor.gov/taxes-forms/individual-income-tax
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Oregon', 'state_occupation': 'NorthCarolina', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def oregon_and_north_dakota(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Oregon:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        North Dakota:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Oregon:
        - https://www.oregon.gov/dor/programs/individuals/Pages/PIT.aspx
          https://www.oregon.gov/dor/programs/individuals/Documents/Full%20year%20resident
         %20Form%20or-40%20filers.pdf
          https://www.oregon.gov/dor/programs/individuals/Documents/Part-year%20and%20nonresident
         %20Form%20or-40-P%20and%20or-40-N%20filers.pdf.pdf
          https://www.oregon.gov/dor/forms/FormsPubs/form-or-40_101-040_2022.pdf
          https://www.oregon.gov/dor/forms/FormsPubs/form-or-40_101-040_2022.pdf
          https://www.oregon.gov/dor/forms/FormsPubs/form-or-40_101-040_2022.pdf
          https://www.portland.gov/revenue/personal-tax
        
        North Dakota:
        -         
        - https://www.tax.nd.gov/forms
          https://www.tax.nd.gov/sites/www/files/documents/forms/individual/2022-iit/2022-individual-income-tax-booklet.pdf
          https://www.tax.nd.gov/sites/www/files/documents/forms/2022-individual-income-tax-booklet.pdf
          https://www.tax.nd.gov/tax-exemptions-credits/income-tax-exemptions-credits
          https://www.tax.nd.gov/sites/www/files/documents/forms/business/ndwrfillable.pdf
          https://www.tax.nd.gov/news/tax-legislative-changes/significant-changes-law/individual-income-tax-history
          https://www.tax.nd.gov/forms
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Oregon', 'state_occupation': 'NorthDakota', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def oregon_and_ohio(self, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Oregon:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Ohio:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Oregon:
        - https://www.oregon.gov/dor/programs/individuals/Pages/PIT.aspx
          https://www.oregon.gov/dor/programs/individuals/Documents/Full%20year%20resident
         %20Form%20or-40%20filers.pdf
          https://www.oregon.gov/dor/programs/individuals/Documents/Part-year%20and%20nonresident
         %20Form%20or-40-P%20and%20or-40-N%20filers.pdf.pdf
          https://www.oregon.gov/dor/forms/FormsPubs/form-or-40_101-040_2022.pdf
          https://www.oregon.gov/dor/forms/FormsPubs/form-or-40_101-040_2022.pdf
          https://www.oregon.gov/dor/forms/FormsPubs/form-or-40_101-040_2022.pdf
          https://www.portland.gov/revenue/personal-tax
        
        Ohio:
        -         
        - https://tax.ohio.gov/individual/resources/annual-tax-rates
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://tax.ohio.gov/static/forms/employer_withholding/generic/wth_it4nr.pdf
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://www.ritaohio.com/TaxRatesTable
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Oregon', 'state_occupation': 'Ohio', "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def oregon_and_oklahoma(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Oregon:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Oklahoma:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Oregon:
        - https://www.oregon.gov/dor/programs/individuals/Pages/PIT.aspx
          https://www.oregon.gov/dor/programs/individuals/Documents/Full%20year%20resident
         %20Form%20or-40%20filers.pdf
          https://www.oregon.gov/dor/programs/individuals/Documents/Part-year%20and%20nonresident
         %20Form%20or-40-P%20and%20or-40-N%20filers.pdf.pdf
          https://www.oregon.gov/dor/forms/FormsPubs/form-or-40_101-040_2022.pdf
          https://www.oregon.gov/dor/forms/FormsPubs/form-or-40_101-040_2022.pdf
          https://www.oregon.gov/dor/forms/FormsPubs/form-or-40_101-040_2022.pdf
          https://www.portland.gov/revenue/personal-tax
        
        Oklahoma:
        -         
        - https://oklahoma.gov/content/dam/ok/en/tax/documents/resources/publications/businesses/withholding-tables/WHTables-2022.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Oregon', 'state_occupation': 'Oklahoma', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def oregon_and_pennsylvania(self, county_residence='Other', state_agi=75000, county_occupation='Other', filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Oregon:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Pennsylvania:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        county_residence : str, optional
            County of residence. Default is Other.
        county_occupation : str, optional
            County of occupation. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Oregon:
        - https://www.oregon.gov/dor/programs/individuals/Pages/PIT.aspx
          https://www.oregon.gov/dor/programs/individuals/Documents/Full%20year%20resident
         %20Form%20or-40%20filers.pdf
          https://www.oregon.gov/dor/programs/individuals/Documents/Part-year%20and%20nonresident
         %20Form%20or-40-P%20and%20or-40-N%20filers.pdf.pdf
          https://www.oregon.gov/dor/forms/FormsPubs/form-or-40_101-040_2022.pdf
          https://www.oregon.gov/dor/forms/FormsPubs/form-or-40_101-040_2022.pdf
          https://www.oregon.gov/dor/forms/FormsPubs/form-or-40_101-040_2022.pdf
          https://www.portland.gov/revenue/personal-tax
        
        Pennsylvania:
        -         
        - https://www.revenue.pa.gov/Tax%20Rates/Pages/default.aspx
          https://www.revenue.pa.gov/FormsandPublications/FormsforIndividuals/PIT/Documents/2022/2022_pa-40in.pdf
          https://www.revenue.pa.gov/FormsandPublications/FormsforBusinesses/EmployerWithholding/Documents/rev-419.pdf
          https://www.revenue.pa.gov/FormsandPublications/FormsforIndividuals/PIT/Documents/2022/2022_pa-40in.pdf
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-25.htm
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        county_occupation = self.county_occupation if hasattr(self, 'county_occupation') else county_occupation
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Oregon', 'state_occupation': 'Pennsylvania', "county_residence": county_residence, "state_agi": state_agi, "county_occupation": county_occupation, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def oregon_and_rhode_island(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Oregon:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Rhode Island:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Oregon:
        - https://www.oregon.gov/dor/programs/individuals/Pages/PIT.aspx
          https://www.oregon.gov/dor/programs/individuals/Documents/Full%20year%20resident
         %20Form%20or-40%20filers.pdf
          https://www.oregon.gov/dor/programs/individuals/Documents/Part-year%20and%20nonresident
         %20Form%20or-40-P%20and%20or-40-N%20filers.pdf.pdf
          https://www.oregon.gov/dor/forms/FormsPubs/form-or-40_101-040_2022.pdf
          https://www.oregon.gov/dor/forms/FormsPubs/form-or-40_101-040_2022.pdf
          https://www.oregon.gov/dor/forms/FormsPubs/form-or-40_101-040_2022.pdf
          https://www.portland.gov/revenue/personal-tax
        
        Rhode Island:
        -         
        - https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2022_1040WE_w_0.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Oregon', 'state_occupation': 'RhodeIsland', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def oregon_and_south_carolina(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Oregon:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        South Carolina:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Oregon:
        - https://www.oregon.gov/dor/programs/individuals/Pages/PIT.aspx
          https://www.oregon.gov/dor/programs/individuals/Documents/Full%20year%20resident
         %20Form%20or-40%20filers.pdf
          https://www.oregon.gov/dor/programs/individuals/Documents/Part-year%20and%20nonresident
         %20Form%20or-40-P%20and%20or-40-N%20filers.pdf.pdf
          https://www.oregon.gov/dor/forms/FormsPubs/form-or-40_101-040_2022.pdf
          https://www.oregon.gov/dor/forms/FormsPubs/form-or-40_101-040_2022.pdf
          https://www.oregon.gov/dor/forms/FormsPubs/form-or-40_101-040_2022.pdf
          https://www.portland.gov/revenue/personal-tax
        
        South Carolina:
        -         
        - https://dor.sc.gov/resources-site/lawandpolicy/Advisory%20Opinions/IL22-15.pdf
          https://dor.sc.gov/forms-site/Forms/IITPacket_2022.pdf
          https://www.scstatehouse.gov/code/t12c006.php
          https://dor.sc.gov/resources-site/lawandpolicy/Advisory%20Opinions/IL22-15.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Oregon', 'state_occupation': 'SouthCarolina', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def oregon_and_south_dakota(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Oregon:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        South Dakota:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Oregon:
        - https://www.oregon.gov/dor/programs/individuals/Pages/PIT.aspx
          https://www.oregon.gov/dor/programs/individuals/Documents/Full%20year%20resident
         %20Form%20or-40%20filers.pdf
          https://www.oregon.gov/dor/programs/individuals/Documents/Part-year%20and%20nonresident
         %20Form%20or-40-P%20and%20or-40-N%20filers.pdf.pdf
          https://www.oregon.gov/dor/forms/FormsPubs/form-or-40_101-040_2022.pdf
          https://www.oregon.gov/dor/forms/FormsPubs/form-or-40_101-040_2022.pdf
          https://www.oregon.gov/dor/forms/FormsPubs/form-or-40_101-040_2022.pdf
          https://www.portland.gov/revenue/personal-tax
        
        South Dakota:
        -         
        - https://dor.sd.gov/individuals/taxes/
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Oregon', 'state_occupation': 'SouthDakota', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def oregon_and_tennessee(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Oregon:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Tennessee:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Oregon:
        - https://www.oregon.gov/dor/programs/individuals/Pages/PIT.aspx
          https://www.oregon.gov/dor/programs/individuals/Documents/Full%20year%20resident
         %20Form%20or-40%20filers.pdf
          https://www.oregon.gov/dor/programs/individuals/Documents/Part-year%20and%20nonresident
         %20Form%20or-40-P%20and%20or-40-N%20filers.pdf.pdf
          https://www.oregon.gov/dor/forms/FormsPubs/form-or-40_101-040_2022.pdf
          https://www.oregon.gov/dor/forms/FormsPubs/form-or-40_101-040_2022.pdf
          https://www.oregon.gov/dor/forms/FormsPubs/form-or-40_101-040_2022.pdf
          https://www.portland.gov/revenue/personal-tax
        
        Tennessee:
        -         
        - https://www.tn.gov/revenue/taxes.html
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Oregon', 'state_occupation': 'Tennessee', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def oregon_and_texas(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Oregon:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Texas:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Oregon:
        - https://www.oregon.gov/dor/programs/individuals/Pages/PIT.aspx
          https://www.oregon.gov/dor/programs/individuals/Documents/Full%20year%20resident
         %20Form%20or-40%20filers.pdf
          https://www.oregon.gov/dor/programs/individuals/Documents/Part-year%20and%20nonresident
         %20Form%20or-40-P%20and%20or-40-N%20filers.pdf.pdf
          https://www.oregon.gov/dor/forms/FormsPubs/form-or-40_101-040_2022.pdf
          https://www.oregon.gov/dor/forms/FormsPubs/form-or-40_101-040_2022.pdf
          https://www.oregon.gov/dor/forms/FormsPubs/form-or-40_101-040_2022.pdf
          https://www.portland.gov/revenue/personal-tax
        
        Texas:
        -         
        - https://comptroller.texas.gov/taxes/
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Oregon', 'state_occupation': 'Texas', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def oregon_and_utah(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Oregon:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Utah:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Oregon:
        - https://www.oregon.gov/dor/programs/individuals/Pages/PIT.aspx
          https://www.oregon.gov/dor/programs/individuals/Documents/Full%20year%20resident
         %20Form%20or-40%20filers.pdf
          https://www.oregon.gov/dor/programs/individuals/Documents/Part-year%20and%20nonresident
         %20Form%20or-40-P%20and%20or-40-N%20filers.pdf.pdf
          https://www.oregon.gov/dor/forms/FormsPubs/form-or-40_101-040_2022.pdf
          https://www.oregon.gov/dor/forms/FormsPubs/form-or-40_101-040_2022.pdf
          https://www.oregon.gov/dor/forms/FormsPubs/form-or-40_101-040_2022.pdf
          https://www.portland.gov/revenue/personal-tax
        
        Utah:
        -         
        - https://incometax.utah.gov/paying/tax-rates
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Oregon', 'state_occupation': 'Utah', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def oregon_and_vermont(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Oregon:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Vermont:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Oregon:
        - https://www.oregon.gov/dor/programs/individuals/Pages/PIT.aspx
          https://www.oregon.gov/dor/programs/individuals/Documents/Full%20year%20resident
         %20Form%20or-40%20filers.pdf
          https://www.oregon.gov/dor/programs/individuals/Documents/Part-year%20and%20nonresident
         %20Form%20or-40-P%20and%20or-40-N%20filers.pdf.pdf
          https://www.oregon.gov/dor/forms/FormsPubs/form-or-40_101-040_2022.pdf
          https://www.oregon.gov/dor/forms/FormsPubs/form-or-40_101-040_2022.pdf
          https://www.oregon.gov/dor/forms/FormsPubs/form-or-40_101-040_2022.pdf
          https://www.portland.gov/revenue/personal-tax
        
        Vermont:
        -         
        - https://tax.vermont.gov/sites/tax/files/documents/GB-1210-2023.pdf
          https://tax.vermont.gov/sites/tax/files/documents/Income%20Booklet-2022.pdf
          https://tax.vermont.gov/sites/tax/files/documents/Income%20Booklet-2022.pdf
          https://tax.vermont.gov/sites/tax/files/documents/Income%20Booklet-2022.pdf
          https://tax.vermont.gov/individuals/personal-income-tax/tax-credits
          https://tax.vermont.gov/individuals/personal-income-tax/taxable-income
          https://tax.vermont.gov/individuals/personal-income-tax/local-option-tax
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Oregon', 'state_occupation': 'Vermont', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def oregon_and_virginia(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Oregon:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Virginia:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Oregon:
        - https://www.oregon.gov/dor/programs/individuals/Pages/PIT.aspx
          https://www.oregon.gov/dor/programs/individuals/Documents/Full%20year%20resident
         %20Form%20or-40%20filers.pdf
          https://www.oregon.gov/dor/programs/individuals/Documents/Part-year%20and%20nonresident
         %20Form%20or-40-P%20and%20or-40-N%20filers.pdf.pdf
          https://www.oregon.gov/dor/forms/FormsPubs/form-or-40_101-040_2022.pdf
          https://www.oregon.gov/dor/forms/FormsPubs/form-or-40_101-040_2022.pdf
          https://www.oregon.gov/dor/forms/FormsPubs/form-or-40_101-040_2022.pdf
          https://www.portland.gov/revenue/personal-tax
        
        Virginia:
        -         
        - https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/reciprocity
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Oregon', 'state_occupation': 'Virginia', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def oregon_and_washington(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Oregon:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Washington:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Oregon:
        - https://www.oregon.gov/dor/programs/individuals/Pages/PIT.aspx
          https://www.oregon.gov/dor/programs/individuals/Documents/Full%20year%20resident
         %20Form%20or-40%20filers.pdf
          https://www.oregon.gov/dor/programs/individuals/Documents/Part-year%20and%20nonresident
         %20Form%20or-40-P%20and%20or-40-N%20filers.pdf.pdf
          https://www.oregon.gov/dor/forms/FormsPubs/form-or-40_101-040_2022.pdf
          https://www.oregon.gov/dor/forms/FormsPubs/form-or-40_101-040_2022.pdf
          https://www.oregon.gov/dor/forms/FormsPubs/form-or-40_101-040_2022.pdf
          https://www.portland.gov/revenue/personal-tax
        
        Washington:
        -         
        - https://dor.wa.gov/taxes-rates/income-tax
          https://dor.wa.gov/taxes-rates/other-taxes/capital-gains-tax
          https://dor.wa.gov/taxes-rates/income-tax
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Oregon', 'state_occupation': 'Washington', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def oregon_and_west_virginia(self, pay_periods=24, capital_gains_short=0, county_occupation='Other', dependents=0, sources=False, state_agi=75000, filers_over_65=0, capital_gains_long=0, filing_status='single', income=100000, credits_state=0):
        """
        --- Parameters ----
        
        Oregon:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        West Virginia:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_occupation : str, optional
            County of occupation. Default is Other.
        pay_periods : int, optional
            Number of pay periods. Options are 13, 24, 26, 52. Default is 24.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Oregon:
        - https://www.oregon.gov/dor/programs/individuals/Pages/PIT.aspx
          https://www.oregon.gov/dor/programs/individuals/Documents/Full%20year%20resident
         %20Form%20or-40%20filers.pdf
          https://www.oregon.gov/dor/programs/individuals/Documents/Part-year%20and%20nonresident
         %20Form%20or-40-P%20and%20or-40-N%20filers.pdf.pdf
          https://www.oregon.gov/dor/forms/FormsPubs/form-or-40_101-040_2022.pdf
          https://www.oregon.gov/dor/forms/FormsPubs/form-or-40_101-040_2022.pdf
          https://www.oregon.gov/dor/forms/FormsPubs/form-or-40_101-040_2022.pdf
          https://www.portland.gov/revenue/personal-tax
        
        West Virginia:
        -         
        - https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://tax.wv.gov/Documents/TSD/tsd438.pdf
          https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2011/Taxes-11-35att.html
        """
        
        # Check if the parameters are set upstream in the instance
        pay_periods = self.pay_periods if hasattr(self, 'pay_periods') else pay_periods
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        county_occupation = self.county_occupation if hasattr(self, 'county_occupation') else county_occupation
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Oregon', 'state_occupation': 'WestVirginia', "pay_periods": pay_periods, "capital_gains_short": capital_gains_short, "county_occupation": county_occupation, "dependents": dependents, "sources": sources, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def oregon_and_wisconsin(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Oregon:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Wisconsin:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Oregon:
        - https://www.oregon.gov/dor/programs/individuals/Pages/PIT.aspx
          https://www.oregon.gov/dor/programs/individuals/Documents/Full%20year%20resident
         %20Form%20or-40%20filers.pdf
          https://www.oregon.gov/dor/programs/individuals/Documents/Part-year%20and%20nonresident
         %20Form%20or-40-P%20and%20or-40-N%20filers.pdf.pdf
          https://www.oregon.gov/dor/forms/FormsPubs/form-or-40_101-040_2022.pdf
          https://www.oregon.gov/dor/forms/FormsPubs/form-or-40_101-040_2022.pdf
          https://www.oregon.gov/dor/forms/FormsPubs/form-or-40_101-040_2022.pdf
          https://www.portland.gov/revenue/personal-tax
        
        Wisconsin:
        -         
        - https://www.revenue.wi.gov/TaxForms2023/2023-Form1-ES-Inst.pdf
          https://www.revenue.wi.gov/TaxForms2023/2023-Form1-ES-Inst.pdf
          https://www.revenue.wi.gov/TaxForms2023/2023-Form1-ES-Inst.pdf
          https://www.revenue.wi.gov/Pages/FAQS/pcs-work.aspx
          https://www.revenue.wi.gov/Dor%20Publications/pb103.pdf
          https://www.revenue.wi.gov/Pages/FAQS/pcs-work.aspx
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Oregon', 'state_occupation': 'Wisconsin', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def oregon_and_wyoming(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Oregon:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Wyoming:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Oregon:
        - https://www.oregon.gov/dor/programs/individuals/Pages/PIT.aspx
          https://www.oregon.gov/dor/programs/individuals/Documents/Full%20year%20resident
         %20Form%20or-40%20filers.pdf
          https://www.oregon.gov/dor/programs/individuals/Documents/Part-year%20and%20nonresident
         %20Form%20or-40-P%20and%20or-40-N%20filers.pdf.pdf
          https://www.oregon.gov/dor/forms/FormsPubs/form-or-40_101-040_2022.pdf
          https://www.oregon.gov/dor/forms/FormsPubs/form-or-40_101-040_2022.pdf
          https://www.oregon.gov/dor/forms/FormsPubs/form-or-40_101-040_2022.pdf
          https://www.portland.gov/revenue/personal-tax
        
        Wyoming:
        -         
        - https://revenue.wyo.gov/
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Oregon', 'state_occupation': 'Wyoming', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def pennsylvania_and_alabama(self, county_residence='Other', state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', county_occupation='Other', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Pennsylvania:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        county_residence : str, optional
            County of residence. Default is Other.
        county_occupation : str, optional
            County of occupation. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        
        Alabama:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Pennsylvania:
        - https://www.revenue.pa.gov/Tax%20Rates/Pages/default.aspx
          https://www.revenue.pa.gov/FormsandPublications/FormsforIndividuals/PIT/Documents/2022/2022_pa-40in.pdf
          https://www.revenue.pa.gov/FormsandPublications/FormsforBusinesses/EmployerWithholding/Documents/rev-419.pdf
          https://www.revenue.pa.gov/FormsandPublications/FormsforIndividuals/PIT/Documents/2022/2022_pa-40in.pdf
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-25.htm
        
        Alabama:
        -         
        - https://www.revenue.alabama.gov/ultraviewer/viewer/basic_viewer/index.html?form=2023/02/22f40.pdf
          https://www.revenue.alabama.gov/faqs/what-is-alabamas-individual-income-tax-rate/
          https://www.revenue.alabama.gov/ultraviewer/viewer/basic_viewer/index.html?form=2023/02/22f40.pdf
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-24.htm?taxmap=true
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-24.htm?taxmap=true
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-24.htm?taxmap=true
          https://www.revenue.alabama.gov/individual-corporate/income-to-be-reported-on-the-alabama-income-tax-return/
          https://www.almonline.org/TaxRates.aspx#OccupationalTax
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        county_occupation = self.county_occupation if hasattr(self, 'county_occupation') else county_occupation
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Pennsylvania', 'state_occupation': 'Alabama', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "county_occupation": county_occupation, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def pennsylvania_and_alaska(self, county_residence='Other', state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, county_occupation='Other', sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Pennsylvania:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        county_residence : str, optional
            County of residence. Default is Other.
        county_occupation : str, optional
            County of occupation. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        
        Alaska:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Pennsylvania:
        - https://www.revenue.pa.gov/Tax%20Rates/Pages/default.aspx
          https://www.revenue.pa.gov/FormsandPublications/FormsforIndividuals/PIT/Documents/2022/2022_pa-40in.pdf
          https://www.revenue.pa.gov/FormsandPublications/FormsforBusinesses/EmployerWithholding/Documents/rev-419.pdf
          https://www.revenue.pa.gov/FormsandPublications/FormsforIndividuals/PIT/Documents/2022/2022_pa-40in.pdf
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-25.htm
        
        Alaska:
        -         
        - https://www.commerce.alaska.gov/web/dcra/officeofthestateassessor/alaskataxfacts.aspx
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        county_occupation = self.county_occupation if hasattr(self, 'county_occupation') else county_occupation
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Pennsylvania', 'state_occupation': 'Alaska', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "county_occupation": county_occupation, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def pennsylvania_and_arizona(self, county_residence='Other', state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', county_occupation='Other', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Pennsylvania:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        county_residence : str, optional
            County of residence. Default is Other.
        county_occupation : str, optional
            County of occupation. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        
        Arizona:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        credits_state : int, optional
            Credits for state taxes. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Pennsylvania:
        - https://www.revenue.pa.gov/Tax%20Rates/Pages/default.aspx
          https://www.revenue.pa.gov/FormsandPublications/FormsforIndividuals/PIT/Documents/2022/2022_pa-40in.pdf
          https://www.revenue.pa.gov/FormsandPublications/FormsforBusinesses/EmployerWithholding/Documents/rev-419.pdf
          https://www.revenue.pa.gov/FormsandPublications/FormsforIndividuals/PIT/Documents/2022/2022_pa-40in.pdf
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-25.htm
        
        Arizona:
        -         
        - https://azdor.gov/forms/individual/form-140-x-y-tables
          https://azdor.gov/forms/individual/form-140-x-y-tables
          https://azdor.gov/forms/individual/form-140-x-y-tables
          https://azdor.gov/businesses-arizona/withholding-tax/withholding-exceptions
          "https://azdor.gov/individuals/income-tax-filing-assistance/identifying-other-taxable-income#:~:text=In%20regards%20to%20capital%20gains
         the%20individuals%20regular%20tax%20rate.
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        county_occupation = self.county_occupation if hasattr(self, 'county_occupation') else county_occupation
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Pennsylvania', 'state_occupation': 'Arizona', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "county_occupation": county_occupation, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def pennsylvania_and_arkansas(self, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', county_occupation='Other', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Pennsylvania:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        county_residence : str, optional
            County of residence. Default is Other.
        county_occupation : str, optional
            County of occupation. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        
        Arkansas:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Pennsylvania:
        - https://www.revenue.pa.gov/Tax%20Rates/Pages/default.aspx
          https://www.revenue.pa.gov/FormsandPublications/FormsforIndividuals/PIT/Documents/2022/2022_pa-40in.pdf
          https://www.revenue.pa.gov/FormsandPublications/FormsforBusinesses/EmployerWithholding/Documents/rev-419.pdf
          https://www.revenue.pa.gov/FormsandPublications/FormsforIndividuals/PIT/Documents/2022/2022_pa-40in.pdf
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-25.htm
        
        Arkansas:
        -         
        - https://www.arkansasedc.com/why-arkansas/business-climate/tax-structure/personal-income-tax
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_and_AR1000NR_Instructions.pdf
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_FullYearResidentIndividualIncomeTaxReturn.pdf
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_and_AR1000NR_Instructions.pdf
          https://www.arkansasedc.com/why-arkansas/business-climate/tax-structure/capital-gains-tax-reduction
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_and_AR1000NR_Instructions.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        county_occupation = self.county_occupation if hasattr(self, 'county_occupation') else county_occupation
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Pennsylvania', 'state_occupation': 'Arkansas', "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "county_occupation": county_occupation, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def pennsylvania_and_california(self, county_residence='Other', state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', county_occupation='Other', sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Pennsylvania:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        county_residence : str, optional
            County of residence. Default is Other.
        county_occupation : str, optional
            County of occupation. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        
        California:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Pennsylvania:
        - https://www.revenue.pa.gov/Tax%20Rates/Pages/default.aspx
          https://www.revenue.pa.gov/FormsandPublications/FormsforIndividuals/PIT/Documents/2022/2022_pa-40in.pdf
          https://www.revenue.pa.gov/FormsandPublications/FormsforBusinesses/EmployerWithholding/Documents/rev-419.pdf
          https://www.revenue.pa.gov/FormsandPublications/FormsforIndividuals/PIT/Documents/2022/2022_pa-40in.pdf
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-25.htm
        
        California:
        -         
        - https://www.ftb.ca.gov/forms/2022/2022-540-tax-rate-schedules.pdf
          https://www.ftb.ca.gov/file/personal/deductions/index.html
          https://www.ftb.ca.gov/about-ftb/newsroom/caleitc/eligibility-and-credit-information.html
          https://www.ftb.ca.gov/file/personal/income-types/capital-gains-and-losses.html#:~:text=California%20does%20not%20have%20a
         are%20taxed%20as%20ordinary%20income
          https://www.ftb.ca.gov/file/personal/
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        county_occupation = self.county_occupation if hasattr(self, 'county_occupation') else county_occupation
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Pennsylvania', 'state_occupation': 'California', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "county_occupation": county_occupation, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def pennsylvania_and_colorado(self, county_residence='Other', state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, county_occupation='Other', sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Pennsylvania:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        county_residence : str, optional
            County of residence. Default is Other.
        county_occupation : str, optional
            County of occupation. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        
        Colorado:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Pennsylvania:
        - https://www.revenue.pa.gov/Tax%20Rates/Pages/default.aspx
          https://www.revenue.pa.gov/FormsandPublications/FormsforIndividuals/PIT/Documents/2022/2022_pa-40in.pdf
          https://www.revenue.pa.gov/FormsandPublications/FormsforBusinesses/EmployerWithholding/Documents/rev-419.pdf
          https://www.revenue.pa.gov/FormsandPublications/FormsforIndividuals/PIT/Documents/2022/2022_pa-40in.pdf
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-25.htm
        
        Colorado:
        -         
        - https://leg.colorado.gov/agencies/legislative-council-staff/individual-income-tax%C2%A0
          https://leg.colorado.gov/agencies/legislative-council-staff/individual-income-tax%C2%A0
          https://tax.colorado.gov/sites/tax/files/documents/ITT_Capital_Gain_Subtraction_Feb_2022.pdf
          https://www.auroragov.org/business_services/taxes/occupational_privilege_tax
          https://www.denvergov.org/Government/Agencies-Departments-Offices/Agencies-Departments-Offices-Directory/Department-of-Finance/Our-Divisions/Treasury/Business-Tax-Information#:~:text=Occupational%20privilege%20tax
         -%28show%20below%29&text=Each%20taxable%20employee%20is%20liable
         month%20for%20each%20taxable%20employee
          https://www.glendale.co.us/355/Occupational-Privilege-Tax
          https://www.greenwoodvillage.com/1220/Occupational-Privilege-Tax-OPT
          https://ci.sheridan.co.us/288/Occupational-Privilege-Tax
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        county_occupation = self.county_occupation if hasattr(self, 'county_occupation') else county_occupation
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Pennsylvania', 'state_occupation': 'Colorado', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "county_occupation": county_occupation, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def pennsylvania_and_connecticut(self, county_residence='Other', state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', county_occupation='Other', sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Pennsylvania:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        county_residence : str, optional
            County of residence. Default is Other.
        county_occupation : str, optional
            County of occupation. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        
        Connecticut:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Pennsylvania:
        - https://www.revenue.pa.gov/Tax%20Rates/Pages/default.aspx
          https://www.revenue.pa.gov/FormsandPublications/FormsforIndividuals/PIT/Documents/2022/2022_pa-40in.pdf
          https://www.revenue.pa.gov/FormsandPublications/FormsforBusinesses/EmployerWithholding/Documents/rev-419.pdf
          https://www.revenue.pa.gov/FormsandPublications/FormsforIndividuals/PIT/Documents/2022/2022_pa-40in.pdf
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-25.htm
        
        Connecticut:
        -         
        - https://portal.ct.gov/-/media/DRS/Forms/2022/Income/2022-CT-1040-Instructions_1222.pdf
          https://www.cga.ct.gov/2022/rpt/pdf/2022-R-0108.pdf
          https://www.cga.ct.gov/current/pub/chap_229.htm#sec_12-703
          https://portal.ct.gov/-/media/DRS/Forms/2022/Income/2022-CT-1040-Instructions_1222.pdf
          "https://portal.ct.gov/DRS/Publications/TSSNs/TSSN-29#:~:text=An%20individuals%20net%20capital%20gains
         portion%20of%20Social%20Security%20benefits"
          "https://portal.ct.gov/DRS/Publications/TSSNs/TSSN-29#:~:text=An%20individuals%20net%20capital%20gains
         portion%20of%20Social%20Security%20benefits"
          https://portal.ct.gov/DRS/Individuals/Resident-Income-Tax/Tax-Information
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        county_occupation = self.county_occupation if hasattr(self, 'county_occupation') else county_occupation
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Pennsylvania', 'state_occupation': 'Connecticut', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "county_occupation": county_occupation, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def pennsylvania_and_delaware(self, county_residence='Other', state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, federal_agi=75000, county_occupation='Other', sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Pennsylvania:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        county_residence : str, optional
            County of residence. Default is Other.
        county_occupation : str, optional
            County of occupation. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        
        Delaware:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        federal_agi : int, optional
            Adjusted Gross Income (AGI) for federal taxes. Default is 0. Sources - https://www.irs.gov/e-file-providers/definition-of-adjusted-gross-income, https://www.irs.gov/newsroom/taxpayers-should-know-the-difference-between-standard-and-itemized-deductions
        income : int, optional
            Gross income. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Pennsylvania:
        - https://www.revenue.pa.gov/Tax%20Rates/Pages/default.aspx
          https://www.revenue.pa.gov/FormsandPublications/FormsforIndividuals/PIT/Documents/2022/2022_pa-40in.pdf
          https://www.revenue.pa.gov/FormsandPublications/FormsforBusinesses/EmployerWithholding/Documents/rev-419.pdf
          https://www.revenue.pa.gov/FormsandPublications/FormsforIndividuals/PIT/Documents/2022/2022_pa-40in.pdf
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-25.htm
        
        Delaware:
        -         
        - https://revenue.delaware.gov/software-developer/tax-rate-changes/
          https://revenuefiles.delaware.gov/2022/PIT-RES_TY22_2022-01_PaperInteractive.pdf
          https://revenuefiles.delaware.gov/2022/PIT-RES_TY22_2022-02_Instructions.pdf
          https://revenue.delaware.gov/frequently-asked-questions/personal-income-tax-faqs/
          https://www.newcastlede.gov/DocumentCenter/View/20147/City-of-Wilmington-Wage-Tax-Information-PDF
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        federal_agi = self.federal_agi if hasattr(self, 'federal_agi') else federal_agi
        county_occupation = self.county_occupation if hasattr(self, 'county_occupation') else county_occupation
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Pennsylvania', 'state_occupation': 'Delaware', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "federal_agi": federal_agi, "county_occupation": county_occupation, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def pennsylvania_and_florida(self, county_residence='Other', state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, county_occupation='Other', sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Pennsylvania:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        county_residence : str, optional
            County of residence. Default is Other.
        county_occupation : str, optional
            County of occupation. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        
        Florida:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Pennsylvania:
        - https://www.revenue.pa.gov/Tax%20Rates/Pages/default.aspx
          https://www.revenue.pa.gov/FormsandPublications/FormsforIndividuals/PIT/Documents/2022/2022_pa-40in.pdf
          https://www.revenue.pa.gov/FormsandPublications/FormsforBusinesses/EmployerWithholding/Documents/rev-419.pdf
          https://www.revenue.pa.gov/FormsandPublications/FormsforIndividuals/PIT/Documents/2022/2022_pa-40in.pdf
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-25.htm
        
        Florida:
        -         
        - https://floridarevenue.com/taxes/taxesfees/Pages/tax_interest_rates.aspx
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        county_occupation = self.county_occupation if hasattr(self, 'county_occupation') else county_occupation
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Pennsylvania', 'state_occupation': 'Florida', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "county_occupation": county_occupation, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def pennsylvania_and_georgia(self, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', county_occupation='Other', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Pennsylvania:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        county_residence : str, optional
            County of residence. Default is Other.
        county_occupation : str, optional
            County of occupation. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        
        Georgia:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Pennsylvania:
        - https://www.revenue.pa.gov/Tax%20Rates/Pages/default.aspx
          https://www.revenue.pa.gov/FormsandPublications/FormsforIndividuals/PIT/Documents/2022/2022_pa-40in.pdf
          https://www.revenue.pa.gov/FormsandPublications/FormsforBusinesses/EmployerWithholding/Documents/rev-419.pdf
          https://www.revenue.pa.gov/FormsandPublications/FormsforIndividuals/PIT/Documents/2022/2022_pa-40in.pdf
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-25.htm
        
        Georgia:
        -         
        - https://dor.georgia.gov/tax-tables-georgia-tax-rate-schedule
          https://apps.dor.ga.gov/FillableForms/PDFViewer/Index?form=2022GA500
          https://apps.dor.ga.gov/FillableForms/PDFViewer/Index?form=2022GA500
          https://apps.dor.ga.gov/FillableForms/PDFViewer/Index?form=2022GA500
          https://dor.georgia.gov/years-individual-income-tax-forms
          https://dor.georgia.gov/years-individual-income-tax-forms
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        county_occupation = self.county_occupation if hasattr(self, 'county_occupation') else county_occupation
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Pennsylvania', 'state_occupation': 'Georgia', "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "county_occupation": county_occupation, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def pennsylvania_and_hawaii(self, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', county_occupation='Other', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Pennsylvania:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        county_residence : str, optional
            County of residence. Default is Other.
        county_occupation : str, optional
            County of occupation. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        
        Hawaii:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Pennsylvania:
        - https://www.revenue.pa.gov/Tax%20Rates/Pages/default.aspx
          https://www.revenue.pa.gov/FormsandPublications/FormsforIndividuals/PIT/Documents/2022/2022_pa-40in.pdf
          https://www.revenue.pa.gov/FormsandPublications/FormsforBusinesses/EmployerWithholding/Documents/rev-419.pdf
          https://www.revenue.pa.gov/FormsandPublications/FormsforIndividuals/PIT/Documents/2022/2022_pa-40in.pdf
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-25.htm
        
        Hawaii:
        -         
        - https://tax.hawaii.gov/forms/d_18table-on/d_18table-on_p13/
          https://files.hawaii.gov/tax/forms/2018/18table-on.pdf
          https://files.hawaii.gov/tax/news/pubs/22outline.pdf
          https://files.hawaii.gov/tax/news/pubs/FastTaxRefGuide_2017.pdf
          https://www.tfhawaii.org/wordpress/state-tax-resources/mini-tax-guides/individual-income-tax-chapter-235/#:~:text=Long%20term%20capital%20gains%20are%20taxed%20at%20a%20maximum%20of%207.25%25
          https://www.grassrootinstitute.org/2023/02/hb337-increasing-tax-on-capital-gains-a-bad-risk/
          https://files.hawaii.gov/tax/news/pubs/22outline.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        county_occupation = self.county_occupation if hasattr(self, 'county_occupation') else county_occupation
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Pennsylvania', 'state_occupation': 'Hawaii', "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "county_occupation": county_occupation, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def pennsylvania_and_idaho(self, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', county_occupation='Other', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Pennsylvania:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        county_residence : str, optional
            County of residence. Default is Other.
        county_occupation : str, optional
            County of occupation. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        
        Idaho:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Pennsylvania:
        - https://www.revenue.pa.gov/Tax%20Rates/Pages/default.aspx
          https://www.revenue.pa.gov/FormsandPublications/FormsforIndividuals/PIT/Documents/2022/2022_pa-40in.pdf
          https://www.revenue.pa.gov/FormsandPublications/FormsforBusinesses/EmployerWithholding/Documents/rev-419.pdf
          https://www.revenue.pa.gov/FormsandPublications/FormsforIndividuals/PIT/Documents/2022/2022_pa-40in.pdf
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-25.htm
        
        Idaho:
        -         
        - https://tax.idaho.gov/taxes/income-tax/individual-income/individual-income-tax-rate-schedule/
          https://tax.idaho.gov/wp-content/uploads/forms/EIN00046/EIN00046_03-01-2023.pdf
          https://tax.idaho.gov/taxes/income-tax/individual-income/idaho-source-income/
          https://tax.idaho.gov/taxes/income-tax/individual-income/idaho-source-income/
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        county_occupation = self.county_occupation if hasattr(self, 'county_occupation') else county_occupation
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Pennsylvania', 'state_occupation': 'Idaho', "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "county_occupation": county_occupation, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def pennsylvania_and_illinois(self, county_residence='Other', state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', county_occupation='Other', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Pennsylvania:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        county_residence : str, optional
            County of residence. Default is Other.
        county_occupation : str, optional
            County of occupation. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        
        Illinois:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Pennsylvania:
        - https://www.revenue.pa.gov/Tax%20Rates/Pages/default.aspx
          https://www.revenue.pa.gov/FormsandPublications/FormsforIndividuals/PIT/Documents/2022/2022_pa-40in.pdf
          https://www.revenue.pa.gov/FormsandPublications/FormsforBusinesses/EmployerWithholding/Documents/rev-419.pdf
          https://www.revenue.pa.gov/FormsandPublications/FormsforIndividuals/PIT/Documents/2022/2022_pa-40in.pdf
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-25.htm
        
        Illinois:
        -         
        - https://tax.illinois.gov/research/taxrates/income.html#IndividualIncome
          https://tax.illinois.gov/content/dam/soi/en/web/tax/forms/incometax/documents/currentyear/individual/il-1040-instr.pdf
          https://tax.illinois.gov/questionsandanswers/12.html
          https://tax.illinois.gov/content/dam/soi/en/web/tax/forms/incometax/documents/currentyear/individual/il-1040-instr.pdf
          https://tax.illinois.gov/research/taxrates/personalproperty.html
          https://tax.illinois.gov/localgovernments/income.html
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        county_occupation = self.county_occupation if hasattr(self, 'county_occupation') else county_occupation
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Pennsylvania', 'state_occupation': 'Illinois', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "county_occupation": county_occupation, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def pennsylvania_and_indiana(self, county_residence='Other', state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, county_occupation='Other', sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Pennsylvania:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        county_residence : str, optional
            County of residence. Default is Other.
        county_occupation : str, optional
            County of occupation. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        
        Indiana:
        


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Pennsylvania:
        - https://www.revenue.pa.gov/Tax%20Rates/Pages/default.aspx
          https://www.revenue.pa.gov/FormsandPublications/FormsforIndividuals/PIT/Documents/2022/2022_pa-40in.pdf
          https://www.revenue.pa.gov/FormsandPublications/FormsforBusinesses/EmployerWithholding/Documents/rev-419.pdf
          https://www.revenue.pa.gov/FormsandPublications/FormsforIndividuals/PIT/Documents/2022/2022_pa-40in.pdf
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-25.htm
        
        Indiana:
        -         
        - https://www.in.gov/dor/tax-forms/2022-individual-income-tax-forms/
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-32.htm
          https://www.in.gov/dor/files/ib28.pdf
          https://www.in.gov/dor/tax-forms/2022-individual-income-tax-forms/
          https://www.in.gov/dor/files/2022-county-tax-rates-and-codes.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        county_occupation = self.county_occupation if hasattr(self, 'county_occupation') else county_occupation
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Pennsylvania', 'state_occupation': 'Indiana', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "county_occupation": county_occupation, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def pennsylvania_and_iowa(self, county_residence='Other', state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', county_occupation='Other', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Pennsylvania:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        county_residence : str, optional
            County of residence. Default is Other.
        county_occupation : str, optional
            County of occupation. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        
        Iowa:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Pennsylvania:
        - https://www.revenue.pa.gov/Tax%20Rates/Pages/default.aspx
          https://www.revenue.pa.gov/FormsandPublications/FormsforIndividuals/PIT/Documents/2022/2022_pa-40in.pdf
          https://www.revenue.pa.gov/FormsandPublications/FormsforBusinesses/EmployerWithholding/Documents/rev-419.pdf
          https://www.revenue.pa.gov/FormsandPublications/FormsforIndividuals/PIT/Documents/2022/2022_pa-40in.pdf
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-25.htm
        
        Iowa:
        -         
        - https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
          https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
          https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
          https://tax.iowa.gov/iowa-illinois-reciprocal-agreement
          https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        county_occupation = self.county_occupation if hasattr(self, 'county_occupation') else county_occupation
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Pennsylvania', 'state_occupation': 'Iowa', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "county_occupation": county_occupation, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def pennsylvania_and_kansas(self, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', county_occupation='Other', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Pennsylvania:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        county_residence : str, optional
            County of residence. Default is Other.
        county_occupation : str, optional
            County of occupation. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        
        Kansas:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Pennsylvania:
        - https://www.revenue.pa.gov/Tax%20Rates/Pages/default.aspx
          https://www.revenue.pa.gov/FormsandPublications/FormsforIndividuals/PIT/Documents/2022/2022_pa-40in.pdf
          https://www.revenue.pa.gov/FormsandPublications/FormsforBusinesses/EmployerWithholding/Documents/rev-419.pdf
          https://www.revenue.pa.gov/FormsandPublications/FormsforIndividuals/PIT/Documents/2022/2022_pa-40in.pdf
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-25.htm
        
        Kansas:
        -         
        - https://www.ksrevenue.gov/taxrates.html
          https://www.ksrevenue.gov/incomebook22.html#0
          https://www.ksrevenue.gov/incomebook22.html#0
          https://www.ksrevenue.gov/incomebook22.html#0
          https://www.ksrevenue.gov/incomebook22.html
          https://www.ksrevenue.gov/incomebook22.html
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        county_occupation = self.county_occupation if hasattr(self, 'county_occupation') else county_occupation
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Pennsylvania', 'state_occupation': 'Kansas', "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "county_occupation": county_occupation, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def pennsylvania_and_kentucky(self, county_residence='Other', state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', county_occupation='Other', sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Pennsylvania:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        county_residence : str, optional
            County of residence. Default is Other.
        county_occupation : str, optional
            County of occupation. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        
        Kentucky:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Pennsylvania:
        - https://www.revenue.pa.gov/Tax%20Rates/Pages/default.aspx
          https://www.revenue.pa.gov/FormsandPublications/FormsforIndividuals/PIT/Documents/2022/2022_pa-40in.pdf
          https://www.revenue.pa.gov/FormsandPublications/FormsforBusinesses/EmployerWithholding/Documents/rev-419.pdf
          https://www.revenue.pa.gov/FormsandPublications/FormsforIndividuals/PIT/Documents/2022/2022_pa-40in.pdf
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-25.htm
        
        Kentucky:
        -         
        - https://revenue.ky.gov/News/Pages/Dor-Announces-Updates-to-Individual-Income-Tax-for-2023-Tax-Year.aspx
          https://revenue.ky.gov/Individual/Individual-Income-Tax/Pages/default.aspx
          https://revenue.ky.gov/News/Pages/Dor-Announces-Updates-to-Individual-Income-Tax-for-2023-Tax-Year.aspx
          https://revenue.ky.gov/Individual/Individual-Income-Tax/Pages/default.aspx
          https://revenue.ky.gov/Dor%20Training%20Materials/103%20KAR%2017.140.%20Individual%20income%20tax%20-%20reciprocity%20-%20nonresidents.pdf
          https://revenue.ky.gov/Individual/Individual-Income-Tax/Pages/default.aspx
          https://nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2021/Taxes-21-32.htm?taxmap=true
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        county_occupation = self.county_occupation if hasattr(self, 'county_occupation') else county_occupation
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Pennsylvania', 'state_occupation': 'Kentucky', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "county_occupation": county_occupation, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def pennsylvania_and_louisiana(self, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', county_occupation='Other', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Pennsylvania:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        county_residence : str, optional
            County of residence. Default is Other.
        county_occupation : str, optional
            County of occupation. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        
        Louisiana:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Pennsylvania:
        - https://www.revenue.pa.gov/Tax%20Rates/Pages/default.aspx
          https://www.revenue.pa.gov/FormsandPublications/FormsforIndividuals/PIT/Documents/2022/2022_pa-40in.pdf
          https://www.revenue.pa.gov/FormsandPublications/FormsforBusinesses/EmployerWithholding/Documents/rev-419.pdf
          https://www.revenue.pa.gov/FormsandPublications/FormsforIndividuals/PIT/Documents/2022/2022_pa-40in.pdf
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-25.htm
        
        Louisiana:
        -         
        - https://revenue.louisiana.gov/individualincometax
          https://revenue.louisiana.gov/individualincometax
          https://www.revenue.louisiana.gov/taxforms/6935(11_02)F.pdf
          https://revenue.louisiana.gov/individualincometax
          https://revenue.louisiana.gov/individualincometax
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        county_occupation = self.county_occupation if hasattr(self, 'county_occupation') else county_occupation
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Pennsylvania', 'state_occupation': 'Louisiana', "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "county_occupation": county_occupation, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def pennsylvania_and_maine(self, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', county_occupation='Other', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Pennsylvania:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        county_residence : str, optional
            County of residence. Default is Other.
        county_occupation : str, optional
            County of occupation. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        
        Maine:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Pennsylvania:
        - https://www.revenue.pa.gov/Tax%20Rates/Pages/default.aspx
          https://www.revenue.pa.gov/FormsandPublications/FormsforIndividuals/PIT/Documents/2022/2022_pa-40in.pdf
          https://www.revenue.pa.gov/FormsandPublications/FormsforBusinesses/EmployerWithholding/Documents/rev-419.pdf
          https://www.revenue.pa.gov/FormsandPublications/FormsforIndividuals/PIT/Documents/2022/2022_pa-40in.pdf
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-25.htm
        
        Maine:
        -         
        - https://www.maine.gov/revenue/tax-return-forms/individual-income-tax-2022
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/ind_tax_rate_sched_2022.pdf
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/22_1040me_book_gen_instruc_revisedFeb23.pdf
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/22_1040me_book_gen_instruc_revisedFeb23.pdf
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/22_1040me_book_gen_instruc_revisedFeb23.pdf
          https://www.maine.gov/revenue/tax-return-forms/individual-income-tax-2022
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        county_occupation = self.county_occupation if hasattr(self, 'county_occupation') else county_occupation
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Pennsylvania', 'state_occupation': 'Maine', "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "county_occupation": county_occupation, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def pennsylvania_and_maryland(self, county_residence='Other', state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, county_occupation='Other', sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Pennsylvania:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        county_residence : str, optional
            County of residence. Default is Other.
        county_occupation : str, optional
            County of occupation. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        
        Maryland:
        


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Pennsylvania:
        - https://www.revenue.pa.gov/Tax%20Rates/Pages/default.aspx
          https://www.revenue.pa.gov/FormsandPublications/FormsforIndividuals/PIT/Documents/2022/2022_pa-40in.pdf
          https://www.revenue.pa.gov/FormsandPublications/FormsforBusinesses/EmployerWithholding/Documents/rev-419.pdf
          https://www.revenue.pa.gov/FormsandPublications/FormsforIndividuals/PIT/Documents/2022/2022_pa-40in.pdf
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-25.htm
        
        Maryland:
        -         
        - https://www.marylandtaxes.gov/individual/income/tax-info/tax-rates.php
          https://www.marylandtaxes.gov/forms/22_forms/Resident_Booklet.pdf
          https://www.marylandtaxes.gov/forms/Tax_Publications/Administrative_Releases/Income_and_Estate_Tax_Releases/ar_it37.pdf
          https://www.marylandtaxes.gov/individual/index.php
          https://www.marylandtaxes.gov/individual/income/tax-info/tax-rates.php
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        county_occupation = self.county_occupation if hasattr(self, 'county_occupation') else county_occupation
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Pennsylvania', 'state_occupation': 'Maryland', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "county_occupation": county_occupation, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def pennsylvania_and_massachusetts(self, county_residence='Other', state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', county_occupation='Other', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Pennsylvania:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        county_residence : str, optional
            County of residence. Default is Other.
        county_occupation : str, optional
            County of occupation. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        
        Massachusetts:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Pennsylvania:
        - https://www.revenue.pa.gov/Tax%20Rates/Pages/default.aspx
          https://www.revenue.pa.gov/FormsandPublications/FormsforIndividuals/PIT/Documents/2022/2022_pa-40in.pdf
          https://www.revenue.pa.gov/FormsandPublications/FormsforBusinesses/EmployerWithholding/Documents/rev-419.pdf
          https://www.revenue.pa.gov/FormsandPublications/FormsforIndividuals/PIT/Documents/2022/2022_pa-40in.pdf
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-25.htm
        
        Massachusetts:
        -         
        - https://www.mass.gov/service-details/massachusetts-tax-rates
          https://www.mass.gov/service-details/view-massachusetts-personal-income-tax-exemptions
          https://www.mass.gov/service-details/view-massachusetts-personal-income-tax-exemptions
          https://www.mass.gov/service-details/view-child-and-dependent-related-credits
          https://www.mass.gov/service-details/massachusetts-tax-rates
          https://www.mass.gov/personal-income-tax
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        county_occupation = self.county_occupation if hasattr(self, 'county_occupation') else county_occupation
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Pennsylvania', 'state_occupation': 'Massachusetts', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "county_occupation": county_occupation, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def pennsylvania_and_michigan(self, county_residence='Other', state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', county_occupation='Other', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Pennsylvania:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        county_residence : str, optional
            County of residence. Default is Other.
        county_occupation : str, optional
            County of occupation. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        
        Michigan:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        county_occupation : str, optional
            County of occupation. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Pennsylvania:
        - https://www.revenue.pa.gov/Tax%20Rates/Pages/default.aspx
          https://www.revenue.pa.gov/FormsandPublications/FormsforIndividuals/PIT/Documents/2022/2022_pa-40in.pdf
          https://www.revenue.pa.gov/FormsandPublications/FormsforBusinesses/EmployerWithholding/Documents/rev-419.pdf
          https://www.revenue.pa.gov/FormsandPublications/FormsforIndividuals/PIT/Documents/2022/2022_pa-40in.pdf
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-25.htm
        
        Michigan:
        -         
        - https://www.michigan.gov/taxes/iit/new-developments-for-tax-year-2022
          "https://www.michigan.gov/taxes/-/media/Project/Websites/taxes/Forms/2023/2023-SUW/446_Withholding-Guide_2023.pdf?rev=177ca3d8cb034392bf84e317ed78c320&hash=19C5F42F43A89A91863D229349F48883#:~:text=The%20exemption%20amount%20is%20%245%2C400
         employees%20Michigan%20income%20tax%20return"
          "https://www.michigan.gov/taxes/-/media/Project/Websites/taxes/Forms/2023/2023-SUW/446_Withholding-Guide_2023.pdf?rev=177ca3d8cb034392bf84e317ed78c320&hash=19C5F42F43A89A91863D229349F48883#:~:text=The%20exemption%20amount%20is%20%245%2C400
         employees%20Michigan%20income%20tax%20return"
          https://www.michigan.gov/taxes/questions/iit/accordion/residency/are-my-wages-earned-in-another-state-taxable-in-michigan-if-i-am-a-michigan-resident-1
          https://www.michigan.gov/taxes/-/media/Project/Websites/taxes/Forms/2022/2022-IIT-Forms/BOOK_MI-1040.pdf?rev=4524f1cbbc0243b194ab05a6680be75b&hash=40DFE71E651F7C1D65C31F6CD03D944F
          https://www.michigan.gov/taxes/questions/iit/accordion/general/what-cities-impose-an-income-tax
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        county_occupation = self.county_occupation if hasattr(self, 'county_occupation') else county_occupation
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Pennsylvania', 'state_occupation': 'Michigan', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "county_occupation": county_occupation, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def pennsylvania_and_minnesota(self, county_residence='Other', state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', county_occupation='Other', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Pennsylvania:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        county_residence : str, optional
            County of residence. Default is Other.
        county_occupation : str, optional
            County of occupation. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        
        Minnesota:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Pennsylvania:
        - https://www.revenue.pa.gov/Tax%20Rates/Pages/default.aspx
          https://www.revenue.pa.gov/FormsandPublications/FormsforIndividuals/PIT/Documents/2022/2022_pa-40in.pdf
          https://www.revenue.pa.gov/FormsandPublications/FormsforBusinesses/EmployerWithholding/Documents/rev-419.pdf
          https://www.revenue.pa.gov/FormsandPublications/FormsforIndividuals/PIT/Documents/2022/2022_pa-40in.pdf
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-25.htm
        
        Minnesota:
        -         
        - https://www.revenue.state.mn.us/minnesota-income-tax-rates-and-brackets
          https://www.revenue.state.mn.us/child-and-dependent-care-credit
          https://www.revenue.state.mn.us/dependent-exemptions
          https://www.revenue.state.mn.us/minnesota-standard-deduction
          https://mn.gov/mmb/assets/mwr_form_tcm1059-128581.pdf
          https://www.revenue.state.mn.us/individual-income-tax
          https://www.revenue.state.mn.us/individual-income-tax
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        county_occupation = self.county_occupation if hasattr(self, 'county_occupation') else county_occupation
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Pennsylvania', 'state_occupation': 'Minnesota', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "county_occupation": county_occupation, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def pennsylvania_and_mississippi(self, county_residence='Other', state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', county_occupation='Other', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Pennsylvania:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        county_residence : str, optional
            County of residence. Default is Other.
        county_occupation : str, optional
            County of occupation. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        
        Mississippi:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Pennsylvania:
        - https://www.revenue.pa.gov/Tax%20Rates/Pages/default.aspx
          https://www.revenue.pa.gov/FormsandPublications/FormsforIndividuals/PIT/Documents/2022/2022_pa-40in.pdf
          https://www.revenue.pa.gov/FormsandPublications/FormsforBusinesses/EmployerWithholding/Documents/rev-419.pdf
          https://www.revenue.pa.gov/FormsandPublications/FormsforIndividuals/PIT/Documents/2022/2022_pa-40in.pdf
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-25.htm
        
        Mississippi:
        -         
        - https://www.dor.ms.gov/individual/tax-rates
          https://www.dor.ms.gov/individual/tax-rates
          https://www.dor.ms.gov/individual/tax-rates
          https://www.dor.ms.gov/individual
          https://www.dor.ms.gov/individual
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        county_occupation = self.county_occupation if hasattr(self, 'county_occupation') else county_occupation
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Pennsylvania', 'state_occupation': 'Mississippi', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "county_occupation": county_occupation, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def pennsylvania_and_missouri(self, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', county_occupation='Other', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Pennsylvania:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        county_residence : str, optional
            County of residence. Default is Other.
        county_occupation : str, optional
            County of occupation. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        
        Missouri:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Pennsylvania:
        - https://www.revenue.pa.gov/Tax%20Rates/Pages/default.aspx
          https://www.revenue.pa.gov/FormsandPublications/FormsforIndividuals/PIT/Documents/2022/2022_pa-40in.pdf
          https://www.revenue.pa.gov/FormsandPublications/FormsforBusinesses/EmployerWithholding/Documents/rev-419.pdf
          https://www.revenue.pa.gov/FormsandPublications/FormsforIndividuals/PIT/Documents/2022/2022_pa-40in.pdf
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-25.htm
        
        Missouri:
        -         
        - https://dor.mo.gov/taxation/individual/tax-types/income/year-changes/
          https://dor.mo.gov/taxation/individual/tax-types/income/year-changes/
          https://dor.mo.gov/taxation/individual/tax-types/income/year-changes/
          https://www.kcmo.gov/city-hall/departments/finance/earnings-tax
          https://www.stlouis-mo.gov/government/departments/collector/earnings-tax/individual-earnings-tax-info.cfm
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        county_occupation = self.county_occupation if hasattr(self, 'county_occupation') else county_occupation
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Pennsylvania', 'state_occupation': 'Missouri', "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "county_occupation": county_occupation, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def pennsylvania_and_montana(self, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', county_occupation='Other', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Pennsylvania:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        county_residence : str, optional
            County of residence. Default is Other.
        county_occupation : str, optional
            County of occupation. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        
        Montana:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Pennsylvania:
        - https://www.revenue.pa.gov/Tax%20Rates/Pages/default.aspx
          https://www.revenue.pa.gov/FormsandPublications/FormsforIndividuals/PIT/Documents/2022/2022_pa-40in.pdf
          https://www.revenue.pa.gov/FormsandPublications/FormsforBusinesses/EmployerWithholding/Documents/rev-419.pdf
          https://www.revenue.pa.gov/FormsandPublications/FormsforIndividuals/PIT/Documents/2022/2022_pa-40in.pdf
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-25.htm
        
        Montana:
        -         
        - https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/2022/12/montana-employees-withholding-allowance-and-exemption-certificate-form-mw-4.pdf
          https://mtrevenue.gov/2021/11/29/simplification-of-montana-income-taxation/?utm_source=rss&utm_medium=rss&utm_campaign=simplification-of-montana-income-taxation
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        county_occupation = self.county_occupation if hasattr(self, 'county_occupation') else county_occupation
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Pennsylvania', 'state_occupation': 'Montana', "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "county_occupation": county_occupation, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def pennsylvania_and_nebraska(self, county_residence='Other', state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', county_occupation='Other', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Pennsylvania:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        county_residence : str, optional
            County of residence. Default is Other.
        county_occupation : str, optional
            County of occupation. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        
        Nebraska:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Pennsylvania:
        - https://www.revenue.pa.gov/Tax%20Rates/Pages/default.aspx
          https://www.revenue.pa.gov/FormsandPublications/FormsforIndividuals/PIT/Documents/2022/2022_pa-40in.pdf
          https://www.revenue.pa.gov/FormsandPublications/FormsforBusinesses/EmployerWithholding/Documents/rev-419.pdf
          https://www.revenue.pa.gov/FormsandPublications/FormsforIndividuals/PIT/Documents/2022/2022_pa-40in.pdf
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-25.htm
        
        Nebraska:
        -         
        - https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2023_Circular_EN_8-429r2023_Whole_Book_11.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        county_occupation = self.county_occupation if hasattr(self, 'county_occupation') else county_occupation
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Pennsylvania', 'state_occupation': 'Nebraska', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "county_occupation": county_occupation, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def pennsylvania_and_nevada(self, county_residence='Other', state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, county_occupation='Other', sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Pennsylvania:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        county_residence : str, optional
            County of residence. Default is Other.
        county_occupation : str, optional
            County of occupation. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        
        Nevada:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Pennsylvania:
        - https://www.revenue.pa.gov/Tax%20Rates/Pages/default.aspx
          https://www.revenue.pa.gov/FormsandPublications/FormsforIndividuals/PIT/Documents/2022/2022_pa-40in.pdf
          https://www.revenue.pa.gov/FormsandPublications/FormsforBusinesses/EmployerWithholding/Documents/rev-419.pdf
          https://www.revenue.pa.gov/FormsandPublications/FormsforIndividuals/PIT/Documents/2022/2022_pa-40in.pdf
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-25.htm
        
        Nevada:
        -         
        - https://www.nvsos.gov/sos/businesses/the-nevada-advantage
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        county_occupation = self.county_occupation if hasattr(self, 'county_occupation') else county_occupation
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Pennsylvania', 'state_occupation': 'Nevada', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "county_occupation": county_occupation, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def pennsylvania_and_new_hampshire(self, county_residence='Other', state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, county_occupation='Other', sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Pennsylvania:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        county_residence : str, optional
            County of residence. Default is Other.
        county_occupation : str, optional
            County of occupation. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        
        New Hampshire:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Pennsylvania:
        - https://www.revenue.pa.gov/Tax%20Rates/Pages/default.aspx
          https://www.revenue.pa.gov/FormsandPublications/FormsforIndividuals/PIT/Documents/2022/2022_pa-40in.pdf
          https://www.revenue.pa.gov/FormsandPublications/FormsforBusinesses/EmployerWithholding/Documents/rev-419.pdf
          https://www.revenue.pa.gov/FormsandPublications/FormsforIndividuals/PIT/Documents/2022/2022_pa-40in.pdf
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-25.htm
        
        New Hampshire:
        -         
        - https://www.revenue.nh.gov/assistance/tax-overview.htm
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        county_occupation = self.county_occupation if hasattr(self, 'county_occupation') else county_occupation
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Pennsylvania', 'state_occupation': 'NewHampshire', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "county_occupation": county_occupation, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def pennsylvania_and_new_jersey(self, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', county_occupation='Other', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Pennsylvania:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        county_residence : str, optional
            County of residence. Default is Other.
        county_occupation : str, optional
            County of occupation. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        
        New Jersey:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Pennsylvania:
        - https://www.revenue.pa.gov/Tax%20Rates/Pages/default.aspx
          https://www.revenue.pa.gov/FormsandPublications/FormsforIndividuals/PIT/Documents/2022/2022_pa-40in.pdf
          https://www.revenue.pa.gov/FormsandPublications/FormsforBusinesses/EmployerWithholding/Documents/rev-419.pdf
          https://www.revenue.pa.gov/FormsandPublications/FormsforIndividuals/PIT/Documents/2022/2022_pa-40in.pdf
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-25.htm
        
        New Jersey:
        -         
        - https://www.state.nj.us/treasury/taxation/taxtables.shtml
          https://www.state.nj.us/treasury/taxation/njit13.shtml
          https://www.nj.gov/treasury/taxation/njit25.shtml
          https://www.state.nj.us/treasury/taxation/individuals/freqqiti.shtml
          https://www.state.nj.us/treasury/taxation/localtax.shtml
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        county_occupation = self.county_occupation if hasattr(self, 'county_occupation') else county_occupation
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Pennsylvania', 'state_occupation': 'NewJersey', "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "county_occupation": county_occupation, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def pennsylvania_and_new_mexico(self, county_residence='Other', state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', county_occupation='Other', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Pennsylvania:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        county_residence : str, optional
            County of residence. Default is Other.
        county_occupation : str, optional
            County of occupation. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        
        New Mexico:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Pennsylvania:
        - https://www.revenue.pa.gov/Tax%20Rates/Pages/default.aspx
          https://www.revenue.pa.gov/FormsandPublications/FormsforIndividuals/PIT/Documents/2022/2022_pa-40in.pdf
          https://www.revenue.pa.gov/FormsandPublications/FormsforBusinesses/EmployerWithholding/Documents/rev-419.pdf
          https://www.revenue.pa.gov/FormsandPublications/FormsforIndividuals/PIT/Documents/2022/2022_pa-40in.pdf
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-25.htm
        
        New Mexico:
        -         
        - https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/fdf3c548-8aba-4b9c-9eb4-bb564c716015/FYI-104.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/0bf943fd-652e-4400-bb1b-cd8397eb5a95/2022pit-adj.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/2f1a6781-9534-4436-b427-1557f9592099/2022pit-adj-ins.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/0bf943fd-652e-4400-bb1b-cd8397eb5a95/2022pit-adj.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/1afc56af-ea90-4d48-82e5-1f9aeb43255a/PITbook2022.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/fdf3c548-8aba-4b9c-9eb4-bb564c716015/FYI-104.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        county_occupation = self.county_occupation if hasattr(self, 'county_occupation') else county_occupation
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Pennsylvania', 'state_occupation': 'NewMexico', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "county_occupation": county_occupation, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def pennsylvania_and_new_york(self, county_residence='Other', state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', county_occupation='Other', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Pennsylvania:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        county_residence : str, optional
            County of residence. Default is Other.
        county_occupation : str, optional
            County of occupation. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        
        New York:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Pennsylvania:
        - https://www.revenue.pa.gov/Tax%20Rates/Pages/default.aspx
          https://www.revenue.pa.gov/FormsandPublications/FormsforIndividuals/PIT/Documents/2022/2022_pa-40in.pdf
          https://www.revenue.pa.gov/FormsandPublications/FormsforBusinesses/EmployerWithholding/Documents/rev-419.pdf
          https://www.revenue.pa.gov/FormsandPublications/FormsforIndividuals/PIT/Documents/2022/2022_pa-40in.pdf
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-25.htm
        
        New York:
        -         
        - https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nys-tax-rate-schedule
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nys-tax-rate-schedule
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nys-tax-rate-schedule
          https://www.tax.ny.gov/pit/
          https://www.tax.ny.gov/pit/file/nonresident-faqs.htm
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#yonkers-tax-rate-schedule
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nyc-tax-rate-schedule
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        county_occupation = self.county_occupation if hasattr(self, 'county_occupation') else county_occupation
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Pennsylvania', 'state_occupation': 'NewYork', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "county_occupation": county_occupation, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def pennsylvania_and_north_carolina(self, county_residence='Other', state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', county_occupation='Other', sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Pennsylvania:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        county_residence : str, optional
            County of residence. Default is Other.
        county_occupation : str, optional
            County of occupation. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        
        North Carolina:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Pennsylvania:
        - https://www.revenue.pa.gov/Tax%20Rates/Pages/default.aspx
          https://www.revenue.pa.gov/FormsandPublications/FormsforIndividuals/PIT/Documents/2022/2022_pa-40in.pdf
          https://www.revenue.pa.gov/FormsandPublications/FormsforBusinesses/EmployerWithholding/Documents/rev-419.pdf
          https://www.revenue.pa.gov/FormsandPublications/FormsforIndividuals/PIT/Documents/2022/2022_pa-40in.pdf
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-25.htm
        
        North Carolina:
        -         
        - https://www.ncdor.gov/taxes-forms/tax-rate-schedules
          https://www.ncdor.gov/taxes-forms/individual-income-tax/north-carolina-standard-deduction-or-north-carolina-itemized-deductions
          https://www.ncdor.gov/taxes-forms/individual-income-tax
          https://www.ncdor.gov/taxes-forms/individual-income-tax
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        county_occupation = self.county_occupation if hasattr(self, 'county_occupation') else county_occupation
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Pennsylvania', 'state_occupation': 'NorthCarolina', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "county_occupation": county_occupation, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def pennsylvania_and_north_dakota(self, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', county_occupation='Other', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Pennsylvania:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        county_residence : str, optional
            County of residence. Default is Other.
        county_occupation : str, optional
            County of occupation. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        
        North Dakota:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Pennsylvania:
        - https://www.revenue.pa.gov/Tax%20Rates/Pages/default.aspx
          https://www.revenue.pa.gov/FormsandPublications/FormsforIndividuals/PIT/Documents/2022/2022_pa-40in.pdf
          https://www.revenue.pa.gov/FormsandPublications/FormsforBusinesses/EmployerWithholding/Documents/rev-419.pdf
          https://www.revenue.pa.gov/FormsandPublications/FormsforIndividuals/PIT/Documents/2022/2022_pa-40in.pdf
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-25.htm
        
        North Dakota:
        -         
        - https://www.tax.nd.gov/forms
          https://www.tax.nd.gov/sites/www/files/documents/forms/individual/2022-iit/2022-individual-income-tax-booklet.pdf
          https://www.tax.nd.gov/sites/www/files/documents/forms/2022-individual-income-tax-booklet.pdf
          https://www.tax.nd.gov/tax-exemptions-credits/income-tax-exemptions-credits
          https://www.tax.nd.gov/sites/www/files/documents/forms/business/ndwrfillable.pdf
          https://www.tax.nd.gov/news/tax-legislative-changes/significant-changes-law/individual-income-tax-history
          https://www.tax.nd.gov/forms
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        county_occupation = self.county_occupation if hasattr(self, 'county_occupation') else county_occupation
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Pennsylvania', 'state_occupation': 'NorthDakota', "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "county_occupation": county_occupation, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def pennsylvania_and_ohio(self, county_residence='Other', state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, county_occupation='Other', sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Pennsylvania:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        county_residence : str, optional
            County of residence. Default is Other.
        county_occupation : str, optional
            County of occupation. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        
        Ohio:
        


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Pennsylvania:
        - https://www.revenue.pa.gov/Tax%20Rates/Pages/default.aspx
          https://www.revenue.pa.gov/FormsandPublications/FormsforIndividuals/PIT/Documents/2022/2022_pa-40in.pdf
          https://www.revenue.pa.gov/FormsandPublications/FormsforBusinesses/EmployerWithholding/Documents/rev-419.pdf
          https://www.revenue.pa.gov/FormsandPublications/FormsforIndividuals/PIT/Documents/2022/2022_pa-40in.pdf
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-25.htm
        
        Ohio:
        -         
        - https://tax.ohio.gov/individual/resources/annual-tax-rates
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://tax.ohio.gov/static/forms/employer_withholding/generic/wth_it4nr.pdf
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://www.ritaohio.com/TaxRatesTable
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        county_occupation = self.county_occupation if hasattr(self, 'county_occupation') else county_occupation
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Pennsylvania', 'state_occupation': 'Ohio', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "county_occupation": county_occupation, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def pennsylvania_and_oklahoma(self, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', county_occupation='Other', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Pennsylvania:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        county_residence : str, optional
            County of residence. Default is Other.
        county_occupation : str, optional
            County of occupation. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        
        Oklahoma:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Pennsylvania:
        - https://www.revenue.pa.gov/Tax%20Rates/Pages/default.aspx
          https://www.revenue.pa.gov/FormsandPublications/FormsforIndividuals/PIT/Documents/2022/2022_pa-40in.pdf
          https://www.revenue.pa.gov/FormsandPublications/FormsforBusinesses/EmployerWithholding/Documents/rev-419.pdf
          https://www.revenue.pa.gov/FormsandPublications/FormsforIndividuals/PIT/Documents/2022/2022_pa-40in.pdf
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-25.htm
        
        Oklahoma:
        -         
        - https://oklahoma.gov/content/dam/ok/en/tax/documents/resources/publications/businesses/withholding-tables/WHTables-2022.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        county_occupation = self.county_occupation if hasattr(self, 'county_occupation') else county_occupation
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Pennsylvania', 'state_occupation': 'Oklahoma', "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "county_occupation": county_occupation, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def pennsylvania_and_oregon(self, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', county_occupation='Other', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Pennsylvania:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        county_residence : str, optional
            County of residence. Default is Other.
        county_occupation : str, optional
            County of occupation. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        
        Oregon:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Pennsylvania:
        - https://www.revenue.pa.gov/Tax%20Rates/Pages/default.aspx
          https://www.revenue.pa.gov/FormsandPublications/FormsforIndividuals/PIT/Documents/2022/2022_pa-40in.pdf
          https://www.revenue.pa.gov/FormsandPublications/FormsforBusinesses/EmployerWithholding/Documents/rev-419.pdf
          https://www.revenue.pa.gov/FormsandPublications/FormsforIndividuals/PIT/Documents/2022/2022_pa-40in.pdf
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-25.htm
        
        Oregon:
        -         
        - https://www.oregon.gov/dor/programs/individuals/Pages/PIT.aspx
          https://www.oregon.gov/dor/programs/individuals/Documents/Full%20year%20resident
         %20Form%20or-40%20filers.pdf
          https://www.oregon.gov/dor/programs/individuals/Documents/Part-year%20and%20nonresident
         %20Form%20or-40-P%20and%20or-40-N%20filers.pdf.pdf
          https://www.oregon.gov/dor/forms/FormsPubs/form-or-40_101-040_2022.pdf
          https://www.oregon.gov/dor/forms/FormsPubs/form-or-40_101-040_2022.pdf
          https://www.oregon.gov/dor/forms/FormsPubs/form-or-40_101-040_2022.pdf
          https://www.portland.gov/revenue/personal-tax
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        county_occupation = self.county_occupation if hasattr(self, 'county_occupation') else county_occupation
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Pennsylvania', 'state_occupation': 'Oregon', "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "county_occupation": county_occupation, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def pennsylvania_and_rhode_island(self, county_residence='Other', state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', county_occupation='Other', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Pennsylvania:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        county_residence : str, optional
            County of residence. Default is Other.
        county_occupation : str, optional
            County of occupation. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        
        Rhode Island:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Pennsylvania:
        - https://www.revenue.pa.gov/Tax%20Rates/Pages/default.aspx
          https://www.revenue.pa.gov/FormsandPublications/FormsforIndividuals/PIT/Documents/2022/2022_pa-40in.pdf
          https://www.revenue.pa.gov/FormsandPublications/FormsforBusinesses/EmployerWithholding/Documents/rev-419.pdf
          https://www.revenue.pa.gov/FormsandPublications/FormsforIndividuals/PIT/Documents/2022/2022_pa-40in.pdf
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-25.htm
        
        Rhode Island:
        -         
        - https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2022_1040WE_w_0.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        county_occupation = self.county_occupation if hasattr(self, 'county_occupation') else county_occupation
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Pennsylvania', 'state_occupation': 'RhodeIsland', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "county_occupation": county_occupation, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def pennsylvania_and_south_carolina(self, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', county_occupation='Other', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Pennsylvania:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        county_residence : str, optional
            County of residence. Default is Other.
        county_occupation : str, optional
            County of occupation. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        
        South Carolina:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Pennsylvania:
        - https://www.revenue.pa.gov/Tax%20Rates/Pages/default.aspx
          https://www.revenue.pa.gov/FormsandPublications/FormsforIndividuals/PIT/Documents/2022/2022_pa-40in.pdf
          https://www.revenue.pa.gov/FormsandPublications/FormsforBusinesses/EmployerWithholding/Documents/rev-419.pdf
          https://www.revenue.pa.gov/FormsandPublications/FormsforIndividuals/PIT/Documents/2022/2022_pa-40in.pdf
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-25.htm
        
        South Carolina:
        -         
        - https://dor.sc.gov/resources-site/lawandpolicy/Advisory%20Opinions/IL22-15.pdf
          https://dor.sc.gov/forms-site/Forms/IITPacket_2022.pdf
          https://www.scstatehouse.gov/code/t12c006.php
          https://dor.sc.gov/resources-site/lawandpolicy/Advisory%20Opinions/IL22-15.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        county_occupation = self.county_occupation if hasattr(self, 'county_occupation') else county_occupation
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Pennsylvania', 'state_occupation': 'SouthCarolina', "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "county_occupation": county_occupation, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def pennsylvania_and_south_dakota(self, county_residence='Other', state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, county_occupation='Other', sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Pennsylvania:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        county_residence : str, optional
            County of residence. Default is Other.
        county_occupation : str, optional
            County of occupation. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        
        South Dakota:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Pennsylvania:
        - https://www.revenue.pa.gov/Tax%20Rates/Pages/default.aspx
          https://www.revenue.pa.gov/FormsandPublications/FormsforIndividuals/PIT/Documents/2022/2022_pa-40in.pdf
          https://www.revenue.pa.gov/FormsandPublications/FormsforBusinesses/EmployerWithholding/Documents/rev-419.pdf
          https://www.revenue.pa.gov/FormsandPublications/FormsforIndividuals/PIT/Documents/2022/2022_pa-40in.pdf
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-25.htm
        
        South Dakota:
        -         
        - https://dor.sd.gov/individuals/taxes/
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        county_occupation = self.county_occupation if hasattr(self, 'county_occupation') else county_occupation
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Pennsylvania', 'state_occupation': 'SouthDakota', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "county_occupation": county_occupation, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def pennsylvania_and_tennessee(self, county_residence='Other', state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, county_occupation='Other', sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Pennsylvania:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        county_residence : str, optional
            County of residence. Default is Other.
        county_occupation : str, optional
            County of occupation. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        
        Tennessee:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Pennsylvania:
        - https://www.revenue.pa.gov/Tax%20Rates/Pages/default.aspx
          https://www.revenue.pa.gov/FormsandPublications/FormsforIndividuals/PIT/Documents/2022/2022_pa-40in.pdf
          https://www.revenue.pa.gov/FormsandPublications/FormsforBusinesses/EmployerWithholding/Documents/rev-419.pdf
          https://www.revenue.pa.gov/FormsandPublications/FormsforIndividuals/PIT/Documents/2022/2022_pa-40in.pdf
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-25.htm
        
        Tennessee:
        -         
        - https://www.tn.gov/revenue/taxes.html
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        county_occupation = self.county_occupation if hasattr(self, 'county_occupation') else county_occupation
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Pennsylvania', 'state_occupation': 'Tennessee', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "county_occupation": county_occupation, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def pennsylvania_and_texas(self, county_residence='Other', state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, county_occupation='Other', sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Pennsylvania:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        county_residence : str, optional
            County of residence. Default is Other.
        county_occupation : str, optional
            County of occupation. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        
        Texas:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Pennsylvania:
        - https://www.revenue.pa.gov/Tax%20Rates/Pages/default.aspx
          https://www.revenue.pa.gov/FormsandPublications/FormsforIndividuals/PIT/Documents/2022/2022_pa-40in.pdf
          https://www.revenue.pa.gov/FormsandPublications/FormsforBusinesses/EmployerWithholding/Documents/rev-419.pdf
          https://www.revenue.pa.gov/FormsandPublications/FormsforIndividuals/PIT/Documents/2022/2022_pa-40in.pdf
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-25.htm
        
        Texas:
        -         
        - https://comptroller.texas.gov/taxes/
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        county_occupation = self.county_occupation if hasattr(self, 'county_occupation') else county_occupation
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Pennsylvania', 'state_occupation': 'Texas', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "county_occupation": county_occupation, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def pennsylvania_and_utah(self, county_residence='Other', state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', county_occupation='Other', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Pennsylvania:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        county_residence : str, optional
            County of residence. Default is Other.
        county_occupation : str, optional
            County of occupation. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        
        Utah:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Pennsylvania:
        - https://www.revenue.pa.gov/Tax%20Rates/Pages/default.aspx
          https://www.revenue.pa.gov/FormsandPublications/FormsforIndividuals/PIT/Documents/2022/2022_pa-40in.pdf
          https://www.revenue.pa.gov/FormsandPublications/FormsforBusinesses/EmployerWithholding/Documents/rev-419.pdf
          https://www.revenue.pa.gov/FormsandPublications/FormsforIndividuals/PIT/Documents/2022/2022_pa-40in.pdf
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-25.htm
        
        Utah:
        -         
        - https://incometax.utah.gov/paying/tax-rates
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        county_occupation = self.county_occupation if hasattr(self, 'county_occupation') else county_occupation
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Pennsylvania', 'state_occupation': 'Utah', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "county_occupation": county_occupation, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def pennsylvania_and_vermont(self, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', county_occupation='Other', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Pennsylvania:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        county_residence : str, optional
            County of residence. Default is Other.
        county_occupation : str, optional
            County of occupation. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        
        Vermont:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Pennsylvania:
        - https://www.revenue.pa.gov/Tax%20Rates/Pages/default.aspx
          https://www.revenue.pa.gov/FormsandPublications/FormsforIndividuals/PIT/Documents/2022/2022_pa-40in.pdf
          https://www.revenue.pa.gov/FormsandPublications/FormsforBusinesses/EmployerWithholding/Documents/rev-419.pdf
          https://www.revenue.pa.gov/FormsandPublications/FormsforIndividuals/PIT/Documents/2022/2022_pa-40in.pdf
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-25.htm
        
        Vermont:
        -         
        - https://tax.vermont.gov/sites/tax/files/documents/GB-1210-2023.pdf
          https://tax.vermont.gov/sites/tax/files/documents/Income%20Booklet-2022.pdf
          https://tax.vermont.gov/sites/tax/files/documents/Income%20Booklet-2022.pdf
          https://tax.vermont.gov/sites/tax/files/documents/Income%20Booklet-2022.pdf
          https://tax.vermont.gov/individuals/personal-income-tax/tax-credits
          https://tax.vermont.gov/individuals/personal-income-tax/taxable-income
          https://tax.vermont.gov/individuals/personal-income-tax/local-option-tax
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        county_occupation = self.county_occupation if hasattr(self, 'county_occupation') else county_occupation
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Pennsylvania', 'state_occupation': 'Vermont', "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "county_occupation": county_occupation, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def pennsylvania_and_virginia(self, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', county_occupation='Other', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Pennsylvania:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        county_residence : str, optional
            County of residence. Default is Other.
        county_occupation : str, optional
            County of occupation. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        
        Virginia:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Pennsylvania:
        - https://www.revenue.pa.gov/Tax%20Rates/Pages/default.aspx
          https://www.revenue.pa.gov/FormsandPublications/FormsforIndividuals/PIT/Documents/2022/2022_pa-40in.pdf
          https://www.revenue.pa.gov/FormsandPublications/FormsforBusinesses/EmployerWithholding/Documents/rev-419.pdf
          https://www.revenue.pa.gov/FormsandPublications/FormsforIndividuals/PIT/Documents/2022/2022_pa-40in.pdf
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-25.htm
        
        Virginia:
        -         
        - https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/reciprocity
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        county_occupation = self.county_occupation if hasattr(self, 'county_occupation') else county_occupation
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Pennsylvania', 'state_occupation': 'Virginia', "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "county_occupation": county_occupation, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def pennsylvania_and_washington(self, county_residence='Other', state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, county_occupation='Other', sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Pennsylvania:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        county_residence : str, optional
            County of residence. Default is Other.
        county_occupation : str, optional
            County of occupation. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        
        Washington:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Pennsylvania:
        - https://www.revenue.pa.gov/Tax%20Rates/Pages/default.aspx
          https://www.revenue.pa.gov/FormsandPublications/FormsforIndividuals/PIT/Documents/2022/2022_pa-40in.pdf
          https://www.revenue.pa.gov/FormsandPublications/FormsforBusinesses/EmployerWithholding/Documents/rev-419.pdf
          https://www.revenue.pa.gov/FormsandPublications/FormsforIndividuals/PIT/Documents/2022/2022_pa-40in.pdf
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-25.htm
        
        Washington:
        -         
        - https://dor.wa.gov/taxes-rates/income-tax
          https://dor.wa.gov/taxes-rates/other-taxes/capital-gains-tax
          https://dor.wa.gov/taxes-rates/income-tax
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        county_occupation = self.county_occupation if hasattr(self, 'county_occupation') else county_occupation
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Pennsylvania', 'state_occupation': 'Washington', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "county_occupation": county_occupation, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def pennsylvania_and_west_virginia(self, county_residence='Other', state_agi=75000, pay_periods=24, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', county_occupation='Other', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Pennsylvania:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        county_residence : str, optional
            County of residence. Default is Other.
        county_occupation : str, optional
            County of occupation. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        
        West Virginia:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_occupation : str, optional
            County of occupation. Default is Other.
        pay_periods : int, optional
            Number of pay periods. Options are 13, 24, 26, 52. Default is 24.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Pennsylvania:
        - https://www.revenue.pa.gov/Tax%20Rates/Pages/default.aspx
          https://www.revenue.pa.gov/FormsandPublications/FormsforIndividuals/PIT/Documents/2022/2022_pa-40in.pdf
          https://www.revenue.pa.gov/FormsandPublications/FormsforBusinesses/EmployerWithholding/Documents/rev-419.pdf
          https://www.revenue.pa.gov/FormsandPublications/FormsforIndividuals/PIT/Documents/2022/2022_pa-40in.pdf
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-25.htm
        
        West Virginia:
        -         
        - https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://tax.wv.gov/Documents/TSD/tsd438.pdf
          https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2011/Taxes-11-35att.html
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        pay_periods = self.pay_periods if hasattr(self, 'pay_periods') else pay_periods
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        county_occupation = self.county_occupation if hasattr(self, 'county_occupation') else county_occupation
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Pennsylvania', 'state_occupation': 'WestVirginia', "county_residence": county_residence, "state_agi": state_agi, "pay_periods": pay_periods, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "county_occupation": county_occupation, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def pennsylvania_and_wisconsin(self, county_residence='Other', state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', county_occupation='Other', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Pennsylvania:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        county_residence : str, optional
            County of residence. Default is Other.
        county_occupation : str, optional
            County of occupation. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        
        Wisconsin:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Pennsylvania:
        - https://www.revenue.pa.gov/Tax%20Rates/Pages/default.aspx
          https://www.revenue.pa.gov/FormsandPublications/FormsforIndividuals/PIT/Documents/2022/2022_pa-40in.pdf
          https://www.revenue.pa.gov/FormsandPublications/FormsforBusinesses/EmployerWithholding/Documents/rev-419.pdf
          https://www.revenue.pa.gov/FormsandPublications/FormsforIndividuals/PIT/Documents/2022/2022_pa-40in.pdf
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-25.htm
        
        Wisconsin:
        -         
        - https://www.revenue.wi.gov/TaxForms2023/2023-Form1-ES-Inst.pdf
          https://www.revenue.wi.gov/TaxForms2023/2023-Form1-ES-Inst.pdf
          https://www.revenue.wi.gov/TaxForms2023/2023-Form1-ES-Inst.pdf
          https://www.revenue.wi.gov/Pages/FAQS/pcs-work.aspx
          https://www.revenue.wi.gov/Dor%20Publications/pb103.pdf
          https://www.revenue.wi.gov/Pages/FAQS/pcs-work.aspx
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        county_occupation = self.county_occupation if hasattr(self, 'county_occupation') else county_occupation
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Pennsylvania', 'state_occupation': 'Wisconsin', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "county_occupation": county_occupation, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def pennsylvania_and_wyoming(self, county_residence='Other', state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, county_occupation='Other', sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Pennsylvania:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        county_residence : str, optional
            County of residence. Default is Other.
        county_occupation : str, optional
            County of occupation. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        
        Wyoming:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Pennsylvania:
        - https://www.revenue.pa.gov/Tax%20Rates/Pages/default.aspx
          https://www.revenue.pa.gov/FormsandPublications/FormsforIndividuals/PIT/Documents/2022/2022_pa-40in.pdf
          https://www.revenue.pa.gov/FormsandPublications/FormsforBusinesses/EmployerWithholding/Documents/rev-419.pdf
          https://www.revenue.pa.gov/FormsandPublications/FormsforIndividuals/PIT/Documents/2022/2022_pa-40in.pdf
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-25.htm
        
        Wyoming:
        -         
        - https://revenue.wyo.gov/
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        county_occupation = self.county_occupation if hasattr(self, 'county_occupation') else county_occupation
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Pennsylvania', 'state_occupation': 'Wyoming', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "county_occupation": county_occupation, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def rhode_island_and_alabama(self, county_residence='Other', state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Rhode Island:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Alabama:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Rhode Island:
        - https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2022_1040WE_w_0.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
        
        Alabama:
        -         
        - https://www.revenue.alabama.gov/ultraviewer/viewer/basic_viewer/index.html?form=2023/02/22f40.pdf
          https://www.revenue.alabama.gov/faqs/what-is-alabamas-individual-income-tax-rate/
          https://www.revenue.alabama.gov/ultraviewer/viewer/basic_viewer/index.html?form=2023/02/22f40.pdf
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-24.htm?taxmap=true
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-24.htm?taxmap=true
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-24.htm?taxmap=true
          https://www.revenue.alabama.gov/individual-corporate/income-to-be-reported-on-the-alabama-income-tax-return/
          https://www.almonline.org/TaxRates.aspx#OccupationalTax
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'RhodeIsland', 'state_occupation': 'Alabama', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def rhode_island_and_alaska(self, state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Rhode Island:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Alaska:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Rhode Island:
        - https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2022_1040WE_w_0.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
        
        Alaska:
        -         
        - https://www.commerce.alaska.gov/web/dcra/officeofthestateassessor/alaskataxfacts.aspx
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'RhodeIsland', 'state_occupation': 'Alaska', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def rhode_island_and_arizona(self, state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Rhode Island:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Arizona:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        credits_state : int, optional
            Credits for state taxes. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Rhode Island:
        - https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2022_1040WE_w_0.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
        
        Arizona:
        -         
        - https://azdor.gov/forms/individual/form-140-x-y-tables
          https://azdor.gov/forms/individual/form-140-x-y-tables
          https://azdor.gov/forms/individual/form-140-x-y-tables
          https://azdor.gov/businesses-arizona/withholding-tax/withholding-exceptions
          "https://azdor.gov/individuals/income-tax-filing-assistance/identifying-other-taxable-income#:~:text=In%20regards%20to%20capital%20gains
         the%20individuals%20regular%20tax%20rate.
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'RhodeIsland', 'state_occupation': 'Arizona', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def rhode_island_and_arkansas(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Rhode Island:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Arkansas:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Rhode Island:
        - https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2022_1040WE_w_0.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
        
        Arkansas:
        -         
        - https://www.arkansasedc.com/why-arkansas/business-climate/tax-structure/personal-income-tax
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_and_AR1000NR_Instructions.pdf
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_FullYearResidentIndividualIncomeTaxReturn.pdf
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_and_AR1000NR_Instructions.pdf
          https://www.arkansasedc.com/why-arkansas/business-climate/tax-structure/capital-gains-tax-reduction
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_and_AR1000NR_Instructions.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'RhodeIsland', 'state_occupation': 'Arkansas', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def rhode_island_and_california(self, state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Rhode Island:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        California:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Rhode Island:
        - https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2022_1040WE_w_0.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
        
        California:
        -         
        - https://www.ftb.ca.gov/forms/2022/2022-540-tax-rate-schedules.pdf
          https://www.ftb.ca.gov/file/personal/deductions/index.html
          https://www.ftb.ca.gov/about-ftb/newsroom/caleitc/eligibility-and-credit-information.html
          https://www.ftb.ca.gov/file/personal/income-types/capital-gains-and-losses.html#:~:text=California%20does%20not%20have%20a
         are%20taxed%20as%20ordinary%20income
          https://www.ftb.ca.gov/file/personal/
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'RhodeIsland', 'state_occupation': 'California', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def rhode_island_and_colorado(self, county_residence='Other', state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Rhode Island:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Colorado:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Rhode Island:
        - https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2022_1040WE_w_0.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
        
        Colorado:
        -         
        - https://leg.colorado.gov/agencies/legislative-council-staff/individual-income-tax%C2%A0
          https://leg.colorado.gov/agencies/legislative-council-staff/individual-income-tax%C2%A0
          https://tax.colorado.gov/sites/tax/files/documents/ITT_Capital_Gain_Subtraction_Feb_2022.pdf
          https://www.auroragov.org/business_services/taxes/occupational_privilege_tax
          https://www.denvergov.org/Government/Agencies-Departments-Offices/Agencies-Departments-Offices-Directory/Department-of-Finance/Our-Divisions/Treasury/Business-Tax-Information#:~:text=Occupational%20privilege%20tax
         -%28show%20below%29&text=Each%20taxable%20employee%20is%20liable
         month%20for%20each%20taxable%20employee
          https://www.glendale.co.us/355/Occupational-Privilege-Tax
          https://www.greenwoodvillage.com/1220/Occupational-Privilege-Tax-OPT
          https://ci.sheridan.co.us/288/Occupational-Privilege-Tax
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'RhodeIsland', 'state_occupation': 'Colorado', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def rhode_island_and_connecticut(self, state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Rhode Island:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Connecticut:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Rhode Island:
        - https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2022_1040WE_w_0.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
        
        Connecticut:
        -         
        - https://portal.ct.gov/-/media/DRS/Forms/2022/Income/2022-CT-1040-Instructions_1222.pdf
          https://www.cga.ct.gov/2022/rpt/pdf/2022-R-0108.pdf
          https://www.cga.ct.gov/current/pub/chap_229.htm#sec_12-703
          https://portal.ct.gov/-/media/DRS/Forms/2022/Income/2022-CT-1040-Instructions_1222.pdf
          "https://portal.ct.gov/DRS/Publications/TSSNs/TSSN-29#:~:text=An%20individuals%20net%20capital%20gains
         portion%20of%20Social%20Security%20benefits"
          "https://portal.ct.gov/DRS/Publications/TSSNs/TSSN-29#:~:text=An%20individuals%20net%20capital%20gains
         portion%20of%20Social%20Security%20benefits"
          https://portal.ct.gov/DRS/Individuals/Resident-Income-Tax/Tax-Information
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'RhodeIsland', 'state_occupation': 'Connecticut', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def rhode_island_and_delaware(self, state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, federal_agi=75000, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Rhode Island:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Delaware:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        federal_agi : int, optional
            Adjusted Gross Income (AGI) for federal taxes. Default is 0. Sources - https://www.irs.gov/e-file-providers/definition-of-adjusted-gross-income, https://www.irs.gov/newsroom/taxpayers-should-know-the-difference-between-standard-and-itemized-deductions
        income : int, optional
            Gross income. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Rhode Island:
        - https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2022_1040WE_w_0.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
        
        Delaware:
        -         
        - https://revenue.delaware.gov/software-developer/tax-rate-changes/
          https://revenuefiles.delaware.gov/2022/PIT-RES_TY22_2022-01_PaperInteractive.pdf
          https://revenuefiles.delaware.gov/2022/PIT-RES_TY22_2022-02_Instructions.pdf
          https://revenue.delaware.gov/frequently-asked-questions/personal-income-tax-faqs/
          https://www.newcastlede.gov/DocumentCenter/View/20147/City-of-Wilmington-Wage-Tax-Information-PDF
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        federal_agi = self.federal_agi if hasattr(self, 'federal_agi') else federal_agi
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'RhodeIsland', 'state_occupation': 'Delaware', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "federal_agi": federal_agi, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def rhode_island_and_florida(self, state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Rhode Island:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Florida:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Rhode Island:
        - https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2022_1040WE_w_0.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
        
        Florida:
        -         
        - https://floridarevenue.com/taxes/taxesfees/Pages/tax_interest_rates.aspx
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'RhodeIsland', 'state_occupation': 'Florida', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def rhode_island_and_georgia(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Rhode Island:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Georgia:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Rhode Island:
        - https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2022_1040WE_w_0.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
        
        Georgia:
        -         
        - https://dor.georgia.gov/tax-tables-georgia-tax-rate-schedule
          https://apps.dor.ga.gov/FillableForms/PDFViewer/Index?form=2022GA500
          https://apps.dor.ga.gov/FillableForms/PDFViewer/Index?form=2022GA500
          https://apps.dor.ga.gov/FillableForms/PDFViewer/Index?form=2022GA500
          https://dor.georgia.gov/years-individual-income-tax-forms
          https://dor.georgia.gov/years-individual-income-tax-forms
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'RhodeIsland', 'state_occupation': 'Georgia', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def rhode_island_and_hawaii(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Rhode Island:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Hawaii:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Rhode Island:
        - https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2022_1040WE_w_0.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
        
        Hawaii:
        -         
        - https://tax.hawaii.gov/forms/d_18table-on/d_18table-on_p13/
          https://files.hawaii.gov/tax/forms/2018/18table-on.pdf
          https://files.hawaii.gov/tax/news/pubs/22outline.pdf
          https://files.hawaii.gov/tax/news/pubs/FastTaxRefGuide_2017.pdf
          https://www.tfhawaii.org/wordpress/state-tax-resources/mini-tax-guides/individual-income-tax-chapter-235/#:~:text=Long%20term%20capital%20gains%20are%20taxed%20at%20a%20maximum%20of%207.25%25
          https://www.grassrootinstitute.org/2023/02/hb337-increasing-tax-on-capital-gains-a-bad-risk/
          https://files.hawaii.gov/tax/news/pubs/22outline.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'RhodeIsland', 'state_occupation': 'Hawaii', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def rhode_island_and_idaho(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Rhode Island:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Idaho:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Rhode Island:
        - https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2022_1040WE_w_0.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
        
        Idaho:
        -         
        - https://tax.idaho.gov/taxes/income-tax/individual-income/individual-income-tax-rate-schedule/
          https://tax.idaho.gov/wp-content/uploads/forms/EIN00046/EIN00046_03-01-2023.pdf
          https://tax.idaho.gov/taxes/income-tax/individual-income/idaho-source-income/
          https://tax.idaho.gov/taxes/income-tax/individual-income/idaho-source-income/
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'RhodeIsland', 'state_occupation': 'Idaho', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def rhode_island_and_illinois(self, state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Rhode Island:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Illinois:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Rhode Island:
        - https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2022_1040WE_w_0.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
        
        Illinois:
        -         
        - https://tax.illinois.gov/research/taxrates/income.html#IndividualIncome
          https://tax.illinois.gov/content/dam/soi/en/web/tax/forms/incometax/documents/currentyear/individual/il-1040-instr.pdf
          https://tax.illinois.gov/questionsandanswers/12.html
          https://tax.illinois.gov/content/dam/soi/en/web/tax/forms/incometax/documents/currentyear/individual/il-1040-instr.pdf
          https://tax.illinois.gov/research/taxrates/personalproperty.html
          https://tax.illinois.gov/localgovernments/income.html
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'RhodeIsland', 'state_occupation': 'Illinois', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def rhode_island_and_indiana(self, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Rhode Island:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Indiana:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Rhode Island:
        - https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2022_1040WE_w_0.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
        
        Indiana:
        -         
        - https://www.in.gov/dor/tax-forms/2022-individual-income-tax-forms/
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-32.htm
          https://www.in.gov/dor/files/ib28.pdf
          https://www.in.gov/dor/tax-forms/2022-individual-income-tax-forms/
          https://www.in.gov/dor/files/2022-county-tax-rates-and-codes.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'RhodeIsland', 'state_occupation': 'Indiana', "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def rhode_island_and_iowa(self, county_residence='Other', state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Rhode Island:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Iowa:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Rhode Island:
        - https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2022_1040WE_w_0.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
        
        Iowa:
        -         
        - https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
          https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
          https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
          https://tax.iowa.gov/iowa-illinois-reciprocal-agreement
          https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'RhodeIsland', 'state_occupation': 'Iowa', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def rhode_island_and_kansas(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Rhode Island:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Kansas:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Rhode Island:
        - https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2022_1040WE_w_0.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
        
        Kansas:
        -         
        - https://www.ksrevenue.gov/taxrates.html
          https://www.ksrevenue.gov/incomebook22.html#0
          https://www.ksrevenue.gov/incomebook22.html#0
          https://www.ksrevenue.gov/incomebook22.html#0
          https://www.ksrevenue.gov/incomebook22.html
          https://www.ksrevenue.gov/incomebook22.html
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'RhodeIsland', 'state_occupation': 'Kansas', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def rhode_island_and_kentucky(self, county_residence='Other', state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Rhode Island:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Kentucky:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Rhode Island:
        - https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2022_1040WE_w_0.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
        
        Kentucky:
        -         
        - https://revenue.ky.gov/News/Pages/Dor-Announces-Updates-to-Individual-Income-Tax-for-2023-Tax-Year.aspx
          https://revenue.ky.gov/Individual/Individual-Income-Tax/Pages/default.aspx
          https://revenue.ky.gov/News/Pages/Dor-Announces-Updates-to-Individual-Income-Tax-for-2023-Tax-Year.aspx
          https://revenue.ky.gov/Individual/Individual-Income-Tax/Pages/default.aspx
          https://revenue.ky.gov/Dor%20Training%20Materials/103%20KAR%2017.140.%20Individual%20income%20tax%20-%20reciprocity%20-%20nonresidents.pdf
          https://revenue.ky.gov/Individual/Individual-Income-Tax/Pages/default.aspx
          https://nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2021/Taxes-21-32.htm?taxmap=true
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'RhodeIsland', 'state_occupation': 'Kentucky', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def rhode_island_and_louisiana(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Rhode Island:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Louisiana:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Rhode Island:
        - https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2022_1040WE_w_0.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
        
        Louisiana:
        -         
        - https://revenue.louisiana.gov/individualincometax
          https://revenue.louisiana.gov/individualincometax
          https://www.revenue.louisiana.gov/taxforms/6935(11_02)F.pdf
          https://revenue.louisiana.gov/individualincometax
          https://revenue.louisiana.gov/individualincometax
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'RhodeIsland', 'state_occupation': 'Louisiana', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def rhode_island_and_maine(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Rhode Island:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Maine:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Rhode Island:
        - https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2022_1040WE_w_0.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
        
        Maine:
        -         
        - https://www.maine.gov/revenue/tax-return-forms/individual-income-tax-2022
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/ind_tax_rate_sched_2022.pdf
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/22_1040me_book_gen_instruc_revisedFeb23.pdf
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/22_1040me_book_gen_instruc_revisedFeb23.pdf
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/22_1040me_book_gen_instruc_revisedFeb23.pdf
          https://www.maine.gov/revenue/tax-return-forms/individual-income-tax-2022
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'RhodeIsland', 'state_occupation': 'Maine', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def rhode_island_and_maryland(self, county_residence='Other', state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Rhode Island:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Maryland:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Rhode Island:
        - https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2022_1040WE_w_0.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
        
        Maryland:
        -         
        - https://www.marylandtaxes.gov/individual/income/tax-info/tax-rates.php
          https://www.marylandtaxes.gov/forms/22_forms/Resident_Booklet.pdf
          https://www.marylandtaxes.gov/forms/Tax_Publications/Administrative_Releases/Income_and_Estate_Tax_Releases/ar_it37.pdf
          https://www.marylandtaxes.gov/individual/index.php
          https://www.marylandtaxes.gov/individual/income/tax-info/tax-rates.php
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'RhodeIsland', 'state_occupation': 'Maryland', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def rhode_island_and_massachusetts(self, state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Rhode Island:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Massachusetts:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Rhode Island:
        - https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2022_1040WE_w_0.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
        
        Massachusetts:
        -         
        - https://www.mass.gov/service-details/massachusetts-tax-rates
          https://www.mass.gov/service-details/view-massachusetts-personal-income-tax-exemptions
          https://www.mass.gov/service-details/view-massachusetts-personal-income-tax-exemptions
          https://www.mass.gov/service-details/view-child-and-dependent-related-credits
          https://www.mass.gov/service-details/massachusetts-tax-rates
          https://www.mass.gov/personal-income-tax
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'RhodeIsland', 'state_occupation': 'Massachusetts', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def rhode_island_and_michigan(self, county_occupation='Other', state_agi=75000, county_residence='Other', capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Rhode Island:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Michigan:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        county_occupation : str, optional
            County of occupation. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Rhode Island:
        - https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2022_1040WE_w_0.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
        
        Michigan:
        -         
        - https://www.michigan.gov/taxes/iit/new-developments-for-tax-year-2022
          "https://www.michigan.gov/taxes/-/media/Project/Websites/taxes/Forms/2023/2023-SUW/446_Withholding-Guide_2023.pdf?rev=177ca3d8cb034392bf84e317ed78c320&hash=19C5F42F43A89A91863D229349F48883#:~:text=The%20exemption%20amount%20is%20%245%2C400
         employees%20Michigan%20income%20tax%20return"
          "https://www.michigan.gov/taxes/-/media/Project/Websites/taxes/Forms/2023/2023-SUW/446_Withholding-Guide_2023.pdf?rev=177ca3d8cb034392bf84e317ed78c320&hash=19C5F42F43A89A91863D229349F48883#:~:text=The%20exemption%20amount%20is%20%245%2C400
         employees%20Michigan%20income%20tax%20return"
          https://www.michigan.gov/taxes/questions/iit/accordion/residency/are-my-wages-earned-in-another-state-taxable-in-michigan-if-i-am-a-michigan-resident-1
          https://www.michigan.gov/taxes/-/media/Project/Websites/taxes/Forms/2022/2022-IIT-Forms/BOOK_MI-1040.pdf?rev=4524f1cbbc0243b194ab05a6680be75b&hash=40DFE71E651F7C1D65C31F6CD03D944F
          https://www.michigan.gov/taxes/questions/iit/accordion/general/what-cities-impose-an-income-tax
        """
        
        # Check if the parameters are set upstream in the instance
        county_occupation = self.county_occupation if hasattr(self, 'county_occupation') else county_occupation
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'RhodeIsland', 'state_occupation': 'Michigan', "county_occupation": county_occupation, "state_agi": state_agi, "county_residence": county_residence, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def rhode_island_and_minnesota(self, state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Rhode Island:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Minnesota:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Rhode Island:
        - https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2022_1040WE_w_0.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
        
        Minnesota:
        -         
        - https://www.revenue.state.mn.us/minnesota-income-tax-rates-and-brackets
          https://www.revenue.state.mn.us/child-and-dependent-care-credit
          https://www.revenue.state.mn.us/dependent-exemptions
          https://www.revenue.state.mn.us/minnesota-standard-deduction
          https://mn.gov/mmb/assets/mwr_form_tcm1059-128581.pdf
          https://www.revenue.state.mn.us/individual-income-tax
          https://www.revenue.state.mn.us/individual-income-tax
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'RhodeIsland', 'state_occupation': 'Minnesota', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def rhode_island_and_mississippi(self, state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Rhode Island:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Mississippi:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Rhode Island:
        - https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2022_1040WE_w_0.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
        
        Mississippi:
        -         
        - https://www.dor.ms.gov/individual/tax-rates
          https://www.dor.ms.gov/individual/tax-rates
          https://www.dor.ms.gov/individual/tax-rates
          https://www.dor.ms.gov/individual
          https://www.dor.ms.gov/individual
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'RhodeIsland', 'state_occupation': 'Mississippi', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def rhode_island_and_missouri(self, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Rhode Island:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Missouri:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Rhode Island:
        - https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2022_1040WE_w_0.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
        
        Missouri:
        -         
        - https://dor.mo.gov/taxation/individual/tax-types/income/year-changes/
          https://dor.mo.gov/taxation/individual/tax-types/income/year-changes/
          https://dor.mo.gov/taxation/individual/tax-types/income/year-changes/
          https://www.kcmo.gov/city-hall/departments/finance/earnings-tax
          https://www.stlouis-mo.gov/government/departments/collector/earnings-tax/individual-earnings-tax-info.cfm
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'RhodeIsland', 'state_occupation': 'Missouri', "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def rhode_island_and_montana(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Rhode Island:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Montana:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Rhode Island:
        - https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2022_1040WE_w_0.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
        
        Montana:
        -         
        - https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/2022/12/montana-employees-withholding-allowance-and-exemption-certificate-form-mw-4.pdf
          https://mtrevenue.gov/2021/11/29/simplification-of-montana-income-taxation/?utm_source=rss&utm_medium=rss&utm_campaign=simplification-of-montana-income-taxation
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'RhodeIsland', 'state_occupation': 'Montana', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def rhode_island_and_nebraska(self, state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Rhode Island:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Nebraska:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Rhode Island:
        - https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2022_1040WE_w_0.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
        
        Nebraska:
        -         
        - https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2023_Circular_EN_8-429r2023_Whole_Book_11.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'RhodeIsland', 'state_occupation': 'Nebraska', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def rhode_island_and_nevada(self, state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Rhode Island:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Nevada:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Rhode Island:
        - https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2022_1040WE_w_0.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
        
        Nevada:
        -         
        - https://www.nvsos.gov/sos/businesses/the-nevada-advantage
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'RhodeIsland', 'state_occupation': 'Nevada', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def rhode_island_and_new_hampshire(self, state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Rhode Island:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        New Hampshire:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Rhode Island:
        - https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2022_1040WE_w_0.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
        
        New Hampshire:
        -         
        - https://www.revenue.nh.gov/assistance/tax-overview.htm
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'RhodeIsland', 'state_occupation': 'NewHampshire', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def rhode_island_and_new_jersey(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Rhode Island:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        New Jersey:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Rhode Island:
        - https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2022_1040WE_w_0.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
        
        New Jersey:
        -         
        - https://www.state.nj.us/treasury/taxation/taxtables.shtml
          https://www.state.nj.us/treasury/taxation/njit13.shtml
          https://www.nj.gov/treasury/taxation/njit25.shtml
          https://www.state.nj.us/treasury/taxation/individuals/freqqiti.shtml
          https://www.state.nj.us/treasury/taxation/localtax.shtml
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'RhodeIsland', 'state_occupation': 'NewJersey', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def rhode_island_and_new_mexico(self, state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Rhode Island:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        New Mexico:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Rhode Island:
        - https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2022_1040WE_w_0.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
        
        New Mexico:
        -         
        - https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/fdf3c548-8aba-4b9c-9eb4-bb564c716015/FYI-104.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/0bf943fd-652e-4400-bb1b-cd8397eb5a95/2022pit-adj.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/2f1a6781-9534-4436-b427-1557f9592099/2022pit-adj-ins.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/0bf943fd-652e-4400-bb1b-cd8397eb5a95/2022pit-adj.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/1afc56af-ea90-4d48-82e5-1f9aeb43255a/PITbook2022.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/fdf3c548-8aba-4b9c-9eb4-bb564c716015/FYI-104.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'RhodeIsland', 'state_occupation': 'NewMexico', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def rhode_island_and_new_york(self, county_residence='Other', state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Rhode Island:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        New York:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Rhode Island:
        - https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2022_1040WE_w_0.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
        
        New York:
        -         
        - https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nys-tax-rate-schedule
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nys-tax-rate-schedule
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nys-tax-rate-schedule
          https://www.tax.ny.gov/pit/
          https://www.tax.ny.gov/pit/file/nonresident-faqs.htm
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#yonkers-tax-rate-schedule
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nyc-tax-rate-schedule
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'RhodeIsland', 'state_occupation': 'NewYork', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def rhode_island_and_north_carolina(self, state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Rhode Island:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        North Carolina:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Rhode Island:
        - https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2022_1040WE_w_0.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
        
        North Carolina:
        -         
        - https://www.ncdor.gov/taxes-forms/tax-rate-schedules
          https://www.ncdor.gov/taxes-forms/individual-income-tax/north-carolina-standard-deduction-or-north-carolina-itemized-deductions
          https://www.ncdor.gov/taxes-forms/individual-income-tax
          https://www.ncdor.gov/taxes-forms/individual-income-tax
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'RhodeIsland', 'state_occupation': 'NorthCarolina', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def rhode_island_and_north_dakota(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Rhode Island:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        North Dakota:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Rhode Island:
        - https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2022_1040WE_w_0.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
        
        North Dakota:
        -         
        - https://www.tax.nd.gov/forms
          https://www.tax.nd.gov/sites/www/files/documents/forms/individual/2022-iit/2022-individual-income-tax-booklet.pdf
          https://www.tax.nd.gov/sites/www/files/documents/forms/2022-individual-income-tax-booklet.pdf
          https://www.tax.nd.gov/tax-exemptions-credits/income-tax-exemptions-credits
          https://www.tax.nd.gov/sites/www/files/documents/forms/business/ndwrfillable.pdf
          https://www.tax.nd.gov/news/tax-legislative-changes/significant-changes-law/individual-income-tax-history
          https://www.tax.nd.gov/forms
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'RhodeIsland', 'state_occupation': 'NorthDakota', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def rhode_island_and_ohio(self, county_residence='Other', state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Rhode Island:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Ohio:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Rhode Island:
        - https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2022_1040WE_w_0.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
        
        Ohio:
        -         
        - https://tax.ohio.gov/individual/resources/annual-tax-rates
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://tax.ohio.gov/static/forms/employer_withholding/generic/wth_it4nr.pdf
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://www.ritaohio.com/TaxRatesTable
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'RhodeIsland', 'state_occupation': 'Ohio', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def rhode_island_and_oklahoma(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Rhode Island:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Oklahoma:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Rhode Island:
        - https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2022_1040WE_w_0.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
        
        Oklahoma:
        -         
        - https://oklahoma.gov/content/dam/ok/en/tax/documents/resources/publications/businesses/withholding-tables/WHTables-2022.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'RhodeIsland', 'state_occupation': 'Oklahoma', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def rhode_island_and_oregon(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Rhode Island:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Oregon:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Rhode Island:
        - https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2022_1040WE_w_0.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
        
        Oregon:
        -         
        - https://www.oregon.gov/dor/programs/individuals/Pages/PIT.aspx
          https://www.oregon.gov/dor/programs/individuals/Documents/Full%20year%20resident
         %20Form%20or-40%20filers.pdf
          https://www.oregon.gov/dor/programs/individuals/Documents/Part-year%20and%20nonresident
         %20Form%20or-40-P%20and%20or-40-N%20filers.pdf.pdf
          https://www.oregon.gov/dor/forms/FormsPubs/form-or-40_101-040_2022.pdf
          https://www.oregon.gov/dor/forms/FormsPubs/form-or-40_101-040_2022.pdf
          https://www.oregon.gov/dor/forms/FormsPubs/form-or-40_101-040_2022.pdf
          https://www.portland.gov/revenue/personal-tax
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'RhodeIsland', 'state_occupation': 'Oregon', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def rhode_island_and_pennsylvania(self, county_residence='Other', state_agi=75000, county_occupation='Other', capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Rhode Island:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Pennsylvania:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        county_residence : str, optional
            County of residence. Default is Other.
        county_occupation : str, optional
            County of occupation. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Rhode Island:
        - https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2022_1040WE_w_0.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
        
        Pennsylvania:
        -         
        - https://www.revenue.pa.gov/Tax%20Rates/Pages/default.aspx
          https://www.revenue.pa.gov/FormsandPublications/FormsforIndividuals/PIT/Documents/2022/2022_pa-40in.pdf
          https://www.revenue.pa.gov/FormsandPublications/FormsforBusinesses/EmployerWithholding/Documents/rev-419.pdf
          https://www.revenue.pa.gov/FormsandPublications/FormsforIndividuals/PIT/Documents/2022/2022_pa-40in.pdf
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-25.htm
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        county_occupation = self.county_occupation if hasattr(self, 'county_occupation') else county_occupation
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'RhodeIsland', 'state_occupation': 'Pennsylvania', "county_residence": county_residence, "state_agi": state_agi, "county_occupation": county_occupation, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def rhode_island_and_south_carolina(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Rhode Island:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        South Carolina:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Rhode Island:
        - https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2022_1040WE_w_0.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
        
        South Carolina:
        -         
        - https://dor.sc.gov/resources-site/lawandpolicy/Advisory%20Opinions/IL22-15.pdf
          https://dor.sc.gov/forms-site/Forms/IITPacket_2022.pdf
          https://www.scstatehouse.gov/code/t12c006.php
          https://dor.sc.gov/resources-site/lawandpolicy/Advisory%20Opinions/IL22-15.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'RhodeIsland', 'state_occupation': 'SouthCarolina', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def rhode_island_and_south_dakota(self, state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Rhode Island:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        South Dakota:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Rhode Island:
        - https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2022_1040WE_w_0.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
        
        South Dakota:
        -         
        - https://dor.sd.gov/individuals/taxes/
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'RhodeIsland', 'state_occupation': 'SouthDakota', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def rhode_island_and_tennessee(self, state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Rhode Island:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Tennessee:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Rhode Island:
        - https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2022_1040WE_w_0.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
        
        Tennessee:
        -         
        - https://www.tn.gov/revenue/taxes.html
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'RhodeIsland', 'state_occupation': 'Tennessee', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def rhode_island_and_texas(self, state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Rhode Island:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Texas:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Rhode Island:
        - https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2022_1040WE_w_0.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
        
        Texas:
        -         
        - https://comptroller.texas.gov/taxes/
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'RhodeIsland', 'state_occupation': 'Texas', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def rhode_island_and_utah(self, state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Rhode Island:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Utah:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Rhode Island:
        - https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2022_1040WE_w_0.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
        
        Utah:
        -         
        - https://incometax.utah.gov/paying/tax-rates
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'RhodeIsland', 'state_occupation': 'Utah', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def rhode_island_and_vermont(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Rhode Island:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Vermont:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Rhode Island:
        - https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2022_1040WE_w_0.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
        
        Vermont:
        -         
        - https://tax.vermont.gov/sites/tax/files/documents/GB-1210-2023.pdf
          https://tax.vermont.gov/sites/tax/files/documents/Income%20Booklet-2022.pdf
          https://tax.vermont.gov/sites/tax/files/documents/Income%20Booklet-2022.pdf
          https://tax.vermont.gov/sites/tax/files/documents/Income%20Booklet-2022.pdf
          https://tax.vermont.gov/individuals/personal-income-tax/tax-credits
          https://tax.vermont.gov/individuals/personal-income-tax/taxable-income
          https://tax.vermont.gov/individuals/personal-income-tax/local-option-tax
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'RhodeIsland', 'state_occupation': 'Vermont', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def rhode_island_and_virginia(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Rhode Island:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Virginia:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Rhode Island:
        - https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2022_1040WE_w_0.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
        
        Virginia:
        -         
        - https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/reciprocity
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'RhodeIsland', 'state_occupation': 'Virginia', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def rhode_island_and_washington(self, state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Rhode Island:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Washington:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Rhode Island:
        - https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2022_1040WE_w_0.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
        
        Washington:
        -         
        - https://dor.wa.gov/taxes-rates/income-tax
          https://dor.wa.gov/taxes-rates/other-taxes/capital-gains-tax
          https://dor.wa.gov/taxes-rates/income-tax
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'RhodeIsland', 'state_occupation': 'Washington', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def rhode_island_and_west_virginia(self, county_occupation='Other', state_agi=75000, pay_periods=24, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Rhode Island:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        West Virginia:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_occupation : str, optional
            County of occupation. Default is Other.
        pay_periods : int, optional
            Number of pay periods. Options are 13, 24, 26, 52. Default is 24.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Rhode Island:
        - https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2022_1040WE_w_0.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
        
        West Virginia:
        -         
        - https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://tax.wv.gov/Documents/TSD/tsd438.pdf
          https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2011/Taxes-11-35att.html
        """
        
        # Check if the parameters are set upstream in the instance
        county_occupation = self.county_occupation if hasattr(self, 'county_occupation') else county_occupation
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        pay_periods = self.pay_periods if hasattr(self, 'pay_periods') else pay_periods
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'RhodeIsland', 'state_occupation': 'WestVirginia', "county_occupation": county_occupation, "state_agi": state_agi, "pay_periods": pay_periods, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def rhode_island_and_wisconsin(self, state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Rhode Island:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Wisconsin:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Rhode Island:
        - https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2022_1040WE_w_0.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
        
        Wisconsin:
        -         
        - https://www.revenue.wi.gov/TaxForms2023/2023-Form1-ES-Inst.pdf
          https://www.revenue.wi.gov/TaxForms2023/2023-Form1-ES-Inst.pdf
          https://www.revenue.wi.gov/TaxForms2023/2023-Form1-ES-Inst.pdf
          https://www.revenue.wi.gov/Pages/FAQS/pcs-work.aspx
          https://www.revenue.wi.gov/Dor%20Publications/pb103.pdf
          https://www.revenue.wi.gov/Pages/FAQS/pcs-work.aspx
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'RhodeIsland', 'state_occupation': 'Wisconsin', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def rhode_island_and_wyoming(self, state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Rhode Island:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Wyoming:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Rhode Island:
        - https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2022_1040WE_w_0.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
        
        Wyoming:
        -         
        - https://revenue.wyo.gov/
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'RhodeIsland', 'state_occupation': 'Wyoming', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def south_carolina_and_alabama(self, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        South Carolina:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Alabama:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        South Carolina:
        - https://dor.sc.gov/resources-site/lawandpolicy/Advisory%20Opinions/IL22-15.pdf
          https://dor.sc.gov/forms-site/Forms/IITPacket_2022.pdf
          https://www.scstatehouse.gov/code/t12c006.php
          https://dor.sc.gov/resources-site/lawandpolicy/Advisory%20Opinions/IL22-15.pdf
        
        Alabama:
        -         
        - https://www.revenue.alabama.gov/ultraviewer/viewer/basic_viewer/index.html?form=2023/02/22f40.pdf
          https://www.revenue.alabama.gov/faqs/what-is-alabamas-individual-income-tax-rate/
          https://www.revenue.alabama.gov/ultraviewer/viewer/basic_viewer/index.html?form=2023/02/22f40.pdf
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-24.htm?taxmap=true
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-24.htm?taxmap=true
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-24.htm?taxmap=true
          https://www.revenue.alabama.gov/individual-corporate/income-to-be-reported-on-the-alabama-income-tax-return/
          https://www.almonline.org/TaxRates.aspx#OccupationalTax
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'SouthCarolina', 'state_occupation': 'Alabama', "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def south_carolina_and_alaska(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        South Carolina:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Alaska:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        South Carolina:
        - https://dor.sc.gov/resources-site/lawandpolicy/Advisory%20Opinions/IL22-15.pdf
          https://dor.sc.gov/forms-site/Forms/IITPacket_2022.pdf
          https://www.scstatehouse.gov/code/t12c006.php
          https://dor.sc.gov/resources-site/lawandpolicy/Advisory%20Opinions/IL22-15.pdf
        
        Alaska:
        -         
        - https://www.commerce.alaska.gov/web/dcra/officeofthestateassessor/alaskataxfacts.aspx
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'SouthCarolina', 'state_occupation': 'Alaska', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def south_carolina_and_arizona(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        South Carolina:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Arizona:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        credits_state : int, optional
            Credits for state taxes. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        South Carolina:
        - https://dor.sc.gov/resources-site/lawandpolicy/Advisory%20Opinions/IL22-15.pdf
          https://dor.sc.gov/forms-site/Forms/IITPacket_2022.pdf
          https://www.scstatehouse.gov/code/t12c006.php
          https://dor.sc.gov/resources-site/lawandpolicy/Advisory%20Opinions/IL22-15.pdf
        
        Arizona:
        -         
        - https://azdor.gov/forms/individual/form-140-x-y-tables
          https://azdor.gov/forms/individual/form-140-x-y-tables
          https://azdor.gov/forms/individual/form-140-x-y-tables
          https://azdor.gov/businesses-arizona/withholding-tax/withholding-exceptions
          "https://azdor.gov/individuals/income-tax-filing-assistance/identifying-other-taxable-income#:~:text=In%20regards%20to%20capital%20gains
         the%20individuals%20regular%20tax%20rate.
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'SouthCarolina', 'state_occupation': 'Arizona', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def south_carolina_and_arkansas(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        South Carolina:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Arkansas:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        South Carolina:
        - https://dor.sc.gov/resources-site/lawandpolicy/Advisory%20Opinions/IL22-15.pdf
          https://dor.sc.gov/forms-site/Forms/IITPacket_2022.pdf
          https://www.scstatehouse.gov/code/t12c006.php
          https://dor.sc.gov/resources-site/lawandpolicy/Advisory%20Opinions/IL22-15.pdf
        
        Arkansas:
        -         
        - https://www.arkansasedc.com/why-arkansas/business-climate/tax-structure/personal-income-tax
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_and_AR1000NR_Instructions.pdf
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_FullYearResidentIndividualIncomeTaxReturn.pdf
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_and_AR1000NR_Instructions.pdf
          https://www.arkansasedc.com/why-arkansas/business-climate/tax-structure/capital-gains-tax-reduction
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_and_AR1000NR_Instructions.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'SouthCarolina', 'state_occupation': 'Arkansas', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def south_carolina_and_california(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        South Carolina:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        California:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        South Carolina:
        - https://dor.sc.gov/resources-site/lawandpolicy/Advisory%20Opinions/IL22-15.pdf
          https://dor.sc.gov/forms-site/Forms/IITPacket_2022.pdf
          https://www.scstatehouse.gov/code/t12c006.php
          https://dor.sc.gov/resources-site/lawandpolicy/Advisory%20Opinions/IL22-15.pdf
        
        California:
        -         
        - https://www.ftb.ca.gov/forms/2022/2022-540-tax-rate-schedules.pdf
          https://www.ftb.ca.gov/file/personal/deductions/index.html
          https://www.ftb.ca.gov/about-ftb/newsroom/caleitc/eligibility-and-credit-information.html
          https://www.ftb.ca.gov/file/personal/income-types/capital-gains-and-losses.html#:~:text=California%20does%20not%20have%20a
         are%20taxed%20as%20ordinary%20income
          https://www.ftb.ca.gov/file/personal/
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'SouthCarolina', 'state_occupation': 'California', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def south_carolina_and_colorado(self, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        South Carolina:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Colorado:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        South Carolina:
        - https://dor.sc.gov/resources-site/lawandpolicy/Advisory%20Opinions/IL22-15.pdf
          https://dor.sc.gov/forms-site/Forms/IITPacket_2022.pdf
          https://www.scstatehouse.gov/code/t12c006.php
          https://dor.sc.gov/resources-site/lawandpolicy/Advisory%20Opinions/IL22-15.pdf
        
        Colorado:
        -         
        - https://leg.colorado.gov/agencies/legislative-council-staff/individual-income-tax%C2%A0
          https://leg.colorado.gov/agencies/legislative-council-staff/individual-income-tax%C2%A0
          https://tax.colorado.gov/sites/tax/files/documents/ITT_Capital_Gain_Subtraction_Feb_2022.pdf
          https://www.auroragov.org/business_services/taxes/occupational_privilege_tax
          https://www.denvergov.org/Government/Agencies-Departments-Offices/Agencies-Departments-Offices-Directory/Department-of-Finance/Our-Divisions/Treasury/Business-Tax-Information#:~:text=Occupational%20privilege%20tax
         -%28show%20below%29&text=Each%20taxable%20employee%20is%20liable
         month%20for%20each%20taxable%20employee
          https://www.glendale.co.us/355/Occupational-Privilege-Tax
          https://www.greenwoodvillage.com/1220/Occupational-Privilege-Tax-OPT
          https://ci.sheridan.co.us/288/Occupational-Privilege-Tax
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'SouthCarolina', 'state_occupation': 'Colorado', "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def south_carolina_and_connecticut(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        South Carolina:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Connecticut:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        South Carolina:
        - https://dor.sc.gov/resources-site/lawandpolicy/Advisory%20Opinions/IL22-15.pdf
          https://dor.sc.gov/forms-site/Forms/IITPacket_2022.pdf
          https://www.scstatehouse.gov/code/t12c006.php
          https://dor.sc.gov/resources-site/lawandpolicy/Advisory%20Opinions/IL22-15.pdf
        
        Connecticut:
        -         
        - https://portal.ct.gov/-/media/DRS/Forms/2022/Income/2022-CT-1040-Instructions_1222.pdf
          https://www.cga.ct.gov/2022/rpt/pdf/2022-R-0108.pdf
          https://www.cga.ct.gov/current/pub/chap_229.htm#sec_12-703
          https://portal.ct.gov/-/media/DRS/Forms/2022/Income/2022-CT-1040-Instructions_1222.pdf
          "https://portal.ct.gov/DRS/Publications/TSSNs/TSSN-29#:~:text=An%20individuals%20net%20capital%20gains
         portion%20of%20Social%20Security%20benefits"
          "https://portal.ct.gov/DRS/Publications/TSSNs/TSSN-29#:~:text=An%20individuals%20net%20capital%20gains
         portion%20of%20Social%20Security%20benefits"
          https://portal.ct.gov/DRS/Individuals/Resident-Income-Tax/Tax-Information
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'SouthCarolina', 'state_occupation': 'Connecticut', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def south_carolina_and_delaware(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, federal_agi=75000, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        South Carolina:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Delaware:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        federal_agi : int, optional
            Adjusted Gross Income (AGI) for federal taxes. Default is 0. Sources - https://www.irs.gov/e-file-providers/definition-of-adjusted-gross-income, https://www.irs.gov/newsroom/taxpayers-should-know-the-difference-between-standard-and-itemized-deductions
        income : int, optional
            Gross income. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        South Carolina:
        - https://dor.sc.gov/resources-site/lawandpolicy/Advisory%20Opinions/IL22-15.pdf
          https://dor.sc.gov/forms-site/Forms/IITPacket_2022.pdf
          https://www.scstatehouse.gov/code/t12c006.php
          https://dor.sc.gov/resources-site/lawandpolicy/Advisory%20Opinions/IL22-15.pdf
        
        Delaware:
        -         
        - https://revenue.delaware.gov/software-developer/tax-rate-changes/
          https://revenuefiles.delaware.gov/2022/PIT-RES_TY22_2022-01_PaperInteractive.pdf
          https://revenuefiles.delaware.gov/2022/PIT-RES_TY22_2022-02_Instructions.pdf
          https://revenue.delaware.gov/frequently-asked-questions/personal-income-tax-faqs/
          https://www.newcastlede.gov/DocumentCenter/View/20147/City-of-Wilmington-Wage-Tax-Information-PDF
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        federal_agi = self.federal_agi if hasattr(self, 'federal_agi') else federal_agi
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'SouthCarolina', 'state_occupation': 'Delaware', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "federal_agi": federal_agi, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def south_carolina_and_florida(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        South Carolina:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Florida:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        South Carolina:
        - https://dor.sc.gov/resources-site/lawandpolicy/Advisory%20Opinions/IL22-15.pdf
          https://dor.sc.gov/forms-site/Forms/IITPacket_2022.pdf
          https://www.scstatehouse.gov/code/t12c006.php
          https://dor.sc.gov/resources-site/lawandpolicy/Advisory%20Opinions/IL22-15.pdf
        
        Florida:
        -         
        - https://floridarevenue.com/taxes/taxesfees/Pages/tax_interest_rates.aspx
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'SouthCarolina', 'state_occupation': 'Florida', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def south_carolina_and_georgia(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        South Carolina:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Georgia:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        South Carolina:
        - https://dor.sc.gov/resources-site/lawandpolicy/Advisory%20Opinions/IL22-15.pdf
          https://dor.sc.gov/forms-site/Forms/IITPacket_2022.pdf
          https://www.scstatehouse.gov/code/t12c006.php
          https://dor.sc.gov/resources-site/lawandpolicy/Advisory%20Opinions/IL22-15.pdf
        
        Georgia:
        -         
        - https://dor.georgia.gov/tax-tables-georgia-tax-rate-schedule
          https://apps.dor.ga.gov/FillableForms/PDFViewer/Index?form=2022GA500
          https://apps.dor.ga.gov/FillableForms/PDFViewer/Index?form=2022GA500
          https://apps.dor.ga.gov/FillableForms/PDFViewer/Index?form=2022GA500
          https://dor.georgia.gov/years-individual-income-tax-forms
          https://dor.georgia.gov/years-individual-income-tax-forms
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'SouthCarolina', 'state_occupation': 'Georgia', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def south_carolina_and_hawaii(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        South Carolina:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Hawaii:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        South Carolina:
        - https://dor.sc.gov/resources-site/lawandpolicy/Advisory%20Opinions/IL22-15.pdf
          https://dor.sc.gov/forms-site/Forms/IITPacket_2022.pdf
          https://www.scstatehouse.gov/code/t12c006.php
          https://dor.sc.gov/resources-site/lawandpolicy/Advisory%20Opinions/IL22-15.pdf
        
        Hawaii:
        -         
        - https://tax.hawaii.gov/forms/d_18table-on/d_18table-on_p13/
          https://files.hawaii.gov/tax/forms/2018/18table-on.pdf
          https://files.hawaii.gov/tax/news/pubs/22outline.pdf
          https://files.hawaii.gov/tax/news/pubs/FastTaxRefGuide_2017.pdf
          https://www.tfhawaii.org/wordpress/state-tax-resources/mini-tax-guides/individual-income-tax-chapter-235/#:~:text=Long%20term%20capital%20gains%20are%20taxed%20at%20a%20maximum%20of%207.25%25
          https://www.grassrootinstitute.org/2023/02/hb337-increasing-tax-on-capital-gains-a-bad-risk/
          https://files.hawaii.gov/tax/news/pubs/22outline.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'SouthCarolina', 'state_occupation': 'Hawaii', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def south_carolina_and_idaho(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        South Carolina:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Idaho:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        South Carolina:
        - https://dor.sc.gov/resources-site/lawandpolicy/Advisory%20Opinions/IL22-15.pdf
          https://dor.sc.gov/forms-site/Forms/IITPacket_2022.pdf
          https://www.scstatehouse.gov/code/t12c006.php
          https://dor.sc.gov/resources-site/lawandpolicy/Advisory%20Opinions/IL22-15.pdf
        
        Idaho:
        -         
        - https://tax.idaho.gov/taxes/income-tax/individual-income/individual-income-tax-rate-schedule/
          https://tax.idaho.gov/wp-content/uploads/forms/EIN00046/EIN00046_03-01-2023.pdf
          https://tax.idaho.gov/taxes/income-tax/individual-income/idaho-source-income/
          https://tax.idaho.gov/taxes/income-tax/individual-income/idaho-source-income/
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'SouthCarolina', 'state_occupation': 'Idaho', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def south_carolina_and_illinois(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        South Carolina:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Illinois:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        South Carolina:
        - https://dor.sc.gov/resources-site/lawandpolicy/Advisory%20Opinions/IL22-15.pdf
          https://dor.sc.gov/forms-site/Forms/IITPacket_2022.pdf
          https://www.scstatehouse.gov/code/t12c006.php
          https://dor.sc.gov/resources-site/lawandpolicy/Advisory%20Opinions/IL22-15.pdf
        
        Illinois:
        -         
        - https://tax.illinois.gov/research/taxrates/income.html#IndividualIncome
          https://tax.illinois.gov/content/dam/soi/en/web/tax/forms/incometax/documents/currentyear/individual/il-1040-instr.pdf
          https://tax.illinois.gov/questionsandanswers/12.html
          https://tax.illinois.gov/content/dam/soi/en/web/tax/forms/incometax/documents/currentyear/individual/il-1040-instr.pdf
          https://tax.illinois.gov/research/taxrates/personalproperty.html
          https://tax.illinois.gov/localgovernments/income.html
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'SouthCarolina', 'state_occupation': 'Illinois', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def south_carolina_and_indiana(self, capital_gains_short=0, dependents=0, sources=False, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_long=0, filing_status='single', income=100000, credits_state=0):
        """
        --- Parameters ----
        
        South Carolina:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Indiana:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        South Carolina:
        - https://dor.sc.gov/resources-site/lawandpolicy/Advisory%20Opinions/IL22-15.pdf
          https://dor.sc.gov/forms-site/Forms/IITPacket_2022.pdf
          https://www.scstatehouse.gov/code/t12c006.php
          https://dor.sc.gov/resources-site/lawandpolicy/Advisory%20Opinions/IL22-15.pdf
        
        Indiana:
        -         
        - https://www.in.gov/dor/tax-forms/2022-individual-income-tax-forms/
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-32.htm
          https://www.in.gov/dor/files/ib28.pdf
          https://www.in.gov/dor/tax-forms/2022-individual-income-tax-forms/
          https://www.in.gov/dor/files/2022-county-tax-rates-and-codes.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'SouthCarolina', 'state_occupation': 'Indiana', "capital_gains_short": capital_gains_short, "dependents": dependents, "sources": sources, "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def south_carolina_and_iowa(self, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        South Carolina:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Iowa:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        South Carolina:
        - https://dor.sc.gov/resources-site/lawandpolicy/Advisory%20Opinions/IL22-15.pdf
          https://dor.sc.gov/forms-site/Forms/IITPacket_2022.pdf
          https://www.scstatehouse.gov/code/t12c006.php
          https://dor.sc.gov/resources-site/lawandpolicy/Advisory%20Opinions/IL22-15.pdf
        
        Iowa:
        -         
        - https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
          https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
          https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
          https://tax.iowa.gov/iowa-illinois-reciprocal-agreement
          https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'SouthCarolina', 'state_occupation': 'Iowa', "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def south_carolina_and_kansas(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        South Carolina:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Kansas:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        South Carolina:
        - https://dor.sc.gov/resources-site/lawandpolicy/Advisory%20Opinions/IL22-15.pdf
          https://dor.sc.gov/forms-site/Forms/IITPacket_2022.pdf
          https://www.scstatehouse.gov/code/t12c006.php
          https://dor.sc.gov/resources-site/lawandpolicy/Advisory%20Opinions/IL22-15.pdf
        
        Kansas:
        -         
        - https://www.ksrevenue.gov/taxrates.html
          https://www.ksrevenue.gov/incomebook22.html#0
          https://www.ksrevenue.gov/incomebook22.html#0
          https://www.ksrevenue.gov/incomebook22.html#0
          https://www.ksrevenue.gov/incomebook22.html
          https://www.ksrevenue.gov/incomebook22.html
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'SouthCarolina', 'state_occupation': 'Kansas', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def south_carolina_and_kentucky(self, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        South Carolina:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Kentucky:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        South Carolina:
        - https://dor.sc.gov/resources-site/lawandpolicy/Advisory%20Opinions/IL22-15.pdf
          https://dor.sc.gov/forms-site/Forms/IITPacket_2022.pdf
          https://www.scstatehouse.gov/code/t12c006.php
          https://dor.sc.gov/resources-site/lawandpolicy/Advisory%20Opinions/IL22-15.pdf
        
        Kentucky:
        -         
        - https://revenue.ky.gov/News/Pages/Dor-Announces-Updates-to-Individual-Income-Tax-for-2023-Tax-Year.aspx
          https://revenue.ky.gov/Individual/Individual-Income-Tax/Pages/default.aspx
          https://revenue.ky.gov/News/Pages/Dor-Announces-Updates-to-Individual-Income-Tax-for-2023-Tax-Year.aspx
          https://revenue.ky.gov/Individual/Individual-Income-Tax/Pages/default.aspx
          https://revenue.ky.gov/Dor%20Training%20Materials/103%20KAR%2017.140.%20Individual%20income%20tax%20-%20reciprocity%20-%20nonresidents.pdf
          https://revenue.ky.gov/Individual/Individual-Income-Tax/Pages/default.aspx
          https://nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2021/Taxes-21-32.htm?taxmap=true
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'SouthCarolina', 'state_occupation': 'Kentucky', "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def south_carolina_and_louisiana(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        South Carolina:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Louisiana:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        South Carolina:
        - https://dor.sc.gov/resources-site/lawandpolicy/Advisory%20Opinions/IL22-15.pdf
          https://dor.sc.gov/forms-site/Forms/IITPacket_2022.pdf
          https://www.scstatehouse.gov/code/t12c006.php
          https://dor.sc.gov/resources-site/lawandpolicy/Advisory%20Opinions/IL22-15.pdf
        
        Louisiana:
        -         
        - https://revenue.louisiana.gov/individualincometax
          https://revenue.louisiana.gov/individualincometax
          https://www.revenue.louisiana.gov/taxforms/6935(11_02)F.pdf
          https://revenue.louisiana.gov/individualincometax
          https://revenue.louisiana.gov/individualincometax
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'SouthCarolina', 'state_occupation': 'Louisiana', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def south_carolina_and_maine(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        South Carolina:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Maine:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        South Carolina:
        - https://dor.sc.gov/resources-site/lawandpolicy/Advisory%20Opinions/IL22-15.pdf
          https://dor.sc.gov/forms-site/Forms/IITPacket_2022.pdf
          https://www.scstatehouse.gov/code/t12c006.php
          https://dor.sc.gov/resources-site/lawandpolicy/Advisory%20Opinions/IL22-15.pdf
        
        Maine:
        -         
        - https://www.maine.gov/revenue/tax-return-forms/individual-income-tax-2022
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/ind_tax_rate_sched_2022.pdf
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/22_1040me_book_gen_instruc_revisedFeb23.pdf
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/22_1040me_book_gen_instruc_revisedFeb23.pdf
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/22_1040me_book_gen_instruc_revisedFeb23.pdf
          https://www.maine.gov/revenue/tax-return-forms/individual-income-tax-2022
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'SouthCarolina', 'state_occupation': 'Maine', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def south_carolina_and_maryland(self, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        South Carolina:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Maryland:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        South Carolina:
        - https://dor.sc.gov/resources-site/lawandpolicy/Advisory%20Opinions/IL22-15.pdf
          https://dor.sc.gov/forms-site/Forms/IITPacket_2022.pdf
          https://www.scstatehouse.gov/code/t12c006.php
          https://dor.sc.gov/resources-site/lawandpolicy/Advisory%20Opinions/IL22-15.pdf
        
        Maryland:
        -         
        - https://www.marylandtaxes.gov/individual/income/tax-info/tax-rates.php
          https://www.marylandtaxes.gov/forms/22_forms/Resident_Booklet.pdf
          https://www.marylandtaxes.gov/forms/Tax_Publications/Administrative_Releases/Income_and_Estate_Tax_Releases/ar_it37.pdf
          https://www.marylandtaxes.gov/individual/index.php
          https://www.marylandtaxes.gov/individual/income/tax-info/tax-rates.php
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'SouthCarolina', 'state_occupation': 'Maryland', "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def south_carolina_and_massachusetts(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        South Carolina:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Massachusetts:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        South Carolina:
        - https://dor.sc.gov/resources-site/lawandpolicy/Advisory%20Opinions/IL22-15.pdf
          https://dor.sc.gov/forms-site/Forms/IITPacket_2022.pdf
          https://www.scstatehouse.gov/code/t12c006.php
          https://dor.sc.gov/resources-site/lawandpolicy/Advisory%20Opinions/IL22-15.pdf
        
        Massachusetts:
        -         
        - https://www.mass.gov/service-details/massachusetts-tax-rates
          https://www.mass.gov/service-details/view-massachusetts-personal-income-tax-exemptions
          https://www.mass.gov/service-details/view-massachusetts-personal-income-tax-exemptions
          https://www.mass.gov/service-details/view-child-and-dependent-related-credits
          https://www.mass.gov/service-details/massachusetts-tax-rates
          https://www.mass.gov/personal-income-tax
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'SouthCarolina', 'state_occupation': 'Massachusetts', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def south_carolina_and_michigan(self, capital_gains_short=0, county_occupation='Other', dependents=0, sources=False, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_long=0, filing_status='single', income=100000, credits_state=0):
        """
        --- Parameters ----
        
        South Carolina:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Michigan:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        county_occupation : str, optional
            County of occupation. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        South Carolina:
        - https://dor.sc.gov/resources-site/lawandpolicy/Advisory%20Opinions/IL22-15.pdf
          https://dor.sc.gov/forms-site/Forms/IITPacket_2022.pdf
          https://www.scstatehouse.gov/code/t12c006.php
          https://dor.sc.gov/resources-site/lawandpolicy/Advisory%20Opinions/IL22-15.pdf
        
        Michigan:
        -         
        - https://www.michigan.gov/taxes/iit/new-developments-for-tax-year-2022
          "https://www.michigan.gov/taxes/-/media/Project/Websites/taxes/Forms/2023/2023-SUW/446_Withholding-Guide_2023.pdf?rev=177ca3d8cb034392bf84e317ed78c320&hash=19C5F42F43A89A91863D229349F48883#:~:text=The%20exemption%20amount%20is%20%245%2C400
         employees%20Michigan%20income%20tax%20return"
          "https://www.michigan.gov/taxes/-/media/Project/Websites/taxes/Forms/2023/2023-SUW/446_Withholding-Guide_2023.pdf?rev=177ca3d8cb034392bf84e317ed78c320&hash=19C5F42F43A89A91863D229349F48883#:~:text=The%20exemption%20amount%20is%20%245%2C400
         employees%20Michigan%20income%20tax%20return"
          https://www.michigan.gov/taxes/questions/iit/accordion/residency/are-my-wages-earned-in-another-state-taxable-in-michigan-if-i-am-a-michigan-resident-1
          https://www.michigan.gov/taxes/-/media/Project/Websites/taxes/Forms/2022/2022-IIT-Forms/BOOK_MI-1040.pdf?rev=4524f1cbbc0243b194ab05a6680be75b&hash=40DFE71E651F7C1D65C31F6CD03D944F
          https://www.michigan.gov/taxes/questions/iit/accordion/general/what-cities-impose-an-income-tax
        """
        
        # Check if the parameters are set upstream in the instance
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        county_occupation = self.county_occupation if hasattr(self, 'county_occupation') else county_occupation
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'SouthCarolina', 'state_occupation': 'Michigan', "capital_gains_short": capital_gains_short, "county_occupation": county_occupation, "dependents": dependents, "sources": sources, "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def south_carolina_and_minnesota(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        South Carolina:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Minnesota:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        South Carolina:
        - https://dor.sc.gov/resources-site/lawandpolicy/Advisory%20Opinions/IL22-15.pdf
          https://dor.sc.gov/forms-site/Forms/IITPacket_2022.pdf
          https://www.scstatehouse.gov/code/t12c006.php
          https://dor.sc.gov/resources-site/lawandpolicy/Advisory%20Opinions/IL22-15.pdf
        
        Minnesota:
        -         
        - https://www.revenue.state.mn.us/minnesota-income-tax-rates-and-brackets
          https://www.revenue.state.mn.us/child-and-dependent-care-credit
          https://www.revenue.state.mn.us/dependent-exemptions
          https://www.revenue.state.mn.us/minnesota-standard-deduction
          https://mn.gov/mmb/assets/mwr_form_tcm1059-128581.pdf
          https://www.revenue.state.mn.us/individual-income-tax
          https://www.revenue.state.mn.us/individual-income-tax
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'SouthCarolina', 'state_occupation': 'Minnesota', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def south_carolina_and_mississippi(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        South Carolina:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Mississippi:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        South Carolina:
        - https://dor.sc.gov/resources-site/lawandpolicy/Advisory%20Opinions/IL22-15.pdf
          https://dor.sc.gov/forms-site/Forms/IITPacket_2022.pdf
          https://www.scstatehouse.gov/code/t12c006.php
          https://dor.sc.gov/resources-site/lawandpolicy/Advisory%20Opinions/IL22-15.pdf
        
        Mississippi:
        -         
        - https://www.dor.ms.gov/individual/tax-rates
          https://www.dor.ms.gov/individual/tax-rates
          https://www.dor.ms.gov/individual/tax-rates
          https://www.dor.ms.gov/individual
          https://www.dor.ms.gov/individual
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'SouthCarolina', 'state_occupation': 'Mississippi', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def south_carolina_and_missouri(self, capital_gains_short=0, dependents=0, sources=False, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_long=0, filing_status='single', income=100000, credits_state=0):
        """
        --- Parameters ----
        
        South Carolina:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Missouri:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        South Carolina:
        - https://dor.sc.gov/resources-site/lawandpolicy/Advisory%20Opinions/IL22-15.pdf
          https://dor.sc.gov/forms-site/Forms/IITPacket_2022.pdf
          https://www.scstatehouse.gov/code/t12c006.php
          https://dor.sc.gov/resources-site/lawandpolicy/Advisory%20Opinions/IL22-15.pdf
        
        Missouri:
        -         
        - https://dor.mo.gov/taxation/individual/tax-types/income/year-changes/
          https://dor.mo.gov/taxation/individual/tax-types/income/year-changes/
          https://dor.mo.gov/taxation/individual/tax-types/income/year-changes/
          https://www.kcmo.gov/city-hall/departments/finance/earnings-tax
          https://www.stlouis-mo.gov/government/departments/collector/earnings-tax/individual-earnings-tax-info.cfm
        """
        
        # Check if the parameters are set upstream in the instance
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'SouthCarolina', 'state_occupation': 'Missouri', "capital_gains_short": capital_gains_short, "dependents": dependents, "sources": sources, "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def south_carolina_and_montana(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        South Carolina:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Montana:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        South Carolina:
        - https://dor.sc.gov/resources-site/lawandpolicy/Advisory%20Opinions/IL22-15.pdf
          https://dor.sc.gov/forms-site/Forms/IITPacket_2022.pdf
          https://www.scstatehouse.gov/code/t12c006.php
          https://dor.sc.gov/resources-site/lawandpolicy/Advisory%20Opinions/IL22-15.pdf
        
        Montana:
        -         
        - https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/2022/12/montana-employees-withholding-allowance-and-exemption-certificate-form-mw-4.pdf
          https://mtrevenue.gov/2021/11/29/simplification-of-montana-income-taxation/?utm_source=rss&utm_medium=rss&utm_campaign=simplification-of-montana-income-taxation
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'SouthCarolina', 'state_occupation': 'Montana', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def south_carolina_and_nebraska(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        South Carolina:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Nebraska:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        South Carolina:
        - https://dor.sc.gov/resources-site/lawandpolicy/Advisory%20Opinions/IL22-15.pdf
          https://dor.sc.gov/forms-site/Forms/IITPacket_2022.pdf
          https://www.scstatehouse.gov/code/t12c006.php
          https://dor.sc.gov/resources-site/lawandpolicy/Advisory%20Opinions/IL22-15.pdf
        
        Nebraska:
        -         
        - https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2023_Circular_EN_8-429r2023_Whole_Book_11.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'SouthCarolina', 'state_occupation': 'Nebraska', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def south_carolina_and_nevada(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        South Carolina:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Nevada:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        South Carolina:
        - https://dor.sc.gov/resources-site/lawandpolicy/Advisory%20Opinions/IL22-15.pdf
          https://dor.sc.gov/forms-site/Forms/IITPacket_2022.pdf
          https://www.scstatehouse.gov/code/t12c006.php
          https://dor.sc.gov/resources-site/lawandpolicy/Advisory%20Opinions/IL22-15.pdf
        
        Nevada:
        -         
        - https://www.nvsos.gov/sos/businesses/the-nevada-advantage
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'SouthCarolina', 'state_occupation': 'Nevada', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def south_carolina_and_new_hampshire(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        South Carolina:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        New Hampshire:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        South Carolina:
        - https://dor.sc.gov/resources-site/lawandpolicy/Advisory%20Opinions/IL22-15.pdf
          https://dor.sc.gov/forms-site/Forms/IITPacket_2022.pdf
          https://www.scstatehouse.gov/code/t12c006.php
          https://dor.sc.gov/resources-site/lawandpolicy/Advisory%20Opinions/IL22-15.pdf
        
        New Hampshire:
        -         
        - https://www.revenue.nh.gov/assistance/tax-overview.htm
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'SouthCarolina', 'state_occupation': 'NewHampshire', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def south_carolina_and_new_jersey(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        South Carolina:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        New Jersey:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        South Carolina:
        - https://dor.sc.gov/resources-site/lawandpolicy/Advisory%20Opinions/IL22-15.pdf
          https://dor.sc.gov/forms-site/Forms/IITPacket_2022.pdf
          https://www.scstatehouse.gov/code/t12c006.php
          https://dor.sc.gov/resources-site/lawandpolicy/Advisory%20Opinions/IL22-15.pdf
        
        New Jersey:
        -         
        - https://www.state.nj.us/treasury/taxation/taxtables.shtml
          https://www.state.nj.us/treasury/taxation/njit13.shtml
          https://www.nj.gov/treasury/taxation/njit25.shtml
          https://www.state.nj.us/treasury/taxation/individuals/freqqiti.shtml
          https://www.state.nj.us/treasury/taxation/localtax.shtml
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'SouthCarolina', 'state_occupation': 'NewJersey', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def south_carolina_and_new_mexico(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        South Carolina:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        New Mexico:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        South Carolina:
        - https://dor.sc.gov/resources-site/lawandpolicy/Advisory%20Opinions/IL22-15.pdf
          https://dor.sc.gov/forms-site/Forms/IITPacket_2022.pdf
          https://www.scstatehouse.gov/code/t12c006.php
          https://dor.sc.gov/resources-site/lawandpolicy/Advisory%20Opinions/IL22-15.pdf
        
        New Mexico:
        -         
        - https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/fdf3c548-8aba-4b9c-9eb4-bb564c716015/FYI-104.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/0bf943fd-652e-4400-bb1b-cd8397eb5a95/2022pit-adj.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/2f1a6781-9534-4436-b427-1557f9592099/2022pit-adj-ins.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/0bf943fd-652e-4400-bb1b-cd8397eb5a95/2022pit-adj.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/1afc56af-ea90-4d48-82e5-1f9aeb43255a/PITbook2022.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/fdf3c548-8aba-4b9c-9eb4-bb564c716015/FYI-104.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'SouthCarolina', 'state_occupation': 'NewMexico', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def south_carolina_and_new_york(self, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        South Carolina:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        New York:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        South Carolina:
        - https://dor.sc.gov/resources-site/lawandpolicy/Advisory%20Opinions/IL22-15.pdf
          https://dor.sc.gov/forms-site/Forms/IITPacket_2022.pdf
          https://www.scstatehouse.gov/code/t12c006.php
          https://dor.sc.gov/resources-site/lawandpolicy/Advisory%20Opinions/IL22-15.pdf
        
        New York:
        -         
        - https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nys-tax-rate-schedule
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nys-tax-rate-schedule
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nys-tax-rate-schedule
          https://www.tax.ny.gov/pit/
          https://www.tax.ny.gov/pit/file/nonresident-faqs.htm
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#yonkers-tax-rate-schedule
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nyc-tax-rate-schedule
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'SouthCarolina', 'state_occupation': 'NewYork', "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def south_carolina_and_north_carolina(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        South Carolina:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        North Carolina:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        South Carolina:
        - https://dor.sc.gov/resources-site/lawandpolicy/Advisory%20Opinions/IL22-15.pdf
          https://dor.sc.gov/forms-site/Forms/IITPacket_2022.pdf
          https://www.scstatehouse.gov/code/t12c006.php
          https://dor.sc.gov/resources-site/lawandpolicy/Advisory%20Opinions/IL22-15.pdf
        
        North Carolina:
        -         
        - https://www.ncdor.gov/taxes-forms/tax-rate-schedules
          https://www.ncdor.gov/taxes-forms/individual-income-tax/north-carolina-standard-deduction-or-north-carolina-itemized-deductions
          https://www.ncdor.gov/taxes-forms/individual-income-tax
          https://www.ncdor.gov/taxes-forms/individual-income-tax
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'SouthCarolina', 'state_occupation': 'NorthCarolina', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def south_carolina_and_north_dakota(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        South Carolina:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        North Dakota:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        South Carolina:
        - https://dor.sc.gov/resources-site/lawandpolicy/Advisory%20Opinions/IL22-15.pdf
          https://dor.sc.gov/forms-site/Forms/IITPacket_2022.pdf
          https://www.scstatehouse.gov/code/t12c006.php
          https://dor.sc.gov/resources-site/lawandpolicy/Advisory%20Opinions/IL22-15.pdf
        
        North Dakota:
        -         
        - https://www.tax.nd.gov/forms
          https://www.tax.nd.gov/sites/www/files/documents/forms/individual/2022-iit/2022-individual-income-tax-booklet.pdf
          https://www.tax.nd.gov/sites/www/files/documents/forms/2022-individual-income-tax-booklet.pdf
          https://www.tax.nd.gov/tax-exemptions-credits/income-tax-exemptions-credits
          https://www.tax.nd.gov/sites/www/files/documents/forms/business/ndwrfillable.pdf
          https://www.tax.nd.gov/news/tax-legislative-changes/significant-changes-law/individual-income-tax-history
          https://www.tax.nd.gov/forms
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'SouthCarolina', 'state_occupation': 'NorthDakota', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def south_carolina_and_ohio(self, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        South Carolina:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Ohio:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        South Carolina:
        - https://dor.sc.gov/resources-site/lawandpolicy/Advisory%20Opinions/IL22-15.pdf
          https://dor.sc.gov/forms-site/Forms/IITPacket_2022.pdf
          https://www.scstatehouse.gov/code/t12c006.php
          https://dor.sc.gov/resources-site/lawandpolicy/Advisory%20Opinions/IL22-15.pdf
        
        Ohio:
        -         
        - https://tax.ohio.gov/individual/resources/annual-tax-rates
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://tax.ohio.gov/static/forms/employer_withholding/generic/wth_it4nr.pdf
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://www.ritaohio.com/TaxRatesTable
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'SouthCarolina', 'state_occupation': 'Ohio', "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def south_carolina_and_oklahoma(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        South Carolina:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Oklahoma:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        South Carolina:
        - https://dor.sc.gov/resources-site/lawandpolicy/Advisory%20Opinions/IL22-15.pdf
          https://dor.sc.gov/forms-site/Forms/IITPacket_2022.pdf
          https://www.scstatehouse.gov/code/t12c006.php
          https://dor.sc.gov/resources-site/lawandpolicy/Advisory%20Opinions/IL22-15.pdf
        
        Oklahoma:
        -         
        - https://oklahoma.gov/content/dam/ok/en/tax/documents/resources/publications/businesses/withholding-tables/WHTables-2022.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'SouthCarolina', 'state_occupation': 'Oklahoma', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def south_carolina_and_oregon(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        South Carolina:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Oregon:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        South Carolina:
        - https://dor.sc.gov/resources-site/lawandpolicy/Advisory%20Opinions/IL22-15.pdf
          https://dor.sc.gov/forms-site/Forms/IITPacket_2022.pdf
          https://www.scstatehouse.gov/code/t12c006.php
          https://dor.sc.gov/resources-site/lawandpolicy/Advisory%20Opinions/IL22-15.pdf
        
        Oregon:
        -         
        - https://www.oregon.gov/dor/programs/individuals/Pages/PIT.aspx
          https://www.oregon.gov/dor/programs/individuals/Documents/Full%20year%20resident
         %20Form%20or-40%20filers.pdf
          https://www.oregon.gov/dor/programs/individuals/Documents/Part-year%20and%20nonresident
         %20Form%20or-40-P%20and%20or-40-N%20filers.pdf.pdf
          https://www.oregon.gov/dor/forms/FormsPubs/form-or-40_101-040_2022.pdf
          https://www.oregon.gov/dor/forms/FormsPubs/form-or-40_101-040_2022.pdf
          https://www.oregon.gov/dor/forms/FormsPubs/form-or-40_101-040_2022.pdf
          https://www.portland.gov/revenue/personal-tax
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'SouthCarolina', 'state_occupation': 'Oregon', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def south_carolina_and_pennsylvania(self, county_residence='Other', state_agi=75000, county_occupation='Other', filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        South Carolina:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Pennsylvania:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        county_residence : str, optional
            County of residence. Default is Other.
        county_occupation : str, optional
            County of occupation. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        South Carolina:
        - https://dor.sc.gov/resources-site/lawandpolicy/Advisory%20Opinions/IL22-15.pdf
          https://dor.sc.gov/forms-site/Forms/IITPacket_2022.pdf
          https://www.scstatehouse.gov/code/t12c006.php
          https://dor.sc.gov/resources-site/lawandpolicy/Advisory%20Opinions/IL22-15.pdf
        
        Pennsylvania:
        -         
        - https://www.revenue.pa.gov/Tax%20Rates/Pages/default.aspx
          https://www.revenue.pa.gov/FormsandPublications/FormsforIndividuals/PIT/Documents/2022/2022_pa-40in.pdf
          https://www.revenue.pa.gov/FormsandPublications/FormsforBusinesses/EmployerWithholding/Documents/rev-419.pdf
          https://www.revenue.pa.gov/FormsandPublications/FormsforIndividuals/PIT/Documents/2022/2022_pa-40in.pdf
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-25.htm
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        county_occupation = self.county_occupation if hasattr(self, 'county_occupation') else county_occupation
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'SouthCarolina', 'state_occupation': 'Pennsylvania', "county_residence": county_residence, "state_agi": state_agi, "county_occupation": county_occupation, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def south_carolina_and_rhode_island(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        South Carolina:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Rhode Island:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        South Carolina:
        - https://dor.sc.gov/resources-site/lawandpolicy/Advisory%20Opinions/IL22-15.pdf
          https://dor.sc.gov/forms-site/Forms/IITPacket_2022.pdf
          https://www.scstatehouse.gov/code/t12c006.php
          https://dor.sc.gov/resources-site/lawandpolicy/Advisory%20Opinions/IL22-15.pdf
        
        Rhode Island:
        -         
        - https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2022_1040WE_w_0.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'SouthCarolina', 'state_occupation': 'RhodeIsland', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def south_carolina_and_south_dakota(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        South Carolina:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        South Dakota:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        South Carolina:
        - https://dor.sc.gov/resources-site/lawandpolicy/Advisory%20Opinions/IL22-15.pdf
          https://dor.sc.gov/forms-site/Forms/IITPacket_2022.pdf
          https://www.scstatehouse.gov/code/t12c006.php
          https://dor.sc.gov/resources-site/lawandpolicy/Advisory%20Opinions/IL22-15.pdf
        
        South Dakota:
        -         
        - https://dor.sd.gov/individuals/taxes/
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'SouthCarolina', 'state_occupation': 'SouthDakota', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def south_carolina_and_tennessee(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        South Carolina:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Tennessee:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        South Carolina:
        - https://dor.sc.gov/resources-site/lawandpolicy/Advisory%20Opinions/IL22-15.pdf
          https://dor.sc.gov/forms-site/Forms/IITPacket_2022.pdf
          https://www.scstatehouse.gov/code/t12c006.php
          https://dor.sc.gov/resources-site/lawandpolicy/Advisory%20Opinions/IL22-15.pdf
        
        Tennessee:
        -         
        - https://www.tn.gov/revenue/taxes.html
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'SouthCarolina', 'state_occupation': 'Tennessee', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def south_carolina_and_texas(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        South Carolina:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Texas:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        South Carolina:
        - https://dor.sc.gov/resources-site/lawandpolicy/Advisory%20Opinions/IL22-15.pdf
          https://dor.sc.gov/forms-site/Forms/IITPacket_2022.pdf
          https://www.scstatehouse.gov/code/t12c006.php
          https://dor.sc.gov/resources-site/lawandpolicy/Advisory%20Opinions/IL22-15.pdf
        
        Texas:
        -         
        - https://comptroller.texas.gov/taxes/
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'SouthCarolina', 'state_occupation': 'Texas', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def south_carolina_and_utah(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        South Carolina:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Utah:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        South Carolina:
        - https://dor.sc.gov/resources-site/lawandpolicy/Advisory%20Opinions/IL22-15.pdf
          https://dor.sc.gov/forms-site/Forms/IITPacket_2022.pdf
          https://www.scstatehouse.gov/code/t12c006.php
          https://dor.sc.gov/resources-site/lawandpolicy/Advisory%20Opinions/IL22-15.pdf
        
        Utah:
        -         
        - https://incometax.utah.gov/paying/tax-rates
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'SouthCarolina', 'state_occupation': 'Utah', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def south_carolina_and_vermont(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        South Carolina:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Vermont:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        South Carolina:
        - https://dor.sc.gov/resources-site/lawandpolicy/Advisory%20Opinions/IL22-15.pdf
          https://dor.sc.gov/forms-site/Forms/IITPacket_2022.pdf
          https://www.scstatehouse.gov/code/t12c006.php
          https://dor.sc.gov/resources-site/lawandpolicy/Advisory%20Opinions/IL22-15.pdf
        
        Vermont:
        -         
        - https://tax.vermont.gov/sites/tax/files/documents/GB-1210-2023.pdf
          https://tax.vermont.gov/sites/tax/files/documents/Income%20Booklet-2022.pdf
          https://tax.vermont.gov/sites/tax/files/documents/Income%20Booklet-2022.pdf
          https://tax.vermont.gov/sites/tax/files/documents/Income%20Booklet-2022.pdf
          https://tax.vermont.gov/individuals/personal-income-tax/tax-credits
          https://tax.vermont.gov/individuals/personal-income-tax/taxable-income
          https://tax.vermont.gov/individuals/personal-income-tax/local-option-tax
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'SouthCarolina', 'state_occupation': 'Vermont', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def south_carolina_and_virginia(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        South Carolina:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Virginia:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        South Carolina:
        - https://dor.sc.gov/resources-site/lawandpolicy/Advisory%20Opinions/IL22-15.pdf
          https://dor.sc.gov/forms-site/Forms/IITPacket_2022.pdf
          https://www.scstatehouse.gov/code/t12c006.php
          https://dor.sc.gov/resources-site/lawandpolicy/Advisory%20Opinions/IL22-15.pdf
        
        Virginia:
        -         
        - https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/reciprocity
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'SouthCarolina', 'state_occupation': 'Virginia', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def south_carolina_and_washington(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        South Carolina:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Washington:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        South Carolina:
        - https://dor.sc.gov/resources-site/lawandpolicy/Advisory%20Opinions/IL22-15.pdf
          https://dor.sc.gov/forms-site/Forms/IITPacket_2022.pdf
          https://www.scstatehouse.gov/code/t12c006.php
          https://dor.sc.gov/resources-site/lawandpolicy/Advisory%20Opinions/IL22-15.pdf
        
        Washington:
        -         
        - https://dor.wa.gov/taxes-rates/income-tax
          https://dor.wa.gov/taxes-rates/other-taxes/capital-gains-tax
          https://dor.wa.gov/taxes-rates/income-tax
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'SouthCarolina', 'state_occupation': 'Washington', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def south_carolina_and_west_virginia(self, pay_periods=24, capital_gains_short=0, county_occupation='Other', dependents=0, sources=False, state_agi=75000, filers_over_65=0, capital_gains_long=0, filing_status='single', income=100000, credits_state=0):
        """
        --- Parameters ----
        
        South Carolina:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        West Virginia:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_occupation : str, optional
            County of occupation. Default is Other.
        pay_periods : int, optional
            Number of pay periods. Options are 13, 24, 26, 52. Default is 24.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        South Carolina:
        - https://dor.sc.gov/resources-site/lawandpolicy/Advisory%20Opinions/IL22-15.pdf
          https://dor.sc.gov/forms-site/Forms/IITPacket_2022.pdf
          https://www.scstatehouse.gov/code/t12c006.php
          https://dor.sc.gov/resources-site/lawandpolicy/Advisory%20Opinions/IL22-15.pdf
        
        West Virginia:
        -         
        - https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://tax.wv.gov/Documents/TSD/tsd438.pdf
          https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2011/Taxes-11-35att.html
        """
        
        # Check if the parameters are set upstream in the instance
        pay_periods = self.pay_periods if hasattr(self, 'pay_periods') else pay_periods
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        county_occupation = self.county_occupation if hasattr(self, 'county_occupation') else county_occupation
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'SouthCarolina', 'state_occupation': 'WestVirginia', "pay_periods": pay_periods, "capital_gains_short": capital_gains_short, "county_occupation": county_occupation, "dependents": dependents, "sources": sources, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def south_carolina_and_wisconsin(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        South Carolina:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Wisconsin:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        South Carolina:
        - https://dor.sc.gov/resources-site/lawandpolicy/Advisory%20Opinions/IL22-15.pdf
          https://dor.sc.gov/forms-site/Forms/IITPacket_2022.pdf
          https://www.scstatehouse.gov/code/t12c006.php
          https://dor.sc.gov/resources-site/lawandpolicy/Advisory%20Opinions/IL22-15.pdf
        
        Wisconsin:
        -         
        - https://www.revenue.wi.gov/TaxForms2023/2023-Form1-ES-Inst.pdf
          https://www.revenue.wi.gov/TaxForms2023/2023-Form1-ES-Inst.pdf
          https://www.revenue.wi.gov/TaxForms2023/2023-Form1-ES-Inst.pdf
          https://www.revenue.wi.gov/Pages/FAQS/pcs-work.aspx
          https://www.revenue.wi.gov/Dor%20Publications/pb103.pdf
          https://www.revenue.wi.gov/Pages/FAQS/pcs-work.aspx
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'SouthCarolina', 'state_occupation': 'Wisconsin', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def south_carolina_and_wyoming(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        South Carolina:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Wyoming:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        South Carolina:
        - https://dor.sc.gov/resources-site/lawandpolicy/Advisory%20Opinions/IL22-15.pdf
          https://dor.sc.gov/forms-site/Forms/IITPacket_2022.pdf
          https://www.scstatehouse.gov/code/t12c006.php
          https://dor.sc.gov/resources-site/lawandpolicy/Advisory%20Opinions/IL22-15.pdf
        
        Wyoming:
        -         
        - https://revenue.wyo.gov/
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'SouthCarolina', 'state_occupation': 'Wyoming', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def south_dakota_and_alabama(self, county_residence='Other', state_agi=75000, capital_gains_short=0, capital_gains_long=0, filing_status='single', income=100000, dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        South Dakota:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        Alabama:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        South Dakota:
        - https://dor.sd.gov/individuals/taxes/
        
        Alabama:
        -         
        - https://www.revenue.alabama.gov/ultraviewer/viewer/basic_viewer/index.html?form=2023/02/22f40.pdf
          https://www.revenue.alabama.gov/faqs/what-is-alabamas-individual-income-tax-rate/
          https://www.revenue.alabama.gov/ultraviewer/viewer/basic_viewer/index.html?form=2023/02/22f40.pdf
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-24.htm?taxmap=true
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-24.htm?taxmap=true
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-24.htm?taxmap=true
          https://www.revenue.alabama.gov/individual-corporate/income-to-be-reported-on-the-alabama-income-tax-return/
          https://www.almonline.org/TaxRates.aspx#OccupationalTax
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'SouthDakota', 'state_occupation': 'Alabama', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def south_dakota_and_alaska(self, sources=False):
        """
        --- Parameters ----
        
        South Dakota:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        Alaska:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        South Dakota:
        - https://dor.sd.gov/individuals/taxes/
        
        Alaska:
        -         
        - https://www.commerce.alaska.gov/web/dcra/officeofthestateassessor/alaskataxfacts.aspx
        """
        
        # Check if the parameters are set upstream in the instance
        sources = self.sources if hasattr(self, 'sources') else sources
        
        # State-specific parameters
        data = {
            'state_residence': 'SouthDakota', 'state_occupation': 'Alaska', "sources": sources
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def south_dakota_and_arizona(self, state_agi=75000, capital_gains_short=0, capital_gains_long=0, filing_status='single', income=100000, dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        South Dakota:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        Arizona:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        credits_state : int, optional
            Credits for state taxes. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        South Dakota:
        - https://dor.sd.gov/individuals/taxes/
        
        Arizona:
        -         
        - https://azdor.gov/forms/individual/form-140-x-y-tables
          https://azdor.gov/forms/individual/form-140-x-y-tables
          https://azdor.gov/forms/individual/form-140-x-y-tables
          https://azdor.gov/businesses-arizona/withholding-tax/withholding-exceptions
          "https://azdor.gov/individuals/income-tax-filing-assistance/identifying-other-taxable-income#:~:text=In%20regards%20to%20capital%20gains
         the%20individuals%20regular%20tax%20rate.
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'SouthDakota', 'state_occupation': 'Arizona', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def south_dakota_and_arkansas(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, capital_gains_long=0, filing_status='single', income=100000, dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        South Dakota:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        Arkansas:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        South Dakota:
        - https://dor.sd.gov/individuals/taxes/
        
        Arkansas:
        -         
        - https://www.arkansasedc.com/why-arkansas/business-climate/tax-structure/personal-income-tax
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_and_AR1000NR_Instructions.pdf
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_FullYearResidentIndividualIncomeTaxReturn.pdf
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_and_AR1000NR_Instructions.pdf
          https://www.arkansasedc.com/why-arkansas/business-climate/tax-structure/capital-gains-tax-reduction
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_and_AR1000NR_Instructions.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'SouthDakota', 'state_occupation': 'Arkansas', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def south_dakota_and_california(self, state_agi=75000, capital_gains_short=0, capital_gains_long=0, filing_status='single', income=100000, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        South Dakota:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        California:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        South Dakota:
        - https://dor.sd.gov/individuals/taxes/
        
        California:
        -         
        - https://www.ftb.ca.gov/forms/2022/2022-540-tax-rate-schedules.pdf
          https://www.ftb.ca.gov/file/personal/deductions/index.html
          https://www.ftb.ca.gov/about-ftb/newsroom/caleitc/eligibility-and-credit-information.html
          https://www.ftb.ca.gov/file/personal/income-types/capital-gains-and-losses.html#:~:text=California%20does%20not%20have%20a
         are%20taxed%20as%20ordinary%20income
          https://www.ftb.ca.gov/file/personal/
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'SouthDakota', 'state_occupation': 'California', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def south_dakota_and_colorado(self, county_residence='Other', state_agi=75000, capital_gains_short=0, capital_gains_long=0, income=100000, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        South Dakota:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        Colorado:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        South Dakota:
        - https://dor.sd.gov/individuals/taxes/
        
        Colorado:
        -         
        - https://leg.colorado.gov/agencies/legislative-council-staff/individual-income-tax%C2%A0
          https://leg.colorado.gov/agencies/legislative-council-staff/individual-income-tax%C2%A0
          https://tax.colorado.gov/sites/tax/files/documents/ITT_Capital_Gain_Subtraction_Feb_2022.pdf
          https://www.auroragov.org/business_services/taxes/occupational_privilege_tax
          https://www.denvergov.org/Government/Agencies-Departments-Offices/Agencies-Departments-Offices-Directory/Department-of-Finance/Our-Divisions/Treasury/Business-Tax-Information#:~:text=Occupational%20privilege%20tax
         -%28show%20below%29&text=Each%20taxable%20employee%20is%20liable
         month%20for%20each%20taxable%20employee
          https://www.glendale.co.us/355/Occupational-Privilege-Tax
          https://www.greenwoodvillage.com/1220/Occupational-Privilege-Tax-OPT
          https://ci.sheridan.co.us/288/Occupational-Privilege-Tax
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        income = self.income if hasattr(self, 'income') else income
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'SouthDakota', 'state_occupation': 'Colorado', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "income": income, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def south_dakota_and_connecticut(self, state_agi=75000, capital_gains_short=0, capital_gains_long=0, filing_status='single', income=100000, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        South Dakota:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        Connecticut:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        South Dakota:
        - https://dor.sd.gov/individuals/taxes/
        
        Connecticut:
        -         
        - https://portal.ct.gov/-/media/DRS/Forms/2022/Income/2022-CT-1040-Instructions_1222.pdf
          https://www.cga.ct.gov/2022/rpt/pdf/2022-R-0108.pdf
          https://www.cga.ct.gov/current/pub/chap_229.htm#sec_12-703
          https://portal.ct.gov/-/media/DRS/Forms/2022/Income/2022-CT-1040-Instructions_1222.pdf
          "https://portal.ct.gov/DRS/Publications/TSSNs/TSSN-29#:~:text=An%20individuals%20net%20capital%20gains
         portion%20of%20Social%20Security%20benefits"
          "https://portal.ct.gov/DRS/Publications/TSSNs/TSSN-29#:~:text=An%20individuals%20net%20capital%20gains
         portion%20of%20Social%20Security%20benefits"
          https://portal.ct.gov/DRS/Individuals/Resident-Income-Tax/Tax-Information
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'SouthDakota', 'state_occupation': 'Connecticut', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def south_dakota_and_delaware(self, capital_gains_long=0, federal_agi=75000, income=100000, sources=False, capital_gains_short=0, credits_state=0):
        """
        --- Parameters ----
        
        South Dakota:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        Delaware:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        federal_agi : int, optional
            Adjusted Gross Income (AGI) for federal taxes. Default is 0. Sources - https://www.irs.gov/e-file-providers/definition-of-adjusted-gross-income, https://www.irs.gov/newsroom/taxpayers-should-know-the-difference-between-standard-and-itemized-deductions
        income : int, optional
            Gross income. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        South Dakota:
        - https://dor.sd.gov/individuals/taxes/
        
        Delaware:
        -         
        - https://revenue.delaware.gov/software-developer/tax-rate-changes/
          https://revenuefiles.delaware.gov/2022/PIT-RES_TY22_2022-01_PaperInteractive.pdf
          https://revenuefiles.delaware.gov/2022/PIT-RES_TY22_2022-02_Instructions.pdf
          https://revenue.delaware.gov/frequently-asked-questions/personal-income-tax-faqs/
          https://www.newcastlede.gov/DocumentCenter/View/20147/City-of-Wilmington-Wage-Tax-Information-PDF
        """
        
        # Check if the parameters are set upstream in the instance
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        federal_agi = self.federal_agi if hasattr(self, 'federal_agi') else federal_agi
        income = self.income if hasattr(self, 'income') else income
        sources = self.sources if hasattr(self, 'sources') else sources
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'SouthDakota', 'state_occupation': 'Delaware', "capital_gains_long": capital_gains_long, "federal_agi": federal_agi, "income": income, "sources": sources, "capital_gains_short": capital_gains_short, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def south_dakota_and_florida(self, sources=False):
        """
        --- Parameters ----
        
        South Dakota:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        Florida:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        South Dakota:
        - https://dor.sd.gov/individuals/taxes/
        
        Florida:
        -         
        - https://floridarevenue.com/taxes/taxesfees/Pages/tax_interest_rates.aspx
        """
        
        # Check if the parameters are set upstream in the instance
        sources = self.sources if hasattr(self, 'sources') else sources
        
        # State-specific parameters
        data = {
            'state_residence': 'SouthDakota', 'state_occupation': 'Florida', "sources": sources
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def south_dakota_and_georgia(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, capital_gains_long=0, filing_status='single', income=100000, dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        South Dakota:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        Georgia:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        South Dakota:
        - https://dor.sd.gov/individuals/taxes/
        
        Georgia:
        -         
        - https://dor.georgia.gov/tax-tables-georgia-tax-rate-schedule
          https://apps.dor.ga.gov/FillableForms/PDFViewer/Index?form=2022GA500
          https://apps.dor.ga.gov/FillableForms/PDFViewer/Index?form=2022GA500
          https://apps.dor.ga.gov/FillableForms/PDFViewer/Index?form=2022GA500
          https://dor.georgia.gov/years-individual-income-tax-forms
          https://dor.georgia.gov/years-individual-income-tax-forms
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'SouthDakota', 'state_occupation': 'Georgia', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def south_dakota_and_hawaii(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, capital_gains_long=0, filing_status='single', income=100000, dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        South Dakota:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        Hawaii:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        South Dakota:
        - https://dor.sd.gov/individuals/taxes/
        
        Hawaii:
        -         
        - https://tax.hawaii.gov/forms/d_18table-on/d_18table-on_p13/
          https://files.hawaii.gov/tax/forms/2018/18table-on.pdf
          https://files.hawaii.gov/tax/news/pubs/22outline.pdf
          https://files.hawaii.gov/tax/news/pubs/FastTaxRefGuide_2017.pdf
          https://www.tfhawaii.org/wordpress/state-tax-resources/mini-tax-guides/individual-income-tax-chapter-235/#:~:text=Long%20term%20capital%20gains%20are%20taxed%20at%20a%20maximum%20of%207.25%25
          https://www.grassrootinstitute.org/2023/02/hb337-increasing-tax-on-capital-gains-a-bad-risk/
          https://files.hawaii.gov/tax/news/pubs/22outline.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'SouthDakota', 'state_occupation': 'Hawaii', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def south_dakota_and_idaho(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, capital_gains_long=0, filing_status='single', income=100000, dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        South Dakota:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        Idaho:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        South Dakota:
        - https://dor.sd.gov/individuals/taxes/
        
        Idaho:
        -         
        - https://tax.idaho.gov/taxes/income-tax/individual-income/individual-income-tax-rate-schedule/
          https://tax.idaho.gov/wp-content/uploads/forms/EIN00046/EIN00046_03-01-2023.pdf
          https://tax.idaho.gov/taxes/income-tax/individual-income/idaho-source-income/
          https://tax.idaho.gov/taxes/income-tax/individual-income/idaho-source-income/
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'SouthDakota', 'state_occupation': 'Idaho', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def south_dakota_and_illinois(self, state_agi=75000, capital_gains_short=0, capital_gains_long=0, filing_status='single', income=100000, dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        South Dakota:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        Illinois:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        South Dakota:
        - https://dor.sd.gov/individuals/taxes/
        
        Illinois:
        -         
        - https://tax.illinois.gov/research/taxrates/income.html#IndividualIncome
          https://tax.illinois.gov/content/dam/soi/en/web/tax/forms/incometax/documents/currentyear/individual/il-1040-instr.pdf
          https://tax.illinois.gov/questionsandanswers/12.html
          https://tax.illinois.gov/content/dam/soi/en/web/tax/forms/incometax/documents/currentyear/individual/il-1040-instr.pdf
          https://tax.illinois.gov/research/taxrates/personalproperty.html
          https://tax.illinois.gov/localgovernments/income.html
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'SouthDakota', 'state_occupation': 'Illinois', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def south_dakota_and_indiana(self, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_short=0, capital_gains_long=0, filing_status='single', income=100000, dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        South Dakota:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        Indiana:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        South Dakota:
        - https://dor.sd.gov/individuals/taxes/
        
        Indiana:
        -         
        - https://www.in.gov/dor/tax-forms/2022-individual-income-tax-forms/
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-32.htm
          https://www.in.gov/dor/files/ib28.pdf
          https://www.in.gov/dor/tax-forms/2022-individual-income-tax-forms/
          https://www.in.gov/dor/files/2022-county-tax-rates-and-codes.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'SouthDakota', 'state_occupation': 'Indiana', "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def south_dakota_and_iowa(self, county_residence='Other', state_agi=75000, capital_gains_short=0, capital_gains_long=0, filing_status='single', income=100000, dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        South Dakota:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        Iowa:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        South Dakota:
        - https://dor.sd.gov/individuals/taxes/
        
        Iowa:
        -         
        - https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
          https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
          https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
          https://tax.iowa.gov/iowa-illinois-reciprocal-agreement
          https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'SouthDakota', 'state_occupation': 'Iowa', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def south_dakota_and_kansas(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, capital_gains_long=0, filing_status='single', income=100000, dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        South Dakota:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        Kansas:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        South Dakota:
        - https://dor.sd.gov/individuals/taxes/
        
        Kansas:
        -         
        - https://www.ksrevenue.gov/taxrates.html
          https://www.ksrevenue.gov/incomebook22.html#0
          https://www.ksrevenue.gov/incomebook22.html#0
          https://www.ksrevenue.gov/incomebook22.html#0
          https://www.ksrevenue.gov/incomebook22.html
          https://www.ksrevenue.gov/incomebook22.html
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'SouthDakota', 'state_occupation': 'Kansas', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def south_dakota_and_kentucky(self, county_residence='Other', state_agi=75000, capital_gains_short=0, capital_gains_long=0, filing_status='single', income=100000, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        South Dakota:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        Kentucky:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        South Dakota:
        - https://dor.sd.gov/individuals/taxes/
        
        Kentucky:
        -         
        - https://revenue.ky.gov/News/Pages/Dor-Announces-Updates-to-Individual-Income-Tax-for-2023-Tax-Year.aspx
          https://revenue.ky.gov/Individual/Individual-Income-Tax/Pages/default.aspx
          https://revenue.ky.gov/News/Pages/Dor-Announces-Updates-to-Individual-Income-Tax-for-2023-Tax-Year.aspx
          https://revenue.ky.gov/Individual/Individual-Income-Tax/Pages/default.aspx
          https://revenue.ky.gov/Dor%20Training%20Materials/103%20KAR%2017.140.%20Individual%20income%20tax%20-%20reciprocity%20-%20nonresidents.pdf
          https://revenue.ky.gov/Individual/Individual-Income-Tax/Pages/default.aspx
          https://nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2021/Taxes-21-32.htm?taxmap=true
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'SouthDakota', 'state_occupation': 'Kentucky', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def south_dakota_and_louisiana(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, capital_gains_long=0, filing_status='single', income=100000, dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        South Dakota:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        Louisiana:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        South Dakota:
        - https://dor.sd.gov/individuals/taxes/
        
        Louisiana:
        -         
        - https://revenue.louisiana.gov/individualincometax
          https://revenue.louisiana.gov/individualincometax
          https://www.revenue.louisiana.gov/taxforms/6935(11_02)F.pdf
          https://revenue.louisiana.gov/individualincometax
          https://revenue.louisiana.gov/individualincometax
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'SouthDakota', 'state_occupation': 'Louisiana', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def south_dakota_and_maine(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, capital_gains_long=0, filing_status='single', income=100000, dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        South Dakota:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        Maine:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        South Dakota:
        - https://dor.sd.gov/individuals/taxes/
        
        Maine:
        -         
        - https://www.maine.gov/revenue/tax-return-forms/individual-income-tax-2022
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/ind_tax_rate_sched_2022.pdf
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/22_1040me_book_gen_instruc_revisedFeb23.pdf
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/22_1040me_book_gen_instruc_revisedFeb23.pdf
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/22_1040me_book_gen_instruc_revisedFeb23.pdf
          https://www.maine.gov/revenue/tax-return-forms/individual-income-tax-2022
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'SouthDakota', 'state_occupation': 'Maine', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def south_dakota_and_maryland(self, county_residence='Other', state_agi=75000, capital_gains_short=0, capital_gains_long=0, filing_status='single', income=100000, dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        South Dakota:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        Maryland:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        South Dakota:
        - https://dor.sd.gov/individuals/taxes/
        
        Maryland:
        -         
        - https://www.marylandtaxes.gov/individual/income/tax-info/tax-rates.php
          https://www.marylandtaxes.gov/forms/22_forms/Resident_Booklet.pdf
          https://www.marylandtaxes.gov/forms/Tax_Publications/Administrative_Releases/Income_and_Estate_Tax_Releases/ar_it37.pdf
          https://www.marylandtaxes.gov/individual/index.php
          https://www.marylandtaxes.gov/individual/income/tax-info/tax-rates.php
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'SouthDakota', 'state_occupation': 'Maryland', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def south_dakota_and_massachusetts(self, state_agi=75000, capital_gains_short=0, capital_gains_long=0, filing_status='single', income=100000, dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        South Dakota:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        Massachusetts:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        South Dakota:
        - https://dor.sd.gov/individuals/taxes/
        
        Massachusetts:
        -         
        - https://www.mass.gov/service-details/massachusetts-tax-rates
          https://www.mass.gov/service-details/view-massachusetts-personal-income-tax-exemptions
          https://www.mass.gov/service-details/view-massachusetts-personal-income-tax-exemptions
          https://www.mass.gov/service-details/view-child-and-dependent-related-credits
          https://www.mass.gov/service-details/massachusetts-tax-rates
          https://www.mass.gov/personal-income-tax
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'SouthDakota', 'state_occupation': 'Massachusetts', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def south_dakota_and_michigan(self, county_residence='Other', state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', county_occupation='Other', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        South Dakota:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        Michigan:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        county_occupation : str, optional
            County of occupation. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        South Dakota:
        - https://dor.sd.gov/individuals/taxes/
        
        Michigan:
        -         
        - https://www.michigan.gov/taxes/iit/new-developments-for-tax-year-2022
          "https://www.michigan.gov/taxes/-/media/Project/Websites/taxes/Forms/2023/2023-SUW/446_Withholding-Guide_2023.pdf?rev=177ca3d8cb034392bf84e317ed78c320&hash=19C5F42F43A89A91863D229349F48883#:~:text=The%20exemption%20amount%20is%20%245%2C400
         employees%20Michigan%20income%20tax%20return"
          "https://www.michigan.gov/taxes/-/media/Project/Websites/taxes/Forms/2023/2023-SUW/446_Withholding-Guide_2023.pdf?rev=177ca3d8cb034392bf84e317ed78c320&hash=19C5F42F43A89A91863D229349F48883#:~:text=The%20exemption%20amount%20is%20%245%2C400
         employees%20Michigan%20income%20tax%20return"
          https://www.michigan.gov/taxes/questions/iit/accordion/residency/are-my-wages-earned-in-another-state-taxable-in-michigan-if-i-am-a-michigan-resident-1
          https://www.michigan.gov/taxes/-/media/Project/Websites/taxes/Forms/2022/2022-IIT-Forms/BOOK_MI-1040.pdf?rev=4524f1cbbc0243b194ab05a6680be75b&hash=40DFE71E651F7C1D65C31F6CD03D944F
          https://www.michigan.gov/taxes/questions/iit/accordion/general/what-cities-impose-an-income-tax
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        county_occupation = self.county_occupation if hasattr(self, 'county_occupation') else county_occupation
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'SouthDakota', 'state_occupation': 'Michigan', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "county_occupation": county_occupation, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def south_dakota_and_minnesota(self, state_agi=75000, capital_gains_short=0, capital_gains_long=0, filing_status='single', income=100000, dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        South Dakota:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        Minnesota:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        South Dakota:
        - https://dor.sd.gov/individuals/taxes/
        
        Minnesota:
        -         
        - https://www.revenue.state.mn.us/minnesota-income-tax-rates-and-brackets
          https://www.revenue.state.mn.us/child-and-dependent-care-credit
          https://www.revenue.state.mn.us/dependent-exemptions
          https://www.revenue.state.mn.us/minnesota-standard-deduction
          https://mn.gov/mmb/assets/mwr_form_tcm1059-128581.pdf
          https://www.revenue.state.mn.us/individual-income-tax
          https://www.revenue.state.mn.us/individual-income-tax
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'SouthDakota', 'state_occupation': 'Minnesota', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def south_dakota_and_mississippi(self, state_agi=75000, capital_gains_short=0, capital_gains_long=0, filing_status='single', income=100000, dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        South Dakota:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        Mississippi:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        South Dakota:
        - https://dor.sd.gov/individuals/taxes/
        
        Mississippi:
        -         
        - https://www.dor.ms.gov/individual/tax-rates
          https://www.dor.ms.gov/individual/tax-rates
          https://www.dor.ms.gov/individual/tax-rates
          https://www.dor.ms.gov/individual
          https://www.dor.ms.gov/individual
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'SouthDakota', 'state_occupation': 'Mississippi', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def south_dakota_and_missouri(self, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_short=0, capital_gains_long=0, filing_status='single', income=100000, dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        South Dakota:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        Missouri:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        South Dakota:
        - https://dor.sd.gov/individuals/taxes/
        
        Missouri:
        -         
        - https://dor.mo.gov/taxation/individual/tax-types/income/year-changes/
          https://dor.mo.gov/taxation/individual/tax-types/income/year-changes/
          https://dor.mo.gov/taxation/individual/tax-types/income/year-changes/
          https://www.kcmo.gov/city-hall/departments/finance/earnings-tax
          https://www.stlouis-mo.gov/government/departments/collector/earnings-tax/individual-earnings-tax-info.cfm
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'SouthDakota', 'state_occupation': 'Missouri', "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def south_dakota_and_montana(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, capital_gains_long=0, filing_status='single', income=100000, dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        South Dakota:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        Montana:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        South Dakota:
        - https://dor.sd.gov/individuals/taxes/
        
        Montana:
        -         
        - https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/2022/12/montana-employees-withholding-allowance-and-exemption-certificate-form-mw-4.pdf
          https://mtrevenue.gov/2021/11/29/simplification-of-montana-income-taxation/?utm_source=rss&utm_medium=rss&utm_campaign=simplification-of-montana-income-taxation
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'SouthDakota', 'state_occupation': 'Montana', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def south_dakota_and_nebraska(self, state_agi=75000, capital_gains_short=0, capital_gains_long=0, filing_status='single', income=100000, dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        South Dakota:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        Nebraska:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        South Dakota:
        - https://dor.sd.gov/individuals/taxes/
        
        Nebraska:
        -         
        - https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2023_Circular_EN_8-429r2023_Whole_Book_11.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'SouthDakota', 'state_occupation': 'Nebraska', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def south_dakota_and_nevada(self, sources=False):
        """
        --- Parameters ----
        
        South Dakota:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        Nevada:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        South Dakota:
        - https://dor.sd.gov/individuals/taxes/
        
        Nevada:
        -         
        - https://www.nvsos.gov/sos/businesses/the-nevada-advantage
        """
        
        # Check if the parameters are set upstream in the instance
        sources = self.sources if hasattr(self, 'sources') else sources
        
        # State-specific parameters
        data = {
            'state_residence': 'SouthDakota', 'state_occupation': 'Nevada', "sources": sources
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def south_dakota_and_new_hampshire(self, sources=False):
        """
        --- Parameters ----
        
        South Dakota:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        New Hampshire:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        South Dakota:
        - https://dor.sd.gov/individuals/taxes/
        
        New Hampshire:
        -         
        - https://www.revenue.nh.gov/assistance/tax-overview.htm
        """
        
        # Check if the parameters are set upstream in the instance
        sources = self.sources if hasattr(self, 'sources') else sources
        
        # State-specific parameters
        data = {
            'state_residence': 'SouthDakota', 'state_occupation': 'NewHampshire', "sources": sources
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def south_dakota_and_new_jersey(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, capital_gains_long=0, filing_status='single', income=100000, dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        South Dakota:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        New Jersey:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        South Dakota:
        - https://dor.sd.gov/individuals/taxes/
        
        New Jersey:
        -         
        - https://www.state.nj.us/treasury/taxation/taxtables.shtml
          https://www.state.nj.us/treasury/taxation/njit13.shtml
          https://www.nj.gov/treasury/taxation/njit25.shtml
          https://www.state.nj.us/treasury/taxation/individuals/freqqiti.shtml
          https://www.state.nj.us/treasury/taxation/localtax.shtml
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'SouthDakota', 'state_occupation': 'NewJersey', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def south_dakota_and_new_mexico(self, state_agi=75000, capital_gains_short=0, capital_gains_long=0, filing_status='single', income=100000, dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        South Dakota:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        New Mexico:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        South Dakota:
        - https://dor.sd.gov/individuals/taxes/
        
        New Mexico:
        -         
        - https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/fdf3c548-8aba-4b9c-9eb4-bb564c716015/FYI-104.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/0bf943fd-652e-4400-bb1b-cd8397eb5a95/2022pit-adj.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/2f1a6781-9534-4436-b427-1557f9592099/2022pit-adj-ins.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/0bf943fd-652e-4400-bb1b-cd8397eb5a95/2022pit-adj.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/1afc56af-ea90-4d48-82e5-1f9aeb43255a/PITbook2022.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/fdf3c548-8aba-4b9c-9eb4-bb564c716015/FYI-104.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'SouthDakota', 'state_occupation': 'NewMexico', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def south_dakota_and_new_york(self, county_residence='Other', state_agi=75000, capital_gains_short=0, capital_gains_long=0, filing_status='single', income=100000, dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        South Dakota:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        New York:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        South Dakota:
        - https://dor.sd.gov/individuals/taxes/
        
        New York:
        -         
        - https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nys-tax-rate-schedule
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nys-tax-rate-schedule
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nys-tax-rate-schedule
          https://www.tax.ny.gov/pit/
          https://www.tax.ny.gov/pit/file/nonresident-faqs.htm
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#yonkers-tax-rate-schedule
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nyc-tax-rate-schedule
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'SouthDakota', 'state_occupation': 'NewYork', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def south_dakota_and_north_carolina(self, state_agi=75000, capital_gains_short=0, capital_gains_long=0, filing_status='single', income=100000, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        South Dakota:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        North Carolina:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        South Dakota:
        - https://dor.sd.gov/individuals/taxes/
        
        North Carolina:
        -         
        - https://www.ncdor.gov/taxes-forms/tax-rate-schedules
          https://www.ncdor.gov/taxes-forms/individual-income-tax/north-carolina-standard-deduction-or-north-carolina-itemized-deductions
          https://www.ncdor.gov/taxes-forms/individual-income-tax
          https://www.ncdor.gov/taxes-forms/individual-income-tax
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'SouthDakota', 'state_occupation': 'NorthCarolina', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def south_dakota_and_north_dakota(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, capital_gains_long=0, filing_status='single', income=100000, dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        South Dakota:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        North Dakota:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        South Dakota:
        - https://dor.sd.gov/individuals/taxes/
        
        North Dakota:
        -         
        - https://www.tax.nd.gov/forms
          https://www.tax.nd.gov/sites/www/files/documents/forms/individual/2022-iit/2022-individual-income-tax-booklet.pdf
          https://www.tax.nd.gov/sites/www/files/documents/forms/2022-individual-income-tax-booklet.pdf
          https://www.tax.nd.gov/tax-exemptions-credits/income-tax-exemptions-credits
          https://www.tax.nd.gov/sites/www/files/documents/forms/business/ndwrfillable.pdf
          https://www.tax.nd.gov/news/tax-legislative-changes/significant-changes-law/individual-income-tax-history
          https://www.tax.nd.gov/forms
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'SouthDakota', 'state_occupation': 'NorthDakota', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def south_dakota_and_ohio(self, county_residence='Other', state_agi=75000, capital_gains_short=0, capital_gains_long=0, filing_status='single', income=100000, dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        South Dakota:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        Ohio:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        South Dakota:
        - https://dor.sd.gov/individuals/taxes/
        
        Ohio:
        -         
        - https://tax.ohio.gov/individual/resources/annual-tax-rates
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://tax.ohio.gov/static/forms/employer_withholding/generic/wth_it4nr.pdf
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://www.ritaohio.com/TaxRatesTable
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'SouthDakota', 'state_occupation': 'Ohio', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def south_dakota_and_oklahoma(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, capital_gains_long=0, filing_status='single', income=100000, dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        South Dakota:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        Oklahoma:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        South Dakota:
        - https://dor.sd.gov/individuals/taxes/
        
        Oklahoma:
        -         
        - https://oklahoma.gov/content/dam/ok/en/tax/documents/resources/publications/businesses/withholding-tables/WHTables-2022.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'SouthDakota', 'state_occupation': 'Oklahoma', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def south_dakota_and_oregon(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, capital_gains_long=0, filing_status='single', income=100000, dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        South Dakota:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        Oregon:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        South Dakota:
        - https://dor.sd.gov/individuals/taxes/
        
        Oregon:
        -         
        - https://www.oregon.gov/dor/programs/individuals/Pages/PIT.aspx
          https://www.oregon.gov/dor/programs/individuals/Documents/Full%20year%20resident
         %20Form%20or-40%20filers.pdf
          https://www.oregon.gov/dor/programs/individuals/Documents/Part-year%20and%20nonresident
         %20Form%20or-40-P%20and%20or-40-N%20filers.pdf.pdf
          https://www.oregon.gov/dor/forms/FormsPubs/form-or-40_101-040_2022.pdf
          https://www.oregon.gov/dor/forms/FormsPubs/form-or-40_101-040_2022.pdf
          https://www.oregon.gov/dor/forms/FormsPubs/form-or-40_101-040_2022.pdf
          https://www.portland.gov/revenue/personal-tax
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'SouthDakota', 'state_occupation': 'Oregon', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def south_dakota_and_pennsylvania(self, county_residence='Other', state_agi=75000, county_occupation='Other', capital_gains_short=0, capital_gains_long=0, income=100000, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        South Dakota:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        Pennsylvania:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        county_residence : str, optional
            County of residence. Default is Other.
        county_occupation : str, optional
            County of occupation. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        South Dakota:
        - https://dor.sd.gov/individuals/taxes/
        
        Pennsylvania:
        -         
        - https://www.revenue.pa.gov/Tax%20Rates/Pages/default.aspx
          https://www.revenue.pa.gov/FormsandPublications/FormsforIndividuals/PIT/Documents/2022/2022_pa-40in.pdf
          https://www.revenue.pa.gov/FormsandPublications/FormsforBusinesses/EmployerWithholding/Documents/rev-419.pdf
          https://www.revenue.pa.gov/FormsandPublications/FormsforIndividuals/PIT/Documents/2022/2022_pa-40in.pdf
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-25.htm
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        county_occupation = self.county_occupation if hasattr(self, 'county_occupation') else county_occupation
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        income = self.income if hasattr(self, 'income') else income
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'SouthDakota', 'state_occupation': 'Pennsylvania', "county_residence": county_residence, "state_agi": state_agi, "county_occupation": county_occupation, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "income": income, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def south_dakota_and_rhode_island(self, state_agi=75000, capital_gains_short=0, capital_gains_long=0, filing_status='single', income=100000, dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        South Dakota:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        Rhode Island:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        South Dakota:
        - https://dor.sd.gov/individuals/taxes/
        
        Rhode Island:
        -         
        - https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2022_1040WE_w_0.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'SouthDakota', 'state_occupation': 'RhodeIsland', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def south_dakota_and_south_carolina(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, capital_gains_long=0, filing_status='single', income=100000, dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        South Dakota:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        South Carolina:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        South Dakota:
        - https://dor.sd.gov/individuals/taxes/
        
        South Carolina:
        -         
        - https://dor.sc.gov/resources-site/lawandpolicy/Advisory%20Opinions/IL22-15.pdf
          https://dor.sc.gov/forms-site/Forms/IITPacket_2022.pdf
          https://www.scstatehouse.gov/code/t12c006.php
          https://dor.sc.gov/resources-site/lawandpolicy/Advisory%20Opinions/IL22-15.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'SouthDakota', 'state_occupation': 'SouthCarolina', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def south_dakota_and_tennessee(self, sources=False):
        """
        --- Parameters ----
        
        South Dakota:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        Tennessee:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        South Dakota:
        - https://dor.sd.gov/individuals/taxes/
        
        Tennessee:
        -         
        - https://www.tn.gov/revenue/taxes.html
        """
        
        # Check if the parameters are set upstream in the instance
        sources = self.sources if hasattr(self, 'sources') else sources
        
        # State-specific parameters
        data = {
            'state_residence': 'SouthDakota', 'state_occupation': 'Tennessee', "sources": sources
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def south_dakota_and_texas(self, sources=False):
        """
        --- Parameters ----
        
        South Dakota:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        Texas:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        South Dakota:
        - https://dor.sd.gov/individuals/taxes/
        
        Texas:
        -         
        - https://comptroller.texas.gov/taxes/
        """
        
        # Check if the parameters are set upstream in the instance
        sources = self.sources if hasattr(self, 'sources') else sources
        
        # State-specific parameters
        data = {
            'state_residence': 'SouthDakota', 'state_occupation': 'Texas', "sources": sources
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def south_dakota_and_utah(self, state_agi=75000, capital_gains_short=0, capital_gains_long=0, filing_status='single', income=100000, dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        South Dakota:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        Utah:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        South Dakota:
        - https://dor.sd.gov/individuals/taxes/
        
        Utah:
        -         
        - https://incometax.utah.gov/paying/tax-rates
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'SouthDakota', 'state_occupation': 'Utah', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def south_dakota_and_vermont(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, capital_gains_long=0, filing_status='single', income=100000, dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        South Dakota:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        Vermont:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        South Dakota:
        - https://dor.sd.gov/individuals/taxes/
        
        Vermont:
        -         
        - https://tax.vermont.gov/sites/tax/files/documents/GB-1210-2023.pdf
          https://tax.vermont.gov/sites/tax/files/documents/Income%20Booklet-2022.pdf
          https://tax.vermont.gov/sites/tax/files/documents/Income%20Booklet-2022.pdf
          https://tax.vermont.gov/sites/tax/files/documents/Income%20Booklet-2022.pdf
          https://tax.vermont.gov/individuals/personal-income-tax/tax-credits
          https://tax.vermont.gov/individuals/personal-income-tax/taxable-income
          https://tax.vermont.gov/individuals/personal-income-tax/local-option-tax
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'SouthDakota', 'state_occupation': 'Vermont', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def south_dakota_and_virginia(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, capital_gains_long=0, filing_status='single', income=100000, dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        South Dakota:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        Virginia:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        South Dakota:
        - https://dor.sd.gov/individuals/taxes/
        
        Virginia:
        -         
        - https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/reciprocity
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'SouthDakota', 'state_occupation': 'Virginia', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def south_dakota_and_washington(self, capital_gains_long=0, sources=False, capital_gains_short=0):
        """
        --- Parameters ----
        
        South Dakota:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        Washington:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        South Dakota:
        - https://dor.sd.gov/individuals/taxes/
        
        Washington:
        -         
        - https://dor.wa.gov/taxes-rates/income-tax
          https://dor.wa.gov/taxes-rates/other-taxes/capital-gains-tax
          https://dor.wa.gov/taxes-rates/income-tax
        """
        
        # Check if the parameters are set upstream in the instance
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        sources = self.sources if hasattr(self, 'sources') else sources
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        
        # State-specific parameters
        data = {
            'state_residence': 'SouthDakota', 'state_occupation': 'Washington', "capital_gains_long": capital_gains_long, "sources": sources, "capital_gains_short": capital_gains_short
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def south_dakota_and_west_virginia(self, pay_periods=24, state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', county_occupation='Other', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        South Dakota:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        West Virginia:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_occupation : str, optional
            County of occupation. Default is Other.
        pay_periods : int, optional
            Number of pay periods. Options are 13, 24, 26, 52. Default is 24.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        South Dakota:
        - https://dor.sd.gov/individuals/taxes/
        
        West Virginia:
        -         
        - https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://tax.wv.gov/Documents/TSD/tsd438.pdf
          https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2011/Taxes-11-35att.html
        """
        
        # Check if the parameters are set upstream in the instance
        pay_periods = self.pay_periods if hasattr(self, 'pay_periods') else pay_periods
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        county_occupation = self.county_occupation if hasattr(self, 'county_occupation') else county_occupation
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'SouthDakota', 'state_occupation': 'WestVirginia', "pay_periods": pay_periods, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "county_occupation": county_occupation, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def south_dakota_and_wisconsin(self, state_agi=75000, capital_gains_short=0, capital_gains_long=0, filing_status='single', income=100000, dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        South Dakota:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        Wisconsin:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        South Dakota:
        - https://dor.sd.gov/individuals/taxes/
        
        Wisconsin:
        -         
        - https://www.revenue.wi.gov/TaxForms2023/2023-Form1-ES-Inst.pdf
          https://www.revenue.wi.gov/TaxForms2023/2023-Form1-ES-Inst.pdf
          https://www.revenue.wi.gov/TaxForms2023/2023-Form1-ES-Inst.pdf
          https://www.revenue.wi.gov/Pages/FAQS/pcs-work.aspx
          https://www.revenue.wi.gov/Dor%20Publications/pb103.pdf
          https://www.revenue.wi.gov/Pages/FAQS/pcs-work.aspx
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'SouthDakota', 'state_occupation': 'Wisconsin', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def south_dakota_and_wyoming(self, sources=False):
        """
        --- Parameters ----
        
        South Dakota:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        Wyoming:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        South Dakota:
        - https://dor.sd.gov/individuals/taxes/
        
        Wyoming:
        -         
        - https://revenue.wyo.gov/
        """
        
        # Check if the parameters are set upstream in the instance
        sources = self.sources if hasattr(self, 'sources') else sources
        
        # State-specific parameters
        data = {
            'state_residence': 'SouthDakota', 'state_occupation': 'Wyoming', "sources": sources
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def tennessee_and_alabama(self, county_residence='Other', state_agi=75000, capital_gains_short=0, capital_gains_long=0, filing_status='single', income=100000, dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Tennessee:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        Alabama:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Tennessee:
        - https://www.tn.gov/revenue/taxes.html
        
        Alabama:
        -         
        - https://www.revenue.alabama.gov/ultraviewer/viewer/basic_viewer/index.html?form=2023/02/22f40.pdf
          https://www.revenue.alabama.gov/faqs/what-is-alabamas-individual-income-tax-rate/
          https://www.revenue.alabama.gov/ultraviewer/viewer/basic_viewer/index.html?form=2023/02/22f40.pdf
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-24.htm?taxmap=true
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-24.htm?taxmap=true
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-24.htm?taxmap=true
          https://www.revenue.alabama.gov/individual-corporate/income-to-be-reported-on-the-alabama-income-tax-return/
          https://www.almonline.org/TaxRates.aspx#OccupationalTax
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Tennessee', 'state_occupation': 'Alabama', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def tennessee_and_alaska(self, sources=False):
        """
        --- Parameters ----
        
        Tennessee:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        Alaska:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Tennessee:
        - https://www.tn.gov/revenue/taxes.html
        
        Alaska:
        -         
        - https://www.commerce.alaska.gov/web/dcra/officeofthestateassessor/alaskataxfacts.aspx
        """
        
        # Check if the parameters are set upstream in the instance
        sources = self.sources if hasattr(self, 'sources') else sources
        
        # State-specific parameters
        data = {
            'state_residence': 'Tennessee', 'state_occupation': 'Alaska', "sources": sources
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def tennessee_and_arizona(self, state_agi=75000, capital_gains_short=0, capital_gains_long=0, filing_status='single', income=100000, dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Tennessee:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        Arizona:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        credits_state : int, optional
            Credits for state taxes. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Tennessee:
        - https://www.tn.gov/revenue/taxes.html
        
        Arizona:
        -         
        - https://azdor.gov/forms/individual/form-140-x-y-tables
          https://azdor.gov/forms/individual/form-140-x-y-tables
          https://azdor.gov/forms/individual/form-140-x-y-tables
          https://azdor.gov/businesses-arizona/withholding-tax/withholding-exceptions
          "https://azdor.gov/individuals/income-tax-filing-assistance/identifying-other-taxable-income#:~:text=In%20regards%20to%20capital%20gains
         the%20individuals%20regular%20tax%20rate.
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Tennessee', 'state_occupation': 'Arizona', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def tennessee_and_arkansas(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, capital_gains_long=0, filing_status='single', income=100000, dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Tennessee:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        Arkansas:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Tennessee:
        - https://www.tn.gov/revenue/taxes.html
        
        Arkansas:
        -         
        - https://www.arkansasedc.com/why-arkansas/business-climate/tax-structure/personal-income-tax
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_and_AR1000NR_Instructions.pdf
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_FullYearResidentIndividualIncomeTaxReturn.pdf
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_and_AR1000NR_Instructions.pdf
          https://www.arkansasedc.com/why-arkansas/business-climate/tax-structure/capital-gains-tax-reduction
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_and_AR1000NR_Instructions.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Tennessee', 'state_occupation': 'Arkansas', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def tennessee_and_california(self, state_agi=75000, capital_gains_short=0, capital_gains_long=0, filing_status='single', income=100000, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Tennessee:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        California:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Tennessee:
        - https://www.tn.gov/revenue/taxes.html
        
        California:
        -         
        - https://www.ftb.ca.gov/forms/2022/2022-540-tax-rate-schedules.pdf
          https://www.ftb.ca.gov/file/personal/deductions/index.html
          https://www.ftb.ca.gov/about-ftb/newsroom/caleitc/eligibility-and-credit-information.html
          https://www.ftb.ca.gov/file/personal/income-types/capital-gains-and-losses.html#:~:text=California%20does%20not%20have%20a
         are%20taxed%20as%20ordinary%20income
          https://www.ftb.ca.gov/file/personal/
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Tennessee', 'state_occupation': 'California', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def tennessee_and_colorado(self, county_residence='Other', state_agi=75000, capital_gains_short=0, capital_gains_long=0, income=100000, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Tennessee:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        Colorado:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Tennessee:
        - https://www.tn.gov/revenue/taxes.html
        
        Colorado:
        -         
        - https://leg.colorado.gov/agencies/legislative-council-staff/individual-income-tax%C2%A0
          https://leg.colorado.gov/agencies/legislative-council-staff/individual-income-tax%C2%A0
          https://tax.colorado.gov/sites/tax/files/documents/ITT_Capital_Gain_Subtraction_Feb_2022.pdf
          https://www.auroragov.org/business_services/taxes/occupational_privilege_tax
          https://www.denvergov.org/Government/Agencies-Departments-Offices/Agencies-Departments-Offices-Directory/Department-of-Finance/Our-Divisions/Treasury/Business-Tax-Information#:~:text=Occupational%20privilege%20tax
         -%28show%20below%29&text=Each%20taxable%20employee%20is%20liable
         month%20for%20each%20taxable%20employee
          https://www.glendale.co.us/355/Occupational-Privilege-Tax
          https://www.greenwoodvillage.com/1220/Occupational-Privilege-Tax-OPT
          https://ci.sheridan.co.us/288/Occupational-Privilege-Tax
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        income = self.income if hasattr(self, 'income') else income
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Tennessee', 'state_occupation': 'Colorado', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "income": income, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def tennessee_and_connecticut(self, state_agi=75000, capital_gains_short=0, capital_gains_long=0, filing_status='single', income=100000, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Tennessee:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        Connecticut:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Tennessee:
        - https://www.tn.gov/revenue/taxes.html
        
        Connecticut:
        -         
        - https://portal.ct.gov/-/media/DRS/Forms/2022/Income/2022-CT-1040-Instructions_1222.pdf
          https://www.cga.ct.gov/2022/rpt/pdf/2022-R-0108.pdf
          https://www.cga.ct.gov/current/pub/chap_229.htm#sec_12-703
          https://portal.ct.gov/-/media/DRS/Forms/2022/Income/2022-CT-1040-Instructions_1222.pdf
          "https://portal.ct.gov/DRS/Publications/TSSNs/TSSN-29#:~:text=An%20individuals%20net%20capital%20gains
         portion%20of%20Social%20Security%20benefits"
          "https://portal.ct.gov/DRS/Publications/TSSNs/TSSN-29#:~:text=An%20individuals%20net%20capital%20gains
         portion%20of%20Social%20Security%20benefits"
          https://portal.ct.gov/DRS/Individuals/Resident-Income-Tax/Tax-Information
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Tennessee', 'state_occupation': 'Connecticut', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def tennessee_and_delaware(self, capital_gains_long=0, federal_agi=75000, income=100000, sources=False, capital_gains_short=0, credits_state=0):
        """
        --- Parameters ----
        
        Tennessee:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        Delaware:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        federal_agi : int, optional
            Adjusted Gross Income (AGI) for federal taxes. Default is 0. Sources - https://www.irs.gov/e-file-providers/definition-of-adjusted-gross-income, https://www.irs.gov/newsroom/taxpayers-should-know-the-difference-between-standard-and-itemized-deductions
        income : int, optional
            Gross income. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Tennessee:
        - https://www.tn.gov/revenue/taxes.html
        
        Delaware:
        -         
        - https://revenue.delaware.gov/software-developer/tax-rate-changes/
          https://revenuefiles.delaware.gov/2022/PIT-RES_TY22_2022-01_PaperInteractive.pdf
          https://revenuefiles.delaware.gov/2022/PIT-RES_TY22_2022-02_Instructions.pdf
          https://revenue.delaware.gov/frequently-asked-questions/personal-income-tax-faqs/
          https://www.newcastlede.gov/DocumentCenter/View/20147/City-of-Wilmington-Wage-Tax-Information-PDF
        """
        
        # Check if the parameters are set upstream in the instance
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        federal_agi = self.federal_agi if hasattr(self, 'federal_agi') else federal_agi
        income = self.income if hasattr(self, 'income') else income
        sources = self.sources if hasattr(self, 'sources') else sources
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Tennessee', 'state_occupation': 'Delaware', "capital_gains_long": capital_gains_long, "federal_agi": federal_agi, "income": income, "sources": sources, "capital_gains_short": capital_gains_short, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def tennessee_and_florida(self, sources=False):
        """
        --- Parameters ----
        
        Tennessee:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        Florida:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Tennessee:
        - https://www.tn.gov/revenue/taxes.html
        
        Florida:
        -         
        - https://floridarevenue.com/taxes/taxesfees/Pages/tax_interest_rates.aspx
        """
        
        # Check if the parameters are set upstream in the instance
        sources = self.sources if hasattr(self, 'sources') else sources
        
        # State-specific parameters
        data = {
            'state_residence': 'Tennessee', 'state_occupation': 'Florida', "sources": sources
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def tennessee_and_georgia(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, capital_gains_long=0, filing_status='single', income=100000, dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Tennessee:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        Georgia:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Tennessee:
        - https://www.tn.gov/revenue/taxes.html
        
        Georgia:
        -         
        - https://dor.georgia.gov/tax-tables-georgia-tax-rate-schedule
          https://apps.dor.ga.gov/FillableForms/PDFViewer/Index?form=2022GA500
          https://apps.dor.ga.gov/FillableForms/PDFViewer/Index?form=2022GA500
          https://apps.dor.ga.gov/FillableForms/PDFViewer/Index?form=2022GA500
          https://dor.georgia.gov/years-individual-income-tax-forms
          https://dor.georgia.gov/years-individual-income-tax-forms
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Tennessee', 'state_occupation': 'Georgia', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def tennessee_and_hawaii(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, capital_gains_long=0, filing_status='single', income=100000, dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Tennessee:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        Hawaii:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Tennessee:
        - https://www.tn.gov/revenue/taxes.html
        
        Hawaii:
        -         
        - https://tax.hawaii.gov/forms/d_18table-on/d_18table-on_p13/
          https://files.hawaii.gov/tax/forms/2018/18table-on.pdf
          https://files.hawaii.gov/tax/news/pubs/22outline.pdf
          https://files.hawaii.gov/tax/news/pubs/FastTaxRefGuide_2017.pdf
          https://www.tfhawaii.org/wordpress/state-tax-resources/mini-tax-guides/individual-income-tax-chapter-235/#:~:text=Long%20term%20capital%20gains%20are%20taxed%20at%20a%20maximum%20of%207.25%25
          https://www.grassrootinstitute.org/2023/02/hb337-increasing-tax-on-capital-gains-a-bad-risk/
          https://files.hawaii.gov/tax/news/pubs/22outline.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Tennessee', 'state_occupation': 'Hawaii', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def tennessee_and_idaho(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, capital_gains_long=0, filing_status='single', income=100000, dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Tennessee:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        Idaho:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Tennessee:
        - https://www.tn.gov/revenue/taxes.html
        
        Idaho:
        -         
        - https://tax.idaho.gov/taxes/income-tax/individual-income/individual-income-tax-rate-schedule/
          https://tax.idaho.gov/wp-content/uploads/forms/EIN00046/EIN00046_03-01-2023.pdf
          https://tax.idaho.gov/taxes/income-tax/individual-income/idaho-source-income/
          https://tax.idaho.gov/taxes/income-tax/individual-income/idaho-source-income/
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Tennessee', 'state_occupation': 'Idaho', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def tennessee_and_illinois(self, state_agi=75000, capital_gains_short=0, capital_gains_long=0, filing_status='single', income=100000, dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Tennessee:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        Illinois:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Tennessee:
        - https://www.tn.gov/revenue/taxes.html
        
        Illinois:
        -         
        - https://tax.illinois.gov/research/taxrates/income.html#IndividualIncome
          https://tax.illinois.gov/content/dam/soi/en/web/tax/forms/incometax/documents/currentyear/individual/il-1040-instr.pdf
          https://tax.illinois.gov/questionsandanswers/12.html
          https://tax.illinois.gov/content/dam/soi/en/web/tax/forms/incometax/documents/currentyear/individual/il-1040-instr.pdf
          https://tax.illinois.gov/research/taxrates/personalproperty.html
          https://tax.illinois.gov/localgovernments/income.html
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Tennessee', 'state_occupation': 'Illinois', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def tennessee_and_indiana(self, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_short=0, capital_gains_long=0, filing_status='single', income=100000, dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Tennessee:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        Indiana:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Tennessee:
        - https://www.tn.gov/revenue/taxes.html
        
        Indiana:
        -         
        - https://www.in.gov/dor/tax-forms/2022-individual-income-tax-forms/
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-32.htm
          https://www.in.gov/dor/files/ib28.pdf
          https://www.in.gov/dor/tax-forms/2022-individual-income-tax-forms/
          https://www.in.gov/dor/files/2022-county-tax-rates-and-codes.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Tennessee', 'state_occupation': 'Indiana', "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def tennessee_and_iowa(self, county_residence='Other', state_agi=75000, capital_gains_short=0, capital_gains_long=0, filing_status='single', income=100000, dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Tennessee:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        Iowa:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Tennessee:
        - https://www.tn.gov/revenue/taxes.html
        
        Iowa:
        -         
        - https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
          https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
          https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
          https://tax.iowa.gov/iowa-illinois-reciprocal-agreement
          https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Tennessee', 'state_occupation': 'Iowa', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def tennessee_and_kansas(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, capital_gains_long=0, filing_status='single', income=100000, dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Tennessee:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        Kansas:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Tennessee:
        - https://www.tn.gov/revenue/taxes.html
        
        Kansas:
        -         
        - https://www.ksrevenue.gov/taxrates.html
          https://www.ksrevenue.gov/incomebook22.html#0
          https://www.ksrevenue.gov/incomebook22.html#0
          https://www.ksrevenue.gov/incomebook22.html#0
          https://www.ksrevenue.gov/incomebook22.html
          https://www.ksrevenue.gov/incomebook22.html
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Tennessee', 'state_occupation': 'Kansas', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def tennessee_and_kentucky(self, county_residence='Other', state_agi=75000, capital_gains_short=0, capital_gains_long=0, filing_status='single', income=100000, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Tennessee:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        Kentucky:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Tennessee:
        - https://www.tn.gov/revenue/taxes.html
        
        Kentucky:
        -         
        - https://revenue.ky.gov/News/Pages/Dor-Announces-Updates-to-Individual-Income-Tax-for-2023-Tax-Year.aspx
          https://revenue.ky.gov/Individual/Individual-Income-Tax/Pages/default.aspx
          https://revenue.ky.gov/News/Pages/Dor-Announces-Updates-to-Individual-Income-Tax-for-2023-Tax-Year.aspx
          https://revenue.ky.gov/Individual/Individual-Income-Tax/Pages/default.aspx
          https://revenue.ky.gov/Dor%20Training%20Materials/103%20KAR%2017.140.%20Individual%20income%20tax%20-%20reciprocity%20-%20nonresidents.pdf
          https://revenue.ky.gov/Individual/Individual-Income-Tax/Pages/default.aspx
          https://nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2021/Taxes-21-32.htm?taxmap=true
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Tennessee', 'state_occupation': 'Kentucky', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def tennessee_and_louisiana(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, capital_gains_long=0, filing_status='single', income=100000, dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Tennessee:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        Louisiana:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Tennessee:
        - https://www.tn.gov/revenue/taxes.html
        
        Louisiana:
        -         
        - https://revenue.louisiana.gov/individualincometax
          https://revenue.louisiana.gov/individualincometax
          https://www.revenue.louisiana.gov/taxforms/6935(11_02)F.pdf
          https://revenue.louisiana.gov/individualincometax
          https://revenue.louisiana.gov/individualincometax
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Tennessee', 'state_occupation': 'Louisiana', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def tennessee_and_maine(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, capital_gains_long=0, filing_status='single', income=100000, dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Tennessee:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        Maine:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Tennessee:
        - https://www.tn.gov/revenue/taxes.html
        
        Maine:
        -         
        - https://www.maine.gov/revenue/tax-return-forms/individual-income-tax-2022
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/ind_tax_rate_sched_2022.pdf
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/22_1040me_book_gen_instruc_revisedFeb23.pdf
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/22_1040me_book_gen_instruc_revisedFeb23.pdf
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/22_1040me_book_gen_instruc_revisedFeb23.pdf
          https://www.maine.gov/revenue/tax-return-forms/individual-income-tax-2022
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Tennessee', 'state_occupation': 'Maine', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def tennessee_and_maryland(self, county_residence='Other', state_agi=75000, capital_gains_short=0, capital_gains_long=0, filing_status='single', income=100000, dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Tennessee:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        Maryland:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Tennessee:
        - https://www.tn.gov/revenue/taxes.html
        
        Maryland:
        -         
        - https://www.marylandtaxes.gov/individual/income/tax-info/tax-rates.php
          https://www.marylandtaxes.gov/forms/22_forms/Resident_Booklet.pdf
          https://www.marylandtaxes.gov/forms/Tax_Publications/Administrative_Releases/Income_and_Estate_Tax_Releases/ar_it37.pdf
          https://www.marylandtaxes.gov/individual/index.php
          https://www.marylandtaxes.gov/individual/income/tax-info/tax-rates.php
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Tennessee', 'state_occupation': 'Maryland', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def tennessee_and_massachusetts(self, state_agi=75000, capital_gains_short=0, capital_gains_long=0, filing_status='single', income=100000, dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Tennessee:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        Massachusetts:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Tennessee:
        - https://www.tn.gov/revenue/taxes.html
        
        Massachusetts:
        -         
        - https://www.mass.gov/service-details/massachusetts-tax-rates
          https://www.mass.gov/service-details/view-massachusetts-personal-income-tax-exemptions
          https://www.mass.gov/service-details/view-massachusetts-personal-income-tax-exemptions
          https://www.mass.gov/service-details/view-child-and-dependent-related-credits
          https://www.mass.gov/service-details/massachusetts-tax-rates
          https://www.mass.gov/personal-income-tax
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Tennessee', 'state_occupation': 'Massachusetts', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def tennessee_and_michigan(self, county_residence='Other', state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', county_occupation='Other', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Tennessee:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        Michigan:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        county_occupation : str, optional
            County of occupation. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Tennessee:
        - https://www.tn.gov/revenue/taxes.html
        
        Michigan:
        -         
        - https://www.michigan.gov/taxes/iit/new-developments-for-tax-year-2022
          "https://www.michigan.gov/taxes/-/media/Project/Websites/taxes/Forms/2023/2023-SUW/446_Withholding-Guide_2023.pdf?rev=177ca3d8cb034392bf84e317ed78c320&hash=19C5F42F43A89A91863D229349F48883#:~:text=The%20exemption%20amount%20is%20%245%2C400
         employees%20Michigan%20income%20tax%20return"
          "https://www.michigan.gov/taxes/-/media/Project/Websites/taxes/Forms/2023/2023-SUW/446_Withholding-Guide_2023.pdf?rev=177ca3d8cb034392bf84e317ed78c320&hash=19C5F42F43A89A91863D229349F48883#:~:text=The%20exemption%20amount%20is%20%245%2C400
         employees%20Michigan%20income%20tax%20return"
          https://www.michigan.gov/taxes/questions/iit/accordion/residency/are-my-wages-earned-in-another-state-taxable-in-michigan-if-i-am-a-michigan-resident-1
          https://www.michigan.gov/taxes/-/media/Project/Websites/taxes/Forms/2022/2022-IIT-Forms/BOOK_MI-1040.pdf?rev=4524f1cbbc0243b194ab05a6680be75b&hash=40DFE71E651F7C1D65C31F6CD03D944F
          https://www.michigan.gov/taxes/questions/iit/accordion/general/what-cities-impose-an-income-tax
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        county_occupation = self.county_occupation if hasattr(self, 'county_occupation') else county_occupation
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Tennessee', 'state_occupation': 'Michigan', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "county_occupation": county_occupation, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def tennessee_and_minnesota(self, state_agi=75000, capital_gains_short=0, capital_gains_long=0, filing_status='single', income=100000, dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Tennessee:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        Minnesota:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Tennessee:
        - https://www.tn.gov/revenue/taxes.html
        
        Minnesota:
        -         
        - https://www.revenue.state.mn.us/minnesota-income-tax-rates-and-brackets
          https://www.revenue.state.mn.us/child-and-dependent-care-credit
          https://www.revenue.state.mn.us/dependent-exemptions
          https://www.revenue.state.mn.us/minnesota-standard-deduction
          https://mn.gov/mmb/assets/mwr_form_tcm1059-128581.pdf
          https://www.revenue.state.mn.us/individual-income-tax
          https://www.revenue.state.mn.us/individual-income-tax
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Tennessee', 'state_occupation': 'Minnesota', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def tennessee_and_mississippi(self, state_agi=75000, capital_gains_short=0, capital_gains_long=0, filing_status='single', income=100000, dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Tennessee:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        Mississippi:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Tennessee:
        - https://www.tn.gov/revenue/taxes.html
        
        Mississippi:
        -         
        - https://www.dor.ms.gov/individual/tax-rates
          https://www.dor.ms.gov/individual/tax-rates
          https://www.dor.ms.gov/individual/tax-rates
          https://www.dor.ms.gov/individual
          https://www.dor.ms.gov/individual
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Tennessee', 'state_occupation': 'Mississippi', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def tennessee_and_missouri(self, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_short=0, capital_gains_long=0, filing_status='single', income=100000, dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Tennessee:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        Missouri:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Tennessee:
        - https://www.tn.gov/revenue/taxes.html
        
        Missouri:
        -         
        - https://dor.mo.gov/taxation/individual/tax-types/income/year-changes/
          https://dor.mo.gov/taxation/individual/tax-types/income/year-changes/
          https://dor.mo.gov/taxation/individual/tax-types/income/year-changes/
          https://www.kcmo.gov/city-hall/departments/finance/earnings-tax
          https://www.stlouis-mo.gov/government/departments/collector/earnings-tax/individual-earnings-tax-info.cfm
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Tennessee', 'state_occupation': 'Missouri', "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def tennessee_and_montana(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, capital_gains_long=0, filing_status='single', income=100000, dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Tennessee:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        Montana:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Tennessee:
        - https://www.tn.gov/revenue/taxes.html
        
        Montana:
        -         
        - https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/2022/12/montana-employees-withholding-allowance-and-exemption-certificate-form-mw-4.pdf
          https://mtrevenue.gov/2021/11/29/simplification-of-montana-income-taxation/?utm_source=rss&utm_medium=rss&utm_campaign=simplification-of-montana-income-taxation
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Tennessee', 'state_occupation': 'Montana', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def tennessee_and_nebraska(self, state_agi=75000, capital_gains_short=0, capital_gains_long=0, filing_status='single', income=100000, dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Tennessee:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        Nebraska:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Tennessee:
        - https://www.tn.gov/revenue/taxes.html
        
        Nebraska:
        -         
        - https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2023_Circular_EN_8-429r2023_Whole_Book_11.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Tennessee', 'state_occupation': 'Nebraska', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def tennessee_and_nevada(self, sources=False):
        """
        --- Parameters ----
        
        Tennessee:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        Nevada:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Tennessee:
        - https://www.tn.gov/revenue/taxes.html
        
        Nevada:
        -         
        - https://www.nvsos.gov/sos/businesses/the-nevada-advantage
        """
        
        # Check if the parameters are set upstream in the instance
        sources = self.sources if hasattr(self, 'sources') else sources
        
        # State-specific parameters
        data = {
            'state_residence': 'Tennessee', 'state_occupation': 'Nevada', "sources": sources
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def tennessee_and_new_hampshire(self, sources=False):
        """
        --- Parameters ----
        
        Tennessee:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        New Hampshire:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Tennessee:
        - https://www.tn.gov/revenue/taxes.html
        
        New Hampshire:
        -         
        - https://www.revenue.nh.gov/assistance/tax-overview.htm
        """
        
        # Check if the parameters are set upstream in the instance
        sources = self.sources if hasattr(self, 'sources') else sources
        
        # State-specific parameters
        data = {
            'state_residence': 'Tennessee', 'state_occupation': 'NewHampshire', "sources": sources
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def tennessee_and_new_jersey(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, capital_gains_long=0, filing_status='single', income=100000, dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Tennessee:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        New Jersey:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Tennessee:
        - https://www.tn.gov/revenue/taxes.html
        
        New Jersey:
        -         
        - https://www.state.nj.us/treasury/taxation/taxtables.shtml
          https://www.state.nj.us/treasury/taxation/njit13.shtml
          https://www.nj.gov/treasury/taxation/njit25.shtml
          https://www.state.nj.us/treasury/taxation/individuals/freqqiti.shtml
          https://www.state.nj.us/treasury/taxation/localtax.shtml
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Tennessee', 'state_occupation': 'NewJersey', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def tennessee_and_new_mexico(self, state_agi=75000, capital_gains_short=0, capital_gains_long=0, filing_status='single', income=100000, dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Tennessee:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        New Mexico:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Tennessee:
        - https://www.tn.gov/revenue/taxes.html
        
        New Mexico:
        -         
        - https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/fdf3c548-8aba-4b9c-9eb4-bb564c716015/FYI-104.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/0bf943fd-652e-4400-bb1b-cd8397eb5a95/2022pit-adj.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/2f1a6781-9534-4436-b427-1557f9592099/2022pit-adj-ins.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/0bf943fd-652e-4400-bb1b-cd8397eb5a95/2022pit-adj.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/1afc56af-ea90-4d48-82e5-1f9aeb43255a/PITbook2022.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/fdf3c548-8aba-4b9c-9eb4-bb564c716015/FYI-104.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Tennessee', 'state_occupation': 'NewMexico', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def tennessee_and_new_york(self, county_residence='Other', state_agi=75000, capital_gains_short=0, capital_gains_long=0, filing_status='single', income=100000, dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Tennessee:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        New York:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Tennessee:
        - https://www.tn.gov/revenue/taxes.html
        
        New York:
        -         
        - https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nys-tax-rate-schedule
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nys-tax-rate-schedule
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nys-tax-rate-schedule
          https://www.tax.ny.gov/pit/
          https://www.tax.ny.gov/pit/file/nonresident-faqs.htm
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#yonkers-tax-rate-schedule
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nyc-tax-rate-schedule
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Tennessee', 'state_occupation': 'NewYork', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def tennessee_and_north_carolina(self, state_agi=75000, capital_gains_short=0, capital_gains_long=0, filing_status='single', income=100000, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Tennessee:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        North Carolina:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Tennessee:
        - https://www.tn.gov/revenue/taxes.html
        
        North Carolina:
        -         
        - https://www.ncdor.gov/taxes-forms/tax-rate-schedules
          https://www.ncdor.gov/taxes-forms/individual-income-tax/north-carolina-standard-deduction-or-north-carolina-itemized-deductions
          https://www.ncdor.gov/taxes-forms/individual-income-tax
          https://www.ncdor.gov/taxes-forms/individual-income-tax
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Tennessee', 'state_occupation': 'NorthCarolina', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def tennessee_and_north_dakota(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, capital_gains_long=0, filing_status='single', income=100000, dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Tennessee:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        North Dakota:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Tennessee:
        - https://www.tn.gov/revenue/taxes.html
        
        North Dakota:
        -         
        - https://www.tax.nd.gov/forms
          https://www.tax.nd.gov/sites/www/files/documents/forms/individual/2022-iit/2022-individual-income-tax-booklet.pdf
          https://www.tax.nd.gov/sites/www/files/documents/forms/2022-individual-income-tax-booklet.pdf
          https://www.tax.nd.gov/tax-exemptions-credits/income-tax-exemptions-credits
          https://www.tax.nd.gov/sites/www/files/documents/forms/business/ndwrfillable.pdf
          https://www.tax.nd.gov/news/tax-legislative-changes/significant-changes-law/individual-income-tax-history
          https://www.tax.nd.gov/forms
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Tennessee', 'state_occupation': 'NorthDakota', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def tennessee_and_ohio(self, county_residence='Other', state_agi=75000, capital_gains_short=0, capital_gains_long=0, filing_status='single', income=100000, dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Tennessee:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        Ohio:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Tennessee:
        - https://www.tn.gov/revenue/taxes.html
        
        Ohio:
        -         
        - https://tax.ohio.gov/individual/resources/annual-tax-rates
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://tax.ohio.gov/static/forms/employer_withholding/generic/wth_it4nr.pdf
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://www.ritaohio.com/TaxRatesTable
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Tennessee', 'state_occupation': 'Ohio', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def tennessee_and_oklahoma(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, capital_gains_long=0, filing_status='single', income=100000, dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Tennessee:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        Oklahoma:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Tennessee:
        - https://www.tn.gov/revenue/taxes.html
        
        Oklahoma:
        -         
        - https://oklahoma.gov/content/dam/ok/en/tax/documents/resources/publications/businesses/withholding-tables/WHTables-2022.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Tennessee', 'state_occupation': 'Oklahoma', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def tennessee_and_oregon(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, capital_gains_long=0, filing_status='single', income=100000, dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Tennessee:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        Oregon:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Tennessee:
        - https://www.tn.gov/revenue/taxes.html
        
        Oregon:
        -         
        - https://www.oregon.gov/dor/programs/individuals/Pages/PIT.aspx
          https://www.oregon.gov/dor/programs/individuals/Documents/Full%20year%20resident
         %20Form%20or-40%20filers.pdf
          https://www.oregon.gov/dor/programs/individuals/Documents/Part-year%20and%20nonresident
         %20Form%20or-40-P%20and%20or-40-N%20filers.pdf.pdf
          https://www.oregon.gov/dor/forms/FormsPubs/form-or-40_101-040_2022.pdf
          https://www.oregon.gov/dor/forms/FormsPubs/form-or-40_101-040_2022.pdf
          https://www.oregon.gov/dor/forms/FormsPubs/form-or-40_101-040_2022.pdf
          https://www.portland.gov/revenue/personal-tax
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Tennessee', 'state_occupation': 'Oregon', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def tennessee_and_pennsylvania(self, county_residence='Other', state_agi=75000, county_occupation='Other', capital_gains_short=0, capital_gains_long=0, income=100000, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Tennessee:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        Pennsylvania:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        county_residence : str, optional
            County of residence. Default is Other.
        county_occupation : str, optional
            County of occupation. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Tennessee:
        - https://www.tn.gov/revenue/taxes.html
        
        Pennsylvania:
        -         
        - https://www.revenue.pa.gov/Tax%20Rates/Pages/default.aspx
          https://www.revenue.pa.gov/FormsandPublications/FormsforIndividuals/PIT/Documents/2022/2022_pa-40in.pdf
          https://www.revenue.pa.gov/FormsandPublications/FormsforBusinesses/EmployerWithholding/Documents/rev-419.pdf
          https://www.revenue.pa.gov/FormsandPublications/FormsforIndividuals/PIT/Documents/2022/2022_pa-40in.pdf
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-25.htm
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        county_occupation = self.county_occupation if hasattr(self, 'county_occupation') else county_occupation
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        income = self.income if hasattr(self, 'income') else income
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Tennessee', 'state_occupation': 'Pennsylvania', "county_residence": county_residence, "state_agi": state_agi, "county_occupation": county_occupation, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "income": income, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def tennessee_and_rhode_island(self, state_agi=75000, capital_gains_short=0, capital_gains_long=0, filing_status='single', income=100000, dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Tennessee:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        Rhode Island:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Tennessee:
        - https://www.tn.gov/revenue/taxes.html
        
        Rhode Island:
        -         
        - https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2022_1040WE_w_0.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Tennessee', 'state_occupation': 'RhodeIsland', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def tennessee_and_south_carolina(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, capital_gains_long=0, filing_status='single', income=100000, dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Tennessee:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        South Carolina:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Tennessee:
        - https://www.tn.gov/revenue/taxes.html
        
        South Carolina:
        -         
        - https://dor.sc.gov/resources-site/lawandpolicy/Advisory%20Opinions/IL22-15.pdf
          https://dor.sc.gov/forms-site/Forms/IITPacket_2022.pdf
          https://www.scstatehouse.gov/code/t12c006.php
          https://dor.sc.gov/resources-site/lawandpolicy/Advisory%20Opinions/IL22-15.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Tennessee', 'state_occupation': 'SouthCarolina', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def tennessee_and_south_dakota(self, sources=False):
        """
        --- Parameters ----
        
        Tennessee:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        South Dakota:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Tennessee:
        - https://www.tn.gov/revenue/taxes.html
        
        South Dakota:
        -         
        - https://dor.sd.gov/individuals/taxes/
        """
        
        # Check if the parameters are set upstream in the instance
        sources = self.sources if hasattr(self, 'sources') else sources
        
        # State-specific parameters
        data = {
            'state_residence': 'Tennessee', 'state_occupation': 'SouthDakota', "sources": sources
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def tennessee_and_texas(self, sources=False):
        """
        --- Parameters ----
        
        Tennessee:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        Texas:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Tennessee:
        - https://www.tn.gov/revenue/taxes.html
        
        Texas:
        -         
        - https://comptroller.texas.gov/taxes/
        """
        
        # Check if the parameters are set upstream in the instance
        sources = self.sources if hasattr(self, 'sources') else sources
        
        # State-specific parameters
        data = {
            'state_residence': 'Tennessee', 'state_occupation': 'Texas', "sources": sources
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def tennessee_and_utah(self, state_agi=75000, capital_gains_short=0, capital_gains_long=0, filing_status='single', income=100000, dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Tennessee:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        Utah:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Tennessee:
        - https://www.tn.gov/revenue/taxes.html
        
        Utah:
        -         
        - https://incometax.utah.gov/paying/tax-rates
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Tennessee', 'state_occupation': 'Utah', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def tennessee_and_vermont(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, capital_gains_long=0, filing_status='single', income=100000, dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Tennessee:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        Vermont:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Tennessee:
        - https://www.tn.gov/revenue/taxes.html
        
        Vermont:
        -         
        - https://tax.vermont.gov/sites/tax/files/documents/GB-1210-2023.pdf
          https://tax.vermont.gov/sites/tax/files/documents/Income%20Booklet-2022.pdf
          https://tax.vermont.gov/sites/tax/files/documents/Income%20Booklet-2022.pdf
          https://tax.vermont.gov/sites/tax/files/documents/Income%20Booklet-2022.pdf
          https://tax.vermont.gov/individuals/personal-income-tax/tax-credits
          https://tax.vermont.gov/individuals/personal-income-tax/taxable-income
          https://tax.vermont.gov/individuals/personal-income-tax/local-option-tax
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Tennessee', 'state_occupation': 'Vermont', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def tennessee_and_virginia(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, capital_gains_long=0, filing_status='single', income=100000, dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Tennessee:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        Virginia:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Tennessee:
        - https://www.tn.gov/revenue/taxes.html
        
        Virginia:
        -         
        - https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/reciprocity
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Tennessee', 'state_occupation': 'Virginia', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def tennessee_and_washington(self, capital_gains_long=0, sources=False, capital_gains_short=0):
        """
        --- Parameters ----
        
        Tennessee:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        Washington:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Tennessee:
        - https://www.tn.gov/revenue/taxes.html
        
        Washington:
        -         
        - https://dor.wa.gov/taxes-rates/income-tax
          https://dor.wa.gov/taxes-rates/other-taxes/capital-gains-tax
          https://dor.wa.gov/taxes-rates/income-tax
        """
        
        # Check if the parameters are set upstream in the instance
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        sources = self.sources if hasattr(self, 'sources') else sources
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        
        # State-specific parameters
        data = {
            'state_residence': 'Tennessee', 'state_occupation': 'Washington', "capital_gains_long": capital_gains_long, "sources": sources, "capital_gains_short": capital_gains_short
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def tennessee_and_west_virginia(self, pay_periods=24, state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', county_occupation='Other', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Tennessee:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        West Virginia:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_occupation : str, optional
            County of occupation. Default is Other.
        pay_periods : int, optional
            Number of pay periods. Options are 13, 24, 26, 52. Default is 24.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Tennessee:
        - https://www.tn.gov/revenue/taxes.html
        
        West Virginia:
        -         
        - https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://tax.wv.gov/Documents/TSD/tsd438.pdf
          https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2011/Taxes-11-35att.html
        """
        
        # Check if the parameters are set upstream in the instance
        pay_periods = self.pay_periods if hasattr(self, 'pay_periods') else pay_periods
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        county_occupation = self.county_occupation if hasattr(self, 'county_occupation') else county_occupation
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Tennessee', 'state_occupation': 'WestVirginia', "pay_periods": pay_periods, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "county_occupation": county_occupation, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def tennessee_and_wisconsin(self, state_agi=75000, capital_gains_short=0, capital_gains_long=0, filing_status='single', income=100000, dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Tennessee:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        Wisconsin:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Tennessee:
        - https://www.tn.gov/revenue/taxes.html
        
        Wisconsin:
        -         
        - https://www.revenue.wi.gov/TaxForms2023/2023-Form1-ES-Inst.pdf
          https://www.revenue.wi.gov/TaxForms2023/2023-Form1-ES-Inst.pdf
          https://www.revenue.wi.gov/TaxForms2023/2023-Form1-ES-Inst.pdf
          https://www.revenue.wi.gov/Pages/FAQS/pcs-work.aspx
          https://www.revenue.wi.gov/Dor%20Publications/pb103.pdf
          https://www.revenue.wi.gov/Pages/FAQS/pcs-work.aspx
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Tennessee', 'state_occupation': 'Wisconsin', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def tennessee_and_wyoming(self, sources=False):
        """
        --- Parameters ----
        
        Tennessee:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        Wyoming:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Tennessee:
        - https://www.tn.gov/revenue/taxes.html
        
        Wyoming:
        -         
        - https://revenue.wyo.gov/
        """
        
        # Check if the parameters are set upstream in the instance
        sources = self.sources if hasattr(self, 'sources') else sources
        
        # State-specific parameters
        data = {
            'state_residence': 'Tennessee', 'state_occupation': 'Wyoming', "sources": sources
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def texas_and_alabama(self, county_residence='Other', state_agi=75000, capital_gains_short=0, capital_gains_long=0, filing_status='single', income=100000, dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Texas:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        Alabama:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Texas:
        - https://comptroller.texas.gov/taxes/
        
        Alabama:
        -         
        - https://www.revenue.alabama.gov/ultraviewer/viewer/basic_viewer/index.html?form=2023/02/22f40.pdf
          https://www.revenue.alabama.gov/faqs/what-is-alabamas-individual-income-tax-rate/
          https://www.revenue.alabama.gov/ultraviewer/viewer/basic_viewer/index.html?form=2023/02/22f40.pdf
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-24.htm?taxmap=true
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-24.htm?taxmap=true
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-24.htm?taxmap=true
          https://www.revenue.alabama.gov/individual-corporate/income-to-be-reported-on-the-alabama-income-tax-return/
          https://www.almonline.org/TaxRates.aspx#OccupationalTax
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Texas', 'state_occupation': 'Alabama', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def texas_and_alaska(self, sources=False):
        """
        --- Parameters ----
        
        Texas:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        Alaska:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Texas:
        - https://comptroller.texas.gov/taxes/
        
        Alaska:
        -         
        - https://www.commerce.alaska.gov/web/dcra/officeofthestateassessor/alaskataxfacts.aspx
        """
        
        # Check if the parameters are set upstream in the instance
        sources = self.sources if hasattr(self, 'sources') else sources
        
        # State-specific parameters
        data = {
            'state_residence': 'Texas', 'state_occupation': 'Alaska', "sources": sources
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def texas_and_arizona(self, state_agi=75000, capital_gains_short=0, capital_gains_long=0, filing_status='single', income=100000, dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Texas:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        Arizona:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        credits_state : int, optional
            Credits for state taxes. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Texas:
        - https://comptroller.texas.gov/taxes/
        
        Arizona:
        -         
        - https://azdor.gov/forms/individual/form-140-x-y-tables
          https://azdor.gov/forms/individual/form-140-x-y-tables
          https://azdor.gov/forms/individual/form-140-x-y-tables
          https://azdor.gov/businesses-arizona/withholding-tax/withholding-exceptions
          "https://azdor.gov/individuals/income-tax-filing-assistance/identifying-other-taxable-income#:~:text=In%20regards%20to%20capital%20gains
         the%20individuals%20regular%20tax%20rate.
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Texas', 'state_occupation': 'Arizona', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def texas_and_arkansas(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, capital_gains_long=0, filing_status='single', income=100000, dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Texas:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        Arkansas:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Texas:
        - https://comptroller.texas.gov/taxes/
        
        Arkansas:
        -         
        - https://www.arkansasedc.com/why-arkansas/business-climate/tax-structure/personal-income-tax
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_and_AR1000NR_Instructions.pdf
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_FullYearResidentIndividualIncomeTaxReturn.pdf
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_and_AR1000NR_Instructions.pdf
          https://www.arkansasedc.com/why-arkansas/business-climate/tax-structure/capital-gains-tax-reduction
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_and_AR1000NR_Instructions.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Texas', 'state_occupation': 'Arkansas', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def texas_and_california(self, state_agi=75000, capital_gains_short=0, capital_gains_long=0, filing_status='single', income=100000, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Texas:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        California:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Texas:
        - https://comptroller.texas.gov/taxes/
        
        California:
        -         
        - https://www.ftb.ca.gov/forms/2022/2022-540-tax-rate-schedules.pdf
          https://www.ftb.ca.gov/file/personal/deductions/index.html
          https://www.ftb.ca.gov/about-ftb/newsroom/caleitc/eligibility-and-credit-information.html
          https://www.ftb.ca.gov/file/personal/income-types/capital-gains-and-losses.html#:~:text=California%20does%20not%20have%20a
         are%20taxed%20as%20ordinary%20income
          https://www.ftb.ca.gov/file/personal/
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Texas', 'state_occupation': 'California', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def texas_and_colorado(self, county_residence='Other', state_agi=75000, capital_gains_short=0, capital_gains_long=0, income=100000, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Texas:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        Colorado:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Texas:
        - https://comptroller.texas.gov/taxes/
        
        Colorado:
        -         
        - https://leg.colorado.gov/agencies/legislative-council-staff/individual-income-tax%C2%A0
          https://leg.colorado.gov/agencies/legislative-council-staff/individual-income-tax%C2%A0
          https://tax.colorado.gov/sites/tax/files/documents/ITT_Capital_Gain_Subtraction_Feb_2022.pdf
          https://www.auroragov.org/business_services/taxes/occupational_privilege_tax
          https://www.denvergov.org/Government/Agencies-Departments-Offices/Agencies-Departments-Offices-Directory/Department-of-Finance/Our-Divisions/Treasury/Business-Tax-Information#:~:text=Occupational%20privilege%20tax
         -%28show%20below%29&text=Each%20taxable%20employee%20is%20liable
         month%20for%20each%20taxable%20employee
          https://www.glendale.co.us/355/Occupational-Privilege-Tax
          https://www.greenwoodvillage.com/1220/Occupational-Privilege-Tax-OPT
          https://ci.sheridan.co.us/288/Occupational-Privilege-Tax
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        income = self.income if hasattr(self, 'income') else income
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Texas', 'state_occupation': 'Colorado', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "income": income, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def texas_and_connecticut(self, state_agi=75000, capital_gains_short=0, capital_gains_long=0, filing_status='single', income=100000, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Texas:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        Connecticut:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Texas:
        - https://comptroller.texas.gov/taxes/
        
        Connecticut:
        -         
        - https://portal.ct.gov/-/media/DRS/Forms/2022/Income/2022-CT-1040-Instructions_1222.pdf
          https://www.cga.ct.gov/2022/rpt/pdf/2022-R-0108.pdf
          https://www.cga.ct.gov/current/pub/chap_229.htm#sec_12-703
          https://portal.ct.gov/-/media/DRS/Forms/2022/Income/2022-CT-1040-Instructions_1222.pdf
          "https://portal.ct.gov/DRS/Publications/TSSNs/TSSN-29#:~:text=An%20individuals%20net%20capital%20gains
         portion%20of%20Social%20Security%20benefits"
          "https://portal.ct.gov/DRS/Publications/TSSNs/TSSN-29#:~:text=An%20individuals%20net%20capital%20gains
         portion%20of%20Social%20Security%20benefits"
          https://portal.ct.gov/DRS/Individuals/Resident-Income-Tax/Tax-Information
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Texas', 'state_occupation': 'Connecticut', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def texas_and_delaware(self, capital_gains_long=0, federal_agi=75000, income=100000, sources=False, capital_gains_short=0, credits_state=0):
        """
        --- Parameters ----
        
        Texas:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        Delaware:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        federal_agi : int, optional
            Adjusted Gross Income (AGI) for federal taxes. Default is 0. Sources - https://www.irs.gov/e-file-providers/definition-of-adjusted-gross-income, https://www.irs.gov/newsroom/taxpayers-should-know-the-difference-between-standard-and-itemized-deductions
        income : int, optional
            Gross income. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Texas:
        - https://comptroller.texas.gov/taxes/
        
        Delaware:
        -         
        - https://revenue.delaware.gov/software-developer/tax-rate-changes/
          https://revenuefiles.delaware.gov/2022/PIT-RES_TY22_2022-01_PaperInteractive.pdf
          https://revenuefiles.delaware.gov/2022/PIT-RES_TY22_2022-02_Instructions.pdf
          https://revenue.delaware.gov/frequently-asked-questions/personal-income-tax-faqs/
          https://www.newcastlede.gov/DocumentCenter/View/20147/City-of-Wilmington-Wage-Tax-Information-PDF
        """
        
        # Check if the parameters are set upstream in the instance
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        federal_agi = self.federal_agi if hasattr(self, 'federal_agi') else federal_agi
        income = self.income if hasattr(self, 'income') else income
        sources = self.sources if hasattr(self, 'sources') else sources
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Texas', 'state_occupation': 'Delaware', "capital_gains_long": capital_gains_long, "federal_agi": federal_agi, "income": income, "sources": sources, "capital_gains_short": capital_gains_short, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def texas_and_florida(self, sources=False):
        """
        --- Parameters ----
        
        Texas:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        Florida:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Texas:
        - https://comptroller.texas.gov/taxes/
        
        Florida:
        -         
        - https://floridarevenue.com/taxes/taxesfees/Pages/tax_interest_rates.aspx
        """
        
        # Check if the parameters are set upstream in the instance
        sources = self.sources if hasattr(self, 'sources') else sources
        
        # State-specific parameters
        data = {
            'state_residence': 'Texas', 'state_occupation': 'Florida', "sources": sources
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def texas_and_georgia(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, capital_gains_long=0, filing_status='single', income=100000, dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Texas:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        Georgia:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Texas:
        - https://comptroller.texas.gov/taxes/
        
        Georgia:
        -         
        - https://dor.georgia.gov/tax-tables-georgia-tax-rate-schedule
          https://apps.dor.ga.gov/FillableForms/PDFViewer/Index?form=2022GA500
          https://apps.dor.ga.gov/FillableForms/PDFViewer/Index?form=2022GA500
          https://apps.dor.ga.gov/FillableForms/PDFViewer/Index?form=2022GA500
          https://dor.georgia.gov/years-individual-income-tax-forms
          https://dor.georgia.gov/years-individual-income-tax-forms
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Texas', 'state_occupation': 'Georgia', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def texas_and_hawaii(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, capital_gains_long=0, filing_status='single', income=100000, dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Texas:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        Hawaii:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Texas:
        - https://comptroller.texas.gov/taxes/
        
        Hawaii:
        -         
        - https://tax.hawaii.gov/forms/d_18table-on/d_18table-on_p13/
          https://files.hawaii.gov/tax/forms/2018/18table-on.pdf
          https://files.hawaii.gov/tax/news/pubs/22outline.pdf
          https://files.hawaii.gov/tax/news/pubs/FastTaxRefGuide_2017.pdf
          https://www.tfhawaii.org/wordpress/state-tax-resources/mini-tax-guides/individual-income-tax-chapter-235/#:~:text=Long%20term%20capital%20gains%20are%20taxed%20at%20a%20maximum%20of%207.25%25
          https://www.grassrootinstitute.org/2023/02/hb337-increasing-tax-on-capital-gains-a-bad-risk/
          https://files.hawaii.gov/tax/news/pubs/22outline.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Texas', 'state_occupation': 'Hawaii', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def texas_and_idaho(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, capital_gains_long=0, filing_status='single', income=100000, dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Texas:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        Idaho:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Texas:
        - https://comptroller.texas.gov/taxes/
        
        Idaho:
        -         
        - https://tax.idaho.gov/taxes/income-tax/individual-income/individual-income-tax-rate-schedule/
          https://tax.idaho.gov/wp-content/uploads/forms/EIN00046/EIN00046_03-01-2023.pdf
          https://tax.idaho.gov/taxes/income-tax/individual-income/idaho-source-income/
          https://tax.idaho.gov/taxes/income-tax/individual-income/idaho-source-income/
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Texas', 'state_occupation': 'Idaho', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def texas_and_illinois(self, state_agi=75000, capital_gains_short=0, capital_gains_long=0, filing_status='single', income=100000, dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Texas:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        Illinois:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Texas:
        - https://comptroller.texas.gov/taxes/
        
        Illinois:
        -         
        - https://tax.illinois.gov/research/taxrates/income.html#IndividualIncome
          https://tax.illinois.gov/content/dam/soi/en/web/tax/forms/incometax/documents/currentyear/individual/il-1040-instr.pdf
          https://tax.illinois.gov/questionsandanswers/12.html
          https://tax.illinois.gov/content/dam/soi/en/web/tax/forms/incometax/documents/currentyear/individual/il-1040-instr.pdf
          https://tax.illinois.gov/research/taxrates/personalproperty.html
          https://tax.illinois.gov/localgovernments/income.html
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Texas', 'state_occupation': 'Illinois', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def texas_and_indiana(self, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_short=0, capital_gains_long=0, filing_status='single', income=100000, dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Texas:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        Indiana:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Texas:
        - https://comptroller.texas.gov/taxes/
        
        Indiana:
        -         
        - https://www.in.gov/dor/tax-forms/2022-individual-income-tax-forms/
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-32.htm
          https://www.in.gov/dor/files/ib28.pdf
          https://www.in.gov/dor/tax-forms/2022-individual-income-tax-forms/
          https://www.in.gov/dor/files/2022-county-tax-rates-and-codes.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Texas', 'state_occupation': 'Indiana', "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def texas_and_iowa(self, county_residence='Other', state_agi=75000, capital_gains_short=0, capital_gains_long=0, filing_status='single', income=100000, dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Texas:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        Iowa:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Texas:
        - https://comptroller.texas.gov/taxes/
        
        Iowa:
        -         
        - https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
          https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
          https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
          https://tax.iowa.gov/iowa-illinois-reciprocal-agreement
          https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Texas', 'state_occupation': 'Iowa', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def texas_and_kansas(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, capital_gains_long=0, filing_status='single', income=100000, dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Texas:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        Kansas:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Texas:
        - https://comptroller.texas.gov/taxes/
        
        Kansas:
        -         
        - https://www.ksrevenue.gov/taxrates.html
          https://www.ksrevenue.gov/incomebook22.html#0
          https://www.ksrevenue.gov/incomebook22.html#0
          https://www.ksrevenue.gov/incomebook22.html#0
          https://www.ksrevenue.gov/incomebook22.html
          https://www.ksrevenue.gov/incomebook22.html
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Texas', 'state_occupation': 'Kansas', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def texas_and_kentucky(self, county_residence='Other', state_agi=75000, capital_gains_short=0, capital_gains_long=0, filing_status='single', income=100000, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Texas:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        Kentucky:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Texas:
        - https://comptroller.texas.gov/taxes/
        
        Kentucky:
        -         
        - https://revenue.ky.gov/News/Pages/Dor-Announces-Updates-to-Individual-Income-Tax-for-2023-Tax-Year.aspx
          https://revenue.ky.gov/Individual/Individual-Income-Tax/Pages/default.aspx
          https://revenue.ky.gov/News/Pages/Dor-Announces-Updates-to-Individual-Income-Tax-for-2023-Tax-Year.aspx
          https://revenue.ky.gov/Individual/Individual-Income-Tax/Pages/default.aspx
          https://revenue.ky.gov/Dor%20Training%20Materials/103%20KAR%2017.140.%20Individual%20income%20tax%20-%20reciprocity%20-%20nonresidents.pdf
          https://revenue.ky.gov/Individual/Individual-Income-Tax/Pages/default.aspx
          https://nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2021/Taxes-21-32.htm?taxmap=true
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Texas', 'state_occupation': 'Kentucky', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def texas_and_louisiana(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, capital_gains_long=0, filing_status='single', income=100000, dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Texas:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        Louisiana:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Texas:
        - https://comptroller.texas.gov/taxes/
        
        Louisiana:
        -         
        - https://revenue.louisiana.gov/individualincometax
          https://revenue.louisiana.gov/individualincometax
          https://www.revenue.louisiana.gov/taxforms/6935(11_02)F.pdf
          https://revenue.louisiana.gov/individualincometax
          https://revenue.louisiana.gov/individualincometax
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Texas', 'state_occupation': 'Louisiana', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def texas_and_maine(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, capital_gains_long=0, filing_status='single', income=100000, dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Texas:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        Maine:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Texas:
        - https://comptroller.texas.gov/taxes/
        
        Maine:
        -         
        - https://www.maine.gov/revenue/tax-return-forms/individual-income-tax-2022
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/ind_tax_rate_sched_2022.pdf
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/22_1040me_book_gen_instruc_revisedFeb23.pdf
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/22_1040me_book_gen_instruc_revisedFeb23.pdf
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/22_1040me_book_gen_instruc_revisedFeb23.pdf
          https://www.maine.gov/revenue/tax-return-forms/individual-income-tax-2022
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Texas', 'state_occupation': 'Maine', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def texas_and_maryland(self, county_residence='Other', state_agi=75000, capital_gains_short=0, capital_gains_long=0, filing_status='single', income=100000, dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Texas:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        Maryland:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Texas:
        - https://comptroller.texas.gov/taxes/
        
        Maryland:
        -         
        - https://www.marylandtaxes.gov/individual/income/tax-info/tax-rates.php
          https://www.marylandtaxes.gov/forms/22_forms/Resident_Booklet.pdf
          https://www.marylandtaxes.gov/forms/Tax_Publications/Administrative_Releases/Income_and_Estate_Tax_Releases/ar_it37.pdf
          https://www.marylandtaxes.gov/individual/index.php
          https://www.marylandtaxes.gov/individual/income/tax-info/tax-rates.php
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Texas', 'state_occupation': 'Maryland', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def texas_and_massachusetts(self, state_agi=75000, capital_gains_short=0, capital_gains_long=0, filing_status='single', income=100000, dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Texas:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        Massachusetts:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Texas:
        - https://comptroller.texas.gov/taxes/
        
        Massachusetts:
        -         
        - https://www.mass.gov/service-details/massachusetts-tax-rates
          https://www.mass.gov/service-details/view-massachusetts-personal-income-tax-exemptions
          https://www.mass.gov/service-details/view-massachusetts-personal-income-tax-exemptions
          https://www.mass.gov/service-details/view-child-and-dependent-related-credits
          https://www.mass.gov/service-details/massachusetts-tax-rates
          https://www.mass.gov/personal-income-tax
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Texas', 'state_occupation': 'Massachusetts', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def texas_and_michigan(self, county_residence='Other', state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', county_occupation='Other', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Texas:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        Michigan:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        county_occupation : str, optional
            County of occupation. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Texas:
        - https://comptroller.texas.gov/taxes/
        
        Michigan:
        -         
        - https://www.michigan.gov/taxes/iit/new-developments-for-tax-year-2022
          "https://www.michigan.gov/taxes/-/media/Project/Websites/taxes/Forms/2023/2023-SUW/446_Withholding-Guide_2023.pdf?rev=177ca3d8cb034392bf84e317ed78c320&hash=19C5F42F43A89A91863D229349F48883#:~:text=The%20exemption%20amount%20is%20%245%2C400
         employees%20Michigan%20income%20tax%20return"
          "https://www.michigan.gov/taxes/-/media/Project/Websites/taxes/Forms/2023/2023-SUW/446_Withholding-Guide_2023.pdf?rev=177ca3d8cb034392bf84e317ed78c320&hash=19C5F42F43A89A91863D229349F48883#:~:text=The%20exemption%20amount%20is%20%245%2C400
         employees%20Michigan%20income%20tax%20return"
          https://www.michigan.gov/taxes/questions/iit/accordion/residency/are-my-wages-earned-in-another-state-taxable-in-michigan-if-i-am-a-michigan-resident-1
          https://www.michigan.gov/taxes/-/media/Project/Websites/taxes/Forms/2022/2022-IIT-Forms/BOOK_MI-1040.pdf?rev=4524f1cbbc0243b194ab05a6680be75b&hash=40DFE71E651F7C1D65C31F6CD03D944F
          https://www.michigan.gov/taxes/questions/iit/accordion/general/what-cities-impose-an-income-tax
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        county_occupation = self.county_occupation if hasattr(self, 'county_occupation') else county_occupation
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Texas', 'state_occupation': 'Michigan', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "county_occupation": county_occupation, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def texas_and_minnesota(self, state_agi=75000, capital_gains_short=0, capital_gains_long=0, filing_status='single', income=100000, dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Texas:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        Minnesota:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Texas:
        - https://comptroller.texas.gov/taxes/
        
        Minnesota:
        -         
        - https://www.revenue.state.mn.us/minnesota-income-tax-rates-and-brackets
          https://www.revenue.state.mn.us/child-and-dependent-care-credit
          https://www.revenue.state.mn.us/dependent-exemptions
          https://www.revenue.state.mn.us/minnesota-standard-deduction
          https://mn.gov/mmb/assets/mwr_form_tcm1059-128581.pdf
          https://www.revenue.state.mn.us/individual-income-tax
          https://www.revenue.state.mn.us/individual-income-tax
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Texas', 'state_occupation': 'Minnesota', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def texas_and_mississippi(self, state_agi=75000, capital_gains_short=0, capital_gains_long=0, filing_status='single', income=100000, dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Texas:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        Mississippi:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Texas:
        - https://comptroller.texas.gov/taxes/
        
        Mississippi:
        -         
        - https://www.dor.ms.gov/individual/tax-rates
          https://www.dor.ms.gov/individual/tax-rates
          https://www.dor.ms.gov/individual/tax-rates
          https://www.dor.ms.gov/individual
          https://www.dor.ms.gov/individual
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Texas', 'state_occupation': 'Mississippi', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def texas_and_missouri(self, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_short=0, capital_gains_long=0, filing_status='single', income=100000, dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Texas:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        Missouri:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Texas:
        - https://comptroller.texas.gov/taxes/
        
        Missouri:
        -         
        - https://dor.mo.gov/taxation/individual/tax-types/income/year-changes/
          https://dor.mo.gov/taxation/individual/tax-types/income/year-changes/
          https://dor.mo.gov/taxation/individual/tax-types/income/year-changes/
          https://www.kcmo.gov/city-hall/departments/finance/earnings-tax
          https://www.stlouis-mo.gov/government/departments/collector/earnings-tax/individual-earnings-tax-info.cfm
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Texas', 'state_occupation': 'Missouri', "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def texas_and_montana(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, capital_gains_long=0, filing_status='single', income=100000, dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Texas:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        Montana:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Texas:
        - https://comptroller.texas.gov/taxes/
        
        Montana:
        -         
        - https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/2022/12/montana-employees-withholding-allowance-and-exemption-certificate-form-mw-4.pdf
          https://mtrevenue.gov/2021/11/29/simplification-of-montana-income-taxation/?utm_source=rss&utm_medium=rss&utm_campaign=simplification-of-montana-income-taxation
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Texas', 'state_occupation': 'Montana', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def texas_and_nebraska(self, state_agi=75000, capital_gains_short=0, capital_gains_long=0, filing_status='single', income=100000, dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Texas:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        Nebraska:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Texas:
        - https://comptroller.texas.gov/taxes/
        
        Nebraska:
        -         
        - https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2023_Circular_EN_8-429r2023_Whole_Book_11.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Texas', 'state_occupation': 'Nebraska', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def texas_and_nevada(self, sources=False):
        """
        --- Parameters ----
        
        Texas:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        Nevada:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Texas:
        - https://comptroller.texas.gov/taxes/
        
        Nevada:
        -         
        - https://www.nvsos.gov/sos/businesses/the-nevada-advantage
        """
        
        # Check if the parameters are set upstream in the instance
        sources = self.sources if hasattr(self, 'sources') else sources
        
        # State-specific parameters
        data = {
            'state_residence': 'Texas', 'state_occupation': 'Nevada', "sources": sources
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def texas_and_new_hampshire(self, sources=False):
        """
        --- Parameters ----
        
        Texas:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        New Hampshire:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Texas:
        - https://comptroller.texas.gov/taxes/
        
        New Hampshire:
        -         
        - https://www.revenue.nh.gov/assistance/tax-overview.htm
        """
        
        # Check if the parameters are set upstream in the instance
        sources = self.sources if hasattr(self, 'sources') else sources
        
        # State-specific parameters
        data = {
            'state_residence': 'Texas', 'state_occupation': 'NewHampshire', "sources": sources
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def texas_and_new_jersey(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, capital_gains_long=0, filing_status='single', income=100000, dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Texas:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        New Jersey:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Texas:
        - https://comptroller.texas.gov/taxes/
        
        New Jersey:
        -         
        - https://www.state.nj.us/treasury/taxation/taxtables.shtml
          https://www.state.nj.us/treasury/taxation/njit13.shtml
          https://www.nj.gov/treasury/taxation/njit25.shtml
          https://www.state.nj.us/treasury/taxation/individuals/freqqiti.shtml
          https://www.state.nj.us/treasury/taxation/localtax.shtml
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Texas', 'state_occupation': 'NewJersey', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def texas_and_new_mexico(self, state_agi=75000, capital_gains_short=0, capital_gains_long=0, filing_status='single', income=100000, dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Texas:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        New Mexico:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Texas:
        - https://comptroller.texas.gov/taxes/
        
        New Mexico:
        -         
        - https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/fdf3c548-8aba-4b9c-9eb4-bb564c716015/FYI-104.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/0bf943fd-652e-4400-bb1b-cd8397eb5a95/2022pit-adj.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/2f1a6781-9534-4436-b427-1557f9592099/2022pit-adj-ins.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/0bf943fd-652e-4400-bb1b-cd8397eb5a95/2022pit-adj.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/1afc56af-ea90-4d48-82e5-1f9aeb43255a/PITbook2022.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/fdf3c548-8aba-4b9c-9eb4-bb564c716015/FYI-104.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Texas', 'state_occupation': 'NewMexico', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def texas_and_new_york(self, county_residence='Other', state_agi=75000, capital_gains_short=0, capital_gains_long=0, filing_status='single', income=100000, dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Texas:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        New York:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Texas:
        - https://comptroller.texas.gov/taxes/
        
        New York:
        -         
        - https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nys-tax-rate-schedule
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nys-tax-rate-schedule
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nys-tax-rate-schedule
          https://www.tax.ny.gov/pit/
          https://www.tax.ny.gov/pit/file/nonresident-faqs.htm
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#yonkers-tax-rate-schedule
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nyc-tax-rate-schedule
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Texas', 'state_occupation': 'NewYork', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def texas_and_north_carolina(self, state_agi=75000, capital_gains_short=0, capital_gains_long=0, filing_status='single', income=100000, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Texas:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        North Carolina:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Texas:
        - https://comptroller.texas.gov/taxes/
        
        North Carolina:
        -         
        - https://www.ncdor.gov/taxes-forms/tax-rate-schedules
          https://www.ncdor.gov/taxes-forms/individual-income-tax/north-carolina-standard-deduction-or-north-carolina-itemized-deductions
          https://www.ncdor.gov/taxes-forms/individual-income-tax
          https://www.ncdor.gov/taxes-forms/individual-income-tax
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Texas', 'state_occupation': 'NorthCarolina', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def texas_and_north_dakota(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, capital_gains_long=0, filing_status='single', income=100000, dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Texas:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        North Dakota:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Texas:
        - https://comptroller.texas.gov/taxes/
        
        North Dakota:
        -         
        - https://www.tax.nd.gov/forms
          https://www.tax.nd.gov/sites/www/files/documents/forms/individual/2022-iit/2022-individual-income-tax-booklet.pdf
          https://www.tax.nd.gov/sites/www/files/documents/forms/2022-individual-income-tax-booklet.pdf
          https://www.tax.nd.gov/tax-exemptions-credits/income-tax-exemptions-credits
          https://www.tax.nd.gov/sites/www/files/documents/forms/business/ndwrfillable.pdf
          https://www.tax.nd.gov/news/tax-legislative-changes/significant-changes-law/individual-income-tax-history
          https://www.tax.nd.gov/forms
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Texas', 'state_occupation': 'NorthDakota', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def texas_and_ohio(self, county_residence='Other', state_agi=75000, capital_gains_short=0, capital_gains_long=0, filing_status='single', income=100000, dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Texas:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        Ohio:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Texas:
        - https://comptroller.texas.gov/taxes/
        
        Ohio:
        -         
        - https://tax.ohio.gov/individual/resources/annual-tax-rates
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://tax.ohio.gov/static/forms/employer_withholding/generic/wth_it4nr.pdf
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://www.ritaohio.com/TaxRatesTable
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Texas', 'state_occupation': 'Ohio', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def texas_and_oklahoma(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, capital_gains_long=0, filing_status='single', income=100000, dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Texas:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        Oklahoma:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Texas:
        - https://comptroller.texas.gov/taxes/
        
        Oklahoma:
        -         
        - https://oklahoma.gov/content/dam/ok/en/tax/documents/resources/publications/businesses/withholding-tables/WHTables-2022.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Texas', 'state_occupation': 'Oklahoma', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def texas_and_oregon(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, capital_gains_long=0, filing_status='single', income=100000, dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Texas:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        Oregon:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Texas:
        - https://comptroller.texas.gov/taxes/
        
        Oregon:
        -         
        - https://www.oregon.gov/dor/programs/individuals/Pages/PIT.aspx
          https://www.oregon.gov/dor/programs/individuals/Documents/Full%20year%20resident
         %20Form%20or-40%20filers.pdf
          https://www.oregon.gov/dor/programs/individuals/Documents/Part-year%20and%20nonresident
         %20Form%20or-40-P%20and%20or-40-N%20filers.pdf.pdf
          https://www.oregon.gov/dor/forms/FormsPubs/form-or-40_101-040_2022.pdf
          https://www.oregon.gov/dor/forms/FormsPubs/form-or-40_101-040_2022.pdf
          https://www.oregon.gov/dor/forms/FormsPubs/form-or-40_101-040_2022.pdf
          https://www.portland.gov/revenue/personal-tax
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Texas', 'state_occupation': 'Oregon', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def texas_and_pennsylvania(self, county_residence='Other', state_agi=75000, county_occupation='Other', capital_gains_short=0, capital_gains_long=0, income=100000, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Texas:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        Pennsylvania:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        county_residence : str, optional
            County of residence. Default is Other.
        county_occupation : str, optional
            County of occupation. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Texas:
        - https://comptroller.texas.gov/taxes/
        
        Pennsylvania:
        -         
        - https://www.revenue.pa.gov/Tax%20Rates/Pages/default.aspx
          https://www.revenue.pa.gov/FormsandPublications/FormsforIndividuals/PIT/Documents/2022/2022_pa-40in.pdf
          https://www.revenue.pa.gov/FormsandPublications/FormsforBusinesses/EmployerWithholding/Documents/rev-419.pdf
          https://www.revenue.pa.gov/FormsandPublications/FormsforIndividuals/PIT/Documents/2022/2022_pa-40in.pdf
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-25.htm
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        county_occupation = self.county_occupation if hasattr(self, 'county_occupation') else county_occupation
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        income = self.income if hasattr(self, 'income') else income
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Texas', 'state_occupation': 'Pennsylvania', "county_residence": county_residence, "state_agi": state_agi, "county_occupation": county_occupation, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "income": income, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def texas_and_rhode_island(self, state_agi=75000, capital_gains_short=0, capital_gains_long=0, filing_status='single', income=100000, dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Texas:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        Rhode Island:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Texas:
        - https://comptroller.texas.gov/taxes/
        
        Rhode Island:
        -         
        - https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2022_1040WE_w_0.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Texas', 'state_occupation': 'RhodeIsland', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def texas_and_south_carolina(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, capital_gains_long=0, filing_status='single', income=100000, dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Texas:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        South Carolina:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Texas:
        - https://comptroller.texas.gov/taxes/
        
        South Carolina:
        -         
        - https://dor.sc.gov/resources-site/lawandpolicy/Advisory%20Opinions/IL22-15.pdf
          https://dor.sc.gov/forms-site/Forms/IITPacket_2022.pdf
          https://www.scstatehouse.gov/code/t12c006.php
          https://dor.sc.gov/resources-site/lawandpolicy/Advisory%20Opinions/IL22-15.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Texas', 'state_occupation': 'SouthCarolina', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def texas_and_south_dakota(self, sources=False):
        """
        --- Parameters ----
        
        Texas:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        South Dakota:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Texas:
        - https://comptroller.texas.gov/taxes/
        
        South Dakota:
        -         
        - https://dor.sd.gov/individuals/taxes/
        """
        
        # Check if the parameters are set upstream in the instance
        sources = self.sources if hasattr(self, 'sources') else sources
        
        # State-specific parameters
        data = {
            'state_residence': 'Texas', 'state_occupation': 'SouthDakota', "sources": sources
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def texas_and_tennessee(self, sources=False):
        """
        --- Parameters ----
        
        Texas:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        Tennessee:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Texas:
        - https://comptroller.texas.gov/taxes/
        
        Tennessee:
        -         
        - https://www.tn.gov/revenue/taxes.html
        """
        
        # Check if the parameters are set upstream in the instance
        sources = self.sources if hasattr(self, 'sources') else sources
        
        # State-specific parameters
        data = {
            'state_residence': 'Texas', 'state_occupation': 'Tennessee', "sources": sources
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def texas_and_utah(self, state_agi=75000, capital_gains_short=0, capital_gains_long=0, filing_status='single', income=100000, dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Texas:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        Utah:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Texas:
        - https://comptroller.texas.gov/taxes/
        
        Utah:
        -         
        - https://incometax.utah.gov/paying/tax-rates
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Texas', 'state_occupation': 'Utah', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def texas_and_vermont(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, capital_gains_long=0, filing_status='single', income=100000, dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Texas:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        Vermont:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Texas:
        - https://comptroller.texas.gov/taxes/
        
        Vermont:
        -         
        - https://tax.vermont.gov/sites/tax/files/documents/GB-1210-2023.pdf
          https://tax.vermont.gov/sites/tax/files/documents/Income%20Booklet-2022.pdf
          https://tax.vermont.gov/sites/tax/files/documents/Income%20Booklet-2022.pdf
          https://tax.vermont.gov/sites/tax/files/documents/Income%20Booklet-2022.pdf
          https://tax.vermont.gov/individuals/personal-income-tax/tax-credits
          https://tax.vermont.gov/individuals/personal-income-tax/taxable-income
          https://tax.vermont.gov/individuals/personal-income-tax/local-option-tax
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Texas', 'state_occupation': 'Vermont', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def texas_and_virginia(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, capital_gains_long=0, filing_status='single', income=100000, dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Texas:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        Virginia:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Texas:
        - https://comptroller.texas.gov/taxes/
        
        Virginia:
        -         
        - https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/reciprocity
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Texas', 'state_occupation': 'Virginia', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def texas_and_washington(self, capital_gains_long=0, sources=False, capital_gains_short=0):
        """
        --- Parameters ----
        
        Texas:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        Washington:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Texas:
        - https://comptroller.texas.gov/taxes/
        
        Washington:
        -         
        - https://dor.wa.gov/taxes-rates/income-tax
          https://dor.wa.gov/taxes-rates/other-taxes/capital-gains-tax
          https://dor.wa.gov/taxes-rates/income-tax
        """
        
        # Check if the parameters are set upstream in the instance
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        sources = self.sources if hasattr(self, 'sources') else sources
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        
        # State-specific parameters
        data = {
            'state_residence': 'Texas', 'state_occupation': 'Washington', "capital_gains_long": capital_gains_long, "sources": sources, "capital_gains_short": capital_gains_short
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def texas_and_west_virginia(self, pay_periods=24, state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', county_occupation='Other', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Texas:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        West Virginia:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_occupation : str, optional
            County of occupation. Default is Other.
        pay_periods : int, optional
            Number of pay periods. Options are 13, 24, 26, 52. Default is 24.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Texas:
        - https://comptroller.texas.gov/taxes/
        
        West Virginia:
        -         
        - https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://tax.wv.gov/Documents/TSD/tsd438.pdf
          https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2011/Taxes-11-35att.html
        """
        
        # Check if the parameters are set upstream in the instance
        pay_periods = self.pay_periods if hasattr(self, 'pay_periods') else pay_periods
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        county_occupation = self.county_occupation if hasattr(self, 'county_occupation') else county_occupation
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Texas', 'state_occupation': 'WestVirginia', "pay_periods": pay_periods, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "county_occupation": county_occupation, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def texas_and_wisconsin(self, state_agi=75000, capital_gains_short=0, capital_gains_long=0, filing_status='single', income=100000, dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Texas:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        Wisconsin:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Texas:
        - https://comptroller.texas.gov/taxes/
        
        Wisconsin:
        -         
        - https://www.revenue.wi.gov/TaxForms2023/2023-Form1-ES-Inst.pdf
          https://www.revenue.wi.gov/TaxForms2023/2023-Form1-ES-Inst.pdf
          https://www.revenue.wi.gov/TaxForms2023/2023-Form1-ES-Inst.pdf
          https://www.revenue.wi.gov/Pages/FAQS/pcs-work.aspx
          https://www.revenue.wi.gov/Dor%20Publications/pb103.pdf
          https://www.revenue.wi.gov/Pages/FAQS/pcs-work.aspx
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Texas', 'state_occupation': 'Wisconsin', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def texas_and_wyoming(self, sources=False):
        """
        --- Parameters ----
        
        Texas:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        Wyoming:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Texas:
        - https://comptroller.texas.gov/taxes/
        
        Wyoming:
        -         
        - https://revenue.wyo.gov/
        """
        
        # Check if the parameters are set upstream in the instance
        sources = self.sources if hasattr(self, 'sources') else sources
        
        # State-specific parameters
        data = {
            'state_residence': 'Texas', 'state_occupation': 'Wyoming', "sources": sources
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def utah_and_alabama(self, county_residence='Other', state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Utah:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Alabama:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Utah:
        - https://incometax.utah.gov/paying/tax-rates
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
        
        Alabama:
        -         
        - https://www.revenue.alabama.gov/ultraviewer/viewer/basic_viewer/index.html?form=2023/02/22f40.pdf
          https://www.revenue.alabama.gov/faqs/what-is-alabamas-individual-income-tax-rate/
          https://www.revenue.alabama.gov/ultraviewer/viewer/basic_viewer/index.html?form=2023/02/22f40.pdf
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-24.htm?taxmap=true
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-24.htm?taxmap=true
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-24.htm?taxmap=true
          https://www.revenue.alabama.gov/individual-corporate/income-to-be-reported-on-the-alabama-income-tax-return/
          https://www.almonline.org/TaxRates.aspx#OccupationalTax
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Utah', 'state_occupation': 'Alabama', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def utah_and_alaska(self, state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Utah:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Alaska:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Utah:
        - https://incometax.utah.gov/paying/tax-rates
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
        
        Alaska:
        -         
        - https://www.commerce.alaska.gov/web/dcra/officeofthestateassessor/alaskataxfacts.aspx
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Utah', 'state_occupation': 'Alaska', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def utah_and_arizona(self, state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Utah:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Arizona:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        credits_state : int, optional
            Credits for state taxes. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Utah:
        - https://incometax.utah.gov/paying/tax-rates
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
        
        Arizona:
        -         
        - https://azdor.gov/forms/individual/form-140-x-y-tables
          https://azdor.gov/forms/individual/form-140-x-y-tables
          https://azdor.gov/forms/individual/form-140-x-y-tables
          https://azdor.gov/businesses-arizona/withholding-tax/withholding-exceptions
          "https://azdor.gov/individuals/income-tax-filing-assistance/identifying-other-taxable-income#:~:text=In%20regards%20to%20capital%20gains
         the%20individuals%20regular%20tax%20rate.
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Utah', 'state_occupation': 'Arizona', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def utah_and_arkansas(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Utah:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Arkansas:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Utah:
        - https://incometax.utah.gov/paying/tax-rates
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
        
        Arkansas:
        -         
        - https://www.arkansasedc.com/why-arkansas/business-climate/tax-structure/personal-income-tax
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_and_AR1000NR_Instructions.pdf
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_FullYearResidentIndividualIncomeTaxReturn.pdf
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_and_AR1000NR_Instructions.pdf
          https://www.arkansasedc.com/why-arkansas/business-climate/tax-structure/capital-gains-tax-reduction
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_and_AR1000NR_Instructions.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Utah', 'state_occupation': 'Arkansas', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def utah_and_california(self, state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Utah:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        California:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Utah:
        - https://incometax.utah.gov/paying/tax-rates
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
        
        California:
        -         
        - https://www.ftb.ca.gov/forms/2022/2022-540-tax-rate-schedules.pdf
          https://www.ftb.ca.gov/file/personal/deductions/index.html
          https://www.ftb.ca.gov/about-ftb/newsroom/caleitc/eligibility-and-credit-information.html
          https://www.ftb.ca.gov/file/personal/income-types/capital-gains-and-losses.html#:~:text=California%20does%20not%20have%20a
         are%20taxed%20as%20ordinary%20income
          https://www.ftb.ca.gov/file/personal/
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Utah', 'state_occupation': 'California', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def utah_and_colorado(self, county_residence='Other', state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Utah:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Colorado:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Utah:
        - https://incometax.utah.gov/paying/tax-rates
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
        
        Colorado:
        -         
        - https://leg.colorado.gov/agencies/legislative-council-staff/individual-income-tax%C2%A0
          https://leg.colorado.gov/agencies/legislative-council-staff/individual-income-tax%C2%A0
          https://tax.colorado.gov/sites/tax/files/documents/ITT_Capital_Gain_Subtraction_Feb_2022.pdf
          https://www.auroragov.org/business_services/taxes/occupational_privilege_tax
          https://www.denvergov.org/Government/Agencies-Departments-Offices/Agencies-Departments-Offices-Directory/Department-of-Finance/Our-Divisions/Treasury/Business-Tax-Information#:~:text=Occupational%20privilege%20tax
         -%28show%20below%29&text=Each%20taxable%20employee%20is%20liable
         month%20for%20each%20taxable%20employee
          https://www.glendale.co.us/355/Occupational-Privilege-Tax
          https://www.greenwoodvillage.com/1220/Occupational-Privilege-Tax-OPT
          https://ci.sheridan.co.us/288/Occupational-Privilege-Tax
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Utah', 'state_occupation': 'Colorado', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def utah_and_connecticut(self, state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Utah:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Connecticut:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Utah:
        - https://incometax.utah.gov/paying/tax-rates
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
        
        Connecticut:
        -         
        - https://portal.ct.gov/-/media/DRS/Forms/2022/Income/2022-CT-1040-Instructions_1222.pdf
          https://www.cga.ct.gov/2022/rpt/pdf/2022-R-0108.pdf
          https://www.cga.ct.gov/current/pub/chap_229.htm#sec_12-703
          https://portal.ct.gov/-/media/DRS/Forms/2022/Income/2022-CT-1040-Instructions_1222.pdf
          "https://portal.ct.gov/DRS/Publications/TSSNs/TSSN-29#:~:text=An%20individuals%20net%20capital%20gains
         portion%20of%20Social%20Security%20benefits"
          "https://portal.ct.gov/DRS/Publications/TSSNs/TSSN-29#:~:text=An%20individuals%20net%20capital%20gains
         portion%20of%20Social%20Security%20benefits"
          https://portal.ct.gov/DRS/Individuals/Resident-Income-Tax/Tax-Information
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Utah', 'state_occupation': 'Connecticut', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def utah_and_delaware(self, state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, federal_agi=75000, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Utah:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Delaware:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        federal_agi : int, optional
            Adjusted Gross Income (AGI) for federal taxes. Default is 0. Sources - https://www.irs.gov/e-file-providers/definition-of-adjusted-gross-income, https://www.irs.gov/newsroom/taxpayers-should-know-the-difference-between-standard-and-itemized-deductions
        income : int, optional
            Gross income. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Utah:
        - https://incometax.utah.gov/paying/tax-rates
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
        
        Delaware:
        -         
        - https://revenue.delaware.gov/software-developer/tax-rate-changes/
          https://revenuefiles.delaware.gov/2022/PIT-RES_TY22_2022-01_PaperInteractive.pdf
          https://revenuefiles.delaware.gov/2022/PIT-RES_TY22_2022-02_Instructions.pdf
          https://revenue.delaware.gov/frequently-asked-questions/personal-income-tax-faqs/
          https://www.newcastlede.gov/DocumentCenter/View/20147/City-of-Wilmington-Wage-Tax-Information-PDF
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        federal_agi = self.federal_agi if hasattr(self, 'federal_agi') else federal_agi
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Utah', 'state_occupation': 'Delaware', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "federal_agi": federal_agi, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def utah_and_florida(self, state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Utah:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Florida:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Utah:
        - https://incometax.utah.gov/paying/tax-rates
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
        
        Florida:
        -         
        - https://floridarevenue.com/taxes/taxesfees/Pages/tax_interest_rates.aspx
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Utah', 'state_occupation': 'Florida', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def utah_and_georgia(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Utah:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Georgia:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Utah:
        - https://incometax.utah.gov/paying/tax-rates
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
        
        Georgia:
        -         
        - https://dor.georgia.gov/tax-tables-georgia-tax-rate-schedule
          https://apps.dor.ga.gov/FillableForms/PDFViewer/Index?form=2022GA500
          https://apps.dor.ga.gov/FillableForms/PDFViewer/Index?form=2022GA500
          https://apps.dor.ga.gov/FillableForms/PDFViewer/Index?form=2022GA500
          https://dor.georgia.gov/years-individual-income-tax-forms
          https://dor.georgia.gov/years-individual-income-tax-forms
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Utah', 'state_occupation': 'Georgia', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def utah_and_hawaii(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Utah:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Hawaii:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Utah:
        - https://incometax.utah.gov/paying/tax-rates
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
        
        Hawaii:
        -         
        - https://tax.hawaii.gov/forms/d_18table-on/d_18table-on_p13/
          https://files.hawaii.gov/tax/forms/2018/18table-on.pdf
          https://files.hawaii.gov/tax/news/pubs/22outline.pdf
          https://files.hawaii.gov/tax/news/pubs/FastTaxRefGuide_2017.pdf
          https://www.tfhawaii.org/wordpress/state-tax-resources/mini-tax-guides/individual-income-tax-chapter-235/#:~:text=Long%20term%20capital%20gains%20are%20taxed%20at%20a%20maximum%20of%207.25%25
          https://www.grassrootinstitute.org/2023/02/hb337-increasing-tax-on-capital-gains-a-bad-risk/
          https://files.hawaii.gov/tax/news/pubs/22outline.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Utah', 'state_occupation': 'Hawaii', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def utah_and_idaho(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Utah:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Idaho:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Utah:
        - https://incometax.utah.gov/paying/tax-rates
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
        
        Idaho:
        -         
        - https://tax.idaho.gov/taxes/income-tax/individual-income/individual-income-tax-rate-schedule/
          https://tax.idaho.gov/wp-content/uploads/forms/EIN00046/EIN00046_03-01-2023.pdf
          https://tax.idaho.gov/taxes/income-tax/individual-income/idaho-source-income/
          https://tax.idaho.gov/taxes/income-tax/individual-income/idaho-source-income/
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Utah', 'state_occupation': 'Idaho', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def utah_and_illinois(self, state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Utah:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Illinois:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Utah:
        - https://incometax.utah.gov/paying/tax-rates
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
        
        Illinois:
        -         
        - https://tax.illinois.gov/research/taxrates/income.html#IndividualIncome
          https://tax.illinois.gov/content/dam/soi/en/web/tax/forms/incometax/documents/currentyear/individual/il-1040-instr.pdf
          https://tax.illinois.gov/questionsandanswers/12.html
          https://tax.illinois.gov/content/dam/soi/en/web/tax/forms/incometax/documents/currentyear/individual/il-1040-instr.pdf
          https://tax.illinois.gov/research/taxrates/personalproperty.html
          https://tax.illinois.gov/localgovernments/income.html
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Utah', 'state_occupation': 'Illinois', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def utah_and_indiana(self, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Utah:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Indiana:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Utah:
        - https://incometax.utah.gov/paying/tax-rates
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
        
        Indiana:
        -         
        - https://www.in.gov/dor/tax-forms/2022-individual-income-tax-forms/
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-32.htm
          https://www.in.gov/dor/files/ib28.pdf
          https://www.in.gov/dor/tax-forms/2022-individual-income-tax-forms/
          https://www.in.gov/dor/files/2022-county-tax-rates-and-codes.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Utah', 'state_occupation': 'Indiana', "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def utah_and_iowa(self, county_residence='Other', state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Utah:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Iowa:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Utah:
        - https://incometax.utah.gov/paying/tax-rates
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
        
        Iowa:
        -         
        - https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
          https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
          https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
          https://tax.iowa.gov/iowa-illinois-reciprocal-agreement
          https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Utah', 'state_occupation': 'Iowa', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def utah_and_kansas(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Utah:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Kansas:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Utah:
        - https://incometax.utah.gov/paying/tax-rates
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
        
        Kansas:
        -         
        - https://www.ksrevenue.gov/taxrates.html
          https://www.ksrevenue.gov/incomebook22.html#0
          https://www.ksrevenue.gov/incomebook22.html#0
          https://www.ksrevenue.gov/incomebook22.html#0
          https://www.ksrevenue.gov/incomebook22.html
          https://www.ksrevenue.gov/incomebook22.html
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Utah', 'state_occupation': 'Kansas', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def utah_and_kentucky(self, county_residence='Other', state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Utah:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Kentucky:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Utah:
        - https://incometax.utah.gov/paying/tax-rates
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
        
        Kentucky:
        -         
        - https://revenue.ky.gov/News/Pages/Dor-Announces-Updates-to-Individual-Income-Tax-for-2023-Tax-Year.aspx
          https://revenue.ky.gov/Individual/Individual-Income-Tax/Pages/default.aspx
          https://revenue.ky.gov/News/Pages/Dor-Announces-Updates-to-Individual-Income-Tax-for-2023-Tax-Year.aspx
          https://revenue.ky.gov/Individual/Individual-Income-Tax/Pages/default.aspx
          https://revenue.ky.gov/Dor%20Training%20Materials/103%20KAR%2017.140.%20Individual%20income%20tax%20-%20reciprocity%20-%20nonresidents.pdf
          https://revenue.ky.gov/Individual/Individual-Income-Tax/Pages/default.aspx
          https://nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2021/Taxes-21-32.htm?taxmap=true
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Utah', 'state_occupation': 'Kentucky', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def utah_and_louisiana(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Utah:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Louisiana:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Utah:
        - https://incometax.utah.gov/paying/tax-rates
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
        
        Louisiana:
        -         
        - https://revenue.louisiana.gov/individualincometax
          https://revenue.louisiana.gov/individualincometax
          https://www.revenue.louisiana.gov/taxforms/6935(11_02)F.pdf
          https://revenue.louisiana.gov/individualincometax
          https://revenue.louisiana.gov/individualincometax
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Utah', 'state_occupation': 'Louisiana', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def utah_and_maine(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Utah:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Maine:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Utah:
        - https://incometax.utah.gov/paying/tax-rates
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
        
        Maine:
        -         
        - https://www.maine.gov/revenue/tax-return-forms/individual-income-tax-2022
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/ind_tax_rate_sched_2022.pdf
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/22_1040me_book_gen_instruc_revisedFeb23.pdf
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/22_1040me_book_gen_instruc_revisedFeb23.pdf
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/22_1040me_book_gen_instruc_revisedFeb23.pdf
          https://www.maine.gov/revenue/tax-return-forms/individual-income-tax-2022
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Utah', 'state_occupation': 'Maine', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def utah_and_maryland(self, county_residence='Other', state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Utah:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Maryland:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Utah:
        - https://incometax.utah.gov/paying/tax-rates
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
        
        Maryland:
        -         
        - https://www.marylandtaxes.gov/individual/income/tax-info/tax-rates.php
          https://www.marylandtaxes.gov/forms/22_forms/Resident_Booklet.pdf
          https://www.marylandtaxes.gov/forms/Tax_Publications/Administrative_Releases/Income_and_Estate_Tax_Releases/ar_it37.pdf
          https://www.marylandtaxes.gov/individual/index.php
          https://www.marylandtaxes.gov/individual/income/tax-info/tax-rates.php
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Utah', 'state_occupation': 'Maryland', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def utah_and_massachusetts(self, state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Utah:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Massachusetts:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Utah:
        - https://incometax.utah.gov/paying/tax-rates
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
        
        Massachusetts:
        -         
        - https://www.mass.gov/service-details/massachusetts-tax-rates
          https://www.mass.gov/service-details/view-massachusetts-personal-income-tax-exemptions
          https://www.mass.gov/service-details/view-massachusetts-personal-income-tax-exemptions
          https://www.mass.gov/service-details/view-child-and-dependent-related-credits
          https://www.mass.gov/service-details/massachusetts-tax-rates
          https://www.mass.gov/personal-income-tax
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Utah', 'state_occupation': 'Massachusetts', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def utah_and_michigan(self, county_occupation='Other', state_agi=75000, county_residence='Other', capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Utah:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Michigan:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        county_occupation : str, optional
            County of occupation. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Utah:
        - https://incometax.utah.gov/paying/tax-rates
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
        
        Michigan:
        -         
        - https://www.michigan.gov/taxes/iit/new-developments-for-tax-year-2022
          "https://www.michigan.gov/taxes/-/media/Project/Websites/taxes/Forms/2023/2023-SUW/446_Withholding-Guide_2023.pdf?rev=177ca3d8cb034392bf84e317ed78c320&hash=19C5F42F43A89A91863D229349F48883#:~:text=The%20exemption%20amount%20is%20%245%2C400
         employees%20Michigan%20income%20tax%20return"
          "https://www.michigan.gov/taxes/-/media/Project/Websites/taxes/Forms/2023/2023-SUW/446_Withholding-Guide_2023.pdf?rev=177ca3d8cb034392bf84e317ed78c320&hash=19C5F42F43A89A91863D229349F48883#:~:text=The%20exemption%20amount%20is%20%245%2C400
         employees%20Michigan%20income%20tax%20return"
          https://www.michigan.gov/taxes/questions/iit/accordion/residency/are-my-wages-earned-in-another-state-taxable-in-michigan-if-i-am-a-michigan-resident-1
          https://www.michigan.gov/taxes/-/media/Project/Websites/taxes/Forms/2022/2022-IIT-Forms/BOOK_MI-1040.pdf?rev=4524f1cbbc0243b194ab05a6680be75b&hash=40DFE71E651F7C1D65C31F6CD03D944F
          https://www.michigan.gov/taxes/questions/iit/accordion/general/what-cities-impose-an-income-tax
        """
        
        # Check if the parameters are set upstream in the instance
        county_occupation = self.county_occupation if hasattr(self, 'county_occupation') else county_occupation
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Utah', 'state_occupation': 'Michigan', "county_occupation": county_occupation, "state_agi": state_agi, "county_residence": county_residence, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def utah_and_minnesota(self, state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Utah:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Minnesota:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Utah:
        - https://incometax.utah.gov/paying/tax-rates
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
        
        Minnesota:
        -         
        - https://www.revenue.state.mn.us/minnesota-income-tax-rates-and-brackets
          https://www.revenue.state.mn.us/child-and-dependent-care-credit
          https://www.revenue.state.mn.us/dependent-exemptions
          https://www.revenue.state.mn.us/minnesota-standard-deduction
          https://mn.gov/mmb/assets/mwr_form_tcm1059-128581.pdf
          https://www.revenue.state.mn.us/individual-income-tax
          https://www.revenue.state.mn.us/individual-income-tax
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Utah', 'state_occupation': 'Minnesota', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def utah_and_mississippi(self, state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Utah:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Mississippi:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Utah:
        - https://incometax.utah.gov/paying/tax-rates
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
        
        Mississippi:
        -         
        - https://www.dor.ms.gov/individual/tax-rates
          https://www.dor.ms.gov/individual/tax-rates
          https://www.dor.ms.gov/individual/tax-rates
          https://www.dor.ms.gov/individual
          https://www.dor.ms.gov/individual
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Utah', 'state_occupation': 'Mississippi', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def utah_and_missouri(self, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Utah:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Missouri:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Utah:
        - https://incometax.utah.gov/paying/tax-rates
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
        
        Missouri:
        -         
        - https://dor.mo.gov/taxation/individual/tax-types/income/year-changes/
          https://dor.mo.gov/taxation/individual/tax-types/income/year-changes/
          https://dor.mo.gov/taxation/individual/tax-types/income/year-changes/
          https://www.kcmo.gov/city-hall/departments/finance/earnings-tax
          https://www.stlouis-mo.gov/government/departments/collector/earnings-tax/individual-earnings-tax-info.cfm
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Utah', 'state_occupation': 'Missouri', "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def utah_and_montana(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Utah:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Montana:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Utah:
        - https://incometax.utah.gov/paying/tax-rates
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
        
        Montana:
        -         
        - https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/2022/12/montana-employees-withholding-allowance-and-exemption-certificate-form-mw-4.pdf
          https://mtrevenue.gov/2021/11/29/simplification-of-montana-income-taxation/?utm_source=rss&utm_medium=rss&utm_campaign=simplification-of-montana-income-taxation
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Utah', 'state_occupation': 'Montana', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def utah_and_nebraska(self, state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Utah:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Nebraska:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Utah:
        - https://incometax.utah.gov/paying/tax-rates
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
        
        Nebraska:
        -         
        - https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2023_Circular_EN_8-429r2023_Whole_Book_11.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Utah', 'state_occupation': 'Nebraska', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def utah_and_nevada(self, state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Utah:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Nevada:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Utah:
        - https://incometax.utah.gov/paying/tax-rates
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
        
        Nevada:
        -         
        - https://www.nvsos.gov/sos/businesses/the-nevada-advantage
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Utah', 'state_occupation': 'Nevada', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def utah_and_new_hampshire(self, state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Utah:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        New Hampshire:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Utah:
        - https://incometax.utah.gov/paying/tax-rates
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
        
        New Hampshire:
        -         
        - https://www.revenue.nh.gov/assistance/tax-overview.htm
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Utah', 'state_occupation': 'NewHampshire', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def utah_and_new_jersey(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Utah:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        New Jersey:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Utah:
        - https://incometax.utah.gov/paying/tax-rates
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
        
        New Jersey:
        -         
        - https://www.state.nj.us/treasury/taxation/taxtables.shtml
          https://www.state.nj.us/treasury/taxation/njit13.shtml
          https://www.nj.gov/treasury/taxation/njit25.shtml
          https://www.state.nj.us/treasury/taxation/individuals/freqqiti.shtml
          https://www.state.nj.us/treasury/taxation/localtax.shtml
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Utah', 'state_occupation': 'NewJersey', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def utah_and_new_mexico(self, state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Utah:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        New Mexico:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Utah:
        - https://incometax.utah.gov/paying/tax-rates
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
        
        New Mexico:
        -         
        - https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/fdf3c548-8aba-4b9c-9eb4-bb564c716015/FYI-104.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/0bf943fd-652e-4400-bb1b-cd8397eb5a95/2022pit-adj.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/2f1a6781-9534-4436-b427-1557f9592099/2022pit-adj-ins.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/0bf943fd-652e-4400-bb1b-cd8397eb5a95/2022pit-adj.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/1afc56af-ea90-4d48-82e5-1f9aeb43255a/PITbook2022.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/fdf3c548-8aba-4b9c-9eb4-bb564c716015/FYI-104.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Utah', 'state_occupation': 'NewMexico', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def utah_and_new_york(self, county_residence='Other', state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Utah:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        New York:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Utah:
        - https://incometax.utah.gov/paying/tax-rates
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
        
        New York:
        -         
        - https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nys-tax-rate-schedule
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nys-tax-rate-schedule
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nys-tax-rate-schedule
          https://www.tax.ny.gov/pit/
          https://www.tax.ny.gov/pit/file/nonresident-faqs.htm
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#yonkers-tax-rate-schedule
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nyc-tax-rate-schedule
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Utah', 'state_occupation': 'NewYork', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def utah_and_north_carolina(self, state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Utah:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        North Carolina:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Utah:
        - https://incometax.utah.gov/paying/tax-rates
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
        
        North Carolina:
        -         
        - https://www.ncdor.gov/taxes-forms/tax-rate-schedules
          https://www.ncdor.gov/taxes-forms/individual-income-tax/north-carolina-standard-deduction-or-north-carolina-itemized-deductions
          https://www.ncdor.gov/taxes-forms/individual-income-tax
          https://www.ncdor.gov/taxes-forms/individual-income-tax
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Utah', 'state_occupation': 'NorthCarolina', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def utah_and_north_dakota(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Utah:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        North Dakota:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Utah:
        - https://incometax.utah.gov/paying/tax-rates
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
        
        North Dakota:
        -         
        - https://www.tax.nd.gov/forms
          https://www.tax.nd.gov/sites/www/files/documents/forms/individual/2022-iit/2022-individual-income-tax-booklet.pdf
          https://www.tax.nd.gov/sites/www/files/documents/forms/2022-individual-income-tax-booklet.pdf
          https://www.tax.nd.gov/tax-exemptions-credits/income-tax-exemptions-credits
          https://www.tax.nd.gov/sites/www/files/documents/forms/business/ndwrfillable.pdf
          https://www.tax.nd.gov/news/tax-legislative-changes/significant-changes-law/individual-income-tax-history
          https://www.tax.nd.gov/forms
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Utah', 'state_occupation': 'NorthDakota', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def utah_and_ohio(self, county_residence='Other', state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Utah:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Ohio:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Utah:
        - https://incometax.utah.gov/paying/tax-rates
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
        
        Ohio:
        -         
        - https://tax.ohio.gov/individual/resources/annual-tax-rates
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://tax.ohio.gov/static/forms/employer_withholding/generic/wth_it4nr.pdf
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://www.ritaohio.com/TaxRatesTable
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Utah', 'state_occupation': 'Ohio', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def utah_and_oklahoma(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Utah:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Oklahoma:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Utah:
        - https://incometax.utah.gov/paying/tax-rates
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
        
        Oklahoma:
        -         
        - https://oklahoma.gov/content/dam/ok/en/tax/documents/resources/publications/businesses/withholding-tables/WHTables-2022.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Utah', 'state_occupation': 'Oklahoma', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def utah_and_oregon(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Utah:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Oregon:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Utah:
        - https://incometax.utah.gov/paying/tax-rates
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
        
        Oregon:
        -         
        - https://www.oregon.gov/dor/programs/individuals/Pages/PIT.aspx
          https://www.oregon.gov/dor/programs/individuals/Documents/Full%20year%20resident
         %20Form%20or-40%20filers.pdf
          https://www.oregon.gov/dor/programs/individuals/Documents/Part-year%20and%20nonresident
         %20Form%20or-40-P%20and%20or-40-N%20filers.pdf.pdf
          https://www.oregon.gov/dor/forms/FormsPubs/form-or-40_101-040_2022.pdf
          https://www.oregon.gov/dor/forms/FormsPubs/form-or-40_101-040_2022.pdf
          https://www.oregon.gov/dor/forms/FormsPubs/form-or-40_101-040_2022.pdf
          https://www.portland.gov/revenue/personal-tax
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Utah', 'state_occupation': 'Oregon', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def utah_and_pennsylvania(self, county_residence='Other', state_agi=75000, county_occupation='Other', capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Utah:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Pennsylvania:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        county_residence : str, optional
            County of residence. Default is Other.
        county_occupation : str, optional
            County of occupation. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Utah:
        - https://incometax.utah.gov/paying/tax-rates
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
        
        Pennsylvania:
        -         
        - https://www.revenue.pa.gov/Tax%20Rates/Pages/default.aspx
          https://www.revenue.pa.gov/FormsandPublications/FormsforIndividuals/PIT/Documents/2022/2022_pa-40in.pdf
          https://www.revenue.pa.gov/FormsandPublications/FormsforBusinesses/EmployerWithholding/Documents/rev-419.pdf
          https://www.revenue.pa.gov/FormsandPublications/FormsforIndividuals/PIT/Documents/2022/2022_pa-40in.pdf
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-25.htm
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        county_occupation = self.county_occupation if hasattr(self, 'county_occupation') else county_occupation
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Utah', 'state_occupation': 'Pennsylvania', "county_residence": county_residence, "state_agi": state_agi, "county_occupation": county_occupation, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def utah_and_rhode_island(self, state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Utah:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Rhode Island:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Utah:
        - https://incometax.utah.gov/paying/tax-rates
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
        
        Rhode Island:
        -         
        - https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2022_1040WE_w_0.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Utah', 'state_occupation': 'RhodeIsland', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def utah_and_south_carolina(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Utah:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        South Carolina:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Utah:
        - https://incometax.utah.gov/paying/tax-rates
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
        
        South Carolina:
        -         
        - https://dor.sc.gov/resources-site/lawandpolicy/Advisory%20Opinions/IL22-15.pdf
          https://dor.sc.gov/forms-site/Forms/IITPacket_2022.pdf
          https://www.scstatehouse.gov/code/t12c006.php
          https://dor.sc.gov/resources-site/lawandpolicy/Advisory%20Opinions/IL22-15.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Utah', 'state_occupation': 'SouthCarolina', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def utah_and_south_dakota(self, state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Utah:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        South Dakota:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Utah:
        - https://incometax.utah.gov/paying/tax-rates
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
        
        South Dakota:
        -         
        - https://dor.sd.gov/individuals/taxes/
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Utah', 'state_occupation': 'SouthDakota', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def utah_and_tennessee(self, state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Utah:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Tennessee:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Utah:
        - https://incometax.utah.gov/paying/tax-rates
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
        
        Tennessee:
        -         
        - https://www.tn.gov/revenue/taxes.html
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Utah', 'state_occupation': 'Tennessee', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def utah_and_texas(self, state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Utah:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Texas:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Utah:
        - https://incometax.utah.gov/paying/tax-rates
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
        
        Texas:
        -         
        - https://comptroller.texas.gov/taxes/
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Utah', 'state_occupation': 'Texas', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def utah_and_vermont(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Utah:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Vermont:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Utah:
        - https://incometax.utah.gov/paying/tax-rates
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
        
        Vermont:
        -         
        - https://tax.vermont.gov/sites/tax/files/documents/GB-1210-2023.pdf
          https://tax.vermont.gov/sites/tax/files/documents/Income%20Booklet-2022.pdf
          https://tax.vermont.gov/sites/tax/files/documents/Income%20Booklet-2022.pdf
          https://tax.vermont.gov/sites/tax/files/documents/Income%20Booklet-2022.pdf
          https://tax.vermont.gov/individuals/personal-income-tax/tax-credits
          https://tax.vermont.gov/individuals/personal-income-tax/taxable-income
          https://tax.vermont.gov/individuals/personal-income-tax/local-option-tax
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Utah', 'state_occupation': 'Vermont', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def utah_and_virginia(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Utah:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Virginia:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Utah:
        - https://incometax.utah.gov/paying/tax-rates
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
        
        Virginia:
        -         
        - https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/reciprocity
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Utah', 'state_occupation': 'Virginia', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def utah_and_washington(self, state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Utah:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Washington:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Utah:
        - https://incometax.utah.gov/paying/tax-rates
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
        
        Washington:
        -         
        - https://dor.wa.gov/taxes-rates/income-tax
          https://dor.wa.gov/taxes-rates/other-taxes/capital-gains-tax
          https://dor.wa.gov/taxes-rates/income-tax
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Utah', 'state_occupation': 'Washington', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def utah_and_west_virginia(self, county_occupation='Other', state_agi=75000, pay_periods=24, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Utah:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        West Virginia:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_occupation : str, optional
            County of occupation. Default is Other.
        pay_periods : int, optional
            Number of pay periods. Options are 13, 24, 26, 52. Default is 24.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Utah:
        - https://incometax.utah.gov/paying/tax-rates
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
        
        West Virginia:
        -         
        - https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://tax.wv.gov/Documents/TSD/tsd438.pdf
          https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2011/Taxes-11-35att.html
        """
        
        # Check if the parameters are set upstream in the instance
        county_occupation = self.county_occupation if hasattr(self, 'county_occupation') else county_occupation
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        pay_periods = self.pay_periods if hasattr(self, 'pay_periods') else pay_periods
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Utah', 'state_occupation': 'WestVirginia', "county_occupation": county_occupation, "state_agi": state_agi, "pay_periods": pay_periods, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def utah_and_wisconsin(self, state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Utah:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Wisconsin:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Utah:
        - https://incometax.utah.gov/paying/tax-rates
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
        
        Wisconsin:
        -         
        - https://www.revenue.wi.gov/TaxForms2023/2023-Form1-ES-Inst.pdf
          https://www.revenue.wi.gov/TaxForms2023/2023-Form1-ES-Inst.pdf
          https://www.revenue.wi.gov/TaxForms2023/2023-Form1-ES-Inst.pdf
          https://www.revenue.wi.gov/Pages/FAQS/pcs-work.aspx
          https://www.revenue.wi.gov/Dor%20Publications/pb103.pdf
          https://www.revenue.wi.gov/Pages/FAQS/pcs-work.aspx
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Utah', 'state_occupation': 'Wisconsin', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def utah_and_wyoming(self, state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Utah:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Wyoming:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Utah:
        - https://incometax.utah.gov/paying/tax-rates
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
        
        Wyoming:
        -         
        - https://revenue.wyo.gov/
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Utah', 'state_occupation': 'Wyoming', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def vermont_and_alabama(self, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Vermont:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Alabama:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Vermont:
        - https://tax.vermont.gov/sites/tax/files/documents/GB-1210-2023.pdf
          https://tax.vermont.gov/sites/tax/files/documents/Income%20Booklet-2022.pdf
          https://tax.vermont.gov/sites/tax/files/documents/Income%20Booklet-2022.pdf
          https://tax.vermont.gov/sites/tax/files/documents/Income%20Booklet-2022.pdf
          https://tax.vermont.gov/individuals/personal-income-tax/tax-credits
          https://tax.vermont.gov/individuals/personal-income-tax/taxable-income
          https://tax.vermont.gov/individuals/personal-income-tax/local-option-tax
        
        Alabama:
        -         
        - https://www.revenue.alabama.gov/ultraviewer/viewer/basic_viewer/index.html?form=2023/02/22f40.pdf
          https://www.revenue.alabama.gov/faqs/what-is-alabamas-individual-income-tax-rate/
          https://www.revenue.alabama.gov/ultraviewer/viewer/basic_viewer/index.html?form=2023/02/22f40.pdf
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-24.htm?taxmap=true
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-24.htm?taxmap=true
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-24.htm?taxmap=true
          https://www.revenue.alabama.gov/individual-corporate/income-to-be-reported-on-the-alabama-income-tax-return/
          https://www.almonline.org/TaxRates.aspx#OccupationalTax
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Vermont', 'state_occupation': 'Alabama', "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def vermont_and_alaska(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Vermont:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Alaska:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Vermont:
        - https://tax.vermont.gov/sites/tax/files/documents/GB-1210-2023.pdf
          https://tax.vermont.gov/sites/tax/files/documents/Income%20Booklet-2022.pdf
          https://tax.vermont.gov/sites/tax/files/documents/Income%20Booklet-2022.pdf
          https://tax.vermont.gov/sites/tax/files/documents/Income%20Booklet-2022.pdf
          https://tax.vermont.gov/individuals/personal-income-tax/tax-credits
          https://tax.vermont.gov/individuals/personal-income-tax/taxable-income
          https://tax.vermont.gov/individuals/personal-income-tax/local-option-tax
        
        Alaska:
        -         
        - https://www.commerce.alaska.gov/web/dcra/officeofthestateassessor/alaskataxfacts.aspx
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Vermont', 'state_occupation': 'Alaska', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def vermont_and_arizona(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Vermont:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Arizona:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        credits_state : int, optional
            Credits for state taxes. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Vermont:
        - https://tax.vermont.gov/sites/tax/files/documents/GB-1210-2023.pdf
          https://tax.vermont.gov/sites/tax/files/documents/Income%20Booklet-2022.pdf
          https://tax.vermont.gov/sites/tax/files/documents/Income%20Booklet-2022.pdf
          https://tax.vermont.gov/sites/tax/files/documents/Income%20Booklet-2022.pdf
          https://tax.vermont.gov/individuals/personal-income-tax/tax-credits
          https://tax.vermont.gov/individuals/personal-income-tax/taxable-income
          https://tax.vermont.gov/individuals/personal-income-tax/local-option-tax
        
        Arizona:
        -         
        - https://azdor.gov/forms/individual/form-140-x-y-tables
          https://azdor.gov/forms/individual/form-140-x-y-tables
          https://azdor.gov/forms/individual/form-140-x-y-tables
          https://azdor.gov/businesses-arizona/withholding-tax/withholding-exceptions
          "https://azdor.gov/individuals/income-tax-filing-assistance/identifying-other-taxable-income#:~:text=In%20regards%20to%20capital%20gains
         the%20individuals%20regular%20tax%20rate.
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Vermont', 'state_occupation': 'Arizona', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def vermont_and_arkansas(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Vermont:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Arkansas:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Vermont:
        - https://tax.vermont.gov/sites/tax/files/documents/GB-1210-2023.pdf
          https://tax.vermont.gov/sites/tax/files/documents/Income%20Booklet-2022.pdf
          https://tax.vermont.gov/sites/tax/files/documents/Income%20Booklet-2022.pdf
          https://tax.vermont.gov/sites/tax/files/documents/Income%20Booklet-2022.pdf
          https://tax.vermont.gov/individuals/personal-income-tax/tax-credits
          https://tax.vermont.gov/individuals/personal-income-tax/taxable-income
          https://tax.vermont.gov/individuals/personal-income-tax/local-option-tax
        
        Arkansas:
        -         
        - https://www.arkansasedc.com/why-arkansas/business-climate/tax-structure/personal-income-tax
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_and_AR1000NR_Instructions.pdf
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_FullYearResidentIndividualIncomeTaxReturn.pdf
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_and_AR1000NR_Instructions.pdf
          https://www.arkansasedc.com/why-arkansas/business-climate/tax-structure/capital-gains-tax-reduction
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_and_AR1000NR_Instructions.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Vermont', 'state_occupation': 'Arkansas', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def vermont_and_california(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Vermont:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        California:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Vermont:
        - https://tax.vermont.gov/sites/tax/files/documents/GB-1210-2023.pdf
          https://tax.vermont.gov/sites/tax/files/documents/Income%20Booklet-2022.pdf
          https://tax.vermont.gov/sites/tax/files/documents/Income%20Booklet-2022.pdf
          https://tax.vermont.gov/sites/tax/files/documents/Income%20Booklet-2022.pdf
          https://tax.vermont.gov/individuals/personal-income-tax/tax-credits
          https://tax.vermont.gov/individuals/personal-income-tax/taxable-income
          https://tax.vermont.gov/individuals/personal-income-tax/local-option-tax
        
        California:
        -         
        - https://www.ftb.ca.gov/forms/2022/2022-540-tax-rate-schedules.pdf
          https://www.ftb.ca.gov/file/personal/deductions/index.html
          https://www.ftb.ca.gov/about-ftb/newsroom/caleitc/eligibility-and-credit-information.html
          https://www.ftb.ca.gov/file/personal/income-types/capital-gains-and-losses.html#:~:text=California%20does%20not%20have%20a
         are%20taxed%20as%20ordinary%20income
          https://www.ftb.ca.gov/file/personal/
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Vermont', 'state_occupation': 'California', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def vermont_and_colorado(self, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Vermont:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Colorado:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Vermont:
        - https://tax.vermont.gov/sites/tax/files/documents/GB-1210-2023.pdf
          https://tax.vermont.gov/sites/tax/files/documents/Income%20Booklet-2022.pdf
          https://tax.vermont.gov/sites/tax/files/documents/Income%20Booklet-2022.pdf
          https://tax.vermont.gov/sites/tax/files/documents/Income%20Booklet-2022.pdf
          https://tax.vermont.gov/individuals/personal-income-tax/tax-credits
          https://tax.vermont.gov/individuals/personal-income-tax/taxable-income
          https://tax.vermont.gov/individuals/personal-income-tax/local-option-tax
        
        Colorado:
        -         
        - https://leg.colorado.gov/agencies/legislative-council-staff/individual-income-tax%C2%A0
          https://leg.colorado.gov/agencies/legislative-council-staff/individual-income-tax%C2%A0
          https://tax.colorado.gov/sites/tax/files/documents/ITT_Capital_Gain_Subtraction_Feb_2022.pdf
          https://www.auroragov.org/business_services/taxes/occupational_privilege_tax
          https://www.denvergov.org/Government/Agencies-Departments-Offices/Agencies-Departments-Offices-Directory/Department-of-Finance/Our-Divisions/Treasury/Business-Tax-Information#:~:text=Occupational%20privilege%20tax
         -%28show%20below%29&text=Each%20taxable%20employee%20is%20liable
         month%20for%20each%20taxable%20employee
          https://www.glendale.co.us/355/Occupational-Privilege-Tax
          https://www.greenwoodvillage.com/1220/Occupational-Privilege-Tax-OPT
          https://ci.sheridan.co.us/288/Occupational-Privilege-Tax
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Vermont', 'state_occupation': 'Colorado', "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def vermont_and_connecticut(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Vermont:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Connecticut:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Vermont:
        - https://tax.vermont.gov/sites/tax/files/documents/GB-1210-2023.pdf
          https://tax.vermont.gov/sites/tax/files/documents/Income%20Booklet-2022.pdf
          https://tax.vermont.gov/sites/tax/files/documents/Income%20Booklet-2022.pdf
          https://tax.vermont.gov/sites/tax/files/documents/Income%20Booklet-2022.pdf
          https://tax.vermont.gov/individuals/personal-income-tax/tax-credits
          https://tax.vermont.gov/individuals/personal-income-tax/taxable-income
          https://tax.vermont.gov/individuals/personal-income-tax/local-option-tax
        
        Connecticut:
        -         
        - https://portal.ct.gov/-/media/DRS/Forms/2022/Income/2022-CT-1040-Instructions_1222.pdf
          https://www.cga.ct.gov/2022/rpt/pdf/2022-R-0108.pdf
          https://www.cga.ct.gov/current/pub/chap_229.htm#sec_12-703
          https://portal.ct.gov/-/media/DRS/Forms/2022/Income/2022-CT-1040-Instructions_1222.pdf
          "https://portal.ct.gov/DRS/Publications/TSSNs/TSSN-29#:~:text=An%20individuals%20net%20capital%20gains
         portion%20of%20Social%20Security%20benefits"
          "https://portal.ct.gov/DRS/Publications/TSSNs/TSSN-29#:~:text=An%20individuals%20net%20capital%20gains
         portion%20of%20Social%20Security%20benefits"
          https://portal.ct.gov/DRS/Individuals/Resident-Income-Tax/Tax-Information
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Vermont', 'state_occupation': 'Connecticut', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def vermont_and_delaware(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, federal_agi=75000, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Vermont:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Delaware:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        federal_agi : int, optional
            Adjusted Gross Income (AGI) for federal taxes. Default is 0. Sources - https://www.irs.gov/e-file-providers/definition-of-adjusted-gross-income, https://www.irs.gov/newsroom/taxpayers-should-know-the-difference-between-standard-and-itemized-deductions
        income : int, optional
            Gross income. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Vermont:
        - https://tax.vermont.gov/sites/tax/files/documents/GB-1210-2023.pdf
          https://tax.vermont.gov/sites/tax/files/documents/Income%20Booklet-2022.pdf
          https://tax.vermont.gov/sites/tax/files/documents/Income%20Booklet-2022.pdf
          https://tax.vermont.gov/sites/tax/files/documents/Income%20Booklet-2022.pdf
          https://tax.vermont.gov/individuals/personal-income-tax/tax-credits
          https://tax.vermont.gov/individuals/personal-income-tax/taxable-income
          https://tax.vermont.gov/individuals/personal-income-tax/local-option-tax
        
        Delaware:
        -         
        - https://revenue.delaware.gov/software-developer/tax-rate-changes/
          https://revenuefiles.delaware.gov/2022/PIT-RES_TY22_2022-01_PaperInteractive.pdf
          https://revenuefiles.delaware.gov/2022/PIT-RES_TY22_2022-02_Instructions.pdf
          https://revenue.delaware.gov/frequently-asked-questions/personal-income-tax-faqs/
          https://www.newcastlede.gov/DocumentCenter/View/20147/City-of-Wilmington-Wage-Tax-Information-PDF
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        federal_agi = self.federal_agi if hasattr(self, 'federal_agi') else federal_agi
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Vermont', 'state_occupation': 'Delaware', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "federal_agi": federal_agi, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def vermont_and_florida(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Vermont:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Florida:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Vermont:
        - https://tax.vermont.gov/sites/tax/files/documents/GB-1210-2023.pdf
          https://tax.vermont.gov/sites/tax/files/documents/Income%20Booklet-2022.pdf
          https://tax.vermont.gov/sites/tax/files/documents/Income%20Booklet-2022.pdf
          https://tax.vermont.gov/sites/tax/files/documents/Income%20Booklet-2022.pdf
          https://tax.vermont.gov/individuals/personal-income-tax/tax-credits
          https://tax.vermont.gov/individuals/personal-income-tax/taxable-income
          https://tax.vermont.gov/individuals/personal-income-tax/local-option-tax
        
        Florida:
        -         
        - https://floridarevenue.com/taxes/taxesfees/Pages/tax_interest_rates.aspx
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Vermont', 'state_occupation': 'Florida', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def vermont_and_georgia(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Vermont:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Georgia:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Vermont:
        - https://tax.vermont.gov/sites/tax/files/documents/GB-1210-2023.pdf
          https://tax.vermont.gov/sites/tax/files/documents/Income%20Booklet-2022.pdf
          https://tax.vermont.gov/sites/tax/files/documents/Income%20Booklet-2022.pdf
          https://tax.vermont.gov/sites/tax/files/documents/Income%20Booklet-2022.pdf
          https://tax.vermont.gov/individuals/personal-income-tax/tax-credits
          https://tax.vermont.gov/individuals/personal-income-tax/taxable-income
          https://tax.vermont.gov/individuals/personal-income-tax/local-option-tax
        
        Georgia:
        -         
        - https://dor.georgia.gov/tax-tables-georgia-tax-rate-schedule
          https://apps.dor.ga.gov/FillableForms/PDFViewer/Index?form=2022GA500
          https://apps.dor.ga.gov/FillableForms/PDFViewer/Index?form=2022GA500
          https://apps.dor.ga.gov/FillableForms/PDFViewer/Index?form=2022GA500
          https://dor.georgia.gov/years-individual-income-tax-forms
          https://dor.georgia.gov/years-individual-income-tax-forms
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Vermont', 'state_occupation': 'Georgia', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def vermont_and_hawaii(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Vermont:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Hawaii:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Vermont:
        - https://tax.vermont.gov/sites/tax/files/documents/GB-1210-2023.pdf
          https://tax.vermont.gov/sites/tax/files/documents/Income%20Booklet-2022.pdf
          https://tax.vermont.gov/sites/tax/files/documents/Income%20Booklet-2022.pdf
          https://tax.vermont.gov/sites/tax/files/documents/Income%20Booklet-2022.pdf
          https://tax.vermont.gov/individuals/personal-income-tax/tax-credits
          https://tax.vermont.gov/individuals/personal-income-tax/taxable-income
          https://tax.vermont.gov/individuals/personal-income-tax/local-option-tax
        
        Hawaii:
        -         
        - https://tax.hawaii.gov/forms/d_18table-on/d_18table-on_p13/
          https://files.hawaii.gov/tax/forms/2018/18table-on.pdf
          https://files.hawaii.gov/tax/news/pubs/22outline.pdf
          https://files.hawaii.gov/tax/news/pubs/FastTaxRefGuide_2017.pdf
          https://www.tfhawaii.org/wordpress/state-tax-resources/mini-tax-guides/individual-income-tax-chapter-235/#:~:text=Long%20term%20capital%20gains%20are%20taxed%20at%20a%20maximum%20of%207.25%25
          https://www.grassrootinstitute.org/2023/02/hb337-increasing-tax-on-capital-gains-a-bad-risk/
          https://files.hawaii.gov/tax/news/pubs/22outline.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Vermont', 'state_occupation': 'Hawaii', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def vermont_and_idaho(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Vermont:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Idaho:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Vermont:
        - https://tax.vermont.gov/sites/tax/files/documents/GB-1210-2023.pdf
          https://tax.vermont.gov/sites/tax/files/documents/Income%20Booklet-2022.pdf
          https://tax.vermont.gov/sites/tax/files/documents/Income%20Booklet-2022.pdf
          https://tax.vermont.gov/sites/tax/files/documents/Income%20Booklet-2022.pdf
          https://tax.vermont.gov/individuals/personal-income-tax/tax-credits
          https://tax.vermont.gov/individuals/personal-income-tax/taxable-income
          https://tax.vermont.gov/individuals/personal-income-tax/local-option-tax
        
        Idaho:
        -         
        - https://tax.idaho.gov/taxes/income-tax/individual-income/individual-income-tax-rate-schedule/
          https://tax.idaho.gov/wp-content/uploads/forms/EIN00046/EIN00046_03-01-2023.pdf
          https://tax.idaho.gov/taxes/income-tax/individual-income/idaho-source-income/
          https://tax.idaho.gov/taxes/income-tax/individual-income/idaho-source-income/
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Vermont', 'state_occupation': 'Idaho', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def vermont_and_illinois(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Vermont:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Illinois:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Vermont:
        - https://tax.vermont.gov/sites/tax/files/documents/GB-1210-2023.pdf
          https://tax.vermont.gov/sites/tax/files/documents/Income%20Booklet-2022.pdf
          https://tax.vermont.gov/sites/tax/files/documents/Income%20Booklet-2022.pdf
          https://tax.vermont.gov/sites/tax/files/documents/Income%20Booklet-2022.pdf
          https://tax.vermont.gov/individuals/personal-income-tax/tax-credits
          https://tax.vermont.gov/individuals/personal-income-tax/taxable-income
          https://tax.vermont.gov/individuals/personal-income-tax/local-option-tax
        
        Illinois:
        -         
        - https://tax.illinois.gov/research/taxrates/income.html#IndividualIncome
          https://tax.illinois.gov/content/dam/soi/en/web/tax/forms/incometax/documents/currentyear/individual/il-1040-instr.pdf
          https://tax.illinois.gov/questionsandanswers/12.html
          https://tax.illinois.gov/content/dam/soi/en/web/tax/forms/incometax/documents/currentyear/individual/il-1040-instr.pdf
          https://tax.illinois.gov/research/taxrates/personalproperty.html
          https://tax.illinois.gov/localgovernments/income.html
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Vermont', 'state_occupation': 'Illinois', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def vermont_and_indiana(self, capital_gains_short=0, dependents=0, sources=False, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_long=0, filing_status='single', income=100000, credits_state=0):
        """
        --- Parameters ----
        
        Vermont:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Indiana:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Vermont:
        - https://tax.vermont.gov/sites/tax/files/documents/GB-1210-2023.pdf
          https://tax.vermont.gov/sites/tax/files/documents/Income%20Booklet-2022.pdf
          https://tax.vermont.gov/sites/tax/files/documents/Income%20Booklet-2022.pdf
          https://tax.vermont.gov/sites/tax/files/documents/Income%20Booklet-2022.pdf
          https://tax.vermont.gov/individuals/personal-income-tax/tax-credits
          https://tax.vermont.gov/individuals/personal-income-tax/taxable-income
          https://tax.vermont.gov/individuals/personal-income-tax/local-option-tax
        
        Indiana:
        -         
        - https://www.in.gov/dor/tax-forms/2022-individual-income-tax-forms/
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-32.htm
          https://www.in.gov/dor/files/ib28.pdf
          https://www.in.gov/dor/tax-forms/2022-individual-income-tax-forms/
          https://www.in.gov/dor/files/2022-county-tax-rates-and-codes.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Vermont', 'state_occupation': 'Indiana', "capital_gains_short": capital_gains_short, "dependents": dependents, "sources": sources, "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def vermont_and_iowa(self, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Vermont:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Iowa:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Vermont:
        - https://tax.vermont.gov/sites/tax/files/documents/GB-1210-2023.pdf
          https://tax.vermont.gov/sites/tax/files/documents/Income%20Booklet-2022.pdf
          https://tax.vermont.gov/sites/tax/files/documents/Income%20Booklet-2022.pdf
          https://tax.vermont.gov/sites/tax/files/documents/Income%20Booklet-2022.pdf
          https://tax.vermont.gov/individuals/personal-income-tax/tax-credits
          https://tax.vermont.gov/individuals/personal-income-tax/taxable-income
          https://tax.vermont.gov/individuals/personal-income-tax/local-option-tax
        
        Iowa:
        -         
        - https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
          https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
          https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
          https://tax.iowa.gov/iowa-illinois-reciprocal-agreement
          https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Vermont', 'state_occupation': 'Iowa', "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def vermont_and_kansas(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Vermont:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Kansas:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Vermont:
        - https://tax.vermont.gov/sites/tax/files/documents/GB-1210-2023.pdf
          https://tax.vermont.gov/sites/tax/files/documents/Income%20Booklet-2022.pdf
          https://tax.vermont.gov/sites/tax/files/documents/Income%20Booklet-2022.pdf
          https://tax.vermont.gov/sites/tax/files/documents/Income%20Booklet-2022.pdf
          https://tax.vermont.gov/individuals/personal-income-tax/tax-credits
          https://tax.vermont.gov/individuals/personal-income-tax/taxable-income
          https://tax.vermont.gov/individuals/personal-income-tax/local-option-tax
        
        Kansas:
        -         
        - https://www.ksrevenue.gov/taxrates.html
          https://www.ksrevenue.gov/incomebook22.html#0
          https://www.ksrevenue.gov/incomebook22.html#0
          https://www.ksrevenue.gov/incomebook22.html#0
          https://www.ksrevenue.gov/incomebook22.html
          https://www.ksrevenue.gov/incomebook22.html
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Vermont', 'state_occupation': 'Kansas', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def vermont_and_kentucky(self, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Vermont:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Kentucky:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Vermont:
        - https://tax.vermont.gov/sites/tax/files/documents/GB-1210-2023.pdf
          https://tax.vermont.gov/sites/tax/files/documents/Income%20Booklet-2022.pdf
          https://tax.vermont.gov/sites/tax/files/documents/Income%20Booklet-2022.pdf
          https://tax.vermont.gov/sites/tax/files/documents/Income%20Booklet-2022.pdf
          https://tax.vermont.gov/individuals/personal-income-tax/tax-credits
          https://tax.vermont.gov/individuals/personal-income-tax/taxable-income
          https://tax.vermont.gov/individuals/personal-income-tax/local-option-tax
        
        Kentucky:
        -         
        - https://revenue.ky.gov/News/Pages/Dor-Announces-Updates-to-Individual-Income-Tax-for-2023-Tax-Year.aspx
          https://revenue.ky.gov/Individual/Individual-Income-Tax/Pages/default.aspx
          https://revenue.ky.gov/News/Pages/Dor-Announces-Updates-to-Individual-Income-Tax-for-2023-Tax-Year.aspx
          https://revenue.ky.gov/Individual/Individual-Income-Tax/Pages/default.aspx
          https://revenue.ky.gov/Dor%20Training%20Materials/103%20KAR%2017.140.%20Individual%20income%20tax%20-%20reciprocity%20-%20nonresidents.pdf
          https://revenue.ky.gov/Individual/Individual-Income-Tax/Pages/default.aspx
          https://nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2021/Taxes-21-32.htm?taxmap=true
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Vermont', 'state_occupation': 'Kentucky', "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def vermont_and_louisiana(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Vermont:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Louisiana:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Vermont:
        - https://tax.vermont.gov/sites/tax/files/documents/GB-1210-2023.pdf
          https://tax.vermont.gov/sites/tax/files/documents/Income%20Booklet-2022.pdf
          https://tax.vermont.gov/sites/tax/files/documents/Income%20Booklet-2022.pdf
          https://tax.vermont.gov/sites/tax/files/documents/Income%20Booklet-2022.pdf
          https://tax.vermont.gov/individuals/personal-income-tax/tax-credits
          https://tax.vermont.gov/individuals/personal-income-tax/taxable-income
          https://tax.vermont.gov/individuals/personal-income-tax/local-option-tax
        
        Louisiana:
        -         
        - https://revenue.louisiana.gov/individualincometax
          https://revenue.louisiana.gov/individualincometax
          https://www.revenue.louisiana.gov/taxforms/6935(11_02)F.pdf
          https://revenue.louisiana.gov/individualincometax
          https://revenue.louisiana.gov/individualincometax
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Vermont', 'state_occupation': 'Louisiana', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def vermont_and_maine(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Vermont:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Maine:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Vermont:
        - https://tax.vermont.gov/sites/tax/files/documents/GB-1210-2023.pdf
          https://tax.vermont.gov/sites/tax/files/documents/Income%20Booklet-2022.pdf
          https://tax.vermont.gov/sites/tax/files/documents/Income%20Booklet-2022.pdf
          https://tax.vermont.gov/sites/tax/files/documents/Income%20Booklet-2022.pdf
          https://tax.vermont.gov/individuals/personal-income-tax/tax-credits
          https://tax.vermont.gov/individuals/personal-income-tax/taxable-income
          https://tax.vermont.gov/individuals/personal-income-tax/local-option-tax
        
        Maine:
        -         
        - https://www.maine.gov/revenue/tax-return-forms/individual-income-tax-2022
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/ind_tax_rate_sched_2022.pdf
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/22_1040me_book_gen_instruc_revisedFeb23.pdf
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/22_1040me_book_gen_instruc_revisedFeb23.pdf
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/22_1040me_book_gen_instruc_revisedFeb23.pdf
          https://www.maine.gov/revenue/tax-return-forms/individual-income-tax-2022
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Vermont', 'state_occupation': 'Maine', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def vermont_and_maryland(self, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Vermont:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Maryland:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Vermont:
        - https://tax.vermont.gov/sites/tax/files/documents/GB-1210-2023.pdf
          https://tax.vermont.gov/sites/tax/files/documents/Income%20Booklet-2022.pdf
          https://tax.vermont.gov/sites/tax/files/documents/Income%20Booklet-2022.pdf
          https://tax.vermont.gov/sites/tax/files/documents/Income%20Booklet-2022.pdf
          https://tax.vermont.gov/individuals/personal-income-tax/tax-credits
          https://tax.vermont.gov/individuals/personal-income-tax/taxable-income
          https://tax.vermont.gov/individuals/personal-income-tax/local-option-tax
        
        Maryland:
        -         
        - https://www.marylandtaxes.gov/individual/income/tax-info/tax-rates.php
          https://www.marylandtaxes.gov/forms/22_forms/Resident_Booklet.pdf
          https://www.marylandtaxes.gov/forms/Tax_Publications/Administrative_Releases/Income_and_Estate_Tax_Releases/ar_it37.pdf
          https://www.marylandtaxes.gov/individual/index.php
          https://www.marylandtaxes.gov/individual/income/tax-info/tax-rates.php
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Vermont', 'state_occupation': 'Maryland', "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def vermont_and_massachusetts(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Vermont:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Massachusetts:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Vermont:
        - https://tax.vermont.gov/sites/tax/files/documents/GB-1210-2023.pdf
          https://tax.vermont.gov/sites/tax/files/documents/Income%20Booklet-2022.pdf
          https://tax.vermont.gov/sites/tax/files/documents/Income%20Booklet-2022.pdf
          https://tax.vermont.gov/sites/tax/files/documents/Income%20Booklet-2022.pdf
          https://tax.vermont.gov/individuals/personal-income-tax/tax-credits
          https://tax.vermont.gov/individuals/personal-income-tax/taxable-income
          https://tax.vermont.gov/individuals/personal-income-tax/local-option-tax
        
        Massachusetts:
        -         
        - https://www.mass.gov/service-details/massachusetts-tax-rates
          https://www.mass.gov/service-details/view-massachusetts-personal-income-tax-exemptions
          https://www.mass.gov/service-details/view-massachusetts-personal-income-tax-exemptions
          https://www.mass.gov/service-details/view-child-and-dependent-related-credits
          https://www.mass.gov/service-details/massachusetts-tax-rates
          https://www.mass.gov/personal-income-tax
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Vermont', 'state_occupation': 'Massachusetts', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def vermont_and_michigan(self, capital_gains_short=0, county_occupation='Other', dependents=0, sources=False, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_long=0, filing_status='single', income=100000, credits_state=0):
        """
        --- Parameters ----
        
        Vermont:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Michigan:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        county_occupation : str, optional
            County of occupation. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Vermont:
        - https://tax.vermont.gov/sites/tax/files/documents/GB-1210-2023.pdf
          https://tax.vermont.gov/sites/tax/files/documents/Income%20Booklet-2022.pdf
          https://tax.vermont.gov/sites/tax/files/documents/Income%20Booklet-2022.pdf
          https://tax.vermont.gov/sites/tax/files/documents/Income%20Booklet-2022.pdf
          https://tax.vermont.gov/individuals/personal-income-tax/tax-credits
          https://tax.vermont.gov/individuals/personal-income-tax/taxable-income
          https://tax.vermont.gov/individuals/personal-income-tax/local-option-tax
        
        Michigan:
        -         
        - https://www.michigan.gov/taxes/iit/new-developments-for-tax-year-2022
          "https://www.michigan.gov/taxes/-/media/Project/Websites/taxes/Forms/2023/2023-SUW/446_Withholding-Guide_2023.pdf?rev=177ca3d8cb034392bf84e317ed78c320&hash=19C5F42F43A89A91863D229349F48883#:~:text=The%20exemption%20amount%20is%20%245%2C400
         employees%20Michigan%20income%20tax%20return"
          "https://www.michigan.gov/taxes/-/media/Project/Websites/taxes/Forms/2023/2023-SUW/446_Withholding-Guide_2023.pdf?rev=177ca3d8cb034392bf84e317ed78c320&hash=19C5F42F43A89A91863D229349F48883#:~:text=The%20exemption%20amount%20is%20%245%2C400
         employees%20Michigan%20income%20tax%20return"
          https://www.michigan.gov/taxes/questions/iit/accordion/residency/are-my-wages-earned-in-another-state-taxable-in-michigan-if-i-am-a-michigan-resident-1
          https://www.michigan.gov/taxes/-/media/Project/Websites/taxes/Forms/2022/2022-IIT-Forms/BOOK_MI-1040.pdf?rev=4524f1cbbc0243b194ab05a6680be75b&hash=40DFE71E651F7C1D65C31F6CD03D944F
          https://www.michigan.gov/taxes/questions/iit/accordion/general/what-cities-impose-an-income-tax
        """
        
        # Check if the parameters are set upstream in the instance
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        county_occupation = self.county_occupation if hasattr(self, 'county_occupation') else county_occupation
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Vermont', 'state_occupation': 'Michigan', "capital_gains_short": capital_gains_short, "county_occupation": county_occupation, "dependents": dependents, "sources": sources, "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def vermont_and_minnesota(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Vermont:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Minnesota:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Vermont:
        - https://tax.vermont.gov/sites/tax/files/documents/GB-1210-2023.pdf
          https://tax.vermont.gov/sites/tax/files/documents/Income%20Booklet-2022.pdf
          https://tax.vermont.gov/sites/tax/files/documents/Income%20Booklet-2022.pdf
          https://tax.vermont.gov/sites/tax/files/documents/Income%20Booklet-2022.pdf
          https://tax.vermont.gov/individuals/personal-income-tax/tax-credits
          https://tax.vermont.gov/individuals/personal-income-tax/taxable-income
          https://tax.vermont.gov/individuals/personal-income-tax/local-option-tax
        
        Minnesota:
        -         
        - https://www.revenue.state.mn.us/minnesota-income-tax-rates-and-brackets
          https://www.revenue.state.mn.us/child-and-dependent-care-credit
          https://www.revenue.state.mn.us/dependent-exemptions
          https://www.revenue.state.mn.us/minnesota-standard-deduction
          https://mn.gov/mmb/assets/mwr_form_tcm1059-128581.pdf
          https://www.revenue.state.mn.us/individual-income-tax
          https://www.revenue.state.mn.us/individual-income-tax
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Vermont', 'state_occupation': 'Minnesota', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def vermont_and_mississippi(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Vermont:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Mississippi:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Vermont:
        - https://tax.vermont.gov/sites/tax/files/documents/GB-1210-2023.pdf
          https://tax.vermont.gov/sites/tax/files/documents/Income%20Booklet-2022.pdf
          https://tax.vermont.gov/sites/tax/files/documents/Income%20Booklet-2022.pdf
          https://tax.vermont.gov/sites/tax/files/documents/Income%20Booklet-2022.pdf
          https://tax.vermont.gov/individuals/personal-income-tax/tax-credits
          https://tax.vermont.gov/individuals/personal-income-tax/taxable-income
          https://tax.vermont.gov/individuals/personal-income-tax/local-option-tax
        
        Mississippi:
        -         
        - https://www.dor.ms.gov/individual/tax-rates
          https://www.dor.ms.gov/individual/tax-rates
          https://www.dor.ms.gov/individual/tax-rates
          https://www.dor.ms.gov/individual
          https://www.dor.ms.gov/individual
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Vermont', 'state_occupation': 'Mississippi', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def vermont_and_missouri(self, capital_gains_short=0, dependents=0, sources=False, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_long=0, filing_status='single', income=100000, credits_state=0):
        """
        --- Parameters ----
        
        Vermont:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Missouri:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Vermont:
        - https://tax.vermont.gov/sites/tax/files/documents/GB-1210-2023.pdf
          https://tax.vermont.gov/sites/tax/files/documents/Income%20Booklet-2022.pdf
          https://tax.vermont.gov/sites/tax/files/documents/Income%20Booklet-2022.pdf
          https://tax.vermont.gov/sites/tax/files/documents/Income%20Booklet-2022.pdf
          https://tax.vermont.gov/individuals/personal-income-tax/tax-credits
          https://tax.vermont.gov/individuals/personal-income-tax/taxable-income
          https://tax.vermont.gov/individuals/personal-income-tax/local-option-tax
        
        Missouri:
        -         
        - https://dor.mo.gov/taxation/individual/tax-types/income/year-changes/
          https://dor.mo.gov/taxation/individual/tax-types/income/year-changes/
          https://dor.mo.gov/taxation/individual/tax-types/income/year-changes/
          https://www.kcmo.gov/city-hall/departments/finance/earnings-tax
          https://www.stlouis-mo.gov/government/departments/collector/earnings-tax/individual-earnings-tax-info.cfm
        """
        
        # Check if the parameters are set upstream in the instance
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Vermont', 'state_occupation': 'Missouri', "capital_gains_short": capital_gains_short, "dependents": dependents, "sources": sources, "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def vermont_and_montana(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Vermont:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Montana:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Vermont:
        - https://tax.vermont.gov/sites/tax/files/documents/GB-1210-2023.pdf
          https://tax.vermont.gov/sites/tax/files/documents/Income%20Booklet-2022.pdf
          https://tax.vermont.gov/sites/tax/files/documents/Income%20Booklet-2022.pdf
          https://tax.vermont.gov/sites/tax/files/documents/Income%20Booklet-2022.pdf
          https://tax.vermont.gov/individuals/personal-income-tax/tax-credits
          https://tax.vermont.gov/individuals/personal-income-tax/taxable-income
          https://tax.vermont.gov/individuals/personal-income-tax/local-option-tax
        
        Montana:
        -         
        - https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/2022/12/montana-employees-withholding-allowance-and-exemption-certificate-form-mw-4.pdf
          https://mtrevenue.gov/2021/11/29/simplification-of-montana-income-taxation/?utm_source=rss&utm_medium=rss&utm_campaign=simplification-of-montana-income-taxation
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Vermont', 'state_occupation': 'Montana', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def vermont_and_nebraska(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Vermont:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Nebraska:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Vermont:
        - https://tax.vermont.gov/sites/tax/files/documents/GB-1210-2023.pdf
          https://tax.vermont.gov/sites/tax/files/documents/Income%20Booklet-2022.pdf
          https://tax.vermont.gov/sites/tax/files/documents/Income%20Booklet-2022.pdf
          https://tax.vermont.gov/sites/tax/files/documents/Income%20Booklet-2022.pdf
          https://tax.vermont.gov/individuals/personal-income-tax/tax-credits
          https://tax.vermont.gov/individuals/personal-income-tax/taxable-income
          https://tax.vermont.gov/individuals/personal-income-tax/local-option-tax
        
        Nebraska:
        -         
        - https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2023_Circular_EN_8-429r2023_Whole_Book_11.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Vermont', 'state_occupation': 'Nebraska', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def vermont_and_nevada(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Vermont:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Nevada:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Vermont:
        - https://tax.vermont.gov/sites/tax/files/documents/GB-1210-2023.pdf
          https://tax.vermont.gov/sites/tax/files/documents/Income%20Booklet-2022.pdf
          https://tax.vermont.gov/sites/tax/files/documents/Income%20Booklet-2022.pdf
          https://tax.vermont.gov/sites/tax/files/documents/Income%20Booklet-2022.pdf
          https://tax.vermont.gov/individuals/personal-income-tax/tax-credits
          https://tax.vermont.gov/individuals/personal-income-tax/taxable-income
          https://tax.vermont.gov/individuals/personal-income-tax/local-option-tax
        
        Nevada:
        -         
        - https://www.nvsos.gov/sos/businesses/the-nevada-advantage
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Vermont', 'state_occupation': 'Nevada', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def vermont_and_new_hampshire(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Vermont:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        New Hampshire:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Vermont:
        - https://tax.vermont.gov/sites/tax/files/documents/GB-1210-2023.pdf
          https://tax.vermont.gov/sites/tax/files/documents/Income%20Booklet-2022.pdf
          https://tax.vermont.gov/sites/tax/files/documents/Income%20Booklet-2022.pdf
          https://tax.vermont.gov/sites/tax/files/documents/Income%20Booklet-2022.pdf
          https://tax.vermont.gov/individuals/personal-income-tax/tax-credits
          https://tax.vermont.gov/individuals/personal-income-tax/taxable-income
          https://tax.vermont.gov/individuals/personal-income-tax/local-option-tax
        
        New Hampshire:
        -         
        - https://www.revenue.nh.gov/assistance/tax-overview.htm
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Vermont', 'state_occupation': 'NewHampshire', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def vermont_and_new_jersey(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Vermont:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        New Jersey:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Vermont:
        - https://tax.vermont.gov/sites/tax/files/documents/GB-1210-2023.pdf
          https://tax.vermont.gov/sites/tax/files/documents/Income%20Booklet-2022.pdf
          https://tax.vermont.gov/sites/tax/files/documents/Income%20Booklet-2022.pdf
          https://tax.vermont.gov/sites/tax/files/documents/Income%20Booklet-2022.pdf
          https://tax.vermont.gov/individuals/personal-income-tax/tax-credits
          https://tax.vermont.gov/individuals/personal-income-tax/taxable-income
          https://tax.vermont.gov/individuals/personal-income-tax/local-option-tax
        
        New Jersey:
        -         
        - https://www.state.nj.us/treasury/taxation/taxtables.shtml
          https://www.state.nj.us/treasury/taxation/njit13.shtml
          https://www.nj.gov/treasury/taxation/njit25.shtml
          https://www.state.nj.us/treasury/taxation/individuals/freqqiti.shtml
          https://www.state.nj.us/treasury/taxation/localtax.shtml
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Vermont', 'state_occupation': 'NewJersey', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def vermont_and_new_mexico(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Vermont:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        New Mexico:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Vermont:
        - https://tax.vermont.gov/sites/tax/files/documents/GB-1210-2023.pdf
          https://tax.vermont.gov/sites/tax/files/documents/Income%20Booklet-2022.pdf
          https://tax.vermont.gov/sites/tax/files/documents/Income%20Booklet-2022.pdf
          https://tax.vermont.gov/sites/tax/files/documents/Income%20Booklet-2022.pdf
          https://tax.vermont.gov/individuals/personal-income-tax/tax-credits
          https://tax.vermont.gov/individuals/personal-income-tax/taxable-income
          https://tax.vermont.gov/individuals/personal-income-tax/local-option-tax
        
        New Mexico:
        -         
        - https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/fdf3c548-8aba-4b9c-9eb4-bb564c716015/FYI-104.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/0bf943fd-652e-4400-bb1b-cd8397eb5a95/2022pit-adj.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/2f1a6781-9534-4436-b427-1557f9592099/2022pit-adj-ins.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/0bf943fd-652e-4400-bb1b-cd8397eb5a95/2022pit-adj.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/1afc56af-ea90-4d48-82e5-1f9aeb43255a/PITbook2022.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/fdf3c548-8aba-4b9c-9eb4-bb564c716015/FYI-104.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Vermont', 'state_occupation': 'NewMexico', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def vermont_and_new_york(self, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Vermont:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        New York:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Vermont:
        - https://tax.vermont.gov/sites/tax/files/documents/GB-1210-2023.pdf
          https://tax.vermont.gov/sites/tax/files/documents/Income%20Booklet-2022.pdf
          https://tax.vermont.gov/sites/tax/files/documents/Income%20Booklet-2022.pdf
          https://tax.vermont.gov/sites/tax/files/documents/Income%20Booklet-2022.pdf
          https://tax.vermont.gov/individuals/personal-income-tax/tax-credits
          https://tax.vermont.gov/individuals/personal-income-tax/taxable-income
          https://tax.vermont.gov/individuals/personal-income-tax/local-option-tax
        
        New York:
        -         
        - https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nys-tax-rate-schedule
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nys-tax-rate-schedule
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nys-tax-rate-schedule
          https://www.tax.ny.gov/pit/
          https://www.tax.ny.gov/pit/file/nonresident-faqs.htm
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#yonkers-tax-rate-schedule
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nyc-tax-rate-schedule
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Vermont', 'state_occupation': 'NewYork', "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def vermont_and_north_carolina(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Vermont:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        North Carolina:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Vermont:
        - https://tax.vermont.gov/sites/tax/files/documents/GB-1210-2023.pdf
          https://tax.vermont.gov/sites/tax/files/documents/Income%20Booklet-2022.pdf
          https://tax.vermont.gov/sites/tax/files/documents/Income%20Booklet-2022.pdf
          https://tax.vermont.gov/sites/tax/files/documents/Income%20Booklet-2022.pdf
          https://tax.vermont.gov/individuals/personal-income-tax/tax-credits
          https://tax.vermont.gov/individuals/personal-income-tax/taxable-income
          https://tax.vermont.gov/individuals/personal-income-tax/local-option-tax
        
        North Carolina:
        -         
        - https://www.ncdor.gov/taxes-forms/tax-rate-schedules
          https://www.ncdor.gov/taxes-forms/individual-income-tax/north-carolina-standard-deduction-or-north-carolina-itemized-deductions
          https://www.ncdor.gov/taxes-forms/individual-income-tax
          https://www.ncdor.gov/taxes-forms/individual-income-tax
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Vermont', 'state_occupation': 'NorthCarolina', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def vermont_and_north_dakota(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Vermont:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        North Dakota:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Vermont:
        - https://tax.vermont.gov/sites/tax/files/documents/GB-1210-2023.pdf
          https://tax.vermont.gov/sites/tax/files/documents/Income%20Booklet-2022.pdf
          https://tax.vermont.gov/sites/tax/files/documents/Income%20Booklet-2022.pdf
          https://tax.vermont.gov/sites/tax/files/documents/Income%20Booklet-2022.pdf
          https://tax.vermont.gov/individuals/personal-income-tax/tax-credits
          https://tax.vermont.gov/individuals/personal-income-tax/taxable-income
          https://tax.vermont.gov/individuals/personal-income-tax/local-option-tax
        
        North Dakota:
        -         
        - https://www.tax.nd.gov/forms
          https://www.tax.nd.gov/sites/www/files/documents/forms/individual/2022-iit/2022-individual-income-tax-booklet.pdf
          https://www.tax.nd.gov/sites/www/files/documents/forms/2022-individual-income-tax-booklet.pdf
          https://www.tax.nd.gov/tax-exemptions-credits/income-tax-exemptions-credits
          https://www.tax.nd.gov/sites/www/files/documents/forms/business/ndwrfillable.pdf
          https://www.tax.nd.gov/news/tax-legislative-changes/significant-changes-law/individual-income-tax-history
          https://www.tax.nd.gov/forms
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Vermont', 'state_occupation': 'NorthDakota', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def vermont_and_ohio(self, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Vermont:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Ohio:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Vermont:
        - https://tax.vermont.gov/sites/tax/files/documents/GB-1210-2023.pdf
          https://tax.vermont.gov/sites/tax/files/documents/Income%20Booklet-2022.pdf
          https://tax.vermont.gov/sites/tax/files/documents/Income%20Booklet-2022.pdf
          https://tax.vermont.gov/sites/tax/files/documents/Income%20Booklet-2022.pdf
          https://tax.vermont.gov/individuals/personal-income-tax/tax-credits
          https://tax.vermont.gov/individuals/personal-income-tax/taxable-income
          https://tax.vermont.gov/individuals/personal-income-tax/local-option-tax
        
        Ohio:
        -         
        - https://tax.ohio.gov/individual/resources/annual-tax-rates
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://tax.ohio.gov/static/forms/employer_withholding/generic/wth_it4nr.pdf
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://www.ritaohio.com/TaxRatesTable
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Vermont', 'state_occupation': 'Ohio', "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def vermont_and_oklahoma(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Vermont:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Oklahoma:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Vermont:
        - https://tax.vermont.gov/sites/tax/files/documents/GB-1210-2023.pdf
          https://tax.vermont.gov/sites/tax/files/documents/Income%20Booklet-2022.pdf
          https://tax.vermont.gov/sites/tax/files/documents/Income%20Booklet-2022.pdf
          https://tax.vermont.gov/sites/tax/files/documents/Income%20Booklet-2022.pdf
          https://tax.vermont.gov/individuals/personal-income-tax/tax-credits
          https://tax.vermont.gov/individuals/personal-income-tax/taxable-income
          https://tax.vermont.gov/individuals/personal-income-tax/local-option-tax
        
        Oklahoma:
        -         
        - https://oklahoma.gov/content/dam/ok/en/tax/documents/resources/publications/businesses/withholding-tables/WHTables-2022.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Vermont', 'state_occupation': 'Oklahoma', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def vermont_and_oregon(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Vermont:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Oregon:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Vermont:
        - https://tax.vermont.gov/sites/tax/files/documents/GB-1210-2023.pdf
          https://tax.vermont.gov/sites/tax/files/documents/Income%20Booklet-2022.pdf
          https://tax.vermont.gov/sites/tax/files/documents/Income%20Booklet-2022.pdf
          https://tax.vermont.gov/sites/tax/files/documents/Income%20Booklet-2022.pdf
          https://tax.vermont.gov/individuals/personal-income-tax/tax-credits
          https://tax.vermont.gov/individuals/personal-income-tax/taxable-income
          https://tax.vermont.gov/individuals/personal-income-tax/local-option-tax
        
        Oregon:
        -         
        - https://www.oregon.gov/dor/programs/individuals/Pages/PIT.aspx
          https://www.oregon.gov/dor/programs/individuals/Documents/Full%20year%20resident
         %20Form%20or-40%20filers.pdf
          https://www.oregon.gov/dor/programs/individuals/Documents/Part-year%20and%20nonresident
         %20Form%20or-40-P%20and%20or-40-N%20filers.pdf.pdf
          https://www.oregon.gov/dor/forms/FormsPubs/form-or-40_101-040_2022.pdf
          https://www.oregon.gov/dor/forms/FormsPubs/form-or-40_101-040_2022.pdf
          https://www.oregon.gov/dor/forms/FormsPubs/form-or-40_101-040_2022.pdf
          https://www.portland.gov/revenue/personal-tax
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Vermont', 'state_occupation': 'Oregon', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def vermont_and_pennsylvania(self, county_residence='Other', state_agi=75000, county_occupation='Other', filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Vermont:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Pennsylvania:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        county_residence : str, optional
            County of residence. Default is Other.
        county_occupation : str, optional
            County of occupation. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Vermont:
        - https://tax.vermont.gov/sites/tax/files/documents/GB-1210-2023.pdf
          https://tax.vermont.gov/sites/tax/files/documents/Income%20Booklet-2022.pdf
          https://tax.vermont.gov/sites/tax/files/documents/Income%20Booklet-2022.pdf
          https://tax.vermont.gov/sites/tax/files/documents/Income%20Booklet-2022.pdf
          https://tax.vermont.gov/individuals/personal-income-tax/tax-credits
          https://tax.vermont.gov/individuals/personal-income-tax/taxable-income
          https://tax.vermont.gov/individuals/personal-income-tax/local-option-tax
        
        Pennsylvania:
        -         
        - https://www.revenue.pa.gov/Tax%20Rates/Pages/default.aspx
          https://www.revenue.pa.gov/FormsandPublications/FormsforIndividuals/PIT/Documents/2022/2022_pa-40in.pdf
          https://www.revenue.pa.gov/FormsandPublications/FormsforBusinesses/EmployerWithholding/Documents/rev-419.pdf
          https://www.revenue.pa.gov/FormsandPublications/FormsforIndividuals/PIT/Documents/2022/2022_pa-40in.pdf
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-25.htm
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        county_occupation = self.county_occupation if hasattr(self, 'county_occupation') else county_occupation
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Vermont', 'state_occupation': 'Pennsylvania', "county_residence": county_residence, "state_agi": state_agi, "county_occupation": county_occupation, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def vermont_and_rhode_island(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Vermont:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Rhode Island:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Vermont:
        - https://tax.vermont.gov/sites/tax/files/documents/GB-1210-2023.pdf
          https://tax.vermont.gov/sites/tax/files/documents/Income%20Booklet-2022.pdf
          https://tax.vermont.gov/sites/tax/files/documents/Income%20Booklet-2022.pdf
          https://tax.vermont.gov/sites/tax/files/documents/Income%20Booklet-2022.pdf
          https://tax.vermont.gov/individuals/personal-income-tax/tax-credits
          https://tax.vermont.gov/individuals/personal-income-tax/taxable-income
          https://tax.vermont.gov/individuals/personal-income-tax/local-option-tax
        
        Rhode Island:
        -         
        - https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2022_1040WE_w_0.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Vermont', 'state_occupation': 'RhodeIsland', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def vermont_and_south_carolina(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Vermont:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        South Carolina:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Vermont:
        - https://tax.vermont.gov/sites/tax/files/documents/GB-1210-2023.pdf
          https://tax.vermont.gov/sites/tax/files/documents/Income%20Booklet-2022.pdf
          https://tax.vermont.gov/sites/tax/files/documents/Income%20Booklet-2022.pdf
          https://tax.vermont.gov/sites/tax/files/documents/Income%20Booklet-2022.pdf
          https://tax.vermont.gov/individuals/personal-income-tax/tax-credits
          https://tax.vermont.gov/individuals/personal-income-tax/taxable-income
          https://tax.vermont.gov/individuals/personal-income-tax/local-option-tax
        
        South Carolina:
        -         
        - https://dor.sc.gov/resources-site/lawandpolicy/Advisory%20Opinions/IL22-15.pdf
          https://dor.sc.gov/forms-site/Forms/IITPacket_2022.pdf
          https://www.scstatehouse.gov/code/t12c006.php
          https://dor.sc.gov/resources-site/lawandpolicy/Advisory%20Opinions/IL22-15.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Vermont', 'state_occupation': 'SouthCarolina', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def vermont_and_south_dakota(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Vermont:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        South Dakota:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Vermont:
        - https://tax.vermont.gov/sites/tax/files/documents/GB-1210-2023.pdf
          https://tax.vermont.gov/sites/tax/files/documents/Income%20Booklet-2022.pdf
          https://tax.vermont.gov/sites/tax/files/documents/Income%20Booklet-2022.pdf
          https://tax.vermont.gov/sites/tax/files/documents/Income%20Booklet-2022.pdf
          https://tax.vermont.gov/individuals/personal-income-tax/tax-credits
          https://tax.vermont.gov/individuals/personal-income-tax/taxable-income
          https://tax.vermont.gov/individuals/personal-income-tax/local-option-tax
        
        South Dakota:
        -         
        - https://dor.sd.gov/individuals/taxes/
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Vermont', 'state_occupation': 'SouthDakota', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def vermont_and_tennessee(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Vermont:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Tennessee:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Vermont:
        - https://tax.vermont.gov/sites/tax/files/documents/GB-1210-2023.pdf
          https://tax.vermont.gov/sites/tax/files/documents/Income%20Booklet-2022.pdf
          https://tax.vermont.gov/sites/tax/files/documents/Income%20Booklet-2022.pdf
          https://tax.vermont.gov/sites/tax/files/documents/Income%20Booklet-2022.pdf
          https://tax.vermont.gov/individuals/personal-income-tax/tax-credits
          https://tax.vermont.gov/individuals/personal-income-tax/taxable-income
          https://tax.vermont.gov/individuals/personal-income-tax/local-option-tax
        
        Tennessee:
        -         
        - https://www.tn.gov/revenue/taxes.html
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Vermont', 'state_occupation': 'Tennessee', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def vermont_and_texas(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Vermont:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Texas:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Vermont:
        - https://tax.vermont.gov/sites/tax/files/documents/GB-1210-2023.pdf
          https://tax.vermont.gov/sites/tax/files/documents/Income%20Booklet-2022.pdf
          https://tax.vermont.gov/sites/tax/files/documents/Income%20Booklet-2022.pdf
          https://tax.vermont.gov/sites/tax/files/documents/Income%20Booklet-2022.pdf
          https://tax.vermont.gov/individuals/personal-income-tax/tax-credits
          https://tax.vermont.gov/individuals/personal-income-tax/taxable-income
          https://tax.vermont.gov/individuals/personal-income-tax/local-option-tax
        
        Texas:
        -         
        - https://comptroller.texas.gov/taxes/
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Vermont', 'state_occupation': 'Texas', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def vermont_and_utah(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Vermont:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Utah:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Vermont:
        - https://tax.vermont.gov/sites/tax/files/documents/GB-1210-2023.pdf
          https://tax.vermont.gov/sites/tax/files/documents/Income%20Booklet-2022.pdf
          https://tax.vermont.gov/sites/tax/files/documents/Income%20Booklet-2022.pdf
          https://tax.vermont.gov/sites/tax/files/documents/Income%20Booklet-2022.pdf
          https://tax.vermont.gov/individuals/personal-income-tax/tax-credits
          https://tax.vermont.gov/individuals/personal-income-tax/taxable-income
          https://tax.vermont.gov/individuals/personal-income-tax/local-option-tax
        
        Utah:
        -         
        - https://incometax.utah.gov/paying/tax-rates
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Vermont', 'state_occupation': 'Utah', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def vermont_and_virginia(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Vermont:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Virginia:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Vermont:
        - https://tax.vermont.gov/sites/tax/files/documents/GB-1210-2023.pdf
          https://tax.vermont.gov/sites/tax/files/documents/Income%20Booklet-2022.pdf
          https://tax.vermont.gov/sites/tax/files/documents/Income%20Booklet-2022.pdf
          https://tax.vermont.gov/sites/tax/files/documents/Income%20Booklet-2022.pdf
          https://tax.vermont.gov/individuals/personal-income-tax/tax-credits
          https://tax.vermont.gov/individuals/personal-income-tax/taxable-income
          https://tax.vermont.gov/individuals/personal-income-tax/local-option-tax
        
        Virginia:
        -         
        - https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/reciprocity
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Vermont', 'state_occupation': 'Virginia', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def vermont_and_washington(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Vermont:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Washington:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Vermont:
        - https://tax.vermont.gov/sites/tax/files/documents/GB-1210-2023.pdf
          https://tax.vermont.gov/sites/tax/files/documents/Income%20Booklet-2022.pdf
          https://tax.vermont.gov/sites/tax/files/documents/Income%20Booklet-2022.pdf
          https://tax.vermont.gov/sites/tax/files/documents/Income%20Booklet-2022.pdf
          https://tax.vermont.gov/individuals/personal-income-tax/tax-credits
          https://tax.vermont.gov/individuals/personal-income-tax/taxable-income
          https://tax.vermont.gov/individuals/personal-income-tax/local-option-tax
        
        Washington:
        -         
        - https://dor.wa.gov/taxes-rates/income-tax
          https://dor.wa.gov/taxes-rates/other-taxes/capital-gains-tax
          https://dor.wa.gov/taxes-rates/income-tax
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Vermont', 'state_occupation': 'Washington', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def vermont_and_west_virginia(self, pay_periods=24, capital_gains_short=0, county_occupation='Other', dependents=0, sources=False, state_agi=75000, filers_over_65=0, capital_gains_long=0, filing_status='single', income=100000, credits_state=0):
        """
        --- Parameters ----
        
        Vermont:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        West Virginia:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_occupation : str, optional
            County of occupation. Default is Other.
        pay_periods : int, optional
            Number of pay periods. Options are 13, 24, 26, 52. Default is 24.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Vermont:
        - https://tax.vermont.gov/sites/tax/files/documents/GB-1210-2023.pdf
          https://tax.vermont.gov/sites/tax/files/documents/Income%20Booklet-2022.pdf
          https://tax.vermont.gov/sites/tax/files/documents/Income%20Booklet-2022.pdf
          https://tax.vermont.gov/sites/tax/files/documents/Income%20Booklet-2022.pdf
          https://tax.vermont.gov/individuals/personal-income-tax/tax-credits
          https://tax.vermont.gov/individuals/personal-income-tax/taxable-income
          https://tax.vermont.gov/individuals/personal-income-tax/local-option-tax
        
        West Virginia:
        -         
        - https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://tax.wv.gov/Documents/TSD/tsd438.pdf
          https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2011/Taxes-11-35att.html
        """
        
        # Check if the parameters are set upstream in the instance
        pay_periods = self.pay_periods if hasattr(self, 'pay_periods') else pay_periods
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        county_occupation = self.county_occupation if hasattr(self, 'county_occupation') else county_occupation
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Vermont', 'state_occupation': 'WestVirginia', "pay_periods": pay_periods, "capital_gains_short": capital_gains_short, "county_occupation": county_occupation, "dependents": dependents, "sources": sources, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def vermont_and_wisconsin(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Vermont:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Wisconsin:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Vermont:
        - https://tax.vermont.gov/sites/tax/files/documents/GB-1210-2023.pdf
          https://tax.vermont.gov/sites/tax/files/documents/Income%20Booklet-2022.pdf
          https://tax.vermont.gov/sites/tax/files/documents/Income%20Booklet-2022.pdf
          https://tax.vermont.gov/sites/tax/files/documents/Income%20Booklet-2022.pdf
          https://tax.vermont.gov/individuals/personal-income-tax/tax-credits
          https://tax.vermont.gov/individuals/personal-income-tax/taxable-income
          https://tax.vermont.gov/individuals/personal-income-tax/local-option-tax
        
        Wisconsin:
        -         
        - https://www.revenue.wi.gov/TaxForms2023/2023-Form1-ES-Inst.pdf
          https://www.revenue.wi.gov/TaxForms2023/2023-Form1-ES-Inst.pdf
          https://www.revenue.wi.gov/TaxForms2023/2023-Form1-ES-Inst.pdf
          https://www.revenue.wi.gov/Pages/FAQS/pcs-work.aspx
          https://www.revenue.wi.gov/Dor%20Publications/pb103.pdf
          https://www.revenue.wi.gov/Pages/FAQS/pcs-work.aspx
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Vermont', 'state_occupation': 'Wisconsin', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def vermont_and_wyoming(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Vermont:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Wyoming:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Vermont:
        - https://tax.vermont.gov/sites/tax/files/documents/GB-1210-2023.pdf
          https://tax.vermont.gov/sites/tax/files/documents/Income%20Booklet-2022.pdf
          https://tax.vermont.gov/sites/tax/files/documents/Income%20Booklet-2022.pdf
          https://tax.vermont.gov/sites/tax/files/documents/Income%20Booklet-2022.pdf
          https://tax.vermont.gov/individuals/personal-income-tax/tax-credits
          https://tax.vermont.gov/individuals/personal-income-tax/taxable-income
          https://tax.vermont.gov/individuals/personal-income-tax/local-option-tax
        
        Wyoming:
        -         
        - https://revenue.wyo.gov/
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Vermont', 'state_occupation': 'Wyoming', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def virginia_and_alabama(self, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Virginia:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Alabama:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Virginia:
        - https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/reciprocity
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
        
        Alabama:
        -         
        - https://www.revenue.alabama.gov/ultraviewer/viewer/basic_viewer/index.html?form=2023/02/22f40.pdf
          https://www.revenue.alabama.gov/faqs/what-is-alabamas-individual-income-tax-rate/
          https://www.revenue.alabama.gov/ultraviewer/viewer/basic_viewer/index.html?form=2023/02/22f40.pdf
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-24.htm?taxmap=true
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-24.htm?taxmap=true
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-24.htm?taxmap=true
          https://www.revenue.alabama.gov/individual-corporate/income-to-be-reported-on-the-alabama-income-tax-return/
          https://www.almonline.org/TaxRates.aspx#OccupationalTax
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Virginia', 'state_occupation': 'Alabama', "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def virginia_and_alaska(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Virginia:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Alaska:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Virginia:
        - https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/reciprocity
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
        
        Alaska:
        -         
        - https://www.commerce.alaska.gov/web/dcra/officeofthestateassessor/alaskataxfacts.aspx
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Virginia', 'state_occupation': 'Alaska', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def virginia_and_arizona(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Virginia:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Arizona:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        credits_state : int, optional
            Credits for state taxes. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Virginia:
        - https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/reciprocity
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
        
        Arizona:
        -         
        - https://azdor.gov/forms/individual/form-140-x-y-tables
          https://azdor.gov/forms/individual/form-140-x-y-tables
          https://azdor.gov/forms/individual/form-140-x-y-tables
          https://azdor.gov/businesses-arizona/withholding-tax/withholding-exceptions
          "https://azdor.gov/individuals/income-tax-filing-assistance/identifying-other-taxable-income#:~:text=In%20regards%20to%20capital%20gains
         the%20individuals%20regular%20tax%20rate.
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Virginia', 'state_occupation': 'Arizona', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def virginia_and_arkansas(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Virginia:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Arkansas:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Virginia:
        - https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/reciprocity
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
        
        Arkansas:
        -         
        - https://www.arkansasedc.com/why-arkansas/business-climate/tax-structure/personal-income-tax
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_and_AR1000NR_Instructions.pdf
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_FullYearResidentIndividualIncomeTaxReturn.pdf
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_and_AR1000NR_Instructions.pdf
          https://www.arkansasedc.com/why-arkansas/business-climate/tax-structure/capital-gains-tax-reduction
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_and_AR1000NR_Instructions.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Virginia', 'state_occupation': 'Arkansas', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def virginia_and_california(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Virginia:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        California:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Virginia:
        - https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/reciprocity
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
        
        California:
        -         
        - https://www.ftb.ca.gov/forms/2022/2022-540-tax-rate-schedules.pdf
          https://www.ftb.ca.gov/file/personal/deductions/index.html
          https://www.ftb.ca.gov/about-ftb/newsroom/caleitc/eligibility-and-credit-information.html
          https://www.ftb.ca.gov/file/personal/income-types/capital-gains-and-losses.html#:~:text=California%20does%20not%20have%20a
         are%20taxed%20as%20ordinary%20income
          https://www.ftb.ca.gov/file/personal/
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Virginia', 'state_occupation': 'California', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def virginia_and_colorado(self, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Virginia:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Colorado:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Virginia:
        - https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/reciprocity
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
        
        Colorado:
        -         
        - https://leg.colorado.gov/agencies/legislative-council-staff/individual-income-tax%C2%A0
          https://leg.colorado.gov/agencies/legislative-council-staff/individual-income-tax%C2%A0
          https://tax.colorado.gov/sites/tax/files/documents/ITT_Capital_Gain_Subtraction_Feb_2022.pdf
          https://www.auroragov.org/business_services/taxes/occupational_privilege_tax
          https://www.denvergov.org/Government/Agencies-Departments-Offices/Agencies-Departments-Offices-Directory/Department-of-Finance/Our-Divisions/Treasury/Business-Tax-Information#:~:text=Occupational%20privilege%20tax
         -%28show%20below%29&text=Each%20taxable%20employee%20is%20liable
         month%20for%20each%20taxable%20employee
          https://www.glendale.co.us/355/Occupational-Privilege-Tax
          https://www.greenwoodvillage.com/1220/Occupational-Privilege-Tax-OPT
          https://ci.sheridan.co.us/288/Occupational-Privilege-Tax
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Virginia', 'state_occupation': 'Colorado', "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def virginia_and_connecticut(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Virginia:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Connecticut:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Virginia:
        - https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/reciprocity
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
        
        Connecticut:
        -         
        - https://portal.ct.gov/-/media/DRS/Forms/2022/Income/2022-CT-1040-Instructions_1222.pdf
          https://www.cga.ct.gov/2022/rpt/pdf/2022-R-0108.pdf
          https://www.cga.ct.gov/current/pub/chap_229.htm#sec_12-703
          https://portal.ct.gov/-/media/DRS/Forms/2022/Income/2022-CT-1040-Instructions_1222.pdf
          "https://portal.ct.gov/DRS/Publications/TSSNs/TSSN-29#:~:text=An%20individuals%20net%20capital%20gains
         portion%20of%20Social%20Security%20benefits"
          "https://portal.ct.gov/DRS/Publications/TSSNs/TSSN-29#:~:text=An%20individuals%20net%20capital%20gains
         portion%20of%20Social%20Security%20benefits"
          https://portal.ct.gov/DRS/Individuals/Resident-Income-Tax/Tax-Information
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Virginia', 'state_occupation': 'Connecticut', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def virginia_and_delaware(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, federal_agi=75000, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Virginia:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Delaware:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        federal_agi : int, optional
            Adjusted Gross Income (AGI) for federal taxes. Default is 0. Sources - https://www.irs.gov/e-file-providers/definition-of-adjusted-gross-income, https://www.irs.gov/newsroom/taxpayers-should-know-the-difference-between-standard-and-itemized-deductions
        income : int, optional
            Gross income. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Virginia:
        - https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/reciprocity
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
        
        Delaware:
        -         
        - https://revenue.delaware.gov/software-developer/tax-rate-changes/
          https://revenuefiles.delaware.gov/2022/PIT-RES_TY22_2022-01_PaperInteractive.pdf
          https://revenuefiles.delaware.gov/2022/PIT-RES_TY22_2022-02_Instructions.pdf
          https://revenue.delaware.gov/frequently-asked-questions/personal-income-tax-faqs/
          https://www.newcastlede.gov/DocumentCenter/View/20147/City-of-Wilmington-Wage-Tax-Information-PDF
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        federal_agi = self.federal_agi if hasattr(self, 'federal_agi') else federal_agi
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Virginia', 'state_occupation': 'Delaware', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "federal_agi": federal_agi, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def virginia_and_florida(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Virginia:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Florida:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Virginia:
        - https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/reciprocity
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
        
        Florida:
        -         
        - https://floridarevenue.com/taxes/taxesfees/Pages/tax_interest_rates.aspx
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Virginia', 'state_occupation': 'Florida', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def virginia_and_georgia(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Virginia:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Georgia:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Virginia:
        - https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/reciprocity
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
        
        Georgia:
        -         
        - https://dor.georgia.gov/tax-tables-georgia-tax-rate-schedule
          https://apps.dor.ga.gov/FillableForms/PDFViewer/Index?form=2022GA500
          https://apps.dor.ga.gov/FillableForms/PDFViewer/Index?form=2022GA500
          https://apps.dor.ga.gov/FillableForms/PDFViewer/Index?form=2022GA500
          https://dor.georgia.gov/years-individual-income-tax-forms
          https://dor.georgia.gov/years-individual-income-tax-forms
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Virginia', 'state_occupation': 'Georgia', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def virginia_and_hawaii(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Virginia:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Hawaii:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Virginia:
        - https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/reciprocity
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
        
        Hawaii:
        -         
        - https://tax.hawaii.gov/forms/d_18table-on/d_18table-on_p13/
          https://files.hawaii.gov/tax/forms/2018/18table-on.pdf
          https://files.hawaii.gov/tax/news/pubs/22outline.pdf
          https://files.hawaii.gov/tax/news/pubs/FastTaxRefGuide_2017.pdf
          https://www.tfhawaii.org/wordpress/state-tax-resources/mini-tax-guides/individual-income-tax-chapter-235/#:~:text=Long%20term%20capital%20gains%20are%20taxed%20at%20a%20maximum%20of%207.25%25
          https://www.grassrootinstitute.org/2023/02/hb337-increasing-tax-on-capital-gains-a-bad-risk/
          https://files.hawaii.gov/tax/news/pubs/22outline.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Virginia', 'state_occupation': 'Hawaii', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def virginia_and_idaho(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Virginia:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Idaho:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Virginia:
        - https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/reciprocity
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
        
        Idaho:
        -         
        - https://tax.idaho.gov/taxes/income-tax/individual-income/individual-income-tax-rate-schedule/
          https://tax.idaho.gov/wp-content/uploads/forms/EIN00046/EIN00046_03-01-2023.pdf
          https://tax.idaho.gov/taxes/income-tax/individual-income/idaho-source-income/
          https://tax.idaho.gov/taxes/income-tax/individual-income/idaho-source-income/
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Virginia', 'state_occupation': 'Idaho', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def virginia_and_illinois(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Virginia:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Illinois:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Virginia:
        - https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/reciprocity
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
        
        Illinois:
        -         
        - https://tax.illinois.gov/research/taxrates/income.html#IndividualIncome
          https://tax.illinois.gov/content/dam/soi/en/web/tax/forms/incometax/documents/currentyear/individual/il-1040-instr.pdf
          https://tax.illinois.gov/questionsandanswers/12.html
          https://tax.illinois.gov/content/dam/soi/en/web/tax/forms/incometax/documents/currentyear/individual/il-1040-instr.pdf
          https://tax.illinois.gov/research/taxrates/personalproperty.html
          https://tax.illinois.gov/localgovernments/income.html
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Virginia', 'state_occupation': 'Illinois', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def virginia_and_indiana(self, capital_gains_short=0, dependents=0, sources=False, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_long=0, filing_status='single', income=100000, credits_state=0):
        """
        --- Parameters ----
        
        Virginia:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Indiana:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Virginia:
        - https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/reciprocity
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
        
        Indiana:
        -         
        - https://www.in.gov/dor/tax-forms/2022-individual-income-tax-forms/
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-32.htm
          https://www.in.gov/dor/files/ib28.pdf
          https://www.in.gov/dor/tax-forms/2022-individual-income-tax-forms/
          https://www.in.gov/dor/files/2022-county-tax-rates-and-codes.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Virginia', 'state_occupation': 'Indiana', "capital_gains_short": capital_gains_short, "dependents": dependents, "sources": sources, "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def virginia_and_iowa(self, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Virginia:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Iowa:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Virginia:
        - https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/reciprocity
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
        
        Iowa:
        -         
        - https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
          https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
          https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
          https://tax.iowa.gov/iowa-illinois-reciprocal-agreement
          https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Virginia', 'state_occupation': 'Iowa', "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def virginia_and_kansas(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Virginia:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Kansas:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Virginia:
        - https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/reciprocity
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
        
        Kansas:
        -         
        - https://www.ksrevenue.gov/taxrates.html
          https://www.ksrevenue.gov/incomebook22.html#0
          https://www.ksrevenue.gov/incomebook22.html#0
          https://www.ksrevenue.gov/incomebook22.html#0
          https://www.ksrevenue.gov/incomebook22.html
          https://www.ksrevenue.gov/incomebook22.html
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Virginia', 'state_occupation': 'Kansas', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def virginia_and_kentucky(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Virginia:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Kentucky:
        


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Virginia:
        - https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/reciprocity
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
        
        Kentucky:
        -         
        - https://revenue.ky.gov/News/Pages/Dor-Announces-Updates-to-Individual-Income-Tax-for-2023-Tax-Year.aspx
          https://revenue.ky.gov/Individual/Individual-Income-Tax/Pages/default.aspx
          https://revenue.ky.gov/News/Pages/Dor-Announces-Updates-to-Individual-Income-Tax-for-2023-Tax-Year.aspx
          https://revenue.ky.gov/Individual/Individual-Income-Tax/Pages/default.aspx
          https://revenue.ky.gov/Dor%20Training%20Materials/103%20KAR%2017.140.%20Individual%20income%20tax%20-%20reciprocity%20-%20nonresidents.pdf
          https://revenue.ky.gov/Individual/Individual-Income-Tax/Pages/default.aspx
          https://nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2021/Taxes-21-32.htm?taxmap=true
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Virginia', 'state_occupation': 'Kentucky', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def virginia_and_louisiana(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Virginia:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Louisiana:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Virginia:
        - https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/reciprocity
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
        
        Louisiana:
        -         
        - https://revenue.louisiana.gov/individualincometax
          https://revenue.louisiana.gov/individualincometax
          https://www.revenue.louisiana.gov/taxforms/6935(11_02)F.pdf
          https://revenue.louisiana.gov/individualincometax
          https://revenue.louisiana.gov/individualincometax
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Virginia', 'state_occupation': 'Louisiana', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def virginia_and_maine(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Virginia:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Maine:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Virginia:
        - https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/reciprocity
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
        
        Maine:
        -         
        - https://www.maine.gov/revenue/tax-return-forms/individual-income-tax-2022
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/ind_tax_rate_sched_2022.pdf
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/22_1040me_book_gen_instruc_revisedFeb23.pdf
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/22_1040me_book_gen_instruc_revisedFeb23.pdf
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/22_1040me_book_gen_instruc_revisedFeb23.pdf
          https://www.maine.gov/revenue/tax-return-forms/individual-income-tax-2022
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Virginia', 'state_occupation': 'Maine', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def virginia_and_maryland(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Virginia:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Maryland:
        


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Virginia:
        - https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/reciprocity
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
        
        Maryland:
        -         
        - https://www.marylandtaxes.gov/individual/income/tax-info/tax-rates.php
          https://www.marylandtaxes.gov/forms/22_forms/Resident_Booklet.pdf
          https://www.marylandtaxes.gov/forms/Tax_Publications/Administrative_Releases/Income_and_Estate_Tax_Releases/ar_it37.pdf
          https://www.marylandtaxes.gov/individual/index.php
          https://www.marylandtaxes.gov/individual/income/tax-info/tax-rates.php
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Virginia', 'state_occupation': 'Maryland', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def virginia_and_massachusetts(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Virginia:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Massachusetts:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Virginia:
        - https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/reciprocity
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
        
        Massachusetts:
        -         
        - https://www.mass.gov/service-details/massachusetts-tax-rates
          https://www.mass.gov/service-details/view-massachusetts-personal-income-tax-exemptions
          https://www.mass.gov/service-details/view-massachusetts-personal-income-tax-exemptions
          https://www.mass.gov/service-details/view-child-and-dependent-related-credits
          https://www.mass.gov/service-details/massachusetts-tax-rates
          https://www.mass.gov/personal-income-tax
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Virginia', 'state_occupation': 'Massachusetts', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def virginia_and_michigan(self, capital_gains_short=0, county_occupation='Other', dependents=0, sources=False, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_long=0, filing_status='single', income=100000, credits_state=0):
        """
        --- Parameters ----
        
        Virginia:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Michigan:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        county_occupation : str, optional
            County of occupation. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Virginia:
        - https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/reciprocity
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
        
        Michigan:
        -         
        - https://www.michigan.gov/taxes/iit/new-developments-for-tax-year-2022
          "https://www.michigan.gov/taxes/-/media/Project/Websites/taxes/Forms/2023/2023-SUW/446_Withholding-Guide_2023.pdf?rev=177ca3d8cb034392bf84e317ed78c320&hash=19C5F42F43A89A91863D229349F48883#:~:text=The%20exemption%20amount%20is%20%245%2C400
         employees%20Michigan%20income%20tax%20return"
          "https://www.michigan.gov/taxes/-/media/Project/Websites/taxes/Forms/2023/2023-SUW/446_Withholding-Guide_2023.pdf?rev=177ca3d8cb034392bf84e317ed78c320&hash=19C5F42F43A89A91863D229349F48883#:~:text=The%20exemption%20amount%20is%20%245%2C400
         employees%20Michigan%20income%20tax%20return"
          https://www.michigan.gov/taxes/questions/iit/accordion/residency/are-my-wages-earned-in-another-state-taxable-in-michigan-if-i-am-a-michigan-resident-1
          https://www.michigan.gov/taxes/-/media/Project/Websites/taxes/Forms/2022/2022-IIT-Forms/BOOK_MI-1040.pdf?rev=4524f1cbbc0243b194ab05a6680be75b&hash=40DFE71E651F7C1D65C31F6CD03D944F
          https://www.michigan.gov/taxes/questions/iit/accordion/general/what-cities-impose-an-income-tax
        """
        
        # Check if the parameters are set upstream in the instance
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        county_occupation = self.county_occupation if hasattr(self, 'county_occupation') else county_occupation
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Virginia', 'state_occupation': 'Michigan', "capital_gains_short": capital_gains_short, "county_occupation": county_occupation, "dependents": dependents, "sources": sources, "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def virginia_and_minnesota(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Virginia:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Minnesota:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Virginia:
        - https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/reciprocity
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
        
        Minnesota:
        -         
        - https://www.revenue.state.mn.us/minnesota-income-tax-rates-and-brackets
          https://www.revenue.state.mn.us/child-and-dependent-care-credit
          https://www.revenue.state.mn.us/dependent-exemptions
          https://www.revenue.state.mn.us/minnesota-standard-deduction
          https://mn.gov/mmb/assets/mwr_form_tcm1059-128581.pdf
          https://www.revenue.state.mn.us/individual-income-tax
          https://www.revenue.state.mn.us/individual-income-tax
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Virginia', 'state_occupation': 'Minnesota', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def virginia_and_mississippi(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Virginia:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Mississippi:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Virginia:
        - https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/reciprocity
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
        
        Mississippi:
        -         
        - https://www.dor.ms.gov/individual/tax-rates
          https://www.dor.ms.gov/individual/tax-rates
          https://www.dor.ms.gov/individual/tax-rates
          https://www.dor.ms.gov/individual
          https://www.dor.ms.gov/individual
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Virginia', 'state_occupation': 'Mississippi', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def virginia_and_missouri(self, capital_gains_short=0, dependents=0, sources=False, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_long=0, filing_status='single', income=100000, credits_state=0):
        """
        --- Parameters ----
        
        Virginia:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Missouri:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Virginia:
        - https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/reciprocity
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
        
        Missouri:
        -         
        - https://dor.mo.gov/taxation/individual/tax-types/income/year-changes/
          https://dor.mo.gov/taxation/individual/tax-types/income/year-changes/
          https://dor.mo.gov/taxation/individual/tax-types/income/year-changes/
          https://www.kcmo.gov/city-hall/departments/finance/earnings-tax
          https://www.stlouis-mo.gov/government/departments/collector/earnings-tax/individual-earnings-tax-info.cfm
        """
        
        # Check if the parameters are set upstream in the instance
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Virginia', 'state_occupation': 'Missouri', "capital_gains_short": capital_gains_short, "dependents": dependents, "sources": sources, "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def virginia_and_montana(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Virginia:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Montana:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Virginia:
        - https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/reciprocity
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
        
        Montana:
        -         
        - https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/2022/12/montana-employees-withholding-allowance-and-exemption-certificate-form-mw-4.pdf
          https://mtrevenue.gov/2021/11/29/simplification-of-montana-income-taxation/?utm_source=rss&utm_medium=rss&utm_campaign=simplification-of-montana-income-taxation
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Virginia', 'state_occupation': 'Montana', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def virginia_and_nebraska(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Virginia:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Nebraska:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Virginia:
        - https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/reciprocity
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
        
        Nebraska:
        -         
        - https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2023_Circular_EN_8-429r2023_Whole_Book_11.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Virginia', 'state_occupation': 'Nebraska', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def virginia_and_nevada(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Virginia:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Nevada:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Virginia:
        - https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/reciprocity
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
        
        Nevada:
        -         
        - https://www.nvsos.gov/sos/businesses/the-nevada-advantage
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Virginia', 'state_occupation': 'Nevada', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def virginia_and_new_hampshire(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Virginia:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        New Hampshire:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Virginia:
        - https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/reciprocity
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
        
        New Hampshire:
        -         
        - https://www.revenue.nh.gov/assistance/tax-overview.htm
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Virginia', 'state_occupation': 'NewHampshire', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def virginia_and_new_jersey(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Virginia:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        New Jersey:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Virginia:
        - https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/reciprocity
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
        
        New Jersey:
        -         
        - https://www.state.nj.us/treasury/taxation/taxtables.shtml
          https://www.state.nj.us/treasury/taxation/njit13.shtml
          https://www.nj.gov/treasury/taxation/njit25.shtml
          https://www.state.nj.us/treasury/taxation/individuals/freqqiti.shtml
          https://www.state.nj.us/treasury/taxation/localtax.shtml
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Virginia', 'state_occupation': 'NewJersey', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def virginia_and_new_mexico(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Virginia:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        New Mexico:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Virginia:
        - https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/reciprocity
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
        
        New Mexico:
        -         
        - https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/fdf3c548-8aba-4b9c-9eb4-bb564c716015/FYI-104.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/0bf943fd-652e-4400-bb1b-cd8397eb5a95/2022pit-adj.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/2f1a6781-9534-4436-b427-1557f9592099/2022pit-adj-ins.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/0bf943fd-652e-4400-bb1b-cd8397eb5a95/2022pit-adj.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/1afc56af-ea90-4d48-82e5-1f9aeb43255a/PITbook2022.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/fdf3c548-8aba-4b9c-9eb4-bb564c716015/FYI-104.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Virginia', 'state_occupation': 'NewMexico', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def virginia_and_new_york(self, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Virginia:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        New York:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Virginia:
        - https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/reciprocity
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
        
        New York:
        -         
        - https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nys-tax-rate-schedule
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nys-tax-rate-schedule
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nys-tax-rate-schedule
          https://www.tax.ny.gov/pit/
          https://www.tax.ny.gov/pit/file/nonresident-faqs.htm
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#yonkers-tax-rate-schedule
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nyc-tax-rate-schedule
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Virginia', 'state_occupation': 'NewYork', "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def virginia_and_north_carolina(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Virginia:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        North Carolina:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Virginia:
        - https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/reciprocity
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
        
        North Carolina:
        -         
        - https://www.ncdor.gov/taxes-forms/tax-rate-schedules
          https://www.ncdor.gov/taxes-forms/individual-income-tax/north-carolina-standard-deduction-or-north-carolina-itemized-deductions
          https://www.ncdor.gov/taxes-forms/individual-income-tax
          https://www.ncdor.gov/taxes-forms/individual-income-tax
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Virginia', 'state_occupation': 'NorthCarolina', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def virginia_and_north_dakota(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Virginia:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        North Dakota:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Virginia:
        - https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/reciprocity
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
        
        North Dakota:
        -         
        - https://www.tax.nd.gov/forms
          https://www.tax.nd.gov/sites/www/files/documents/forms/individual/2022-iit/2022-individual-income-tax-booklet.pdf
          https://www.tax.nd.gov/sites/www/files/documents/forms/2022-individual-income-tax-booklet.pdf
          https://www.tax.nd.gov/tax-exemptions-credits/income-tax-exemptions-credits
          https://www.tax.nd.gov/sites/www/files/documents/forms/business/ndwrfillable.pdf
          https://www.tax.nd.gov/news/tax-legislative-changes/significant-changes-law/individual-income-tax-history
          https://www.tax.nd.gov/forms
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Virginia', 'state_occupation': 'NorthDakota', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def virginia_and_ohio(self, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Virginia:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Ohio:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Virginia:
        - https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/reciprocity
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
        
        Ohio:
        -         
        - https://tax.ohio.gov/individual/resources/annual-tax-rates
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://tax.ohio.gov/static/forms/employer_withholding/generic/wth_it4nr.pdf
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://www.ritaohio.com/TaxRatesTable
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Virginia', 'state_occupation': 'Ohio', "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def virginia_and_oklahoma(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Virginia:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Oklahoma:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Virginia:
        - https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/reciprocity
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
        
        Oklahoma:
        -         
        - https://oklahoma.gov/content/dam/ok/en/tax/documents/resources/publications/businesses/withholding-tables/WHTables-2022.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Virginia', 'state_occupation': 'Oklahoma', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def virginia_and_oregon(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Virginia:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Oregon:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Virginia:
        - https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/reciprocity
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
        
        Oregon:
        -         
        - https://www.oregon.gov/dor/programs/individuals/Pages/PIT.aspx
          https://www.oregon.gov/dor/programs/individuals/Documents/Full%20year%20resident
         %20Form%20or-40%20filers.pdf
          https://www.oregon.gov/dor/programs/individuals/Documents/Part-year%20and%20nonresident
         %20Form%20or-40-P%20and%20or-40-N%20filers.pdf.pdf
          https://www.oregon.gov/dor/forms/FormsPubs/form-or-40_101-040_2022.pdf
          https://www.oregon.gov/dor/forms/FormsPubs/form-or-40_101-040_2022.pdf
          https://www.oregon.gov/dor/forms/FormsPubs/form-or-40_101-040_2022.pdf
          https://www.portland.gov/revenue/personal-tax
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Virginia', 'state_occupation': 'Oregon', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def virginia_and_pennsylvania(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Virginia:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Pennsylvania:
        


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Virginia:
        - https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/reciprocity
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
        
        Pennsylvania:
        -         
        - https://www.revenue.pa.gov/Tax%20Rates/Pages/default.aspx
          https://www.revenue.pa.gov/FormsandPublications/FormsforIndividuals/PIT/Documents/2022/2022_pa-40in.pdf
          https://www.revenue.pa.gov/FormsandPublications/FormsforBusinesses/EmployerWithholding/Documents/rev-419.pdf
          https://www.revenue.pa.gov/FormsandPublications/FormsforIndividuals/PIT/Documents/2022/2022_pa-40in.pdf
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-25.htm
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Virginia', 'state_occupation': 'Pennsylvania', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def virginia_and_rhode_island(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Virginia:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Rhode Island:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Virginia:
        - https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/reciprocity
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
        
        Rhode Island:
        -         
        - https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2022_1040WE_w_0.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Virginia', 'state_occupation': 'RhodeIsland', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def virginia_and_south_carolina(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Virginia:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        South Carolina:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Virginia:
        - https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/reciprocity
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
        
        South Carolina:
        -         
        - https://dor.sc.gov/resources-site/lawandpolicy/Advisory%20Opinions/IL22-15.pdf
          https://dor.sc.gov/forms-site/Forms/IITPacket_2022.pdf
          https://www.scstatehouse.gov/code/t12c006.php
          https://dor.sc.gov/resources-site/lawandpolicy/Advisory%20Opinions/IL22-15.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Virginia', 'state_occupation': 'SouthCarolina', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def virginia_and_south_dakota(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Virginia:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        South Dakota:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Virginia:
        - https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/reciprocity
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
        
        South Dakota:
        -         
        - https://dor.sd.gov/individuals/taxes/
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Virginia', 'state_occupation': 'SouthDakota', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def virginia_and_tennessee(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Virginia:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Tennessee:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Virginia:
        - https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/reciprocity
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
        
        Tennessee:
        -         
        - https://www.tn.gov/revenue/taxes.html
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Virginia', 'state_occupation': 'Tennessee', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def virginia_and_texas(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Virginia:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Texas:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Virginia:
        - https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/reciprocity
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
        
        Texas:
        -         
        - https://comptroller.texas.gov/taxes/
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Virginia', 'state_occupation': 'Texas', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def virginia_and_utah(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Virginia:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Utah:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Virginia:
        - https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/reciprocity
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
        
        Utah:
        -         
        - https://incometax.utah.gov/paying/tax-rates
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Virginia', 'state_occupation': 'Utah', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def virginia_and_vermont(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Virginia:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Vermont:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Virginia:
        - https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/reciprocity
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
        
        Vermont:
        -         
        - https://tax.vermont.gov/sites/tax/files/documents/GB-1210-2023.pdf
          https://tax.vermont.gov/sites/tax/files/documents/Income%20Booklet-2022.pdf
          https://tax.vermont.gov/sites/tax/files/documents/Income%20Booklet-2022.pdf
          https://tax.vermont.gov/sites/tax/files/documents/Income%20Booklet-2022.pdf
          https://tax.vermont.gov/individuals/personal-income-tax/tax-credits
          https://tax.vermont.gov/individuals/personal-income-tax/taxable-income
          https://tax.vermont.gov/individuals/personal-income-tax/local-option-tax
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Virginia', 'state_occupation': 'Vermont', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def virginia_and_washington(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Virginia:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Washington:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Virginia:
        - https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/reciprocity
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
        
        Washington:
        -         
        - https://dor.wa.gov/taxes-rates/income-tax
          https://dor.wa.gov/taxes-rates/other-taxes/capital-gains-tax
          https://dor.wa.gov/taxes-rates/income-tax
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Virginia', 'state_occupation': 'Washington', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def virginia_and_west_virginia(self, pay_periods=24, capital_gains_short=0, county_occupation='Other', dependents=0, sources=False, state_agi=75000, filers_over_65=0, capital_gains_long=0, filing_status='single', income=100000, credits_state=0):
        """
        --- Parameters ----
        
        Virginia:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        West Virginia:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_occupation : str, optional
            County of occupation. Default is Other.
        pay_periods : int, optional
            Number of pay periods. Options are 13, 24, 26, 52. Default is 24.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Virginia:
        - https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/reciprocity
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
        
        West Virginia:
        -         
        - https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://tax.wv.gov/Documents/TSD/tsd438.pdf
          https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2011/Taxes-11-35att.html
        """
        
        # Check if the parameters are set upstream in the instance
        pay_periods = self.pay_periods if hasattr(self, 'pay_periods') else pay_periods
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        county_occupation = self.county_occupation if hasattr(self, 'county_occupation') else county_occupation
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Virginia', 'state_occupation': 'WestVirginia', "pay_periods": pay_periods, "capital_gains_short": capital_gains_short, "county_occupation": county_occupation, "dependents": dependents, "sources": sources, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def virginia_and_wisconsin(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Virginia:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Wisconsin:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Virginia:
        - https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/reciprocity
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
        
        Wisconsin:
        -         
        - https://www.revenue.wi.gov/TaxForms2023/2023-Form1-ES-Inst.pdf
          https://www.revenue.wi.gov/TaxForms2023/2023-Form1-ES-Inst.pdf
          https://www.revenue.wi.gov/TaxForms2023/2023-Form1-ES-Inst.pdf
          https://www.revenue.wi.gov/Pages/FAQS/pcs-work.aspx
          https://www.revenue.wi.gov/Dor%20Publications/pb103.pdf
          https://www.revenue.wi.gov/Pages/FAQS/pcs-work.aspx
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Virginia', 'state_occupation': 'Wisconsin', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def virginia_and_wyoming(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Virginia:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.
        
        Wyoming:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Virginia:
        - https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/reciprocity
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
        
        Wyoming:
        -         
        - https://revenue.wyo.gov/
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Virginia', 'state_occupation': 'Wyoming', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def washington_and_alabama(self, county_residence='Other', state_agi=75000, capital_gains_short=0, capital_gains_long=0, filing_status='single', income=100000, dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Washington:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        
        Alabama:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Washington:
        - https://dor.wa.gov/taxes-rates/income-tax
          https://dor.wa.gov/taxes-rates/other-taxes/capital-gains-tax
          https://dor.wa.gov/taxes-rates/income-tax
        
        Alabama:
        -         
        - https://www.revenue.alabama.gov/ultraviewer/viewer/basic_viewer/index.html?form=2023/02/22f40.pdf
          https://www.revenue.alabama.gov/faqs/what-is-alabamas-individual-income-tax-rate/
          https://www.revenue.alabama.gov/ultraviewer/viewer/basic_viewer/index.html?form=2023/02/22f40.pdf
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-24.htm?taxmap=true
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-24.htm?taxmap=true
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-24.htm?taxmap=true
          https://www.revenue.alabama.gov/individual-corporate/income-to-be-reported-on-the-alabama-income-tax-return/
          https://www.almonline.org/TaxRates.aspx#OccupationalTax
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Washington', 'state_occupation': 'Alabama', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def washington_and_alaska(self, capital_gains_long=0, sources=False, capital_gains_short=0):
        """
        --- Parameters ----
        
        Washington:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        
        Alaska:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Washington:
        - https://dor.wa.gov/taxes-rates/income-tax
          https://dor.wa.gov/taxes-rates/other-taxes/capital-gains-tax
          https://dor.wa.gov/taxes-rates/income-tax
        
        Alaska:
        -         
        - https://www.commerce.alaska.gov/web/dcra/officeofthestateassessor/alaskataxfacts.aspx
        """
        
        # Check if the parameters are set upstream in the instance
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        sources = self.sources if hasattr(self, 'sources') else sources
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        
        # State-specific parameters
        data = {
            'state_residence': 'Washington', 'state_occupation': 'Alaska', "capital_gains_long": capital_gains_long, "sources": sources, "capital_gains_short": capital_gains_short
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def washington_and_arizona(self, state_agi=75000, capital_gains_short=0, capital_gains_long=0, filing_status='single', income=100000, dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Washington:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        
        Arizona:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        credits_state : int, optional
            Credits for state taxes. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Washington:
        - https://dor.wa.gov/taxes-rates/income-tax
          https://dor.wa.gov/taxes-rates/other-taxes/capital-gains-tax
          https://dor.wa.gov/taxes-rates/income-tax
        
        Arizona:
        -         
        - https://azdor.gov/forms/individual/form-140-x-y-tables
          https://azdor.gov/forms/individual/form-140-x-y-tables
          https://azdor.gov/forms/individual/form-140-x-y-tables
          https://azdor.gov/businesses-arizona/withholding-tax/withholding-exceptions
          "https://azdor.gov/individuals/income-tax-filing-assistance/identifying-other-taxable-income#:~:text=In%20regards%20to%20capital%20gains
         the%20individuals%20regular%20tax%20rate.
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Washington', 'state_occupation': 'Arizona', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def washington_and_arkansas(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, capital_gains_long=0, filing_status='single', income=100000, dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Washington:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        
        Arkansas:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Washington:
        - https://dor.wa.gov/taxes-rates/income-tax
          https://dor.wa.gov/taxes-rates/other-taxes/capital-gains-tax
          https://dor.wa.gov/taxes-rates/income-tax
        
        Arkansas:
        -         
        - https://www.arkansasedc.com/why-arkansas/business-climate/tax-structure/personal-income-tax
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_and_AR1000NR_Instructions.pdf
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_FullYearResidentIndividualIncomeTaxReturn.pdf
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_and_AR1000NR_Instructions.pdf
          https://www.arkansasedc.com/why-arkansas/business-climate/tax-structure/capital-gains-tax-reduction
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_and_AR1000NR_Instructions.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Washington', 'state_occupation': 'Arkansas', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def washington_and_california(self, state_agi=75000, capital_gains_short=0, capital_gains_long=0, filing_status='single', income=100000, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Washington:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        
        California:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Washington:
        - https://dor.wa.gov/taxes-rates/income-tax
          https://dor.wa.gov/taxes-rates/other-taxes/capital-gains-tax
          https://dor.wa.gov/taxes-rates/income-tax
        
        California:
        -         
        - https://www.ftb.ca.gov/forms/2022/2022-540-tax-rate-schedules.pdf
          https://www.ftb.ca.gov/file/personal/deductions/index.html
          https://www.ftb.ca.gov/about-ftb/newsroom/caleitc/eligibility-and-credit-information.html
          https://www.ftb.ca.gov/file/personal/income-types/capital-gains-and-losses.html#:~:text=California%20does%20not%20have%20a
         are%20taxed%20as%20ordinary%20income
          https://www.ftb.ca.gov/file/personal/
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Washington', 'state_occupation': 'California', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def washington_and_colorado(self, county_residence='Other', state_agi=75000, capital_gains_short=0, capital_gains_long=0, income=100000, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Washington:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        
        Colorado:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Washington:
        - https://dor.wa.gov/taxes-rates/income-tax
          https://dor.wa.gov/taxes-rates/other-taxes/capital-gains-tax
          https://dor.wa.gov/taxes-rates/income-tax
        
        Colorado:
        -         
        - https://leg.colorado.gov/agencies/legislative-council-staff/individual-income-tax%C2%A0
          https://leg.colorado.gov/agencies/legislative-council-staff/individual-income-tax%C2%A0
          https://tax.colorado.gov/sites/tax/files/documents/ITT_Capital_Gain_Subtraction_Feb_2022.pdf
          https://www.auroragov.org/business_services/taxes/occupational_privilege_tax
          https://www.denvergov.org/Government/Agencies-Departments-Offices/Agencies-Departments-Offices-Directory/Department-of-Finance/Our-Divisions/Treasury/Business-Tax-Information#:~:text=Occupational%20privilege%20tax
         -%28show%20below%29&text=Each%20taxable%20employee%20is%20liable
         month%20for%20each%20taxable%20employee
          https://www.glendale.co.us/355/Occupational-Privilege-Tax
          https://www.greenwoodvillage.com/1220/Occupational-Privilege-Tax-OPT
          https://ci.sheridan.co.us/288/Occupational-Privilege-Tax
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        income = self.income if hasattr(self, 'income') else income
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Washington', 'state_occupation': 'Colorado', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "income": income, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def washington_and_connecticut(self, state_agi=75000, capital_gains_short=0, capital_gains_long=0, filing_status='single', income=100000, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Washington:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        
        Connecticut:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Washington:
        - https://dor.wa.gov/taxes-rates/income-tax
          https://dor.wa.gov/taxes-rates/other-taxes/capital-gains-tax
          https://dor.wa.gov/taxes-rates/income-tax
        
        Connecticut:
        -         
        - https://portal.ct.gov/-/media/DRS/Forms/2022/Income/2022-CT-1040-Instructions_1222.pdf
          https://www.cga.ct.gov/2022/rpt/pdf/2022-R-0108.pdf
          https://www.cga.ct.gov/current/pub/chap_229.htm#sec_12-703
          https://portal.ct.gov/-/media/DRS/Forms/2022/Income/2022-CT-1040-Instructions_1222.pdf
          "https://portal.ct.gov/DRS/Publications/TSSNs/TSSN-29#:~:text=An%20individuals%20net%20capital%20gains
         portion%20of%20Social%20Security%20benefits"
          "https://portal.ct.gov/DRS/Publications/TSSNs/TSSN-29#:~:text=An%20individuals%20net%20capital%20gains
         portion%20of%20Social%20Security%20benefits"
          https://portal.ct.gov/DRS/Individuals/Resident-Income-Tax/Tax-Information
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Washington', 'state_occupation': 'Connecticut', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def washington_and_delaware(self, capital_gains_short=0, capital_gains_long=0, federal_agi=75000, income=100000, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Washington:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        
        Delaware:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        federal_agi : int, optional
            Adjusted Gross Income (AGI) for federal taxes. Default is 0. Sources - https://www.irs.gov/e-file-providers/definition-of-adjusted-gross-income, https://www.irs.gov/newsroom/taxpayers-should-know-the-difference-between-standard-and-itemized-deductions
        income : int, optional
            Gross income. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Washington:
        - https://dor.wa.gov/taxes-rates/income-tax
          https://dor.wa.gov/taxes-rates/other-taxes/capital-gains-tax
          https://dor.wa.gov/taxes-rates/income-tax
        
        Delaware:
        -         
        - https://revenue.delaware.gov/software-developer/tax-rate-changes/
          https://revenuefiles.delaware.gov/2022/PIT-RES_TY22_2022-01_PaperInteractive.pdf
          https://revenuefiles.delaware.gov/2022/PIT-RES_TY22_2022-02_Instructions.pdf
          https://revenue.delaware.gov/frequently-asked-questions/personal-income-tax-faqs/
          https://www.newcastlede.gov/DocumentCenter/View/20147/City-of-Wilmington-Wage-Tax-Information-PDF
        """
        
        # Check if the parameters are set upstream in the instance
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        federal_agi = self.federal_agi if hasattr(self, 'federal_agi') else federal_agi
        income = self.income if hasattr(self, 'income') else income
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Washington', 'state_occupation': 'Delaware', "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "federal_agi": federal_agi, "income": income, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def washington_and_florida(self, capital_gains_long=0, sources=False, capital_gains_short=0):
        """
        --- Parameters ----
        
        Washington:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        
        Florida:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Washington:
        - https://dor.wa.gov/taxes-rates/income-tax
          https://dor.wa.gov/taxes-rates/other-taxes/capital-gains-tax
          https://dor.wa.gov/taxes-rates/income-tax
        
        Florida:
        -         
        - https://floridarevenue.com/taxes/taxesfees/Pages/tax_interest_rates.aspx
        """
        
        # Check if the parameters are set upstream in the instance
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        sources = self.sources if hasattr(self, 'sources') else sources
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        
        # State-specific parameters
        data = {
            'state_residence': 'Washington', 'state_occupation': 'Florida', "capital_gains_long": capital_gains_long, "sources": sources, "capital_gains_short": capital_gains_short
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def washington_and_georgia(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, capital_gains_long=0, filing_status='single', income=100000, dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Washington:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        
        Georgia:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Washington:
        - https://dor.wa.gov/taxes-rates/income-tax
          https://dor.wa.gov/taxes-rates/other-taxes/capital-gains-tax
          https://dor.wa.gov/taxes-rates/income-tax
        
        Georgia:
        -         
        - https://dor.georgia.gov/tax-tables-georgia-tax-rate-schedule
          https://apps.dor.ga.gov/FillableForms/PDFViewer/Index?form=2022GA500
          https://apps.dor.ga.gov/FillableForms/PDFViewer/Index?form=2022GA500
          https://apps.dor.ga.gov/FillableForms/PDFViewer/Index?form=2022GA500
          https://dor.georgia.gov/years-individual-income-tax-forms
          https://dor.georgia.gov/years-individual-income-tax-forms
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Washington', 'state_occupation': 'Georgia', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def washington_and_hawaii(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, capital_gains_long=0, filing_status='single', income=100000, dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Washington:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        
        Hawaii:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Washington:
        - https://dor.wa.gov/taxes-rates/income-tax
          https://dor.wa.gov/taxes-rates/other-taxes/capital-gains-tax
          https://dor.wa.gov/taxes-rates/income-tax
        
        Hawaii:
        -         
        - https://tax.hawaii.gov/forms/d_18table-on/d_18table-on_p13/
          https://files.hawaii.gov/tax/forms/2018/18table-on.pdf
          https://files.hawaii.gov/tax/news/pubs/22outline.pdf
          https://files.hawaii.gov/tax/news/pubs/FastTaxRefGuide_2017.pdf
          https://www.tfhawaii.org/wordpress/state-tax-resources/mini-tax-guides/individual-income-tax-chapter-235/#:~:text=Long%20term%20capital%20gains%20are%20taxed%20at%20a%20maximum%20of%207.25%25
          https://www.grassrootinstitute.org/2023/02/hb337-increasing-tax-on-capital-gains-a-bad-risk/
          https://files.hawaii.gov/tax/news/pubs/22outline.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Washington', 'state_occupation': 'Hawaii', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def washington_and_idaho(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, capital_gains_long=0, filing_status='single', income=100000, dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Washington:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        
        Idaho:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Washington:
        - https://dor.wa.gov/taxes-rates/income-tax
          https://dor.wa.gov/taxes-rates/other-taxes/capital-gains-tax
          https://dor.wa.gov/taxes-rates/income-tax
        
        Idaho:
        -         
        - https://tax.idaho.gov/taxes/income-tax/individual-income/individual-income-tax-rate-schedule/
          https://tax.idaho.gov/wp-content/uploads/forms/EIN00046/EIN00046_03-01-2023.pdf
          https://tax.idaho.gov/taxes/income-tax/individual-income/idaho-source-income/
          https://tax.idaho.gov/taxes/income-tax/individual-income/idaho-source-income/
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Washington', 'state_occupation': 'Idaho', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def washington_and_illinois(self, state_agi=75000, capital_gains_short=0, capital_gains_long=0, filing_status='single', income=100000, dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Washington:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        
        Illinois:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Washington:
        - https://dor.wa.gov/taxes-rates/income-tax
          https://dor.wa.gov/taxes-rates/other-taxes/capital-gains-tax
          https://dor.wa.gov/taxes-rates/income-tax
        
        Illinois:
        -         
        - https://tax.illinois.gov/research/taxrates/income.html#IndividualIncome
          https://tax.illinois.gov/content/dam/soi/en/web/tax/forms/incometax/documents/currentyear/individual/il-1040-instr.pdf
          https://tax.illinois.gov/questionsandanswers/12.html
          https://tax.illinois.gov/content/dam/soi/en/web/tax/forms/incometax/documents/currentyear/individual/il-1040-instr.pdf
          https://tax.illinois.gov/research/taxrates/personalproperty.html
          https://tax.illinois.gov/localgovernments/income.html
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Washington', 'state_occupation': 'Illinois', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def washington_and_indiana(self, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_short=0, capital_gains_long=0, filing_status='single', income=100000, dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Washington:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        
        Indiana:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Washington:
        - https://dor.wa.gov/taxes-rates/income-tax
          https://dor.wa.gov/taxes-rates/other-taxes/capital-gains-tax
          https://dor.wa.gov/taxes-rates/income-tax
        
        Indiana:
        -         
        - https://www.in.gov/dor/tax-forms/2022-individual-income-tax-forms/
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-32.htm
          https://www.in.gov/dor/files/ib28.pdf
          https://www.in.gov/dor/tax-forms/2022-individual-income-tax-forms/
          https://www.in.gov/dor/files/2022-county-tax-rates-and-codes.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Washington', 'state_occupation': 'Indiana', "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def washington_and_iowa(self, county_residence='Other', state_agi=75000, capital_gains_short=0, capital_gains_long=0, filing_status='single', income=100000, dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Washington:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        
        Iowa:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Washington:
        - https://dor.wa.gov/taxes-rates/income-tax
          https://dor.wa.gov/taxes-rates/other-taxes/capital-gains-tax
          https://dor.wa.gov/taxes-rates/income-tax
        
        Iowa:
        -         
        - https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
          https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
          https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
          https://tax.iowa.gov/iowa-illinois-reciprocal-agreement
          https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Washington', 'state_occupation': 'Iowa', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def washington_and_kansas(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, capital_gains_long=0, filing_status='single', income=100000, dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Washington:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        
        Kansas:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Washington:
        - https://dor.wa.gov/taxes-rates/income-tax
          https://dor.wa.gov/taxes-rates/other-taxes/capital-gains-tax
          https://dor.wa.gov/taxes-rates/income-tax
        
        Kansas:
        -         
        - https://www.ksrevenue.gov/taxrates.html
          https://www.ksrevenue.gov/incomebook22.html#0
          https://www.ksrevenue.gov/incomebook22.html#0
          https://www.ksrevenue.gov/incomebook22.html#0
          https://www.ksrevenue.gov/incomebook22.html
          https://www.ksrevenue.gov/incomebook22.html
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Washington', 'state_occupation': 'Kansas', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def washington_and_kentucky(self, county_residence='Other', state_agi=75000, capital_gains_short=0, capital_gains_long=0, filing_status='single', income=100000, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Washington:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        
        Kentucky:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Washington:
        - https://dor.wa.gov/taxes-rates/income-tax
          https://dor.wa.gov/taxes-rates/other-taxes/capital-gains-tax
          https://dor.wa.gov/taxes-rates/income-tax
        
        Kentucky:
        -         
        - https://revenue.ky.gov/News/Pages/Dor-Announces-Updates-to-Individual-Income-Tax-for-2023-Tax-Year.aspx
          https://revenue.ky.gov/Individual/Individual-Income-Tax/Pages/default.aspx
          https://revenue.ky.gov/News/Pages/Dor-Announces-Updates-to-Individual-Income-Tax-for-2023-Tax-Year.aspx
          https://revenue.ky.gov/Individual/Individual-Income-Tax/Pages/default.aspx
          https://revenue.ky.gov/Dor%20Training%20Materials/103%20KAR%2017.140.%20Individual%20income%20tax%20-%20reciprocity%20-%20nonresidents.pdf
          https://revenue.ky.gov/Individual/Individual-Income-Tax/Pages/default.aspx
          https://nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2021/Taxes-21-32.htm?taxmap=true
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Washington', 'state_occupation': 'Kentucky', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def washington_and_louisiana(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, capital_gains_long=0, filing_status='single', income=100000, dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Washington:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        
        Louisiana:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Washington:
        - https://dor.wa.gov/taxes-rates/income-tax
          https://dor.wa.gov/taxes-rates/other-taxes/capital-gains-tax
          https://dor.wa.gov/taxes-rates/income-tax
        
        Louisiana:
        -         
        - https://revenue.louisiana.gov/individualincometax
          https://revenue.louisiana.gov/individualincometax
          https://www.revenue.louisiana.gov/taxforms/6935(11_02)F.pdf
          https://revenue.louisiana.gov/individualincometax
          https://revenue.louisiana.gov/individualincometax
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Washington', 'state_occupation': 'Louisiana', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def washington_and_maine(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, capital_gains_long=0, filing_status='single', income=100000, dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Washington:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        
        Maine:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Washington:
        - https://dor.wa.gov/taxes-rates/income-tax
          https://dor.wa.gov/taxes-rates/other-taxes/capital-gains-tax
          https://dor.wa.gov/taxes-rates/income-tax
        
        Maine:
        -         
        - https://www.maine.gov/revenue/tax-return-forms/individual-income-tax-2022
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/ind_tax_rate_sched_2022.pdf
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/22_1040me_book_gen_instruc_revisedFeb23.pdf
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/22_1040me_book_gen_instruc_revisedFeb23.pdf
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/22_1040me_book_gen_instruc_revisedFeb23.pdf
          https://www.maine.gov/revenue/tax-return-forms/individual-income-tax-2022
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Washington', 'state_occupation': 'Maine', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def washington_and_maryland(self, county_residence='Other', state_agi=75000, capital_gains_short=0, capital_gains_long=0, filing_status='single', income=100000, dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Washington:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        
        Maryland:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Washington:
        - https://dor.wa.gov/taxes-rates/income-tax
          https://dor.wa.gov/taxes-rates/other-taxes/capital-gains-tax
          https://dor.wa.gov/taxes-rates/income-tax
        
        Maryland:
        -         
        - https://www.marylandtaxes.gov/individual/income/tax-info/tax-rates.php
          https://www.marylandtaxes.gov/forms/22_forms/Resident_Booklet.pdf
          https://www.marylandtaxes.gov/forms/Tax_Publications/Administrative_Releases/Income_and_Estate_Tax_Releases/ar_it37.pdf
          https://www.marylandtaxes.gov/individual/index.php
          https://www.marylandtaxes.gov/individual/income/tax-info/tax-rates.php
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Washington', 'state_occupation': 'Maryland', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def washington_and_massachusetts(self, state_agi=75000, capital_gains_short=0, capital_gains_long=0, filing_status='single', income=100000, dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Washington:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        
        Massachusetts:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Washington:
        - https://dor.wa.gov/taxes-rates/income-tax
          https://dor.wa.gov/taxes-rates/other-taxes/capital-gains-tax
          https://dor.wa.gov/taxes-rates/income-tax
        
        Massachusetts:
        -         
        - https://www.mass.gov/service-details/massachusetts-tax-rates
          https://www.mass.gov/service-details/view-massachusetts-personal-income-tax-exemptions
          https://www.mass.gov/service-details/view-massachusetts-personal-income-tax-exemptions
          https://www.mass.gov/service-details/view-child-and-dependent-related-credits
          https://www.mass.gov/service-details/massachusetts-tax-rates
          https://www.mass.gov/personal-income-tax
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Washington', 'state_occupation': 'Massachusetts', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def washington_and_michigan(self, county_residence='Other', state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', county_occupation='Other', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Washington:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        
        Michigan:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        county_occupation : str, optional
            County of occupation. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Washington:
        - https://dor.wa.gov/taxes-rates/income-tax
          https://dor.wa.gov/taxes-rates/other-taxes/capital-gains-tax
          https://dor.wa.gov/taxes-rates/income-tax
        
        Michigan:
        -         
        - https://www.michigan.gov/taxes/iit/new-developments-for-tax-year-2022
          "https://www.michigan.gov/taxes/-/media/Project/Websites/taxes/Forms/2023/2023-SUW/446_Withholding-Guide_2023.pdf?rev=177ca3d8cb034392bf84e317ed78c320&hash=19C5F42F43A89A91863D229349F48883#:~:text=The%20exemption%20amount%20is%20%245%2C400
         employees%20Michigan%20income%20tax%20return"
          "https://www.michigan.gov/taxes/-/media/Project/Websites/taxes/Forms/2023/2023-SUW/446_Withholding-Guide_2023.pdf?rev=177ca3d8cb034392bf84e317ed78c320&hash=19C5F42F43A89A91863D229349F48883#:~:text=The%20exemption%20amount%20is%20%245%2C400
         employees%20Michigan%20income%20tax%20return"
          https://www.michigan.gov/taxes/questions/iit/accordion/residency/are-my-wages-earned-in-another-state-taxable-in-michigan-if-i-am-a-michigan-resident-1
          https://www.michigan.gov/taxes/-/media/Project/Websites/taxes/Forms/2022/2022-IIT-Forms/BOOK_MI-1040.pdf?rev=4524f1cbbc0243b194ab05a6680be75b&hash=40DFE71E651F7C1D65C31F6CD03D944F
          https://www.michigan.gov/taxes/questions/iit/accordion/general/what-cities-impose-an-income-tax
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        county_occupation = self.county_occupation if hasattr(self, 'county_occupation') else county_occupation
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Washington', 'state_occupation': 'Michigan', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "county_occupation": county_occupation, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def washington_and_minnesota(self, state_agi=75000, capital_gains_short=0, capital_gains_long=0, filing_status='single', income=100000, dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Washington:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        
        Minnesota:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Washington:
        - https://dor.wa.gov/taxes-rates/income-tax
          https://dor.wa.gov/taxes-rates/other-taxes/capital-gains-tax
          https://dor.wa.gov/taxes-rates/income-tax
        
        Minnesota:
        -         
        - https://www.revenue.state.mn.us/minnesota-income-tax-rates-and-brackets
          https://www.revenue.state.mn.us/child-and-dependent-care-credit
          https://www.revenue.state.mn.us/dependent-exemptions
          https://www.revenue.state.mn.us/minnesota-standard-deduction
          https://mn.gov/mmb/assets/mwr_form_tcm1059-128581.pdf
          https://www.revenue.state.mn.us/individual-income-tax
          https://www.revenue.state.mn.us/individual-income-tax
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Washington', 'state_occupation': 'Minnesota', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def washington_and_mississippi(self, state_agi=75000, capital_gains_short=0, capital_gains_long=0, filing_status='single', income=100000, dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Washington:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        
        Mississippi:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Washington:
        - https://dor.wa.gov/taxes-rates/income-tax
          https://dor.wa.gov/taxes-rates/other-taxes/capital-gains-tax
          https://dor.wa.gov/taxes-rates/income-tax
        
        Mississippi:
        -         
        - https://www.dor.ms.gov/individual/tax-rates
          https://www.dor.ms.gov/individual/tax-rates
          https://www.dor.ms.gov/individual/tax-rates
          https://www.dor.ms.gov/individual
          https://www.dor.ms.gov/individual
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Washington', 'state_occupation': 'Mississippi', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def washington_and_missouri(self, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_short=0, capital_gains_long=0, filing_status='single', income=100000, dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Washington:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        
        Missouri:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Washington:
        - https://dor.wa.gov/taxes-rates/income-tax
          https://dor.wa.gov/taxes-rates/other-taxes/capital-gains-tax
          https://dor.wa.gov/taxes-rates/income-tax
        
        Missouri:
        -         
        - https://dor.mo.gov/taxation/individual/tax-types/income/year-changes/
          https://dor.mo.gov/taxation/individual/tax-types/income/year-changes/
          https://dor.mo.gov/taxation/individual/tax-types/income/year-changes/
          https://www.kcmo.gov/city-hall/departments/finance/earnings-tax
          https://www.stlouis-mo.gov/government/departments/collector/earnings-tax/individual-earnings-tax-info.cfm
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Washington', 'state_occupation': 'Missouri', "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def washington_and_montana(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, capital_gains_long=0, filing_status='single', income=100000, dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Washington:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        
        Montana:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Washington:
        - https://dor.wa.gov/taxes-rates/income-tax
          https://dor.wa.gov/taxes-rates/other-taxes/capital-gains-tax
          https://dor.wa.gov/taxes-rates/income-tax
        
        Montana:
        -         
        - https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/2022/12/montana-employees-withholding-allowance-and-exemption-certificate-form-mw-4.pdf
          https://mtrevenue.gov/2021/11/29/simplification-of-montana-income-taxation/?utm_source=rss&utm_medium=rss&utm_campaign=simplification-of-montana-income-taxation
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Washington', 'state_occupation': 'Montana', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def washington_and_nebraska(self, state_agi=75000, capital_gains_short=0, capital_gains_long=0, filing_status='single', income=100000, dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Washington:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        
        Nebraska:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Washington:
        - https://dor.wa.gov/taxes-rates/income-tax
          https://dor.wa.gov/taxes-rates/other-taxes/capital-gains-tax
          https://dor.wa.gov/taxes-rates/income-tax
        
        Nebraska:
        -         
        - https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2023_Circular_EN_8-429r2023_Whole_Book_11.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Washington', 'state_occupation': 'Nebraska', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def washington_and_nevada(self, capital_gains_long=0, sources=False, capital_gains_short=0):
        """
        --- Parameters ----
        
        Washington:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        
        Nevada:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Washington:
        - https://dor.wa.gov/taxes-rates/income-tax
          https://dor.wa.gov/taxes-rates/other-taxes/capital-gains-tax
          https://dor.wa.gov/taxes-rates/income-tax
        
        Nevada:
        -         
        - https://www.nvsos.gov/sos/businesses/the-nevada-advantage
        """
        
        # Check if the parameters are set upstream in the instance
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        sources = self.sources if hasattr(self, 'sources') else sources
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        
        # State-specific parameters
        data = {
            'state_residence': 'Washington', 'state_occupation': 'Nevada', "capital_gains_long": capital_gains_long, "sources": sources, "capital_gains_short": capital_gains_short
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def washington_and_new_hampshire(self, capital_gains_long=0, sources=False, capital_gains_short=0):
        """
        --- Parameters ----
        
        Washington:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        
        New Hampshire:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Washington:
        - https://dor.wa.gov/taxes-rates/income-tax
          https://dor.wa.gov/taxes-rates/other-taxes/capital-gains-tax
          https://dor.wa.gov/taxes-rates/income-tax
        
        New Hampshire:
        -         
        - https://www.revenue.nh.gov/assistance/tax-overview.htm
        """
        
        # Check if the parameters are set upstream in the instance
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        sources = self.sources if hasattr(self, 'sources') else sources
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        
        # State-specific parameters
        data = {
            'state_residence': 'Washington', 'state_occupation': 'NewHampshire', "capital_gains_long": capital_gains_long, "sources": sources, "capital_gains_short": capital_gains_short
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def washington_and_new_jersey(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, capital_gains_long=0, filing_status='single', income=100000, dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Washington:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        
        New Jersey:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Washington:
        - https://dor.wa.gov/taxes-rates/income-tax
          https://dor.wa.gov/taxes-rates/other-taxes/capital-gains-tax
          https://dor.wa.gov/taxes-rates/income-tax
        
        New Jersey:
        -         
        - https://www.state.nj.us/treasury/taxation/taxtables.shtml
          https://www.state.nj.us/treasury/taxation/njit13.shtml
          https://www.nj.gov/treasury/taxation/njit25.shtml
          https://www.state.nj.us/treasury/taxation/individuals/freqqiti.shtml
          https://www.state.nj.us/treasury/taxation/localtax.shtml
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Washington', 'state_occupation': 'NewJersey', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def washington_and_new_mexico(self, state_agi=75000, capital_gains_short=0, capital_gains_long=0, filing_status='single', income=100000, dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Washington:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        
        New Mexico:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Washington:
        - https://dor.wa.gov/taxes-rates/income-tax
          https://dor.wa.gov/taxes-rates/other-taxes/capital-gains-tax
          https://dor.wa.gov/taxes-rates/income-tax
        
        New Mexico:
        -         
        - https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/fdf3c548-8aba-4b9c-9eb4-bb564c716015/FYI-104.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/0bf943fd-652e-4400-bb1b-cd8397eb5a95/2022pit-adj.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/2f1a6781-9534-4436-b427-1557f9592099/2022pit-adj-ins.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/0bf943fd-652e-4400-bb1b-cd8397eb5a95/2022pit-adj.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/1afc56af-ea90-4d48-82e5-1f9aeb43255a/PITbook2022.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/fdf3c548-8aba-4b9c-9eb4-bb564c716015/FYI-104.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Washington', 'state_occupation': 'NewMexico', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def washington_and_new_york(self, county_residence='Other', state_agi=75000, capital_gains_short=0, capital_gains_long=0, filing_status='single', income=100000, dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Washington:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        
        New York:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Washington:
        - https://dor.wa.gov/taxes-rates/income-tax
          https://dor.wa.gov/taxes-rates/other-taxes/capital-gains-tax
          https://dor.wa.gov/taxes-rates/income-tax
        
        New York:
        -         
        - https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nys-tax-rate-schedule
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nys-tax-rate-schedule
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nys-tax-rate-schedule
          https://www.tax.ny.gov/pit/
          https://www.tax.ny.gov/pit/file/nonresident-faqs.htm
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#yonkers-tax-rate-schedule
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nyc-tax-rate-schedule
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Washington', 'state_occupation': 'NewYork', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def washington_and_north_carolina(self, state_agi=75000, capital_gains_short=0, capital_gains_long=0, filing_status='single', income=100000, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Washington:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        
        North Carolina:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Washington:
        - https://dor.wa.gov/taxes-rates/income-tax
          https://dor.wa.gov/taxes-rates/other-taxes/capital-gains-tax
          https://dor.wa.gov/taxes-rates/income-tax
        
        North Carolina:
        -         
        - https://www.ncdor.gov/taxes-forms/tax-rate-schedules
          https://www.ncdor.gov/taxes-forms/individual-income-tax/north-carolina-standard-deduction-or-north-carolina-itemized-deductions
          https://www.ncdor.gov/taxes-forms/individual-income-tax
          https://www.ncdor.gov/taxes-forms/individual-income-tax
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Washington', 'state_occupation': 'NorthCarolina', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def washington_and_north_dakota(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, capital_gains_long=0, filing_status='single', income=100000, dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Washington:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        
        North Dakota:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Washington:
        - https://dor.wa.gov/taxes-rates/income-tax
          https://dor.wa.gov/taxes-rates/other-taxes/capital-gains-tax
          https://dor.wa.gov/taxes-rates/income-tax
        
        North Dakota:
        -         
        - https://www.tax.nd.gov/forms
          https://www.tax.nd.gov/sites/www/files/documents/forms/individual/2022-iit/2022-individual-income-tax-booklet.pdf
          https://www.tax.nd.gov/sites/www/files/documents/forms/2022-individual-income-tax-booklet.pdf
          https://www.tax.nd.gov/tax-exemptions-credits/income-tax-exemptions-credits
          https://www.tax.nd.gov/sites/www/files/documents/forms/business/ndwrfillable.pdf
          https://www.tax.nd.gov/news/tax-legislative-changes/significant-changes-law/individual-income-tax-history
          https://www.tax.nd.gov/forms
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Washington', 'state_occupation': 'NorthDakota', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def washington_and_ohio(self, county_residence='Other', state_agi=75000, capital_gains_short=0, capital_gains_long=0, filing_status='single', income=100000, dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Washington:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        
        Ohio:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Washington:
        - https://dor.wa.gov/taxes-rates/income-tax
          https://dor.wa.gov/taxes-rates/other-taxes/capital-gains-tax
          https://dor.wa.gov/taxes-rates/income-tax
        
        Ohio:
        -         
        - https://tax.ohio.gov/individual/resources/annual-tax-rates
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://tax.ohio.gov/static/forms/employer_withholding/generic/wth_it4nr.pdf
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://www.ritaohio.com/TaxRatesTable
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Washington', 'state_occupation': 'Ohio', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def washington_and_oklahoma(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, capital_gains_long=0, filing_status='single', income=100000, dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Washington:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        
        Oklahoma:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Washington:
        - https://dor.wa.gov/taxes-rates/income-tax
          https://dor.wa.gov/taxes-rates/other-taxes/capital-gains-tax
          https://dor.wa.gov/taxes-rates/income-tax
        
        Oklahoma:
        -         
        - https://oklahoma.gov/content/dam/ok/en/tax/documents/resources/publications/businesses/withholding-tables/WHTables-2022.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Washington', 'state_occupation': 'Oklahoma', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def washington_and_oregon(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, capital_gains_long=0, filing_status='single', income=100000, dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Washington:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        
        Oregon:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Washington:
        - https://dor.wa.gov/taxes-rates/income-tax
          https://dor.wa.gov/taxes-rates/other-taxes/capital-gains-tax
          https://dor.wa.gov/taxes-rates/income-tax
        
        Oregon:
        -         
        - https://www.oregon.gov/dor/programs/individuals/Pages/PIT.aspx
          https://www.oregon.gov/dor/programs/individuals/Documents/Full%20year%20resident
         %20Form%20or-40%20filers.pdf
          https://www.oregon.gov/dor/programs/individuals/Documents/Part-year%20and%20nonresident
         %20Form%20or-40-P%20and%20or-40-N%20filers.pdf.pdf
          https://www.oregon.gov/dor/forms/FormsPubs/form-or-40_101-040_2022.pdf
          https://www.oregon.gov/dor/forms/FormsPubs/form-or-40_101-040_2022.pdf
          https://www.oregon.gov/dor/forms/FormsPubs/form-or-40_101-040_2022.pdf
          https://www.portland.gov/revenue/personal-tax
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Washington', 'state_occupation': 'Oregon', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def washington_and_pennsylvania(self, county_residence='Other', state_agi=75000, county_occupation='Other', capital_gains_short=0, capital_gains_long=0, income=100000, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Washington:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        
        Pennsylvania:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        county_residence : str, optional
            County of residence. Default is Other.
        county_occupation : str, optional
            County of occupation. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Washington:
        - https://dor.wa.gov/taxes-rates/income-tax
          https://dor.wa.gov/taxes-rates/other-taxes/capital-gains-tax
          https://dor.wa.gov/taxes-rates/income-tax
        
        Pennsylvania:
        -         
        - https://www.revenue.pa.gov/Tax%20Rates/Pages/default.aspx
          https://www.revenue.pa.gov/FormsandPublications/FormsforIndividuals/PIT/Documents/2022/2022_pa-40in.pdf
          https://www.revenue.pa.gov/FormsandPublications/FormsforBusinesses/EmployerWithholding/Documents/rev-419.pdf
          https://www.revenue.pa.gov/FormsandPublications/FormsforIndividuals/PIT/Documents/2022/2022_pa-40in.pdf
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-25.htm
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        county_occupation = self.county_occupation if hasattr(self, 'county_occupation') else county_occupation
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        income = self.income if hasattr(self, 'income') else income
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Washington', 'state_occupation': 'Pennsylvania', "county_residence": county_residence, "state_agi": state_agi, "county_occupation": county_occupation, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "income": income, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def washington_and_rhode_island(self, state_agi=75000, capital_gains_short=0, capital_gains_long=0, filing_status='single', income=100000, dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Washington:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        
        Rhode Island:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Washington:
        - https://dor.wa.gov/taxes-rates/income-tax
          https://dor.wa.gov/taxes-rates/other-taxes/capital-gains-tax
          https://dor.wa.gov/taxes-rates/income-tax
        
        Rhode Island:
        -         
        - https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2022_1040WE_w_0.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Washington', 'state_occupation': 'RhodeIsland', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def washington_and_south_carolina(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, capital_gains_long=0, filing_status='single', income=100000, dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Washington:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        
        South Carolina:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Washington:
        - https://dor.wa.gov/taxes-rates/income-tax
          https://dor.wa.gov/taxes-rates/other-taxes/capital-gains-tax
          https://dor.wa.gov/taxes-rates/income-tax
        
        South Carolina:
        -         
        - https://dor.sc.gov/resources-site/lawandpolicy/Advisory%20Opinions/IL22-15.pdf
          https://dor.sc.gov/forms-site/Forms/IITPacket_2022.pdf
          https://www.scstatehouse.gov/code/t12c006.php
          https://dor.sc.gov/resources-site/lawandpolicy/Advisory%20Opinions/IL22-15.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Washington', 'state_occupation': 'SouthCarolina', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def washington_and_south_dakota(self, capital_gains_long=0, sources=False, capital_gains_short=0):
        """
        --- Parameters ----
        
        Washington:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        
        South Dakota:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Washington:
        - https://dor.wa.gov/taxes-rates/income-tax
          https://dor.wa.gov/taxes-rates/other-taxes/capital-gains-tax
          https://dor.wa.gov/taxes-rates/income-tax
        
        South Dakota:
        -         
        - https://dor.sd.gov/individuals/taxes/
        """
        
        # Check if the parameters are set upstream in the instance
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        sources = self.sources if hasattr(self, 'sources') else sources
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        
        # State-specific parameters
        data = {
            'state_residence': 'Washington', 'state_occupation': 'SouthDakota', "capital_gains_long": capital_gains_long, "sources": sources, "capital_gains_short": capital_gains_short
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def washington_and_tennessee(self, capital_gains_long=0, sources=False, capital_gains_short=0):
        """
        --- Parameters ----
        
        Washington:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        
        Tennessee:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Washington:
        - https://dor.wa.gov/taxes-rates/income-tax
          https://dor.wa.gov/taxes-rates/other-taxes/capital-gains-tax
          https://dor.wa.gov/taxes-rates/income-tax
        
        Tennessee:
        -         
        - https://www.tn.gov/revenue/taxes.html
        """
        
        # Check if the parameters are set upstream in the instance
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        sources = self.sources if hasattr(self, 'sources') else sources
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        
        # State-specific parameters
        data = {
            'state_residence': 'Washington', 'state_occupation': 'Tennessee', "capital_gains_long": capital_gains_long, "sources": sources, "capital_gains_short": capital_gains_short
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def washington_and_texas(self, capital_gains_long=0, sources=False, capital_gains_short=0):
        """
        --- Parameters ----
        
        Washington:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        
        Texas:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Washington:
        - https://dor.wa.gov/taxes-rates/income-tax
          https://dor.wa.gov/taxes-rates/other-taxes/capital-gains-tax
          https://dor.wa.gov/taxes-rates/income-tax
        
        Texas:
        -         
        - https://comptroller.texas.gov/taxes/
        """
        
        # Check if the parameters are set upstream in the instance
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        sources = self.sources if hasattr(self, 'sources') else sources
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        
        # State-specific parameters
        data = {
            'state_residence': 'Washington', 'state_occupation': 'Texas', "capital_gains_long": capital_gains_long, "sources": sources, "capital_gains_short": capital_gains_short
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def washington_and_utah(self, state_agi=75000, capital_gains_short=0, capital_gains_long=0, filing_status='single', income=100000, dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Washington:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        
        Utah:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Washington:
        - https://dor.wa.gov/taxes-rates/income-tax
          https://dor.wa.gov/taxes-rates/other-taxes/capital-gains-tax
          https://dor.wa.gov/taxes-rates/income-tax
        
        Utah:
        -         
        - https://incometax.utah.gov/paying/tax-rates
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Washington', 'state_occupation': 'Utah', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def washington_and_vermont(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, capital_gains_long=0, filing_status='single', income=100000, dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Washington:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        
        Vermont:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Washington:
        - https://dor.wa.gov/taxes-rates/income-tax
          https://dor.wa.gov/taxes-rates/other-taxes/capital-gains-tax
          https://dor.wa.gov/taxes-rates/income-tax
        
        Vermont:
        -         
        - https://tax.vermont.gov/sites/tax/files/documents/GB-1210-2023.pdf
          https://tax.vermont.gov/sites/tax/files/documents/Income%20Booklet-2022.pdf
          https://tax.vermont.gov/sites/tax/files/documents/Income%20Booklet-2022.pdf
          https://tax.vermont.gov/sites/tax/files/documents/Income%20Booklet-2022.pdf
          https://tax.vermont.gov/individuals/personal-income-tax/tax-credits
          https://tax.vermont.gov/individuals/personal-income-tax/taxable-income
          https://tax.vermont.gov/individuals/personal-income-tax/local-option-tax
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Washington', 'state_occupation': 'Vermont', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def washington_and_virginia(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, capital_gains_long=0, filing_status='single', income=100000, dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Washington:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        
        Virginia:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Washington:
        - https://dor.wa.gov/taxes-rates/income-tax
          https://dor.wa.gov/taxes-rates/other-taxes/capital-gains-tax
          https://dor.wa.gov/taxes-rates/income-tax
        
        Virginia:
        -         
        - https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/reciprocity
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Washington', 'state_occupation': 'Virginia', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def washington_and_west_virginia(self, pay_periods=24, state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', county_occupation='Other', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Washington:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        
        West Virginia:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_occupation : str, optional
            County of occupation. Default is Other.
        pay_periods : int, optional
            Number of pay periods. Options are 13, 24, 26, 52. Default is 24.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Washington:
        - https://dor.wa.gov/taxes-rates/income-tax
          https://dor.wa.gov/taxes-rates/other-taxes/capital-gains-tax
          https://dor.wa.gov/taxes-rates/income-tax
        
        West Virginia:
        -         
        - https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://tax.wv.gov/Documents/TSD/tsd438.pdf
          https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2011/Taxes-11-35att.html
        """
        
        # Check if the parameters are set upstream in the instance
        pay_periods = self.pay_periods if hasattr(self, 'pay_periods') else pay_periods
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        county_occupation = self.county_occupation if hasattr(self, 'county_occupation') else county_occupation
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Washington', 'state_occupation': 'WestVirginia', "pay_periods": pay_periods, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "county_occupation": county_occupation, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def washington_and_wisconsin(self, state_agi=75000, capital_gains_short=0, capital_gains_long=0, filing_status='single', income=100000, dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Washington:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        
        Wisconsin:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Washington:
        - https://dor.wa.gov/taxes-rates/income-tax
          https://dor.wa.gov/taxes-rates/other-taxes/capital-gains-tax
          https://dor.wa.gov/taxes-rates/income-tax
        
        Wisconsin:
        -         
        - https://www.revenue.wi.gov/TaxForms2023/2023-Form1-ES-Inst.pdf
          https://www.revenue.wi.gov/TaxForms2023/2023-Form1-ES-Inst.pdf
          https://www.revenue.wi.gov/TaxForms2023/2023-Form1-ES-Inst.pdf
          https://www.revenue.wi.gov/Pages/FAQS/pcs-work.aspx
          https://www.revenue.wi.gov/Dor%20Publications/pb103.pdf
          https://www.revenue.wi.gov/Pages/FAQS/pcs-work.aspx
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Washington', 'state_occupation': 'Wisconsin', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def washington_and_wyoming(self, capital_gains_long=0, sources=False, capital_gains_short=0):
        """
        --- Parameters ----
        
        Washington:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        
        Wyoming:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Washington:
        - https://dor.wa.gov/taxes-rates/income-tax
          https://dor.wa.gov/taxes-rates/other-taxes/capital-gains-tax
          https://dor.wa.gov/taxes-rates/income-tax
        
        Wyoming:
        -         
        - https://revenue.wyo.gov/
        """
        
        # Check if the parameters are set upstream in the instance
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        sources = self.sources if hasattr(self, 'sources') else sources
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        
        # State-specific parameters
        data = {
            'state_residence': 'Washington', 'state_occupation': 'Wyoming', "capital_gains_long": capital_gains_long, "sources": sources, "capital_gains_short": capital_gains_short
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def west_virginia_and_alabama(self, pay_periods=24, capital_gains_short=0, county_occupation='Other', dependents=0, sources=False, county_residence='Other', state_agi=75000, capital_gains_long=0, filing_status='single', income=100000, credits_state=0):
        """
        --- Parameters ----
        
        West Virginia:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_occupation : str, optional
            County of occupation. Default is Other.
        pay_periods : int, optional
            Number of pay periods. Options are 13, 24, 26, 52. Default is 24.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Alabama:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        West Virginia:
        - https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://tax.wv.gov/Documents/TSD/tsd438.pdf
          https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2011/Taxes-11-35att.html
        
        Alabama:
        -         
        - https://www.revenue.alabama.gov/ultraviewer/viewer/basic_viewer/index.html?form=2023/02/22f40.pdf
          https://www.revenue.alabama.gov/faqs/what-is-alabamas-individual-income-tax-rate/
          https://www.revenue.alabama.gov/ultraviewer/viewer/basic_viewer/index.html?form=2023/02/22f40.pdf
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-24.htm?taxmap=true
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-24.htm?taxmap=true
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-24.htm?taxmap=true
          https://www.revenue.alabama.gov/individual-corporate/income-to-be-reported-on-the-alabama-income-tax-return/
          https://www.almonline.org/TaxRates.aspx#OccupationalTax
        """
        
        # Check if the parameters are set upstream in the instance
        pay_periods = self.pay_periods if hasattr(self, 'pay_periods') else pay_periods
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        county_occupation = self.county_occupation if hasattr(self, 'county_occupation') else county_occupation
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'WestVirginia', 'state_occupation': 'Alabama', "pay_periods": pay_periods, "capital_gains_short": capital_gains_short, "county_occupation": county_occupation, "dependents": dependents, "sources": sources, "county_residence": county_residence, "state_agi": state_agi, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def west_virginia_and_alaska(self, county_occupation='Other', pay_periods=24, state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        West Virginia:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_occupation : str, optional
            County of occupation. Default is Other.
        pay_periods : int, optional
            Number of pay periods. Options are 13, 24, 26, 52. Default is 24.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Alaska:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        West Virginia:
        - https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://tax.wv.gov/Documents/TSD/tsd438.pdf
          https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2011/Taxes-11-35att.html
        
        Alaska:
        -         
        - https://www.commerce.alaska.gov/web/dcra/officeofthestateassessor/alaskataxfacts.aspx
        """
        
        # Check if the parameters are set upstream in the instance
        county_occupation = self.county_occupation if hasattr(self, 'county_occupation') else county_occupation
        pay_periods = self.pay_periods if hasattr(self, 'pay_periods') else pay_periods
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'WestVirginia', 'state_occupation': 'Alaska', "county_occupation": county_occupation, "pay_periods": pay_periods, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def west_virginia_and_arizona(self, county_occupation='Other', pay_periods=24, state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        West Virginia:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_occupation : str, optional
            County of occupation. Default is Other.
        pay_periods : int, optional
            Number of pay periods. Options are 13, 24, 26, 52. Default is 24.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Arizona:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        credits_state : int, optional
            Credits for state taxes. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        West Virginia:
        - https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://tax.wv.gov/Documents/TSD/tsd438.pdf
          https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2011/Taxes-11-35att.html
        
        Arizona:
        -         
        - https://azdor.gov/forms/individual/form-140-x-y-tables
          https://azdor.gov/forms/individual/form-140-x-y-tables
          https://azdor.gov/forms/individual/form-140-x-y-tables
          https://azdor.gov/businesses-arizona/withholding-tax/withholding-exceptions
          "https://azdor.gov/individuals/income-tax-filing-assistance/identifying-other-taxable-income#:~:text=In%20regards%20to%20capital%20gains
         the%20individuals%20regular%20tax%20rate.
        """
        
        # Check if the parameters are set upstream in the instance
        county_occupation = self.county_occupation if hasattr(self, 'county_occupation') else county_occupation
        pay_periods = self.pay_periods if hasattr(self, 'pay_periods') else pay_periods
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'WestVirginia', 'state_occupation': 'Arizona', "county_occupation": county_occupation, "pay_periods": pay_periods, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def west_virginia_and_arkansas(self, pay_periods=24, capital_gains_short=0, county_occupation='Other', dependents=0, sources=False, state_agi=75000, filers_over_65=0, capital_gains_long=0, filing_status='single', income=100000, credits_state=0):
        """
        --- Parameters ----
        
        West Virginia:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_occupation : str, optional
            County of occupation. Default is Other.
        pay_periods : int, optional
            Number of pay periods. Options are 13, 24, 26, 52. Default is 24.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Arkansas:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        West Virginia:
        - https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://tax.wv.gov/Documents/TSD/tsd438.pdf
          https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2011/Taxes-11-35att.html
        
        Arkansas:
        -         
        - https://www.arkansasedc.com/why-arkansas/business-climate/tax-structure/personal-income-tax
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_and_AR1000NR_Instructions.pdf
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_FullYearResidentIndividualIncomeTaxReturn.pdf
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_and_AR1000NR_Instructions.pdf
          https://www.arkansasedc.com/why-arkansas/business-climate/tax-structure/capital-gains-tax-reduction
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_and_AR1000NR_Instructions.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        pay_periods = self.pay_periods if hasattr(self, 'pay_periods') else pay_periods
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        county_occupation = self.county_occupation if hasattr(self, 'county_occupation') else county_occupation
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'WestVirginia', 'state_occupation': 'Arkansas', "pay_periods": pay_periods, "capital_gains_short": capital_gains_short, "county_occupation": county_occupation, "dependents": dependents, "sources": sources, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def west_virginia_and_california(self, county_occupation='Other', pay_periods=24, state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        West Virginia:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_occupation : str, optional
            County of occupation. Default is Other.
        pay_periods : int, optional
            Number of pay periods. Options are 13, 24, 26, 52. Default is 24.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        California:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        West Virginia:
        - https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://tax.wv.gov/Documents/TSD/tsd438.pdf
          https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2011/Taxes-11-35att.html
        
        California:
        -         
        - https://www.ftb.ca.gov/forms/2022/2022-540-tax-rate-schedules.pdf
          https://www.ftb.ca.gov/file/personal/deductions/index.html
          https://www.ftb.ca.gov/about-ftb/newsroom/caleitc/eligibility-and-credit-information.html
          https://www.ftb.ca.gov/file/personal/income-types/capital-gains-and-losses.html#:~:text=California%20does%20not%20have%20a
         are%20taxed%20as%20ordinary%20income
          https://www.ftb.ca.gov/file/personal/
        """
        
        # Check if the parameters are set upstream in the instance
        county_occupation = self.county_occupation if hasattr(self, 'county_occupation') else county_occupation
        pay_periods = self.pay_periods if hasattr(self, 'pay_periods') else pay_periods
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'WestVirginia', 'state_occupation': 'California', "county_occupation": county_occupation, "pay_periods": pay_periods, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def west_virginia_and_colorado(self, county_occupation='Other', pay_periods=24, state_agi=75000, county_residence='Other', capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        West Virginia:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_occupation : str, optional
            County of occupation. Default is Other.
        pay_periods : int, optional
            Number of pay periods. Options are 13, 24, 26, 52. Default is 24.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Colorado:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        West Virginia:
        - https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://tax.wv.gov/Documents/TSD/tsd438.pdf
          https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2011/Taxes-11-35att.html
        
        Colorado:
        -         
        - https://leg.colorado.gov/agencies/legislative-council-staff/individual-income-tax%C2%A0
          https://leg.colorado.gov/agencies/legislative-council-staff/individual-income-tax%C2%A0
          https://tax.colorado.gov/sites/tax/files/documents/ITT_Capital_Gain_Subtraction_Feb_2022.pdf
          https://www.auroragov.org/business_services/taxes/occupational_privilege_tax
          https://www.denvergov.org/Government/Agencies-Departments-Offices/Agencies-Departments-Offices-Directory/Department-of-Finance/Our-Divisions/Treasury/Business-Tax-Information#:~:text=Occupational%20privilege%20tax
         -%28show%20below%29&text=Each%20taxable%20employee%20is%20liable
         month%20for%20each%20taxable%20employee
          https://www.glendale.co.us/355/Occupational-Privilege-Tax
          https://www.greenwoodvillage.com/1220/Occupational-Privilege-Tax-OPT
          https://ci.sheridan.co.us/288/Occupational-Privilege-Tax
        """
        
        # Check if the parameters are set upstream in the instance
        county_occupation = self.county_occupation if hasattr(self, 'county_occupation') else county_occupation
        pay_periods = self.pay_periods if hasattr(self, 'pay_periods') else pay_periods
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'WestVirginia', 'state_occupation': 'Colorado', "county_occupation": county_occupation, "pay_periods": pay_periods, "state_agi": state_agi, "county_residence": county_residence, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def west_virginia_and_connecticut(self, county_occupation='Other', pay_periods=24, state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        West Virginia:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_occupation : str, optional
            County of occupation. Default is Other.
        pay_periods : int, optional
            Number of pay periods. Options are 13, 24, 26, 52. Default is 24.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Connecticut:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        West Virginia:
        - https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://tax.wv.gov/Documents/TSD/tsd438.pdf
          https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2011/Taxes-11-35att.html
        
        Connecticut:
        -         
        - https://portal.ct.gov/-/media/DRS/Forms/2022/Income/2022-CT-1040-Instructions_1222.pdf
          https://www.cga.ct.gov/2022/rpt/pdf/2022-R-0108.pdf
          https://www.cga.ct.gov/current/pub/chap_229.htm#sec_12-703
          https://portal.ct.gov/-/media/DRS/Forms/2022/Income/2022-CT-1040-Instructions_1222.pdf
          "https://portal.ct.gov/DRS/Publications/TSSNs/TSSN-29#:~:text=An%20individuals%20net%20capital%20gains
         portion%20of%20Social%20Security%20benefits"
          "https://portal.ct.gov/DRS/Publications/TSSNs/TSSN-29#:~:text=An%20individuals%20net%20capital%20gains
         portion%20of%20Social%20Security%20benefits"
          https://portal.ct.gov/DRS/Individuals/Resident-Income-Tax/Tax-Information
        """
        
        # Check if the parameters are set upstream in the instance
        county_occupation = self.county_occupation if hasattr(self, 'county_occupation') else county_occupation
        pay_periods = self.pay_periods if hasattr(self, 'pay_periods') else pay_periods
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'WestVirginia', 'state_occupation': 'Connecticut', "county_occupation": county_occupation, "pay_periods": pay_periods, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def west_virginia_and_delaware(self, county_occupation='Other', pay_periods=24, state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, federal_agi=75000, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        West Virginia:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_occupation : str, optional
            County of occupation. Default is Other.
        pay_periods : int, optional
            Number of pay periods. Options are 13, 24, 26, 52. Default is 24.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Delaware:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        federal_agi : int, optional
            Adjusted Gross Income (AGI) for federal taxes. Default is 0. Sources - https://www.irs.gov/e-file-providers/definition-of-adjusted-gross-income, https://www.irs.gov/newsroom/taxpayers-should-know-the-difference-between-standard-and-itemized-deductions
        income : int, optional
            Gross income. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        West Virginia:
        - https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://tax.wv.gov/Documents/TSD/tsd438.pdf
          https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2011/Taxes-11-35att.html
        
        Delaware:
        -         
        - https://revenue.delaware.gov/software-developer/tax-rate-changes/
          https://revenuefiles.delaware.gov/2022/PIT-RES_TY22_2022-01_PaperInteractive.pdf
          https://revenuefiles.delaware.gov/2022/PIT-RES_TY22_2022-02_Instructions.pdf
          https://revenue.delaware.gov/frequently-asked-questions/personal-income-tax-faqs/
          https://www.newcastlede.gov/DocumentCenter/View/20147/City-of-Wilmington-Wage-Tax-Information-PDF
        """
        
        # Check if the parameters are set upstream in the instance
        county_occupation = self.county_occupation if hasattr(self, 'county_occupation') else county_occupation
        pay_periods = self.pay_periods if hasattr(self, 'pay_periods') else pay_periods
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        federal_agi = self.federal_agi if hasattr(self, 'federal_agi') else federal_agi
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'WestVirginia', 'state_occupation': 'Delaware', "county_occupation": county_occupation, "pay_periods": pay_periods, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "federal_agi": federal_agi, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def west_virginia_and_florida(self, county_occupation='Other', pay_periods=24, state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        West Virginia:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_occupation : str, optional
            County of occupation. Default is Other.
        pay_periods : int, optional
            Number of pay periods. Options are 13, 24, 26, 52. Default is 24.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Florida:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        West Virginia:
        - https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://tax.wv.gov/Documents/TSD/tsd438.pdf
          https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2011/Taxes-11-35att.html
        
        Florida:
        -         
        - https://floridarevenue.com/taxes/taxesfees/Pages/tax_interest_rates.aspx
        """
        
        # Check if the parameters are set upstream in the instance
        county_occupation = self.county_occupation if hasattr(self, 'county_occupation') else county_occupation
        pay_periods = self.pay_periods if hasattr(self, 'pay_periods') else pay_periods
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'WestVirginia', 'state_occupation': 'Florida', "county_occupation": county_occupation, "pay_periods": pay_periods, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def west_virginia_and_georgia(self, pay_periods=24, capital_gains_short=0, county_occupation='Other', dependents=0, sources=False, state_agi=75000, filers_over_65=0, capital_gains_long=0, filing_status='single', income=100000, credits_state=0):
        """
        --- Parameters ----
        
        West Virginia:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_occupation : str, optional
            County of occupation. Default is Other.
        pay_periods : int, optional
            Number of pay periods. Options are 13, 24, 26, 52. Default is 24.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Georgia:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        West Virginia:
        - https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://tax.wv.gov/Documents/TSD/tsd438.pdf
          https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2011/Taxes-11-35att.html
        
        Georgia:
        -         
        - https://dor.georgia.gov/tax-tables-georgia-tax-rate-schedule
          https://apps.dor.ga.gov/FillableForms/PDFViewer/Index?form=2022GA500
          https://apps.dor.ga.gov/FillableForms/PDFViewer/Index?form=2022GA500
          https://apps.dor.ga.gov/FillableForms/PDFViewer/Index?form=2022GA500
          https://dor.georgia.gov/years-individual-income-tax-forms
          https://dor.georgia.gov/years-individual-income-tax-forms
        """
        
        # Check if the parameters are set upstream in the instance
        pay_periods = self.pay_periods if hasattr(self, 'pay_periods') else pay_periods
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        county_occupation = self.county_occupation if hasattr(self, 'county_occupation') else county_occupation
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'WestVirginia', 'state_occupation': 'Georgia', "pay_periods": pay_periods, "capital_gains_short": capital_gains_short, "county_occupation": county_occupation, "dependents": dependents, "sources": sources, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def west_virginia_and_hawaii(self, pay_periods=24, capital_gains_short=0, county_occupation='Other', dependents=0, sources=False, state_agi=75000, filers_over_65=0, capital_gains_long=0, filing_status='single', income=100000, credits_state=0):
        """
        --- Parameters ----
        
        West Virginia:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_occupation : str, optional
            County of occupation. Default is Other.
        pay_periods : int, optional
            Number of pay periods. Options are 13, 24, 26, 52. Default is 24.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Hawaii:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        West Virginia:
        - https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://tax.wv.gov/Documents/TSD/tsd438.pdf
          https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2011/Taxes-11-35att.html
        
        Hawaii:
        -         
        - https://tax.hawaii.gov/forms/d_18table-on/d_18table-on_p13/
          https://files.hawaii.gov/tax/forms/2018/18table-on.pdf
          https://files.hawaii.gov/tax/news/pubs/22outline.pdf
          https://files.hawaii.gov/tax/news/pubs/FastTaxRefGuide_2017.pdf
          https://www.tfhawaii.org/wordpress/state-tax-resources/mini-tax-guides/individual-income-tax-chapter-235/#:~:text=Long%20term%20capital%20gains%20are%20taxed%20at%20a%20maximum%20of%207.25%25
          https://www.grassrootinstitute.org/2023/02/hb337-increasing-tax-on-capital-gains-a-bad-risk/
          https://files.hawaii.gov/tax/news/pubs/22outline.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        pay_periods = self.pay_periods if hasattr(self, 'pay_periods') else pay_periods
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        county_occupation = self.county_occupation if hasattr(self, 'county_occupation') else county_occupation
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'WestVirginia', 'state_occupation': 'Hawaii', "pay_periods": pay_periods, "capital_gains_short": capital_gains_short, "county_occupation": county_occupation, "dependents": dependents, "sources": sources, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def west_virginia_and_idaho(self, pay_periods=24, capital_gains_short=0, county_occupation='Other', dependents=0, sources=False, state_agi=75000, filers_over_65=0, capital_gains_long=0, filing_status='single', income=100000, credits_state=0):
        """
        --- Parameters ----
        
        West Virginia:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_occupation : str, optional
            County of occupation. Default is Other.
        pay_periods : int, optional
            Number of pay periods. Options are 13, 24, 26, 52. Default is 24.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Idaho:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        West Virginia:
        - https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://tax.wv.gov/Documents/TSD/tsd438.pdf
          https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2011/Taxes-11-35att.html
        
        Idaho:
        -         
        - https://tax.idaho.gov/taxes/income-tax/individual-income/individual-income-tax-rate-schedule/
          https://tax.idaho.gov/wp-content/uploads/forms/EIN00046/EIN00046_03-01-2023.pdf
          https://tax.idaho.gov/taxes/income-tax/individual-income/idaho-source-income/
          https://tax.idaho.gov/taxes/income-tax/individual-income/idaho-source-income/
        """
        
        # Check if the parameters are set upstream in the instance
        pay_periods = self.pay_periods if hasattr(self, 'pay_periods') else pay_periods
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        county_occupation = self.county_occupation if hasattr(self, 'county_occupation') else county_occupation
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'WestVirginia', 'state_occupation': 'Idaho', "pay_periods": pay_periods, "capital_gains_short": capital_gains_short, "county_occupation": county_occupation, "dependents": dependents, "sources": sources, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def west_virginia_and_illinois(self, county_occupation='Other', pay_periods=24, state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        West Virginia:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_occupation : str, optional
            County of occupation. Default is Other.
        pay_periods : int, optional
            Number of pay periods. Options are 13, 24, 26, 52. Default is 24.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Illinois:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        West Virginia:
        - https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://tax.wv.gov/Documents/TSD/tsd438.pdf
          https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2011/Taxes-11-35att.html
        
        Illinois:
        -         
        - https://tax.illinois.gov/research/taxrates/income.html#IndividualIncome
          https://tax.illinois.gov/content/dam/soi/en/web/tax/forms/incometax/documents/currentyear/individual/il-1040-instr.pdf
          https://tax.illinois.gov/questionsandanswers/12.html
          https://tax.illinois.gov/content/dam/soi/en/web/tax/forms/incometax/documents/currentyear/individual/il-1040-instr.pdf
          https://tax.illinois.gov/research/taxrates/personalproperty.html
          https://tax.illinois.gov/localgovernments/income.html
        """
        
        # Check if the parameters are set upstream in the instance
        county_occupation = self.county_occupation if hasattr(self, 'county_occupation') else county_occupation
        pay_periods = self.pay_periods if hasattr(self, 'pay_periods') else pay_periods
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'WestVirginia', 'state_occupation': 'Illinois', "county_occupation": county_occupation, "pay_periods": pay_periods, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def west_virginia_and_indiana(self, pay_periods=24, capital_gains_short=0, county_occupation='Other', dependents=0, sources=False, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_long=0, filing_status='single', income=100000, credits_state=0):
        """
        --- Parameters ----
        
        West Virginia:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_occupation : str, optional
            County of occupation. Default is Other.
        pay_periods : int, optional
            Number of pay periods. Options are 13, 24, 26, 52. Default is 24.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Indiana:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        West Virginia:
        - https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://tax.wv.gov/Documents/TSD/tsd438.pdf
          https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2011/Taxes-11-35att.html
        
        Indiana:
        -         
        - https://www.in.gov/dor/tax-forms/2022-individual-income-tax-forms/
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-32.htm
          https://www.in.gov/dor/files/ib28.pdf
          https://www.in.gov/dor/tax-forms/2022-individual-income-tax-forms/
          https://www.in.gov/dor/files/2022-county-tax-rates-and-codes.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        pay_periods = self.pay_periods if hasattr(self, 'pay_periods') else pay_periods
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        county_occupation = self.county_occupation if hasattr(self, 'county_occupation') else county_occupation
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'WestVirginia', 'state_occupation': 'Indiana', "pay_periods": pay_periods, "capital_gains_short": capital_gains_short, "county_occupation": county_occupation, "dependents": dependents, "sources": sources, "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def west_virginia_and_iowa(self, pay_periods=24, capital_gains_short=0, county_occupation='Other', dependents=0, sources=False, county_residence='Other', state_agi=75000, capital_gains_long=0, filing_status='single', income=100000, credits_state=0):
        """
        --- Parameters ----
        
        West Virginia:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_occupation : str, optional
            County of occupation. Default is Other.
        pay_periods : int, optional
            Number of pay periods. Options are 13, 24, 26, 52. Default is 24.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Iowa:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        West Virginia:
        - https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://tax.wv.gov/Documents/TSD/tsd438.pdf
          https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2011/Taxes-11-35att.html
        
        Iowa:
        -         
        - https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
          https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
          https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
          https://tax.iowa.gov/iowa-illinois-reciprocal-agreement
          https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
        """
        
        # Check if the parameters are set upstream in the instance
        pay_periods = self.pay_periods if hasattr(self, 'pay_periods') else pay_periods
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        county_occupation = self.county_occupation if hasattr(self, 'county_occupation') else county_occupation
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'WestVirginia', 'state_occupation': 'Iowa', "pay_periods": pay_periods, "capital_gains_short": capital_gains_short, "county_occupation": county_occupation, "dependents": dependents, "sources": sources, "county_residence": county_residence, "state_agi": state_agi, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def west_virginia_and_kansas(self, pay_periods=24, capital_gains_short=0, county_occupation='Other', dependents=0, sources=False, state_agi=75000, filers_over_65=0, capital_gains_long=0, filing_status='single', income=100000, credits_state=0):
        """
        --- Parameters ----
        
        West Virginia:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_occupation : str, optional
            County of occupation. Default is Other.
        pay_periods : int, optional
            Number of pay periods. Options are 13, 24, 26, 52. Default is 24.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Kansas:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        West Virginia:
        - https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://tax.wv.gov/Documents/TSD/tsd438.pdf
          https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2011/Taxes-11-35att.html
        
        Kansas:
        -         
        - https://www.ksrevenue.gov/taxrates.html
          https://www.ksrevenue.gov/incomebook22.html#0
          https://www.ksrevenue.gov/incomebook22.html#0
          https://www.ksrevenue.gov/incomebook22.html#0
          https://www.ksrevenue.gov/incomebook22.html
          https://www.ksrevenue.gov/incomebook22.html
        """
        
        # Check if the parameters are set upstream in the instance
        pay_periods = self.pay_periods if hasattr(self, 'pay_periods') else pay_periods
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        county_occupation = self.county_occupation if hasattr(self, 'county_occupation') else county_occupation
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'WestVirginia', 'state_occupation': 'Kansas', "pay_periods": pay_periods, "capital_gains_short": capital_gains_short, "county_occupation": county_occupation, "dependents": dependents, "sources": sources, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def west_virginia_and_kentucky(self, county_occupation='Other', pay_periods=24, state_agi=75000, county_residence='Other', capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        West Virginia:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_occupation : str, optional
            County of occupation. Default is Other.
        pay_periods : int, optional
            Number of pay periods. Options are 13, 24, 26, 52. Default is 24.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Kentucky:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        West Virginia:
        - https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://tax.wv.gov/Documents/TSD/tsd438.pdf
          https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2011/Taxes-11-35att.html
        
        Kentucky:
        -         
        - https://revenue.ky.gov/News/Pages/Dor-Announces-Updates-to-Individual-Income-Tax-for-2023-Tax-Year.aspx
          https://revenue.ky.gov/Individual/Individual-Income-Tax/Pages/default.aspx
          https://revenue.ky.gov/News/Pages/Dor-Announces-Updates-to-Individual-Income-Tax-for-2023-Tax-Year.aspx
          https://revenue.ky.gov/Individual/Individual-Income-Tax/Pages/default.aspx
          https://revenue.ky.gov/Dor%20Training%20Materials/103%20KAR%2017.140.%20Individual%20income%20tax%20-%20reciprocity%20-%20nonresidents.pdf
          https://revenue.ky.gov/Individual/Individual-Income-Tax/Pages/default.aspx
          https://nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2021/Taxes-21-32.htm?taxmap=true
        """
        
        # Check if the parameters are set upstream in the instance
        county_occupation = self.county_occupation if hasattr(self, 'county_occupation') else county_occupation
        pay_periods = self.pay_periods if hasattr(self, 'pay_periods') else pay_periods
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'WestVirginia', 'state_occupation': 'Kentucky', "county_occupation": county_occupation, "pay_periods": pay_periods, "state_agi": state_agi, "county_residence": county_residence, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def west_virginia_and_louisiana(self, pay_periods=24, capital_gains_short=0, county_occupation='Other', dependents=0, sources=False, state_agi=75000, filers_over_65=0, capital_gains_long=0, filing_status='single', income=100000, credits_state=0):
        """
        --- Parameters ----
        
        West Virginia:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_occupation : str, optional
            County of occupation. Default is Other.
        pay_periods : int, optional
            Number of pay periods. Options are 13, 24, 26, 52. Default is 24.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Louisiana:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        West Virginia:
        - https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://tax.wv.gov/Documents/TSD/tsd438.pdf
          https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2011/Taxes-11-35att.html
        
        Louisiana:
        -         
        - https://revenue.louisiana.gov/individualincometax
          https://revenue.louisiana.gov/individualincometax
          https://www.revenue.louisiana.gov/taxforms/6935(11_02)F.pdf
          https://revenue.louisiana.gov/individualincometax
          https://revenue.louisiana.gov/individualincometax
        """
        
        # Check if the parameters are set upstream in the instance
        pay_periods = self.pay_periods if hasattr(self, 'pay_periods') else pay_periods
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        county_occupation = self.county_occupation if hasattr(self, 'county_occupation') else county_occupation
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'WestVirginia', 'state_occupation': 'Louisiana', "pay_periods": pay_periods, "capital_gains_short": capital_gains_short, "county_occupation": county_occupation, "dependents": dependents, "sources": sources, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def west_virginia_and_maine(self, pay_periods=24, capital_gains_short=0, county_occupation='Other', dependents=0, sources=False, state_agi=75000, filers_over_65=0, capital_gains_long=0, filing_status='single', income=100000, credits_state=0):
        """
        --- Parameters ----
        
        West Virginia:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_occupation : str, optional
            County of occupation. Default is Other.
        pay_periods : int, optional
            Number of pay periods. Options are 13, 24, 26, 52. Default is 24.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Maine:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        West Virginia:
        - https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://tax.wv.gov/Documents/TSD/tsd438.pdf
          https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2011/Taxes-11-35att.html
        
        Maine:
        -         
        - https://www.maine.gov/revenue/tax-return-forms/individual-income-tax-2022
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/ind_tax_rate_sched_2022.pdf
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/22_1040me_book_gen_instruc_revisedFeb23.pdf
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/22_1040me_book_gen_instruc_revisedFeb23.pdf
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/22_1040me_book_gen_instruc_revisedFeb23.pdf
          https://www.maine.gov/revenue/tax-return-forms/individual-income-tax-2022
        """
        
        # Check if the parameters are set upstream in the instance
        pay_periods = self.pay_periods if hasattr(self, 'pay_periods') else pay_periods
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        county_occupation = self.county_occupation if hasattr(self, 'county_occupation') else county_occupation
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'WestVirginia', 'state_occupation': 'Maine', "pay_periods": pay_periods, "capital_gains_short": capital_gains_short, "county_occupation": county_occupation, "dependents": dependents, "sources": sources, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def west_virginia_and_maryland(self, pay_periods=24, capital_gains_short=0, county_occupation='Other', dependents=0, sources=False, county_residence='Other', state_agi=75000, capital_gains_long=0, filing_status='single', income=100000, credits_state=0):
        """
        --- Parameters ----
        
        West Virginia:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_occupation : str, optional
            County of occupation. Default is Other.
        pay_periods : int, optional
            Number of pay periods. Options are 13, 24, 26, 52. Default is 24.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Maryland:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        West Virginia:
        - https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://tax.wv.gov/Documents/TSD/tsd438.pdf
          https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2011/Taxes-11-35att.html
        
        Maryland:
        -         
        - https://www.marylandtaxes.gov/individual/income/tax-info/tax-rates.php
          https://www.marylandtaxes.gov/forms/22_forms/Resident_Booklet.pdf
          https://www.marylandtaxes.gov/forms/Tax_Publications/Administrative_Releases/Income_and_Estate_Tax_Releases/ar_it37.pdf
          https://www.marylandtaxes.gov/individual/index.php
          https://www.marylandtaxes.gov/individual/income/tax-info/tax-rates.php
        """
        
        # Check if the parameters are set upstream in the instance
        pay_periods = self.pay_periods if hasattr(self, 'pay_periods') else pay_periods
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        county_occupation = self.county_occupation if hasattr(self, 'county_occupation') else county_occupation
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'WestVirginia', 'state_occupation': 'Maryland', "pay_periods": pay_periods, "capital_gains_short": capital_gains_short, "county_occupation": county_occupation, "dependents": dependents, "sources": sources, "county_residence": county_residence, "state_agi": state_agi, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def west_virginia_and_massachusetts(self, county_occupation='Other', pay_periods=24, state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        West Virginia:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_occupation : str, optional
            County of occupation. Default is Other.
        pay_periods : int, optional
            Number of pay periods. Options are 13, 24, 26, 52. Default is 24.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Massachusetts:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        West Virginia:
        - https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://tax.wv.gov/Documents/TSD/tsd438.pdf
          https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2011/Taxes-11-35att.html
        
        Massachusetts:
        -         
        - https://www.mass.gov/service-details/massachusetts-tax-rates
          https://www.mass.gov/service-details/view-massachusetts-personal-income-tax-exemptions
          https://www.mass.gov/service-details/view-massachusetts-personal-income-tax-exemptions
          https://www.mass.gov/service-details/view-child-and-dependent-related-credits
          https://www.mass.gov/service-details/massachusetts-tax-rates
          https://www.mass.gov/personal-income-tax
        """
        
        # Check if the parameters are set upstream in the instance
        county_occupation = self.county_occupation if hasattr(self, 'county_occupation') else county_occupation
        pay_periods = self.pay_periods if hasattr(self, 'pay_periods') else pay_periods
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'WestVirginia', 'state_occupation': 'Massachusetts', "county_occupation": county_occupation, "pay_periods": pay_periods, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def west_virginia_and_michigan(self, pay_periods=24, capital_gains_short=0, county_occupation='Other', dependents=0, sources=False, county_residence='Other', state_agi=75000, capital_gains_long=0, filing_status='single', income=100000, credits_state=0):
        """
        --- Parameters ----
        
        West Virginia:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_occupation : str, optional
            County of occupation. Default is Other.
        pay_periods : int, optional
            Number of pay periods. Options are 13, 24, 26, 52. Default is 24.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Michigan:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        county_occupation : str, optional
            County of occupation. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        West Virginia:
        - https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://tax.wv.gov/Documents/TSD/tsd438.pdf
          https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2011/Taxes-11-35att.html
        
        Michigan:
        -         
        - https://www.michigan.gov/taxes/iit/new-developments-for-tax-year-2022
          "https://www.michigan.gov/taxes/-/media/Project/Websites/taxes/Forms/2023/2023-SUW/446_Withholding-Guide_2023.pdf?rev=177ca3d8cb034392bf84e317ed78c320&hash=19C5F42F43A89A91863D229349F48883#:~:text=The%20exemption%20amount%20is%20%245%2C400
         employees%20Michigan%20income%20tax%20return"
          "https://www.michigan.gov/taxes/-/media/Project/Websites/taxes/Forms/2023/2023-SUW/446_Withholding-Guide_2023.pdf?rev=177ca3d8cb034392bf84e317ed78c320&hash=19C5F42F43A89A91863D229349F48883#:~:text=The%20exemption%20amount%20is%20%245%2C400
         employees%20Michigan%20income%20tax%20return"
          https://www.michigan.gov/taxes/questions/iit/accordion/residency/are-my-wages-earned-in-another-state-taxable-in-michigan-if-i-am-a-michigan-resident-1
          https://www.michigan.gov/taxes/-/media/Project/Websites/taxes/Forms/2022/2022-IIT-Forms/BOOK_MI-1040.pdf?rev=4524f1cbbc0243b194ab05a6680be75b&hash=40DFE71E651F7C1D65C31F6CD03D944F
          https://www.michigan.gov/taxes/questions/iit/accordion/general/what-cities-impose-an-income-tax
        """
        
        # Check if the parameters are set upstream in the instance
        pay_periods = self.pay_periods if hasattr(self, 'pay_periods') else pay_periods
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        county_occupation = self.county_occupation if hasattr(self, 'county_occupation') else county_occupation
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'WestVirginia', 'state_occupation': 'Michigan', "pay_periods": pay_periods, "capital_gains_short": capital_gains_short, "county_occupation": county_occupation, "dependents": dependents, "sources": sources, "county_residence": county_residence, "state_agi": state_agi, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def west_virginia_and_minnesota(self, county_occupation='Other', pay_periods=24, state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        West Virginia:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_occupation : str, optional
            County of occupation. Default is Other.
        pay_periods : int, optional
            Number of pay periods. Options are 13, 24, 26, 52. Default is 24.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Minnesota:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        West Virginia:
        - https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://tax.wv.gov/Documents/TSD/tsd438.pdf
          https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2011/Taxes-11-35att.html
        
        Minnesota:
        -         
        - https://www.revenue.state.mn.us/minnesota-income-tax-rates-and-brackets
          https://www.revenue.state.mn.us/child-and-dependent-care-credit
          https://www.revenue.state.mn.us/dependent-exemptions
          https://www.revenue.state.mn.us/minnesota-standard-deduction
          https://mn.gov/mmb/assets/mwr_form_tcm1059-128581.pdf
          https://www.revenue.state.mn.us/individual-income-tax
          https://www.revenue.state.mn.us/individual-income-tax
        """
        
        # Check if the parameters are set upstream in the instance
        county_occupation = self.county_occupation if hasattr(self, 'county_occupation') else county_occupation
        pay_periods = self.pay_periods if hasattr(self, 'pay_periods') else pay_periods
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'WestVirginia', 'state_occupation': 'Minnesota', "county_occupation": county_occupation, "pay_periods": pay_periods, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def west_virginia_and_mississippi(self, county_occupation='Other', pay_periods=24, state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        West Virginia:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_occupation : str, optional
            County of occupation. Default is Other.
        pay_periods : int, optional
            Number of pay periods. Options are 13, 24, 26, 52. Default is 24.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Mississippi:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        West Virginia:
        - https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://tax.wv.gov/Documents/TSD/tsd438.pdf
          https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2011/Taxes-11-35att.html
        
        Mississippi:
        -         
        - https://www.dor.ms.gov/individual/tax-rates
          https://www.dor.ms.gov/individual/tax-rates
          https://www.dor.ms.gov/individual/tax-rates
          https://www.dor.ms.gov/individual
          https://www.dor.ms.gov/individual
        """
        
        # Check if the parameters are set upstream in the instance
        county_occupation = self.county_occupation if hasattr(self, 'county_occupation') else county_occupation
        pay_periods = self.pay_periods if hasattr(self, 'pay_periods') else pay_periods
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'WestVirginia', 'state_occupation': 'Mississippi', "county_occupation": county_occupation, "pay_periods": pay_periods, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def west_virginia_and_missouri(self, pay_periods=24, capital_gains_short=0, county_occupation='Other', dependents=0, sources=False, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_long=0, filing_status='single', income=100000, credits_state=0):
        """
        --- Parameters ----
        
        West Virginia:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_occupation : str, optional
            County of occupation. Default is Other.
        pay_periods : int, optional
            Number of pay periods. Options are 13, 24, 26, 52. Default is 24.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Missouri:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        West Virginia:
        - https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://tax.wv.gov/Documents/TSD/tsd438.pdf
          https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2011/Taxes-11-35att.html
        
        Missouri:
        -         
        - https://dor.mo.gov/taxation/individual/tax-types/income/year-changes/
          https://dor.mo.gov/taxation/individual/tax-types/income/year-changes/
          https://dor.mo.gov/taxation/individual/tax-types/income/year-changes/
          https://www.kcmo.gov/city-hall/departments/finance/earnings-tax
          https://www.stlouis-mo.gov/government/departments/collector/earnings-tax/individual-earnings-tax-info.cfm
        """
        
        # Check if the parameters are set upstream in the instance
        pay_periods = self.pay_periods if hasattr(self, 'pay_periods') else pay_periods
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        county_occupation = self.county_occupation if hasattr(self, 'county_occupation') else county_occupation
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'WestVirginia', 'state_occupation': 'Missouri', "pay_periods": pay_periods, "capital_gains_short": capital_gains_short, "county_occupation": county_occupation, "dependents": dependents, "sources": sources, "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def west_virginia_and_montana(self, pay_periods=24, capital_gains_short=0, county_occupation='Other', dependents=0, sources=False, state_agi=75000, filers_over_65=0, capital_gains_long=0, filing_status='single', income=100000, credits_state=0):
        """
        --- Parameters ----
        
        West Virginia:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_occupation : str, optional
            County of occupation. Default is Other.
        pay_periods : int, optional
            Number of pay periods. Options are 13, 24, 26, 52. Default is 24.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Montana:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        West Virginia:
        - https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://tax.wv.gov/Documents/TSD/tsd438.pdf
          https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2011/Taxes-11-35att.html
        
        Montana:
        -         
        - https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/2022/12/montana-employees-withholding-allowance-and-exemption-certificate-form-mw-4.pdf
          https://mtrevenue.gov/2021/11/29/simplification-of-montana-income-taxation/?utm_source=rss&utm_medium=rss&utm_campaign=simplification-of-montana-income-taxation
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        pay_periods = self.pay_periods if hasattr(self, 'pay_periods') else pay_periods
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        county_occupation = self.county_occupation if hasattr(self, 'county_occupation') else county_occupation
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'WestVirginia', 'state_occupation': 'Montana', "pay_periods": pay_periods, "capital_gains_short": capital_gains_short, "county_occupation": county_occupation, "dependents": dependents, "sources": sources, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def west_virginia_and_nebraska(self, county_occupation='Other', pay_periods=24, state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        West Virginia:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_occupation : str, optional
            County of occupation. Default is Other.
        pay_periods : int, optional
            Number of pay periods. Options are 13, 24, 26, 52. Default is 24.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Nebraska:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        West Virginia:
        - https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://tax.wv.gov/Documents/TSD/tsd438.pdf
          https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2011/Taxes-11-35att.html
        
        Nebraska:
        -         
        - https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2023_Circular_EN_8-429r2023_Whole_Book_11.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        county_occupation = self.county_occupation if hasattr(self, 'county_occupation') else county_occupation
        pay_periods = self.pay_periods if hasattr(self, 'pay_periods') else pay_periods
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'WestVirginia', 'state_occupation': 'Nebraska', "county_occupation": county_occupation, "pay_periods": pay_periods, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def west_virginia_and_nevada(self, county_occupation='Other', pay_periods=24, state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        West Virginia:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_occupation : str, optional
            County of occupation. Default is Other.
        pay_periods : int, optional
            Number of pay periods. Options are 13, 24, 26, 52. Default is 24.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Nevada:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        West Virginia:
        - https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://tax.wv.gov/Documents/TSD/tsd438.pdf
          https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2011/Taxes-11-35att.html
        
        Nevada:
        -         
        - https://www.nvsos.gov/sos/businesses/the-nevada-advantage
        """
        
        # Check if the parameters are set upstream in the instance
        county_occupation = self.county_occupation if hasattr(self, 'county_occupation') else county_occupation
        pay_periods = self.pay_periods if hasattr(self, 'pay_periods') else pay_periods
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'WestVirginia', 'state_occupation': 'Nevada', "county_occupation": county_occupation, "pay_periods": pay_periods, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def west_virginia_and_new_hampshire(self, county_occupation='Other', pay_periods=24, state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        West Virginia:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_occupation : str, optional
            County of occupation. Default is Other.
        pay_periods : int, optional
            Number of pay periods. Options are 13, 24, 26, 52. Default is 24.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        New Hampshire:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        West Virginia:
        - https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://tax.wv.gov/Documents/TSD/tsd438.pdf
          https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2011/Taxes-11-35att.html
        
        New Hampshire:
        -         
        - https://www.revenue.nh.gov/assistance/tax-overview.htm
        """
        
        # Check if the parameters are set upstream in the instance
        county_occupation = self.county_occupation if hasattr(self, 'county_occupation') else county_occupation
        pay_periods = self.pay_periods if hasattr(self, 'pay_periods') else pay_periods
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'WestVirginia', 'state_occupation': 'NewHampshire', "county_occupation": county_occupation, "pay_periods": pay_periods, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def west_virginia_and_new_jersey(self, pay_periods=24, capital_gains_short=0, county_occupation='Other', dependents=0, sources=False, state_agi=75000, filers_over_65=0, capital_gains_long=0, filing_status='single', income=100000, credits_state=0):
        """
        --- Parameters ----
        
        West Virginia:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_occupation : str, optional
            County of occupation. Default is Other.
        pay_periods : int, optional
            Number of pay periods. Options are 13, 24, 26, 52. Default is 24.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        New Jersey:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        West Virginia:
        - https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://tax.wv.gov/Documents/TSD/tsd438.pdf
          https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2011/Taxes-11-35att.html
        
        New Jersey:
        -         
        - https://www.state.nj.us/treasury/taxation/taxtables.shtml
          https://www.state.nj.us/treasury/taxation/njit13.shtml
          https://www.nj.gov/treasury/taxation/njit25.shtml
          https://www.state.nj.us/treasury/taxation/individuals/freqqiti.shtml
          https://www.state.nj.us/treasury/taxation/localtax.shtml
        """
        
        # Check if the parameters are set upstream in the instance
        pay_periods = self.pay_periods if hasattr(self, 'pay_periods') else pay_periods
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        county_occupation = self.county_occupation if hasattr(self, 'county_occupation') else county_occupation
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'WestVirginia', 'state_occupation': 'NewJersey', "pay_periods": pay_periods, "capital_gains_short": capital_gains_short, "county_occupation": county_occupation, "dependents": dependents, "sources": sources, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def west_virginia_and_new_mexico(self, county_occupation='Other', pay_periods=24, state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        West Virginia:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_occupation : str, optional
            County of occupation. Default is Other.
        pay_periods : int, optional
            Number of pay periods. Options are 13, 24, 26, 52. Default is 24.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        New Mexico:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        West Virginia:
        - https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://tax.wv.gov/Documents/TSD/tsd438.pdf
          https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2011/Taxes-11-35att.html
        
        New Mexico:
        -         
        - https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/fdf3c548-8aba-4b9c-9eb4-bb564c716015/FYI-104.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/0bf943fd-652e-4400-bb1b-cd8397eb5a95/2022pit-adj.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/2f1a6781-9534-4436-b427-1557f9592099/2022pit-adj-ins.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/0bf943fd-652e-4400-bb1b-cd8397eb5a95/2022pit-adj.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/1afc56af-ea90-4d48-82e5-1f9aeb43255a/PITbook2022.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/fdf3c548-8aba-4b9c-9eb4-bb564c716015/FYI-104.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        county_occupation = self.county_occupation if hasattr(self, 'county_occupation') else county_occupation
        pay_periods = self.pay_periods if hasattr(self, 'pay_periods') else pay_periods
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'WestVirginia', 'state_occupation': 'NewMexico', "county_occupation": county_occupation, "pay_periods": pay_periods, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def west_virginia_and_new_york(self, pay_periods=24, capital_gains_short=0, county_occupation='Other', dependents=0, sources=False, county_residence='Other', state_agi=75000, capital_gains_long=0, filing_status='single', income=100000, credits_state=0):
        """
        --- Parameters ----
        
        West Virginia:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_occupation : str, optional
            County of occupation. Default is Other.
        pay_periods : int, optional
            Number of pay periods. Options are 13, 24, 26, 52. Default is 24.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        New York:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        West Virginia:
        - https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://tax.wv.gov/Documents/TSD/tsd438.pdf
          https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2011/Taxes-11-35att.html
        
        New York:
        -         
        - https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nys-tax-rate-schedule
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nys-tax-rate-schedule
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nys-tax-rate-schedule
          https://www.tax.ny.gov/pit/
          https://www.tax.ny.gov/pit/file/nonresident-faqs.htm
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#yonkers-tax-rate-schedule
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nyc-tax-rate-schedule
        """
        
        # Check if the parameters are set upstream in the instance
        pay_periods = self.pay_periods if hasattr(self, 'pay_periods') else pay_periods
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        county_occupation = self.county_occupation if hasattr(self, 'county_occupation') else county_occupation
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'WestVirginia', 'state_occupation': 'NewYork', "pay_periods": pay_periods, "capital_gains_short": capital_gains_short, "county_occupation": county_occupation, "dependents": dependents, "sources": sources, "county_residence": county_residence, "state_agi": state_agi, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def west_virginia_and_north_carolina(self, county_occupation='Other', pay_periods=24, state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        West Virginia:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_occupation : str, optional
            County of occupation. Default is Other.
        pay_periods : int, optional
            Number of pay periods. Options are 13, 24, 26, 52. Default is 24.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        North Carolina:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        West Virginia:
        - https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://tax.wv.gov/Documents/TSD/tsd438.pdf
          https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2011/Taxes-11-35att.html
        
        North Carolina:
        -         
        - https://www.ncdor.gov/taxes-forms/tax-rate-schedules
          https://www.ncdor.gov/taxes-forms/individual-income-tax/north-carolina-standard-deduction-or-north-carolina-itemized-deductions
          https://www.ncdor.gov/taxes-forms/individual-income-tax
          https://www.ncdor.gov/taxes-forms/individual-income-tax
        """
        
        # Check if the parameters are set upstream in the instance
        county_occupation = self.county_occupation if hasattr(self, 'county_occupation') else county_occupation
        pay_periods = self.pay_periods if hasattr(self, 'pay_periods') else pay_periods
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'WestVirginia', 'state_occupation': 'NorthCarolina', "county_occupation": county_occupation, "pay_periods": pay_periods, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def west_virginia_and_north_dakota(self, pay_periods=24, capital_gains_short=0, county_occupation='Other', dependents=0, sources=False, state_agi=75000, filers_over_65=0, capital_gains_long=0, filing_status='single', income=100000, credits_state=0):
        """
        --- Parameters ----
        
        West Virginia:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_occupation : str, optional
            County of occupation. Default is Other.
        pay_periods : int, optional
            Number of pay periods. Options are 13, 24, 26, 52. Default is 24.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        North Dakota:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        West Virginia:
        - https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://tax.wv.gov/Documents/TSD/tsd438.pdf
          https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2011/Taxes-11-35att.html
        
        North Dakota:
        -         
        - https://www.tax.nd.gov/forms
          https://www.tax.nd.gov/sites/www/files/documents/forms/individual/2022-iit/2022-individual-income-tax-booklet.pdf
          https://www.tax.nd.gov/sites/www/files/documents/forms/2022-individual-income-tax-booklet.pdf
          https://www.tax.nd.gov/tax-exemptions-credits/income-tax-exemptions-credits
          https://www.tax.nd.gov/sites/www/files/documents/forms/business/ndwrfillable.pdf
          https://www.tax.nd.gov/news/tax-legislative-changes/significant-changes-law/individual-income-tax-history
          https://www.tax.nd.gov/forms
        """
        
        # Check if the parameters are set upstream in the instance
        pay_periods = self.pay_periods if hasattr(self, 'pay_periods') else pay_periods
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        county_occupation = self.county_occupation if hasattr(self, 'county_occupation') else county_occupation
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'WestVirginia', 'state_occupation': 'NorthDakota', "pay_periods": pay_periods, "capital_gains_short": capital_gains_short, "county_occupation": county_occupation, "dependents": dependents, "sources": sources, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def west_virginia_and_ohio(self, pay_periods=24, capital_gains_short=0, county_occupation='Other', dependents=0, sources=False, county_residence='Other', state_agi=75000, capital_gains_long=0, filing_status='single', income=100000, credits_state=0):
        """
        --- Parameters ----
        
        West Virginia:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_occupation : str, optional
            County of occupation. Default is Other.
        pay_periods : int, optional
            Number of pay periods. Options are 13, 24, 26, 52. Default is 24.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Ohio:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        West Virginia:
        - https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://tax.wv.gov/Documents/TSD/tsd438.pdf
          https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2011/Taxes-11-35att.html
        
        Ohio:
        -         
        - https://tax.ohio.gov/individual/resources/annual-tax-rates
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://tax.ohio.gov/static/forms/employer_withholding/generic/wth_it4nr.pdf
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://www.ritaohio.com/TaxRatesTable
        """
        
        # Check if the parameters are set upstream in the instance
        pay_periods = self.pay_periods if hasattr(self, 'pay_periods') else pay_periods
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        county_occupation = self.county_occupation if hasattr(self, 'county_occupation') else county_occupation
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'WestVirginia', 'state_occupation': 'Ohio', "pay_periods": pay_periods, "capital_gains_short": capital_gains_short, "county_occupation": county_occupation, "dependents": dependents, "sources": sources, "county_residence": county_residence, "state_agi": state_agi, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def west_virginia_and_oklahoma(self, pay_periods=24, capital_gains_short=0, county_occupation='Other', dependents=0, sources=False, state_agi=75000, filers_over_65=0, capital_gains_long=0, filing_status='single', income=100000, credits_state=0):
        """
        --- Parameters ----
        
        West Virginia:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_occupation : str, optional
            County of occupation. Default is Other.
        pay_periods : int, optional
            Number of pay periods. Options are 13, 24, 26, 52. Default is 24.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Oklahoma:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        West Virginia:
        - https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://tax.wv.gov/Documents/TSD/tsd438.pdf
          https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2011/Taxes-11-35att.html
        
        Oklahoma:
        -         
        - https://oklahoma.gov/content/dam/ok/en/tax/documents/resources/publications/businesses/withholding-tables/WHTables-2022.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        pay_periods = self.pay_periods if hasattr(self, 'pay_periods') else pay_periods
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        county_occupation = self.county_occupation if hasattr(self, 'county_occupation') else county_occupation
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'WestVirginia', 'state_occupation': 'Oklahoma', "pay_periods": pay_periods, "capital_gains_short": capital_gains_short, "county_occupation": county_occupation, "dependents": dependents, "sources": sources, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def west_virginia_and_oregon(self, pay_periods=24, capital_gains_short=0, county_occupation='Other', dependents=0, sources=False, state_agi=75000, filers_over_65=0, capital_gains_long=0, filing_status='single', income=100000, credits_state=0):
        """
        --- Parameters ----
        
        West Virginia:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_occupation : str, optional
            County of occupation. Default is Other.
        pay_periods : int, optional
            Number of pay periods. Options are 13, 24, 26, 52. Default is 24.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Oregon:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        West Virginia:
        - https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://tax.wv.gov/Documents/TSD/tsd438.pdf
          https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2011/Taxes-11-35att.html
        
        Oregon:
        -         
        - https://www.oregon.gov/dor/programs/individuals/Pages/PIT.aspx
          https://www.oregon.gov/dor/programs/individuals/Documents/Full%20year%20resident
         %20Form%20or-40%20filers.pdf
          https://www.oregon.gov/dor/programs/individuals/Documents/Part-year%20and%20nonresident
         %20Form%20or-40-P%20and%20or-40-N%20filers.pdf.pdf
          https://www.oregon.gov/dor/forms/FormsPubs/form-or-40_101-040_2022.pdf
          https://www.oregon.gov/dor/forms/FormsPubs/form-or-40_101-040_2022.pdf
          https://www.oregon.gov/dor/forms/FormsPubs/form-or-40_101-040_2022.pdf
          https://www.portland.gov/revenue/personal-tax
        """
        
        # Check if the parameters are set upstream in the instance
        pay_periods = self.pay_periods if hasattr(self, 'pay_periods') else pay_periods
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        county_occupation = self.county_occupation if hasattr(self, 'county_occupation') else county_occupation
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'WestVirginia', 'state_occupation': 'Oregon', "pay_periods": pay_periods, "capital_gains_short": capital_gains_short, "county_occupation": county_occupation, "dependents": dependents, "sources": sources, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def west_virginia_and_pennsylvania(self, county_occupation='Other', pay_periods=24, state_agi=75000, county_residence='Other', capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        West Virginia:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_occupation : str, optional
            County of occupation. Default is Other.
        pay_periods : int, optional
            Number of pay periods. Options are 13, 24, 26, 52. Default is 24.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Pennsylvania:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        county_residence : str, optional
            County of residence. Default is Other.
        county_occupation : str, optional
            County of occupation. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        West Virginia:
        - https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://tax.wv.gov/Documents/TSD/tsd438.pdf
          https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2011/Taxes-11-35att.html
        
        Pennsylvania:
        -         
        - https://www.revenue.pa.gov/Tax%20Rates/Pages/default.aspx
          https://www.revenue.pa.gov/FormsandPublications/FormsforIndividuals/PIT/Documents/2022/2022_pa-40in.pdf
          https://www.revenue.pa.gov/FormsandPublications/FormsforBusinesses/EmployerWithholding/Documents/rev-419.pdf
          https://www.revenue.pa.gov/FormsandPublications/FormsforIndividuals/PIT/Documents/2022/2022_pa-40in.pdf
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-25.htm
        """
        
        # Check if the parameters are set upstream in the instance
        county_occupation = self.county_occupation if hasattr(self, 'county_occupation') else county_occupation
        pay_periods = self.pay_periods if hasattr(self, 'pay_periods') else pay_periods
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'WestVirginia', 'state_occupation': 'Pennsylvania', "county_occupation": county_occupation, "pay_periods": pay_periods, "state_agi": state_agi, "county_residence": county_residence, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def west_virginia_and_rhode_island(self, county_occupation='Other', pay_periods=24, state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        West Virginia:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_occupation : str, optional
            County of occupation. Default is Other.
        pay_periods : int, optional
            Number of pay periods. Options are 13, 24, 26, 52. Default is 24.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Rhode Island:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        West Virginia:
        - https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://tax.wv.gov/Documents/TSD/tsd438.pdf
          https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2011/Taxes-11-35att.html
        
        Rhode Island:
        -         
        - https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2022_1040WE_w_0.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        county_occupation = self.county_occupation if hasattr(self, 'county_occupation') else county_occupation
        pay_periods = self.pay_periods if hasattr(self, 'pay_periods') else pay_periods
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'WestVirginia', 'state_occupation': 'RhodeIsland', "county_occupation": county_occupation, "pay_periods": pay_periods, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def west_virginia_and_south_carolina(self, pay_periods=24, capital_gains_short=0, county_occupation='Other', dependents=0, sources=False, state_agi=75000, filers_over_65=0, capital_gains_long=0, filing_status='single', income=100000, credits_state=0):
        """
        --- Parameters ----
        
        West Virginia:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_occupation : str, optional
            County of occupation. Default is Other.
        pay_periods : int, optional
            Number of pay periods. Options are 13, 24, 26, 52. Default is 24.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        South Carolina:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        West Virginia:
        - https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://tax.wv.gov/Documents/TSD/tsd438.pdf
          https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2011/Taxes-11-35att.html
        
        South Carolina:
        -         
        - https://dor.sc.gov/resources-site/lawandpolicy/Advisory%20Opinions/IL22-15.pdf
          https://dor.sc.gov/forms-site/Forms/IITPacket_2022.pdf
          https://www.scstatehouse.gov/code/t12c006.php
          https://dor.sc.gov/resources-site/lawandpolicy/Advisory%20Opinions/IL22-15.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        pay_periods = self.pay_periods if hasattr(self, 'pay_periods') else pay_periods
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        county_occupation = self.county_occupation if hasattr(self, 'county_occupation') else county_occupation
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'WestVirginia', 'state_occupation': 'SouthCarolina', "pay_periods": pay_periods, "capital_gains_short": capital_gains_short, "county_occupation": county_occupation, "dependents": dependents, "sources": sources, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def west_virginia_and_south_dakota(self, county_occupation='Other', pay_periods=24, state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        West Virginia:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_occupation : str, optional
            County of occupation. Default is Other.
        pay_periods : int, optional
            Number of pay periods. Options are 13, 24, 26, 52. Default is 24.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        South Dakota:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        West Virginia:
        - https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://tax.wv.gov/Documents/TSD/tsd438.pdf
          https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2011/Taxes-11-35att.html
        
        South Dakota:
        -         
        - https://dor.sd.gov/individuals/taxes/
        """
        
        # Check if the parameters are set upstream in the instance
        county_occupation = self.county_occupation if hasattr(self, 'county_occupation') else county_occupation
        pay_periods = self.pay_periods if hasattr(self, 'pay_periods') else pay_periods
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'WestVirginia', 'state_occupation': 'SouthDakota', "county_occupation": county_occupation, "pay_periods": pay_periods, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def west_virginia_and_tennessee(self, county_occupation='Other', pay_periods=24, state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        West Virginia:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_occupation : str, optional
            County of occupation. Default is Other.
        pay_periods : int, optional
            Number of pay periods. Options are 13, 24, 26, 52. Default is 24.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Tennessee:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        West Virginia:
        - https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://tax.wv.gov/Documents/TSD/tsd438.pdf
          https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2011/Taxes-11-35att.html
        
        Tennessee:
        -         
        - https://www.tn.gov/revenue/taxes.html
        """
        
        # Check if the parameters are set upstream in the instance
        county_occupation = self.county_occupation if hasattr(self, 'county_occupation') else county_occupation
        pay_periods = self.pay_periods if hasattr(self, 'pay_periods') else pay_periods
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'WestVirginia', 'state_occupation': 'Tennessee', "county_occupation": county_occupation, "pay_periods": pay_periods, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def west_virginia_and_texas(self, county_occupation='Other', pay_periods=24, state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        West Virginia:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_occupation : str, optional
            County of occupation. Default is Other.
        pay_periods : int, optional
            Number of pay periods. Options are 13, 24, 26, 52. Default is 24.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Texas:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        West Virginia:
        - https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://tax.wv.gov/Documents/TSD/tsd438.pdf
          https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2011/Taxes-11-35att.html
        
        Texas:
        -         
        - https://comptroller.texas.gov/taxes/
        """
        
        # Check if the parameters are set upstream in the instance
        county_occupation = self.county_occupation if hasattr(self, 'county_occupation') else county_occupation
        pay_periods = self.pay_periods if hasattr(self, 'pay_periods') else pay_periods
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'WestVirginia', 'state_occupation': 'Texas', "county_occupation": county_occupation, "pay_periods": pay_periods, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def west_virginia_and_utah(self, county_occupation='Other', pay_periods=24, state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        West Virginia:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_occupation : str, optional
            County of occupation. Default is Other.
        pay_periods : int, optional
            Number of pay periods. Options are 13, 24, 26, 52. Default is 24.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Utah:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        West Virginia:
        - https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://tax.wv.gov/Documents/TSD/tsd438.pdf
          https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2011/Taxes-11-35att.html
        
        Utah:
        -         
        - https://incometax.utah.gov/paying/tax-rates
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        county_occupation = self.county_occupation if hasattr(self, 'county_occupation') else county_occupation
        pay_periods = self.pay_periods if hasattr(self, 'pay_periods') else pay_periods
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'WestVirginia', 'state_occupation': 'Utah', "county_occupation": county_occupation, "pay_periods": pay_periods, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def west_virginia_and_vermont(self, pay_periods=24, capital_gains_short=0, county_occupation='Other', dependents=0, sources=False, state_agi=75000, filers_over_65=0, capital_gains_long=0, filing_status='single', income=100000, credits_state=0):
        """
        --- Parameters ----
        
        West Virginia:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_occupation : str, optional
            County of occupation. Default is Other.
        pay_periods : int, optional
            Number of pay periods. Options are 13, 24, 26, 52. Default is 24.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Vermont:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        West Virginia:
        - https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://tax.wv.gov/Documents/TSD/tsd438.pdf
          https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2011/Taxes-11-35att.html
        
        Vermont:
        -         
        - https://tax.vermont.gov/sites/tax/files/documents/GB-1210-2023.pdf
          https://tax.vermont.gov/sites/tax/files/documents/Income%20Booklet-2022.pdf
          https://tax.vermont.gov/sites/tax/files/documents/Income%20Booklet-2022.pdf
          https://tax.vermont.gov/sites/tax/files/documents/Income%20Booklet-2022.pdf
          https://tax.vermont.gov/individuals/personal-income-tax/tax-credits
          https://tax.vermont.gov/individuals/personal-income-tax/taxable-income
          https://tax.vermont.gov/individuals/personal-income-tax/local-option-tax
        """
        
        # Check if the parameters are set upstream in the instance
        pay_periods = self.pay_periods if hasattr(self, 'pay_periods') else pay_periods
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        county_occupation = self.county_occupation if hasattr(self, 'county_occupation') else county_occupation
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'WestVirginia', 'state_occupation': 'Vermont', "pay_periods": pay_periods, "capital_gains_short": capital_gains_short, "county_occupation": county_occupation, "dependents": dependents, "sources": sources, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def west_virginia_and_virginia(self, pay_periods=24, capital_gains_short=0, county_occupation='Other', dependents=0, sources=False, state_agi=75000, filers_over_65=0, capital_gains_long=0, filing_status='single', income=100000, credits_state=0):
        """
        --- Parameters ----
        
        West Virginia:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_occupation : str, optional
            County of occupation. Default is Other.
        pay_periods : int, optional
            Number of pay periods. Options are 13, 24, 26, 52. Default is 24.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Virginia:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        West Virginia:
        - https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://tax.wv.gov/Documents/TSD/tsd438.pdf
          https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2011/Taxes-11-35att.html
        
        Virginia:
        -         
        - https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/reciprocity
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        pay_periods = self.pay_periods if hasattr(self, 'pay_periods') else pay_periods
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        county_occupation = self.county_occupation if hasattr(self, 'county_occupation') else county_occupation
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'WestVirginia', 'state_occupation': 'Virginia', "pay_periods": pay_periods, "capital_gains_short": capital_gains_short, "county_occupation": county_occupation, "dependents": dependents, "sources": sources, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def west_virginia_and_washington(self, county_occupation='Other', pay_periods=24, state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        West Virginia:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_occupation : str, optional
            County of occupation. Default is Other.
        pay_periods : int, optional
            Number of pay periods. Options are 13, 24, 26, 52. Default is 24.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Washington:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        West Virginia:
        - https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://tax.wv.gov/Documents/TSD/tsd438.pdf
          https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2011/Taxes-11-35att.html
        
        Washington:
        -         
        - https://dor.wa.gov/taxes-rates/income-tax
          https://dor.wa.gov/taxes-rates/other-taxes/capital-gains-tax
          https://dor.wa.gov/taxes-rates/income-tax
        """
        
        # Check if the parameters are set upstream in the instance
        county_occupation = self.county_occupation if hasattr(self, 'county_occupation') else county_occupation
        pay_periods = self.pay_periods if hasattr(self, 'pay_periods') else pay_periods
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'WestVirginia', 'state_occupation': 'Washington', "county_occupation": county_occupation, "pay_periods": pay_periods, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def west_virginia_and_wisconsin(self, county_occupation='Other', pay_periods=24, state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        West Virginia:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_occupation : str, optional
            County of occupation. Default is Other.
        pay_periods : int, optional
            Number of pay periods. Options are 13, 24, 26, 52. Default is 24.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Wisconsin:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        West Virginia:
        - https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://tax.wv.gov/Documents/TSD/tsd438.pdf
          https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2011/Taxes-11-35att.html
        
        Wisconsin:
        -         
        - https://www.revenue.wi.gov/TaxForms2023/2023-Form1-ES-Inst.pdf
          https://www.revenue.wi.gov/TaxForms2023/2023-Form1-ES-Inst.pdf
          https://www.revenue.wi.gov/TaxForms2023/2023-Form1-ES-Inst.pdf
          https://www.revenue.wi.gov/Pages/FAQS/pcs-work.aspx
          https://www.revenue.wi.gov/Dor%20Publications/pb103.pdf
          https://www.revenue.wi.gov/Pages/FAQS/pcs-work.aspx
        """
        
        # Check if the parameters are set upstream in the instance
        county_occupation = self.county_occupation if hasattr(self, 'county_occupation') else county_occupation
        pay_periods = self.pay_periods if hasattr(self, 'pay_periods') else pay_periods
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'WestVirginia', 'state_occupation': 'Wisconsin', "county_occupation": county_occupation, "pay_periods": pay_periods, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def west_virginia_and_wyoming(self, county_occupation='Other', pay_periods=24, state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        West Virginia:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_occupation : str, optional
            County of occupation. Default is Other.
        pay_periods : int, optional
            Number of pay periods. Options are 13, 24, 26, 52. Default is 24.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Wyoming:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        West Virginia:
        - https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://tax.wv.gov/Documents/TSD/tsd438.pdf
          https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2011/Taxes-11-35att.html
        
        Wyoming:
        -         
        - https://revenue.wyo.gov/
        """
        
        # Check if the parameters are set upstream in the instance
        county_occupation = self.county_occupation if hasattr(self, 'county_occupation') else county_occupation
        pay_periods = self.pay_periods if hasattr(self, 'pay_periods') else pay_periods
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'WestVirginia', 'state_occupation': 'Wyoming', "county_occupation": county_occupation, "pay_periods": pay_periods, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def wisconsin_and_alabama(self, county_residence='Other', state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Wisconsin:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Alabama:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Wisconsin:
        - https://www.revenue.wi.gov/TaxForms2023/2023-Form1-ES-Inst.pdf
          https://www.revenue.wi.gov/TaxForms2023/2023-Form1-ES-Inst.pdf
          https://www.revenue.wi.gov/TaxForms2023/2023-Form1-ES-Inst.pdf
          https://www.revenue.wi.gov/Pages/FAQS/pcs-work.aspx
          https://www.revenue.wi.gov/Dor%20Publications/pb103.pdf
          https://www.revenue.wi.gov/Pages/FAQS/pcs-work.aspx
        
        Alabama:
        -         
        - https://www.revenue.alabama.gov/ultraviewer/viewer/basic_viewer/index.html?form=2023/02/22f40.pdf
          https://www.revenue.alabama.gov/faqs/what-is-alabamas-individual-income-tax-rate/
          https://www.revenue.alabama.gov/ultraviewer/viewer/basic_viewer/index.html?form=2023/02/22f40.pdf
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-24.htm?taxmap=true
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-24.htm?taxmap=true
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-24.htm?taxmap=true
          https://www.revenue.alabama.gov/individual-corporate/income-to-be-reported-on-the-alabama-income-tax-return/
          https://www.almonline.org/TaxRates.aspx#OccupationalTax
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Wisconsin', 'state_occupation': 'Alabama', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def wisconsin_and_alaska(self, state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Wisconsin:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Alaska:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Wisconsin:
        - https://www.revenue.wi.gov/TaxForms2023/2023-Form1-ES-Inst.pdf
          https://www.revenue.wi.gov/TaxForms2023/2023-Form1-ES-Inst.pdf
          https://www.revenue.wi.gov/TaxForms2023/2023-Form1-ES-Inst.pdf
          https://www.revenue.wi.gov/Pages/FAQS/pcs-work.aspx
          https://www.revenue.wi.gov/Dor%20Publications/pb103.pdf
          https://www.revenue.wi.gov/Pages/FAQS/pcs-work.aspx
        
        Alaska:
        -         
        - https://www.commerce.alaska.gov/web/dcra/officeofthestateassessor/alaskataxfacts.aspx
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Wisconsin', 'state_occupation': 'Alaska', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def wisconsin_and_arizona(self, state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Wisconsin:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Arizona:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        credits_state : int, optional
            Credits for state taxes. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Wisconsin:
        - https://www.revenue.wi.gov/TaxForms2023/2023-Form1-ES-Inst.pdf
          https://www.revenue.wi.gov/TaxForms2023/2023-Form1-ES-Inst.pdf
          https://www.revenue.wi.gov/TaxForms2023/2023-Form1-ES-Inst.pdf
          https://www.revenue.wi.gov/Pages/FAQS/pcs-work.aspx
          https://www.revenue.wi.gov/Dor%20Publications/pb103.pdf
          https://www.revenue.wi.gov/Pages/FAQS/pcs-work.aspx
        
        Arizona:
        -         
        - https://azdor.gov/forms/individual/form-140-x-y-tables
          https://azdor.gov/forms/individual/form-140-x-y-tables
          https://azdor.gov/forms/individual/form-140-x-y-tables
          https://azdor.gov/businesses-arizona/withholding-tax/withholding-exceptions
          "https://azdor.gov/individuals/income-tax-filing-assistance/identifying-other-taxable-income#:~:text=In%20regards%20to%20capital%20gains
         the%20individuals%20regular%20tax%20rate.
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Wisconsin', 'state_occupation': 'Arizona', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def wisconsin_and_arkansas(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Wisconsin:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Arkansas:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Wisconsin:
        - https://www.revenue.wi.gov/TaxForms2023/2023-Form1-ES-Inst.pdf
          https://www.revenue.wi.gov/TaxForms2023/2023-Form1-ES-Inst.pdf
          https://www.revenue.wi.gov/TaxForms2023/2023-Form1-ES-Inst.pdf
          https://www.revenue.wi.gov/Pages/FAQS/pcs-work.aspx
          https://www.revenue.wi.gov/Dor%20Publications/pb103.pdf
          https://www.revenue.wi.gov/Pages/FAQS/pcs-work.aspx
        
        Arkansas:
        -         
        - https://www.arkansasedc.com/why-arkansas/business-climate/tax-structure/personal-income-tax
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_and_AR1000NR_Instructions.pdf
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_FullYearResidentIndividualIncomeTaxReturn.pdf
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_and_AR1000NR_Instructions.pdf
          https://www.arkansasedc.com/why-arkansas/business-climate/tax-structure/capital-gains-tax-reduction
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_and_AR1000NR_Instructions.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Wisconsin', 'state_occupation': 'Arkansas', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def wisconsin_and_california(self, state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Wisconsin:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        California:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Wisconsin:
        - https://www.revenue.wi.gov/TaxForms2023/2023-Form1-ES-Inst.pdf
          https://www.revenue.wi.gov/TaxForms2023/2023-Form1-ES-Inst.pdf
          https://www.revenue.wi.gov/TaxForms2023/2023-Form1-ES-Inst.pdf
          https://www.revenue.wi.gov/Pages/FAQS/pcs-work.aspx
          https://www.revenue.wi.gov/Dor%20Publications/pb103.pdf
          https://www.revenue.wi.gov/Pages/FAQS/pcs-work.aspx
        
        California:
        -         
        - https://www.ftb.ca.gov/forms/2022/2022-540-tax-rate-schedules.pdf
          https://www.ftb.ca.gov/file/personal/deductions/index.html
          https://www.ftb.ca.gov/about-ftb/newsroom/caleitc/eligibility-and-credit-information.html
          https://www.ftb.ca.gov/file/personal/income-types/capital-gains-and-losses.html#:~:text=California%20does%20not%20have%20a
         are%20taxed%20as%20ordinary%20income
          https://www.ftb.ca.gov/file/personal/
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Wisconsin', 'state_occupation': 'California', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def wisconsin_and_colorado(self, county_residence='Other', state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Wisconsin:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Colorado:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Wisconsin:
        - https://www.revenue.wi.gov/TaxForms2023/2023-Form1-ES-Inst.pdf
          https://www.revenue.wi.gov/TaxForms2023/2023-Form1-ES-Inst.pdf
          https://www.revenue.wi.gov/TaxForms2023/2023-Form1-ES-Inst.pdf
          https://www.revenue.wi.gov/Pages/FAQS/pcs-work.aspx
          https://www.revenue.wi.gov/Dor%20Publications/pb103.pdf
          https://www.revenue.wi.gov/Pages/FAQS/pcs-work.aspx
        
        Colorado:
        -         
        - https://leg.colorado.gov/agencies/legislative-council-staff/individual-income-tax%C2%A0
          https://leg.colorado.gov/agencies/legislative-council-staff/individual-income-tax%C2%A0
          https://tax.colorado.gov/sites/tax/files/documents/ITT_Capital_Gain_Subtraction_Feb_2022.pdf
          https://www.auroragov.org/business_services/taxes/occupational_privilege_tax
          https://www.denvergov.org/Government/Agencies-Departments-Offices/Agencies-Departments-Offices-Directory/Department-of-Finance/Our-Divisions/Treasury/Business-Tax-Information#:~:text=Occupational%20privilege%20tax
         -%28show%20below%29&text=Each%20taxable%20employee%20is%20liable
         month%20for%20each%20taxable%20employee
          https://www.glendale.co.us/355/Occupational-Privilege-Tax
          https://www.greenwoodvillage.com/1220/Occupational-Privilege-Tax-OPT
          https://ci.sheridan.co.us/288/Occupational-Privilege-Tax
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Wisconsin', 'state_occupation': 'Colorado', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def wisconsin_and_connecticut(self, state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Wisconsin:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Connecticut:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Wisconsin:
        - https://www.revenue.wi.gov/TaxForms2023/2023-Form1-ES-Inst.pdf
          https://www.revenue.wi.gov/TaxForms2023/2023-Form1-ES-Inst.pdf
          https://www.revenue.wi.gov/TaxForms2023/2023-Form1-ES-Inst.pdf
          https://www.revenue.wi.gov/Pages/FAQS/pcs-work.aspx
          https://www.revenue.wi.gov/Dor%20Publications/pb103.pdf
          https://www.revenue.wi.gov/Pages/FAQS/pcs-work.aspx
        
        Connecticut:
        -         
        - https://portal.ct.gov/-/media/DRS/Forms/2022/Income/2022-CT-1040-Instructions_1222.pdf
          https://www.cga.ct.gov/2022/rpt/pdf/2022-R-0108.pdf
          https://www.cga.ct.gov/current/pub/chap_229.htm#sec_12-703
          https://portal.ct.gov/-/media/DRS/Forms/2022/Income/2022-CT-1040-Instructions_1222.pdf
          "https://portal.ct.gov/DRS/Publications/TSSNs/TSSN-29#:~:text=An%20individuals%20net%20capital%20gains
         portion%20of%20Social%20Security%20benefits"
          "https://portal.ct.gov/DRS/Publications/TSSNs/TSSN-29#:~:text=An%20individuals%20net%20capital%20gains
         portion%20of%20Social%20Security%20benefits"
          https://portal.ct.gov/DRS/Individuals/Resident-Income-Tax/Tax-Information
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Wisconsin', 'state_occupation': 'Connecticut', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def wisconsin_and_delaware(self, state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, federal_agi=75000, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Wisconsin:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Delaware:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        federal_agi : int, optional
            Adjusted Gross Income (AGI) for federal taxes. Default is 0. Sources - https://www.irs.gov/e-file-providers/definition-of-adjusted-gross-income, https://www.irs.gov/newsroom/taxpayers-should-know-the-difference-between-standard-and-itemized-deductions
        income : int, optional
            Gross income. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Wisconsin:
        - https://www.revenue.wi.gov/TaxForms2023/2023-Form1-ES-Inst.pdf
          https://www.revenue.wi.gov/TaxForms2023/2023-Form1-ES-Inst.pdf
          https://www.revenue.wi.gov/TaxForms2023/2023-Form1-ES-Inst.pdf
          https://www.revenue.wi.gov/Pages/FAQS/pcs-work.aspx
          https://www.revenue.wi.gov/Dor%20Publications/pb103.pdf
          https://www.revenue.wi.gov/Pages/FAQS/pcs-work.aspx
        
        Delaware:
        -         
        - https://revenue.delaware.gov/software-developer/tax-rate-changes/
          https://revenuefiles.delaware.gov/2022/PIT-RES_TY22_2022-01_PaperInteractive.pdf
          https://revenuefiles.delaware.gov/2022/PIT-RES_TY22_2022-02_Instructions.pdf
          https://revenue.delaware.gov/frequently-asked-questions/personal-income-tax-faqs/
          https://www.newcastlede.gov/DocumentCenter/View/20147/City-of-Wilmington-Wage-Tax-Information-PDF
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        federal_agi = self.federal_agi if hasattr(self, 'federal_agi') else federal_agi
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Wisconsin', 'state_occupation': 'Delaware', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "federal_agi": federal_agi, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def wisconsin_and_florida(self, state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Wisconsin:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Florida:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Wisconsin:
        - https://www.revenue.wi.gov/TaxForms2023/2023-Form1-ES-Inst.pdf
          https://www.revenue.wi.gov/TaxForms2023/2023-Form1-ES-Inst.pdf
          https://www.revenue.wi.gov/TaxForms2023/2023-Form1-ES-Inst.pdf
          https://www.revenue.wi.gov/Pages/FAQS/pcs-work.aspx
          https://www.revenue.wi.gov/Dor%20Publications/pb103.pdf
          https://www.revenue.wi.gov/Pages/FAQS/pcs-work.aspx
        
        Florida:
        -         
        - https://floridarevenue.com/taxes/taxesfees/Pages/tax_interest_rates.aspx
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Wisconsin', 'state_occupation': 'Florida', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def wisconsin_and_georgia(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Wisconsin:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Georgia:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Wisconsin:
        - https://www.revenue.wi.gov/TaxForms2023/2023-Form1-ES-Inst.pdf
          https://www.revenue.wi.gov/TaxForms2023/2023-Form1-ES-Inst.pdf
          https://www.revenue.wi.gov/TaxForms2023/2023-Form1-ES-Inst.pdf
          https://www.revenue.wi.gov/Pages/FAQS/pcs-work.aspx
          https://www.revenue.wi.gov/Dor%20Publications/pb103.pdf
          https://www.revenue.wi.gov/Pages/FAQS/pcs-work.aspx
        
        Georgia:
        -         
        - https://dor.georgia.gov/tax-tables-georgia-tax-rate-schedule
          https://apps.dor.ga.gov/FillableForms/PDFViewer/Index?form=2022GA500
          https://apps.dor.ga.gov/FillableForms/PDFViewer/Index?form=2022GA500
          https://apps.dor.ga.gov/FillableForms/PDFViewer/Index?form=2022GA500
          https://dor.georgia.gov/years-individual-income-tax-forms
          https://dor.georgia.gov/years-individual-income-tax-forms
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Wisconsin', 'state_occupation': 'Georgia', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def wisconsin_and_hawaii(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Wisconsin:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Hawaii:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Wisconsin:
        - https://www.revenue.wi.gov/TaxForms2023/2023-Form1-ES-Inst.pdf
          https://www.revenue.wi.gov/TaxForms2023/2023-Form1-ES-Inst.pdf
          https://www.revenue.wi.gov/TaxForms2023/2023-Form1-ES-Inst.pdf
          https://www.revenue.wi.gov/Pages/FAQS/pcs-work.aspx
          https://www.revenue.wi.gov/Dor%20Publications/pb103.pdf
          https://www.revenue.wi.gov/Pages/FAQS/pcs-work.aspx
        
        Hawaii:
        -         
        - https://tax.hawaii.gov/forms/d_18table-on/d_18table-on_p13/
          https://files.hawaii.gov/tax/forms/2018/18table-on.pdf
          https://files.hawaii.gov/tax/news/pubs/22outline.pdf
          https://files.hawaii.gov/tax/news/pubs/FastTaxRefGuide_2017.pdf
          https://www.tfhawaii.org/wordpress/state-tax-resources/mini-tax-guides/individual-income-tax-chapter-235/#:~:text=Long%20term%20capital%20gains%20are%20taxed%20at%20a%20maximum%20of%207.25%25
          https://www.grassrootinstitute.org/2023/02/hb337-increasing-tax-on-capital-gains-a-bad-risk/
          https://files.hawaii.gov/tax/news/pubs/22outline.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Wisconsin', 'state_occupation': 'Hawaii', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def wisconsin_and_idaho(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Wisconsin:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Idaho:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Wisconsin:
        - https://www.revenue.wi.gov/TaxForms2023/2023-Form1-ES-Inst.pdf
          https://www.revenue.wi.gov/TaxForms2023/2023-Form1-ES-Inst.pdf
          https://www.revenue.wi.gov/TaxForms2023/2023-Form1-ES-Inst.pdf
          https://www.revenue.wi.gov/Pages/FAQS/pcs-work.aspx
          https://www.revenue.wi.gov/Dor%20Publications/pb103.pdf
          https://www.revenue.wi.gov/Pages/FAQS/pcs-work.aspx
        
        Idaho:
        -         
        - https://tax.idaho.gov/taxes/income-tax/individual-income/individual-income-tax-rate-schedule/
          https://tax.idaho.gov/wp-content/uploads/forms/EIN00046/EIN00046_03-01-2023.pdf
          https://tax.idaho.gov/taxes/income-tax/individual-income/idaho-source-income/
          https://tax.idaho.gov/taxes/income-tax/individual-income/idaho-source-income/
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Wisconsin', 'state_occupation': 'Idaho', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def wisconsin_and_illinois(self, state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Wisconsin:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Illinois:
        


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Wisconsin:
        - https://www.revenue.wi.gov/TaxForms2023/2023-Form1-ES-Inst.pdf
          https://www.revenue.wi.gov/TaxForms2023/2023-Form1-ES-Inst.pdf
          https://www.revenue.wi.gov/TaxForms2023/2023-Form1-ES-Inst.pdf
          https://www.revenue.wi.gov/Pages/FAQS/pcs-work.aspx
          https://www.revenue.wi.gov/Dor%20Publications/pb103.pdf
          https://www.revenue.wi.gov/Pages/FAQS/pcs-work.aspx
        
        Illinois:
        -         
        - https://tax.illinois.gov/research/taxrates/income.html#IndividualIncome
          https://tax.illinois.gov/content/dam/soi/en/web/tax/forms/incometax/documents/currentyear/individual/il-1040-instr.pdf
          https://tax.illinois.gov/questionsandanswers/12.html
          https://tax.illinois.gov/content/dam/soi/en/web/tax/forms/incometax/documents/currentyear/individual/il-1040-instr.pdf
          https://tax.illinois.gov/research/taxrates/personalproperty.html
          https://tax.illinois.gov/localgovernments/income.html
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Wisconsin', 'state_occupation': 'Illinois', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def wisconsin_and_indiana(self, state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Wisconsin:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Indiana:
        


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Wisconsin:
        - https://www.revenue.wi.gov/TaxForms2023/2023-Form1-ES-Inst.pdf
          https://www.revenue.wi.gov/TaxForms2023/2023-Form1-ES-Inst.pdf
          https://www.revenue.wi.gov/TaxForms2023/2023-Form1-ES-Inst.pdf
          https://www.revenue.wi.gov/Pages/FAQS/pcs-work.aspx
          https://www.revenue.wi.gov/Dor%20Publications/pb103.pdf
          https://www.revenue.wi.gov/Pages/FAQS/pcs-work.aspx
        
        Indiana:
        -         
        - https://www.in.gov/dor/tax-forms/2022-individual-income-tax-forms/
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-32.htm
          https://www.in.gov/dor/files/ib28.pdf
          https://www.in.gov/dor/tax-forms/2022-individual-income-tax-forms/
          https://www.in.gov/dor/files/2022-county-tax-rates-and-codes.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Wisconsin', 'state_occupation': 'Indiana', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def wisconsin_and_iowa(self, state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Wisconsin:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Iowa:
        


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Wisconsin:
        - https://www.revenue.wi.gov/TaxForms2023/2023-Form1-ES-Inst.pdf
          https://www.revenue.wi.gov/TaxForms2023/2023-Form1-ES-Inst.pdf
          https://www.revenue.wi.gov/TaxForms2023/2023-Form1-ES-Inst.pdf
          https://www.revenue.wi.gov/Pages/FAQS/pcs-work.aspx
          https://www.revenue.wi.gov/Dor%20Publications/pb103.pdf
          https://www.revenue.wi.gov/Pages/FAQS/pcs-work.aspx
        
        Iowa:
        -         
        - https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
          https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
          https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
          https://tax.iowa.gov/iowa-illinois-reciprocal-agreement
          https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Wisconsin', 'state_occupation': 'Iowa', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def wisconsin_and_kansas(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Wisconsin:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Kansas:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Wisconsin:
        - https://www.revenue.wi.gov/TaxForms2023/2023-Form1-ES-Inst.pdf
          https://www.revenue.wi.gov/TaxForms2023/2023-Form1-ES-Inst.pdf
          https://www.revenue.wi.gov/TaxForms2023/2023-Form1-ES-Inst.pdf
          https://www.revenue.wi.gov/Pages/FAQS/pcs-work.aspx
          https://www.revenue.wi.gov/Dor%20Publications/pb103.pdf
          https://www.revenue.wi.gov/Pages/FAQS/pcs-work.aspx
        
        Kansas:
        -         
        - https://www.ksrevenue.gov/taxrates.html
          https://www.ksrevenue.gov/incomebook22.html#0
          https://www.ksrevenue.gov/incomebook22.html#0
          https://www.ksrevenue.gov/incomebook22.html#0
          https://www.ksrevenue.gov/incomebook22.html
          https://www.ksrevenue.gov/incomebook22.html
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Wisconsin', 'state_occupation': 'Kansas', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def wisconsin_and_kentucky(self, state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Wisconsin:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Kentucky:
        


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Wisconsin:
        - https://www.revenue.wi.gov/TaxForms2023/2023-Form1-ES-Inst.pdf
          https://www.revenue.wi.gov/TaxForms2023/2023-Form1-ES-Inst.pdf
          https://www.revenue.wi.gov/TaxForms2023/2023-Form1-ES-Inst.pdf
          https://www.revenue.wi.gov/Pages/FAQS/pcs-work.aspx
          https://www.revenue.wi.gov/Dor%20Publications/pb103.pdf
          https://www.revenue.wi.gov/Pages/FAQS/pcs-work.aspx
        
        Kentucky:
        -         
        - https://revenue.ky.gov/News/Pages/Dor-Announces-Updates-to-Individual-Income-Tax-for-2023-Tax-Year.aspx
          https://revenue.ky.gov/Individual/Individual-Income-Tax/Pages/default.aspx
          https://revenue.ky.gov/News/Pages/Dor-Announces-Updates-to-Individual-Income-Tax-for-2023-Tax-Year.aspx
          https://revenue.ky.gov/Individual/Individual-Income-Tax/Pages/default.aspx
          https://revenue.ky.gov/Dor%20Training%20Materials/103%20KAR%2017.140.%20Individual%20income%20tax%20-%20reciprocity%20-%20nonresidents.pdf
          https://revenue.ky.gov/Individual/Individual-Income-Tax/Pages/default.aspx
          https://nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2021/Taxes-21-32.htm?taxmap=true
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Wisconsin', 'state_occupation': 'Kentucky', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def wisconsin_and_louisiana(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Wisconsin:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Louisiana:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Wisconsin:
        - https://www.revenue.wi.gov/TaxForms2023/2023-Form1-ES-Inst.pdf
          https://www.revenue.wi.gov/TaxForms2023/2023-Form1-ES-Inst.pdf
          https://www.revenue.wi.gov/TaxForms2023/2023-Form1-ES-Inst.pdf
          https://www.revenue.wi.gov/Pages/FAQS/pcs-work.aspx
          https://www.revenue.wi.gov/Dor%20Publications/pb103.pdf
          https://www.revenue.wi.gov/Pages/FAQS/pcs-work.aspx
        
        Louisiana:
        -         
        - https://revenue.louisiana.gov/individualincometax
          https://revenue.louisiana.gov/individualincometax
          https://www.revenue.louisiana.gov/taxforms/6935(11_02)F.pdf
          https://revenue.louisiana.gov/individualincometax
          https://revenue.louisiana.gov/individualincometax
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Wisconsin', 'state_occupation': 'Louisiana', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def wisconsin_and_maine(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Wisconsin:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Maine:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Wisconsin:
        - https://www.revenue.wi.gov/TaxForms2023/2023-Form1-ES-Inst.pdf
          https://www.revenue.wi.gov/TaxForms2023/2023-Form1-ES-Inst.pdf
          https://www.revenue.wi.gov/TaxForms2023/2023-Form1-ES-Inst.pdf
          https://www.revenue.wi.gov/Pages/FAQS/pcs-work.aspx
          https://www.revenue.wi.gov/Dor%20Publications/pb103.pdf
          https://www.revenue.wi.gov/Pages/FAQS/pcs-work.aspx
        
        Maine:
        -         
        - https://www.maine.gov/revenue/tax-return-forms/individual-income-tax-2022
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/ind_tax_rate_sched_2022.pdf
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/22_1040me_book_gen_instruc_revisedFeb23.pdf
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/22_1040me_book_gen_instruc_revisedFeb23.pdf
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/22_1040me_book_gen_instruc_revisedFeb23.pdf
          https://www.maine.gov/revenue/tax-return-forms/individual-income-tax-2022
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Wisconsin', 'state_occupation': 'Maine', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def wisconsin_and_maryland(self, county_residence='Other', state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Wisconsin:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Maryland:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Wisconsin:
        - https://www.revenue.wi.gov/TaxForms2023/2023-Form1-ES-Inst.pdf
          https://www.revenue.wi.gov/TaxForms2023/2023-Form1-ES-Inst.pdf
          https://www.revenue.wi.gov/TaxForms2023/2023-Form1-ES-Inst.pdf
          https://www.revenue.wi.gov/Pages/FAQS/pcs-work.aspx
          https://www.revenue.wi.gov/Dor%20Publications/pb103.pdf
          https://www.revenue.wi.gov/Pages/FAQS/pcs-work.aspx
        
        Maryland:
        -         
        - https://www.marylandtaxes.gov/individual/income/tax-info/tax-rates.php
          https://www.marylandtaxes.gov/forms/22_forms/Resident_Booklet.pdf
          https://www.marylandtaxes.gov/forms/Tax_Publications/Administrative_Releases/Income_and_Estate_Tax_Releases/ar_it37.pdf
          https://www.marylandtaxes.gov/individual/index.php
          https://www.marylandtaxes.gov/individual/income/tax-info/tax-rates.php
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Wisconsin', 'state_occupation': 'Maryland', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def wisconsin_and_massachusetts(self, state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Wisconsin:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Massachusetts:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Wisconsin:
        - https://www.revenue.wi.gov/TaxForms2023/2023-Form1-ES-Inst.pdf
          https://www.revenue.wi.gov/TaxForms2023/2023-Form1-ES-Inst.pdf
          https://www.revenue.wi.gov/TaxForms2023/2023-Form1-ES-Inst.pdf
          https://www.revenue.wi.gov/Pages/FAQS/pcs-work.aspx
          https://www.revenue.wi.gov/Dor%20Publications/pb103.pdf
          https://www.revenue.wi.gov/Pages/FAQS/pcs-work.aspx
        
        Massachusetts:
        -         
        - https://www.mass.gov/service-details/massachusetts-tax-rates
          https://www.mass.gov/service-details/view-massachusetts-personal-income-tax-exemptions
          https://www.mass.gov/service-details/view-massachusetts-personal-income-tax-exemptions
          https://www.mass.gov/service-details/view-child-and-dependent-related-credits
          https://www.mass.gov/service-details/massachusetts-tax-rates
          https://www.mass.gov/personal-income-tax
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Wisconsin', 'state_occupation': 'Massachusetts', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def wisconsin_and_michigan(self, state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Wisconsin:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Michigan:
        


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Wisconsin:
        - https://www.revenue.wi.gov/TaxForms2023/2023-Form1-ES-Inst.pdf
          https://www.revenue.wi.gov/TaxForms2023/2023-Form1-ES-Inst.pdf
          https://www.revenue.wi.gov/TaxForms2023/2023-Form1-ES-Inst.pdf
          https://www.revenue.wi.gov/Pages/FAQS/pcs-work.aspx
          https://www.revenue.wi.gov/Dor%20Publications/pb103.pdf
          https://www.revenue.wi.gov/Pages/FAQS/pcs-work.aspx
        
        Michigan:
        -         
        - https://www.michigan.gov/taxes/iit/new-developments-for-tax-year-2022
          "https://www.michigan.gov/taxes/-/media/Project/Websites/taxes/Forms/2023/2023-SUW/446_Withholding-Guide_2023.pdf?rev=177ca3d8cb034392bf84e317ed78c320&hash=19C5F42F43A89A91863D229349F48883#:~:text=The%20exemption%20amount%20is%20%245%2C400
         employees%20Michigan%20income%20tax%20return"
          "https://www.michigan.gov/taxes/-/media/Project/Websites/taxes/Forms/2023/2023-SUW/446_Withholding-Guide_2023.pdf?rev=177ca3d8cb034392bf84e317ed78c320&hash=19C5F42F43A89A91863D229349F48883#:~:text=The%20exemption%20amount%20is%20%245%2C400
         employees%20Michigan%20income%20tax%20return"
          https://www.michigan.gov/taxes/questions/iit/accordion/residency/are-my-wages-earned-in-another-state-taxable-in-michigan-if-i-am-a-michigan-resident-1
          https://www.michigan.gov/taxes/-/media/Project/Websites/taxes/Forms/2022/2022-IIT-Forms/BOOK_MI-1040.pdf?rev=4524f1cbbc0243b194ab05a6680be75b&hash=40DFE71E651F7C1D65C31F6CD03D944F
          https://www.michigan.gov/taxes/questions/iit/accordion/general/what-cities-impose-an-income-tax
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Wisconsin', 'state_occupation': 'Michigan', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def wisconsin_and_minnesota(self, state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Wisconsin:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Minnesota:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Wisconsin:
        - https://www.revenue.wi.gov/TaxForms2023/2023-Form1-ES-Inst.pdf
          https://www.revenue.wi.gov/TaxForms2023/2023-Form1-ES-Inst.pdf
          https://www.revenue.wi.gov/TaxForms2023/2023-Form1-ES-Inst.pdf
          https://www.revenue.wi.gov/Pages/FAQS/pcs-work.aspx
          https://www.revenue.wi.gov/Dor%20Publications/pb103.pdf
          https://www.revenue.wi.gov/Pages/FAQS/pcs-work.aspx
        
        Minnesota:
        -         
        - https://www.revenue.state.mn.us/minnesota-income-tax-rates-and-brackets
          https://www.revenue.state.mn.us/child-and-dependent-care-credit
          https://www.revenue.state.mn.us/dependent-exemptions
          https://www.revenue.state.mn.us/minnesota-standard-deduction
          https://mn.gov/mmb/assets/mwr_form_tcm1059-128581.pdf
          https://www.revenue.state.mn.us/individual-income-tax
          https://www.revenue.state.mn.us/individual-income-tax
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Wisconsin', 'state_occupation': 'Minnesota', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def wisconsin_and_mississippi(self, state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Wisconsin:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Mississippi:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Wisconsin:
        - https://www.revenue.wi.gov/TaxForms2023/2023-Form1-ES-Inst.pdf
          https://www.revenue.wi.gov/TaxForms2023/2023-Form1-ES-Inst.pdf
          https://www.revenue.wi.gov/TaxForms2023/2023-Form1-ES-Inst.pdf
          https://www.revenue.wi.gov/Pages/FAQS/pcs-work.aspx
          https://www.revenue.wi.gov/Dor%20Publications/pb103.pdf
          https://www.revenue.wi.gov/Pages/FAQS/pcs-work.aspx
        
        Mississippi:
        -         
        - https://www.dor.ms.gov/individual/tax-rates
          https://www.dor.ms.gov/individual/tax-rates
          https://www.dor.ms.gov/individual/tax-rates
          https://www.dor.ms.gov/individual
          https://www.dor.ms.gov/individual
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Wisconsin', 'state_occupation': 'Mississippi', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def wisconsin_and_missouri(self, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Wisconsin:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Missouri:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Wisconsin:
        - https://www.revenue.wi.gov/TaxForms2023/2023-Form1-ES-Inst.pdf
          https://www.revenue.wi.gov/TaxForms2023/2023-Form1-ES-Inst.pdf
          https://www.revenue.wi.gov/TaxForms2023/2023-Form1-ES-Inst.pdf
          https://www.revenue.wi.gov/Pages/FAQS/pcs-work.aspx
          https://www.revenue.wi.gov/Dor%20Publications/pb103.pdf
          https://www.revenue.wi.gov/Pages/FAQS/pcs-work.aspx
        
        Missouri:
        -         
        - https://dor.mo.gov/taxation/individual/tax-types/income/year-changes/
          https://dor.mo.gov/taxation/individual/tax-types/income/year-changes/
          https://dor.mo.gov/taxation/individual/tax-types/income/year-changes/
          https://www.kcmo.gov/city-hall/departments/finance/earnings-tax
          https://www.stlouis-mo.gov/government/departments/collector/earnings-tax/individual-earnings-tax-info.cfm
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Wisconsin', 'state_occupation': 'Missouri', "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def wisconsin_and_montana(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Wisconsin:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Montana:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Wisconsin:
        - https://www.revenue.wi.gov/TaxForms2023/2023-Form1-ES-Inst.pdf
          https://www.revenue.wi.gov/TaxForms2023/2023-Form1-ES-Inst.pdf
          https://www.revenue.wi.gov/TaxForms2023/2023-Form1-ES-Inst.pdf
          https://www.revenue.wi.gov/Pages/FAQS/pcs-work.aspx
          https://www.revenue.wi.gov/Dor%20Publications/pb103.pdf
          https://www.revenue.wi.gov/Pages/FAQS/pcs-work.aspx
        
        Montana:
        -         
        - https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/2022/12/montana-employees-withholding-allowance-and-exemption-certificate-form-mw-4.pdf
          https://mtrevenue.gov/2021/11/29/simplification-of-montana-income-taxation/?utm_source=rss&utm_medium=rss&utm_campaign=simplification-of-montana-income-taxation
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Wisconsin', 'state_occupation': 'Montana', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def wisconsin_and_nebraska(self, state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Wisconsin:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Nebraska:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Wisconsin:
        - https://www.revenue.wi.gov/TaxForms2023/2023-Form1-ES-Inst.pdf
          https://www.revenue.wi.gov/TaxForms2023/2023-Form1-ES-Inst.pdf
          https://www.revenue.wi.gov/TaxForms2023/2023-Form1-ES-Inst.pdf
          https://www.revenue.wi.gov/Pages/FAQS/pcs-work.aspx
          https://www.revenue.wi.gov/Dor%20Publications/pb103.pdf
          https://www.revenue.wi.gov/Pages/FAQS/pcs-work.aspx
        
        Nebraska:
        -         
        - https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2023_Circular_EN_8-429r2023_Whole_Book_11.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Wisconsin', 'state_occupation': 'Nebraska', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def wisconsin_and_nevada(self, state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Wisconsin:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Nevada:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Wisconsin:
        - https://www.revenue.wi.gov/TaxForms2023/2023-Form1-ES-Inst.pdf
          https://www.revenue.wi.gov/TaxForms2023/2023-Form1-ES-Inst.pdf
          https://www.revenue.wi.gov/TaxForms2023/2023-Form1-ES-Inst.pdf
          https://www.revenue.wi.gov/Pages/FAQS/pcs-work.aspx
          https://www.revenue.wi.gov/Dor%20Publications/pb103.pdf
          https://www.revenue.wi.gov/Pages/FAQS/pcs-work.aspx
        
        Nevada:
        -         
        - https://www.nvsos.gov/sos/businesses/the-nevada-advantage
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Wisconsin', 'state_occupation': 'Nevada', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def wisconsin_and_new_hampshire(self, state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Wisconsin:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        New Hampshire:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Wisconsin:
        - https://www.revenue.wi.gov/TaxForms2023/2023-Form1-ES-Inst.pdf
          https://www.revenue.wi.gov/TaxForms2023/2023-Form1-ES-Inst.pdf
          https://www.revenue.wi.gov/TaxForms2023/2023-Form1-ES-Inst.pdf
          https://www.revenue.wi.gov/Pages/FAQS/pcs-work.aspx
          https://www.revenue.wi.gov/Dor%20Publications/pb103.pdf
          https://www.revenue.wi.gov/Pages/FAQS/pcs-work.aspx
        
        New Hampshire:
        -         
        - https://www.revenue.nh.gov/assistance/tax-overview.htm
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Wisconsin', 'state_occupation': 'NewHampshire', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def wisconsin_and_new_jersey(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Wisconsin:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        New Jersey:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Wisconsin:
        - https://www.revenue.wi.gov/TaxForms2023/2023-Form1-ES-Inst.pdf
          https://www.revenue.wi.gov/TaxForms2023/2023-Form1-ES-Inst.pdf
          https://www.revenue.wi.gov/TaxForms2023/2023-Form1-ES-Inst.pdf
          https://www.revenue.wi.gov/Pages/FAQS/pcs-work.aspx
          https://www.revenue.wi.gov/Dor%20Publications/pb103.pdf
          https://www.revenue.wi.gov/Pages/FAQS/pcs-work.aspx
        
        New Jersey:
        -         
        - https://www.state.nj.us/treasury/taxation/taxtables.shtml
          https://www.state.nj.us/treasury/taxation/njit13.shtml
          https://www.nj.gov/treasury/taxation/njit25.shtml
          https://www.state.nj.us/treasury/taxation/individuals/freqqiti.shtml
          https://www.state.nj.us/treasury/taxation/localtax.shtml
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Wisconsin', 'state_occupation': 'NewJersey', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def wisconsin_and_new_mexico(self, state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Wisconsin:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        New Mexico:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Wisconsin:
        - https://www.revenue.wi.gov/TaxForms2023/2023-Form1-ES-Inst.pdf
          https://www.revenue.wi.gov/TaxForms2023/2023-Form1-ES-Inst.pdf
          https://www.revenue.wi.gov/TaxForms2023/2023-Form1-ES-Inst.pdf
          https://www.revenue.wi.gov/Pages/FAQS/pcs-work.aspx
          https://www.revenue.wi.gov/Dor%20Publications/pb103.pdf
          https://www.revenue.wi.gov/Pages/FAQS/pcs-work.aspx
        
        New Mexico:
        -         
        - https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/fdf3c548-8aba-4b9c-9eb4-bb564c716015/FYI-104.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/0bf943fd-652e-4400-bb1b-cd8397eb5a95/2022pit-adj.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/2f1a6781-9534-4436-b427-1557f9592099/2022pit-adj-ins.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/0bf943fd-652e-4400-bb1b-cd8397eb5a95/2022pit-adj.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/1afc56af-ea90-4d48-82e5-1f9aeb43255a/PITbook2022.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/fdf3c548-8aba-4b9c-9eb4-bb564c716015/FYI-104.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Wisconsin', 'state_occupation': 'NewMexico', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def wisconsin_and_new_york(self, county_residence='Other', state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Wisconsin:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        New York:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Wisconsin:
        - https://www.revenue.wi.gov/TaxForms2023/2023-Form1-ES-Inst.pdf
          https://www.revenue.wi.gov/TaxForms2023/2023-Form1-ES-Inst.pdf
          https://www.revenue.wi.gov/TaxForms2023/2023-Form1-ES-Inst.pdf
          https://www.revenue.wi.gov/Pages/FAQS/pcs-work.aspx
          https://www.revenue.wi.gov/Dor%20Publications/pb103.pdf
          https://www.revenue.wi.gov/Pages/FAQS/pcs-work.aspx
        
        New York:
        -         
        - https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nys-tax-rate-schedule
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nys-tax-rate-schedule
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nys-tax-rate-schedule
          https://www.tax.ny.gov/pit/
          https://www.tax.ny.gov/pit/file/nonresident-faqs.htm
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#yonkers-tax-rate-schedule
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nyc-tax-rate-schedule
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Wisconsin', 'state_occupation': 'NewYork', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def wisconsin_and_north_carolina(self, state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Wisconsin:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        North Carolina:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Wisconsin:
        - https://www.revenue.wi.gov/TaxForms2023/2023-Form1-ES-Inst.pdf
          https://www.revenue.wi.gov/TaxForms2023/2023-Form1-ES-Inst.pdf
          https://www.revenue.wi.gov/TaxForms2023/2023-Form1-ES-Inst.pdf
          https://www.revenue.wi.gov/Pages/FAQS/pcs-work.aspx
          https://www.revenue.wi.gov/Dor%20Publications/pb103.pdf
          https://www.revenue.wi.gov/Pages/FAQS/pcs-work.aspx
        
        North Carolina:
        -         
        - https://www.ncdor.gov/taxes-forms/tax-rate-schedules
          https://www.ncdor.gov/taxes-forms/individual-income-tax/north-carolina-standard-deduction-or-north-carolina-itemized-deductions
          https://www.ncdor.gov/taxes-forms/individual-income-tax
          https://www.ncdor.gov/taxes-forms/individual-income-tax
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Wisconsin', 'state_occupation': 'NorthCarolina', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def wisconsin_and_north_dakota(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Wisconsin:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        North Dakota:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Wisconsin:
        - https://www.revenue.wi.gov/TaxForms2023/2023-Form1-ES-Inst.pdf
          https://www.revenue.wi.gov/TaxForms2023/2023-Form1-ES-Inst.pdf
          https://www.revenue.wi.gov/TaxForms2023/2023-Form1-ES-Inst.pdf
          https://www.revenue.wi.gov/Pages/FAQS/pcs-work.aspx
          https://www.revenue.wi.gov/Dor%20Publications/pb103.pdf
          https://www.revenue.wi.gov/Pages/FAQS/pcs-work.aspx
        
        North Dakota:
        -         
        - https://www.tax.nd.gov/forms
          https://www.tax.nd.gov/sites/www/files/documents/forms/individual/2022-iit/2022-individual-income-tax-booklet.pdf
          https://www.tax.nd.gov/sites/www/files/documents/forms/2022-individual-income-tax-booklet.pdf
          https://www.tax.nd.gov/tax-exemptions-credits/income-tax-exemptions-credits
          https://www.tax.nd.gov/sites/www/files/documents/forms/business/ndwrfillable.pdf
          https://www.tax.nd.gov/news/tax-legislative-changes/significant-changes-law/individual-income-tax-history
          https://www.tax.nd.gov/forms
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Wisconsin', 'state_occupation': 'NorthDakota', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def wisconsin_and_ohio(self, county_residence='Other', state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Wisconsin:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Ohio:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Wisconsin:
        - https://www.revenue.wi.gov/TaxForms2023/2023-Form1-ES-Inst.pdf
          https://www.revenue.wi.gov/TaxForms2023/2023-Form1-ES-Inst.pdf
          https://www.revenue.wi.gov/TaxForms2023/2023-Form1-ES-Inst.pdf
          https://www.revenue.wi.gov/Pages/FAQS/pcs-work.aspx
          https://www.revenue.wi.gov/Dor%20Publications/pb103.pdf
          https://www.revenue.wi.gov/Pages/FAQS/pcs-work.aspx
        
        Ohio:
        -         
        - https://tax.ohio.gov/individual/resources/annual-tax-rates
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://tax.ohio.gov/static/forms/employer_withholding/generic/wth_it4nr.pdf
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://www.ritaohio.com/TaxRatesTable
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Wisconsin', 'state_occupation': 'Ohio', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def wisconsin_and_oklahoma(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Wisconsin:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Oklahoma:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Wisconsin:
        - https://www.revenue.wi.gov/TaxForms2023/2023-Form1-ES-Inst.pdf
          https://www.revenue.wi.gov/TaxForms2023/2023-Form1-ES-Inst.pdf
          https://www.revenue.wi.gov/TaxForms2023/2023-Form1-ES-Inst.pdf
          https://www.revenue.wi.gov/Pages/FAQS/pcs-work.aspx
          https://www.revenue.wi.gov/Dor%20Publications/pb103.pdf
          https://www.revenue.wi.gov/Pages/FAQS/pcs-work.aspx
        
        Oklahoma:
        -         
        - https://oklahoma.gov/content/dam/ok/en/tax/documents/resources/publications/businesses/withholding-tables/WHTables-2022.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Wisconsin', 'state_occupation': 'Oklahoma', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def wisconsin_and_oregon(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Wisconsin:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Oregon:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Wisconsin:
        - https://www.revenue.wi.gov/TaxForms2023/2023-Form1-ES-Inst.pdf
          https://www.revenue.wi.gov/TaxForms2023/2023-Form1-ES-Inst.pdf
          https://www.revenue.wi.gov/TaxForms2023/2023-Form1-ES-Inst.pdf
          https://www.revenue.wi.gov/Pages/FAQS/pcs-work.aspx
          https://www.revenue.wi.gov/Dor%20Publications/pb103.pdf
          https://www.revenue.wi.gov/Pages/FAQS/pcs-work.aspx
        
        Oregon:
        -         
        - https://www.oregon.gov/dor/programs/individuals/Pages/PIT.aspx
          https://www.oregon.gov/dor/programs/individuals/Documents/Full%20year%20resident
         %20Form%20or-40%20filers.pdf
          https://www.oregon.gov/dor/programs/individuals/Documents/Part-year%20and%20nonresident
         %20Form%20or-40-P%20and%20or-40-N%20filers.pdf.pdf
          https://www.oregon.gov/dor/forms/FormsPubs/form-or-40_101-040_2022.pdf
          https://www.oregon.gov/dor/forms/FormsPubs/form-or-40_101-040_2022.pdf
          https://www.oregon.gov/dor/forms/FormsPubs/form-or-40_101-040_2022.pdf
          https://www.portland.gov/revenue/personal-tax
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Wisconsin', 'state_occupation': 'Oregon', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def wisconsin_and_pennsylvania(self, county_residence='Other', state_agi=75000, county_occupation='Other', capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Wisconsin:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Pennsylvania:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        county_residence : str, optional
            County of residence. Default is Other.
        county_occupation : str, optional
            County of occupation. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Wisconsin:
        - https://www.revenue.wi.gov/TaxForms2023/2023-Form1-ES-Inst.pdf
          https://www.revenue.wi.gov/TaxForms2023/2023-Form1-ES-Inst.pdf
          https://www.revenue.wi.gov/TaxForms2023/2023-Form1-ES-Inst.pdf
          https://www.revenue.wi.gov/Pages/FAQS/pcs-work.aspx
          https://www.revenue.wi.gov/Dor%20Publications/pb103.pdf
          https://www.revenue.wi.gov/Pages/FAQS/pcs-work.aspx
        
        Pennsylvania:
        -         
        - https://www.revenue.pa.gov/Tax%20Rates/Pages/default.aspx
          https://www.revenue.pa.gov/FormsandPublications/FormsforIndividuals/PIT/Documents/2022/2022_pa-40in.pdf
          https://www.revenue.pa.gov/FormsandPublications/FormsforBusinesses/EmployerWithholding/Documents/rev-419.pdf
          https://www.revenue.pa.gov/FormsandPublications/FormsforIndividuals/PIT/Documents/2022/2022_pa-40in.pdf
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-25.htm
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        county_occupation = self.county_occupation if hasattr(self, 'county_occupation') else county_occupation
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Wisconsin', 'state_occupation': 'Pennsylvania', "county_residence": county_residence, "state_agi": state_agi, "county_occupation": county_occupation, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def wisconsin_and_rhode_island(self, state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Wisconsin:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Rhode Island:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Wisconsin:
        - https://www.revenue.wi.gov/TaxForms2023/2023-Form1-ES-Inst.pdf
          https://www.revenue.wi.gov/TaxForms2023/2023-Form1-ES-Inst.pdf
          https://www.revenue.wi.gov/TaxForms2023/2023-Form1-ES-Inst.pdf
          https://www.revenue.wi.gov/Pages/FAQS/pcs-work.aspx
          https://www.revenue.wi.gov/Dor%20Publications/pb103.pdf
          https://www.revenue.wi.gov/Pages/FAQS/pcs-work.aspx
        
        Rhode Island:
        -         
        - https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2022_1040WE_w_0.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Wisconsin', 'state_occupation': 'RhodeIsland', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def wisconsin_and_south_carolina(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Wisconsin:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        South Carolina:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Wisconsin:
        - https://www.revenue.wi.gov/TaxForms2023/2023-Form1-ES-Inst.pdf
          https://www.revenue.wi.gov/TaxForms2023/2023-Form1-ES-Inst.pdf
          https://www.revenue.wi.gov/TaxForms2023/2023-Form1-ES-Inst.pdf
          https://www.revenue.wi.gov/Pages/FAQS/pcs-work.aspx
          https://www.revenue.wi.gov/Dor%20Publications/pb103.pdf
          https://www.revenue.wi.gov/Pages/FAQS/pcs-work.aspx
        
        South Carolina:
        -         
        - https://dor.sc.gov/resources-site/lawandpolicy/Advisory%20Opinions/IL22-15.pdf
          https://dor.sc.gov/forms-site/Forms/IITPacket_2022.pdf
          https://www.scstatehouse.gov/code/t12c006.php
          https://dor.sc.gov/resources-site/lawandpolicy/Advisory%20Opinions/IL22-15.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Wisconsin', 'state_occupation': 'SouthCarolina', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def wisconsin_and_south_dakota(self, state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Wisconsin:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        South Dakota:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Wisconsin:
        - https://www.revenue.wi.gov/TaxForms2023/2023-Form1-ES-Inst.pdf
          https://www.revenue.wi.gov/TaxForms2023/2023-Form1-ES-Inst.pdf
          https://www.revenue.wi.gov/TaxForms2023/2023-Form1-ES-Inst.pdf
          https://www.revenue.wi.gov/Pages/FAQS/pcs-work.aspx
          https://www.revenue.wi.gov/Dor%20Publications/pb103.pdf
          https://www.revenue.wi.gov/Pages/FAQS/pcs-work.aspx
        
        South Dakota:
        -         
        - https://dor.sd.gov/individuals/taxes/
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Wisconsin', 'state_occupation': 'SouthDakota', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def wisconsin_and_tennessee(self, state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Wisconsin:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Tennessee:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Wisconsin:
        - https://www.revenue.wi.gov/TaxForms2023/2023-Form1-ES-Inst.pdf
          https://www.revenue.wi.gov/TaxForms2023/2023-Form1-ES-Inst.pdf
          https://www.revenue.wi.gov/TaxForms2023/2023-Form1-ES-Inst.pdf
          https://www.revenue.wi.gov/Pages/FAQS/pcs-work.aspx
          https://www.revenue.wi.gov/Dor%20Publications/pb103.pdf
          https://www.revenue.wi.gov/Pages/FAQS/pcs-work.aspx
        
        Tennessee:
        -         
        - https://www.tn.gov/revenue/taxes.html
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Wisconsin', 'state_occupation': 'Tennessee', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def wisconsin_and_texas(self, state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Wisconsin:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Texas:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Wisconsin:
        - https://www.revenue.wi.gov/TaxForms2023/2023-Form1-ES-Inst.pdf
          https://www.revenue.wi.gov/TaxForms2023/2023-Form1-ES-Inst.pdf
          https://www.revenue.wi.gov/TaxForms2023/2023-Form1-ES-Inst.pdf
          https://www.revenue.wi.gov/Pages/FAQS/pcs-work.aspx
          https://www.revenue.wi.gov/Dor%20Publications/pb103.pdf
          https://www.revenue.wi.gov/Pages/FAQS/pcs-work.aspx
        
        Texas:
        -         
        - https://comptroller.texas.gov/taxes/
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Wisconsin', 'state_occupation': 'Texas', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def wisconsin_and_utah(self, state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Wisconsin:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Utah:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Wisconsin:
        - https://www.revenue.wi.gov/TaxForms2023/2023-Form1-ES-Inst.pdf
          https://www.revenue.wi.gov/TaxForms2023/2023-Form1-ES-Inst.pdf
          https://www.revenue.wi.gov/TaxForms2023/2023-Form1-ES-Inst.pdf
          https://www.revenue.wi.gov/Pages/FAQS/pcs-work.aspx
          https://www.revenue.wi.gov/Dor%20Publications/pb103.pdf
          https://www.revenue.wi.gov/Pages/FAQS/pcs-work.aspx
        
        Utah:
        -         
        - https://incometax.utah.gov/paying/tax-rates
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Wisconsin', 'state_occupation': 'Utah', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def wisconsin_and_vermont(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Wisconsin:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Vermont:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Wisconsin:
        - https://www.revenue.wi.gov/TaxForms2023/2023-Form1-ES-Inst.pdf
          https://www.revenue.wi.gov/TaxForms2023/2023-Form1-ES-Inst.pdf
          https://www.revenue.wi.gov/TaxForms2023/2023-Form1-ES-Inst.pdf
          https://www.revenue.wi.gov/Pages/FAQS/pcs-work.aspx
          https://www.revenue.wi.gov/Dor%20Publications/pb103.pdf
          https://www.revenue.wi.gov/Pages/FAQS/pcs-work.aspx
        
        Vermont:
        -         
        - https://tax.vermont.gov/sites/tax/files/documents/GB-1210-2023.pdf
          https://tax.vermont.gov/sites/tax/files/documents/Income%20Booklet-2022.pdf
          https://tax.vermont.gov/sites/tax/files/documents/Income%20Booklet-2022.pdf
          https://tax.vermont.gov/sites/tax/files/documents/Income%20Booklet-2022.pdf
          https://tax.vermont.gov/individuals/personal-income-tax/tax-credits
          https://tax.vermont.gov/individuals/personal-income-tax/taxable-income
          https://tax.vermont.gov/individuals/personal-income-tax/local-option-tax
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Wisconsin', 'state_occupation': 'Vermont', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def wisconsin_and_virginia(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Wisconsin:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Virginia:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Wisconsin:
        - https://www.revenue.wi.gov/TaxForms2023/2023-Form1-ES-Inst.pdf
          https://www.revenue.wi.gov/TaxForms2023/2023-Form1-ES-Inst.pdf
          https://www.revenue.wi.gov/TaxForms2023/2023-Form1-ES-Inst.pdf
          https://www.revenue.wi.gov/Pages/FAQS/pcs-work.aspx
          https://www.revenue.wi.gov/Dor%20Publications/pb103.pdf
          https://www.revenue.wi.gov/Pages/FAQS/pcs-work.aspx
        
        Virginia:
        -         
        - https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/reciprocity
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Wisconsin', 'state_occupation': 'Virginia', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def wisconsin_and_washington(self, state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Wisconsin:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Washington:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Wisconsin:
        - https://www.revenue.wi.gov/TaxForms2023/2023-Form1-ES-Inst.pdf
          https://www.revenue.wi.gov/TaxForms2023/2023-Form1-ES-Inst.pdf
          https://www.revenue.wi.gov/TaxForms2023/2023-Form1-ES-Inst.pdf
          https://www.revenue.wi.gov/Pages/FAQS/pcs-work.aspx
          https://www.revenue.wi.gov/Dor%20Publications/pb103.pdf
          https://www.revenue.wi.gov/Pages/FAQS/pcs-work.aspx
        
        Washington:
        -         
        - https://dor.wa.gov/taxes-rates/income-tax
          https://dor.wa.gov/taxes-rates/other-taxes/capital-gains-tax
          https://dor.wa.gov/taxes-rates/income-tax
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Wisconsin', 'state_occupation': 'Washington', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def wisconsin_and_west_virginia(self, county_occupation='Other', state_agi=75000, pay_periods=24, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Wisconsin:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        West Virginia:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_occupation : str, optional
            County of occupation. Default is Other.
        pay_periods : int, optional
            Number of pay periods. Options are 13, 24, 26, 52. Default is 24.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Wisconsin:
        - https://www.revenue.wi.gov/TaxForms2023/2023-Form1-ES-Inst.pdf
          https://www.revenue.wi.gov/TaxForms2023/2023-Form1-ES-Inst.pdf
          https://www.revenue.wi.gov/TaxForms2023/2023-Form1-ES-Inst.pdf
          https://www.revenue.wi.gov/Pages/FAQS/pcs-work.aspx
          https://www.revenue.wi.gov/Dor%20Publications/pb103.pdf
          https://www.revenue.wi.gov/Pages/FAQS/pcs-work.aspx
        
        West Virginia:
        -         
        - https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://tax.wv.gov/Documents/TSD/tsd438.pdf
          https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2011/Taxes-11-35att.html
        """
        
        # Check if the parameters are set upstream in the instance
        county_occupation = self.county_occupation if hasattr(self, 'county_occupation') else county_occupation
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        pay_periods = self.pay_periods if hasattr(self, 'pay_periods') else pay_periods
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Wisconsin', 'state_occupation': 'WestVirginia', "county_occupation": county_occupation, "state_agi": state_agi, "pay_periods": pay_periods, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def wisconsin_and_wyoming(self, state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Wisconsin:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        
        Wyoming:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Wisconsin:
        - https://www.revenue.wi.gov/TaxForms2023/2023-Form1-ES-Inst.pdf
          https://www.revenue.wi.gov/TaxForms2023/2023-Form1-ES-Inst.pdf
          https://www.revenue.wi.gov/TaxForms2023/2023-Form1-ES-Inst.pdf
          https://www.revenue.wi.gov/Pages/FAQS/pcs-work.aspx
          https://www.revenue.wi.gov/Dor%20Publications/pb103.pdf
          https://www.revenue.wi.gov/Pages/FAQS/pcs-work.aspx
        
        Wyoming:
        -         
        - https://revenue.wyo.gov/
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Wisconsin', 'state_occupation': 'Wyoming', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def wyoming_and_alabama(self, county_residence='Other', state_agi=75000, capital_gains_short=0, capital_gains_long=0, filing_status='single', income=100000, dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Wyoming:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        Alabama:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Wyoming:
        - https://revenue.wyo.gov/
        
        Alabama:
        -         
        - https://www.revenue.alabama.gov/ultraviewer/viewer/basic_viewer/index.html?form=2023/02/22f40.pdf
          https://www.revenue.alabama.gov/faqs/what-is-alabamas-individual-income-tax-rate/
          https://www.revenue.alabama.gov/ultraviewer/viewer/basic_viewer/index.html?form=2023/02/22f40.pdf
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-24.htm?taxmap=true
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-24.htm?taxmap=true
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-24.htm?taxmap=true
          https://www.revenue.alabama.gov/individual-corporate/income-to-be-reported-on-the-alabama-income-tax-return/
          https://www.almonline.org/TaxRates.aspx#OccupationalTax
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Wyoming', 'state_occupation': 'Alabama', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def wyoming_and_alaska(self, sources=False):
        """
        --- Parameters ----
        
        Wyoming:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        Alaska:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Wyoming:
        - https://revenue.wyo.gov/
        
        Alaska:
        -         
        - https://www.commerce.alaska.gov/web/dcra/officeofthestateassessor/alaskataxfacts.aspx
        """
        
        # Check if the parameters are set upstream in the instance
        sources = self.sources if hasattr(self, 'sources') else sources
        
        # State-specific parameters
        data = {
            'state_residence': 'Wyoming', 'state_occupation': 'Alaska', "sources": sources
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def wyoming_and_arizona(self, state_agi=75000, capital_gains_short=0, capital_gains_long=0, filing_status='single', income=100000, dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Wyoming:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        Arizona:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        credits_state : int, optional
            Credits for state taxes. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Wyoming:
        - https://revenue.wyo.gov/
        
        Arizona:
        -         
        - https://azdor.gov/forms/individual/form-140-x-y-tables
          https://azdor.gov/forms/individual/form-140-x-y-tables
          https://azdor.gov/forms/individual/form-140-x-y-tables
          https://azdor.gov/businesses-arizona/withholding-tax/withholding-exceptions
          "https://azdor.gov/individuals/income-tax-filing-assistance/identifying-other-taxable-income#:~:text=In%20regards%20to%20capital%20gains
         the%20individuals%20regular%20tax%20rate.
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Wyoming', 'state_occupation': 'Arizona', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def wyoming_and_arkansas(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, capital_gains_long=0, filing_status='single', income=100000, dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Wyoming:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        Arkansas:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Wyoming:
        - https://revenue.wyo.gov/
        
        Arkansas:
        -         
        - https://www.arkansasedc.com/why-arkansas/business-climate/tax-structure/personal-income-tax
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_and_AR1000NR_Instructions.pdf
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_FullYearResidentIndividualIncomeTaxReturn.pdf
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_and_AR1000NR_Instructions.pdf
          https://www.arkansasedc.com/why-arkansas/business-climate/tax-structure/capital-gains-tax-reduction
          https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_and_AR1000NR_Instructions.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Wyoming', 'state_occupation': 'Arkansas', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def wyoming_and_california(self, state_agi=75000, capital_gains_short=0, capital_gains_long=0, filing_status='single', income=100000, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Wyoming:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        California:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Wyoming:
        - https://revenue.wyo.gov/
        
        California:
        -         
        - https://www.ftb.ca.gov/forms/2022/2022-540-tax-rate-schedules.pdf
          https://www.ftb.ca.gov/file/personal/deductions/index.html
          https://www.ftb.ca.gov/about-ftb/newsroom/caleitc/eligibility-and-credit-information.html
          https://www.ftb.ca.gov/file/personal/income-types/capital-gains-and-losses.html#:~:text=California%20does%20not%20have%20a
         are%20taxed%20as%20ordinary%20income
          https://www.ftb.ca.gov/file/personal/
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Wyoming', 'state_occupation': 'California', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def wyoming_and_colorado(self, county_residence='Other', state_agi=75000, capital_gains_short=0, capital_gains_long=0, income=100000, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Wyoming:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        Colorado:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Wyoming:
        - https://revenue.wyo.gov/
        
        Colorado:
        -         
        - https://leg.colorado.gov/agencies/legislative-council-staff/individual-income-tax%C2%A0
          https://leg.colorado.gov/agencies/legislative-council-staff/individual-income-tax%C2%A0
          https://tax.colorado.gov/sites/tax/files/documents/ITT_Capital_Gain_Subtraction_Feb_2022.pdf
          https://www.auroragov.org/business_services/taxes/occupational_privilege_tax
          https://www.denvergov.org/Government/Agencies-Departments-Offices/Agencies-Departments-Offices-Directory/Department-of-Finance/Our-Divisions/Treasury/Business-Tax-Information#:~:text=Occupational%20privilege%20tax
         -%28show%20below%29&text=Each%20taxable%20employee%20is%20liable
         month%20for%20each%20taxable%20employee
          https://www.glendale.co.us/355/Occupational-Privilege-Tax
          https://www.greenwoodvillage.com/1220/Occupational-Privilege-Tax-OPT
          https://ci.sheridan.co.us/288/Occupational-Privilege-Tax
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        income = self.income if hasattr(self, 'income') else income
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Wyoming', 'state_occupation': 'Colorado', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "income": income, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def wyoming_and_connecticut(self, state_agi=75000, capital_gains_short=0, capital_gains_long=0, filing_status='single', income=100000, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Wyoming:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        Connecticut:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Wyoming:
        - https://revenue.wyo.gov/
        
        Connecticut:
        -         
        - https://portal.ct.gov/-/media/DRS/Forms/2022/Income/2022-CT-1040-Instructions_1222.pdf
          https://www.cga.ct.gov/2022/rpt/pdf/2022-R-0108.pdf
          https://www.cga.ct.gov/current/pub/chap_229.htm#sec_12-703
          https://portal.ct.gov/-/media/DRS/Forms/2022/Income/2022-CT-1040-Instructions_1222.pdf
          "https://portal.ct.gov/DRS/Publications/TSSNs/TSSN-29#:~:text=An%20individuals%20net%20capital%20gains
         portion%20of%20Social%20Security%20benefits"
          "https://portal.ct.gov/DRS/Publications/TSSNs/TSSN-29#:~:text=An%20individuals%20net%20capital%20gains
         portion%20of%20Social%20Security%20benefits"
          https://portal.ct.gov/DRS/Individuals/Resident-Income-Tax/Tax-Information
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Wyoming', 'state_occupation': 'Connecticut', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def wyoming_and_delaware(self, capital_gains_long=0, federal_agi=75000, income=100000, sources=False, capital_gains_short=0, credits_state=0):
        """
        --- Parameters ----
        
        Wyoming:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        Delaware:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        federal_agi : int, optional
            Adjusted Gross Income (AGI) for federal taxes. Default is 0. Sources - https://www.irs.gov/e-file-providers/definition-of-adjusted-gross-income, https://www.irs.gov/newsroom/taxpayers-should-know-the-difference-between-standard-and-itemized-deductions
        income : int, optional
            Gross income. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Wyoming:
        - https://revenue.wyo.gov/
        
        Delaware:
        -         
        - https://revenue.delaware.gov/software-developer/tax-rate-changes/
          https://revenuefiles.delaware.gov/2022/PIT-RES_TY22_2022-01_PaperInteractive.pdf
          https://revenuefiles.delaware.gov/2022/PIT-RES_TY22_2022-02_Instructions.pdf
          https://revenue.delaware.gov/frequently-asked-questions/personal-income-tax-faqs/
          https://www.newcastlede.gov/DocumentCenter/View/20147/City-of-Wilmington-Wage-Tax-Information-PDF
        """
        
        # Check if the parameters are set upstream in the instance
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        federal_agi = self.federal_agi if hasattr(self, 'federal_agi') else federal_agi
        income = self.income if hasattr(self, 'income') else income
        sources = self.sources if hasattr(self, 'sources') else sources
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Wyoming', 'state_occupation': 'Delaware', "capital_gains_long": capital_gains_long, "federal_agi": federal_agi, "income": income, "sources": sources, "capital_gains_short": capital_gains_short, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def wyoming_and_florida(self, sources=False):
        """
        --- Parameters ----
        
        Wyoming:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        Florida:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Wyoming:
        - https://revenue.wyo.gov/
        
        Florida:
        -         
        - https://floridarevenue.com/taxes/taxesfees/Pages/tax_interest_rates.aspx
        """
        
        # Check if the parameters are set upstream in the instance
        sources = self.sources if hasattr(self, 'sources') else sources
        
        # State-specific parameters
        data = {
            'state_residence': 'Wyoming', 'state_occupation': 'Florida', "sources": sources
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def wyoming_and_georgia(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, capital_gains_long=0, filing_status='single', income=100000, dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Wyoming:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        Georgia:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Wyoming:
        - https://revenue.wyo.gov/
        
        Georgia:
        -         
        - https://dor.georgia.gov/tax-tables-georgia-tax-rate-schedule
          https://apps.dor.ga.gov/FillableForms/PDFViewer/Index?form=2022GA500
          https://apps.dor.ga.gov/FillableForms/PDFViewer/Index?form=2022GA500
          https://apps.dor.ga.gov/FillableForms/PDFViewer/Index?form=2022GA500
          https://dor.georgia.gov/years-individual-income-tax-forms
          https://dor.georgia.gov/years-individual-income-tax-forms
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Wyoming', 'state_occupation': 'Georgia', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def wyoming_and_hawaii(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, capital_gains_long=0, filing_status='single', income=100000, dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Wyoming:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        Hawaii:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Wyoming:
        - https://revenue.wyo.gov/
        
        Hawaii:
        -         
        - https://tax.hawaii.gov/forms/d_18table-on/d_18table-on_p13/
          https://files.hawaii.gov/tax/forms/2018/18table-on.pdf
          https://files.hawaii.gov/tax/news/pubs/22outline.pdf
          https://files.hawaii.gov/tax/news/pubs/FastTaxRefGuide_2017.pdf
          https://www.tfhawaii.org/wordpress/state-tax-resources/mini-tax-guides/individual-income-tax-chapter-235/#:~:text=Long%20term%20capital%20gains%20are%20taxed%20at%20a%20maximum%20of%207.25%25
          https://www.grassrootinstitute.org/2023/02/hb337-increasing-tax-on-capital-gains-a-bad-risk/
          https://files.hawaii.gov/tax/news/pubs/22outline.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Wyoming', 'state_occupation': 'Hawaii', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def wyoming_and_idaho(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, capital_gains_long=0, filing_status='single', income=100000, dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Wyoming:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        Idaho:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Wyoming:
        - https://revenue.wyo.gov/
        
        Idaho:
        -         
        - https://tax.idaho.gov/taxes/income-tax/individual-income/individual-income-tax-rate-schedule/
          https://tax.idaho.gov/wp-content/uploads/forms/EIN00046/EIN00046_03-01-2023.pdf
          https://tax.idaho.gov/taxes/income-tax/individual-income/idaho-source-income/
          https://tax.idaho.gov/taxes/income-tax/individual-income/idaho-source-income/
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Wyoming', 'state_occupation': 'Idaho', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def wyoming_and_illinois(self, state_agi=75000, capital_gains_short=0, capital_gains_long=0, filing_status='single', income=100000, dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Wyoming:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        Illinois:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Wyoming:
        - https://revenue.wyo.gov/
        
        Illinois:
        -         
        - https://tax.illinois.gov/research/taxrates/income.html#IndividualIncome
          https://tax.illinois.gov/content/dam/soi/en/web/tax/forms/incometax/documents/currentyear/individual/il-1040-instr.pdf
          https://tax.illinois.gov/questionsandanswers/12.html
          https://tax.illinois.gov/content/dam/soi/en/web/tax/forms/incometax/documents/currentyear/individual/il-1040-instr.pdf
          https://tax.illinois.gov/research/taxrates/personalproperty.html
          https://tax.illinois.gov/localgovernments/income.html
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Wyoming', 'state_occupation': 'Illinois', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def wyoming_and_indiana(self, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_short=0, capital_gains_long=0, filing_status='single', income=100000, dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Wyoming:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        Indiana:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Wyoming:
        - https://revenue.wyo.gov/
        
        Indiana:
        -         
        - https://www.in.gov/dor/tax-forms/2022-individual-income-tax-forms/
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-32.htm
          https://www.in.gov/dor/files/ib28.pdf
          https://www.in.gov/dor/tax-forms/2022-individual-income-tax-forms/
          https://www.in.gov/dor/files/2022-county-tax-rates-and-codes.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Wyoming', 'state_occupation': 'Indiana', "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def wyoming_and_iowa(self, county_residence='Other', state_agi=75000, capital_gains_short=0, capital_gains_long=0, filing_status='single', income=100000, dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Wyoming:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        Iowa:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Wyoming:
        - https://revenue.wyo.gov/
        
        Iowa:
        -         
        - https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
          https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
          https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
          https://tax.iowa.gov/iowa-illinois-reciprocal-agreement
          https://tax.iowa.gov/2023-changes-iowa-individual-income-tax#:~:text=Individual%20Income%20Tax%20Rates%20(HF%202317)&text=Beginning%20in%20tax%20year%202026
         who%20file%20a%20joint%20return
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Wyoming', 'state_occupation': 'Iowa', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def wyoming_and_kansas(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, capital_gains_long=0, filing_status='single', income=100000, dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Wyoming:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        Kansas:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Wyoming:
        - https://revenue.wyo.gov/
        
        Kansas:
        -         
        - https://www.ksrevenue.gov/taxrates.html
          https://www.ksrevenue.gov/incomebook22.html#0
          https://www.ksrevenue.gov/incomebook22.html#0
          https://www.ksrevenue.gov/incomebook22.html#0
          https://www.ksrevenue.gov/incomebook22.html
          https://www.ksrevenue.gov/incomebook22.html
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Wyoming', 'state_occupation': 'Kansas', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def wyoming_and_kentucky(self, county_residence='Other', state_agi=75000, capital_gains_short=0, capital_gains_long=0, filing_status='single', income=100000, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Wyoming:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        Kentucky:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Wyoming:
        - https://revenue.wyo.gov/
        
        Kentucky:
        -         
        - https://revenue.ky.gov/News/Pages/Dor-Announces-Updates-to-Individual-Income-Tax-for-2023-Tax-Year.aspx
          https://revenue.ky.gov/Individual/Individual-Income-Tax/Pages/default.aspx
          https://revenue.ky.gov/News/Pages/Dor-Announces-Updates-to-Individual-Income-Tax-for-2023-Tax-Year.aspx
          https://revenue.ky.gov/Individual/Individual-Income-Tax/Pages/default.aspx
          https://revenue.ky.gov/Dor%20Training%20Materials/103%20KAR%2017.140.%20Individual%20income%20tax%20-%20reciprocity%20-%20nonresidents.pdf
          https://revenue.ky.gov/Individual/Individual-Income-Tax/Pages/default.aspx
          https://nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2021/Taxes-21-32.htm?taxmap=true
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Wyoming', 'state_occupation': 'Kentucky', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def wyoming_and_louisiana(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, capital_gains_long=0, filing_status='single', income=100000, dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Wyoming:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        Louisiana:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Wyoming:
        - https://revenue.wyo.gov/
        
        Louisiana:
        -         
        - https://revenue.louisiana.gov/individualincometax
          https://revenue.louisiana.gov/individualincometax
          https://www.revenue.louisiana.gov/taxforms/6935(11_02)F.pdf
          https://revenue.louisiana.gov/individualincometax
          https://revenue.louisiana.gov/individualincometax
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Wyoming', 'state_occupation': 'Louisiana', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def wyoming_and_maine(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, capital_gains_long=0, filing_status='single', income=100000, dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Wyoming:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        Maine:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Wyoming:
        - https://revenue.wyo.gov/
        
        Maine:
        -         
        - https://www.maine.gov/revenue/tax-return-forms/individual-income-tax-2022
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/ind_tax_rate_sched_2022.pdf
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/22_1040me_book_gen_instruc_revisedFeb23.pdf
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/22_1040me_book_gen_instruc_revisedFeb23.pdf
          https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/22_1040me_book_gen_instruc_revisedFeb23.pdf
          https://www.maine.gov/revenue/tax-return-forms/individual-income-tax-2022
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Wyoming', 'state_occupation': 'Maine', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def wyoming_and_maryland(self, county_residence='Other', state_agi=75000, capital_gains_short=0, capital_gains_long=0, filing_status='single', income=100000, dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Wyoming:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        Maryland:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Wyoming:
        - https://revenue.wyo.gov/
        
        Maryland:
        -         
        - https://www.marylandtaxes.gov/individual/income/tax-info/tax-rates.php
          https://www.marylandtaxes.gov/forms/22_forms/Resident_Booklet.pdf
          https://www.marylandtaxes.gov/forms/Tax_Publications/Administrative_Releases/Income_and_Estate_Tax_Releases/ar_it37.pdf
          https://www.marylandtaxes.gov/individual/index.php
          https://www.marylandtaxes.gov/individual/income/tax-info/tax-rates.php
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Wyoming', 'state_occupation': 'Maryland', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def wyoming_and_massachusetts(self, state_agi=75000, capital_gains_short=0, capital_gains_long=0, filing_status='single', income=100000, dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Wyoming:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        Massachusetts:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Wyoming:
        - https://revenue.wyo.gov/
        
        Massachusetts:
        -         
        - https://www.mass.gov/service-details/massachusetts-tax-rates
          https://www.mass.gov/service-details/view-massachusetts-personal-income-tax-exemptions
          https://www.mass.gov/service-details/view-massachusetts-personal-income-tax-exemptions
          https://www.mass.gov/service-details/view-child-and-dependent-related-credits
          https://www.mass.gov/service-details/massachusetts-tax-rates
          https://www.mass.gov/personal-income-tax
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Wyoming', 'state_occupation': 'Massachusetts', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def wyoming_and_michigan(self, county_residence='Other', state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', county_occupation='Other', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Wyoming:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        Michigan:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        county_occupation : str, optional
            County of occupation. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Wyoming:
        - https://revenue.wyo.gov/
        
        Michigan:
        -         
        - https://www.michigan.gov/taxes/iit/new-developments-for-tax-year-2022
          "https://www.michigan.gov/taxes/-/media/Project/Websites/taxes/Forms/2023/2023-SUW/446_Withholding-Guide_2023.pdf?rev=177ca3d8cb034392bf84e317ed78c320&hash=19C5F42F43A89A91863D229349F48883#:~:text=The%20exemption%20amount%20is%20%245%2C400
         employees%20Michigan%20income%20tax%20return"
          "https://www.michigan.gov/taxes/-/media/Project/Websites/taxes/Forms/2023/2023-SUW/446_Withholding-Guide_2023.pdf?rev=177ca3d8cb034392bf84e317ed78c320&hash=19C5F42F43A89A91863D229349F48883#:~:text=The%20exemption%20amount%20is%20%245%2C400
         employees%20Michigan%20income%20tax%20return"
          https://www.michigan.gov/taxes/questions/iit/accordion/residency/are-my-wages-earned-in-another-state-taxable-in-michigan-if-i-am-a-michigan-resident-1
          https://www.michigan.gov/taxes/-/media/Project/Websites/taxes/Forms/2022/2022-IIT-Forms/BOOK_MI-1040.pdf?rev=4524f1cbbc0243b194ab05a6680be75b&hash=40DFE71E651F7C1D65C31F6CD03D944F
          https://www.michigan.gov/taxes/questions/iit/accordion/general/what-cities-impose-an-income-tax
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        county_occupation = self.county_occupation if hasattr(self, 'county_occupation') else county_occupation
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Wyoming', 'state_occupation': 'Michigan', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "county_occupation": county_occupation, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def wyoming_and_minnesota(self, state_agi=75000, capital_gains_short=0, capital_gains_long=0, filing_status='single', income=100000, dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Wyoming:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        Minnesota:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Wyoming:
        - https://revenue.wyo.gov/
        
        Minnesota:
        -         
        - https://www.revenue.state.mn.us/minnesota-income-tax-rates-and-brackets
          https://www.revenue.state.mn.us/child-and-dependent-care-credit
          https://www.revenue.state.mn.us/dependent-exemptions
          https://www.revenue.state.mn.us/minnesota-standard-deduction
          https://mn.gov/mmb/assets/mwr_form_tcm1059-128581.pdf
          https://www.revenue.state.mn.us/individual-income-tax
          https://www.revenue.state.mn.us/individual-income-tax
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Wyoming', 'state_occupation': 'Minnesota', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def wyoming_and_mississippi(self, state_agi=75000, capital_gains_short=0, capital_gains_long=0, filing_status='single', income=100000, dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Wyoming:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        Mississippi:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Wyoming:
        - https://revenue.wyo.gov/
        
        Mississippi:
        -         
        - https://www.dor.ms.gov/individual/tax-rates
          https://www.dor.ms.gov/individual/tax-rates
          https://www.dor.ms.gov/individual/tax-rates
          https://www.dor.ms.gov/individual
          https://www.dor.ms.gov/individual
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Wyoming', 'state_occupation': 'Mississippi', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def wyoming_and_missouri(self, county_residence='Other', state_agi=75000, filers_over_65=0, capital_gains_short=0, capital_gains_long=0, filing_status='single', income=100000, dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Wyoming:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        Missouri:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Wyoming:
        - https://revenue.wyo.gov/
        
        Missouri:
        -         
        - https://dor.mo.gov/taxation/individual/tax-types/income/year-changes/
          https://dor.mo.gov/taxation/individual/tax-types/income/year-changes/
          https://dor.mo.gov/taxation/individual/tax-types/income/year-changes/
          https://www.kcmo.gov/city-hall/departments/finance/earnings-tax
          https://www.stlouis-mo.gov/government/departments/collector/earnings-tax/individual-earnings-tax-info.cfm
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Wyoming', 'state_occupation': 'Missouri', "county_residence": county_residence, "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def wyoming_and_montana(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, capital_gains_long=0, filing_status='single', income=100000, dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Wyoming:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        Montana:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Wyoming:
        - https://revenue.wyo.gov/
        
        Montana:
        -         
        - https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
          https://mtrevenue.gov/wp-content/uploads/2022/12/montana-employees-withholding-allowance-and-exemption-certificate-form-mw-4.pdf
          https://mtrevenue.gov/2021/11/29/simplification-of-montana-income-taxation/?utm_source=rss&utm_medium=rss&utm_campaign=simplification-of-montana-income-taxation
          https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/02/Form_2_2022-1.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Wyoming', 'state_occupation': 'Montana', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def wyoming_and_nebraska(self, state_agi=75000, capital_gains_short=0, capital_gains_long=0, filing_status='single', income=100000, dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Wyoming:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        Nebraska:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Wyoming:
        - https://revenue.wyo.gov/
        
        Nebraska:
        -         
        - https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2023_Circular_EN_8-429r2023_Whole_Book_11.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
          https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_6.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Wyoming', 'state_occupation': 'Nebraska', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def wyoming_and_nevada(self, sources=False):
        """
        --- Parameters ----
        
        Wyoming:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        Nevada:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Wyoming:
        - https://revenue.wyo.gov/
        
        Nevada:
        -         
        - https://www.nvsos.gov/sos/businesses/the-nevada-advantage
        """
        
        # Check if the parameters are set upstream in the instance
        sources = self.sources if hasattr(self, 'sources') else sources
        
        # State-specific parameters
        data = {
            'state_residence': 'Wyoming', 'state_occupation': 'Nevada', "sources": sources
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def wyoming_and_new_hampshire(self, sources=False):
        """
        --- Parameters ----
        
        Wyoming:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        New Hampshire:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Wyoming:
        - https://revenue.wyo.gov/
        
        New Hampshire:
        -         
        - https://www.revenue.nh.gov/assistance/tax-overview.htm
        """
        
        # Check if the parameters are set upstream in the instance
        sources = self.sources if hasattr(self, 'sources') else sources
        
        # State-specific parameters
        data = {
            'state_residence': 'Wyoming', 'state_occupation': 'NewHampshire', "sources": sources
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def wyoming_and_new_jersey(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, capital_gains_long=0, filing_status='single', income=100000, dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Wyoming:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        New Jersey:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Wyoming:
        - https://revenue.wyo.gov/
        
        New Jersey:
        -         
        - https://www.state.nj.us/treasury/taxation/taxtables.shtml
          https://www.state.nj.us/treasury/taxation/njit13.shtml
          https://www.nj.gov/treasury/taxation/njit25.shtml
          https://www.state.nj.us/treasury/taxation/individuals/freqqiti.shtml
          https://www.state.nj.us/treasury/taxation/localtax.shtml
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Wyoming', 'state_occupation': 'NewJersey', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def wyoming_and_new_mexico(self, state_agi=75000, capital_gains_short=0, capital_gains_long=0, filing_status='single', income=100000, dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Wyoming:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        New Mexico:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Wyoming:
        - https://revenue.wyo.gov/
        
        New Mexico:
        -         
        - https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/fdf3c548-8aba-4b9c-9eb4-bb564c716015/FYI-104.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/0bf943fd-652e-4400-bb1b-cd8397eb5a95/2022pit-adj.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/2f1a6781-9534-4436-b427-1557f9592099/2022pit-adj-ins.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/0bf943fd-652e-4400-bb1b-cd8397eb5a95/2022pit-adj.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/1afc56af-ea90-4d48-82e5-1f9aeb43255a/PITbook2022.pdf
          https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/fdf3c548-8aba-4b9c-9eb4-bb564c716015/FYI-104.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Wyoming', 'state_occupation': 'NewMexico', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def wyoming_and_new_york(self, county_residence='Other', state_agi=75000, capital_gains_short=0, capital_gains_long=0, filing_status='single', income=100000, dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Wyoming:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        New York:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Wyoming:
        - https://revenue.wyo.gov/
        
        New York:
        -         
        - https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nys-tax-rate-schedule
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nys-tax-rate-schedule
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nys-tax-rate-schedule
          https://www.tax.ny.gov/pit/
          https://www.tax.ny.gov/pit/file/nonresident-faqs.htm
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#yonkers-tax-rate-schedule
          https://www.tax.ny.gov/forms/html-instructions/2022/it/it201i-2022.htm#nyc-tax-rate-schedule
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Wyoming', 'state_occupation': 'NewYork', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def wyoming_and_north_carolina(self, state_agi=75000, capital_gains_short=0, capital_gains_long=0, filing_status='single', income=100000, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Wyoming:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        North Carolina:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Wyoming:
        - https://revenue.wyo.gov/
        
        North Carolina:
        -         
        - https://www.ncdor.gov/taxes-forms/tax-rate-schedules
          https://www.ncdor.gov/taxes-forms/individual-income-tax/north-carolina-standard-deduction-or-north-carolina-itemized-deductions
          https://www.ncdor.gov/taxes-forms/individual-income-tax
          https://www.ncdor.gov/taxes-forms/individual-income-tax
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Wyoming', 'state_occupation': 'NorthCarolina', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def wyoming_and_north_dakota(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, capital_gains_long=0, filing_status='single', income=100000, dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Wyoming:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        North Dakota:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Wyoming:
        - https://revenue.wyo.gov/
        
        North Dakota:
        -         
        - https://www.tax.nd.gov/forms
          https://www.tax.nd.gov/sites/www/files/documents/forms/individual/2022-iit/2022-individual-income-tax-booklet.pdf
          https://www.tax.nd.gov/sites/www/files/documents/forms/2022-individual-income-tax-booklet.pdf
          https://www.tax.nd.gov/tax-exemptions-credits/income-tax-exemptions-credits
          https://www.tax.nd.gov/sites/www/files/documents/forms/business/ndwrfillable.pdf
          https://www.tax.nd.gov/news/tax-legislative-changes/significant-changes-law/individual-income-tax-history
          https://www.tax.nd.gov/forms
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Wyoming', 'state_occupation': 'NorthDakota', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def wyoming_and_ohio(self, county_residence='Other', state_agi=75000, capital_gains_short=0, capital_gains_long=0, filing_status='single', income=100000, dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Wyoming:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        Ohio:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_residence : str, optional
            County of residence. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Wyoming:
        - https://revenue.wyo.gov/
        
        Ohio:
        -         
        - https://tax.ohio.gov/individual/resources/annual-tax-rates
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://tax.ohio.gov/static/forms/employer_withholding/generic/wth_it4nr.pdf
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://tax.ohio.gov/static/forms/ohio_individual/individual/2022/it1040-sd100-instruction-booklet.pdf
          https://www.ritaohio.com/TaxRatesTable
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Wyoming', 'state_occupation': 'Ohio', "county_residence": county_residence, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def wyoming_and_oklahoma(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, capital_gains_long=0, filing_status='single', income=100000, dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Wyoming:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        Oklahoma:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Wyoming:
        - https://revenue.wyo.gov/
        
        Oklahoma:
        -         
        - https://oklahoma.gov/content/dam/ok/en/tax/documents/resources/publications/businesses/withholding-tables/WHTables-2022.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
          https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-NR-Pkt.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Wyoming', 'state_occupation': 'Oklahoma', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def wyoming_and_oregon(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, capital_gains_long=0, filing_status='single', income=100000, dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Wyoming:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        Oregon:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Wyoming:
        - https://revenue.wyo.gov/
        
        Oregon:
        -         
        - https://www.oregon.gov/dor/programs/individuals/Pages/PIT.aspx
          https://www.oregon.gov/dor/programs/individuals/Documents/Full%20year%20resident
         %20Form%20or-40%20filers.pdf
          https://www.oregon.gov/dor/programs/individuals/Documents/Part-year%20and%20nonresident
         %20Form%20or-40-P%20and%20or-40-N%20filers.pdf.pdf
          https://www.oregon.gov/dor/forms/FormsPubs/form-or-40_101-040_2022.pdf
          https://www.oregon.gov/dor/forms/FormsPubs/form-or-40_101-040_2022.pdf
          https://www.oregon.gov/dor/forms/FormsPubs/form-or-40_101-040_2022.pdf
          https://www.portland.gov/revenue/personal-tax
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Wyoming', 'state_occupation': 'Oregon', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def wyoming_and_pennsylvania(self, county_residence='Other', state_agi=75000, county_occupation='Other', capital_gains_short=0, capital_gains_long=0, income=100000, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Wyoming:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        Pennsylvania:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        county_residence : str, optional
            County of residence. Default is Other.
        county_occupation : str, optional
            County of occupation. Default is Other.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Wyoming:
        - https://revenue.wyo.gov/
        
        Pennsylvania:
        -         
        - https://www.revenue.pa.gov/Tax%20Rates/Pages/default.aspx
          https://www.revenue.pa.gov/FormsandPublications/FormsforIndividuals/PIT/Documents/2022/2022_pa-40in.pdf
          https://www.revenue.pa.gov/FormsandPublications/FormsforBusinesses/EmployerWithholding/Documents/rev-419.pdf
          https://www.revenue.pa.gov/FormsandPublications/FormsforIndividuals/PIT/Documents/2022/2022_pa-40in.pdf
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2022/Taxes-22-25.htm
        """
        
        # Check if the parameters are set upstream in the instance
        county_residence = self.county_residence if hasattr(self, 'county_residence') else county_residence
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        county_occupation = self.county_occupation if hasattr(self, 'county_occupation') else county_occupation
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        income = self.income if hasattr(self, 'income') else income
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Wyoming', 'state_occupation': 'Pennsylvania', "county_residence": county_residence, "state_agi": state_agi, "county_occupation": county_occupation, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "income": income, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def wyoming_and_rhode_island(self, state_agi=75000, capital_gains_short=0, capital_gains_long=0, filing_status='single', income=100000, dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Wyoming:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        Rhode Island:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Wyoming:
        - https://revenue.wyo.gov/
        
        Rhode Island:
        -         
        - https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2022_1040WE_w_0.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
          https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2023%20RI-1040ES_w.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Wyoming', 'state_occupation': 'RhodeIsland', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def wyoming_and_south_carolina(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, capital_gains_long=0, filing_status='single', income=100000, dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Wyoming:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        South Carolina:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Wyoming:
        - https://revenue.wyo.gov/
        
        South Carolina:
        -         
        - https://dor.sc.gov/resources-site/lawandpolicy/Advisory%20Opinions/IL22-15.pdf
          https://dor.sc.gov/forms-site/Forms/IITPacket_2022.pdf
          https://www.scstatehouse.gov/code/t12c006.php
          https://dor.sc.gov/resources-site/lawandpolicy/Advisory%20Opinions/IL22-15.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Wyoming', 'state_occupation': 'SouthCarolina', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def wyoming_and_south_dakota(self, sources=False):
        """
        --- Parameters ----
        
        Wyoming:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        South Dakota:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Wyoming:
        - https://revenue.wyo.gov/
        
        South Dakota:
        -         
        - https://dor.sd.gov/individuals/taxes/
        """
        
        # Check if the parameters are set upstream in the instance
        sources = self.sources if hasattr(self, 'sources') else sources
        
        # State-specific parameters
        data = {
            'state_residence': 'Wyoming', 'state_occupation': 'SouthDakota', "sources": sources
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def wyoming_and_tennessee(self, sources=False):
        """
        --- Parameters ----
        
        Wyoming:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        Tennessee:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Wyoming:
        - https://revenue.wyo.gov/
        
        Tennessee:
        -         
        - https://www.tn.gov/revenue/taxes.html
        """
        
        # Check if the parameters are set upstream in the instance
        sources = self.sources if hasattr(self, 'sources') else sources
        
        # State-specific parameters
        data = {
            'state_residence': 'Wyoming', 'state_occupation': 'Tennessee', "sources": sources
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def wyoming_and_texas(self, sources=False):
        """
        --- Parameters ----
        
        Wyoming:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        Texas:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Wyoming:
        - https://revenue.wyo.gov/
        
        Texas:
        -         
        - https://comptroller.texas.gov/taxes/
        """
        
        # Check if the parameters are set upstream in the instance
        sources = self.sources if hasattr(self, 'sources') else sources
        
        # State-specific parameters
        data = {
            'state_residence': 'Wyoming', 'state_occupation': 'Texas', "sources": sources
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def wyoming_and_utah(self, state_agi=75000, capital_gains_short=0, capital_gains_long=0, filing_status='single', income=100000, dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Wyoming:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        Utah:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Wyoming:
        - https://revenue.wyo.gov/
        
        Utah:
        -         
        - https://incometax.utah.gov/paying/tax-rates
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
          https://tax.utah.gov/forms/current/tc-40.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Wyoming', 'state_occupation': 'Utah', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def wyoming_and_vermont(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, capital_gains_long=0, filing_status='single', income=100000, dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Wyoming:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        Vermont:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Wyoming:
        - https://revenue.wyo.gov/
        
        Vermont:
        -         
        - https://tax.vermont.gov/sites/tax/files/documents/GB-1210-2023.pdf
          https://tax.vermont.gov/sites/tax/files/documents/Income%20Booklet-2022.pdf
          https://tax.vermont.gov/sites/tax/files/documents/Income%20Booklet-2022.pdf
          https://tax.vermont.gov/sites/tax/files/documents/Income%20Booklet-2022.pdf
          https://tax.vermont.gov/individuals/personal-income-tax/tax-credits
          https://tax.vermont.gov/individuals/personal-income-tax/taxable-income
          https://tax.vermont.gov/individuals/personal-income-tax/local-option-tax
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Wyoming', 'state_occupation': 'Vermont', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def wyoming_and_virginia(self, state_agi=75000, filers_over_65=0, capital_gains_short=0, capital_gains_long=0, filing_status='single', income=100000, dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Wyoming:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        Virginia:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2
        filers_over_65 : int, optional
            Number of filers over 65. Options are 0, 1, 2. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Wyoming:
        - https://revenue.wyo.gov/
        
        Virginia:
        -         
        - https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/reciprocity
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
          https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2022-760-instructions.pdf
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        filers_over_65 = self.filers_over_65 if hasattr(self, 'filers_over_65') else filers_over_65
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Wyoming', 'state_occupation': 'Virginia', "state_agi": state_agi, "filers_over_65": filers_over_65, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def wyoming_and_washington(self, capital_gains_long=0, sources=False, capital_gains_short=0):
        """
        --- Parameters ----
        
        Wyoming:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        Washington:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Wyoming:
        - https://revenue.wyo.gov/
        
        Washington:
        -         
        - https://dor.wa.gov/taxes-rates/income-tax
          https://dor.wa.gov/taxes-rates/other-taxes/capital-gains-tax
          https://dor.wa.gov/taxes-rates/income-tax
        """
        
        # Check if the parameters are set upstream in the instance
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        sources = self.sources if hasattr(self, 'sources') else sources
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        
        # State-specific parameters
        data = {
            'state_residence': 'Wyoming', 'state_occupation': 'Washington', "capital_gains_long": capital_gains_long, "sources": sources, "capital_gains_short": capital_gains_short
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def wyoming_and_west_virginia(self, pay_periods=24, state_agi=75000, capital_gains_short=0, income=100000, capital_gains_long=0, filing_status='single', county_occupation='Other', dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Wyoming:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        West Virginia:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        county_occupation : str, optional
            County of occupation. Default is Other.
        pay_periods : int, optional
            Number of pay periods. Options are 13, 24, 26, 52. Default is 24.
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Wyoming:
        - https://revenue.wyo.gov/
        
        West Virginia:
        -         
        - https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://tax.wv.gov/Documents/TSD/tsd438.pdf
          https://tax.wv.gov/Individuals/Pages/PersonalIncomeTaxReductionBill.aspx
          https://www.nfc.usda.gov/Publications/HR_Payroll/Taxes/Bulletins/2011/Taxes-11-35att.html
        """
        
        # Check if the parameters are set upstream in the instance
        pay_periods = self.pay_periods if hasattr(self, 'pay_periods') else pay_periods
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        income = self.income if hasattr(self, 'income') else income
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        county_occupation = self.county_occupation if hasattr(self, 'county_occupation') else county_occupation
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Wyoming', 'state_occupation': 'WestVirginia', "pay_periods": pay_periods, "state_agi": state_agi, "capital_gains_short": capital_gains_short, "income": income, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "county_occupation": county_occupation, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    

    def wyoming_and_wisconsin(self, state_agi=75000, capital_gains_short=0, capital_gains_long=0, filing_status='single', income=100000, dependents=0, sources=False, credits_state=0):
        """
        --- Parameters ----
        
        Wyoming:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        
        Wisconsin:
        sources : bool, optional
            Whether to return the sources of the tax details. Default is False.
        filing_status : str, optional
            Filing status for taxes. Options are 'single', 'married_jointly', 'married_seperately', 'head_of_household', 'widow(er)'. Default is 'single'. Source - https://www.irs.gov/newsroom/taxpayers-should-use-the-correct-filing-status-for-accuracy-and-to-avoid-
        capital_gains_short : int, optional
            Net short term capital gains. Default is 0.
        capital_gains_long : int, optional
            Net long term capital gains. Default is 0.
        credits_state : int, optional
            Credits for state taxes. Default is 0.
        state_agi : int, optional
            Adjusted Gross Income (AGI) for state taxes. Default is 0.
        income : int, optional
            Gross income. Default is 0.
        dependents : int, optional
            Number of dependents. Default is 0. Source - https://www.irs.gov/faqs/filing-requirements-status-dependents/dependents/dependents-2


        Returns:
        pandas.DataFrame: A DataFrame containing the detailed tax breakdown.
        or
        dict: where 'data' contains the DataFrame and 'sources' contains a dict of the sources.
        
        Note:
        The function fetches the latest tax details from external sources to perform the calculation.

        --- Sources ---

        Wyoming:
        - https://revenue.wyo.gov/
        
        Wisconsin:
        -         
        - https://www.revenue.wi.gov/TaxForms2023/2023-Form1-ES-Inst.pdf
          https://www.revenue.wi.gov/TaxForms2023/2023-Form1-ES-Inst.pdf
          https://www.revenue.wi.gov/TaxForms2023/2023-Form1-ES-Inst.pdf
          https://www.revenue.wi.gov/Pages/FAQS/pcs-work.aspx
          https://www.revenue.wi.gov/Dor%20Publications/pb103.pdf
          https://www.revenue.wi.gov/Pages/FAQS/pcs-work.aspx
        """
        
        # Check if the parameters are set upstream in the instance
        state_agi = self.state_agi if hasattr(self, 'state_agi') else state_agi
        capital_gains_short = self.capital_gains_short if hasattr(self, 'capital_gains_short') else capital_gains_short
        capital_gains_long = self.capital_gains_long if hasattr(self, 'capital_gains_long') else capital_gains_long
        filing_status = self.filing_status if hasattr(self, 'filing_status') else filing_status
        income = self.income if hasattr(self, 'income') else income
        dependents = self.dependents if hasattr(self, 'dependents') else dependents
        sources = self.sources if hasattr(self, 'sources') else sources
        credits_state = self.credits_state if hasattr(self, 'credits_state') else credits_state
        
        # State-specific parameters
        data = {
            'state_residence': 'Wyoming', 'state_occupation': 'Wisconsin', "state_agi": state_agi, "capital_gains_short": capital_gains_short, "capital_gains_long": capital_gains_long, "filing_status": filing_status, "income": income, "dependents": dependents, "sources": sources, "credits_state": credits_state
        }
        
        return self.__fetch__(data, 'state-taxes')
    
